#!/bin/bash

# This software is available to you under a choice of one of two
# licenses.  You may choose to be licensed under the terms of the GNU
# General Public License (GPL) Version 2, available at
# <http://www.fsf.org/copyleft/gpl.html>, or the OpenIB.org BSD
# license, available in the LICENSE.TXT file accompanying this
# software.  These details are also available at
# <http://openib.org/license.html>.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS                               
# BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
# ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# Copyright (c) 2004 Mellanox Technologies Ltd.  All rights reserved.
#
# Description: IBGD RPMs building script
#

usage()
{
cat << EOF

    `basename $0` [ -c <config_file> ]

EOF
}

unalias -a

cd ${0%*/*}
CWD=`pwd`

INTERACTIVE=1

if [ ! -e ./build_env.sh ]; then
    echo "ERROR: $CWD/build_env.sh file is missing"
    exit 1
fi

while [ $# -gt 0 ]
do
    case $1 in
            -c)
                shift
                CONFIG=$1
                INTERACTIVE=0
            ;;
            *)
                usage
                exit 1
            ;;
    esac
    shift
done

. $CWD/build_env.sh

# Set RPM build tree  
if [ -d /usr/src/redhat ]; then
        RPM_DIR=/usr/src/redhat
elif [ -d /usr/src/packages ]; then
        RPM_DIR=/usr/src/packages
else
        echo Build rpm on this system is not supported...
        exit 1
fi


SELECTED_PACKAGES=""
OPENIB_PACKAGES=""
EXTRA_PACKAGES=""


# Environment variables definition
BUILD_COUNTER=0
EXISTING_RPMS=0


# Create selected from openib RPM packages (ib-verbs, ib-ipoib, ...)
openib()
{
        if ! check_build_req openib err; then
            exit 2
        fi    
        cd $BUILD_TMP
        ex_silent "$RM -f ${OPENIB_SOURCE##*/}"
        echo
        echo "Building ${OPENIB_NAME} RPMs. Please wait..."
        echo
        ex_silent "tar xzvf ${OPENIB_SOURCE}"
        local OPENIB_SRCDIR=$BUILD_TMP/`echo ${OPENIB_SOURCE##*/} | sed -e 's/.tgz//' | tr -d '[:space:]'`
	if ( echo -n $OPENIB_PACKAGES | grep ib_srp_target > $NULL 2>&1 ) && [ ! -d ${OPENIB_SRCDIR}/ib_srp_target ]; then
		OPENIB_PACKAGES=$(echo -n $OPENIB_PACKAGES | sed -e "s/ ib_srp_target//")
		export OPENIB_PACKAGES
		echo "IB_SRP_TARGET is not included in this package"
	fi
        ex "${OPENIB_SRCDIR}/build_rpm.sh --prefix ${STACK_PREFIX} --build_root ${BUILD_ROOT} --packages ${OPENIB_PACKAGES} -- -kver ${K_VER} --ksrc ${K_SRC} ${OPENIB_PARAMS}"
        ex "$MV -f ${RPM_DIR}/RPMS/$build_arch/${OPENIB_NAME}-*-${OPENIB_VER}-${RPMK_VER}.${build_arch}.rpm $RPMS"
        ex_silent "$RM -rf ${OPENIB_SRCDIR}"
        ex_silent cd -
        let BUILD_COUNTER++
        
        return 0
}


# OpenSM installation
opensm()
{
        if ! check_build_req opensm err; then
            exit 2
        fi    
        cd $BUILD_TMP
        ex_silent "$RM -f ${OPENSM_SOURCE##*/}"
        echo
        echo "Building ${OPENSM_NAME} RPM. Please wait..."
        echo
        ex_silent "tar xzvf ${OPENSM_SOURCE}"
        local OPENSM_SRCDIR=$BUILD_TMP/`echo ${OPENSM_SOURCE##*/} | sed -e 's/.tgz//' | tr -d '[:space:]'`
        MTHOME=${K_SRC}/drivers/infiniband/ib_verbs/hw/mellanox-hca
        TSHOME=${K_SRC}/drivers/infiniband/include
        ex "${OPENSM_SRCDIR}/build_rpm.sh -b ${BUILD_ROOT} -d ${OPENSM_DEST} -m ${BUILD_ROOT}${OPENIB_DEST} -t ${BUILD_ROOT}${TSHOME} -v ${VENDOR}"
        ex "$MV -f ${RPM_DIR}/RPMS/$build_arch/${OPENSM_NAME}-${OPENSM_VER}-${RPMK_VER}.${build_arch}.rpm $RPMS"
        ex_silent "$RM -rf ${OPENSM_SRCDIR}"
        ex_silent cd -    
        let BUILD_COUNTER++
        
        return 0
}

# HCA Tools installation
mst()
{
        if ( $RPM -q ${MST_NAME}-${ARCH} > $NULL 2>&1 ); then
            ex "$RPM -e ${MST_NAME}-${ARCH}"
        fi
        if ! check_build_req mst err; then
            return 2
        fi    
        cd $BUILD_TMP
        ex_silent "$RM -f ${MST_SOURCE##*/}"
        echo
        echo "Building ${MST_NAME}-${ARCH} RPM. Please wait..."
        echo
        ex_silent "tar xzvf ${MST_SOURCE}"
        local MST_SRCDIR=$BUILD_TMP/`echo ${MST_SOURCE##*/} | sed -e 's/.tgz//' | tr -d '[:space:]'`
        ex "${MST_SRCDIR}/build_rpm.sh"
        ex "$RPM -ihv ${RPM_DIR}/RPMS/$build_arch/${MST_NAME}-${ARCH}-${MST_VER}-${RPMK_VER}.${build_arch}.rpm"
        ex "$MV -f ${RPM_DIR}/RPMS/$build_arch/${MST_NAME}-${ARCH}-${MST_VER}-${RPMK_VER}.${build_arch}.rpm $RPMS"
        ex_silent "$RM -rf ${MST_SRCDIR}"
        ex_silent cd -
        let BUILD_COUNTER++
        
        return 0
}

# IBADM installation
ibadm()
{
        # Build/Install MST
        if ! mst; then
                err_echo "Failed to build MST (Mellanox Software Tools) required for IBADM functionality"
                exit 1
        fi
        if ! check_build_req ibadm err; then
            exit 2
        fi    
        cd $BUILD_TMP
        ex_silent "$RM -f ${IBADM_SOURCE##*/}"
        echo
        echo "Building ${IBADM_NAME} RPM. Please wait..."
        echo
        ex_silent "tar xzvf ${IBADM_SOURCE}"
        local IBADM_SRCDIR=$BUILD_TMP/`echo ${IBADM_SOURCE##*/} | sed -e 's/.tgz//' | tr -d '[:space:]'`
        ex "${IBADM_SRCDIR}/build_rpm.sh  -b ${BUILD_ROOT} -d ${IBADM_DEST} -stack ibgd --prefix ${STACK_PREFIX}"
        ex "$MV -f ${RPM_DIR}/RPMS/$build_arch/${IBADM_NAME}-${IBADM_VER}-${RPMK_VER}.${build_arch}.rpm $RPMS"
        ex_silent "$RM -rf ${IBADM_SRCDIR}"
        ex_silent cd -    
        let BUILD_COUNTER++
        
        return 0
}

             
# NCSA MPI installation
mpi_ncsa()
{
        local VMI_RPM=${MPI_NCSA_VMI_NAME}-${MPI_NCSA_VMI_VER}-${RPMK_VER}.${build_arch}.rpm
        local NCSA_RPM=${MPI_NCSA_MPICH_NAME}-${MPI_NCSA_MPICH_VER}-${RPMK_VER}.${build_arch}.rpm
        
        if ! check_build_req mpi_ncsa err; then
            return 2
        fi    
        
            # Add libcurl.so link (W/A for Rocks 0.3.3)
            if [ -d /usr/lib ]; then
                    if [ ! -h /usr/lib/libcurl.so ]; then
                            cd /usr/lib
                            ln -snf libcurl.so.2.0.2 libcurl.so
                            cd -
                    fi
            fi
        
            if [ -d /usr/lib64 ]; then
                    if [ ! -h /usr/lib64/libcurl.so ]; then
                            cd /usr/lib64
                            ln -snf libcurl.so.2.0.2 libcurl.so
                            cd -
                    fi
            fi
        
        
        cd $BUILD_TMP
        ex_silent "$RM -f ${NCSA_SOURCE##*/}"
        echo
        echo "Building NCSA RPM. Please wait..."
        echo
        ex_silent "tar xzvf ${NCSA_SOURCE}"
        local NCSA_SRCDIR=$BUILD_TMP/`echo ${NCSA_SOURCE##*/} | sed -e 's/.tgz//' | tr -d '[:space:]'`
        ex "${NCSA_SRCDIR}/build_vmi_rpm.sh -b ${BUILD_ROOT} --prefix ${STACK_PREFIX}"
        ex "$MV -f ${RPM_DIR}/RPMS/$build_arch/${VMI_RPM} $RPMS"
        ex "env PGF_77=$PGF_77 PGF_90=$PGF_90 ${NCSA_SRCDIR}/build_mpich_rpm.sh -b ${BUILD_ROOT} --prefix ${STACK_PREFIX}"
        ex "$MV -f ${RPM_DIR}/RPMS/$build_arch/${NCSA_RPM} $RPMS"
        ex_silent "$RM -rf ${NCSA_SRCDIR}"
        ex_silent cd -
        let BUILD_COUNTER++
        
        return 0
}

# OSU MPI installation
mpi_osu()
{
        if ! check_build_req mpi_osu err; then
            return 2
        fi    

        echo
        echo "Building OSU RPM. Please wait..."
        echo

        for mpi_comp in ${MPI_COMPILER}
        do
            echo "Using ${mpi_comp} compiler"
            OSU_RPM=${MPI_OSU_MPICH_NAME}_${mpi_comp}-${MPI_OSU_MPICH_VER}-${RPMK_VER}.${build_arch}.rpm
            cd $BUILD_TMP
            ex_silent "$RM -f ${OSU_SOURCE##*/}"
            ex_silent "tar xzvf ${OSU_SOURCE}"
            local OSU_SRCDIR=$BUILD_TMP/`echo ${OSU_SOURCE##*/} | sed -e 's/.tgz//' | tr -d '[:space:]'`
            # ex "env PGF_77=$PGF_77 PGF_90=$PGF_90 ${OSU_SRCDIR}/build_rpm.sh -b ${BUILD_ROOT} --prefix ${STACK_PREFIX}"
            ex "${OSU_SRCDIR}/build_rpm.sh --install_root ${BUILD_ROOT} --prefix ${STACK_PREFIX} --comp ${mpi_comp} --csize ${CLUSTER_SIZE}"
            ex "$MV -f ${RPM_DIR}/RPMS/$build_arch/${OSU_RPM} $RPMS"
            ex_silent "$RM -rf ${OSU_SRCDIR}"
            ex_silent cd -
            let BUILD_COUNTER++
        done        
        return 0
}

# Build PDSH
pdsh()
{
        if [ -e $PDSH_SRC_RPM ]; then
            ex "$RPM -ihv $PDSH_SRC_RPM"
        else
            return 1    
        fi

        if [ -f /etc/SuSE-release ]; then
            if [ `grep VERSION /etc/SuSE-release | cut -d = -f 2 | cut -d . -f 1 | tr -d '[:space:]'` -ge 9 ]; then
                 ex "rpmbuild -bb ${RPM_DIR}/SPECS/${PDSH_NAME}.spec"
            else
                 ex "rpm -bb ${RPM_DIR}/SPECS/${PDSH_NAME}.spec"
            fi
        else
            ex "rpmbuild -bb ${RPM_DIR}/SPECS/${PDSH_NAME}.spec"
        fi

        ex "$MV -f ${RPM_DIR}/RPMS/$build_arch/${PDSH_NAME}*.rpm $RPMS"
        let BUILD_COUNTER++
        
        return 0
            
}

build_packages()
{
        # # Install packages required for IBGD operation
        # # Like ( swig1.1p5, tcl8.3.2 )
        # install_packages

        echo   
        echo "Building InfiniBand Software RPMs. Please wait..."
        echo
        
        if [ -n "$OPENIB_PACKAGES" ]; then
                # Build openib
                openib
        fi        
        
        for package in $EXTRA_PACKAGES
        do
                $package
        done

        return 0
}

# Set configuration parameters and make basic checks
pre_build()
{
        # Check that user is root
        check_user
        
        # Set network environment configuration                  
        if [ -f /etc/SuSE-release ]; then
            DISTRIBUTION="SuSE"
            
            # Check termcap library link
            if [[ ! -e /usr/lib/libtermcap.so && -e /usr/lib/libtermcap.so.2.0.8 ]]; then
                ex_silent "ln -s /usr/lib/libtermcap.so.2.0.8 /usr/lib/libtermcap.so"
            fi
            
        elif [ -f /etc/redhat-release ]; then
            DISTRIBUTION="redhat"
        elif [ -f /etc/rocks-release ]; then
            DISTRIBUTION="Rocks"
        else
            DISTRIBUTION="Unknown"
        fi
        
#         # Check for Infiniband software of other vendors
#         local vendors="topspin iview_fm"
#         local vrpms=""
#
#         for vendor in $vendors
#         do
#            vrpms=`$RPM -qa | grep $vendor`
#            if [ -n "$vrpms" ]; then
#                echo
#                err_echo "You have to remove ${vendor} Infiniband software before continuing:"
#                echo
#                echo "$vrpms"
#                echo
#                exit 9
#            fi
#         done

}

# Run interactive build mode
interactive()
{
           show_menu "build_menu"
        	
                read ans_build
                case $ans_build in
                        1)
                                view_readme
                        ;;    
                        2)
                                if ! select_packages; then
                                   interactive
                                fi    
                                set_package_deps
                                check_build_dependencies
                                
                                echo
                                echo "Going to build IBGD RPMs"
                                echo "The RPM build process uses a temporary directory"
                                echo
                                read -p "Please enter the temporary directory [${BUILD_ROOT}]:"
                                BUILD_ROOT=${REPLY:-$BUILD_ROOT}
                                
                                if [ "${BUILD_ROOT}" != "${BUILD_ROOT_DFT}" ]; then
                                       BUILD_ROOT=${BUILD_ROOT}/IBGD
                                fi
                                
                                [ -d $BUILD_ROOT ] && $RM -rf $BUILD_ROOT
                                
                                check_space_req "$BUILD_SPACE_REQ" "$BUILD_ROOT"
                                
                                echo
                                echo "The build process requires the target installation directory."
                                echo
                                echo "   Note: The build process does not create nor touch the target installation directory."
                                echo "         The directory will only be created when you actually install the IBGD RPMs."
                                echo
                                read -p "Please enter the IBGD installation directory [${STACK_PREFIX}]:"
                                STACK_PREFIX=${REPLY:-$STACK_PREFIX}
                                echo   
                                # Set  parameters with new values
                                . $CWD/build_env.sh
                                echo
                                pass_echo "Chosen temporary directory for IBGD RPMs is: $BUILD_ROOT"
                                pass_echo "RPM packages will be stored under: $PKG_SOURCE_DIR/RPMS"
                                pass_echo "IBGD Software is configured to be installed under: ${STACK_PREFIX}"
                                echo
                                read -p "Do you want to continue? [Y/n]:" ans_r
                                if ! [[ "$ans_r" == "" || "$ans_r" == "y" || "$ans_r" == "Y" || "$ans_r" == "yes" ]]; then
                                    return 1
                                fi
                                
                                
                                if [ ! -d "$BUILD_TMP" ]; then
                                   ex_silent "$MKDIR -p $BUILD_TMP"
                                fi
                                
                                echo
                                pass_echo "It may take up to 40 minutes to complete the installation process"
                                echo   
        
                                
                                if build_packages; then
                                    echo
                                    if [ $BUILD_COUNTER -gt 0 ]; then
                                       pass_echo "$BUILD_COUNTER packages were built"
                                    fi
                                    echo
                                    pass_echo "Build process finished ..."
                                
                                    echo
                                    read -p  "Do you want to remove the temporary build directory ${BUILD_ROOT}? [Y/n]:" ans_r
                                    if [[ "$ans_r" == "" || "$ans_r" == "y" || "$ans_r" == "Y" || "$ans_r" == "yes" ]]; then
                                        ex_silent "$RM -rf $BUILD_ROOT"
                                    fi
                                    
                                    getch
                                fi

        		;;
                        Q|q)
                                close_log    
                                exit 0
                        ;;
                        *)
                                return 1
                        ;;
                esac
}

# Run unattended build rpms mode
unattended()
{

                select_packages
                set_package_deps
                check_build_dependencies

                
                if [ ! -d  "$BUILD_ROOT" ]; then
                    ex_silent "$MKDIR -p $BUILD_ROOT"
                fi
        
                check_space_req "$BUILD_SPACE_REQ" "$BUILD_ROOT"
                
                if [ ! -d "$BUILD_TMP" ]; then
                   ex_silent "$MKDIR -p $BUILD_TMP"
                fi
           
                if build_packages; then
                     echo
                     if [ $BUILD_COUNTER -gt 0 ]; then
                        pass_echo "$BUILD_COUNTER packages were built"
                     fi
                     if [ $EXISTING_RPMS -gt 0 ]; then
                        pass_echo "$EXISTING_RPMS RPMs already exist"
                     fi
                     echo
                     
                     pass_echo "Building process finished ..."
                fi
        
            # Remove BUILD_TMP directory
            if [ "$BUILD_ROOT" == "${BUILD_ROOT_DFT}" ]; then
               ex_silent "$RM -rf $BUILD_ROOT"
            fi   
        #     if [ $NEED_REBOOT -ne 0 ]; then
        #         echo
        #         echo    "Please reboot the system"
        #         echo
        #     fi
}

###
### Main
###

main()
{

        pre_build
         
        if [ $INTERACTIVE -eq 1 ]; then
            interactive
        else
            unattended
        fi
        
        return 0
}

open_log
while true
do
    main
    if [ $INTERACTIVE -eq 0 ]; then
        exit 0
    fi    
done
