# This software is available to you under a choice of one of two
# licenses.  You may choose to be licensed under the terms of the GNU
# General Public License (GPL) Version 2, available at
# <http://www.fsf.org/copyleft/gpl.html>, or the OpenIB.org BSD
# license, available in the LICENSE.TXT file accompanying this
# software.  These details are also available at
# <http://openib.org/license.html>.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS                               
# BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
# ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# Copyright (c) 2004 Mellanox Technologies Ltd.  All rights reserved.
#
# Description: IBGD building environment
#


ARCH=${ARCH:-$(uname -m)}
K_VER=${K_VER:-$(uname -r)}
K_SRC=${K_SRC:-/lib/modules/${K_VER}/build}
KERNEL_TREE_ROOT=${K_SRC%*/*}

if [ -d ${KERNEL_TREE_ROOT}/source/ ]; then
   K_SRC=${KERNEL_TREE_ROOT}/source
fi

RPMK_VER=$(echo -n ${K_VER} | tr -s '-' '_')
IB_PORTS_NUM=2

NULL=/dev/null
curr_dir=$PWD

RM=/bin/rm
CP=/bin/cp
MV=/bin/mv
MKDIR=/bin/mkdir
RPM=/bin/rpm
IFCONFIG=/sbin/ifconfig
CHMOD=/bin/chmod
LSMOD=/sbin/lsmod
WHICH=/usr/bin/which
DF=/bin/df
REBOOT="/sbin/shutdown -r 0"
SUPPORTED=0
NEED_REBOOT=0
EXECUTING=0
RUNNING_CLEANUP=0
TYPICAL=1
MINIMAL=2
ALL=3
CUSTOM=4

prog=`basename $0`

# Disk space required for building IBGD RPMs (300 MB)
BUILD_SPACE_REQ=307200

# Disk space required for installing IBGD RPMs (200 MB)
RUN_SPACE_REQ=204800

# Default vendor is "ts" (second - mtl)
VENDOR="ts"
build_arch=`$RPM --showrc | grep -w "build arch" | cut -d : -f 2 | tr -d '[:space:]'`

# Default Fortran compiler for MPI installation
MPI_SUPPORTED_COMPILERS="gcc pgi intel pathscale"

# Versions definitions:

OPENIB_NAME="ib"
OPENIB_VER="1.8.1"

# OpenIB sub packages
IB_ALL_PACKAGES="ib_verbs ib_ipoib ib_cm ib_sdp ib_dapl ib_srp"
IB_MINIMAL_PACKAGES="ib_verbs"


OPENSM_NAME="opensm"
OPENSM_VER="1.8.0_1"

MST_NAME="mst"
MST_VER="4.3.1"

IBADM_NAME="ibadm"
IBADM_VER="1.8.1"

IBGFVS_NAME="ibgfvs"
IBGFVS_VERSION="1.0.0"

MPI_OSU_NAME="mpi_osu"
MPI_OSU_VER="0.9.5_mlx1.0.2"
MPI_OSU_MPICH_NAME="mpich_mlx"
MPI_OSU_MPICH_VER="0.9.5_mlx1.0.2"
CLUSTER_SIZE=${CLUSTER_SIZE:-small}

MPI_NCSA_NAME="MPI_NCSA"
MPI_NCSA_VER=""
MPI_NCSA_MPICH_NAME="mpich_mlx-ch_vmi"
MPI_NCSA_MPICH_VER="2.0.1"
MPI_NCSA_VMI_NAME="vmi_mlx"
MPI_NCSA_VMI_VER="2.0.1"

PDSH_NAME="pdsh"
PDSH_VER="2.3-1"

IBFW_NAME="ibgd1.8.1.fwrel.tgz"

# Packages description
ib_verbs_descr="IB HCA driver and Access Layer"
ib_ipoib_descr="IPoIB Protocol"
ib_cm_descr="Connection Manager"
ib_sdp_descr="Socket Direct Protocol"
ib_dapl_descr="DAT Interface"
ib_srp_descr="SCSI Remote Protocol"
ib_srp_target_descr="SCSI Remote Protocol Target"
opensm_descr="Open Subnet Manager"
ibadm_descr="IB Administration Tools"
mpi_osu_descr="MPI OSU Protocol and Basic Tests"
pdsh_descr="Multithreaded Remote Shell Client"


# Sources path definitions
PKG_SOURCE_DIR=${PWD}

OPENIB_SOURCE=${PKG_SOURCE_DIR}/SOURCES/${OPENIB_NAME}-${OPENIB_VER}.tgz
OPENSM_SOURCE=${PKG_SOURCE_DIR}/SOURCES/${OPENSM_NAME}-${OPENSM_VER}.tgz
MST_SOURCE=${PKG_SOURCE_DIR}/SOURCES/${MST_NAME}-${ARCH}-${MST_VER}.tgz
IBADM_SOURCE=${PKG_SOURCE_DIR}/SOURCES/${IBADM_NAME}-${IBADM_VER}.tgz
IBGFVS_SOURCES=${PKG_SOURCE_DIR}/SOURCES/${IBGFVS_NAME}-${IBGFVS_VERSION}.tgz
NCSA_SOURCE=${PKG_SOURCE_DIR}/SOURCES/${MPI_NCSA_NAME}.tgz
OSU_SOURCE=${PKG_SOURCE_DIR}/SOURCES/${MPI_OSU_NAME}-${MPI_OSU_VER}.tgz
PDSH_SRC_RPM=${PKG_SOURCE_DIR}/SOURCES/${PDSH_NAME}-${PDSH_VER}.src.rpm
SCRIPTS_SOURCE=${PKG_SOURCE_DIR}
DOCS_SOURCE=${PKG_SOURCE_DIR}/docs
#FW_SOURCE_DIR=${PKG_SOURCE_DIR}/SOURCES/FW

PROPEL=${PKG_SOURCE_DIR}/SOURCES/propel.sh

IBFW=${PKG_SOURCE_DIR}/SOURCES/$IBFW_NAME

RPMS=${PKG_SOURCE_DIR}/RPMS

if [ ! -d ${RPMS} ]; then
	$MKDIR -p ${RPMS} || exit 1
fi

###
### Check configuration file and include it's info if exist
###
[ -z "$CONFIG" ] && CONFIG=${PKG_SOURCE_DIR}/ibgd.conf

SET_CONFIG_NET=0
if [ -z "$CONFIG_NET" ]; then
        CONFIG_NET=${PKG_SOURCE_DIR}/ibgd_net.conf
        touch $CONFIG_NET
        cat $NULL > ${CONFIG_NET}
	SET_CONFIG_NET=1
else
        if [ -s "$CONFIG_NET" ]; then
		. $CONFIG_NET
		SET_CONFIG_NET=0
	else
		SET_CONFIG_NET=1
	fi
fi

if [ $INTERACTIVE -eq 0 ]; then
	
        if [ -s "$CONFIG" ]; then
        	. $CONFIG 
        else	
        	echo "$CONFIG file does not exist or empty"
             exit 1
        fi
        TT_COLOR=no
else
        touch $CONFIG || exit 1
        TT_COLOR=yes
fi

###
### Build Environment definition
###
BUILD_ROOT_DFT=${BUILD_ROOT:-/var/tmp/IBGD}
BUILD_ROOT=${BUILD_ROOT_DFT}
BUILD_TMP=/tmp

###
### Destination path definitions
###
STACK_PREFIX=${STACK_PREFIX:-/usr/local/ibgd}

OPENIB_DEST=${STACK_PREFIX}/driver/infinihost
OPENSM_DEST=${STACK_PREFIX}/apps/osm
MST_DEST=/usr/mst
IBADM_DEST=${STACK_PREFIX}/apps/ibadm
NCSA_DEST=${STACK_PREFIX}/mpi/ncsa
OSU_DEST=${STACK_PREFIX}/mpi/osu
SCRIPTS_DEST=${STACK_PREFIX}/scripts
FW_DEST=${STACK_PREFIX}

###
### Installation Types
###
MINIMAL_PACKAGES="${IB_MINIMAL_PACKAGES}"
TYPICAL_PACKAGES="ib_verbs ib_ipoib opensm ibadm mpi_osu"
ALL_PACKAGES="${IB_ALL_PACKAGES} opensm ibadm mpi_osu pdsh"
PREV_RELEASE_PACKAGES="openib mpi_ncsa thca"

###
### Requirements definitions per package
###
ib_verbs_RUN_REQ="glib glibc readline ncurses"
ib_verbs_BUILD_REQ="gcc glib glibc glibc-devel termcap readline readline-devel ncurses ncurses-devel"
    
ib_ipoib_RUN_REQ=""
ib_ipoib_BUILD_REQ=""

ib_cm_RUN_REQ=""
ib_cm_BUILD_REQ=""

ib_sdp_RUN_REQ=""
ib_sdp_BUILD_REQ=""

ib_dapl_RUN_REQ=""
ib_dapl_BUILD_REQ=""

ib_srp_RUN_REQ=""
ib_srp_BUILD_REQ=""

ib_srp_target_RUN_REQ=""
ib_srp_target_BUILD_REQ=""

opensm_RUN_REQ=""
opensm_BUILD_REQ="glib glib-devel glibc glibc-devel readline readline-devel"

mst_RUN_REQ=""
mst_BUILD_REQ=""

ibadm_RUN_REQ="expat"
ibadm_BUILD_REQ="automake autoconf libtool expat zlib-devel bison"

mpi_ncsa_RUN_REQ="curl"
mpi_ncsa_BUILD_REQ="curl curl-devel libtool autoconf_2.52 automake_1.5"    

mpi_osu_RUN_REQ=""
mpi_osu_BUILD_REQ=""    

pdsh_RUN_REQ=""
pdsh_BUILD_REQ=""    

LOG=/tmp/IBGD.$$.log

# COLOR SETTINGS
SETCOLOR_SUCCESS="echo -en \\033[1;34m"
SETCOLOR_FAILURE="echo -en \\033[1;31m"
SETCOLOR_WARNING="echo -en \\033[1;35m"
SETCOLOR_NORMAL="echo -en \\033[0;39m"

# Print message with "ERROR:" prefix
err_echo()
{
   [ "$TT_COLOR" = "yes" ] && $SETCOLOR_FAILURE
   echo -n "ERROR: $@" | tee -a $LOG
   [ "$TT_COLOR" = "yes" ] && $SETCOLOR_NORMAL
   echo
   return 0
}

# Print message with "WARNING:" prefix
warn_echo()
{
   [ "$TT_COLOR" = "yes" ] && $SETCOLOR_WARNING
   echo -n "WARNING: $@" | tee -a $LOG
   [ "$TT_COLOR" = "yes" ] && $SETCOLOR_NORMAL
   echo
   return 0
}

# Print message (bold)
pass_echo()
{
   [ "$TT_COLOR" = "yes" ] && $SETCOLOR_SUCCESS
   echo -n "$@" | tee -a $LOG
   [ "$TT_COLOR" = "yes" ] && $SETCOLOR_NORMAL
   echo
   return 0
}
                    

# Remove LOG file if exist and create a new one                    
open_log()
{
        $RM -f $LOG    
        touch $LOG
}

close_log()
{
	echo "`basename $0` finished: `date`" >> $LOG    
}

# Wait for user to press any key and continue
getch()
{
        local ans
        
        echo -n "Press Enter to continue..."
        read ans
        return 0
}


# Clean temporary files on failure or after cancelling installation
cleanup()
{
        local ACTION
        local ans

        if [ $RUNNING_CLEANUP -eq 1 ]; then
            return 0
        fi
        
        RUNNING_CLEANUP=1
        
        if [ $INTERACTIVE -eq 1 ]; then
                # IF in the install mode run uninstall
                if [ "$prog" == "install.sh" ]; then
                        echo
                        read -p  "Do you want to remove already installed IBGD packages? [Y/n]:" ans
                        if ! [[ "$ans" == "" || "$ans" == "y" || "$ans" == "Y" || "$ans" == "yes" ]]; then
                                return 0
                        fi   
                        ACTION="uninstall"
                else
                        # IF in the build mode delete BUILD_ROOT
                        echo
                        if [ -d "$BUILD_ROOT" ]; then
                                read -p  "Do you want to remove temporary build directory ${BUILD_ROOT}? [Y/n]:" ans
                                if ! [[ "$ans" == "" || "$ans" == "y" || "$ans" == "Y" || "$ans" == "yes" ]]; then
                                        return 0
                                fi   
                        fi
                        ACTION="delete"
                fi
        else # Non interactive mode
                if [ "$prog" == "install.sh" ]; then
                        ACTION="leave" # ACTION="uninstall"
                else
                        ACTION="leave"
                fi
        fi
        
        case "$ACTION" in
              "uninstall")
                      rmInstalled
              ;;
              "delete")
                      [ -d "$BUILD_ROOT" ] && $RM -rf $BUILD_ROOT
              ;;
              *)
              ;;
        esac
        
        if [ $INTERACTIVE -eq 1 ]; then
                getch
        fi
        
}

# Returns "0" if $1 is integer and "1" otherwise
is_integer() 
{
        printf "%s\n" $1 |grep -E "^[+-]?[0-9]+$" > $NULL
        return $?
}

# Stop running propeller
kill_propel()
{
        [ -z $PR_PID ] && return 0
        if ! is_integer $PR_PID; then
                return 0
        fi
        if ! ( ps -p $PR_PID > $NULL 2>&1 ); then
                return 0
        fi
        
        kill -15 $PR_PID  > $NULL 2>&1
        sleep 1
}

trap_handler()
{
        local ans
        
        # Disable traps while cleaning system
        trap 'echo Please wait ...' 2 9 15
        
        if [ $EXECUTING -eq 0 ]; then
                echo
                read -p "Do you want to Cancel Installation? [y/N]:" ans
                if [ "$ans" == "y" ]; then
                        kill_propel
                        err_echo "Installation aborted by user"
                        cleanup
                        exit 1
                else
                        echo "Continuing from the same point..."
                        return 0
                fi
        else
                kill_propel
                err_echo "Installation aborted by user"
                cleanup
                exit 1    
        fi
}

# Get user's trap (Ctrl-C or kill signals)
trap 'trap_handler' 2 9 15

# Execute command $@ and check exit status
ex()
{
        EXECUTING=1
        
        $PROPEL &
        PR_PID=$!
     
        echo Running $@
        eval "$@" >> $LOG 2>&1
        if [ $? -ne 0 ]; then
             echo
             err_echo Failed executing \"$@\"
             echo
             cd ${curr_dir}
             echo See log file: $LOG
             echo
     
             kill_propel
     
             if [ $INTERACTIVE -eq 1 ]; then
                getch
             fi
             cleanup
             exit 5
        fi
     
        kill_propel
        
        EXECUTING=0   
}

# Execute command $@ in silent mode and check exit status
ex_silent()
{
        EXECUTING=1
        
        #   echo Running $@
        eval "$@" >> $LOG 2>&1
        if [ $? -ne 0 ]; then
             echo
             err_echo Failed executing \"$@\"
             echo
             cd ${curr_dir}
             echo See log file: $LOG
             echo
        
             if [ $INTERACTIVE -eq 1 ]; then
                getch
             fi
             cleanup
             exit 5
        fi
        
        EXECUTING=0   
}

# Check that "root" is the current user
check_user()
{
        local user
        user=`/usr/bin/id -u`
        if [ $user -ne 0 ]; then
             err_echo "You must be root to run this install program"
             exit 1
        fi
}

# Check input from user (default is "No")
get_ans()
{
read ans
case $ans in
    Y|YES|y|yes)
        return 0
    ;;    
    *)
        return 1    
     ;;
esac        
}

# Display installation/build menues
show_menu()
{

    clear
     case "$1" in
                 "main_menu")
cat << EOF_INST

         InfiniBand Gold Distribution (IBGD) Software Installation Menu

          1) View IBGD Installation Guide
          2) Install IBGD Software 
          3) Show Installed Software
          4) Configure IPoIB Network Interface, IBADM Server, and OpenSM Server
          5) Uninstall IBGD Software
          6) Build IBGD Software RPMs

          Q) Exit

EOF_INST
          
	echo -n   "Select Option [1-6]:"
                         ;;
                 "build_menu")
cat << EOF_BUILD

         InfiniBand Software Build Menu

          1) View README (read "Building IB SW RPM packages")
          2) Build InfiniBand Software RPMs

          Q) Exit

EOF_BUILD

	echo -n   "Select Option [1-2]:"

                 ;;
                 "select_mode_menu")
cat << EOF_MODE_MENU

          Select IBGD Software

          1) Typical (ib-verbs, ib-ipoib, opensm, ibadm and mpi)
          2) Minimal (ib-verbs only)
          3) All packages (ib-verbs, ib-ipoib, ib-cm, ib-sdp, ib-dapl, ib-srp, opensm, ibadm, mpi, pdsh)
          4) Customize

          Q) Exit

EOF_MODE_MENU

	echo -n   "Select Option [1-4]:"

                 ;;
                 *)
                     return 1
                 ;;
     esac
    return 0
}

# Show README or IBGD Installation guide file
view_readme()
{
        local prog=`basename $0`
        
        if [ "$prog" == "install.sh" ]; then
                if [ -e $PKG_SOURCE_DIR/docs/IBGD_Installation_Guide.txt ]; then
        	        less $PKG_SOURCE_DIR/docs/IBGD_Installation_Guide.txt
                else
                        warn_echo $PKG_SOURCE_DIR/docs/IBGD_Installation_Guide.txt does not exist
                        getch
                fi                        
        else
                if [ -e $PKG_SOURCE_DIR/docs/README.txt ]; then
        	        less $PKG_SOURCE_DIR/README.txt
                else
                        warn_echo $PKG_SOURCE_DIR/README.txt does not exist
                        getch
                fi                        
        	
        fi
        
        return 0
}

# Check disk space requirments $1 - required size $2 - directory to check
check_space_req()
{
        local space_req=$1
        local dir=$2
        
        shift 2
        
        while [ ! -d $dir ]
        do
                dir=${dir%*/*}
        done
        
        local avail_space=`$DF $dir | tail -1 | awk '{print$4}' | tr -d '[space:]'`
        
        if ! is_integer $avail_space; then
                # Wrong avail_space found
                return 0
        fi
        
        if [ $avail_space -lt $space_req ]; then
                echo
                err_echo "Not enough disk space in the ${dir} directory. Required ${space_req}KB"
                echo
                if [ $INTERACTIVE -eq 1 ]; then
                      	getch
                fi
                exit 1
        fi

        return 0
}

# Compare package versions $1 and $2. Return 0 if $1=$2, 1 if $1 > $2, 2 - if $1 < $2
compare()
{
        local ver1=`echo $1 | tr -d '[:alpha:]'`
        local ver2=`echo $2 | tr -d '[:alpha:]'`
        local dots1=0
        local dots2=0
        local dots=0
        local diff=0
        
        if [ "$ver1" == "$ver2" ]; then
            return 0
        fi
        
        [ "$ver1" == "" ] && ver1=0
        [ "$ver2" == "" ] && ver2=0
        
        dots1=`echo $ver1 | tr -d '[:digit:]' | tr -d '[space:]' | wc -c | tr -d '[space:]'`
        dots2=`echo $ver2 | tr -d '[:digit:]' | tr -d '[space:]' | wc -c | tr -d '[space:]'`
        
        if [ $dots1 -gt $dots2 ]; then
            dots=$dots2
            diff=1
        elif [ $dots1 -lt $dots2 ]; then
            dots=$dots1
            diff=2
        else
            dots=$dots1
            diff=0    
        fi
        
        let dots++
        for (( i=1 ; i <= $dots ; i++ ))
        do
            [ `echo $ver1 | cut -d . -f $i | tr -d '[space:]'` -gt `echo $ver2 | cut -d . -f $i | tr -d '[space:]'` ] && return 1
            [ `echo $ver1 | cut -d . -f $i | tr -d '[space:]'` -lt `echo $ver2 | cut -d . -f $i | tr -d '[space:]'` ] && return 2
        done
        
        [ $diff -eq 1 ] && return 1
        [ $diff -eq 2 ] && return 2
        
        return 3
}

# View installed packages
showInstalled()
{
        local RC=0
        for package in $ALL_PACKAGES $PREV_RELEASE_PACKAGES
        do
            pack=$package
            if [ "$pack" == "thca" ]; then
		THCA_NAME="thca_linux"
                if ( $RPM -q ${THCA_NAME} > $NULL 2>&1 ); then
                    local DESCR=`$RPM -qi ${THCA_NAME} | grep -A 1 Description | tail -1`
                    local VERSION=`$RPM -q ${THCA_NAME}`
                    echo
                    echo "NAME: $DESCR"  
                    echo "VERSION: $VERSION"
                    echo
                    let RC++
                elif [[ ! -z $MTHOME && -e $MTHOME/BUILD_ID ]]; then
                    cat $MTHOME/BUILD_ID
                    echo "THCA driver installed not in RPM format"
                fi 
            elif [ "$pack" == "mst" ]; then
                if ( $RPM -q ${MST_NAME}-${ARCH} > $NULL 2>&1 ); then
                    local DESCR=`$RPM -qi ${MST_NAME}-${ARCH} | grep -A 1 Description | tail -1`
                    local VERSION=`$RPM -q ${MST_NAME}-${ARCH}`
                    echo "NAME: $DESCR"  
                    echo "VERSION: $VERSION"
                    echo
                    let RC++
                fi    
            elif [ "$pack" == "ib_verbs" ]; then
                if ( $RPM -q ib-verbs > $NULL 2>&1 ); then
                    local DESCR=`$RPM -qi ib-verbs | grep -A 1 Description | tail -1`
                    local VERSION=`$RPM -q ib-verbs`
                    echo "NAME: $DESCR"  
                    echo "VERSION: $VERSION"
                    echo
                    let RC++
                fi    
            elif [ "$pack" == "ib_ipoib" ]; then
                if ( $RPM -q ib-ipoib > $NULL 2>&1 ); then
                    local DESCR=`$RPM -qi ib-ipoib | grep -A 1 Description | tail -1`
                    local VERSION=`$RPM -q ib-ipoib`
                    echo "NAME: $DESCR"  
                    echo "VERSION: $VERSION"
                    echo
                    let RC++
                fi    
            elif [ "$pack" == "ib_cm" ]; then
                if ( $RPM -q ib-cm > $NULL 2>&1 ); then
                    local DESCR=`$RPM -qi ib-cm | grep -A 1 Description | tail -1`
                    local VERSION=`$RPM -q ib-cm`
                    echo "NAME: $DESCR"  
                    echo "VERSION: $VERSION"
                    echo
                    let RC++
                fi    
            elif [ "$pack" == "ib_sdp" ]; then
                if ( $RPM -q ib-sdp > $NULL 2>&1 ); then
                    local DESCR=`$RPM -qi ib-sdp | grep -A 1 Description | tail -1`
                    local VERSION=`$RPM -q ib-sdp`
                    echo "NAME: $DESCR"  
                    echo "VERSION: $VERSION"
                    echo
                    let RC++
                fi    
            elif [ "$pack" == "ib_dapl" ]; then
                if ( $RPM -q ib-dapl > $NULL 2>&1 ); then
                    local DESCR=`$RPM -qi ib-dapl | grep -A 1 Description | tail -1`
                    local VERSION=`$RPM -q ib-dapl`
                    echo "NAME: $DESCR"  
                    echo "VERSION: $VERSION"
                    echo
                    let RC++
                fi    
            elif [ "$pack" == "ib_srp" ]; then
                if ( $RPM -q ib-srp > $NULL 2>&1 ); then
                    local DESCR=`$RPM -qi ib-srp | grep -A 1 Description | tail -1`
                    local VERSION=`$RPM -q ib-srp`
                    echo "NAME: $DESCR"  
                    echo "VERSION: $VERSION"
                    echo
                    let RC++
                fi    
            elif [ "$pack" == "ib_srp_target" ]; then
                if ( $RPM -q ib-srp_target > $NULL 2>&1 ); then
                    local DESCR=`$RPM -qi ib-srp_target | grep -A 1 Description | tail -1`
                    local VERSION=`$RPM -q ib-srp_target`
                    echo "NAME: $DESCR"  
                    echo "VERSION: $VERSION"
                    echo
                    let RC++
                fi    
           elif [ "$pack" == "mpi_ncsa" ]; then
                if ( $RPM -q  ${MPI_NCSA_MPICH_NAME} > $NULL 2>&1 ); then
                    local DESCR=`$RPM -qi ${MPI_NCSA_VMI_NAME} | grep -A 1 Description | tail -1`
                    local VERSION=`$RPM -q ${MPI_NCSA_VMI_NAME}`
                    echo "NAME: $DESCR"  
                    echo "VERSION: $VERSION"
                    echo
                    let RC++
                    local DESCR=`$RPM -qi ${MPI_NCSA_MPICH_NAME} | grep -A 1 Description | tail -1`
                    local VERSION=`$RPM -q ${MPI_NCSA_MPICH_NAME}`
                    echo "NAME: $DESCR"  
                    echo "VERSION: $VERSION"
                    echo
                    let RC++
                fi
            elif [ "$pack" == "mpi_osu" ]; then
                for mpi_comp in ${MPI_SUPPORTED_COMPILERS}
                do
                    if ( $RPM -q  ${MPI_OSU_MPICH_NAME}_${mpi_comp} > $NULL 2>&1 ); then
                        local DESCR=`$RPM -qi ${MPI_OSU_MPICH_NAME}_${mpi_comp} | grep -A 1 Description | tail -1`
                        local VERSION=`$RPM -q ${MPI_OSU_MPICH_NAME}_${mpi_comp}`
                        echo "NAME: $DESCR"  
                        echo "VERSION: $VERSION"
                        echo
                        let RC++
                    fi
                done    
            else
                if ( $RPM -q $pack > $NULL 2>&1 ); then
                    local DESCR=`$RPM -qi $pack | grep -A 1 Description | tail -1`
                    local VERSION=`$RPM -q $pack`
                    echo "NAME: $DESCR"  
                    echo "VERSION: $VERSION"
                    echo
                    let RC++
                fi    
            fi    
        done
        
        if [ $RC -eq 0 ]; then
            echo
            echo    There is no IBGD Software Installed
            echo
        fi
  
        return 0
}


# Check $1 package RPM Build Requires ( Like kernel sources ... )
check_build_req()
{
        local RC=0
        local mode=$2
        local breq_pack
        local breq_pack_name
        local breq_pack_ver
        local breq_pack_rel
        local curr_pack_ver
            
        for breq_pack in `eval echo \$\{${1}_BUILD_REQ\}`
        do
                if ( echo $breq_pack | grep '_' > $NULL 2>&1 ); then
                        breq_pack_name=`echo $breq_pack | cut -d '_' -f 1`
                        breq_pack_ver=`echo $breq_pack | cut -d '_' -f 2`
                        breq_pack_rel=`echo $breq_pack | cut -d '_' -f 3`
                else
                        breq_pack_name=$breq_pack
                        breq_pack_ver=""
                        breq_pack_rel=""
                fi    
                
                if ! ( rpm -q $breq_pack_name > $NULL 2>&1 ); then
                        [ "$mode" == "warn" ] && warn_echo "The $breq_pack_name package is required to build $1 RPM"
                        [ "$mode" == "err" ] && err_echo "The $breq_pack_name package is required to build $1 RPM"
                        let RC++
                else
                    if [ "$breq_pack_ver" != "" ]; then
                            curr_pack_ver=`$RPM --queryformat "[%{VERSION}]" -q $breq_pack_name`
                            compare $breq_pack_ver $curr_pack_ver
                            if [ $? -eq 1 ]; then
                                    [ "$mode" == "warn" ] && warn_echo "The $breq_pack_name >= $breq_pack_ver package is required to build $1"
                                    [ "$mode" == "err" ] && err_echo "The $breq_pack_name >= $breq_pack_ver package is required to build $1"
                                    let RC++
                            fi
                    fi
                
                fi
        done
            
        return $RC    
}

# Check if required compiler - $1 exist
is_compiler()
{
    local compiler=$1
    shift

    case ${compiler} in
        gcc)
            if ( $WHICH g77 > $NULL 2>&1 ); then
                return 0
            fi
        ;;
        pathscale)
            if ( $WHICH pathcc > $NULL 2>&1 && $WHICH pathCC > $NULL 2>&1 && $WHICH pathf90 > $NULL 2>&1 ); then
                return 0
            fi
        ;;
        pgi)
            if ( $WHICH pgf77 > $NULL 2>&1 && $WHICH pgf90 > $NULL 2>&1 ); then
                return 0
            fi
        ;;
        intel)
            if ( $WHICH icc > $NULL 2>&1 && $WHICH icpc > $NULL 2>&1 && $WHICH ifort > $NULL 2>&1 ); then
                return 0
            fi
        ;;
        *)
            warn_echo "Not supported compiler"
            return 1
        ;;
    esac
    
    return 1
}

# Check if required MPI rpms are exist
is_mpi_rpms()
{
    for mpi_comp in ${MPI_COMPILER}
    do
        [ ! -f ${RPMS}/${MPI_OSU_MPICH_NAME}_${mpi_comp}-${MPI_OSU_MPICH_VER}-${RPMK_VER}.${build_arch}.rpm ] && return 1
    done    

    return 0
}

# Set Compilation environment for OSU MPI
set_mpi_env()
{
        local COMPILERS_FOUND=""
        local EXCLUDED_COMPILERS=""

        if [ $INTERACTIVE -eq 1 ]; then
        
            for mpi_compiler in ${MPI_SUPPORTED_COMPILERS}
            do
                if is_compiler ${mpi_compiler}; then
                    COMPILERS_FOUND="$COMPILERS_FOUND ${mpi_compiler}"
                fi
            done
                
            # COMPILERS_FOUND=$(echo $COMPILERS_FOUND | tr -s ' ' '\n' | sort -rn | uniq | tr -s '\n' ' ')
            if [ -n "$COMPILERS_FOUND" ]; then
                echo
                echo "The following compiler(s) on your system can be used to build/install MPI: $COMPILERS_FOUND"
                echo

                if [ $(echo -n ${COMPILERS_FOUND} | wc -w) -gt 1 ]; then
                    echo "Next you will be prompted to choose the compiler(s) with which to build/install the MPI RPM(s)"
                    echo
                fi
            fi

            MPI_COMPILER=""
                                   
            for mpi_compiler in $COMPILERS_FOUND
            do
                read -p "Do you wish to create/install an MPI RPM with ${mpi_compiler}? [Y/n]:" ans
                if [[ "$ans" == "" || "$ans" == "y" || "$ans" == "Y" || "$ans" == "yes" ]]; then
                    MPI_COMPILER="$MPI_COMPILER ${mpi_compiler}"
                fi    
            done
        
        else # Unattended mode
	    if [ ! -n "${MPI_COMPILER}" ]; then
		MPI_COMPILER=gcc
		echo "MPI_COMPILER variable is not defined. Using default: MPI_COMPILER=${MPI_COMPILER}"
	    fi
            for mpi_compiler in ${MPI_COMPILER}
            do
                if ! is_compiler ${mpi_compiler}; then
                    EXCLUDED_COMPILERS="$EXCLUDED_COMPILERS ${mpi_compiler}"
                else
                    COMPILERS_FOUND="$COMPILERS_FOUND ${mpi_compiler}"
                fi
            done

            MPI_COMPILER=$(echo $COMPILERS_FOUND | tr -s ' ' '\n' | sort -rn | uniq | tr -s '\n' ' ')
        
        fi

        if [ ! -n "${COMPILERS_FOUND}" ]; then
            warn_echo "No compilers for MPI were found"
            return 1
        fi

        if [ ! -n "${MPI_COMPILER}" ]; then
            warn_echo "No compiler was selected for MPI"
            return 1
        fi
            
        if [ -n "${EXCLUDED_COMPILERS}" ]; then
            warn_echo "The following compilers were not found on the system: ${EXCLUDED_COMPILERS}"
            return 0
        fi

        if [ "$prog" == "build.sh" ]; then
            echo
            echo "The following compiler(s) will be used to build the MPI RPM(s): $MPI_COMPILER"
            echo
        fi    
        
        if ! ( grep -w MPI_COMPILER $CONFIG > $NULL 2>&1 ); then
            echo "MPI_COMPILER=\"${MPI_COMPILER}\"" >> $CONFIG
        fi

        # Set Cluster Size
        if [ $INTERACTIVE -eq 1 ]; then
            if ! is_mpi_rpms || [ "$prog" == "build.sh" ]; then
                printf "\nNext you will be prompted to enter the number of CPUs in your cluster.\n"
                printf "\n\tsmall:\t1 - 63 CPUs"
                printf "\n\tmedium:\t64 - 255 CPUs"
                printf "\n\tbig:\t256+ CPUs"
                printf "\n\nPlease select the size of your cluster [small/medium/big]:"
                read CLUSTER_SIZE
                while true
                do
                    case $CLUSTER_SIZE in
                        small|medium|big)
                            break
                        ;;
                        *)
                            echo -n "Please choose from the following list [small/medium/big]:"
                            read CLUSTER_SIZE
                        ;;
                    esac    
                done
                echo "CLUSTER_SIZE=${CLUSTER_SIZE}" >> $CONFIG
            fi    
        fi
        
        return 0
}

# Select which packages to install
select_packages()
{

        if [ $INTERACTIVE -eq 1 ]; then
        
                # Configuration parameters initialization
                touch $CONFIG
                
                # Save previous configuration file
                [ -s $CONFIG ] && $MV -f $CONFIG ${CONFIG}.save
                
                cat $NULL > $CONFIG
                
                echo STACK_PREFIX=$STACK_PREFIX >> $CONFIG
                echo BUILD_ROOT=$BUILD_ROOT >> $CONFIG

                show_menu "select_mode_menu"
                
                read ans_select
                case $ans_select in
                        $TYPICAL)
                                SELECTED_PACKAGES="$TYPICAL_PACKAGES"
                                for package in $TYPICAL_PACKAGES
                                do
					echo "${package}=y" >> $CONFIG
				done
                        ;;
                        $MINIMAL)
                                SELECTED_PACKAGES="$MINIMAL_PACKAGES"
                                for package in $MINIMAL_PACKAGES
                                do
					echo "${package}=y" >> $CONFIG
				done
                        ;;
                        $ALL)
                                SELECTED_PACKAGES="$ALL_PACKAGES"
                                for package in $ALL_PACKAGES
                                do
					echo "${package}=y" >> $CONFIG
				done
                        ;;
                        $CUSTOM)
                                echo
                                for package in $ALL_PACKAGES
                                do
                                        pack_descr=$(eval echo \$\{${package}_descr\})
                                        case ${package} in
                                            ib_verbs|ib_ipoib|ib_cm|ib_sdp|ib_dapl|ib_srp|ib_srp_target)
                                                    echo -n "Install ${package}: ${pack_descr}? [y/N]:" | sed -e "s/ib_/ib-/"
                                                ;;
                                                *)
                                                    echo -n "Install ${package}: ${pack_descr}? [y/N]:"
                                                ;;
                                        esac
                                                
                                        if get_ans; then
					    if [ "$SELECTED_PACKAGES" == "" ]; then
						SELECTED_PACKAGES="$package"
					    else
                                            	SELECTED_PACKAGES="$SELECTED_PACKAGES $package"
					    fi
                                            echo "${package}=y" >> $CONFIG
                                        else
                                            echo "${package}=n" >> $CONFIG
                                        fi
                                done
                        ;;
                        q|Q)
                                return 1
                        ;;
                        *)
                                select_packages
                        ;;
                esac
        else # Unatended mode
        
                # Read configuration from CONFIG file
                    if [ "$ib_verbs" == "y" ]; then
                        SELECTED_PACKAGES="$SELECTED_PACKAGES ib_verbs"
                    fi
                
                    if [ "$ib_ipoib" == "y" ]; then
                        SELECTED_PACKAGES="$SELECTED_PACKAGES ib_ipoib"
                    fi
                
                    if [ "$ib_cm" == "y" ]; then
                        SELECTED_PACKAGES="$SELECTED_PACKAGES ib_cm"
                    fi
                
                    if [ "$ib_sdp" == "y" ]; then
                        SELECTED_PACKAGES="$SELECTED_PACKAGES ib_sdp"
                    fi
                
                    if [ "$ib_dapl" == "y" ]; then
                        SELECTED_PACKAGES="$SELECTED_PACKAGES ib_dapl"
                    fi
                
                    if [ "$ib_srp" == "y" ]; then
                        SELECTED_PACKAGES="$SELECTED_PACKAGES ib_srp"
                    fi
                
                    if [ "$ib_srp_target" == "y" ]; then
                        SELECTED_PACKAGES="$SELECTED_PACKAGES ib_srp_target"
                    fi
                    
                    if [ "$opensm" == "y" ]; then
                        SELECTED_PACKAGES="$SELECTED_PACKAGES opensm"
                    fi
                    
                    if [ "$ibadm" == "y" ]; then
                        SELECTED_PACKAGES="$SELECTED_PACKAGES ibadm"
                    fi
                    
                    if [ "$mpi_osu" == "y" ]; then
                        SELECTED_PACKAGES="$SELECTED_PACKAGES mpi_osu"
                    fi
                    
                    if [ "$pdsh" == "y" ]; then
                        SELECTED_PACKAGES="$SELECTED_PACKAGES pdsh"
                    fi
        fi
}

# Select packages following their dependencies inside IBGD
set_package_deps()
{        
        OPENIB_PACKAGES=""
        EXTRA_PACKAGES=""
        
        # Create list of packages to build
        for package in $SELECTED_PACKAGES
        do
		package=$(echo -n ${package} | tr -d '[:space:]')
                case $package in
                        ib_verbs)
                                OPENIB_PACKAGES=$(echo "$OPENIB_PACKAGES ib_verbs" | tr -s ' ' '\n' | sort -n | uniq)
                        ;;                                
                        ib_ipoib)
                                OPENIB_PACKAGES=$(echo "$OPENIB_PACKAGES ib_verbs ib_ipoib" | tr -s ' ' '\n' | sort -n | uniq)
                        ;;
                        ib_cm)
                                OPENIB_PACKAGES=$(echo "$OPENIB_PACKAGES ib_verbs ib_ipoib ib_cm" | tr -s ' ' '\n' | sort -n | uniq)
                        ;;
                        ib_sdp)
                                OPENIB_PACKAGES=$(echo "$OPENIB_PACKAGES ib_verbs ib_ipoib ib_cm ib_sdp" | tr -s ' ' '\n' | sort -n | uniq)
                        ;;
                        ib_dapl)
                                OPENIB_PACKAGES=$(echo "$OPENIB_PACKAGES ib_verbs ib_ipoib ib_cm ib_dapl" | tr -s ' ' '\n' | sort -n | uniq)
                        ;;
                        ib_srp)
                                OPENIB_PACKAGES=$(echo "$OPENIB_PACKAGES ib_verbs ib_ipoib ib_cm ib_srp" | tr -s ' ' '\n' | sort -n | uniq)
                        ;;
                        ib_srp_target)
                                OPENIB_PACKAGES=$(echo "$OPENIB_PACKAGES ib_verbs ib_ipoib ib_cm ib_srp_target" | tr -s ' ' '\n' | sort -n | uniq)
                        ;;
                        opensm)
                                EXTRA_PACKAGES=$(echo "$EXTRA_PACKAGES opensm" | tr -s ' ' '\n' | sort -rn | uniq)
                                OPENIB_PACKAGES=$(echo "$OPENIB_PACKAGES ib_verbs" | tr -s ' ' '\n' | sort -n | uniq)
                        ;;
                        ibadm)
                                # Sort "-rn" in order to build opensm before ibadm
                                EXTRA_PACKAGES=$(echo "$EXTRA_PACKAGES opensm ibadm" | tr -s ' ' '\n' | sort -rn | uniq)
                                OPENIB_PACKAGES=$(echo "$OPENIB_PACKAGES ib_verbs" | tr -s ' ' '\n' | sort -n | uniq)
                        ;;
                        mpi_osu)
                                # Set fortran environment for OSU MPI                                    
                                if ! set_mpi_env; then # && [ "$prog" == "build.sh" ]; then
                                        warn_echo "OSU MPI will not be installed"
                                        export mpi_osu=n
                                else                                     
                                        EXTRA_PACKAGES=$(echo "$EXTRA_PACKAGES mpi_osu" | tr -s ' ' '\n' | sort -rn | uniq)
                                        OPENIB_PACKAGES=$(echo "$OPENIB_PACKAGES ib_verbs" | tr -s ' ' '\n' | sort -n | uniq)
                                fi   
                        ;;
                        pdsh)
                                EXTRA_PACKAGES=$(echo "$EXTRA_PACKAGES pdsh" | tr -s ' ' '\n' | sort -rn | uniq)
                        ;;
                        *)     
                        ;;
                esac
        done

        EXTRA_PACKAGES=$(echo -n $EXTRA_PACKAGES | tr -s '\n' ' ')
        OPENIB_PACKAGES=$(echo -n $OPENIB_PACKAGES | tr -s '\n' ' ')
        # SELECTED_PACKAGES="$OPENIB_PACKAGES $EXTRA_PACKAGES"
        if [ $INTERACTIVE -eq 1 ]; then
            echo
            pass_echo "Following is the list of IBGD packages that you have chosen 
            (some may have been added by the installation program due to package dependencies):"
            echo "$OPENIB_PACKAGES $EXTRA_PACKAGES" | tr -s ' ' '\n' | sed -e "s/ib_/ib-/"
        fi    

}



# Checking dependencies for selected packages
check_build_dependencies()
{
        echo
        echo "Checking dependencies. Please wait ..."
        echo

        # Check if kernel sources exists (needed for packages compilation)
        if ! test -d ${K_SRC}; then
            echo
            warn_echo "There is no sources for ${K_VER} kernel installed.
                       Please install ${K_VER} kernel sources to build RPMs on this system"
            echo
            exit 1
        fi
        
        
        # Check dependencies and print Warning messages
        local req_status=0
        
        for package in $SELECTED_PACKAGES
        do
            check_build_req $package err
            let req_status=$req_status+$?
        done

        # Exit if there is missing required software for mandatory packages
        if [ $req_status -gt 0 ]; then
#           getch
           exit 7
        fi   

}

check_run_dependencies()
{
        # Check dependencies and print Warning messages
        local req_status=0
        
        for package in $SELECTED_PACKAGES
        do
                check_run_req $package err
                let req_status=$req_status+$?
        done
        
        # Exit if there is missing required software for mandatory packages
        if [ $req_status -gt 0 ]; then
           exit 7
        fi   
}

count_ib_ports()
{
    local cnt=0
    local tmp_cnt=0
    
    tmp_cnt=$(/sbin/lspci -n | grep "15b3:6282" | wc -l | tr -d '[:space:]') # Arbel mode
    cnt=$[ $cnt + 2*${tmp_cnt} ]
    
    tmp_cnt=$(/sbin/lspci -n | grep -E "15b3:5e8c|15b3:6274" | wc -l | tr -d '[:space:]') # Sinai
    cnt=$[ $cnt + ${tmp_cnt} ]

    tmp_cnt=$(/sbin/lspci -n | grep -E "15b3:5a44|15b3:6278" | wc -l | tr -d '[:space:]') # Tavor mode
    cnt=$[ $cnt + 2*${tmp_cnt} ]
    
    return $cnt
}
