#!/bin/bash                                 

# This software is available to you under a choice of one of two
# licenses.  You may choose to be licensed under the terms of the GNU
# General Public License (GPL) Version 2, available at
# <http://www.fsf.org/copyleft/gpl.html>, or the OpenIB.org BSD
# license, available in the LICENSE.TXT file accompanying this
# software.  These details are also available at
# <http://openib.org/license.html>.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS                               
# BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
# ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# Copyright (c) 2004 Mellanox Technologies Ltd.  All rights reserved.
#
# Description: IBGD installation script
#

usage()
{
cat << EOF

    `basename $0` [ -c <packages config_file> ] [ -net <network config_file> ]

EOF
}

unalias -a

cd ${0%*/*}
CWD=`pwd`

INTERACTIVE=1
CONFIG=""
CONFIG_NET=""

if [ ! -e ./build_env.sh ]; then
    echo "ERROR: $CWD/build_env.sh file is missing"
    exit 1
fi

while [ $# -gt 0 ]
do
    case $1 in
            -c)
                shift
                CONFIG=$1
                INTERACTIVE=0
            ;;
	    -net)
	    	shift
		CONFIG_NET=$1
	    ;;
            *)
                usage
                exit 1
            ;;
    esac
    shift
done
                    

. $CWD/build_env.sh    

SELECTED_PACKAGES=""
OPENIB_PACKAGES=""
EXTRA_PACKAGES=""


# Remove installed packages in case of upgrade
rmInstalled()
{
        local RC=0
        local OLD_PREFIX=""
        echo
        echo "Removing previous IBGD Software installations"
        echo
        
            MPI_OSU_LIST=$(rpm -qa | grep ${MPI_OSU_MPICH_NAME})

            if [ -n "$MPI_OSU_LIST" ]; then
                if [ "$mpi_osu" == "y" ]; then
                    echo "OSU MPI will be reinstalled"
                else
                    echo "OSU MPI will be uninstalled"
                fi    
                
                for mpi_name in $MPI_OSU_LIST
                do 
                    if ( $RPM -q ${mpi_name} > $NULL 2>&1 ); then
                        ex "$RPM -e ${mpi_name}"
                    fi
                done    
            fi
        
            if ( $RPM -q ${MPI_NCSA_MPICH_NAME} > $NULL 2>&1 || $RPM -q ${MPI_NCSA_VMI_NAME} > $NULL 2>&1 ); then
                if [ "$mpi_ncsa" == "y" ]; then
                    echo "NCSA MPI will be reinstalled"
                else
                    echo "NCSA MPI will be uninstalled"
                fi        
                $RPM -q ${MPI_NCSA_MPICH_NAME} > $NULL 2>&1 && ex "$RPM -e ${MPI_NCSA_MPICH_NAME}"
                $RPM -q ${MPI_NCSA_VMI_NAME} > $NULL 2>&1 && ex "$RPM -e ${MPI_NCSA_VMI_NAME}"
            fi    
          
            if ( $RPM -q hmst-${ARCH} > $NULL 2>&1 ); then
                ex "$RPM -e hmst-${ARCH}"
            fi  
            
            if ( $RPM -q ${IBADM_NAME} > $NULL 2>&1 ); then
                if [ "$ibadm" == "y" ]; then
                    echo "IBADM will be reinstalled"
                else
                    echo "IBADM will be uninstalled"    
                fi    
                ex "$RPM -e ${IBADM_NAME}"
            fi    
         
            if ( $RPM -q ${MST_NAME}-${ARCH} > $NULL 2>&1 ); then
                if [ "$mst" == "y" ]; then
                    echo "MST will be reinstalled"
                else
                    echo "MST will be uninstalled"
                fi        
                ex "$RPM -e ${MST_NAME}-${ARCH}"
            fi    
           
            if ( $RPM -q ${OPENSM_NAME} > $NULL 2>&1 ); then
                if [ "$opensm" == "y" ]; then
                    echo "Opensm will be reinstalled"
                else
                    echo "Opensm will be uninstalled"    
                fi    
                ex "$RPM -e ${OPENSM_NAME}"
            fi    
        
            if ( $RPM -q ibtsal > $NULL 2>&1 ); then
                ex "$RPM -e ibtsal"
                let RC++
            fi    
        
            if ( $RPM -q thca-linux > $NULL 2>&1 ); then
                ex "$RPM -e thca-linux"
                let RC++
            fi    
            
            if ( $RPM -q openib > $NULL 2>&1 ); then
                if [ "$openib" == "y" ]; then
                    echo "OPENIB will be reinstalled"
                else
                    echo "OPENIB will be uninstalled"
                fi    
        	OLD_PREFIX=`$RPM -ql openib | grep "utils/create_devs" | sed -e 's/\/utils\/create_devs//'`
                ex "$RPM -e openib"
                let RC++
            elif [[ ! -z $MTHOME && -d $MTHOME ]]; then
                if [ -e $MTHOME/uninstall.sh ]; then
                    if [ "$openib" == "y" ]; then
                        echo "OPENIB will be reinstalled"
                    else
                        echo "OPENIB will be uninstalled"
                    fi    
                    echo
                    echo "  An old version of OPENIB driver is detected and will be removed now"
                    ex "yes | env MTHOME=$MTHOME $MTHOME/uninstall.sh"
                else
                    echo
                    warn_echo "Found an MTHOME variable pointing to $MTHOME. Probably some old Infiniband Sotware ..."
                    echo
                fi    
                let RC++
            elif [ -d /usr/mellanox ]; then
                if [ -e /usr/mellanox/uninstall.sh ]; then
                    if [ "$openib" == "y" ]; then
                        echo "MVAPI will be reinstalled with IBGD"
                    else
                        echo "MVAPI will be uninstalled"
                    fi    
                    echo
                    echo "  Removing MVAPI..."
                    ex "yes | /usr/mellanox/uninstall.sh"
                else
                    echo
                    warn_echo "Found /usr/mellanox directory. Probably some old Infiniband Sotware ..."
                    echo
                fi  
            fi    


            if ( $RPM -q ib-srp_target > $NULL 2>&1 ); then
                ex "$RPM -e ib-srp_target"
                let RC++
            fi    

            if ( $RPM -q ib-srp > $NULL 2>&1 ); then
                ex "$RPM -e ib-srp"
                let RC++
            fi    

            if ( $RPM -q ib-dapl > $NULL 2>&1 ); then
                ex "$RPM -e ib-dapl"
                let RC++
            fi    

            if ( $RPM -q ib-sdp > $NULL 2>&1 ); then
                ex "$RPM -e ib-sdp"
                let RC++
            fi    

            if ( $RPM -q ib-cm > $NULL 2>&1 ); then
                ex "$RPM -e ib-cm"
                let RC++
            fi    

            if ( $RPM -q ib-ipoib > $NULL 2>&1 ); then
                ex "$RPM -e ib-ipoib"
                let RC++
            fi    

            if ( $RPM -q ib-verbs > $NULL 2>&1 ); then
        	OLD_PREFIX=`$RPM -ql ib-verbs | grep "utils/create_devs" | sed -e 's/\/utils\/create_devs//'`
                ex "$RPM -e ib-verbs"
                let RC++
            fi    

        
            [ -d /usr/local/ib_hpc ] && ex_silent "$RM -rf /usr/local/ib_hpc"
            [ -d /usr/local/openib ] && ex_silent "$RM -rf /usr/local/openib"
            [ -d /usr/local/ibgd ] && ex_silent "$RM -rf /usr/local/ibgd"
            [ -d $STACK_PREFIX ] && ex_silent "$RM -rf $STACK_PREFIX"
            [ -d $OLD_PREFIX ] && ex_silent "$RM -rf $OLD_PREFIX"
           
            if [[ ! -d /etc/ibfw || ! -h /etc/ibfw ]]; then
                $RM -rf /etc/ibfw    
            fi  
        
            if [ $RC -gt 0 ]; then
                let NEED_REBOOT++
            fi
        return 0
}

# Check if RPMs for selected packages exists
is_rpms()
{

        for package in $OPENIB_PACKAGES
        do
                package=$(echo $package | sed -e 's/ib_/ib-/g')
                [ ! -f ${RPMS}/${package}-${OPENIB_VER}-${RPMK_VER}.${build_arch}.rpm ] && return 1
        
        done

        for package in $EXTRA_PACKAGES
        do
                case $package in
                     opensm)
                        [ ! -f ${RPMS}/${OPENSM_NAME}-${OPENSM_VER}-${RPMK_VER}.${build_arch}.rpm ] && return 1
                     ;;
                     ibadm)
                        [ ! -f ${RPMS}/${IBADM_NAME}-${IBADM_VER}-${RPMK_VER}.${build_arch}.rpm ] && return 1
                        [ ! -f ${RPMS}/${MST_NAME}-${ARCH}-${MST_VER}-${RPMK_VER}.${build_arch}.rpm ] && return 1
                     ;;
                     mpi_osu)
                        for mpi_comp in ${MPI_COMPILER}
                        do
                            [ ! -f ${RPMS}/${MPI_OSU_MPICH_NAME}_${mpi_comp}-${MPI_OSU_MPICH_VER}-${RPMK_VER}.${build_arch}.rpm ] && return 1
                        done    
                     ;;
                     mpi_ncsa)
                        [ ! -f ${RPMS}/${MPI_NCSA_VMI_NAME}-${MPI_NCSA_VMI_VER}-${RPMK_VER}.${build_arch}.rpm ] && return 1 
                        [ ! -f ${RPMS}/${MPI_NCSA_MPICH_NAME}-${MPI_NCSA_MPICH_VER}-${RPMK_VER}.${build_arch}.rpm ] && return 1
                     ;;
                     pdsh)
                        [ ! -f ${RPMS}/${PDSH_NAME}-${PDSH_VER}.${build_arch}.rpm ] && return 1
                     ;;
                     *)
                        echo "Unknown package !!!"
                        return 1
                     ;;
                esac    
        
        done

        return 0
}

# Check $1 package requirments for running
check_run_req()
{
        local RC=0
        local mode=$2
        local rreq_pack
        local rreq_pack_name
        local rreq_pack_ver
        local rreq_pack_rel
        local curr_pack_ver
        
            for rreq_pack in  `eval echo \$\{${1}_RUN_REQ\}`
            do  
                if ( echo $rreq_pack | grep '_' > $NULL 2>&1 ); then
                    rreq_pack_name=`echo $rreq_pack | cut -d '_' -f 1`          
                    rreq_pack_ver=`echo $rreq_pack | cut -d '_' -f 2`
                    rreq_pack_rel=`echo $rreq_pack | cut -d '_' -f 3`
                else
                    rreq_pack_name=$rreq_pack
                    rreq_pack_ver=""
                    rreq_pack_rel=""
                fi
                
                if ! ( $RPM -q $rreq_pack_name > $NULL 2>&1 ); then
                    [ "$mode" == "warn" ] && warn_echo "The $rreq_pack_name package is required to run $1"
                    [ "$mode" == "err" ] && err_echo "The $rreq_pack_name package is required to run $1"
                    let RC++
                else
                    if [ "$rreq_pack_ver" != "" ]; then
                        curr_pack_ver=`$RPM --queryformat "[%{VERSION}]" -q $rreq_pack_name`
                        compare $rreq_pack_ver $curr_pack_ver
                        if [ $? -eq 1 ]; then
                            [ "$mode" == "warn" ] && warn_echo "The $rreq_pack_name >= $rreq_pack_ver package is required  to run $1"
                            [ "$mode" == "err" ] && err_echo "The $rreq_pack_name >= $rreq_pack_ver package is required to run $1"
                            let RC++
                        fi
                    fi
                fi
            done
        return $RC    
}

# Check if the package $1 is installed
isInstalled()
{
        case $1 in 
            mpi_osu)
                MPI_OSU_LIST=$(rpm -qa | grep ${MPI_OSU_MPICH_NAME})
                if [ -n "$MPI_OSU_LIST" ]; then
                    return 0
                fi    
            ;;
            mpi_ncsa)
                if ( $RPM -q ${MPI_NCSA_MPICH_NAME} > $NULL 2>&1 || $RPM -q ${MPI_NCSA_VMI_NAME} > $NULL 2>&1 ); then
                    return 0
                fi    
            ;;
            ibadm)
                if ( $RPM -q ${IBADM_NAME} > $NULL 2>&1 ); then
                    return 0
                fi    
            ;;
            mst)
                if ( $RPM -q ${MST_NAME}-${ARCH} > $NULL 2>&1 ); then
                    return 0
                fi    
            ;;
            opensm)
                if ( $RPM -q ${OPENSM_NAME} > $NULL 2>&1 ); then
                    return 0
                fi    
            ;;
            openib)
                if ( $RPM -q openib > $NULL 2>&1 ); then
                    return 0
                fi    
            ;;    
            ib-verbs)
                if ( $RPM -q ib-verbs > $NULL 2>&1 ); then
                    return 0
                fi    
            ;;    
            ib-ipoib)
                if ( $RPM -q ib-ipoib > $NULL 2>&1 ); then
                    return 0
                fi    
            ;;    
            ib-cm)
                if ( $RPM -q ib-cm > $NULL 2>&1 ); then
                    return 0
                fi    
            ;;    
            ib-sdp)
                if ( $RPM -q ib-sdp > $NULL 2>&1 ); then
                    return 0
                fi    
            ;;    
            ib-dapl)
                if ( $RPM -q ib-dapl > $NULL 2>&1 ); then
                    return 0
                fi    
            ;;    
            ib-srp)
                if ( $RPM -q ib-srp > $NULL 2>&1 ); then
                    return 0
                fi    
            ;;    
            ib-srp_target)
                if ( $RPM -q ib-srp_target > $NULL 2>&1 ); then
                    return 0
                fi    
            ;;    
            *)
                if ( $RPM -q $1 > $NULL 2>&1 ); then
                    return 0
                fi    
            ;;    
        esac  
        
        return 1
}

# Check if given IP address is valid
is_valid()
{
        local IP=$1
        # Check for 3 dots a.b.c.d
        if [ "`echo $IP | tr -d '[:space:]' | tr -d '[:digit:]' | wc -c`" -ne 3 ]; then
            return 1
        fi
        
        local A=`echo $IP | tr -d '[:space:]' | cut -d . -f 1`
        if [ "`echo $A | tr -d '[:digit:]'`" != "" ]; then return 1;fi
        local B=`echo $IP | tr -d '[:space:]' | cut -d . -f 2`
        if [ "`echo $B | tr -d '[:digit:]'`" != "" ]; then return 1;fi
        local C=`echo $IP | tr -d '[:space:]' | cut -d . -f 3`
        if [ "`echo $C | tr -d '[:digit:]'`" != "" ]; then return 1;fi
        local D=`echo $IP | tr -d '[:space:]' | cut -d . -f 4`    
        if [ "`echo $D | tr -d '[:digit:]'`" != "" ]; then return 1;fi
        
        return 0
}

# Configure OpenSM parameters
osmconf()
{
        local OSM_HOST_tmp
        local LAN_INTERFACE_tmp
        local OSM_HOSTS_LIST_tmp
        
             LAN_INTERFACE_tmp=`$IFCONFIG | grep eth | awk '{print$1}' | sort -n | head -1 | tr -d '[:space:]'`
             LAN_INTERFACE=${LAN_INTERFACE:-$LAN_INTERFACE_tmp}
            
             OSM_HOST_tmp=`$IFCONFIG | sed -e "/\($LAN_INTERFACE\).*/{ N; s/.*inet addr:\([0-9]*\.[0-9]*\.[0-9]*\.[0-9]*\).*/\1/; }" | grep ^[0-9] | head -1`
             echo -n "Enter IP Addresses of all OpenSM Servers in the IB subnet  [${OSM_HOST_tmp} [...]]:"
             read OSM_HOSTS_LIST_tmp

             if [ "${OSM_HOSTS_LIST_tmp}" == "" ]; then
                 OSM_HOSTS_LIST=$OSM_HOST_tmp
             else
                 for osm_host in ${OSM_HOSTS_LIST_tmp}
                 do
                     if ! is_valid $osm_host; then
                         while ! is_valid $osm_host; do
                             echo -n "Enter Valid IP address like [$OSM_HOST_tmp] instead of ${osm_host}:"
                             read osm_host
                             [ "$osm_host" == "" ] && osm_host=$OSM_HOST_tmp
                         done
                         OSM_HOSTS_LIST="${OSM_HOSTS_LIST} ${osm_host}"
                     else
                         OSM_HOSTS_LIST="${OSM_HOSTS_LIST} ${osm_host}"
                     fi
                 done
             fi    

        # Clean previous IBADM configuration in order to prevent duplicate IP
        perl -ni -e 'if (/\#\# OpenSM \#\#$/) { $filter = 1 }' -e 'if (!$filter) { print }' -e 'if (/\#\# END OpenSM \#\#$/){ $filter = 0 }' $CONFIG
        
        echo "## OpenSM ##" >> $CONFIG
        echo OSM_HOSTS_LIST=\"$OSM_HOSTS_LIST\" >> $CONFIG
        echo "## END OpenSM ##" >> $CONFIG
        
        return 0
}

# Configure IBADM parameters
ibadmconf()
{
        local NS_HOST_tmp
        local IBBS_NAME_tmp
        local LAN_INTERFACE_tmp
        
             LAN_INTERFACE_tmp=`$IFCONFIG | grep eth | awk '{print$1}' | sort -n | head -1 | tr -d '[:space:]'`
             LAN_INTERFACE=${LAN_INTERFACE:-$LAN_INTERFACE_tmp}
            
             NS_HOST_tmp=`$IFCONFIG | sed -e "/\($LAN_INTERFACE\).*/{ N; s/.*inet addr:\([0-9]*\.[0-9]*\.[0-9]*\.[0-9]*\).*/\1/; }" | grep ^[0-9] | head -1`
             if [ "${NS_HOST_tmp}" == "" ]; then
             	NS_HOST_tmp=11.0.0.1
             fi	
             IBBS_NAME_tmp="H-1"
             
             echo 
             echo "Please provide IBADM (Infiniband Administration Package) configuration:"
             echo 
             
             echo -n "Enter IBADM Name Server and In-Band Server IP Address (one per IB subnet) [$NS_HOST_tmp]:"
             read NS_HOST
             if [ "$NS_HOST" == "" ]; then
                 NS_HOST=$NS_HOST_tmp
             else     
                 while ! is_valid $NS_HOST; do
                     echo -n "Enter Valid Network like [$NS_HOST_tmp]:"
                     read NS_HOST
                     [ "$NS_HOST" == "" ] && NS_HOST=$NS_HOST_tmp
                 done    
             fi    
        
             IBBS_HOST=$NS_HOST
             
             echo -n "Enter In-Band Server Hostname [$IBBS_NAME_tmp]:"
             read IBBS_NAME
             if [ "$IBBS_NAME" == "" ]; then
                IBBS_NAME=$IBBS_NAME_tmp
             fi   
                
             # Set FW_WORK_DIR
             echo -n "Enter firmware work directory [/tmp]:"
             read FW_WORK_DIR
             if [ "$FW_WORK_DIR" == "" ]; then
                   FW_WORK_DIR="/tmp"
             else      
                 while [[ "$FW_WORK_DIR" == "/" || "$FW_WORK_DIR" == "/root" ]]; do
                     err_echo "Invalid directory."
                     echo -n "Enter firmware work directory [/tmp]:"
                     read FW_WORK_DIR
                     [ "$FW_WORK_DIR" == "" ] && FW_WORK_DIR="/tmp"
                 done
             fi    
        
             
        # Clean previous IBADM configuration in order to prevent duplicate IP
        perl -ni -e 'if (/\#\# IBADM \#\#$/) { $filter = 1 }' -e 'if (!$filter) { print }' -e 'if (/\#\# END IBADM \#\#$/){ $filter = 0 }' $CONFIG
        
        echo "## IBADM ##" >> $CONFIG
        echo LAN_INTERFACE=$LAN_INTERFACE >> $CONFIG
        echo NS_HOST=$NS_HOST >> $CONFIG
        echo IBBS_HOST=$IBBS_HOST >> $CONFIG
        echo IBBS_NAME=$IBBS_NAME >> $CONFIG
        echo FW_WORK_DIR=$FW_WORK_DIR >> $CONFIG
        echo "## END IBADM ##" >> $CONFIG
        
        return 0
}

# Check if ETH interface $1 exists
is_eth()
{
    local eth_name=$1
    shift
    
    $IFCONFIG $eth_name > $NULL 2>&1   
    return $?
}

# Configure Driver Autostart
ipconfig()
{
        local device=$1
              shift
        local target=$1
              shift
        
        # Set default network configuration
        local IPADDR IPADDR_tmp
        local NETMASK NETMASK_tmp
        local BROADCAST BROADCAST_tmp
        local NETWORK NETWORK_tmp 
        local ONBOOT ONBOOT_tmp=1
        local LAN_INTERFACE LAN_INTERFACE_tmp
        local INTERFACES INTERFACE ADDRS
        
        
        INTERFACES=`$IFCONFIG | grep eth | awk '{print$1}' | sort -n | tr -s '\n' '\ '`
        for INTERFACE in $INTERFACES
        do
                if ( $IFCONFIG $INTERFACE | grep "inet addr" > $NULL 2>&1 ); then
                      LAN_INTERFACE_tmp=$INTERFACE
                      break
                fi
        done
        
        LAN_INTERFACE_tmp=${LAN_INTERFACE_tmp:-eth0}
        
            # Set Ethernet device for IPoIB configuration
            LAN_INTERFACE=`eval echo \$\{LAN_INTERFACE_$device\}`
        
            if [ $INTERACTIVE -eq 1 ]; then
               echo
               read -p "Enter LAN interface to be used for setting $device interface [$LAN_INTERFACE_tmp]:"
               LAN_INTERFACE=${REPLY:-eth0}
               while ! is_eth $LAN_INTERFACE; do
                    read -p "Enter existing LAN interface to be used for setting $device interface [$LAN_INTERFACE_tmp]:"
                    LAN_INTERFACE=${REPLY:-eth0}
               done 
            else
               LAN_INTERFACE=${LAN_INTERFACE:-$LAN_INTERFACE_tmp}
            fi
        
        
        
        local high_IP=`$IFCONFIG | sed -e "/\($LAN_INTERFACE\).*/{ N; s/.*inet addr:\([0-9]*\).*/\1/; }" | grep ^[0-9] | head -1`
        if [ "${high_IP}" == "" ]; then
        	high_IP=10
        fi
        let high_IP=$high_IP+$IP_cnt
        
        local low_IP=`$IFCONFIG | sed -e "/\($LAN_INTERFACE\).*/{ N; s/.*inet addr:[0-9]*\.\([0-9]*\.[0-9]*\.[0-9]*\).*/\1/; }" | grep ^[0-9] | head -1`
        if [ "${low_IP}" == "" ]; then
        	low_IP=0.0.1
        fi
        IPADDR_tmp=${high_IP}.${low_IP}
        
        NETMASK_tmp=`$IFCONFIG | sed -e "/\($LAN_INTERFACE\).*/{ N; s/.*Mask:\([0-9]*\.[0-9]*\.[0-9]*\.[0-9]*\).*/\1/; }" | grep ^[0-9] | head -1`
        if [ "${NETMASK_tmp}" == "" ]; then
        	NETMASK_tmp=255.0.0.0
        fi
        
        local low_Bcast=`$IFCONFIG | sed -e "/\($LAN_INTERFACE\).*/{ N; s/.*Bcast:[0-9]*\.\([0-9]*\.[0-9]*\.[0-9]*\).*/\1/; }" | grep ^[0-9] | head -1`
        if [ "${low_Bcast}" == "" ]; then
        	low_Bcast=0.255.255
        fi
        BROADCAST_tmp=${high_IP}.${low_Bcast}
        
        #################################################################
        
        IP_A=`echo ${IPADDR_tmp} | cut -d '.' -f 1 | tr -d '[:space:]'`
        IP_B=`echo ${IPADDR_tmp} | cut -d '.' -f 2 | tr -d '[:space:]'`
        IP_C=`echo ${IPADDR_tmp} | cut -d '.' -f 3 | tr -d '[:space:]'`
        IP_D=`echo ${IPADDR_tmp} | cut -d '.' -f 4 | tr -d '[:space:]'`
        
        NM_A=`echo ${NETMASK_tmp} | cut -d '.' -f 1 | tr -d '[:space:]'`
        NM_B=`echo ${NETMASK_tmp} | cut -d '.' -f 2 | tr -d '[:space:]'`
        NM_C=`echo ${NETMASK_tmp} | cut -d '.' -f 3 | tr -d '[:space:]'`
        NM_D=`echo ${NETMASK_tmp} | cut -d '.' -f 4 | tr -d '[:space:]'`
        
        NW_A=$[ $IP_A & $NM_A ]
        NW_B=$[ $IP_B & $NM_B ]
        NW_C=$[ $IP_C & $NM_C ]
        NW_D=$[ $IP_D & $NM_D ]
        
        NETWORK_tmp=${NW_A}.${NW_B}.${NW_C}.${NW_D}
        
        local ip=""
        local nm=""
        local nw=""
        local bc=""
         
        if [ $INTERACTIVE -eq 1 ]; then
           echo
           echo "$device configuration:"
        
                if [ -s $NETWORK_CONF_DIR/ifcfg-$device ]; then
                    echo
                    echo "  Current IPOIB configuration for $device"
                    echo
                    cat $NETWORK_CONF_DIR/ifcfg-$device
                    echo
                    echo -n "Do you want to change this configuration? [y/N]:"
                    if ! get_ans; then
                        # echo
                        # pass_echo "Saved current configuration (/etc/infiniband/ifcfg-${device}.rpmsave) for $device"
                        # echo
                        return 0
                    fi
                fi
                
        
                    echo -n "Enter IP Address [$IPADDR_tmp]:"
                    read ip
                    if [ "$ip" == "" ]; then
                        IPADDR=$IPADDR_tmp            
                    else
                        while ! is_valid $ip; do
                            echo -n "Enter valid IP Address like [$IPADDR_tmp]:"
                            read ip
                        done    
                        IPADDR=$ip
                    fi
                    
                    echo -n "Enter Netmask [$NETMASK_tmp]:"
                    read nm
                    if [ "$nm" == "" ]; then
                        NETMASK=$NETMASK_tmp
                    else                  
                        while ! is_valid $nm; do
                            echo -n "Enter Valid Netmask like [$NETMASK_tmp]:"
                            read nm
                        done    
                        NETMASK=$nm
                    fi    
                        
                    echo -n "Enter Network [$NETWORK_tmp]:"
                    read nw
                    if [ "$nw" == "" ]; then
                        NETWORK=$NETWORK_tmp
                    else     
                        while ! is_valid $nw; do
                            echo -n "Enter Valid Network like [$NETWORK_tmp]:"
                            read nw
                        done    
                        NETWORK=$nw
                    fi    
                    
                    echo -n "Enter Broadcast [$BROADCAST_tmp]:"
                    read bc
                    if [ "$bc" == "" ]; then
                        BROADCAST=$BROADCAST_tmp
                    else     
                        while ! is_valid $bc; do
                            echo -n "Enter Valid Network like [$BROADCAST_tmp]:"
                            read bc
                        done    
                        BROADCAST=$bc
                    fi    
                    
                    read -p "Start Device On Boot? [Y/n]:" ans
                    if ! [[ "$ans" == "" || "$ans" == "y" || "$ans" == "Y" || "$ans" == "yes" ]]; then
                        ONBOOT=0
                    else
                        ONBOOT=1    
                    fi
        #         fi
        
                    
                    echo
                    echo Selected configuration:
                    echo
                    echo IPADDR=$IPADDR
                    echo NETMASK=$NETMASK
                    echo NETWORK=$NETWORK
                    echo BROADCAST=$BROADCAST
                    if [ $ONBOOT -eq 1 ]; then
                         echo ONBOOT=yes
                    else
                         echo ONBOOT=no
                    fi
        
                    echo
                    echo -n "Do you want to save selected configuration? [Y/n]:"
                    read ans
                    if ! [[ "$ans" == "" || "$ans" == "y" || "$ans" == "Y" || "$ans" == "yes" ]]; then
                        return 1
                    fi    
                    
        else # Non interactive mode
        
        if [ -e ${target}.rpmsave ]; then
        	if ( /usr/bin/diff ${target}.rpmsave ${target} > $NULL 2>&1 ); then
        		pass_echo "Used the previous configuration file (${target}.rpmsave) for $device interface"
        		return 0
        	fi	
        fi
        	
            # Set device configuration from config file
            IPADDR=`eval echo \$\{IPADDR_$device\}`
            NETMASK=`eval echo \$\{NETMASK_$device\}`
            NETWORK=`eval echo \$\{NETWORK_$device\}`
            BROADCAST=`eval echo \$\{BROADCAST_$device\}`
            ONBOOT=`eval echo \$\{ONBOOT_$device\}`
                       
        
            # If "A" in calculated IP (A.B.C.D) address same as IP from config file use calculated IP, otherwise use an IP from config file
            if [[ -z ${IPADDR} || "${IPADDR%*.*.*.*}" == "${IPADDR_tmp%*.*.*.*}" ]]; then
                IPADDR=$IPADDR_tmp
            fi    
            
            NETMASK=${NETMASK:-$NETMASK_tmp}
            NETWORK=${NETWORK:-$NETWORK_tmp}
            BROADCAST=${BROADCAST:-$BROADCAST_tmp}
            ONBOOT=${ONBOOT:-$ONBOOT_tmp}
        
            echo
            echo "IPoIB configuration for $device:"
            echo
            echo IPADDR=$IPADDR
            echo NETMASK=$NETMASK
            echo NETWORK=$NETWORK
            echo BROADCAST=$BROADCAST
            if [ $ONBOOT -eq 1 ]; then
                echo ONBOOT=yes
            else
                echo ONBOOT=no
            fi
        fi    
        
        cat $NULL > $target
        
        if [ "$DISTRIBUTION" == "SuSE" ]; then
            echo BOOTPROTO=\'static\' >> $target
            echo IPADDR=\'${IPADDR}\' >> $target
            echo NETMASK=\'$NETMASK\' >> $target
            echo NETWORK=\'${NETWORK}\' >> $target
            echo BROADCAST=\'${BROADCAST}\' >> $target
            echo REMOTE_IPADDR=\'\' >> $target
            if [ $ONBOOT -eq 1 ]; then
                echo STARTMODE=\'onboot\' >> $target
            else
                echo STARTMODE=\'manual\' >> $target
            fi    
            echo WIRELESS=\'no\' >> $target
        else
            echo DEVICE=$DEVICE >> $target
            echo BOOTPROTO=static >> $target
            echo IPADDR=${IPADDR} >> $target
            echo NETMASK=$NETMASK >> $target
            echo NETWORK=${NETWORK} >> $target
            echo BROADCAST=${BROADCAST} >> $target
            if [ $ONBOOT -eq 1 ]; then
                echo ONBOOT=yes >> $target
            else
                echo ONBOOT=no >> $target    
            fi        
        fi
        
        
        if [ $SET_CONFIG_NET -eq 1 ]; then                                
            echo LAN_INTERFACE_$device=$LAN_INTERFACE >> $CONFIG_NET
            echo IPADDR_$device=$IPADDR >> $CONFIG_NET
            echo NETMASK_$device=$NETMASK >> $CONFIG_NET
            echo NETWORK_$device=$NETWORK >> $CONFIG_NET
            echo BROADCAST_$device=$BROADCAST >> $CONFIG_NET
            echo ONBOOT_$device=$ONBOOT >> $CONFIG_NET
        fi
        
        chown root $target
        chgrp root $target
        chmod ugo=r,u=rw $target

}


set_ip()
{
        count_ib_ports
        ports_num=$?

        DEVICES=""
                           
        for (( i=0 ; i < $ports_num ; i++ ))
        do
            DEVICES="$DEVICES ib${i}"
        done    

        # Set default if no IB devices found                                
        [ "$DEVICES" == "" ] && DEVICES="ib0 ib1"
        
        IP_cnt=1
        
        if [ $SET_CONFIG_NET -eq 1 ]; then                                
            # Clean IPOIB configuration file
        #    perl -ni -e 'if (/\#\# IPOIB \#\#$/) { $filter = 1 }' -e 'if (!$filter) { print }' -e 'if (/\#\# END IPOIB \#\#$/){ $filter = 0 }' $CONFIG_NET
        
            echo "## IPOIB ##" > $CONFIG_NET
        fi
        
        if [ $INTERACTIVE -eq 1 ]; then
            echo
            echo "The default IPoIB interface configuration is based on a LAN interface configuration."
            echo "You may change this default configuration in the following steps."    
        fi
        
        for DEVICE in $DEVICES
        do
           target=$NETWORK_CONF_DIR/ifcfg-${DEVICE}
           ipconfig $DEVICE $target
           let IP_cnt++
        done
        
        if [ $SET_CONFIG_NET -eq 1 ]; then                                
            echo "## END IPOIB ##" >> $CONFIG_NET
        fi    
        
        if [ $INTERACTIVE -eq 1 ]; then
            echo
            pass_echo "IPOIB interface configured successfully"
            echo
        fi
}

# HCA driver and Ulps RPMs installation
openib()
{
	local rpms_to_install=""

        for package in $OPENIB_PACKAGES
        do
                package=$(echo $package | sed -e 's/ib_/ib-/g')
                if [ ! -f ${RPMS}/${package}-${OPENIB_VER}-${RPMK_VER}.${build_arch}.rpm ]; then
			if [ "${package}" == "ib-srp_target" ]; then
				OPENIB_PACKAGES=$(echo -n $OPENIB_PACKAGES | sed -e "s/ ib_srp_target//")
                		export OPENIB_PACKAGES
                		echo "IB_SRP_TARGET is not included in this package"
				continue
			else
                        	err_echo "${RPMS}/${package}-${OPENIB_VER}-${RPMK_VER}.${build_arch}.rpm not found"
                        	exit 1
			fi
                fi                        
                
                if ! check_run_req $(echo $package | sed -e 's/ib-//g') err; then
                        exit 2
                fi    
                rpms_to_install="$rpms_to_install ${RPMS}/${package}-${OPENIB_VER}-${RPMK_VER}.${build_arch}.rpm"
        
        done

        ex "$RPM -ihv ${rpms_to_install}"

        # Set permissions
        $CHMOD -R +rw ${OPENIB_DEST}

        return 0
}


# OpenSM RPM installation
opensm()
{
        if [ -f ${RPMS}/${OPENSM_NAME}-${OPENSM_VER}-${RPMK_VER}.${build_arch}.rpm ]; then
            if ! check_run_req opensm err; then
                exit 2
            fi    
            ex "$RPM -Uhv ${RPMS}/${OPENSM_NAME}-${OPENSM_VER}-${RPMK_VER}.${build_arch}.rpm"
        else
            err_echo "${RPMS}/${OPENSM_NAME}-${OPENSM_VER}-${RPMK_VER}.${build_arch}.rpm not found"
            exit 1
        fi
        
        return 0
}

# HCA Tools installation
mst()
{
        if [ -f ${RPMS}/${MST_NAME}-${ARCH}-${MST_VER}-${RPMK_VER}.${build_arch}.rpm ]; then
            if ! check_run_req mst err; then
                return 2
            fi    
            ex "$RPM -Uhv ${RPMS}/${MST_NAME}-${ARCH}-${MST_VER}-${RPMK_VER}.${build_arch}.rpm"
        else
            err_echo "${RPMS}/${MST_NAME}-${ARCH}-${MST_VER}-${RPMK_VER}.${build_arch}.rpm not found"
            return 1
        fi
        
        return 0
}

# IBADM RPM installation (Install mst as a part of ibadm installation)
ibadm()
{

        if ! mst; then
                err_echo "Failed to build MST (Mellanox Software Tools) required for IBADM functionality"
                exit 1
        fi
        
        if [ -f ${RPMS}/${IBADM_NAME}-${IBADM_VER}-${RPMK_VER}.${build_arch}.rpm ]; then
            if ! check_run_req ibadm err; then
                exit 2
            fi    
            ex "$RPM -Uhv ${RPMS}/${IBADM_NAME}-${IBADM_VER}-${RPMK_VER}.${build_arch}.rpm"
        else
            err_echo "${RPMS}/${IBADM_NAME}-${IBADM_VER}-${RPMK_VER}.${build_arch}.rpm not found"
            return 1
        fi
        
        return 0
}

# IBGranite installation
ibgfvs()
{
	# Check if ibgfvs-x.x.x.tgz exist
	if [ ! -f ${IBGFVS_SOURCES} ]; then
		warn_echo "${IBGFVS_SOURCES} is not found. IBGranite will not be installed."
		return 1
	fi
	
	if [ ! -f ${IBADM_SOURCE} ]; then
		warn_echo "${IBADM_SOURCE} is not found (required for IBGranite installation). IBGranite will not be installed."
	fi
	
	if [ ! -f ${MST_SOURCE} ]; then
		warn_echo "${MST_SOURCE} is not found (required for IBGranite installation). IBGranite will not be installed."
	fi

	cd $BUILD_TMP
	ex_silent "$RM -f ${IBGFVS_SOURCES##*/}"
        echo
        echo "Installing ${IBGFVS_NAME} . Please wait..."
        echo

	# Open ibgfvs-x.x.x.tgz under $BUILD_TMP directory
        ex_silent "tar xzvf ${IBGFVS_SOURCES}"
        local IBGFVS_SRCDIR=$BUILD_TMP/$(echo ${IBGFVS_SOURCES##*/} | sed -e 's/.tgz//' | tr -d '[:space:]')

	# Create ibadm tgz including mst under $IBGFVS_SRCDIR directory
	cd $IBGFVS_SRCDIR
	ex_silent "tar xzvf ${IBADM_SOURCE}"
	local IBADM_SRCDIR=$IBGFVS_SRCDIR/$(echo ${IBADM_SOURCE##*/} | sed -e 's/.tgz//' | tr -d '[:space:]')
	# Copy MST tgz to $IBADM_SRCDIR
	ex_silent "$CP -a ${PKG_SOURCE_DIR}/SOURCES/${MST_NAME}*.tgz ${IBADM_SRCDIR}"
	ex_silent "tar czvf ${IBADM_NAME}-${IBADM_VER}.tgz ${IBADM_NAME}-${IBADM_VER}" 
	ex_silent "$RM -rf ${IBADM_SRCDIR}"
	
	# Run IBGranite installation
	ex_silent "${IBGFVS_SRCDIR}/install.sh --prefix ${STACK_PREFIX} --batch"
	if [ -x ${STACK_PREFIX}/bin/ibgfvs ] && [ "${STACK_PREFIX}" != "/usr" ]; then
		ln -snf ${STACK_PREFIX}/bin/ibgfvs /usr/bin
	fi
	ex_silent "$RM -rf ${IBGFVS_SRCDIR}"
        echo "IBGFVS installation done. Please run \"ibgfvs setup\"."
	echo
}
             
# NCSA MPI RPM installation (Not applied in the IBGD-1.8.0)
mpi_ncsa()
{
        local VMI_RPM=${MPI_NCSA_VMI_NAME}-${MPI_NCSA_VMI_VER}-${RPMK_VER}.${build_arch}.rpm
        local NCSA_RPM=${MPI_NCSA_MPICH_NAME}-${MPI_NCSA_MPICH_VER}-${RPMK_VER}.${build_arch}.rpm
        
        # Add libcurl.so link (W/A for Rocks 0.3.3)
        
        if [ -d /usr/lib ]; then
                if [ ! -h /usr/lib/libcurl.so ]; then
                        cd /usr/lib
                        ln -snf libcurl.so.2.0.2 libcurl.so
                        cd -
                fi
        fi
        
        if [ -d /usr/lib64 ]; then
                if [ ! -h /usr/lib64/libcurl.so ]; then
                        cd /usr/lib64
                        ln -snf libcurl.so.2.0.2 libcurl.so
                        cd -
                fi
        fi
        
        
        if [[ -f ${RPMS}/${VMI_RPM} && -f ${RPMS}/${NCSA_RPM} ]]; then
            if ! check_run_req mpi_ncsa err; then
                return 2
            fi    
            ex "$RPM -Uhv ${RPMS}/${VMI_RPM} ${RPMS}/${NCSA_RPM}"
        else
            err_echo "${RPMS}/${VMI_RPM} and ${RPMS}/${NCSA_RPM} not found"
            return 1
        fi
        
        return 0
}

# pdsh RPM installation (In case of error - continue)
pdsh()
{
    if ( $RPM -q ${PDSH_NAME}-${PDSH_VER} > /dev/null 2>&1 ); then
        echo
        echo "${PDSH_NAME}-${PDSH_VER} is already installed"
        return 0
    fi

    if [ -f ${RPMS}/${PDSH_NAME}-${PDSH_VER}*.${build_arch}.rpm ]; then
        $RPM -Uhv ${RPMS}/${PDSH_NAME}-*${PDSH_VER}.${build_arch}.rpm >> $LOG 2>&1
        if [ $? -ne 0 ]; then
            err_echo "Failed to install pdsh"
            echo "Continuing installation..."
        fi    
    fi    

    return 0
}

# OSU MPI RPM installation (MPI_COMPILER variable should contain the list of compilers)
mpi_osu()
{
        local rpm_cnt=0

        if ! check_run_req mpi_osu err; then
            return 2
        fi    

        for mpi_comp in ${MPI_COMPILER}
        do
            OSU_RPM=${MPI_OSU_MPICH_NAME}_${mpi_comp}-${MPI_OSU_MPICH_VER}-${RPMK_VER}.${build_arch}.rpm
        
            if [ -f ${RPMS}/${OSU_RPM} ]; then
                ex "$RPM -Uhv ${RPMS}/${OSU_RPM}"
                let rpm_cnt++
            else
                echo "${RPMS}/${OSU_RPM} not found"
            fi
        done        

        if [ $rpm_cnt -eq 0 ]; then
            err_echo "No MPI RPMs installed"
            return 1
        fi    

        return 0
}

#
#   Remove previously installed stack (if exist)
#   Install the new stack RPMs. Also copy Documentation and useful scripts
#
make_install()
{
        if [ -n "$OPENIB_PACKAGES" ]; then
                rmInstalled
                
                echo
                echo "Installing IBGD software"
                echo
                
                # install openib RPMs
                openib

                ex_silent "$CP -rf $DOCS_SOURCE $STACK_PREFIX"
                ex_silent "$CP $PKG_SOURCE_DIR/README.txt $STACK_PREFIX/README.txt"
                ex_silent "$CP $PKG_SOURCE_DIR/BUILD_ID $STACK_PREFIX/BUILD_ID"
                ex_silent "$CP $PKG_SOURCE_DIR/LICENSE $STACK_PREFIX/LICENSE"
                
                ex_silent "$CP $CONFIG $STACK_PREFIX/"
                ex_silent "$CP $PKG_SOURCE_DIR/uninstall.sh $STACK_PREFIX/uninstall.sh"
                
                perl -ni -e "s@(STACK_PREFIX=).*@\$1${STACK_PREFIX}@; print" $STACK_PREFIX/uninstall.sh
                
                # Create scripts directory and links from /usr/bin
                ex_silent "$CP -rf $SCRIPTS_SOURCE/scripts $STACK_PREFIX"
                
                cd /usr/bin
                
                for bin in $STACK_PREFIX/scripts/*
                do
                     ex_silent "ln -snf $bin"
                done
                
                ex_silent cd -
        fi        
        
        for package in $EXTRA_PACKAGES
        do
                $package
        done
        
        return 0
}

# Uninstall Software
uninstall()
{
    # Get list of installed packages and Select packages to uninstall
    if isInstalled ib-verbs || isInstalled openib; then
        echo -n "Do you want to remove all IBGD software? [y/N]:"
        if ! get_ans; then
            echo
            echo "Checking for installed IBGD Software..."
            echo

#                 # Check PDSH
#                 if isInstalled pdsh; then
#                      echo -n "Remove PDSH [y/N]?"
#                      if ! get_ans; then
#                          pdsh=y
#                      else
#                          pdsh=n
#                      fi
#                 fi
                
                # Check MPI OSU            
                if isInstalled mpi_osu; then
                     echo -n "Remove MPI OSU [y/N]?"
                     if ! get_ans; then
                         mpi_osu=y
                     else
                         mpi_osu=n
                     fi        
                fi

                # Check NCSA MPI
                if isInstalled mpi_ncsa; then
                     echo -n "Remove MPI NCSA [y/N]?"
                     if ! get_ans; then
                         mpi_ncsa=y
                     else
                         mpi_ncsa=n
                     fi        
                fi    

                # Check ibadm (handle mst as a part of ibadm)
                if isInstalled ibadm; then
                     echo -n "Remove ibadm [y/N]?"
                     if ! get_ans; then
                         ibadm=y
                         mst=y
                     else
                         ibadm=n
                         mst=n
                     fi        
                fi    

                # Check opensm
                if isInstalled opensm; then
                     echo -n "Remove OpenSM [y/N]?"
                     if ! get_ans; then
                         opensm=y
                     else
                         opensm=n
                     fi        
                fi    

                # Check openib / ib package
                if isInstalled openib; then
                     echo -n "Remove OpenIB [y/N]?"
                     if ! get_ans; then
                        openib=y
                     else
                        openib=n
                     fi        
                fi
                
                if isInstalled ib-srp_target; then
                     echo -n "Remove SRP Target [y/N]?"
                     if ! get_ans; then
                        ib_srp_target=y
                     else
                        ib_srp_target=n
                     fi        
                fi
                
                if isInstalled ib-srp; then
                     echo -n "Remove SRP Initiator [y/N]?"
                     if ! get_ans; then
                        ib_srp=y
                     else
                        ib_srp=n
                     fi        
                fi    

                if isInstalled ib-dapl; then
                     echo -n "Remove DAPL [y/N]?"
                     if ! get_ans; then
                        ib_dapl=y
                     else
                        ib_dapl=n
                     fi        
                fi    

                if isInstalled ib-sdp; then
                     echo -n "Remove SDP [y/N]?"
                     if ! get_ans; then
                        ib_sdp=y
                     else
                        ib_sdp=n
                     fi        
                fi    

                if isInstalled ib-cm; then
                     echo -n "Remove CM [y/N]?"
                     if ! get_ans; then
                        ib_cm=y
                     else
                        ib_sdp=n
                        ib_dapl=n
                        ib_srp=n
                        ib_srp_target=n
                        ib_cm=n
                     fi        
                fi    

                if isInstalled ib-ipoib; then
                     echo -n "Remove IPoIB [y/N]?"
                     if ! get_ans; then
                        ib_ipoib=y
                     else
                        ib_cm=n
                        ib_sdp=n
                        ib_dapl=n
                        ib_srp=n
                        ib_srp_target=n
                        ib_ipoib=n
                     fi        
                fi    

                if isInstalled ib-verbs; then
                     echo -n "Remove Verbs and Access Layer [y/N]?"
                     if ! get_ans; then
                        ib_verbs=y
                     else
                        echo
                        echo "Removing ib-verbs will cause to all IBGD Software to be removed."
                        echo -n "Are you sure? [y/N]:"
                        if get_ans; then
                            return 1
                        else    
                            mpi_osu=n
                            mpi_ncsa=n
                            ibadm=n
                            mst=n
                            opensm=n
                            openib=n
                            ib_verbs=n
                            ib_ipoib=n
                            ib_cm=n
                            ib_sdp=n
                            ib_dapl=n
                            ib_srp=n
                            ib_srp_target=n
                        fi    
                     fi        
                fi    
    
        else # All IBGD Software selected to be removed
                echo -n "Are you sure? [y/N]:"
                if get_ans; then
                    # Set all packages to be removed
                    mpi_osu=n
                    mpi_ncsa=n
                    ibadm=n
                    mst=n
                    opensm=n
                    openib=n
                    ib_verbs=n
                    ib_ipoib=n
                    ib_cm=n
                    ib_sdp=n
                    ib_dapl=n
                    ib_srp=n
                    ib_srp_target=n
                    
                else
                    return 1
                fi
        fi
    else
        warn_echo There is no IBGD packages installed
        getch
        return 1
    fi  
      
#    read -p "Do you want to continue? [Y/n]:" ans_r
#    if ! [[ "$ans_r" == "" || "$ans_r" == "y" || "$ans_r" == "Y" || "$ans_r" == "yes" ]]; then
#        return 1
#    fi
        
    # Run Uninstall
    
        local RC=0

#            if [ "$pdsh" != "y" ]; then
#                 pdsh_rpms=$(rpm -qa | grep pdsh)
#                 ex "$RPM -e ${pdsh_rpms}"
#                 let RC++
#            fi
           
           if [ "$mpi_osu" != "y" ]; then
                MPI_OSU_LIST=$(rpm -qa | grep ${MPI_OSU_MPICH_NAME})
                
                for mpi_name in $MPI_OSU_LIST
                do 
                    ex "$RPM -e ${mpi_name}"
                    let RC++
                done    
           fi 
           
           if [ "$mpi_ncsa" != "y" ]; then
                if ( $RPM -q ${MPI_NCSA_MPICH_NAME} > $NULL 2>&1 ); then
                    ex "$RPM -e ${MPI_NCSA_MPICH_NAME}"
                    let RC++
                fi    
                 
                if ( $RPM -q ${MPI_NCSA_VMI_NAME} > $NULL 2>&1 ); then
                    ex "$RPM -e ${MPI_NCSA_VMI_NAME}"
                fi   
           fi 
        
           if [ "$ibadm" != "y" ]; then
                if ( $RPM -q ${IBADM_NAME} > $NULL 2>&1 ); then
                    ex "$RPM -e ${IBADM_NAME}"
                    let RC++
                fi    
           fi 
           
           if [ "$mst" != "y" ]; then
                if ( $RPM -q hmst-${ARCH} > $NULL 2>&1 ); then
                    ex "$RPM -e hmst-${ARCH}"
                fi 
        	
                if ( $RPM -q ${MST_NAME}-${ARCH} > $NULL 2>&1 ); then
                    ex "$RPM -e ${MST_NAME}-${ARCH}"
                    let RC++
                fi    
           fi 
           
           if [ "$opensm" != "y" ]; then
                if ( $RPM -q ${OPENSM_NAME} > $NULL 2>&1 ); then
                    ex "$RPM -e ${OPENSM_NAME}"
                    let RC++
                fi    
           fi 
        
           if [ "$openib" != "y" ]; then
                if ( $RPM -q openib > $NULL 2>&1 ); then
                    ex "$RPM -e openib"
                    let RC++
                fi
                # Remove IBGD software the from system    
                $RM -rf $STACK_PREFIX
                $RM -f $NETWORK_CONF_DIR/ifcfg-ib*
                if [[ ! -d /etc/ibfw || ! -h /etc/ibfw ]]; then
                    $RM -rf /etc/ibfw    
                fi 
           fi

           if [ "$ib_srp_target" != "y" ]; then
                   if ( $RPM -q ib-srp_target > $NULL 2>&1 ); then
                       ex "$RPM -e ib-srp_target"
                       let RC++
                   fi    
           fi 

           if [ "$ib_srp" != "y" ]; then
                   if ( $RPM -q ib-srp > $NULL 2>&1 ); then
                       ex "$RPM -e ib-srp"
                       let RC++
                   fi    
           fi 

           if [ "$ib_dapl" != "y" ]; then
                   if ( $RPM -q ib-dapl > $NULL 2>&1 ); then
                       ex "$RPM -e ib-dapl"
                       let RC++
                   fi    
           fi 

           if [ "$ib_sdp" != "y" ]; then
                   if ( $RPM -q ib-sdp > $NULL 2>&1 ); then
                       ex "$RPM -e ib-sdp"
                       let RC++
                   fi    
           fi 

           if [ "$ib_cm" != "y" ]; then
                   if ( $RPM -q ib-cm > $NULL 2>&1 ); then
                       ex "$RPM -e ib-cm"
                       let RC++
                   fi    
           fi 

           if [ "$ib_ipoib" != "y" ]; then
                   if ( $RPM -q ib-ipoib > $NULL 2>&1 ); then
                       ex "$RPM -e ib-ipoib"
                       let RC++
                   fi    
           fi 

           if [ "$ib_verbs" != "y" ]; then
                   if ( $RPM -q ib-verbs > $NULL 2>&1 ); then
                       ex "$RPM -e ib-verbs"
                       let RC++
                   fi    
           fi 
        
            if [ $RC -gt 0 ]; then
            
                echo
                pass_echo "$RC packages were removed successfully"
                echo
        #         echo "  Please reboot the system ..."
        #         echo
                getch
            else 
                echo
                warn_echo "No packages selected to be removed"
                getch    
            fi    

        return 0
}

#
# Edit /etc/infiniband/openib.conf
# It is possible to redefine the default values of modules' on boot load parameters
# by defining the new value in the ibgd.conf file when running unattended installation
# or by exporting those variables in the shell before running an interactive installation
#
set_config_openib()
{
    local OPENIB_CONF="/etc/infiniband/openib.conf"

    if [ ! -e ${OPENIB_CONF} ]; then
        return 1
    fi

    # Define OPENIB_ONBOOT="yes" or "no" (default "yes")
    if [ ! -z "${OPENIB_ONBOOT}" ]; then
        case "${OPENIB_ONBOOT}" in
            yes|no)
                perl -ni -e "s@^(ONBOOT=).*@\${1}${OPENIB_ONBOOT}@; print" ${OPENIB_CONF}
            ;;
        esac
    fi

    # Define IPOIB_LOAD="yes" or "no" (default "yes")
    if [ ! -z "${IPOIB_LOAD}" ]; then
        case "${IPOIB_LOAD}" in
            yes|no)
                perl -ni -e "s@^(IPOIB_LOAD=).*@\${1}${IPOIB_LOAD}@; print" ${OPENIB_CONF}
            ;;
        esac
    fi

    # Define USER_ACCESS_CM_LOAD="yes" or "no" (default "no")
    if [ ! -z "${USER_ACCESS_CM_LOAD}" ]; then
        case "${USER_ACCESS_CM_LOAD}" in
            yes|no)
                perl -ni -e "s@^(USER_ACCESS_CM_LOAD=).*@\${1}${USER_ACCESS_CM_LOAD}@; print" ${OPENIB_CONF}
            ;;
        esac
    fi

    # Define UDAPL_LOAD="yes" or "no" (default "no")
    if [ ! -z "${UDAPL_LOAD}" ]; then
        case "${UDAPL_LOAD}" in
            yes|no)
                perl -ni -e "s@^(UDAPL_LOAD=).*@\${1}${UDAPL_LOAD}@; print" ${OPENIB_CONF}
            ;;
        esac
    fi

    # Define KDAPL_LOAD="yes" or "no" (default "no")
    if [ ! -z "${KDAPL_LOAD}" ]; then
        case "${KDAPL_LOAD}" in
            yes|no)
                perl -ni -e "s@^(KDAPL_LOAD=).*@\${1}${KDAPL_LOAD}@; print" ${OPENIB_CONF}
            ;;
        esac
    fi

    # Define SDP_LOAD="yes" or "no" (default "no")
    if [ ! -z "${SDP_LOAD}" ]; then
        case "${SDP_LOAD}" in
            yes|no)
                perl -ni -e "s@^(SDP_LOAD=).*@\${1}${SDP_LOAD}@; print" ${OPENIB_CONF}
            ;;
        esac
    fi

    # Define SRP_LOAD="yes" or "no" (default "no")
    if [ ! -z "${SRP_LOAD}" ]; then
        case "${SRP_LOAD}" in
            yes|no)
                perl -ni -e "s@^(SRP_LOAD=).*@\${1}${SRP_LOAD}@; print" ${OPENIB_CONF}
            ;;
        esac
    fi

    # Define SRP_TARGET_LOAD="yes" or "no" (default "no")
    if [ ! -z "${SRP_TARGET_LOAD}" ]; then
        case "${SRP_TARGET_LOAD}" in
            yes|no)
                perl -ni -e "s@^(SRP_TARGET_LOAD=).*@\${1}${SRP_TARGET_LOAD}@; print" ${OPENIB_CONF}
            ;;
        esac
    fi

}

# Edit /etc/opensm.conf
set_config_opensm()
{
        # Edit /etc/opensm.conf
    if [ ! -z "${OSM_HOSTS_LIST}" ]; then
        for OPENSM_HOST in ${OSM_HOSTS_LIST}
        do
                if [[ -n "`$IFCONFIG | grep -w $OPENSM_HOST`"  || -n "`grep -rw $OPENSM_HOST ${NETWORK_CONF_DIR}`" ]]; then
                    perl -ni -e "s@^(ONBOOT=).*@\${1}yes@; print" /etc/opensm.conf
                    
                    # Enlarge QP1 size on OpenSM server
                    if [ -e /etc/modules.conf ]; then
                            perl -ni -e "s@^(options ib_mad ib_outs_mad_recv_entries=).*@\${1}15000@; print" /etc/modules.conf
                    fi
                    
                    if [ -e /etc/modprobe.conf ]; then
                            perl -ni -e "s@^(options ib_mad ib_outs_mad_recv_entries=).*@\${1}15000@; print" /etc/modprobe.conf
                    fi
                    
                    break
                else
                    perl -ni -e "s@^(ONBOOT=).*@\${1}no@; print" /etc/opensm.conf
                fi
        done
        
        perl -ni -e "s@^(OSM_HOSTS=).*@\${1}\"${OSM_HOSTS_LIST}\"@; print" /etc/opensm.conf
    fi
    
    return 0
}

# Edit /etc/ibadm.conf
set_config_ibadm()
{
        # Edit /etc/ibadm.conf
        if [[ ! -z "$IBBS_HOST" && ! -z "$NS_HOST" && ! -z "$IBBS_NAME" ]]; then
            	eval "sed -e 's/IBBS_HOST=.*/IBBS_HOST=${IBBS_HOST}/g' -e 's/IBBS_NAME=.*/IBBS_NAME=${IBBS_NAME}/g' -e 's/NS_HOST=.*/NS_HOST=${NS_HOST}/g' -e 's/OBBS_HOST=.*/OBBS_HOST=127.0.0.1/g' /etc/ibadm.conf" > /tmp/ibadm.conf.tmp
        	ex_silent "$MV -f /tmp/ibadm.conf.tmp /etc/ibadm.conf"
            
            # Open ibfw tgz in the IBBS Host
            if [[ -n "`$IFCONFIG | grep -w $IBBS_HOST`" || -n "`grep -rw $IBBS_HOST ${NETWORK_CONF_DIR}`" ]]; then
                [ ! -d $FW_DEST/FW ] && ex_silent "$MKDIR -p $FW_DEST/FW"
                echo
                echo "Creating FW directory to be used by IBADM server"
                echo
                
                cd $FW_DEST/FW
                ex "tar xzvf $IBFW"
                pass_echo "$FW_DEST/FW directory updated with new FW release"
                ex_silent cd -
                if [[ ! -d /etc/ibfw || ! -h /etc/ibfw ]]; then
                    ex_silent "ln -snf $FW_DEST/FW/${IBFW_NAME%*.*}  /etc/ibfw"
                fi    
		if [ $INTERACTIVE -eq 1 ]; then
			echo
			echo -n "Do you want to install IBGranite Cluster Verification Suite (ibgfvs-1.0.0 - beta version) [y/N]?"
			if get_ans; then
				ibgfvs
				echo "ibgfvs=y" >> $CONFIG
                        fi
		else 	# Unattended installation
			# Install IBGranite if ibgfvs=y in ibgd.conf file
			if [ "$ibgfvs" == "y" ]; then
				ibgfvs
			fi
		fi
            fi
        fi
        
        # Check FW_WORK_DIR
        if [ ! -z "$FW_WORK_DIR" ]; then
            if [ ! -d  "$FW_WORK_DIR" ]; then
                mkdir -p $FW_WORK_DIR
                if [ $? -ne 0 ]; then
                    err_echo "Failed to create $FW_WORK_DIR. Please create it manually"
                fi    
            fi
            perl -ni -e "s@(FW_WORK_DIR=).*@\$1${FW_WORK_DIR}@; print" /etc/ibadm.conf
        fi
        
        
}

# Set configuration parameters and make basic checks
pre_install()
{
        check_user
        
        # Set network environment configuration
        NETWORK_CONF_DIR="/etc/infiniband"
        
        if [ -f /etc/SuSE-release ]; then
            DISTRIBUTION="SuSE"
            
            # Check termcap library link
            if [[ ! -e /usr/lib/libtermcap.so && -e /usr/lib/libtermcap.so.2.0.8 ]]; then
                ex_silent "ln -snf /usr/lib/libtermcap.so.2.0.8 /usr/lib/libtermcap.so"
            fi
            
        elif [ -f /etc/redhat-release ]; then
            DISTRIBUTION="redhat"
        
        elif [ -f /etc/rocks-release ]; then
            DISTRIBUTION="Rocks"
        else
            DISTRIBUTION=`ls /etc/*-release | head -n 1 | xargs -iXXX basename XXX -release 2> /dev/null`
        fi
        
        # Check if Infiniband driver was loaded
        if ( $LSMOD | grep -w mosal > $NULL ); then
            let NEED_REBOOT++
        fi
        
        		     
        # Check STACK_PREFIX for existing RPMs
        if [ -e ${RPMS}/ib-verbs-${OPENIB_VER}-${RPMK_VER}.${build_arch}.rpm ]; then
        	STACK_PREFIX=`$RPM -qlp ${RPMS}/ib-verbs-${OPENIB_VER}-${RPMK_VER}.${build_arch}.rpm | grep "utils/create_devs" | sed -e 's/\/utils\/create_devs//'`
        	. $CWD/build_env.sh    
        fi
        
        # Check for Infiniband software of other vendors
        local vendors="topspin|iview_fm|ibhost-hpc"
        local vrpms=""
        
        vrpms=`$RPM -qa | grep -E "${vendors}"`
        if [ -n "$vrpms" ]; then
            echo
            err_echo "You have to remove the following Infiniband software before continuing:"
            echo
            echo "$vrpms"
            echo
            exit 9
        fi

}

# Run interactive installation mode
interactive()
{
           show_menu "main_menu"
        	
                read ans_inst
                case $ans_inst in
                        2)   
                             if ! select_packages; then
                                 interactive
                             fi    
                             set_package_deps
                             check_run_dependencies        
                          
                             if isInstalled "thca-linux" || isInstalled "openib" || isInstalled "ib-verbs"; then
                                     echo
                                     warn_echo "This installation program will remove any previously installed IB packages on your machine."
                                     echo
                                     read -p "Do you want to continue? [Y/n]:" ans_r
                                     if ! [[ "$ans_r" == "" || "$ans_r" == "y" || "$ans_r" == "Y" || "$ans_r" == "yes" ]]; then
                                         return 1
                                     fi
                             fi
                             
        		     
                             # Check if all required RPM packages exists
                             # If not, run build.sh script with created configuration file
                             if ! is_rpms; then
                                
                                     echo
                                     pass_echo "Preparing to build the IBGD RPMs:"
                                     echo
        
        
                                     echo
                                     echo "RPM build process uses a temporary directory."
                                     echo
                                     read -p "Please enter the temporary directory [${BUILD_ROOT}]:"
                                     BUILD_ROOT=${REPLY:-$BUILD_ROOT}
             
                                     if [ "${BUILD_ROOT}" != "${BUILD_ROOT_DFT}" ]; then
                                            BUILD_ROOT=${BUILD_ROOT}/IBGD
                                     fi
                                     
                                     [ -d $BUILD_ROOT ] && ex_silent "$RM -rf $BUILD_ROOT"
                                     
                                     check_space_req "$BUILD_SPACE_REQ" "$BUILD_ROOT"
                                     
                                     echo
                                     read -p "Please enter IBGD installation directory [${STACK_PREFIX}]:"
                                     STACK_PREFIX=${REPLY:-$STACK_PREFIX}
             
                                     # Set  parameters with new values
                                     . $CWD/build_env.sh
                                     
                                     check_space_req "$RUN_SPACE_REQ" "$STACK_PREFIX"
             
                                     # Update CONFIG file
                                     perl -ni -e "s@(BUILD_ROOT=).*@\$1${BUILD_ROOT}@; print" $CONFIG
                                     perl -ni -e "s@(STACK_PREFIX=).*@\$1${STACK_PREFIX}@; print" $CONFIG
                                     
                                     ${CWD}/build.sh -c $CONFIG
                                     if [ $? -ne 0 ]; then
                                         exit 1
                                     fi
                                     
                                     if [ "${BUILD_ROOT}" != "${BUILD_ROOT_DFT}" ]; then
                                         read -p  "Do you want to remove temporary build directory $BUILD_ROOT [Y/n]?" ans_r
                                         if [[ "$ans_r" == "" || "$ans_r" == "y" || "$ans_r" == "Y" || "$ans_r" == "yes" ]]; then
                                             $RM -rf $BUILD_ROOT
                                         fi
                                     fi    
                             else
                                     # Check STACK_PREFIX for existing RPMs
                		     if [ -e ${RPMS}/ib-verbs-${OPENIB_VER}-${RPMK_VER}.${build_arch}.rpm ]; then
                			        STACK_PREFIX=`$RPM -qlp ${RPMS}/ib-verbs-${OPENIB_VER}-${RPMK_VER}.${build_arch}.rpm | grep "utils/create_devs" | sed -e 's/\/utils\/create_devs//'`
                			        . $CWD/build_env.sh    
                		     fi
                                     
                                     echo   
                                     echo
                                     echo "The default installation directory for IBGD Software is ${STACK_PREFIX}"
                                     read -p "Do you want to continue? [Y/n]:" ans_r
                                     if ! [[ "$ans_r" == "" || "$ans_r" == "y" || "$ans_r" == "Y" || "$ans_r" == "yes" ]]; then
                                         return 1
                                     fi
                                
                             fi
                             
                             if make_install; then

                                    # Set ipoib configuration
                                    if isInstalled ib-ipoib && ( echo "$OPENIB_PACKAGES" | grep ipoib > $NULL 2>&1 ); then
                                        echo
                                        pass_echo "Configuring IPoIB:"
                                        echo
                                    
#                                         read -p "Do you want to configure IPoIB? [Y/n]:" ans_conf
#                                         if [[ "$ans_conf" == "" || "$ans_conf" == "y" || "$ans_conf" == "Y" || "$ans_conf" == "yes" ]]; then
                                         set_ip
#                                         fi
                                    fi                                       
                                    
                                    # Set OpenSM configuration
                                    if isInstalled opensm && ( echo "$EXTRA_PACKAGES" | grep opensm > $NULL 2>&1 ); then
                                        echo
                                        pass_echo "Configuring OpenSM:"
                                        echo
                                    
#                                         read -p "Do you want to configure OpenSM now? [Y/n]:" ans_conf
#                                         if [[ "$ans_conf" == "" || "$ans_conf" == "y" || "$ans_conf" == "Y" || "$ans_conf" == "yes" ]]; then
                                        osmconf
                                        set_config_opensm
#                                         fi
                                    fi    

                                    # Set ibadm configuration       
                                    if isInstalled ibadm && ( echo "$EXTRA_PACKAGES" | grep ibadm > $NULL 2>&1 ); then
                                        echo
                                        pass_echo "Configuring IBADM:"
                                        echo
#                                         read -p "Do you want to configure IBADM now? [Y/n]:" ans_conf
#                                         if [[ "$ans_conf" == "" || "$ans_conf" == "y" || "$ans_conf" == "Y" || "$ans_conf" == "yes" ]]; then
                                        ibadmconf
                                        set_config_ibadm
#                                         fi    
                                    fi
                                    
                                    pass_echo "Installation finished successfully..."
                                    getch
                                    
           #                          if [ $NEED_REBOOT -ne 0 ]; then
           #                              echo
           #                              echo    "Please reboot the system after installation"
           #                              echo
           #                              read -p "Do you want to reboot now? [Y/n]:" ans_r
           #                              if [[ "$ans_r" == "" || "$ans_r" == "y" || "$ans_r" == "Y" || "$ans_r" == "yes" ]]; then
           #                                 $REBOOT
           #                                 exit 0
           #                              fi
           #                          fi
                             fi
                        ;;
                        3)
                            echo
                            showInstalled
                            echo
                            getch
        
                        ;;
                        4)
                            if ! isInstalled ib-verbs; then
                                echo
                                err_echo "You need first to install the InfiniBand Software"
                                echo
                                getch
                                break
                            fi
                            
                            if isInstalled ib-ipoib; then
                                set_ip
                            fi

                            if isInstalled opensm; then
                                osmconf
                                set_config_opensm
                            fi
                                
                            if isInstalled ibadm; then
                                ibadmconf
                                set_config_ibadm
                            fi    

                            getch
                        ;;
                        5)
                            echo
                            uninstall
                            let NEED_REBOOT++
                        ;;
                        1)
                            view_readme
                        ;;
                        6)
                            ${CWD}/build.sh
                        ;;    
                        Q|q)
                            close_log    
                            exit 0
                        ;;
                        *)
                            return 1
                        ;;
                esac

}

# Run unattended installation mode
unattended()
{
        check_space_req "$RUN_SPACE_REQ" "$STACK_PREFIX"
        select_packages
        set_package_deps
        check_run_dependencies        
        
        # Check if all required RPM packages exists
        # If not, run build.sh script with created configuration file
        if ! is_rpms; then
           ${CWD}/build.sh -c $CONFIG
           if [ $? -ne 0 ]; then
               exit 1
           fi
           # Reread config file (need if the config file was changed by build.sh )
           . $CONFIG
        fi
        make_install
        if isInstalled ib-ipoib && ( echo "$OPENIB_PACKAGES" | grep ipoib > $NULL 2>&1 ); then
            set_ip
        fi

        if isInstalled opensm && ( echo "$EXTRA_PACKAGES" | grep opensm > $NULL 2>&1 ); then
            set_config_opensm
        fi
            
        if isInstalled ibadm && ( echo "$EXTRA_PACKAGES" | grep ibadm > $NULL 2>&1 ); then
            set_config_ibadm
        fi    

#     if [ $NEED_REBOOT -ne 0 ]; then
#         echo
#         echo    "Please reboot the system"
#         echo
#     fi

}

###
### Main
###

main()
{
        
        pre_install
        
        if [ $INTERACTIVE -eq 1 ]; then
            interactive
        else
            unattended
        fi
        
        set_config_openib
        
        return 0
}

open_log
while true
do
    main
    if [ $INTERACTIVE -eq 0 ]; then
        exit 0
    fi    
done    
