eval '(exit $?0)' && eval 'exec perl -w -S $0 ${1+"$@"}'
    & eval 'exec perl -w -S $0 $argv:q'
    if 0;

use Net::Ping; 

sub usage
{
   print "\n Usage: $0 -ip_list <file_of_ip_list> [-dead_only] [-t timeout] [-c retry_counter] [-s] [-fork] [-v]\n";
   print "\n";
   print "  -ip_list       - list of IP addresses to ping in a file\n";
   print "  -dead_only     - shows only the hosts that do not respond\n";
   print "  -t             - timeout for icmp ping. Must be > 0. \n";
   print "  -c             - maximum number of icmp ping retry operations. 1 to 65534.\n";
   print "  -s             - show statistics: no. of host alive, no. of hosts dead\n";
   print "  -fork          - use fork to run icmp ping\n";
   print "  -v             - verbose output\n";
   print "\n";
}

my $ip_list = "";
my $timeout = 1;
my $ping_counter = 3;
my $no_alive = 0;
my $statistic = 0;
my $alive_counter = 0;
my $dead_counter = 0;
my $use_fork = 0;
my $verbose = 0;

while ( $#ARGV >= 0 ) {

   my $cmd_flag = shift(@ARGV);

   if ($cmd_flag eq "-ip_list") {
      $ip_list = shift(@ARGV);
   } elsif ($cmd_flag eq "-dead_only") {
      $dead_only = 1;
   } elsif ($cmd_flag eq "-t") {
      $timeout = shift(@ARGV);
   } elsif ($cmd_flag eq "-c") {
      $ping_counter = shift(@ARGV);
   } elsif ($cmd_flag eq "-s") {
      $statistic = 1;
   } elsif ($cmd_flag eq "-fork") {
      $use_fork = 1;
   } elsif ($cmd_flag eq "-v") {
      $verbose = 1;
   } else {
      &usage();
      exit 1;
   }
}

if ( $ping_counter < 1 or $ping_counter >= 65534 ) {
	&usage();
	exit 1;
}

if ( $timeout <= 0 ) {
	&usage();
	exit 1;
}

if ( $ip_list eq "" ) {
   &usage();
   exit 1;
}

my @hosts; 

open(INFILE, $ip_list) || die "Cannot open $ip_list file: $!";
while(<INFILE>)
{
   if ( not /^#.*/ and not /^$/ ) {
      push @hosts, split(' ',$_);
   }
}
close(INFILE) || die "Failed to close $ip_list file: $!";

my %inet; 

if ( $use_fork ) {

   for my $ip (@hosts) {
         $inet{$ip}{count} = 0;
   
         my $mypid = fork();
         die "Failed to fork: $!" if not defined $mypid;
   
         $inet{$ip}{pid} = $mypid;
   
         if ( not $mypid ) {

   	      my $icmp = Net::Ping->new ("icmp", $timeout); 
   
   	      print "Child $$\n" if ( $verbose );
   	      my $res = 0;
   
   	      while ( ( $inet{$ip}{count}++ < $ping_counter ) and not ( $res = $icmp->ping ($ip) ) ) {
   		      print "Child $$, count=$inet{$ip}{count}++\n" if ( $verbose );
                      print "Running /sbin/arp -d $ip\n" if ( $verbose );
   		      system "/sbin/arp -d $ip > /dev/null 2>&1";
   		      print "Child $$, after system call\n" if ( $verbose );
   	      }
   
   	      print "Child $$: exit status: $res\n" if ( $verbose );
   	      $icmp->close;
   	      exit($inet{$ip}{count});
         }
         print "Child launched: ip=$ip child=$mypid\n" if ( $verbose );
   }
   
   for my $ip (@hosts) { 
      my $res = 0;
      my $sig = 0;
   
      print "waiting for: ip=$ip child=$inet{$ip}{pid}\n" if ( $verbose );
      waitpid ( $inet{$ip}{pid}, 0 );
      $res = (($? >> 8) - 1);
      $sig = $? & 255;
      $inet{$ip}{count} = $res;
      
      if ( not $sig and not $res == -1 and $res < $ping_counter ) {
         $inet{$ip}{icmp} = 1;
         $alive_counter ++;
      } elsif ( $sig or $res == -1 ) {
         $inet{$ip}{icmp} = -1;
         $dead_counter ++;
      } else {
         $inet{$ip}{icmp} = 0;
         $dead_counter ++;
      }
   }
   
   
   if ( ( $alive_counter and not $dead_only ) or $dead_counter ) {
      print   "   IP address     ICMP  Retry\n",
   	   "---------------- -----  -----\n";
   }
   
   for my $ip (@hosts) { 
      my %p = %{$inet{$ip}};
      my $dead = 0 ;
      my $str = "";

      if ( $p{icmp} == 1 ) {
         $str = "alive";
         $dead = 0;
      }  elsif ( $p{icmp} == 0 ) {
         $str = "dead";
         $dead = 1;
      }  else {
         $str = "killed";
         $dead = 1;
      }

      if ( $dead or not $dead_only ) {
         printf "%-15s %6s %6d\n", $ip, $str , $p{count}; 
      }
   }
   
} else {
   # Running icmp ping serially
   my $icmp = Net::Ping->new ("icmp", $timeout); 
   
   for my $ip (@hosts) { 
   
      my $cnt = 0;
      my $res = 0;
   
      $res = $icmp->ping ($ip);
   
      while ( not ( $res = $icmp->ping ($ip) ) and ( $cnt < $ping_counter ) ) {
         $cnt ++;
         print "Running /sbin/arp -d $ip\n" if ( $verbose );
         system "/sbin/arp -d $ip > /dev/null 2>&1";
      }
   
      if ( $res ) {
         $inet{$ip}{icmp} = 1;
         $alive_counter ++;
      } else {
         $inet{$ip}{icmp} = 0;
         $dead_counter ++;
      }
      $inet{$ip}{count} = $cnt;
   }
   $icmp->close;
   
   if ( ( $alive_counter and not $dead_only ) or $dead_counter ) {
      print   "   IP address     ICMP  Retry\n",
   	   "---------------- -----  -----\n";
   }
   
   for my $ip (@hosts) { 
   
      my %p = %{$inet{$ip}};
      my $dead = not $p{icmp};
   
      if ( $dead or not $dead_only ) {
         printf "%-15s %6s %6d\n", $ip, $dead ? "dead" : "alive", $p{count}; 
      }
   
   }
   
}

if ( ( $alive_counter and not $dead_only ) or $dead_counter ) {
   print   "---------------- -----  -----\n";
}

if ( $statistic ) {
   print "$alive_counter alive\n";
   print "$dead_counter dead\n";
}
   

