#!/bin/bash

# This software is available to you under a choice of one of two
# licenses.  You may choose to be licensed under the terms of the GNU
# General Public License (GPL) Version 2, available at
# <http://www.fsf.org/copyleft/gpl.html>, or the OpenIB.org BSD
# license, available in the LICENSE.TXT file accompanying this
# software.  These details are also available at
# <http://openib.org/license.html>.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
# BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
# ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# Copyright (c) 2004 Mellanox Technologies Ltd.  All rights reserved.
#
# Description: RPM build script
#

build_root="/var/tmp/IBGD"
STACK_PREFIX=/usr/local/ibgd
PACKAGES=""
ALL_PACKAGES="ib_verbs ib_ipoib ib_cm ib_sdp ib_dapl ib_srp"
MEMTRACK=0

verbs=n
ipoib=n
cm=n
sdp=n
dapl=n
srp=n
srp_target=n

usage()
{
cat << EOF

    `basename $0` [ --prefix <install-dir> ] [ --build_root <BUILD_ROOT> ] [-kver <kernel version>] [-k|--ksrc DIR]
                  [--packages package1 package2 ... -- | --all] [--memtrack]
	
	The default software stack prefix is $STACK_PREFIX
        The default BUILD_ROOT is $build_root
        Available Packages: ib_verbs ib_ipoib ib_cm ib_sdp ib_dapl ib_srp

EOF
}

while [ $# -gt 0 ]
do
    case $1 in
            -p|--prefix)
                shift
                STACK_PREFIX=$1
            ;;
            --build_root)
                shift
                build_root=$1
            ;; 
    	    --packages)
    		shift
    		while [ "$1" != "--" ]
    		do
    		    PACKAGES="$PACKAGES $1"
    		    shift
    		done    
    		if [ -z "$PACKAGES" ]; then
    		    usage
    		    exit 1
    		fi
    	    ;;
    	    --all)
   		PACKAGES=${ALL_PACKAGES}
    	    ;;
	    -kver)
		shift
		KVERSION=$1
	    ;;
	    -k|--ksrc)
		shift
		KSRC=$1
	    ;;
	    --memtrack)
		MEMTRACK=1
	    ;;	
            *)
                usage
                exit 1
            ;;
    esac            
    shift
done

# Set default value for PACKAGES  
if [ -z "$PACKAGES" ]; then
    PACKAGES=$ALL_PACKAGES
fi

#PACKAGES=($PACKAGES)
KVERSION=${KVERSION:-$(uname -r)}
KSRC=${KSRC:-/lib/modules/${KVERSION}/build}
KERNEL_TREE_ROOT=${KSRC%*/*}

if [ -d ${KERNEL_TREE_ROOT}/source/ ]; then
   KSRC=${KERNEL_TREE_ROOT}/source
fi

BUILD_ENV="env build_root=${build_root} STACK_PREFIX=${STACK_PREFIX} KVERSION=${KVERSION} KSRC=${KSRC} MEMTRACK=${MEMTRACK}"

cd ${0%*/*}

TMPFILE=/tmp/$$.tmp
SOURCE_DIR=`pwd`
KERNEL=`uname -r | sed -e "s/\-/\_/g"`
OpenIB_DIR=${PWD##*/}
#NAME=`echo ${PWD##*/} | cut -d \- -f 1`
NAME="ib"
VERSION=`echo ${PWD##*/} | cut -d \- -f 2- | tr -s '-' '_'`
RELVER=`echo ${PWD##*/} | cut -d \- -f 2-`
BUILD_REQUIRES="glib glibc glibc-devel termcap readline readline-devel"

# Check kernel sources existence
if ! test -d ${KSRC}; then
	echo In order to build RPM kernel sources need to be installed
	exit 1
fi

# Check dependencies
for pack in $BUILD_REQUIRES
do            
    if ! ( rpm -q $pack > /dev/null 2>&1 ); then
	echo In order to build RPM $pack need to be installed
	exit 1
    fi
done


build_arch=`rpm --showrc | grep -w "build arch" | cut -d : -f 2 | tr -d " "`
spec="openib.spec"

if [ -d /usr/src/redhat ]; then
	RPM_DIR=/usr/src/redhat
elif [ -d /usr/src/packages ]; then
	RPM_DIR=/usr/src/packages
else
	echo Build rpm on this system is not supported...
	exit 1
fi

if [ -f $RPM_DIR/SPECS/$spec ]; then
	rm -f $RPM_DIR/SPECS/$spec 
fi

# Update Package name, version and release in the spec file
sed -e "s/XXXXXX/$VERSION/g" -e "s/YYYYYY/$NAME/g" -e "s/ZZZZZZ/$RELVER/g" $spec > $RPM_DIR/SPECS/$spec

cd ..
rm -f /tmp/${NAME}-${RELVER}.tgz
tar czvf /tmp/${NAME}-${RELVER}.tgz $OpenIB_DIR > /dev/null 2>&1
mv -f /tmp/${NAME}-${RELVER}.tgz $RPM_DIR/SOURCES/

if [ -f /etc/SuSE-release ]; then
   if [ `grep VERSION /etc/SuSE-release | cut -d = -f 2 | cut -d . -f 1 | tr -d '[:space:]'` -ge 9 ]; then
	$BUILD_ENV  PACKAGES="$PACKAGES" rpmbuild -bb $RPM_DIR/SPECS/$spec # > /dev/null 2>&1
   else	
	$BUILD_ENV  PACKAGES="$PACKAGES" rpm -bb $RPM_DIR/SPECS/$spec # > /dev/null 2>&1
   fi
elif [[ -f /etc/redhat-release || -f /etc/rocks-release ]]; then
	$BUILD_ENV PACKAGES="$PACKAGES" rpmbuild -bb $RPM_DIR/SPECS/$spec # > /dev/null 2>&1
else
        echo Build rpm on this system is not supported...
        exit 1
fi

if [ $? -ne 0 ]; then
    echo RPM building failed...
    exit 1
fi 

rm -f $RPM_DIR/SOURCES/$OpenIB_DIR.tgz
rm -f $RPM_DIR/SPECS/$spec

rpms_required=`echo "$PACKAGES" | wc -w | tr -d '[:space:]'`
rpms_done=`find $RPM_DIR/RPMS/$build_arch -name ${NAME}-*-${VERSION}-${KERNEL}.${build_arch}.rpm | wc -l | tr -d '[:space:]'`

if [ $rpms_done -lt $rpms_required ] ; then
	echo "$rpms_done RPMs from $rpms_required created"
	exit 1
else
	echo "$rpms_build created"
fi

exit 0

