/*
  This software is available to you under a choice of one of two
  licenses.  You may choose to be licensed under the terms of the GNU
  General Public License (GPL) Version 2, available at
  <http://www.fsf.org/copyleft/gpl.html>, or the OpenIB.org BSD
  license, available in the LICENSE.TXT file accompanying this
  software.  These details are also available at
  <http://openib.org/license.html>.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
  BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
  ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
  CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
  SOFTWARE.

  Copyright (c) 2004 Topspin Communications.  All rights reserved.
  Copyright (c) 2005 Mellanox Technologies.   All rights reserved.


  $Id: useraccess_cm.h 340 2004-07-01 03:14:20Z roland $
*/

#ifndef _USERACCESS_CM_H
#define _USERACCESS_CM_H

#include <linux/sched.h>
#include <linux/list.h>
#include <linux/fs.h>
#include <linux/poll.h>
#if !defined(TS_KERNEL_2_6)
#  include <linux/devfs_fs_kernel.h>
#endif

#include "ts_ib_core.h"
#include "ts_ib_cm.h"
#include "ts_ib_useraccess_cm.h"

struct ib_useraccess_cm_device {
	struct ib_device *ib_device;
	struct ib_pd     *pd;
#if !defined(TS_KERNEL_2_6)
	devfs_handle_t devfs_handle;
#endif
};

/* this is the private context allocated for each process
   that opens the CM device file */
struct ib_useraccess_cm_private {
	/* pointer to the entry in the static array that identifies
	  the device being worked on */
	struct ib_useraccess_cm_device *device;

	struct list_head cm_comp_list;
	wait_queue_head_t cm_comp_wait;
	spinlock_t cm_comp_lock;

	struct list_head cm_conn_list;
	struct list_head free_conn_list;
	struct semaphore cm_conn_sem;

	/* unique value identifying the connection - IB spec 12.7.1 */
	tTS_IB_CM_COMM_ID    last_comm_id;
    struct task_struct * callbck_holder;
	wait_queue_head_t cleanup_wait;
};

/*
 * We use this to keep track of connections an application owns and destroy
 * then when the fd is closed
 */
struct ib_cm_user_connection {
	struct list_head list;
	struct list_head free_list;

	atomic_t refcnt;

	struct ib_useraccess_cm_private *priv;
	tTS_IB_CM_COMM_ID comm_id;
	tTS_IB_LISTEN_HANDLE listen_handle;
	unsigned long cm_arg;

	VAPI_hca_hndl_t v_device;

	struct ib_qp    *qp;
	VAPI_k_qp_hndl_t vk_qp;
};

int tsIbCmUserIoctl(struct inode *inode,
		    struct file *filp, unsigned int cmd, unsigned long arg);

/* User Mode CM Filter Routines */
int tsIbCmUserGetServiceId(unsigned long arg);

int tsIbCmUserConnect(struct ib_useraccess_cm_private *priv, unsigned long arg);

int tsIbCmUserListen(struct ib_useraccess_cm_private *priv, unsigned long arg);

int tsIbCmUserAccept(struct ib_useraccess_cm_private *priv, unsigned long arg);

int tsIbCmUserDropConsumer(struct ib_useraccess_cm_private *priv,
			   unsigned long arg);

int tsIbCmUserGetCompletion(struct ib_useraccess_cm_private *priv,
			    unsigned long arg);

int tsIbCmUserClose(struct ib_useraccess_cm_private *priv);

#endif				/* _USERACCESS_CM_H */
