#ifndef CM_TEST_H
#define CM_TEST_H

#include <unistd.h>
#include <stdio.h>
#include "ts_ib_useraccess_cm.h"	/* ioctl API to kernel */
#include "ts_ib_mad_types.h"
#include <asm/atomic.h>

///////////////////////////////////////////////////
// MACRO DECLERATIONS
///////////////////////////////////////////////////

#define HCA_NAME "InfiniHost0"
#define HCA_PORT 1
#define SID 0x10

#define MAX_PRIV_DATA_SIZE  128

#define NUM_QP 100

#define INFO
#define DEBUG
#ifdef INFO
#define info(args...)          \
        do {                    \
            printf(args);     \
        } while (0)
#else
#define info(args...)     // nothing
#endif

#ifdef DEBUG
#define debug(args...)          \
        do {                    \
            printf ("Debug:In %s:%d %s() =>",__FILE__, __LINE__, __FUNCTION__); \
            printf(args);     \
            printf("\n");       \
        } while (0)
#else
#define debug(args...)     // nothing
#endif


#define ERROR
#ifdef ERROR
#define error(args...)          \
        do {                    \
            printf ("\nError in file %s func %s in line %d =>",__FILE__, __FUNCTION__, __LINE__); \
            printf(args);     \
            printf("\n");       \
        } while (0)
#else
#define error(args...)     // nothing
#endif

#define err_exit(args...)          \
        do {                    \
            printf ("\nError and exit in file %s func %s in line %d =>",__FILE__, __FUNCTION__, __LINE__); \
            printf(args);     \
            printf("\n");       \
            exit(-1);            \
        } while (0)





#define TEST_TABLE  {                                                                         \
    [TEST_TYPE_1] = { CONN_STAGE_CLNT_CREATION , test_1_clnt_conn_code, test_1_listen_code  } , \
    [TEST_TYPE_2] = { CONN_STAGE_CLNT_CREATION , test_2_clnt_conn_code, test_2_listen_code  } , \
    [TEST_TYPE_3] = { CONN_STAGE_CLNT_CREATION , test_3_clnt_conn_code, test_3_listen_code  } , \
    [TEST_TYPE_4] = { CONN_STAGE_CLNT_CREATION , test_4_clnt_conn_code, test_4_listen_code  } , \
    [TEST_TYPE_5] = { CONN_STAGE_CLNT_CREATION , test_5_clnt_conn_code, test_5_listen_code  } , \
    [TEST_TYPE_6] = { CONN_STAGE_CLNT_CREATION , test_6_clnt_conn_code, test_6_listen_code  }   \
}

///////////////////////////////////////////////////
// STRUCT/ENUM DECLERATIONS
///////////////////////////////////////////////////

typedef enum {
    TEST_TYPE_1     = 0x01,
	TEST_TYPE_2     ,
    TEST_TYPE_3     ,
    TEST_TYPE_4     ,
    TEST_TYPE_5     ,
    TEST_TYPE_6     

} TEST_TYPE_enum;


typedef enum {
	CONN_STAGE_SERV_ACCEPT        ,
    CONN_STAGE_SERV_ESTB          ,
    CONN_STAGE_SERV_ESTB_DROP     ,
    CONN_STAGE_SERV_DISCONN       ,
    CONN_STAGE_SERV_IDLE          ,
	CONN_STAGE_CLNT_CREATION      ,
    CONN_STAGE_CLNT_CONFIRM       ,
    CONN_STAGE_CLNT_CONFIRM_FAIL  ,
    CONN_STAGE_CLNT_CONFIRM_DROP  ,
    CONN_STAGE_CLNT_ESTB          ,
    CONN_STAGE_CLNT_DISCONN       ,
    CONN_STAHE_CLNT_REJ           ,
    CONN_STAGE_CLNT_IDLE       

} CONN_STAGE_enum;


typedef struct QP_HANDLE_st {

    VAPI_qp_num_t   qpn;
    VAPI_qp_hndl_t  qp;
    VAPI_cq_hndl_t  rcq;
    VAPI_cq_hndl_t  scq;

}QP_HANDLE_t;

typedef struct QP_QUEUE_ELEM_st {

    QP_HANDLE_t * handle_p;

}QP_QUEUE_ELEM_t;


typedef struct QUEUE_st {
    
    int inx;

    int size;

    QP_HANDLE_t ** p_arr;

}QUEUE_t;


typedef struct CONNECTION_INFO_st {

    QP_HANDLE_t * qp_hndl_p;

    u_int64_t           sid;

    tTS_IB_CM_COMM_ID   cid;

    struct ib_path_record *path_record;
    
    TEST_TYPE_enum test_type;

    CONN_STAGE_enum      stage;
    
}CONNECTION_INFO_t;


typedef struct LISTEN_INFO_st {

    tTS_IB_LISTEN_HANDLE srv;

    u_int64_t sid;

}LISTEN_INFO_t;


typedef struct TEST_OPT_st {

    u_int64_t start_sid;
    u_int64_t end_sid;
    //int num_qp;

} TEST_OPT_t;


typedef struct SID_CONN_INFO_st {

    in_addr_t ip_addr;

    u_int64_t sid;

    struct ib_path_record path_record;

} SID_CONN_INFO_t;


typedef struct CM_EVENT_TABLE_st{
	//tTS_IB_CM_EVENT mod;
	char *desc;

} CM_EVENT_TABLE_t;

typedef struct cmtest_context_st
{
    VAPI_hca_hndl_t v_hca;

    VAPI_qp_hndl_t act_qp;
    VAPI_qp_num_t  act_qpn;

    //qp_pool_info;

} cmtest_context_t;


typedef struct TEST_FLOW_st
{
    CONN_STAGE_enum                 init_stage;

    tTS_IB_U_CM_COMPLETION_FUNC     clnt_func;

    tTS_IB_U_CM_COMPLETION_FUNC     serv_func;

} TEST_FLOW_t;

typedef struct TEST_PRIV_DATA_st
{
    TEST_TYPE_enum test_type;

} TEST_PRIV_DATA_t;


///////////////////////////////////////////////////
// FUNCTION DECLERATIONS
///////////////////////////////////////////////////

void init_qp_queue(QUEUE_t * q, unsigned int size);

int enQ_qp(QUEUE_t * q, QP_HANDLE_t * elem_p) ;

int deQ_qp(QUEUE_t * q, QP_HANDLE_t ** elem_p_p);

int get_port_gid(VAPI_hca_hndl_t hca_hndl, IB_port_t port, IB_gid_t * gid);

void init_hca(void) ;

void init_pd(void);

int ib_qp_alloc_init(VAPI_hca_hndl_t hca, IB_port_t port, QP_HANDLE_t * handle_p);

int reset_qp_state(QP_HANDLE_t * handle_p);

int qp_state_to_err(QP_HANDLE_t * handle_p);

void ib_qp_dealloc(VAPI_hca_hndl_t hca,QP_HANDLE_t * handle_p);

void qp_cleanup(QUEUE_t * q);

int init_qp_pool(QUEUE_t * q, QP_HANDLE_t *qp_hndl_arr) ;

unsigned int mlx_rand(unsigned int x, unsigned int y);



/********
*  TEST *
********/

int callcbk_is_err(tTS_IB_U_QP_HANDLE   consumer_qp,
                         tTS_IB_CM_COMM_ID    comm_id,
                         tTS_IB_CM_EVENT      event,
                         void                *params,
                         void                *arg);

void test_1_listen_code(tTS_IB_U_QP_HANDLE   consumer_qp,
                     tTS_IB_CM_COMM_ID    comm_id,
                     tTS_IB_CM_EVENT      event,
                     void                *params,
                     void                *arg);

int test_1_clnt_conn_code(tTS_IB_U_QP_HANDLE consumer_qp, tTS_IB_CM_COMM_ID comm_id, 
             tTS_IB_CM_EVENT event, void *params, void  *arg);

int test_2_clnt_conn_code(tTS_IB_U_QP_HANDLE consumer_qp, tTS_IB_CM_COMM_ID comm_id, 
                          tTS_IB_CM_EVENT event, void *params, void  *arg);

void test_2_listen_code(tTS_IB_U_QP_HANDLE   consumer_qp,
                        tTS_IB_CM_COMM_ID    comm_id,
                        tTS_IB_CM_EVENT      event,
                        void                *params,
                        void                *arg);
int test_3_clnt_conn_code(tTS_IB_U_QP_HANDLE consumer_qp, tTS_IB_CM_COMM_ID comm_id, 
                          tTS_IB_CM_EVENT event, void *params, void  *arg);

void test_3_listen_code(tTS_IB_U_QP_HANDLE   consumer_qp,
                        tTS_IB_CM_COMM_ID    comm_id,
                        tTS_IB_CM_EVENT      event,
                        void                *params,
                        void                *arg);

void test_4_listen_code(tTS_IB_U_QP_HANDLE   consumer_qp,
                     tTS_IB_CM_COMM_ID    comm_id,
                     tTS_IB_CM_EVENT      event,
                     void                *params,
                     void                *arg);

int test_4_clnt_conn_code(tTS_IB_U_QP_HANDLE consumer_qp, tTS_IB_CM_COMM_ID comm_id, 
             tTS_IB_CM_EVENT event, void *params, void  *arg);

void test_5_listen_code(tTS_IB_U_QP_HANDLE   consumer_qp,
                     tTS_IB_CM_COMM_ID    comm_id,
                     tTS_IB_CM_EVENT      event,
                     void                *params,
                     void                *arg);

int test_5_clnt_conn_code(tTS_IB_U_QP_HANDLE consumer_qp, tTS_IB_CM_COMM_ID comm_id, 
             tTS_IB_CM_EVENT event, void *params, void  *arg);

void test_6_listen_code(tTS_IB_U_QP_HANDLE   consumer_qp,
                     tTS_IB_CM_COMM_ID    comm_id,
                     tTS_IB_CM_EVENT      event,
                     void                *params,
                     void                *arg);

int test_6_clnt_conn_code(tTS_IB_U_QP_HANDLE consumer_qp, tTS_IB_CM_COMM_ID comm_id, 
             tTS_IB_CM_EVENT event, void *params, void  *arg);




#endif
