#include <cm_test.h> 
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>


VAPI_hca_hndl_t hca;

char * hca_name;

VAPI_hca_port_t hca_port;

IB_gid_t gid;

VAPI_pd_hndl_t        pd = 0;

#define MAX_SID_2_CONN 100

SID_CONN_INFO_t sid_ip_arr[MAX_SID_2_CONN];
int sid_inx = 0;

QP_HANDLE_t        * clnt_qp_hndl_arr;

QUEUE_t q;


CONNECTION_INFO_t * conn_arr;
int num_conn;


atomic_t clnt_refcnt;

TEST_FLOW_t test_table[] = TEST_TABLE;


inline void usage(const char *prog_name)
{
    printf("Usage: %s -h <hca name> -f <sid file name>  -n <num of connections> -s <seed> -i <server ip> -d <server sid>\n", prog_name);
    
}

int str2ip_num(char* ip_str, in_addr_t * in_num)
{
    char *cur_char,*next_char;
    int i;
    *in_num= 0;

    for (i= 0, cur_char= ip_str; i < 4; i++)
    {
        *in_num <<= 8;

        *in_num |= (u_int8_t)strtoul(cur_char, &next_char , 10) ;


        if ((next_char == cur_char) ||
            ((i < 3) && (*next_char != '.')) )
        {
            printf("decimal dotted format must be used for IP , e.g. 10.0.0.229 (4 numbers):\n-g ");
            cur_char= ip_str;
            while ((*cur_char != ' ') && (*cur_char != 0))
            {
                printf("%c",*cur_char); cur_char++;
            }
            printf("\n   ");
            for (; next_char > ip_str; next_char--)  printf(" ");
            printf("^\n");

            return 1;
        }
        cur_char= next_char+1;  /* skip dot */
    }

    return 0;

}


int get_from_file_sid(char *file_name)
{
    FILE * file;

    char * tmp;

    int sid;

    char * str, str_buff[80];

    char * curr_str;

    in_addr_t ip;

    file = fopen(file_name, "r");

    str =  str_buff;

    while (fgets(str, 80, file))
    {
        curr_str = strsep( &str, " \t");

        //if (str2ip_num(curr_str, &ip))
        if ((ip = inet_addr(curr_str)) == INADDR_NONE )
        {
            return 1;
        }
        //printf("the ip is 0x%x\n", ip);

        curr_str = strsep( &str, " \t");

        sid = strtoll(curr_str, &tmp , 10) ;

        if (tmp[0] == '\o')
        {
            return 1;
        }

        sid_ip_arr[sid_inx].sid = sid;

        sid_ip_arr[sid_inx].ip_addr = ip;

        sid_inx++;

        str =  str_buff;
    }

    return 0;
}

int get_test_opt(int argc, char *argv[], TEST_OPT_t * test_opt)
{
    int c;

    char *sid_file_name = NULL;
    char *num_conn_string = NULL;
    char *tmp;
    char *seed_str;
    char *server_ip = NULL;
    char *sid_str = NULL;

    int seed;

    hca_name = HCA_NAME;

    while ( (c= getopt(argc, argv, "h:f:n:s:i:d:")) != -1)
    {
        switch (c)
        {
        case 'h':
            hca_name = optarg;
            break;
        case 'f': 
            sid_file_name = optarg;
            break;
        case 'n':
            num_conn_string = optarg;
            break;
        case 's':
            seed_str = optarg;
            break;
        case 'i':
            server_ip = optarg;
            break;
        case 'd':
             sid_str = optarg;
            break;
        case ':': /* missing param. */
        case '?': /* Unknown option */
        default:
            usage(argv[0]);
            return -1;
        }
    }

    if (seed_str)
    {
        seed = strtoll(seed_str, &tmp, 0);

        if (seed <= 0)
        {
            goto err_exit;
        }

        //if (tmp) goto err_exit;
    }
    else
        seed = 1;

    srand(seed);

    if (num_conn_string)
    {
        num_conn = strtoll(num_conn_string, &tmp, 0);

        if (num_conn <= 0)
        {
            goto err_exit;
        }

        //if (tmp) goto err_exit;
    }
    else
        goto err_exit;

    conn_arr = malloc(sizeof(CONNECTION_INFO_t) * num_conn);

    printf("conn_arr = %p, sizeof(CONNECTION_INFO_t) %d, num_conn %d\n", 
           conn_arr, sizeof(CONNECTION_INFO_t), num_conn);

    clnt_qp_hndl_arr = malloc(sizeof(QP_HANDLE_t) * num_conn);

    printf("clnt_qp_hndl_arr = %p, sizeof(QP_HANDLE_t) %d, num_conn %d\n", 
           clnt_qp_hndl_arr, sizeof(QP_HANDLE_t), num_conn);

    init_qp_queue(&q, num_conn);

    printf("line %d\n", __LINE__);
    
    if (sid_file_name)
    {
        if (get_from_file_sid(sid_file_name))
        {
            goto err_exit;
        }
    } 
    else if (sid_str && server_ip)
    {
        in_addr_t ip;
        int sid;

        if ((ip = inet_addr(server_ip)) == INADDR_NONE )
        {
            goto err_exit;
        }

        sid = strtoll(sid_str, &tmp , 10) ;

        if (tmp[0] == '\o')
        {
            goto err_exit;
        }

        printf("line %d sid = %d, ip= 0x%x\n", __LINE__, sid, ip);

        sid_ip_arr[sid_inx].sid = sid;

        sid_ip_arr[sid_inx].ip_addr = ip;

        sid_inx++;

    } 
    else
        goto err_exit;

    return 0;


    err_exit:

    usage(argv[0]);

    return -1;

}


int
ib_connect(VAPI_hca_hndl_t hca, IB_lid_t lid, CONNECTION_INFO_t * conn_p)
{
    int                      status;
    //struct ib_path_record    path_record;
    struct ib_u_cm_qp_param  qp_param;

    TEST_PRIV_DATA_t priv_data;

    TEST_TYPE_enum test_type = conn_p->test_type;

    tTS_IB_U_CM_COMPLETION_FUNC callbck_func = test_table[test_type].clnt_func;

    priv_data.test_type = test_type;


    qp_param.responder_resources = 4;
    qp_param.initiator_depth     = 4;
    qp_param.retry_count         = 4;
    qp_param.rnr_retry_count     = 4;
    qp_param.cm_response_timeout = 20;
    qp_param.max_cm_retries      = 4;


    //memcpy(&path_record.sgid, gid, sizeof(IB_gid_t));
    //memcpy(&path_record.dgid, gid, sizeof(IB_gid_t));
    //path_record.slid        = lid;
    //path_record.dlid        = lid;
    //path_record.flowlabel   = 0;
    //path_record.hoplmt      = 64;
    //path_record.tclass      = 0;
    //path_record.pkey        = 0xffff;
    //path_record.sl          = 0;
    //path_record.rate        = IB_RATE_10GB;
    //path_record.packet_life = 0;
    //path_record.mtu = IB_MTU_1024; /* Best performance is obtained at this MTU */

    status = tsIbUCmConnect(hca, 
                            conn_p->qp_hndl_p->qp,conn_p->qp_hndl_p->qpn,
                            &qp_param, conn_p->path_record, NULL, 
                            conn_p->sid, &priv_data, sizeof(priv_data),
                            callbck_func, conn_p, &conn_p->cid);


    if (status)
    {
        debug("Failed in connect call status %d", status);

        perror(NULL);

        return(-1);
    }

    return 0;
}





int main(int argc, char *argv[])
{
    TEST_OPT_t test_opt;

    int i;

    if (get_test_opt(argc, argv, &test_opt) != 0)
    {
        err_exit("couldn't get test param\n");
    }

    init_hca();

    init_pd();

    if (init_qp_pool(&q, clnt_qp_hndl_arr))
    {
        VAPI_dealloc_pd(hca, pd);

        exit(-1);
    }

    for (i=0;i< sid_inx; i++)
    {
        int ret = tsIbUCmPathRecord( sid_ip_arr[i].ip_addr, &sid_ip_arr[i].path_record);

        if (ret)
            err_exit("Could not find path record form ip 0x%x", sid_ip_arr[i].ip_addr);
    }

    for (i=0;i < num_conn ; i++)
    {
        if (deQ_qp(&q, &conn_arr[i].qp_hndl_p))
            err_exit("Error in text Q not managed well\n");

        //printf("conn_arr[i].qp_hndl_p = %p\n", conn_arr[i].qp_hndl_p);

        int rnd = mlx_rand(sid_inx,0);

        conn_arr[i].sid = sid_ip_arr[rnd].sid;

        conn_arr[i].path_record = &sid_ip_arr[rnd].path_record;

        {
            TEST_TYPE_enum test_type = mlx_rand(TEST_TYPE_1, TEST_TYPE_6); 
            conn_arr[i].test_type = test_type;
            conn_arr[i].stage = test_table[test_type].init_stage;
        }
    }


    atomic_set(&clnt_refcnt, 0);    

    for (i=0;i < num_conn ; i++)
    {

        if (ib_connect(hca, hca_port.lid, &conn_arr[i]))
        {
            exit(-1);
        }

        atomic_inc(&clnt_refcnt);
    }

    while (1)
    {
        int tmp_ref;

        sleep(1);

        if ((tmp_ref = atomic_read(&clnt_refcnt)) <= 0 )
        {
            debug("tmp_ref %d\n", tmp_ref);

            break;
        }
    }




    return 0;

}
