#include <unistd.h>
#include <stdio.h>

#include "ts_ib_useraccess_cm.h"	/* ioctl API to kernel */

#define HCA_NAME "InfiniHost0"
#define HCA_PORT 1
#define SID 0x10

#define MAX_PRIV_DATA_SIZE  128


char pasv_priv_data[MAX_PRIV_DATA_SIZE];
char actv_priv_data[MAX_PRIV_DATA_SIZE];

volatile int test_done = 0;



struct cmtest_context
{
    VAPI_hca_hndl_t v_hca;

    VAPI_qp_hndl_t act_qp;
    VAPI_qp_num_t  act_qpn;

    VAPI_qp_hndl_t psv_qp;
    VAPI_qp_num_t  psv_qpn;

} cm_test_ctx = { 0 };



/*
 *  alloc_get_all_port_gids
 */
#define MAX_GID_TBL 0x20

int get_port_gid(VAPI_hca_hndl_t hca_hndl, IB_port_t port, IB_gid_t * gid)
{
  VAPI_ret_t rc;
  u_int16_t tbl_sz;
  IB_gid_t gid_tbl[MAX_GID_TBL] = { {0} };
  

  rc = VAPI_query_hca_gid_tbl(hca_hndl, port, MAX_GID_TBL, &tbl_sz, gid_tbl);
  if ( rc != VAPI_OK )
      return(-1);

  memcpy(gid, gid_tbl[0], sizeof(IB_gid_t));


  return(0);

}


#define QP_CQ_SIZE  0x20
int ib_qp_alloc_init(VAPI_hca_hndl_t hca, IB_port_t port, VAPI_qp_hndl_t * qp_out, VAPI_qp_num_t * qpn)
{
    //VAPI_ret_t            ib_status;
    VAPI_qp_init_attr_t   qp_init_attr;
    VAPI_qp_prop_t        qp_prop;
    
    VAPI_qp_attr_t        qp_attr = { 0 };
    VAPI_qp_attr_mask_t   qp_attr_mask = 0;
    VAPI_qp_cap_t         qp_cap;

    VAPI_qp_hndl_t        qp = 0;
    VAPI_pd_hndl_t        pd = 0;
    VAPI_cq_hndl_t        cq_recv = 0;
    VAPI_cq_hndl_t        cq_send = 0;
    

    int cq_sz;


    if(VAPI_alloc_pd(hca, &pd))
        goto qp_alloc_out;

    if(VAPI_create_cq(hca, QP_CQ_SIZE, &cq_recv, &cq_sz))
        goto qp_alloc_out;

    if(VAPI_create_cq(hca, QP_CQ_SIZE, &cq_send, &cq_sz))
        goto qp_alloc_out; 

    qp_init_attr.cap.max_inline_data_sq    = 0;
    qp_init_attr.cap.max_oust_wr_sq        = 0x10;
    qp_init_attr.cap.max_oust_wr_rq        = 0x10;
    qp_init_attr.cap.max_sg_size_sq        = 0x10;
    qp_init_attr.cap.max_sg_size_rq        = 0x10;
    qp_init_attr.pd_hndl                   = pd;
    qp_init_attr.rdd_hndl                  = 0;
    qp_init_attr.rq_cq_hndl                = cq_recv;
    qp_init_attr.rq_sig_type               = VAPI_SIGNAL_ALL_WR;
    qp_init_attr.sq_cq_hndl                = cq_send;
    qp_init_attr.sq_sig_type               = VAPI_SIGNAL_REQ_WR;
    qp_init_attr.ts_type                   = VAPI_TS_RC;

    if(VAPI_create_qp(hca, &qp_init_attr, &qp, &qp_prop))
        goto qp_alloc_out;

    qp_attr.qp_state = VAPI_INIT;  
    QP_ATTR_MASK_SET (qp_attr_mask, QP_ATTR_QP_STATE);

    qp_attr.remote_atomic_flags = VAPI_EN_REM_WRITE |  VAPI_EN_REM_READ;
    QP_ATTR_MASK_SET (qp_attr_mask, QP_ATTR_REMOTE_ATOMIC_FLAGS);

    qp_attr.pkey_ix = 0;
    QP_ATTR_MASK_SET (qp_attr_mask, QP_ATTR_PKEY_IX);

    qp_attr.port = port;

    
    QP_ATTR_MASK_SET (qp_attr_mask, QP_ATTR_PORT);

    if(VAPI_modify_qp(hca, qp, &qp_attr, &qp_attr_mask, &qp_cap))
        return(-1);

    *qpn = qp_prop.qp_num;
    *qp_out = qp;

    return 0;


qp_alloc_out:
    if(qp)
        VAPI_destroy_qp(hca, qp);

    if(cq_recv)
        VAPI_destroy_cq(hca, cq_recv);

    if(cq_send)
        VAPI_destroy_cq(hca, cq_send);

    if(pd)
        VAPI_dealloc_pd(hca, pd);

    return(-1);
}






static int do_active_idle(struct ib_u_cm_idle_param *idle_param)
{
    switch (idle_param->reason)
    {
    case TS_IB_CM_IDLE_REMOTE_REJECT:

        printf("Rejected\n");
        break;

    case TS_IB_CM_IDLE_REMOTE_TIMEOUT:

        printf("timeout\n");
        break;

    case TS_IB_CM_IDLE_TIME_WAIT_DONE:

        printf("disconnect\n");
        break;

    case TS_IB_CM_IDLE_LOCAL_REJECT:

    default:
        printf("failure\n");
        break;

    }

    return 0;
}




int active_callback(tTS_IB_U_QP_HANDLE consumer_qp, tTS_IB_CM_COMM_ID comm_id, 
             tTS_IB_CM_EVENT event, void *params, void  *arg)
{
    int     ret = 0;


    switch (event)
    {
    case TS_IB_CM_IDLE:

        do_active_idle(params);
        break;

    case TS_IB_CM_REP_RECEIVED:

        printf("Rep received\n");

        /* Showing remote gid */
        {
//            tTS_IB_GID rgid;
//
//            ret = ib_cm_get_remote_gid(comm_id, &rgid);
//            if(ret < 0)
//                printf("failed to get remote gid\n");
//            else
//                printf("Remote id is %02x:%02x:%02x:%02x:%02x:%02x\n",
//                       0xff&rgid[0], 0xff&rgid[1], 
//                       0xff&rgid[2], 0xff&rgid[3], 
//                       0xff&rgid[4], 0xff&rgid[5]);
            
        }


        tsIbUCmConfirm(comm_id, NULL, 0);
        break;

    case TS_IB_CM_ESTABLISHED:

        printf("Active established...\n");

        break;

    case TS_IB_CM_REQ_RECEIVED:
        printf("REQ RE\n");
    default:

        printf("Unknown\n");
        break;

    }

    return ret;
}




int
ib_connect(VAPI_hca_hndl_t hca, IB_gid_t gid, IB_lid_t lid,
           VAPI_qp_hndl_t qp, VAPI_qp_num_t qpn, tTS_IB_SERVICE_ID service_id)
{
    int                      status;
    struct ib_path_record    path_record;
    struct ib_u_cm_qp_param  qp_param;
    tTS_IB_CM_COMM_ID        comm_id;



    qp_param.responder_resources = 4;
    qp_param.initiator_depth     = 4;
    qp_param.retry_count         = 4;
    qp_param.rnr_retry_count     = 4;
    qp_param.cm_response_timeout = 4;
    qp_param.max_cm_retries      = 4;


    memcpy(&path_record.sgid, gid, sizeof(IB_gid_t));
    memcpy(&path_record.dgid, gid, sizeof(IB_gid_t));

    path_record.slid        = lid;
    path_record.dlid        = lid;
    path_record.flowlabel   = 0;
    path_record.hoplmt      = 64;
    path_record.tclass      = 0;
    path_record.pkey        = 0xffff;
    path_record.sl          = 0;
    path_record.rate        = IB_RATE_10GB;
    path_record.packet_life = 0;

    path_record.mtu = IB_MTU_1024; /* Best performance is obtained at this MTU */

    status = tsIbUCmConnect(hca, qp, qpn, &qp_param, &path_record, NULL, service_id,
                            0, 0, active_callback, 0, &comm_id);

    if (status)
    {
        printf("Failed in connect call");
        return(-1);
    }

    return 0;
}


int passive_established(tTS_IB_U_QP_HANDLE consumer_qp, tTS_IB_CM_COMM_ID comm_id,  
                        tTS_IB_CM_EVENT event, void *params, void  *arg)
{
    printf("Passive established...\n");
    test_done = 1;
    return(0);
}


int passive_callback(tTS_IB_U_QP_HANDLE   consumer_qp,
                     tTS_IB_CM_COMM_ID    comm_id,
                     tTS_IB_CM_EVENT      event,
                     void                *params,
                     void                *arg)
{
    int ret = 0;
    VAPI_qp_hndl_t qp;
    VAPI_qp_num_t  qpn;



    switch (event)
    {
    case TS_IB_CM_IDLE:

        printf("RCV: TS_IB_CM_IDLE\n");
        break;

    case TS_IB_CM_REQ_RECEIVED:

        if (ib_qp_alloc_init(cm_test_ctx.v_hca, HCA_PORT, &qp, &qpn))
            return(-1);

        ret = tsIbUCmAccept(comm_id, cm_test_ctx.v_hca, qp, qpn, passive_established, pasv_priv_data, 
                            pasv_priv_data, MAX_PRIV_DATA_SIZE);         

        break;

    case TS_IB_CM_ESTABLISHED:

        printf("RCV: Establish\n");
        break;

    default:

        printf("RCV: Unkown\n");
        break;

    }

    return ret;
}


int ib_listen(u_int64_t sid, tTS_IB_LISTEN_HANDLE srv)
{
    int status;

    status = tsIbUCmListen(sid, TS_IB_CM_SERVICE_EXACT_MASK, passive_callback,
                           &cm_test_ctx, &srv);



    return(status);

}




int main()
{
    VAPI_hca_hndl_t hca;
    VAPI_hca_port_t hca_port;
    VAPI_qp_hndl_t qp;
    VAPI_qp_num_t  qpn;
    tTS_IB_LISTEN_HANDLE srv;


    IB_gid_t gid;

    if (EVAPI_get_hca_hndl(HCA_NAME, &hca))
        exit(-1);

    cm_test_ctx.v_hca = hca;


    if (get_port_gid(hca, HCA_PORT, &gid))
        exit(-1);

    if (VAPI_query_hca_port_prop(hca, HCA_PORT, &hca_port))
        exit(-1);

    if (ib_qp_alloc_init(hca, HCA_PORT, &qp, &qpn))
        exit(-1);

    cm_test_ctx.act_qp = qp;
    cm_test_ctx.act_qpn = qpn;


    ib_listen(SID, &srv);

    if (ib_connect(hca, gid, hca_port.lid, qp, qpn, SID))
        exit(-1);


    sleep(1);

    if(test_done)
    {
        printf("Test OK\n");
    }
    else 
    {
        printf("Test Failed\n");
        exit(1);
    }

    return(0);
}
