/*
 * Copyright (c) 2002-2003, Network Appliance, Inc. All rights reserved.
 * Copyright (c) 2005 Mellanox Technologies.   All rights reserved.

 *
 * This Software is licensed under either one of the following two licenses:
 *
 * 1) under the terms of the "Common Public License 1.0" a copy of which is
 *    in the file LICENSE.txt in the root directory. The license is also
 *    available from the Open Source Initiative, see
 *    http://www.opensource.org/licenses/cpl.php.
 * OR
 *
 * 2) under the terms of the "The BSD License" a copy of which is in the file
 *    LICENSE2.txt in the root directory. The license is also available from
 *    the Open Source Initiative, see
 *    http://www.opensource.org/licenses/bsd-license.php.
 *
 * Licensee has the right to choose either one of the above two licenses.
 *
 * Redistributions of source code must retain both the above copyright
 * notice and either one of the license notices.
 *
 * Redistributions in binary form must reproduce both the above copyright
 * notice, either one of the license notices in the documentation
 * and/or other materials provided with the distribution.
 */

/**********************************************************************
 *
 * MODULE: kdat.c
 *
 * PURPOSE: DAT Provider and Consumer registry functions.
 *
 * $Id: kdat_api.c,v 1.4 2005/02/11 20:17:04 jlentini Exp $
 **********************************************************************/

#include "dat_osd.h"
#include <dat/dat_registry.h>
#include "dat_init.h"

#define UDAT_IS_BAD_HANDLE(h) ( NULL == (p) )

DAT_RETURN dat_lmr_kcreate (
	IN      DAT_IA_HANDLE		ia_handle,
	IN      DAT_MEM_TYPE		mem_type,
	IN      DAT_REGION_DESCRIPTION	region_description,
	IN      DAT_VLEN		length,
	IN      DAT_PZ_HANDLE		pz_handle,
	IN      DAT_MEM_PRIV_FLAGS	privileges,
	IN      DAT_MEM_OPTIMIZE_FLAGS	mem_optimization,
	OUT     DAT_LMR_HANDLE	 	*lmr_handle,
	OUT     DAT_LMR_CONTEXT		*lmr_context,
	OUT     DAT_RMR_CONTEXT		*rmr_context,
	OUT     DAT_VLEN 		*registered_length,
	OUT     DAT_VADDR 		*registered_address)
{
    DAT_IA_HANDLE	dapl_ia_handle;
    DAT_RETURN		dat_status;

    dat_status = dats_get_ia_handle((unsigned long)ia_handle,
				    &dapl_ia_handle);
    if (dat_status == DAT_SUCCESS)
    {
	dat_status = DAT_LMR_KCREATE (dapl_ia_handle,
				      mem_type,
				      region_description,
				      length,
				      pz_handle,
				      privileges,
				      mem_optimization,
				      lmr_handle,
				      lmr_context,
				      rmr_context,
				      registered_length,
				      registered_address);
    }

    return dat_status;
}


DAT_RETURN dat_ia_memtype_hint (
	IN      DAT_IA_HANDLE		ia_handle,
	IN      DAT_MEM_TYPE		mem_type,
	IN      DAT_VLEN		length,
	IN      DAT_MEM_OPTIMIZE_FLAGS	mem_optimization,
	OUT     DAT_VLEN *		preferred_length,
	OUT     DAT_VADDR * 		preferred_alignment)
{
    DAT_IA_HANDLE	dapl_ia_handle;
    DAT_RETURN		dat_status;

    dat_status = dats_get_ia_handle((unsigned long)ia_handle,
				    &dapl_ia_handle);
    if (dat_status == DAT_SUCCESS)
    {
	dat_status = DAT_IA_MEMTYPE_HINT (dapl_ia_handle,
					  mem_type,
					  length,
					  mem_optimization,
					  preferred_length,
					  preferred_alignment);
    }

    return dat_status;
}


DAT_RETURN dat_evd_kcreate (
	IN      DAT_IA_HANDLE		ia_handle,
	IN      DAT_COUNT		evd_min_qlen,
	IN      DAT_UPCALL_POLICY	upcall_policy,
	IN      const DAT_UPCALL_OBJECT *upcall,
	IN      DAT_EVD_FLAGS		evd_flags,
	OUT     DAT_EVD_HANDLE		*evd_handle)
{
    DAT_IA_HANDLE	dapl_ia_handle;
    DAT_RETURN		dat_status;

    dat_status = dats_get_ia_handle((unsigned long)ia_handle,
				    &dapl_ia_handle);
    if (dat_status == DAT_SUCCESS)
    {
	dat_status = DAT_EVD_KCREATE (dapl_ia_handle,
				      evd_min_qlen,
				      upcall_policy,
				      upcall,
				      evd_flags,
				      evd_handle);
    }

    return dat_status;
}


DAT_RETURN dat_evd_modify_upcall (
	IN      DAT_EVD_HANDLE		evd_handle,
	IN      DAT_UPCALL_POLICY	upcall_policy,
	IN      const DAT_UPCALL_OBJECT	*upcall)
{
    if (evd_handle == NULL)
    {
	return DAT_ERROR(DAT_INVALID_HANDLE, DAT_INVALID_HANDLE_EVD_REQUEST);
    }
    return DAT_EVD_MODIFY_UPCALL (evd_handle,
				  upcall_policy,
				  upcall);
}


DAT_RETURN dat_cr_handoff (
    IN          DAT_CR_HANDLE		cr_handle,
    IN          DAT_CONN_QUAL		handoff)
{
    if (cr_handle == NULL)
    {
	return DAT_ERROR(DAT_INVALID_HANDLE, DAT_INVALID_HANDLE_CR);
    }
    return DAT_CR_HANDOFF (cr_handle,
			   handoff);
}


DAT_RETURN dat_evd_kquery (
	IN      DAT_EVD_HANDLE		evd_handle,
	IN      DAT_EVD_PARAM_MASK	evd_param_mask,
	OUT     DAT_EVD_PARAM		*evd_param)
{
    if (evd_handle == NULL)
    {
	return DAT_ERROR(DAT_INVALID_HANDLE, DAT_INVALID_HANDLE_EVD_REQUEST);
    }
    return DAT_EVD_QUERY (evd_handle,
			  evd_param_mask,
			  evd_param);
}


DAT_RETURN dat_lmr_query (
	IN      DAT_LMR_HANDLE		lmr_handle,
	IN      DAT_LMR_PARAM_MASK	lmv_param_mask,
	OUT     DAT_LMR_PARAM		*lmr_param)
{
    if (lmr_handle == NULL)
    {
	return DAT_ERROR(DAT_INVALID_HANDLE, DAT_INVALID_HANDLE_LMR);
    }
    return DAT_LMR_QUERY (lmr_handle,
			  lmv_param_mask,
			  lmr_param);
}

/*
 * Local variables:
 *  c-indent-level: 4
 *  c-basic-offset: 4
 *  tab-width: 8
 * End:
 */
