/*
 * Copyright (c) 2002-2003, Network Appliance, Inc. All rights reserved.
 * Copyright (c) 2005 Mellanox Technologies.   All rights reserved.

 *
 * This Software is licensed under both of the following two licenses:
 *
 * 1) under the terms of the "Common Public License 1.0" a copy of which is
 *    in the file LICENSE.txt in the root directory. The license is also
 *    available from the Open Source Initiative, see
 *    http://www.opensource.org/licenses/cpl.php.
 * OR
 *
 * 2) under the terms of the "The BSD License" a copy of which is in the file
 *    LICENSE2.txt in the root directory. The license is also available from
 *    the Open Source Initiative, see
 *    http://www.opensource.org/licenses/bsd-license.php.
 *
 * Licensee has the right to choose either one of the above two licenses.
 *
 * Redistributions of source code must retain both the above copyright
 * notice and either one of the license notices.
 *
 * Redistributions in binary form must reproduce both the above copyright
 * notice, either one of the license notices in the documentation
 * and/or other materials provided with the distribution.
 */

/**********************************************************************
 *
 * MODULE: kdat_module.c
 *
 * PURPOSE: DAT registry module implementation, exists in the Linux
 * kernel
 * Description: a linux module implementation
 *
 * $Id: kdat_module.c,v 1.2 2005/01/28 22:33:05 jlentini Exp $
 **********************************************************************/

#define EXPORT_SYMTAB
#include <dat/kdat.h>
#include <dat/dat_registry.h>
#include "dat_osd.h"
#include "dat_init.h"

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/init.h>
#include <linux/version.h>
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
#include <linux/moduleparam.h>
#endif
MODULE_LICENSE("Dual BSD/GPL");
MODULE_DESCRIPTION("kDAT registry for Linux kernel REVISION 1.8.0");
MODULE_AUTHOR("Fujitsu PST, Network Appliance, Inc.");

int             DbgLvl = 0;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
module_param(DbgLvl, int, 0444);
#else
MODULE_PARM ( DbgLvl, "i" );
#endif
MODULE_PARM_DESC(DbgLvl, "Bit mask to enable dat debugging, equivelent to DAT_DBG_TYPE");


/***********************************************************************
 * init_module
 *
 * Entry point for a Linux module, performs simple initialization
 *
 * Input:
 *	none
 *
 * Output:
 *	none
 *
 * Returns:
 *	SUCCESS
 ***********************************************************************/

static int
kdat_module_init (void)
{
    dat_os_printf ("<1>DAT Registry: Started\n");
    dat_init ();

    return ((int)0);		/* success */
}

/***********************************************************************
 * cleanup_module
 *
 * Entry point for a Linux module, cleans up the module on exit
 *
 * Input:
 *	none
 *
 * Output:
 *	none
 *
 * Returns:
 *	void
 ***********************************************************************/
static void
kdat_module_exit (void)
{
    dat_os_printf ("<1>DAT Registry: Stopped\n");
    dat_fini ();

    return;
}

module_init(kdat_module_init);
module_exit(kdat_module_exit);


/* Module exports */
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
EXPORT_SYMBOL (dat_ia_openv);
EXPORT_SYMBOL (dat_ia_close);

EXPORT_SYMBOL (dat_ia_query);
EXPORT_SYMBOL (dat_set_consumer_context);
EXPORT_SYMBOL (dat_get_consumer_context);
EXPORT_SYMBOL (dat_get_handle_type);
EXPORT_SYMBOL (dat_cr_query);
EXPORT_SYMBOL (dat_cr_accept);
EXPORT_SYMBOL (dat_cr_reject);
EXPORT_SYMBOL (dat_cr_handoff);
EXPORT_SYMBOL (dat_evd_resize);
EXPORT_SYMBOL (dat_evd_post_se);
EXPORT_SYMBOL (dat_evd_dequeue);
EXPORT_SYMBOL (dat_evd_query);
EXPORT_SYMBOL (dat_evd_free);
EXPORT_SYMBOL (dat_ep_create);
EXPORT_SYMBOL (dat_ep_query);
EXPORT_SYMBOL (dat_ep_modify);
EXPORT_SYMBOL (dat_ep_connect);
EXPORT_SYMBOL (dat_ep_dup_connect);
EXPORT_SYMBOL (dat_ep_disconnect);
EXPORT_SYMBOL (dat_ep_post_send);
EXPORT_SYMBOL (dat_ep_post_recv);
EXPORT_SYMBOL (dat_ep_post_rdma_read);
EXPORT_SYMBOL (dat_ep_post_rdma_write);
EXPORT_SYMBOL (dat_ep_get_status);
EXPORT_SYMBOL (dat_ep_free);
EXPORT_SYMBOL (dat_ep_reset);
EXPORT_SYMBOL (dat_ep_create_with_srq);
EXPORT_SYMBOL (dat_ep_recv_query);
EXPORT_SYMBOL (dat_ep_set_watermark);
EXPORT_SYMBOL (dat_lmr_query);
EXPORT_SYMBOL (dat_lmr_free);
EXPORT_SYMBOL (dat_lmr_sync_rdma_read);
EXPORT_SYMBOL (dat_lmr_sync_rdma_write);
EXPORT_SYMBOL (dat_rmr_create);
EXPORT_SYMBOL (dat_rmr_query);
EXPORT_SYMBOL (dat_rmr_bind);
EXPORT_SYMBOL (dat_rmr_free);
EXPORT_SYMBOL (dat_psp_create);
EXPORT_SYMBOL (dat_psp_create_any);
EXPORT_SYMBOL (dat_psp_query);
EXPORT_SYMBOL (dat_psp_free);
EXPORT_SYMBOL (dat_rsp_create);
EXPORT_SYMBOL (dat_rsp_query);
EXPORT_SYMBOL (dat_rsp_free);
EXPORT_SYMBOL (dat_pz_create);
EXPORT_SYMBOL (dat_pz_query);
EXPORT_SYMBOL (dat_pz_free);
EXPORT_SYMBOL (dat_srq_create);
EXPORT_SYMBOL (dat_srq_free);
EXPORT_SYMBOL (dat_srq_post_recv);
EXPORT_SYMBOL (dat_srq_query);
EXPORT_SYMBOL (dat_srq_resize);
EXPORT_SYMBOL (dat_srq_set_lw);

EXPORT_SYMBOL (dat_lmr_kcreate);
EXPORT_SYMBOL (dat_ia_memtype_hint);
EXPORT_SYMBOL (dat_evd_kcreate);
EXPORT_SYMBOL (dat_evd_modify_upcall);

EXPORT_SYMBOL (dat_registry_remove_provider);
EXPORT_SYMBOL (dat_registry_add_provider);

EXPORT_SYMBOL (dat_strerror);
#else
EXPORT_SYMBOL_NOVERS (dat_ia_openv);
EXPORT_SYMBOL_NOVERS (dat_ia_close);

EXPORT_SYMBOL_NOVERS (dat_ia_query);
EXPORT_SYMBOL_NOVERS (dat_set_consumer_context);
EXPORT_SYMBOL_NOVERS (dat_get_consumer_context);
EXPORT_SYMBOL_NOVERS (dat_get_handle_type);
EXPORT_SYMBOL_NOVERS (dat_cr_query);
EXPORT_SYMBOL_NOVERS (dat_cr_accept);
EXPORT_SYMBOL_NOVERS (dat_cr_reject);
EXPORT_SYMBOL_NOVERS (dat_cr_handoff);
EXPORT_SYMBOL_NOVERS (dat_evd_resize);
EXPORT_SYMBOL_NOVERS (dat_evd_post_se);
EXPORT_SYMBOL_NOVERS (dat_evd_dequeue);
EXPORT_SYMBOL_NOVERS (dat_evd_query);
EXPORT_SYMBOL_NOVERS (dat_evd_free);
EXPORT_SYMBOL_NOVERS (dat_ep_create);
EXPORT_SYMBOL_NOVERS (dat_ep_query);
EXPORT_SYMBOL_NOVERS (dat_ep_modify);
EXPORT_SYMBOL_NOVERS (dat_ep_connect);
EXPORT_SYMBOL_NOVERS (dat_ep_dup_connect);
EXPORT_SYMBOL_NOVERS (dat_ep_disconnect);
EXPORT_SYMBOL_NOVERS (dat_ep_post_send);
EXPORT_SYMBOL_NOVERS (dat_ep_post_recv);
EXPORT_SYMBOL_NOVERS (dat_ep_post_rdma_read);
EXPORT_SYMBOL_NOVERS (dat_ep_post_rdma_write);
EXPORT_SYMBOL_NOVERS (dat_ep_get_status);
EXPORT_SYMBOL_NOVERS (dat_ep_free);
EXPORT_SYMBOL_NOVERS (dat_ep_reset);
EXPORT_SYMBOL_NOVERS (dat_ep_create_with_srq);
EXPORT_SYMBOL_NOVERS (dat_ep_recv_query);
EXPORT_SYMBOL_NOVERS (dat_ep_set_watermark);
EXPORT_SYMBOL_NOVERS (dat_lmr_query);
EXPORT_SYMBOL_NOVERS (dat_lmr_free);
EXPORT_SYMBOL_NOVERS (dat_lmr_sync_rdma_read);
EXPORT_SYMBOL_NOVERS (dat_lmr_sync_rdma_write);
EXPORT_SYMBOL_NOVERS (dat_rmr_create);
EXPORT_SYMBOL_NOVERS (dat_rmr_query);
EXPORT_SYMBOL_NOVERS (dat_rmr_bind);
EXPORT_SYMBOL_NOVERS (dat_rmr_free);
EXPORT_SYMBOL_NOVERS (dat_psp_create);
EXPORT_SYMBOL_NOVERS (dat_psp_create_any);
EXPORT_SYMBOL_NOVERS (dat_psp_query);
EXPORT_SYMBOL_NOVERS (dat_psp_free);
EXPORT_SYMBOL_NOVERS (dat_rsp_create);
EXPORT_SYMBOL_NOVERS (dat_rsp_query);
EXPORT_SYMBOL_NOVERS (dat_rsp_free);
EXPORT_SYMBOL_NOVERS (dat_pz_create);
EXPORT_SYMBOL_NOVERS (dat_pz_query);
EXPORT_SYMBOL_NOVERS (dat_pz_free);
EXPORT_SYMBOL_NOVERS (dat_srq_create);
EXPORT_SYMBOL_NOVERS (dat_srq_free);
EXPORT_SYMBOL_NOVERS (dat_srq_post_recv);
EXPORT_SYMBOL_NOVERS (dat_srq_query);
EXPORT_SYMBOL_NOVERS (dat_srq_resize);
EXPORT_SYMBOL_NOVERS (dat_srq_set_lw);

EXPORT_SYMBOL_NOVERS (dat_lmr_kcreate);
EXPORT_SYMBOL_NOVERS (dat_ia_memtype_hint);
EXPORT_SYMBOL_NOVERS (dat_evd_kcreate);
EXPORT_SYMBOL_NOVERS (dat_evd_modify_upcall);

EXPORT_SYMBOL_NOVERS (dat_registry_remove_provider);
EXPORT_SYMBOL_NOVERS (dat_registry_add_provider);

EXPORT_SYMBOL_NOVERS (dat_strerror);

#endif

/*
 * Local variables:
 *  c-indent-level: 4
 *  c-basic-offset: 4
 *  tab-width: 8
 * End:
 */
