/*
 * Copyright (c) 2002-2003, Network Appliance, Inc. All rights reserved.
 * Copyright (c) 2005 Mellanox Technologies.   All rights reserved.

 *
 * This Software is licensed under both of the following two licenses:
 *
 * 1) under the terms of the "Common Public License 1.0" a copy of which is
 *    in the file LICENSE.txt in the root directory. The license is also
 *    available from the Open Source Initiative, see
 *    http://www.opensource.org/licenses/cpl.php.
 * OR
 *
 * 2) under the terms of the "The BSD License" a copy of which is in the file
 *    LICENSE2.txt in the root directory. The license is also available from
 *    the Open Source Initiative, see
 *    http://www.opensource.org/licenses/bsd-license.php.
 *
 * Licensee has the right to choose either one of the above two licenses.
 *
 * Redistributions of source code must retain both the above copyright
 * notice and either one of the license notices.
 *
 * Redistributions in binary form must reproduce both the above copyright
 * notice, either one of the license notices in the documentation
 * and/or other materials provided with the distribution.
 */

/**********************************************************************
 * 
 * MODULE: dapl_evd_un_async_error_callback.c
 *
 * PURPOSE: implements Unaffiliated callbacks from verbs
 *
 * $Id: dapl_evd_un_async_error_callb.c,v 1.2 2004/08/02 15:14:26 rabin Exp $
 **********************************************************************/

#include "dapl.h"
#include "dapl_evd_util.h"
#include "dapl_adapter_util.h"

/*
 * dapl_evd_un_async_error_callback
 *
 * The callback function registered with verbs for unaffiliated async errors
 *
 * Input:
 * 	ib_cm_handle,
 * 	ib_cm_event
 * 	cause_ptr
 * 	context (evd)
 *
 * Output:
 * 	None
 *
 */

void 
dapl_evd_un_async_error_callback (
	IN	ib_hca_handle_t		ib_hca_handle,
	IN	ib_error_record_t *	cause_ptr,
	IN	void *			context)

{
    DAPL_EVD		*async_evd;
    DAT_EVENT_NUMBER	async_event;
    DAT_RETURN		dat_status;

    dapl_dbg_log (DAPL_DBG_TYPE_CALLBACK | DAPL_DBG_TYPE_EXCEPTION,
		  "dapl_evd_un_async_error_callback (%p, %p, %p)\n", 
		  ib_hca_handle, 
		  cause_ptr, 
		  context);

    if ( NULL == context )
    {
	dapl_os_panic ("NULL == context\n");
	return;
    }

    async_evd = (DAPL_EVD *) context;

    dat_status = dapls_ib_get_async_event(cause_ptr, &async_event);

    if ( dat_status == DAT_SUCCESS )
    {
	/*
	 * If dapls_ib_get_async_event is not successful,
	 * an event has been generated by the provider that
	 * we are not interested in. E.g. LINK_UP.
	 */
	dapls_evd_post_async_error_event( async_evd,
					  async_event,
					  async_evd->header.owner_ia);
    }
    dapl_dbg_log (DAPL_DBG_TYPE_CALLBACK | DAPL_DBG_TYPE_EXCEPTION,
		  "dapl_evd_un_async_error_callback () returns\n");
}

