/*
 * Copyright (c) 2002-2003, Network Appliance, Inc. All rights reserved.
 * Copyright (c) 2005 Mellanox Technologies.   All rights reserved.

 *
 * This Software is licensed under both of the following two licenses:
 *
 * 1) under the terms of the "Common Public License 1.0" a copy of which is
 *    in the file LICENSE.txt in the root directory. The license is also
 *    available from the Open Source Initiative, see
 *    http://www.opensource.org/licenses/cpl.php.
 * OR
 *
 * 2) under the terms of the "The BSD License" a copy of which is in the file
 *    LICENSE2.txt in the root directory. The license is also available from
 *    the Open Source Initiative, see
 *    http://www.opensource.org/licenses/bsd-license.php.
 *
 * Licensee has the right to choose either one of the above two licenses.
 *
 * Redistributions of source code must retain both the above copyright
 * notice and either one of the license notices.
 *
 * Redistributions in binary form must reproduce both the above copyright
 * notice, either one of the license notices in the documentation
 * and/or other materials provided with the distribution.
 */

#include "dapl_mr_util.h"

/**********************************************************************
 * 
 * MODULE: dapl_mr_util.c
 *
 * PURPOSE: Common Memory Management functions and data structures
 *
 * $Id: dapl_mr_util.c,v 1.3 2004/08/02 15:14:27 rabin Exp $
 **********************************************************************/

/*********************************************************************
 *                                                                   *
 * Function Definitions	                                             *
 *                                                                   *
 *********************************************************************/

/*
 * dapl_mr_get_address
 *
 * Returns the memory address associated with the given memory descriptor
 *
 * Input:
 * 	desc		memory descriptor
 * 	type		type of memory represented by desc
 *
 * Output:
 * 	None
 *
 */

DAT_VADDR
dapl_mr_get_address (
	IN  DAT_REGION_DESCRIPTION   desc, 
	IN  DAT_MEM_TYPE             type)
{
    switch (type)
    {
	case DAT_MEM_TYPE_VIRTUAL:
	{
	    return (DAT_VADDR) (uintptr_t) desc.for_va;
	}
	case DAT_MEM_TYPE_LMR:
	{
	    DAPL_LMR 	*lmr;

	    lmr = (DAPL_LMR *) desc.for_lmr_handle;

	    /* Since this function is recoursive we cannot inline it */
	    return dapl_mr_get_address (lmr->param.region_desc, 
                                   lmr->param.mem_type);
	}
#if defined(__KDAPL__)
    case DAT_MEM_TYPE_PLATFORM:
	case DAT_MEM_TYPE_IA:
	case DAT_MEM_TYPE_PHYSICAL:
	{
	    /* XXX FIXME!!
         * XXX support 64 bit phys addrs.
         * XXX support vector of size>1 - need to check all addresses and 
         * XXX sizes in dapl_mr_bounds_check
         */
        return ((DAT_VADDR)(uintptr_t)((struct iovec *)desc.for_pointer)->iov_base);
	}
    case DAT_MEM_TYPE_BYPASS:
	{
	    dapl_os_assert (0); /* DAT_NOT_IMPLEMENTED */
	    return 0;
	}
#else
	case DAT_MEM_TYPE_SHARED_VIRTUAL:
	{
	    /* multi-cast necessary to convert a pvoid to a DAT_VADDR on
	     * all architectures
	     */
	    return (DAT_VADDR) (uintptr_t) desc.for_shared_memory.virtual_address;
	}
#endif	/* defined(__KDAPL__) */
	default:
	{
	    dapl_os_assert (0);
	    return 0;
	}
    }
}

/*
 * Local variables:
 *  c-indent-level: 4
 *  c-basic-offset: 4
 *  c-brace-offset: -4
 *  tab-width: 8
 * End:
 */
