/*
 * Copyright (c) 2002-2003, Network Appliance, Inc. All rights reserved.
 * Copyright (c) 2005 Mellanox Technologies.   All rights reserved.

 *
 * This Software is licensed under both of the following two licenses:
 *
 * 1) under the terms of the "Common Public License 1.0" a copy of which is
 *    in the file LICENSE.txt in the root directory. The license is also
 *    available from the Open Source Initiative, see
 *    http://www.opensource.org/licenses/cpl.php.
 * OR
 *
 * 2) under the terms of the "The BSD License" a copy of which is in the file
 *    LICENSE2.txt in the root directory. The license is also available from
 *    the Open Source Initiative, see
 *    http://www.opensource.org/licenses/bsd-license.php.
 *
 * Licensee has the right to choose either one of the above two licenses.
 *
 * Redistributions of source code must retain both the above copyright
 * notice and either one of the license notices.
 *
 * Redistributions in binary form must reproduce both the above copyright
 * notice, either one of the license notices in the documentation
 * and/or other materials provided with the distribution.
 */

/**********************************************************************
 * 
 * MODULE: dapl_pz_create.c
 *
 * PURPOSE: Memory management
 * Description: Interfaces in this file are completely described in
 *		the DAPL 1.1 API, Chapter 6, section 6
 *
 * $Id: dapl_pz_create.c,v 1.2 2004/08/02 15:14:27 rabin Exp $
 **********************************************************************/

#include "dapl_pz_util.h"
#include "dapl_adapter_util.h"

/*
 * dapl_pz_create
 *
 * DAPL Requirements Version xxx, 6.6.2.1
 *
 * Create an instance of a protection zone
 *
 * Input:
 * 	ia_handle
 *
 * Output:
 * 	pz_handle
 *
 * Returns:
 * 	DAT_SUCCESS
 * 	DAT_INSUFFICIENT_RESOURCES
 * 	DAT_INVALID_PARAMETER
 */
DAT_RETURN
dapl_pz_create (
	IN	DAT_IA_HANDLE	   ia_handle,
	OUT	DAT_PZ_HANDLE 	   *pz_handle)
{
    DAPL_IA 		*ia;
    DAPL_PZ 		*pz;
    DAT_RETURN		dat_status;
#ifdef __KDAPL__
    mem_err_t       ret;
#endif

    dapl_dbg_log (DAPL_DBG_TYPE_API,
		  "dapl_pz_create (%p, %p)\n",
		  ia_handle, 
		  pz_handle);

    dat_status = DAT_SUCCESS;
    if ( DAPL_BAD_HANDLE (ia_handle, DAPL_MAGIC_IA) )
    {
	dat_status = DAT_ERROR (DAT_INVALID_HANDLE,DAT_INVALID_HANDLE_IA);
	goto bail;
    }

    if (NULL == pz_handle)
    {
	dat_status = DAT_ERROR (DAT_INVALID_PARAMETER,DAT_INVALID_ARG2);
	goto bail;
    }

    ia = (DAPL_IA *) ia_handle;

    pz = dapl_pz_alloc (ia);
    if ( pz == NULL )
    {
	dat_status = DAT_ERROR (DAT_INSUFFICIENT_RESOURCES,DAT_RESOURCE_MEMORY);
	goto bail;
    }

    dat_status = dapls_ib_pd_alloc (ia, pz);

    if ( dat_status != DAT_SUCCESS )
    {
	dapl_pz_dealloc (pz);
	pz = NULL;
    }

    *pz_handle = pz;

#ifdef __KDAPL__
    if (active_fmr)
    {
        ret = mem_fmr_init_module (pz);
        
        if  (ret != MEM_RET_SUCCESS)
        {
            dat_status = DAT_ERROR (DAT_INSUFFICIENT_RESOURCES,DAT_RESOURCE_MEMORY);
            goto bail;
        }
    }
#endif //__KDAPL__
 bail:
    return dat_status;
}
