/*
 * Copyright (c) 2002-2003, Network Appliance, Inc. All rights reserved.
 * Copyright (c) 2005 Mellanox Technologies.   All rights reserved.

 *
 * This Software is licensed under both of the following two licenses:
 *
 * 1) under the terms of the "Common Public License 1.0" a copy of which is
 *    in the file LICENSE.txt in the root directory. The license is also
 *    available from the Open Source Initiative, see
 *    http://www.opensource.org/licenses/cpl.php.
 * OR
 *
 * 2) under the terms of the "The BSD License" a copy of which is in the file
 *    LICENSE2.txt in the root directory. The license is also available from
 *    the Open Source Initiative, see
 *    http://www.opensource.org/licenses/bsd-license.php.
 *
 * Licensee has the right to choose either one of the above two licenses.
 *
 * Redistributions of source code must retain both the above copyright
 * notice and either one of the license notices.
 *
 * Redistributions in binary form must reproduce both the above copyright
 * notice, either one of the license notices in the documentation
 * and/or other materials provided with the distribution.
 */

/**********************************************************************
 * 
 * MODULE: dapl_pz_create.c
 *
 * PURPOSE: Memory management
 * Description: Interfaces in this file are completely described in
 *		the DAPL 1.1 API, Chapter 6, section 6
 *
 * $Id$
 **********************************************************************/

#include "dapl.h"
#include "dapl_pz_util.h"
#include "dapl_adapter_util.h"
#include "dapl_ia_util.h"

/*
 * dapl_pz_free
 *
 * DAPL Requirements Version xxx, 6.6.2.1
 *
 * Remove an instance of a protection zone
 *
 * Input:
 * 	pz_handle
 *
 * Output:
 * 	None.
 *
 * Returns:
 * 	DAT_SUCCESS
 * 	DAT_INVALID_PARAMETER
 * 	DAT_INVALID_STATE
 */
DAT_RETURN
dapl_pz_free (
	IN	DAT_PZ_HANDLE 	   pz_handle)
{
    DAPL_PZ		*pz;
    DAT_RETURN		dat_status;
#ifdef __KDAPL__ 
    mem_err_t       retcode;
#endif

    pz = (DAPL_PZ *) pz_handle;

    dapl_dbg_log (DAPL_DBG_TYPE_API, "dapl_pz_free (%p)\n", pz_handle);

#ifdef __KDAPL__    
    if (active_fmr)
    {
        retcode = mem_fmr_cleanup (&pz->mem_fmr_ctrl);
        if (retcode != MEM_RET_SUCCESS)
        {
            dapl_dbg_log (DAPL_DBG_TYPE_ERR, 
                          "%s: mem_fmr_cleanup returned error (0x%x)\n",
                          __FUNCTION__, retcode);
        }
        
    }
#endif //__KDPAL__

    dat_status = DAT_SUCCESS;
    if ( DAPL_BAD_HANDLE (pz_handle, DAPL_MAGIC_PZ) )
    {
	dat_status = DAT_ERROR (DAT_INVALID_HANDLE,DAT_INVALID_HANDLE_PZ);
	goto bail;
    }

    if ( 0 != dapl_os_atomic_read (&pz->pz_ref_count) )
    {
	dat_status = DAT_ERROR (DAT_INVALID_STATE,DAT_INVALID_STATE_PZ_IN_USE);
	goto bail;
    }

    dat_status = dapls_ib_pd_free (pz);

    if ( dat_status == DAT_SUCCESS )
    {
	dapl_pz_dealloc (pz);
    }

 bail:
    return dat_status;
}
