/*
 * Copyright (c) 2002-2003, Network Appliance, Inc. All rights reserved.
 * Copyright (c) 2005 Mellanox Technologies.   All rights reserved.

 *
 * This Software is licensed under both of the following two licenses:
 *
 * 1) under the terms of the "Common Public License 1.0" a copy of which is
 *    in the file LICENSE.txt in the root directory. The license is also
 *    available from the Open Source Initiative, see
 *    http://www.opensource.org/licenses/cpl.php.
 * OR
 *
 * 2) under the terms of the "The BSD License" a copy of which is in the file
 *    LICENSE2.txt in the root directory. The license is also available from
 *    the Open Source Initiative, see
 *    http://www.opensource.org/licenses/bsd-license.php.
 *
 * Licensee has the right to choose either one of the above two licenses.
 *
 * Redistributions of source code must retain both the above copyright
 * notice and either one of the license notices.
 *
 * Redistributions in binary form must reproduce both the above copyright
 * notice, either one of the license notices in the documentation
 * and/or other materials provided with the distribution.
 */

/**********************************************************************
 * 
 * MODULE: dapl_pz_query.c
 *
 * PURPOSE: Memory management
 * Description: Interfaces in this file are completely described in
 *		the DAPL 1.1 API, Chapter 6, section 6
 *
 * $Id: dapl_pz_query.c,v 1.2 2004/08/02 15:14:27 rabin Exp $
 **********************************************************************/

#include "dapl.h"

/*
 * dapl_pz_query
 *
 * DAPL Requirements Version xxx, 6.6.2.1
 *
 * Return the ia associated with the protection zone pz
 *
 * Input:
 * 	pz_handle
 *      pz_param_mask
 *
 * Output:
 * 	pz_param
 *
 * Returns:
 * 	DAT_SUCCESS
 *      DAT_INVALID_HANDLE
 * 	DAT_INVALID_PARAMETER
 */
DAT_RETURN
dapl_pz_query (
	IN	DAT_PZ_HANDLE		pz_handle,
	IN	DAT_PZ_PARAM_MASK	pz_param_mask,
	OUT	DAT_PZ_PARAM		*pz_param)
{
    DAPL_PZ		*pz;
    DAT_RETURN		dat_status;

    dapl_dbg_log (DAPL_DBG_TYPE_API,
		  "dapl_pz_query (%p, %x, %p)\n", 
		  pz_handle, 
		  pz_param_mask, 
		  pz_param);

    dat_status = DAT_SUCCESS;

    if ( DAPL_BAD_HANDLE (pz_handle, DAPL_MAGIC_PZ) )
    {
	dat_status = DAT_ERROR (DAT_INVALID_HANDLE,DAT_INVALID_HANDLE_PZ);
	goto bail;
    }
    if (NULL == pz_param)
    {
	dat_status = DAT_ERROR (DAT_INVALID_PARAMETER,DAT_INVALID_ARG3);
	goto bail;
    }

    pz = (DAPL_PZ *) pz_handle;

    /* Since the DAT_PZ_ARGS values are easily accessible, */
    /* don't bother checking the DAT_PZ_ARGS_MASK value    */
    pz_param->ia_handle = (DAT_IA_HANDLE) pz->header.owner_ia;

 bail:
    return dat_status;
}
