/*
 * Copyright (c) 2002-2003, Network Appliance, Inc. All rights reserved.
 * Copyright (c) 2005 Mellanox Technologies.   All rights reserved.

 *
 * This Software is licensed under both of the following two licenses:
 *
 * 1) under the terms of the "Common Public License 1.0" a copy of which is
 *    in the file LICENSE.txt in the root directory. The license is also
 *    available from the Open Source Initiative, see
 *    http://www.opensource.org/licenses/cpl.php.
 * OR
 *
 * 2) under the terms of the "The BSD License" a copy of which is in the file
 *    LICENSE2.txt in the root directory. The license is also available from
 *    the Open Source Initiative, see
 *    http://www.opensource.org/licenses/bsd-license.php.
 *
 * Licensee has the right to choose either one of the above two licenses.
 *
 * Redistributions of source code must retain both the above copyright
 * notice and either one of the license notices.
 *
 * Redistributions in binary form must reproduce both the above copyright
 * notice, either one of the license notices in the documentation
 * and/or other materials provided with the distribution.
 */

/**********************************************************************
 * 
 * HEADER: dapl_dummy_util.h
 *
 * PURPOSE: 
 *	This files is useful for porting and understanding the provider
 *	interface. DAPL is designed to accomodate multiple transport
 *	providers and currently supports InfiniBand, Myrinet, iWarp,
 *	MPI, and others. This file contains the essential data types and
 *	#defines needed for a port.
 *
 *	This file is an example file only. The typdefs below are all
 *	mapped to DAT_UINT32 as an example. Most transports have complex
 *	data types and structures for some of these values, the
 *	implementer must be aware of the requirements of their
 *	transport.
 *
 * $Id: dapl_dummy_util.h,v 1.1 2004/08/03 06:01:01 rabin Exp $
 *
 **********************************************************************/
#ifndef _DAPL_DUMMY_UTIL_H_
#define _DAPL_DUMMY_UTIL_H_

/*
 * ib_cm_handle_t is used in callbacks and in communication directly
 * from and to the provider. It is a Connection Manager handle. Opaque
 * to the common layer.
 */
typedef DAT_UINT32	ib_cm_handle_t;

/*
 * ib_cm_events_t is used in callbacks and in communication directly
 * from and to the provider. Primarily deals with events for completions
 * or notifications.
 */
typedef DAT_UINT32	ib_cm_events_t;

/*
 * ib_hca_handle_t is used in callbacks as a parameter, and by the 
 * provider layer. Most providers API's specify this parameter.
 */
typedef DAT_UINT32	ib_hca_handle_t;

/*
 * Provider defined structure to maintain specific data types and transport
 * parameters. Defined in <transport>_util.h, and included in dapl.h in
 * the DAPL_HCA structure.
 */
typedef DAT_UINT32	ib_hca_tranport_t;

/*
 * Provider defined structure to maintain specific data types to support
 * shared memory. The reference implementation providers both require a
 * memory region exist before it can be made shared, so we store the
 * backing memory region handle here. This is not be neccesary if a
 * more direct mr_register_shared() exists.
 *
 */
typedef DAT_UINT32	ib_shm_transport_t;

/*
 * ib_cq_handle_t is used in callbacks and in communication directly
 * from and to the provider. It is the providers handle for a specific
 * CQ.
 */
typedef DAT_UINT32	ib_cq_handle_t;

/*
 * ib_comp_handle_t is required for the VAPI provider, it is unused on
 * all other providers.
 */
typedef DAT_UINT32	ib_comp_handle_t;

/*
 * ib_qp_handle_t is used in callbacks and in communication directly
 * from and to the provider. It is the providers handle for a specific
 * QP.
 */
typedef DAT_UINT32	ib_qp_handle_t;

/*
 * ib_qp_state_t is used to keep track of the current state of the QP.
 * It is primarily used by the provider layer, with one exception.
 * DAPL_QP_STATE_UNATTACHED is a DAPL state that indicates no QP has 
 * been assigned to an EP. DAPL will lazy attach the QP when it lacks
 * sufficient information, waiting until all of the required parameters
 * are in place. The value of DAPL_QP_STATE_UNATTACHED must not conflict
 * with valid QP state values.
 */
typedef DAT_UINT32	ib_qp_state_t;

/*
 * ib_data_segment_t is the provider definition of the scatter/gather
 * list data structure use for posting DTOs. It typically includes
 * a length, a key/handle, and a virtual address. It is specific to
 * the provider implementation, and put as a generic type here.
 */
typedef DAT_UINT32	ib_data_segment_t;

/*
 * ib_pd_handle_t is the provider handle to a Protection Domain, 
 * supporting the DAPL Protection Zone. It is maintained in the DAPL_PZ
 * structure, but it is used exclusively by the provider layer.
 */
typedef DAT_UINT32	ib_pd_handle_t;

/*
 * ib_mr_handle_t is the provider handle to a Memory Region,
 * supporting the DAPL LMR. It is maintained in the DAPL_LMR
 * structure, but it is used exclusively by the provider layer.
 */
typedef DAT_UINT32	ib_mr_handle_t;


/*
 * ib_mw_handle_t is the provider handle to a Memory Window,
 * supporting the DAPL RMR. It is maintained in the DAPL_RMR
 * structure, but it is used exclusively by the provider layer.
 */
typedef DAT_UINT32	ib_mw_handle_t;

/*
 * ib_cm_srvc_handle_t is required for the VAPI provider, it is unused
 * on all other providers. Needed by the CM.
 */
typedef DAT_UINT32	ib_cm_srvc_handle_t;

/*
 * ib_bool_t is a boolean typedef. Every provider seems to have a 
 * different notion of what a boolean value should be, so we provide
 * a generic type to allow clean compilation.
 */
typedef DAT_UINT32	ib_bool_t;

/*
 * ib_work_completion_t is a completion record for the provider.
 * Common code must define one of these and then ask the provider to
 * fill it in. Macros obtain the necessary fields from the completion
 * structure, which is translated into DAT completions codes.
 */
typedef DAT_UINT32	ib_work_completion_t;

/*
 * ib_notification_type_t is only used as a parameter to pass to the
 * provider layer in dapls_ib_completion_notify(). The two CQ
 * NOTIFICATION TYPE values defined below are the only two values. These
 * values direct the provider CQ behavior.
 */
typedef DAT_UINT32	ib_notification_type_t;

/*
 * ib_error_record_t is the provider notion of the error causing a callback.
 * It is opaque to the common layer, translated by the provider layer with
 * dapls_ib_get_async_event().
 */
typedef DAT_UINT32	ib_error_record_t;

/*
 * ib_async_handler_t is the typedef prototype for asynchronous handler
 * callback functions. Handlers exist in the common code, and pointers
 * are passed to the provider to set up the error mechanism.
 */
typedef void (*ib_async_handler_t)(
	DAT_UINT32	hca_hndl,
        DAT_UINT32 	*event_record_p,
        void		*private_data );


/*
 * CQ NOTIFICATION TYPE. Used when configuring a CQ, must be values defined
 * by the provider.
 */
#define IB_NOTIFY_ON_NEXT_COMP    1
#define IB_NOTIFY_ON_SOLIC_COMP   2

/* Generic INVALID_HANDLE value. Some providers may use 0, but all 0xF's
 * is usually safe
 */
#define IB_INVALID_HANDLE         (~0)

/*
 * provider defined DTO op values. Note that some providers are capable
 * of a richer set of operations, but DAPL only uses these.
 */
#define	OP_RDMA_READ	          0
#define	OP_RDMA_WRITE	          1
#define OP_SEND		          2
#define OP_RECEIVE	          3
#define OP_BIND_MW                4


#endif /* _DAPL_DUMMY_UTIL_H_ */
