/*
 * Copyright (c) 2002-2003, Network Appliance, Inc. All rights reserved.
 * Copyright (c) 2005 Mellanox Technologies.   All rights reserved.

 *
 * This Software is licensed under both of the following two licenses:
 *
 * 1) under the terms of the "Common Public License 1.0" a copy of which is
 *    in the file LICENSE.txt in the root directory. The license is also
 *    available from the Open Source Initiative, see
 *    http://www.opensource.org/licenses/cpl.php.
 * OR
 *
 * 2) under the terms of the "The BSD License" a copy of which is in the file
 *    LICENSE2.txt in the root directory. The license is also available from
 *    the Open Source Initiative, see
 *    http://www.opensource.org/licenses/bsd-license.php.
 *
 * Licensee has the right to choose either one of the above two licenses.
 *
 * Redistributions of source code must retain both the above copyright
 * notice and either one of the license notices.
 *
 * Redistributions in binary form must reproduce both the above copyright
 * notice, either one of the license notices in the documentation
 * and/or other materials provided with the distribution.
 */

/**********************************************************************
 *
 * MODULE: dapl_evd_modify_upcall.c
 *
 * PURPOSE: EVENT management
 *
 * Description: Interfaces in this file are completely defined in
 *              the kDAPL 1.1 API, Chapter 6, section 3.2.4
 *
 * $Id: dapl_evd_modify_upcall.c,v 1.2 2004/08/02 15:14:28 rabin Exp $
 **********************************************************************/

#include "dapl.h"

/*
 * dapl_evd_modify_upcall
 *
 * kDAPL Requirements Version 1.1, 6.3.2.4
 *
 * Modify the upcall instance on an EVD
 *
 * Input:
 *	evd_handle
 *	upcall_policy
 *	upcall
 *
 * Output:
 * 	none
 *
 * Returns:
 * 	DAT_SUCCESS
 * 	DAT_INVALID_HANDLE
 * 	DAT_INVALID_PARAMETER
 */

DAT_RETURN dapl_evd_modify_upcall (
	IN    DAT_EVD_HANDLE		evd_handle,
	IN    DAT_UPCALL_POLICY		upcall_policy,	/* ignore */
	IN    const DAT_UPCALL_OBJECT	*upcall)
{
    DAPL_EVD    *evd_ptr;
    DAPL_CNO	*cno_ptr;
    DAT_RETURN  dat_status;

    dapl_dbg_log (DAPL_DBG_TYPE_API, "dapl_modify_upcall (%p)\n", evd_handle);

    dat_status = DAT_SUCCESS;

    evd_ptr = (DAPL_EVD *)evd_handle;
    cno_ptr = evd_ptr->cno_ptr;

    if (DAPL_BAD_HANDLE (evd_handle, DAPL_MAGIC_EVD))
    {
	dat_status = DAT_ERROR (DAT_INVALID_HANDLE, 0);
	goto bail;
    }

    if (cno_ptr != NULL)
    {
	cno_ptr->cno_upcall_policy = upcall_policy;
	cno_ptr->cno_upcall        = *upcall;
    }

bail:
    return dat_status;
}


/*
 * Local variables:
 *  c-indent-level: 4
 *  c-basic-offset: 4
 *  tab-width: 8
 * End:
 */
