/*
 * Copyright (c) 2002-2003, Network Appliance, Inc. All rights reserved.
 * Copyright (c) 2005 Mellanox Technologies.   All rights reserved.

 *
 * This Software is licensed under both of the following two licenses:
 *
 * 1) under the terms of the "Common Public License 1.0" a copy of which is
 *    in the file LICENSE.txt in the root directory. The license is also
 *    available from the Open Source Initiative, see
 *    http://www.opensource.org/licenses/cpl.php.
 * OR
 *
 * 2) under the terms of the "The BSD License" a copy of which is in the file
 *    LICENSE2.txt in the root directory. The license is also available from
 *    the Open Source Initiative, see
 *    http://www.opensource.org/licenses/bsd-license.php.
 *
 * Licensee has the right to choose either one of the above two licenses.
 *
 * Redistributions of source code must retain both the above copyright
 * notice and either one of the license notices.
 *
 * Redistributions in binary form must reproduce both the above copyright
 * notice, either one of the license notices in the documentation
 * and/or other materials provided with the distribution.
 */

/**********************************************************************
 *
 * MODULE: dapl_ia_memtype_hint.c
 *
 * PURPOSE: EVENT management
 *
 * Description: Interfaces in this file are completely defined in
 *              the kDAPL 1.1 API, Chapter 6, section 6.1.1
 *
 * $Id: dapl_ia_memtype_hint.c,v 1.2 2004/08/02 15:14:28 rabin Exp $
 **********************************************************************/

#include "dapl.h"

/*
 * dapl_ia_memtype_hint
 *
 * kDAPL Requirements Version 1.1, 6.6.1.1
 *
 * Provide a hint about optimal memory alignment and size
 *
 * Input:
 *	ia_handle,
 *	mem_type,
 *	length,
 *	mem_optimization,
 *	*suggested_length,
 *	*suggested_alignment)
 *
 * Output:
 * 	none
 *
 * Returns:
 * 	DAT_SUCCESS
 * 	DAT_INVALID_PARAMETER
 * 	DAT_INVALID_HANDLE
 * 	DAT_MODEL_NOT_SUPPORTED
 */
DAT_RETURN
dapl_ia_memtype_hint (
	IN    DAT_IA_HANDLE		ia_handle,
	IN    DAT_MEM_TYPE		mem_type,
	IN    DAT_VLEN			length,
	IN    DAT_MEM_OPT		mem_optimization,
	OUT   DAT_VLEN			*suggested_length,
	OUT   DAT_VADDR			*suggested_alignment)
{
    DAPL_IA		*ia_ptr;
    DAT_RETURN		dat_status;

    dapl_dbg_log (DAPL_DBG_TYPE_API, "dapl_ia_memtype_hint (%p)\n", ia_handle);

    ia_ptr = (DAPL_IA *)ia_handle;
    dat_status = DAT_SUCCESS;

    if (DAPL_BAD_HANDLE (ia_ptr, DAPL_MAGIC_IA))
    {
	dat_status = DAT_ERROR (DAT_INVALID_HANDLE, DAT_INVALID_HANDLE_IA);
	goto bail;
    }

    if ( suggested_length == NULL)
    {
	dat_status = DAT_ERROR (DAT_INVALID_PARAMETER, DAT_INVALID_ARG5);
	goto bail;
    }

    if ( suggested_alignment == NULL)
    {
	dat_status = DAT_ERROR (DAT_INVALID_PARAMETER, DAT_INVALID_ARG6);
	goto bail;
    }
    dat_status = DAT_ERROR (DAT_NOT_IMPLEMENTED, 0);

bail:
    return dat_status;
}


/*
 * Local variables:
 *  c-indent-level: 4
 *  c-basic-offset: 4
 *  tab-width: 8
 * End:
 */
