/*
 * Copyright (c) 2002, Network Appliance, Inc. All rights reserved.
 * Copyright (c) 2005 Mellanox Technologies.   All rights reserved.

 *
 * This Software is licensed under the terms of the "IBM Common Public
 * License 0.5" a copy of which is in the file LICENSE.txt in the
 * root directory. The license is also available from the Open Source
 * Initiative, see  http://www.opensource.org/licenses/ibmpl.html.
 *
 */

/**********************************************************************
 *
 * MODULE: dapl_module.c
 *
 * PURPOSE: DAPL module implementation, exists in the Linux
 * kernel
 * Description: a linux module implementation
 *
 **********************************************************************/

#define EXPORT_SYMTAB
#include "dapl_osd.h"
#include "dapl.h"
#include "dapl_init.h"
#include <linux/errno.h>
#include <linux/fs.h>
#include <linux/version.h>
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
#include <linux/moduleparam.h>
#endif
MODULE_LICENSE("Dual BSD/GPL");
MODULE_DESCRIPTION("kDAPL transport for Linux kernel REVISION 1.8.0");
MODULE_AUTHOR("Fujitsu PST, Network Appliance, Inc.");

int             DbgLvl = 0;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
module_param(DbgLvl, int, 0444);
#else
MODULE_PARM ( DbgLvl, "i" );
#endif
MODULE_PARM_DESC(DbgLvl, "Bit mask to enable dapl debugging, equivelent to DAPL_DBG_TYPE");


/***********************************************************************
 * init_module
 *
 * Entry point for a Linux module, performs simple initialization
 *
 * Input:
 *	none
 *
 * Output:
 *	none
 *
 * Returns:
 *	SUCCESS
 ***********************************************************************/

int init_module (void)
{
    dapl_os_printf ("<1>kDAPL: Started\n");
    dapl_init ();
    return (0);		/* success */
}

/***********************************************************************
 * cleanup_module
 *
 * Entry point for a Linux module, cleans up the module on exit
 *
 * Input:
 *	none
 *
 * Output:
 *	none
 *
 * Returns:
 *	void
 ***********************************************************************/
void cleanup_module (void)
{
    dapl_fini ();
    dapl_os_printf ("<1>kDAPL: Stopped\n");
    return;
}


/*
 * Local variables:
 *  c-indent-level: 4
 *  c-basic-offset: 4
 *  tab-width: 8
 * End:
 */
