/*
 * Copyright (C) Voltaire Ltd. 2001-2005.  ALL RIGHTS RESERVED.
 * Copyright (c) 2005 Mellanox Technologies.   All rights reserved.

 *
 * Maintained by: vtrmaint1@voltaire.com
 *
 * This Software is licensed under both of the following two licenses:
 *
 * 1) under the terms of the "Common Public License 1.0" a copy of which is
 *    in the file LICENSE.txt in the root directory. The license is also
 *    available from the Open Source Initiative, see
 *    http://www.opensource.org/licenses/cpl.php.
 * OR
 *
 * 2) under the terms of the "The BSD License" a copy of which is in the file
 *    LICENSE2.txt in the root directory. The license is also available from
 *    the Open Source Initiative, see
 *    http://www.opensource.org/licenses/bsd-license.php.
 *
 * Licensee has the right to choose either one of the above two licenses.
 *
 * Redistributions of source code must retain both the above copyright
 * notice and either one of the license notices.
 *
 * Redistributions in binary form must reproduce both the above copyright
 * notice, either one of the license notices in the documentation
 * and/or other materials provided with the distribution.
 */
/**********************************************************************
 * MODULE: dapl_mem_fmr.h                                 
 *
 * PURPOSE: defines common data structures and API's for the dapl_mem_fmr.c
 * Description: 
 *
 * $Id: dapl_mem_fmr.h,v 1.1.2.12 2005/02/15 14:49:11 guyg Exp $
 **********************************************************************/

#ifndef _MEM_FMR__
#define _MEM_FMR__

#include <linux/list.h>
#include <linux/spinlock.h>
#include "dapl_debug.h"

# undef IN
# undef OUT
#include "vapi_types.h"
#include "vapi.h"
#include "evapi.h"
# undef IN
# undef OUT
# define IN
# define OUT

typedef void *mem_fmr_hndl_t;

extern int active_fmr;

#define FMR_DEFUALT_MAX_OUTSTANDING_MAPS  100
#define FMR_DEFUALT_NUMBER_OF_HANDLES     1000
#define FMR_DEFUALT_MAX_UNMAP_AT_ONCE     2000    // this default value is set according to mellanox
#define FMR_DEFUALT_WATER_MARK            10      // the default is 10% of total allocation

#define FMR_MAX_OPTIONS_TO_MAP            65

#define FMR_UNMAP_URGENT_FLAG_OFF 0x0
#define FMR_UNMAP_URGENT_FLAG_ON  0x1

#define ACCESS_CONTROL_MASK_CLR_ALL(mask)   ((mask)=0)
#define ACCESS_CONTROL_MASK_SET(mask,attr)  ((mask)|=(attr))
#define ACCESS_CONTROL_MASK_CLR(mask,attr)  ((mask)&=(~(attr)))
#define ACCESS_CONTROL_IS_SET(mask,attr)    (((mask)&(attr))!=0)
#define ACCESS_CONTROL_MASK_SET_ALL(mask)   ((mask)|= \
                                               VAPI_EN_LOCAL_WRITE  \
                                             | VAPI_EN_REMOTE_WRITE \
                                             | VAPI_EN_REMOTE_READ  \
                                             | VAPI_EN_REMOTE_ATOM  \
                                             | VAPI_EN_MEMREG_BIND)


//FMR mapping flags for virtual address
#define  FMR_VA_AUTO_GEN  0x1   // the function generates virtual address 
#define  FMR_VA_FROM_USER 0x2   // the user supplied his own virtual addres


/**************************************
 * BEGIN of internal FMR structures   *
 **************************************/

typedef struct
{
    VAPI_hca_hndl_t     hca_hdnl;
	EVAPI_fmr_hndl_t    fmr_hndl;
	int                 mapped_pages;
    int                 mapped_pages_req;
	u_int64_t           page_arr[FMR_MAX_OPTIONS_TO_MAP];
	int                 map_count;	                        // counting the map operation on the fmr handl
    VAPI_size_t         data_size;
	struct list_head    list;
} fmr_hndl_impl;

typedef struct
{
	int                 allocted_pages;
	int                 total_hndls;           // total handles allocated
	int                 num_free_hndls;        // total free handles
	int                 min_free_hndl_needed;  // minimum number of handles needed (watermark)
	struct list_head    free_fmr_hndl_list;
	struct list_head    used_fmr_hndl_list;
	spinlock_t          lock;                  // lock for the used fmr
}fmr_pool_entry; 


typedef struct
{
	fmr_hndl_impl       *fmr_hndl[FMR_DEFUALT_MAX_UNMAP_AT_ONCE];
	EVAPI_fmr_hndl_t    v_hndl[FMR_DEFUALT_MAX_UNMAP_AT_ONCE];
	int                 idx;
	spinlock_t          hndl_vec_lock;      // lock for the attributes above.
	int                 max_unmap_at_once;  // maximum number of handles that can be mapped at one time
	int                 urgent_flag;        // if the flag is on, perform unmap as soon as you can.
                                            // (even if you didn't reach MAX_UNMAP_AT_ONCE) 
	spinlock_t          urgent_flag_lock;   // diffrent lock, for efficiency.

}need_to_unmap;

typedef struct
{
	fmr_pool_entry     fmr_pool[FMR_MAX_OPTIONS_TO_MAP];
	need_to_unmap      unmap_vector;
	int                max_fmr_num;             //number of handles available is limited.
	int                current_fmr_num;
	int                max_outstanding_maps;    //max maps before unmap is required.
}fmr_ctrl_struct;


/* error codes */
enum {
    MEM_RET_SUCCESS = 0,
    MEM_RET_UNKNOWN_ERR ,
    MEM_RET_INVALID_ARG,
    MEM_RET_FMR_OUT_OF_RESOURCE,
    MEM_RET_FMR_LOWER_LEVEL_ERROR
};
// mem_err;

/* FMR handler*/
typedef u_int32_t mem_err_t;

mem_err_t  
mem_fmr_get_fmr(IN   int                 num_pages,
                IN   fmr_ctrl_struct     *mem_fmr_ctrl,
				OUT  mem_fmr_hndl_t      *mem_fmr_hndl);

mem_err_t
mem_fmr_free_fmr (IN  int                num_to_free,
				  IN  mem_fmr_hndl_t     fmr_hndl_arr,
                  IN  fmr_ctrl_struct    *mem_fmr_ctrl,
				  OUT int                *num_success_free);

mem_err_t
mem_fmr_map_fmr(IN  mem_fmr_hndl_t      mem_fmr_hndl,
                OUT VAPI_lkey_t         *l_key,
				OUT VAPI_rkey_t         *r_key);

// initalization and exit functions
mem_err_t
mem_fmr_init_module (IN void  *pz_h);

mem_err_t
mem_fmr_cleanup (IN   fmr_ctrl_struct     *mem_fmr_ctrl);

// proc support functions
void mem_fmr_print_all(fmr_ctrl_struct    *mem_fmr_ctrl);

void fmr_set_watermark (int                 mapped_pages, 
                        int                 new_watermark,
                        fmr_ctrl_struct     *mem_fmr_ctrl);

void fmr_set_unmap_at_once (int                 new_unmap_value,
                            fmr_ctrl_struct     *mem_fmr_ctrl);

void mem_fmr_show_param (fmr_ctrl_struct    *mem_fmr_ctrl);

extern void mem_show_fmr_pool(char             *func,
                       fmr_ctrl_struct  *mem_fmr_ctrl);
#if 0
int show_proc_fmr_pool(char *page, 
                       fmr_ctrl_struct *mem_fmr_ctrl);
#endif
#endif /*_MEM_FMR__*/
