/*
 * This Software is licensed under both of the following two licenses:
 *
 * 1) under the terms of the "Common Public License 1.0" a copy of which is
 *    in the file LICENSE.txt in the root directory. The license is also
 *    available from the Open Source Initiative, see
 *    http://www.opensource.org/licenses/cpl.php.
 * OR
 *
 * 2) under the terms of the "The BSD License" a copy of which is in the file
 *    LICENSE2.txt in the root directory. The license is also available from
 *    the Open Source Initiative, see
 *    http://www.opensource.org/licenses/bsd-license.php.
 *
 * Licensee has the right to choose either one of the above two licenses.
 *
 * Redistributions of source code must retain both the above copyright
 * notice and either one of the license notices.
 *
 * Redistributions in binary form must reproduce both the above copyright
 * notice, either one of the license notices in the documentation
 * and/or other materials provided with the distribution.
 */

/**************************************************************************
 *
 *   Module:		 uDAPL
 *
 *   Filename:		 dapl_topspin.h
 *
 *   Author:		 Ariel Cohen
 *
 *   Created:		 10/24/2002
 *
 *   Description:
 *
 *   Header file for the Topspin uDAPL enhancements
 *
 ***************************************************************************
 *
 *		   Source Control System Information
 *
 *    $Id: dapl_openib.h,v 1.1 2004/09/16 13:37:22 sjs2 Exp $
 *
 ***************************************************************************
 *
 ***************************************************************************
 *			 (c) COPYRIGHT 2002
 *			     Topspin Communications, Inc.
 *			     Mountain View, CA
 **************************************************************************/

#ifndef _DAPL_TOPSPIN_H
#define _DAPL_TOPSPIN_H

#undef IN
#undef OUT

#include "dapl_openib_cm.h"

#include "vapi_types.h"
#define IN
#define OUT
#include "ib_legacy_types.h"
#include "ts_ib_core_types.h"
#ifndef __KDAPL__
#include "dapl_srv.h"
#endif
#include "dat/dat.h"


struct ia_prov_data {
    VAPI_hca_hndl_t            ib_khca_handle;
    void                       *ts_hca_handle;
    void                       *global_pd;
    IB_guid_t  node_GUID;

    IB_gid_t   gid;
    IB_lid_t   lid;
    
};


/*
 * This function are for both udapl & kdapl
 * udapl implementation is in dapl_openib.c
 * kdapl implementation is in dapl_srv.c (dapl_srv module in openib driver)
 */
int ib_path_record_lookup(DAT_IA_ADDRESS_PTR      remote_ia_address,
                          tTS_IB_GID              src_gid,
                          struct ib_device       *hca_handle,
                          tTS_IB_PORT             port,
                          tUINT32                 ats_flag,
                          struct ib_path_record  *path_record);
int ib_ats_ipaddr_lookup(uint8_t *gid, struct ib_device * hca_handle,
			 tTS_IB_PORT port, uint8_t *ip_addr);
int ib_ats_set_ipaddr(uint8_t *gid, struct ib_device *ts_hca_handle,
		      tTS_IB_PORT port, uint8_t *ip_addr);
int ib_get_hca_ipaddr (uint8_t *gid, tUINT32 *ip_addr);
int ib_get_khca_hndl(VAPI_hca_id_t hca_id, VAPI_hca_hndl_t *k_hca_handle, void **ts_hca_handle);
int ib_release_khca_hndl(VAPI_hca_hndl_t k_hca_handle);

#ifndef __KDAPL__
int ts_init(void);
int ts_cleanup(void);
int ts_smr_insert(DAT_LMR_COOKIE cookie, tUINT32 *exists);
int ts_smr_add_mrh(DAT_LMR_COOKIE cookie, VAPI_mr_hndl_t mr_handle);
int ts_smr_del_mrh(DAT_LMR_COOKIE cookie, VAPI_mr_hndl_t mr_handle);
int ts_smr_query(DAT_LMR_COOKIE cookie, tUINT32 *ready, VAPI_mr_hndl_t *mr_handle);
int ts_smr_dec(DAT_LMR_COOKIE cookie);
int ts_smr_mutex_lock(void);
int ts_smr_mutex_unlock(void);
#endif


#endif /* _DAPL_TOPSPIN_H */

