/*
 * This Software is licensed under both of the following two licenses:
 *
 * 1) under the terms of the "Common Public License 1.0" a copy of which is
 *    in the file LICENSE.txt in the root directory. The license is also
 *    available from the Open Source Initiative, see
 *    http://www.opensource.org/licenses/cpl.php.
 * OR
 *
 * 2) under the terms of the "The BSD License" a copy of which is in the file
 *    LICENSE2.txt in the root directory. The license is also available from
 *    the Open Source Initiative, see
 *    http://www.opensource.org/licenses/bsd-license.php.
 *
 * Licensee has the right to choose either one of the above two licenses.
 *
 * Redistributions of source code must retain both the above copyright
 * notice and either one of the license notices.
 *
 * Redistributions in binary form must reproduce both the above copyright
 * notice, either one of the license notices in the documentation
 * and/or other materials provided with the distribution.
 */

/***************************************************************************
 *
 *   Module:		 uDAPL
 *
 *   Filename:		 dapl_topspin_cm.c
 *
 *   Author:		 Ariel Cohen
 *
 *   Created:		 03/06/2003
 *
 *   Description:
 *
 *   This is the CM support for uDAPL.
 *
 ****************************************************************************
 *		   Source Control System Information
 *
 *    $Id: dapl_openib_cm.c,v 1.3 2004/09/24 16:37:17 sjs2 Exp $
 *
 ****************************************************************************
 *
 ***************************************************************************
 *			 (c) COPYRIGHT 2003
 *			     Topspin Communications, Inc.
 *			     Mountain View, CA
 **************************************************************************/

#ifdef RCSID
static const char rcsid[] = "$Id$";
#endif


#ifdef __KERNEL__
#include "linux/types.h"
#endif

#include "dapl_openib_cm.h"
#include "dapl.h"           
#include "dapl_adapter_util.h"
#include "dapl_evd_util.h"
#include "dapl_sp_util.h"
#include "dapl_cr_util.h"
#include "dapl_openib.h"



#ifdef IBHOSTS_NAMING
#include "dapl_name_service.h"
#endif

/* FIXME: These should be queried */
#define IB_TARGET_MAX      4  /* responder resources (max_qp_ous_rd_atom) */
#define IB_INITIATOR_DEPTH 4  /* initiator depth (max_qp_init_rd_atom) */



static int do_active_idle(DAPL_EP *ep_ptr, struct ib_u_cm_idle_param *idle_param)
{
    ib_cm_events_t         ib_cm_event;
    //syslog(LOG_NOTICE, "Active went to idle on comm_id = %d\n", ep_ptr->cm_handle);

    if (ep_ptr->param.connect_evd_handle)
    {
	switch (idle_param->reason) 
        {
	case TS_IB_CM_IDLE_REMOTE_REJECT:
	    //syslog(LOG_NOTICE, "do_active_idle: TS_IB_CM_IDLE_REMOTE_REJECT for cm_handle=%x\n", ep_ptr->cm_handle);
	    if (idle_param->rej_info_len && idle_param->rej_info) 
	    {
            ib_cm_event = IB_CME_DESTINATION_REJECT_PRIVATE_DATA;
            dapl_evd_connection_callback (ep_ptr->cm_handle,
                                          ib_cm_event,
                                          idle_param->rej_info,
                                          (void *)ep_ptr );
            return 0;
	    } else

		ib_cm_event = IB_CME_DESTINATION_REJECT;

	    break;

	case TS_IB_CM_IDLE_REMOTE_TIMEOUT:
	    //syslog(LOG_NOTICE, "do_active_idle: TS_IB_CM_IDLE_REMOTE_TIMEOUT for cm_handle=%x\n", ep_ptr->cm_handle);
	    ib_cm_event = IB_CME_DESTINATION_UNREACHABLE;
	    break;

        case TS_IB_CM_IDLE_TIME_WAIT_DONE:
            return(0);
	    break;
	    
	case TS_IB_CM_IDLE_LOCAL_REJECT:
	default:
	    //syslog(LOG_NOTICE, "do_active_idle: TS_IB_CM_IDLE_LOCAL_REJECT or other for cm_handle=%x\n", ep_ptr->cm_handle);
	    ib_cm_event = IB_CME_BROKEN;
	    break;

	}

	dapl_evd_connection_callback (ep_ptr->cm_handle,
				      ib_cm_event,
				      NULL,
				      (void *)ep_ptr );
    }

    return 0;
}


static int do_active_rep_recv(DAPL_EP                           *ep_ptr, 
			      tTS_IB_CM_COMM_ID                  comm_id,
			      struct ib_u_cm_rep_received_param *rep_recv_param)
{
    int status;

    //syslog(LOG_NOTICE, "REP received on comm_id = %d\n", ep_ptr->cm_handle);


    // ETBD - Is this correct - First confirm and later callback
    status = ib_cm_vapi_confirm (comm_id, NULL, 0);
    // in user there is a sync function i dont think we need it (we work with the right handle)

    if (status)
        dapl_dbg_log (DAPL_DBG_TYPE_ERR, " tsIbUCmConfirm failed: %d\n", 
                      status);



    if (ep_ptr->param.connect_evd_handle) 

	dapl_evd_connection_callback (ep_ptr->cm_handle,
				      IB_CME_CONNECTED,
				      rep_recv_param->remote_private_data,
				      (void *) ep_ptr );

       return 0;
}


static int do_active_established(DAPL_EP *ep_ptr)
{

    //syslog(LOG_NOTICE, "Active established on comm_id = %d\n", ep_ptr->cm_handle);

//  ETBD - This is already called when rep is received.
//
//    if (ep_ptr->param.connect_evd_handle)
//	        dapl_evd_connection_callback (ep_ptr->cm_handle,
//		                    		      IB_CME_CONNECTED,
//				                          NULL,
//				                          (void *) ep_ptr );

    return 0;
}

int dapls_cm_active_cb_handler(tTS_IB_U_QP_HANDLE   consumer_qp,
			       tTS_IB_CM_COMM_ID    comm_id,
			       tTS_IB_CM_EVENT      event,
			       void                *params,
			       void                *arg);


int dapls_cm_active_cb_handler(tTS_IB_U_QP_HANDLE   consumer_qp,
			       tTS_IB_CM_COMM_ID    comm_id,
			       tTS_IB_CM_EVENT      event,
			       void                *params,
			       void                *arg)
{
    DAPL_EP *ep_ptr;
    int     ret = 0;
    ep_ptr = (DAPL_EP *) arg;

    if (ep_ptr == NULL) 
    {
	dapl_dbg_log (DAPL_DBG_TYPE_ERR, " Called with a NULL context\n");
	return 0;
    }

    switch (event)
    {
    case TS_IB_CM_IDLE:
	
	ret = do_active_idle(ep_ptr, (struct ib_u_cm_idle_param *) params);
	break;
	
    case TS_IB_CM_REP_RECEIVED:

	ret = do_active_rep_recv(ep_ptr, comm_id,
				 (struct ib_u_cm_rep_received_param *) params);
	break;

    case TS_IB_CM_ESTABLISHED:

	ret = do_active_established(ep_ptr);
	break;

    case TS_IB_CM_DISCONNECTED:
        dapl_evd_connection_callback (comm_id,
				      IB_CME_DISCONNECTED,
				      NULL,
				      (void *) ep_ptr );


        break;


    case TS_IB_CM_REQ_RECEIVED:
    default:

        //dapl_dbg_log (DAPL_DBG_TYPE_ERR, " Unexpected CM state: %d\n", state);
	break;

    }

    return ret;
}


static int do_passive_idle(tTS_IB_CM_COMM_ID          comm_id, 
			   struct ib_u_cm_idle_param *idle_param,
			   void                      *context)
{
    ib_cm_events_t         ib_cm_event = 0;
    //syslog(LOG_NOTICE, "Passive went to idle on comm_id = %d\n", comm_id);
    switch (idle_param->reason) 
    {
    case TS_IB_CM_IDLE_REMOTE_REJECT:
	//syslog(LOG_NOTICE, "do_passive_idle: TS_IB_CM_IDLE_REMOTE_REJECT for comm_id=%x\n", comm_id);
	if (idle_param->rej_info_len && idle_param->rej_info) 
	{
        ib_cm_event = IB_CME_DESTINATION_REJECT_PRIVATE_DATA;
        dapls_cr_callback(comm_id,
                      ib_cm_event,
                      idle_param->rej_info,
                      context);
        return 0;

	} else {
	    ib_cm_event = IB_CME_DESTINATION_REJECT;
	}

	break;
	
    case TS_IB_CM_IDLE_REMOTE_TIMEOUT:
	//syslog(LOG_NOTICE, "do_passive_idle: TS_IB_CM_IDLE_REMOTE_TIMEOUT for comm_id=%x\n", comm_id);
	ib_cm_event = IB_CME_DESTINATION_UNREACHABLE;
	break;
	
    case TS_IB_CM_IDLE_TIME_WAIT_DONE:
	//syslog(LOG_NOTICE, "do_passive_idle: TS_IB_CM_IDLE_TIME_WAIT_TIMEOUT for comm_id=%x\n", comm_id);
        ib_cm_event = IB_CME_DISCONNECTED;
        return 0;
         

	break;
	
    case TS_IB_CM_IDLE_LOCAL_REJECT:
    default:
	//syslog(LOG_NOTICE, "do_passive_idle: TS_IB_CM_IDLE_LOCAL_REJECT for comm_id=%x\n", comm_id);
	ib_cm_event = IB_CME_LOCAL_FAILURE;
	break;
	
    }
    dapls_cr_callback(comm_id,
                  ib_cm_event,
                  NULL,
                  context);
    return 0;
}


static int do_passive_req_recv(tTS_IB_CM_COMM_ID                  comm_id, 
			       struct ib_u_cm_req_received_param *req_recv_param,
			       void                              *context)
{
    ib_cm_events_t ib_cm_event;
    //syslog(LOG_NOTICE, "Passive REQ received on comm_id = %d\n", comm_id);
    if (req_recv_param->remote_private_data_len && 
	req_recv_param->remote_private_data)
        ib_cm_event = IB_CME_CONNECTION_REQUEST_PENDING_PRIVATE_DATA;
    else
        ib_cm_event = IB_CME_CONNECTION_REQUEST_PENDING;
   
    dapls_cr_callback (comm_id,
                       ib_cm_event,
                       req_recv_param->remote_private_data,
                       context);
    return 0;
}


static int do_passive_established(tTS_IB_CM_COMM_ID  comm_id,
				  void              *context)
{
    //syslog(LOG_NOTICE, "Passive established on comm_id = %d\n", comm_id);
    dapls_cr_callback (comm_id,
                       IB_CME_CONNECTED,
                       NULL,
                       (void *) context );
    return 0;
}


static int do_passive_disc(tTS_IB_CM_COMM_ID comm_id, void * context)
{  
    //syslog(LOG_NOTICE, "Passive established on comm_id = %d\n", comm_id);
    dapls_cr_callback (comm_id,
                       IB_CME_DISCONNECTED,
                       NULL,
                       (void *) context );
        return 0;

}

int dapls_cm_passive_cb_handler(tTS_IB_U_QP_HANDLE   consumer_qp,
				tTS_IB_CM_COMM_ID    comm_id,
				tTS_IB_CM_EVENT      event,
				void                *params,
				void                *arg);

int dapls_cm_passive_cb_handler(tTS_IB_U_QP_HANDLE   consumer_qp,
				tTS_IB_CM_COMM_ID    comm_id,
				tTS_IB_CM_EVENT      event,
				void                *params,
				void                *arg)
{
int ret = 0;


    switch (event)
    {
    case TS_IB_CM_IDLE:
	
	ret = do_passive_idle(comm_id, (struct ib_u_cm_idle_param *) params,
			      arg);
	break;
	
    case TS_IB_CM_REQ_RECEIVED:

	ret = do_passive_req_recv(comm_id, 
				  (struct ib_u_cm_req_received_param *) params,
				  arg);
	break;

    case TS_IB_CM_ESTABLISHED:

	ret = do_passive_established(comm_id, arg);
	break;

    case TS_IB_CM_DISCONNECTED:

        ret = do_passive_disc(comm_id, arg);
        break;


    default:

        /* 
         * ETBD - Where state was declared at?
         *
         */

        // dapl_dbg_log (DAPL_DBG_TYPE_ERR, " Unexpected CM state: %d\n", state);
	break;

    }

    return ret;
}

 
/*
 * dapls_ib_connect
 *
 * Initiate a connection with the passive listener on another node
 *
 * Input:
 *        ep_handle,
 *        remote_ia_address,
 *        remote_conn_qual,
 *        private_data_size    size of private data and structure
 *        private_data         pointats_entry->semer to private data structure
 *
 * Output:
 *         none
 *
 * Returns:
 *         DAT_SUCCESS
 *         DAT_INSUFFICIENT_RESOURCES
 *         DAT_INVALID_PARAMETER
 * 
*/



DAT_RETURN dapls_ib_connect (
	IN  DAT_EP_HANDLE		ep_handle,
	IN  DAT_IA_ADDRESS_PTR		remote_ia_address,
	IN  DAT_CONN_QUAL		remote_conn_qual,
	IN  DAT_COUNT			private_data_size,
	IN  DAT_PVOID			*private_data )
{
    tTS_IB_SERVICE_ID            service_id;
    DAPL_IA                      *ia_ptr;
    DAPL_EP                      *ep_ptr;
    int                          status;
    struct ib_path_record        path_record;
    struct ib_cm_active_param    qp_param;


    ep_ptr = (DAPL_EP *) ep_handle;
    ia_ptr = ep_ptr->header.owner_ia;
    qp_param.req_private_data     = private_data;
    qp_param.req_private_data_len = private_data_size;
    qp_param.responder_resources = IB_TARGET_MAX;
    qp_param.initiator_depth     = IB_INITIATOR_DEPTH;
    qp_param.retry_count         = TS_UDAPL_RC_RETRY_COUNT;
    qp_param.rnr_retry_count     = TS_UDAPL_RNR_RETRY_COUNT;
    qp_param.cm_response_timeout = TS_UDAPL_CM_RESPONSE_TIMEOUT;
    qp_param.max_cm_retries      = TS_UDAPL_MAX_CM_RETRIES;


    if (((struct sockaddr_in *)remote_ia_address)->sin_addr.s_addr ==
	((struct sockaddr_in *) &ia_ptr->hca_ptr->hca_address)->sin_addr.s_addr) 
    {
	dapl_os_memcpy(&path_record.sgid, ia_ptr->hca_ptr->prov_data->gid, sizeof(iba_gid_t));
	dapl_os_memcpy(&path_record.dgid, ia_ptr->hca_ptr->prov_data->gid, sizeof(iba_gid_t));

	path_record.slid        = ia_ptr->hca_ptr->prov_data->lid;
	path_record.dlid        = ia_ptr->hca_ptr->prov_data->lid;
	path_record.flowlabel   = 0;
	path_record.hoplmt      = 64;
	path_record.tclass      = 0;
	path_record.pkey        = 0xffff;
	path_record.sl          = 0;
	path_record.rate        = IB_RATE_10GB;
	path_record.packet_life = 0;
	
    } else {
	
#ifdef IBHOSTS_NAMING
        
	dapl_os_memcpy(&path_record.sgid, ia_ptr->hca_ptr->gid, sizeof(iba_gid_t));
	path_record.slid = ia_ptr->hca_ptr->lid;
	
	if (((struct sockaddr_in *)remote_ia_address)->sin_addr.s_addr ==
	    ((struct sockaddr_in *) &ia_ptr->hca_ptr->hca_address)->sin_addr.s_addr) 
	{
	    dapl_os_memcpy(&path_record.dgid, ia_ptr->hca_ptr->gid, sizeof(iba_gid_t));
	    path_record.dlid    = ia_ptr->hca_ptr->lid;
	} else {
	    DAT_RETURN                   dat_status;
	    IB_gid_t                     gid;

	    dat_status = dapls_ns_lookup_address (ia_ptr, remote_ia_address,
						  (ib_gid_t *) &gid, 
						  &path_record.dlid);   
	    if (status != DAT_SUCCESS)
	    {
		dapl_dbg_log (DAPL_DBG_TYPE_ERR, 
			      " dapls_ns_lookup_address() failed: %d\n", 
			      dat_status);
		return DAT_INTERNAL_ERROR;
	    }

	    path_record.dgid[0] = gid[3];
	    path_record.dgid[1] = gid[2];
	    path_record.dgid[2] = gid[1];
	    path_record.dgid[3] = gid[0];
	    path_record.dgid[4] = gid[4];
	    path_record.dgid[5] = gid[5];
	    path_record.dgid[6] = gid[6];
	    path_record.dgid[7] = gid[7];
	    path_record.dgid[8] = gid[15];
	    path_record.dgid[9] = gid[14];
	    path_record.dgid[10] = gid[13];
	    path_record.dgid[11] = gid[12];
	    path_record.dgid[12] = gid[11];
	    path_record.dgid[13] = gid[10];
	    path_record.dgid[14] = gid[9];
	    path_record.dgid[15] = gid[8];

	}

	path_record.flowlabel   = 0;
	path_record.hoplmt      = 64;
	path_record.tclass      = 0;
	path_record.pkey        = 0xffff;
	path_record.sl          = 0;
	path_record.rate        = IB_RATE_10GB;
	path_record.packet_life = 0;
	
#else

#ifdef DAPL_ATS
	status = ib_path_record_lookup(remote_ia_address, ia_ptr->hca_ptr->prov_data->gid,
                                   ia_ptr->hca_ptr->prov_data->ts_hca_handle,
                                   ia_ptr->hca_ptr->port_num, 1,
                                   &path_record);
#else
	status = ib_path_record_lookup(remote_ia_address, ia_ptr->hca_ptr->gid,
				       ia_ptr->hca_ptr->ts_hca_handle,
				       ia_ptr->hca_ptr->port_num, 0, 
				       &path_record);
#endif

	if (status < 0)
	{
	    dapl_dbg_log (DAPL_DBG_TYPE_ERR, " Path record lookup failed\n");
	    return DAT_INTERNAL_ERROR;
	}

#endif

    }
    
    path_record.mtu = IB_MTU_1024; /* Best performance is obtained at this MTU */

    dapl_os_memcpy (&service_id, &remote_conn_qual, 
		    sizeof (tTS_IB_SERVICE_ID));

    status = ib_cm_vapi_connect ( ia_ptr->hca_ptr->ib_hca_handle,
                                  ia_ptr->hca_ptr->prov_data->global_pd,
                  ep_ptr->qp_handle,
			      ep_ptr->qpn,
			      &qp_param,
			      &path_record,
			      NULL,
			      service_id,
                  0,
			      dapls_cm_active_cb_handler,
			      (void *) ep_ptr,
			      &ep_ptr->cm_handle );

    if (status)
    {
        dapl_dbg_log (DAPL_DBG_TYPE_ERR, " ib_cm_vapi_connect failed: %d\n", status);
        return DAT_INTERNAL_ERROR;
    }

    //syslog(LOG_NOTICE, "Connect issued to conn qual: %lld comm_id: %d\n", remote_conn_qual, ep_ptr->cm_handle);

    return DAT_SUCCESS;
}


/*
 * dapls_ib_disconnect
 *
 * Disconnect an EP
 *
 * Input:
 *        ep_handle,
 *        disconnect_flags
 *
 * Output:
 *         none
 *
 * Returns:
 *        DAT_SUCCESS
 *        DAT_INSUFFICIENT_RESOURCES
 *        DAT_INVALID_PARAMETER
 *
 */
DAT_RETURN
dapls_ib_disconnect (
		     IN  DAPL_EP              *ep_ptr,
		     IN  DAT_CLOSE_FLAGS       cm_flag )
{
    int status;


    //syslog(LOG_NOTICE, "Disconnecting comm_id: %d\n", ep_ptr->cm_handle);

    if(cm_flag == DAT_CLOSE_ABRUPT_FLAG) {
        // in user there is listen stop call
        // do i need EVAPI_k_clear_destroy_qp_cbk 
        status = ib_cm_vapi_kill(ep_ptr->cm_handle);
    }  else {
        status = ib_cm_vapi_disconnect (ep_ptr->cm_handle);
    }


// remove the warning, it may happen in good flow that the connection
// has been freed, no need here for error
//  if (status)
//  {
//      dapl_dbg_log (DAPL_DBG_TYPE_ERR, " tsIbUCmDisconnect failed: %d\n", 
//                    status);
//
//      return DAT_ERROR (DAT_INSUFFICIENT_RESOURCES,0);
//  }
    
    return DAT_SUCCESS;
}


/*                                                                              
 * dapls_ib_disconnect_clean                                                    
 *                                                                              
 * Clean up outstanding connection data. This routine is invoked                
 * after the final disconnect callback has occurred. Only on the                
 * ACTIVE side of a connection.                                                 
 *                                                                              
 * Input:                                                                       
 *      ep_ptr          DAPL_EP                                                 
 *      active          Indicates active side of connection                     
 *                                                                              
 * Output:                                                                      
 *      none                                                                    
 *                                                                              
 * Returns:                                                                     
 *      void                                                                    
 *                                                                              
 */ 
// ETBD - This is the new function, old function is right below.
// Need to review all ib_cm_events
void
dapls_ib_disconnect_clean (
	IN  DAPL_EP			*ep_ptr,
	IN  DAT_BOOLEAN			active,
	IN  const ib_cm_events_t	ib_cm_event )
{
    DAPL_IA			*ia_ptr;
    int ib_status;

    ia_ptr = ep_ptr->header.owner_ia;
#ifndef	CM_BUSTED

    

    /*
     * Clean up outstanding connection state
     */
    if ( ep_ptr->cm_handle != IB_INVALID_HANDLE )
    {
	dapl_dbg_log (DAPL_DBG_TYPE_CM, 
                      "  >>> ib_disconnect_clean: CM handle: %p\n", 
                      ep_ptr->cm_handle);
	if ( active )
	{
            ib_status = ib_cm_vapi_reject (ep_ptr->cm_handle, NULL, 0);
// remove the warning, it may happen in good flow that the connection
// has been freed, no need here for error
//          if (ib_status) 
//          {
//          	dapl_dbg_log (DAPL_DBG_TYPE_ERR,
//                            " Connection abort reject send failed: %d\n", ib_status);
//          }
        }


	ep_ptr->cm_handle = IB_INVALID_HANDLE;
	/* Certain events  will leave the QP in the ERROR state, per
	 * the IB spec
	 */
	if ( ib_cm_event == IB_CME_DISCONNECTED ||
	     ib_cm_event == IB_CME_DISCONNECTED_ON_LINK_DOWN ||
	     ib_cm_event == IB_CME_DESTINATION_REJECT ||
	     ib_cm_event == IB_CME_DESTINATION_REJECT_PRIVATE_DATA )
	{
	    ep_ptr->qp_state = VAPI_ERR;
	}

	if ( ib_cm_event == IB_CME_CONNECTION_REQUEST_PENDING )
	{
	    /*
	     * Abort a pending connection; move the EP to the error
	     * state. We dont care about the return code, it will work
	     * or it will be in the ERROR state already
	     */
           ib_status = dapls_modify_qp_state_to_error(ia_ptr->hca_ptr->ib_hca_handle,
                                                      ep_ptr->qp_handle);
    
    
            if ( ib_status != VAPI_OK )
            {
                dapl_dbg_log (DAPL_DBG_TYPE_ERR, 
                              " dapls_modify_qp_state_to_error failed: %d\n", 
                              ib_status);
            }


        }

    }
#endif /* CM_BUSTED */

    return;
}


// ETBD - This is the old function. We need to review that the reject is not needed
//void
//dapls_ib_disconnect_clean (
//	IN  DAPL_EP			*ep_ptr,
//	IN  DAT_BOOLEAN			passive,
//	IN  const ib_cm_events_t	ib_cm_event )
//{
//    DAPL_IA    *ia_ptr;
//    VAPI_ret_t ib_status;
//    int        status;
//
//    
//    ia_ptr = ep_ptr->header.owner_ia;
//
//    ib_status = dapls_modify_qp_state_to_error (
//	ia_ptr->hca_ptr->ib_hca_handle,
//	ep_ptr->qp_handle);
//    
//    if ( ib_status != VAPI_OK )
//    {
//	dapl_dbg_log (DAPL_DBG_TYPE_ERR, 
//		      " dapls_modify_qp_state_to_error failed: %d\n", 
//		      ib_status);
//    }240219
//    
//    status = tsIbUCmReject(ep_ptr->cm_handle, NULL, 0);
//    
//    if (status) 
//    {
//	dapl_dbg_log (DAPL_DBG_TYPE_ERR, 
//		      " Connection abort reject send failed: %d\n", 
//		      status);
//    }
//
//}


/*
 * dapls_ib_reinit_ep
 *
 * Move the QP to INIT state again.
 *
 * Input:
 *      ep_ptr          DAPL_EP
 *
 * Output:
 *      none
 *
 * Returns:
 *      void
 *
 */
void
dapls_ib_reinit_ep (
		    IN  DAPL_EP *ep_ptr)
{
    DAPL_IA    *ia_ptr;
    VAPI_ret_t ib_status;


    ia_ptr = ep_ptr->header.owner_ia;
    ib_status = dapls_modify_qp_state_to_reset (
	ia_ptr->hca_ptr->ib_hca_handle,
	ep_ptr->qp_handle);

    if ( ib_status != VAPI_OK )
    {
	dapl_dbg_log (DAPL_DBG_TYPE_ERR, 
		      " dapls_modify_qp_state_to_reset failed: %d\n", 
		      ib_status);
	return;
    }
    ib_status = dapls_modify_qp_state_to_init (
                               ia_ptr->hca_ptr->ib_hca_handle,
                               ia_ptr->hca_ptr->port_num,
                               ep_ptr->qp_handle);

    if ( ib_status != VAPI_OK )
    {
        dapl_dbg_log (DAPL_DBG_TYPE_ERR, 
		      " dapls_modify_qp_state_to_init failed: %d\n", 
		      ib_status);
        return;
    }

    return;
}


/*
 * dapl_ib_setup_conn_listener
 *
 * Have the CM set up a connection listener.
 *
 * Input:
 *        ibm_hca_handle           HCA handle
 *        qp_handle                QP handle
 *
 * Output:
 *         none
 *
 * Returns:
 *        DAT_SUCCESS
 *        DAT_INSUFFICIENT_RESOURCES
 *        DAT_INVALID_PARAMETER
 *
 */
DAT_RETURN
dapls_ib_setup_conn_listener (
			      IN  DAPL_IA    *ia_ptr,
			      IN  DAT_UINT64 ServiceID,
			      IN  DAPL_SP    *sp_ptr )
{
    int status;


    //syslog(LOG_NOTICE, "Listening on service id = %llx\n", ServiceID);
    status = ib_cm_vapi_listen    (ServiceID,
			   TS_IB_CM_SERVICE_EXACT_MASK,
			   dapls_cm_passive_cb_handler,
			   (void *) sp_ptr,
			   &sp_ptr->cm_srvc_handle);
    if (status)
    {
        sp_ptr->cm_srvc_handle = 0;
//	dapl_dbg_log (DAPL_DBG_TYPE_ERR, " tsIbUCmListen failed: %d\n", 
//					  status);

	if (status == -EADDRINUSE)
	    return DAT_CONN_QUAL_IN_USE;
	else
	    return DAT_ERROR (DAT_INSUFFICIENT_RESOURCES,0);
    }

    return DAT_SUCCESS;

}


/*
 * dapl_ib_remove_conn_listener
 *
 * Have the CM remove a connection listener.
 *
 * Input:
 *      ia_handle               IA handle
 *      ServiceID               IB Channel Service ID
 *
 * Output:
 *      none
 *
 * Returns:
 *      DAT_SUCCESS
 *      DAT_INVALID_PARAMETER
 *
 */
DAT_RETURN dapls_ib_remove_conn_listener (
					  IN  DAPL_IA *ia_ptr,
					  IN  DAPL_SP *sp_ptr)
{
    int status;


    //syslog(LOG_NOTICE, "Stopping listen on service id = %llx\n", sp_ptr->conn_qual);

    status = ib_cm_vapi_listen_stop(sp_ptr->cm_srvc_handle);

    if (status) 
    {
	dapl_dbg_log (DAPL_DBG_TYPE_ERR, " tsIbUCmListenStop failed: %d\n", 
				     status);
	return DAT_ERROR (DAT_INSUFFICIENT_RESOURCES,0);
    }

    return DAT_SUCCESS;
}


/*
 * dapls_ib_reject_connection
 *
 * Perform necessary steps to reject a connection
 *
 * Input:
 *        cr_handle
 *
 * Output:
 *         none
 *
 * Returns:
 *        DAT_SUCCESS
 *        DAT_INSUFFICIENT_RESOURCES
 *        DAT_INVALID_PARAMETER
 *
 */
DAT_RETURN
dapls_ib_reject_connection (
			    IN  ib_cm_handle_t ib_cm_handle,
			    IN  int            reject_reason )
{
    int status;


    //syslog(LOG_NOTICE, "Rejecting connection comm_id = %x\n", ib_cm_handle);

    status = ib_cm_vapi_reject(ib_cm_handle, NULL, 0);

    if (status) 
    {
	dapl_dbg_log (DAPL_DBG_TYPE_ERR, " ib_cm_vapi_reject failed: %d\n", 
		      status);
	return DAT_ERROR (DAT_INSUFFICIENT_RESOURCES,0);
    }

    return DAT_SUCCESS;
}


/*
 * dapls_ib_accept_connection
 *
 * Perform necessary steps to accept a connection
 *
 * Input:
 *        cr_handle
 *        ep_handle
 *        private_data
 *
 * Output:
 *         none
 *
 * Returns:
 *        DAT_SUCCESS
 *        DAT_INSUFFICIENT_RESOURCES
 *        DAT_INVALID_PARAMETER
 *
 */

/* 
 * ETBD - We added private data size but we are not using it.
 *
 */
DAT_RETURN dapls_ib_accept_connection (
			    IN  DAT_CR_HANDLE cr_handle,
			    IN  DAT_EP_HANDLE ep_handle,
                            IN  DAT_COUNT private_data_size,
                            IN  const DAT_PVOID prd_ptr )
{
    DAPL_CR                *cr_ptr;
    DAPL_EP                *ep_ptr;
    DAPL_IA                *ia_ptr;
    DAT_RETURN             dat_status;
    int                    status;

    cr_ptr = (DAPL_CR *) cr_handle;
    ep_ptr = (DAPL_EP *) ep_handle;
    ia_ptr = ep_ptr->header.owner_ia;

    /* Obtain size of private data structure & contents */
    if(private_data_size >  IB_MAX_REP_PDATA_SIZE)
        return(DAT_LENGTH_ERROR);
    

    if ( ep_ptr->qp_state == DAPL_QP_STATE_UNATTACHED )
    {
        /* 
         * If we are lazy attaching the QP then we may need to
         * hook it up here. Typically, we run this code only for
         * DAT_PSP_PROVIDER_FLAG
         */
        dat_status = dapls_ib_qp_alloc ( ia_ptr, ep_ptr, NULL );

        if ( dat_status != DAT_SUCCESS)
        {
            dapl_dbg_log (DAPL_DBG_TYPE_ERR, " dapls_ib_qp_alloc failed: %d\n", 
					 dat_status);
            return dat_status;
        }
    }

    
    cr_ptr->param.local_ep_handle = ep_handle;
    ep_ptr->cm_handle             = cr_ptr->ib_cm_handle;

    //syslog(LOG_NOTICE, "Accepting connection service id = %llx comm_id = %d\n", cr_ptr->sp_ptr->conn_qual, cr_ptr->ib_cm_handle);
    {
        struct ib_cm_passive_param passive_params = {
            .reply_private_data = prd_ptr,
            .reply_private_data_len = private_data_size,
        };

        status = ib_cm_vapi_accept( ia_ptr->hca_ptr->ib_hca_handle,
                                    ia_ptr->hca_ptr->prov_data->global_pd,
                                    ep_ptr->qp_handle,
                                    ep_ptr->qpn,
                                    dapls_cm_passive_cb_handler,
                                    (void *) cr_ptr->sp_ptr,
                                    cr_ptr->ib_cm_handle,
                                    &passive_params);
    };

    if (status)
    {
        dapl_dbg_log (DAPL_DBG_TYPE_ERR, "ib_cm_accept  failed: %d\n",
		      status);
        return DAT_ERROR (DAT_INSUFFICIENT_RESOURCES,0);
    }

    return DAT_SUCCESS;
}




DAT_RETURN dapls_ib_comm_established(IN DAPL_EP *ep_ptr )
{
    int          status;
    DAT_RETURN   dat_status = 0;

    status = ib_cm_vapi_establish(ep_ptr->cm_handle ,1);

    if (status)
    {
        dapl_dbg_log (DAPL_DBG_TYPE_ERR, " ib_cm_vapi_establish failed: %d cm_handle: %x\n", 
		      status, ep_ptr->cm_handle);
        return DAT_ERROR (DAT_INSUFFICIENT_RESOURCES,0);
    }
/*
    ep_ptr->param.ep_state = DAT_EP_STATE_CONNECTED;

    dat_status = dapls_evd_post_connection_event (
        (DAPL_EVD *) ep_ptr->param.connect_evd_handle,
        DAT_CONNECTION_EVENT_ESTABLISHED,
        (DAT_HANDLE) ep_ptr,
        0,
        NULL);
*/        
    return dat_status;
}

/*
 * dapls_ib_cm_remote_addr
 *
 * Obtain the remote IP address given a connection
 *
 * Input:
 *	cr_handle
 *	private data structure handle (only if not using ATS)
 *
 * Output:
 *	remote_ia_address: where to place the remote address
 *
 * Returns:
 * 	DAT_SUCCESS
 *	DAT_INSUFFICIENT_RESOURCES
 *	DAT_INVALID_PARAMETER
 *
 */

//DAT_RETURN
//dapls_ib_cm_remote_addr (
//	IN      DAPL_CR			*cr_ptr,
//	IN      DAPL_PRIVATE		*prd_ptr, // ETBD - we are not supporting no-ATS mode
//	OUT	DAT_SOCK_ADDR6		*remote_ia_address )

DAT_RETURN dapls_ib_cm_remote_addr (
	IN      DAT_HANDLE		dat_handle,
	OUT	DAT_SOCK_ADDR6		*remote_ia_address )

{
#ifdef DAPL_ATS
    int                     status = 0;
    uint8_t                 ip_addr[16];
    struct sockaddr_in      *ipv4_addr;
    DAPL_IA                 *ia_ptr;
#endif

    int rc;
    tTS_IB_GID remote_gid;

    DAPL_CR * cr_ptr = (DAPL_CR*)dat_handle;

#ifndef DAPL_ATS
    if (prd_ptr != NULL) 
    {
	*(DAT_SOCK_ADDR *)remote_ia_address = prd_ptr->hca_address;
    }
    return DAT_SUCCESS;
#else

    ia_ptr = (DAPL_IA *)cr_ptr->sp_ptr->ia_handle;

    rc = ib_cm_get_remote_gid(cr_ptr->ib_cm_handle, &remote_gid);
    if(rc < 0)
        return(rc);

    if (!memcmp(ia_ptr->hca_ptr->prov_data->gid, remote_gid, 16)) {
	memcpy(remote_ia_address, &ia_ptr->hca_ptr->hca_address, sizeof(DAT_SOCK_ADDR));
	return DAT_SUCCESS;
    }

    /* Obtain the remote address from the provider */
    status = ib_ats_ipaddr_lookup (remote_gid, ia_ptr->hca_ptr->prov_data->ts_hca_handle, 
				   ia_ptr->hca_ptr->port_num, ip_addr);

    if (status < 0)
    {
        return DAT_ERROR (DAT_INSUFFICIENT_RESOURCES,0);
    }

    ipv4_addr = (struct sockaddr_in *) remote_ia_address;
        
    ipv4_addr->sin_family = AF_INET;
    dapl_os_memcpy (&ipv4_addr->sin_addr.s_addr,
		    &ip_addr[12],
		    4);

    return DAT_SUCCESS;
#endif /* !DAPL_ATS */
}

#ifdef DAPL_ATS
DAT_RETURN dapls_ib_cm_set_addr (
        IN      DAPL_HCA                *hca_ptr );


DAT_RETURN dapls_ib_cm_set_addr (
        IN      DAPL_HCA                *hca_ptr )
{
    int                 status;
    struct sockaddr_in *ipv4_addr;
    uint8_t             ip_addr[16];

    ipv4_addr = (struct sockaddr_in *) &hca_ptr->hca_address;
    memset(ip_addr, 0, 12);
    dapl_os_memcpy(&ip_addr[12], &ipv4_addr->sin_addr.s_addr, 4);

    status = ib_ats_set_ipaddr(hca_ptr->prov_data->gid, hca_ptr->prov_data->ts_hca_handle, 
			       hca_ptr->port_num, ip_addr);

    if (status < 0)
    {
        return DAT_ERROR (DAT_INSUFFICIENT_RESOURCES,0);
    }

    return DAT_SUCCESS;
}
#endif


/*
 * dapls_ib_private_data_size
 *
 * Return the size of private data given a connection op type
 *
 * Input:
 *	prd_ptr		private data pointer
 *	conn_op		connection operation type
 *
 * If prd_ptr is NULL, this is a query for the max size supported by
 * the provider, otherwise it is the actual size of the private data
 * contained in prd_ptr.
 *
 * Infiniband has fixed size private data, so prd_ptr is ignored.
 *
 * Output:
 *	None
 *
 * Returns:
 * 	length of private data
 *
 */
int
dapls_ib_private_data_size (
	IN      DAPL_PRIVATE		*prd_ptr,
	IN	DAPL_PDATA_OP		conn_op)
{
    int  size;

    switch (conn_op)
    {
        case DAPL_PDATA_CONN_REQ:
	{
	    size = IB_MAX_REQ_PDATA_SIZE;
	    break;
	}
        case DAPL_PDATA_CONN_REP:
	{
	    size = IB_MAX_REP_PDATA_SIZE;
	    break;
	}
        case DAPL_PDATA_CONN_REJ:
	{
	    size = IB_MAX_REJ_PDATA_SIZE;
	    break;
	}
        case DAPL_PDATA_CONN_DREQ:
	{
	    size = IB_MAX_DREQ_PDATA_SIZE;
	    break;
	}
        case DAPL_PDATA_CONN_DREP:
	{
	    size = IB_MAX_DREP_PDATA_SIZE;
	    break;
	}
        default:
	{
	    size = 0;
	}
    } /* end case */

#ifdef CM_BUSTED
    size = 0;			/* force to size of 0 */
#endif

    return size;
}



