/*
 * This Software is licensed under both of the following two licenses:
 *
 * 1) under the terms of the "Common Public License 1.0" a copy of which is
 *    in the file LICENSE.txt in the root directory. The license is also
 *    available from the Open Source Initiative, see
 *    http://www.opensource.org/licenses/cpl.php.
 * OR
 *
 * 2) under the terms of the "The BSD License" a copy of which is in the file
 *    LICENSE2.txt in the root directory. The license is also available from
 *    the Open Source Initiative, see
 *    http://www.opensource.org/licenses/bsd-license.php.
 *
 * Licensee has the right to choose either one of the above two licenses.
 *
 * Redistributions of source code must retain both the above copyright
 * notice and either one of the license notices.
 *
 * Redistributions in binary form must reproduce both the above copyright
 * notice, either one of the license notices in the documentation
 * and/or other materials provided with the distribution.
 */

/**************************************************************************
 *
 *   Module:		 uDAPL
 *
 *   Filename:		 dapl_topspin_cm.h
 *
 *   Author:		 Ariel Cohen
 *
 *   Created:		 03/05/2003
 *
 *   Description:
 *
 *   Header file for the uDAPL Topspin CM
 *
 ***************************************************************************
 *
 *		   Source Control System Information
 *
 *    $Id: dapl_openib_cm.h,v 1.1 2004/09/16 13:37:22 sjs2 Exp $
 *
 ***************************************************************************
 *
 ***************************************************************************
 *			 (c) COPYRIGHT 2003
 *			     Topspin Communications, Inc.
 *			     Mountain View, CA
 **************************************************************************/
#ifndef _DAPL_TOPSPIN_CM_H
#define _DAPL_TOPSPIN_CM_H

#define u64 u_int64_t 

#include "ts_ib_sa_client.h"
//#include "ts_ib_useraccess_cm.h"
#include "dapl_openib_cm_util.h"

#define TS_UDAPL_RC_RETRY_COUNT      7
#define TS_UDAPL_RNR_RETRY_COUNT     6
#define TS_UDAPL_CM_RESPONSE_TIMEOUT 20  /* 4 sec */
#define TS_UDAPL_MAX_CM_RETRIES      15

typedef enum {
    IB_CME_CONNECTED,
    IB_CME_DISCONNECTED,
    IB_CME_DISCONNECTED_ON_LINK_DOWN,
    IB_CME_CONNECTION_REQUEST_PENDING,
    IB_CME_CONNECTION_REQUEST_PENDING_PRIVATE_DATA,
    IB_CME_CONNECTION_REQUEST_ACKED,
    IB_CME_DESTINATION_REJECT,
    IB_CME_DESTINATION_REJECT_PRIVATE_DATA,
    IB_CME_DESTINATION_UNREACHABLE,
    IB_CME_TOO_MANY_CONNECTION_REQUESTS,
    IB_CME_LOCAL_FAILURE,
    IB_CME_BROKEN
} ib_cm_events_t;

#define IB_CM_LOCAL_FAILURE IB_CME_LOCAL_FAILURE
#define IB_CM_HANDLE_FMT  "%x"
typedef tTS_IB_CM_COMM_ID    ib_cm_handle_t;
typedef tTS_IB_LISTEN_HANDLE ib_cm_srvc_handle_t;

#define IB_MAX_REQ_PDATA_SIZE   92
#define IB_MAX_REP_PDATA_SIZE  196
#define IB_MAX_REJ_PDATA_SIZE  148
#define IB_MAX_DREQ_PDATA_SIZE 220
#define IB_MAX_DREP_PDATA_SIZE 224

#define TS_REQ_MRA_TIMEOUT 27  /* a little over 9 minutes */
    

/************************
* Function declarations *
************************/

int dapls_cm_passive_cb_handler(tTS_IB_U_QP_HANDLE   consumer_qp,
				tTS_IB_CM_COMM_ID    comm_id,
				tTS_IB_CM_EVENT      event,
				void                *params,
				void                *arg);

int dapls_cm_active_cb_handler(tTS_IB_U_QP_HANDLE   consumer_qp,
			       tTS_IB_CM_COMM_ID    comm_id,
			       tTS_IB_CM_EVENT      event,
			       void                *params,
			       void                *arg); 

#endif /* _DAPL_TOPSPIN_CM_H */
