/*************************************************************************
*   Copyright (c) 2004 Mellanox Technologies Ltd.  All rights reserved.  *
*   Copyright (c) 2005 Mellanox Technologies.   All rights reserved.

*************************************************************************/

#include "dapl.h"

/*
 * TS specific includes
 */
#include "dapl_openib_cm_util.h"
#include "ts_ib_core.h"
#include "ts_ib_tavor_provider.h"

/* Mellanox stuff */
#include "vapi_types.h"
#include "vapi.h"
#include "evapi.h"



typedef struct ib_cm_vapi_function_s {
    tTS_IB_U_CM_COMPLETION_FUNC  function;
    void                        *arg;

} ib_cm_vapi_function_t;

tTS_IB_CM_CALLBACK_RETURN ib_cm_vapi_cb_handler(tTS_IB_U_CM_COMPLETION_FUNC function,
                                                tTS_IB_CM_EVENT   event,
                                                tTS_IB_CM_COMM_ID comm_id,
                                                void             *param,
                                                void             *arg);


tTS_IB_CM_CALLBACK_RETURN ib_cm_vapi_cb_handler(tTS_IB_U_CM_COMPLETION_FUNC function,
                                                tTS_IB_CM_EVENT   event,
                                                tTS_IB_CM_COMM_ID comm_id,
                                                void             *param,
                                                void             *arg) {

    tTS_IB_CM_CALLBACK_RETURN ret = TS_IB_CM_CALLBACK_PROCEED;    

    void * buffer;
    int params_size;

    

    switch (event) {
    case TS_IB_CM_REQ_RECEIVED:
        {

            struct ib_cm_req_received_param *req_param =
            (struct ib_cm_req_received_param *) param;

            struct ib_u_cm_req_received_param *u_req_param;

            params_size = sizeof(*u_req_param) +
                          req_param->remote_private_data_len;
            buffer = dapl_os_alloc(params_size);
            u_req_param =
            (struct ib_u_cm_req_received_param *) buffer;

            u_req_param->listen_handle = req_param->listen_handle;
            u_req_param->service_id = req_param->service_id;
            u_req_param->local_qpn = req_param->local_qpn;
            u_req_param->remote_qpn = req_param->remote_qpn;

            memcpy(u_req_param->remote_guid, req_param->remote_guid,
                   sizeof(u_req_param->remote_guid));
            memcpy(u_req_param->dgid, req_param->dgid,
                   sizeof(u_req_param->dgid));
            memcpy(u_req_param->sgid, req_param->sgid,
                   sizeof(u_req_param->sgid));

            u_req_param->dlid = req_param->dlid;
            u_req_param->slid = req_param->slid;
            u_req_param->port = req_param->port;

            u_req_param->remote_private_data_len =
            req_param->remote_private_data_len;

            memcpy(buffer + sizeof(*u_req_param),
                   req_param->remote_private_data,
                   req_param->remote_private_data_len);

            u_req_param->remote_private_data = buffer + sizeof(*u_req_param);

            ret = function( VAPI_INVAL_HNDL, 
                                           comm_id,
                                           event,
                                           u_req_param,
                                           arg);
            /* TBD do i need to check the return value ?*/
			ret = TS_IB_CM_CALLBACK_DEFER;
            dapl_os_free(buffer,params_size);
        }
        break;

    case TS_IB_CM_REP_RECEIVED:
        {

            struct ib_cm_rep_received_param *rep_param =
            (struct ib_cm_rep_received_param *) param;

            struct ib_u_cm_rep_received_param *u_rep_param;

            params_size = sizeof(*u_rep_param) +
                          rep_param->remote_private_data_len;
            buffer = dapl_os_alloc(params_size);
            u_rep_param = (struct ib_u_cm_rep_received_param *) buffer;

            u_rep_param->local_qpn = rep_param->local_qpn;
            u_rep_param->remote_qpn = rep_param->remote_qpn;

            u_rep_param->remote_private_data_len =
            rep_param->remote_private_data_len;

            memcpy(buffer + sizeof(*u_rep_param),
                   rep_param->remote_private_data,
                   rep_param->remote_private_data_len);


            u_rep_param->remote_private_data = buffer + sizeof(*u_rep_param);

            ret = function( VAPI_INVAL_HNDL, 
                                           comm_id,
                                           event,
                                           u_rep_param,
                                           arg);

            /* TBD di i need to check the return value*/
			ret = TS_IB_CM_CALLBACK_DEFER;
            dapl_os_free(buffer,params_size);
        }
        break;

    case TS_IB_CM_DISCONNECTED:
        {
            struct ib_cm_disconnected_param *disconnected_param =
            (struct ib_cm_disconnected_param *) param;
            struct ib_u_cm_disconnected_param *u_disconnected_param;

            params_size = sizeof(*u_disconnected_param);
            buffer = dapl_os_alloc(params_size);

            u_disconnected_param =
            (struct ib_u_cm_disconnected_param *) buffer;

            u_disconnected_param->reason =disconnected_param->reason;

            ret = function( VAPI_INVAL_HNDL, 
                                           comm_id,
                                           event,
                                           u_disconnected_param,
                                           arg);

            dapl_os_free(buffer,params_size);
        }
        break;

    case TS_IB_CM_LAP_RECEIVED:
    case TS_IB_CM_APR_RECEIVED:
    case TS_IB_CM_ESTABLISHED:
    {
        ret = function( VAPI_INVAL_HNDL, 
                                       comm_id,
                                       event,
                                       NULL,
                                       arg);
    };
    break;
    case TS_IB_CM_IDLE:{
            struct ib_cm_idle_param *idle_param =
            (struct ib_cm_idle_param *) param;
            struct ib_u_cm_idle_param *u_idle_param;

            params_size = sizeof(*u_idle_param) + idle_param->rej_info_len;
            buffer = dapl_os_alloc(params_size);
            // if (!buffer) {
            //     TS_REPORT_WARN(MOD_KERNEL_IB,
            //                "Unable to allocate memory for IDLE param struct");
            //     kfree(cm_comp);
            //     goto out;
            // }

            u_idle_param = (struct ib_u_cm_idle_param *) buffer;

            u_idle_param->reason = idle_param->reason;
            u_idle_param->rej_reason = idle_param->rej_reason;

            u_idle_param->rej_info_len = idle_param->rej_info_len;

            memcpy(buffer + sizeof(*u_idle_param),
                   idle_param->rej_info, idle_param->rej_info_len);


            u_idle_param->rej_info = buffer +sizeof(*u_idle_param);

            //TBD do we need this
            //EVAPI_k_clear_destroy_qp_cbk(cm_conn->v_device,cm_conn->vk_qp);
            //cm_conn->vk_qp = VAPI_INVAL_HNDL;

            ret = function( VAPI_INVAL_HNDL, 
                                           comm_id,
                                           event,
                                           u_idle_param,
                                           arg);
            dapl_os_free(buffer,params_size);
        }
        break;
    }   
    return ret;
}


tTS_IB_CM_CALLBACK_RETURN ib_cm_vapi_passive_cb_handler(tTS_IB_CM_EVENT   event,
                                                tTS_IB_CM_COMM_ID comm_id,
                                                void             *param,
                                                void             *arg);


tTS_IB_CM_CALLBACK_RETURN ib_cm_vapi_passive_cb_handler(tTS_IB_CM_EVENT   event,
                                                tTS_IB_CM_COMM_ID comm_id,
                                                void             *param,
                                                void             *arg) {

    return ib_cm_vapi_cb_handler(dapls_cm_passive_cb_handler,
                                 event,
                                 comm_id,
                                 param,
                                 arg);
}


tTS_IB_CM_CALLBACK_RETURN ib_cm_vapi_active_cb_handler(tTS_IB_CM_EVENT   event,
                                                tTS_IB_CM_COMM_ID comm_id,
                                                void             *param,
                                                void             *arg);


tTS_IB_CM_CALLBACK_RETURN ib_cm_vapi_active_cb_handler(tTS_IB_CM_EVENT   event,
                                                tTS_IB_CM_COMM_ID comm_id,
                                                void             *param,
                                                void             *arg) {
    return ib_cm_vapi_cb_handler(dapls_cm_active_cb_handler,
                                 event,
                                 comm_id,
                                 param,
                                 arg);
}






int ib_cm_vapi_disconnect(tTS_IB_CM_COMM_ID comm_id ) {
    return ib_cm_disconnect(comm_id);
}

int ib_cm_vapi_confirm(tTS_IB_CM_COMM_ID comm_id,
                       void             *rtu_private_data,
                       int               rtu_private_data_len) {
    return  ib_cm_confirm (comm_id,
                           rtu_private_data,
                           rtu_private_data_len);
}


int ib_cm_vapi_kill(tTS_IB_CM_COMM_ID comm_id) {
    return ib_cm_kill(comm_id);
}

int ib_cm_vapi_establish(tTS_IB_CM_COMM_ID comm_id,
                         int               immediate) {
    return ib_cm_establish(comm_id,
                           immediate);
}

int ib_cm_vapi_listen_stop(tTS_IB_LISTEN_HANDLE listen_handle) {
    return  ib_cm_listen_stop(listen_handle);
}


int ib_cm_vapi_listen(tTS_IB_SERVICE_ID           service_id,
                      tTS_IB_SERVICE_ID           service_mask,
                      tTS_IB_U_CM_COMPLETION_FUNC function, /*This param is disregarded*/
                      void                       *arg,
                      tTS_IB_LISTEN_HANDLE       *listen_handle) {

    return ib_cm_listen(service_id,
                        service_mask,
                        ib_cm_vapi_passive_cb_handler,
                        arg,
                        listen_handle);
}


int  ib_cm_vapi_connect (VAPI_hca_hndl_t                hca_handle,
                         void*                          pd,
                         tTS_IB_U_QP_HANDLE             vqp_handle,
                         __u32                          qpn,
                         struct ib_cm_active_param  *param,
                         struct ib_path_record      *primary_path,
                         struct ib_path_record      *alternate_path,
                         tTS_IB_SERVICE_ID           service_id,
                         int                         peer_to_peer,
                         tTS_IB_U_CM_COMPLETION_FUNC function, /*This param is disregarded*/
                         void                       *arg,
                         tTS_IB_CM_COMM_ID          *comm_id) {

    int status;
    // change vapi handle to ts handle
    struct ib_tavor_qp_create_param tavor_qp_create;
    struct ib_qp_create_param qp_create = {
        .pd = (struct ib_pd*)pd,
        .transport = IB_TRANSPORT_RC,
        .device_specific = &tavor_qp_create,
    };
    struct ib_qp *qp_handle;
    u32 qp_number;
    tavor_qp_create.is_user     = 0;
    tavor_qp_create.vapi_handle =
    (void *) (unsigned long)  vqp_handle;
    tavor_qp_create.qpn         = qpn;

    //TODO itamar what is rhe qp state
    //tavor_qp_create.qp_state    = ;

    status = ib_qp_create(&qp_create, &qp_handle, &qp_number);
    if (status) {
        dapl_dbg_log (DAPL_DBG_TYPE_ERR," ib_qp_create failed: %d\n",status);
        return status;
    }
    param->qp=qp_handle; 
    
    // TODO itamar do i need EVAPI_k_set_destroy_qp_cbk ?

    status =ib_cm_connect(param,
                          primary_path,
                          NULL,
                          service_id,
                          0,
                          ib_cm_vapi_active_cb_handler,
                          arg,
                          comm_id );
    if (status) {
        dapl_dbg_log (DAPL_DBG_TYPE_ERR, " ib_cm_connect failed: %d\n", status);
    }
    return status;
}



int ib_cm_vapi_accept(VAPI_hca_hndl_t                hca_handle,
                      void*                          pd,
                      tTS_IB_U_QP_HANDLE             vqp_handle,
                      __u32                          qpn,
                      tTS_IB_U_CM_COMPLETION_FUNC    function,
                      void                          *comp_arg,
                      tTS_IB_CM_COMM_ID              comm_id,
                      struct ib_cm_passive_param    *param) {
    // itamar change from user to kernel 
    // change vapi handle to ts handle
    int status;
    struct ib_tavor_qp_create_param tavor_qp_create;
    struct ib_qp_create_param qp_create = {
        .pd = (struct ib_pd*)pd,
        .transport = IB_TRANSPORT_RC,
        .device_specific = &tavor_qp_create,
    };
    struct ib_qp *qp_handle;
    u32 qp_number;
    tavor_qp_create.is_user     = 0;
    tavor_qp_create.vapi_handle =
    (void *) (unsigned long)  vqp_handle;
    tavor_qp_create.qpn         = qpn;

    //TODO what is rhe qp state
    //tavor_qp_create.qp_state    = ;

    status = ib_qp_create(&qp_create, &qp_handle, &qp_number);
    if (status) {
        dapl_dbg_log (DAPL_DBG_TYPE_ERR," ib_qp_create failed: %d\n",status);
        return status;
    }
    param->qp = qp_handle;

    status  = ib_cm_accept(comm_id,param);
    if (status) {
        dapl_dbg_log (DAPL_DBG_TYPE_ERR, "ib_cm_accept_wapper  failed: %d\n",
                      status);
        return status;
    }

    // do i need to use EVAPI_k_set_destroy_qp_cbk ?? 

    return(0);
}

int ib_cm_vapi_reject(tTS_IB_CM_COMM_ID comm_id,
                      void             *rej_private_data,
                      int               rej_private_data_len) {

    int status;
    status =  ib_cm_reject (comm_id,
                            rej_private_data,
                            rej_private_data_len);
    if (status) {
        dapl_dbg_log (DAPL_DBG_TYPE_ERR, "ib_cm_reject  failed: %d\n",
                      status);
        return status;
    }
    ib_cm_callback_modify(comm_id, NULL, NULL);
    return status;
}

tTS_IB_SERVICE_ID ib_cm_vapi_service_assign(void)
{
    return ib_cm_service_assign();
}


