/*
 * Copyright (c) 1999-2002, Mellanox Technologies, Inc. All rights reserved. 
 * Copyright (c) 2005 Mellanox Technologies.   All rights reserved.

 *
 * This Software is licensed under either one of the following two licenses:
 *
 * 1) under the terms of the "Common Public License 1.0" a copy of which is
 *    in the file LICENSE.txt in the root directory. The license is also
 *    available from the Open Source Initiative, see
 *    http://www.opensource.org/licenses/cpl.php.
 * OR
 *
 * 2) under the terms of the "The BSD License" a copy of which is in the file
 *    LICENSE2.txt in the root directory. The license is also available from
 *    the Open Source Initiative, see
 *    http://www.opensource.org/licenses/bsd-license.php.
 *
 * Licensee has the right to choose either one of the above two licenses.
 *
 * Redistributions of source code must retain both the above copyright
 * notice and either one of the license notices.
 *
 * Redistributions in binary form must reproduce both the above copyright
 * notice, either one of the license notices in the documentation
 * and/or other materials provided with the distribution.
 */

/**********************************************************************
 * 
 * MODULE: dapl_vapi_util.c
 *
 * PURPOSE: Utility routines for access to MELLANOX Tavor Verbs
 *
 * $Id: dapl_openib_util.c,v 1.3 2004/07/28 13:09:31 rabin Exp $
 *
 **********************************************************************/

#ifdef __KERNEL__
#include <linux/types.h>
#include <linux/netdevice.h>


#include "ts_ib_core.h"
#include "ipoib_proto.h"
#endif

#include "dapl_openib_cm.h"
#include "dapl.h"
#include "dapl_adapter_util.h"
#include "dapl_evd_util.h"
#include "dapl_cr_util.h"
#include "dapl_lmr_util.h"
#include "dapl_rmr_util.h"
#include "dapl_cookie.h"
#include "dapl_vendor.h"
#include "dapl_openib.h"

#ifdef __KDAPL__

#define PAGE_SIZE_64             ((DAT_UINT64)PAGE_SIZE)
#define PAGE_LIMIT_64            (PAGE_SIZE_64 - 1)
#define PAGE_MASK_64       	     (~PAGE_LIMIT_64)
#define NOT_PAGE_ALLIGNED(addr)  ((addr)&PAGE_LIMIT_64)
#define PAGE_ALIGN_UP_64(size)   (size)=(((size)+PAGE_LIMIT_64)&PAGE_MASK_64)
#define PAGE_ALIGN_DOWN_64(size) (size)&=PAGE_MASK_64

#endif // __KDAPL__

extern int dapls_ib_cm_set_addr (DAPL_HCA  *hca_ptr);
DAT_RETURN dapls_ib_get_hca_ids(ib_hca_handle_t hca, IB_port_t port, 
                                IB_gid_t  gid, IB_lid_t * lid, IB_guid_t guid);

int g_dapl_loopback_connection = 0;
dapl_tavor_async_cb_t tavor_async_cb;

/*
 * dapls_ib_convert_mem_privileges
 *
 * Convert DAPL mem privileges to provider equivelents
 *
 * Input:
 * 	DAT_MEM_PRIV_FLAGS
 *
 * Output:
 *	none
 *
 * Returns:
 * 	provider equivelent memory flags
 *
 */
STATIC _INLINE_ VAPI_mrw_acl_t
dapls_ib_convert_mem_privileges (
    IN DAT_MEM_PRIV_FLAGS	privileges)
{
    VAPI_mrw_acl_t value = 0;

    /*
     * if (DAT_MEM_PRIV_LOCAL_READ_FLAG & privileges) do nothing
     */
    if (DAT_MEM_PRIV_LOCAL_WRITE_FLAG & privileges)
    {
	value |= VAPI_EN_LOCAL_WRITE;
    }

    if (DAT_MEM_PRIV_REMOTE_READ_FLAG & privileges)
    {
	value |= VAPI_EN_REMOTE_READ;
    }

    if (DAT_MEM_PRIV_REMOTE_WRITE_FLAG & privileges)
    {
	value |= VAPI_EN_REMOTE_WRITE;
    }

    return value;
}


/*
 * dapls_ib_init
 *
 * Initialization and Verb related items
 *
 * Input:
 * 	none
 *
 * Output:
 *	none
 *
 * Returns:
 * 	DAT_SUCCESS
 *
 */
int32_t dapls_ib_init (void)
{
#ifndef __KDAPL__
    int status;
    status = ts_init();                                                                                                                                      
    if (status < 0)
    {
        dapl_dbg_log (DAPL_DBG_TYPE_ERR, "ts_init () failed: %d\n", status);
        return(DAT_INTERNAL_ERROR);
    }
#endif

    dapl_os_memzero (&tavor_async_cb, sizeof (dapl_tavor_async_cb_t));
    tavor_async_cb.set_async_cb_handler = FALSE;
    return 0;
};

int32_t dapls_ib_release (void)
{
    return 0;
};

/*
 * dapls_ib_enum_hcas
 *
 * Enumerate all HCAs on the system
 *
 * Input:
 * 	none
 *
 * Output:
 *	hca_names	Array of hca names
 *	total_hca_count	
 *
 * Returns:
 * 	DAT_SUCCESS
 *	DAT_INSUFFICIENT_RESOURCES
 *
 */
#ifdef __KDAPL__
#define MAX_NET_DEVICES         64
DAPL_HCA_NAME ib_dev_name_arr[MAX_NET_DEVICES];

DAT_RETURN
dapls_ib_enum_hcas(
    OUT  DAPL_HCA_NAME		**hca_names,
    OUT  DAT_COUNT			*total_hca_count,
    IN   const char                 *vendor )
{
    uint32_t i, count;
    struct net_device *dev;
    
    for (i = 0, count = 0; i < MAX_NET_DEVICES; i++) {

	dev = dev_get_by_index(i);
        
        if ((dev != NULL) && (!strncmp(dev->name, "ib", 2))) {
                strcpy(ib_dev_name_arr[count++], dev->name);
                //ITBD edding this put in order to free the dev
                // if need to hold it need to free it in the exit of the module
                dev_put(dev);
                continue;
        }

        if(dev)
            dev_put(dev);
    }

    *hca_names = ib_dev_name_arr;
    *total_hca_count = count;

    return(DAT_SUCCESS);
    
}

DAT_RETURN 
dapls_ib_get_instance_data(
    OUT char *instance,
    IN  DAPL_HCA_NAME hca_name
    )
{
    uint32_t status;
    uint8_t dev_gid[16];

    struct net_device *dev;
    struct ib_device *ca;
    struct ib_device_properties *props;

    tTS_IB_PORT hw_port;
    uint16_t pkey;

    if(strncmp(hca_name, "ib", 2))
       return(DAT_INTERNAL_ERROR);
    
    dev = dev_get_by_name(hca_name);
    if(!dev)
        return(DAT_INTERNAL_ERROR);
    
    status = ipoib_device_handle(dev, &ca, &hw_port, dev_gid, &pkey);
    if (status) 
    {
        dev_put(dev);
        return(DAT_INTERNAL_ERROR);
    }

    props = dapl_os_alloc(sizeof(struct ib_device_properties));
    if (props == NULL) {
        dev_put(dev);
        return(DAT_INTERNAL_ERROR);
    }

    status = ca->device_query(ca, props);
    if (status)
    {
        dev_put(dev);
        return(DAT_INTERNAL_ERROR);
    }

    sprintf(instance, "%s %d", props->name, hw_port);

    printk("ETBD: %s %d\n", props->name, hw_port);

    dev_put(dev);
    dapl_os_free(props,sizeof(struct ib_device_properties));
    return(DAT_SUCCESS);
}
#else
DAT_RETURN
dapls_ib_enum_hcas(
    OUT  DAPL_HCA_NAME		**hca_names,
    OUT  DAT_COUNT			*total_hca_count,
    IN   const char                 *vendor )
{
    return(DAT_INTERNAL_ERROR);
} 
 
#endif









/*
 * dapls_ib_open_hca
 *
 * Open HCA
 *
 * Input:
 *      *hca_name         pointer to provider device name
 *      *ib_hca_handle_p  pointer to provide HCA handle
 *
 * Output:
 *      none
 *
 * Return:
 *      DAT_SUCCESS
 *      DAT_INSUFFICIENT_RESOURCES
 *
 */
DAT_RETURN dapls_ib_open_hca (
	IN   char               *name,
	IN   DAPL_HCA		       *hca_ptr)
{
    VAPI_ret_t                     vapi_ret;
    DAT_RETURN                     dat_ret;

    uint8_t             ip_addr[16] = { 0 };
#ifdef __KERNEL__
    struct ib_pd* g_pd;
#endif

    struct ia_prov_data  prov_data = {
        .ib_khca_handle = VAPI_INVAL_HNDL,
        .ts_hca_handle = 0,
        .global_pd =0,
        .lid = 0,
        .node_GUID = { 0 },
        .gid = { 0 }
    };

    hca_ptr->ib_hca_handle = VAPI_INVAL_HNDL;

    

    /* register with the HW */
    vapi_ret = EVAPI_get_hca_hndl(name, &hca_ptr->ib_hca_handle);
    if (vapi_ret) 
    {
        dat_ret = dapl_ib_status_convert (vapi_ret);
        goto bail_out;
    }

    dat_ret = ib_get_khca_hndl(name, &prov_data.ib_khca_handle, &prov_data.ts_hca_handle);
    if(dat_ret)
        goto bail_out;
#ifdef __KERNEL__
    dat_ret = ib_pd_create(prov_data.ts_hca_handle,NULL,&g_pd);
    prov_data.global_pd = g_pd;
    if(dat_ret)
        goto bail_out;
#endif
    dat_ret = dapls_ib_get_hca_ids(hca_ptr->ib_hca_handle, hca_ptr->port_num, 
                                   prov_data.gid, &prov_data.lid, prov_data.node_GUID);
    if(dat_ret)
        goto bail_out;


    dat_ret = ib_get_hca_ipaddr(prov_data.gid, (u_int32_t*)&ip_addr[12]);
    if (dat_ret != DAT_SUCCESS)
        goto bail_out;



    dat_ret = ib_ats_set_ipaddr(prov_data.gid, prov_data.ts_hca_handle, 
			       hca_ptr->port_num, ip_addr);
    if(dat_ret != DAT_SUCCESS) {
        dat_ret = DAT_ERROR(DAT_INSUFFICIENT_RESOURCES,DAT_RESOURCE_DEVICE);
        goto bail_out;
    }
    

    hca_ptr->prov_data = dapl_os_alloc(sizeof(struct ia_prov_data));
    if(hca_ptr->prov_data == NULL)
    {
        dat_ret = DAT_ERROR (DAT_INSUFFICIENT_RESOURCES,DAT_RESOURCE_MEMORY);
        goto bail_out;
    }


                
    


    memcpy(hca_ptr->prov_data, &prov_data, sizeof(prov_data));


    return DAT_SUCCESS;

bail_out:
#ifdef __KERNEL__
    if (prov_data.global_pd != 0)
        ib_pd_destroy((struct ib_pd*)(prov_data.global_pd));
#endif    
    if(hca_ptr->ib_hca_handle != VAPI_INVAL_HNDL)
        EVAPI_release_hca_hndl (hca_ptr->ib_hca_handle);
    hca_ptr->ib_hca_handle = VAPI_INVAL_HNDL;
    if (prov_data.ib_khca_handle != VAPI_INVAL_HNDL)
        ib_release_khca_hndl(prov_data.ib_khca_handle);
    prov_data.ib_khca_handle = VAPI_INVAL_HNDL;


    dapl_dbg_log (DAPL_DBG_TYPE_ERR, "failed opening device\n");
    return dat_ret;

    

}



/*
 * dapls_ib_close_hca
 *
 * Open HCA
 *
 * Input:
 *      ib_hca_handle   provide HCA handle
 *
 * Output:
 *      none
 *
 * Return:
 *      DAT_SUCCESS
 *      DAT_INSUFFICIENT_RESOURCES
 *
 */
DAT_RETURN dapls_ib_close_hca (
	IN   DAPL_HCA		       *hca_ptr)
{
    VAPI_ret_t   ib_status;
    int          status;
#ifdef __KERNEL__
    ib_pd_destroy((struct ib_pd*)(hca_ptr->prov_data->global_pd));
#endif
    ib_status = EVAPI_release_hca_hndl (hca_ptr->ib_hca_handle);

    if (ib_status != VAPI_OK)
        return dapl_ib_status_convert (ib_status);

    status = ib_release_khca_hndl(hca_ptr->prov_data->ib_khca_handle);


    dapl_os_free(hca_ptr->prov_data, sizeof(struct ia_prov_data));
    

    if (status)
        return DAT_ERROR (DAT_INSUFFICIENT_RESOURCES,DAT_RESOURCE_MEMORY);
    else
        return DAT_SUCCESS;
}


/*
 * dapl_ib_cq_alloc
 *
 * Alloc a CQ
 *
 * Input:
 *	ia_handle		IA handle
 *	evd_ptr			pointer to EVD struct
 *	cqlen			minimum QLen
 *
 * Output:
 * 	none
 *
 * Returns:
 * 	DAT_SUCCESS
 *	DAT_INSUFFICIENT_RESOURCES
 *
 */
DAT_RETURN
dapls_ib_cq_alloc (
	IN  DAPL_IA		*ia_ptr,
	IN  DAPL_EVD		*evd_ptr,
	IN  DAT_COUNT		*cqlen)
{
    VAPI_ret_t             ib_status;
    VAPI_cqe_num_t         actual_ib_qlen;

    ib_status = VAPI_create_cq (
                        ia_ptr->hca_ptr->ib_hca_handle,
                        *cqlen,
                        &evd_ptr->ib_cq_handle,
                        &actual_ib_qlen);

    if (ib_status == VAPI_OK)
    {
	*cqlen = actual_ib_qlen;
    } else
        return dapl_ib_status_convert (ib_status);


    return DAT_SUCCESS;
}

/*
 * dapl_ib_cq_free
 *
 * Alloc a CQ
 *
 * Input:
 *	ia_handle		IA handle
 *	evd_ptr			pointer to EVD struct
 *	cqlen			minimum QLen
 *
 * Output:
 * 	none
 *
 * Returns:
 * 	DAT_SUCCESS
 *	DAT_INVALID_HANDLE
 *	DAT_INSUFFICIENT_RESOURCES
 *
 */
DAT_RETURN
dapls_ib_cq_free (
	IN  DAPL_IA		*ia_ptr,
	IN  DAPL_EVD		*evd_ptr)
{
    VAPI_ret_t             ib_status;

#ifdef __KDAPL__
    int                    status;

//   EVAPI_clear_comp_eventh will be done in the wait_obj_destroy
        status = EVAPI_clear_comp_eventh(ia_ptr->hca_ptr->ib_hca_handle, evd_ptr->ib_cq_handle);
        
	if (status)
	{
	    dapl_dbg_log (DAPL_DBG_TYPE_ERR,
			    " EVAPI_clear_comp_eventh() failed\n");
	    return DAT_ERROR (DAT_INSUFFICIENT_RESOURCES,DAT_RESOURCE_MEMORY);
    }
#endif
    if (evd_ptr->ib_cq_handle != IB_INVALID_HANDLE)
    {
/*   EVAPI_clear_comp_eventh will be done in the wait_obj_destroy
        status = EVAPI_clear_comp_eventh(ia_ptr->hca_ptr->ib_hca_handle, evd_ptr->ib_cq_handle);
        
	if (status)
	{
	    dapl_dbg_log (DAPL_DBG_TYPE_ERR,
			    " EVAPI_clear_comp_eventh() failed\n");
	    return DAT_ERROR (DAT_INSUFFICIENT_RESOURCES,DAT_RESOURCE_MEMORY);
        */
	}
	else
	{
	    evd_ptr->ib_comp_handle = IB_INVALID_HANDLE;	
    }
    
    ib_status = VAPI_destroy_cq(ia_ptr->hca_ptr->ib_hca_handle, evd_ptr->ib_cq_handle);
    
    return dapl_ib_status_convert (ib_status);
}



DAT_RETURN
dapls_ib_cq_resize (
        IN  DAPL_IA             *ia_ptr,
        IN  DAPL_EVD            *evd_ptr,
        IN  DAT_COUNT           *cqlen)
{
    VAPI_ret_t             ib_status;
    VAPI_cqe_num_t         actual_ib_qlen;

    ib_status = VAPI_resize_cq (
                                ia_ptr->hca_ptr->ib_hca_handle,
                                evd_ptr->ib_cq_handle,
                                *cqlen,
                                &actual_ib_qlen);

    if (ib_status) {
        dapl_dbg_log (DAPL_DBG_TYPE_ERR, " VAPI_resize_cq() failed: %d\n", ib_status);
        return dapl_ib_status_convert (ib_status);
    }

    *cqlen = actual_ib_qlen;

    return DAT_SUCCESS;

}

/*
 * dapl_set_cq_notify
 *
 * Set up CQ completion notifications
 *
 * Input:
 *	ia_handle		IA handle
 *	evd_ptr			pointer to EVD struct
 *
 * Output:
 * 	none
 *
 * Returns:
 * 	DAT_SUCCESS
 *	DAT_INVALID_HANDLE
 *	DAT_INSUFFICIENT_RESOURCES
 *
 */
DAT_RETURN
dapls_set_cq_notify (
	IN  DAPL_IA		*ia_ptr,
	IN  DAPL_EVD		*evd_ptr)
{
    VAPI_ret_t             ib_status;

    ib_status = VAPI_req_comp_notif ( 
                         ia_ptr->hca_ptr->ib_hca_handle,
                         evd_ptr->ib_cq_handle,
                         VAPI_NEXT_COMP );

    return dapl_ib_status_convert (ib_status);
}


/*
 * dapls_ib_cqd_create
 *
 * Set up CQ notification event thread
 *
 * Input:
 *	hca_handle		HCA handle
 *
 * Output:
 * 	none
 *
 * Returns:
 * 	DAT_SUCCESS
 *	DAT_INVALID_HANDLE
 *	DAT_INSUFFICIENT_RESOURCES
 *
 */
DAT_RETURN
dapls_ib_cqd_create (
	IN  DAPL_HCA		*hca_ptr);

DAT_RETURN
dapls_ib_cqd_create (
	IN  DAPL_HCA		*hca_ptr)
{

    // hca_ptr->ib_cqd_handle = IB_INVALID_HANDLE;
    return DAT_SUCCESS;
}


/*
 * dapl_cqd_destroy
 *
 * Destroy CQ notification event thread
 *
 * Input:
 *	hca_handle		HCA handle
 *
 * Output:
 * 	none
 *
 * Returns:
 * 	DAT_SUCCESS
 *	DAT_INVALID_HANDLE
 *	DAT_INSUFFICIENT_RESOURCES
 *
 */
DAT_RETURN
dapls_ib_cqd_destroy (
	IN  DAPL_HCA		*hca_ptr);

DAT_RETURN
dapls_ib_cqd_destroy (
	IN  DAPL_HCA		*hca_ptr)
{
    VAPI_ret_t    ib_status;


    if ( hca_ptr->ib_trans.null_ib_cq_handle  != IB_INVALID_HANDLE )
    {
	(void) VAPI_destroy_cq (
				hca_ptr->ib_hca_handle,
				hca_ptr->ib_trans.null_ib_cq_handle );
    }

    ib_status = VAPI_OK;

    /*
     * EBD - field doesn't exist in current implementaiton
     *
     */
    //hca_ptr->ib_cqd_handle = IB_INVALID_HANDLE;

    if (tavor_async_cb.set_async_cb_handler == TRUE)
    {
#if 0
	ib_status = VAPI_set_async_event_handler (
                              hca_ptr->ib_hca_handle,
                              NULL,
                              NULL);
#endif	
	if (ib_status == VAPI_OK)
	{
	    tavor_async_cb.set_async_cb_handler = FALSE;
	}
    }

    return dapl_ib_status_convert (ib_status);
}


/*
 * dapl_ib_pd_alloc
 *
 * Alloc a PD
 *
 * Input:
 *	ia_handle		IA handle
 *	PZ_ptr			pointer to PZEVD struct
 *
 * Output:
 * 	none
 *
 * Returns:
 * 	DAT_SUCCESS
 *	DAT_INSUFFICIENT_RESOURCES
 *
 */
DAT_RETURN
dapls_ib_pd_alloc (
	IN  DAPL_IA 		*ia,
	IN  DAPL_PZ 		*pz)
{
    VAPI_pd_hndl_t 	pd;
    VAPI_ret_t 		ib_status;

    ib_status = VAPI_alloc_pd (
                      ia->hca_ptr->ib_hca_handle, 
                      &pd);

    if (ib_status != VAPI_OK)
    {
	dapl_dbg_log (DAPL_DBG_TYPE_ERR, 
                      " error code return = %d\n", 
		      ib_status);
	return dapl_ib_status_convert (ib_status);
    }

    pz->pd_handle = pd;

    return (DAT_SUCCESS);
}


/*
 * dapl_ib_pd_free
 *
 * Free a PD
 *
 * Input:
 *	ia_handle		IA handle
 *	PZ_ptr			pointer to PZ struct
 *
 * Output:
 * 	none
 *
 * Returns:
 * 	DAT_SUCCESS
 *	DAT_INSUFFICIENT_RESOURCES
 *
 */
DAT_RETURN
dapls_ib_pd_free (
	IN  DAPL_PZ 		*pz)
{
    VAPI_ret_t 		ib_status;

    ib_status = VAPI_dealloc_pd (
                  pz->header.owner_ia->hca_ptr->ib_hca_handle, 
                  pz->pd_handle);

    if (ib_status != VAPI_OK)
    {
	dapl_dbg_log (DAPL_DBG_TYPE_ERR, 
                      " error code return = %d\n", 
		      ib_status);
	return dapl_ib_status_convert (ib_status);
    }

    pz->pd_handle = IB_INVALID_HANDLE;

    return (DAT_SUCCESS);
}


/*
 * dapl_ib_mr_register
 *
 * Register a virtual memory region
 *
 * Input:
 *	ia_handle		IA handle
 *	lmr			pointer to dapl_lmr struct
 *	virt_addr		virtual address of beginning of mem region
 *	length			length of memory region
 *
 * Output:
 * 	none
 *
 * Returns:
 * 	DAT_SUCCESS
 *	DAT_INSUFFICIENT_RESOURCES
 *
 */
DAT_RETURN
dapls_ib_mr_register (
	IN  DAPL_IA 		*ia,
        IN  DAPL_LMR		*lmr,
	IN  DAT_PVOID		virt_addr,
	IN  DAT_VLEN		length,
	IN  DAT_MEM_PRIV_FLAGS  privileges)
{
    VAPI_ret_t 		ib_status;
    VAPI_mr_hndl_t 	mr_handle;
    DAT_RETURN      dat_status = DAT_SUCCESS;
    VAPI_mr_t       *mr_prop_in = NULL;
    VAPI_mr_t       *mr_prop_out = NULL;
    mr_prop_in = dapl_os_alloc (sizeof(VAPI_mr_t));
    if ( mr_prop_in == NULL )
    {
    dat_status = DAT_ERROR (DAT_INSUFFICIENT_RESOURCES, DAT_RESOURCE_MEMORY);
    goto bail;
    }
    mr_prop_out = dapl_os_alloc (sizeof(VAPI_mr_t));
    if ( mr_prop_out == NULL )
    {
    dat_status = DAT_ERROR (DAT_INSUFFICIENT_RESOURCES, DAT_RESOURCE_MEMORY);
    goto bail;
    }
    mr_prop_in->type    = VAPI_MR;

    mr_prop_in->pd_hndl = ((DAPL_PZ *)lmr->param.pz_handle)->pd_handle;
    mr_prop_in->start   = (VAPI_virt_addr_t) ((uintptr_t) virt_addr);
    mr_prop_in->size    = length;
    mr_prop_in->acl     = dapls_ib_convert_mem_privileges (privileges);
    mr_prop_in->acl    |= VAPI_EN_MEMREG_BIND;

    dapl_dbg_log (DAPL_DBG_TYPE_UTIL, " access_ctr = 0x%x\n", mr_prop_in->acl);

    mr_prop_in->l_key   = 0;
    mr_prop_in->r_key   = 0;

    ib_status = VAPI_register_mr (
                     ia->hca_ptr->ib_hca_handle,
                     mr_prop_in,
                     &mr_handle,
                     mr_prop_out);

    if (ib_status != VAPI_OK)
    {
	dapl_dbg_log (DAPL_DBG_TYPE_ERR, 
		      " error code return = %d\n", 
		      ib_status);
	dat_status =  dapl_ib_status_convert (ib_status);
    goto bail;
    }

    lmr->param.lmr_context = mr_prop_out->l_key;
    if ((mr_prop_in->acl & VAPI_EN_REMOTE_READ) || 
        (mr_prop_in->acl & VAPI_EN_REMOTE_WRITE)) 
    {
    lmr->param.rmr_context = mr_prop_out->r_key;
    } 
    lmr->param.registered_size = length;
    lmr->param.registered_address = (DAT_VADDR) (uintptr_t) virt_addr;
    lmr->mr_handle         = mr_handle;
bail:
    if (mr_prop_in != NULL) {
        dapl_os_free(mr_prop_in,0);
    }
    if (mr_prop_out != NULL) {
        dapl_os_free(mr_prop_out,0);
    }
    return dat_status;
}



#ifdef __KDAPL__

DAT_RETURN dapls_ib_mr_register_physical (
		IN  DAPL_IA                     *ia_ptr,
		INOUT  DAPL_LMR                 *lmr,
		IN  DAT_PVOID                   phys_addr,
		IN  DAT_VLEN                    length,
		IN  DAT_MEM_PRIV_FLAGS          privileges)
{
    VAPI_ret_t 		    ib_status;
    VAPI_mr_hndl_t 	    mr_handle;
    DAT_RETURN      dat_status = DAT_SUCCESS;
    VAPI_mr_t       *mr_prop_in = NULL;
    VAPI_mr_t       *mr_prop_out = NULL;
    VAPI_phy_buf_t      *buf_list;
    
    buf_list = phys_addr;
    
    if (NULL == buf_list) {
        return (DAT_INVALID_PARAMETER);
    }
    mr_prop_in = dapl_os_alloc (sizeof(VAPI_mr_t));
    if ( mr_prop_in == NULL )
    {
    dat_status = DAT_ERROR (DAT_INSUFFICIENT_RESOURCES, DAT_RESOURCE_MEMORY);
    goto bail;
    }
    mr_prop_out = dapl_os_alloc (sizeof(VAPI_mr_t));
    if ( mr_prop_out == NULL )
    {
    dat_status = DAT_ERROR (DAT_INSUFFICIENT_RESOURCES, DAT_RESOURCE_MEMORY);
    goto bail;
    }
    dapl_os_memzero(mr_prop_in, sizeof(VAPI_mr_t));
    mr_prop_in->pbuf_list_len = length;
    mr_prop_in->pbuf_list_p = buf_list;
    mr_prop_in->iova_offset = buf_list[0].start & PAGE_LIMIT_64;
    mr_prop_in->type = VAPI_MPR;
    mr_prop_in->pd_hndl = ((DAPL_PZ *)lmr->param.pz_handle)->pd_handle;
    mr_prop_in->start = buf_list[0].start;
    PAGE_ALIGN_DOWN_64(buf_list[0].start);
    PAGE_ALIGN_UP_64 (buf_list[0].size);
    if (length>1) {
        PAGE_ALIGN_UP_64 (buf_list[length-1].size);
    }
    mr_prop_in->size = lmr->param.length; 
    mr_prop_in->acl = (VAPI_mrw_acl_t) dapls_ib_convert_mem_privileges (privileges);
    mr_prop_in->acl |= VAPI_EN_MEMREG_BIND;
    ib_status = VAPI_register_mr (
                     ia_ptr->hca_ptr->ib_hca_handle,
                     mr_prop_in,
                     &mr_handle,
                     mr_prop_out);

    if (ib_status != VAPI_OK)
    {
        dapl_dbg_log (DAPL_DBG_TYPE_ERR, 
                      "(VAPI_MPR) returned error %d %s "
                      "[size=0x%llx acl=0x%x pd_hndl=0x%x pbuf_list_len=0x%llx iova_offset=0x%llx]\n", 
                      ib_status,VAPI_strerror(ib_status),
                      mr_prop_in->size, mr_prop_in->acl, mr_prop_in->pd_hndl, 
                      mr_prop_in->pbuf_list_len, mr_prop_in->iova_offset);
        dat_status = (dapl_ib_status_convert (ib_status));
        goto bail;
    }
    dapl_dbg_log(DAPL_DBG_TYPE_UTIL, 
                 "dapls_ib_mr_register_physical(%p %d) got handle %x\n", 
                 phys_addr,(DAT_UINT32)length,mr_handle); 

    lmr->param.lmr_context        = mr_prop_out->l_key;
    lmr->param.rmr_context        = mr_prop_out->r_key;
    lmr->param.registered_size    = mr_prop_out->size;
    lmr->param.registered_address = mr_prop_out->start;
    lmr->mr_handle                = mr_handle;
bail:
    if (mr_prop_in != NULL) {
        dapl_os_free(mr_prop_in,0);
    }
    if (mr_prop_out != NULL) {
        dapl_os_free(mr_prop_out,0);
    }
    return dat_status;
}

/* the function copies the buflist to a page array vector,
   which is the vector format VAPI expects for FMR.
   The the fuction returns the number of elements writen 
   to page_arr.returns 0 if error. */
DAT_RETURN dapls_span_2_addr_list (
                        VAPI_phy_addr_t *page_arr,
                        VAPI_phy_buf_t  *buf_list,
                        DAT_UINT32      list_length)
{        
    DAT_UINT32              ii, jj, kk;
    VAPI_phy_addr_t         size, start_addr, end_addr, cur_addr;
    
    kk = 0;
    for (ii=0; ii<list_length; ii++) {
        start_addr = buf_list[ii].start ;
        size = buf_list[ii].size;
        end_addr = start_addr + size;
        
        jj = 0;
        do {   
            cur_addr = start_addr + (PAGE_SIZE_64 * (jj++));
            page_arr[kk] = cur_addr;
            PAGE_ALIGN_DOWN_64 (page_arr[kk]);
            dapl_dbg_log(DAPL_DBG_TYPE_UTIL, "%s: page_arr[%d]=0x%llx (cur_addr=0x%llx end_add=0x%llx)\n", 
                 __FUNCTION__, (kk), page_arr[kk], cur_addr, end_addr);
            kk++;
        } while (((cur_addr&PAGE_MASK_64)+PAGE_SIZE_64) < (end_addr&PAGE_MASK_64));
        
    }
    return (kk);
}

DAT_RETURN dapls_ib_fmr (
		IN  DAPL_IA                     *ia_ptr,
		INOUT  DAPL_LMR                 *lmr,
		IN  DAT_PVOID                   phys_addr,
		IN  DAT_VLEN                    length,
		IN  DAT_MEM_PRIV_FLAGS          privileges)
{
    mem_err_t 		    mem_status;
    DAT_RETURN          rc;
    int                 ii;
    mem_fmr_hndl_t      mem_fmr_hndl;
    fmr_hndl_impl       *fmr_hndl;
    VAPI_lkey_t         l_key;
    VAPI_rkey_t         r_key;
    fmr_ctrl_struct     *mem_fmr_ctrl;
    int                 num_pages;
    VAPI_phy_buf_t      *buf_list;
    VAPI_phy_addr_t     total_size=0, end_addr;

    buf_list = (VAPI_phy_buf_t *)phys_addr;

    mem_fmr_ctrl = &((DAPL_PZ *)lmr->param.pz_handle)->mem_fmr_ctrl;

    num_pages = 0;

    for (ii=0;ii<length;ii++) {
        end_addr = buf_list[ii].start + buf_list[ii].size;
        total_size = buf_list[ii].size + (buf_list[ii].start & PAGE_LIMIT_64) + 
                    (PAGE_LIMIT_64 - ((end_addr-1) & PAGE_LIMIT_64));
        num_pages += (total_size / PAGE_SIZE);
    }
    
    mem_fmr_get_fmr(num_pages, mem_fmr_ctrl, &mem_fmr_hndl);

    fmr_hndl = (fmr_hndl_impl *)mem_fmr_hndl;
    

    dapl_dbg_log(DAPL_DBG_TYPE_UTIL, "%s: num_pages=%d -> fmr_hndl=%p  "
                 "mapped_pages=%d hca_hdnl=0x%x lmr_len=0x%llx", 
                 __FUNCTION__, num_pages, mem_fmr_hndl, fmr_hndl->mapped_pages, 
                 fmr_hndl->hca_hdnl, lmr->param.length);
    
    dapl_os_memzero ((void *)fmr_hndl->page_arr, FMR_MAX_OPTIONS_TO_MAP * sizeof(u_int64_t));
    
    rc = dapls_span_2_addr_list (fmr_hndl->page_arr, (VAPI_phy_buf_t *)phys_addr, length);
    
    if (rc != num_pages) {
        dapl_dbg_log (DAPL_DBG_TYPE_ERR, 
                      "%s: Error in copying struct iovec vector to page array (wrote %d calculated %d)\n",
                      __FUNCTION__, rc, num_pages);
        return (DAT_INTERNAL_ERROR);
    }

    fmr_hndl->data_size = lmr->param.length;
    PAGE_ALIGN_UP_64(fmr_hndl->data_size);
    fmr_hndl->mapped_pages_req = num_pages;
    
    mem_status = mem_fmr_map_fmr(mem_fmr_hndl,
                                 &l_key,
                                 &r_key);
    
    if (mem_status != VAPI_OK) {
        dapl_dbg_log (DAPL_DBG_TYPE_ERR, 
                      "%s: mem_fmr_map_fmr returned error %d\n", 
                      __FUNCTION__, mem_status);

        return (DAT_INTERNAL_ERROR);
    }
    
    lmr->param.lmr_context        = l_key;
    lmr->param.rmr_context        = r_key;
    lmr->param.registered_size    = fmr_hndl->data_size;
    lmr->param.registered_address = buf_list[0].start; // BUGBUG: doesn't support the case where it is not zbva (i.e -  registered_address != start_addr) - check with mellanox...
    lmr->mr_handle                = fmr_hndl->fmr_hndl;
    lmr->mem_fmr_hndl             = mem_fmr_hndl; 

    return (DAT_SUCCESS);
}


#endif // __KDAPL__



/*
 * dapl_ib_mr_deregister
 *
 * Free a memory region
 *
 * Input:
 *	lmr			pointer to dapl_lmr struct
 *
 * Output:
 * 	none
 *
 * Returns:
 * 	DAT_SUCCESS
 *	DAT_INSUFFICIENT_RESOURCES
 *
 */
DAT_RETURN
dapls_ib_mr_deregister (
	IN  DAPL_LMR		*lmr)
{

    VAPI_ret_t		ib_status;

#ifdef __KDAPL__
    mem_fmr_hndl_t      mem_fmr_hndl;
    fmr_ctrl_struct     *mem_fmr_ctrl;
    int                 num_success_free;
    
    mem_fmr_ctrl = &((DAPL_PZ *)lmr->param.pz_handle)->mem_fmr_ctrl;

    if (DAT_MEM_TYPE_PLATFORM == lmr->param.mem_type) 
    {                                 
        if (active_fmr)
        {   
            mem_fmr_hndl = lmr->mem_fmr_hndl;

            dapl_dbg_log (DAPL_DBG_TYPE_UTIL, 
                              "%s: mem_fmr_hndl=%p\n", 
                              __FUNCTION__, mem_fmr_hndl);

            mem_fmr_free_fmr (1, mem_fmr_hndl,
                              mem_fmr_ctrl,
                              &num_success_free);

            dapl_dbg_log (DAPL_DBG_TYPE_UTIL, 
                              "%s: fmr num_success_free=%d\n", 
                              __FUNCTION__, num_success_free);
        }
        else
        {
            return (DAT_INVALID_PARAMETER);
        }
    }
    else // DAT_MEM_TYPE_PHYSICAL || DAT_MEM_TYPE_IA
#endif //__KDAPL__
    {
        ib_status = VAPI_deregister_mr (
                       lmr->header.owner_ia->hca_ptr->ib_hca_handle, 
                       lmr->mr_handle);
            
        if (ib_status != VAPI_OK)
        {
            dapl_dbg_log (DAPL_DBG_TYPE_ERR, 
                          " VAPI_deregister_mr(%d) returned error %d %s\n", 
                          ib_status,VAPI_strerror(ib_status));
            return dapl_ib_status_convert (ib_status);
        }
#ifndef __KDAPL__
    /* If this region was SHARED_VIRTUAL, there will be a backing
     * memory region that needs to be released as well
     */
    if (lmr->ib_trans.mr_handle != 0)
    {
	(void) VAPI_deregister_mr (
                       lmr->header.owner_ia->hca_ptr->ib_hca_handle, 
		       lmr->ib_trans.mr_handle);
	lmr->ib_trans.mr_handle = 0;
    }

#endif

    dapl_dbg_log(DAPL_DBG_TYPE_UTIL,
                     "dapls_ib_mr_deregister(%x)\n",lmr->mr_handle); 
    }

    lmr->param.lmr_context = 0;
    lmr->mr_handle         = IB_INVALID_HANDLE;

    return (DAT_SUCCESS);
}

/*
 * dapl_ib_mr_register_shared
 *
 * Register a virtual memory region
 * VAPI does not have an API to directly register memory as shared
 * memory. Instead, we must register memory, then register that memory
 * as shared. This gives us a new mr_handle, as well as a new l_key and
 * r_key, so we need to save the original handle until it can be
 * disposed of when the user does an lmr_free. The l_key and r_key are
 * not available and will disappear when we deallocate the mr_handle, so
 * we don't save them.
 *
 * Input:
 *	ia_handle		IA handle
 *	lmr			pointer to dapl_lmr struct
 *	virt_addr		virtual address of beginning of mem region
 *	length			length of memory region
 *
 * Output:
 * 	none
 *
 * Returns:
 * 	DAT_SUCCESS
 *	DAT_INSUFFICIENT_RESOURCES
 *
 */

// ETBD - This is the original shared
///*                                                                              
// * dapl_ib_smr_register                                                         
// *                                                                              
// * Register a virtual memory region                                             
// *                                                                              
// * Input:                                                                       
// *      ia_handle               IA handle                                       
// *      lmr                     pointer to dapl_lmr struct                      
// *      virt_addr               virtual address of beginning of mem region      
// *      length                  length of memory region                         
// *                                                                              
// * Output:                                                                      
// *      none                                                                    
// *                                                                              
// * Returns:                                                                     
// *      DAT_SUCCESS                                                             
// *      DAT_INSUFFICIENT_RESOURCES                                              
// *                                                                              
// */
//
#ifndef __KDAPL__
DAT_RETURN
dapls_ib_smr_register (
        IN  DAPL_IA             *ia,
        IN  DAPL_LMR            *lmr,
        IN  ib_mr_handle_t      orig_mr_handle,
        IN  DAT_PVOID           virt_addr,
        IN  DAT_VLEN            length,
        IN  DAT_MEM_PRIV_FLAGS  privileges);

DAT_RETURN
dapls_ib_smr_register (
        IN  DAPL_IA             *ia,
        IN  DAPL_LMR            *lmr,
        IN  ib_mr_handle_t      orig_mr_handle,
        IN  DAT_PVOID           virt_addr,
        IN  DAT_VLEN            length,
        IN  DAT_MEM_PRIV_FLAGS  privileges)
{
    VAPI_ret_t          ib_status;
    VAPI_mr_hndl_t      mr_handle;
    VAPI_mr_t           mr_prop_in, mr_prop_out;

    mr_prop_in.type    = VAPI_MSHAR;

    mr_prop_in.pd_hndl = ((DAPL_PZ *)lmr->param.pz_handle)->pd_handle;
    mr_prop_in.start   = (VAPI_virt_addr_t) ((uintptr_t) virt_addr);
    mr_prop_in.size    = length;
    mr_prop_in.acl     = dapls_ib_convert_mem_privileges (privileges);
    mr_prop_in.acl    |= VAPI_EN_MEMREG_BIND;

    dapl_dbg_log (DAPL_DBG_TYPE_UTIL, " access_ctr = 0x%x\n", mr_prop_in.acl);

    mr_prop_in.l_key   = 0;
    mr_prop_in.r_key   = 0;

    ib_status = VAPI_register_smr (
                                   ia->hca_ptr->ib_hca_handle,
                                   orig_mr_handle,
                                   &mr_prop_in,
                                   &mr_handle,
                                   &mr_prop_out);

    dapl_dbg_log (DAPL_DBG_TYPE_UTIL, 
		  "Registration: SMR start = %llx size = %llx handle = %x orig_handle = %x\n",
		  mr_prop_in.start, mr_prop_in.size, mr_handle, orig_mr_handle);
    
    if (ib_status != VAPI_OK)
    {
        dapl_dbg_log (DAPL_DBG_TYPE_ERR, " error code return = %d\n", ib_status);
        return dapl_ib_status_convert (ib_status);
    }

//    lmr->param.lmr_context = mr_prop_out.l_key;
//    lmr->mr_handle         = mr_handle;

    lmr->param.lmr_context = mr_prop_out.l_key;
    lmr->param.rmr_context = mr_prop_out.r_key;
    lmr->param.registered_size = length;
    lmr->param.registered_address = (DAT_VADDR) (uintptr_t) virt_addr;
    lmr->mr_handle         = mr_handle;

    return (DAT_SUCCESS);
}
#endif
///*
// * dapl_ib_mr_register_shared
// *
// * Register a virtual memory region
// *
// * Input:
// *	ia_handle		IA handle
// *	lmr			pointer to dapl_lmr struct
// *	virt_addr		virtual address of beginning of mem region
// *	length			length of memory region
// *
// * Output:
// * 	none
// *
// * Returns:
// * 	DAT_SUCCESS
// *	DAT_INSUFFICIENT_RESOURCES
// *
// */
DAT_RETURN
dapls_ib_mr_register_shared (
	IN  DAPL_IA 			*ia,
        IN  DAPL_LMR			*lmr,
	IN  DAT_MEM_PRIV_FLAGS		privileges)
{
    DAT_PVOID			virt_addr;
    VAPI_mr_hndl_t 		mr_handle;
    VAPI_ret_t			ib_status;
    VAPI_mr_hndl_t              new_mr_handle;

    DAT_RETURN      dat_status = DAT_SUCCESS;
    VAPI_mr_t       *mr_prop_in = NULL;
    VAPI_mr_t       *mr_prop_out = NULL;
    
    mr_prop_in = dapl_os_alloc (sizeof(VAPI_mr_t));
    if ( mr_prop_in == NULL )
    {
    dat_status = DAT_ERROR (DAT_INSUFFICIENT_RESOURCES, DAT_RESOURCE_MEMORY);
    goto bail;
    }
    mr_prop_out = dapl_os_alloc (sizeof(VAPI_mr_t));
    if ( mr_prop_out == NULL )
    {
    dat_status = DAT_ERROR (DAT_INSUFFICIENT_RESOURCES, DAT_RESOURCE_MEMORY);
    goto bail;
    }
    dapl_os_memzero(mr_prop_in, sizeof(VAPI_mr_t));
    mr_handle = (VAPI_mr_hndl_t) lmr->mr_handle;

    virt_addr = (u_int8_t *) 
	(uintptr_t) dapl_mr_get_address (lmr->param.region_desc,
					lmr->param.mem_type);
return 
    mr_prop_in->type    = VAPI_MSHAR;
    mr_prop_in->pd_hndl = ((DAPL_PZ *)lmr->param.pz_handle)->pd_handle;
    mr_prop_in->start   = (uintptr_t) virt_addr;

    mr_prop_in->acl     = dapls_ib_convert_mem_privileges (privileges);

    ib_status = VAPI_register_smr (
                      ia->hca_ptr->ib_hca_handle,
                      mr_handle,
                      mr_prop_in,
                      &new_mr_handle,
                      mr_prop_out);

    if (ib_status != VAPI_OK)
    {
	dapl_dbg_log (DAPL_DBG_TYPE_ERR,
		      " error code return = %d\n", 
		      ib_status);
	dat_status = dapl_ib_status_convert (ib_status);
    goto bail;
    }

    lmr->mr_handle         = new_mr_handle;
    lmr->param.lmr_context = (uint32_t) mr_prop_out->l_key;
bail:
    if (mr_prop_in != NULL) {
        dapl_os_free(mr_prop_in,0);
    }
    if (mr_prop_out != NULL) {
        dapl_os_free(mr_prop_out,0);
    }
    return dat_status;
}

/*
 * dapls_ib_mw_alloc
 *
 * Bind a protection domain to a memory window
 *
 * Input:
 *	rmr			Initialized rmr to hold binding handles
 *
 * Output:
 * 	none
 *
 * Returns:
 * 	DAT_SUCCESS
 *	DAT_INSUFFICIENT_RESOURCES
 *
 */
DAT_RETURN
dapls_ib_mw_alloc (
	IN  DAPL_RMR	 		*rmr)
{
    VAPI_ret_t	 	ib_status;
    VAPI_rkey_t 	r_key;
    VAPI_mw_hndl_t 	mw_handle;

    ib_status = VAPI_alloc_mw (
                  rmr->header.owner_ia->hca_ptr->ib_hca_handle,
                  ((DAPL_PZ *)rmr->param.pz_handle)->pd_handle,
                  &mw_handle,
                  &r_key);

    if (ib_status != VAPI_OK)
    {
	dapl_dbg_log (DAPL_DBG_TYPE_ERR,
		      " error code return = %d\n", 
		      ib_status);
	return dapl_ib_status_convert (ib_status);
    }

    rmr->mw_handle         = mw_handle;
    rmr->param.rmr_context = (DAT_RMR_CONTEXT) r_key;

    dapl_dbg_log (DAPL_DBG_TYPE_UTIL,
	          " created mw_handle = 0x%lx, r_key = 0x%x\n",
                  mw_handle, 
		  r_key);

    return (DAT_SUCCESS);
}


/*
 * dapls_ib_mw_free
 *
 * Release bindings of a protection domain to a memory window
 *
 * Input:
 *	rmr			Initialized rmr to hold binding handles
 *
 * Output:
 * 	none
 *
 * Returns:
 * 	DAT_SUCCESS
 *	DAT_INSUFFICIENT_RESOURCES
 *
 */
DAT_RETURN
dapls_ib_mw_free (
	IN  DAPL_RMR	 		*rmr)
{
    VAPI_ret_t 	ib_status;

    ib_status = VAPI_dealloc_mw (
                  rmr->header.owner_ia->hca_ptr->ib_hca_handle, 
                  rmr->mw_handle);

    if (ib_status != VAPI_OK)
    {
	dapl_dbg_log (DAPL_DBG_TYPE_ERR, 
		      " error code return = %d\n", 
		      ib_status);
	return dapl_ib_status_convert (ib_status);
    }

    rmr->param.rmr_context = 0;
    rmr->mw_handle         = IB_INVALID_HANDLE;

    return (DAT_SUCCESS);
}

/*
 * dapls_ib_mw_bind
 *
 * Bind a protection domain to a memory window
 *
 * Input:
 *	rmr			Initialized rmr to hold binding handles
 *
 * Output:
 * 	none
 *
 * Returns:
 * 	DAT_SUCCESS
 *	DAT_INSUFFICIENT_RESOURCES
 *
 */
DAT_RETURN
dapls_ib_mw_bind (
	IN  DAPL_RMR		*rmr,
	IN  DAPL_LMR		*lmr,
	IN  DAPL_EP		*ep,
	IN  DAPL_COOKIE		*cookie,
	IN  DAT_VADDR		virtual_address,
	IN  DAT_VLEN		length,
	IN  DAT_MEM_PRIV_FLAGS	mem_priv,
	IN  ib_bool_t		is_signaled)
{
    VAPI_ret_t       ib_status;
    VAPI_mw_bind_t   mw_bind_prop;
    VAPI_rkey_t      new_rkey;
    VAPI_comp_type_t comp_type;

    mw_bind_prop.start          = (VAPI_virt_addr_t) virtual_address;
    mw_bind_prop.size           = length;
    mw_bind_prop.acl            = (dapls_ib_convert_mem_privileges (mem_priv) |
	    			  VAPI_EN_REMOTE_ATOM);
    mw_bind_prop.mr_lkey        = (VAPI_lkey_t) lmr->param.lmr_context;

    comp_type = (is_signaled == FALSE) ? VAPI_UNSIGNALED : VAPI_SIGNALED;

    dapl_dbg_log (DAPL_DBG_TYPE_UTIL,
	          " access_ctr = 0x%x, comp_type = %d\n",
                  mw_bind_prop.acl, 
		  comp_type);

    ib_status = VAPI_OK;
    new_rkey = 0;

    dapl_dbg_log (DAPL_DBG_TYPE_UTIL, 
	    	  " BEFORE BIND old_rkey = 0x%x mw_handle = 0x%lx\n",
                 rmr->param.rmr_context, 
		 rmr->mw_handle);

    ib_status = VAPI_bind_mw (
		    rmr->pz->header.owner_ia->hca_ptr->ib_hca_handle,
		    rmr->mw_handle,
		    &mw_bind_prop,
		    ep->qp_handle,
		    (VAPI_wr_id_t) (uintptr_t) cookie,
		    comp_type,
		    &new_rkey);

    if (ib_status != VAPI_OK)
    {
	dapl_dbg_log (DAPL_DBG_TYPE_ERR, 
		      " error code return = %d\n", 
		      ib_status);
	return (dapl_ib_status_convert (ib_status));
    }

    rmr->param.rmr_context      = (DAT_RMR_CONTEXT) new_rkey;

    dapl_dbg_log (DAPL_DBG_TYPE_UTIL, " new_rkey = 0x%x\n", new_rkey);

    return (DAT_SUCCESS);
}

/*
 * dapls_ib_mw_unbind
 *
 * Unbind a protection domain from a memory window
 *
 * Input:
 *	rmr			Initialized rmr to hold binding handles
 *
 * Output:
 * 	none
 *
 * Returns:
 * 	DAT_SUCCESS
 *	DAT_INSUFFICIENT_RESOURCES
 *
 */
DAT_RETURN
dapls_ib_mw_unbind (
	IN  DAPL_RMR		*rmr,
	IN  DAPL_EP		*ep,
	IN  DAPL_COOKIE		*cookie,
	IN  ib_bool_t		is_signaled)
{
    VAPI_ret_t       ib_status;
    VAPI_mw_bind_t   mw_bind_prop;
    VAPI_rkey_t      new_rkey;
    VAPI_comp_type_t comp_type;

    mw_bind_prop.start          = 0;
    mw_bind_prop.size           = 0;
    mw_bind_prop.acl            = 0;
    mw_bind_prop.mr_lkey        = 0;

    comp_type = (is_signaled == FALSE) ? VAPI_UNSIGNALED : VAPI_SIGNALED;

    dapl_dbg_log (DAPL_DBG_TYPE_UTIL, 
	    	  " access_ctr = 0x%x, comp_type = %d\n",
                  mw_bind_prop.acl, 
		  comp_type);

    ib_status = VAPI_OK;
    new_rkey = 0;

    dapl_dbg_log (DAPL_DBG_TYPE_UTIL,
	          " BEFORE BIND old_rkey = 0x%x mw_handle = 0x%lx\n",
                  rmr->param.rmr_context, 
		  rmr->mw_handle);

    ib_status = VAPI_bind_mw (
		    rmr->pz->header.owner_ia->hca_ptr->ib_hca_handle,
		    rmr->mw_handle,
		    &mw_bind_prop,
		    ep->qp_handle,
		    (VAPI_wr_id_t) (uintptr_t) cookie,
		    comp_type,
		    &new_rkey);

    if (ib_status != VAPI_OK)
    {
	dapl_dbg_log (DAPL_DBG_TYPE_ERR,
		      " error code return = %d\n", 
		      ib_status);
	return (dapl_ib_status_convert (ib_status));
    }

    rmr->param.rmr_context      = (DAT_RMR_CONTEXT) new_rkey;

    dapl_dbg_log (DAPL_DBG_TYPE_UTIL, " new_rkey = 0x%x\n", new_rkey);

    return (DAT_SUCCESS);
}

///*
// * dapls_ib_ca_async_callback
// *
// * General asynchronous error callback of various events for the HCA
// *
// * Input:
// *	hca_handle		HCA handle
// *	event_record		async event record
// *	context			private context
// *
// * Output:
// * 	none
// *
// *
// * Returns:
// *
// */
//static void
//dapls_ib_ca_async_callback (
//        IN    VAPI_hca_hndl_t     hca_handle,
//        IN    VAPI_event_record_t *event_ptr,
//        IN    void                *context)
//{
//    DAPL_IA		*ia_ptr;
//    DAPL_EP		*ep_ptr;
//    DAPL_LLIST_ENTRY	*cur_ent;
//
//    ia_ptr = (DAPL_IA *) context;
//
//    dapl_os_assert (event_ptr != NULL);
//
//    dapl_dbg_log (DAPL_DBG_TYPE_CALLBACK,
//		  " get ASYNC ERROR = %d\n", event_ptr->type);
//
//    switch (event_ptr->type)
//    {
//	case VAPI_QP_PATH_MIGRATED:
//	case VAPI_QP_COMM_ESTABLISHED:
//	case VAPI_SEND_QUEUE_DRAINED:
//	case VAPI_LOCAL_WQ_INV_REQUEST_ERROR:
//	case VAPI_LOCAL_WQ_CATASTROPHIC_ERROR:
//	case VAPI_PATH_MIG_REQ_ERROR:
//	{
//	    if (tavor_async_cb.qp_async_error_cb)
//	    {
//		/*
//		 * Walk the EPs to match this QP, then invoke the
//		 * routine when we have the EP we need
//		 */
//		dapl_os_assert (!dapl_llist_is_empty (&ia_ptr->ep_list_head));
//		dapl_os_lock (&ia_ptr->header.lock);
//
//		cur_ent = (DAPL_LLIST_ENTRY *) dapl_llist_next_entry (
//					     &ia_ptr->ep_list_head,
//					     NULL );
//		ep_ptr = dapl_llist_peek_head (&cur_ent);
//		while ( cur_ent != NULL )
//		{
//		    if ( ep_ptr->qp_handle == event_ptr->modifier.qp_hndl )
//		    {
//			break;
//		    }
//		    cur_ent = (DAPL_LLIST_ENTRY *) dapl_llist_next_entry (
//					     &ia_ptr->ep_list_head,
//					     cur_ent );
//		    ep_ptr = dapl_llist_peek_head (&cur_ent);
//		}
//		dapl_os_unlock (&ia_ptr->header.lock);
//
//		dapl_os_assert ( ep_ptr != NULL );
//
//		(tavor_async_cb.qp_async_error_cb) (
//                                 hca_handle,
//                                 event_ptr->modifier.qp_hndl,
//                                 event_ptr,
//                                 (void *)ep_ptr);
//	    }
//	    break;
//	}
//	case VAPI_CQ_ERROR:
//	{
//	    if (tavor_async_cb.cq_async_error_cb)
//	    {
//		(tavor_async_cb.cq_async_error_cb) (
//                                 hca_handle,
//                                 event_ptr->modifier.cq_hndl,
//                                 event_ptr,
//                                 context);
//	    }
//	    break;
//	}
//	case VAPI_PORT_ERROR:
//	{
//	    if (tavor_async_cb.async_error_cb)
//	    {
//		(tavor_async_cb.async_error_cb) (
//                                               hca_handle,
//                                               event_ptr,
//                                               context);
//	    }
//	    break;
//	}
//	case VAPI_PORT_ACTIVE:
//	{
//	    // ETBD __ActivePort = event_ptr->modifier.port_num;
//            break;
//	}
//	default:
//	{
//	    dapl_dbg_log (DAPL_DBG_TYPE_ERR, 
//			  "UNKNOWN ASYNC ERROR\n");
//	    break;
//	}
//    }
//    return;
//}

/*
 * dapls_ib_setup_async_callback
 *
 * Set up an asynchronous callbacks of various kinds
 *
 * Input:
 *	ia_handle		IA handle
 *	handler_type		type of handler to set up
 *	callback_handle		handle param for completion callbacks
 *	callback		callback routine pointer
 *	context			argument for callback routine
 *
 * Output:
 * 	none
 *
 * Returns:
 * 	DAT_SUCCESS
 *	DAT_INSUFFICIENT_RESOURCES
 *	DAT_INVALID_PARAMETER
 *
 */
DAT_RETURN
dapls_ib_setup_async_callback (
	IN  DAPL_IA			*ia_ptr,
	IN  DAPL_ASYNC_HANDLER_TYPE	handler_type,
	IN  unsigned int		*callback_handle,
	IN  ib_async_handler_t		callback,
	IN  void			*context )
{
    VAPI_ret_t                  ib_status = VAPI_OK;
    int                         status;
    DAPL_EVD                    *evd_ptr;

    evd_ptr = (DAPL_EVD *) context;
    
    switch (handler_type)
    {
	case DAPL_ASYNC_UNAFILIATED:
	case DAPL_ASYNC_QP_ERROR:
	case DAPL_ASYNC_CQ_ERROR:
	{
#if defined (__linux__)
	    if (tavor_async_cb.set_async_cb_handler == FALSE)
	    {
		/*
		ib_status = VAPI_set_async_event_handler (
                                  ia_ptr->hca_ptr->ib_hca_handle,
                                  dapls_ib_ca_async_callback,
                                  context);
		*/		  
		if (ib_status == VAPI_OK)
		{
		    tavor_async_cb.set_async_cb_handler = TRUE;
		}
		else
		{
		    break;
		}
	    }

	    if (handler_type == DAPL_ASYNC_CQ_ERROR)
	    {
		tavor_async_cb.cq_async_error_cb = (async_cq_err_cb_t) callback;
	    }
	    else if (handler_type == DAPL_ASYNC_QP_ERROR)
	    {
		tavor_async_cb.qp_async_error_cb = (async_qp_err_cb_t) callback;
	    }
	    else
	    {
		tavor_async_cb.async_error_cb = (async_err_cb_t) callback;
	    }
#endif
	    break;
	}
	case DAPL_ASYNC_CQ_COMPLETION:
	{
        EVAPI_compl_handler_hndl_t comp_handle;

        /*
         * ETBD - Currently the handle return is identical to the cq handle
         * so we just check that it is true and move on
         *
         */
        status = EVAPI_set_comp_eventh(ia_ptr->hca_ptr->ib_hca_handle, 
                                       (VAPI_cq_hndl_t)callback_handle,
                                        (VAPI_completion_event_handler_t)callback,
                                        context, &comp_handle);

        
        if(status) 
	    {
            dapl_dbg_log (DAPL_DBG_TYPE_ERR, " EVAPI_set_comp_eventh() failed\n");
            return DAT_ERROR (DAT_INSUFFICIENT_RESOURCES,DAT_RESOURCE_MEMORY);
        }

        if(callback_handle != (unsigned int *)comp_handle)
        {
            dapl_dbg_log (DAPL_DBG_TYPE_ERR, " Completion handle should be eq to cq handle\n");
           return DAT_ERROR (DAT_INSUFFICIENT_RESOURCES,DAT_RESOURCE_ERROR_HANDLER);
        }

        return(DAT_SUCCESS);
        

	    break;
	}
	default:
	{
	    break;
	}
    }

    if (ib_status != VAPI_OK)
    {
	dapl_dbg_log (DAPL_DBG_TYPE_ERR, 
		      "dapls_ib_setup_asycn_cb return = %d\n", 
                      ib_status);
	return dapl_ib_status_convert (ib_status);
    }

    return DAT_SUCCESS;

}

/*
 * dapls_ib_query_hca
 *
 * Set up an asynchronous callbacks of various kinds
 *
 * Input:
 *	hca_handl		hca handle	
 *	ep_attr			attribute of the ep
 *
 * Output:
 * 	none
 *
 * Returns:
 * 	DAT_SUCCESS
 *	DAT_INVALID_PARAMETER
 */

DAT_RETURN dapls_ib_query_hca (
	IN  DAPL_HCA			*hca_ptr,
	OUT DAT_IA_ATTR	   		*ia_attr,
	OUT DAT_EP_ATTR			*ep_attr,
	OUT DAT_SOCK_ADDR6		*ip_addr )
{
    VAPI_hca_hndl_t   hca_handle;
    VAPI_ret_t        ib_status;
    VAPI_hca_cap_t    *ib_hca_cap_info      =NULL;
    VAPI_hca_vendor_t *ib_hca_vendor_info   =NULL;
    VAPI_hca_port_t   *ib_hca_port_info     =NULL;

    DAT_RETURN				dat_status = DAT_SUCCESS;

    DAT_BOOLEAN				ib_hca_cap_info_allocated = DAT_FALSE;
    DAT_BOOLEAN				ib_hca_vendor_info_allocated = DAT_FALSE;
    DAT_BOOLEAN				ib_hca_port_info_allocated   = DAT_FALSE;

    ib_hca_cap_info = dapl_os_alloc(sizeof(VAPI_hca_cap_t));
    if (ib_hca_cap_info == NULL) {
        dat_status =  DAT_ERROR (DAT_INSUFFICIENT_RESOURCES,DAT_RESOURCE_MEMORY);
        goto bail;
    }
    ib_hca_cap_info_allocated =DAT_TRUE;

    ib_hca_vendor_info = dapl_os_alloc(sizeof(VAPI_hca_vendor_t));
    if (ib_hca_vendor_info == NULL) {
        dat_status =  DAT_ERROR (DAT_INSUFFICIENT_RESOURCES,DAT_RESOURCE_MEMORY);
        goto bail;
    }
    ib_hca_vendor_info_allocated =DAT_TRUE;

    ib_hca_port_info = dapl_os_alloc(sizeof(VAPI_hca_cap_t));
    if (ib_hca_port_info == NULL) {
        dat_status =  DAT_ERROR (DAT_INSUFFICIENT_RESOURCES,DAT_RESOURCE_MEMORY);
         goto bail;
     }
    ib_hca_port_info_allocated =DAT_TRUE;


    hca_handle = hca_ptr->ib_hca_handle;

    ib_status = VAPI_query_hca_cap (hca_handle,
				    ib_hca_vendor_info,
				    ib_hca_cap_info);

    if (ib_status != VAPI_OK)
    {
	dapl_dbg_log (DAPL_DBG_TYPE_ERR, 
		      " error code return = %d\n", 
		      ib_status);
    dat_status = (dapl_ib_status_convert (ib_status));
    goto bail;
    }

    ib_status = VAPI_query_hca_port_prop (hca_handle,
					  hca_ptr->port_num,
					  ib_hca_port_info);

    if (ib_status != VAPI_OK)
    {
	dapl_dbg_log (DAPL_DBG_TYPE_ERR,
		      " error code return = %d\n", 
		      ib_status);
    dat_status = (dapl_ib_status_convert (ib_status));
    goto bail;
    }

       
    if ( ia_attr != NULL )
    {
	strncpy (ia_attr->adapter_name,
		 VN_ADAPTER_NAME,
		 DAT_NAME_MAX_LENGTH );

	strncpy (ia_attr->vendor_name,
		 VN_VENDOR_NAME,
		 DAT_NAME_MAX_LENGTH );

	ia_attr->ia_address_ptr		   = (DAT_IA_ADDRESS_PTR)&hca_ptr->hca_address;
	ia_attr->hardware_version_major    = ib_hca_vendor_info->hw_ver;
	ia_attr->hardware_version_minor    = 0;
	ia_attr->max_eps                   = ib_hca_cap_info->max_num_qp;
	ia_attr->max_dto_per_ep            = ib_hca_cap_info->max_qp_ous_wr;
	ia_attr->max_rdma_read_per_ep      = ib_hca_cap_info->max_qp_ous_wr;
    /*  The Driver use 6 cq's 4-ipoib 1-qp0 1-qp1 */ 
    ia_attr->max_evds                  = ib_hca_cap_info->max_num_cq-6;
	ia_attr->max_evd_qlen              = ib_hca_cap_info->max_num_ent_cq;
	ia_attr->max_iov_segments_per_dto  = ib_hca_cap_info->max_num_sg_ent;
    /*  The Driver use 1030 lmr's for the core and ipoib */ 
    ia_attr->max_lmrs                  = ib_hca_cap_info->max_num_mr-1030;
	ia_attr->max_lmr_block_size        = ib_hca_cap_info->max_mr_size;
	ia_attr->max_lmr_virtual_address   = 0xFFFFFFFF;
	ia_attr->max_pzs                   = ib_hca_cap_info->max_pd_num;
	/* 
	 * RDMA and message size is defined in the spec:
	 * IBTA 1.1, 9.4.3, 9.4.4, 9.7.7.
	 * The chip does not report this value
	 */
	ia_attr->max_mtu_size              = 0x80000000;
	ia_attr->max_rdma_size		   = 0x80000000;
	ia_attr->max_rmrs                  = ib_hca_cap_info->max_num_mr;
	ia_attr->max_rmr_target_address    = 0xFFFFFFFF;

	/*
	 * VENDOR Specific Values.
	 *
	 * Assign transport specific and vendor specific values here.
	 * Both attributes are of type:
	 *    typedef struct dat_named_attr
	 *    {
	 *    		const char *    name;      * Name of attribute  *
	 *		const char *    value;     * Value of attribute *
	 *    } DAT_NAMED_ATTR;
	 */
	ia_attr->num_transport_attr        = 0;
	ia_attr->transport_attr            = NULL;
	ia_attr->num_vendor_attr           = 0;
	ia_attr->vendor_attr               = NULL;
    }

    if ( ep_attr != NULL )
    {
	/* Sizes fixed in IB as per IBTA 1.1, 9.4.3, 9.4.4, 9.7.7.  */
	ep_attr->service_type	  = DAT_SERVICE_TYPE_RC;
	ep_attr->max_mtu_size     = 0x80000000UL;
	ep_attr->max_rdma_size    = 0x80000000UL;
	ep_attr->max_recv_dtos    = ib_hca_cap_info->max_qp_ous_wr;
	ep_attr->max_request_dtos = ib_hca_cap_info->max_qp_ous_wr;
	ep_attr->max_recv_iov     = ib_hca_cap_info->max_num_sg_ent;
	ep_attr->max_request_iov  = ib_hca_cap_info->max_num_sg_ent;
	ep_attr->max_rdma_read_in  = ib_hca_cap_info->max_qp_ous_wr; 
	ep_attr->max_rdma_read_out = ib_hca_cap_info->max_qp_ous_wr;
    }

    if(ip_addr) {
        struct sockaddr_in * ipv4_addr   = (struct sockaddr_in *) ip_addr;
        ipv4_addr->sin_family = AF_INET;

        memset(ip_addr, 0, sizeof(ip_addr));

        dat_status = ib_get_hca_ipaddr(hca_ptr->prov_data->gid, (u_int32_t*)&ipv4_addr->sin_addr);
        if (dat_status != DAT_SUCCESS)
            goto bail;
    }

bail:
    if (ib_hca_cap_info_allocated) {
        dapl_os_free(ib_hca_cap_info,sizeof(VAPI_hca_cap_t));
    }
    if (ib_hca_vendor_info_allocated) {
        dapl_os_free(ib_hca_vendor_info,sizeof(VAPI_hca_vendor_t));
    }
    if (ib_hca_port_info_allocated) {
        dapl_os_free(ib_hca_port_info,sizeof(VAPI_hca_port_t));
    }
    return dat_status;
}


DAT_RETURN
dapls_ib_completion_poll (
	IN  DAPL_HCA			*hca_ptr,
	IN  ib_cq_handle_t		cq_handle,
	IN  ib_work_completion_t	*cqe_ptr)
{
    VAPI_ret_t        ib_status;

    ib_status = VAPI_poll_cq (hca_ptr->ib_hca_handle, cq_handle, cqe_ptr);

    return dapl_ib_status_convert (ib_status);
}

DAT_RETURN
dapls_ib_completion_poll_block (
	IN  DAPL_HCA			*hca_ptr,
	IN  ib_cq_handle_t		cq_handle,
	IN  ib_work_completion_t	*cqe_ptr,
    IN  DAT_TIMEOUT             timeout)
{
    VAPI_ret_t        ib_status;

    ib_status = EVAPI_poll_cq_block (hca_ptr->ib_hca_handle, cq_handle,timeout , cqe_ptr);

    return dapl_ib_status_convert (ib_status);
}

#ifdef VAPI_VERSION_CODE 
DAT_RETURN
dapls_ib_completion_poll_nblock (
	IN  DAPL_HCA			*hca_ptr,
	IN  ib_cq_handle_t		cq_handle,
	IN  ib_work_completion_t	*cqe_ptr,
    IN  DAT_TIMEOUT             timeout,
    IN  DAT_COUNT               cqe_num)
{

    VAPI_ret_t        ib_status;
    ib_status = EVAPI_poll_cq_nblock (hca_ptr->ib_hca_handle, cq_handle,timeout, cqe_num,cqe_ptr);

    return dapl_ib_status_convert (ib_status);
}
#endif

DAT_RETURN
dapls_ib_completion_poll_unblock (
	IN  DAPL_HCA   	*hca_ptr,
	IN  ib_cq_handle_t		cq_handle)	
{
    VAPI_ret_t        ib_status;

    ib_status = EVAPI_poll_cq_unblock (hca_ptr->ib_hca_handle, cq_handle);
    return dapl_ib_status_convert (ib_status);
}

DAT_RETURN
dapls_ib_peek_cq (
	IN  DAPL_HCA			*hca_ptr,
	IN  ib_cq_handle_t		cq_handle,
	IN  DAT_COUNT	        cqe_num)
{
    VAPI_ret_t        ib_status;

    ib_status = EVAPI_peek_cq (hca_ptr->ib_hca_handle, cq_handle,cqe_num);

    return dapl_ib_status_convert (ib_status);
}

// I'm assuming that the cqe_num in previous peek failed.

DAT_RETURN
dapls_ib_peek_cq_count (
	IN  DAPL_HCA			*hca_ptr,
	IN  ib_cq_handle_t		cq_handle,
	IN  DAT_COUNT	        threshold,
    IN  DAT_BOOLEAN         count_en,
    OUT DAT_COUNT           *cq_count)
{
    VAPI_ret_t        ib_status;
    DAT_COUNT         cqe_num_check,cqe_num_half;    
    cqe_num_half   = (threshold+1)/2;
    cqe_num_check = threshold - cqe_num_half;

    ib_status = EVAPI_peek_cq (hca_ptr->ib_hca_handle, cq_handle,threshold);

    if (ib_status == VAPI_OK || !count_en) {
       *cq_count = threshold-1;
       return dapl_ib_status_convert (ib_status);
    }

    if (cqe_num_check==0) {
        *cq_count = 0;
        return dapl_ib_status_convert(VAPI_CQ_EMPTY);
    }

    do {
    ib_status = EVAPI_peek_cq (hca_ptr->ib_hca_handle, cq_handle,cqe_num_check);
    if ((ib_status != VAPI_OK) && (ib_status !=VAPI_CQ_EMPTY)) {
        return dapl_ib_status_convert (ib_status);
    }    
    if (ib_status == VAPI_OK) {
        cqe_num_half=cqe_num_half/2;
        cqe_num_check = cqe_num_check+cqe_num_half;
    } else {
        cqe_num_half=(cqe_num_half+1)/2;
        cqe_num_check =cqe_num_check-cqe_num_half;
    }

    } while (cqe_num_half!=0  );

    if (ib_status == VAPI_OK) {
     *cq_count=cqe_num_check;        
    } else {
     *cq_count=0;
    }
    
    return dapl_ib_status_convert (VAPI_CQ_EMPTY);
}


DAT_RETURN
dapls_ib_completion_notify (
		IN ib_hca_handle_t hca_handle,
		IN ib_cq_handle_t cq_handle,
		IN ib_notification_type_t type)
{
    VAPI_ret_t        ib_status;

    ib_status = VAPI_req_comp_notif (hca_handle, cq_handle, type);

    return dapl_ib_status_convert (ib_status);
}


//Webman add new evd_wait function

DAT_RETURN 
dapl_evd_wait_object_init (
//    IN DAPL_OS_WAIT_OBJECT *wait_obj, 
    IN DAPL_EVD		       *evd_ptr)    
{
    DAT_RETURN dat_status;
#ifndef __KDAPL__
    int status;
    if (evd_ptr->ib_cq_handle != IB_INVALID_HANDLE) {
        /* condtion varaible not used */
        status = EVAPI_set_comp_eventh(evd_ptr->header.owner_ia->hca_ptr->ib_hca_handle, 
                                   evd_ptr->ib_cq_handle,
                                    EVAPI_POLL_CQ_UNBLOCK_HANDLER,
                                     NULL, &evd_ptr->ib_comp_handle);

        
        if(status) 
        {
            dapl_dbg_log (DAPL_DBG_TYPE_ERR, " EVAPI_set_comp_eventh() failed\n");
            return DAT_ERROR (DAT_INSUFFICIENT_RESOURCES,DAT_RESOURCE_MEMORY);
        }
    } else
    {
//        wait_obj->signaled = DAT_FALSE;
//        if ( 0 != pthread_cond_init ( &wait_obj->cv, NULL ) )
//        {
//            return DAT_ERROR (DAT_INTERNAL_ERROR,0);
//        }
//
//        /* Always returns 0.  */
//        pthread_mutex_init ( &wait_obj->lock, NULL );
#endif
        dat_status = dapl_os_wait_object_init(&evd_ptr->wait_object);
        return dat_status;
#ifndef __KDAPL__

    }
    


    return DAT_SUCCESS;
#endif
}


/* Wait on the supplied wait object, up to the specified time_out.
 * A timeout of DAT_TIMEOUT_INFINITE will wait indefinitely.
 * Timeout should be specified in micro seconds.
 *
 * Functional returns:
 *	DAT_SUCCESS -- another thread invoked dapl_os_wait object_wakeup
 * 	DAT_INVALID_STATE -- someone else is already waiting in this wait
 * 	object.
 *			     only one waiter is allowed at a time.
 *	DAT_ABORT -- another thread invoked dapl_os_wait_object_destroy
 *	DAT_TIMEOUT -- the specified time limit was reached.
 */

DAT_RETURN 
dapl_evd_wait_object_wait (
    IN	DAPL_EVD *evd_ptr, 
    IN  DAT_TIMEOUT timeout_val,
    IN  DAT_COUNT   cqe_num)
{
    DAT_RETURN 		dat_status = DAT_SUCCESS;
    // the flag is for kernel only need to add in order not to break compile
//    int 		pthread_status;
//    struct timespec 	future;

#ifndef __KDAPL__
    int flag;
    if (evd_ptr->ib_cq_handle == IB_INVALID_HANDLE) 
        {
        dat_status = dapl_os_wait_object_wait(&evd_ptr->wait_object,timeout_val,&flag);
    } else {
        dat_status = dapls_evd_cq_poll_block_to_event(evd_ptr, timeout_val, cqe_num);
    }
#endif 
        return dat_status;
}


/*
 * dapl_os_wait_object_wakeup
 *
 * Wakeup a thread waiting on a wait object
 *
 * Input:
 *      wait_obj
 *
 * Returns:
 *	DAT_SUCCESS
 *	DAT_INTERNAL_ERROR
 */
DAT_RETURN 
dapl_evd_wait_object_wakeup (
    IN	DAPL_EVD *evd_ptr)
{
    DAT_RETURN dat_status;
#ifndef __KDAPL__
    if (evd_ptr->ib_cq_handle == IB_INVALID_HANDLE) 
    {
#endif
    dat_status = dapl_os_wait_object_wakeup(&evd_ptr->wait_object);
#ifndef __KDAPL__
    } else {
        dat_status = dapls_ib_completion_poll_unblock(evd_ptr->header.owner_ia->hca_ptr,evd_ptr->ib_cq_handle);
    }
#endif

    return dat_status;
}


/*
 * dapl_os_wait_object_destroy
 *
 * Destroy a wait object
 *
 * Input:
 *      wait_obj
 *
 * Returns:
 *	DAT_SUCCESS
 *	DAT_INTERNAL_ERROR
 */
DAT_RETURN 
dapl_evd_wait_object_destroy (
    IN	DAPL_EVD *evd_ptr)
{
    DAT_RETURN dat_status;
#ifndef __KDAPL__
    int status;
    if (evd_ptr->ib_cq_handle == IB_INVALID_HANDLE) {
#endif
dat_status = dapl_os_wait_object_destroy(&evd_ptr->wait_object);
#ifndef __KDAPL__
    } else {
        
        status = EVAPI_clear_comp_eventh(evd_ptr->header.owner_ia->hca_ptr->ib_hca_handle, evd_ptr->ib_comp_handle);
        
        if (status)
        {
            dapl_dbg_log (DAPL_DBG_TYPE_ERR,
                          " EVAPI_clear_comp_eventh() failed\n");
            return  DAT_ERROR (DAT_INSUFFICIENT_RESOURCES,DAT_RESOURCE_MEMORY);
        }
        dat_status = DAT_SUCCESS;
    }

#endif

    return dat_status;
}


/*
 * dapls_ib_get_async_event
 *
 * Translate an asynchronous event type to the DAT event.
 * Note that different providers have different sets of errors.
 *
 * Input:
 *	cause_ptr		provider event cause
 *
 * Output:
 * 	async_event		DAT mapping of error
 *
 * Returns:
 * 	DAT_SUCCESS
 *	DAT_NOT_IMPLEMENTED	Caller is not interested this event
 */

DAT_RETURN dapls_ib_get_async_event (
	IN  VAPI_event_record_t		*cause_ptr,
	OUT DAT_EVENT_NUMBER		*async_event)
{
    VAPI_event_record_type_t	event_id;
    DAT_RETURN			dat_status;

    dat_status = DAT_SUCCESS;
    event_id = cause_ptr->type;

    switch (event_id )
    {
	/* VAPI does not detect OVERFLOW errors? */

	/* INTERNAL errors */
	case VAPI_CQ_ERROR:
	case VAPI_LOCAL_WQ_ACCESS_VIOL_ERROR:
	case VAPI_LOCAL_WQ_INV_REQUEST_ERROR:
	{
	    *async_event = DAT_ASYNC_ERROR_PROVIDER_INTERNAL_ERROR;
	    break;
	}

	/* CATASTROPHIC errors */
	case VAPI_LOCAL_WQ_CATASTROPHIC_ERROR:
	case VAPI_LOCAL_EEC_CATASTROPHIC_ERROR:
	case VAPI_LOCAL_CATASTROPHIC_ERROR:
	case VAPI_PORT_ERROR:
	{
	    *async_event = DAT_ASYNC_ERROR_IA_CATASTROPHIC;
	    break;
	}
	default:
	{
	    /*
	     * Errors we are not interested in reporting:
	     * VAPI_QP_PATH_MIGRATED
	     * VAPI_EEC_PATH_MIGRATED
	     * VAPI_PATH_MIG_REQ_ERROR
	     * VAPI_QP_COMM_ESTABLISHED
	     * VAPI_EEC_COMM_ESTABLISHED
	     * VAPI_SEND_QUEUE_DRAINED
	     * VAPI_PORT_ACTIVE
	     */
	    dat_status = DAT_NOT_IMPLEMENTED;
	}

    }

    return dat_status;
}


DAT_RETURN
dapls_ib_ncompletion_notify (
                IN ib_hca_handle_t hca_handle,
                IN ib_cq_handle_t cq_handle,
                IN DAT_COUNT num);


DAT_RETURN
dapls_ib_ncompletion_notify (
                IN ib_hca_handle_t hca_handle,
                IN ib_cq_handle_t cq_handle,
                IN DAT_COUNT num)
{
    VAPI_ret_t        ib_status;

    ib_status = EVAPI_req_ncomp_notif (hca_handle, cq_handle, num);

    return dapl_ib_status_convert (ib_status);
}

// this old function
/*DAT_RETURN
dapls_ib_peek_cq (
                IN ib_hca_handle_t hca_handle,
                IN ib_cq_handle_t cq_handle,
                IN DAT_COUNT num);



DAT_RETURN
dapls_ib_peek_cq (
                IN ib_hca_handle_t hca_handle,
                IN ib_cq_handle_t cq_handle,
                IN DAT_COUNT num)
{
    VAPI_ret_t        ib_status;

    ib_status = EVAPI_peek_cq (hca_handle, cq_handle, num);

    return dapl_ib_status_convert (ib_status);
} */



DAT_RETURN 
dapls_ib_get_hca_ids(ib_hca_handle_t hca, IB_port_t port, IB_gid_t  gid, IB_lid_t * lid,
                     IB_guid_t guid)
{
    int             status;
    VAPI_hca_cap_t    *cap_info     = NULL;
    VAPI_hca_vendor_t *vendor_info  = NULL;
    VAPI_hca_port_t   *port_info    = NULL;
    DAT_RETURN				dat_status = DAT_SUCCESS;
    /* Flag indicating we've allocated one of these.  */
    DAT_BOOLEAN				cap_info_allocated = DAT_FALSE;
    DAT_BOOLEAN				vendor_info_allocated = DAT_FALSE;
    DAT_BOOLEAN				port_info_allocated   = DAT_FALSE;

    cap_info = dapl_os_alloc(sizeof(VAPI_hca_cap_t));
    if (cap_info == NULL) {
        dat_status =  DAT_ERROR (DAT_INSUFFICIENT_RESOURCES,DAT_RESOURCE_MEMORY);
        goto bail;
    }
    cap_info_allocated =DAT_TRUE;

    vendor_info = dapl_os_alloc(sizeof(VAPI_hca_vendor_t));
    if (vendor_info == NULL) {
        dat_status =  DAT_ERROR (DAT_INSUFFICIENT_RESOURCES,DAT_RESOURCE_MEMORY);
        goto bail;
    }
    vendor_info_allocated =DAT_TRUE;

    port_info = dapl_os_alloc(sizeof(VAPI_hca_cap_t));
    if (port_info == NULL) {
        dat_status =  DAT_ERROR (DAT_INSUFFICIENT_RESOURCES,DAT_RESOURCE_MEMORY);
         goto bail;
     }
    port_info_allocated =DAT_TRUE;

    status = VAPI_query_hca_cap (hca, vendor_info, cap_info);
    if(status != VAPI_OK){
        dat_status =  DAT_ERROR (DAT_INTERNAL_ERROR,DAT_NO_SUBTYPE);
        goto bail;
    }

    status = VAPI_query_hca_port_prop(hca, port, port_info);
    if (status != VAPI_OK) {
        dat_status =  DAT_ERROR (DAT_INTERNAL_ERROR,DAT_NO_SUBTYPE);
        goto bail;
    }
    dapl_os_memcpy (guid, &cap_info->node_guid, sizeof (IB_guid_t));
    
    {
	tUINT16  tbl_len_out = -1;
	IB_gid_t gid_tbl[port_info->gid_tbl_len];
	
	status = VAPI_query_hca_gid_tbl(hca, port, port_info->gid_tbl_len,
                                         &tbl_len_out, gid_tbl);
	if (status != VAPI_OK)
	{
	    dapl_dbg_log (DAPL_DBG_TYPE_ERR,
			  " VAPI_query_hca_gid_tbl() failed: %d\n",
			  status);
        dat_status =  DAT_ERROR (DAT_INTERNAL_ERROR,DAT_NO_SUBTYPE);
        goto bail;
	}
	
	memcpy(gid, &gid_tbl[0], sizeof(IB_gid_t));
    }


    *lid = port_info->lid;

bail:
    if (cap_info_allocated) {
        dapl_os_free(cap_info,sizeof(VAPI_hca_cap_t));
    }
    if (vendor_info_allocated) {
        dapl_os_free(vendor_info,sizeof(VAPI_hca_vendor_t));
    }
    if (port_info_allocated) {
        dapl_os_free(port_info,sizeof(VAPI_hca_port_t));
    }
   
    return dat_status;
}



/*
 * Map all VAPI DTO completion codes to the DAT equivelent.
 *
 * Not returned by IB:     DAT_DTO_ERR_PARTIAL_PACKET
 */
static struct ib_status_map
{
    int				ib_status;
    DAT_DTO_COMPLETION_STATUS	dat_status;
} ib_status_map[] = {
    /* 0  */  { VAPI_SUCCESS, 			DAT_DTO_SUCCESS},
    /* 01 */  { VAPI_LOC_LEN_ERR,		DAT_DTO_ERR_LOCAL_LENGTH},
    /* 02 */  { VAPI_LOC_QP_OP_ERR,		DAT_DTO_ERR_LOCAL_EP},
    /* 03 */  { VAPI_LOC_EE_OP_ERR,		DAT_DTO_ERR_TRANSPORT},
    /* 04 */  { VAPI_LOC_PROT_ERR,		DAT_DTO_ERR_LOCAL_PROTECTION},
    /* 05 */  { VAPI_WR_FLUSH_ERR,		DAT_DTO_ERR_FLUSHED},
    /* 06 */  { VAPI_MW_BIND_ERR,		DAT_RMR_OPERATION_FAILED},
    /* 07 */  { VAPI_BAD_RESP_ERR,		DAT_DTO_ERR_BAD_RESPONSE},
    /* 08 */  { VAPI_LOC_ACCS_ERR,		DAT_DTO_ERR_LOCAL_EP},
    /* 09 */  { VAPI_REM_INV_REQ_ERR,		DAT_DTO_ERR_REMOTE_RESPONDER},
    /* 10 */  { VAPI_REM_ACCESS_ERR,		DAT_DTO_ERR_REMOTE_ACCESS},
    /* 11 */  { VAPI_REM_OP_ERR,		DAT_DTO_ERR_REMOTE_RESPONDER},
    /* 12 */  { VAPI_RETRY_EXC_ERR,		DAT_DTO_ERR_TRANSPORT},
    /* 13 */  { VAPI_RNR_RETRY_EXC_ERR,		DAT_DTO_ERR_RECEIVER_NOT_READY},
    /* 14 */  { VAPI_LOC_RDD_VIOL_ERR,		DAT_DTO_ERR_REMOTE_ACCESS},
    /* 15 */  { VAPI_REM_INV_RD_REQ_ERR,	DAT_DTO_FAILURE},
    /* 16 */  { VAPI_REM_ABORT_ERR,		DAT_DTO_ERR_REMOTE_RESPONDER},
    /* 17 */  { VAPI_INV_EECN_ERR,		DAT_DTO_ERR_TRANSPORT},
    /* 18 */  { VAPI_INV_EEC_STATE_ERR,		DAT_DTO_ERR_TRANSPORT},
    /* 19 */  { VAPI_COMP_FATAL_ERR,		DAT_DTO_FAILURE},
    /* 20 */  { VAPI_COMP_GENERAL_ERR,		DAT_DTO_FAILURE}
};


/*
 * dapls_ib_get_dto_status
 *
 * Return the DAT status of a DTO operation
 *
 * Input:
 *	cqe_ptr			pointer to completion queue entry
 *
 * Output:
 * 	none
 *
 * Returns:
 * 	Value from ib_status_map table above
 */

DAT_DTO_COMPLETION_STATUS
dapls_ib_get_dto_status (
	IN ib_work_completion_t		*cqe_ptr)
{
    uint32_t		ib_status;
    int			i;

    ib_status = DAPL_GET_CQE_STATUS (cqe_ptr);

    /*
     * Due to the implementation of VAPI completion code, we need to
     * search the table for the correct value rather than assuming
     * linear distribution.
     * N.B.  This could be done with a case statement, but for the common
     *	     case of a successful completion, this is faster.
     *
     */
    for (i = 0; i < 21; i++)
    {
	if (ib_status == ib_status_map[i].ib_status)
	{
	    return ib_status_map[i].dat_status;
	}
    }

    dapl_dbg_log (DAPL_DBG_TYPE_DTO_COMP_ERR,
		  " DTO completion ERROR: %d: op %#x\n", 
                  ib_status,
		  DAPL_GET_CQE_OPTYPE (cqe_ptr));

    return DAT_DTO_FAILURE;
}





/*
 * Map all VAPI DTO completion codes to the DAT equivelent.
 *
 * Not returned by IB:     DAT_DTO_ERR_PARTIAL_PACKET
 */
#define DAPL_IB_EVENT_CNT	12

static struct ib_cm_event_map
{
    const ib_cm_events_t	ib_cm_event;
    DAT_EVENT_NUMBER		dat_event_num;
} ib_cm_event_map[DAPL_IB_EVENT_CNT] = {  
    { IB_CME_CONNECTED, DAT_CONNECTION_EVENT_ESTABLISHED }, 
    { IB_CME_DISCONNECTED, DAT_CONNECTION_EVENT_DISCONNECTED }, 
    { IB_CME_DISCONNECTED_ON_LINK_DOWN, DAT_CONNECTION_EVENT_DISCONNECTED },
    { IB_CME_CONNECTION_REQUEST_PENDING, DAT_CONNECTION_REQUEST_EVENT },
    { IB_CME_CONNECTION_REQUEST_PENDING_PRIVATE_DATA, DAT_CONNECTION_REQUEST_EVENT },
    { IB_CME_CONNECTION_REQUEST_ACKED, 0 },
    { IB_CME_DESTINATION_REJECT, DAT_CONNECTION_EVENT_NON_PEER_REJECTED }, 
    { IB_CME_DESTINATION_REJECT_PRIVATE_DATA, DAT_CONNECTION_EVENT_NON_PEER_REJECTED },
    { IB_CME_DESTINATION_UNREACHABLE,  DAT_CONNECTION_EVENT_UNREACHABLE },
    { IB_CME_TOO_MANY_CONNECTION_REQUESTS, DAT_CONNECTION_EVENT_NON_PEER_REJECTED },
    { IB_CME_LOCAL_FAILURE, DAT_CONNECTION_EVENT_BROKEN },
    { IB_CME_BROKEN, DAT_CONNECTION_EVENT_BROKEN }
};

/*
 * dapls_ib_get_dat_event
 *
 * Return a DAT connection event given a provider CM event.
 *
 * N.B.	Some architectures combine async and CM events into a
 *	generic async event. In that case, dapls_ib_get_cm_event()
 *	and dapls_ib_get_async_event() should be entry points that
 *	call into a common routine.
 *
 * Input:
 *	ib_cm_event	event provided to the dapl callback routine
 *	active		switch indicating active or passive connection
 *
 * Output:
 * 	none
 *
 * Returns:
 * 	DAT_EVENT_NUMBER of translated provider value
 */

DAT_EVENT_NUMBER
dapls_ib_get_dat_event (
	IN    const ib_cm_events_t	ib_cm_event,
	IN    DAT_BOOLEAN		active)
{
    DAT_EVENT_NUMBER	dat_event_num;
    int			index;
    int			i;

    active = active;			/* keep compiler happy */

    if (ib_cm_event > IB_CME_BROKEN)
    {
	return (DAT_EVENT_NUMBER) 0;
    }
    
    dat_event_num = 0;
    index        = ib_cm_event - IB_CME_CONNECTED;

    if (ib_cm_event == ib_cm_event_map[index].ib_cm_event)
    {
	dat_event_num = ib_cm_event_map[index].dat_event_num;
    }
    else
    {
	/*
	 * Do a linear walk of the list, just in case the table has been
	 * messed up or someone changes the definitions; paranoia, but
	 * recoverable and robust.
	 */
	for (i = 0; i < DAPL_IB_EVENT_CNT; i++)
	{
	    if (ib_cm_event == ib_cm_event_map[i].ib_cm_event)
	    {
		dat_event_num = ib_cm_event_map[i].dat_event_num;
	    }
	}
    }

    dapl_dbg_log (DAPL_DBG_TYPE_CALLBACK,
		  " dapls_ib_get_dat_event:ibapi event translation: (%s) ib_event 0x%x dat_event 0x%x\n",
		  active ? "active" : "passive",
		  ib_cm_event,
		  dat_event_num);

    return dat_event_num;
}


/*
 * dapls_ib_get_dat_event
 *
 * Return a DAT connection event given a provider CM event.
 *
 * N.B.	Some architectures combine async and CM events into a
 *	generic async event. In that case, dapls_ib_get_cm_event()
 *	and dapls_ib_get_async_event() should be entry points that
 *	call into a common routine.
 *
 * Input:
 *	dat_event_num	DAT event we need an equivelent CM event for
 *
 * Output:
 * 	none
 *
 * Returns:
 * 	ib_cm_event of translated DAPL value
 */
ib_cm_events_t
dapls_ib_get_cm_event (
	IN    DAT_EVENT_NUMBER		dat_event_num)
{
    ib_cm_events_t	ib_cm_event;
    int			i;

    ib_cm_event = 0;

    for (i = 0; i < DAPL_IB_EVENT_CNT; i++)
    {
	if (dat_event_num == ib_cm_event_map[i].dat_event_num)
	{
            ib_cm_event = ib_cm_event_map[i].ib_cm_event;
            break;
	}
    }

    return ib_cm_event;


}

/*
 * Local variables:
 *  c-indent-level: 4
 *  c-basic-offset: 4
 *  tab-width: 8
 * End:
 */

