/*
 * Copyright (c) 1999-2002, Mellanox Technologies, Inc. All rights reserved. 
 * Copyright (c) 2005 Mellanox Technologies.   All rights reserved.

 *
 * This Software is licensed under either one of the following two licenses:
 *
 * 1) under the terms of the "Common Public License 1.0" a copy of which is
 *    in the file LICENSE.txt in the root directory. The license is also
 *    available from the Open Source Initiative, see
 *    http://www.opensource.org/licenses/cpl.php.
 * OR
 *
 * 2) under the terms of the "The BSD License" a copy of which is in the file
 *    LICENSE2.txt in the root directory. The license is also available from
 *    the Open Source Initiative, see
 *    http://www.opensource.org/licenses/bsd-license.php.
 *
 * Licensee has the right to choose either one of the above two licenses.
 *
 * Redistributions of source code must retain both the above copyright
 * notice and either one of the license notices.
 *
 * Redistributions in binary form must reproduce both the above copyright
 * notice, either one of the license notices in the documentation
 * and/or other materials provided with the distribution.
 */

/**********************************************************************
 * 
 * HEADER: dapl_vapi_util.h
 *
 * PURPOSE: Utility defs & routines for access to MELLANOX Tavor Verbs
 *
 * $Id: dapl_openib_util.h,v 1.3 2004/07/28 13:09:31 rabin Exp $
 *
 **********************************************************************/

#ifndef _DAPL_VAPI_UTIL_H_
#define _DAPL_VAPI_UTIL_H_

#include "vapi_types.h"
#include "vapi.h"
#include "evapi.h"
#ifdef __KDAPL__
#include "dapl_mem_fmr.h"
#endif
STATIC _INLINE_ DAT_RETURN 
dapl_ib_status_convert (
    IN 	int32_t 	ib_status);


/* A version of DAT_ERROR is redefined here as the order of include
 * files makes it unavailable at this point
 */
#define DAT_ERRORP(Type,SubType) ((DAT_RETURN)(DAT_CLASS_ERROR | Type | SubType))

/*
 * Typedefs to map VAPI types to more generic 'ib' types
 */
typedef          VAPI_hca_hndl_t               ib_hca_handle_t;
typedef          VAPI_cq_hndl_t                ib_cq_handle_t;
typedef          VAPI_k_cq_hndl_t              ib_kcq_handle_t;
typedef          VAPI_qp_hndl_t                ib_qp_handle_t;
typedef          VAPI_pd_hndl_t                ib_pd_handle_t;
typedef          VAPI_mr_hndl_t                ib_mr_handle_t;
typedef          VAPI_mw_hndl_t                ib_mw_handle_t;
typedef          VAPI_qp_state_t               ib_qp_state_t;
typedef          VAPI_hca_id_t                 ib_hca_name_t;
typedef          VAPI_event_record_t           ib_error_record_t;
typedef          VAPI_wc_desc_t                ib_work_completion_t;
typedef          VAPI_cq_notif_type_t          ib_notification_type_t;
typedef          VAPI_async_event_handler_t    ib_async_handler_t;
typedef          EVAPI_compl_handler_hndl_t    ib_comp_handle_t;
typedef		 VAPI_sg_lst_entry_t	       ib_data_segment_t;
typedef		 VAPI_event_record_type_t      ib_async_event_type;


// ETBD - Remove this later
typedef char iba_gid_t[16];

typedef struct _ib_gid_t
{
    uint64_t gid_prefix;
    uint64_t guid;
} ib_gid_t;

typedef          MT_bool                       ib_bool_t;

/*
 * ib_hca_transport_t structure. Define fields specific to this
 * provider implementation necessary to maintain the HCA
 *
 * NOTE: null_ib_hca_handle is strictly internal to IB. DAPL allows a
 * 	 NULL DTO EVD handle, but IB does not. Create a CQ under the 
 *	 hood and make sure an error is generated if the user every
 *	 tries to post, by setting the WQ length to 0 in ep_create
 *	 and/or ep_modify.
 */
typedef struct ib_hca_transport
{
    ib_cq_handle_t	null_ib_cq_handle;   /* CQ handle with 0 entries */
    IB_guid_t           node_GUID;	     /* HCA Address */
#if defined(_TOPSPIN_)
    VAPI_hca_hndl_t	ib_khca_handle;
    VAPI_hca_hndl_t	ts_hca_handle;
#endif /* _TOPSPIN_ */
} ib_hca_tranport_t;

/*
 * ib_shm_transport_t structure. Define fields specific to this
 * provider implementation necessary to maintain shared memory state
 *
 * VAPI does not have an API to directly register memory as shared
 * memory. Instead, we must register memory, then register that memory
 * as shared. This gives us a new mr_handle, as well as a new l_key and
 * r_key, so we need to save the original handle until it can be
 * disposed of when the user does an lmr_free. The l_key and r_key are
 * not available and will disappear when we deallocate the mr_handle, so
 * we don't save them.
 */
typedef struct ib_shm_transport
{
    ib_mr_handle_t	mr_handle;
} ib_shm_transport_t;


typedef void (*async_cq_err_cb_t)
               (ib_hca_handle_t, ib_cq_handle_t, ib_error_record_t *, void *); 

typedef void (*async_qp_err_cb_t)
               (ib_hca_handle_t, ib_qp_handle_t, ib_error_record_t *, void *); 

typedef void (*async_err_cb_t)
               (ib_hca_handle_t, ib_error_record_t *, void *); 

typedef struct _dapl_tavor_async_cb
{
    ib_bool_t                   set_async_cb_handler;
    async_cq_err_cb_t           cq_async_error_cb;
    async_qp_err_cb_t           qp_async_error_cb;
    async_err_cb_t              async_error_cb;
} dapl_tavor_async_cb_t;



#define          IB_INVALID_HANDLE             VAPI_INVAL_HNDL

#define          IB_MAX_REQ_PDATA_SIZE      92
#define          IB_MAX_REP_PDATA_SIZE      196
#define          IB_MAX_REJ_PDATA_SIZE      148
#define          IB_MAX_DREQ_PDATA_SIZE     220
#define          IB_MAX_DREP_PDATA_SIZE     224


/*
 * Definitions to map DTO OPs
 */
#define	OP_RDMA_READ	 VAPI_RDMA_READ
#define	OP_RDMA_WRITE	 VAPI_RDMA_WRITE
#define OP_SEND		 VAPI_SEND
#define OP_RECEIVE	 VAPI_RECEIVE
#define OP_COMP_AND_SWAP VAPI_ATOMIC_CMP_AND_SWP
#define OP_FETCH_AND_ADD VAPI_ATOMIC_FETCH_AND_ADD
#define OP_BIND_MW       8

/*
 * CQ NOTIFICATION TYPE
 */
#define IB_NOTIFY_ON_NEXT_COMP    VAPI_NEXT_COMP
#define IB_NOTIFY_ON_SOLIC_COMP   VAPI_SOLIC_COMP

/*
 * Macro to check the state of an EP/QP
 */
#define DAPLIB_NEEDS_INIT(ep)  ((ep)->qp_state == VAPI_ERR)


STATIC _INLINE_ char *
dapl_ib_convert_name(
    IN  char 	*name)
{
    /* no need to convert */
    return dapl_os_strdup(name);
}

/*
 * Resolve VAPI return codes to their DAPL equivelent.
 * Do not return invalid Handles, the user is not able
 * to deal with them.
 */
STATIC _INLINE_ DAT_RETURN 
dapl_ib_status_convert (
    IN 	int32_t 	ib_status)
{
    switch ( ib_status )
    {
	case VAPI_OK:
	{
	    return DAT_SUCCESS;
	}
	case VAPI_EAGAIN:
	case VAPI_EBUSY:
	{
	    return DAT_ERRORP (DAT_INTERNAL_ERROR, DAT_NO_SUBTYPE);
	}
	case VAPI_ENOMEM:
	{
	    return DAT_ERRORP (DAT_INSUFFICIENT_RESOURCES, DAT_NO_SUBTYPE);
	}
    case VAPI_EINVAL_CQ_HNDL:
	case VAPI_EINVAL_HCA_HNDL:
	{
	    return DAT_ERRORP (DAT_INVALID_HANDLE, DAT_NO_SUBTYPE);
	}
	case VAPI_CQ_EMPTY:
	{
	    return DAT_ERRORP (DAT_QUEUE_EMPTY, DAT_NO_SUBTYPE);
	}
    case VAPI_ETIMEOUT:
	{
	    return DAT_ERRORP (DAT_TIMEOUT_EXPIRED, DAT_NO_SUBTYPE);
	}    
    case VAPI_EINTR:
	{
	    return DAT_ERRORP (DAT_INTERRUPTED_CALL, DAT_NO_SUBTYPE);
	}    
	default:
	{
	    return DAT_ERRORP (DAT_INTERNAL_ERROR, DAT_NO_SUBTYPE);
	}
    }
}


/*
 * Prototype
 */

VAPI_ret_t 
dapls_modify_qp_state_to_init (
	VAPI_hca_hndl_t, IB_port_t, VAPI_qp_hndl_t);

VAPI_ret_t 
dapls_modify_qp_state_to_error (
    VAPI_hca_hndl_t               hca_handle,
    VAPI_qp_hndl_t                qp );

VAPI_ret_t
dapls_modify_qp_state_to_reset (
    VAPI_hca_hndl_t               hca_handle,
    VAPI_qp_hndl_t                qp );

STATIC _INLINE_ void
dapl_ib_release_name(
    IN  char *	name)
{
    /* free the space allocated by strdup */
    dapl_os_free(name, dapl_os_strlen(name));
}



extern int32_t dapls_ib_init (void);
extern int32_t dapls_ib_release (void);



#endif /*  _DAPL_VAPI_UTIL_H_ */
