/* * Copyright (c) 2002-2003, Network Appliance, Inc. All rights reserved.  *
 * Copyright (c) 2005 Mellanox Technologies.   All rights reserved.
 * This Software is licensed under both of the following two licenses:
 *
 * 1) under the terms of the "Common Public License 1.0" a copy of which is
 *    in the file LICENSE.txt in the root directory. The license is also
 *    available from the Open Source Initiative, see
 *    http://www.opensource.org/licenses/cpl.php.
 * OR
 *
 * 2) under the terms of the "The BSD License" a copy of which is in the file
 *    LICENSE2.txt in the root directory. The license is also available from
 *    the Open Source Initiative, see
 *    http://www.opensource.org/licenses/bsd-license.php.
 *
 * Licensee has the right to choose either one of the above two licenses.
 *
 * Redistributions of source code must retain both the above copyright
 * notice and either one of the license notices.
 *
 * Redistributions in binary form must reproduce both the above copyright
 * notice, either one of the license notices in the documentation
 * and/or other materials provided with the distribution.
 */

#include "dapl_proto.h"
#include "kdapl_ioctl.h"
#include "kdapl_tdep.h"
#include <linux/sched.h>

static DT_Tdep_Print_Head	*DT_Print_List;
static DT_Mdep_LockType	 	 DT_Print_List_Lock;

void
KDT_Print_Init (void)
{
    DT_Print_List = NULL;
    DT_Mdep_LockInit (&DT_Print_List_Lock);
}

void
KDT_Print_Destroy (void)
{
    /* TODO: Must free mem */
    DT_Mdep_LockDestroy (&DT_Print_List_Lock);
}

DT_Tdep_Print_Head *
KDT_Print_Alloc (int instance)
{
    DT_Tdep_Print_Head *phead;

    phead = DT_Mdep_Malloc (sizeof (DT_Tdep_Print_Head));
    if (phead)
    {
	phead->next = NULL;
	phead->instance = instance;
	phead->head = NULL;
	phead->tail = NULL;
	DT_Mdep_LockInit (&phead->lock);
	DT_Mdep_wait_object_init (&phead->wait_object);
	
	/* add new print phead to DT_Print_List */
	if (DT_Print_List == NULL)
	{
	    DT_Print_List = phead;
	}
	else
	{
	    phead->next = DT_Print_List;
	    DT_Print_List = phead;
	}
    }
    return (phead);
}

void
KDT_Print_Free (DT_Tdep_Print_Head *free_phead)
{
    DT_Tdep_Print_Head *phead;

    if (free_phead == DT_Print_List)
    {
	DT_Print_List = DT_Print_List->next;
	DT_Mdep_Free (free_phead);
    }
    else
    {
	phead = DT_Print_List;
	while (phead)
	{
	    if (phead->next == free_phead)
	    {
		phead->next = free_phead->next;
		DT_Mdep_Free (free_phead);
		break;
	    }
	    phead = phead->next;
	}
    }
}

void
DT_Tdep_PT_Printf (DT_Tdep_Print_Head * phead, const char * fmt, ...)
{
    va_list		args;
    int			len;
    Tdep_Print_Entry 	*entry;

    entry = DT_Mdep_Malloc (sizeof (Tdep_Print_Entry));

    va_start (args, fmt);
    len = vsnprintf (entry->buffer, PRINT_MAX, fmt, args);
    va_end(args);

    if (len >= PRINT_MAX-1)
    {
	DT_Tdep_PT_Printf (phead,
			   "WARNING: Possible truncation of print Line\n");
    }
    entry->next = NULL;

    DT_Mdep_Lock (&phead->lock);

    if (phead->head == NULL)
    {
	phead->head = entry;
	phead->tail = entry;
    }
    else
    {
	phead->tail->next = entry;
	phead->tail = entry;
    }
    DT_Mdep_wait_object_wakeup (&phead->wait_object);
    DT_Mdep_Unlock (&phead->lock);
}

int
KDT_Get_Print_Line (int instance, char * buffer)
{
    unsigned int	 len;
    DT_Tdep_Print_Head  *phead;
    Tdep_Print_Entry    *entry;
    int rc;

    buffer[0] = '\0';

    DT_Mdep_Lock (&DT_Print_List_Lock);
    phead = DT_Print_List;
    while (phead)
    {
	if (phead->instance == instance)
	{
	    break;
	}
	phead = phead->next;
    }
    DT_Mdep_Unlock (&DT_Print_List_Lock);

    if (!phead)
    {
	printk ("<1> Kdapltest_module called with invalid instance\n");
	strncpy (buffer, "Invalid Cookie\n", PRINT_MAX);
    }
    else
    {
	rc = wait_event_interruptible (phead->wait_object, phead->head != NULL);
	if (rc == -ERESTARTSYS)
	{
	    /* user hit ^C, just exit */
	    return 0;
	}

	DT_Mdep_Lock (&phead->lock);
	entry = phead->head;
	if (entry)
	{
	    phead->head = entry->next;
	    if (phead->head == NULL)
	    {
		phead->tail = NULL;
	    }
	    strncpy (buffer, entry->buffer, PRINT_MAX);
	    DT_Mdep_Free (entry);
	}
	DT_Mdep_Unlock (&phead->lock);
    }

    len = strlen (buffer);
    if (len == 0)
    {
	KDT_Print_Free (phead);
    }
    return (len);
}
