/*
 * Copyright (c) 2002-2003, Network Appliance, Inc. All rights reserved.
 *
 * This Software is licensed under both of the following two licenses:
 *
 * 1) under the terms of the "Common Public License 1.0" a copy of which is
 *    in the file LICENSE.txt in the root directory. The license is also
 *    available from the Open Source Initiative, see
 *    http://www.opensource.org/licenses/cpl.php.
 * OR
 *
 * 2) under the terms of the "The BSD License" a copy of which is in the file
 *    LICENSE2.txt in the root directory. The license is also available from
 *    the Open Source Initiative, see
 *    http://www.opensource.org/licenses/bsd-license.php.
 *
 * Licensee has the right to choose either one of the above two licenses.
 *
 * Redistributions of source code must retain both the above copyright
 * notice and either one of the license notices.
 *
 * Redistributions in binary form must reproduce both the above copyright
 * notice, either one of the license notices in the documentation
 * and/or other materials provided with the distribution.
 */

#include "dapl_proto.h"

void
DT_performance_stats_init (
    Performance_Stats_t 		*stats)
{
    stats->num_ops = 0;
    stats->bytes = 0;
    stats->post_ctxt_switch_num = 0;
    stats->reap_ctxt_switch_num = 0;

    stats->time_ts = 0;

    stats->posts_sans_ctxt.num = 0;
    stats->posts_sans_ctxt.total_ts = 0;
    stats->posts_sans_ctxt.max_ts = 0;
    stats->posts_sans_ctxt.min_ts = ~0;

    stats->posts_with_ctxt.num = 0;
    stats->posts_with_ctxt.total_ts = 0;
    stats->posts_with_ctxt.max_ts = 0;
    stats->posts_with_ctxt.min_ts = ~0;

    stats->reaps_sans_ctxt.num = 0;
    stats->reaps_sans_ctxt.total_ts = 0;
    stats->reaps_sans_ctxt.max_ts = 0;
    stats->reaps_sans_ctxt.min_ts = ~0;

    stats->reaps_with_ctxt.num = 0;
    stats->reaps_with_ctxt.total_ts = 0;
    stats->reaps_with_ctxt.max_ts = 0;
    stats->reaps_with_ctxt.min_ts = ~0;

    stats->latency.num = 0;
    stats->latency.total_ts = 0;
    stats->latency.max_ts = 0;
    stats->latency.min_ts = ~0;
}


void
DT_performance_stats_record_post (
    Performance_Stats_t 		*stats,
    unsigned long 			ctxt_switch_num,
    DT_Mdep_TimeStamp 			ts)
{
    if ( ctxt_switch_num )
    {
	stats->posts_with_ctxt.num++;
	stats->posts_with_ctxt.total_ts += ts;
	stats->posts_with_ctxt.max_ts =
	    DT_max (stats->posts_with_ctxt.max_ts, ts);
	stats->posts_with_ctxt.min_ts =
	    DT_min (stats->posts_with_ctxt.min_ts, ts);

	stats->post_ctxt_switch_num += ctxt_switch_num;
    }
    else
    {
	stats->posts_sans_ctxt.num++;
	stats->posts_sans_ctxt.total_ts += ts;
	stats->posts_sans_ctxt.max_ts =
	    DT_max (stats->posts_sans_ctxt.max_ts, ts);
	stats->posts_sans_ctxt.min_ts =
	    DT_min (stats->posts_sans_ctxt.min_ts, ts);
    }
}


void
DT_performance_stats_record_reap (
    Performance_Stats_t 		*stats,
    unsigned long 			ctxt_switch_num,
    DT_Mdep_TimeStamp 			ts)
{
    if ( ctxt_switch_num )
    {
	stats->reaps_with_ctxt.num++;
	stats->reaps_with_ctxt.total_ts += ts;
	stats->reaps_with_ctxt.max_ts =
	    DT_max (stats->reaps_with_ctxt.max_ts, ts);
	stats->reaps_with_ctxt.min_ts =
	    DT_min (stats->reaps_with_ctxt.min_ts, ts);

	stats->reap_ctxt_switch_num += ctxt_switch_num;
    }
    else
    {
	stats->reaps_sans_ctxt.num++;
	stats->reaps_sans_ctxt.total_ts += ts;
	stats->reaps_sans_ctxt.max_ts =
	    DT_max (stats->reaps_sans_ctxt.max_ts, ts);
	stats->reaps_sans_ctxt.min_ts =
	    DT_min (stats->reaps_sans_ctxt.min_ts, ts);
    }
}


void
DT_performance_stats_record_latency (
    Performance_Stats_t 		*stats,
    DT_Mdep_TimeStamp 			ts)
{
    stats->latency.num++;
    stats->latency.total_ts += ts;
    stats->latency.max_ts =
	DT_max (stats->latency.max_ts, ts);
    stats->latency.min_ts =
	DT_min (stats->latency.min_ts, ts);
}

void
DT_performance_stats_data_combine (
    Performance_Stats_Data_t 		*dest,
    Performance_Stats_Data_t 		*src_a,
    Performance_Stats_Data_t 		*src_b)
{
    dest->num = src_a->num + src_b->num;
    dest->total_ts = src_a->total_ts + src_b->total_ts;
    dest->max_ts = DT_max (src_a->max_ts, src_b->max_ts);
    dest->min_ts = DT_min (src_a->min_ts, src_b->min_ts);
}


void
DT_performance_stats_combine (
    Performance_Stats_t 		*dest,
    Performance_Stats_t 		*src_a,
    Performance_Stats_t 		*src_b)
{
    dest->num_ops =
	src_a->num_ops + src_b->num_ops;

    dest->bytes =
	src_a->bytes + src_b->bytes;

    dest->post_ctxt_switch_num =
	src_a->post_ctxt_switch_num + src_b->post_ctxt_switch_num;

    dest->reap_ctxt_switch_num =
	src_b->reap_ctxt_switch_num + src_b->reap_ctxt_switch_num;

    dest->time_ts = DT_max (src_a->time_ts, src_b->time_ts);

    DT_performance_stats_data_combine (&dest->posts_sans_ctxt,
				      &src_a->posts_sans_ctxt,
				      &src_b->posts_sans_ctxt);

    DT_performance_stats_data_combine (&dest->posts_with_ctxt,
				      &src_a->posts_with_ctxt,
				      &src_b->posts_with_ctxt);

    DT_performance_stats_data_combine (&dest->reaps_sans_ctxt,
				      &src_a->reaps_sans_ctxt,
				      &src_b->reaps_sans_ctxt);

    DT_performance_stats_data_combine (&dest->reaps_with_ctxt,
				      &src_a->reaps_with_ctxt,
				      &src_b->reaps_with_ctxt);

    DT_performance_stats_data_combine (&dest->latency,
				      &src_a->latency,
				      &src_b->latency);
}


