/*
 * Copyright (c) 2002-2003, Network Appliance, Inc. All rights reserved.
 *
 * This Software is licensed under both of the following two licenses:
 *
 * 1) under the terms of the "Common Public License 1.0" a copy of which is
 *    in the file LICENSE.txt in the root directory. The license is also
 *    available from the Open Source Initiative, see
 *    http://www.opensource.org/licenses/cpl.php.
 * OR
 *
 * 2) under the terms of the "The BSD License" a copy of which is in the file
 *    LICENSE2.txt in the root directory. The license is also available from
 *    the Open Source Initiative, see
 *    http://www.opensource.org/licenses/bsd-license.php.
 *
 * Licensee has the right to choose either one of the above two licenses.
 *
 * Redistributions of source code must retain both the above copyright
 * notice and either one of the license notices.
 *
 * Redistributions in binary form must reproduce both the above copyright
 * notice, either one of the license notices in the documentation
 * and/or other materials provided with the distribution.
 */

#include "dapl_proto.h"

void
DT_init_transaction_stats (Transaction_Stats_t * transaction_stats, unsigned int num)
{
    DT_Mdep_LockInit (&transaction_stats->lock);

    transaction_stats->wait_count = num;
    transaction_stats->num_ops = 0;
    transaction_stats->time_ms = 0;
    transaction_stats->bytes_send = 0;
    transaction_stats->bytes_recv = 0;
    transaction_stats->bytes_rdma_read = 0;
    transaction_stats->bytes_rdma_write = 0;
}

void
DT_transaction_stats_set_ready (DT_Tdep_Print_Head *phead, 
				Transaction_Stats_t * transaction_stats)
{
    DT_Mdep_Lock (&transaction_stats->lock);
    transaction_stats->wait_count--;

    DT_Tdep_PT_Debug (1,(phead,"Received Sync Message from server (%d left)\n",
		    transaction_stats->wait_count));
    DT_Mdep_Unlock (&transaction_stats->lock);
}

bool
DT_transaction_stats_wait_for_all (DT_Tdep_Print_Head *phead,
				   Transaction_Stats_t * transaction_stats)
{
    unsigned int    loop_count;
    loop_count = 100 * 10;	/* 100 * 10ms * 10  = 10 seconds */
    while (transaction_stats->wait_count != 0 && loop_count != 0)
    {
	DT_Mdep_Sleep (10);
	loop_count--;
    }
    if (loop_count == 0)
    {
	DT_Tdep_PT_Printf (phead, "FAIL: %d Server test connections did not report ready.\n",
		    transaction_stats->wait_count);
	return false;
    }
    return true;
}


/*
 *
 */
void
DT_update_transaction_stats (Transaction_Stats_t * transaction_stats,
		       unsigned int num_ops,
		       unsigned int time_ms,
		       unsigned int bytes_send,
		       unsigned int bytes_recv,
		       unsigned int bytes_rdma_read,
		       unsigned int bytes_rdma_write)
{
    DT_Mdep_Lock (&transaction_stats->lock);

    /* look for the longest time... */
    if (time_ms > transaction_stats->time_ms)
    {
	transaction_stats->time_ms = time_ms;
    }

    transaction_stats->num_ops += num_ops;
    transaction_stats->bytes_send += bytes_send;
    transaction_stats->bytes_recv += bytes_recv;
    transaction_stats->bytes_rdma_read += bytes_rdma_read;
    transaction_stats->bytes_rdma_write += bytes_rdma_write;
    DT_Mdep_Unlock (&transaction_stats->lock);
}

