/*
 * This Software is licensed under both of the following two licenses:
 *
 * 1) under the terms of the "Common Public License 1.0" a copy of which is
 *    in the file LICENSE.txt in the root directory. The license is also
 *    available from the Open Source Initiative, see
 *    http://www.opensource.org/licenses/cpl.php.
 * OR
 *
 * 2) under the terms of the "The BSD License" a copy of which is in the file
 *    LICENSE2.txt in the root directory. The license is also available from
 *    the Open Source Initiative, see
 *    http://www.opensource.org/licenses/bsd-license.php.
 *
 * Licensee has the right to choose either one of the above two licenses.
 *
 * Redistributions of source code must retain both the above copyright
 * notice and either one of the license notices.
 *
 * Redistributions in binary form must reproduce both the above copyright
 * notice, either one of the license notices in the documentation
 * and/or other materials provided with the distribution.
 */

/***************************************************************************
 *
 *   Module:		 uDAPL
 *
 *   Filename:		 dapl_topspin.c
 *
 *   Author:		 Ariel Cohen
 *
 *   Created:		 10/24/2002
 *
 *   Description:
 *
 *   The Topspin uDAPL enhancements
 *
 ****************************************************************************
 *		   Source Control System Information
 *
 *    $Id: dapl_openib.c,v 1.1 2004/09/16 13:37:22 sjs2 Exp $
 *
 ****************************************************************************
 *
 ***************************************************************************
 *			 (c) COPYRIGHT 2002
 *			     Topspin Communications, Inc.
 *			     Mountain View, CA
 **************************************************************************/

#ifdef RCSID
static const char rcsid[] = "$Id$";
#endif


#ifndef __KDAPL__

#include <unistd.h>
#include <fcntl.h>
#include <sys/ioctl.h>

#endif



 
#include "vapi_types.h"
#include "vapi.h"
#include "evapi.h"

#include "dapl_openib_cm.h"

#include "dapl.h"

#include "dapl_openib.h"



static int udapl_fd = -1;


int ts_init(void)
{
    udapl_fd = open(UDAPL_DEVFILE, O_RDONLY);

    if (udapl_fd < 0) {
        dapl_dbg_log (DAPL_DBG_TYPE_ERR, "ERROR %s: failed to open uDAPL device\n", __func__);
        return -1;
    }

    return 0;
}


int ts_cleanup(void)
{
    if (udapl_fd != -1) 
	close(udapl_fd);

    return 0;
}


int ib_path_record_lookup(DAT_IA_ADDRESS_PTR     remote_ia_address,
                          tTS_IB_GID             src_gid,
                          struct ib_device      *ts_hca_handle,
                          tTS_IB_PORT            port,
                          tUINT32                ats_flag,
                          struct ib_path_record *path_record)
{
    path_record_param_t param;
    int                 status;
    //uint8_t dst_ip_addr[16];
    
    if ((remote_ia_address == NULL) || (path_record == NULL)) {
        dapl_dbg_log (DAPL_DBG_TYPE_ERR, "DAPL: %s called with a NULL parameter\n", __func__);
        return -1;
    }

   // param.dst_ip_addr = dst_ip_addr;
   // if (param.dst_ip_addr == NULL) {
   //     dapl_dbg_log (DAPL_DBG_TYPE_ERR, "DAPL: %s failed to allocate memory\n", __func__);
   //     return -1;
   // }

    //memset(param.dst_ip_addr, 0, 12);
   // memcpy(param.dst_ip_addr + 12, 
   //    &(((struct sockaddr_in *)remote_ia_address)->sin_addr.s_addr),
	//   4);
    param.dst_ip_addr   = remote_ia_address;
    param.src_gid       = &src_gid[0];
    param.ts_hca_handle = ts_hca_handle;
    param.port          = port;
    param.ats_flag      = ats_flag;
    param.path_record   = path_record;

    do {
	status = ioctl(udapl_fd, T_PATH_RECORD, &param);
    } while ((status < 0) && (errno == EINTR));


    if (status) {
        dapl_dbg_log (DAPL_DBG_TYPE_ERR, "DAPL: %s ioctl failed: %d\n", __func__, status);
        return -1;
    }
    
    return 0;
}


int ib_ats_ipaddr_lookup(uint8_t *gid, struct ib_device *hca_handle,
			 tTS_IB_PORT port, uint8_t *ip_addr)
{
    ats_ipaddr_lookup_param_t param;
    int                       status;

    if ((gid == NULL) || (ip_addr == NULL)) {
        dapl_dbg_log (DAPL_DBG_TYPE_ERR, "DAPL: %s called with a NULL parameter\n", __func__);
        return -1;
    }

    param.gid           = gid;
    param.ts_hca_handle = hca_handle;
    param.port          = port;
    param.ip_addr       = ip_addr;

    do {
	status = ioctl(udapl_fd, T_ATS_IPADDR_LOOKUP, &param);
    } while ((status < 0) && (errno == EINTR));

    if (status) {
        dapl_dbg_log (DAPL_DBG_TYPE_ERR, "DAPL: %s ioctl failed: %d\n", __func__, status);
        return -1;
    }
    
    return 0;
}

int ib_ats_set_ipaddr(uint8_t *gid, struct ib_device * ts_hca_handle, 
		      tTS_IB_PORT port, uint8_t *ip_addr)
{
    ats_set_ipaddr_param_t param;
    int                    status;

    if ((gid == NULL) || (ip_addr == NULL)) {
        dapl_dbg_log (DAPL_DBG_TYPE_ERR, "DAPL: %s called with a NULL parameter\n", __func__);
        return -1;
    }

    param.gid           = gid;
    param.ts_hca_handle = ts_hca_handle;
    param.port          = port;
    param.ip_addr       = ip_addr;

    do {
	status = ioctl(udapl_fd, T_ATS_SET_IPADDR, &param);
    } while ((status < 0) && (errno == EINTR));

    if (status) {
        dapl_dbg_log (DAPL_DBG_TYPE_ERR, "DAPL: %s ioctl failed: %d\n", __func__, status);
        return -1;
    }
    
    return 0;
}


int ib_release_khca_hndl(VAPI_hca_hndl_t k_hca_handle)
{
    int status;
    
    status = ioctl(udapl_fd, T_RELEASE_KHCA_HNDL, k_hca_handle);

    if (status) {
        dapl_dbg_log (DAPL_DBG_TYPE_ERR, "DAPL: %s ioctl failed: %d\n", __func__, status);
        return -1;
    }

    return 0;
}

int ts_smr_insert(DAT_LMR_COOKIE cookie, tUINT32 *exists)
{
    smr_insert_param_t param;
    int                status;


    param.cookie = cookie;
    
    status = ioctl(udapl_fd, T_SMR_INSERT, &param);

    if (status) {
        dapl_dbg_log (DAPL_DBG_TYPE_ERR, "DAPL: %s ioctl failed: %d\n", __func__, status);
        return -1;
    }
    
    *exists = param.exists;

    return 0;
}

int ts_smr_add_mrh(DAT_LMR_COOKIE cookie, VAPI_mr_hndl_t mr_handle)
{
    smr_add_mrh_param_t param;
    int                 status;


    param.cookie    = cookie;
    param.mr_handle = mr_handle;
    
    status = ioctl(udapl_fd, T_SMR_ADD_MRH, &param);

    if (status) {
        dapl_dbg_log (DAPL_DBG_TYPE_ERR, "DAPL: %s ioctl failed: %d\n", __func__, status);
        return -1;
    }
    
    return 0;
}

int ts_smr_del_mrh(DAT_LMR_COOKIE cookie, VAPI_mr_hndl_t mr_handle)
{
    smr_del_mrh_param_t param;
    int                 status;


    param.cookie    = cookie;
    param.mr_handle = mr_handle;
    
    status = ioctl(udapl_fd, T_SMR_DEL_MRH, &param);

    if (status) {
        dapl_dbg_log (DAPL_DBG_TYPE_ERR, "DAPL: %s ioctl failed: %d\n", __func__, status);
        return -1;
    }
    
    return 0;
}

int ts_smr_query(DAT_LMR_COOKIE cookie, tUINT32 *ready, VAPI_mr_hndl_t *mr_handle)
{
    smr_query_param_t param;
    int               status;


    param.cookie = cookie;
    
    status = ioctl(udapl_fd, T_SMR_QUERY, &param);

    if (status) {
        dapl_dbg_log (DAPL_DBG_TYPE_ERR, "DAPL: %s ioctl failed: %d\n", __func__, status);
        return -1;
    }
    
    *ready     = param.ready;
    *mr_handle = param.mr_handle;

    return 0;
}

int ts_smr_dec(DAT_LMR_COOKIE cookie)
{
    smr_dec_param_t param;
    int             status;


    param.cookie = cookie;
    
    status = ioctl(udapl_fd, T_SMR_DEC, &param);

    if (status) {
        dapl_dbg_log (DAPL_DBG_TYPE_ERR, "DAPL: %s ioctl failed: %d\n", __func__, status);
        return -1;
    }
    
    return 0;
}

int ts_smr_mutex_lock(void)
{
    int             status;


    status = ioctl(udapl_fd, T_SMR_MUTEX_LOCK);

    if (status) {
        dapl_dbg_log (DAPL_DBG_TYPE_ERR, "DAPL: %s ioctl failed: %d\n", __func__, status);
        return -1;
    }
    
    return 0;
}

int ts_smr_mutex_unlock(void)
{
    int             status;


    status = ioctl(udapl_fd, T_SMR_MUTEX_UNLOCK);

    if (status) {
        dapl_dbg_log (DAPL_DBG_TYPE_ERR, "DAPL: %s ioctl failed: %d\n", __func__, status);
        return -1;
    }
    
    return 0;
}


int ib_get_khca_hndl(VAPI_hca_id_t hca_id, VAPI_hca_hndl_t *k_hca_handle, void **ts_hca_handle)
{
    get_khca_hndl_param_t param;
    int                   status;

    memcpy(param.hca_id, hca_id, strlen(hca_id) + 1);
    status = ioctl(udapl_fd, T_GET_KHCA_HNDL, &param);

    if (status) {
        dapl_dbg_log (DAPL_DBG_TYPE_ERR, "DAPL: %s ioctl failed: %d\n", __func__, status);
        return -1;
    }

    *k_hca_handle  = param.k_hca_handle;
    *ts_hca_handle = param.ts_hca_handle;

    return 0;
}

int ib_get_hca_ipaddr (uint8_t *gid, tUINT32 *ip_addr)
{
    get_hca_ipaddr_param_t param;
    int                    status;

    
    param.gid = gid;
    
    status = ioctl(udapl_fd, T_GET_HCA_IPADDR, &param);
    
    if (status) {
        dapl_dbg_log (DAPL_DBG_TYPE_ERR, "DAPL: %s ioctl failed: %d\n", __func__, status);
        return -1;
    }
    
    *ip_addr = param.ip_addr;
    
    return 0;
}


