/*************************************************************************
*   Copyright (c) 2004 Mellanox Technologies Ltd.  All rights reserved.  *
*   Copyright (c) 2005 Mellanox Technologies.   All rights reserved.

*************************************************************************/

/*
 * System includes
 */
#include <unistd.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <pthread.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/ioctl.h>
#include <arpa/inet.h>

/*
 * Local includes
 */
#include "dapl.h"

/*
 * TS specific includes
 */
#include "dapl_openib_cm_util.h"


/* Mellanox stuff */
#include "vapi_types.h"
#include "vapi.h"
#include "evapi.h"


int ib_cm_vapi_disconnect(tTS_IB_CM_COMM_ID comm_id ) {
    return tsIbUCmDisconnect(comm_id);
};


int ib_cm_vapi_reject(tTS_IB_CM_COMM_ID comm_id,
                      void             *rej_private_data,
                      int               rej_private_data_len) {
    return tsIbUCmReject (comm_id,
                          rej_private_data,
                          rej_private_data_len);
};



int ib_cm_vapi_confirm(tTS_IB_CM_COMM_ID comm_id,
                       void             *rtu_private_data,
                       int               rtu_private_data_len) {
    return tsIbUCmConfirm(comm_id,
                          rtu_private_data,
                          rtu_private_data_len);
};





int ib_cm_vapi_kill(tTS_IB_CM_COMM_ID comm_id) {
    return tsIbUCmDropConsumer(comm_id);
};


int ib_cm_vapi_listen_stop (tTS_IB_LISTEN_HANDLE listen_handle){
    return tsIbUCmListenStop(listen_handle);
};


int ib_cm_vapi_establish(tTS_IB_CM_COMM_ID comm_id,
                         int               immediate) {
    return tsIbUCmEstablish (comm_id);
}

int ib_cm_vapi_accept(VAPI_hca_hndl_t                hca_handle,
                      void*                          pd,
                      tTS_IB_U_QP_HANDLE             vqp_handle,
                      __u32                          qpn,
                      tTS_IB_U_CM_COMPLETION_FUNC    function,
                      void                          *comp_arg,
                      tTS_IB_CM_COMM_ID              comm_id,
                      struct ib_cm_passive_param    *param)

{

  return tsIbUCmAccept (comm_id,
                        hca_handle,
                        vqp_handle,
                        qpn,
                        function,
                        comp_arg,
                        param->reply_private_data,
                        param->reply_private_data_len);
}



int  ib_cm_vapi_connect (VAPI_hca_hndl_t                hca_handle,
                         void*                          pd,
                         tTS_IB_U_QP_HANDLE             vqp_handle,
                         __u32                          qpn,
                         struct ib_cm_active_param  *param,
                         struct ib_path_record      *primary_path,
                         struct ib_path_record      *alternate_path,
                         tTS_IB_SERVICE_ID           service_id,
                         int                         peer_to_peer,
                         tTS_IB_U_CM_COMPLETION_FUNC function,
                         void                       *arg,
                         tTS_IB_CM_COMM_ID          *comm_id)

{
    {
       
    struct ib_u_cm_qp_param      qp_param;
    qp_param.responder_resources = param->responder_resources;
    qp_param.initiator_depth     = param->initiator_depth;
    qp_param.retry_count         = param->retry_count;
    qp_param.rnr_retry_count     = param->rnr_retry_count;
    qp_param.cm_response_timeout = param->cm_response_timeout;
    qp_param.max_cm_retries      = param->max_cm_retries;
     
    return tsIbUCmConnect(hca_handle,
                          vqp_handle,
                          qpn,
                          &qp_param,
                          primary_path,
                          alternate_path,
                          service_id,
                          param->req_private_data,
                          param->req_private_data_len,
                          function,
                          arg,
                          comm_id
                        );
    }
}


int ib_cm_vapi_listen(tTS_IB_SERVICE_ID           service_id,
                      tTS_IB_SERVICE_ID           service_mask,
                      tTS_IB_U_CM_COMPLETION_FUNC function,
                      void                       *arg,
                      tTS_IB_LISTEN_HANDLE       *listen_handle)
{
  
    return tsIbUCmListen(service_id,
                         service_mask,
                         function,
                         arg,
                         listen_handle);
}
tTS_IB_SERVICE_ID ib_cm_vapi_service_assign(void)
{
    return tsIbUCmServiceAssign();
}

