/*
 * Copyright (c) 2002-2003, Network Appliance, Inc. All rights reserved.
 * Copyright (c) 2005 Mellanox Technologies.   All rights reserved.

 *
 * This Software is licensed under both of the following two licenses:
 *
 * 1) under the terms of the "Common Public License 1.0" a copy of which is
 *    in the file LICENSE.txt in the root directory. The license is also
 *    available from the Open Source Initiative, see
 *    http://www.opensource.org/licenses/cpl.php.
 * OR
 *
 * 2) under the terms of the "The BSD License" a copy of which is in the file
 *    LICENSE2.txt in the root directory. The license is also available from
 *    the Open Source Initiative, see
 *    http://www.opensource.org/licenses/bsd-license.php.
 *
 * Licensee has the right to choose either one of the above two licenses.
 *
 * Redistributions of source code must retain both the above copyright
 * notice and either one of the license notices.
 *
 * Redistributions in binary form must reproduce both the above copyright
 * notice, either one of the license notices in the documentation
 * and/or other materials provided with the distribution.
 */

/**********************************************************************
 * 
 * MODULE: dapl_evd_set_unwaitable.c
 *
 * PURPOSE: EVENT management
 * Description: Interfaces in this file are completely described in
 *		the DAPL 1.1 API, Chapter 6, section 3.4.7
 *
 * $Id: dapl_evd_set_unwaitable.c,v 1.2 2004/08/02 15:14:28 rabin Exp $
 **********************************************************************/

#include "dapl.h"
#include "dapl_adapter_util.h"

/*
 * dapl_evd_set_unwaitable
 *
 * DAPL Requirements Version 1.1, 6.3.4.7
 *
 * Transition the Event Dispatcher into an unwaitable state
 *
 * Input:
 * 	evd_handle
 *
 * Output:
 *	none
 *
 * Returns:
 *	DAT_SUCCESS
 *	DAT_INVALID_HANDLE
 */
DAT_RETURN
dapl_evd_set_unwaitable (
	IN    DAT_EVD_HANDLE	evd_handle )
{
    DAPL_EVD		*evd_ptr;
    DAT_RETURN		dat_status;

    evd_ptr    = (DAPL_EVD *)evd_handle;
    dat_status = DAT_SUCCESS;

    if ( DAPL_BAD_HANDLE (evd_handle, DAPL_MAGIC_EVD) )

    {
	dat_status = DAT_ERROR (DAT_INVALID_HANDLE,0);
	goto bail;
    }
    dapl_os_lock ( &evd_ptr->header.lock );
    evd_ptr->evd_waitable = DAT_FALSE;
    dapl_os_unlock ( &evd_ptr->header.lock );

    /*
     * If this evd is waiting, wake it up. There is an obvious race
     * condition here where we may wakeup the waiter before it goes to
     * sleep; but the wait_object allows this and will do the right
     * thing.
     */
    if (evd_ptr->evd_state == DAPL_EVD_STATE_WAITED)
    {
//	dapl_os_wait_object_wakeup (&evd_ptr->wait_object);
	dapl_evd_wait_object_wakeup (evd_ptr);
    }
bail:
    return dat_status;
}

/*
 * Local variables:
 *  c-indent-level: 4
 *  c-basic-offset: 4
 *  tab-width: 8
 * End:
 */
