/*
 * Copyright (c) 2002-2003, Network Appliance, Inc. All rights reserved.
 * Copyright (c) 2005 Mellanox Technologies.   All rights reserved.
 *
 * This Software is licensed under both of the following two licenses:
 *
 * 1) under the terms of the "Common Public License 1.0" a copy of which is
 *    in the file LICENSE.txt in the root directory. The license is also
 *    available from the Open Source Initiative, see
 *    http://www.opensource.org/licenses/cpl.php.
 * OR
 *
 * 2) under the terms of the "The BSD License" a copy of which is in the file
 *    LICENSE2.txt in the root directory. The license is also available from
 *    the Open Source Initiative, see
 *    http://www.opensource.org/licenses/bsd-license.php.
 *
 * Licensee has the right to choose either one of the above two licenses.
 *
 * Redistributions of source code must retain both the above copyright
 * notice and either one of the license notices.
 *
 * Redistributions in binary form must reproduce both the above copyright
 * notice, either one of the license notices in the documentation
 * and/or other materials provided with the distribution.
 */

#include "dapl_proto.h"

/* Main Entry Point */
int
main (int argc, char *argv[])
{
    return ( dapltest (argc, argv) );
}


/*
 * dapltest main program
 */
int
dapltest (int argc, char *argv[])
{
    Params_t       	*params_ptr;
    int status;

    /* check memory leaking */
    /*
     * DT_Mdep_LockInit(&Alloc_Count_Lock); alloc_count = 0;
     */

#if 0
#if defined(WIN32)
    /* Cannot be done from DT_Mdep_Init as dapl_init makes some socket
     * calls....So need to do this before calling dapl_init */
    WORD            wv = MAKEWORD (1, 1);
    WSADATA         wsaData;
    WSAStartup (wv, &wsaData);
#endif
#endif

#ifdef GPROF
    {
	extern void dapl_init (void);
	dapl_init ();
    }
#endif
    DT_dapltest_debug = 0;

    params_ptr = (Params_t *) DT_Mdep_Malloc (sizeof (Params_t));

    if (!params_ptr)
    {
	DT_Mdep_printf ("Cannot allocate memory for Params structure\n");
	return ( 1 );
    }
    memset((void *)params_ptr, 0, sizeof (Params_t));
    DT_Tdep_Init ();	/* init (kdapl/udapl)test-dependent code */
    DT_Endian_Init ();	/* init endian of local machine */
    DT_Mdep_Init ();	/* init OS, libraries, etc.     */

    params_ptr->local_is_little_endian = DT_local_is_little_endian;

    /*
     * parse command line arguments
     */
    if (!DT_Params_Parse (argc, argv, params_ptr))
    {
	DT_Mdep_printf ("Command line syntax error\n");
	return 1;
    }
    /* call the test-dependent code for invoking the actual test */

    status = DT_Tdep_Execute_Test (params_ptr);

    printf("TEST return code = %d \n",status); 
    /* cleanup */

    DT_Mdep_End ();
    DT_Mdep_Free (params_ptr);
    DT_Tdep_End ();
#ifdef GPROF
    {
	extern void dapl_fini (void);
	dapl_fini ();
    }
#endif

    /*
     * check memory leaking DT_Mdep_printf("App allocated Memory left: %d\n",
     * alloc_count); DT_Mdep_LockDestroy(&Alloc_Count_Lock);
     */

    return status;
}


void
Dapltest_Main_Usage (void)
{
    DT_Mdep_printf ("USAGE:\n");
    DT_Mdep_printf ("USAGE:     dapltest -T <Test_Type> [test-specific args]\n");
    DT_Mdep_printf ("USAGE:         where <Test_Type>\n");
    DT_Mdep_printf ("USAGE:         S = Run as a server\n");
    DT_Mdep_printf ("USAGE:         T = Transaction Test\n");
    DT_Mdep_printf ("USAGE:         Q = Quit Test\n");
    DT_Mdep_printf ("USAGE:         P = Performance Test\n");
    DT_Mdep_printf ("USAGE:         L = Limit Test\n");
    DT_Mdep_printf ("USAGE:         F = FFT Test\n");
    DT_Mdep_printf ("USAGE:\n");
    DT_Mdep_printf ("NOTE:\tRun as server taking defaults (dapltest -T S)\n");
    DT_Mdep_printf ("NOTE:         dapltest\n");
    DT_Mdep_printf ("NOTE:\n");
    DT_Mdep_printf ("NOTE:\tdapltest arguments may be supplied in a script file\n");
    DT_Mdep_printf ("NOTE:\tdapltest -f <script_file>\n");
    DT_Mdep_printf ("USAGE:\n");
}
