/*
 * Copyright (c) 2002-2003, Network Appliance, Inc. All rights reserved.
 * Copyright (c) 2005 Mellanox Technologies.   All rights reserved.
 *
 * This Software is licensed under both of the following two licenses:
 *
 * 1) under the terms of the "Common Public License 1.0" a copy of which is
 *    in the file LICENSE.txt in the root directory. The license is also
 *    available from the Open Source Initiative, see
 *    http://www.opensource.org/licenses/cpl.php.
 * OR
 *
 * 2) under the terms of the "The BSD License" a copy of which is in the file
 *    LICENSE2.txt in the root directory. The license is also available from
 *    the Open Source Initiative, see
 *    http://www.opensource.org/licenses/bsd-license.php.
 *
 * Licensee has the right to choose either one of the above two licenses.
 *
 * Redistributions of source code must retain both the above copyright
 * notice and either one of the license notices.
 *
 * Redistributions in binary form must reproduce both the above copyright
 * notice, either one of the license notices in the documentation
 * and/or other materials provided with the distribution.
 */

#include "dapl_proto.h"

/*
 * Parse a QOS argument into a DAT_QOS.
 *
 * Returns no errors: defaults to best effort.
 */
DAT_QOS
DT_ParseQoS (char *arg)
{
    if (0 == strcmp (arg, "HT"))
    {
	return ( DAT_QOS_HIGH_THROUGHPUT );
    }

    if (0 == strcmp (arg, "LL"))
    {
	return ( DAT_QOS_LOW_LATENCY );
    }

    if (0 == strcmp (arg, "EC"))
    {
	return ( DAT_QOS_ECONOMY );
    }

    if (0 == strcmp (arg, "PM"))
    {
	return ( DAT_QOS_PREMIUM );
    }
    /*
     * Default to "BE" so no point in checking further
     */
    return ( DAT_QOS_BEST_EFFORT );
}

/* cant use the enum because we are in user and we need it for kernel  */
/* typedef enum dat_mem_type               */
/* {                                       */
/*      Shared between udat and kdat   */
/*     DAT_MEM_TYPE_VIRTUAL        = 0x00, */
/*     DAT_MEM_TYPE_LMR            = 0x01, */
/*      kdat specific                  */
/*     DAT_MEM_TYPE_PHYSICAL       = 0x10, */
/*     DAT_MEM_TYPE_PLATFORM       = 0x20, */
/*     DAT_MEM_TYPE_IA             = 0x40, */
/*     DAT_MEM_TYPE_BYPASS         = 0x80  */
/* } DAT_MEM_TYPE;                         */
/*                                         */




DAT_MEM_TYPE
DT_ParseMemType (char *arg)
{
#ifndef __KDAPLTEST__
#else
    if (0 == strcmp (arg, "PHY"))
    {
	return (0x10  );/*DAT_MEM_TYPE_PHYSICAL */
    }

    if (0 == strcmp (arg, "IA"))
    {
	return (0x40  );/* DAT_MEM_TYPE_IA  */
    }

    if (0 == strcmp (arg, "FMR"))
    {
	return (0x20  );/* DAT_MEM_TYPE_PLATFORM */
    }

    if (0 == strcmp (arg, "BP"))
    {
	return (0x80  );/* DAT_MEM_TYPE_BYPASS  */
    }
#endif
    /*
     * Default to "VIR" so no point in checking further
     */
	return ( DAT_MEM_TYPE_VIRTUAL );
}


