/*
 * Copyright (c) 2002-2003, Network Appliance, Inc. All rights reserved.
 * Copyright (c) 2005 Mellanox Technologies.   All rights reserved.
 *
 * This Software is licensed under both of the following two licenses:
 *
 * 1) under the terms of the "Common Public License 1.0" a copy of which is
 *    in the file LICENSE.txt in the root directory. The license is also
 *    available from the Open Source Initiative, see
 *    http://www.opensource.org/licenses/cpl.php.
 * OR
 *
 * 2) under the terms of the "The BSD License" a copy of which is in the file
 *    LICENSE2.txt in the root directory. The license is also available from
 *    the Open Source Initiative, see
 *    http://www.opensource.org/licenses/bsd-license.php.
 *
 * Licensee has the right to choose either one of the above two licenses.
 *
 * Redistributions of source code must retain both the above copyright
 * notice and either one of the license notices.
 *
 * Redistributions in binary form must reproduce both the above copyright
 * notice, either one of the license notices in the documentation
 * and/or other materials provided with the distribution.
 */

#include "dapl_proto.h"
#include "kdapl_ioctl.h"
#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <fcntl.h>
#include <errno.h>

int	fd = -1;

void
DT_Tdep_Init (void)
{
    fd = open ("/dev/kdapltest", O_RDWR);
    if (fd == -1)
    {
	fprintf (stderr, "Cannot open device \"/dev/kdapltest\"\n");
	fprintf (stderr, "You must first insmod kdapltest_module.o\n");
	exit (1);
    }
}

void
DT_Tdep_End (void)
{
    close (fd);
}

int
DT_Tdep_Execute_Test (Params_t *params_ptr)
{
    DT_get_printf_ioctl print_ioctl;
    char buffer[PRINT_MAX];
    int status;
    Transaction_Stats2_t Client_Stats_T;
    Performance_Stats_t Client_Stats_P;
    print_ioctl.buffer = &buffer[0];
    print_ioctl.cookie = ioctl (fd, KDAPL_IOCTL_START_TEST, params_ptr);
    if (print_ioctl.cookie < 0)
    {
	fprintf (stderr, "IOCTL failed: errno %d\n", errno);
	exit (1);
    }
    while (ioctl (fd, KDAPL_IOCTL_GET_PRINTF, &print_ioctl) > 0)
    {
	printf ("%s", &buffer[0]);
    }
    if (params_ptr->test_type == TRANSACTION_TEST) {
        print_ioctl.cookie = ioctl (fd, KDAPL_IOCTL_GET_STAT_T, &Client_Stats_T);
        DT_print_transaction_stats(&Client_Stats_T,params_ptr->u.Transaction_Cmd.num_iterations,
                                   params_ptr->u.Transaction_Cmd.eps_per_thread);
    }

    if (params_ptr->test_type == PERFORMANCE_TEST) {
        print_ioctl.cookie = ioctl (fd, KDAPL_IOCTL_GET_STAT_P, &Client_Stats_P);
        DT_performance_stats_print (&Client_Stats_P, &(params_ptr->u.Performance_Cmd));
    }
    print_ioctl.cookie = ioctl (fd, KDAPL_IOCTL_END_TEST, &status);
    printf ("TEST INSTANCE %d\n", print_ioctl.cookie);
    if (print_ioctl.cookie < 0)
    {
	fprintf (stderr, "IOCTL failed: errno %d\n", errno);
	exit (1);
    }
    return status;
}
