/*
  This software is available to you under a choice of one of two
  licenses.  You may choose to be licensed under the terms of the GNU
  General Public License (GPL) Version 2, available at
  <http://www.fsf.org/copyleft/gpl.html>, or the OpenIB.org BSD
  license, available in the LICENSE.TXT file accompanying this
  software.  These details are also available at
  <http://openib.org/license.html>.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
  BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
  ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
  CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
  SOFTWARE.

  Copyright (c) 2004 Topspin Communications.  All rights reserved.
  Copyright (c) 2005 Mellanox Technologies.   All rights reserved.


  $Id: ipoib.h 432 2004-07-12 17:27:37Z roland $
*/

#ifndef _IPOIB_H
#define _IPOIB_H

#include <linux/list.h>
#include <linux/skbuff.h>
#include <linux/netdevice.h>	/* struct device, and other headers */
#include <linux/proc_fs.h>
#include <linux/completion.h>

#include <asm/atomic.h>
#include <asm/semaphore.h>

#include "ipoib_proto.h"

#include <ts_ib_core.h>
#include <ts_ib_header_types.h>

#include <ts_ib_sa_client.h>

#include <ts_kernel_services.h>
#include <ts_kernel_thread.h>

#ifdef TS_KERNEL_2_6
#define tq_struct work_struct
#define INIT_TQUEUE(x,y,z) INIT_WORK(x,y,z)
#define schedule_task schedule_work
#endif

extern int ipoib_tx_ring_size;
#define IPOIB_RX_RING_SIZE (ipoib_tx_ring_size * 2)
#define IPOIB_TX_RING_SIZE (ipoib_tx_ring_size)

/* constants */
#ifndef MAC_20_BYTES

#define ARPHRD_INFINIBAND 32
#define IPOIB_PACKET_SIZE	2048

enum {

	IPOIB_BUF_SIZE = IPOIB_PACKET_SIZE + sizeof(struct ib_grh),

	IPOIB_ADDRESS_HASH_BYTES = ETH_ALEN,
	IPOIB_ENCAP_LEN = 4,
	IPOIB_HW_ADDR_LEN = 20,

	IPOIB_FLAG_TX_FULL = 0,
	IPOIB_FLAG_TIMEOUT = 1,
	IPOIB_FLAG_OPER_UP = 2,
	IPOIB_FLAG_ADMIN_UP = 3,
	IPOIB_PKEY_ASSIGNED = 4,
	IPOIB_FLAG_SUBINTERFACE = 5,
  

	IPOIB_MAX_BACKOFF_SECONDS = 32,

	IPOIB_MCAST_FLAG_FOUND = 0,	/* used in set_multicast_list */
	IPOIB_MCAST_FLAG_SENDONLY = 1,
	IPOIB_MCAST_FLAG_BUSY = 2,	/* joining or already joined */
	IPOIB_MCAST_FLAG_ATTACHED = 3,
	IPOIB_MCAST_FLAG_REJOIN   = 4, /* Rejoining should start*/
};

#else
union ib_gid {
	u8	raw[16];
	struct {
		u64	subnet_prefix;
		u64	interface_id;
	} global;
};

enum {
    IPOIB_PACKET_SIZE     = 2048,
	IPOIB_BUF_SIZE 		  = IPOIB_PACKET_SIZE + sizeof(struct ib_grh),
	IPOIB_ENCAP_LEN 	  = 4,
	IPOIB_HW_ADDR_LEN 	  = 20, 
	
	IPOIB_MAX_PATH_REC_QUEUE  = 3,
	IPOIB_MAX_MCAST_QUEUE     = 3,
	
	IPOIB_FLAG_TIMEOUT = 1,

	IPOIB_FLAG_TX_FULL        = 0,
 	IPOIB_FLAG_OPER_UP 	  = 1,
	IPOIB_FLAG_ADMIN_UP       = 2,
	IPOIB_PKEY_ASSIGNED 	  = 3,
	IPOIB_PKEY_STOP 	  = 4,
	IPOIB_FLAG_SUBINTERFACE   = 5,
	IPOIB_MCAST_STOP 	  = 6,
	IPOIB_MAX_BACKOFF_SECONDS = 32,
	IPOIB_MAX_UPDATE_MULTI = 900,
 


	IPOIB_MCAST_FLAG_FOUND    = 0,	/* used in set_multicast_list */
	IPOIB_MCAST_FLAG_SENDONLY = 1,
	IPOIB_MCAST_FLAG_BUSY     = 2,	/* joining or already joined */
	IPOIB_MCAST_FLAG_ATTACHED = 3,
	IPOIB_MCAST_FLAG_REJOIN   = 4, /* Rejoining should start*/
};
#endif
/* structs */

struct ipoib_mcast;

#ifndef MAC_20_BYTES
typedef void (*ipoib_tx_callback_t)(void *);
struct ipoib_sarp;
struct ipoib_tx_buf {
        struct sk_buff *skb;
        ipoib_tx_callback_t callback;
        void *ptr;
        DECLARE_PCI_UNMAP_ADDR(mapping)
        int skb_not_alloc;
};

struct ipoib_rx_buf {
        struct sk_buff *skb;
        DECLARE_PCI_UNMAP_ADDR(mapping)
        int skb_not_alloc;
};

#else
struct ipoib_header {
        u16 proto;
        u16 reserved;
};

struct ipoib_pseudoheader {
	u8  hwaddr[IPOIB_HW_ADDR_LEN];
};

struct ipoib_dummy_header {
	u8  hwaddr[10];
};

struct ipoib_buf {	
        struct sk_buff *skb;
        DECLARE_PCI_UNMAP_ADDR(mapping)
        int skb_not_alloc;
};


#endif

struct ipoib_neigh_ops {
        struct list_head list;
        struct neigh_ops *ops;
};

struct ipoib_dev_priv {
        spinlock_t lock;
        spinlock_t rq_lock;

        struct net_device dev;
        struct list_head list;

        struct list_head child_intfs;
#ifdef MAC_20_BYTES
        struct list_head created_ahs;
        struct list_head ops_list;
#endif        
        unsigned long flags;

        struct semaphore mcast_mutex;

        tTS_KERNEL_THREAD mcast_thread;
#ifdef MAC_20_BYTES        
        tTS_KERNEL_THREAD mcast_update_thread;
        wait_queue_head_t mcast_update_wait;
        int mcast_update_stop;
        int mcast_rejoin;
#endif
        tTS_IB_CLIENT_QUERY_TID mcast_tid;
        int mcast_status;
        struct completion mcast_complete;

        struct ipoib_mcast *broadcast;
        struct list_head multicast_list;
        rb_root_t multicast_tree;

        atomic_t mcast_joins;
 
        struct tq_struct restart_task;

        struct ib_device *ca;
        tTS_IB_PORT port;
        struct ib_pd *pd;
        struct ib_mr *mr;
        u32 lkey;
        struct ib_cq *cq_tq;
        struct ib_cq *cq_rq;  
        struct ib_qp *qp;
        u32 qkey;
        u16 pkey;
        tTS_KERNEL_THREAD pkey_thread;
#ifndef MAC_20_BYTES
        tTS_IB_GID local_gid;
#else
        union ib_gid local_gid;
#endif        
        u16        local_lid;
        u32        local_qpn;

       unsigned int admin_mtu;
       unsigned int mcast_mtu;

#ifndef MAC_20_BYTES
       tTS_IB_GID bcast_gid;
       struct ipoib_rx_buf *rx_ring;
       struct ipoib_tx_buf *tx_ring;
       struct ipoib_sarp_cache *sarp_cache;
       struct proc_dir_entry *arp_proc_entry;
#else
       struct ipoib_buf *rx_ring;
       struct ipoib_buf *tx_ring;
       int    ah_amount;      
#endif 
       int tx_head;
       volatile int tx_free;
       int tx_last_wri_handel;
       int tx_hanging;
       int rx_last_post_receive;
       int rx_last_used;

       struct proc_dir_entry *mcast_proc_entry;

       struct ib_async_event_handler *active_handler;
       struct ib_async_event_handler *mcast_rejoin_handler;

       struct net_device_stats stats;
};
#ifdef MAC_20_BYTES
//struct ipoib_ah {
//	struct net_device *dev;
//	struct ib_address  *ah;
//	struct list_head   list;
//  	unsigned           last_send;
//};

struct ipoib_path {
	struct ib_address   *ah;
   	u32                 qpn;
	struct sk_buff_head queue;
	struct net_device  *dev;
	struct neighbour   *neighbour;
	struct list_head   list;
	union  ib_gid      gid;
	u8                 unused_ah;
	u8                 get_counter;
	tTS_IB_CLIENT_QUERY_TID tid;
	u32                 ipaddr;
};


#define IPOIB_PATH(neigh) (*(struct ipoib_path **) ((neigh)->ha + 24))


#endif

/* list of IPoIB network devices */
extern struct semaphore ipoib_device_mutex;
extern struct list_head ipoib_device_list;

extern tTS_IB_GID ipoib_broadcast_mgid;

/* functions */

void ipoib_ib_rq_completion(struct ib_cq *cq, struct ib_cq_entry *entry,
			 void *dev_ptr);

void ipoib_ib_tq_completion(struct ib_cq *cq, struct ib_cq_entry *entry,
			 void *dev_ptr);

void ipoib_ib_tq_event_handler(struct ib_cq *cq, void *dev_ptr);

void ipoib_ib_finish_completion(struct net_device *dev);
#ifndef MAC_20_BYTES
int ipoib_dev_send(struct net_device *dev, struct sk_buff *skb,
		   ipoib_tx_callback_t callback,
		   void *ptr, struct ib_address *address, u32 qpn);

#else
void ipoib_send(struct net_device *dev, struct sk_buff *skb,
		struct ib_address *address, u32 qpn);

void ipoib_free_ah(struct ipoib_path *dpath);
struct ipoib_path *ipoib_get_path(struct net_device *dev, u8 *hwaddr, u32 ipaddr);
struct ipoib_path *ipoib_path_look_up(struct net_device *dev, u8 *hwaddr, u32 ipaddr);
int ipoib_reproduce_ah(struct ipoib_path *new_path, struct ipoib_path *old_path);
void ipoib_path_cache_old(struct net_device *dev);
void ipoib_delete_unused_ah(struct net_device *dev);

#endif
struct ipoib_dev_priv *ipoib_intf_alloc(void);

int ipoib_ib_dev_init(struct net_device *dev, struct ib_device *ca, int port);
void ipoib_ib_dev_flush(void *dev);
void ipoib_ib_dev_cleanup(struct net_device *dev);

int ipoib_ib_dev_open(struct net_device *dev);
int ipoib_ib_dev_up(struct net_device *dev);
int ipoib_ib_dev_down(struct net_device *dev);
int ipoib_ib_dev_stop(struct net_device *dev);

int ipoib_dev_init(struct net_device *dev, struct ib_device *ca, int port);
void ipoib_dev_cleanup(struct net_device *dev);

#ifndef MAC_20_BYTES
void ipoib_sarp_get(struct ipoib_sarp *entry);
void ipoib_sarp_put(struct ipoib_sarp *entry);
struct ipoib_sarp *ipoib_sarp_add(struct net_device *dev, tTS_IB_GID gid,
				  u32 qpn);
struct ipoib_sarp *ipoib_sarp_local_add(struct net_device *dev, tTS_IB_GID gid,
					u32 qpn);
int ipoib_sarp_delete(struct net_device *dev, const uint8_t *hash);
int ipoib_sarp_lookup(struct net_device *dev, uint8_t *hash,
		      struct ipoib_sarp **entry);
int ipoib_sarp_queue_packet(struct ipoib_sarp *entry, struct sk_buff *skb);
int ipoib_sarp_send(struct net_device *dev, struct ipoib_sarp *entry,
		    struct sk_buff *skb);
int ipoib_sarp_rewrite_receive(struct net_device *dev, struct sk_buff *skb);
int ipoib_sarp_rewrite_send(struct net_device *dev, struct sk_buff *skb);
int ipoib_sarp_dev_init(struct net_device *dev);
void ipoib_sarp_dev_flush(struct net_device *dev);
void ipoib_sarp_dev_cleanup(struct net_device *dev);

struct ipoib_sarp_iter *ipoib_sarp_iter_init(struct net_device *dev);
void ipoib_sarp_iter_free(struct ipoib_sarp_iter *iter);
int ipoib_sarp_iter_next(struct ipoib_sarp_iter *iter);
void ipoib_sarp_iter_read(struct ipoib_sarp_iter *iter, uint8_t *hash,
			  tTS_IB_GID gid, u32 *qpn,
			  unsigned long *created,
			  unsigned long *last_verify,
			  unsigned int *queuelen, unsigned int *complete);

#endif

int ipoib_proc_dev_init(struct net_device *dev);
void ipoib_proc_dev_cleanup(struct net_device *dev);

void ipoib_mcast_get(struct ipoib_mcast *mcast);
void ipoib_mcast_put(struct ipoib_mcast *mcast);
int ipoib_mcast_lookup(struct net_device *dev, tTS_IB_GID mgid,
		       struct ipoib_mcast **mcast);
int ipoib_mcast_queue_packet(struct ipoib_mcast *mcast, struct sk_buff *skb);
#ifndef MAC_20_BYTES
int ipoib_mcast_send(struct net_device *dev, struct ipoib_mcast *mcast,
		     struct sk_buff *skb);
#else
void ipoib_mcast_send(struct net_device *dev, union ib_gid *mgid,
		      struct sk_buff *skb);

int ipoib_mcast_update_stop_thread(struct net_device *dev);
int ipoib_mcast_update_start_thread(struct net_device *dev);
void ipoib_mcast_update_thread(void *dev_ptr);
#endif

void ipoib_mcast_dev_exit(struct net_device *dev);

void ipoib_mcast_restart_task(void *dev_ptr);

int ipoib_mcast_start_thread(struct net_device *dev);
void ipoib_mcast_restart_thread(struct ib_async_event_record * event, void *arg);
int ipoib_mcast_stop_thread(struct net_device *dev);

void ipoib_mcast_dev_down(struct net_device *dev);
void ipoib_mcast_dev_flush(struct net_device *dev);

struct ipoib_mcast_iter *ipoib_mcast_iter_init(struct net_device *dev);
void ipoib_mcast_iter_free(struct ipoib_mcast_iter *iter);
int ipoib_mcast_iter_next(struct ipoib_mcast_iter *iter);

#ifndef MAC_20_BYTES
void ipoib_mcast_iter_read(struct ipoib_mcast_iter *iter,
				  tTS_IB_GID gid,
				  unsigned long *created,
				  unsigned int *queuelen,
				  unsigned int *complete,
				  unsigned int *send_only);

int ipoib_mcast_attach(struct net_device *dev, u16 mlid,
		       tTS_IB_GID mgid);
int ipoib_mcast_detach(struct net_device *dev, u16 mlid,
		       tTS_IB_GID mgid);
#else
void ipoib_mcast_iter_read(struct ipoib_mcast_iter *iter,
				  union ib_gid *mgid,
				  unsigned long *created,
				  unsigned int *queuelen,
				  unsigned int *complete,
				  unsigned int *send_only);

int ipoib_mcast_attach(struct net_device *dev, u16 mlid,
		       union ib_gid *mgid);
int ipoib_mcast_detach(struct net_device *dev, u16 mlid,
		       union ib_gid *mgid);
#endif

int ipoib_qp_create(struct net_device *dev);
void ipoib_qp_destroy(struct net_device *dev);
int ipoib_transport_dev_init(struct net_device *dev, struct ib_device *ca);
void ipoib_transport_dev_cleanup(struct net_device *dev);

int ipoib_add_port(const char *format, struct ib_device *device,
		   tTS_IB_PORT port);

int ipoib_transport_create_devices(void);

void ipoib_transport_cleanup(void);

int ipoib_port_monitor_dev_start(struct net_device *dev);
void ipoib_port_monitor_dev_stop(struct net_device *dev);

int ipoib_vlan_init(void);
void ipoib_vlan_cleanup(void);

int ipoib_pkey_dev_start_thread(struct net_device *dev);
int ipoib_pkey_dev_stop_thread(struct net_device *dev);
void ipoib_pkey_dev_check_presence(struct net_device *dev);
int ipoib_pkey_dev_delay_open(struct net_device *dev);

#ifdef MAC_20_BYTES
void ipoib_ah_delete(struct net_device *dev);
#endif

#define IPOIB_GID_FMT		"%02x%02x%02x%02x%02x%02x%02x%02x" \
				"%02x%02x%02x%02x%02x%02x%02x%02x"

#ifdef MAC_20_BYTES
#define IPOIB_GID_ARG(gid)	(gid).raw[ 0], (gid).raw[ 1], (gid).raw[ 2], (gid).raw[ 3], \
				(gid).raw[ 4], (gid).raw[ 5], (gid).raw[ 6], (gid).raw[ 7], \
				(gid).raw[ 8], (gid).raw[ 9], (gid).raw[10], (gid).raw[11], \
				(gid).raw[12], (gid).raw[13], (gid).raw[14], (gid).raw[15]
                 
#define IPOIB_GID_ARG_OLD(gid)	gid[ 0], gid[ 1], gid[ 2], gid[ 3], \
				gid[ 4], gid[ 5], gid[ 6], gid[ 7], \
				gid[ 8], gid[ 9], gid[10], gid[11], \
				gid[12], gid[13], gid[14], gid[15]

#else
#define IPOIB_GID_ARG(gid)	gid[ 0], gid[ 1], gid[ 2], gid[ 3], \
				gid[ 4], gid[ 5], gid[ 6], gid[ 7], \
				gid[ 8], gid[ 9], gid[10], gid[11], \
				gid[12], gid[13], gid[14], gid[15]
    
#endif

#define IPOIB_MAX_PACKET_HANGING 2
#define IPOIB_MIN_DOORBELL 1
#define IPOIB_MAX_POST_REC 8
#define IPOIB_MAX_REPRODUCE_AH 10
#define IPOIB_MAX_RE_AH_NO_SA 1


#endif /* _IPOIB_H */
