/*
  This software is available to you under a choice of one of two
  licenses.  You may choose to be licensed under the terms of the GNU
  General Public License (GPL) Version 2, available at
  <http://www.fsf.org/copyleft/gpl.html>, or the OpenIB.org BSD
  license, available in the LICENSE.TXT file accompanying this
  software.  These details are also available at
  <http://openib.org/license.html>.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
  BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
  ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
  CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
  SOFTWARE.

  Copyright (c) 2004 Topspin Communications.  All rights reserved.
  Copyright (c) 2005 Mellanox Technologies.   All rights reserved.


  $Id: ipoib_ib.c 432 2004-07-12 17:27:37Z roland $
*/

#include "ipoib.h"

#include "ts_ib_sa_client.h"

#include "ts_kernel_trace.h"


#include <asm/io.h>

tTS_IB_GID broadcast_mgid = {
	0xff, 0x12, 0x40, 0x1b, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0xff, 0xff, 0xff, 0xff
};

#ifdef MAC_20_BYTES
int ipoib_reproduce_ah(struct ipoib_path *new_path, struct ipoib_path *old_path)
{

	struct ipoib_dev_priv *priv = new_path->dev->priv;
  	struct ipoib_path *path, *tpath;
	unsigned long flags;
	
	if (!old_path)
		return 1;
	
	if (old_path->get_counter >= IPOIB_MAX_REPRODUCE_AH)
	{
		TS_TRACE(MOD_IB_NET, T_VERBOSE, TRACE_IB_NET_ARP,
			 "ipoib_reproduce_ah - the get counter is bigger then  IPOIB_MAX_REPRODUCE_AH ");
		
		ipoib_free_ah(old_path);
		kfree(old_path);
		return 1; 
	}
	
	new_path->ah = old_path->ah;
	new_path->unused_ah = 0;
	new_path->get_counter = old_path->get_counter + 1;
	
	spin_lock_irqsave(&priv->lock, flags);
	list_for_each_entry_safe(path, tpath, &priv->created_ahs, list) {
		if (old_path->ah == path->ah)            
		{    
			list_del(&path->list);         
			TS_TRACE(MOD_IB_NET, T_VERBOSE, TRACE_IB_NET_ARP,
				 "ipoib_reproduce_ah succed  the old ah %p found in the list and will be removed",
				 old_path->ah);
		}
	}
	spin_unlock_irqrestore(&priv->lock, flags);
	
	kfree(old_path);
	return 0;
	
}

void ipoib_delete_unused_ah(struct net_device *dev)
{
	struct ipoib_dev_priv *priv = dev->priv;
	struct ipoib_path *path, *tpath;
 	LIST_HEAD(remove_list);
	
    TS_TRACE(MOD_IB_NET,T_VERBOSE, TRACE_IB_NET_ARP,
		 "%s: Reaping unused ahs",dev->name);   
	
	spin_lock_irq(&priv->lock);
	list_for_each_entry_safe(path, tpath, &priv->created_ahs, list) {
		if (path->unused_ah == 1)
        {
			list_del(&path->list);
            list_add_tail(&path->list, &remove_list);	       
		} 
	}
	spin_unlock_irq(&priv->lock);
	
	// Destroy the AHs
	list_for_each_entry_safe(path, tpath, &remove_list, list) {
		TS_TRACE(MOD_IB_NET, T_VERBOSE, TRACE_IB_NET_ARP,
			 "%s: Reaping unused ah %x",dev->name, path->ah);        
		if (ib_address_destroy(path->ah) < 0){
			TS_REPORT_WARN(MOD_IB_NET,
				       "%s: ib_address_destroy failed",
				   dev->name);    
		} else {
			--priv->ah_amount;
		}
		kfree(path);
	}		
}

void ipoib_free_ah(struct ipoib_path *dpath)
{
	struct ipoib_dev_priv *priv = dpath->dev->priv;
	struct net_device *dev = dpath->dev;
	struct ipoib_path *path, *tpath;
	unsigned long flags;
	
	spin_lock_irqsave(&priv->lock, flags);
	list_for_each_entry_safe(path, tpath, &priv->created_ahs, list) {
		if (dpath->ah == path->ah)            
		{    
			list_del(&path->list);         
			TS_TRACE(MOD_IB_NET, T_VERBOSE, TRACE_IB_NET_ARP,
				 "ah %p found in the list and will be removed",
				 dpath->ah);
		}
	}
	spin_unlock_irqrestore(&priv->lock, flags);
	if (ib_address_destroy(dpath->ah) < 0){
        TS_REPORT_WARN(MOD_IB_NET,
		       "%s: ib_address_destroy failed",
		       dev->name);
	} else {
		spin_lock(&priv->lock);
		--priv->ah_amount;
		spin_unlock(&priv->lock);
	}
	
}
// With out lock
struct ipoib_path *_ipoib_get_path(struct net_device *dev, u8 *hwaddr, u32 ipaddr)
{
	struct ipoib_dev_priv *priv = dev->priv;
	struct ipoib_path *path, *tpath;
	
	list_for_each_entry_safe(path, tpath, &priv->created_ahs, list) {        
			
		TS_TRACE(MOD_IB_NET, T_VERY_VERBOSE, TRACE_IB_NET_ARP,
			 "%s: check path is  " IPOIB_GID_FMT "\n",
			 dev->name,
			 path->gid.raw[0], path->gid.raw[1], path->gid.raw[2],
			 path->gid.raw[3], path->gid.raw[4], path->gid.raw[5],
			 path->gid.raw[6], path->gid.raw[7], path->gid.raw[8],
			 path->gid.raw[9], path->gid.raw[10], path->gid.raw[11],
			 path->gid.raw[12], path->gid.raw[13], path->gid.raw[14],
			 path->gid.raw[15]);
		
		TS_TRACE(MOD_IB_NET, T_VERY_VERBOSE, TRACE_IB_NET_ARP,
			 "%s: Needed gid is  " IPOIB_GID_FMT "\n",
			 dev->name,
			 hwaddr[4], hwaddr[5], hwaddr[6],
			 hwaddr[7], hwaddr[8], hwaddr[9],
			 hwaddr[10], hwaddr[11], hwaddr[12],
			 hwaddr[13], hwaddr[14], hwaddr[15],
			 hwaddr[16], hwaddr[17], hwaddr[18],
			 hwaddr[19]);
		
		// Check QP and gid
		if ((!memcmp(path->gid.raw , hwaddr + 4, IPOIB_HW_ADDR_LEN - 4)) &&
		    (hwaddr[3] == ((path->qpn      ) & 0xff)) &&  
		    (hwaddr[2] == ((path->qpn >> 8 ) & 0xff)) &&
		    (hwaddr[1] == ((path->qpn >> 16) & 0xff)) &&  
		    (hwaddr[0] == ((path->qpn >> 24) & 0xff)) &&
		    ((ipaddr == 0) || (ipaddr == path->ipaddr))) {
			
			TS_TRACE(MOD_IB_NET, T_VERBOSE, MOD_IB_NET,
				 "%s: AH found in the list",dev->name);
			return path;
		}
	}
	return NULL;
	
}

/// Find the AH in our list of created AHs.
struct ipoib_path *ipoib_get_path(struct net_device *dev, u8 *hwaddr, u32 ipaddr)
{
	struct ipoib_dev_priv *priv = dev->priv;
	unsigned long flags;
	struct ipoib_path *rpath = NULL;
	spin_lock_irqsave(&priv->lock, flags);
	rpath = _ipoib_get_path(dev, hwaddr, ipaddr);
	spin_unlock_irqrestore(&priv->lock, flags); 
	return rpath;

}
/// move the chache to be old
void ipoib_path_cache_old(struct net_device *dev)
{
	struct ipoib_dev_priv *priv = dev->priv;
	struct ipoib_path *path, *tpath;
	unsigned long flags;
	TS_TRACE(MOD_IB_NET, T_VERBOSE, MOD_IB_NET,
		 "%s:ipoib_path_cache_old moving the AHs to be already used ",dev->name);
	
	spin_lock_irqsave(&priv->lock, flags);
	list_for_each_entry_safe(path, tpath, &priv->created_ahs, list) {
		if (((IPOIB_MAX_REPRODUCE_AH - 2) > 0) && 
		    (path->get_counter < (IPOIB_MAX_REPRODUCE_AH - 2 ))) {
			path->get_counter = IPOIB_MAX_REPRODUCE_AH - 2;
		}
	}
	spin_unlock_irqrestore(&priv->lock, flags);
	
}
/// Find the AH in our list of created AHs.
struct ipoib_path *ipoib_path_look_up(struct net_device *dev, u8 *hwaddr, u32 ipaddr)
{
	struct ipoib_path *rpath = NULL;
	rpath = ipoib_get_path(dev, hwaddr, ipaddr);
	
	if (!rpath || (!rpath->ah) || 
	    (rpath->get_counter >= IPOIB_MAX_RE_AH_NO_SA)) {
		
		return NULL;
	}
	else
	{
		TS_TRACE(MOD_IB_NET, T_VERBOSE, MOD_IB_NET,
			 "%s:ipoib_path_look_up AH was found and will be used ",dev->name);
		
		return rpath;
	}
}


#endif

static int _ipoib_ib_receive(struct ipoib_dev_priv *priv,
			     u64 work_request_id,
			     dma_addr_t addr)
{
	struct ib_gather_scatter list = {
		.address = addr,
		.length  = IPOIB_BUF_SIZE,
		.key     = priv->lkey,
	};
	struct ib_receive_param param = {
		.work_request_id = work_request_id,
		.scatter_list = &list,
		.num_scatter_entries = 1,
		.device_specific = NULL,
		.signaled = 1,
	};

	return ib_receive(priv->qp, &param, 1, NULL);
}

/* =============================================================== */
/*.._ipoib_ib_post_receive -- post a receive buffer                */
static int _ipoib_ib_post_receive(struct net_device *dev, int id)
{
	struct ipoib_dev_priv *priv = dev->priv;
	struct sk_buff *skb;
	dma_addr_t addr;
	int ret;

	skb = dev_alloc_skb(IPOIB_BUF_SIZE + 4);
	if (!skb) {
		TS_REPORT_WARN(MOD_IB_NET,
			       "%s: failed to allocate receive buffer",
			       dev->name);

		priv->rx_ring[id].skb = NULL;
		priv->rx_ring[id].skb_not_alloc = 1;
		return -ENOMEM;
	}
	skb_reserve(skb, 4);	/* 16 byte align IP header */
	priv->rx_ring[id].skb = skb;
	priv->rx_ring[id].skb_not_alloc =  0;

	addr = pci_map_single(priv->ca->dma_device,
			      skb->data, IPOIB_BUF_SIZE,
			      PCI_DMA_FROMDEVICE);
	pci_unmap_addr_set(&priv->rx_ring[id], mapping, addr);

	ret = _ipoib_ib_receive(priv, id, addr);
	if (ret)
		TS_REPORT_WARN(MOD_IB_NET,
			       "%s: _ipoib_ib_receive failed for buf %d",
			       dev->name, id);

	return ret;
}
/* =============================================================== */
/*.._ipoib_ib_post_receive -- post a receive few buffers           */
static int _ipoib_ib_post_receive_few(struct net_device *dev, int id, int post_rec_num)
{
	struct ipoib_dev_priv *priv = dev->priv;
	int i, ret;
	int post_actual_num = 0;
	int current_id; 
	struct ib_gather_scatter *list = NULL;
	struct ib_receive_param *param = NULL;
	struct sk_buff **skb = NULL;
	
	list = kmalloc((sizeof *list) * IPOIB_MAX_POST_REC, GFP_ATOMIC);
	if ( !list ) {
		ret = -ENOMEM;
		goto exit;
	}

	param = kmalloc((sizeof *param) * IPOIB_MAX_POST_REC, GFP_ATOMIC);
	if ( !param ) {
		ret = -ENOMEM;
		goto exit;
	}

	skb = kmalloc(sizeof(struct sk_buff *) * IPOIB_MAX_POST_REC, GFP_ATOMIC);
	if ( !skb) {
		ret = -ENOMEM;
		goto exit;
	}
	
	
	if (post_rec_num > IPOIB_MAX_POST_REC)
		post_rec_num = IPOIB_MAX_POST_REC;
	
	
	for (i = 0 ; i < post_rec_num ; i++)
	{  
		dma_addr_t addr;
		
		current_id = (priv->rx_last_post_receive + 1 + i)%IPOIB_RX_RING_SIZE;
		skb[i] = dev_alloc_skb(IPOIB_BUF_SIZE + 4);
		
		if (!skb[i]) {
			TS_REPORT_WARN(MOD_IB_NET,
				       "%s: failed to allocate receive buffer",
				       dev->name);
			
			priv->rx_ring[current_id].skb = NULL;
			priv->rx_ring[current_id].skb_not_alloc = 1;
			ret = -ENOMEM;
			goto exit;
		}
		
		skb_reserve(skb[i], 4);	/* 16 byte align IP header */
		priv->rx_ring[current_id].skb = skb[i];
		priv->rx_ring[current_id].skb_not_alloc = 0;
		addr = pci_map_single(priv->ca->dma_device,
				      skb[i]->data, IPOIB_BUF_SIZE,
				      PCI_DMA_FROMDEVICE);
		pci_unmap_addr_set(&priv->rx_ring[current_id], mapping, addr);
		
		list[i].address = addr;
		list[i].key = priv->lkey;
		list[i].length = IPOIB_BUF_SIZE;
		
		param[i].work_request_id = current_id;
		param[i].scatter_list = &list[i];
		param[i].num_scatter_entries = 1;
		param[i].device_specific = NULL;
		param[i].signaled = 1;   
	}
	
	ret = ib_receive(priv->qp, param, post_rec_num, &post_actual_num);
	
	if (post_actual_num < post_rec_num)
	{
        // We will free the buffers that didn't got WQE
		for (i = post_actual_num ; i < post_rec_num ; i++)
		{
			dev_kfree_skb(skb[i]);
		}
	}
	
	priv->rx_last_post_receive = (priv->rx_last_post_receive + post_actual_num)%IPOIB_RX_RING_SIZE;
	
	
	if (ret)
		TS_REPORT_WARN(MOD_IB_NET,
			       "%s: _ipoib_ib_receive failed for buf %d",
			       dev->name, id);
exit:	
	if ( !list ) kfree(list);
	if ( !param ) kfree(param);
	if ( !skb ) kfree(skb);
	return ret;
}

/* =============================================================== */
/*.._ipoib_ib_post_receives -- post all receive buffers            */
static int _ipoib_ib_post_receives(struct net_device *dev)
{
	int i;

	for (i = 0; i < IPOIB_RX_RING_SIZE; ++i) {
		if (_ipoib_ib_post_receive(dev, i)) {
			TS_REPORT_FATAL(MOD_IB_NET,
					"%s: _ipoib_ib_post_receive failed for buf %d",
					dev->name, i);
			return -EIO;
		}
	}

	return 0;
}
/* =============================================================== */
/*..ipoib_ib_completion_tq -- analogous to interrupt handler          */
void ipoib_ib_tq_completion(struct ib_cq *cq, struct ib_cq_entry *entry,
		         void *dev_ptr)
{
    
	struct net_device *dev = (struct net_device *)dev_ptr;
	struct ipoib_dev_priv *priv = dev->priv;
	unsigned int work_request_id = (unsigned int) entry->work_request_id;
	int wake_queue = 0;
	int wqe_comp,comp_gap; 
	unsigned int pres_work_request_id;
	unsigned long flags;
#ifdef MAC_20_BYTES
	struct ipoib_buf *tx_req;
#else
	struct ipoib_tx_buf *tx_req;
#endif			
	
	TS_REPORT_DATA(MOD_IB_NET,
                   "%s: called: id %d, op %d, status: %d",
                   dev->name, work_request_id, entry->op, entry->status);
 	
	if (entry->op != IB_COMPLETION_OP_SEND) {
		TS_REPORT_WARN(MOD_IB_NET,
			       "%s: got unexpected completion event (status=%d, op=%d)",
			       dev->name, entry->status, entry->op);
		return;
	}

	if (entry->status) {
		TS_REPORT_WARN(MOD_IB_NET,
			       "%s: got failed completion event (status=%d, wrid=%d, op=%d)",
			       dev->name, entry->status, work_request_id,
			       entry->op);

		if (work_request_id < IPOIB_TX_RING_SIZE) {

#ifdef MAC_20_BYTES
			struct ipoib_buf *tx_req;
#else
			struct ipoib_tx_buf *tx_req;
#endif			
			tx_req = &priv->tx_ring[work_request_id];
			
			dev_kfree_skb(tx_req->skb);
			tx_req->skb = NULL;
		}
		
		return;
	}
	
	
	if (work_request_id >= IPOIB_TX_RING_SIZE) {
		TS_REPORT_WARN(MOD_IB_NET,
			       "%s: completion event with wrid %d",
			       dev->name,
			       work_request_id);
		return;
	}
	
	spin_lock_irqsave(&priv->lock, flags);
	
	if (work_request_id > priv->tx_last_wri_handel)
		comp_gap = work_request_id - priv->tx_last_wri_handel;
	else
		comp_gap = (IPOIB_TX_RING_SIZE - priv->tx_last_wri_handel) + work_request_id;
	
	pres_work_request_id = priv->tx_last_wri_handel;
   	
	priv->tx_last_wri_handel = (pres_work_request_id + comp_gap)%(IPOIB_TX_RING_SIZE);
	spin_unlock_irqrestore(&priv->lock, flags);
 	
	for (wqe_comp = 0 ; wqe_comp < comp_gap ; wqe_comp++) {
		
		pres_work_request_id = (pres_work_request_id + 1)%(IPOIB_TX_RING_SIZE);
		
		tx_req = &priv->tx_ring[pres_work_request_id];
		
		pci_unmap_single(priv->ca->dma_device,
				 pci_unmap_addr(tx_req, mapping),
				 tx_req->skb->len,
				 PCI_DMA_TODEVICE);
#ifndef MAC_20_BYTES  	
		clear_bit(IPOIB_FLAG_TIMEOUT, &priv->flags);
#endif   	
		++priv->stats.tx_packets;
		priv->stats.tx_bytes += tx_req->skb->len;
		
		dev_kfree_skb(tx_req->skb);
		tx_req->skb = NULL;

#ifndef MAC_20_BYTES  		
		tx_req->callback(tx_req->ptr);
		tx_req->callback = NULL;
		tx_req->ptr = NULL;
#endif   	
	}
	
	spin_lock_irqsave(&priv->lock, flags);
	priv->tx_free += wqe_comp;
	if (priv->tx_free > IPOIB_TX_RING_SIZE / 2)
		wake_queue = 1;
	spin_unlock_irqrestore(&priv->lock, flags);
	if (wake_queue && netif_queue_stopped(dev))
		netif_wake_queue(dev); 
		
}
/* =============================================================== */
/*.ipoib_ib_tq_event_handler. -- analogous to event handler          */
void ipoib_ib_tq_event_handler(struct ib_cq *cq, void *dev_ptr)
{
	struct net_device *dev = (struct net_device *)dev_ptr;
	struct ib_cq_entry      entry;
    
	while(ib_cq_poll_multi(cq, 1, &entry))
		ipoib_ib_tq_completion(cq ,&entry ,dev);
	
	if (netif_queue_stopped(dev))  
		ib_cq_request_notification(cq,0);
	
	while (ib_cq_poll_multi(cq, 1, &entry))
		ipoib_ib_tq_completion(cq, &entry, dev); 
           
}


/* =============================================================== */
/*..ipoib_ib_completion_rq -- analogous to interrupt handler          */
void ipoib_ib_rq_completion(struct ib_cq *cq, struct ib_cq_entry *entry,
		         void *dev_ptr)
{
	struct net_device *dev = (struct net_device *)dev_ptr;
	struct ipoib_dev_priv *priv = dev->priv;
	unsigned int work_request_id = (unsigned int) entry->work_request_id;
	int post_rec_num = 0;  
	unsigned long flags;
	struct sk_buff *skb;

	TS_REPORT_DATA(MOD_IB_NET,
		       "%s: called: id %d, op %d, status: %d",
		       dev->name, work_request_id, entry->op, entry->status);
	
	if (entry->status) {
		TS_REPORT_WARN(MOD_IB_NET,
			       "%s: got failed completion event (status=%d, wrid=%d, op=%d)",
			       dev->name, entry->status, work_request_id,
			       entry->op);
		return;
	}
	
	if (entry->op != IB_COMPLETION_OP_RECEIVE) {
		TS_REPORT_WARN(MOD_IB_NET,
			       "%s: got unexpected completion event (status=%d, op=%d)",
			       dev->name, entry->status, entry->op);
		return;
	}
	
	if (work_request_id >= IPOIB_RX_RING_SIZE) {
		TS_REPORT_WARN(MOD_IB_NET,
			       "%s: completion event with wrid %d",
			       dev->name, work_request_id);	
	}
	
	skb = priv->rx_ring[work_request_id].skb;
	
	pci_unmap_single(priv->ca->dma_device,
			 pci_unmap_addr(&priv->rx_ring[work_request_id],
					mapping),
			 IPOIB_BUF_SIZE,
			 PCI_DMA_FROMDEVICE);
	
	
	TS_REPORT_DATA(MOD_IB_NET,
		       "%s: received %d bytes, SLID 0x%04x",
		       dev->name,
		       entry->bytes_transferred,
		       entry->slid);
	
	skb_put(skb, entry->bytes_transferred);
	skb_pull(skb, TS_IB_GRH_BYTES);
	
	if (entry->slid != priv->local_lid
	    || entry->sqpn != priv->local_qpn) {
#ifdef MAC_20_BYTES
		skb->protocol = ((struct ipoib_header *) skb->data)->proto;
		skb_pull(skb, IPOIB_ENCAP_LEN);
#else
		struct ethhdr *header;
		
		skb->protocol = *(uint16_t *)skb->data;
		
		/* pull the IPoIB header and add an ethernet header */
		skb_pull(skb, IPOIB_ENCAP_LEN);
		
		header = (struct ethhdr *)skb_push(skb,
						   ETH_HLEN);
		
		/*
		 * We could figure out the MAC address from
		 * the IPoIB header and matching, but it's
		 * probably too much effort for what it's worth
		 */
		memset(header->h_dest, 0,
		       sizeof(header->h_dest));
		memset(header->h_source, 0,
		       sizeof(header->h_source));
		header->h_proto = skb->protocol;
		
		skb->mac.raw = skb->data;
		skb_pull(skb, ETH_HLEN);
#endif		
		dev->last_rx = jiffies;
		++priv->stats.rx_packets;
		priv->stats.rx_bytes += skb->len;

#ifdef MAC_20_BYTES
		skb->dev = dev;
		if (skb->mac.raw == 0)
			skb->mac.raw = skb->head;	
		
		/* XXX get correct PACKET_ type here */
		skb->pkt_type = PACKET_HOST;
		netif_rx_ni(skb);
#else   		
		if (skb->protocol == htons(ETH_P_ARP)) {
			if (ipoib_sarp_rewrite_receive(dev, skb))
				TS_REPORT_WARN(MOD_IB_NET,
					       "%s: ipoib_sarp_rewrite_receive failed",
					       dev->name);
		} else {
			skb->dev = dev;
			skb->pkt_type = PACKET_HOST;
			netif_rx_ni(skb);
		}
#endif		
	} else {
		TS_REPORT_DATA(MOD_IB_NET,
			       "%s: dropping loopback packet",
			       dev->name);
		dev_kfree_skb(skb);
	}
	
	spin_lock_irqsave(&priv->rq_lock, flags);
	
	priv->rx_ring[work_request_id].skb_not_alloc = 1;


	if (_ipoib_ib_post_receive(dev, work_request_id))
        TS_REPORT_FATAL(MOD_IB_NET,
                        "%s: _ipoib_ib_post_receive failed for buf %d",
                        dev->name,
                        work_request_id);


//if (work_request_id > priv->rx_last_post_receive)
//		post_rec_num = work_request_id - priv->rx_last_post_receive;
//	else
//		post_rec_num = (IPOIB_RX_RING_SIZE - priv->rx_last_post_receive) + work_request_id;
	
	/* repost receive */
	
//	if (post_rec_num >= IPOIB_MAX_POST_REC)
//	{
//		if (_ipoib_ib_post_receive_few(dev, work_request_id, post_rec_num))
//			TS_REPORT_FATAL(MOD_IB_NET,
//					"%s: _ipoib_ib_post_receive_few failed for buf %d",
//					dev->name,
//					work_request_id);
//	}   
	
	priv->rx_last_used = work_request_id;
	
	spin_unlock_irqrestore(&priv->rq_lock, flags);
	
}

/* =============================================================== */
/*..ipoib_ib_completion -- analogous to interrupt handler          */
void ipoib_ib_finish_completion(struct net_device *dev)
{
	struct ipoib_dev_priv *priv = dev->priv;
	unsigned int work_request_id = (unsigned int)(priv->tx_head - 1);
	int wake_queue = 0;
	int wqe_comp,comp_gap; 
	unsigned int pres_work_request_id;
	unsigned long flags;

#ifdef MAC_20_BYTES
	struct ipoib_buf *tx_req;
#else   		
	struct ipoib_tx_buf *tx_req;
#endif
	
	if (priv->tx_head == 0)
		work_request_id = IPOIB_TX_RING_SIZE - 1;
	
	TS_TRACE(MOD_IB_NET, T_VERBOSE, MOD_IB_NET,
		 "%s: called: dev get down finish completions from %d up to %d tx head is %d",
		 dev->name,priv->tx_last_wri_handel,work_request_id,priv->tx_head);
	
	if (work_request_id >= priv->tx_last_wri_handel)
		comp_gap = work_request_id - priv->tx_last_wri_handel;
	else
		comp_gap = (IPOIB_TX_RING_SIZE - priv->tx_last_wri_handel) + work_request_id;
	
	pres_work_request_id = priv->tx_last_wri_handel;
	
	for (wqe_comp = 0 ; wqe_comp < comp_gap ; wqe_comp++) {
		
		pres_work_request_id = (pres_work_request_id + 1)%(IPOIB_TX_RING_SIZE);
		
		tx_req = &priv->tx_ring[pres_work_request_id];
		
		/* continue if skb has already been freed, due to 
		 * error completions  -- see ipoib_ib_tq_completion,
		 * where on a completion with error,  tx_last_wri_handel
		 * does not get updated, but the skb is freed */
		if (!(tx_req->skb)) 
			continue;

		pci_unmap_single(priv->ca->dma_device,
				 pci_unmap_addr(tx_req, mapping),
				 tx_req->skb->len,
				 PCI_DMA_TODEVICE);
		
#ifndef MAC_20_BYTES  	
		clear_bit(IPOIB_FLAG_TIMEOUT, &priv->flags);
#endif   	
		++priv->stats.tx_packets;
		priv->stats.tx_bytes += tx_req->skb->len;
		
		dev_kfree_skb(tx_req->skb);
		tx_req->skb = NULL;

#ifndef MAC_20_BYTES  		
		tx_req->callback(tx_req->ptr);
		tx_req->callback = NULL;
		tx_req->ptr = NULL;
#endif   	
		
		spin_lock_irqsave(&priv->lock, flags);
		++priv->tx_free;
		if (priv->tx_free > IPOIB_TX_RING_SIZE / 2)
			wake_queue = 1;
		priv->tx_last_wri_handel = pres_work_request_id;
		spin_unlock_irqrestore(&priv->lock, flags);
		if (wake_queue)
			netif_wake_queue(dev);	
	}
	
}

static int _ipoib_ib_send(struct ipoib_dev_priv *priv,
			  u64 work_request_id,
			  struct ib_address *address, u32 qpn,
			  dma_addr_t addr, int len)
{
	int signaled, ret;
	struct ib_gather_scatter list = {
		.address = addr,
		.length  = len,
		.key     = priv->lkey,
	};

	struct ib_send_param *param;

	param = kmalloc(sizeof *param, GFP_ATOMIC);
	if ( !param ) {
		return -ENOMEM;
	}

	memset(param, 0, sizeof *param);
	param->work_request_id = work_request_id;
	param->op = IB_OP_SEND;
	param->gather_list = &list;
	param->num_gather_entries = 1;
	param->dest_qpn = qpn;
	param->dest_qkey = priv->qkey;
	param->dest_address = address;
	param->signaled = 1;
    
#ifdef MAC_20_BYTES

    if (priv->tx_hanging >= IPOIB_MAX_PACKET_HANGING) {
		signaled = 1;
		priv->tx_hanging = 0;
	} else {
		signaled = 0;
		++priv->tx_hanging;
	}
    param->signaled = signaled;
#else
    param->signaled = 1;
#endif
    ret = ib_send(priv->qp, param, 1, NULL);
	kfree(param);
	return ret;
}
#ifndef MAC_20_BYTES
/* =============================================================== */
/*..ipoib_dev_send -- schedule an IB send work request             */
int ipoib_dev_send(struct net_device *dev, struct sk_buff *skb,
		   ipoib_tx_callback_t callback, void *ptr,
		   struct ib_address *address, u32 qpn)
{
	struct ipoib_dev_priv *priv = dev->priv;
	struct ipoib_tx_buf *tx_req;
	dma_addr_t addr;
	int stop_queue = 0;
	unsigned long flags;
	int packet_send = 0;

	if (skb->len > dev->mtu + IPOIB_HW_ADDR_LEN) {
		TS_REPORT_WARN(MOD_IB_NET,
			       "%s: packet len %d too long to send, dropping",
			       dev->name, skb->len);
		++priv->stats.tx_dropped;
		++priv->stats.tx_errors;
		
		goto err;
	}
	
	if (!(skb = skb_unshare(skb, GFP_ATOMIC))) {
		TS_REPORT_WARN(MOD_IB_NET,
			       "%s: failed to unshare sk_buff. Dropping",
			       dev->name);
		++priv->stats.tx_dropped;
		++priv->stats.tx_errors;
		
		goto out;
	}

	/*
	 * We check if there is doorbell available in order to send this packet
	 * We will also check if this is the last doorbell we have
	 */
    
	spin_lock_irqsave(&priv->lock, flags);
	if (priv->tx_free < IPOIB_MIN_DOORBELL) {
		++priv->stats.tx_dropped;
		++priv->stats.tx_errors;
		if (!netif_queue_stopped(dev))
			netif_stop_queue(dev);
		spin_unlock_irqrestore(&priv->lock, flags);
		TS_REPORT_WARN(MOD_IB_NET,
			       "%s: packet arrived without door bell available, dropping",
			       dev->name);
		goto err;	
	}
	--priv->tx_free;
	if (priv->tx_free <= IPOIB_MIN_DOORBELL && (!netif_queue_stopped(dev))) {
		netif_stop_queue(dev);
		stop_queue = 1;
	}
	spin_unlock_irqrestore(&priv->lock, flags);
	
	if (stop_queue)
		TS_TRACE(MOD_IB_NET, T_TERSE, TRACE_IB_NET_GEN,
			 "%s: TX ring full, stopping Linux queue",
			 dev->name);
	
	TS_REPORT_DATA(MOD_IB_NET,
		       "%s: sending packet, length=%d address=0x%08x qpn=0x%06x",
		       dev->name, skb->len, address, qpn);
	
	/* make the skb look like an IPoIB packet again */
	{
		struct ethhdr *header = (struct ethhdr *)skb->data;
		uint16_t *reserved, *ether_type;
		
		skb_pull(skb, ETH_HLEN);
		reserved = (uint16_t *)skb_push(skb, 2);
		ether_type = (uint16_t *)skb_push(skb, 2);
		
		*ether_type = header->h_proto;
		*reserved = 0;
	}
	
	
	/*
	 * We put the skb into the tx_ring _before_ we call _ipoib_ib_send()
	 * because it's entirely possible that the completion handler will
	 * run before we execute anything after the _ipoib_ib_send().  That
	 * means we have to make sure everything is properly recorded and
	 * our state is consistent before we call _ipoib_ib_send().
	 */
	tx_req = &priv->tx_ring[priv->tx_head];
	tx_req->skb = skb;
	tx_req->callback = callback;
	tx_req->ptr = ptr;
	addr = pci_map_single(priv->ca->dma_device,
			      skb->data, skb->len,
			      PCI_DMA_TODEVICE);
	pci_unmap_addr_set(tx_req, mapping, addr);

	if (_ipoib_ib_send(priv, priv->tx_head, address, qpn, addr, skb->len)) {
		TS_REPORT_WARN(MOD_IB_NET, "%s: _ipoib_ib_send failed",
			       dev->name);
		++priv->stats.tx_errors;
		tx_req->skb = NULL;
		tx_req->callback = NULL;
		tx_req->ptr = NULL;
		spin_lock_irqsave(&priv->lock, flags);
		++priv->tx_free;
		spin_unlock_irqrestore(&priv->lock, flags);
		
	} else {

		dev->trans_start = jiffies;

		priv->tx_head = (priv->tx_head + 1) % IPOIB_TX_RING_SIZE;
		packet_send = 1;
	}
	{
		struct ib_cq_entry      entry;
		
		while(ib_cq_poll_multi(priv->cq_tq, 1, &entry) == 1) {
			ipoib_ib_tq_completion(priv->cq_tq ,&entry ,dev);
 		}
		
		if (netif_queue_stopped(dev)) {
			TS_TRACE(MOD_IB_NET, T_TERSE, TRACE_IB_NET_GEN,
				 "%s: Call ipoib_ib_tq_event_handler in order to arm the CQ",
				 dev->name);
			ipoib_ib_tq_event_handler(priv->cq_tq, dev);         
		}
	}
	if (packet_send == 1) {
		return 0;
	}
	
err:
	dev_kfree_skb(skb);

out:
	callback(ptr);

	return 0;
}

#else
/* =============================================================== */
/*..ipoib_dev_send -- schedule an IB send work request             */
void ipoib_send(struct net_device *dev, struct sk_buff *skb,
		struct ib_address *address, u32 qpn)
{
	struct ipoib_dev_priv *priv = dev->priv;
   	dma_addr_t addr;
	int stop_queue = 0;
	unsigned long flags;
	int packet_send = 0;
	struct ipoib_buf *tx_req;
    
	if (skb->len > dev->mtu + IPOIB_HW_ADDR_LEN) {
		TS_REPORT_WARN(MOD_IB_NET,
			       "%s: packet len %d too long to send, dropping",
			       dev->name, skb->len);
		++priv->stats.tx_dropped;
		++priv->stats.tx_errors;
        dev_kfree_skb_any(skb);
		return;
	}
	
	if (!(skb = skb_unshare(skb, GFP_ATOMIC))) {
		TS_REPORT_WARN(MOD_IB_NET,
			       "%s: failed to unshare sk_buff. Dropping",
			       dev->name);
		++priv->stats.tx_dropped;
		++priv->stats.tx_errors;
        return;
	}

	/*
	 * We check if there is doorbell available in order to send this packet
	 * We will also check if this is the last doorbell we have
	 */
    
	spin_lock_irqsave(&priv->lock, flags);
	if (priv->tx_free < IPOIB_MIN_DOORBELL) {
		++priv->stats.tx_dropped;
		++priv->stats.tx_errors;
		if (!netif_queue_stopped(dev))
			netif_stop_queue(dev);
		spin_unlock_irqrestore(&priv->lock, flags);
		TS_REPORT_WARN(MOD_IB_NET,
			       "%s: packet arrived without door bell available, dropping",
			       dev->name);
		dev_kfree_skb(skb);
		return;
	}
	--priv->tx_free;
	if (priv->tx_free <= IPOIB_MIN_DOORBELL && (!netif_queue_stopped(dev))) {
		netif_stop_queue(dev);
		stop_queue = 1;
	}
	spin_unlock_irqrestore(&priv->lock, flags);
	
	if (stop_queue)
		TS_TRACE(MOD_IB_NET, T_TERSE, TRACE_IB_NET_GEN,
			 "%s: TX ring full, stopping Linux queue",
			 dev->name);
	
	TS_REPORT_DATA(MOD_IB_NET,
		       "%s: sending packet, length=%d address=0x%08x qpn=0x%06x",
		       dev->name, skb->len, address, qpn);
	
	
	/*
	 * We put the skb into the tx_ring _before_ we call _ipoib_ib_send()
	 * because it's entirely possible that the completion handler will
	 * run before we execute anything after the _ipoib_ib_send().  That
	 * means we have to make sure everything is properly recorded and
	 * our state is consistent before we call _ipoib_ib_send().
	 */
	tx_req = &priv->tx_ring[priv->tx_head];
	tx_req->skb = skb;

	addr = pci_map_single(priv->ca->dma_device,
			      skb->data, skb->len,
			      PCI_DMA_TODEVICE);
	pci_unmap_addr_set(tx_req, mapping, addr);
	
	if (_ipoib_ib_send(priv, priv->tx_head, address, qpn, addr, skb->len)) {
		TS_REPORT_WARN(MOD_IB_NET, "%s: _ipoib_ib_send failed",
			       dev->name);
		++priv->stats.tx_errors;
		tx_req->skb = NULL;
		
		dev_kfree_skb_any(skb);
		spin_lock_irqsave(&priv->lock, flags);
		++priv->tx_free;
		spin_unlock_irqrestore(&priv->lock, flags);
		
	} else {
		
		dev->trans_start = jiffies;
		
		priv->tx_head = (priv->tx_head + 1) % IPOIB_TX_RING_SIZE;
		packet_send = 1;
	}
	{
		struct ib_cq_entry      entry;
		
		while(ib_cq_poll_multi(priv->cq_tq, 1, &entry) == 1) {
			ipoib_ib_tq_completion(priv->cq_tq ,&entry ,dev);
 		}
		
		if (netif_queue_stopped(dev)) {
			TS_TRACE(MOD_IB_NET, T_TERSE, TRACE_IB_NET_GEN,
				 "%s: Call ipoib_ib_tq_event_handler in order to arm the CQ",
				 dev->name);
			ipoib_ib_tq_event_handler(priv->cq_tq, dev);         
		}
	}   
}
#endif

int ipoib_ib_dev_open(struct net_device *dev)
{
	int ret;

	ret = ipoib_qp_create(dev);
	if (ret) {
		TS_REPORT_FATAL(MOD_IB_NET,
				"%s: ipoib_qp_create returned %d",
				dev->name, ret);
		return -1;
	}

	ret = _ipoib_ib_post_receives(dev);
	if (ret) {
		TS_REPORT_FATAL(MOD_IB_NET,
				"%s: _ipoib_ib_post_receives returned %d",
				dev->name, ret);
		return -1;
	}

	return 0;
}

int ipoib_ib_dev_up(struct net_device *dev)
{
	struct ipoib_dev_priv *priv = dev->priv;

	set_bit(IPOIB_FLAG_OPER_UP, &priv->flags);

	return ipoib_mcast_start_thread(dev);
}

/* =============================================================== */
/*..ipoib_ib_dev_down -- remove from multicast, etc                */
int ipoib_ib_dev_down(struct net_device *dev)
{
	struct ipoib_dev_priv *priv = dev->priv;
	int count = 0;

	clear_bit(IPOIB_FLAG_OPER_UP, &priv->flags);
	netif_carrier_off(dev);

	/* Shutdown the P_Key thread if still active */
	if (!test_bit(IPOIB_PKEY_ASSIGNED, &priv->flags))
		ipoib_pkey_dev_stop_thread(dev);

	if (priv->mcast_rejoin_handler) {
		ib_async_event_handler_deregister(priv->mcast_rejoin_handler);
		priv->mcast_rejoin_handler = NULL;
	}
    
	ipoib_mcast_stop_thread(dev);

	/*
	 * Flush the multicast groups first so we stop any multicast joins. The
	 * completion thread may have already died and we may deadlock waiting
	 * for the completion thread to finish some multicast joins.
	 */
	ipoib_mcast_dev_flush(dev);

	/* Wait for all joins to finish first */
	while (atomic_read(&priv->mcast_joins)) {
		if (count == 5) {
			TS_REPORT_WARN(MOD_IB_NET,
				       "%s: still waiting on %d multicast joins still to complete",
				       dev->name,
				       atomic_read(&priv->mcast_joins));
		}

		set_current_state(TASK_INTERRUPTIBLE);
		schedule_timeout(HZ);
		set_current_state(TASK_RUNNING);
		count++;
	}

	/* Delete broadcast and local addresses since they will be recreated */
	ipoib_mcast_dev_down(dev);

#ifndef MAC_20_BYTES 
	ipoib_sarp_delete(dev, dev->dev_addr);
	/* Invalidate all address vectors */
	ipoib_sarp_dev_flush(dev);
#else
	ipoib_ah_delete(dev); 
#endif

	return 0;
}

/* =============================================================== */
/*..ipoib_ib_dev_stop -- cleanup QP and RX ring                    */
int ipoib_ib_dev_stop(struct net_device *dev)
{
	struct ipoib_dev_priv *priv = dev->priv;
	int i;
 	
	/* Kill the existing QP and allocate a new one */
	if (priv && priv->qp != NULL)
		ipoib_qp_destroy(dev);
	
	for (i = 0; i < IPOIB_RX_RING_SIZE; ++i) {
		if (priv->rx_ring && priv->rx_ring[i].skb ) {
			if (priv->rx_ring[i].skb_not_alloc == 0)  {
				dev_kfree_skb(priv->rx_ring[i].skb);
				priv->rx_ring[i].skb_not_alloc = 1;
				priv->rx_ring[i].skb = NULL;
			}   			          
		}
	}
	priv->rx_last_post_receive = IPOIB_RX_RING_SIZE - 1;
	priv->rx_last_used = IPOIB_RX_RING_SIZE - 1;
	return 0;
}

/* =============================================================== */
/*..ipoib_ib_dev_init -- set up IB resources for iface             */
int ipoib_ib_dev_init(struct net_device *dev, struct ib_device *ca, int port)
{
	struct ipoib_dev_priv *priv = dev->priv;

	priv->ca = ca;
	priv->port = port;
	priv->qp = NULL;
	priv->mcast_tid = TS_IB_CLIENT_QUERY_TID_INVALID;

	if (ipoib_transport_dev_init(dev, ca)) {
		TS_REPORT_WARN(MOD_IB_NET,
			       "%s: ipoib_transport_dev_init failed",
			       dev->name);
		return -ENODEV;
	}

	if (dev->flags & IFF_UP) {
		if (ipoib_ib_dev_open(dev)) {
			ipoib_transport_dev_cleanup(dev);
			return -ENODEV;
		}
	}

	return 0;
}

/* =============================================================== */
/*..ipoib_ib_dev_flush -- flush and rejoin multicast               */
void ipoib_ib_dev_flush(void *_dev)
{
	struct net_device *dev = (struct net_device *)_dev;
	struct ipoib_dev_priv *priv = dev->priv, *cpriv;

	if (!test_bit(IPOIB_FLAG_ADMIN_UP, &priv->flags))
		return;

	TS_TRACE(MOD_IB_NET, T_VERBOSE, TRACE_IB_NET_GEN, "%s: flushing",
		 dev->name);

	ipoib_ib_dev_down(dev);

	/*
	 * The device could have been brought down between the start and when
	 * we get here, don't bring it back up if it's not configured up
	 */
	if (test_bit(IPOIB_FLAG_ADMIN_UP, &priv->flags))
		ipoib_ib_dev_up(dev);

	/* Flush any child interfaces too */
	list_for_each_entry(cpriv, &priv->child_intfs, list)
		ipoib_ib_dev_flush(&cpriv->dev);
}

/* =============================================================== */
/*..ipoib_ib_dev_cleanup -- clean up IB resources for iface        */
void ipoib_ib_dev_cleanup(struct net_device *dev)
{
	struct ipoib_dev_priv *priv = dev->priv;

	TS_REPORT_CLEANUP(MOD_IB_NET,
			  "%s: cleaning up IB resources", dev->name);

	ipoib_mcast_stop_thread(dev);
#ifdef MAC_20_BYTES 
	ipoib_mcast_update_stop_thread(dev);
#endif
	/* Wait until all multicast joins are done */
	while (atomic_read(&priv->mcast_joins)) {
		set_current_state(TASK_RUNNING);
		schedule();
	}

	/* Delete the broadcast address and the local address */
	ipoib_mcast_dev_down(dev);

#ifndef MAC_20_BYTES 	
	ipoib_sarp_delete(dev, dev->dev_addr);
#endif
	
	ipoib_transport_dev_cleanup(dev);
}

/*
 * Delayed P_Key Assigment Interim Support
 *
 * The following is initial implementation of delayed P_Key assigment
 * mechanism. It is using the same approach implemented for the multicast
 * group join. The single goal of this implementation is to quickly address
 * Bug #2507. This implementation will probably be removed when the P_Key
 * change async notification is available.
 */
static DECLARE_MUTEX(pkey_sem);
static void _ipoib_pkey_thread(void *dev_ptr);
int ipoib_dev_open(struct net_device *dev);

/* =============================================================== */
/*..ipoib_pkey_dev_start_thread -- Start the P_Key thread          */
int ipoib_pkey_dev_start_thread(struct net_device *dev)
{
	struct ipoib_dev_priv *priv = dev->priv;
	char thread_name[sizeof("ibX.YYYY_pkey")];
	int ret = 0;

	TS_TRACE(MOD_IB_NET, T_VERBOSE, TRACE_IB_NET_GEN,
		 "%s: starting P_Key thread", dev->name);

	down(&pkey_sem);

	if (priv->pkey_thread)
		goto out;

	snprintf(thread_name, sizeof(thread_name), "%s_pkey", dev->name);

	if (tsKernelThreadStart(thread_name,
				_ipoib_pkey_thread,
				dev, &priv->pkey_thread)) {
		TS_REPORT_FATAL(MOD_IB_NET,
				"%s: failed to start the P_Key thread",
				dev->name);
		ret = -1;
	}

out:
	up(&pkey_sem);

	return ret;
}

/* =============================================================== */
/*..ipoib_pkey_dev_stop_thread -- Stop the P_Key thread            */
int ipoib_pkey_dev_stop_thread(struct net_device *dev)
{
	struct ipoib_dev_priv *priv = dev->priv;
	int ret = 0;

	TS_TRACE(MOD_IB_NET, T_VERBOSE, TRACE_IB_NET_GEN,
		 "%s: stopping P_Key thread", dev->name);

	down(&pkey_sem);

	if (priv->pkey_thread) {
		tsKernelThreadStop(priv->pkey_thread);
		priv->pkey_thread = NULL;

		ret = 1;
	}

	up(&pkey_sem);

	return ret;
}

/* =============================================================== */
/*.._ipoib_pkey_thread -- poll until pkey set                      */
static void _ipoib_pkey_thread(void *dev_ptr)
{
	struct net_device *dev = dev_ptr;
	struct ipoib_dev_priv *priv = dev->priv;

	/* P_Key already assigned */
	if (test_bit(IPOIB_PKEY_ASSIGNED, &priv->flags))
		return;

	/* Wait for the IB SM to program the interface P_Key */
	for (;;) {
		ipoib_pkey_dev_check_presence(dev);

		if (test_bit(IPOIB_PKEY_ASSIGNED, &priv->flags)) {
			ipoib_dev_open(dev);
			return;
		}

		/* Pause */
		set_current_state(TASK_INTERRUPTIBLE);
		schedule_timeout(3 * HZ);
		set_current_state(TASK_RUNNING);

		if (signal_pending(current))
			return;
	}
}

/* ======================================================================== */
/*..ipoib_pkey_dev_delay_open -- wait for pkey to be set                    */
int ipoib_pkey_dev_delay_open(struct net_device *dev)
{
	struct ipoib_dev_priv *priv = dev->priv;

	/* Look for the interface pkey value in the IB Port P_Key table and */
	/* set the interface pkey assigment flag                            */
	ipoib_pkey_dev_check_presence(dev);

	/* P_Key value not assigned yet - start the P_Key thread */
	if (!test_bit(IPOIB_PKEY_ASSIGNED, &priv->flags)) {
		ipoib_pkey_dev_start_thread(dev);
		return 1;
	}

	return 0;
}

/*
  Local Variables:
  c-file-style: "linux"
  indent-tabs-mode: t
  End:
*/
