/*
  This software is available to you under a choice of one of two
  licenses.  You may choose to be licensed under the terms of the GNU
  General Public License (GPL) Version 2, available at
  <http://www.fsf.org/copyleft/gpl.html>, or the OpenIB.org BSD
  license, available in the LICENSE.TXT file accompanying this
  software.  These details are also available at
  <http://openib.org/license.html>.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
  BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
  ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
  CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
  SOFTWARE.

  Copyright (c) 2004 Topspin Communications.  All rights reserved.
  Copyright (c) 2005 Mellanox Technologies.   All rights reserved.


  $Id: sdp_conn.h 432 2004-07-12 17:27:37Z roland $
*/

#ifndef _TS_SDP_CONN_H
#define _TS_SDP_CONN_H
/*
 * topspin specific includes.
 */
#include "sdp_types.h"
#include "sdp_advt.h"
#include "sdp_iocb.h"
#include "sdp_dev.h"
#include "sdp_os.h"

/* ---------------------------------------------------------------------- */
/* SDP connection specific definitions                                    */
/* ---------------------------------------------------------------------- */
/*
 * definitions
 */
#define TS_SDP_CONN_SEND_POST_NAME "tx_sdp_post"	/* name for pool */
#define TS_SDP_CONN_RECV_POST_NAME "rx_sdp_post"	/* name for pool */
#define TS_SDP_CONN_SEND_DATA_NAME "tx_sdp_data"	/* name for pool */
#define TS_SDP_CONN_SEND_CTRL_NAME "tx_sdp_ctrl"	/* name for pool */
#define TS_SDP_CONN_RDMA_READ_NAME "rx_sdp_rdma"	/* name for pool */
#define TS_SDP_CONN_RDMA_SEND_NAME "tx_sdp_rdma"	/* name for pool */
#define TS_SDP_SOCK_RECV_DATA_NAME "rx_lnx_data"
#define TS_SDP_INET_PEEK_DATA_NAME "rx_lnx_peek"
#define TS_SDP_CONN_COMM_ID_NULL   0xFFFFFFFF	/* when no id is available */

#define TS_SDP_CONN_F_SRC_CANCEL_L 0x0001	/* source cancel was issued */
#define TS_SDP_CONN_F_SRC_CANCEL_R 0x0002	/* source cancel was received */
#define TS_SDP_CONN_F_SRC_CANCEL_C 0x0004	/* source data cancelled */
#define TS_SDP_CONN_F_SNK_CANCEL   0x0008	/* sink cancel issued */
#define TS_SDP_CONN_F_DIS_HOLD     0x0010	/* Hold pending disconnects. */
#define TS_SDP_CONN_F_OOB_SEND     0x0020	/* OOB notification pending. */
#define TS_SDP_CONN_F_RECV_CQ_PEND 0x0040	/* recv CQ event is pending */
#define TS_SDP_CONN_F_SEND_CQ_PEND 0x0080	/* send CQ event is pending */
#define TS_SDP_CONN_F_DEAD         0xFFFF	/* connection has been deleted */

#define TS_SDP_CONN_F_MASK_EVENT   (TS_SDP_CONN_F_RECV_CQ_PEND | \
				    TS_SDP_CONN_F_SEND_CQ_PEND)
/*
 * sdp connection table proc entries
 */
#define TS_SDP_SOPT_PROC_DUMP_SIZE   55	/* output line size. */
#define TS_SDP_CONN_PROC_MAIN_SIZE  183	/* output line size. */
#define TS_SDP_CONN_PROC_DATA_SIZE  176	/* output line size. */
#define TS_SDP_CONN_PROC_RDMA_SIZE  101	/* output line size. */
/*
 * Work Request ID mask to identify IOCB READs and BUFF RECVs.
 */
#define TS_SDP_WRID_READ_FLAG 0x80000000
#define TS_SDP_WRID_RECV_MASK 0x7FFFFFFF
/*
 * SDP states.
 */
typedef enum {
	TS_SDP_MODE_BUFF = 0x00,
	TS_SDP_MODE_COMB = 0x01,
	TS_SDP_MODE_PIPE = 0x02,
	TS_SDP_MODE_ERROR = 0x03
} tSDP_MODE;
/*
 * First two bytes are the primary state values. Third byte is a bit
 * field used for different mask operations, defined below. Fourth
 * byte are the mib values for the different states.
 */
#define TS_SDP_CONN_ST_LISTEN      0x0100 /* listening */

#define TS_SDP_CONN_ST_ESTABLISHED 0x11B1 /* connected */

#define TS_SDP_CONN_ST_REQ_PATH    0x2100 /* active open, path record lookup */
#define TS_SDP_CONN_ST_REQ_SENT    0x2200 /* active open, Hello msg sent */
#define TS_SDP_CONN_ST_REQ_RECV    0x2300 /* passive open, Hello msg recv'd */
#define TS_SDP_CONN_ST_REP_SENT    0x2480 /* passive open, Hello ack sent */
#define TS_SDP_CONN_ST_REP_RECV    0x2500 /* active open, Hello ack recv'd */
#define TS_SDP_CONN_ST_RTU_SENT    0x2680 /* active open, Hello ack, acked */

#define TS_SDP_CONN_ST_DIS_RECV_1  0x31B1 /* recv disconnect, passive close */
#define TS_SDP_CONN_ST_DIS_PEND_1  0x32F1 /* pending disconn, active close */
#define TS_SDP_CONN_ST_DIS_SEND_1  0x33B1 /* send disconnect, active close */
#define TS_SDP_CONN_ST_DIS_SENT_1  0x34A1 /* disconnect sent, active close */
#define TS_SDP_CONN_ST_DIS_PEND_R  0x35F1 /* disconnect recv, active close */
#define TS_SDP_CONN_ST_DIS_RECV_R  0x36B1 /* disconnect recv, active close */
#define TS_SDP_CONN_ST_DIS_PEND_2  0x37F1 /* pending disconn, passive close */
#define TS_SDP_CONN_ST_DIS_SEND_2  0x38B1 /* send disconnect, passive close */
#define TS_SDP_CONN_ST_DIS_SENT_2  0x3901 /* disconnect sent, passive close */
#define TS_SDP_CONN_ST_TIME_WAIT_1 0x3A01 /* IB/gateway disconnect */
#define TS_SDP_CONN_ST_TIME_WAIT_2 0x3B01 /* waiting for idle close */

#define TS_SDP_CONN_ST_ERROR_CM    0xFB03 /* CM error, waiting on gateway */
#define TS_SDP_CONN_ST_ERROR_CQ    0xFC03 /* gateway error, waiting on CM */
#define TS_SDP_CONN_ST_ERROR_STRM  0xFD01 /* gateway error, waiting on CM */
#define TS_SDP_CONN_ST_CLOSED      0xFE03 /* not connected */
#define TS_SDP_CONN_ST_INVALID     0xFF03 /* not connected */
/*
 * states masks for SDP
 */
#define TS_SDP_ST_MASK_EVENTS    0x0001	/* event processing is allowed. */
#define TS_SDP_ST_MASK_ERROR     0x0002	/* protocol in error state. */
#define TS_SDP_ST_MASK_SEND_OK   0x0010	/* posting data for send */
#define TS_SDP_ST_MASK_CTRL_OK   0x0020	/* posting control for send */
#define TS_SDP_ST_MASK_DIS_PEND  0x0040	/* disconnect transmission pending. */
#define TS_SDP_ST_MASK_RCV_POST  0x0080	/* posting IB recv's is allowed. */
/*
 * transition one of the disconnect pending states to disconnect send
 */
#define TS_SDP_ST_PEND_2_SEND(conn) \
        (conn)->state = ((conn)->state + 0x0100) & ~TS_SDP_ST_MASK_DIS_PEND;
/*
 * internal connection structure
 */
#define TS_SDP_SOCK_ST_CLOSED      0x0101
#define TS_SDP_SOCK_ST_CONNECT     0x0204
#define TS_SDP_SOCK_ST_ACCEPTING   0x0300
#define TS_SDP_SOCK_ST_ACCEPTED    0x0404 /* writable ? */
#define TS_SDP_SOCK_ST_ESTABLISHED 0x050E
#define TS_SDP_SOCK_ST_DISCONNECT  0x0602 /* active close request */
#define TS_SDP_SOCK_ST_CLOSE       0x070C /* passive close request */
#define TS_SDP_SOCK_ST_CLOSING     0x0800
#define TS_SDP_SOCK_ST_LISTEN      0x0901
#define TS_SDP_SOCK_ST_ERROR       0xFF01
/*
 * state masks.
 */
#define TS_SDP_ST_MASK_CLOSED 0x0001 /* socket is not referenced by the GW. */
#define TS_SDP_ST_MASK_SEND   0x0004 /* valid state for API send */
#define TS_SDP_ST_MASK_OPEN   0x0008 /* send window is valid (writeable) */
/*
 * shutdown
 */
#define TS_SDP_SHUTDOWN_RECV   0x01 /* recv connection half close */
#define TS_SDP_SHUTDOWN_SEND   0x02 /* send connection half close */
#define TS_SDP_SHUTDOWN_MASK   0x03
#define TS_SDP_SHUTDOWN_NONE   0x00
/*
 * event dispatch table
 */
#define TS_SDP_MSG_EVENT_TABLE_SIZE 0x20
/* --------------------------------------------------------------------- */
/* state transition information recording                                */
/* --------------------------------------------------------------------- */
#ifdef _TS_SDP_CONN_STATE_REC

#define TS_SDP_CONN_STATE_MAX 16 /* maximum state transitions recorded. */

struct sdp_conn_state {
	__u8  value;
	__u16 state[TS_SDP_CONN_STATE_MAX];
	void *file[TS_SDP_CONN_STATE_MAX];
	__s32 line[TS_SDP_CONN_STATE_MAX];
}; /* struct sdp_conn_state */

#define TS_SDP_CONN_ST_SET(conn, val) \
do { \
  (conn)->state = (val); \
  if (TS_SDP_CONN_STATE_MAX > (conn)->state_rec.value) { \
    (conn)->state_rec.state[(conn)->state_rec.value] = (val); \
    (conn)->state_rec.file[(conn)->state_rec.value] = __FILE__; \
    (conn)->state_rec.line[(conn)->state_rec.value] = __LINE__; \
    (conn)->state_rec.value++; \
  } \
}  while(0)

#define TS_SDP_CONN_ST_INIT(conn) \
do { \
  (conn)->state = TS_SDP_CONN_ST_INVALID; \
  for ((conn)->state_rec.value = 0; \
       TS_SDP_CONN_STATE_MAX > (conn)->state_rec.value; \
       (conn)->state_rec.value++) { \
    (conn)->state_rec.state[(conn)->state_rec.value] = TS_SDP_CONN_ST_INVALID;\
    (conn)->state_rec.file[(conn)->state_rec.value] = NULL; \
    (conn)->state_rec.line[(conn)->state_rec.value] = 0; \
  } \
  (conn)->state_rec.value = 0; \
}  while(0)
#else
#define TS_SDP_CONN_ST_SET(conn, val) (conn)->state = (val)
#define TS_SDP_CONN_ST_INIT(conn)     (conn)->state = TS_SDP_CONN_ST_INVALID
#endif
/*
 * connection lock
 */
struct sdp_conn_lock {
	__u32 users;
	spinlock_t slock;
	wait_queue_head_t waitq;
}; /* struct sdp_conn_lock */
/*
 * SDP Connection structure.
 */
struct sdp_opt {
	__s32 hashent;     /* connection ID/hash entry */
	atomic_t refcnt;   /* connection reference count. */

	struct sock *sk;
	/*
	 * SDP specific data
	 */
	__u32 send_buf;
	__u32 send_qud;
	__u32 send_pipe;	/* buffered bytes in the local send queue */
	__s32 oob_offset;	/* bytes till OOB byte is sent. */

	__s16 send_usig;  /* number of unsignalled sends in the pipe. */
	__s16 send_cons;  /* number of consecutive unsignalled sends. */
	__s16 usig_max;   /* maximum unsignalled back-to-back sends. */


	struct sdpc_desc_q send_queue;	/* queue of send objects. */
	struct sdpc_buff_q send_post;	/* posted sends */

	__u32 send_seq;	/* sequence number of last message sent */
	__u32 recv_seq;	/* sequence number of last message received */
	__u32 advt_seq;	/* sequence number of last message acknowledged */

	struct sdpc_buff_q recv_pool;	/* pool of received buffer */
	struct sdpc_buff_q recv_post;	/* posted receives */

	__s32 byte_strm;  /* buffered bytes in the local recv queue */
	__s32 rwin_max;   /* maximum recveive window size */

	__u16 state;      /* connection state */
	__u16 istate;     /* inet connection state */

	__u8  flags;      /* single bit flags. */
	__u8  shutdown;   /* shutdown flag */
	__u8  recv_mode;  /* current flow control mode */
	__u8  send_mode;  /* current flow control mode */

	__u16 recv_max;   /* max posted/used receive buffers */
	__u16 send_max;   /* max posted/used send buffers */

	__u16 recv_size;  /* local recv buffer size */
	__u16 send_size;  /* remote recv buffer size */

	__u8  l_max_adv;  /* local maximum zcopy advertisments */
	__u8  r_max_adv;  /* remote maximum zcopy advertisments */
	__u8  s_cur_adv;  /* current source advertisments (slow start) */
	__u8  s_par_adv;  /* current source advertisments (slow start) */

	__u16 r_recv_bf;  /* number of recv buffers remote currently has */
	__u16 l_recv_bf;  /* number of recv buffers local currently has */
	__u16 l_advt_bf;  /* number of recv buffers local has advertised */

	__u8  s_wq_size;  /* current number of posted sends. */

	__u8  s_wq_cur;   /* buffered transmission limit current */
	__u8  s_wq_par;   /* buffered transmission limit increment */

	__u8  src_recv;   /* outstanding remote source advertisments */
	__u8  snk_recv;   /* outstanding remote sink advertisments */
	__u8  src_sent;   /* outstanding local source advertisments */
	__u8  snk_sent;   /* outstanding local sink advertisments */

	__u8  sink_actv;

	__u8  src_cncl;   /* local source advertisments cancelled by user */
	__u32 src_cseq;   /* sequence number of source cancel message */
	/*
	 * work request ID's used to double-check queue consistency
	 */
	__u32 send_wrid;
	__u32 recv_wrid;

	__u32 send_cq_size;
	__u32 recv_cq_size;
	/*
	 * stale SnkAvail detection
	 */
	__u32 nond_recv;	/* non discarded buffers received. */
	__u32 nond_send;	/* non discarded buffers sent */

	__s32 error;		/* error value on connection. */
	/*
	 * OOB/URG data transfer.
	 */
	__s16 rcv_urg_cnt;	/* queued urgent data */
	/*
	 * listen backlog
	 */
	__u16 backlog_cnt;	/* depth of the listen backlog queue */
	__u16 backlog_max;	/* max length of the listen backlog queue */
	/*
	 * memory specific data
	 */
	struct sdpc_desc_q send_ctrl;	/* control messages waiting to
						   be transmitted, which do not
						   depend on data ordering */
	/*
	 * advertisment managment
	 */
	struct sdpc_advt_q src_pend;	/* pending remote source advertisments */
	struct sdpc_advt_q src_actv;	/* active remote source advertisments */
	struct sdpc_advt_q snk_pend;	/* pending remote sink advertisments */
	/*
	 * outstanding IOCBs/BUFFs
	 */
	struct sdpc_iocb_q r_pend;	/* pending user read IOCBs */
	struct sdpc_iocb_q r_snk;	/* active user read sink IOCBs */
	struct sdpc_iocb_q w_src;   /* active user write source IOCBs */
    struct sdpc_iocb_q r_comp;
  
	struct sdpc_desc_q r_src;	/* active user read source IOCBs */
	struct sdpc_desc_q w_snk;	/* active user write sink IOCBs */
	/*
	 * addresses
	 */
	__u32 src_addr;	/* ipv4 address on the stream interface */
	__u32 dst_addr;	/* ipv4 address of the remote SDP client */
	__u16 dst_port;	/* tcp port of the remote SDP client */
	__u16 src_port;	/* tcp port on the stream interface */
	/*
	 * IB specific data
	 */
	tTS_IB_GID d_gid;
	tTS_IB_GID s_gid;
	__u16 d_lid;
	__u16 s_lid;
	__u32 d_qpn;
	__u32 s_qpn;

	tIP2PR_PATH_LOOKUP_ID plid;	/* path record lookup */

	struct ib_device *ca;	/* hca that we'll be using for sdp */
	struct ib_qp *qp;	/* queue pair for the SDP connection */
	struct ib_pd *pd;	/* protection domain used by the kernel */
	struct ib_cq *send_cq;	/* send completion queue */
	struct ib_cq *recv_cq;	/* recv completion queue */
	tTS_IB_PORT hw_port;	/* hca port */
	u32 l_key;		/* local key for buffered memory */
	struct ib_fmr_pool *fmr_pool;	/* fast memory for Zcopy */

	tTS_IB_CM_COMM_ID comm_id;	/* CM handle durring async connection setup */
	/*
	 * timer for defered execution. Used to call CM functions from a
	 * non-interupt context.
	 */
	tTS_KERNEL_TIMER_STRUCT cm_exec;
	/*
	 * SDP connection lock
	 */
	struct sdp_conn_lock lock;
	/*
	 * table managment
	 */
	struct sdp_opt *lstn_next;	/* next connection in the chain */
	struct sdp_opt **lstn_p_next;	/* previous next connection in the chain */

	struct sdp_opt *bind_next;	/* next connection in the chain */
	struct sdp_opt **bind_p_next;	/* previous next connection in the chain */
	/*
	 * listen/accept managment
	 */
	struct sdp_opt *parent;	/* listening socket queuing. */
	struct sdp_opt *accept_next;	/* sockets waiting for acceptance. */
	struct sdp_opt *accept_prev;	/* sockets waiting for acceptance. */
	/*
	 * OS info
	 */
	__u16 pid;		/* process ID of creator */
	/*
	 * TCP specific socket options
	 */
	__u8  nodelay;		/* socket nodelay is set */
	__u32 src_zthresh;	/* source zero copy threshold */
	__u32 snk_zthresh;	/* sink zero copy threshold */
	/*
	 * stats
	 */
	__u32 send_mid[TS_SDP_MSG_EVENT_TABLE_SIZE];	/* send event stats */
	__u32 recv_mid[TS_SDP_MSG_EVENT_TABLE_SIZE];	/* recv event stats */

	__u64 send_bytes;	/* socket bytes sent */
	__u64 recv_bytes;	/* socket bytes received */
	__u64 write_bytes;	/* AIO bytes sent */
	__u64 read_bytes;	/* AIO bytes received */

	__u32 read_queued;	/* reads queued for reception */
	__u32 write_queued;	/* writes queued for transmission */

	__u32 src_serv;	/* source advertisments completed. */
	__u32 snk_serv;	/* sink advertisments completed. */
    __u32   s_pend_copied;
    __u32  use_zcopy;


    __u32 buffered_bytes;
    __u32   r_zcopy_called;
    __u32   wait_buff_before_src;
    __u32   r_pend_copied;
    __u32  flying_bytes; //bytes on the air
    __u32  inside_read;


 
#ifdef _TS_SDP_CONN_STATE_REC
	struct sdp_conn_state state_rec;
#endif
}; /* struct sdp_opt */

#define TS_SDP_WRID_GT(x, y) ((signed int)((x) - (y)) > 0)
#define TS_SDP_WRID_LT(x, y) ((signed int)((x) - (y)) < 0)
#define TS_SDP_WRID_GTE(x, y) ((signed int)((x) - (y)) >= 0)
#define TS_SDP_WRID_LTE(x, y) ((signed int)((x) - (y)) <= 0)

#define TS_SDP_SEQ_GT(x, y) ((signed int)((x) - (y)) > 0)
#define TS_SDP_SEQ_LT(x, y) ((signed int)((x) - (y)) < 0)
#define TS_SDP_SEQ_GTE(x, y) ((signed int)((x) - (y)) >= 0)
#define TS_SDP_SEQ_LTE(x, y) ((signed int)((x) - (y)) <= 0)
/*
 * statistics.
 */
#ifdef _TS_SDP_CONN_STATS_REC
#define TS_SDP_CONN_STAT_SEND_INC(conn, size)  ((conn)->send_bytes += (size))
#define TS_SDP_CONN_STAT_RECV_INC(conn, size)  ((conn)->recv_bytes += (size))
#define TS_SDP_CONN_STAT_READ_INC(conn, size)  ((conn)->read_bytes += (size))
#define TS_SDP_CONN_STAT_WRITE_INC(conn, size) ((conn)->write_bytes += (size))

#define TS_SDP_CONN_STAT_RQ_INC(conn, size) ((conn)->read_queued  += (size))
#define TS_SDP_CONN_STAT_WQ_INC(conn, size) ((conn)->write_queued += (size))
#define TS_SDP_CONN_STAT_RQ_DEC(conn, size) ((conn)->read_queued  -= (size))
#define TS_SDP_CONN_STAT_WQ_DEC(conn, size) ((conn)->write_queued -= (size))

#define TS_SDP_CONN_STAT_SRC_INC(conn) ((conn)->src_serv++)
#define TS_SDP_CONN_STAT_SNK_INC(conn) ((conn)->snk_serv++)

#define TS_SDP_CONN_STAT_SEND_MID_INC(conn, mid) \
        ((conn)->send_mid[(mid)]++)
#define TS_SDP_CONN_STAT_RECV_MID_INC(conn, mid) \
        ((conn)->recv_mid[(mid)]++)
#else
#define TS_SDP_CONN_STAT_SEND_INC(conn, size)
#define TS_SDP_CONN_STAT_RECV_INC(conn, size)
#define TS_SDP_CONN_STAT_READ_INC(conn, size)
#define TS_SDP_CONN_STAT_WRITE_INC(conn, size)

#define TS_SDP_CONN_STAT_RQ_INC(conn, size)
#define TS_SDP_CONN_STAT_WQ_INC(conn, size)
#define TS_SDP_CONN_STAT_RQ_DEC(conn, size)
#define TS_SDP_CONN_STAT_WQ_DEC(conn, size)

#define TS_SDP_CONN_STAT_SRC_INC(conn)
#define TS_SDP_CONN_STAT_SNK_INC(conn)

#define TS_SDP_CONN_STAT_SEND_MID_INC(conn, mid)
#define TS_SDP_CONN_STAT_RECV_MID_INC(conn, mid)

#endif
/* ---------------------------------------------------------------------- */
/*                                                                        */
/* SDP connection lock                                                    */
/*                                                                        */
/* ---------------------------------------------------------------------- */
#if !defined(TS_KERNEL_2_6)

#define TS_SDP_CONN_LOCK_BH(conn)   spin_lock_bh(&((conn)->lock.slock))
#define TS_SDP_CONN_UNLOCK_BH(conn) spin_unlock_bh(&((conn)->lock.slock))

#define TS_SDP_CONN_LOCK(conn)           \
do {                                     \
  spin_lock_bh(&((conn)->lock.slock));   \
  if ((conn)->lock.users != 0) {         \
     sdp_conn_internal_lock(conn);    \
  } /* if */                             \
  (conn)->lock.users = 1;                \
  spin_unlock_bh(&((conn)->lock.slock)); \
} while(0)

#define TS_SDP_CONN_UNLOCK(conn)         \
do {                                     \
  spin_lock_bh(&((conn)->lock.slock));   \
  if (0 < (TS_SDP_CONN_F_MASK_EVENT & conn->flags) && \
      0 < (TS_SDP_ST_MASK_EVENTS & conn->state)) { \
     sdp_conn_internal_unlock(conn);  \
  } /* if */                             \
  (conn)->lock.users = 0;                \
  wake_up(&((conn)->lock.waitq));        \
  spin_unlock_bh(&((conn)->lock.slock)); \
} while(0)

#define TS_SDP_CONN_RELOCK(conn)         \
do {                                     \
  spin_lock_bh(&((conn)->lock.slock));   \
     sdp_conn_internal_relock(conn);  \
  spin_unlock_bh(&((conn)->lock.slock)); \
} while(0)

#else				/* kernel 2.6 */

/* Work around the temporary workaround for THCA locking bugs :) */

#define TS_SDP_CONN_LOCK_BH(conn)   spin_lock_irqsave(&((conn)->lock.slock), flags)
#define TS_SDP_CONN_UNLOCK_BH(conn) spin_unlock_irqrestore(&((conn)->lock.slock), flags)

#define TS_SDP_CONN_LOCK(conn)                          \
do {                                                    \
  unsigned long flags;                                  \
  spin_lock_irqsave(&((conn)->lock.slock), flags);      \
  if ((conn)->lock.users != 0) {                        \
     sdp_conn_internal_lock(conn, &flags);           \
  } /* if */                                            \
  (conn)->lock.users = 1;                               \
  spin_unlock_irqrestore(&((conn)->lock.slock), flags); \
} while(0)

#define TS_SDP_CONN_UNLOCK(conn)                        \
do {                                                    \
  unsigned long flags;                                  \
  spin_lock_irqsave(&((conn)->lock.slock), flags);      \
  if (0 < (TS_SDP_CONN_F_MASK_EVENT & conn->flags) &&   \
      0 < (TS_SDP_ST_MASK_EVENTS & conn->state)) {      \
     sdp_conn_internal_unlock(conn);                 \
  } /* if */                                            \
  (conn)->lock.users = 0;                               \
  wake_up(&((conn)->lock.waitq));                       \
  spin_unlock_irqrestore(&((conn)->lock.slock), flags); \
} while(0)

#define TS_SDP_CONN_RELOCK(conn)                        \
do {                                                    \
  unsigned long flags;                                  \
  spin_lock_irqsave(&((conn)->lock.slock), flags);      \
     sdp_conn_internal_relock(conn);                 \
  spin_unlock_irqrestore(&((conn)->lock.slock), flags); \
} while(0)

#endif				/* kernel 2.6 */


#define TS_SDP_CONN_HOLD(conn) __sdp_conn_hold((conn))
#define TS_SDP_CONN_PUT(conn)  sdp_conn_put((conn))
#define TS_SDP_CONN_ERROR(conn)  __sdp_conn_error((conn))

#endif /* _TS_SDP_CONN_H */









