/*
  This software is available to you under a choice of one of two
  licenses.  You may choose to be licensed under the terms of the GNU
  General Public License (GPL) Version 2, available at
  <http://www.fsf.org/copyleft/gpl.html>, or the OpenIB.org BSD
  license, available in the LICENSE.TXT file accompanying this
  software.  These details are also available at
  <http://openib.org/license.html>.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
  BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
  ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
  CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
  SOFTWARE.

  Copyright (c) 2004 Topspin Communications.  All rights reserved.
  Copyright (c) 2005 Mellanox Technologies.   All rights reserved.


  $Id: sdp_kvec.c 159 2004-05-14 00:16:54Z libor $
*/

#include "sdp_main.h"

/* ------------------------------------------------------------------- */
/*                                                                     */
/* static kvec internal functions                                      */
/*                                                                     */
/* ------------------------------------------------------------------- */
/* ========================================================================= */
/*.._sdp_iocb_q_cancel_lookup_func -- lookup function for cancelation */
static int _sdp_iocb_q_cancel_lookup_func(struct sdpc_desc *element, void *arg)
{
	TS_CHECK_NULL(element, -EINVAL);

	return ((TS_SDP_GENERIC_TYPE_IOCB == element->type) ? 0 : -ERANGE);
} /* _sdp_iocb_q_cancel_lookup_func */

/* ------------------------------------------------------------------- */
/*                                                                     */
/* Public write functions                                              */
/*                                                                     */
/* ------------------------------------------------------------------- */

/* ------------------------------------------------------------------- */
/*                                                                     */
/* Cancel operations                                                   */
/*                                                                     */
/* ------------------------------------------------------------------- */

/* ========================================================================= */
/*.._sdp_iocb_q_cancel_read_pending -- cancel all pending read AIOs */
static int _sdp_iocb_q_cancel_read_pending(struct sdp_opt *conn, ssize_t error)
{
	TS_CHECK_NULL(conn, -EINVAL);
		return sdp_iocb_q_cancel(&conn->r_pend, TS_SDP_IOCB_F_ALL,
					error);
} /* _sdp_iocb_q_cancel_read_pending */

/* ========================================================================= */
/*.._sdp_iocb_q_cancel_read_source -- cancel all pending read AIOs */
static int _sdp_iocb_q_cancel_read_source(struct sdp_opt *conn, ssize_t error)
{
	struct sdpc_iocb *iocb;
	int result;

	TS_CHECK_NULL(conn, -EINVAL);
   	while (NULL != (iocb = (struct sdpc_iocb *) sdp_desc_q_lookup(&conn->r_src,
								   _sdp_iocb_q_cancel_lookup_func,
								   NULL))) {

		result = sdp_iocb_q_remove(iocb);
		TS_EXPECT(MOD_LNX_SDP, !(0 > result));

        if (iocb->aio==1)
        {
            result = sdp_iocb_complete(iocb, error);
        }  else {
            result = sdp_zcopy_complete(iocb, error);
        }
		TS_EXPECT(MOD_LNX_SDP, !(0 > result));
	}

	return 0;
} /* _sdp_iocb_q_cancel_read_source */

/* ========================================================================= */
/*.._sdp_iocb_q_cancel_read_snk -- cancel all pending read AIOs */
static int _sdp_iocb_q_cancel_read_snk(struct sdp_opt *conn, ssize_t error)
{
	TS_CHECK_NULL(conn, -EINVAL);

	return sdp_iocb_q_cancel(&conn->r_snk, TS_SDP_IOCB_F_ALL, error);
} /* _sdp_iocb_q_cancel_read_snk */

/* ========================================================================= */
/*.._sdp_iocb_q_cancel_write_pending -- cancel all pending read AIOs */
static int _sdp_iocb_q_cancel_write_pending(struct sdp_opt *conn, ssize_t error)
{
	struct sdpc_iocb *iocb;
	int result;

	TS_CHECK_NULL(conn, -EINVAL);

	while (NULL !=
	       (iocb =
		(struct sdpc_iocb *) sdp_desc_q_lookup(&conn->send_queue,
						    _sdp_iocb_q_cancel_lookup_func,
						    NULL))) {

		result = sdp_iocb_q_remove(iocb);
		TS_EXPECT(MOD_LNX_SDP, !(0 > result));

		result = sdp_iocb_complete(iocb, error);
		TS_EXPECT(MOD_LNX_SDP, !(0 > result));
	}

	return 0;
} /* _sdp_iocb_q_cancel_write_pending */

/* ========================================================================= */
/*.._sdp_iocb_q_cancel_write_source -- cancel all pending source AIOs */
static int _sdp_iocb_q_cancel_write_source(struct sdp_opt *conn, ssize_t error)
{
	TS_CHECK_NULL(conn, -EINVAL);

	return sdp_iocb_q_cancel(&conn->w_src, TS_SDP_IOCB_F_ALL, error);
} /* _sdp_iocb_q_cancel_write_source */

    

/* ========================================================================= */
/*.._sdp_iocb_q_cancel_write_snk -- cancel all pending sink AIOs */
static int _sdp_iocb_q_cancel_write_snk(struct sdp_opt *conn, ssize_t error)
{
	struct sdpc_iocb *iocb;
	int result;

	TS_CHECK_NULL(conn, -EINVAL);

	while (NULL != (iocb = (struct sdpc_iocb *) sdp_desc_q_lookup(&conn->w_snk,
								   _sdp_iocb_q_cancel_lookup_func,
								   NULL))) {

		result = sdp_iocb_q_remove(iocb);
		TS_EXPECT(MOD_LNX_SDP, !(0 > result));

		result = sdp_iocb_complete(iocb, error);
		TS_EXPECT(MOD_LNX_SDP, !(0 > result));
	}

	return 0;
} /* _sdp_iocb_q_cancel_write_snk */

/* ========================================================================= */
/*..sdp_iocb_q_cancel_all_read -- cancel all outstanding read AIOs */
int sdp_iocb_q_cancel_all_read(struct sdp_opt *conn, ssize_t error)
{
	int result;

	TS_CHECK_NULL(conn, -EINVAL);

	result = _sdp_iocb_q_cancel_read_pending(conn, error);
	TS_EXPECT(MOD_LNX_SDP, !(0 > result));

	result = _sdp_iocb_q_cancel_read_snk(conn, error);
	TS_EXPECT(MOD_LNX_SDP, !(0 > result));

	result = _sdp_iocb_q_cancel_read_source(conn, error);
	TS_EXPECT(MOD_LNX_SDP, !(0 > result));

	return 0;
} /* sdp_iocb_q_cancel_all_read */

/* ========================================================================= */
/*..sdp_iocb_q_cancel_all_write -- cancel all outstanding write AIOs */
int sdp_iocb_q_cancel_all_write(struct sdp_opt *conn, ssize_t error)
{
	int result;

	TS_CHECK_NULL(conn, -EINVAL);

	result = _sdp_iocb_q_cancel_write_pending(conn, error);
	TS_EXPECT(MOD_LNX_SDP, !(0 > result));

	result = _sdp_iocb_q_cancel_write_source(conn, error);
	TS_EXPECT(MOD_LNX_SDP, !(0 > result));

    


	result = _sdp_iocb_q_cancel_write_snk(conn, error);
	TS_EXPECT(MOD_LNX_SDP, !(0 > result));

	return 0;
} /* sdp_iocb_q_cancel_all_write */

/* ========================================================================= */
/*..sdp_iocb_q_cancel_all -- cancel all outstanding AIOs */
int sdp_iocb_q_cancel_all(struct sdp_opt *conn, ssize_t error)
{
	int result;

	TS_CHECK_NULL(conn, -EINVAL);

	result = sdp_iocb_q_cancel_all_read(conn, error);
	TS_EXPECT(MOD_LNX_SDP, !(0 > result));

	result = sdp_iocb_q_cancel_all_write(conn, error);
	TS_EXPECT(MOD_LNX_SDP, !(0 > result));

	return 0;
} /* sdp_iocb_q_cancel_all */
