/*
  This software is available to you under a choice of one of two
  licenses.  You may choose to be licensed under the terms of the GNU
  General Public License (GPL) Version 2, available at
  <http://www.fsf.org/copyleft/gpl.html>, or the OpenIB.org BSD
  license, available in the LICENSE.TXT file accompanying this
  software.  These details are also available at
  <http://openib.org/license.html>.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
  BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
  ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
  CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
  SOFTWARE.

  Copyright (c) 2004 Topspin Communications.  All rights reserved.
  Copyright (c) 2005 Mellanox Technologies.   All rights reserved.


  $Id: sdp_proto.h 379 2004-07-08 02:24:32Z roland $
*/

#ifndef _TS_SDP_PROTO_H
#define _TS_SDP_PROTO_H
/*
 * types used in the prototype file.                               
 */
#include <ts_ib_core.h>

#include "sdp_types.h"
#include "sdp_conn.h"
#include "sdp_buff.h"
#include "sdp_advt.h"
#include "sdp_iocb.h"
#include "sdp_queue.h"

/* --------------------------------------------------------------------- */
/* Buffer managment                                                      */
/* --------------------------------------------------------------------- */
struct sdpc_buff *sdp_buff_pool_get(void);

int sdp_buff_pool_put(struct sdpc_buff *buff);

int sdp_buff_pool_chain_put(struct sdpc_buff *buff, u32 count);

int sdp_buff_pool_chain_link(struct sdpc_buff *head, struct sdpc_buff *buff);

int sdp_buff_pool_size(void);

int sdp_buff_pool_buff_size(void);

int sdp_buff_q_init(struct sdpc_buff_q *pool, char *name, u32 size);

int sdp_buff_q_clear(struct sdpc_buff_q *pool);

int sdp_buff_q_remove(struct sdpc_buff *buff);

struct sdpc_buff *sdp_buff_q_get(struct sdpc_buff_q *pool);

struct sdpc_buff *sdp_buff_q_get_head(struct sdpc_buff_q *pool);

struct sdpc_buff *sdp_buff_q_get_tail(struct sdpc_buff_q *pool);

struct sdpc_buff *sdp_buff_q_look_head(struct sdpc_buff_q *pool);

struct sdpc_buff *sdp_buff_q_look_tail(struct sdpc_buff_q *pool);

int sdp_buff_q_put(struct sdpc_buff_q *pool, struct sdpc_buff *buff);

int sdp_buff_q_put_head(struct sdpc_buff_q *pool, struct sdpc_buff *buff);

int sdp_buff_q_put_tail(struct sdpc_buff_q *pool, struct sdpc_buff *buff);

struct sdpc_buff *sdp_buff_q_fetch_head(struct sdpc_buff_q *pool,
				 tSDP_BUFF_TEST_FUNC test_func, 
				 void *usr_arg);

struct sdpc_buff *sdp_buff_q_fetch_tail(struct sdpc_buff_q *pool,
				 tSDP_BUFF_TEST_FUNC test_func,
				 void *usr_arg);

int sdp_buff_q_trav_tail(struct sdpc_buff_q *pool,
				 tSDP_BUFF_TRAV_FUNC trav_func,
				 void *usr_arg);

int sdp_buff_q_trav_head(struct sdpc_buff_q *pool,
				 tSDP_BUFF_TRAV_FUNC trav_func,
				 void *usr_arg);

struct sdpc_buff *sdp_buff_q_fetch(struct sdpc_buff_q *pool,
			     tSDP_BUFF_TEST_FUNC test_func,
			     void *usr_arg);

int sdp_buff_pool_init(tUINT32 buff_min, tUINT32 buff_max);

void sdp_buff_pool_destroy(void);

int sdp_proc_dump_buff_pool(char *buffer,
			    int   max_size,
			    off_t start_index,
			    long *end_index);

/* --------------------------------------------------------------------- */
/* Wall between userspace protocol and SDP protocol proper               */
/* --------------------------------------------------------------------- */

int sdp_wall_send_connect(struct sdp_opt *conn);

int sdp_wall_send_accept(struct sdp_opt *conn);

int sdp_wall_send_reject(struct sdp_opt *conn);

int sdp_wall_send_confirm(struct sdp_opt *conn);

int sdp_wall_send_failed(struct sdp_opt *conn);

int sdp_wall_send_close(struct sdp_opt *conn);

int sdp_wall_send_closing(struct sdp_opt *conn);

int sdp_wall_send_abort(struct sdp_opt *conn);

int sdp_wall_recv_connect(struct sdp_opt *conn);

int sdp_wall_recv_accept(struct sdp_opt *conn);

int sdp_wall_recv_reject(struct sdp_opt *conn, int error);

int sdp_wall_recv_confirm(struct sdp_opt *conn);

int sdp_wall_recv_failed(struct sdp_opt *conn, int error);

int sdp_wall_recv_close(struct sdp_opt *conn);

int sdp_wall_recv_closing(struct sdp_opt *conn);

int sdp_wall_recv_abort(struct sdp_opt *conn);

int sdp_wall_recv_drop(struct sdp_opt *conn);

int sdp_wall_abort(struct sdp_opt *conn);

tINT32 sdp_recv_buff(struct sdp_opt *conn, struct sdpc_buff *buff);
/* --------------------------------------------------------------------- */
/* Zcopy advertisment managment                                          */
/* --------------------------------------------------------------------- */
tINT32 sdp_main_advt_init(void);

tINT32 sdp_main_advt_cleanup(void);

struct sdpc_advt_q *sdp_advt_q_create(tINT32 * result);

int sdp_advt_q_init(struct sdpc_advt_q *table);

int sdp_advt_q_clear(struct sdpc_advt_q *table);

int sdp_advt_q_destroy(struct sdpc_advt_q *table);

struct sdpc_advt *sdp_advt_create(void);

int sdp_advt_destroy(struct sdpc_advt *advt);

struct sdpc_advt *sdp_advt_q_get(struct sdpc_advt_q *table);

struct sdpc_advt *sdp_advt_q_look(struct sdpc_advt_q *table);

int sdp_advt_q_put(struct sdpc_advt_q *table, struct sdpc_advt *advt);
/* --------------------------------------------------------------------- */
/* Zcopy IOCB managment                                                  */
/* --------------------------------------------------------------------- */
tINT32 sdp_main_iocb_init(void);

tINT32 sdp_main_iocb_cleanup(void);

void sdp_mark_dirty(struct page *page);

struct sdpc_iocb_q *sdp_iocb_q_create(tINT32 * result);

int sdp_iocb_q_init(struct sdpc_iocb_q *table);

int sdp_iocb_q_clear(struct sdpc_iocb_q *table);

int sdp_iocb_q_destroy(struct sdpc_iocb_q *table);

struct sdpc_iocb *sdp_iocb_create(void);

struct sdpc_iocb *sdp_zcopy_create(void);

int sdp_iocb_destroy(struct sdpc_iocb *iocb);

struct sdpc_iocb *sdp_iocb_q_look(struct sdpc_iocb_q *table);

struct sdpc_iocb *sdp_iocb_q_get_head(struct sdpc_iocb_q *table);

struct sdpc_iocb *sdp_iocb_q_get_tail(struct sdpc_iocb_q *table);

int sdp_iocb_q_put_head(struct sdpc_iocb_q *table, struct sdpc_iocb *iocb);

int sdp_iocb_q_put_tail(struct sdpc_iocb_q *table, struct sdpc_iocb *iocb);

struct sdpc_iocb *sdp_iocb_q_get_key(struct sdpc_iocb_q *table, tUINT32 key);

struct sdpc_iocb *sdp_iocb_q_lookup(struct sdpc_iocb_q *table, tUINT32 key);

int sdp_iocb_q_cancel(struct sdpc_iocb_q *table, u32 mask, ssize_t comp);

int sdp_iocb_q_remove(struct sdpc_iocb *iocb);

int sdp_iocb_register(struct sdpc_iocb *iocb, struct sdp_opt *conn);

int sdp_iocb_release(struct sdpc_iocb *iocb);

int sdp_iocb_complete(struct sdpc_iocb *iocb, ssize_t status);

int sdp_zcopy_release(struct sdpc_iocb *iocb);
int sdp_zcopy_complete(struct sdpc_iocb *iocb, ssize_t status);
int sdp_zcopy_register(struct sdpc_iocb *iocb, struct sdp_opt *conn);

int sdp_signal_abort(struct sdp_opt *conn);

/* --------------------------------------------------------------------- */
/* Generic object managment                                              */
/* --------------------------------------------------------------------- */
int sdp_desc_q_remove(struct sdpc_desc *element);

struct sdpc_desc *sdp_desc_q_get_all(struct sdpc_desc_q *table);

struct sdpc_desc *sdp_desc_q_get_head(struct sdpc_desc_q *table);

struct sdpc_desc *sdp_desc_q_get_tail(struct sdpc_desc_q *table);

int sdp_desc_q_put_head(struct sdpc_desc_q *table,
				struct sdpc_desc *element);

int sdp_desc_q_put_tail(struct sdpc_desc_q *table,
				struct sdpc_desc *element);

struct sdpc_desc *sdp_desc_q_look_head(struct sdpc_desc_q *table);

struct sdpc_desc *sdp_desc_q_look_tail(struct sdpc_desc_q *table);

int sdp_desc_q_type_head(struct sdpc_desc_q *table);

int sdp_desc_q_type_tail(struct sdpc_desc_q *table);

struct sdpc_desc *sdp_desc_q_look_type_head(struct sdpc_desc_q *table,
					   tSDP_GENERIC_TYPE type);

struct sdpc_desc *sdp_desc_q_look_type_tail(struct sdpc_desc_q *table,
					   tSDP_GENERIC_TYPE type);

struct sdpc_desc *sdp_desc_q_lookup(struct sdpc_desc_q *table,
				     tSDP_GENERIC_LOOKUP_FUNC lookup_func,
				     void *arg);

int sdp_desc_q_types_size(struct sdpc_desc_q *table,
			  tSDP_GENERIC_TYPE type);

int sdp_desc_q_size(struct sdpc_desc_q *table);

struct sdpc_desc_q *sdp_desc_q_create(tINT32 * result);

int sdp_desc_q_init(struct sdpc_desc_q *table);

int sdp_desc_q_clear(struct sdpc_desc_q *table);

int sdp_desc_q_destroy(struct sdpc_desc_q *table);

tINT32 sdp_main_desc_init(void);

tINT32 sdp_main_desc_cleanup(void);
/* --------------------------------------------------------------------- */
/* proc entry managment                                                  */
/* --------------------------------------------------------------------- */
tINT32 sdp_main_proc_init(void);

tINT32 sdp_main_proc_cleanup(void);
/* --------------------------------------------------------------------- */
/* connection table                                                      */
/* --------------------------------------------------------------------- */
int sdp_conn_table_init(int proto_family,
			int conn_size,
			int buff_min,
			int buff_max,
			int recv_post_max,
			int recv_buff_max,
			int send_post_max,
			int send_buff_max,
			int send_usig_max);

tINT32 sdp_conn_table_clear(void);

int sdp_proc_dump_conn_main(char *buffer,
			    int   max_size,
			    off_t start_index,
			    long *end_index);

int sdp_proc_dump_conn_data(char *buffer,
			    int   max_size,
			    off_t start_index,
			    long *end_index);

int sdp_proc_dump_conn_rdma(char *buffer,
			    int   max_size,
			    off_t start_index,
			    long *end_index);

int sdp_proc_dump_conn_sopt(char *buffer,
			    int   max_size,
			    off_t start_index,
			    long *end_index);

int sdp_proc_dump_device(char *buffer,
			 int   max_size,
			 off_t start_index,
			 long *end_index);

struct sdp_opt *sdp_conn_table_lookup(tINT32 entry);

struct sdp_opt *sdp_conn_alloc(tINT32 priority, tTS_IB_CM_COMM_ID comm_id);

int sdp_conn_alloc_ib(struct sdp_opt *conn,
		      struct ib_device *device, 
		      tTS_IB_PORT hw_port);

int sdp_conn_put(struct sdp_opt *conn);

void sdp_inet_wake_send(struct sock *sk);

void sdp_inet_wake_generic(struct sock *sk);

void sdp_inet_wake_recv(struct sock *sk, int len);

void sdp_inet_wake_error(struct sock *sk);

void sdp_inet_wake_urg(struct sock *sk);
/* --------------------------------------------------------------------- */
/* port/queue managment                                                  */
/* --------------------------------------------------------------------- */
tINT32 sdp_inet_accept_q_put(struct sdp_opt *listen_conn, struct sdp_opt *accept_conn);

struct sdp_opt *sdp_inet_accept_q_get(struct sdp_opt *listen_conn);

tINT32 sdp_inet_accept_q_remove(struct sdp_opt *accept_conn);

tINT32 sdp_inet_listen_start(struct sdp_opt *listen_conn);

tINT32 sdp_inet_listen_stop(struct sdp_opt *listen_conn);

struct sdp_opt *sdp_inet_listen_lookup(tUINT32 addr, tUINT16 port);

tINT32 sdp_inet_port_get(struct sdp_opt *conn, tUINT16 port);

tINT32 sdp_inet_port_put(struct sdp_opt *conn);

tINT32 sdp_inet_port_inherit(struct sdp_opt *parent, struct sdp_opt *child);

/* --------------------------------------------------------------------- */
/* post functions                                                        */
/* --------------------------------------------------------------------- */
int sdp_cm_listen_start(struct sdev_root *dev_root);

int sdp_cm_listen_stop(struct sdev_root *dev_root);

int sdp_cm_connect(struct sdp_opt *conn);

int sdp_cm_accept(struct sdp_opt *conn);

int sdp_cm_disconnect(struct sdp_opt *conn);

int sdp_cm_reject(struct sdp_opt *conn);

int sdp_cm_failed(struct sdp_opt *conn);

int sdp_cm_confirm(struct sdp_opt *conn);

tINT32 sdp_recv_flush(struct sdp_opt *conn);

tINT32 sdp_send_flush(struct sdp_opt *conn);

tINT32 sdp_send_ctrl_ack(struct sdp_opt *conn);

tINT32 sdp_send_ctrl_disconnect(struct sdp_opt *conn);

tINT32 sdp_send_ctrl_abort(struct sdp_opt *conn);

tINT32 sdp_send_ctrl_send_sm(struct sdp_opt *conn);

tINT32 sdp_send_ctrl_snk_avail(struct sdp_opt *conn,
			     tUINT32 size, 
			     tUINT32 rkey,
			     tUINT64 addr);

tINT32 sdp_send_ctrl_resize_buff_ack(struct sdp_opt *conn, tUINT32 size);

tINT32 sdp_send_ctrl_rdma_rd(struct sdp_opt *conn, tINT32 size);

tINT32 sdp_send_ctrl_rdma_wr(struct sdp_opt *conn, tUINT32 size);

tINT32 sdp_send_ctrl_mode_ch(struct sdp_opt *conn, tUINT8 mode);

tINT32 sdp_send_ctrl_src_cancel(struct sdp_opt *conn);

tINT32 sdp_send_ctrl_snk_cancel(struct sdp_opt *conn);

tINT32 sdp_send_ctrl_snk_cancel_ack(struct sdp_opt *conn);

/* --------------------------------------------------------------------- */
/* inet functions                                                        */
/* --------------------------------------------------------------------- */
/* --------------------------------------------------------------------- */
/* event functions                                                       */
/* --------------------------------------------------------------------- */
tINT32 sdp_cq_event_locked(struct ib_cq_entry *comp, struct sdp_opt *conn);

void sdp_cq_event_handler(struct ib_cq *cq, void *arg);

tTS_IB_CM_CALLBACK_RETURN sdp_cm_event_handler(tTS_IB_CM_EVENT event,
					      tTS_IB_CM_COMM_ID comm_id,
					      void *params, void *arg);

tINT32 sdp_event_recv(struct sdp_opt *conn, struct ib_cq_entry *comp);

tINT32 sdp_event_send(struct sdp_opt *conn, struct ib_cq_entry *comp);

tINT32 sdp_event_read(struct sdp_opt *conn, struct ib_cq_entry *comp);

tINT32 sdp_event_write(struct sdp_opt *conn, struct ib_cq_entry *comp);


/*-----------------------------------------------------------------------*/
int sdp_zcopy_threshold(void);

int sdp_zcopy_use(void);
/* --------------------------------------------------------------------- */
/* internal connection lock functions                                    */
/* --------------------------------------------------------------------- */
void sdp_conn_internal_lock(struct sdp_opt *conn
#if defined(TS_KERNEL_2_6)
/* Work around the temporary workaround for THCA locking bugs :) */
			       , unsigned long *flags
#endif
    );

void sdp_conn_internal_unlock(struct sdp_opt *conn);

void sdp_conn_internal_relock(struct sdp_opt *conn);

tINT32 sdp_conn_cq_drain(struct ib_cq *cq, struct sdp_opt *conn);  

/* --------------------------------------------------------------------- */
/* DATA transport                                                        */
/* --------------------------------------------------------------------- */
int sdp_inet_send(
#ifdef TS_KERNEL_2_6
	          struct kiocb *iocb,
		  struct socket *sock, 
		  struct msghdr *msg,
		  size_t size
#else
		  struct socket *sock,
		  struct msghdr *msg,
		  int    size, 
		  struct scm_cookie *scm
#endif
    );

int sdp_inet_recv(
#ifdef TS_KERNEL_2_6
	 	  struct kiocb *iocb,
		  struct socket *sock,
		  struct msghdr *msg, 
		  size_t size, 
		  int    flags
#else
		  struct socket *sock,
		  struct msghdr *msg,
		  int    size, 
		  int    flags, 
		  struct scm_cookie *scm
#endif
    );

#ifdef _TS_SDP_AIO_SUPPORT
int sdp_inet_write(struct socket *sock, kvec_cb_t cb, size_t size);

int sdp_inet_read(struct socket *sock, kvec_cb_t cb, size_t size);
#endif

int sdp_iocb_q_cancel_all_read(struct sdp_opt *conn, ssize_t error);

int sdp_iocb_q_cancel_all_write(struct sdp_opt *conn, ssize_t error);

int sdp_iocb_q_cancel_all(struct sdp_opt *conn, ssize_t error);
/* --------------------------------------------------------------------- */
/* Function types.                                                       */
/* --------------------------------------------------------------------- */
/*
 * Event handling function, demultiplexed base on Message ID
 */
typedef int (*tGW_SDP_EVENT_CB_FUNC)(struct sdp_opt *conn, 
				     struct sdpc_buff *buff);

/* --------------------------------------------------------------------- */
/* Inline functions.                                                     */
/* --------------------------------------------------------------------- */
/* ====================================================================== */
/*..__sdp_inet_write_space -- writable space on send side. */
static __inline__ tINT32 __sdp_inet_write_space(struct sdp_opt *conn, tINT32 urg)
{
	tINT32 size;

	TS_CHECK_NULL(conn, -EINVAL);
	/*
	 * Allow for more space if Urgent data is being considered
	 */
	size = (conn->send_buf - conn->send_qud);
	/*
	 * write space is determined by amount of outstanding bytes of data
	 * and number of buffers used for transmission by this connection
	 */
	if (0 < (TS_SDP_ST_MASK_OPEN & conn->istate) &&
	    (conn->send_max >
	     sdp_desc_q_types_size(&conn->send_queue,
				   TS_SDP_GENERIC_TYPE_BUFF))) {

		return ((TS_SDP_DEV_INET_THRSH < size || 1 < urg) ? size : 0);
	}
	else {

		return 0;
	}
} /* __sdp_inet_write_space */

/* ====================================================================== */
/*..__sdp_inet_writable -- return non-zero if socket is writable. */
static __inline__ tINT32 __sdp_inet_writable(struct sdp_opt *conn)
{
	TS_CHECK_NULL(conn, -EINVAL);

	if (0 < (TS_SDP_ST_MASK_OPEN & conn->istate)) {

		return (__sdp_inet_write_space(conn, 0) <
			(conn->send_qud / 2)) ? 0 : 1;
	}
	else {

		return 0;
	}
} /* __sdp_inet_writable */

/* ======================================================================== */
/*..__sdp_conn_stat_dump -- dump stats to the log */
static __inline__ tINT32 __sdp_conn_stat_dump(struct sdp_opt *conn)
{
#ifdef _TS_SDP_CONN_STATS_REC
	tUINT32 counter;

	TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_INOUT,
		 "STAT: src <%u> snk <%u>", conn->src_serv, conn->snk_serv);

	for (counter = 0; counter < 0x20; counter++) {

		if (0 < conn->send_mid[counter] || 0 < conn->recv_mid[counter]) {

			TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_INOUT,
				 "STAT: MID send <%02x> <%u>", counter,
				 conn->send_mid[counter]);
			TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_INOUT,
				 "STAT: MID recv <%02x> <%u>", counter,
				 conn->recv_mid[counter]);
		}
	}
#endif
	return 0;
} /* __sdp_conn_stat_dump */

/* ======================================================================== */
/*..__sdp_conn_state_dump -- dump state information to the log */
static __inline__ tINT32 __sdp_conn_state_dump(struct sdp_opt *conn)
{
#ifdef _TS_SDP_CONN_STATE_REC
	tUINT32 counter;

	TS_CHECK_NULL(conn, -EINVAL);

	TS_TRACE(MOD_LNX_SDP, T_TERSE, TRACE_FLOW_INOUT,
		 "STATE: Connection <%04x> state:", conn->hashent);

	if (TS_SDP_CONN_ST_INVALID == conn->state_rec.state[0]) {

		TS_TRACE(MOD_LNX_SDP, T_TERSE, TRACE_FLOW_INOUT,
			 "STATE:   No state history. <%d>",
			 conn->state_rec.value);
	}
	else {

		for (counter = 0;
		     TS_SDP_CONN_ST_INVALID != conn->state_rec.state[counter];
		     counter++) {

			TS_TRACE(MOD_LNX_SDP, T_TERSE, TRACE_FLOW_INOUT,
				 "STATE:   counter <%02x> state <%04x> <%s:%d>",
				 counter,
				 conn->state_rec.state[counter],
				 conn->state_rec.file[counter],
				 conn->state_rec.line[counter]);
		}
	}
#endif
	return 0;
} /* __sdp_conn_state_dump */

/* ======================================================================== */
/*..__sdp_conn_hold -- increment reference count */
static inline void __sdp_conn_hold(struct sdp_opt *conn)
{
	atomic_inc(&conn->refcnt);
} /* __sdp_conn_hold */

/* ======================================================================== */
/*..sdp_conn_put_locked -- decrement reference count */
/* Only safe to call on a locked connection */
static inline void sdp_conn_put_locked(struct sdp_opt *conn)
{
	int result = 0;
    result = (atomic_dec_and_test(&conn->refcnt));
	TS_EXPECT(MOD_LNX_SDP, !result);

} /* __sdp_conn_hold */

/* ======================================================================== */
/*..__sdp_conn_error -- get the connections error value destructively. */
static inline tINT32 __sdp_conn_error(struct sdp_opt *conn)
{
	/*
	 * The connection error parameter is set and read under the connection
	 * lock, however the linux socket error, needs to be xchg'd since the
	 * SO_ERROR getsockopt happens outside of the connection lock.
	 */
	tINT32 error = xchg(&TS_SDP_OS_SK_ERR(conn->sk), 0);
	TS_SDP_OS_CONN_SET_ERR(conn, 0);

	return -error;
} /* __sdp_conn_error */

#endif /* _TS_SDP_PROTO_H */
