/*
  This software is available to you under a choice of one of two
  licenses.  You may choose to be licensed under the terms of the GNU
  General Public License (GPL) Version 2, available at
  <http://www.fsf.org/copyleft/gpl.html>, or the OpenIB.org BSD
  license, available in the LICENSE.TXT file accompanying this
  software.  These details are also available at
  <http://openib.org/license.html>.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
  BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
  ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
  CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
  SOFTWARE.

  Copyright (c) 2004 Topspin Communications.  All rights reserved.
  Copyright (c) 2005 Mellanox Technologies.   All rights reserved.


  $Id: sdp_queue.h 159 2004-05-14 00:16:54Z libor $
*/

#ifndef _TS_SDP_QUEUE_H
#define _TS_SDP_QUEUE_H
/*
 * topspin specific includes.
 */
#include "sdp_types.h"
/*
 * defines for object types.
 */

typedef enum {
	TS_SDP_GENERIC_TYPE_UNKOWN = 0x00,
	TS_SDP_GENERIC_TYPE_BUFF = 0x01,
	TS_SDP_GENERIC_TYPE_IOCB = 0x02,
	TS_SDP_GENERIC_TYPE_ADVT = 0x03,
	TS_SDP_GENERIC_TYPE_ZCOPY = 0x04,
	TS_SDP_GENERIC_TYPE_NONE
} tSDP_GENERIC_TYPE;

/* ----------------------------------------------------------------------- */
/* SDP generic queue for multiple object types.                            */
/* ----------------------------------------------------------------------- */

struct sdpc_desc {
	struct sdpc_desc *next;	/* next structure in table */
	struct sdpc_desc *prev;	/* previous structure in table */
	tUINT32 type;		/* element type. (for generic queue) */
	struct sdpc_desc_q *table; /* table to which this object belongs */
	int (*release) (struct sdpc_desc *element); /* release the object */
}; /* struct sdpc_desc */
/*
 * table for holding SDP advertisments.
 */
struct sdpc_desc_q {
	struct sdpc_desc *head;	/* double linked list of advertisments */
	tINT32 size;		/* current number of advertisments in table */
	tUINT16 count[TS_SDP_GENERIC_TYPE_NONE];	/* object specific counter */
}; /* struct sdpc_desc_q */
/*
 * object destruction callback type
 */
typedef int (*tSDP_GENERIC_DESTRUCT_FUNC) (struct sdpc_desc *element);
typedef int (*tSDP_GENERIC_LOOKUP_FUNC) (struct sdpc_desc *element, void *arg);

/* ----------------------------------------------------------------------- */
/*                                                                         */
/* SDP generic queue inline functions.                                     */
/*                                                                         */
/* ----------------------------------------------------------------------- */

/* ========================================================================= */
/*..__sdp_desc_q_size - return the number of elements in the table */
static __inline__ int __sdp_desc_q_size(struct sdpc_desc_q *table)
{
	TS_CHECK_NULL(table, -EINVAL);

	return table->size;
} /* __sdp_desc_q_size */

/* ========================================================================= */
/*..__sdp_desc_q_member - return non-zero if element is in a table */
static __inline__ int __sdp_desc_q_member(struct sdpc_desc *element)
{
	TS_CHECK_NULL(element, -EINVAL);

	return ((NULL == element->table) ? 0 : 1);
} /* __sdp_desc_q_member */

#define sdp_desc_q_size(x)   __sdp_desc_q_size(x)
#define sdp_desc_q_member(x) __sdp_desc_q_member(x)

#endif /* _TS_SDP_QUEUE_H */
