/*
  This software is available to you under a choice of one of two
  licenses.  You may choose to be licensed under the terms of the GNU
  General Public License (GPL) Version 2, available at
  <http://www.fsf.org/copyleft/gpl.html>, or the OpenIB.org BSD
  license, available in the LICENSE.TXT file accompanying this
  software.  These details are also available at
  <http://openib.org/license.html>.

  This source code may incorporate intellectual property owned by
  Microsoft Corporation. Our provision of this source code does not
  include any licenses or any other rights to you under any Microsoft
  intellectual property. If you would like a license from Microsoft
  (e.g., to rebrand, redistribute), you need to contact Microsoft
  directly.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
  BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
  ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
  CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
  SOFTWARE.

  Copyright (c) 2004 Topspin Communications.  All rights reserved.
  Copyright (c) 2005 Mellanox Technologies.   All rights reserved.


  $Id: sdp_rcvd.c 379 2004-07-08 02:24:32Z roland $
*/

#include "sdp_main.h"

/* --------------------------------------------------------------------- */
/*                                                                       */
/* Specific MID handler functions. (RECV)                                */
/*                                                                       */
/* --------------------------------------------------------------------- */
/* ========================================================================= */
/*.._sdp_rcvd_disconnect --  */
static int _sdp_rcvd_disconnect(struct sdp_opt *conn, struct sdpc_buff *buff)
{
	int result = 0;

	TS_CHECK_NULL(conn, -EINVAL);
	TS_CHECK_NULL(buff, -EINVAL);

	TS_TRACE(MOD_LNX_SDP, T_VERY_VERBOSE, TRACE_FLOW_INOUT,
		 "STRM: <%d> <%08x:%04x><%08x:%04x> disconnect received <%04x>",
		 conn->hashent, conn->src_addr, conn->src_port, conn->dst_addr,
		 conn->dst_port, conn->state);
	TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_INOUT," conn->state %d",conn->state);
	switch (conn->state) {
	case TS_SDP_CONN_ST_ESTABLISHED:
		/*
		 * transition state
		 */
		TS_SDP_CONN_ST_SET(conn, TS_SDP_CONN_ST_DIS_RECV_1);
		/*
		 * initiate disconnect to framework
		 */
		result = sdp_wall_recv_close(conn);
		if (0 > result) {

			TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_WARN,
				 "STRM: Error initiating close to stream interface. <%d>",
				 result);
			goto error;
		}

		break;
	case TS_SDP_CONN_ST_DIS_PEND_1:

		TS_SDP_CONN_ST_SET(conn, TS_SDP_CONN_ST_DIS_PEND_R);

		break;
	case TS_SDP_CONN_ST_DIS_SEND_1:

		TS_SDP_CONN_ST_SET(conn, TS_SDP_CONN_ST_DIS_RECV_R);

		break;
	case TS_SDP_CONN_ST_DIS_SENT_1:
		/*
		 * After receiving the final disconnet and posting the DREQ,
		 * the next step is TIME_WAIT from the CM
		 */
		TS_SDP_CONN_ST_SET(conn, TS_SDP_CONN_ST_DIS_RECV_R);
		TS_SDP_CONN_ST_SET(conn, TS_SDP_CONN_ST_TIME_WAIT_1);
		/*
		 * acknowledge disconnect to framework (we're in active
		 * disconnect)
		 */
		result = sdp_wall_recv_closing(conn);
		if (0 > result) {

			TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_WARN,
				 "STRM: Error confirming close to stream interface. <%d>",
				 result);
			goto error;
		}
		/*
		 * Begin IB/CM disconnect
		 */
		result = sdp_cm_disconnect(conn);
		/*
		 * if the remote DREQ was already received, but unprocessed, 
		 * do not treat it as an error
		 */
		if (0 > result) {

			if (-EPROTO != result) {

				TS_TRACE(MOD_LNX_SDP, T_VERBOSE,
					 TRACE_FLOW_WARN,
					 "STRM: Error posting CM disconnect. <%d>",
					 result);
				goto error;
			}
			else {

				TS_TRACE(MOD_LNX_SDP, T_VERBOSE,
					 TRACE_FLOW_WARN,
					 "STRM: CM disconnect of disconnected conn. OK");
			}
		}

		break;
	default:

		TS_TRACE(MOD_LNX_SDP, T_TERSE, TRACE_FLOW_WARN,
			 "STRM: Received disconnect in an unexpected state. <%04x>",
			 conn->state);
		result = -EFAULT;
		goto error;
		break;
	}

	return 0;
error:
	return result;
} /* _sdp_rcvd_disconnect */

/* ========================================================================= */
/*.._sdp_rcvd_abort --  */
static int _sdp_rcvd_abort(struct sdp_opt *conn, struct sdpc_buff *buff)
{
	int result = 0;

	TS_CHECK_NULL(conn, -EINVAL);
	TS_CHECK_NULL(buff, -EINVAL);

	TS_TRACE(MOD_LNX_SDP, T_VERY_VERBOSE, TRACE_FLOW_INOUT,
		 "STRM: <%d> <%08x:%04x><%08x:%04x> abort received in state <%04x>",
		 conn->hashent, conn->src_addr, conn->src_port, conn->dst_addr,
		 conn->dst_port, conn->state);
	/*
	 * Connection should be in some post DisConn recveived state.
	 * Notify gateway interface about abort
	 */
	switch (conn->state) {
	case TS_SDP_CONN_ST_DIS_RECV_1:
	case TS_SDP_CONN_ST_DIS_PEND_R:
	case TS_SDP_CONN_ST_DIS_RECV_R:
	case TS_SDP_CONN_ST_DIS_PEND_2:
	case TS_SDP_CONN_ST_DIS_SEND_2:
	case TS_SDP_CONN_ST_DIS_SENT_2:

		result = sdp_wall_abort(conn);
		if (0 > result) {

			TS_TRACE(MOD_LNX_SDP, T_TERSE, TRACE_FLOW_WARN,
				 "EVENT: error <%d> during abort in state. <%04x>",
				 result, conn->state);
		}

		break;
	default:

		TS_TRACE(MOD_LNX_SDP, T_TERSE, TRACE_FLOW_WARN,
			 "EVENT: received ABORT message in incorrect state. <%04x>",
			 conn->state);
		result = -EPROTO;
	}

	return result;
} /* _sdp_rcvd_abort */

/* ========================================================================= */
/*.._sdp_rcvd_send_sm --  */
static int _sdp_rcvd_send_sm(struct sdp_opt *conn, struct sdpc_buff *buff)
{
	struct sdpc_iocb *iocb;
	int result=0;

	TS_CHECK_NULL(conn, -EINVAL);
	TS_CHECK_NULL(buff, -EINVAL);
	/*
	 * 1) Conn is not in source cancel mode. Send active IOCB
	 *    using buffered mode
	 * 2) Conn is in source cancel, and this message acks the cancel.
	 *    Release all active IOCBs in the source queue.
	 * 3) Conn is in source cancel, but this message doesn't ack the 
	 *    cancel.
	 *
	 *    Do nothing, can't send since the IOCB is being cancelled, but
	 *    cannot release the IOCB since the cancel has yet to be ack'd
	 */
	TS_TRACE(MOD_LNX_SDP, T_VERY_VERBOSE, TRACE_FLOW_INOUT,
		 "EVENT: SendSM. active <%d> count <%d> flags <%08x>",
		 conn->src_sent, conn->src_cncl, conn->flags);

	if (0 < (TS_SDP_CONN_F_SRC_CANCEL_L & conn->flags) &&
	    TS_SDP_SEQ_GTE(buff->bsdh_hdr->seq_ack, conn->src_cseq)) {
		/*
		 * drain the active source queue
		 */
		while (NULL != (iocb = sdp_iocb_q_get_tail(&conn->w_src))) {

			TS_EXPECT(MOD_LNX_SDP,
				  (0 < (TS_SDP_IOCB_F_ACTIVE & iocb->flags)));
			TS_EXPECT(MOD_LNX_SDP,
				  (0 < (TS_SDP_IOCB_F_CANCEL & iocb->flags)));

			conn->src_sent--;
            if (iocb->aio==0)  {
                result = sdp_zcopy_complete(iocb, 0);
            }  else {
                result = sdp_iocb_complete(iocb, 0);
            }
            TS_EXPECT(MOD_LNX_SDP, !(0 > result));
		}
		/*
		 * Cancel complete, clear the state.
		 */
		conn->src_cncl = 0;
		conn->flags &= ~(TS_SDP_CONN_F_SRC_CANCEL_L);
	}

	return result;
} /* _sdp_rcvd_send_sm */

/* ========================================================================= */
/*.._sdp_rcvd_rdma_wr --  */
static int _sdp_rcvd_rdma_wr(struct sdp_opt *conn, struct sdpc_buff *buff)
{
	struct sdp_msg_hdr_rwch *rwch;
	struct sdpc_iocb *iocb;
	int result;

	TS_CHECK_NULL(conn, -EINVAL);
	TS_CHECK_NULL(buff, -EINVAL);

	rwch = (struct sdp_msg_hdr_rwch *) buff->data;
	buff->data = buff->data + sizeof(struct sdp_msg_hdr_rwch);

	result = _sdp_msg_net_to_cpu_rwch(rwch);
	TS_EXPECT(MOD_LNX_SDP, !(0 > result));
	/*
	 * lookup active IOCB read.
	 */
	iocb = sdp_iocb_q_look(&conn->r_snk);
	if (NULL == iocb) {

		TS_TRACE(MOD_LNX_SDP, T_VERY_VERBOSE, TRACE_FLOW_INOUT,
			 "EVENT: Cannot find outstanding IOCB for RdmaRead Completion.");
		result = -EPROTO;
		goto error;
	}

	TS_EXPECT(MOD_LNX_SDP, (0 < (TS_SDP_IOCB_F_RDMA_W & iocb->flags)));

#ifdef _TS_SDP_DATA_PATH_DEBUG
	TS_TRACE(MOD_LNX_SDP, T_VERY_VERBOSE, TRACE_FLOW_INOUT,
		 "EVENT: RDMA Write Complete <%d> <%d:%d:%d> mode <%d> active <%d>",
		 rwch->size, iocb->len, iocb->size, iocb->key,
		 conn->recv_mode, conn->snk_sent);
#endif
	/*
	 * update IOCB
	 */
	if (rwch->size > iocb->len) {

		TS_TRACE(MOD_LNX_SDP, T_VERY_VERBOSE, TRACE_FLOW_INOUT,
			 "EVENT: IOCB and RdmaWrite Completion size mismatch. <%d:%d>",
			 rwch->size, iocb->len);
		result = -EPROTO;
		goto error;
	}
	/*
	 * Iocb is done, deregister memory, and generate completion.
	 */
	iocb = sdp_iocb_q_get_head(&conn->r_snk);
	TS_EXPECT(MOD_LNX_SDP, (NULL != iocb));

	conn->snk_sent--;

	iocb->len -= rwch->size;
	iocb->post += rwch->size;

	TS_SDP_CONN_STAT_SNK_INC(conn);
	TS_SDP_CONN_STAT_READ_INC(conn, iocb->post);
	TS_SDP_CONN_STAT_RQ_DEC(conn, iocb->size);

	result = sdp_iocb_complete(iocb, 0);
	if (0 > result) {

		TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_WARN,
			 "EVENT: Error <%d> completing iocb. <%d>", result,
			 iocb->key);
		(void)sdp_iocb_destroy(iocb);
		goto error;
	}

	return 0;
error:
	return result;
} /* _sdp_rcvd_rdma_wr */

/* ========================================================================= */
/*.._sdp_rcvd_rdma_rd --  */
static int _sdp_rcvd_rdma_rd(struct sdp_opt *conn, struct sdpc_buff *buff)
{
	struct sdp_msg_hdr_rrch *rrch;
	struct sdpc_iocb *iocb;
	int result;

	TS_CHECK_NULL(conn, -EINVAL);
	TS_CHECK_NULL(buff, -EINVAL);

	rrch = (struct sdp_msg_hdr_rrch *) buff->data;
	buff->data = buff->data + sizeof(struct sdp_msg_hdr_rrch);

	result = _sdp_msg_net_to_cpu_rrch(rrch);
	TS_EXPECT(MOD_LNX_SDP, !(0 > result));
	/*
	 * lookup IOCB read.
	 */
	iocb = sdp_iocb_q_look(&conn->w_src);
	if (NULL == iocb) {

		TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_INOUT,
			 "EVENT: Cannot find outstanding IOCB for RdmaRead Completion.");
		result = -EPROTO;
		goto error;
	}

	TS_SDP_CONN_STAT_SRC_INC(conn);
#ifdef _TS_SDP_DATA_PATH_DEBUG
	TS_TRACE(MOD_LNX_SDP, T_VERY_VERBOSE, TRACE_FLOW_INOUT,
		 "EVENT: RDMA Read Complete <%d> <%d:%d:%d> mode <%d> active <%d>",
		 rrch->size, iocb->len, iocb->size, iocb->key,
		 conn->recv_mode, conn->src_sent);
#endif
	/*
	 * update IOCB
	 */
	if (rrch->size > iocb->len) {

		TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_INOUT,
			 "EVENT: IOCB and RdmaRead Completion size mismatch. <%d:%d>",
			 rrch->size, iocb->len);
		result = -EPROTO;
		goto error;
	}
	/*
	 * In combined mode the total RDMA size is going to be the buffer
	 * size minus the size sent in the SrcAvail. We could fix up the
	 * iocb->post in the SrcAvailSend function, but it's better to do
	 * it on the first successful RDMA to make sure we don't get a
	 * false positive of data sent. (specification ambiguity/pain)
	 */
	iocb->post += (0 == iocb->post) ? (iocb->size - iocb->len) : 0;
	iocb->len -= rrch->size;
	iocb->post += rrch->size;

	conn->send_pipe -= rrch->size;
	conn->oob_offset -= (0 < conn->oob_offset) ? rrch->size : 0;

	/*
	 * If iocb is done, deregister memory, and generate completion.
	 */
	if (!(0 < iocb->len)) {
        iocb = sdp_iocb_q_get_head(&conn->w_src);
		TS_EXPECT(MOD_LNX_SDP, (NULL != iocb));
       
		conn->src_sent--;

		TS_SDP_CONN_STAT_WRITE_INC(conn, iocb->post);
		TS_SDP_CONN_STAT_WQ_DEC(conn, iocb->size);
		
		if (iocb->aio == 1) {
		  result = sdp_iocb_complete(iocb, 0);
		} else {
            result = sdp_zcopy_complete(iocb, 0);
            conn->s_pend_copied+=iocb->size;
            TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_WARN,"%d bytes where completed",iocb->size); 
              //sdp_iocb_q_put_tail(&conn->w_comp,iocb);
	  		}
		if (0 > result) {

			TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_WARN,
				 "EVENT: Error <%d> completing iocb. <%d>",
				 result, iocb->key);
			(void)sdp_iocb_destroy(iocb);
			goto error;
		}
	}
	/*
	 * If Source Cancel was in process, and there are no more outstanding
	 * advertisments, then it should now be cleared.
	 */
	if (0 < (TS_SDP_CONN_F_SRC_CANCEL_L & conn->flags) &&
	    0 == sdp_iocb_q_size(&conn->w_src)) {

		conn->src_cncl = 0;
		conn->flags &= ~(TS_SDP_CONN_F_SRC_CANCEL_L);
	}

	return 0;
error:
	return result;
} /* _sdp_rcvd_rdma_rd */

/* ========================================================================= */
/*.._sdp_rcvd_mode_change --  */
static int _sdp_rcvd_mode_change(struct sdp_opt *conn, struct sdpc_buff *buff)
{
	struct sdp_msg_hdr_mch *mch;
	int result;

	TS_CHECK_NULL(conn, -EINVAL);
	TS_CHECK_NULL(buff, -EINVAL);

	mch = (struct sdp_msg_hdr_mch *) buff->data;
	buff->data = buff->data + sizeof(struct sdp_msg_hdr_mch);

	result = _sdp_msg_net_to_cpu_mch(mch);
	TS_EXPECT(MOD_LNX_SDP, !(0 > result));

	TS_TRACE(MOD_LNX_SDP, T_VERY_VERBOSE, TRACE_FLOW_INOUT,
		 "EVENT: Mode transition request <%d> from current mode. <%d:%d>",
		 TS_SDP_MSG_MCH_GET_MODE(mch), conn->recv_mode,
		 conn->send_mode);
	/*
	 * Check if the mode change is to the same mode.
	 */
	if (((TS_SDP_MSG_MCH_GET_MODE(mch) & 0x7) ==
	     ((0 < (TS_SDP_MSG_MCH_GET_MODE(mch) & 0x8)) ?
	      conn->send_mode : conn->recv_mode))) {

		TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_WARN,
			 "EVENT: Mode transition <%d> is a nop. <%d:%d>",
			 TS_SDP_MSG_MCH_GET_MODE(mch), conn->recv_mode,
			 conn->send_mode);
		result = -EPROTO;
		goto error;
	}
	/*
	 * process mode change requests based on which state we're in
	 */
	switch (TS_SDP_MSG_MCH_GET_MODE(mch)) {

	case TS_SDP_MSG_MCH_BUFF_RECV:	/* source to sink */

		if (TS_SDP_MODE_COMB != conn->recv_mode) {

			goto mode_error;
			result = -EPROTO;
		}

		if (0 < conn->src_recv) {

			TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_WARN,
				 "EVENT: Illegal mode transition <%d> SrcAvail pending. <%d>",
				 TS_SDP_MSG_MCH_GET_MODE(mch), conn->src_recv);
			result = -EPROTO;
			goto error;
		}

		break;
	case TS_SDP_MSG_MCH_COMB_SEND:	/* sink to source */

		if (TS_SDP_MODE_BUFF != conn->send_mode) {

			goto mode_error;
			result = -EPROTO;
		}

		break;
	case TS_SDP_MSG_MCH_PIPE_RECV:	/* source to sink */

		if (TS_SDP_MODE_COMB != conn->recv_mode) {

			goto mode_error;
			result = -EPROTO;
		}

		break;
	case TS_SDP_MSG_MCH_COMB_RECV:	/* source to sink */

		if (TS_SDP_MODE_PIPE != conn->recv_mode) {

			goto mode_error;
			result = -EPROTO;
		}

		/* if */
		/*
		 * drop all srcAvail message, they will be reissued, with
		 * combined mode constraints. No snkAvails outstanding on 
		 * this half of the connection. How do I know which srcAvail
		 * RDMA's completed?
		 */
		break;
	default:

		TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_WARN,
			 "EVENT: Invalid mode transition <%d> requested.",
			 TS_SDP_MSG_MCH_GET_MODE(mch));
		result = -EPROTO;
		goto error;
	}
	/*
	 * assign new mode
	 */
	if (0 < (TS_SDP_MSG_MCH_GET_MODE(mch) & 0x8)) {

		conn->send_mode = TS_SDP_MSG_MCH_GET_MODE(mch) & 0x7;
	}
	else {

		conn->recv_mode = TS_SDP_MSG_MCH_GET_MODE(mch) & 0x7;
	}

	return 0;

mode_error:
	TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_WARN,
		 "EVENT: Invalid mode <%d:%d> for requested transition. <%d>",
		 conn->recv_mode, conn->send_mode,
		 TS_SDP_MSG_MCH_GET_MODE(mch));
error:
	return result;
} /* _sdp_rcvd_mode_change */

/* ========================================================================= */
/*.._sdp_rcvd_src_cancel --  */
static int _sdp_rcvd_src_cancel(struct sdp_opt *conn, struct sdpc_buff *buff)
{
	struct sdpc_advt *advt;
	int result;

	TS_CHECK_NULL(conn, -EINVAL);
	TS_CHECK_NULL(buff, -EINVAL);

	TS_TRACE(MOD_LNX_SDP, T_VERY_VERBOSE, TRACE_FLOW_INOUT,
		 "EVENT: RDMA Source Cancel. active <%d> pending <%d> mode <%d>",
		 sdp_advt_q_size(&conn->src_actv),
		 sdp_advt_q_size(&conn->src_pend), conn->send_mode);
	/*
	 * If there are no outstanding advertisments, then there is nothing
	 * to do.
	 */
	if (!(0 < conn->src_recv)) {

		TS_TRACE(MOD_LNX_SDP, T_VERY_VERBOSE, TRACE_FLOW_INOUT,
			 "EVENT: No SrcAvail advertisments to cancel.");
		result = 0;
		goto done;
	}
	/*
	 * Get and terminate the remainder of the oldest advertisment, only
	 * if it's already processed data.
	 */
	advt = sdp_advt_q_look(&conn->src_pend);
	if (NULL != advt && 0 < advt->post) {
		/*
		 * If active, move to the active queue. Otherwise generate an
		 * immediate completion
		 */
		if (0 < (TS_SDP_ADVT_F_READ & advt->flag)) {

			advt = sdp_advt_q_get(&conn->src_pend);
			TS_EXPECT(MOD_LNX_SDP, (NULL != advt));

			result = sdp_advt_q_put(&conn->src_actv, advt);
			TS_EXPECT(MOD_LNX_SDP, !(0 > result));
			/*
			 * keep track of cancellations
			 */
			conn->flags |= TS_SDP_CONN_F_SRC_CANCEL_C;
		}
		else {

			result = sdp_send_ctrl_rdma_rd(conn, advt->post);
			if (0 > result) {

				TS_TRACE(MOD_LNX_SDP, T_VERBOSE,
					 TRACE_FLOW_WARN,
					 "EVENT: Error posting RDMA read completion. <%d>",
					 result);
				goto done;
			}
		}
	}
	/*
	 * drop the pending advertisment queue.
	 */
	while (NULL != (advt = sdp_advt_q_get(&conn->src_pend))) {

		conn->flags |= TS_SDP_CONN_F_SRC_CANCEL_C;

		conn->src_recv--;

		result = sdp_advt_destroy(advt);
		TS_EXPECT(MOD_LNX_SDP, !(0 > result));
	}
	/*
	 * If there are active reads, mark the connection as being in
	 * source cancel. Otherwise
	 */
	if (0 < sdp_advt_q_size(&conn->src_actv)) {
		/*
		 * Set flag. Adjust sequence number ack. (spec dosn't want the
		 * seq ack in subsequent messages updated until the cancel has
		 * been processed. all would be simpler with an explicit cancel
		 * ack, but...)
		 */
		conn->flags |= TS_SDP_CONN_F_SRC_CANCEL_R;
		conn->advt_seq--;
	}
	else {
		/*
		 * If a source was dropped, generate an ack.
		 */
		if (0 < (TS_SDP_CONN_F_SRC_CANCEL_C & conn->flags)) {

			result = sdp_send_ctrl_send_sm(conn);
			if (0 > result) {

				TS_TRACE(MOD_LNX_SDP, T_VERBOSE,
					 TRACE_FLOW_WARN,
					 "EVENT: Error posting SendSm in response to SrcCancel. <%d>",
					 result);
				goto done;
			}

			conn->flags &= ~TS_SDP_CONN_F_SRC_CANCEL_C;
		}
	}

	return 0;
done:
	return result;
} /* _sdp_rcvd_src_cancel */

/* ========================================================================= */
/*.._sdp_rcvd_snk_cancel --  */
static int _sdp_rcvd_snk_cancel(struct sdp_opt *conn, struct sdpc_buff *buff)
{
	struct sdpc_advt *advt;
	tINT32 counter;
	int result;

	TS_CHECK_NULL(conn, -EINVAL);
	TS_CHECK_NULL(buff, -EINVAL);

	TS_TRACE(MOD_LNX_SDP, T_VERY_VERBOSE, TRACE_FLOW_INOUT,
		 "EVENT: RDMA Sink Cancel. active <%d> mode <%d>",
		 conn->snk_recv, conn->send_mode);
	/*
	 * If there are no outstanding advertisments, they we've completed
	 * since the message was sent, and there is nothing to do.
	 */
	if (!(0 < conn->snk_recv)) {

		TS_TRACE(MOD_LNX_SDP, T_VERY_VERBOSE, TRACE_FLOW_INOUT,
			 "EVENT: No SnkAvail advertisments to cancel.");
		result = 0;
		goto done;
	}
	/*
	 * Get the oldest advertisment, and complete it if it's partially
	 * consumed. Throw away all unprocessed advertisments, and ack
	 * the cancel. Since all the active writes and sends are fenced,
	 * it's possible to handle the entire Cancel here.
	 */
	advt = sdp_advt_q_look(&conn->snk_pend);
	if (NULL != advt && 0 < advt->post) {
		/*
		 * Generate completion
		 */
		result = sdp_send_ctrl_rdma_wr(conn, advt->post);
		if (0 > result) {

			TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_WARN,
				 "EVENT: Error posting RDMA write completion. <%d>",
				 result);
			goto done;
		}
		/*
		 * reduce cancel counter
		 */
		counter = -1;
	}
	else {
		/*
		 * cancel count.
		 */
		counter = 0;
	}
	/*
	 * drain the advertisments which have yet to be processed.
	 */
	while (NULL != (advt = sdp_advt_q_get(&conn->snk_pend))) {

		counter++;
		conn->snk_recv--;

		result = sdp_advt_destroy(advt);
		TS_EXPECT(MOD_LNX_SDP, !(0 > result));
	}
	/*
	 * A cancel ack is sent only if we cancelled an advertisment without
	 * sending a completion
	 */
	if (0 < counter) {

		result = sdp_send_ctrl_snk_cancel_ack(conn);
		if (0 > result) {

			TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_WARN,
				 "EVENT: Error posting SnkCacelAck response to SnkCancel. <%d>",
				 result);
			goto done;
		}
	}

	return 0;
done:
	return result;
} /* _sdp_rcvd_snk_cancel */

/* ========================================================================= */
/*.._sdp_rcvd_snk_cancel_ack -- sink cancel confirmantion */
static int _sdp_rcvd_snk_cancel_ack(struct sdp_opt *conn, struct sdpc_buff *buff)
{
	struct sdpc_iocb *iocb;
	int result;

	TS_CHECK_NULL(conn, -EINVAL);
	TS_CHECK_NULL(buff, -EINVAL);

	TS_TRACE(MOD_LNX_SDP, T_VERY_VERBOSE, TRACE_FLOW_INOUT,
		 "EVENT: RDMA Sink Cancel Ack. active <%d> mode <%d> flags <%08x>",
		 conn->snk_sent, conn->recv_mode, conn->flags);

	if (0 == (TS_SDP_CONN_F_SNK_CANCEL & conn->flags)) {

		TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_WARN,
			 "EVENT: Connection not in sink cancel processing. <%08x>",
			 conn->flags);
		result = -EPROTO;
		goto done;
	}
	/*
	 * drain and complete all active IOCBs
	 */
	while (NULL != (iocb = sdp_iocb_q_get_head(&conn->r_snk))) {

		conn->snk_sent--;

		result = sdp_iocb_complete(iocb, 0);
		if (0 > result) {

			TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_WARN,
				 "EVENT: Error <%d> completing iocb. <%d>",
				 result, iocb->key);
			(void)sdp_iocb_destroy(iocb);
			goto done;
		}
	}
	/*
	 * cancellation is complete. Cancel flag is cleared in RECV post.
	 */
	return 0;
done:
	return result;
} /* _sdp_rcvd_snk_cancel_ack */

/* ========================================================================= */
/*.._sdp_rcvd_resize_buff_ack --  buffer size change request */
static int _sdp_rcvd_resize_buff_ack(struct sdp_opt *conn, struct sdpc_buff *buff)
{
	struct sdp_msg_hdr_crbh *crbh;
	int result;

	TS_CHECK_NULL(conn, -EINVAL);
	TS_CHECK_NULL(buff, -EINVAL);

	crbh = (struct sdp_msg_hdr_crbh *) buff->data;
	buff->data = buff->data + sizeof(struct sdp_msg_hdr_crbh);

	result = _sdp_msg_net_to_cpu_crbh(crbh);
	TS_EXPECT(MOD_LNX_SDP, !(0 > result));
	/*
	 * request to change our recv buffer size, we're pretty much locked
	 * into the size we're using, once the connection is set up, so we
	 * reject the request.
	 */
	TS_TRACE(MOD_LNX_SDP, T_VERY_VERBOSE, TRACE_FLOW_INOUT,
		 "EVENT: connection receive buffer size change request. <%d:%d>",
		 crbh->size, conn->recv_size);

	result = sdp_send_ctrl_resize_buff_ack(conn, conn->recv_size);
	if (0 > result) {

		TS_TRACE(MOD_LNX_SDP, T_VERY_VERBOSE, TRACE_FLOW_INOUT,
			 "EVENT: Error acking buffer size change request. <%d>",
			 result);
		goto error;
	}

	return 0;
error:
	return result;
} /* _sdp_rcvd_resize_buff_ack */

/* ========================================================================= */
/*.._sdp_rcvd_suspend --  */
static int _sdp_rcvd_suspend(struct sdp_opt *conn, struct sdpc_buff *buff)
{
	struct sdp_msg_hdr_sch *sch;
	int result;

	TS_CHECK_NULL(conn, -EINVAL);
	TS_CHECK_NULL(buff, -EINVAL);

	sch = (struct sdp_msg_hdr_sch *) buff->data;
	buff->data = buff->data + sizeof(struct sdp_msg_hdr_sch);

	result = _sdp_msg_net_to_cpu_sch(sch);
	TS_EXPECT(MOD_LNX_SDP, !(0 > result));

	return 0;
} /* _sdp_rcvd_suspend */

/* ========================================================================= */
/*.._sdp_rcvd_suspend_Ack --  */
static int _sdp_rcvd_suspend_Ack(struct sdp_opt *conn, struct sdpc_buff *buff)
{
	TS_CHECK_NULL(conn, -EINVAL);
	TS_CHECK_NULL(buff, -EINVAL);

	return 0;
} /* _sdp_rcvd_suspend_Ack */

/* ========================================================================= */
/*.._sdp_rcvd_snk_avail --  */
static int _sdp_rcvd_snk_avail(struct sdp_opt *conn, struct sdpc_buff *buff)
{
	struct sdp_msg_hdr_snkah *snkah;
	struct sdpc_advt *advt;
	struct sdpc_iocb *iocb;
	int result;

	TS_CHECK_NULL(conn, -EINVAL);
	TS_CHECK_NULL(buff, -EINVAL);

	snkah = (struct sdp_msg_hdr_snkah *) buff->data;
	buff->data = buff->data + sizeof(struct sdp_msg_hdr_snkah);

	result = _sdp_msg_net_to_cpu_snkah(snkah);
	TS_EXPECT(MOD_LNX_SDP, !(0 > result));

#ifdef _TS_SDP_DATA_PATH_DEBUG
	TS_TRACE(MOD_LNX_SDP, T_VERY_VERBOSE, TRACE_FLOW_INOUT,
		 "EVENT: SnkAvail received. <%d:%d:%016llx> mode <%d>",
		 snkah->size, snkah->r_key, snkah->addr, conn->send_mode);
#endif
	/*
	 * check our send mode, and make sure parameters are within reason.
	 */
	if (TS_SDP_MODE_PIPE != conn->send_mode) {

		TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_WARN,
			 "EVENT: SinkAvail, source mode <%d> is not pipelined.",
			 conn->send_mode);
		result = -EPROTO;
		goto error;
	}

	if (TS_SDP_MSG_MAX_ADVS == (conn->src_recv + conn->snk_recv)) {

		TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_WARN,
			 "EVENT: SinkAvail, too many RDMA advertisments. <%d>",
			 (conn->src_recv + conn->snk_recv));
		result = -EPROTO;
		goto error;
	}
	/*
	 * Save the advertisment, if it's not stale.
	 */
	if (conn->nond_send == snkah->non_disc) {
		/*
		 * check advertisment size.
		 */
		if (snkah->size < conn->send_size) {

			TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_WARN,
				 "EVENT: SinkAvail, advertisment size <%d> too small. <%d>",
				 snkah->size, conn->send_size);
			result = -EPROTO;
			goto error;
		}
		/*
		 * If there are outstanding SrcAvail messages, they are now 
		 * invalid and the queue needs to be fixed up.
		 */
		if (0 < conn->src_sent) {

			while (NULL !=
			       (iocb =
				sdp_iocb_q_get_tail(&conn->w_src))) {

				TS_EXPECT(MOD_LNX_SDP,
					  (0 <
					   (TS_SDP_IOCB_F_ACTIVE & iocb->
					    flags)));

				iocb->flags &= ~TS_SDP_IOCB_F_ACTIVE;
				conn->src_sent--;
				/*
				 * Either move the active queue, back to the
				 * pending queue, or if the operations are
				 * in cancel processing they need to be
				 * completed.
				 */
				if (0 == (TS_SDP_IOCB_F_CANCEL & iocb->flags)) {

					result =
					    sdp_desc_q_put_head(&conn->
								     send_queue,
								     (struct sdpc_desc *)
								     iocb);
					TS_EXPECT(MOD_LNX_SDP, !(0 > result));
				}
				else {

					result = sdp_iocb_complete(iocb, 0);
					TS_EXPECT(MOD_LNX_SDP, !(0 > result));
				}
			}
			/*
			 * If Source Cancel was in process, it should now 
			 * be cleared.
			 */
			if (0 < (TS_SDP_CONN_F_SRC_CANCEL_L & conn->flags)) {

				conn->src_cncl = 0;
				conn->flags &= ~(TS_SDP_CONN_F_SRC_CANCEL_L);
			}
		}
		/*
		 * create and queue new advertisment
		 */
		advt = sdp_advt_create();
		if (NULL == advt) {

			TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_WARN,
				 "EVENT: SrcAvail cannot be copied.");
			result = -ENOMEM;
			goto error;
		}

		advt->post = 0;
        advt->size = snkah->size;
		advt->addr = snkah->addr;
		advt->rkey = snkah->r_key;

		conn->snk_recv++;
		conn->sink_actv++;

		conn->s_cur_adv = 1;
		conn->s_par_adv = 0;

		result = sdp_advt_q_put(&conn->snk_pend, advt);
		if (0 > result) {

			TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_WARN,
				 "EVENT: SnkAvail cannot be saved. <%d>",
				 result);
			goto advt_error;
		}
	}
	else {

		conn->nond_send--;
	}

	conn->s_wq_cur = TS_SDP_DEV_SEND_POST_SLOW;
	conn->s_wq_par = 0;
	/*
	 * consume any data in the advertisment for the other direction.
	 */
	if (0 < (buff->tail - buff->data)) {

		result = sdp_recv_buff(conn, buff);
		if (0 < result) {
			/*
			 * count number of bytes buffered by the connection,
			 * zero byte buffers or errors can be returned, the
			 * buffer will be dispossed of by the caller.
			 */
			conn->byte_strm += result;
		}
		else {

			if (0 > result) {

				TS_TRACE(MOD_LNX_SDP, T_VERBOSE,
					 TRACE_FLOW_WARN,
					 "EVENT: gateway buffer recv error. <%d>",
					 result);
			}
		}
	}
	else {

		result = 0;
	}
	/*
	 * PostRecv will take care of consuming this advertisment, based
	 * on result.
	 */
	return result;
advt_error:
	(void)sdp_advt_destroy(advt);
error:
	return result;
} /* _sdp_rcvd_snk_avail */

/* ========================================================================= */
/*.._sdp_rcvd_src_avail --  */
static int _sdp_rcvd_src_avail(struct sdp_opt *conn, struct sdpc_buff *buff)
{
	struct sdp_msg_hdr_srcah *srcah;
	struct sdpc_advt *advt;
	int result;
	tINT32 size;

	TS_CHECK_NULL(conn, -EINVAL);
	TS_CHECK_NULL(buff, -EINVAL);

	srcah = (struct sdp_msg_hdr_srcah *) buff->data;
	buff->data = buff->data + sizeof(struct sdp_msg_hdr_srcah);
	result = _sdp_msg_net_to_cpu_srcah(srcah);
	TS_EXPECT(MOD_LNX_SDP, !(0 > result));

	size = buff->tail - buff->data;

#ifdef _TS_SDP_DATA_PATH_DEBUG
	TS_TRACE(MOD_LNX_SDP, T_VERY_VERBOSE, TRACE_FLOW_INOUT,
		 "EVENT: SrcAvail received. <%d:%d:%d:%016llx> mode <%d>",
		 srcah->size, srcah->r_key, size, srcah->addr, conn->recv_mode);
#endif

	if (0 < conn->snk_sent) {
		/*
		 * crossed SrcAvail and SnkAvail, the source message is 
		 * discarded.
		 */
#ifdef _TS_SDP_DATA_PATH_DEBUG
		TS_TRACE(MOD_LNX_SDP, T_VERY_VERBOSE, TRACE_FLOW_WARN,
			 "EVENT: Dropping SrcAvail because <%d> SnkAvail in mode <%d>",
			 conn->snk_sent, conn->recv_mode);
#endif
		result = 0;
		goto done;
	}

	if (0 < (TS_SDP_CONN_F_SRC_CANCEL_R & conn->flags)) {

		TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_WARN,
			 "EVENT: SrcAvail during SrcAvailCancel processing. <%d>",
			 conn->src_recv);
		result = -EFAULT;
		goto done;
	}
	/*
	 * To emulate RFC 1122 (page 88) a connection should be reset/aborted
	 * if data is received and the receive half of the connection has been
	 * closed. This notifies the peer that the data was not received.
	 */
	if (0 < (TS_SDP_SHUTDOWN_RECV & conn->shutdown)) {

		TS_TRACE(MOD_LNX_SDP, T_VERY_VERBOSE, TRACE_FLOW_INOUT,
			 "EVENT: Received SrcAvail, receive path closed. <%02x:%04x>",
			 conn->shutdown, conn->state);
		/*
		 * abort connection (send reset)
		 */
		result = sdp_wall_abort(conn);
		TS_EXPECT(MOD_LNX_SDP, !(0 > result));
		/*
		 * drop packet
		 */
		result = 0;
		goto done;
	}
	/*
	 * save the advertisment
	 */
	advt = sdp_advt_create();
	if (NULL == advt) {

		TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_WARN,
			 "EVENT: SrcAvail cannot be copied.");
		result = -ENOMEM;
		goto done;
	}
	/*
	 * consume the advertisment, if it's allowed, first check the recv 
	 * path mode to determine if all is cool for the advertisment.
	 */
	switch (conn->recv_mode) {

	case TS_SDP_MODE_BUFF:

		TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_WARN,
			 "EVENT: SrcAvail message received in buffered mode.");

		result = -EPROTO;
		goto advt_error;

		break;
	case TS_SDP_MODE_COMB:

		if (0 < conn->src_recv || !(0 < size) || !(srcah->size > size)) {

			TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_WARN,
				 "EVENT: SrcAvail mode <%d> mismatch. <%d:%d:%d>",
				 conn->recv_mode, conn->src_recv, size,
				 srcah->size);

			result = -EPROTO;
			goto advt_error;
		}

		advt->rkey = srcah->r_key;
		advt->post =
		    0 - ((TS_SDP_SRC_AVAIL_THRESHOLD > size) ? size : 0);
        advt->size =
		    srcah->size -
		    ((TS_SDP_SRC_AVAIL_THRESHOLD > size) ? 0 : size);
		advt->addr =
		    srcah->addr +
		    ((TS_SDP_SRC_AVAIL_THRESHOLD > size) ? 0 : size);

		break;
	case TS_SDP_MODE_PIPE:

		if (TS_SDP_MSG_MAX_ADVS == (conn->src_recv + conn->snk_recv) ||
		    0 != size) {

			TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_WARN,
				 "EVENT: SrcAvail mode <%d> mismatch. <%d:%d>",
				 conn->recv_mode,
				 (conn->src_recv + conn->snk_recv), size);

			result = -EPROTO;
			goto advt_error;
		}

		advt->post = 0;

		advt->size = srcah->size;
		advt->addr = srcah->addr;
		advt->rkey = srcah->r_key;

		break;
	default:

		TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_WARN,
			 "EVENT: SrcAvail message in unknown mode. <%d>",
			 conn->recv_mode);
		result = -EPROTO;
		goto advt_error;
	}
	/*
	 * save advertisment
	 */
	conn->src_recv++;

	result = sdp_advt_q_put(&conn->src_pend, advt);
	if (0 > result) {

		TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_WARN,
			 "EVENT: SrcAvail cannot be saved. <%d>", result);
		goto advt_error;
	}
	/*
	 * process the ULP data in the message
	 */
	if (0 < size) {
		/*
		 * update non-discard for sink advertisment management
		 */
		conn->nond_recv++;

		if (!(TS_SDP_SRC_AVAIL_THRESHOLD > size)) {

			result = sdp_recv_buff(conn, buff);
			if (0 < result) {
				/*
				 * count number of bytes buffered by the
				 * connection, zero byte buffers or errors
				 * can be returned, the buffer will be
				 * dispossed of by the caller.
				 */
				conn->byte_strm += result;
			}
			else {

				if (0 > result) {

					TS_TRACE(MOD_LNX_SDP, T_VERBOSE,
						 TRACE_FLOW_WARN,
						 "EVENT: gateway buffer recv error. <%d>",
						 result);
				}
			}
		}
		else {

			result = 0;
		}
	}
	/*
	 * PostRecv will take care of consuming this advertisment.
	 */
	return result;

advt_error:
	(void)sdp_advt_destroy(advt);
done:
	return result;
} /* _sdp_rcvd_src_avail */

/* ========================================================================= */
/*.._sdp_rcvd_data -- SDP data message event received */
static int _sdp_rcvd_data(struct sdp_opt *conn, struct sdpc_buff *buff)
{
	s32 ret_val = 0;

	TS_CHECK_NULL(conn, -EINVAL);
	TS_CHECK_NULL(buff, -EINVAL);

	if (buff->tail > buff->data) {
		/*
		 * If we are processing a SrcAvail, there should be no
		 * buffered data
		 */
		if (0 < conn->src_recv) {

			TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_WARN,
				 "EVENT: Error, received buffered data, with SrcAvail active.");
			ret_val = -EPROTO;
			goto done;
		}
		/*
		 * check for out-of-band data, and mark the buffer if there
		 * is a pending urgent message. If the OOB data is in this
		 * buffer, pull it out.
		 */
		if (TS_SDP_MSG_HDR_GET_OOB_PEND(buff->bsdh_hdr)) {

			buff->flags |= TS_SDP_BUFF_F_OOB_PEND;
		}

		if (TS_SDP_MSG_HDR_GET_OOB_PRES(buff->bsdh_hdr)) {

			buff->flags |= TS_SDP_BUFF_F_OOB_PRES;
		}
		/*
		 * update non-discard for sink advertisment management
		 */
		conn->nond_recv++;

		ret_val = sdp_recv_buff(conn, buff);
		if (0 > ret_val) {

			TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_WARN,
				 "EVENT: gateway buffer recv error. <%d>",
				 ret_val);
		}
		/*
		 * result contains the number of bytes in the buffer which
		 * are being kept by the connection. (zero buffered means
		 * me can dispose of the buffer.
		 */
		conn->byte_strm += ret_val;
	}

done:
	return (int)ret_val;
} /* _sdp_rcvd_sata */

/* ========================================================================= */
/*.._sdp_rcvd_unsupported -- Valid messages we're not expecting */
static int _sdp_rcvd_unsupported(struct sdp_opt *conn, struct sdpc_buff *buff)
{
	TS_CHECK_NULL(conn, -EINVAL);
	TS_CHECK_NULL(buff, -EINVAL);
	/*
	 * Since the gateway only initates RDMA's but is never a target, and
	 * for a few other reasons, there are certain valid SDP messages
	 * which we never expect to see.
	 */
	TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_WARN,
		 "EVENT: Unexpected SDP message <%02x> received!",
		 buff->bsdh_hdr->mid);

	return 0;
} /* _sdp_rcvd_unsupported */

/*
 * Event Dispatch table. For performance a dispatch table is used to avoid
 * a giant case statment for every single SDP event. This is a bit more
 * confusing, relies on the layout of the Message IDs, and is less
 * flexable. However, it is faster.
 *
 * Sparse table, the full table would be 16x16, where the low 4 bits, of
 * the MID byte, are one dimension, and the high 4 bits are the other
 * dimension. Since all rows, except for the first and last, are empty,
 * only those are represented in the table.
 */
#define TS_SDP_MSG_EVENT_TABLE_SIZE 0x20

static tGW_SDP_EVENT_CB_FUNC recv_event_funcs[TS_SDP_MSG_EVENT_TABLE_SIZE] = {
	NULL,                       /* TS_SDP_MID_HELLO            0x00 */
	NULL,                       /* TS_SDP_MID_HELLO_ACK        0x01 */
	_sdp_rcvd_disconnect,       /* TS_SDP_MID_DISCONNECT       0x02 */
	_sdp_rcvd_abort,            /* TS_SDP_MID_ABORT_CONN       0x03 */
	_sdp_rcvd_send_sm,          /* TS_SDP_MID_SEND_SM          0x04 */
	_sdp_rcvd_rdma_wr,          /* TS_SDP_MID_RDMA_WR_COMP     0x05 */
	_sdp_rcvd_rdma_rd,          /* TS_SDP_MID_RDMA_RD_COMP     0x06 */
	_sdp_rcvd_mode_change,      /* TS_SDP_MID_MODE_CHANGE      0x07 */
	_sdp_rcvd_src_cancel,       /* TS_SDP_MID_SRC_CANCEL       0x08 */
	_sdp_rcvd_snk_cancel,       /* TS_SDP_MID_SNK_CANCEL       0x09 */
	_sdp_rcvd_snk_cancel_ack,   /* TS_SDP_MID_SNK_CANCEL_ACK   0x0A */
	_sdp_rcvd_resize_buff_ack,  /* TS_SDP_MID_CH_RECV_BUF      0x0B */
	_sdp_rcvd_unsupported,      /* TS_SDP_MID_CH_RECV_BUF_ACK  0x0C */
	_sdp_rcvd_suspend,          /* TS_SDP_MID_SUSPEND          0x0D */
	_sdp_rcvd_suspend_Ack,      /* TS_SDP_MID_SUSPEND_ACK      0x0E */
	NULL,                       /* reserved                    0x0F */
	NULL,                       /* reserved                    0xF0 */
	NULL,                       /* reserved                    0xF1 */
	NULL,                       /* reserved                    0xF2 */
	NULL,                       /* reserved                    0xF3 */
	NULL,                       /* reserved                    0xF4 */
	NULL,                       /* reserved                    0xF5 */
	NULL,                       /* reserved                    0xF6 */
	NULL,                       /* reserved                    0xF7 */
	NULL,                       /* reserved                    0xF8 */
	NULL,                       /* reserved                    0xF9 */
	NULL,                       /* reserved                    0xFA */
	NULL,                       /* reserved                    0xFB */
	NULL,                       /* reserved                    0xFC */
	_sdp_rcvd_snk_avail,        /* TS_SDP_MID_SNK_AVAIL        0xFD */
	_sdp_rcvd_src_avail,        /* TS_SDP_MID_SRC_AVAIL        0xFE */
	_sdp_rcvd_data              /* TS_SDP_MID_DATA             0xFF */
}; /* recv_event_funcs */

/* ========================================================================= */
/*..sdp_event_recv -- recv event demultiplexing into sdp messages. */
tINT32 sdp_event_recv(struct sdp_opt *conn, struct ib_cq_entry *comp)
{
	tGW_SDP_EVENT_CB_FUNC dispatch_func;
	struct sdpc_buff *buff;
	u32 offset;
	int result;

	TS_CHECK_NULL(conn, -EINVAL);
	TS_CHECK_NULL(comp, -EINVAL);
	/*
	 * error handling
	 */
	if (IB_COMPLETION_STATUS_SUCCESS != comp->status) {

		switch (comp->status) {
		case IB_COMPLETION_STATUS_WORK_REQUEST_FLUSHED_ERROR:
			/*
			 * clear posted buffers from error'd queue
			 */
			result = sdp_buff_q_clear(&conn->recv_post);
			TS_EXPECT(MOD_LNX_SDP, !(0 > result));

			result = 0;
			break;
		default:
			TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_WARN,
				 "EVENT: <%04x> unhandled RECV error status <%d>.",
				 conn->hashent, comp->status);
			result = -EIO;
		}

		goto done;
	}
	/*
	 * get data
	 */
	buff = sdp_buff_q_get_head(&conn->recv_post);
	if (NULL == buff) {

		TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_WARN,
			 "EVENT: receive event, but no posted receive?!");
		result = -EINVAL;
		goto done;
	}
	if (comp->work_request_id != buff->ib_wrid) {

		TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_WARN,
			 "EVENT: completion, work request ID mismatch. <%u:%u>",
			 comp->work_request_id, buff->ib_wrid);

		result = -ERANGE;
		goto drop;
	}
	/*
	 * endian swap
	 */
	conn->l_recv_bf--;
	conn->l_advt_bf--;

	buff->bsdh_hdr = (struct sdp_msg_hdr_bsdh *) buff->data;

	result = _sdp_msg_net_to_cpu_bsdh(buff->bsdh_hdr);
	TS_EXPECT(MOD_LNX_SDP, !(0 > result));

	if (comp->bytes_transferred != buff->bsdh_hdr->size) {

		TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_WARN,
			 "EVENT: receive completion, message size mismatch <%d:%d>",
			 comp->bytes_transferred, buff->bsdh_hdr->size);

		result = -EINVAL;
		goto drop;
	}

	buff->tail = buff->data + buff->bsdh_hdr->size;
	buff->data = buff->data + sizeof(struct sdp_msg_hdr_bsdh);
	/*
	 * Do not update the advertised sequence number, until the 
	 * SrcAvailCancel message has been processed.
	 */
	conn->recv_seq = buff->bsdh_hdr->seq_num;
	conn->advt_seq = (((TS_SDP_CONN_F_SRC_CANCEL_R & conn->flags) > 0) ?
			  conn->advt_seq : conn->recv_seq);
	/*
	 * buffers advertised minus the difference in buffer count between
	 * the number we've sent and the remote host has received.
	 */
	conn->r_recv_bf = (buff->bsdh_hdr->recv_bufs -
			   abs((tINT32) conn->send_seq -
			       (tINT32) buff->bsdh_hdr->seq_ack));
	/*
	 * dispatch
	 */
#ifdef _TS_SDP_DATA_PATH_DEBUG
	TS_TRACE(MOD_LNX_SDP, T_VERY_VERBOSE, TRACE_FLOW_INOUT,
		 "EVENT: recv BSDH <%04x:%02x:%02x:%08x:%08x:%08x>",
		 buff->bsdh_hdr->recv_bufs,
		 buff->bsdh_hdr->flags,
		 buff->bsdh_hdr->mid,
		 buff->bsdh_hdr->size,
		 buff->bsdh_hdr->seq_num, buff->bsdh_hdr->seq_ack);
#endif
	/*
	 * fast path data messages
	 */
	if (TS_SDP_MSG_MID_DATA == buff->bsdh_hdr->mid) {

		result = _sdp_rcvd_data(conn, buff);
	}
	else {

		offset = buff->bsdh_hdr->mid & 0x1F;

		if (!(offset < TS_SDP_MSG_EVENT_TABLE_SIZE) ||
		    NULL == recv_event_funcs[offset]) {

			TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_WARN,
				 "EVENT: receive completion, unknown message ID <%d>",
				 buff->bsdh_hdr->mid);
			result = -EINVAL;
			goto drop;
		}

		TS_SDP_CONN_STAT_RECV_MID_INC(conn, offset);
		TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_WARN,
				 "EVENT: received opcode  <%d>",
				 offset);
		dispatch_func = recv_event_funcs[offset];
		result = dispatch_func(conn, buff);
	}
	/*
	 * process result.
	 */
	if (0 == result) {

		result = sdp_buff_pool_put(buff);
		TS_EXPECT(MOD_LNX_SDP, !(0 > result));
		/*
		 * If this buffer was consumed, then make sure sufficient
		 * recv buffers are posted. Also we might be able to move
		 * data with a new RDMA SrcAvail advertisment.
		 */
		result = sdp_recv_flush(conn);
		if (0 > result) {

			TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_WARN,
				 "EVENT: Receive buffer post failed freed message. <%d>",
				 result);
			goto done;
		}
	}
	else {

		if (0 > result) {

			TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_WARN,
				 "EVENT: receive completion, dispatch error. <%d>",
				 result);

			goto drop;
		}
		else {
			/*
			 * If data was consumed by the protocol, signal
			 * the user.
			 */
			sdp_inet_wake_recv(conn->sk, conn->byte_strm);
		}
	}
	/*
	 * It's possible that a new recv buffer advertisment opened up the 
	 * recv window and we can flush buffered send data
	 */
	result = sdp_send_flush(conn);
	if (0 > result) {

		TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_WARN,
			 "EVENT: Flush failure during receive completion. <%d>",
			 result);
		goto done;
	}

	return 0;
drop:
	(void)sdp_buff_pool_put(buff);
done:
	return result;
} /* sdp_event_recv */
