/*
  This software is available to you under a choice of one of two
  licenses.  You may choose to be licensed under the terms of the GNU
  General Public License (GPL) Version 2, available at
  <http://www.fsf.org/copyleft/gpl.html>, or the OpenIB.org BSD
  license, available in the LICENSE.TXT file accompanying this
  software.  These details are also available at
  <http://openib.org/license.html>.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
  BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
  ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
  CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
  SOFTWARE.

  Copyright (c) 2004 Topspin Communications.  All rights reserved.
  Copyright (c) 2005 Mellanox Technologies.   All rights reserved.


  $Id: hostoptions.c 405 2004-07-10 02:00:39Z roland $
*/

#include "srp_host.h"

#define kWWNLength 8
#define kWWNStringLength (kWWNLength * 2)
#define kDLIDLength 4
#define kTargetIndexStringLength (kDLIDLength * 2)


#define kLoadError 1
#define kNoError 0
#define TS_FAILURE -1
#define TS_SUCCESS 0

static char cfilename[80] = "/etc/mlx_srp_bindings.conf";

void ConvertToLowerCase(char *stringPtr)
{
	char tempChar;

	tempChar = *stringPtr;

	while (tempChar != 0) {
		if ((tempChar >= 'A') && (tempChar <= 'Z')) {
			tempChar += 'a' - 'A';
			*stringPtr = tempChar;
		}

		stringPtr++;
		tempChar = *stringPtr;
	}
}

/*
 *  This routine copies all the characters from the source string into the dest
 * string up to but not including a delimiter.  It returns the number of
 * characters which I have processed in the source string (including the
 * delimiter character if it exists).  If the delimiter does not exist in the
 * sourc string, the entire source string is copied to the destination string.
 * If the destination string is too small to hold the characters which need
 * to be copied, then the destination string is filled and the count returned
 * is the number of characters copied to the destination string.
 */
void
GetString(char *sourceString, char *destString,
	  unsigned long destStringLength, char delimiter,
	  unsigned long *sourceCharsUsedPtr)
{
	unsigned long index;

	for (index = 0; index < destStringLength; index++) {
		destString[index] = 0;
	}

	index = 0;

	/*  Copy the string fragment into the destination string. */
	while ((sourceString[index] != 0) && (sourceString[index] != delimiter)
	       && (index < destStringLength)) {
		destString[index] = sourceString[index];

		index++;
	}

	/*
	 * If the string fragment was too big for the destination string,
	 * return a count of only what fits.
	 */
	if (sourceString[index] == delimiter) {
		index++;
	} else if (index == destStringLength) {
		index--;
		destString[index] = 0;
	}

	*sourceCharsUsedPtr = index;
}

int StringToHex32(char *stringPtr, u32 *hexptr)
{
	int firsttime = 1;
	long isError = kNoError;
	char *s = stringPtr;
	unsigned char Nibble = 0;

	*hexptr = 0;

	if (*s == '0' && *(s + 1) == 'x')
		s += 2;

	while (*s != '\0') {
		if ((*s >= '0') && (*s <= '9'))
			Nibble = *s - '0';
		else if ((*s >= 'a') && (*s <= 'f'))
			Nibble = *s - 'a' + 10;
		else
			isError = kLoadError;

		if (isError == kNoError) {
			if (firsttime)
				*hexptr = 0;
			else
				*hexptr <<= 4;;

			*hexptr |= Nibble;
		} else
			break;

		s++;
		firsttime = 0;
	}
	return isError;
}

int StringToHex64(char *stringPtr, u64 * hexptr)
{
	int firsttime = 1;
	long isError = kNoError;
	char *s = stringPtr;
	unsigned char Nibble = 0;

	*hexptr = 0;

	if (*s == '0' && *(s + 1) == 'x')
		s += 2;

	while (*s != '\0') {
		if ((*s >= '0') && (*s <= '9'))
			Nibble = *s - '0';
		else if ((*s >= 'a') && (*s <= 'f'))
			Nibble = *s - 'a' + 10;
		else if ((*s >= 'A') && (*s <= 'F'))
			Nibble = *s - 'A' + 10;
		else
			isError = kLoadError;

		if (isError == kNoError) {
			if (firsttime)
				*hexptr = 0;
			else
				*hexptr <<= 4;;

			*hexptr |= Nibble;
		} else
			break;

		s++;
		firsttime = 0;
	}
	return isError;
}

u32 parse_target_binding_parameters(char *parameters)
{
	char *curr_loc;
	unsigned long chars_copied = 0;
	char delimeter;
	char hex64str[kWWNStringLength + 1];
	char target_index_str[kTargetIndexStringLength + 1];
	u64 wwn, service_id;
	unsigned int target_index, value;
	unsigned int pidx, hidx, dlid;
	srp_target_t *target;
	int status;
	long result;
	u32 i, j;
	struct ib_path_record path_record;
	u64 tid;

	/* first convert to lower case to make life easier */
	ConvertToLowerCase(parameters);

	curr_loc = parameters;
	i = pidx = hidx = dlid = target_index = 0;

	while ((*curr_loc != 0) && (i < max_srp_targets)) {

		printk("Target Binding %d ", i);

		/* first get the wwn as a string */
		delimeter = '.';
		for (j = 0; j < 2; j++) {
			memset(hex64str, 0, kWWNStringLength + 1);

			GetString(curr_loc, hex64str, kWWNStringLength, delimeter,
				  &chars_copied);

			if (chars_copied > (kWWNStringLength + 1)) {
				printk("Exit here %d\n",__LINE__);
				return (TS_FAILURE);
			} else {
				curr_loc += chars_copied;
			}
		
			if (j)
				result = StringToHex64(hex64str, &wwn);
			else
				result = StringToHex64(hex64str, &service_id);

			if (result != kNoError) {
				printk("Exit here %d\n",__LINE__);
				goto err;
			}
		}

		for (j = 0; j < 4; j++) {
			memset(target_index_str, 0, kTargetIndexStringLength + 1);

			if (j == 3)
				delimeter = ':';

			GetString(curr_loc, target_index_str,
				  kTargetIndexStringLength,
				  delimeter, &chars_copied);

			if (chars_copied > (kTargetIndexStringLength + 1)) {
				printk("Exit here %d\n",__LINE__);
				goto err;
			} else {
				curr_loc += chars_copied;
			}

			result = StringToHex32(target_index_str, &value);

			switch (j) {
			case 0:
				dlid = value;
				break;
			case 1:
				target_index = value;
				break;
			case 2:
				pidx = value;
				break;
			case 3:
				hidx = value;
				break;
			default:
				break;
			}
		
			if (result != kNoError) {
				printk("Exit here %d\n",__LINE__);
				goto err;
			}

		}

		if ((target_index > max_srp_targets) || 
		    !pidx || !hidx || (pidx > MAX_LOCAL_PORTS_PER_HCA) ||
		    (hidx > MAX_HCAS)) {
				printk("Exit here %d %d %d %d %d\n",__LINE__, target_index, pidx, hidx, *curr_loc);
				goto err;
		}

		target = &hca_params[hidx - 1].srp_targets[target_index];

		printk("Binding Target 0x%lx.0x%lx.%x.%d.%d.%d\n",
		       (unsigned long)wwn, (unsigned long)service_id, dlid,
		       target_index, pidx, hidx);
		
		*(u64 *) & (target->service_name) = cpu_to_be64(wwn);
		target->port = &hca_params[hidx - 1].port[pidx - 1];
		target->state = TARGET_POTENTIAL_CONNECTION;
		target->target_index = target_index;

		status = ib_get_path_record_from_lids(target->port->hca->ca_hndl,
						      target->port->local_port,
						      target->port->slid,
						      dlid,
						      20 * HZ,
						      srp_path_query_wakeup,
						      (void *)&path_record,
						      &tid);

		if (!status) {
			down(&driver_params.path_query_sema);
		}
		
		srp_update_target_ioc(target, target->port, dlid,
				      cpu_to_be64(service_id), &path_record);

		status = srp_host_alloc_pkts(target);
		if (status) {
			TS_REPORT_FATAL(MOD_SRPTP,
					"Target %d, packet allocation failure");
			target->valid = FALSE;
			goto err;
		} else {
			target->valid = TRUE;
			target->port->new_tgt_insvc++;
			target->rescan_scsi = TRUE;
		}
		i++;
	}

	return (TS_SUCCESS);

      err:
	return (TS_FAILURE);
}

int get_target_bindings(void)
{
	struct file *cf;
	u32 size = 0;
	char c, *buf = NULL;
	int err = 0;
	mm_segment_t fs;

	fs = get_fs();
	set_fs(get_ds());
	
	cf = filp_open(cfilename, O_RDONLY | O_CREAT, 0600);

	if (!cf || !cf->f_op || !cf->f_op->read) {
		printk("failed to open %s file\n", cfilename);
		err = -EINVAL;
		goto out;
	}

	while(vfs_read(cf, &c, 1, &cf->f_pos) > 0)
		size++;

	if (!size) {
		printk("File %s has 0 byte\n", cfilename);
		err = -ENOMEM;
		goto close_out;
	} 

	buf = kmalloc(size, GFP_KERNEL);
	if (!buf) {
		printk("failed to kmalloc buffer to read %s\n", cfilename);
		err = -ENOMEM;
		goto close_out;
	}

	printk("kmalloc %d-byte buffer to read %s\n", size, cfilename);

	vfs_llseek(cf, 0, 0);

	if (vfs_read(cf, buf, size, &cf->f_pos) > 0) {
		buf[size - 1] = 0;
		err = parse_target_binding_parameters(buf);
	} else
		printk("failed to read %d bytes from file %s\n",
		       size, cfilename);
		
	kfree(buf);

       close_out:
	fput(cf);
       out:
	set_fs(fs);
	return err;
}

void print_target_bindings(void)
{
	srp_target_t *target;
	int not_first_entry = FALSE;
	srp_host_hca_params_t *hca;

	printk("target_bindings=");

	for (hca = &hca_params[0];
	     hca < &hca_params[MAX_HCAS]; hca++) {
		if (hca->valid == FALSE)
			continue;

		for (target = &hca->srp_targets[0];
		     target < &hca->srp_targets[max_srp_targets];
		     target++) {

			if (target->valid == FALSE)
				continue;

			/* don't print colon on first guy */
			if (not_first_entry == TRUE) {
				printk(":");
			} else {
				not_first_entry = TRUE;
			}

			printk("%lx.%lx.%x.%d.%d.%d",
			       (unsigned long)cpu_to_be64(target->
							  service_name),
			       (unsigned long)cpu_to_be64(target->ioc.
							  service_id),
			       target->ioc.iou_path_record.dlid,
	       		       target->target_index, target->port->local_port,
			       hca->hca_index + 1);
		}
	}
	printk("\n");
}
