/*
  This software is available to you under a choice of one of two
  licenses.  You may choose to be licensed under the terms of the GNU
  General Public License (GPL) Version 2, available at
  <http://www.fsf.org/copyleft/gpl.html>, or the OpenIB.org BSD
  license, available in the LICENSE.TXT file accompanying this
  software.  These details are also available at
  <http://openib.org/license.html>.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
  BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
  ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
  CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
  SOFTWARE.

  Copyright (c) 2004 Topspin Communications.  All rights reserved.

  $Id: client_query_main.c 477 2004-07-16 16:40:34Z roland $
*/

#include "client_query.h"
#include "ts_ib_mad.h"
#include "ts_kernel_trace.h"
#include "openib_version.h"

#include <linux/version.h>
#include <linux/module.h>
#include <linux/init.h>
#include <linux/errno.h>
#include <linux/kernel.h>
#include <linux/param.h>
#include <asm/system.h>
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
#include <linux/moduleparam.h>
#endif

		      
MODULE_AUTHOR("Carl Yang");
MODULE_DESCRIPTION("kernel IB client query service" OPENIB_REVISION);
MODULE_LICENSE("Dual BSD/GPL");

/////Module Param/////////////
int ib_query_throttle_threshhold = 0;
int ib_query_throttle_waitqueue_limit = 0;

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
module_param(ib_query_throttle_threshhold, int, 0444);
module_param(ib_query_throttle_waitqueue_limit, int, 0444);
#else
MODULE_PARM(ib_query_throttle_threshhold,"i");
MODULE_PARM(ib_query_throttle_waitqueue_limit,"i");
#endif
MODULE_PARM_DESC(ib_query_throttle_threshhold,"The max number of outstanding ib queries allowed. 0=unlimited");
MODULE_PARM_DESC(ib_query_throttle_waitqueue_limit,"The max number of queued ib queries - more queries cause timeout returns. 0=unlimited");
/////End of Module Param//////

enum {
	TS_IB_CLIENT_MAX_MGMT_CLASS = 1 << 8
};

static struct {
	ib_mad_dispatch_func function;
	void *arg;
	tTS_IB_MAD_FILTER_HANDLE filter;
} async_mad_table[TS_IB_CLIENT_MAX_MGMT_CLASS];

int ib_client_mad_handler_register(u8 mgmt_class,
				   ib_mad_dispatch_func function,
				   void *arg)
{
	int ret;

	if (function) {
		struct ib_mad_filter filter = { 0 };

		filter.qpn = 1;
		filter.mgmt_class = mgmt_class;
		filter.direction = TS_IB_MAD_DIRECTION_IN;
		filter.mask =
		    TS_IB_MAD_FILTER_QPN | TS_IB_MAD_FILTER_MGMT_CLASS |
		    TS_IB_MAD_FILTER_DIRECTION;
		snprintf(filter.name, sizeof filter.name,
			 "query client (class 0x%02x)", mgmt_class);

		ret = tsIbMadHandlerRegister(&filter,
					     ib_client_mad_handler,
					     NULL,
					     &async_mad_table[mgmt_class].
					     filter);
		if (ret) {
			TS_REPORT_WARN(MOD_KERNEL_IB,
				       "Failed to register MAD filter");
			return ret;
		}
	} else {
		tsIbMadHandlerDeregister(async_mad_table[mgmt_class].filter);
	}

	async_mad_table[mgmt_class].function = function;
	async_mad_table[mgmt_class].arg = arg;

	return 0;
}
EXPORT_SYMBOL(ib_client_mad_handler_register);

ib_mad_dispatch_func ib_client_async_mad_handler_get(u8 mgmt_class)
{
	return async_mad_table[mgmt_class].function;
}

void *ib_client_async_mad_handler_arg_get(u8 mgmt_class)
{
	return async_mad_table[mgmt_class].arg;
}

static int __init ib_client_query_init_module(void)
{
	if (ib_client_query_init()) {
		return -EINVAL;
	}

	TS_TRACE(MOD_KERNEL_IB, T_VERBOSE, TRACE_KERNEL_IB_GEN, 
		 "throttle threshold  = %d",ib_query_throttle_threshhold );
	return 0;
}

static void __exit ib_client_query_cleanup_module(void)
{
	ib_client_query_cleanup();
}

module_init(ib_client_query_init_module);
module_exit(ib_client_query_cleanup_module);

