/*
 * This software is available to you under a choice of one of two
 * licenses.  You may choose to be licensed under the terms of the GNU
 * General Public License (GPL) Version 2, available at
 * <http://www.fsf.org/copyleft/gpl.html>, or the OpenIB.org BSD
 * license, available in the LICENSE.TXT file accompanying this
 * software.  These details are also available at
 * <http://openib.org/license.html>.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 * Copyright (c) 2004 Topspin Communications.  All rights reserved.
 *
 * $Id: core_mw.c 475 2004-07-16 16:29:15Z roland $
 */

#include <linux/errno.h>
#include <linux/slab.h>
#include <linux/module.h>

#include "core_priv.h"

#include "ts_kernel_trace.h"
#include "ts_kernel_services.h"

int ib_mw_create(struct ib_pd     *pd,
		 struct ib_mw    **mw,
		 u32              *rkey)
{
	int ret;

	IB_CHECK_MAGIC(pd, PD);

	if (!pd->device->mw_create)
		return -ENOSYS;

	*mw = kmalloc(sizeof **mw, GFP_KERNEL);
	if (!*mw) {
		TS_REPORT_WARN(MOD_KERNEL_IB,"mw struct kmalloc failure");
		return -ENOMEM;
        }

	(*mw)->device = pd->device;
	(*mw)->pd     = pd;

	ret = pd->device->mw_create(pd, mw, &(*mw)->rkey);
	if (ret) {
		kfree(mw);
		return ret;
	}

	IB_SET_MAGIC(*mw, MW);
	*rkey = (*mw)->rkey;
	
	return 0;
}
EXPORT_SYMBOL(ib_mw_create);

int ib_mw_destroy(struct ib_mw *mw)
{
	int ret;

	IB_CHECK_MAGIC(mw, MW);

	if (!mw->device->mw_destroy)
		return -ENOSYS;

	ret = mw->device->mw_destroy(mw);
	if (ret)
		return ret;

	IB_CLEAR_MAGIC(mw);
	kfree(mw);

	return 0;
}
EXPORT_SYMBOL(ib_mw_destroy);

int ib_mw_bind(struct ib_qp            *qp,
	       struct ib_mw            *mw,
	       struct ib_mw_bind_param *param,
	       u32                     *new_rkey)
{
	IB_CHECK_MAGIC(qp, QP);
	IB_CHECK_MAGIC(mw, MW);

	if (qp->device != mw->device)
		return -EINVAL;

	return mw->device->mw_bind ?
		mw->device->mw_bind(qp, mw, param, new_rkey) : -ENOSYS;
}
EXPORT_SYMBOL(ib_mw_bind);

/*
 * Local Variables:
 * c-file-style: "linux"
 * indent-tabs-mode: t
 * End:
 */
