/*
  This software is available to you under a choice of one of two
  licenses.  You may choose to be licensed under the terms of the GNU
  General Public License (GPL) Version 2, available at
  <http://www.fsf.org/copyleft/gpl.html>, or the OpenIB.org BSD
  license, available in the LICENSE.TXT file accompanying this
  software.  These details are also available at
  <http://openib.org/license.html>.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
  BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
  ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
  CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
  SOFTWARE.

  Copyright (c) 2004 Topspin Communications.  All rights reserved.

  $Id: core_pd.c 475 2004-07-16 16:29:15Z roland $
*/

#include "core_priv.h"

#include "ts_kernel_trace.h"
#include "ts_kernel_services.h"

#include <linux/version.h>
#include <linux/module.h>

#include <linux/errno.h>
#include <linux/slab.h>

int __ib_pd_create(struct ib_device    *device,
		   void                *device_specific,
		   struct ib_pd       **pd_handle)
{
	struct ib_pd     *pd;
	int               ret;

	IB_CHECK_MAGIC(device, DEVICE);

	if (!device->pd_create)
		return -ENOSYS;

	pd = kmalloc(sizeof *pd, GFP_KERNEL);
	if (!pd) {
		TS_REPORT_WARN(MOD_KERNEL_IB,"pd struct kmalloc failure");
		return -ENOMEM;
	}

	ret = device->pd_create(device, device_specific, pd);

	if (!ret) {
		IB_SET_MAGIC(pd, PD);
		pd->device = device;
		*pd_handle = pd;
	} else {
		kfree(pd);
	}

	return ret;
}
EXPORT_SYMBOL(__ib_pd_create);

int __ib_pd_destroy(struct ib_pd *pd)
{
	int           ret;

	IB_CHECK_MAGIC(pd, PD);

	if (!pd->device->pd_destroy)
		return -ENOSYS;

	ret = pd->device->pd_destroy(pd);
	if (!ret) {
		IB_CLEAR_MAGIC(pd);
		kfree(pd);
	}

	return ret;
}
EXPORT_SYMBOL(__ib_pd_destroy);

int ib_pd_create(struct ib_device    *device,
                 void                *device_specific,
                 struct ib_pd       **pd_handle)
{
	IB_CHECK_MAGIC(device, DEVICE);

	if (!try_module_get(device->owner))
		return -ENODEV;

	return __ib_pd_create(device, device_specific, pd_handle);
}
EXPORT_SYMBOL(ib_pd_create);

int ib_pd_destroy(struct ib_pd *pd)
{
	struct module *owner;
	int            ret;

	IB_CHECK_MAGIC(pd, PD);

	owner = pd->device->owner;

	ret = __ib_pd_destroy(pd);
	if (!ret)
		module_put(owner);

	return ret;
}
EXPORT_SYMBOL(ib_pd_destroy);

/*
  Local Variables:
  c-file-style: "linux"
  indent-tabs-mode: t
  End:
*/
