/*
  This software is available to you under a choice of one of two
  licenses.  You may choose to be licensed under the terms of the GNU
  General Public License (GPL) Version 2, available at
  <http://www.fsf.org/copyleft/gpl.html>, or the OpenIB.org BSD
  license, available in the LICENSE.TXT file accompanying this
  software.  These details are also available at
  <http://openib.org/license.html>.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
  BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
  ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
  CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
  SOFTWARE.

  Copyright (c) 2004 Topspin Communications.  All rights reserved.

  $Id: dm_client_main.c 479 2004-07-16 18:01:50Z roland $
*/

#include "dm_client.h"

#include "ts_ib_mad.h"
#include "ts_ib_dm_client.h"
#include "ts_ib_dm_client_host.h"
#include "ts_ib_client_query.h"

#include "ts_kernel_trace.h"
#include "openib_version.h"

#include <linux/version.h>
#include <linux/module.h>
#include <linux/init.h>
#include <linux/errno.h>
#include <linux/kernel.h>
#include <linux/param.h>
#include <asm/system.h>
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
#include <linux/moduleparam.h>
#endif

int use_port_info_tbl = 1;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
module_param(use_port_info_tbl, int, 0444);
#else
MODULE_PARM(use_port_info_tbl, "i");
#endif

MODULE_AUTHOR("Carl Yang");
MODULE_DESCRIPTION("kernel IB Device Managment client" OPENIB_REVISION);
MODULE_LICENSE("Dual BSD/GPL");

static int __init ib_dm_client_init_module(void)
{
	if (ib_dm_client_query_init()) {
		return -EINVAL;
	}

	ib_client_mad_handler_register(IB_MGMT_CLASS_DEV_MGT,
				     ib_dm_async_notify_handler, NULL);

	return 0;
}

static void __exit ib_dm_client_cleanup_module(void)
{
	ib_client_mad_handler_register(IB_MGMT_CLASS_DEV_MGT, NULL, NULL);

	ib_dm_client_query_cleanup();
}

module_init(ib_dm_client_init_module);
module_exit(ib_dm_client_cleanup_module);
