/*
  This software is available to you under a choice of one of two
  licenses.  You may choose to be licensed under the terms of the GNU
  General Public License (GPL) Version 2, available at
  <http://www.fsf.org/copyleft/gpl.html>, or the OpenIB.org BSD
  license, available in the LICENSE.TXT file accompanying this
  software.  These details are also available at
  <http://openib.org/license.html>.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
  BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
  ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
  CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
  SOFTWARE.

  Copyright (c) 2004 Topspin Communications.  All rights reserved.

  $Id: dm_client_query.c 479 2004-07-16 18:01:50Z roland $
*/

#include "dm_client.h"

#include "ts_ib_mad.h"
#include "ts_ib_client_query.h"

#include "ts_kernel_trace.h"
#include "ts_kernel_hash.h"
#include "ts_kernel_timer.h"

#include <linux/list.h>
#include <linux/slab.h>
#include <linux/random.h>
#include <linux/spinlock.h>
#include "dm_client_priv.h"

LIST_HEAD(dm_outs_query_q);
spinlock_t        dm_outs_query_lock = SPIN_LOCK_UNLOCKED;

extern void ib_client_query_set_in_cleanup_wait(void);

void ib_dm_client_mad_init(struct ib_mad *packet, struct ib_device *device,
			   tTS_IB_PORT port, u16 dst_lid, u32 dst_qpn,
			   u32 r_method, u16 attribute_id,
			   u32 attribute_modifier)
{
	memset(packet, 0, sizeof *packet);

	packet->format_version = 1;
	packet->mgmt_class = IB_MGMT_CLASS_DEV_MGT;
	packet->class_version = TS_IB_DM_CLASS_VERSION;
	packet->transaction_id = ib_client_alloc_tid();

	packet->device = device;
	packet->pkey_index = 0;
	packet->port = port;
	packet->slid = 0xffff;
	packet->dlid = dst_lid;
	packet->sl = 0;
	packet->sqpn = IB_GSI_QP;
	packet->dqpn = dst_qpn;
	packet->r_method = r_method;
	packet->attribute_id = cpu_to_be16(attribute_id);
	packet->attribute_modifier = cpu_to_be32(attribute_modifier);

	packet->has_grh = 0;

	packet->completion_func = NULL;

}

int ib_dm_client_query_init(void)
{
	return 0;
}

void ib_dm_client_query_cleanup(void)
{
	/* cancel all queries (which also deletes outstanding timers) */
	struct list_head *ptr;
	struct list_head *tmp;
	struct ib_dm_query_hdr *entry;

	ib_client_query_set_in_cleanup_wait();
	list_for_each_safe(ptr, tmp, &dm_outs_query_q) {
		entry = list_entry(ptr, struct ib_dm_query_hdr, outs_query);
		ib_client_query_cancel(entry->transaction_id);
	}
}
