/*
  This software is available to you under a choice of one of two
  licenses.  You may choose to be licensed under the terms of the GNU
  General Public License (GPL) Version 2, available at
  <http://www.fsf.org/copyleft/gpl.html>, or the OpenIB.org BSD
  license, available in the LICENSE.TXT file accompanying this
  software.  These details are also available at
  <http://openib.org/license.html>.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
  BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
  ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
  CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
  SOFTWARE.

  Copyright (c) 2004 Topspin Communications.  All rights reserved.

  $Id: sa_client_port_info.c 477 2004-07-16 16:40:34Z roland $
*/

#include "sa_client.h"

#include "ts_ib_mad.h"
#include "ts_ib_rmpp_mad_types.h"

#include "ts_kernel_trace.h"

#include <linux/kernel.h>
#include <linux/slab.h>
#include <linux/errno.h>
#include <asm/system.h>
#include <asm/byteorder.h>
#include "sa_client_priv.h"

struct ib_sa_port_info {
	u16 port_lid;
	tTS_IB_PORT port_num;
	u8 reserved;
	tTS_IB_MKEY mkey;
	u8 gid_prefix[8];
	u16 lid;
	u16 master_sm_lid;
	u32 capability_mask;
	u16 diag_code;
	u16 mkey_lease_period;
	u8 local_port_num;
	u8 link_width_enable;
	u8 link_width_supported;
	u8 link_width_active;
	u8 link_speed_supported__port_state;
	u8 port_phy_state__link_down_default_state;
	u8 mkey_protect_bits__reserve__lmc;
	u8 link_speed_active__linke_speed_enabled;
	u8 nighbor_mtu__master_smsl;
	u8 vlcap__reserved;
	u8 vlhigh_limit;
	u8 vl_arbitration_high_cap;
	u8 vl_arbitration_low_cap;
	u8 reserve__mtu_cap;
	u8 vl_stall_count__hoqlife;
	u8 operational_vl__pin__pout__fin__fout;
	u16 mkey_violation;
	u16 pkey_violation;
	u16 qkey_violation;
	u8 guid_cap;
	u8 reserve__subnet_timeout;
	u8 reserve__resp_time_value;
	u8 local_phy_error__overrun_error;
} __attribute__ ((packed));

struct ib_sa_port_info_query {
	struct ib_sa_query_hdr hdr;
	tTS_IB_PORT_INFO_QUERY_COMPLETION_FUNC completion_func;
	void *completion_arg;
};

#define ORACLE_DEMO_LANE15_HACK

static void _tsIbPortInfoResponse(tTS_IB_CLIENT_RESPONSE_STATUS status,
				  struct ib_mad *packet, void *query_ptr)
{
	struct ib_sa_port_info_query *query = query_ptr;

	switch (status) {
	case TS_IB_CLIENT_RESPONSE_OK:
		{
			struct ib_sa_payload *sa_payload =
			    (struct ib_sa_payload *) & packet->payload;
			struct ib_sa_port_info *port_info_ptr =
			    (struct ib_sa_port_info *) sa_payload->admin_data;
			struct ib_port_info port_info;

			memcpy(port_info.mkey, port_info_ptr->mkey, 8);
			memcpy(port_info.gid_prefix, port_info_ptr->gid_prefix,
			       8);
			port_info.lid = be16_to_cpu(port_info_ptr->lid);
			port_info.master_sm_lid =
			    be16_to_cpu(port_info_ptr->master_sm_lid);
			port_info.capability_mask =
			    be32_to_cpu(port_info_ptr->capability_mask);
			port_info.local_port_num =
			    port_info_ptr->local_port_num;
			port_info.link_width_enabled =
			    port_info_ptr->link_width_enable;
			port_info.link_width_supported =
			    port_info_ptr->link_width_supported;
			port_info.link_width_active =
			    port_info_ptr->link_width_active;
			port_info.link_speed_supported =
			    (port_info_ptr->
			     link_speed_supported__port_state & 0xF0) >> 4;
			port_info.port_state =
			    port_info_ptr->
			    link_speed_supported__port_state & 0x0F;
			port_info.phy_state =
			    (port_info_ptr->
			     port_phy_state__link_down_default_state & 0xF0) >>
			    4;
			port_info.down_default_state =
			    port_info_ptr->
			    port_phy_state__link_down_default_state & 0x0F;
			port_info.mkey_protect =
			    port_info_ptr->mkey_protect_bits__reserve__lmc >> 6;
			port_info.lmc =
			    port_info_ptr->
			    mkey_protect_bits__reserve__lmc & 0x07;
			port_info.link_speed_active =
			    (port_info_ptr->
			     link_speed_active__linke_speed_enabled & 0xF0) >>
			    4;
			port_info.link_speed_enabled =
			    port_info_ptr->
			    link_speed_active__linke_speed_enabled & 0x0F;
			port_info.neighbor_mtu =
			    port_info_ptr->nighbor_mtu__master_smsl >> 4;
			port_info.master_sm_sl =
			    port_info_ptr->nighbor_mtu__master_smsl & 0x0F;
			port_info.vl_cap = port_info_ptr->vlcap__reserved >> 4;
			port_info.vl_high_limit = port_info_ptr->vlhigh_limit;
			port_info.vl_arb_high_cap =
			    port_info_ptr->vl_arbitration_high_cap;
			port_info.vl_arb_low_cap =
			    port_info_ptr->vl_arbitration_low_cap;
			port_info.mtu_cap =
			    port_info_ptr->reserve__mtu_cap & 0x0F;
			port_info.vl_stall_count =
			    port_info_ptr->vl_stall_count__hoqlife >> 3;
			port_info.hoq_life =
			    port_info_ptr->vl_stall_count__hoqlife & 0x7;
			port_info.operational_vl =
			    port_info_ptr->
			    operational_vl__pin__pout__fin__fout >> 4;
			port_info.partition_enforcement_inbound =
			    (port_info_ptr->
			     operational_vl__pin__pout__fin__fout & 0x08) >> 3;
			port_info.partition_enforcement_outbound =
			    (port_info_ptr->
			     operational_vl__pin__pout__fin__fout & 0x04) >> 2;
			port_info.filter_raw_inbound =
			    (port_info_ptr->
			     operational_vl__pin__pout__fin__fout & 0x02) >> 1;
			port_info.filter_raw_outbound =
			    port_info_ptr->
			    operational_vl__pin__pout__fin__fout & 0x01;
			port_info.mkey_violations =
			    be16_to_cpu(port_info_ptr->mkey_violation);
			port_info.pkey_violations =
			    be16_to_cpu(port_info_ptr->pkey_violation);
			port_info.qkey_violations =
			    be16_to_cpu(port_info_ptr->qkey_violation);
			port_info.guid_cap = port_info_ptr->guid_cap;
			port_info.subnet_timeout =
			    port_info_ptr->reserve__subnet_timeout & 0x1F;
			port_info.resp_time_val =
			    port_info_ptr->reserve__resp_time_value & 0x1F;
			port_info.local_phy_errs =
			    port_info_ptr->local_phy_error__overrun_error >> 4;
			port_info.overrun_errs =
			    port_info_ptr->
			    local_phy_error__overrun_error & 0x0F;

			TS_TRACE(MOD_KERNEL_IB, T_VERBOSE, TRACE_KERNEL_IB_GEN,
				 "SA client Port Info response: "
				 "lid=0x%04x, capability mask=0x%08x",
				 port_info.lid, port_info.capability_mask);

			if (query->completion_func) {
				query->completion_func(query->hdr.transaction_id,
						       0,
						       &port_info,
						       query->completion_arg);
			}
		}
		break;

	case TS_IB_CLIENT_RESPONSE_ERROR:
		TS_TRACE(MOD_KERNEL_IB, T_VERBOSE, TRACE_KERNEL_IB_GEN,
			 "SA client port info MAD status 0x%04x",
			 be16_to_cpu(packet->status));
		if (query->completion_func) {
			query->completion_func(query->hdr.transaction_id,
					       -EINVAL,
					       NULL, query->completion_arg);
		}
		break;

	case TS_IB_CLIENT_RESPONSE_TIMEOUT:
		TS_TRACE(MOD_KERNEL_IB, T_VERBOSE, TRACE_KERNEL_IB_GEN,
			 "SA client port info query timed out");
		if (query->completion_func) {
			query->completion_func(query->hdr.transaction_id,
					       -ETIMEDOUT,
					       NULL, query->completion_arg);
		}
		break;

	case TS_IB_CLIENT_RESPONSE_CANCEL:
		TS_TRACE(MOD_KERNEL_IB, T_VERBOSE, TRACE_KERNEL_IB_GEN,
			 "SA client port info query canceled");
		break;

	default:
		TS_REPORT_WARN(MOD_KERNEL_IB, "Unknown status %d", status);
		break;
	}

	ib_sa_outs_query_remove(&query->hdr.outs_query);
	kfree(query);
}

static void _tsIbPortInfoTblResponse(tTS_IB_CLIENT_RESPONSE_STATUS status,
				     u8 * header,
				     u8 * data,
				     u32 data_size, void *query_ptr)
{
	struct ib_sa_port_info_query *query =
				(struct ib_sa_port_info_query *) query_ptr;

	struct ib_port_info *port_info = NULL;
	
	switch (status) {
	case TS_IB_CLIENT_RESPONSE_OK:
		{
			struct ib_sa_header *sa_header =
			    (struct ib_sa_header *) header;
			s16 attrib_offset =
			    be16_to_cpu(sa_header->attrib_offset) * 8;
			s32 port_info_size =
			    sizeof(struct ib_sa_port_info);
			s32 num_entries = 0;
			s32 i;

			TS_TRACE(MOD_KERNEL_IB, T_VERBOSE, TRACE_KERNEL_IB_GEN,
				 "_tsIbPortInfoTblResponse(data_size= %d, port-info-size= %d, attrib_offset= %d, num-entries= %d)\n",
				 data_size, port_info_size, attrib_offset,
				 num_entries);

			if ((attrib_offset > 0) && ((port_info = kmalloc(sizeof *port_info, GFP_ATOMIC)) != NULL)){

				num_entries = data_size / attrib_offset;

				for (i = 0; i < num_entries; i++) {
					struct ib_sa_port_info *port_info_ptr =
					    (struct ib_sa_port_info *) (data +
								   i *
								   attrib_offset);

					memcpy(port_info->mkey,
					       port_info_ptr->mkey, 8);
					memcpy(port_info->gid_prefix,
					       port_info_ptr->gid_prefix, 8);
					port_info->lid =
					    be16_to_cpu(port_info_ptr->lid);
					port_info->master_sm_lid =
					    be16_to_cpu(port_info_ptr->
							master_sm_lid);
					port_info->capability_mask =
					    be32_to_cpu(port_info_ptr->
							capability_mask);
					port_info->local_port_num =
					    port_info_ptr->local_port_num;
					port_info->link_width_enabled =
					    port_info_ptr->link_width_enable;
					port_info->link_width_supported =
					    port_info_ptr->link_width_supported;
					port_info->link_width_active =
					    port_info_ptr->link_width_active;
					port_info->link_speed_supported =
					    (port_info_ptr->
					     link_speed_supported__port_state &
					     0xF0) >> 4;
					port_info->port_state =
					    port_info_ptr->
					    link_speed_supported__port_state &
					    0x0F;
					port_info->phy_state =
					    (port_info_ptr->
					     port_phy_state__link_down_default_state
					     & 0xF0) >> 4;
					port_info->down_default_state =
					    port_info_ptr->
					    port_phy_state__link_down_default_state
					    & 0x0F;
					port_info->mkey_protect =
					    port_info_ptr->
					    mkey_protect_bits__reserve__lmc >>
					    6;
					port_info->lmc =
					    port_info_ptr->
					    mkey_protect_bits__reserve__lmc &
					    0x07;
					port_info->link_speed_active =
					    (port_info_ptr->
					     link_speed_active__linke_speed_enabled
					     & 0xF0) >> 4;
					port_info->link_speed_enabled =
					    port_info_ptr->
					    link_speed_active__linke_speed_enabled
					    & 0x0F;
					port_info->neighbor_mtu =
					    port_info_ptr->
					    nighbor_mtu__master_smsl >> 4;
					port_info->master_sm_sl =
					    port_info_ptr->
					    nighbor_mtu__master_smsl & 0x0F;
					port_info->vl_cap =
					    port_info_ptr->vlcap__reserved >> 4;
					port_info->vl_high_limit =
					    port_info_ptr->vlhigh_limit;
					port_info->vl_arb_high_cap =
					    port_info_ptr->
					    vl_arbitration_high_cap;
					port_info->vl_arb_low_cap =
					    port_info_ptr->
					    vl_arbitration_low_cap;
					port_info->mtu_cap =
					    port_info_ptr->
					    reserve__mtu_cap & 0x0F;
					port_info->vl_stall_count =
					    port_info_ptr->
					    vl_stall_count__hoqlife >> 3;
					port_info->hoq_life =
					    port_info_ptr->
					    vl_stall_count__hoqlife & 0x7;
					port_info->operational_vl =
					    port_info_ptr->
					    operational_vl__pin__pout__fin__fout
					    >> 4;
					port_info->
					    partition_enforcement_inbound =
					    (port_info_ptr->
					     operational_vl__pin__pout__fin__fout
					     & 0x08) >> 3;
					port_info->
					    partition_enforcement_outbound =
					    (port_info_ptr->
					     operational_vl__pin__pout__fin__fout
					     & 0x04) >> 2;
					port_info->filter_raw_inbound =
					    (port_info_ptr->
					     operational_vl__pin__pout__fin__fout
					     & 0x02) >> 1;
					port_info->filter_raw_outbound =
					    port_info_ptr->
					    operational_vl__pin__pout__fin__fout
					    & 0x01;
					port_info->mkey_violations =
					    be16_to_cpu(port_info_ptr->
							mkey_violation);
					port_info->pkey_violations =
					    be16_to_cpu(port_info_ptr->
							pkey_violation);
					port_info->qkey_violations =
					    be16_to_cpu(port_info_ptr->
							qkey_violation);
					port_info->guid_cap =
					    port_info_ptr->guid_cap;
					port_info->subnet_timeout =
					    port_info_ptr->
					    reserve__subnet_timeout & 0x1F;
					port_info->resp_time_val =
					    port_info_ptr->
					    reserve__resp_time_value & 0x1F;
					port_info->local_phy_errs =
					    port_info_ptr->
					    local_phy_error__overrun_error >> 4;
					port_info->overrun_errs =
					    port_info_ptr->
					    local_phy_error__overrun_error &
					    0x0F;

					TS_TRACE(MOD_KERNEL_IB, T_VERBOSE,
						 TRACE_KERNEL_IB_GEN,
						 "SA client Port Info table response: "
						 "lid=0x%04x, capability mask=0x%08x",
						 port_info->lid,
						 port_info->capability_mask);

					if (query->completion_func) {
						query->completion_func(query->
								       hdr.transaction_id,
								       0,
								       port_info,
								       query->
								       completion_arg);
					}
				}
				kfree(port_info);
				port_info = NULL;
			} else {
				if (attrib_offset > 0) {
					TS_TRACE(MOD_KERNEL_IB, T_VERBOSE,
						 TRACE_KERNEL_IB_GEN,
						 "SA client port info table with attrib_offset = 0");
					if (query->completion_func) {
						query->completion_func(query->
								       hdr.transaction_id,
								       -EINVAL, NULL,
								       query->
								       completion_arg);
					}
				} else {
					TS_REPORT_WARN(MOD_KERNEL_IB, "%s: kmalloc of port_info failed", __func__);
					if (query->completion_func) {
						query->completion_func(query->
								       hdr.transaction_id,
								       -ENOMEM, NULL,
								       query->
								       completion_arg);
					}
				}
			}
		}
		break;

	case TS_IB_CLIENT_RESPONSE_ERROR:
		TS_TRACE(MOD_KERNEL_IB, T_VERBOSE, TRACE_KERNEL_IB_GEN,
			 "SA client port info table ERROR status");
		if (query->completion_func) {
			query->completion_func(query->hdr.transaction_id,
					       -EINVAL,
					       NULL, query->completion_arg);
		}
		break;

	case TS_IB_CLIENT_RESPONSE_TIMEOUT:
		TS_TRACE(MOD_KERNEL_IB, T_VERBOSE, TRACE_KERNEL_IB_GEN,
			 "SA client port info table query timed out");
		if (query->completion_func) {
			query->completion_func(query->hdr.transaction_id,
					       -ETIMEDOUT,
					       NULL, query->completion_arg);
		}
		break;

	case TS_IB_CLIENT_RESPONSE_CANCEL:
		TS_TRACE(MOD_KERNEL_IB, T_VERBOSE, TRACE_KERNEL_IB_GEN,
			 "SA client port info table query canceled");
		break;

	default:
		TS_REPORT_WARN(MOD_KERNEL_IB, "Unknown status %d", status);
		break;
	}

	ib_rmpp_client_free(data);

	ib_sa_outs_query_remove(&query->hdr.outs_query);
	kfree(query);
}

int tsIbPortInfoQuery(struct ib_device *device,
		      tTS_IB_PORT port,
		      u16 port_lid,
		      tTS_IB_PORT port_num,
		      int timeout_jiffies,
		      tTS_IB_PORT_INFO_QUERY_COMPLETION_FUNC completion_func,
		      void *completion_arg,
		      tTS_IB_CLIENT_QUERY_TID * transaction_id)
{
	struct ib_mad *mad;
	struct ib_sa_payload *sa_payload = NULL;
	struct ib_sa_port_info *port_info = NULL;
	struct ib_sa_port_info_query *query;
	int ret = 0;
	tTS_IB_CLIENT_QUERY_TID temp_tid;

	query = kmalloc(sizeof *query, GFP_ATOMIC);
	if (!query) {
		return -ENOMEM;
	}

	mad = kmalloc(sizeof *mad, GFP_ATOMIC);
	if (!mad) {
		kfree(query);
		return -ENOMEM;
	}
	sa_payload = (struct ib_sa_payload *) &mad->payload;
	port_info = (struct ib_sa_port_info *) sa_payload->admin_data;

	tsIbSaClientMadInit(mad, device, port);
	mad->r_method = IB_MGMT_METHOD_GET;
	mad->attribute_id = cpu_to_be16(TS_IB_SA_ATTRIBUTE_PORT_INFO_RECORD);
	mad->attribute_modifier = 0;

	sa_payload->sa_header.component_mask = cpu_to_be64(0x1ULL);	/* port LID  */

	port_info->port_lid = cpu_to_be16(port_lid);

	temp_tid = query->hdr.transaction_id = mad->transaction_id;
	query->completion_func = completion_func;
	query->completion_arg = completion_arg;

	ib_sa_outs_query_add(&query->hdr.outs_query);
	ret = ib_client_query(mad, timeout_jiffies, _tsIbPortInfoResponse, query);
	if (ret) {
		ib_sa_outs_query_remove(&query->hdr.outs_query);
		kfree(mad);
		kfree(query);
		goto proc_exit;
	}

	*transaction_id = temp_tid;

proc_exit:	
	return ret;
}

int tsIbPortInfoTblQuery(struct ib_device *device,
			 tTS_IB_PORT port,
			 int timeout_jiffies,
			 tTS_IB_PORT_INFO_QUERY_COMPLETION_FUNC completion_func,
			 void *completion_arg,
			 tTS_IB_CLIENT_QUERY_TID * transaction_id)
{
	struct ib_mad *mad;
	struct ib_sa_port_info_query *query;
	struct ib_client_rmpp_mad *rmpp_mad;
	int ret = 0;
	tTS_IB_CLIENT_QUERY_TID temp_tid;

	query = kmalloc(sizeof *query, GFP_ATOMIC);
	if (!query) {
		return -ENOMEM;
	}

	mad = kmalloc(sizeof *mad, GFP_ATOMIC);
	if (!mad) {
		kfree(query);
		return -ENOMEM;
	}

	tsIbSaClientMadInit(mad, device, port);
	mad->r_method = TS_IB_SA_METHOD_GET_TABLE;
	mad->attribute_id = cpu_to_be16(TS_IB_SA_ATTRIBUTE_PORT_INFO_RECORD);
	mad->attribute_modifier = 0;

	/* rmpp header init */
	rmpp_mad = (struct ib_client_rmpp_mad *) mad;
	rmpp_mad->version = 1;
	rmpp_mad->type = TS_IB_CLIENT_RMPP_TYPE_DATA;

	temp_tid = query->hdr.transaction_id = mad->transaction_id;
	query->completion_func = completion_func;
	query->completion_arg = completion_arg;

	ib_sa_outs_query_add(&query->hdr.outs_query);
	ret = ib_rmpp_client_query(mad, timeout_jiffies, sizeof(struct ib_sa_header),
			     _tsIbPortInfoTblResponse, query);

	if (ret) {
		ib_sa_outs_query_remove(&query->hdr.outs_query);
		kfree(mad);
		kfree(query);
		goto proc_exit;
	}
	
	*transaction_id = temp_tid;

proc_exit:	
	return ret;
}
