/*
  This software is available to you under a choice of one of two
  licenses.  You may choose to be licensed under the terms of the GNU
  General Public License (GPL) Version 2, available at
  <http://www.fsf.org/copyleft/gpl.html>, or the OpenIB.org BSD
  license, available in the LICENSE.TXT file accompanying this
  software.  These details are also available at
  <http://openib.org/license.html>.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
  BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
  ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
  CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
  SOFTWARE.

  Copyright (c) 2004 Mellanox Technologies Ltd.  All rights reserved.
*/

/*================ macro definitions ===============================================*/
#ifdef XHH_CMD_TIME_TRACK
extern u_int64_t XHH_eventp_last_cmdif_interrupt;
#endif

#define PRM_RSC_DELTA 32
#define MAX_CMD_OBJS 32 /* max number of objects allowed to be created */
#define CMD_ETIME_EVENTS 1000000 /* max execution time is fixed when we're using events */
#define MAX_ITER_ON_EINTR 10
#define MAX_UC_FOR_GO 100 /* time in microsec to do busy wait on go bit */
#define MAX_UC_SLEEP_FOR_GO 300000000 /* time in microsec to do busy wait on go bit */
#define SHORT_UC_DELAY_FOR_GO 10 /* time in microsec to do busy wait on go bit after first time */

#ifdef IN
  #undef IN
#endif
#define IN

#ifdef OUT
  #undef OUT
#endif
#define OUT

#define HCR_DW_BYTE_OFFSET(field) (MT_BYTE_OFFSET(ADD_DEV_PRM_PREFIX(hca_command_register_st),field) & (~3))
#define HCR_DW_BIT_OFFSET(field) (MT_BIT_OFFSET(ADD_DEV_PRM_PREFIX(hca_command_register_st),field) - HCR_DW_BYTE_OFFSET(field) * 8)
#define HCR_BIT_SIZE(field) MT_BIT_SIZE(ADD_DEV_PRM_PREFIX(hca_command_register_st),field)


#define PREP_TOKEN_DW(token_dw, val) MT_INSERT32((token_dw),(val),HCR_DW_BIT_OFFSET(token),MT_BIT_SIZE(ADD_DEV_PRM_PREFIX(hca_command_register_st),token)); \
                                     token_dw = MOSAL_cpu_to_be32(token_dw);



#define DEFAULT_TOKEN 0x1234 /* default token value to use when there can be no outstanding commands */
#define FREE_LIST_EOL ((u_int32_t)(-1))

#define NEW_EQE_FORMAT

/*================ static variables definitions ====================================*/

/*================ common private functions prototypes =====================================*/
XHH_cmd_status_t XHH_cmd_main_flow(struct cmd_if_context_st *entry, command_fields_t *cmd_prms);
MT_bool XHH_cmd_cmdif_is_free(struct cmd_if_context_st *entry);
XHH_cmd_status_t cmd_invoke(XHH_cmd_t cmd_if, command_fields_t *cmd_prms);
const char *str_XHH_cmd(ADD_DEV_PREFIX_LC(if_cmd_t) opcode);


/* ==== inline functions ===============*/

static HH_ret_t inline get_ctx_by_idx(struct cmd_if_context_st *entry, u_int16_t idx, cmd_ctx_t **ctx_pp)
{
  if ( (idx<entry->ctx_obj.num) && (entry->ctx_obj.ctx_arr[idx].ref_cnt>0) ) {
    *ctx_pp = &entry->ctx_obj.ctx_arr[idx];
    return HH_OK;
  }
  return HH_EAGAIN;
}

static XHH_eqn_t inline eqn_set(struct cmd_if_context_st *entry, XHH_eqn_t new_eqn)
{
  XHH_eqn_t old_eqn;

  MOSAL_spinlock_irq_lock(&entry->eqn_spl);
  old_eqn = entry->eqn;
  entry->eqn = new_eqn;
  MOSAL_spinlock_unlock(&entry->eqn_spl);
  return old_eqn;
}


/*
 *  write_command_dw
 */
static inline void  write_command_dw(u_int32_t *dst_buf, u_int8_t go,
                                     u_int8_t use_event, u_int8_t op_mod, u_int16_t opcode)
{
  u_int32_t cmd=0;

  MT_INSERT32(cmd,opcode,HCR_DW_BIT_OFFSET(opcode),HCR_BIT_SIZE(opcode));
  MT_INSERT32(cmd,op_mod,HCR_DW_BIT_OFFSET(opcode_modifier),HCR_BIT_SIZE(opcode_modifier));
  MT_INSERT32(cmd,use_event,HCR_DW_BIT_OFFSET(e),HCR_BIT_SIZE(e));
  MT_INSERT32(cmd,go,HCR_DW_BIT_OFFSET(go),HCR_BIT_SIZE(go));
  MT_INSERT32(cmd,0,HCR_DW_BIT_OFFSET(status),HCR_BIT_SIZE(status)); /* status */
  *dst_buf = MOSAL_cpu_to_be32(cmd);
}


/*
 *  ptr_to_mailbox_ptr
 */
static inline void ptr_to_mailbox_ptr(MT_phys_addr_t ptr, addr_64bit_t *mbx_pmtr)
{
  if ( sizeof(ptr) == 4 ) {
    MTL_DEBUG4("pointer is 32 bit\n");
    mbx_pmtr->addr_h = 0;
    mbx_pmtr->addr_l = (u_int32_t)ptr;
  }
  else if ( sizeof(ptr) == 8 ) {
    MTL_DEBUG4("pointer is 64 bit\n");
    mbx_pmtr->addr_h = (u_int32_t)(((u_int64_t)ptr)>>32);
    mbx_pmtr->addr_l = (u_int32_t)ptr;
  }
  else {
    MTL_ERROR1("bad address pointer size: %d\n",(int)sizeof(ptr));
  }
  mbx_pmtr->addr_h = MOSAL_cpu_to_be32(mbx_pmtr->addr_h);
  mbx_pmtr->addr_l = MOSAL_cpu_to_be32(mbx_pmtr->addr_l);
}

/*
 *  cmd_if_status
 */
static inline XHH_cmd_status_t cmd_if_status(struct cmd_if_context_st *entry)
{
  u_int32_t *offset = (u_int32_t *)entry->hcr_virt_base;
  u_int32_t cmd;

  cmd = offset[HCR_DW_BYTE_OFFSET(status)>>2];
  cmd = MOSAL_be32_to_cpu(cmd);
  return (XHH_cmd_status_t)MT_EXTRACT32(cmd, HCR_DW_BIT_OFFSET(status), HCR_BIT_SIZE(status));
}


/*
 *  set_mailbox
 */
static inline void set_mailbox(u_int32_t *dst_buf, MT_phys_addr_t mbx_pa)
{
  addr_64bit_t addr;

  ptr_to_mailbox_ptr(mbx_pa, &addr);
  dst_buf[0] = addr.addr_h;
  dst_buf[1] = addr.addr_l;
}


/*
 *  cvt_be32_to_cpu
 *  size in bytes
 */
static inline void cvt_be32_to_cpu(void *buf, u_int32_t size)
{
  u_int32_t i, *p=(u_int32_t *)(buf);
  for ( i=0; i<(size>>2); ++i ) {
    *p = MOSAL_be32_to_cpu(*p);
    p++;
  }
}


/*
 *  cvt_cpu_to_be32
 *  size in bytes
 */
static inline void cvt_cpu_to_be32(void *buf, u_int32_t size)
{
  u_int32_t i, *p=(u_int32_t *)(buf);
  for ( i=0; i<(size>>2); ++i ) {
    *p = MOSAL_cpu_to_be32(*p);
    p++;
  }
}


/*
 *  d32_to_s64
 */
static inline u_int64_t d32_to_s64(u_int32_t hi, u_int32_t lo)
{
  return(((u_int64_t)hi) << 32) | (u_int64_t)(lo);
}


