/*
  This software is available to you under a choice of one of two
  licenses.  You may choose to be licensed under the terms of the GNU
  General Public License (GPL) Version 2, available at
  <http://www.fsf.org/copyleft/gpl.html>, or the OpenIB.org BSD
  license, available in the LICENSE.TXT file accompanying this
  software.  These details are also available at
  <http://openib.org/license.html>.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
  BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
  ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
  CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
  SOFTWARE.

  Copyright (c) 2004 Mellanox Technologies Ltd.  All rights reserved.
*/

#ifndef H_CMDS_WRAP_PRIV_H
#define H_CMDS_WRAP_PRIV_H

/* T.D.(matan): change the method of reporting SQ_DRAINING state to a more cosisntent one.*/
#define XHH_CMD_QP_DRAINING_FLAG	0x80

/* For the READ_MTT and WRITE_MTT, the mailbox contains a 16-byte preamble */
#define MTT_CMD_PA_PREAMBLE_SIZE    (16)

#define EX_FLD(dst, a, st, fld) dst->fld = MT_EXTRACT_ARRAY32(a, MT_BIT_OFFSET(st, fld), MT_BIT_SIZE(st, fld))
#define INS_FLD(src, a, st, fld) MT_INSERT_ARRAY32(a, src->fld, MT_BIT_OFFSET(st, fld), MT_BIT_SIZE(st, fld))
#define EX_FLD_SH(dst, a, st, fld) dst->fld = (MT_EXTRACT_ARRAY32(a, MT_BIT_OFFSET(st, fld), MT_BIT_SIZE(st, fld)) << (MT_BIT_OFFSET(st, fld)& MASK32(5)))
#define INS_FLD_SH(src, a, st, fld) MT_INSERT_ARRAY32(a, (u_int32_t)(src->fld >> (MT_BIT_OFFSET(st, fld) & MASK32(5))), MT_BIT_OFFSET(st, fld), MT_BIT_SIZE(st, fld))
#define INS_FLD_IMMED(src, a, st, fld) MT_INSERT_ARRAY32(a, src, MT_BIT_OFFSET(st, fld), MT_BIT_SIZE(st, fld))

#define EX_BOOL_FLD(dst, a, st, fld) dst->fld = ((MT_EXTRACT_ARRAY32(a, MT_BIT_OFFSET(st, fld), MT_BIT_SIZE(st, fld))==0) ? FALSE : TRUE)
#define INS_BOOL_FLD(src, a, st, fld) MT_INSERT_ARRAY32(a, ((src->fld==FALSE)?0:1), MT_BIT_OFFSET(st, fld), MT_BIT_SIZE(st, fld))

#define QP_EX_FLD(dst, a, st, fld) dst->fld = MT_EXTRACT_ARRAY32(a, MT_BIT_OFFSET(st, qpc_eec_data.fld), MT_BIT_SIZE(st, qpc_eec_data.fld))
#define QP_INS_FLD(src, a, st, fld) MT_INSERT_ARRAY32(a, src->fld, MT_BIT_OFFSET(st, qpc_eec_data.fld), MT_BIT_SIZE(st, qpc_eec_data.fld))

#define QP_EX_BOOL_FLD(dst, a, st, fld) dst->fld = ((MT_EXTRACT_ARRAY32(a, MT_BIT_OFFSET(st, qpc_eec_data.fld), MT_BIT_SIZE(st, qpc_eec_data.fld))==0) ? FALSE : TRUE)
#define QP_INS_BOOL_FLD(src, a, st, fld) MT_INSERT_ARRAY32(a, ((src->fld==FALSE)?0:1), MT_BIT_OFFSET(st, qpc_eec_data.fld), MT_BIT_SIZE(st, qpc_eec_data.fld))

#define EX_FLD64(dst, a, st, fld) dst->fld = (MT_EXTRACT_ARRAY32(a, MT_BIT_OFFSET(st, fld##_l), MT_BIT_SIZE(st, fld##_l)) | \
                                               (u_int64_t)(MT_EXTRACT_ARRAY32(a, MT_BIT_OFFSET(st, fld##_h), MT_BIT_SIZE(st, fld##_h))) << 32) 

#define INS_FLD64(src, a, st, fld) MT_INSERT_ARRAY32(a, (u_int32_t)src->fld, MT_BIT_OFFSET(st, fld##_l), MT_BIT_SIZE(st, fld##_l)) ; \
                                   MT_INSERT_ARRAY32(a, (u_int32_t)((src->fld) >> 32), MT_BIT_OFFSET(st, fld##_h), MT_BIT_SIZE(st, fld##_h))

#define INS_FLD64_IMMED(src, a, st, fld) MT_INSERT_ARRAY32(a, (u_int32_t)src, MT_BIT_OFFSET(st, fld##_l), MT_BIT_SIZE(st, fld##_l)) ; \
                                   MT_INSERT_ARRAY32(a, (u_int32_t)((src) >> 32), MT_BIT_OFFSET(st, fld##_h), MT_BIT_SIZE(st, fld##_h))

#define INS_FLD64_SH(src, a, st, fld) MT_INSERT_ARRAY32(a, (u_int32_t)((src->fld) >> (MT_BIT_OFFSET(st, fld##_l) & MASK32(5))), MT_BIT_OFFSET(st, fld##_l), MT_BIT_SIZE(st, fld##_l)) ; \
                                   MT_INSERT_ARRAY32(a, (u_int32_t)((src->fld) >> 32), MT_BIT_OFFSET(st, fld##_h), MT_BIT_SIZE(st, fld##_h))

#define INS_FLD64_IMMED_SH(src, a, st, fld) MT_INSERT_ARRAY32(a, (u_int32_t)(src >> (MT_BIT_OFFSET(st, fld##_l) & MASK32(5))), MT_BIT_OFFSET(st, fld##_l), MT_BIT_SIZE(st, fld##_l)) ; \
                                   MT_INSERT_ARRAY32(a, (u_int32_t)(src >> 32), MT_BIT_OFFSET(st, fld##_h), MT_BIT_SIZE(st, fld##_h))

#define EX_FLD64_SH(dst, a, st, fld) dst->fld = ((MT_EXTRACT_ARRAY32(a, MT_BIT_OFFSET(st, fld##_l), MT_BIT_SIZE(st, fld##_l))) << ((MT_BIT_OFFSET(st, fld##_l)& MASK32(5)))  | \
                                               (u_int64_t)(MT_EXTRACT_ARRAY32(a, MT_BIT_OFFSET(st, fld##_h), MT_BIT_SIZE(st, fld##_h))) << 32) 

#define QP_EX_FLD64(dst, a, st, fld) dst->fld = (MT_EXTRACT_ARRAY32(a, MT_BIT_OFFSET(st, qpc_eec_data.fld##_1), MT_BIT_SIZE(st, qpc_eec_data.fld##_1)) | \
                                               (u_int64_t)(MT_EXTRACT_ARRAY32(a, MT_BIT_OFFSET(st, qpc_eec_data.fld##_0), MT_BIT_SIZE(st, qpc_eec_data.fld##_0))) << 32) 

#define QP_INS_FLD64(src, a, st, fld) MT_INSERT_ARRAY32(a, (u_int32_t)src->fld, MT_BIT_OFFSET(st, qpc_eec_data.fld##_1), MT_BIT_SIZE(st, qpc_eec_data.fld##_1)) ; \
                                   MT_INSERT_ARRAY32(a, (u_int32_t)((src->fld) >> 32), MT_BIT_OFFSET(st, qpc_eec_data.fld##_0), MT_BIT_SIZE(st, qpc_eec_data.fld##_0))

#define XHH_CMDS_WRAP_DEBUG_LEVEL 4
#define CMDS_DBG    MTL_DEBUG4

#endif

