/*
  This software is available to you under a choice of one of two
  licenses.  You may choose to be licensed under the terms of the GNU
  General Public License (GPL) Version 2, available at
  <http://www.fsf.org/copyleft/gpl.html>, or the OpenIB.org BSD
  license, available in the LICENSE.TXT file accompanying this
  software.  These details are also available at
  <http://openib.org/license.html>.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
  BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
  ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
  CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
  SOFTWARE.

  Copyright (c) 2004 Mellanox Technologies Ltd.  All rights reserved.
*/

#if !defined(H_CQM_PRIV_H)
#define H_CQM_PRIV_H
#include <epool.h>
#include <cmdif.h>
#include <dev_cqm.h>

#define logIfErr(f) \
  if (rc != HH_OK) { MTL_ERROR1("%s: rc=%s\n", f, HH_strerror_sym(rc)); }

/* macro for translating cmd_rc return codes for non-destroy procs */
#define CMDRC2HH_ND(cmd_rc) ((cmd_rc == XHH_CMD_STAT_OK) ? HH_OK : \
                          (cmd_rc == XHH_CMD_STAT_EINTR) ? HH_EINTR : HH_EFATAL)

#define CMDRC2HH_BUSY(cmd_rc) ((cmd_rc == XHH_CMD_STAT_OK) ? HH_OK :   \
                          (cmd_rc == XHH_CMD_STAT_EINTR) ? HH_EINTR :  \
                          ((cmd_rc == XHH_CMD_STAT_RESOURCE_BUSY) ||   \
                            (cmd_rc == XHH_CMD_STAT_REG_BOUND)) ? HH_EBUSY : HH_EFATAL)
                            
#define TCQM_CQN(cqm,cqc_index) \
          ( ( ((cqm)->entries[cqc_index].cqn_prefix) << (cqm)->log2_max_cq ) | (cqc_index) )

enum
{
  CQE_size      = sizeof(struct ADD_DEV_PRM_PREFIX(completion_queue_entry_st))/8, /* 32 */
  CQE_size_log2 = 5,
  CQE_size_mask = (1ul << CQE_size_log2) - 1
};

typedef struct Completion_Queue_Context  Completion_Queue_Context_t;

typedef struct
{
   unsigned long  prev;
   unsigned long  next;
} _free_offs_t;


/* Completion Queue Context Manager - entry info */
typedef struct CQCM_entry_s
{
  /* XHH_cq_props_t  props; / * May be optimzed out, using CmdIf output */
  unsigned long   n_cq;  /* With buf_sz, may just recompute and save needed */
  VAPI_lkey_t     lkey;
  MOSAL_protection_ctx_t  user_protection_context; /*Save protection context to be used on resize*/
  XHH_DEV_SPECIFIC_CQCM_ENTRY_DATA
#if defined(MT_SUSPEND_QP)
  MT_bool         is_suspended;
#endif
} CQCM_entry_t;


/* Completion Queue Context Manager - entry */
typedef struct
{
  union
  {
    CQCM_entry_t  used;
    _free_offs_t  freelist;
  } u;
  unsigned char cqn_prefix:7; /* CQ number - avoid ghost CQ events (FM issue #15134) */
  unsigned char in_use    :1;
} CQCM_entry_ut;

static const EPool_meta_t  fl_meta =
  {
    sizeof(CQCM_entry_ut),
    (unsigned int)(MT_ulong_ptr_t)(&(((CQCM_entry_ut*)(0))->u.freelist.prev)),
    (unsigned int)(MT_ulong_ptr_t)(&(((CQCM_entry_ut*)(0))->u.freelist.next))
  };


/* The main CQ-manager structure */
typedef struct XHH_cqm_st
{
  XHH_hob_t       hob;
  u_int8_t        log2_max_cq;
  u_int32_t       max_cq;
  CQCM_entry_ut*  entries;
  EPool_t         flist;

  /* convenient handle saving  */
  XHH_cmd_t       cmd_if;
  XHH_mrwm_t      mrwm_internal;

  MT_bool cq_resize_fixed; /* FW fix for FM issue #16966/#17002: comp. events during resize */
  XHH_DEV_SPECIFIC_CQM_DATA
} TCQM_t;


/************************************************************************/
/************************************************************************/
/*                         private functions                            */
/************************************************************************/

#if MAX_TRACE >= 1
static char*  ulr_print(char* buf, const XHH_cq_ul_resources_t* p)
{
  sprintf(buf, "{CQulRes: buf="VIRT_ADDR_FMT", sz="SIZE_T_FMT", uar=%d}",
          p->cqe_buf, p->cqe_buf_sz, p->uar_index);  
  return buf;
} /* ulr_print */
#endif


/************************************************************************/
static inline MT_bool  in_use_cq(TCQM_t* cqm, HH_cq_hndl_t  cq)
{
  u_int32_t cqc_index= cq & MASK32(cqm->log2_max_cq);
  MT_bool  in_use = (cqc_index < cqm->max_cq) && cqm->entries[cqc_index].in_use;
  if (!in_use) { MTL_ERROR2(MT_FLFMT("unused cq=0x%x"), cq); }
  return in_use;
} /* in_use_cq */


/************************************************************************/

HH_ret_t dev_specific_cqm_create_cq(XHH_cqm_t  cqm, 
                                    XHH_cq_ul_resources_t*  cq_ul_resources_p,
                                    u_int32_t  cqc_index);

HH_ret_t dev_specific_cqm_destroy_cq(XHH_cqm_t  cqm,
                                     u_int32_t  cqc_index);


#endif
