/*
  This software is available to you under a choice of one of two
  licenses.  You may choose to be licensed under the terms of the GNU
  General Public License (GPL) Version 2, available at
  <http://www.fsf.org/copyleft/gpl.html>, or the OpenIB.org BSD
  license, available in the LICENSE.TXT file accompanying this
  software.  These details are also available at
  <http://openib.org/license.html>.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
  BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
  ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
  CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
  SOFTWARE.

  Copyright (c) 2004 Mellanox Technologies Ltd.  All rights reserved.
*/


#ifndef H_EVENTP_PRIV_H
#define H_EVENTP_PRIV_H

#include <mosal.h>
#include <eventp.h>
#include <vapi_types.h>
#include <hh.h>
#include <xhh.h>

/*================ macro definitions ===============================================*/
#define EQP_MAX_EQS 4 /* we use 4 EQs when MSIX is enabled and 3 if not*/
#define EQE_HW_OWNER 0x80
#define EQE_SW_OWNER 0x00
#define EQE_OWNER_BYTE_OFFSET 31

#define PRM_EVENT_QUEUE_ENTRY_ST ADD_DEV_PRM_PREFIX(event_queue_entry_st)
#define EQE_DATA_BYTE_SIZE   (MT_BYTE_SIZE(PRM_EVENT_QUEUE_ENTRY_ST, event_data))  /* in bytes */
#define EQE_OWNER_OFFSET (MT_BIT_OFFSET(PRM_EVENT_QUEUE_ENTRY_ST, owner) /32 ) /* in DWORDS relay on owner field to be first in DWORD*/
#define EQE_EVENT_TYPE_OFFSET (MT_BIT_OFFSET(PRM_EVENT_QUEUE_ENTRY_ST, event_type) /32 ) /* in DWORDS relay on owner field to be first in DWORD*/
#define EQE_DATA_OFFSET (MT_BIT_OFFSET(PRM_EVENT_QUEUE_ENTRY_ST, event_data) /32 ) /* in DWORDS relay on owner field to be first in DWORD*/
#define EQE_DWORD_SIZE         (sizeof(struct PRM_EVENT_QUEUE_ENTRY_ST)/32)  /* in DWORDS */
#define EQ_ENTRY_SIZE    (sizeof(struct PRM_EVENT_QUEUE_ENTRY_ST) / 8) /* in bytes */


/*================ type definitions ================================================*/

typedef enum {
  EQP_CQ_COMP_EVENT,
  EQP_IB_EVENT,
  EQP_CMD_IF_EVENT,
  EQP_CATAS_ERR_EVENT,
  EQP_MLX_EVENT
}EQ_type_t;

/* entry saved for each EQ */
typedef struct EQP_eq_entry_st {
  MT_virt_addr_t         alloc_mem_addr_p; /* need to save the allocated pointer for free at destroy */
  VAPI_size_t            alloc_mem_bytes_size;
  void*                  eq_buff;
  VAPI_size_t            eq_buff_entry_num; /* if ==0 entry is invalid */
  HH_pd_hndl_t           pd;
  VAPI_lkey_t            mem_lkey;
  void                   *priv_context;
  XHH_eventp_mtmask_t    events_mask;
  EQ_type_t              eq_type;
  XHH_eqn_t              eqn;            /* eq number - needed by DPC */
  XHH_eventp_handler_t   handler;
  XHH_eventp_t           eventp_p;
  MOSAL_DPC_t            polling_dpc;
  u_int32_t              cons_cntr;		    /* the index is the lower bits as indicated by log_eq_size */
  u_int32_t              cons_cntr_mask;    /* mask to take index from counter */
  MT_bool                virtual_eq;
  atomic_t               dpc_cntr;       /* Outstanding DPCs counter (possible values:0,1,2)*/
  MOSAL_spinlock_t       dpc_lock;       /* ensure that only one DPC is running - WINDOWS allows 
                                            more than one DPC to run at the same time */
  DEV_EQ_ENTRY_FIELDS
}EQP_eq_entry_t;


/* The main EVENT processor structure */
typedef struct XHH_eventp_st {
  XHH_hob_t hob;
  XHH_ver_info_t version; 
  XHH_eventp_res_t event_resources; 
  XHH_cmd_t cmd_if; 
  XHH_mrwm_t mrwm_internal;
  XHH_lamm_t lamm;
  EQP_eq_entry_t eq_table[EQP_MAX_EQS]; /* static table of eqs */
  MOSAL_mutex_t    mtx;     /* used internally */
  MOSAL_protection_ctx_t ctx_internal;
  MOSAL_ISR_t            isr_obj[EQP_MAX_EQS];
  HH_hca_hndl_t          hh_hca_hndl;
  volatile MT_virt_addr_t   intr_clr_reg;   /* can be the high or the low but not both */
  MOSAL_iomem_t             intr_clr_reg_iomem;
  u_int32_t                 intr_clr_mask;
  volatile u_int8_t         max_eq_num_used;
  volatile MT_bool          have_fatal;
  volatile u_int32_t        free_eq_num; /*next free EQ number*/
  DEV_EVENTP_FIELDS
}XHH_eventp_internal_t;


/*================ external functions ================================================*/



#endif /* H_EVENTP_PRIV_H  */
