/*
  This software is available to you under a choice of one of two
  licenses.  You may choose to be licensed under the terms of the GNU
  General Public License (GPL) Version 2, available at
  <http://www.fsf.org/copyleft/gpl.html>, or the OpenIB.org BSD
  license, available in the LICENSE.TXT file accompanying this
  software.  These details are also available at
  <http://openib.org/license.html>.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
  BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
  ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
  CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
  SOFTWARE.

  Copyright (c) 2004 Mellanox Technologies Ltd.  All rights reserved.
*/

#if !defined(H_QPM_PRIV_H)
#define H_QPM_PRIV_H


extern void MadBufPrint(void *madbuf);
 
static void printGIDTable(XHH_qpm_t  qpm);
static void printPKeyTable(XHH_qpm_t  qpm);

#define CMDRC2HH_ND(cmd_rc) ((cmd_rc == XHH_CMD_STAT_OK) ? HH_OK : \
                          (cmd_rc == XHH_CMD_STAT_EINTR) ? HH_EINTR : HH_EFATAL)

#define ELSE_ACQ_ERROR(f) else { MTL_ERROR1("%s MOSAL_mutex_acq failed\n", f); }
#define logIfErr(f) \
  if (rc != HH_OK) { MTL_ERROR1("%s: rc=%s\n", f, HH_strerror_sym(rc)); }


#if !defined(ARR_SIZE)
# define ARR_SIZE(a)  (sizeof(a)/sizeof(a[0]))
#endif

/* These are the tunebale parameters */
enum
{
  TUNABLE_ACK_REQ_FREQ = 10,
  TUNABLE_FLIGHT_LIMIT = 9
};


enum {WQE_CHUNK_SIZE_LOG2 = 6,
      WQE_CHUNK_SIZE      = 1ul << WQE_CHUNK_SIZE_LOG2,
      WQE_CHUNK_MASK      = WQE_CHUNK_SIZE - 1
     };



typedef struct
{
  IB_wqpn_t       qpn;
  HH_srq_hndl_t   srqn; /* If invalid, equals HH_EINVAL_SRQ_HNDL */   
  u_int32_t       pd;
  u_int32_t       rdd;
  u_int32_t       cqn_snd;
  u_int32_t       cqn_rcv;

  /* The following is kept merely to support query */
  VAPI_qp_cap_t   cap;

  /* kept to support modify-qp/query-qp */
  VAPI_rdma_atom_acl_t  remote_atomic_flags;/* Enable/Disable RDMA and atomic */
  u_int8_t              qp_ous_rd_atom;      /* Maximum number of oust. RDMA read/atomic as target */
  u_int8_t              ous_dst_rd_atom;     /* Number of outstanding RDMA rd/atomic ops at destination */
  
  u_int8_t        st; /* sufficient for  XHH_service_type_t - enum */
  VAPI_qp_state_t state;
  unsigned int    ssc:1;
  unsigned int    rsc:1;

  MT_virt_addr_t  wqes_buf;     /* WQEs buffer virtual address */
  MT_size_t       wqes_buf_sz;
  VAPI_lkey_t     lkey;
  XHH_uar_index_t uar_index;    /* index of UAR used for this QP */
  MT_bool         external_wqe_buf;

  XHH_DEV_SPECIFIC_QPC_ENTRY_DATA

#if defined(MT_SUSPEND_QP)
  MT_bool         is_suspended;
#endif

} TQPM_sw_qpc_t;


typedef struct
{
  signed char  tab[VAPI_ERR+1][VAPI_ERR+1]; /* -1 or XHH_qpee_transition_t */
} State_machine;

static const VAPI_special_qp_t qp_types[] = {
  VAPI_SMI_QP, VAPI_GSI_QP,          /* _Used_ for special QPs */
  VAPI_RAW_IPV6_QP, VAPI_RAW_ETY_QP  /* Not used: but supported for query */
};
enum {n_qp_types = ARR_SIZE(qp_types)};

typedef struct
{
  u_int8_t                 n_ports;
  MT_bool                  configured;
  IB_wqpn_t                first_sqp_qpn;     /* Above FW reserved QPCs */
  XHH_port_init_props_t*   port_props;
  TQPM_sw_qpc_t**          sqp_ctx;   /* Context for special QPs is outside of the VIP_array */
} Special_QPs;


/* The main QP-manager structure 
 * Note that {(EPool_t flist), (XHH_lamm_t lamm)}
 * have their own mutex-es controlling multi-threads.
 */
typedef struct XHH_qpm_st
{
  /* Capabilities */
  u_int8_t        log2_max_qp;
  u_int32_t       rdb_base_index;
  u_int8_t        log2_max_outs_rdma_atom;
  u_int8_t        log2_max_outs_dst_rd_atom;
  u_int32_t       max_outs_rdma_atom; /* convenient 2^log2_max_outs_rdma_atom */
  u_int32_t       idx_mask;           /* convenient (2^log2_max_qp) - 1 */

  /* SW resources tables */
  Special_QPs     sqp_info;
  VIP_array_p_t   qp_tbl;             /* Index 0 into the table is first_sqp_qpn+NUM_SQP */
  u_int32_t       first_rqp;          /* First regular QP */
  u_int8_t*       qpn_prefix;         /* persistant 8 bit prefix change for QP numbers */
  /* In order to avoid holding to much memory for QP numbers prefix we:
   * 1) Manipulate only upper 8 bits of the 24 bit QPN 
   * 2) Hold only MAX_QPN_PREFIX numbers to be shared among QPs with the same lsb of index */
  MOSAL_mutex_t   mtx; /* protect sqp_ctx and pkey/sgid tables */

  /* convenient handle saving  */
  XHH_hob_t       hob;
  XHH_cmd_t       cmd_if;
  
  XHH_mrwm_t      mrwm_internal;
  XHH_lamm_t      lamm;
  
  /* mirror of the port gid tbl - for special qps*/
  IB_gid_t* sgid_tbl[NUM_PORTS];
  u_int16_t num_sgids[NUM_PORTS];

  /* mirror of the port qp1 pkey - values are kept CPU endian - little endian*/
  VAPI_pkey_t*    pkey_tbl[NUM_PORTS];   
  u_int16_t       pkey_tbl_sz[NUM_PORTS];
  VAPI_pkey_ix_t  qp1_pkey_idx[NUM_PORTS];
  MT_bool         port_active[NUM_PORTS];

  XHH_DEV_SPECIFIC_QPM_DATA
  
} TQPM_t;

#endif
