/*
  This software is available to you under a choice of one of two
  licenses.  You may choose to be licensed under the terms of the GNU
  General Public License (GPL) Version 2, available at
  <http://www.fsf.org/copyleft/gpl.html>, or the OpenIB.org BSD
  license, available in the LICENSE.TXT file accompanying this
  software.  These details are also available at
  <http://openib.org/license.html>.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
  BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
  ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
  CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
  SOFTWARE.

  Copyright (c) 2004 Mellanox Technologies Ltd.  All rights reserved.
*/

#define C_CIDB_C
#include <ring_dbell.h>
#include <MT25218.h>
#include "cidb.h"

/* Doorbells dword offsets */
#define UAR_RD_SEND_DBELL_OFFSET (MT_BYTE_OFFSET(arbelprm_uar_st,rd_send_doorbell)>>2)
#define UAR_SEND_DBELL_OFFSET (MT_BYTE_OFFSET(arbelprm_uar_st,send_doorbell)>>2)
#define UAR_CQ_CMD_DBELL_OFFSET (MT_BYTE_OFFSET(arbelprm_uar_st,cq_command_doorbell)>>2)
/* Doorbells dword offsets */
#define UAR_RD_SEND_DBELL_SZ (MT_BYTE_SIZE(arbelprm_uar_st,rd_send_doorbell)>>2)
#define UAR_SEND_DBELL_SZ (MT_BYTE_SIZE(arbelprm_uar_st,send_doorbell)>>2)
#define UAR_CQ_CMD_DBELL_SZ (MT_BYTE_SIZE(arbelprm_uar_st,cq_command_doorbell)>>2)

/* UAR Context doorbell record size constants (in dwords) */
#define UAR_CTX_CQ_CI_DBELL_SZ (MT_BYTE_SIZE(arbelprm_arbel_prm_st,cq_ci_db_record)>>2)
#define UAR_CTX_CQ_ARM_DBELL_SZ (MT_BYTE_SIZE(arbelprm_arbel_prm_st,cq_arm_db_record)>>2)
#define UAR_CTX_QP_DBELL_SZ (MT_BYTE_SIZE(arbelprm_arbel_prm_st,qp_db_record)>>2)

struct XHH_cidb_st {
  XHH_ver_info_t  ver_info;
  XHH_uar_index_t   uar_index;
  volatile u_int32_t *uar_base;
  MOSAL_spinlock_t  uar_lock;
} /* *XHH_cidb_t */;



/************************************************************************/
/*                    Public functions                                  */ 
/************************************************************************/



HH_ret_t  XHH_cidb_create(
  /*IN*/  XHH_ver_info_t  *version_p, 
  /*IN*/  XHH_uar_index_t uar_index, 
  /*IN*/  void            *uar_base, 
  /*OUT*/ XHH_cidb_t       *cidb_p
  )
{
  XHH_cidb_t new_cidb;

  new_cidb= (XHH_cidb_t)MALLOC(sizeof(struct XHH_cidb_st));
  if (new_cidb == NULL) {
    MTL_ERROR1(MT_FLFMT("%s: Failed allocating object memory."), __func__);
    return HH_EAGAIN;
  }
  new_cidb->uar_index= uar_index;
  new_cidb->uar_base= (volatile u_int32_t*)uar_base;
  memcpy(&(new_cidb->ver_info),version_p,sizeof(XHH_ver_info_t));
  MOSAL_spinlock_init(&(new_cidb->uar_lock));
  
  *cidb_p= new_cidb;
  return HH_OK;
}


void XHH_cidb_destroy(/*IN*/ XHH_cidb_t cidb)
{
  FREE(cidb);
}


HH_ret_t XHH_cidb_get_uar_index(
  /*IN*/ XHH_cidb_t cidb, 
  /*OUT*/ XHH_uar_index_t *uar_index_p
  )
{
  if (cidb == NULL)  return HH_EINVAL;
  *uar_index_p= cidb->uar_index;
  return HH_OK;
}


HH_ret_t  XHH_cidb_qp_sendq(
  /*IN*/  XHH_cidb_t              cidb, 
  /*IN*/  void*                   db_record_p,
  /*IN*/  IB_wqpn_t               qpn,
  /*IN*/  XHH_cidb_sendq_dbell_t* sendq_dbell_p
  )
{
#if 0
  return HH_ENOSYS;
#else
  volatile u_int32_t dbell_draft[UAR_SEND_DBELL_SZ]= {0};
  
  // ER: Update dbell record first
  XHH_cidb_qp_update_counter(db_record_p,sendq_dbell_p->first_wqe_counter +
							 sendq_dbell_p->wqe_cnt); 

  /*now, ring the UAR doorbell */
  
  /* first dword */
  MT_INSERT_ARRAY32(dbell_draft, sendq_dbell_p->wqe_cnt,
    MT_BIT_OFFSET(arbelprm_send_doorbell_st, wqe_cnt),
    MT_BIT_SIZE(arbelprm_send_doorbell_st, wqe_cnt));
  
  MT_INSERT_ARRAY32(dbell_draft, sendq_dbell_p->first_wqe_counter,
    MT_BIT_OFFSET(arbelprm_send_doorbell_st, wqe_counter),
    MT_BIT_SIZE(arbelprm_send_doorbell_st, wqe_counter));
  
  MT_INSERT_ARRAY32(dbell_draft, sendq_dbell_p->fence ? 1 : 0,
    MT_BIT_OFFSET(arbelprm_send_doorbell_st, f),
    MT_BIT_SIZE(arbelprm_send_doorbell_st, f));
  
  MT_INSERT_ARRAY32(dbell_draft, sendq_dbell_p->nopcode,
    MT_BIT_OFFSET(arbelprm_send_doorbell_st, nopcode),
    MT_BIT_SIZE(arbelprm_send_doorbell_st, nopcode));
  
  /* second dword */
  MT_INSERT_ARRAY32(dbell_draft, qpn,
    MT_BIT_OFFSET(arbelprm_send_doorbell_st, qpn),
    MT_BIT_SIZE(arbelprm_send_doorbell_st, qpn));

  MT_INSERT_ARRAY32(dbell_draft, sendq_dbell_p->next_size,
    MT_BIT_OFFSET(arbelprm_send_doorbell_st, nds),
    MT_BIT_SIZE(arbelprm_send_doorbell_st ,nds));

  RING_DBELL_2DW(cidb,UAR_SEND_DBELL_OFFSET,dbell_draft);

  return HH_OK;
#endif
} 



HH_ret_t  XHH_cidb_sendq_rd(
  /*IN*/  XHH_cidb_t              cidb, 
  /*IN*/  void*                   db_record_p,
  /*IN*/  IB_wqpn_t               qpn,
  /*IN*/  IB_eecn_t               een,             
  /*IN*/  XHH_cidb_sendq_dbell_t* sendq_dbell_p
  )
{
#if 1
  return HH_ENOSYS;
#else
  volatile u_int32_t dbell_draft[UAR_RD_SEND_DBELL_SZ]= {0};
  
  MT_INSERT_ARRAY32(dbell_draft, een,
    MT_BIT_OFFSET(tavorprm_rd_send_doorbell_st,een),
    MT_BIT_SIZE(tavorprm_rd_send_doorbell_st,qpn));
  MT_INSERT_ARRAY32(dbell_draft, sendq_dbell_p->qpn,
    MT_BIT_OFFSET(tavorprm_rd_send_doorbell_st,qpn),
    MT_BIT_SIZE(tavorprm_rd_send_doorbell_st,qpn));
  dbell_draft[MT_BYTE_OFFSET(tavorprm_rd_send_doorbell_st,snd_params.nda)>>2]=
    sendq_dbell_p->next_addr_32lsb; /* 6 ls-bits will be masked anyway by f and nopcode */
  MT_INSERT_ARRAY32(dbell_draft, sendq_dbell_p->fence ? 1 : 0,
    MT_BIT_OFFSET(tavorprm_rd_send_doorbell_st,snd_params.f),
    MT_BIT_SIZE(tavorprm_rd_send_doorbell_st,snd_params.f));
  MT_INSERT_ARRAY32(dbell_draft, sendq_dbell_p->nopcode,
    MT_BIT_OFFSET(tavorprm_rd_send_doorbell_st,snd_params.nopcode),
    MT_BIT_SIZE(tavorprm_rd_send_doorbell_st,snd_params.nopcode));
  MT_INSERT_ARRAY32(dbell_draft, sendq_dbell_p->qpn,
    MT_BIT_OFFSET(tavorprm_rd_send_doorbell_st,snd_params.qpn),
    MT_BIT_SIZE(tavorprm_rd_send_doorbell_st,snd_params.qpn));
  MT_INSERT_ARRAY32(dbell_draft, sendq_dbell_p->next_size,
    MT_BIT_OFFSET(tavorprm_rd_send_doorbell_st,snd_params.nds),
    MT_BIT_SIZE(tavorprm_rd_send_doorbell_st,snd_params.nds));

  RING_DBELL_4DW(uar,UAR_RD_SEND_DBELL_OFFSET,dbell_draft);

  return HH_OK;
#endif
} 


HH_ret_t XHH_cidb_qp_init_uar_ctx_entry(
  /*IN*/  void*                  db_record_p,
  /*IN*/  IB_wqpn_t              qpn,
  /*IN*/  XHH_cmd_res_type_t     res
  )
{
#if 0
  return HH_ENOSYS;
#else
   volatile u_int32_t ctx_dbell_draft[UAR_CTX_QP_DBELL_SZ] = {0};

   MT_INSERT_ARRAY32(ctx_dbell_draft, (u_int32_t)res,
     MT_BIT_OFFSET(arbelprm_qp_db_record_st,res),
     MT_BIT_SIZE(arbelprm_qp_db_record_st,res));
   
   MT_INSERT_ARRAY32(ctx_dbell_draft, qpn,
     MT_BIT_OFFSET(arbelprm_qp_db_record_st,qp_number),
     MT_BIT_SIZE(arbelprm_qp_db_record_st,qp_number));
   
   /* counter word is initialized to zero */
   (*(volatile u_int32_t*)(db_record_p)) = 0;
   (*(volatile u_int32_t*)(((u_int32_t *)db_record_p)+1)) = MOSAL_cpu_to_be32(*(ctx_dbell_draft+1));
   
   return HH_OK;
#endif    
} 


HH_ret_t XHH_cidb_cq_init_cntr(
  /*IN*/  void*             db_record_p,
  /*IN*/  HH_cq_hndl_t      cqn
)
{
#if 0
  return HH_ENOSYS;
#else
   volatile u_int32_t dbell_draft[UAR_CTX_CQ_CI_DBELL_SZ] = {0};

   MT_INSERT_ARRAY32(dbell_draft, (u_int32_t)ARBEL_IF_DB_RECORD_CI,
     MT_BIT_OFFSET(arbelprm_cq_ci_db_record_st,res),
     MT_BIT_SIZE(arbelprm_cq_ci_db_record_st,res));
   
   MT_INSERT_ARRAY32(dbell_draft, cqn,
     MT_BIT_OFFSET(arbelprm_cq_ci_db_record_st,cq_number),
     MT_BIT_SIZE(arbelprm_cq_ci_db_record_st,cq_number));
   
   /* counter word is initialized to zero */
   (*(volatile u_int32_t*)(db_record_p)) = 0;
   (*(volatile u_int32_t*)(((u_int32_t *)db_record_p)+1)) = MOSAL_cpu_to_be32(*(dbell_draft+1));

   return HH_OK;
#endif    
}

HH_ret_t XHH_cidb_cq_init_arm(
  /*IN*/  void*             db_record_p,
  /*IN*/  HH_cq_hndl_t      cqn
)
{
   volatile u_int32_t dbell_draft[UAR_CTX_CQ_ARM_DBELL_SZ] = {0};

   MT_INSERT_ARRAY32(dbell_draft, (u_int32_t)ARBEL_IF_DB_RECORD_ARM,
     MT_BIT_OFFSET(arbelprm_cq_ci_db_record_st,res),
     MT_BIT_SIZE(arbelprm_cq_ci_db_record_st,res));
   
   MT_INSERT_ARRAY32(dbell_draft, cqn,
     MT_BIT_OFFSET(arbelprm_cq_ci_db_record_st,cq_number),
     MT_BIT_SIZE(arbelprm_cq_ci_db_record_st,cq_number));
   
   /* counter word is initialized to zero */
   (*(volatile u_int32_t*)(db_record_p)) = 0;
   (*(volatile u_int32_t*)(((u_int32_t *)db_record_p)+1)) = MOSAL_cpu_to_be32(*(dbell_draft+1));

   return HH_OK;
}




/************************************************************************
 *  Function: XHH_cidb_cq_arm
 *
 *  Arguments:
 *    cidb - The XHH_cidb object handle
 *    db_record_p - Pointer to doorbell record 
 *    cqn  - CQ number
 *    arm_type - CQ arm type (solicited/next/N)
 *    cq_cntr - CQ counter (For ARM-N should be the counter to trigger event)
 *
 *  Returns:
 *    HH_OK
 *    HH_EINVAL -Invalid handles
 *
 *  Description:
 *    Ring CQ_ARM doorbell
 */
HH_ret_t XHH_cidb_cq_arm(
  /*IN*/  XHH_cidb_t        cidb, 
  /*IN*/  void*             db_record_p,
  /*IN*/  HH_cq_hndl_t      cqn,
  /*IN*/  XHH_uar_cq_cmd_t  arm_type,
  /*IN*/  u_int32_t         cq_cntr,
  /*IN*/  u_int8_t          command_seq_no
)
{
#if 0
  return HH_ENOSYS;
#else
  volatile u_int32_t dbell_draft[UAR_CQ_CMD_DBELL_SZ]= {0};
  volatile u_int32_t ctx_dbell_draft[UAR_CTX_CQ_CI_DBELL_SZ] = {0};

  // ER: Update dbell context record first

  MT_INSERT_ARRAY32(ctx_dbell_draft, cq_cntr,
    MT_BIT_OFFSET(arbelprm_cq_arm_db_record_st,counter),
    MT_BIT_SIZE(arbelprm_cq_arm_db_record_st,counter));

  /* create UAR context doorbell */
  MT_INSERT_ARRAY32(ctx_dbell_draft, cqn,
    MT_BIT_OFFSET(arbelprm_cq_arm_db_record_st,cq_number),
    MT_BIT_SIZE(arbelprm_cq_arm_db_record_st,cq_number));
  
  MT_INSERT_ARRAY32(ctx_dbell_draft, (u_int32_t)ARBEL_IF_DB_RECORD_ARM,
    MT_BIT_OFFSET(arbelprm_cq_arm_db_record_st,res),
    MT_BIT_SIZE(arbelprm_cq_arm_db_record_st,res));
  
  MT_INSERT_ARRAY32(ctx_dbell_draft, (u_int32_t)arm_type,
    MT_BIT_OFFSET(arbelprm_cq_arm_db_record_st,cmd),
    MT_BIT_SIZE(arbelprm_cq_arm_db_record_st,cmd));

  MT_INSERT_ARRAY32(ctx_dbell_draft, (u_int32_t)command_seq_no,
    MT_BIT_OFFSET(arbelprm_cq_arm_db_record_st,cmd_sn),
    MT_BIT_SIZE(arbelprm_cq_arm_db_record_st,cmd_sn));
  
  /* create UAR doorbell */
  MT_INSERT_ARRAY32(dbell_draft, arm_type,
    MT_BIT_OFFSET(arbelprm_cq_cmd_doorbell_st,cmd),
    MT_BIT_SIZE(arbelprm_cq_cmd_doorbell_st,cmd));
  MT_INSERT_ARRAY32(dbell_draft, cqn,
    MT_BIT_OFFSET(arbelprm_cq_cmd_doorbell_st,cqn),
    MT_BIT_SIZE(arbelprm_cq_cmd_doorbell_st,cqn));
  MT_INSERT_ARRAY32(dbell_draft, cq_cntr,
    MT_BIT_OFFSET(arbelprm_cq_cmd_doorbell_st,cq_param),
    MT_BIT_SIZE(arbelprm_cq_cmd_doorbell_st,cq_param));

  MT_INSERT_ARRAY32(dbell_draft, (u_int32_t)command_seq_no,
    MT_BIT_OFFSET(arbelprm_cq_cmd_doorbell_st,cmd_sn),
    MT_BIT_SIZE(arbelprm_cq_cmd_doorbell_st,cmd_sn));
  
  /*now, ring the UAR doorbells */
  (*(volatile u_int32_t*)(db_record_p)) = MOSAL_cpu_to_be32(*ctx_dbell_draft);
  (*(volatile u_int32_t*)(((u_int32_t *)db_record_p)+1)) = MOSAL_cpu_to_be32(*(ctx_dbell_draft+1));
  
  RING_DBELL_2DW(cidb,UAR_CQ_CMD_DBELL_OFFSET,dbell_draft);

  return HH_OK;
#endif
}

  
