/*
  This software is available to you under a choice of one of two
  licenses.  You may choose to be licensed under the terms of the GNU
  General Public License (GPL) Version 2, available at
  <http://www.fsf.org/copyleft/gpl.html>, or the OpenIB.org BSD
  license, available in the LICENSE.TXT file accompanying this
  software.  These details are also available at
  <http://openib.org/license.html>.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
  BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
  ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
  CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
  SOFTWARE.

  Copyright (c) 2004 Mellanox Technologies Ltd.  All rights reserved.
*/

#ifndef H_CIDB_H
#define H_CIDB_H

#include <mtl_common.h>
#include <ib_defs.h>
#include <mosal.h>
#include <vapi_types.h>
#include <hh.h>
#include <xhh.h>
#include <dev_common.h>
#include <arbel_if_defs.h>

/* Invalidate doorbell record at given address. To be used before destroying QP and CQ */
//#define XHH_INVALIDATE_DB_RECORD(db_record_p)  memset((void*)db_record_p,0,8)

//typedef struct XHH_cidb_st *XHH_cidb_t;
typedef arbel_if_uar_cq_db_cmd_t   XHH_uar_cq_cmd_t;

typedef arbel_if_db_record   XHH_cmd_res_type_t;

#define SENDQ_DBELL_NEXT_SIZE_TYPE	u_int8_t

typedef struct {
  arbel_if_nopcode_t nopcode; /* Next Send descriptor opcode (encoded) */
  MT_bool fence;              /* Fence bit set */
  SENDQ_DBELL_NEXT_SIZE_TYPE next_size  ;       /* Size of next WQE (16-byte chunks) */
  u_int8_t wqe_cnt;           /* Number of WQEs posted with this doorbell */
  u_int16_t first_wqe_counter; /* Counter (mod 64K) of first WQE in attached chain */
} XHH_cidb_sendq_dbell_t;

typedef XHH_cidb_sendq_dbell_t XHH_uar_sendq_dbell_t;
/************************************************************************
 *  Function: XHH_cidb_create
 *
 *  Arguments:
 *     version_p 
 *     uar_index
 *     uar_base - Virtual address mapped to associated UAR
 *     cidb_p -    Created XHH_cidb object handle
 *
 *  Returns:
 *    HH_OK
 *    HH_EAGAIN - Not enough resources to create object
 *    HH_EINVAL - Invalid parameters (NULL ptrs.etc.)
 *
 *  Description: Create the XHH_uar object.
 */
HH_ret_t  XHH_cidb_create(
  /*IN*/  XHH_ver_info_t  *version_p, 
  /*IN*/  XHH_uar_index_t uar_index, 
  /*IN*/  void            *uar_base, 
  /*OUT*/ XHH_cidb_t       *cidb_p
  );


/************************************************************************
 *
 *  Function: XHH_cidb_destroy
 *
 *  Arguments:
 *    cidb - Object handle
 *
 *  Returns:
 *
 *  Description: Free UAR object context.
 */
void XHH_cidb_destroy(/*IN*/ XHH_cidb_t cidb);


/************************************************************************
 *
 *  Function: XHH_cidb_get_uar_index
 *
 *  Arguments:
 *    cidb - Object handle
 *    uar_index_p - Returned UAR page index of UAR associated with this object
 *
 *  Returns:
 *    HH_OK
 *    HH_EINVAL 
 *
 *  Description: Get associated UAR page index.
 */
HH_ret_t XHH_cidb_get_uar_index(
  /*IN*/ XHH_cidb_t cidb, 
  /*OUT*/ XHH_uar_index_t *uar_index_p
  );


/************************************************************************
 *  Function: XHH_cidb_qp_sendq
 *
 *  Arguments:
 *    cidb -           The XHH_uar object handle
 *    db_record_p - Pointer to doorbell record 
 *    qpn  -          QP number to ring doorbell for
 *    sendq_dbell_p - Send queue doorbel data
 *
 *  Returns:
 *    HH_OK
 *    HH_EINVAL -Invalid handles or NULL pointer
 *
 *  Description:
 *    Ring the  send   section of the UAR.
 */
HH_ret_t  XHH_cidb_qp_sendq(
  /*IN*/  XHH_cidb_t              cidb, 
  /*IN*/  void*                   db_record_p,
  /*IN*/  IB_wqpn_t               qpn,
  /*IN*/  XHH_cidb_sendq_dbell_t* sendq_dbell_p
  );


/************************************************************************
 *  Function: XHH_cidb_sendq_rd
 *
 *  Arguments:
 *    cidb -           The XHH_cidb object handle
 *    db_record_p - Pointer to doorbell record 
 *    qpn -           The QP number to ring for
 *    een -           The EE context number for posted request
 *    sendq_dbell_p - Send queue doorbell data
 *
 *  Returns:
 *    HH_OK HH_EINVAL -Invalid handles or NULL pointer
 *  Description:
 *    Ring the  rdd-send   section of the UAR.
 */
HH_ret_t  XHH_cidb_sendq_rd(
  /*IN*/  XHH_cidb_t              cidb, 
  /*IN*/  void*                   db_record_p,
  /*IN*/  IB_wqpn_t               qpn,
  /*IN*/  IB_eecn_t               een,             
  /*IN*/  XHH_cidb_sendq_dbell_t* sendq_dbell_p
  );


/************************************************************************
 *  Function: XHH_cidb_qp_init_uar_ctx_entry
 *
 *  Arguments:
 *    db_record_p - Pointer to doorbell record 
 *    qpn - Receive queue to update
 *    res - Type of resource (SendQ or RecvQ  or SRQ resource)
 *
 *  Returns:
 *    HH_OK
 *    HH_EINVAL -Invalid handles
 *
 *  Description:
 *    initializes a qp UAR context entry (at qp creation time).
 */
HH_ret_t XHH_cidb_qp_init_uar_ctx_entry(
  /*IN*/  void*                  db_record_p,
  /*IN*/  IB_wqpn_t              qpn,
  /*IN*/  XHH_cmd_res_type_t     res
  );


/************************************************************************
 *  Function: XHH_cidb_cq_init_cntr
 *
 *  Arguments:
 *    cidb - The XHH_cidb object handle
 *    db_record_p - Pointer to CI doorbell record 
 *    cqn  - CQ number
 *
 *  Returns:
 *    HH_OK
 *    HH_EINVAL -Invalid handles
 *
 *  Description:
 *    Initialize CQ consumer counter in UAR context.
 */
HH_ret_t XHH_cidb_cq_init_cntr(
  /*IN*/  void*             db_record_p,
  /*IN*/  HH_cq_hndl_t      cqn
);


/************************************************************************
 *  Function: XHH_cidb_cq_init_arm
 *
 *  Arguments:
 *    cidb - The XHH_cidb object handle
 *    db_record_p - Pointer to ARM doorbell record 
 *
 *  Returns:
 *    HH_OK
 *    HH_EINVAL -Invalid handles
 *
 *  Description:
 *    Initialize CQ ARM record in UAR context.
 */
HH_ret_t XHH_cidb_cq_init_arm(
  /*IN*/  void*             db_record_p,
  /*IN*/  HH_cq_hndl_t      cqn
);
/************************************************************************
 *  Function: XHH_cidb_cq_arm
 *
 *  Arguments:
 *    cidb - The XHH_cidb object handle
 *    db_record_p - Pointer to doorbell record 
 *    cqn  - CQ number
 *    arm_type - CQ arm type (solicited/next/N)
 *    cq_cntr - CQ counter (For ARM-N should be the counter to trigger event)
 *
 *  Returns:
 *    HH_OK
 *    HH_EINVAL -Invalid handles
 *
 *  Description:
 *    Ring CQ_ARM doorbell
 */
HH_ret_t XHH_cidb_cq_arm(
  /*IN*/  XHH_cidb_t        cidb, 
  /*IN*/  void*             db_record_p,
  /*IN*/  HH_cq_hndl_t      cqn,
  /*IN*/  XHH_uar_cq_cmd_t  arm_type,
  /*IN*/  u_int32_t         cq_cntr,
  /*IN*/  u_int8_t          command_seq_no
);


/************************************************************************
 *  Function: XHH_cidb_cq_update_cntr
 *
 *  Arguments:
 *    cidb - The XHH_cidb object handle
 *    db_record_p - Pointer to CI doorbell record 
 *    cqn  - CQ number
 *    new_cq_cntr - New CQ counter
 *
 *  Returns:
 *    HH_OK
 *    HH_EINVAL -Invalid handles
 *
 *  Description:
 *    updates CQ consumer counter in UAR context.
 */
inline static void XHH_cidb_cq_update_cntr(
  /*IN*/  void*             db_record_p,
  /*IN*/  u_int32_t         new_cq_cntr
)
{
   (*(volatile u_int32_t*)(db_record_p)) = MOSAL_cpu_to_be32(new_cq_cntr);
}


inline static void XHH_cidb_qp_update_counter(
  /*IN*/  void*             db_record_p,
  /*IN*/  u_int16_t         new_wqe_cntr
)
{
  MTL_DEBUG5(MT_FLFMT("%s: db_r at 0x%p is: 0x%08x, 0x%08x. new cntr=0x%08x"),__func__,
			db_record_p, MOSAL_be32_to_cpu((*(volatile u_int32_t*)(db_record_p))), 
			 MOSAL_be32_to_cpu((*(volatile u_int32_t*)(((u_int32_t *)db_record_p)+1))),
			(u_int32_t)new_wqe_cntr);
   (*(volatile u_int32_t*)(db_record_p)) = MOSAL_cpu_to_be32(((u_int32_t)new_wqe_cntr) & 0x0000FFFF);
}
#endif /* H_CIDB_H */
