/*
  This software is available to you under a choice of one of two
  licenses.  You may choose to be licensed under the terms of the GNU
  General Public License (GPL) Version 2, available at
  <http://www.fsf.org/copyleft/gpl.html>, or the OpenIB.org BSD
  license, available in the LICENSE.TXT file accompanying this
  software.  These details are also available at
  <http://openib.org/license.html>.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
  BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
  ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
  CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
  SOFTWARE.

  Copyright (c) 2004 Mellanox Technologies Ltd.  All rights reserved.
*/

#include <cmdif.h>
#include <cmd_types.h>
#include <cmdif_priv.h>
#include <mtl_common.h>
#include <cmds_wrap_priv.h>
#include <dev_cmds_wrap.h>
 
#if defined(MAX_DEBUG) && XHH_CMDS_WRAP_DEBUG_LEVEL <= MAX_DEBUG
#define XHH_CMD_PRINT_VPM_ENTRIES(a,b)      XHH_cmd_print_vpm_entries(a,b)
#define XHH_CMD_PRINT_QUERY_LAM(a)          XHH_cmd_print_query_lam(a)
#define XHH_CMD_MAILBOX_PRINT(a,b,c)        XHH_cmd_mailbox_print(a,b,c)

static void XHH_cmd_print_vpm_entries( void *mailbox, int num_entries) {}
inline static void XHH_cmd_print_query_lam( XHH_lam_props_t * lam_props_p) 
{
    CMDS_DBG( "LAM start_address = "U64_FMT"\n", lam_props_p->lam_start_adr);
    CMDS_DBG( "LAM lam_end_adr = "U64_FMT"\n", lam_props_p->lam_end_adr);
}

static void XHH_cmd_mailbox_print( void *mailbox, int size, const char * func )
{
    int i, j, maxlines, bytes_left, this_line;
    char linebuf[200], tempout[20];
    u_int8_t *iterator;

    if (mailbox == NULL) return;
    
    iterator = (u_int8_t *)mailbox;
    bytes_left = size;
    MTL_DEBUG4("MailboxPrint from function %s, starting at addr 0x%p, size=%d:\n",
               func, mailbox, size);
    
    if (size <= 0) {
        return;
    }

    maxlines = (size / 16) + ((size % 16) ? 1 : 0);

    for (i = 0; i < maxlines; i++) {
        memset(linebuf, 0, sizeof(linebuf));
        this_line = (bytes_left > 16 ? 16 : bytes_left);

        for (j = 0; j < this_line; j++) {
            if ((j % 4) == 0) {
                strcat(linebuf," ");
            }
            sprintf(tempout, "%02x", *iterator);
            iterator++; bytes_left--;
            strcat(linebuf,tempout);
        }
        MTL_DEBUG4("%s\n", linebuf);
    }
    MTL_DEBUG3("MailboxPrint END\n");
}


#else
#define XHH_CMD_PRINT_VPM_ENTRIES(a,b)
#define XHH_CMD_PRINT_QUERY_LAM(a)
#define XHH_CMD_MAILBOX_PRINT(a,b,c)
#endif

XHH_cmd_status_t XHH_cmd_MAP_FA(/*IN*/ XHH_cmd_t cmd_if, /*IN*/ u_int32_t num_pa_vec_entries,
                                /*IN*/ XHH_cmd_pa_vec_t * pa_vec_p)
{
    command_fields_t cmd_desc;
    u_int8_t *inprm, *iterator;
    int i, local_num_elts, elts_this_loop, max_elts_per_buffer;
    u_int32_t buf_size;
    XHH_cmd_status_t rc = XHH_CMD_STAT_OK;
    u_int64_t  pa;

    FUNC_IN;
    if ( !STACK_OK ) MT_RETURN(XHH_CMD_STAT_EFATAL);
    if (!num_pa_vec_entries) {
        MT_RETURN(XHH_CMD_STAT_BAD_PARAM);
    }

    inprm = (u_int8_t*)MALLOC(MAX_IN_PRM_SIZE);
    if ( !inprm ) {
      MT_COV_MEM_FLTR_START
      MT_RETURN(XHH_CMD_STAT_EAGAIN);
      MT_COV_MEM_FLTR_END
    }
    local_num_elts = num_pa_vec_entries;
    max_elts_per_buffer = MAX_IN_PRM_SIZE / ( PSEUDO_MT_BYTE_SIZE(arbelprm_virtual_physical_mapping_st));

    MTL_DEBUG4("XHH_cmd_MAP_FA: num_pa_vec_entries=%d, local_num_elts = %d, max_elts_per_buffer = %d\n",
               num_pa_vec_entries,local_num_elts, max_elts_per_buffer);
    while(local_num_elts > 0) {

        elts_this_loop = (local_num_elts > max_elts_per_buffer ? max_elts_per_buffer : local_num_elts);

        buf_size = elts_this_loop*PSEUDO_MT_BYTE_SIZE(arbelprm_virtual_physical_mapping_st);
        memset(inprm, 0, buf_size);
        iterator = inprm;
        
        MTL_DEBUG4("XHH_cmd_MAP_FA: elts_this_loop = %d, buf_size = %d\n",elts_this_loop, buf_size);
    
        cmd_desc.in_param = inprm;
        cmd_desc.in_param_size = buf_size;
        cmd_desc.in_trans = TRANS_MAILBOX;
        cmd_desc.input_modifier = elts_this_loop;
        cmd_desc.out_param = 0;
        cmd_desc.out_param_size = 0;
        cmd_desc.out_trans = TRANS_NA;
        cmd_desc.opcode = ARBEL_IF_CMD_MAP_FA;
        cmd_desc.opcode_modifier = 0;
        cmd_desc.exec_time_micro = ARBEL_IF_CMD_ETIME_MAP_FA;
    
        /* copy  */
        for (i = 0; i < elts_this_loop; i++, pa_vec_p++,iterator += PSEUDO_MT_BYTE_SIZE(arbelprm_virtual_physical_mapping_st)) {
            pa = (u_int64_t)(pa_vec_p->phys_addr);
            MTL_DEBUG4("XHH_cmd_MAP_FA: loop %d: lamm phys_addr = "PHYS_ADDR_FMT ",pa ="U64_FMT", log2size = 0x%x\n",
                       i,pa_vec_p->phys_addr,pa, pa_vec_p->log2size);
            INS_FLD64_IMMED_SH(pa, iterator, arbelprm_virtual_physical_mapping_st, pa);
            INS_FLD(pa_vec_p, iterator, arbelprm_virtual_physical_mapping_st, log2size);
        }
        

        XHH_CMD_MAILBOX_PRINT(inprm, buf_size, __func__);
        XHH_CMD_PRINT_VPM_ENTRIES(inprm, elts_this_loop);
        rc = cmd_invoke(cmd_if, &cmd_desc);
        if (rc != XHH_CMD_STAT_OK) {
            FREE(inprm);
            MT_RETURN(rc);
        }
        /* update loop parameters */
        local_num_elts -= elts_this_loop;
    }
    FREE(inprm);
    MT_RETURN(rc);
}

XHH_cmd_status_t XHH_cmd_UNMAP_FA(XHH_cmd_t cmd_if)
{
  command_fields_t cmd_desc;
  XHH_cmd_status_t rc;

  FUNC_IN;
  cmd_desc.in_param = 0;
  cmd_desc.in_param_size = 0;
  cmd_desc.in_trans = TRANS_NA;
  cmd_desc.input_modifier = 0;
  cmd_desc.out_param = 0;
  cmd_desc.out_param_size = 0;
  cmd_desc.out_trans = TRANS_NA;
  cmd_desc.opcode = ARBEL_IF_CMD_UNMAP_FA;
  cmd_desc.opcode_modifier = 0;
  cmd_desc.exec_time_micro = ARBEL_IF_CMD_ETIME_UNMAP_FA;

  rc = cmd_invoke(cmd_if, &cmd_desc);
  MT_RETURN(rc);
}

XHH_cmd_status_t XHH_cmd_MAP_ICM_AUX(/*IN*/ XHH_cmd_t cmd_if, /*IN*/ u_int32_t num_pa_vec_entries,
                                     /*IN*/ XHH_cmd_pa_vec_t * pa_vec_p)
{
    command_fields_t cmd_desc;
    u_int8_t *inprm, *iterator;
    int i, local_num_elts, elts_this_loop, max_elts_per_buffer;
    u_int32_t buf_size;
    XHH_cmd_status_t rc = XHH_CMD_STAT_OK;
    u_int64_t  pa;

    /* TBD:  need to:  a. limit number of entries in the mailbox */
    /*                 b. for performance, if the initial addr is odd, */
    /*                    do one loop of an odd number of elements, then do the rest */

    FUNC_IN;
    if ( !STACK_OK ) MT_RETURN(XHH_CMD_STAT_EFATAL);
    if (!num_pa_vec_entries) {
        MT_RETURN(XHH_CMD_STAT_BAD_PARAM);
    }

    inprm = (u_int8_t*)MALLOC(MAX_IN_PRM_SIZE);
    if ( !inprm ) {
      MT_COV_MEM_FLTR_START
      MT_RETURN(XHH_CMD_STAT_EAGAIN);
      MT_COV_MEM_FLTR_END
    }
    local_num_elts = num_pa_vec_entries;
    max_elts_per_buffer = MAX_IN_PRM_SIZE / ( PSEUDO_MT_BYTE_SIZE(arbelprm_virtual_physical_mapping_st));

    MTL_DEBUG4("XHH_cmd_MAP_ICM_AUX: num_pa_vec_entries=%d, local_num_elts = %d, max_elts_per_buffer = %d\n",
               num_pa_vec_entries,local_num_elts, max_elts_per_buffer);
    while(local_num_elts > 0) {

        elts_this_loop = (local_num_elts > max_elts_per_buffer ? max_elts_per_buffer : local_num_elts);

        buf_size = elts_this_loop*PSEUDO_MT_BYTE_SIZE(arbelprm_virtual_physical_mapping_st);
        memset(inprm, 0, buf_size);
        iterator = inprm;
        
        MTL_DEBUG4("XHH_cmd_MAP_ICM_AUX: elts_this_loop = %d, buf_size = %d\n",elts_this_loop, buf_size);
    
        cmd_desc.in_param = inprm;
        cmd_desc.in_param_size = buf_size;
        cmd_desc.in_trans = TRANS_MAILBOX;
        cmd_desc.input_modifier = elts_this_loop;
        cmd_desc.out_param = 0;
        cmd_desc.out_param_size = 0;
        cmd_desc.out_trans = TRANS_NA;
        cmd_desc.opcode = ARBEL_IF_CMD_MAP_ICM_AUX;
        cmd_desc.opcode_modifier = 0;
        cmd_desc.exec_time_micro = ARBEL_IF_CMD_ETIME_MAP_ICM_AUX;
    
        /* copy  */
        for (i = 0; i < elts_this_loop; i++, pa_vec_p++,iterator += PSEUDO_MT_BYTE_SIZE(arbelprm_virtual_physical_mapping_st)) {
            pa = (u_int64_t)(pa_vec_p->phys_addr);
            INS_FLD64_IMMED_SH(pa, iterator, arbelprm_virtual_physical_mapping_st, pa);
            INS_FLD(pa_vec_p, iterator, arbelprm_virtual_physical_mapping_st, log2size);
        }
        

        XHH_CMD_MAILBOX_PRINT(inprm, buf_size, __func__);
        XHH_CMD_PRINT_VPM_ENTRIES(inprm, elts_this_loop );
        rc = cmd_invoke(cmd_if, &cmd_desc);
        if (rc != XHH_CMD_STAT_OK) {
            FREE(inprm);
            MT_RETURN(rc);
        }
        /* update loop parameters */
        local_num_elts -= elts_this_loop;
    }
    FREE(inprm);
    MT_RETURN(rc);
}

XHH_cmd_status_t XHH_cmd_UNMAP_ICM_AUX(XHH_cmd_t cmd_if)
{
  command_fields_t cmd_desc;
  XHH_cmd_status_t rc;

  FUNC_IN;
  cmd_desc.in_param = 0;
  cmd_desc.in_param_size = 0;
  cmd_desc.in_trans = TRANS_NA;
  cmd_desc.input_modifier = 0;
  cmd_desc.out_param = 0;
  cmd_desc.out_param_size = 0;
  cmd_desc.out_trans = TRANS_NA;
  cmd_desc.opcode = ARBEL_IF_CMD_UNMAP_ICM_AUX;
  cmd_desc.opcode_modifier = 0;
  cmd_desc.exec_time_micro = ARBEL_IF_CMD_ETIME_UNMAP_ICM_AUX;

  rc = cmd_invoke(cmd_if, &cmd_desc);
  MT_RETURN(rc);
}


/*
 *  XHH_cmd_LAM_ENABLE
 */ 
XHH_cmd_status_t XHH_cmd_ENABLE_LAM(XHH_cmd_t cmd_if, XHH_lam_props_t *lam_props)
{
  command_fields_t cmd_desc;
  u_int8_t *outprm;
  XHH_cmd_status_t rc;
  u_int32_t buf_size = PSEUDO_MT_BYTE_SIZE(arbelprm_enable_lam_st);
/*  MOSAL_syncobj_t  sync; */

  FUNC_IN;
  outprm = TNMALLOC(u_int8_t, buf_size);
  if (outprm == NULL) {
    MT_COV_MEM_FLTR_START
    MT_RETURN(XHH_CMD_STAT_EAGAIN);
    MT_COV_MEM_FLTR_END
  }
  memset(outprm, 0, buf_size);
  
/*  MOSAL_syncobj_init(&sync); */
  cmd_desc.in_param = 0;
  cmd_desc.in_param_size = 0;
  cmd_desc.in_trans = TRANS_NA;
  cmd_desc.input_modifier = 0;
  cmd_desc.out_param = outprm;
  cmd_desc.out_param_size = buf_size;
  cmd_desc.out_trans = TRANS_MAILBOX;
  cmd_desc.opcode = ARBEL_IF_CMD_ENABLE_LAM;
  cmd_desc.opcode_modifier = 0;
  cmd_desc.exec_time_micro = ARBEL_IF_CMD_ETIME_ENABLE_LAM;
  
  rc = cmd_invoke(cmd_if, &cmd_desc);
  if ( rc != XHH_CMD_STAT_OK ) {
    FREE(outprm);
    MT_RETURN(rc);
  }

  if ( lam_props ) {
    EX_FLD64(lam_props, outprm, arbelprm_enable_lam_st, lam_start_adr);
    EX_FLD64(lam_props, outprm, arbelprm_enable_lam_st, lam_end_adr);

    EX_BOOL_FLD(lam_props, outprm, arbelprm_enable_lam_st, dh);
    EX_FLD(lam_props, outprm, arbelprm_enable_lam_st, di);
    EX_FLD(lam_props, outprm, arbelprm_enable_lam_st, ap);

    /* TBD handle dimm structs here */
    EX_FLD(lam_props, outprm, arbelprm_enable_lam_st, dimm0.dimmsize);
    EX_FLD(lam_props, outprm, arbelprm_enable_lam_st, dimm0.dimmstatus);


    EX_FLD(lam_props, outprm, arbelprm_enable_lam_st, dimm1.dimmsize);
    EX_FLD(lam_props, outprm, arbelprm_enable_lam_st, dimm1.dimmstatus);


    XHH_CMD_MAILBOX_PRINT(outprm, buf_size, __func__);
    XHH_CMD_PRINT_QUERY_LAM(lam_props);
  }
  FREE(outprm);
/*  MOSAL_syncobj_waiton(&sync,MOSAL_SYNC_TIMEOUT_INFINITE); */
  MT_RETURN(rc);
}
/*
 *  XHH_cmd_LAM_DISABLE
 */ 
XHH_cmd_status_t XHH_cmd_DISABLE_LAM(XHH_cmd_t cmd_if)
{
  command_fields_t cmd_desc;
  XHH_cmd_status_t rc;

  FUNC_IN;
 
  cmd_desc.in_param = 0;
  cmd_desc.in_param_size = 0;
  cmd_desc.in_trans = TRANS_NA;
  cmd_desc.input_modifier = 0;
  cmd_desc.out_param = 0;
  cmd_desc.out_param_size = 0;
  cmd_desc.out_trans = TRANS_NA;
  cmd_desc.opcode = ARBEL_IF_CMD_DISABLE_LAM;
  cmd_desc.opcode_modifier = 0;
  cmd_desc.exec_time_micro = ARBEL_IF_CMD_ETIME_DISABLE_LAM;
  
  rc = cmd_invoke(cmd_if, &cmd_desc);
  MT_RETURN(rc);
}

XHH_cmd_status_t XHH_cmd_ACCESS_LAM(/*IN*/ XHH_cmd_t cmd_if, MT_bool read_flag, u_int32_t num_bytes,
                                    /*INOUT*/ u_int8_t *iobuf_p)
{
  MT_RETURN(XHH_CMD_STAT_BAD_OP);
}

XHH_cmd_status_t XHH_cmd_SET_ICM_SIZE(XHH_cmd_t cmd_if, u_int64_t icm_size, 
                                      u_int64_t *icm_aux_size_p)
{
  command_fields_t cmd_desc;
  XHH_cmd_status_t rc;
  u_int32_t in_param_tmp[2], out_param_tmp[2];
  u_int64_t     ret_val;

  FUNC_IN;

  if (icm_aux_size_p == NULL) {
      MT_RETURN(XHH_CMD_STAT_BAD_PARAM);
  }
  
  in_param_tmp[0]= (u_int32_t)(icm_size >> 32);               /* MS-DWORD */
  in_param_tmp[1]= (u_int32_t)(icm_size & 0xFFFFFFFF);        /* LS-DWORD */  

  cmd_desc.in_param = (u_int8_t *)&(in_param_tmp[0]);
  cmd_desc.in_param_size = sizeof(in_param_tmp);
  cmd_desc.in_trans = TRANS_IMMEDIATE;
  cmd_desc.input_modifier = 0;
  
  cmd_desc.out_trans = TRANS_IMMEDIATE;
  cmd_desc.out_param = (u_int8_t*)out_param_tmp;
  cmd_desc.out_param_size = sizeof(out_param_tmp);
  cmd_desc.opcode = ARBEL_IF_CMD_SET_ICM_SIZE;
  
  cmd_desc.exec_time_micro = ARBEL_IF_CMD_ETIME_SET_ICM_SIZE;

  
  rc = cmd_invoke(cmd_if, &cmd_desc);
  if (rc == XHH_CMD_STAT_OK) {
      ret_val = ( (((u_int64_t)out_param_tmp[0]) << 32) & MAKE_ULONGLONG(0xFFFFFFFF00000000) ) |
                (((u_int64_t)out_param_tmp[1]) & MAKE_ULONGLONG(0x00000000FFFFFFFF));
      *icm_aux_size_p = ret_val;
  }
  MT_RETURN(rc);
}

XHH_cmd_status_t XHH_cmd_MAP_ICM(/*IN*/ XHH_cmd_t cmd_if, 
                                 /*IN*/ XHH_icm_va_t  icm_start_addr,
                                 /*IN*/ u_int32_t num_pa_vec_entries,
                                 /*IN*/ XHH_cmd_pa_vec_t * pa_vec_p)
{
    command_fields_t cmd_desc;
    u_int8_t *inprm, *iterator;
    u_int64_t icm_va = (u_int64_t)icm_start_addr;
    int i, local_num_elts, elts_this_loop, max_elts_per_buffer;
    u_int32_t buf_size;
    XHH_cmd_status_t rc = XHH_CMD_STAT_OK;
    u_int64_t  pa;

    /* TBD:  need to:  a. limit number of entries in the mailbox */
    /*                 b. for performance, if the initial addr is odd, */
    /*                    do one loop of an odd number of elements, then do the rest */

    FUNC_IN;
    if ( !STACK_OK ) MT_RETURN(XHH_CMD_STAT_EFATAL);
    if (!num_pa_vec_entries) {
        MT_RETURN(XHH_CMD_STAT_BAD_PARAM);
    }

    if ((icm_start_addr & MAKE_ULONGLONG(0xFFF))!= 0) {
        /* bad alignment */
        MT_RETURN(XHH_CMD_STAT_BAD_PARAM);
    }

    inprm = (u_int8_t*)MALLOC(MAX_IN_PRM_SIZE);
    if ( !inprm ) {
      MT_COV_MEM_FLTR_START
      MT_RETURN(XHH_CMD_STAT_EAGAIN);
      MT_COV_MEM_FLTR_END
    }
    local_num_elts = num_pa_vec_entries;
    max_elts_per_buffer = MAX_IN_PRM_SIZE / ( PSEUDO_MT_BYTE_SIZE(arbelprm_virtual_physical_mapping_st));

    MTL_DEBUG2("XHH_cmd_MAP_ICM: num_pa_vec_entries=%d, local_num_elts = %d, max_elts_per_buffer = %d\n",
               num_pa_vec_entries,local_num_elts, max_elts_per_buffer);
    while(local_num_elts > 0) {

        elts_this_loop = (local_num_elts > max_elts_per_buffer ? max_elts_per_buffer : local_num_elts);

        buf_size = elts_this_loop*PSEUDO_MT_BYTE_SIZE(arbelprm_virtual_physical_mapping_st);
        memset(inprm, 0, buf_size);
        iterator = inprm;
        
        MTL_DEBUG2("XHH_cmd_MAP_ICM: elts_this_loop = %d, buf_size = %d\n",elts_this_loop, buf_size);
    
        cmd_desc.in_param = inprm;
        cmd_desc.in_param_size = buf_size;
        cmd_desc.in_trans = TRANS_MAILBOX;
        cmd_desc.input_modifier = elts_this_loop;
        cmd_desc.out_param = 0;
        cmd_desc.out_param_size = 0;
        cmd_desc.out_trans = TRANS_NA;
        cmd_desc.opcode = ARBEL_IF_CMD_MAP_ICM;
        cmd_desc.opcode_modifier = 0;
        cmd_desc.exec_time_micro = ARBEL_IF_CMD_ETIME_MAP_ICM_AUX;
    
        /* copy  */
        for (i = 0; i < elts_this_loop; i++, pa_vec_p++,iterator += PSEUDO_MT_BYTE_SIZE(arbelprm_virtual_physical_mapping_st)) {
            pa = (u_int64_t)(pa_vec_p->phys_addr);
            INS_FLD64_IMMED_SH(pa, iterator, arbelprm_virtual_physical_mapping_st, pa);
            INS_FLD64_IMMED_SH(icm_va, iterator, arbelprm_virtual_physical_mapping_st, va);
            INS_FLD(pa_vec_p, iterator, arbelprm_virtual_physical_mapping_st, log2size);
            icm_va += (MAKE_ULONGLONG(1) << (pa_vec_p->log2size + 12));
        }

        XHH_CMD_MAILBOX_PRINT(inprm, buf_size, __func__);
        XHH_CMD_PRINT_VPM_ENTRIES(inprm, elts_this_loop);
        rc = cmd_invoke(cmd_if, &cmd_desc);
        if (rc != XHH_CMD_STAT_OK) {
            FREE(inprm);
            MT_RETURN(rc);
        }
        /* update loop parameters */
        local_num_elts -= elts_this_loop;
    }
    FREE(inprm);
    MT_RETURN(rc);
}

XHH_cmd_status_t XHH_cmd_UNMAP_ICM(/*IN*/ XHH_cmd_t cmd_if, /*IN*/ XHH_icm_va_t  icm_start_addr,
                                   /*IN*/ u_int32_t page_count)
{
    command_fields_t cmd_desc;
    XHH_cmd_status_t rc;
    u_int32_t in_param_tmp[2];

    FUNC_IN;

    in_param_tmp[0]= (u_int32_t)(icm_start_addr >> 32);               /* MS-DWORD */
    in_param_tmp[1]= (u_int32_t)(icm_start_addr & 0xFFFFFFFF);        /* LS-DWORD */  

    cmd_desc.in_param = (u_int8_t *)&(in_param_tmp[0]);
    cmd_desc.in_param_size = sizeof(in_param_tmp);
    cmd_desc.in_trans = TRANS_IMMEDIATE;
    cmd_desc.input_modifier = page_count;

    cmd_desc.out_trans = TRANS_NA;
    cmd_desc.out_param = 0;
    cmd_desc.out_param_size = 0;
    cmd_desc.opcode = ARBEL_IF_CMD_UNMAP_ICM;

    cmd_desc.exec_time_micro = ARBEL_IF_CMD_ETIME_UNMAP_ICM;


    rc = cmd_invoke(cmd_if, &cmd_desc);
    MT_RETURN(rc);
}

XHH_cmd_status_t XHH_cmd_NOP(/*IN*/ XHH_cmd_t cmd_if, /*IN*/ u_int8_t log2interval)
{
    command_fields_t cmd_desc;
    XHH_cmd_status_t rc;

    FUNC_IN;
    
    if ((log2interval > 0x14) && (log2interval != 0x1F)) {
        MT_RETURN(XHH_CMD_STAT_BAD_PARAM);
    }

    cmd_desc.in_param = 0;
    cmd_desc.in_param_size = 0;
    cmd_desc.in_trans = TRANS_NA;
    cmd_desc.input_modifier = log2interval;

    cmd_desc.out_trans = TRANS_NA;
    cmd_desc.out_param = 0;
    cmd_desc.out_param_size = 0;
    cmd_desc.opcode = ARBEL_IF_CMD_NOP;

    cmd_desc.exec_time_micro = ARBEL_IF_CMD_ETIME_NOP;


    rc = cmd_invoke(cmd_if, &cmd_desc);
    MT_RETURN(rc);
}

XHH_cmd_status_t XHH_cmd_RUN_FW(/*IN*/ XHH_cmd_t cmd_if)
{
    command_fields_t cmd_desc;
    XHH_cmd_status_t rc;

    FUNC_IN;
    
    cmd_desc.in_param = 0;
    cmd_desc.in_param_size = 0;
    cmd_desc.in_trans = TRANS_NA;
    cmd_desc.input_modifier = 0;

    cmd_desc.out_trans = TRANS_NA;
    cmd_desc.out_param = 0;
    cmd_desc.out_param_size = 0;
    cmd_desc.opcode = ARBEL_IF_CMD_RUN_FW;

    cmd_desc.exec_time_micro = ARBEL_IF_CMD_ETIME_RUN_FW;


    rc = cmd_invoke(cmd_if, &cmd_desc);
    MT_RETURN(rc);
}

 
/*
 *  XHH_cmd_ACCESS_DDR
 */ 
static inline void build_inject_hcamem_err_input_modifier(u_int8_t * inmod,
							XHH_acc_hcamem_inj_err_inp_mod_t *access_hcamem_p)
{
    INS_BOOL_FLD(access_hcamem_p, inmod, arbelprm_access_lam_inject_errors_input_modifier_st, q0);
    INS_FLD(access_hcamem_p, inmod, arbelprm_access_lam_inject_errors_input_modifier_st, index0);
    INS_BOOL_FLD(access_hcamem_p, inmod, arbelprm_access_lam_inject_errors_input_modifier_st, q1);
    INS_FLD(access_hcamem_p, inmod, arbelprm_access_lam_inject_errors_input_modifier_st, index1);
    INS_BOOL_FLD(access_hcamem_p, inmod, arbelprm_access_lam_inject_errors_input_modifier_st, q2);
    INS_FLD(access_hcamem_p, inmod, arbelprm_access_lam_inject_errors_input_modifier_st, index2);
    INS_BOOL_FLD(access_hcamem_p, inmod, arbelprm_access_lam_inject_errors_input_modifier_st, q3);
    INS_FLD(access_hcamem_p, inmod, arbelprm_access_lam_inject_errors_input_modifier_st, index3);

}

XHH_cmd_status_t XHH_cmd_INJECT_HCAMEM_ERR(XHH_cmd_t cmd_if, 
				    XHH_acc_hcamem_inj_err_inp_param_t *acc_hcamem_inj_err_param_p, 
				    XHH_acc_hcamem_inj_err_inp_mod_t *acc_hcamem_inj_err_mod_p)
{
  command_fields_t cmd_desc;
  XHH_cmd_status_t rc;
  u_int32_t  inp_param[2] = {0};
  u_int8_t   *inparam = (u_int8_t *)&(inp_param[0]);

  FUNC_IN;

  INS_FLD(acc_hcamem_inj_err_param_p, inparam, arbelprm_access_lam_inject_errors_input_parameter_st, da);
  INS_FLD(acc_hcamem_inj_err_param_p, inparam, arbelprm_access_lam_inject_errors_input_parameter_st, ba);
  INS_FLD(acc_hcamem_inj_err_param_p, inparam, arbelprm_access_lam_inject_errors_input_parameter_st, ca);
  INS_FLD(acc_hcamem_inj_err_param_p, inparam, arbelprm_access_lam_inject_errors_input_parameter_st, ra);

  cmd_desc.in_param = (u_int8_t *)&(inp_param[0]);
  cmd_desc.in_param_size = sizeof(inp_param);
  cmd_desc.in_trans = TRANS_IMMEDIATE;
  build_inject_hcamem_err_input_modifier((u_int8_t *)&(cmd_desc.input_modifier),acc_hcamem_inj_err_mod_p);
  cmd_desc.out_param = 0;
  cmd_desc.out_param_size = 0;
  cmd_desc.out_trans = TRANS_NA;
  cmd_desc.opcode = ADD_DEV_PREFIX(IF_CMD_ACCESS_LAM);
  cmd_desc.opcode_modifier = 1;
  cmd_desc.exec_time_micro = ADD_DEV_PREFIX(IF_CMD_ETIME_CLASS_C);

  rc = cmd_invoke(cmd_if, &cmd_desc);
  MT_RETURN(rc);
}


