/*
  This software is available to you under a choice of one of two
  licenses.  You may choose to be licensed under the terms of the GNU
  General Public License (GPL) Version 2, available at
  <http://www.fsf.org/copyleft/gpl.html>, or the OpenIB.org BSD
  license, available in the LICENSE.TXT file accompanying this
  software.  These details are also available at
  <http://openib.org/license.html>.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
  BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
  ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
  CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
  SOFTWARE.

  Copyright (c) 2004 Mellanox Technologies Ltd.  All rights reserved.
*/

#ifndef H_DEV_CMD_TYPES_H
#define H_DEV_CMD_TYPES_H

#include <arbel_if_defs.h>

#define XHH_DEV_SPECIFIC_FW_PROPS \
  u_int8_t  dt;             /* Debug Trace Support: 0-Debug trace is not supported. 1-Debug trace is supported.*/ \
  u_int16_t fw_pages; /* Amount of physical memory to be allocated for FW usage in 4KB blocks */ \
  MT_phys_addr_t clr_int_base_addr;\
  MT_phys_addr_t eq_arm_base_addr; \
  MT_phys_addr_t eq_set_ci_base_addr; \
  u_int8_t wqe_h_mode; \
  u_int8_t zb_wq_cq;   
  
enum {
  LAM_STAT_OK = 0,
  LAM_STAT_CAL_1_ERR = 1,
  LAM_STAT_CAL_2_ERR = 2
};
typedef u_int32_t XHH_lam_cal_status_t;

enum {
  DDR_NO_AUTO=0,
  DDR_AUTO_PRECHARGE_PER_TRANSLATION=1,
  DDR_AUTO_PRECHARGE_PER_64BIT=2
};
typedef u_int32_t XHH_dim_apm_t;

#define XHH_ADAPTER_MEM_PROPS_TYPE XHH_lam_props_t 

#define XHH_DEV_SPECIFIC_ADAPTER_MEM_PROPS \
  u_int64_t lam_start_adr;  /* LAM memory start address */\
  u_int64_t lam_end_adr;	/* LAM memory end address   */
  
#define XHH_DEV_SPECIFIC_DEV_LIMS \
  u_int8_t log_max_mtts;	/* Log2 of the Maximum number of MTT entries  */\
  MT_bool   bmme; /* Base Memory Management Extension Support             */\
  MT_bool   win_type;/* Bound Type 2 Memory Window Association mechanism: */\
                            /* 0 - Type 2A - QP Number Association; or    */\
                           /* 1 - Type 2B - QP Number and PD Association. */\
  MT_bool   mps;  /* Ability of this HCA to support multiple page sizes   */\
                  /* per Memory Region.                                   */\
  MT_bool   bl;   /* Ability of this HCA to support                       */\
                  /* Block List Physical Buffer Lists.                    */\
  MT_bool   zb;   /* Zero Based region/windows supported                  */\
  MT_bool   lif;  /* Ability of this HCA to support                       */\
                  /* Local Invalidate Fencing.                            */\
  u_int8_t  log_pbl_sz; /* Log2 of the Maximum Physical Buffer List size  */\
                        /* in Bytes supported by this HCA when invoking   */\
                        /* the Allocate L_Key verb.                       */\
  u_int32_t resd_lkey;  /* The value of the reserved Lkey for             */\
                        /* Base Memory Management Extension               */\
  MT_bool   lamr; /* When set the device requires local attached memory   */\
                  /* in order to operate. In this case, ICM pages,        */\
                  /* Firmware Area and ICM auxiliary pages must be        */\
                  /* allocated in the local attached memory.              */\
  u_int64_t max_icm_size;  /* maximum ICM size InfiniHost III Ex support  */\
                           /* in bytes                                    */\
  u_int16_t mtt_entry_sz;                                                   \
  u_int16_t mpt_entry_sz;                                                   \
  u_int8_t  log2_rsvd_rdbs;                                                 \
  u_int8_t  max_sg_sq;                                                      \
  u_int8_t  max_sg_rq;                                                      \
  u_int16_t max_desc_sz_sq;                                                 \
  u_int16_t max_desc_sz_rq;


#define  XHH_DEV_SPECIFIC_CQC \
  u_int32_t consumer_counter; /* Consumer counter is a 32bits counter     */\
                              /* that is incremented for each CQE pooled  */\
                         /* from the CQ. Must be 0x0 in CQ initialization.*/\
  u_int32_t producer_counter; /* Producer counter is a 32bits counter that*/\
                              /* is incremented for each CQE that is      */\
                              /* written by the HW to the CQ. CQ overrun  */\
                              /* is reported if Producer_counter + 1      */\
                              /* equals to Consumer_counter. Maintained   */\
                              /* by HW (valid for the QUERY_CQ and        */\
                              /* HW2SW_CQ commands only)                  */\
  u_int32_t cq_ci_db_record;  /* Index in the UAR Context Table Entry. HW */\
                              /* uses this index as an offset from the UAR*/\
							  /* Context Table Entry in order to read this*/\
							  /* CQ Consumer Counter doorbell record. This*/\
							  /* value can be retrieved from the HW in the*/\
							  /* QUERY_CQ command.						  */\
  u_int32_t cq_state_db_record;  /* Index in the UAR Context Table Entry. HW */\
                              /* uses this index as an offset from the UAR*/\
							  /* Context Table Entry in order to read this*/\
							  /* CQ State doorbell record. This value can */\
							  /* be retrieved from the HW in the QUERY_CQ command.*/
  
#define  XHH_DEV_SPECIFIC_QPEE_CONTEXT  \
  u_int8_t log_sq_stride; /* Stride (max WQE size) on the send queue.      */\
                          /* WQ entry is 16*(2^log_SQ_stride) bytes.       */\
                          /*Stride must be equal or bigger than 64 bytes   */\
                          /*(minimum log_RQ_stride value allowed is 2).    */\
  u_int8_t log_sq_size;  /*Log2 of the Number of WQEs in the Send Queue.   */\
  u_int8_t log_rq_stride; /* Stride (max WQE size) on the receive queue.   */\
                          /* WQ entry is 16*(2^log_RQ_stride) bytes.       */\
                          /*Stride must be equal or bigger than 64 bytes   */\
                          /*(minimum log_RQ_stride value allowed is 2).    */\
  u_int8_t log_rq_size;  /*Log2 of the Number of WQEs in the Recv Queue.   */\
  u_int32_t wqe_base_adr_h;/* Bits 63:32 of WQE address for both SQ and RQ.*/\
                           /* Reserved for EE context.                     */\
  u_int32_t	snd_wqe_base_adr_l; /* While opening (creating) the WQ,        */\
                                /* this field should contain the address   */\
                                /* of first descriptor to be posted.       */\
                                /* Not valid (reserved) in EE context.     */\
  u_int32_t	rcv_wqe_base_adr_l; /* While opening (creating) the WQ,        */\
                                /* this field should contain the address   */\
                                /* of first descriptor to be posted.       */\
                                /* Not valid (reserved) in EE context.     */\
  u_int32_t snd_db_record_index;/* Index in the UAR Context Table Entry.   */\
                                /*HW uses this index as an offset from the */\
                                /*UAR Context Table Entry in order to read */\
                                /*this SQ doorbell record.The entry is     */\
                                /*obtained via the usr_page field.         */\
                                /* Not valid for EE.                       */\
  u_int32_t rcv_db_record_index;/* Index in the UAR Context Table Entry.   */\
                                /*HW uses this index as an offset from the */\
                                /*UAR Context Table Entry in order to read */\
                                /*this RQ doorbell record.The entry is     */\
                                /*obtained via the usr_page field.         */\
                                /* Not valid for EE.                       */\
  u_int32_t	sq_wqe_counter;/* A 16bits counter that is incremented for each*/\
                           /* WQE posted to the SQ.                        */\
                           /* Must be 0x0 in SQ initialization.            */\
                           /*(QUERY_QPEE only).                            */\
  u_int32_t	rq_wqe_counter;/* A 16bits counter that is incremented for each*/\
                           /* WQE posted to the RQ.                        */\
                           /* Must be 0x0 in RQ initialization.            */\
                           /*(QUERY_QPEE only).                            */\
  MT_bool  fre;  /* Fast Registration Work Request Enabled.                */\
                 /*(Reserved for EE).                                      */\
  MT_bool rlky;  /* When set this QP can access the Reserved L_Key         */          
  

#define  XHH_DEV_SPECIFIC_SRQ_CONTEXT \
  u_int32_t srq_db_record_index; /* HW uses RQ UAR index to access the      */\
                           /* UAR table in order to read this RQ doorbell   */\
                           /* record. The lsbs of the UAR index are selected*/\
                           /* according to log_uar_entry_sz and then        */\
                           /* concatenated with the usr page number. This value */\
                           /* can not be retrieved from the HW.             */\
  u_int32_t srq_wqe_base_adr_l; /* While opening (creating) the WQ,        */\
                                /* this field should contain the address   */\
                                /* of first descriptor to be posted.       */\
                                /* Not valid (reserved) in EE context.     */\
  u_int32_t usr_page;           /* UAR index  of this SRQ */                 \
  u_int32_t srqn;           /* number  of this SRQ */                        \
  u_int32_t srq_wqe_counter;/* A 16bits counter that is incremented for each*/\
                           /* WQE posted to the RQ.                        */\
                           /* Must be 0x0 in RQ initialization.            */\
                           /*(QUERY_QPEE only).                            */\
  u_int8_t  log_rq_stride; /* Stride (max WQE size) on the receive queue.   */\
                           /* WQ entry is 16*(2^log_RQ_stride) bytes.       */\
  u_int8_t  log_srq_size;  /*Log2 of the Number of WQEs in the Receive Queue*/ \
  u_int8_t  eq; /*Event Queue SRQ reached limit water mark events are reported to.*/
  

#define  XHH_DEV_SPECIFIC_HCA_PROPS   \
  u_int8_t time_stamp_granularity;/* This field controls the granularity    */\
                           /* in which CQE Timestamp counter is incremented.*/\
                     /* The TimeStampGranularity units is 0.25 microseconds.*/\
                     /* (e.g is TimeStampGranularity is configured to 0x2,  */\
                     /* CQE Timestamp will be incremented every microsecond)*/\
                     /* When sets to Zero, timestamp reporting in the CQE   */\
                     /* is disabled.                                        */\
                                                                              \
  MT_bool ce;  /* Checksum Enabled - when Set IPoverIB checksum generation  */\
               /* & checking is enabled                                     */\
  MT_bool sph; /* 0 - SW calculates TCP/UDP Pseudo-Header checksum and      */\
               /*     inserts it into the TCP/UDP checksum field when       */\
               /*     sending a packet                                      */\
               /* 1 - HW calculates TCP/UDP Pseudo-Header checksum when     */\
               /*     sending a packet                                      */\
  MT_bool rph; /* 0 - Not HW calculation of TCP/UDP Pseudo-Header checksum  */\
               /*     are done when receiving a packet                      */\
               /* 1 - HW calculates TCP/UDP Pseudo-Header checksum when     */\
               /*     receiving a packet                                    */\
  MT_bool bmme; /* When set Base Memory Management Extension is Enabled     */\
  u_int8_t responder_exu;/*How many execution engines are dedicated to the  */\
                         /* responder.                                      */\
                         /*    Legal values are 0x0-0xF. 0 is "auto"        */\
  u_int32_t wqe_quota; /* Maximum number of WQEs that are executed prior to */\
                       /* preemption of execution unit.                     */\
                       /*  0 - reserved.                                    */\
  MT_bool wqe_quota_en;  /*If set - wqe_quota field is used. If cleared,    */\
                         /* WQE quota is set to "auto" value                */ 
                         

#define  XHH_DEV_SPECIFIC_CONTEXTS_PRMS  \
  u_int64_t ecqc_base_addr;  /* Extended CQC Base Address. Has same         */\
                             /* num entries as CQC table.                   */\
                             /* Table must be aligned on its size           */\
  u_int32_t eq_arm_base_addr; /* EQ Arm DBs offset in the configuration     */\
                              /* space. Points to 64 bit register. Setting  */\
                              /* bit x in the offset, arms EQ number x.     */\
                              /* (QUERY_HCA only)                           */\
  u_int32_t eq_set_ci_base_addr;/* EQ Set CI DBs offset in the              */\
                                /* configuration space. (QUERY_HCA only)    */ 

#define  XHH_DEV_SPECIFIC_MPT_ENTRY  \
  u_int64_t mtt_adr;	/* Base (first) address of the MTT segment, aligned on segment_size boundary */

#define  XHH_DEV_SPECIFIC_TPT_PRMS

#define XHH_CMD_STAT_LAM_NOT_PRE_STR  "XHH_CMD_STAT_LAM_NOT_PRE - LAM not present"

#define  XHH_CMD_DEV_SPECIFIC_STATS \
 XHH_CMD_STAT_LAM_NOT_PRE = ARBEL_IF_CMD_STAT_LAM_NOT_PRE,

#define 	XHH_DEV_SPECIFIC_CMD_STATUS_CASES \
    case XHH_CMD_STAT_LAM_NOT_PRE:			  \
	  return XHH_CMD_STAT_LAM_NOT_PRE_STR;

typedef struct {
  u_int64_t uar_base_addr; /* UAR Base Address (QUERY_HCA only) */
  u_int8_t log_uar_entry_sz; /* Number of 4K pages per UAR Context entry is 2^log_uar_sz  */
  u_int8_t log_max_uars;     /* Number of UARs supported is 2^log_max_UARs*/
  u_int8_t uar_page_sz;	/* This field defines the size of each UAR page. Size of UAR Page is 4KB*2^UAR_Page_Size */
  u_int64_t uar_scratch_base_addr;	/* Base address of UAR scratchpad. Number of entries in table is UAR BAR size divided by UAR Page Size. Table must be aligned to entry size. */
  u_int64_t uar_context_base_addr;	/* Base address of UAR Context [63:32]. Number of entries in table is 2^log_max_uars. Table must be aligned to its size.*/
}
XHH_uar_prms_t;

#endif
