/*
  This software is available to you under a choice of one of two
  licenses.  You may choose to be licensed under the terms of the GNU
  General Public License (GPL) Version 2, available at
  <http://www.fsf.org/copyleft/gpl.html>, or the OpenIB.org BSD
  license, available in the LICENSE.TXT file accompanying this
  software.  These details are also available at
  <http://openib.org/license.html>.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
  BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
  ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
  CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
  SOFTWARE.

  Copyright (c) 2004 Mellanox Technologies Ltd.  All rights reserved.
*/

#ifndef H_DEV_COMMON_H
#define H_DEV_COMMON_H


#include <mtl_common.h>
#include <vapi.h>
#include <hh.h>
#include <log2.h>
#include <MT25218.h>
#include <xhh.h>

#define HCA_ARBEL_DEV_ID_PREFIX "InfiniHost_III_Ex"
#define HCA_SINAI_DEV_ID_PREFIX "InfiniHost_III_Lx"

#define XHH_NAME "RHH"

/* resources reserved in driver */
#define XHH_NUM_RSVD_PD     2
#define XHH_NUM_RSVD_QP     8

#define RHH_KERNEL_MAX_AV 0x4000 /* default max number of AVs in kernel for Arbel - 16K*/

#define MT_ARBEL_LOG2_MAX_ICM_SIZE  (38)
#define MT_ARBEL_PAGE_SHIFT         (12)

#define ADD_DEV_PREFIX(a)       ARBEL_ ## a
#define ADD_DEV_PREFIX_LC(a)    arbel_ ## a
#define ADD_DEV_PRM_PREFIX(a)   arbelprm_ ## a

typedef struct
{
  MT_virt_addr_t   cqe_buf;     /* CQE buffer virtual addr. CQE size aligned */
  MT_size_t        cqe_buf_sz;  /* Buffer size in bytes (mult of CQE size) */
  XHH_uar_index_t  uar_index;   /* Index of UAR used for this CQ. */
  u_int32_t        new_producer_index;     /* New producer index after "resize_cq" (OUT)*/
  MT_virt_addr_t   ci_db_record_p;         /* for doorbell */
  MT_virt_addr_t   arm_db_record_p;        /* for doorbell */
  u_int8_t* next_cmd_sn_p;
} XHH_cq_ul_resources_t;

typedef struct 
{
  MT_virt_addr_t   wqes_buf;     /* WQEs buffer virtual address */
  MT_size_t        wqes_buf_sz;  /* Buffer size in bytes */
  XHH_uar_index_t  uar_index;    /* index of UAR used for this QP */
  MT_virt_addr_t   sendq_db_record_p;         /* for doorbell */
  MT_virt_addr_t   recvq_db_record_p;        /* for doorbell */
  u_int8_t   log_sq_stride;
  u_int8_t   log_sq_size;
  u_int8_t   log_rq_stride;
  u_int8_t   log_rq_size;
  u_int32_t	 snd_wqe_base_adr_l;
  u_int32_t	 rcv_wqe_base_adr_l; 
  VAPI_lkey_t qp_mem_lkey;       /* to be used in kernel when buffer is allocated externaly */
} XHH_qp_ul_resources_t;

typedef struct 
{
  MT_virt_addr_t      wqes_buf;     /* WQEs buffer virtual address */
  MT_size_t           wqes_buf_sz;  /* Buffer size in bytes */
  MT_size_t           wqe_sz;       /* WQE (descriptor) size in bytes */
  XHH_uar_index_t     uar_index;    /* index of UAR used for this SRQ */
  MT_virt_addr_t      recvq_db_record_p;        /* for doorbell */
  u_int32_t         srq_limit;	   /* SRQ's limit (modify-srq only) */
  VAPI_srq_attr_mask_t srq_attr_mask; /* modify-srq only */
} XHH_srq_ul_resources_t;

typedef struct XHH_pdm_st     *XHH_pdm_t;  
typedef struct XHH_uarm_st    *XHH_uarm_t;  
typedef struct XHH_icmm_st    *XHH_icmm_t;
typedef struct XHH_mapm_st    *XHH_mapm_t;
typedef struct XHH_cidb_st    *XHH_cidb_t;

typedef u_int64_t             XHH_icm_va_t;

#endif  /* H_XHH_COMMON_H */
