/*
  This software is available to you under a choice of one of two
  licenses.  You may choose to be licensed under the terms of the GNU
  General Public License (GPL) Version 2, available at
  <http://www.fsf.org/copyleft/gpl.html>, or the OpenIB.org BSD
  license, available in the LICENSE.TXT file accompanying this
  software.  These details are also available at
  <http://openib.org/license.html>.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
  BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
  ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
  CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
  SOFTWARE.

  Copyright (c) 2004 Mellanox Technologies Ltd.  All rights reserved.
*/

#include "cqm.h"
#include "cqm_priv.h"

HH_ret_t dev_specific_cqm_create_cq(XHH_cqm_t  cqm, 
                                                  XHH_cq_ul_resources_t*  cq_ul_resources_p,
                                                  u_int32_t  cqc_index)
{
    HH_ret_t rc;
    u_int32_t ci_db_record_idx, arm_db_record_idx;
    void * ci_db_record_va;
    void * arm_db_record_va;
    FUNC_IN;
    rc= XHH_mapm_alloc_rsc(cqm->mapm,cqc_index); 
    if (rc!= HH_OK) {
      MT_RETURN(rc);
    }
    /* allocate doorbell records for ci and arm */
    rc =  XHH_uarm_alloc_db_record( cqm->uarm, 
                                    cq_ul_resources_p->uar_index, 
                                    UARM_DBRECORD_SET_CI,
                                    &ci_db_record_idx,
                                    &ci_db_record_va);
    if (rc!= HH_OK)  goto failed_alloc_ci_db;

    rc =  XHH_uarm_alloc_db_record( cqm->uarm, 
                                    cq_ul_resources_p->uar_index, 
                                    UARM_DBRECORD_ARM_CQ,
                                    &arm_db_record_idx,
                                    &arm_db_record_va);
    if (rc!= HH_OK) goto failed_alloc_arm_db;

    rc= XHH_uarm_alloc_cmdsn_byte( cqm->uarm, cq_ul_resources_p->uar_index, 
				   &cqm->entries[cqc_index].u.used.next_cmd_sn_p, 
				   &cq_ul_resources_p->next_cmd_sn_p);
    if (rc!= HH_OK) goto failed_alloc_cmdsn;

    cqm->entries[cqc_index].u.used.uar_index = cq_ul_resources_p->uar_index;
    cqm->entries[cqc_index].u.used.ci_db_record_idx = ci_db_record_idx;
    cqm->entries[cqc_index].u.used.arm_db_record_idx = arm_db_record_idx;
    cq_ul_resources_p->ci_db_record_p = (MT_virt_addr_t) ci_db_record_va;
    cq_ul_resources_p->arm_db_record_p = (MT_virt_addr_t) arm_db_record_va;
    MT_RETURN(HH_OK);

    
    failed_alloc_cmdsn:
    XHH_uarm_free_db_record( cqm->uarm, 
			     cq_ul_resources_p->uar_index, 
			     UARM_DBRECORD_ARM_CQ,
			     arm_db_record_idx);
    failed_alloc_arm_db:
    XHH_uarm_free_db_record( cqm->uarm, 
			     cq_ul_resources_p->uar_index, 
			     UARM_DBRECORD_SET_CI,
			     ci_db_record_idx);
    failed_alloc_ci_db:
    XHH_mapm_free_rsc(cqm->mapm,cqc_index);
    MT_RETURN(rc);
}

HH_ret_t dev_specific_cqm_destroy_cq(XHH_cqm_t  cqm,
                                                   u_int32_t  cqc_index)
{
    FUNC_IN;
    XHH_uarm_free_cmdsn_byte(cqm->uarm,
                             cqm->entries[cqc_index].u.used.uar_index, 
			     cqm->entries[cqc_index].u.used.next_cmd_sn_p);

    XHH_uarm_free_db_record( cqm->uarm, 
                             cqm->entries[cqc_index].u.used.uar_index, 
                             UARM_DBRECORD_SET_CI,
                             cqm->entries[cqc_index].u.used.ci_db_record_idx);
    
    XHH_uarm_free_db_record( cqm->uarm, 
                             cqm->entries[cqc_index].u.used.uar_index, 
                             UARM_DBRECORD_ARM_CQ,
                             cqm->entries[cqc_index].u.used.arm_db_record_idx);
    XHH_mapm_free_rsc(cqm->mapm,cqc_index);
    MT_RETURN(HH_OK);
}


HH_ret_t XHH_cqm_update_cmd_sn( XHH_cqm_t    cqm,                     /* IN */
				HH_cq_hndl_t cq)                      /* IN */
{
	u_int32_t cqc_index= cq & MASK32(cqm->log2_max_cq);
	HH_ret_t  rc = (in_use_cq(cqm, cq) ? HH_OK : HH_EINVAL_CQ_HNDL);

	if (rc == HH_OK) {
		(*(cqm->entries[cqc_index].u.used.next_cmd_sn_p))++; /* increment cmd_sn */
		/* Note: no need for atomicity. All completion events of a CQ arrive on the
		 * same EQ, so they are processed in the same DPC context.		*/
	}
	return rc;
}

