/*
  This software is available to you under a choice of one of two
  licenses.  You may choose to be licensed under the terms of the GNU
  General Public License (GPL) Version 2, available at
  <http://www.fsf.org/copyleft/gpl.html>, or the OpenIB.org BSD
  license, available in the LICENSE.TXT file accompanying this
  software.  These details are also available at
  <http://openib.org/license.html>.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
  BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
  ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
  CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
  SOFTWARE.

  Copyright (c) 2004 Mellanox Technologies Ltd.  All rights reserved.
*/

#ifndef H_XHH_DEV_CQM_H
#define H_XHH_DEV_CQM_H

#include <mapm.h>
#include <uarm.h>

#define CQM_MAPM_LOG2_BLKSZ 7   

#define XHH_DEV_SPECIFIC_CQM_INIT_PARAMS \
  u_int64_t   cqc_tbl_base;              \
  u_int8_t    log2_cqc_entry_sz;

#define XHH_DEV_SPECIFIC_CQM_DATA  \
 XHH_mapm_t             mapm; \
 XHH_uarm_t             uarm;

#define XHH_DEV_SPECIFIC_CQCM_ENTRY_DATA  \
    XHH_uar_index_t uar_index;  /* for call to XHH_uarm_free_db_record */  	\
    u_int32_t	     ci_db_record_idx;       /* for free */               	\
    u_int32_t       arm_db_record_idx;      /* for free */   			\
    u_int8_t *next_cmd_sn_p ;
 
#define XHH_DEV_SPECIFIC_CQM_INIT  \
  rc= XHH_hob_get_uarm(hob, &cqm->uarm);\
  if (rc != HH_OK) {\
      FREE(cqm);\
      VFREE(entries);\
      return rc;\
  }\
  rc= XHH_mapm_create(p_init_params->cqc_tbl_base,\
                          p_init_params->log2_max_cq,\
                          CQM_MAPM_LOG2_BLKSZ,\
                          p_init_params->log2_cqc_entry_sz,\
                          hob,\
                          (1 << p_init_params->log2_rsvd_cqs),\
                          &(cqm->mapm));\
  if (rc != HH_OK) {\
      FREE(cqm);\
      VFREE(entries);\
      return HH_EAGAIN;\
  }

#define XHH_DEV_SPECIFIC_CQM_DONE \
        XHH_mapm_destroy(cqm->mapm);
        
#define XHH_DEV_SPECIFIC_CQM_CREATE_CQ \
  rc= dev_specific_cqm_create_cq(cqm,cq_ul_resources_p,cqc_index); \
  if (rc!= HH_OK) {                                                    \
    epool_free(&cqm->flist,cqc_index);                                 \
    FREE(mr_internal_p);                                               \
    return rc;                                                         \
  }
  
#define XHH_DEV_SPECIFIC_CQM_DESTROY_CQ \
        dev_specific_cqm_destroy_cq(cqm,cqc_index); 


/* Arbel specific function to increment expected cmd_sn after an event is retrieved by eventp */
HH_ret_t XHH_cqm_update_cmd_sn( XHH_cqm_t    cqm,                     /* IN */
				HH_cq_hndl_t cq);                     /* IN */

#endif
