/*
  This software is available to you under a choice of one of two
  licenses.  You may choose to be licensed under the terms of the GNU
  General Public License (GPL) Version 2, available at
  <http://www.fsf.org/copyleft/gpl.html>, or the OpenIB.org BSD
  license, available in the LICENSE.TXT file accompanying this
  software.  These details are also available at
  <http://openib.org/license.html>.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
  BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
  ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
  CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
  SOFTWARE.

  Copyright (c) 2004 Mellanox Technologies Ltd.  All rights reserved.
*/

#ifndef H_XHH_DEV_CQM_INLINE_H
#define H_XHH_DEV_CQM_INLINE_H


/************************************************************************/
static HH_ret_t  sw2hw_cq
(
  XHH_cmd_t               cmd_if,
  XHH_cq_ul_resources_t*  user_prot_ctx_p,
  u_int32_t               cqn,   
  unsigned long           n_cq_entries,
  VAPI_lkey_t             lkey,
  XHH_eqn_t               comp_eqn,
  XHH_eqn_t               error_eqn,
  XHH_cqm_t               cqm,       
  u_int32_t               cqc_index
)
{
  XHH_cqc_t          *cqc_p = NULL;
  XHH_cmd_status_t   cmd_rc;

  cqc_p = (XHH_cqc_t *)MALLOC(sizeof(XHH_cqc_t));
  if (cqc_p == NULL) {
       MT_COV_MEM_FLTR_START
       MTL_ERROR1(MT_FLFMT("%s: Could not allocate memory for cqc_p"),__func__);
       return HH_EAGAIN;
       MT_COV_MEM_FLTR_END
  }

  memset(cqc_p, 0, sizeof(XHH_cqc_t));
  cqc_p->st           = 0; /* disarmed */
#ifdef NO_CQ_CI_DBELL
  /* Use this option carefully - CQ overrun may cause unexpected behavior */
  /* It is recommended to set CQ size to be the total of max. outstanding */
  /* WQEs of all attached work queues.                                    */
  cqc_p->oi            = 1;/*CQ's consumer index update DBells are not used - must ignore CQ overrun*/
#else
  cqc_p->oi            = 0;/* Enforce CQ overrun detection based on consumer index doorbells updates*/
#endif
  cqc_p->tr            = 1;
  cqc_p->status        = 0;
  cqc_p->start_address = user_prot_ctx_p->cqe_buf;
  cqc_p->usr_page      = user_prot_ctx_p->uar_index;
  cqc_p->log_cq_size   = (u_int8_t)floor_log2(n_cq_entries);
  cqc_p->e_eqn         = error_eqn;
  cqc_p->c_eqn         = comp_eqn;
  cqc_p->pd            = XHH_RESERVED_PD(cqm->hob);
  cqc_p->l_key         = lkey;
  cqc_p->cqn           = cqn;
  cqc_p->cq_ci_db_record =    cqm->entries[cqc_index].u.used.ci_db_record_idx;
  cqc_p->cq_state_db_record = cqm->entries[cqc_index].u.used.arm_db_record_idx;


  cmd_rc = XHH_cmd_SW2HW_CQ(cmd_if, cqn, cqc_p);
  MTL_DEBUG4(MT_FLFMT("cmd_rc=%d=%s"), cmd_rc, str_XHH_cmd_status_t(cmd_rc));
  FREE(cqc_p);
  return (CMDRC2HH_ND(cmd_rc));
} /* sw2hw_cq */

#endif
