/*
  This software is available to you under a choice of one of two
  licenses.  You may choose to be licensed under the terms of the GNU
  General Public License (GPL) Version 2, available at
  <http://www.fsf.org/copyleft/gpl.html>, or the OpenIB.org BSD
  license, available in the LICENSE.TXT file accompanying this
  software.  These details are also available at
  <http://openib.org/license.html>.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
  BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
  ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
  CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
  SOFTWARE.

  Copyright (c) 2004 Mellanox Technologies Ltd.  All rights reserved.
*/

#ifndef H_XHH_DEV_SRQM_H
#define H_XHH_DEV_SRQM_H

#include <mapm.h>
#include <pdm.h>
#include <uarm.h>

#define SRQM_MAPM_LOG2_BLKSZ  7   /*blk size: one host page */

#define XHH_DEV_SPECIFIC_SRQM_INIT_PARAMS \
  XHH_icm_va_t   srqc_tbl_base;    \
  u_int8_t      log2_srqc_entry_sz; \
  XHH_eqn_t            ib_eq;

#define XHH_DEV_SPECIFIC_SRQM_DATA  \
 XHH_mapm_t             mapm;\
 XHH_uarm_t             uarm;\
 XHH_pdm_t              prot_module; \
 XHH_eqn_t              ib_eq;
 
#define XHH_DEV_SPECIFIC_SRQ_DATA \
 XHH_uar_index_t  uar_index;   /* Index of UAR used for this SRQ. */ \
 u_int32_t  rcv_db_record_index; 
				     
#define XHH_DEV_SPECIFIC_SRQM_INIT  \
  (*srqm_p)->ib_eq = p_init_params->ib_eq; \
  ret= XHH_hob_get_uarm(hob, &((*srqm_p)->uarm));\
  if (ret != HH_OK) {\
     FREE(srqm_p);\
     return ret;\
  }\
  ret= XHH_mapm_create(p_init_params->srqc_tbl_base,      \
                       p_init_params->log2_max_srq,       \
      ((SRQM_MAPM_LOG2_BLKSZ < (MOSAL_SYS_PAGE_SHIFT - p_init_params->log2_srqc_entry_sz)) ? \
      (MOSAL_SYS_PAGE_SHIFT - p_init_params->log2_srqc_entry_sz):SRQM_MAPM_LOG2_BLKSZ), \
                       p_init_params->log2_srqc_entry_sz, \
                       hob,                               \
                       (1 <<p_init_params->log2_rsvd_srq),\
                       &((*srqm_p)->mapm));                  \
  if (ret != HH_OK) {                                     \
      MTL_ERROR1(MT_FLFMT("%s: Failed allocation of XHH_mapm object"),__func__);\
      FREE(srqm_p);                                                             \
      return HH_EAGAIN;                                                         \
  }

#define XHH_DEV_SPECIFIC_SRQM_DONE \
        XHH_mapm_destroy(srqm->mapm);
	
#define XHH_DEV_SPECIFIC_SRQM_CREATE_SRQ_FAIL_LABEL \
dev_specific_fail:
        
#define XHH_DEV_SPECIFIC_SRQM_CREATE_SRQ  \
      ret= dev_specific_srqm_create_srq(srqm,srq_ul_resources_p,srqn, srq); \
      if (ret != HH_OK) {                                  \
        goto dev_specific_fail;                            \
      }                                                    \
      xhh_srqc.log_rq_stride= (u_int8_t)ceil_log2(srq_ul_resources_p->wqe_sz >> 4); /* 16B chunks */    \
      if (xhh_srqc.log_rq_stride > 0x6) {				    \
	  xhh_srqc.log_rq_stride=6; /* Stride may be 1024, but max WQE size is 1008 (ds is 6bit) */\
      }\
      xhh_srqc.log_srq_size = (u_int8_t)ceil_log2(srq_ul_resources_p->wqes_buf_sz);          \
      xhh_srqc.srq_wqe_base_adr_l = (u_int32_t)srq_ul_resources_p->wqes_buf;  \
      xhh_srqc.srq_db_record_index = srq->rcv_db_record_index; \
      xhh_srqc.usr_page =  srq->uar_index; \
      xhh_srqc.eq = (u_int8_t) (srqm->ib_eq); \
      MTL_DEBUG1(MT_FLFMT("%s: srq_wqe_base_adr_l = 0x%x, logsize=0x%x,logstride=0x%x, rec_ix=0x%x, pg=0x%x,eq=0x%x") ,\
      __func__,xhh_srqc.srq_wqe_base_adr_l,xhh_srqc.log_srq_size,xhh_srqc.log_rq_stride,\
      xhh_srqc.srq_db_record_index,xhh_srqc.usr_page,xhh_srqc.eq); \
      

#define XHH_DEV_SPECIFIC_SRQM_DESTROY_SRQ     \
        XHH_uarm_free_db_record( srqm->uarm,  		    \
			       srq->uar_index, 	    \
			       UARM_DBRECORD_RECEIVE,	    \
			       srq->rcv_db_record_index); \
        XHH_mapm_free_rsc(srqm->mapm,srqn);
							    
#define XHH_DEV_SPECIFIC_SRQM_GET_PROT_MODULE\
        XHH_hob_get_pdm(hob, &(*srqm_p)->prot_module);


#endif
