/*
  This software is available to you under a choice of one of two
  licenses.  You may choose to be licensed under the terms of the GNU
  General Public License (GPL) Version 2, available at
  <http://www.fsf.org/copyleft/gpl.html>, or the OpenIB.org BSD
  license, available in the LICENSE.TXT file accompanying this
  software.  These details are also available at
  <http://openib.org/license.html>.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
  BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
  ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
  CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
  SOFTWARE.

  Copyright (c) 2004 Mellanox Technologies Ltd.  All rights reserved.
*/

#ifndef H_XHH_DEV_TPTM_H
#define H_XHH_DEV_TPTM_H

#include <arbel_if_defs.h>
#include <mapm.h>
#include <pdm.h>

/* Shuffle mode */
#define TPTM_CURRENT_MEMKEY(mrwm,mpt_seg,mpt_index) \
  (( ((mrwm)->key_prefix[mpt_seg][mpt_index-(mrwm)->offset[mpt_seg]] << (mrwm)->props.log2_mpt_sz) | (mpt_index)) << 8)
  
/* Shuffle mode */
#define TPTM_MEMKEY2INDEX(mrwm,memkey)  ( ((memkey) >> 8) & MASK32((mrwm)->props.log2_mpt_sz) )
#define TPTM_MEMKEY2PREFIX(mrmwm,memkey) ( (memkey) >> (mrwm->props.log2_mpt_sz + 8) )

#define TPTM_MEMKEY_HW_FMT(memkey) ( ((memkey)>>8) | ((memkey)<<24) )

#define TPTM_CURRENT_FMR_MEMKEY(mrwm,mpt_seg,mpt_index,fmr_info_p)  \
  ( TPTM_CURRENT_MEMKEY(mrwm,mpt_seg,mpt_index) | (fmr_info_p)->cur_key_suffix )
  
#define TPTM_ADVANCE_FMR_MEMKEY  (fmr_info_p)->cur_key_suffix++
#define TPTM_REVERT_FMR_MEMKEY   (fmr_info_p)->cur_key_suffix--

#define TPTM_MAPM_LOG2_BLKSZ_MPT 7 /*log of number of entries in block */
#define TPTM_MAPM_LOG2_BLKSZ_MTT 10 /*log of number of entries in block */

#define TPTM_LOG_MPT_PG_SZ_SHIFT 12

#define XHH_DEV_SPECIFIC_TPTM_INIT_PARAMS \
            XHH_icm_va_t  mtt_base;        /* Physical address of MTT */  \
            XHH_icm_va_t  mpt_base;        /* Physical address of MPT */

  
#define XHH_DEV_SPECIFIC_TPTM_DATA  \
         XHH_mapm_t mapm_mpt;\
         XHH_mapm_t mapm_mtt;\
         XHH_pdm_t       prot_module;

#define XHH_DEV_SPECIFIC_TPTM_FMR_RSC_MAP_TYPE XHH_icmm_kernel_map_t
#define XHH_DEV_SPECIFIC_TPTM_FMR_INFO u_int8_t cur_key_suffix;

/* TBD : log 2 mpt sz, log2 mtt sz */
#define XHH_DEV_SPECIFIC_TPTM_INIT  \
  rc= XHH_mapm_create(mrwm_props->mtt_base,\
                      mrwm_props->log2_mtt_sz,\
                      TPTM_MAPM_LOG2_BLKSZ_MTT,    \
                      mrwm_props->log2_mtt_seg_sz + MTT_LOG_MTT_ENTRY_SIZE,\
                      hob,                         \
                      (1<< mrwm_props->log2_rsvd_mtt_segs),&mrwm->mapm_mtt);\
  if (rc != HH_OK) {\
          MTL_ERROR1(MT_FLFMT("%s: Failed allocation of XHH_mapm object"),__func__);\
          FREE(mrwm);\
          return HH_EAGAIN;\
      }   \
     \
  rc= XHH_mapm_create(mrwm_props->mpt_base,\
                      mrwm_props->log2_mpt_sz,\
                      TPTM_MAPM_LOG2_BLKSZ_MPT,\
                      6,\
                      hob,\
                      (1<< mrwm_props->log2_rsvd_mpts),&mrwm->mapm_mpt);\
  if (rc != HH_OK) {\
      MTL_ERROR1(MT_FLFMT("%s: Failed allocation of XHH_mapm object"),__func__);\
      XHH_mapm_destroy(mrwm->mapm_mtt);\
      FREE(mrwm);\
      return HH_EAGAIN;\
  }                                                                             
  
  
#define XHH_DEV_SPECIFIC_TPTM_DONE \
        XHH_mapm_destroy(mrwm->mapm_mpt);\
        XHH_mapm_destroy(mrwm->mapm_mtt)
        
#define XHH_DEV_SPECIFIC_TPTM_ALLOC_MPTE            \
          rc = XHH_mapm_alloc_rsc(mrwm->mapm_mpt,(u_int32_t)mpt_index); \
          if (rc != HH_OK) MTL_ERROR1(MT_FLFMT("%s: failed allocating mapm_mpt. mpt_index=0x%x"),__func__, (u_int32_t)mpt_index);\
        
        
#define XHH_DEV_SPECIFIC_TPTM_ALLOC_MPTE4FMR         \
          rc = XHH_mapm_alloc_rsc(mrwm->mapm_mpt,(u_int32_t)mpt_index); \
          if (rc != HH_OK) MTL_ERROR1(MT_FLFMT("%s: failed allocating mapm_mpt. mpt_index=0x%X"),__func__, (u_int32_t)mpt_index);\
	  else 	{					\
		rc= XHH_mapm_rscs_map(mrwm->mapm_mpt, (u_int32_t)mpt_index, 1, &new_fmr_p->mpt_entry);	\
		if (rc != HH_OK) {									\
			MTL_ERROR1(MT_FLFMT("Failed mapping for FMR mpt_index=0x%X)"), (u_int32_t)mpt_index);	\
			XHH_mapm_free_rsc(mrwm->mapm_mpt,(u_int32_t)mpt_index);				\
		}											\
	  }

#define XHH_DEV_SPECIFIC_TPTM_RETRIEVE_MPTE\
        XHH_mapm_free_rsc(mrwm->mapm_mpt,(u_int32_t)mpt_index)

#define XHH_DEV_SPECIFIC_TPTM_RETRIEVE_MPTE4FMR(fmr_p)\
	  XHH_mapm_rscs_unmap(mrwm->mapm_mpt, fmr_p->mpt_entry);	\
	  XHH_mapm_free_rsc(mrwm->mapm_mpt,(u_int32_t)mpt_index)


#define XHH_DEV_SPECIFIC_TPTM_GET_PROT_MODULE\
        XHH_hob_get_pdm(hob, &mrwm->prot_module)
  
/* in mtt table: the mapm "entries" are segments not mtt_entries   */   
#define XHH_DEV_SPECIFIC_TPTM_ALLOC_MTT_SEGS           \
	  rc= XHH_mapm_alloc_rscs(mrwm->mapm_mtt, seg_start, 1<<log2_segs);						\
	  if (rc != HH_OK) MTL_ERROR1(MT_FLFMT("%s: failed allocating mapm_mtt. seg_start=0x%x"),__func__, seg_start)	
	  
        
#define XHH_DEV_SPECIFIC_TPTM_ALLOC_MTT_SEGS4FMR        \
	  rc= XHH_mapm_alloc_rscs(mrwm->mapm_mtt, seg_start, 1<<log2_segs);						\
	  if (rc != HH_OK) MTL_ERROR1(MT_FLFMT("%s: failed allocating mapm_mtt. seg_start=0x%x"),__func__, seg_start);	\
	  else 	{					\
		rc= XHH_mapm_rscs_map(mrwm->mapm_mtt, (u_int32_t)seg_start, 1<<log2_segs, &new_fmr_p->mtt_entries);	\
		if (rc != HH_OK) {											\
			MTL_ERROR1(MT_FLFMT("Failed mapping for FMR MTTs at seg_start=0x%X log2_segs=0x%X)"),		\
				(u_int32_t)seg_start, log2_segs);							\
			XHH_mapm_free_rscs(mrwm->mapm_mtt,(u_int32_t)seg_start, 1<<log2_segs);				\
		}													\
	  }
        
#define XHH_DEV_SPECIFIC_TPTM_FREE_MTT_SEGS\
	  XHH_mapm_free_rscs(mrwm->mapm_mtt, seg_start, 1<<log2_segs)

#define XHH_DEV_SPECIFIC_TPTM_FREE_MTT_SEGS4FMR(mrwm,fmr,seg_start,log2_segs)\
	  XHH_mapm_rscs_unmap(mrwm->mapm_mtt, fmr->mtt_entries);	\
	  XHH_mapm_free_rscs(mrwm->mapm_mtt, seg_start, 1<<log2_segs)

#define XHH_DEV_SPECIFIC_TPTM_MPT_ENTRY_INIT do {} while(0)
#define XHH_DEV_SPECIFIC_TPTM_FMR_MPT_ENTRY_INIT do {} while(0)

#define XHH_DEV_SPECIFIC_TPTM_FMR_INFO_INIT(fmr_info_p) fmr_info_p->cur_key_suffix= 0

#define MPT_MTT_ADR(a)   (a->mtt_adr)
#define MPT_MTT_ADR_DOT(a)   (a.mtt_adr)

#define XHH_DEV_SPECIFIC_TPTM_COMMIT_MTT_ENTRIES \
	rc= XHH_icmm_kernel_map_write_buf(fmr_info_p->mtt_entries, \
					fmr_info_p->mtt_entries_draft, \
					real_num_of_pages << MTT_LOG_MTT_ENTRY_SIZE)

#define XHH_DEV_SPECIFIC_TPTM_MODIFY_MPT_QWORD(mpt_entry, offset, val) \
				rc= XHH_icmm_kernel_map_write_qword(mpt_entry, offset, val)

#define XHH_DEV_SPECIFIC_TPTM_MODIFY_MPT_DWORD(mpt_entry, offset, val) \
				rc= XHH_icmm_kernel_map_write_dword(mpt_entry, offset, val)

#define XHH_DEV_SPECIFIC_TPTM_MODIFY_MPT_BYTE(mpt_entry, offset, val) \
				rc= XHH_icmm_kernel_map_write_byte(mpt_entry, offset, val)

#endif
