/*
  This software is available to you under a choice of one of two
  licenses.  You may choose to be licensed under the terms of the GNU
  General Public License (GPL) Version 2, available at
  <http://www.fsf.org/copyleft/gpl.html>, or the OpenIB.org BSD
  license, available in the LICENSE.TXT file accompanying this
  software.  These details are also available at
  <http://openib.org/license.html>.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
  BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
  ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
  CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
  SOFTWARE.

  Copyright (c) 2004 Mellanox Technologies Ltd.  All rights reserved.
*/

#if !defined(H_DEV_UDAV_H)
#define H_DEV_UDAV_H

#include <ul_hob.h>

/************************************************************************
 *  Function: XHH_udavm_create
 *  
 *  Arguments:
 *  version_p - Version information (see ...) 
 *  pdm - pd manager object
 *  max_kernel_avs - Max number of AVs which can be allocated in kernel space.
 *                   Ignored in user-space apps.
 *  uadvm_p - Allocated object 
 *  
 *  Returns:
 *  HH_OK 
 *  HH_EINVAL -Invalid parameters (NULLs) 
 *  HH_EAGAIN -Not enough resources to allocate object 
 *
 *  Description: 
 *  Create the XHH_udavm_t class instance. 
 ************************************************************************/
 
extern HH_ret_t XHH_udavm_create( /*IN */ XHH_ver_info_t *version_p, 
                                  /*IN */ XHHUL_pdm_t pdm,
				  /*IN */ u_int32_t  max_kernel_avs,
                                  /*OUT*/ XHH_udavm_t *udavm_p);


/************************************************************************
 *  Function: XHH_udavm_put_av
 *  
 *  Arguments:
 *  udavm - the udav manager object
 *  ah - The address handle of UD AV to query
 *  wqe_ud_segment_p - The location in the WQE to copy UD-AV to
 *  
 *  Returns:
 *  HH_OK 
 *  HH_EINVAL -Invalid parameters 
 *  HH_EINVAL_AH - Invalid address handle (no such handle)
 *
 *  Description: 
 *  Put (copy) given UD-AV into the given location (a pointer to WQEs UD segment).
 ************************************************************************/

extern HH_ret_t XHH_udavm_put_av(/*IN*/ XHH_udavm_t udavm,
                                 /*IN*/ HHUL_ud_av_hndl_t ah,
                                 /*OUT*/ volatile u_int32_t *wqe_ud_segment_p);


#define UDAVM_HCA_HNDL HHUL_hca_hndl_t 
#define UDAVM_PD_HNDL HHUL_pd_hndl_t
#define UDAVM_AV_HNDL HHUL_ud_av_hndl_t
  
#endif
