/*
  This software is available to you under a choice of one of two
  licenses.  You may choose to be licensed under the terms of the GNU
  General Public License (GPL) Version 2, available at
  <http://www.fsf.org/copyleft/gpl.html>, or the OpenIB.org BSD
  license, available in the LICENSE.TXT file accompanying this
  software.  These details are also available at
  <http://openib.org/license.html>.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
  BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
  ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
  CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
  SOFTWARE.

  Copyright (c) 2004 Mellanox Technologies Ltd.  All rights reserved.
*/

#ifndef CI_XHHUL_CQM_DEV_CI
#define CI_XHHUL_CQM_DEV_CI

/* Given CQ must be locked when calling this function */
inline static void sync_consumer_index(XHHUL_cq_t *cq_p, const char* caller)
{
  if (cq_p->pending_cq_dbell > 0) {
    XHH_cidb_cq_update_cntr(cq_p->ci_db_record_p, cq_p->cur_buf.consumer_counter);
    cq_p->pending_cq_dbell= 0;
  }
#if 0
  MTL_ERROR1(MT_FLFMT("(from %s) Updated UAR context for CQn=0x%X at address %p to consumer_counter=0x%X (%08X %08X)"
		      " - log2_num_o_cqes=%u cur_spare_cqes=%u pending_cq_dbell=%u"),
	     caller, cq_p->cq_num, cq_p->ci_db_record_p, cq_p->cur_buf.consumer_counter, 
	     ((u_int32_t*)cq_p->ci_db_record_p)[0], ((u_int32_t*)cq_p->ci_db_record_p)[1],
	     cq_p->cur_buf.log2_num_o_cqes, cq_p->cur_spare_cqes, cq_p->pending_cq_dbell);
#endif
}

inline static HH_ret_t init_uar_info(HHUL_hca_hndl_t hca, XHHUL_cq_t *cq, XHH_cq_ul_resources_t *ul_res_p)
{
    HH_ret_t rc;

    rc= XHHUL_hob_get_cidb(hca,&(cq->cidb));
    if (rc != HH_OK) {
      MTL_ERROR1(MT_FLFMT("%s: Failed getting XHHUL_hob's cidb (err=%d)."),__func__,rc);
      return rc;
    }
    rc= XHH_cidb_get_uar_index(cq->cidb,&(ul_res_p->uar_index)); 
    if (rc != HH_OK) {
      MTL_ERROR1(MT_FLFMT("%s: Failed getting UAR index (err=%d)"), __func__, rc);
    }
    return rc;
}

inline static HH_ret_t uar_arm_cq_cmd(XHHUL_cq_t *cq_p, XHH_uar_cq_cmd_t cmd, u_int32_t param)
{
	u_int32_t counter = ((cmd == ARBEL_IF_UAR_CQ_NOTIF_NCOMP) ? param : cq_p->cur_buf.consumer_counter);
	/*separate test for most common case */
	if (cq_p->last_sn != *(cq_p->next_cmd_sn_p)) {
		cq_p->last_sn = *(cq_p->next_cmd_sn_p);
		cq_p->last_cmd = cmd;
		cq_p->last_counter = counter;
		return (XHH_cidb_cq_arm(cq_p->cidb, cq_p->arm_db_record_p, cq_p->cq_num, cmd, 
					counter, cq_p->last_sn));
	} else if ((cq_p->last_cmd == ARBEL_IF_UAR_CQ_NOTIF_SOLIC_COMP && 
				cmd == ARBEL_IF_UAR_CQ_NOTIF_NEXT_COMP)                  ||
			   (cq_p->last_cmd == cmd && cmd == ARBEL_IF_UAR_CQ_NOTIF_NCOMP && 
				cq_p->last_counter > counter )){
		cq_p->last_sn = *(cq_p->next_cmd_sn_p);
		cq_p->last_cmd = cmd;
		cq_p->last_counter = counter;
		return (XHH_cidb_cq_arm(cq_p->cidb, cq_p->arm_db_record_p, cq_p->cq_num, cmd, 
					counter, cq_p->last_sn));
	} else {
		/* no doorbell ring needed */
		return HH_OK;
	}
}

inline static u_int32_t compute_arm_n_param(XHHUL_cq_t *cq_p, u_int32_t cqe_num)
{
	return cq_p->cur_buf.consumer_counter + cqe_num;
}

#endif /* CI_XHHUL_CQM_DEV_CI */
