/*
  This software is available to you under a choice of one of two
  licenses.  You may choose to be licensed under the terms of the GNU
  General Public License (GPL) Version 2, available at
  <http://www.fsf.org/copyleft/gpl.html>, or the OpenIB.org BSD
  license, available in the LICENSE.TXT file accompanying this
  software.  These details are also available at
  <http://openib.org/license.html>.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
  BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
  ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
  CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
  SOFTWARE.

  Copyright (c) 2004 Mellanox Technologies Ltd.  All rights reserved.
*/

#ifndef H_EVENTP_DEV_FUNCS_H
#define H_EVENTP_DEV_FUNCS_H

#include <eventp_priv.h>
#include <cmd_types.h>
#include <cqm.h>


static inline HH_ret_t init_eventp_dev(XHH_eventp_t eventp,
				       XHH_eventp_res_t *event_res_p)
{
	if (XHH_hob_get_cqm (eventp->hob, &eventp->cqm) != HH_OK){
	  MTL_ERROR4("%s(rhh): Cannot get cqm object handle.\n", __func__);
	  return HH_ERR;
	}
	return HH_OK;
}

static inline HH_ret_t init_eq_dev(XHH_eventp_t eventp,
				   XHH_eqc_t    *eq_context,
				   EQP_eq_entry_t *new_eq)
{
	/* 8 = 4 bytes for each register and 4 bytes for the reserved */
	new_eq->eq_set_ci_reg = eventp->eq_set_ci_reg_base + new_eq->eqn *8; 
	return HH_OK;
}                             


static inline HH_ret_t init_dev_regs(XHH_eventp_t eventp,
				    XHH_eventp_res_t *eventp_res)
{
	/* add 4 since we eq_arm_x_l - since we have only 3 EQs */
        call_result_t rc;
	MT_virt_addr_t  temp_addr;
	 
	rc = MOSAL_iomem_map(MOSAL_get_kernel_prot_ctx(),
		         MOSAL_MEM_FLAGS_NO_CACHE | MOSAL_MEM_FLAGS_PERM_READ | 
		         MOSAL_MEM_FLAGS_PERM_WRITE,
                         eventp_res->eq_arm_base_addr+4, 
	                 (MT_size_t)sizeof(u_int32_t),
	                 &(eventp->eq_arm_reg_iomem),
	                 &(temp_addr));	 //TK ???
	if (rc != MT_OK) {
		MTL_ERROR1("%s: Failed to MOSAL_io_remap for eq_arm_reg(%d)\n", __func__,rc);
		return HH_ERR;
	}
	eventp->eq_arm_reg = temp_addr;
	/* map ci for all EQs - each one have ci and reserved */
	
	rc = MOSAL_iomem_map(MOSAL_get_kernel_prot_ctx(),
		         MOSAL_MEM_FLAGS_NO_CACHE | MOSAL_MEM_FLAGS_PERM_READ | 
		         MOSAL_MEM_FLAGS_PERM_WRITE,
	                 eventp_res->eq_set_ci_base_addr, 
			 (MT_size_t)(EQP_MAX_EQS*2*sizeof(u_int32_t)),
	                 &(eventp->eq_set_ci_reg_base_iomem),
	                 &(temp_addr));
	if (rc != MT_OK) {
		MTL_ERROR1("%s: Failed to MOSAL_io_remap for eq_set_ci_reg_base (%d)\n", __func__,rc);
		MOSAL_iomem_unmap(eventp->eq_arm_reg_iomem,MOSAL_get_kernel_prot_ctx());
		eventp->eq_arm_reg = 0;
		eventp->eq_arm_reg_iomem = (MOSAL_iomem_t) NULL;
		return HH_ERR;
	}
	eventp->eq_set_ci_reg_base = temp_addr;
	return HH_OK;
}


static inline void remove_dev_regs(XHH_eventp_t eventp)
{
	MOSAL_iomem_unmap(eventp->eq_arm_reg_iomem,MOSAL_get_kernel_prot_ctx());
	MOSAL_iomem_unmap(eventp->eq_set_ci_reg_base_iomem,MOSAL_get_kernel_prot_ctx());
	eventp->eq_arm_reg = 0;
	eventp->eq_arm_reg_iomem = (MOSAL_iomem_t) NULL;
	eventp->eq_set_ci_reg_base = 0;
	eventp->eq_set_ci_reg_base_iomem = (MOSAL_iomem_t) NULL;
	
}



#define EQ_UPDATE_CONS_CNTR(eventp, eqe_p) \
        MOSAL_MMAP_IO_WRITE_DWORD(eqe_p->eq_set_ci_reg, MOSAL_cpu_to_be32(eqe_p->cons_cntr))

#define RE_ARM_EQ(eventp, eq_num, mask_val) (mask_val) |= (1 << (eq_num))

#define ARM_EQS(eventp, mask_val) \
		MOSAL_MMAP_IO_WRITE_DWORD(eventp->eq_arm_reg, MOSAL_cpu_to_be32(mask_val))
	
#define ARM_EQ_AT_END(eventp, eq_num)\
        MOSAL_MMAP_IO_WRITE_DWORD(eventp->eq_arm_reg, MOSAL_cpu_to_be32(1 << (eq_num)))

#define OVERRUN_CHECK(label) 
#define OVERRUN_CHECK_FAILURE(label)		 

#define DISARM_CQ(eventp, eq_num, cqnum) \
	XHH_cqm_update_cmd_sn(eventp->cqm, cqnum)



#endif
