/*
  This software is available to you under a choice of one of two
  licenses.  You may choose to be licensed under the terms of the GNU
  General Public License (GPL) Version 2, available at
  <http://www.fsf.org/copyleft/gpl.html>, or the OpenIB.org BSD
  license, available in the LICENSE.TXT file accompanying this
  software.  These details are also available at
  <http://openib.org/license.html>.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
  BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
  ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
  CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
  SOFTWARE.

  Copyright (c) 2004 Mellanox Technologies Ltd.  All rights reserved.
*/

#ifndef H_XHH_HOB_PRIV_H
#define H_XHH_HOB_PRIV_H

#include <mosal.h>
#include <vapi.h>
#include <hh.h>
#include <arbel_dev_defs.h>
#include <xhh.h>
#include <hob.h>
#include <epool.h>
#include <lamm.h>
#include <cmdif.h>
#include <uarm.h>
#include <pdm.h>
#include <tptm.h>
#include <cqm.h>
#include <qpm.h>
#include <mcgm.h>
#include <eventp.h>
#include <vip_delay_unlock.h>
#include <mem_utils.h>

/********************* TEMPORARY DEFINES, INCLUDED UNTIL THINGS GET FIXED UP ****************/


/* ******************* END TEMPORARY DEFINES ******************* */

#define XHHOBP(hca_hndl)   ((XHH_hob_t)(hca_hndl->device))
#define XHHOB(hca_id)    ((XHH_hob_t)(HH_hca_dev_tbl[hca_id].device))

#define XHH_DEV_LIM_MCG_ENABLED(hob)      (hob->dev_lims.udm == 1)
#define XHH_ARBEL_PAGE_SIZE        (0x2000)
#define XHH_ARBEL_LOG_PAGE_SIZE    (12)

typedef struct XHH_ib_props_st {
        int dummy;
} XHH_ib_props_t;

/* minimum required firmware -- major_rev=1, minor_rev = 0x0015, sub_minor (patch) = 0 */
#define  XHH_MIN_FW_VERSION  MAKE_ULONGLONG(0x0000000100000000)

/* Definitions for default profile */
/* NOTE:  All table sizes MUST be a power of 2 */

typedef struct XHH_icm_allocation_vector_st {
    MT_size_t   log2_uarc_size;
    MT_size_t   log2_mtt_size;
    MT_size_t   log2_mpt_size;
    MT_size_t   log2_qpc_size;
    MT_size_t   log2_eqpc_size; /* QPC alt path */
    MT_size_t   log2_srqc_size;
    MT_size_t   log2_cqc_size;
    MT_size_t   log2_ecqc_size;
    MT_size_t   log2_rdb_size;      /* in-flight rdma */
    MT_size_t   log2_uar_scratch_size;
    MT_size_t   log2_eqc_size;
    MT_size_t   log2_mcg_size;
    MT_size_t   log2_eec_size;
    MT_size_t   log2_eeec_size;    /* EEC alt path */
} XHH_icm_allocation_vector_t;

typedef struct XHH_icm_base_addr_vector_st {
    XHH_icm_va_t   uarc_base_addr;
    XHH_icm_va_t   mtt_base_addr;
    XHH_icm_va_t   mpt_base_addr;
    XHH_icm_va_t   qpc_base_addr;
    XHH_icm_va_t   eqpc_base_addr;
    XHH_icm_va_t   srqc_base_addr;
    XHH_icm_va_t   cqc_base_addr;
    XHH_icm_va_t   ecqc_base_addr;
    XHH_icm_va_t   rdb_base_addr;      /* in-flight rdma */
    XHH_icm_va_t   uar_scratch_base_addr;
    XHH_icm_va_t   eqc_base_addr;
    XHH_icm_va_t   mcg_base_addr;
    XHH_icm_va_t   eec_base_addr;
    XHH_icm_va_t   eeec_base_addr;
} XHH_icm_base_addr_vector_t;

/* casts to guard against inadvertent 64-bit divides */
#define XHH_ICM_ALLOCATION_VEC_SIZE    \
               ((u_int32_t)(((u_int32_t)sizeof(XHH_icm_base_addr_vector_t))/((u_int32_t)sizeof(XHH_icm_va_t))))

typedef enum  {
    XHH_LAM_SIZE_BAD,
    XHH_LAM_SIZE_32M = 1,
    XHH_LAM_SIZE_64M,
    XHH_LAM_SIZE_128M,
    XHH_LAM_SIZE_256M,
    XHH_LAM_SIZE_512M,
    XHH_LAM_SIZE_1024M,
    XHH_LAM_SIZE_2048M,
    XHH_LAM_SIZE_4096M,
    XHH_LAM_SIZE_BIG,
} XHH_lam_size_enum_t;

typedef struct XHH_non_ddr_defaults_st {
    XHH_lam_size_enum_t   lam_size_code;
    MT_size_t             lam_size;
    MT_size_t             max_num_pds;
    MT_size_t             num_external_mem_regions;
    MT_size_t             num_mem_windows;
    MT_size_t             icm_alloc_vec_size;
    MT_size_t             log2_max_uar;
    MT_size_t             log2_max_qps;
    MT_size_t             max_num_qps;
    MT_size_t             log2_max_srqs;
    MT_size_t             max_num_srqs;
    MT_size_t             log2_max_cqs;
    MT_size_t             max_num_cqs;
    MT_size_t             log2_max_eecs;
    MT_size_t             log2_max_mcgs;
    MT_size_t             log2_mcg_entry_size;
    MT_size_t             log2_mcg_hash_size;
    MT_size_t             qps_per_mcg;
    MT_size_t             log2_max_mpt_entries;
    MT_size_t             log2_max_mtt_entries;
    MT_size_t             log2_mtt_entries_per_seg;
    MT_size_t             log2_mtt_segs_per_region;
    u_int8_t              log2_max_eqs;
    u_int8_t              log2_uar_pg_size;
    u_int8_t              log2_inflight_rdma_per_qp;
} XHH_profile_t;

#define XHH_DEF_CQ_PER_QP       1

#define ARBEL_MAX_EQ            64
#define XHH_COMPL_EQ_IX          0
#define XHH_IB_EQ_IX             1
#define XHH_CMD_EQ_IX            2
#define XHH_MT_EQ_IX             3

typedef struct XHH_hob_port_info_st {
       u_int32_t    capability_bits;
}XHH_hob_port_info_t;

/* catastrophic error thread structure */
typedef struct XHH_hob_cat_err_thread_st {
    MOSAL_thread_t      mto;
    MOSAL_mutex_t       mutex;
    MOSAL_syncobj_t     start_sync; /* sync object needed on start of thread */
    MOSAL_syncobj_t     stop_sync; /* sync object needed on exit of thread */
    MOSAL_syncobj_t     fatal_err_sync; /* wait on fatal_err_sync object */
    struct XHH_hob_st   *hob;         /* pointer to this thread's HOB object */
    volatile MT_bool    have_fatal; /*TRUE ==> catastrophic error has occurred */
                           /*FALSE ==> just exit. */
} XHH_hob_cat_err_thread_t;

typedef struct XHH_hob_pci_info_st {
    MT_bool     is_valid;
    u_int8_t    bus;
    u_int8_t    dev_func;
    MOSAL_pci_dev_t pci_dev;
    XHH_pci_config_space_container_t config;
} XHH_hob_pci_info_t;

typedef struct XHH_hob_st {
    /* XHH_hob_create parameters */
    u_int32_t            hca_seq_num;
    XHH_module_flags_t   module_flags;
    XHH_hw_props_t       hw_props;

    char                 dev_name[HCA_MAX_NAME];
    u_int32_t            dev_id;
    HH_hca_hndl_t        hh_hca_hndl;

    XHH_fw_props_t       fw_props;   /* QUERY_FW */
    MT_bool              have_lam;
    XHH_lam_props_t      lam_props;  /* QUERY_LAM */
    XHH_dev_lim_t        dev_lims;      /* QUERY_DEV_LIM */
    XHH_adapter_props_t  adapter_props; /* QUERY_ADAPTER */

    XHH_port_init_props_t *init_ib_props; /* VMALLOCed. One entry per port */
    /* HCA Props */
    XHH_hca_props_t      hca_props;

    VAPI_hca_cap_t       hca_capabilities;   /* filled at end of open_hca, and saved for Query_hca */
    u_int64_t            fw_phys_start_addr;
    MOSAL_phys_ctg_t     *fw_phys_mem_vec_p; /* pointer to array of physical memory allocations */
    u_int16_t            fw_phys_mem_vec_size;   /* number of physical allocation "chunks" */
    MOSAL_iomem_t        fw_error_iomem_obj;
    MT_virt_addr_t       fw_error_buf_start_va;
    u_int32_t *          fw_error_buf;  /* kmalloced buffer ready to hold info at cat error */
    MOSAL_phys_ctg_t     *icm_aux_phys_mem_vec_p; /* pointer to array of physical memory allocations */
    u_int16_t            icm_aux_phys_mem_vec_size;   /* number of physical allocation "chunks" */
    
    XHH_icm_allocation_vector_t  icm_alloc_size_vec;
    XHH_icm_base_addr_vector_t   icm_alloc_base_addrs_vec;
    u_int64_t            icm_size;
    u_int64_t            icm_aux_size;
    
    XHH_profile_t        profile;

    /* HH Interface */
    HH_if_ops_t          if_ops;

    /* Version information */
    XHH_ver_info_t       version_info;

    /* EQ handles */
    XHH_eqn_t            compl_eq;
    XHH_eqn_t            ib_eq;

    /* Mutexes, etc */
    MOSAL_mutex_t        mtx;     /* used internally */ 

    /* CONTAINED OBJECTS HANDLES */
    XHH_cmd_t            cmd;
    XHH_lamm_t           lamm;
    XHH_icmm_t           icmm;
    XHH_uarm_t           uarm;
    XHH_pdm_t            pdm;
    XHH_mrwm_t           mrwm;
    XHH_cqm_t            cqm;
    /* XHH_eecm_t           eecm; */
    XHH_qpm_t            qpm;
    XHH_srqm_t           srqm;
    XHH_mcgm_t           mcgm;
	XHH_eventp_t         eventp;

    /* for XHH_get_debug_info() */
    XHH_mrwm_props_t     mrwm_props;

   /* fatal error handling fields */
    VAPI_event_syndrome_t fatal_syndrome;
	MT_bool              from_msix;
    XHH_hob_state_t      xhh_state;
    MOSAL_syncobj_t	     xhh_fatal_complete_syncobj;
    HH_async_eventh_t    async_eventh;   /* saved handler and context, registered by VIP */
    void*                async_ev_private_context;
    MOSAL_spinlock_t     async_spl;
    MOSAL_spinlock_t     fatal_spl;

    VIP_delay_unlock_t   delay_unlocks;
    XHH_hob_cat_err_thread_t  fatal_thread_obj;
	u_int32_t            kar_indx;
    MT_virt_addr_t       kar_addr;
    XHH_hob_pci_info_t   pci_bridge_info;
    XHH_hob_pci_info_t   pci_hca_info;

} XHH_hob_dev_t;

typedef struct {
  XHH_hw_props_t               hw_props;
  XHH_profile_t                profile;
  XHH_icm_base_addr_vector_t   icm_addr_vec;
  XHH_icm_allocation_vector_t  icm_size_vec;
  u_int32_t                    num_icm_addrs;
  XHH_mrwm_props_t             mrwm_props;
  MT_bool                      hide_lam;

  /* Allocated resources count */
  u_int32_t allocated_ul_res; 	/* From ULDM */
  /* (in current implementation ul_res num. is the same as UAR num., excluding UAR1) */
  u_int32_t allocated_pd;	  	/* From ULDM */
  u_int32_t allocated_cq;		/* From TCQM */
  u_int32_t allocated_qp;		/* From TQPM */
  u_int32_t allocated_mr_int;		/* From TMRWM */
  u_int32_t allocated_mr_ext;		/* From TMRWM */
  u_int32_t allocated_mw;		/* From TMRWM */
  u_int32_t allocated_mcg;		/* From MCGM */
} XHH_debug_info_t ;	


HH_ret_t XHH_get_debug_info(
	HH_hca_hndl_t hca_hndl, 		/*IN*/
	XHH_debug_info_t *debug_info_p	/*OUT*/
);


#endif  /* H_XHH_H */
