
/*
  This software is available to you under a choice of one of two
  licenses.  You may choose to be licensed under the terms of the GNU
  General Public License (GPL) Version 2, available at
  <http://www.fsf.org/copyleft/gpl.html>, or the OpenIB.org BSD
  license, available in the LICENSE.TXT file accompanying this
  software.  These details are also available at
  <http://openib.org/license.html>.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
  BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
  ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
  CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
  SOFTWARE.

  Copyright (c) 2004 Mellanox Technologies Ltd.  All rights reserved.
*/

#include "pdm.h"

#define PDM_PD_VIPH2H(hdl,num_rsvd) (hdl+num_rsvd + 1)

#define PDM_PD_H2VIPH(idx,num_rsvd) (idx-num_rsvd -1)


/******************************************************************************
 *  Function:     XHH_pdm_create
 *****************************************************************************/
HH_ret_t XHH_pdm_create( XHH_hob_t    hob, 
                         XHH_pdm_init_t* init_p, 
                         XHH_pdm_t   *pdm_p )
{
    XHH_pdm_t  new_pdm_obj;
    VIP_common_ret_t vcr;

    /* create new pdm object */
    new_pdm_obj = (XHH_pdm_t) MALLOC(sizeof(struct XHH_pdm_st));
    if (!new_pdm_obj) {
        MT_COV_MEM_FLTR_START
        MTL_ERROR1("XHH_pdm_create: MALLOC of new_pdm_obj failed\n");
        return HH_ENOMEM;
        MT_COV_MEM_FLTR_END
    }

    new_pdm_obj->max_pd = init_p->max_pds;
    new_pdm_obj->num_rsvd_pds = init_p->num_rsvd_pds;

    vcr =  VIP_array_create_maxsize(new_pdm_obj->max_pd>>5,new_pdm_obj->max_pd,&(new_pdm_obj->pdm_pd_table));
    if (vcr != VIP_OK) {
        MTL_ERROR1("XHH_pdm_create: creating of pd table failed\n");
        FREE(new_pdm_obj);
        return HH_EAGAIN;
    }

    
    /* save HOB handle */
    new_pdm_obj->hob = hob;

    /* return the object handle */
    *pdm_p =  new_pdm_obj;

    MTL_DEBUG4("LEAVING XHH_pdm_create - OK\n");

    return HH_OK;
}

/******************************************************************************
 *  Function:     XHH_pdm_destroy
 *****************************************************************************/
HH_ret_t XHH_pdm_destroy( /*IN*/ XHH_pdm_t pdm )
{
    VIP_common_ret_t vcr;
        
    MTL_DEBUG4("==> XHH_pdm_destroy\n");

    vcr= VIP_array_destroy(pdm->pdm_pd_table,NULL);
    if (vcr != VIP_OK) {
        MTL_ERROR1("failed destroying pd array \n");
    }

    FREE(pdm);
    MTL_DEBUG4("<== XHH_uarm_destroy\n");
    return HH_OK;

}


/******************************************************************************
 *  Function:     XHH_pdm_alloc_pd
 *****************************************************************************/
HH_ret_t XHH_pdm_alloc_pd( /*IN*/ XHH_pdm_t                   pdm, 
                            /*IN*/ MOSAL_protection_ctx_t       prot_ctx, 
                            /*IN-OUT*/ XHH_pd_ul_resources_t    *pd_ul_resources_p, 
                            /*OUT*/ HH_pd_hndl_t                *p_pd_h )
{   
    VIP_common_ret_t vcr;
    VIP_array_handle_t h;
    
    MTL_DEBUG2("==> XHH_pdm_alloc_pd. pdm = %p, prot_ctx = 0x%x, resources_p = %p\n", (void *) pdm, prot_ctx, (void *) pd_ul_resources_p);
    
    if (pdm == (XHH_pdm_t)XHH_INVALID_HNDL) {
        MTL_ERROR1("<==XHH_pdm_alloc_pd: ERROR : HCA device has not yet been opened\n");
        return HH_EINVAL;
    }


    vcr= VIP_array_insert(pdm->pdm_pd_table,(VIP_array_obj_t)prot_ctx,&h);
    if (vcr != VIP_OK) {
        MT_RETURN(HH_EINVAL);
    } 
                
    *p_pd_h = PDM_PD_VIPH2H(h,pdm->num_rsvd_pds);
    MT_RETURN(HH_OK);
}


 /******************************************************************************
 *  Function:     XHH_pdm_free_pd 
 *****************************************************************************/
HH_ret_t XHH_pdm_free_pd( /*IN*/ XHH_pdm_t    pdm, 
                           /*IN*/ HH_pd_hndl_t  pd )
{

    HH_ret_t           ret = HH_OK;
    VIP_common_ret_t     vcr;
    VIP_array_handle_t h = PDM_PD_H2VIPH(pd,pdm->num_rsvd_pds);

    MTL_DEBUG4("==> XHH_pdm_free_pd\n");
    if (pdm == (XHH_pdm_t)XHH_INVALID_HNDL) {
        MTL_ERROR1("XHH_pdm_free_pd: ERROR : HCA device has not yet been opened\n");
        return HH_EINVAL;
    }
    
    vcr= VIP_array_erase(pdm->pdm_pd_table,h,NULL);
    if (vcr != VIP_OK) {
        return HH_ERR;
    }
    
    MTL_DEBUG4("<== XHH_pdm_free_pd. ret = %d\n", ret);
    return ret;
}

/******************************************************************************
 *  Function:     XHH_pdm_get_protection_ctx 
 *****************************************************************************/
HH_ret_t XHH_get_protection_ctx( /*IN*/ XHH_pdm_t                 pdm, 
                                      /*IN*/ HH_pd_hndl_t               pd, 
                                      /*OUT*/ MOSAL_protection_ctx_t    *prot_ctx_p )
{
    VIP_common_ret_t vcr;
    VIP_array_obj_t obj;
    VIP_array_handle_t h = PDM_PD_H2VIPH(pd,pdm->num_rsvd_pds);
    
    MTL_DEBUG4("==> XHH_pdm_get_protection_ctx. pdm = %p, pd = %d, prot_ctx_p = %p\n", 
               (void *) pdm, pd, (void *)prot_ctx_p);
    if (pdm == (XHH_pdm_t)XHH_INVALID_HNDL) {
        MTL_ERROR1("XHH_pdm_get_protection_ctx: ERROR : HCA device has not yet been opened\n");
        return HH_EINVAL;
    }
    
    /* protect against bad calls */
    if (pdm == (XHH_pdm_t)0) {
        MTL_ERROR1("XHH_pdm_get_protection_ctx: ERROR : pdm handle is ZERO\n");
        return HH_EINVAL;
    }

    vcr= VIP_array_find(pdm->pdm_pd_table,h,&obj);
    if (vcr != VIP_OK) {
        return HH_EINVAL;
    }
    *prot_ctx_p = (MOSAL_protection_ctx_t)obj;

    return HH_OK;
}


