/*
  This software is available to you under a choice of one of two
  licenses.  You may choose to be licensed under the terms of the GNU
  General Public License (GPL) Version 2, available at
  <http://www.fsf.org/copyleft/gpl.html>, or the OpenIB.org BSD
  license, available in the LICENSE.TXT file accompanying this
  software.  These details are also available at
  <http://openib.org/license.html>.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
  BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
  ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
  CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
  SOFTWARE.

  Copyright (c) 2004 Mellanox Technologies Ltd.  All rights reserved.
*/

#ifndef H_XHH_DEV_QPM_H
#define H_XHH_DEV_QPM_H

#include <arbel_if_defs.h>
#include <mapm.h>
#include <pdm.h>
#include <uarm.h>

#define QPM_MAPM_LOG2_BLKSZ 7 /*log number of entries in blk */

#define XHH_DEV_SPECIFIC_QPM_INIT_PARAMS \
  u_int64_t   qpc_tbl_base;      \
  u_int8_t    log2_qpc_entry_sz; \
  u_int64_t   rdb_tbl_base;      \
  u_int8_t    log2_rdb_entry_sz; \
  u_int8_t    log2_rsvd_rdbs;	 \
  u_int64_t   eqpc_tbl_base;     \
  u_int8_t    log2_eqpc_entry_sz; 

#define XHH_DEV_SPECIFIC_QPM_DATA  \
 XHH_mapm_t      qpc_mapm; \
 XHH_mapm_t      eqpc_mapm;\
 XHH_mapm_t      rdb_mapm; \
 XHH_uarm_t      uarm;     \
 XHH_pdm_t       prot_module;
 
#define XHH_DEV_SPECIFIC_QPC_ENTRY_DATA  \
     u_int32_t	snd_db_record_index; \
	 u_int32_t  rcv_db_record_index; \
     u_int8_t   log_sq_stride;       \
     u_int8_t   log_sq_size;         \
     u_int8_t   log_rq_stride;       \
     u_int8_t   log_rq_size;         \
     u_int32_t	snd_wqe_base_adr_l;  \
     u_int32_t	rcv_wqe_base_adr_l; 
 

#define XHH_DEV_SPECIFIC_QPM_INIT  \
  rc= XHH_hob_get_uarm(hob, &qpm->uarm);\
  if (rc != HH_OK) {\
      FREE(qpm);\
      return rc;\
  }\
  rc= XHH_mapm_create(init_attr_p->qpc_tbl_base,     \
                      init_attr_p->log2_max_qp,      \
	                  ((QPM_MAPM_LOG2_BLKSZ < (MOSAL_SYS_PAGE_SHIFT - init_attr_p->log2_qpc_entry_sz+1)) ? \
		              (MOSAL_SYS_PAGE_SHIFT - init_attr_p->log2_qpc_entry_sz+1):QPM_MAPM_LOG2_BLKSZ), \
                      init_attr_p->log2_qpc_entry_sz,\
                      hob,                           \
                      (1 << init_attr_p->log2_rsvd_qps),&(qpm->qpc_mapm));      \
  if (rc != HH_OK) {                                                            \
      MTL_ERROR1(MT_FLFMT("%s: Failed allocation of qpc XHH_mapm object"),__func__);\
      FREE(qpm);                                                                \
      return HH_EAGAIN;                                                         \
  }    \
       \
  rc= XHH_mapm_create(init_attr_p->eqpc_tbl_base,     \
		       init_attr_p->log2_max_qp,      \
	               ((QPM_MAPM_LOG2_BLKSZ < (MOSAL_SYS_PAGE_SHIFT - init_attr_p->log2_eqpc_entry_sz+1)) ? \
		       (MOSAL_SYS_PAGE_SHIFT - init_attr_p->log2_eqpc_entry_sz+1):QPM_MAPM_LOG2_BLKSZ), \
		       init_attr_p->log2_eqpc_entry_sz,\
		       hob,                            \
		       (1 << init_attr_p->log2_rsvd_qps),&(qpm->eqpc_mapm));    \
  if (rc != HH_OK) {                                                            \
	  MTL_ERROR1(MT_FLFMT("%s: Failed allocation of eqpc XHH_mapm object"),__func__);\
	  XHH_mapm_destroy(qpm->qpc_mapm);                                          \
	  FREE(qpm);                                                                \
	  return HH_EAGAIN;                                                         \
  }    \
		   \
  rc= XHH_mapm_create(init_attr_p->rdb_tbl_base,             \
                      init_attr_p->log2_max_qp,              \
                      ((QPM_MAPM_LOG2_BLKSZ < (MOSAL_SYS_PAGE_SHIFT - init_attr_p->log2_rdb_entry_sz+1)) ? \
                           (MOSAL_SYS_PAGE_SHIFT - init_attr_p->log2_rdb_entry_sz+1):QPM_MAPM_LOG2_BLKSZ), \
                      init_attr_p->log2_rdb_entry_sz,        \
                      hob,                                   \
                      (1 << init_attr_p->log2_rsvd_rdbs),    \
                      &(qpm->rdb_mapm));                     \
  if (rc != HH_OK) {                                         \
      MTL_ERROR1(MT_FLFMT("%s: Failed allocation of rdb XHH_mapm object"),__func__);\
      XHH_mapm_destroy(qpm->eqpc_mapm);                                         \
      XHH_mapm_destroy(qpm->qpc_mapm);                                          \
      FREE(qpm);                                                                \
      return HH_EAGAIN;                                                         \
  }
  
  
#define XHH_DEV_SPECIFIC_QPM_DONE \
        XHH_mapm_destroy(qpm->qpc_mapm);  \
        XHH_mapm_destroy(qpm->eqpc_mapm); \
        XHH_mapm_destroy(qpm->rdb_mapm);

#define XHH_DEV_SPECIFIC_QPM_CREATE_QP_SRQ_CHECK do{} while(0)
        
#define XHH_DEV_SPECIFIC_QPM_CREATE_QP(qpix,qp_p) \
      rc= dev_specific_qpm_create_qp(qpm,qp_ul_resources_p,qpix, qp_p); \
      if (rc!= HH_OK) {                                             \
            goto failed_create_qp;                                  \
      }
  
#define XHH_DEV_SPECIFIC_QPM_CREATE_SQP(qpix,qp_p) \
      rc= dev_specific_qpm_create_qp(qpm,qp_ul_resources_p,qpix, qp_p); 

#define XHH_DEV_SPECIFIC_CREATE_QP_UL_RES \
      new_qp_p->log_sq_stride = qp_ul_resources_p->log_sq_stride;           \
      new_qp_p->log_sq_size = qp_ul_resources_p->log_sq_size;               \
      new_qp_p->log_rq_stride = qp_ul_resources_p->log_rq_stride;           \
      new_qp_p->log_rq_size = qp_ul_resources_p->log_rq_size;               \
      new_qp_p->snd_wqe_base_adr_l = qp_ul_resources_p->snd_wqe_base_adr_l; \
      new_qp_p->rcv_wqe_base_adr_l = qp_ul_resources_p->rcv_wqe_base_adr_l; 
      
#define XHH_DEV_SPECIFIC_QPM_DESTROY_QP \
        dev_specific_qpm_destroy_qp(qpm,qp_idx,qp2destroy); 

#define XHH_DEV_SPECIFIC_QPM_GET_PROT_MODULE \
        XHH_hob_get_pdm(hob, &qpm->prot_module)

#define XHH_DEV_SPECIFIC_QPM_INIT_WQE_BASE \
        qpc_p->wqe_base_adr_h = (sizeof(MT_virt_addr_t) <= 4                \
                         ? (u_int32_t)0                                     \
                         : (u_int32_t)(((u_int64_t)qp_p->wqes_buf) >> 32)); \
        qpc_p->log_sq_stride = qp_p->log_sq_stride-4;                       \
        qpc_p->log_rq_stride = qp_p->log_rq_stride-4;                       \
        qpc_p->log_sq_size = qp_p->log_sq_size;                             \
        qpc_p->log_rq_size = qp_p->log_rq_size;                             \
        qpc_p->snd_wqe_base_adr_l = qp_p->snd_wqe_base_adr_l;               \
        qpc_p->rcv_wqe_base_adr_l = qp_p->rcv_wqe_base_adr_l;               \
        qpc_p->rcv_db_record_index = qp_p->rcv_db_record_index;             \
        qpc_p->snd_db_record_index = qp_p->snd_db_record_index;
        
#endif
