/*
  This software is available to you under a choice of one of two
  licenses.  You may choose to be licensed under the terms of the GNU
  General Public License (GPL) Version 2, available at
  <http://www.fsf.org/copyleft/gpl.html>, or the OpenIB.org BSD
  license, available in the LICENSE.TXT file accompanying this
  software.  These details are also available at
  <http://openib.org/license.html>.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
  BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
  ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
  CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
  SOFTWARE.

  Copyright (c) 2004 Mellanox Technologies Ltd.  All rights reserved.
*/

#ifndef CI_XHH_DEV_QPM_CI
#define CI_XHH_DEV_QPM_CI

inline static HH_ret_t dev_specific_qpm_create_qp(XHH_qpm_t  qpm, 
                                                  XHH_qp_ul_resources_t*  qp_ul_resources_p,
                                                  u_int32_t  qpc_index,
                                                  TQPM_sw_qpc_t * new_qp_p)
{
    HH_ret_t rc;
    u_int32_t snd_db_record_index, rcv_db_record_index;
    void * sendq_db_record_va;
    void * recvq_db_record_va;
    FUNC_IN;
	rc= XHH_mapm_alloc_rsc(qpm->qpc_mapm,qpc_index); 
    if (rc!= HH_OK) {
      MTL_ERROR1(MT_FLFMT("%s: failed allocating qpc. qpc_index=0x%x"),__func__, qpc_index);
      MT_RETURN(rc);
    }
    
    rc= XHH_mapm_alloc_rsc(qpm->eqpc_mapm,qpc_index); 
    if (rc!= HH_OK) {
      MTL_ERROR1(MT_FLFMT("%s: failed allocating eqpc. qpc_index=0x%x"),__func__, qpc_index);
      goto alloc_eqpc_fail;
    }
    
	rc= XHH_mapm_alloc_rsc(qpm->rdb_mapm,qpc_index); 
    if (rc!= HH_OK) {
	MTL_ERROR1(MT_FLFMT("%s: failed allocating rdb. qpc_index=0x%x"),__func__, qpc_index);
        goto alloc_rdb_fail;
    }
    
    /* allocate doorbell records for ci and arm */
    rc =  XHH_uarm_alloc_db_record( qpm->uarm, 
                                    qp_ul_resources_p->uar_index, 
                                    UARM_DBRECORD_SEND,
                                    &snd_db_record_index,
                                    &sendq_db_record_va);
    if (rc!= HH_OK) {
        goto alloc_sendq_dbell_fail;
    }

    rc =  XHH_uarm_alloc_db_record( qpm->uarm, 
                                    qp_ul_resources_p->uar_index, 
                                    UARM_DBRECORD_RECEIVE,
                                    &rcv_db_record_index,
                                    &recvq_db_record_va);
    if (rc!= HH_OK) {
        goto alloc_recvq_dbell_fail;
    }

    new_qp_p->uar_index = qp_ul_resources_p->uar_index;
    new_qp_p->snd_db_record_index = snd_db_record_index;
    new_qp_p->rcv_db_record_index = rcv_db_record_index;
    qp_ul_resources_p->sendq_db_record_p = (MT_virt_addr_t) sendq_db_record_va;
    qp_ul_resources_p->recvq_db_record_p = (MT_virt_addr_t) recvq_db_record_va;

    MT_RETURN(HH_OK);

alloc_recvq_dbell_fail:
    XHH_uarm_free_db_record( qpm->uarm, 
                             qp_ul_resources_p->uar_index, 
                             UARM_DBRECORD_SEND,
                             snd_db_record_index);
alloc_sendq_dbell_fail:
    XHH_mapm_free_rsc(qpm->rdb_mapm,qpc_index);

alloc_rdb_fail:
    XHH_mapm_free_rsc(qpm->eqpc_mapm,qpc_index);
      
alloc_eqpc_fail:
    XHH_mapm_free_rsc(qpm->qpc_mapm,qpc_index);
      
    MT_RETURN(rc);
}

inline static HH_ret_t dev_specific_qpm_destroy_qp(XHH_qpm_t  qpm,
                                                   u_int32_t  qpc_index,
                                                  TQPM_sw_qpc_t * qp_p)
{
    FUNC_IN;
    XHH_uarm_free_db_record( qpm->uarm, 
                             qp_p->uar_index, 
                             UARM_DBRECORD_SEND,
                             qp_p->snd_db_record_index);
    
    XHH_uarm_free_db_record( qpm->uarm, 
                             qp_p->uar_index, 
                             UARM_DBRECORD_RECEIVE,
                             qp_p->rcv_db_record_index);
    XHH_mapm_free_rsc(qpm->qpc_mapm,qpc_index);
    XHH_mapm_free_rsc(qpm->eqpc_mapm,qpc_index);
    XHH_mapm_free_rsc(qpm->rdb_mapm,qpc_index);
    MT_RETURN(HH_OK);
}
#endif
