/*
  This software is available to you under a choice of one of two
  licenses.  You may choose to be licensed under the terms of the GNU
  General Public License (GPL) Version 2, available at
  <http://www.fsf.org/copyleft/gpl.html>, or the OpenIB.org BSD
  license, available in the LICENSE.TXT file accompanying this
  software.  These details are also available at
  <http://openib.org/license.html>.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
  BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
  ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
  CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
  SOFTWARE.

  Copyright (c) 2004 Mellanox Technologies Ltd.  All rights reserved.
*/

#include "uarm.h"

#define UARM_UAR_VIPH2IDX(hdl,num_rsvd) (hdl+num_rsvd)

#define UARM_UAR_IDX2VIPH(idx,num_rsvd) (idx-num_rsvd)


HH_ret_t XHH_uarm_get_pg_va(XHH_uarm_t uarm, 
                                   u_int32_t pg_idx, 
                                   XHH_uarm_uar_entry_t* p_uar,
                                   XHH_uar_index_t uar_idx,
                                   OUT MT_virt_addr_t* p_pg_va);

static HH_ret_t XHH_uarm_set_db_record_invalid(  /*IN*/ XHH_uarm_t    uarm, 
                                    /*IN*/ XHH_uar_index_t     uar_idx, 
                                    /*IN*/ u_int32_t           db_record_idx);

static void unmap_icmm_entry(VIP_hash_key_t key, VIP_hashp_value_t hash_val, void* data)
{
    HH_ret_t hh_ret;
    XHH_icmm_t icmm = (XHH_icmm_t)data;
    XHH_uarm_uar_page_entry_t* p_pg = (XHH_uarm_uar_page_entry_t*)hash_val;
    /* unmap the pages in icm */
    hh_ret= XHH_icmm_unmap(icmm,p_pg->icm_va,1<<MOSAL_SYS_PAGE_SHIFT);
    if (hh_ret != HH_OK) {
        MTL_ERROR1(MT_FLFMT("%s: failed to unmap page "),__func__);
    }
    FREE(p_pg);
}

static void XHH_uarm_free_uar_objs(XHH_uarm_t uarm,XHH_uarm_uar_entry_t* uar_p)
{
    HH_ret_t hh_ret= HH_OK;
    call_result_t cr;
    VIP_common_ret_t vcr;
    XHH_icmm_t icmm = NULL;
    XHH_uarm_cmdsn_page_t *cur_cmdsn_page, *next_cmdsn_page;


    /*destroy uar db records */
    vcr= VIP_array_destroy(uar_p->uar_dbrecords_pd,NULL);
    if (vcr != VIP_OK) {
        MTL_ERROR1(MT_FLFMT("%s: failed to destroy db records array "),__func__);
    }

    vcr= VIP_array_destroy(uar_p->uar_dbrecords_wt,NULL);
    if (vcr != VIP_OK) {
        MTL_ERROR1(MT_FLFMT("%s: failed to destroy db records array "),__func__);
    }



    hh_ret= XHH_hob_get_icmm(uarm->hob,&icmm);
    if (hh_ret != HH_OK) {
        MTL_ERROR1("could not get icmm handle\n");
    }

    /* destroy the page hash table, unmapping each entry as you go*/
    vcr = VIP_hashp_destroy(uar_p->uar_hash_pages,unmap_icmm_entry, icmm);

    /* unmap uar from user's context virtual adrs*/
    cr= MOSAL_iomem_unmap(uar_p->iomem_obj,uar_p->prot_ctx);
    if (cr != MT_OK) {
        MTL_ERROR1("failed unmapping uar virt_adrs \n");    
    }

    for (cur_cmdsn_page= uar_p->cmdsn_arrays; cur_cmdsn_page != NULL; cur_cmdsn_page= next_cmdsn_page) {
	    next_cmdsn_page= cur_cmdsn_page->next;
	    cr= MOSAL_mem_phys_ctg_free(cur_cmdsn_page->allocated_page);
	    if (cr != MT_OK) {
		    MTL_ERROR1(MT_STDFMT("Failed freeing cmd_sn page (cur_cmdsn_page=%p) - %s . Possible memory leak !"),
			       cur_cmdsn_page, mtl_strerror_sym(cr));
		    /* Continue to clean - nothing else we can do but complete cleanup */
	    }
	    FREE(cur_cmdsn_page);
    }

    MOSAL_mutex_free(&uar_p->dbrecords_mutex);
    MOSAL_mutex_free(&uar_p->cmdsn_mutex);
    FREE(uar_p);
}

/******************************************************************************
 *  Function:     XHH_uarm_create
 *****************************************************************************/
HH_ret_t XHH_uarm_create( /*IN*/  XHH_hob_t    hob, 
                          /*IN*/  XHH_uarm_init_t* init_p,
                          /*OUT*/ XHH_uarm_t   *uarm_p )
{
    u_int32_t   max_uar;
    XHH_uarm_t  new_uarm_obj;
    VIP_common_ret_t vcr;

    /* create new uarm object */
    new_uarm_obj = (XHH_uarm_t) MALLOC(sizeof(XHH_uarm_obj_t));
    if (!new_uarm_obj) {
        MTL_ERROR1("XHH_uarm_create: MALLOC of new_uarm_obj failed\n");
        return HH_EAGAIN;
    }

    max_uar = (1 << init_p->log2_max_uar);

    /* create uar table */
    vcr =  VIP_array_create_maxsize(max_uar>>5,max_uar,&(new_uarm_obj->uarm_uar_table));
    if (vcr != VIP_OK) {
        MTL_ERROR1("XHH_uarm_create: creating of uar table failed\n");
        FREE(new_uarm_obj);
        return HH_EAGAIN;
    }

    new_uarm_obj->uar_base = init_p->uar_base;
    new_uarm_obj->uar_ctx_base = init_p->uar_ctx_base;
    new_uarm_obj->log2_max_uar = init_p->log2_max_uar;
    new_uarm_obj->log2_uar_pg_sz = init_p->log2_uar_pg_sz;
    new_uarm_obj->log2_uar_ctx_sz = init_p->log2_uar_ctx_sz - init_p->log2_max_uar;
    new_uarm_obj->num_rsvd_uars = init_p->num_rsvd_uars;
    new_uarm_obj->hob = hob;

    *uarm_p =  new_uarm_obj;

    return HH_OK;
}

/******************************************************************************
 *  Function:     XHH_uarm_destroy
 *****************************************************************************/
HH_ret_t XHH_uarm_destroy( /*IN*/ XHH_uarm_t uarm )
{
    VIP_common_ret_t vcr;
    XHH_uarm_uar_entry_t* uar_p= NULL;
    VIP_array_obj_t obj;
    VIP_array_handle_t hdl;

    MTL_DEBUG4("==> XHH_uarm_destroy\n");

    VIP_ARRAY_FOREACH(uarm->uarm_uar_table,vcr,hdl,&obj) {
        uar_p= (XHH_uarm_uar_entry_t*)obj;
        XHH_uarm_free_uar_objs(uarm,uar_p);
    }

    vcr= VIP_array_destroy(uarm->uarm_uar_table,NULL);
    if (vcr != VIP_OK) {
        MTL_ERROR1("failed destroying uar table\n");
    }
    FREE(uarm);
    MTL_DEBUG4("<== XHH_uarm_destroy\n");
    return HH_OK;
}

/*************************************************************************
 * Function: XHH_uarm_alloc_ul_res
 *************************************************************************/ 
HH_ret_t XHH_uarm_alloc_ul_res( /*IN*/ XHH_uarm_t              uarm, 
                                /*IN*/ MOSAL_protection_ctx_t  prot_ctx, 
                                /*OUT*/ XHH_hca_ul_resources_t *hca_ul_resources_p )
{
    XHH_uar_index_t    uar_index;
    MT_virt_addr_t  uar_map;
    HH_ret_t     hh_ret = HH_OK;

    MTL_DEBUG4("==> XHH_uarm_alloc_ul_res. prot_ctx = 0x%x\n", prot_ctx);

    hh_ret = XHH_uarm_alloc_uar(uarm,prot_ctx, &uar_index, &uar_map);
    if (hh_ret != HH_OK) {
        MTL_ERROR1("%s: failed allocating UAR. ERROR = %d\n", __func__, hh_ret);
        goto Exit_err;
    }

    hca_ul_resources_p->uar_index = uar_index;
    hca_ul_resources_p->uar_map   = uar_map;

    Exit_err:
    MTL_DEBUG4("<== XHH_uarm_alloc_ul_res. ret = %d\n", hh_ret);
    return hh_ret;
}

/*************************************************************************
 * Function: XHH_uarm_free_ul_res
 *************************************************************************/ 
HH_ret_t XHH_uarm_free_ul_res( /*IN*/ XHH_uarm_t             uarm, 
                               /*IN*/ XHH_hca_ul_resources_t *hca_ul_resources_p)
{
    HH_ret_t     hh_ret = HH_OK;

    MTL_DEBUG4("==> XHH_uarm_free_ul_res. resources ptr = %p\n", (void *) hca_ul_resources_p);
    if (hca_ul_resources_p->uar_index > 1) {
        hh_ret = XHH_uarm_free_uar(uarm, hca_ul_resources_p->uar_index);
        if (hh_ret != HH_OK) {
            MTL_ERROR1("%s: failed freeing UAR index %d. ERROR = %d\n", __func__, hca_ul_resources_p->uar_index, hh_ret);
            goto Exit_err;
        }
    }

    hca_ul_resources_p->uar_index = 0;
    hca_ul_resources_p->uar_map   = 0;

    Exit_err:
    MTL_DEBUG4("<== XHH_uarm_free_ul_res. ret = %d\n", hh_ret);
    return hh_ret;
}

/*************************************************************************
 * Function: XHH_uarm_alloc_uar
 *************************************************************************/
HH_ret_t XHH_uarm_alloc_uar( /*IN*/ XHH_uarm_t              uarm, 
                             /*IN*/ MOSAL_protection_ctx_t  prot_ctx, 
                             /*OUT*/ XHH_uar_index_t       *uar_idx_p, 
                             /*OUT*/ MT_virt_addr_t            *uar_map )
{
    HH_ret_t     hh_ret = HH_OK;
    XHH_uarm_uar_entry_t* p_uar_ent;
    VIP_common_ret_t  vcr;
    VIP_array_handle_t uar_hdl;
    u_int32_t uar_max_dbr, uar_max_pages;
    MT_phys_addr_t uar_phys_adrs;
    call_result_t cr;

    UARM_CHECK(uarm)

    p_uar_ent = MALLOC(sizeof(XHH_uarm_uar_entry_t));
    if (p_uar_ent == NULL) {
        return HH_ENOMEM;
    }

    vcr= VIP_array_insert(uarm->uarm_uar_table,p_uar_ent,&uar_hdl);
    if (vcr != VIP_OK) {
        hh_ret= HH_ERR;
        goto Exit_err4;
    } else *uar_idx_p= UARM_UAR_VIPH2IDX(uar_hdl,uarm->num_rsvd_uars);

    p_uar_ent->prot_ctx = prot_ctx;
    MTL_DEBUG4(MT_FLFMT("%s: allocated uar hdl: 0x%x \n"),__func__,uar_hdl);
    MTL_DEBUG4(MT_FLFMT("%s: rsvd_uars:0x%x     uar_idx:0x%x  \n"),__func__,uarm->num_rsvd_uars,*uar_idx_p);

    /* map uar and get its virtual address in user's ctx*/
    uar_phys_adrs= uarm->uar_base + ( *uar_idx_p << uarm->log2_uar_pg_sz);
    cr= MOSAL_iomem_map(p_uar_ent->prot_ctx,
                        MOSAL_MEM_FLAGS_NO_CACHE | MOSAL_MEM_FLAGS_PERM_WRITE, 
                        uar_phys_adrs,
                        1<<uarm->log2_uar_pg_sz,
                        &(p_uar_ent->iomem_obj),
                        &p_uar_ent->va);
    if (cr != MT_OK) {
        hh_ret= HH_EINVAL;
        goto Exit_err3;
    }else *uar_map = p_uar_ent->va; 
    
    /* create pages array for managing the uar ctx */
    /* uar ctx is devided into pages in sys_page size */
    uar_max_pages = 1 << (uarm->log2_uar_ctx_sz - MOSAL_SYS_PAGE_SHIFT);
    vcr =  VIP_hashp_create_maxsize(UARM_UAR_PAGES_INIT_NUM,uar_max_pages,
                                    &p_uar_ent->uar_hash_pages);
    if (vcr != VIP_OK) {
        MTL_ERROR1("%s: creating of pages hash table failed\n", __func__);
        hh_ret= HH_ENOMEM;
        goto Exit_err2;
    }

    /* create the 2 db records arrays for the 2 types: work through & TBD */
    
    uar_max_dbr = 1 << (uarm->log2_uar_ctx_sz -  UARM_LOG2_DB_RECORD_SZ);
    vcr =  VIP_array_create_maxsize(UARM_UAR_DB_RECORDS_INIT_NUM,uar_max_dbr,
                                    &p_uar_ent->uar_dbrecords_pd);
    if (vcr != VIP_OK) {
        MTL_ERROR1("XHH_uarm_create: creating of db records array failed\n");
        hh_ret= HH_ENOMEM;
        goto Exit_err1;
    }
    p_uar_ent->uar_dbrecords_pd_cnt = 0;
    p_uar_ent->uar_dbrecords_pd_max_idx = 0xffffffff;
    

    vcr =  VIP_array_create_maxsize(UARM_UAR_DB_RECORDS_INIT_NUM,uar_max_dbr,
                                    &p_uar_ent->uar_dbrecords_wt);
    if (vcr != VIP_OK) {
        MTL_ERROR1("XHH_uarm_create: creating of db records array failed\n");
        hh_ret= HH_ENOMEM;
        goto Exit_err0;
    }
    p_uar_ent->uar_dbrecords_wt_cnt = 0;
    p_uar_ent->uar_dbrecords_wt_max_idx = 0xffffffff;

    p_uar_ent->cmdsn_arrays= NULL;
    
    MOSAL_mutex_init(&p_uar_ent->dbrecords_mutex);
    MOSAL_mutex_init(&p_uar_ent->cmdsn_mutex);
    
    return HH_OK;

    Exit_err0:
        VIP_array_destroy(p_uar_ent->uar_dbrecords_pd,NULL);
    Exit_err1:   
        VIP_hashp_destroy(p_uar_ent->uar_hash_pages,NULL,NULL);
    Exit_err2:            
    MOSAL_iomem_unmap(p_uar_ent->iomem_obj,p_uar_ent->prot_ctx);
    Exit_err3:            
    VIP_array_erase(uarm->uarm_uar_table,uar_hdl,NULL); 
    Exit_err4:            
    FREE(p_uar_ent);
    return hh_ret;

}

/******************************************************************************
 *  Function:     XHH_uarm_free_uar
 *****************************************************************************/
HH_ret_t XHH_uarm_free_uar( /*IN*/ XHH_uarm_t    uarm, 
                            /*IN*/ XHH_uar_index_t    uar_index )
{
    HH_ret_t       hh_ret = HH_OK;
    XHH_uarm_uar_entry_t* p_uar = NULL;
    VIP_common_ret_t vcr = VIP_OK;
    VIP_array_obj_t uar_obj;
    VIP_array_handle_t uar_hdl = UARM_UAR_IDX2VIPH(uar_index,uarm->num_rsvd_uars);

    UARM_CHECK(uarm)
    
    vcr= VIP_array_erase_prepare(uarm->uarm_uar_table,uar_hdl,&uar_obj);
    if (vcr != VIP_OK) {
        MTL_ERROR1(MT_FLFMT("%s: Failed finding uar to hdl: 0x%x \n"),__func__,uar_hdl);
        return HH_EINVAL;
    }
      
    p_uar= (XHH_uarm_uar_entry_t*)uar_obj;
    MTL_DEBUG4(MT_FLFMT("%s: found uar to hdl: 0x%x \n"),__func__,uar_hdl);

    XHH_uarm_free_uar_objs(uarm, p_uar);
    VIP_array_erase_done(uarm->uarm_uar_table,uar_hdl,NULL);

    return hh_ret;
}


/******************************************************************************
 *  Function:     XHH_uarm_alloc_db_record
 *****************************************************************************/
HH_ret_t XHH_uarm_alloc_db_record(  /*IN*/ XHH_uarm_t                   uarm, 
                                    /*IN*/ XHH_uar_index_t              uar_idx, 
                                    /*IN*/ XHH_uarm_dbrecord_type_t     dbr_type,
                                    /*OUT*/ u_int32_t*                  p_db_record_idx,
                                    /*OUT*/ void**                      db_record_va_p)

{   
    XHH_uarm_uar_entry_t* p_uar=NULL;
    HH_ret_t hh_ret= HH_OK;
    VIP_common_ret_t vcr;
    VIP_array_handle_t dbr_hdl;
    u_int32_t pg_idx;
    VIP_array_obj_t obj;
    MT_virt_addr_t pg_va;
    u_int32_t max_dbrs_in_uar;
    u_int32_t db_record_idx;
    VIP_array_handle_t uar_hdl = UARM_UAR_IDX2VIPH(uar_idx,uarm->num_rsvd_uars);
    u_int32_t db_rec_pg_offset;

    UARM_CHECK(uarm)

    /* find the uar - keep the uar from being destroyed*/
    vcr= VIP_array_find_hold(uarm->uarm_uar_table,uar_hdl,&obj);
    if (vcr != VIP_OK) {
        return HH_EINVAL;
    } else p_uar= (XHH_uarm_uar_entry_t*)obj;

    max_dbrs_in_uar = 1 << (uarm->log2_uar_ctx_sz - UARM_LOG2_DB_RECORD_SZ); 
    
    /* no empty db record entries were left +need to save space for the terminator */
    if ((p_uar->uar_dbrecords_pd_cnt + p_uar->uar_dbrecords_wt_cnt) == 
                ( max_dbrs_in_uar -1)){
        hh_ret= HH_EBUSY;
        goto Exit1;
    }
    
    /* allocate new db record entry */
    switch (dbr_type) {
    case UARM_DBRECORD_SEND:
    case UARM_DBRECORD_ARM_CQ:
            vcr= VIP_array_insert(p_uar->uar_dbrecords_wt,NULL,&dbr_hdl);
            if (vcr != VIP_OK) {
                hh_ret= HH_ERR;
                goto Exit1;
            } else {
                        *p_db_record_idx= (u_int32_t)dbr_hdl;         
                        p_uar->uar_dbrecords_wt_cnt++;
            }
        break;
    case UARM_DBRECORD_SET_CI:
    case UARM_DBRECORD_RECEIVE:
            vcr= VIP_array_insert(p_uar->uar_dbrecords_pd,NULL,&dbr_hdl);
            if (vcr != VIP_OK) {
                hh_ret= HH_ERR;
                goto Exit1;
            } else {
                *p_db_record_idx= max_dbrs_in_uar - (u_int32_t)dbr_hdl -1 ;         
                p_uar->uar_dbrecords_pd_cnt++;
            }
        break;
    default: hh_ret= HH_EINVAL;
             goto Exit1;
    }
    
    db_record_idx = *p_db_record_idx;
    /* page index in the uar ctx*/
    pg_idx = db_record_idx >> ( MOSAL_SYS_PAGE_SHIFT - UARM_LOG2_DB_RECORD_SZ);   
	db_rec_pg_offset= ((db_record_idx & ((1<<(MOSAL_SYS_PAGE_SHIFT - UARM_LOG2_DB_RECORD_SZ))-1)) <<
						UARM_LOG2_DB_RECORD_SZ);
    hh_ret= XHH_uarm_get_pg_va(uarm, pg_idx, p_uar, uar_idx, &pg_va);
    if (hh_ret== HH_OK) {
        *db_record_va_p = (void*)(pg_va + db_rec_pg_offset);
	 MTL_DEBUG1(MT_FLFMT("%s: OK. type=%d, record_va=%p, record_ix=%d, uar_ix=%u, pg_va="VIRT_ADDR_FMT),
		   __func__, dbr_type, *db_record_va_p, db_record_idx, uar_idx, pg_va);
    }else {
        switch (dbr_type) {
        case UARM_DBRECORD_SEND:
        case UARM_DBRECORD_ARM_CQ:
            vcr= VIP_array_erase(p_uar->uar_dbrecords_wt,dbr_hdl,NULL);
            if (vcr == VIP_OK) {
                p_uar->uar_dbrecords_wt_cnt--;
            }
           break;
        case UARM_DBRECORD_SET_CI:
        case UARM_DBRECORD_RECEIVE:
                   vcr= VIP_array_erase(p_uar->uar_dbrecords_pd,dbr_hdl,NULL);
                    if (vcr == VIP_OK) {
                        p_uar->uar_dbrecords_pd_cnt--;
                    }
                   break;
        default: ;
        }
            
   }

  Exit1:    
    VIP_array_find_release(uarm->uarm_uar_table,uar_hdl);
    return hh_ret;
}

/******************************************************************************
 *  Function:     XHH_uarm_free_db_record 
 *****************************************************************************/
HH_ret_t XHH_uarm_free_db_record(XHH_uarm_t uarm, XHH_uar_index_t uar_idx, 
                                 XHH_uarm_dbrecord_type_t dbr_type, u_int32_t db_record_idx)

{   
    VIP_common_ret_t vcr;
    VIP_array_obj_t obj;
    XHH_uarm_uar_entry_t* p_uar= NULL;
    VIP_array_handle_t dbr_h;
    HH_ret_t ret = HH_OK;
    VIP_array_handle_t uar_hdl = UARM_UAR_IDX2VIPH(uar_idx,uarm->num_rsvd_uars);

    UARM_CHECK(uarm)

    /* find uar*/
    vcr= VIP_array_find(uarm->uarm_uar_table,uar_hdl,&obj);
    if (vcr != VIP_OK) {
        return HH_EINVAL;
    } else p_uar= (XHH_uarm_uar_entry_t*)obj;

    switch (dbr_type) {
    case UARM_DBRECORD_SEND:
    case UARM_DBRECORD_ARM_CQ:
            dbr_h = db_record_idx;
            /* need erase_prep/done because do not want to be able to allocate the record */
            /* again until setting the record invalid has succeeded */
            vcr= VIP_array_erase_prepare(p_uar->uar_dbrecords_wt,dbr_h,NULL);
            if (vcr != VIP_OK) {
                ret= HH_EINVAL;
                break;
            }
            /* invalidate the record */
            ret=XHH_uarm_set_db_record_invalid(uarm, uar_idx, db_record_idx);
            if (ret != HH_OK) {
                /* JPM:  SHOULD THIS BE A FATAL ERROR ??*/
                MTL_ERROR1(MT_FLFMT("%s:XHH_uarm_set_db_record_invalid for type %d failed(%d:%s)"),
                           __func__, dbr_type, ret, HH_strerror_sym(ret));
            }
            
            /* JPM: should we undo the "prep" here if setting record invalid failed ?*/
            vcr= VIP_array_erase_done(p_uar->uar_dbrecords_wt, dbr_h, NULL);
            if (vcr != VIP_OK) {
                ret= HH_EINVAL;
            }else p_uar->uar_dbrecords_wt_cnt--;

           break;
    case UARM_DBRECORD_SET_CI:
    case UARM_DBRECORD_RECEIVE:
            dbr_h = (1 << (uarm->log2_uar_ctx_sz - UARM_LOG2_DB_RECORD_SZ)) - db_record_idx -1;
            
            /* need erase_prep/done because do not want to be able to allocate the record */
            /* again until setting the record invalid has succeeded */
            vcr= VIP_array_erase_prepare(p_uar->uar_dbrecords_pd,dbr_h,NULL);
            if (vcr != VIP_OK) {
                ret= HH_EINVAL;
                break;
            }
            /* invalidate the record */
            ret=XHH_uarm_set_db_record_invalid(uarm, uar_idx, db_record_idx);
            if (ret != HH_OK) {
                /* JPM:  SHOULD THIS BE A FATAL ERROR ??*/
                MTL_ERROR1(MT_FLFMT("%s:XHH_uarm_set_db_record_invalid for type %d failed(%d:%s)"),
                           __func__, dbr_type, ret, HH_strerror_sym(ret));
            }
            
            /* JPM: should we undo the "prep" here if setting record invalid failed ?*/
            vcr= VIP_array_erase_done(p_uar->uar_dbrecords_pd, dbr_h, NULL);
            if (vcr != VIP_OK) {
                ret= HH_EINVAL;
            }else p_uar->uar_dbrecords_pd_cnt--;

        break;
    default: ret= HH_EINVAL;
    }
    
    VIP_array_find_release(uarm->uarm_uar_table,uar_hdl);
    return ret;

}

HH_ret_t 
XHH_uarm_get_pg_va(XHH_uarm_t uarm, 
                            u_int32_t pg_idx, 
                            XHH_uarm_uar_entry_t* p_uar,
                            XHH_uar_index_t uar_idx,
                            OUT MT_virt_addr_t* p_pg_va)
{
    VIP_common_ret_t vcr;
    VIP_hashp_value_t obj;
    XHH_icmm_t icmm = NULL;
    XHH_icm_va_t  uar_ctx_icm_va;
    u_int64_t pg_icm_va;
    HH_ret_t ret= HH_EINVAL;
    XHH_uarm_uar_page_entry_t* p_pg_ent;
    static int num_times_called = 0;

    num_times_called++;
    /* TBD - move the mutex*/
    MOSAL_mutex_acq(&p_uar->dbrecords_mutex,TRUE);

    uar_ctx_icm_va= uarm->uar_ctx_base + (u_int64_t)(((u_int64_t)uar_idx) << uarm->log2_uar_ctx_sz); 
    pg_icm_va = uar_ctx_icm_va + (u_int64_t)(((u_int64_t)pg_idx) << MOSAL_SYS_PAGE_SHIFT); 
    
    /* is the page has been mapped yet? */
    vcr= VIP_hashp_find(p_uar->uar_hash_pages,(VIP_hashp_key_t)(pg_icm_va>>MT_ARBEL_PAGE_SHIFT),&obj);
    switch (vcr) {
    case VIP_OK:
        /* page is mapped and was found*/
        p_pg_ent= (XHH_uarm_uar_page_entry_t*)obj;
        *p_pg_va = p_pg_ent->va;
        ret= HH_OK;
        break;

    case VIP_EINVAL_HNDL:
         /*page does not have va in icm, needs to be mapped */  
        ret= XHH_hob_get_icmm(uarm->hob,&icmm);
        if (ret != HH_OK) {
            MTL_ERROR1("could not get icmm handle\n");
            break;
        }
        
        p_pg_ent= MALLOC(sizeof(XHH_uarm_uar_page_entry_t));
        if (p_pg_ent == NULL) {
            ret= HH_EAGAIN;
            break;
        }

        
        ret= XHH_icmm_set_n_map(icmm, 
                                pg_icm_va,
                                (1 << MOSAL_SYS_PAGE_SHIFT),
                                ICMM_MAP_FORCE_NONE,
                                p_uar->prot_ctx,
                                UARM_DBR_TERMINATOR,
                                p_pg_va);
        if (ret != HH_OK) {
            FREE(p_pg_ent);
            break;
        }
        p_pg_ent->icm_va = pg_icm_va;
        p_pg_ent->va= *p_pg_va;

        vcr= VIP_hashp_insert(p_uar->uar_hash_pages,(VIP_hashp_key_t)(pg_icm_va>>MT_ARBEL_PAGE_SHIFT),
                              (VIP_hashp_value_t)p_pg_ent);
        if (vcr != VIP_OK) {
            XHH_icmm_unmap(icmm,pg_icm_va,(1 << MOSAL_SYS_PAGE_SHIFT));
            FREE(p_pg_ent);
            ret= HH_ERR;
        }
        break;

    default: ;
    }
            
    MOSAL_mutex_rel(&p_uar->dbrecords_mutex);
    return ret;

}

static HH_ret_t XHH_uarm_set_db_record_invalid(  /*IN*/ XHH_uarm_t    uarm, 
                                    /*IN*/ XHH_uar_index_t     uar_idx, 
                                    /*IN*/ u_int32_t           db_record_idx)
{
    HH_ret_t    rc = HH_OK;
    XHH_icmm_t  icmm = NULL;
    u_int64_t   icm_va, uar_ctx_icm_va;

    /* Base of this UAR context entry*/
    uar_ctx_icm_va= uarm->uar_ctx_base + (u_int64_t) (uar_idx << uarm->log2_uar_ctx_sz);
    
    /* add in the offset of the record */
    icm_va = uar_ctx_icm_va + (u_int64_t)(db_record_idx << UARM_LOG2_DB_RECORD_SZ);

    rc= XHH_hob_get_icmm(uarm->hob,&icmm);
    if (rc != HH_OK) {
        MTL_ERROR1(MT_FLFMT("%s: XHH_hob_get_icmm failed (%d:%s)"),
                   __func__, rc, HH_strerror_sym(rc));
        return rc;
    }
    
    /* set the record to all zeroes to invalidate it*/
    rc = XHH_icmm_memset(icmm, icm_va, 0, (1UL << UARM_LOG2_DB_RECORD_SZ)) ;
    if (rc != HH_OK) {
        MTL_ERROR1(MT_FLFMT("%s: XHH_icmm_memset failed (%d:%s)"),
                   __func__, rc, HH_strerror_sym(rc));
        return rc;
    }
    return rc;
}


/* Allocate cmdsn and link to the pages list (assumed cmdsn_mutex is held) */
static XHH_uarm_cmdsn_page_t* alloc_cmdsn_page(XHH_uarm_uar_entry_t* uar_p)
{
	XHH_uarm_cmdsn_page_t *new_page;
	call_result_t rc;
	unsigned int index;
	MT_virt_addr_t tmp_va;

	new_page= MALLOC(sizeof(XHH_uarm_cmdsn_page_t));
	if (new_page == NULL) {
		MTL_ERROR4(MT_STDFMT("Failed allocating XHH_uarm_cmdsn_page_t"));
		goto failed_malloc;
	}

	rc= MOSAL_mem_phys_ctg_alloc(MOSAL_SYS_PAGE_SHIFT, MOSAL_SYS_PAGE_SHIFT, 
				     MOSAL_MEM_ALLOC_FLAGS_NONE, &new_page->allocated_page);
	if (rc != MT_OK) {
		MTL_ERROR4(MT_STDFMT("Failed allocating physical page"));
		goto failed_phys_alloc;
	}

	rc= MOSAL_mem_phys_ctg_map( new_page->allocated_page, mosal_get_kernel_prot_ctx(), 
				    MOSAL_MEM_FLAGS_PERM_READ | MOSAL_MEM_FLAGS_PERM_WRITE, 
				    &tmp_va);
	if (rc != MT_OK) {
		MTL_ERROR4(MT_STDFMT("Failed mapping physical page to kernel space"));
		goto failed_phys_map;
	}
	new_page->page_va= (void*)tmp_va;


	if (uar_p->prot_ctx != mosal_get_kernel_prot_ctx()) {
		rc= MOSAL_mem_phys_ctg_map( new_page->allocated_page, uar_p->prot_ctx, 
					    MOSAL_MEM_FLAGS_PERM_READ, 
					    &tmp_va);
		if (rc != MT_OK) {
			MTL_ERROR4(MT_STDFMT("Failed mapping physical page to user space"));
			goto failed_phys_map;
		}
		new_page->consumer_page_va= (void*)tmp_va;

	} else {
		new_page->consumer_page_va= new_page->page_va; /* Consumer is in the same address space */
	}

	/* Initialize free lists */
	for (index= 0; index < MOSAL_SYS_PAGE_SIZE; index++) {
		new_page->page_va[index]= (index+1) & 0xFF; /* Each byte points to the next byte */
	}

	/* Link new page to linked list in the UAR object */
	new_page->next= uar_p->cmdsn_arrays;
	uar_p->cmdsn_arrays= new_page;

	return new_page;

	failed_phys_map:
		MOSAL_mem_phys_ctg_free(new_page->allocated_page);
	failed_phys_alloc:
		FREE(new_page);
	failed_malloc:
		return NULL;
}

/* Allocate a byte for the expected cmd_sn of a CQ */
HH_ret_t XHH_uarm_alloc_cmdsn_byte( XHH_uarm_t          uarm,		/*IN*/ 
                                    XHH_uar_index_t     uar_idx,	/*IN*/ 
				    u_int8_t**		kernel_cmdsn_pp,	/*OUT*/
				    u_int8_t**		consumer_cmdsn_pp)	/*OUT*/
{
	VIP_common_ret_t vrc;
	VIP_array_obj_t obj;
	HH_ret_t rc= HH_OK;
	VIP_array_handle_t uar_hdl = UARM_UAR_IDX2VIPH(uar_idx,uarm->num_rsvd_uars);
	XHH_uarm_uar_entry_t* uar_p= NULL;
	unsigned int segment_index= 0;
	unsigned int free_index= 0;
	XHH_uarm_cmdsn_page_t *cur_cmdsn_page;
	
	UARM_CHECK(uarm)

	/* find uar*/
	vrc= VIP_array_find(uarm->uarm_uar_table,uar_hdl,&obj);
	if (vrc != VIP_OK) {
	    return HH_EINVAL;
	} else uar_p= (XHH_uarm_uar_entry_t*)obj;

	if (MOSAL_mutex_acq(&uar_p->cmdsn_mutex, TRUE) != MT_OK) {
		rc= HH_EINTR;
		goto find_release_exit;
	}
	/* Scan all allocated pages for a free byte */
	for (cur_cmdsn_page= uar_p->cmdsn_arrays; cur_cmdsn_page != NULL; 
	      cur_cmdsn_page= cur_cmdsn_page->next) {
		/* scan each (256B) segment and try to get a byte from "free_list" */
		for (segment_index= 0; segment_index < MOSAL_SYS_PAGE_SIZE; segment_index+= 256) {
			free_index= cur_cmdsn_page->page_va[segment_index]; /* Head of free list */
			if (free_index != 0) break; /* Found a non-empty bytes list */
		}
		if (free_index != 0) break; /* Found a non-empty bytes list */
	}

	if (free_index == 0) { /* No free byte found - allocate new page */
		cur_cmdsn_page= alloc_cmdsn_page(uar_p);
		if (cur_cmdsn_page == NULL) {
			rc= HH_EAGAIN;
			goto mutex_release_exit;
		}
		/* New page - take first byte in first (256) segment */
		segment_index= 0;
		free_index= 1; 
	}

	/* Remove byte from free list (next is given in the free byte removed) */
	cur_cmdsn_page->page_va[segment_index]= 
		cur_cmdsn_page->page_va[segment_index+free_index];
	cur_cmdsn_page->page_va[segment_index+free_index]= 1; /* Initialize cmd_sn */ 
	/* (ER: there is an update to PRM 0.45 that requires first cmd_sn to be 1) */
	/* Addresses to return */
	*kernel_cmdsn_pp= 
		cur_cmdsn_page->page_va + segment_index + free_index;
	*consumer_cmdsn_pp= 
		cur_cmdsn_page->consumer_page_va + segment_index + free_index;

	mutex_release_exit:
	MOSAL_mutex_rel(&uar_p->cmdsn_mutex);
	find_release_exit:
	VIP_array_find_release(uarm->uarm_uar_table, uar_hdl);
	return rc;
}

/* Free cmd_sn byte */
HH_ret_t XHH_uarm_free_cmdsn_byte( XHH_uarm_t          uarm,		/*IN*/ 
                                    XHH_uar_index_t     uar_idx,	/*IN*/ 
				    u_int8_t*		kernel_cmdsn_p)	/*IN*/
{
	VIP_common_ret_t vrc;
	VIP_array_obj_t obj;
	HH_ret_t rc= HH_EINVAL_VA; /* Default error when given cmdsn address is not found */
	VIP_array_handle_t uar_hdl = UARM_UAR_IDX2VIPH(uar_idx,uarm->num_rsvd_uars);
	XHH_uarm_uar_entry_t* uar_p= NULL;
	unsigned int full_index, segment_index;
	XHH_uarm_cmdsn_page_t *cur_cmdsn_page;
	
	UARM_CHECK(uarm)

	/* find uar*/
	vrc= VIP_array_find(uarm->uarm_uar_table,uar_hdl,&obj);
	if (vrc != VIP_OK) {
	    return HH_EINVAL;
	} else uar_p= (XHH_uarm_uar_entry_t*)obj;

	if (MOSAL_mutex_acq(&uar_p->cmdsn_mutex, TRUE) != MT_OK) {
		rc= HH_EINTR;
		goto find_release_exit;
	}

	/* Lookup the page that holds this cmdsn byte */
	for (cur_cmdsn_page= uar_p->cmdsn_arrays; cur_cmdsn_page != NULL; 
	      cur_cmdsn_page= cur_cmdsn_page->next) {
		if ((kernel_cmdsn_p >= cur_cmdsn_page->page_va) && 
		    (kernel_cmdsn_p < (cur_cmdsn_page->page_va + MOSAL_SYS_PAGE_SIZE)) ) {
			/* given cmdsn is within this page */
			full_index= (unsigned int)(MT_ulong_ptr_t)(kernel_cmdsn_p - cur_cmdsn_page->page_va);
			segment_index= full_index & ~0xFF;
			/* Insert byte into free list of its segment */
			 /* Link freed byte to first in segment's free list */
			cur_cmdsn_page->page_va[full_index]= cur_cmdsn_page->page_va[segment_index];
			/* Then make freed byte be the first */
			cur_cmdsn_page->page_va[segment_index]= full_index & 0xFF;
			rc= HH_OK;
			break;
		}
	}


	MOSAL_mutex_rel(&uar_p->cmdsn_mutex);
	find_release_exit:
	VIP_array_find_release(uarm->uarm_uar_table, uar_hdl);
	return rc;
}

