/*
  This software is available to you under a choice of one of two
  licenses.  You may choose to be licensed under the terms of the GNU
  General Public License (GPL) Version 2, available at
  <http://www.fsf.org/copyleft/gpl.html>, or the OpenIB.org BSD
  license, available in the LICENSE.TXT file accompanying this
  software.  These details are also available at
  <http://openib.org/license.html>.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
  BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
  ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
  CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
  SOFTWARE.

  Copyright (c) 2004 Mellanox Technologies Ltd.  All rights reserved.
*/

#ifndef _UARM_PRIV_H
#define _UARM_PRIV_H

#include <mtl_common.h>
#include <mosal.h>
#include <xhh.h>
#include <dev_common.h>
#include <hob.h>
#include <icmm.h>

#include <vip_array.h>

#define UARM_LOG2_DB_RECORD_SZ 3
#define UARM_UAR_DB_RECORDS_INIT_NUM 128
#define UARM_UAR_PAGES_INIT_NUM 128
#define UARM_DBR_TERMINATOR  0xFF  /* guarantees that separator is writted (res=7) */

#define UARM_CHECK(uarm) if (uarm == (XHH_uarm_t)XHH_INVALID_HNDL) {    \
                            MTL_ERROR1(": ERROR : HCA device has not yet been opened\n");   \
                            return HH_EINVAL;   \
                         }

typedef struct
{
    MT_phys_addr_t  uar_base; 
    u_int8_t     log2_max_uar; 
    u_int8_t     log2_uar_pg_sz;
    XHH_icm_va_t uar_ctx_base;
    u_int8_t     log2_uar_ctx_sz;
    u_int32_t    num_rsvd_uars;
} XHH_uarm_init_t;

typedef struct XHH_uarm_cmdsn_page_st {
/* This is an item in a list that manages cmd_sn array for CQs associated with a UAR 
   The item manages a single page (mapping granularity).
   Each pages is divided into PAGE_SIZE/256 free lists. 
   The head of each free list is in the first byte (index 0) of the array.
   Each byte in the list holds the index of the next free byte (0 is end of list).
*/   
	u_int8_t* page_va;	/* pointer to first byte in the page */
	u_int8_t* consumer_page_va; /* The location of mapping to consumer space 
					(different than page_va if consumer in user space) */
	MOSAL_phys_ctg_t allocated_page;
	struct XHH_uarm_cmdsn_page_st *next; /* more pages... */
} XHH_uarm_cmdsn_page_t;

typedef struct XHH_uarm_uar_entry_st {
    MOSAL_prot_ctx_t    prot_ctx;
    MT_virt_addr_t      va;
    
    /* send+arm DB records (upper in UAR context table) */
    VIP_array_p_t         uar_dbrecords_wt;
    u_int32_t             uar_dbrecords_wt_cnt;
    u_int32_t             uar_dbrecords_wt_max_idx;

    /* ci+recv DB records (lower in UAR context table) */
    VIP_array_p_t         uar_dbrecords_pd;
    u_int32_t             uar_dbrecords_pd_cnt;
    u_int32_t             uar_dbrecords_pd_max_idx;

    VIP_hashp_p_t         uar_hash_pages;

    XHH_uarm_cmdsn_page_t *cmdsn_arrays; 
    
    /* protect the pages lookup*/
    MOSAL_mutex_t         dbrecords_mutex;
    MOSAL_mutex_t         cmdsn_mutex;
    MOSAL_iomem_t         iomem_obj;

} XHH_uarm_uar_entry_t;

typedef struct XHH_uarm_page_entry_st {
    XHH_icm_va_t        icm_va;
    MT_virt_addr_t      va;
}XHH_uarm_uar_page_entry_t;

typedef struct XHH_uarm_st {
      XHH_hob_t               hob;         /* saved during uldm create */
      MT_phys_addr_t          uar_base; 
      u_int8_t                log2_max_uar;
      u_int8_t                log2_uar_pg_sz;
      XHH_icm_va_t            uar_ctx_base;
      u_int8_t                log2_uar_ctx_sz;  
      VIP_array_p_t           uarm_uar_table;
      u_int32_t               num_rsvd_uars;
}XHH_uarm_obj_t;



#endif
