/*
  This software is available to you under a choice of one of two
  licenses.  You may choose to be licensed under the terms of the GNU
  General Public License (GPL) Version 2, available at
  <http://www.fsf.org/copyleft/gpl.html>, or the OpenIB.org BSD
  license, available in the LICENSE.TXT file accompanying this
  software.  These details are also available at
  <http://openib.org/license.html>.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
  BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
  ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
  CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
  SOFTWARE.

  Copyright (c) 2004 Mellanox Technologies Ltd.  All rights reserved.
*/

#define C_XHHUL_HOB_C

#include <mosal.h>
#include "ul_hob.h"
#include <ul_pdm.h>
#include <ul_cqm.h>
#include <ul_srqm.h>
#include <ul_qpm.h>
#include <ul_mwm.h>
#include <udavm.h>
#include <cidb.h>

#define GET_HOB(hhul_dev) (hhul_dev ? ((XHHUL_hob_t)(hhul_dev->device)) : NULL)

struct XHHUL_hob_st { /* *XHHUL_hob_t; */
  HHUL_hca_dev_t hhul_hca;        /* HHUL's device context */
  XHH_hca_ul_resources_t ul_res;  /* Resources allocated by HH_alloc_ul_resources() */
  /* Included objects */
  XHH_cidb_t cidb;
  XHHUL_pdm_t pdm;
  XHHUL_cqm_t cqm;
  XHHUL_qpm_t qpm;
  XHHUL_srqm_t srqm;
  XHHUL_mwm_t mwm;
  XHH_udavm_t udavm;

  /* global_resource_cnt:
   * A counter for resources set up for within this XHHUL_hob context.
   * This counter enables avoiding object destruction in case there are still 
   * resources that were not freed (e.g. QPs).
   * Each XHHUL object is responsible to update this counter using XHHUL_hob_res_add/rem */
  u_int32_t global_resource_cnt;
  MOSAL_mutex_t cntr_lock;  /* A lock for assuring atomicity of the counter lock */
};

static HHUL_if_ops_t xhhul_ops=
{ 
  XHHUL_hob_destroy         /* HHULIF_cleanup_user_level*/,  
  XHHUL_pdm_alloc_pd_avs_prep   /* HHULIF_alloc_pd_avs_prep     */,
  XHHUL_pdm_alloc_pd_done   /* HHULIF_alloc_pd_done     */,
  XHHUL_pdm_free_pd_prep    /* HHULIF_free_pd_prep      */,
  XHHUL_pdm_free_pd_done    /* HHULIF_free_pd_done      */,
  XHHUL_mwm_alloc_mw        /* HHULIF_alloc_mw          */,
  XHHUL_mwm_bind_mw         /* HHULIF_bind_mw           */,
  XHHUL_mwm_free_mw         /* HHULIF_free_mw           */,
  XHH_udavm_create_av 	 /* HHULIF_create_ud_av 		 */,
  XHH_udavm_modify_av 	 /* HHULIF_modify_ud_av 		 */,
  XHH_udavm_query_av		 /* HHULIF_query_ud_av			 */,
  XHH_udavm_destroy_av	 /* HHULIF_destroy_ud_av		 */,
  XHHUL_cqm_create_cq_prep  /* HHULIF_create_cq_prep    */,
  XHHUL_cqm_create_cq_done  /* HHULIF_create_cq_done    */,
  XHHUL_cqm_resize_cq_prep  /* HHULIF_resize_cq_prep    */,
  XHHUL_cqm_resize_cq_done  /* HHULIF_resize_cq_done    */,
  XHHUL_cqm_poll4cqe        /* HHULIF_poll4cqe          */,
  XHHUL_cqm_peek_cq         /* HHULIF_peek_cq           */,
  XHHUL_cqm_req_comp_notif  /* HHULIF_req_comp_notif    */,
  XHHUL_cqm_req_ncomp_notif /* HHULIF_req_ncomp_notif   */,
  XHHUL_cqm_destroy_cq_done /* HHULIF_destroy_cq_done   */,
  XHHUL_qpm_create_qp_prep  /* HHULIF_create_qp_prep    */,
  XHHUL_qpm_special_qp_prep /* HHULIF_special_qp_prep   */,
  XHHUL_qpm_create_qp_done  /* HHULIF_create_qp_done    */,
  XHHUL_qpm_modify_qp_done  /* HHULIF_modify_qp_done    */,
  XHHUL_qpm_post_send_req   /* HHULIF_post_send_req     */,
  XHHUL_qpm_post_inline_send_req /* HHULIF_post_inline_send_req */,
  XHHUL_qpm_post_send_reqs  /* HHULIF_post_send_reqs    */,
  XHHUL_qpm_post_gsi_send_req /* HHULIF_post_gsi_send_req */,
  XHHUL_qpm_post_recv_req   /* HHULIF_post_recv_req     */,
  XHHUL_qpm_post_recv_reqs  /* HHULIF_post_recv_reqs    */,
  XHHUL_qpm_destroy_qp_done /* HHULIF_destroy_qp_done   */,
	
  XHHUL_srqm_create_srq_prep/* HHULIF_create_srq_prep   */,
  XHHUL_srqm_create_srq_done/* HHULIF_create_srq_done   */,
  XHHUL_srqm_modify_srq_prep/* HHULIF_modify_srq_prep   */,
  XHHUL_srqm_modify_srq_done/* HHULIF_modify_srq_done   */,
  XHHUL_srqm_destroy_srq_done/* HHULIF_destroy_srq_done */,
  XHHUL_srqm_post_recv_reqs  /* HHULIF_post_srq         */ 
};

/* Private functions prototypes */
static HH_ret_t alloc_hob_context(
  XHHUL_hob_t *new_hob_p, 
  XHH_hca_ul_resources_t *hca_ul_resources_p
);


/**********************************************************************************************
 *                    Public API Functions (defined in xhhul_hob.h)
 **********************************************************************************************/

HH_ret_t XHHUL_hob_create(
  /*IN*/ void/*XHH_hca_ul_resources_t*/ *hca_ul_resources_p,
  /*OUT*/ HHUL_hca_hndl_t *hca_p 
)
{
  XHHUL_hob_t new_hob;
  HH_ret_t rc;

  if (hca_ul_resources_p == NULL) {
    MTL_ERROR1("XHHUL_hob_create: NULL hca_ul_resources_p.\n");
    return HH_EINVAL;
  }

  /* Allocate associated memory resources and included objects */
  rc= alloc_hob_context(&new_hob,(XHH_hca_ul_resources_t*)hca_ul_resources_p);
  if (rc != HH_OK)  return rc;

  /* Fill the HHUL_hca_dev_t structure */
  new_hob->hhul_hca.hh_hndl= ((XHH_hca_ul_resources_t*)hca_ul_resources_p)->hh_hca_hndl;
  new_hob->hhul_hca.dev_desc= "InfiniHost(Arbel)";
  new_hob->hhul_hca.vendor_id= MT_MELLANOX_IEEE_VENDOR_ID;
  new_hob->hhul_hca.dev_id= ((XHH_hca_ul_resources_t*)hca_ul_resources_p)->device_id;
  new_hob->hhul_hca.pci_bus = ((XHH_hca_ul_resources_t*)hca_ul_resources_p)->pci_bus;
  new_hob->hhul_hca.pci_devfn = ((XHH_hca_ul_resources_t*)hca_ul_resources_p)->pci_devfn; 
  new_hob->hhul_hca.pci_device_p = ((XHH_hca_ul_resources_t*)hca_ul_resources_p)->pci_device_p; 

  new_hob->hhul_hca.hw_ver= ((XHH_hca_ul_resources_t*)hca_ul_resources_p)->version.hw_ver;
  new_hob->hhul_hca.fw_ver= 
    ((XHH_hca_ul_resources_t*)hca_ul_resources_p)->version.fw_ver_major; 
  new_hob->hhul_hca.fw_ver= (new_hob->hhul_hca.fw_ver << 16) |
    ((XHH_hca_ul_resources_t*)hca_ul_resources_p)->version.fw_ver_minor; 
  new_hob->hhul_hca.fw_ver= (new_hob->hhul_hca.fw_ver << 16) |
    ((XHH_hca_ul_resources_t*)hca_ul_resources_p)->version.fw_ver_subminor; 
  new_hob->hhul_hca.if_ops= &xhhul_ops;
  new_hob->hhul_hca.hca_ul_resources_sz= sizeof(XHH_hca_ul_resources_t);
  new_hob->hhul_hca.pd_ul_resources_sz= sizeof(XHH_pd_ul_resources_t);
  new_hob->hhul_hca.cq_ul_resources_sz= sizeof(XHH_cq_ul_resources_t);
  new_hob->hhul_hca.srq_ul_resources_sz= sizeof(XHH_srq_ul_resources_t);
  new_hob->hhul_hca.qp_ul_resources_sz= sizeof(XHH_qp_ul_resources_t);
  /* Get a copy of allocated resources */
  memcpy(&(new_hob->ul_res),hca_ul_resources_p,sizeof(XHH_hca_ul_resources_t));
  new_hob->hhul_hca.hca_ul_resources_p= &(new_hob->ul_res);
  
  new_hob->hhul_hca.device= new_hob;  /* Connect to new XHHUL_hob */
  
  memcpy(new_hob->hhul_hca.vsd,
	 ((XHH_hca_ul_resources_t*)hca_ul_resources_p)->vsd,
	 sizeof(new_hob->hhul_hca.vsd));

#if !defined(MT_LITTLE_ENDIAN)
  /* the string needs to be "endian-swapped" again, since it is returned with improper byte ordering.
   * the string is treated like a collection of unsigned integer dwords, and is returned with the
   * dwords in big-endian format.  This means that the order of each set 4 characters must be
   * un-garbled
   */
  {
      u_int32_t temp_psid_dw[VAPI_PSID_LEN >> 2 ];
      int i;
      for (i = 0; i < (VAPI_PSID_LEN >> 2); i++) {
	  temp_psid_dw[i] = mt_generic_32_swap(new_hob->hhul_hca.vsd[VAPI_VSD_NUM_DWORDS-(VAPI_PSID_LEN>>2) + i]);
      }
      memcpy(new_hob->hhul_hca.spid, temp_psid_dw, VAPI_PSID_LEN);
  }
#else  
  memcpy(new_hob->hhul_hca.psid, &(new_hob->hhul_hca.vsd[VAPI_VSD_NUM_DWORDS-(VAPI_PSID_LEN>>2)]), VAPI_PSID_LEN);
#endif
  
  new_hob->hhul_hca.psid[VAPI_PSID_LEN] = 0;

  /* Return allocated HHUL device context */
  *hca_p= &(new_hob->hhul_hca);

  /*save pointer to UL object ops table field in HH object for kernel apps */
#ifdef MT_KERNEL
  new_hob->hhul_hca.hh_hndl->ul_if_ops_p = &(new_hob->hhul_hca.if_ops);
#endif
    
  return HH_OK;
}


HH_ret_t XHHUL_hob_destroy(/*IN*/ HHUL_hca_hndl_t hca)
{
  XHHUL_hob_t hob= GET_HOB(hca);

  if (hob == NULL)  return HH_EINVAL; /* Invalid handle */

  /* set pointer to UL object ops table field in HH object to NULL for kernel apps, so that
   * HH will not attempt to access the UL object ops table field after it has been deallocated
   */
#ifdef MT_KERNEL
  hca->hh_hndl->ul_if_ops_p = NULL;
#endif

  XHHUL_mwm_destroy(hob->mwm);
  XHHUL_qpm_destroy(hob->qpm);
#ifndef NO_UL_SRQM
  XHHUL_srqm_destroy(hob->srqm);
#endif
  XHHUL_cqm_destroy(hob->cqm);
  XHH_cidb_destroy(hob->cidb);
  XHH_udavm_destroy(hob->udavm);
  XHHUL_pdm_destroy(hob->pdm);
  FREE(hob);
  return HH_OK;
}


HH_ret_t XHHUL_hob_query_version(
  /*IN*/ XHHUL_hob_t hob,
  /*OUT*/ XHH_ver_info_t *version_p 
)
{
  if ((hob == NULL) || (version_p == NULL)) return HH_EINVAL; /* Invalid handle/pointer */
  memcpy(version_p,&(hob->ul_res.version),sizeof(XHH_ver_info_t));
  return HH_OK;
}

HH_ret_t XHHUL_hob_get_hca_ul_handle
(
  /*IN*/ XHHUL_hob_t 		hob,
  /*OUT*/ HHUL_hca_hndl_t 	*hca_ul_p
)
{
  if ((hob == NULL) || (hca_ul_p == NULL)){ 
	  return HH_EINVAL; /* Invalid handle/pointer */
  }
  
  *hca_ul_p = &hob->hhul_hca;
  return HH_OK;
}

HH_ret_t XHHUL_hob_get_hca_ul_res_p(
  /*IN*/ HHUL_hca_hndl_t hca, 
  /*OUT*/ XHH_hca_ul_resources_t **hca_ul_res_p
)
{
  XHHUL_hob_t hob= (XHHUL_hob_t)(hca->device);
  
  if ((hob == NULL) || (hca_ul_res_p == NULL)) {
	  MTL_ERROR1("%s Wrong parameters: hob = %p, hca_ul_res_p=%p\n", __func__, hob, hca_ul_res_p); 
	  return HH_EINVAL; /* Invalid handle/pointer */
  }
  *hca_ul_res_p = &(hob->ul_res);
  return HH_OK;
}

HH_ret_t XHHUL_hob_get_hca_ul_res(
  /*IN*/ HHUL_hca_hndl_t hca, 
  /*OUT*/ XHH_hca_ul_resources_t *hca_ul_res_p
)
{
  XHHUL_hob_t hob= (XHHUL_hob_t)(hca->device);
  
  if ((hob == NULL) || (hca_ul_res_p == NULL)) {
	  MTL_ERROR1("%s Wrong parameters: hob = %p, hca_ul_res_p=%p\n", __func__, hob, hca_ul_res_p); 
	  return HH_EINVAL; /* Invalid handle/pointer */
  }
  memcpy(hca_ul_res_p,&(hob->ul_res),sizeof(XHH_hca_ul_resources_t));
  return HH_OK;
}

HH_ret_t XHHUL_hob_get_hh_hca_hndl(
  /*IN*/ HHUL_hca_hndl_t hca, 
  /*OUT*/ HH_hca_hndl_t *hh_hndl_p
)
{
  XHHUL_hob_t hob= (XHHUL_hob_t)(hca->device);
  
  if ((hob == NULL) || (hh_hndl_p == NULL)) {
	  MTL_ERROR1("%s Wrong parameters: hob = %p, hh_hndl_p=%p\n", __func__, hob, hh_hndl_p); 
	  return HH_EINVAL; /* Invalid handle/pointer */
  }
  *hh_hndl_p = hob->ul_res.hh_hca_hndl;
  return HH_OK;
}

HH_ret_t XHHUL_hob_get_pdm(/*IN*/ HHUL_hca_hndl_t hca,
                           /*OUT*/ XHHUL_pdm_t *pdm_p)
{
  XHHUL_hob_t hob= (XHHUL_hob_t)(hca->device);

  if ((hob == NULL) || (pdm_p == NULL)) return HH_EINVAL; /* Invalid handle/pointer */
  *pdm_p= hob->pdm;
  return HH_OK;
}

HH_ret_t XHHUL_hob_get_cqm (/*IN*/ HHUL_hca_hndl_t hca,
                            /*OUT*/ XHHUL_cqm_t *cqm_p)
{
  XHHUL_hob_t hob= (XHHUL_hob_t)(hca->device);
  
  if ((hob == NULL) || (cqm_p == NULL)) return HH_EINVAL; /* Invalid handle/pointer */
  *cqm_p= hob->cqm;
  return HH_OK;
}

HH_ret_t XHHUL_hob_get_qpm (/*IN*/ HHUL_hca_hndl_t hca, 
                            /*OUT*/ XHHUL_qpm_t *qpm_p)
{
  XHHUL_hob_t hob= (XHHUL_hob_t)(hca->device);
  
  if ((hob == NULL) || (qpm_p == NULL)) return HH_EINVAL; /* Invalid handle/pointer */
  *qpm_p= hob->qpm;
  return HH_OK;
}

HH_ret_t XHHUL_hob_get_srqm (/*IN*/ HHUL_hca_hndl_t hca, 
                             /*OUT*/ XHHUL_srqm_t *srqm_p)
{
  XHHUL_hob_t hob= (XHHUL_hob_t)(hca->device);
  
  if ((hob == NULL) || (srqm_p == NULL)) return HH_EINVAL; /* Invalid handle/pointer */
  *srqm_p= hob->srqm;
  return HH_OK;
}


HH_ret_t XHHUL_hob_get_mwm (
  /*IN*/ HHUL_hca_hndl_t hca, 
  /*OUT*/ XHHUL_mwm_t *mwm_p
)
{
  XHHUL_hob_t hob= (XHHUL_hob_t)(hca->device);
  
  if ((hob == NULL) || (mwm_p == NULL)) return HH_EINVAL; /* Invalid handle/pointer */
  *mwm_p= hob->mwm;
  return HH_OK;
}


HH_ret_t XHHUL_hob_get_udavm (/*IN*/ HHUL_hca_hndl_t hca,
                            /*OUT*/ XHH_udavm_t *udavm_p)
{
  XHHUL_hob_t hob= (XHHUL_hob_t)(hca->device);
  
  if ((hob == NULL) || (udavm_p == NULL)) return HH_EINVAL; /* Invalid handle/pointer */
  *udavm_p= hob->udavm;
  return HH_OK;
}

HH_ret_t XHHUL_hob_get_cidb (/*IN*/ HHUL_hca_hndl_t hca, 
                            /*OUT*/ XHH_cidb_t *cidb_p)
{
  XHHUL_hob_t hob= (XHHUL_hob_t)(hca->device);
  
  if ((hob == NULL) || (cidb_p == NULL)) return HH_EINVAL; /* Invalid handle/pointer */
  *cidb_p = hob->cidb;
  return HH_OK;
}


/**********************************************************************************************
 *                    Private Functions
 **********************************************************************************************/


/*******************************************************
 * Function: alloc_hob_context
 *
 * Description:  Allocate the XHHUL_hob object memory and included objects
 *
 * Arguments: new_hob_p - Object to allocate for
 *            hca_ul_resources_p - As given to XHHUL_hob_create()
 *
 * Returns: HH_OK
 *          HH_EAGAIN
 *******************************************************/
static HH_ret_t alloc_hob_context(
  XHHUL_hob_t *new_hob_p, 
  XHH_hca_ul_resources_t *hca_ul_resources_p
)
{
  HH_ret_t rc;

  /* Allocate XHHUL_hob own context */
  *new_hob_p= (XHHUL_hob_t)MALLOC(sizeof(struct XHHUL_hob_st));
  if (*new_hob_p == NULL) return HH_EAGAIN;

  rc= XHHUL_pdm_create((*new_hob_p),&((*new_hob_p)->pdm));
  if (rc != HH_OK) {
    MTL_ERROR1("XHHUL_hob_create: Failed creating XHHUL_pdm (%d=%s).\n", rc, HH_strerror_sym(rc));
    goto failed_pdm;
  }

  /* Create included objects */
  rc= XHH_udavm_create(&hca_ul_resources_p->version,(*new_hob_p)->pdm,
		       hca_ul_resources_p->max_kernel_avs,&((*new_hob_p)->udavm));
  if (rc != HH_OK) {
    MTL_ERROR1("XHHUL_hob_create: Failed creating XHHUL_udavm (%d=%s).\n", rc, HH_strerror_sym(rc));
    goto failed_udavm;
  }

  rc= XHH_cidb_create(&hca_ul_resources_p->version, hca_ul_resources_p->uar_index,
                       (void *) hca_ul_resources_p->uar_map, &((*new_hob_p)->cidb));
  if (rc != HH_OK) {
    MTL_ERROR1("XHHUL_hob_create: Failed creating XHH_cidb_create (%d=%s).\n", rc, HH_strerror_sym(rc));
    goto failed_cidb;
  }
  
  rc= XHHUL_cqm_create((*new_hob_p),&((*new_hob_p)->cqm));
  if (rc != HH_OK) {
    MTL_ERROR1("XHHUL_hob_create: Failed creating XHHUL_cqm (%d=%s).\n", rc, HH_strerror_sym(rc));
    goto failed_cqm;
  }

#ifndef NO_UL_SRQM
  rc= XHHUL_srqm_create((*new_hob_p),&((*new_hob_p)->srqm));
  if (rc != HH_OK) {
    MTL_ERROR1("XHHUL_hob_create: Failed creating XHHUL_srqm (%d=%s).\n", rc, HH_strerror_sym(rc));
    goto failed_srqm;
  }
#endif

  rc= XHHUL_qpm_create((*new_hob_p), (*new_hob_p)->srqm, &((*new_hob_p)->qpm));
  if (rc != HH_OK) {
    MTL_ERROR1("XHHUL_hob_create: Failed creating XHHUL_qpm (%d=%s).\n", rc, HH_strerror_sym(rc));
    goto failed_qpm;
  }
  rc= XHHUL_mwm_create((*new_hob_p),hca_ul_resources_p->log2_mpt_size,&((*new_hob_p)->mwm));
  if (rc != HH_OK) {
    MTL_ERROR1("XHHUL_hob_create: Failed creating XHHUL_mwm (%d=%s).\n", rc, HH_strerror_sym(rc));
    goto failed_mwm;
  }

  return HH_OK;

  /* Failure cleanup (error exit flow) */
  failed_mwm:
    XHHUL_qpm_destroy((*new_hob_p)->qpm);
  failed_qpm:
#ifndef NO_UL_SRQM
    XHHUL_srqm_destroy((*new_hob_p)->srqm);
  failed_srqm:
#endif		
    XHHUL_cqm_destroy((*new_hob_p)->cqm);
  failed_cqm:
     XHH_cidb_destroy((*new_hob_p)->cidb);
  failed_cidb:
    XHH_udavm_destroy((*new_hob_p)->udavm);
  failed_udavm:		
	  XHHUL_pdm_destroy((*new_hob_p)->pdm);
  failed_pdm:
    FREE(*new_hob_p);
    return rc;
}

