/*
  This software is available to you under a choice of one of two
  licenses.  You may choose to be licensed under the terms of the GNU
  General Public License (GPL) Version 2, available at
  <http://www.fsf.org/copyleft/gpl.html>, or the OpenIB.org BSD
  license, available in the LICENSE.TXT file accompanying this
  software.  These details are also available at
  <http://openib.org/license.html>.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
  BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
  ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
  CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
  SOFTWARE.

  Copyright (c) 2004 Mellanox Technologies Ltd.  All rights reserved.
*/

#ifndef H_XHHUL_QPM_DEV_H
#define H_XHHUL_QPM_DEV_H

#include <cidb.h>

/* #define MAX_WQE_SZ 496 */

#define WQE_SEGMENT_NEXT_ST             arbelprm_wqe_segment_next_st
#define WQE_SEGMENT_CTRL_SEND_ST        arbelprm_wqe_segment_ctrl_send_st
#define WQE_SEGMENT_RD_ST               arbelprm_wqe_segment_rd_st
#define WQE_SEGMENT_UD_ST               arbelprm_wqe_segment_ud_st
#define WQE_SEGMENT_REMOTE_ADDRESS_ST   arbelprm_wqe_segment_remote_address_st
#define WQE_SEGMENT_ATOMIC_ST           arbelprm_wqe_segment_atomic_st
#define WQE_SEGMENT_BIND_ST             arbelprm_wqe_segment_bind_st
#define WQE_SEGMENT_DATA_PTR_ST         arbelprm_wqe_segment_data_ptr_st
#define WQE_SEGMENT_CTRL_MLX_ST         arbelprm_wqe_segment_ctrl_mlx_st
#define WQE_SEGMENT_CTRL_RECV_ST        arbelprm_wqe_segment_ctrl_recv_st

#define XHH_DEV_SPECIFIC_UL_QP_T_FIELDS \
    XHH_cidb_t  cidb;                \
    void *      recvq_db_record_p;   \
    void *      sendq_db_record_p;   \
    u_int16_t   wqe_send_counter;    \
    u_int16_t   wqe_recv_counter;    \
    XHH_udavm_t udavm;

#define XHH_DEV_SPECIFIC_INIT_QP \
    rc= XHHUL_hob_get_udavm(hca,&(new_qp->udavm)); \
    if (rc != HH_OK) {                             \
      MTL_ERROR2(MT_FLFMT("init_qp: Failed XHHUL_hob_get_udavm (%d:%s)"),rc, HH_strerror_sym(rc));  \
      return rc;                                   \
    }



#define XHH_DEV_SPECIFIC_CREATE_QP_DONE \
    qp->recvq_db_record_p = (void *)(((XHH_qp_ul_resources_t *) qp_ul_resources_p)->recvq_db_record_p);   \
    qp->sendq_db_record_p = (void *)(((XHH_qp_ul_resources_t *) qp_ul_resources_p)->sendq_db_record_p);   \
    XHH_cidb_qp_init_uar_ctx_entry(qp->recvq_db_record_p, hh_qp,  ARBEL_IF_DB_RECORD_RQ);                 \
    XHH_cidb_qp_init_uar_ctx_entry(qp->sendq_db_record_p, hh_qp,  ARBEL_IF_DB_RECORD_SQ)

#define XHH_DEV_SPECIFIC_RESET_RQ_DONE \
    qp->wqe_recv_counter= 0;		\
    XHH_cidb_qp_init_uar_ctx_entry(qp->recvq_db_record_p, qp->qpn,  ARBEL_IF_DB_RECORD_RQ)
#define XHH_DEV_SPECIFIC_RESET_SQ_DONE \
    qp->wqe_send_counter= 0;		\
    XHH_cidb_qp_init_uar_ctx_entry(qp->sendq_db_record_p, qp->qpn,  ARBEL_IF_DB_RECORD_SQ)

#define XHH_DEV_SPECIFIC_WQE_INIT_NEXT \
  MT_INSERT_ARRAY32(segment_p,1,     \
    MT_BIT_OFFSET(WQE_SEGMENT_NEXT_ST,always1),MT_BIT_SIZE(WQE_SEGMENT_NEXT_ST,always1));

#define XHH_DEV_SPECIFIC_PACK_RECV_NEXT \
	  segment_p[MT_BYTE_OFFSET(WQE_SEGMENT_NEXT_ST,nda_31_6)>>2]= ( next_wqe_32lsb & (~MASK32(6)) ) ; \
      MT_INSERT_ARRAY32(segment_p,1,                                                                  \
        MT_BIT_OFFSET(WQE_SEGMENT_NEXT_ST,always1),MT_BIT_SIZE(WQE_SEGMENT_NEXT_ST,always1));

#define XHHUL_DEV_SPECIFIC_WQE_BUILD_RECV \
  if (recv_req_p->sg_lst_len < qp->rq_res.max_sg_sz) {               \
	 cur_loc_p+= WQE_pack_dummy_sg_entry((u_int32_t*)cur_loc_p);	 \
  }


#define XHH_DEV_SPECIFIC_WQE_PACK_UD \
      {                                          \
          HH_ret_t rc;                           \
          rc= XHH_udavm_put_av(qp->udavm, ah ,segment_p); \
          if (rc != HH_OK) {                     \
            MTL_ERROR2(MT_FLFMT("WQE_pack_ud: Failed XHH_udavm_put_av (%d:%s)"),rc, HH_strerror_sym(rc));  \
            return ((u_int32_t) rc);             \
          }                                      \
      }
      
#define XHH_DEV_SPECIFIC_EXTRACT_DBD \
      return 1; 

#define XHHUL_DEV_SPECIFIC_ALLOC_WQE_BUF_INIT \
 /* Require that number of WQEs be a power of 2 (because of log_sq_size, */ \
 /* log_rq_size in Arbel QP context */                                      \
 if (new_qp->rq_res.max_outs) {new_qp->rq_res.max_outs = (1 << ceil_log2(new_qp->rq_res.max_outs));} \
 if (new_qp->sq_res.max_outs) {new_qp->sq_res.max_outs = (1 << ceil_log2(new_qp->sq_res.max_outs));}

#define XHH_DEV_SPECIFIC_QPC_ENTRY_DATA  \
     u_int32_t	snd_db_record_index; \
	 u_int32_t  rcv_db_record_index; \
     u_int8_t   log_sq_stride;       \
     u_int8_t   log_sq_size;         \
     u_int8_t   log_rq_stride;       \
     u_int8_t   log_rq_size;         \
     u_int32_t	snd_wqe_base_adr_l;  \
     u_int32_t	rcv_wqe_base_adr_l; 

#define XHHUL_DEV_SPECIFIC_ALLOC_WQE_BUF_DONE \
 qp_ul_resources_p->log_rq_stride = (u_int8_t)new_qp->rq_res.log2_max_wqe_sz;			\
 qp_ul_resources_p->log_sq_stride = (u_int8_t)new_qp->sq_res.log2_max_wqe_sz;			\
 qp_ul_resources_p->log_rq_size = (u_int8_t)ceil_log2(new_qp->rq_res.max_outs);			\
 qp_ul_resources_p->log_sq_size = (u_int8_t)ceil_log2(new_qp->sq_res.max_outs);			\
 qp_ul_resources_p->snd_wqe_base_adr_l = (buf_sz_sq != 0) ? (u_int32_t) (new_qp->sq_res.wqe_buf):0;		\
 qp_ul_resources_p->rcv_wqe_base_adr_l = (buf_sz_rq != 0) ? (u_int32_t) (new_qp->rq_res.wqe_buf):0;		\
 {											\
  int i,j;										\
  MT_virt_addr_t cur_wqe, next_wqe;							\
  u_int32_t next_draft[WQE_SEG_SZ_NEXT>>2], *cur_wqe_p, *cur_wqe_end_p; /* Build "next" segment here */	\
  /* initialize "next" segments for SQ (if present) and RQ */				\
  if (buf_sz_sq != 0) {							\
      for (i=0, cur_wqe = new_qp->sq_res.wqe_buf; i<(int)new_qp->sq_res.max_outs; 		\
            i++, cur_wqe += ((MT_virt_addr_t)(1<<new_qp->sq_res.log2_max_wqe_sz))) {	\
          next_wqe= new_qp->sq_res.wqe_buf + 						\
                (MT_virt_addr_t)(((i+1) % new_qp->sq_res.max_outs)<<new_qp->sq_res.log2_max_wqe_sz);  \
          cur_wqe_p = (u_int32_t *) cur_wqe;						\
          WQE_initialize_next(next_draft, (u_int32_t)next_wqe, 0);			\
          for (j= 0;j < (WQE_SEG_SZ_NEXT>>2) ;j++) {  					\
            /* This copy assures big-endian as well as that NDS is written last */	\
            cur_wqe_p[j]= MOSAL_cpu_to_be32(next_draft[j]);				\
          }										\
     }                                                                                 \
      initialize_sq_last_posted(&(new_qp->sq_res));					\
  }											\
  if (buf_sz_rq != 0) {							                \
      for (i=0, cur_wqe = new_qp->rq_res.wqe_buf; i<(int)new_qp->rq_res.max_outs; 		\
	    i++, cur_wqe += ((MT_virt_addr_t)(1<<new_qp->rq_res.log2_max_wqe_sz))) {	\
	  next_wqe= new_qp->rq_res.wqe_buf + 						\
		(MT_virt_addr_t)(( (i+1) % new_qp->rq_res.max_outs) << new_qp->rq_res.log2_max_wqe_sz) ;  \
	  cur_wqe_p = (u_int32_t *) cur_wqe;						\
	  cur_wqe_end_p = (u_int32_t *)(cur_wqe + ((MT_virt_addr_t)(1<<new_qp->rq_res.log2_max_wqe_sz))); \
	  WQE_initialize_next(next_draft, (u_int32_t)next_wqe, new_qp->rq_res.max_sg_sz+1 );\
	  for (j= 0;j < (WQE_SEG_SZ_NEXT>>2) ;j++) {  					\
	    /* This copy assures big-endian as well as that NDS is written last */	\
	    cur_wqe_p[j]= MOSAL_cpu_to_be32(next_draft[j]);			       	\
	  }									       	\
 	  cur_wqe_p += ((WQE_SEG_SZ_NEXT + WQE_SEG_SZ_CTRL)>>2);                        \
	  while (cur_wqe_p < cur_wqe_end_p) {						\
	     cur_wqe_p+= WQE_pack_dummy_sg_entry((u_int32_t*)cur_wqe_p)>>2;             \
	  }                                                                             \
      }                                                                                 \
  }											\
 }

#define XHH_DEV_SPECIFIC_SQ_TEST_PREV_WQE_NOT_NULL 


#define XHH_DEV_SPECIFIC_SQ_ALLOC_WQE_LINK_AND_RING \
          sq_dbell.first_wqe_counter = qp->wqe_send_counter; \
          sq_dbell.wqe_cnt = 1;

#define XHH_DEV_SPECIFIC_POST_SEND_REQS_INTERNAL \
          sq_dbell.first_wqe_counter = qp->wqe_send_counter; \
          sq_dbell.wqe_cnt = (u_int8_t)num_of_requests;

#define XHH_DEV_SPECIFIC_POST_SEND_REQS \
      {  									      \
	      u_int32_t  remaining_reqs =  num_of_requests;			      \
              u_int32_t  reqs_this_loop =  num_of_requests;			      \
              u_int32_t  reqi = 0;						      \
										      \
	      if (num_of_requests > 255) {                                            \
		    for (reqi= 0; reqi < num_of_requests; reqi++) {                   \
			if (qp->sq_res.max_sg_sz < send_req_array[reqi].sg_lst_len) { \
			  MTL_ERROR2(                                                 \
				    "%s: S/G list of request %d is too large (%d entries > max_sg_sz=%d)\n", __func__, \
				    reqi,send_req_array[reqi].sg_lst_len,qp->sq_res.max_sg_sz); \
			  MOSAL_spinlock_unlock(&(qp->sq_res.q_lock));                \
			  return HH_EINVAL_SG_NUM;                                    \
			}					                      \
		    }						                      \
		  do_sgsize_check = FALSE;			                      \
	      }							                      \
	      while (remaining_reqs) {				                      \
		     reqs_this_loop = (remaining_reqs > 255 ? 255 : remaining_reqs);  \
		     ret = XHHUL_qpm_post_send_reqs_internal(hca, qp, reqs_this_loop, \
		                                   send_req_array, do_sgsize_check);  \
		     if (ret) {							      \
			     MOSAL_spinlock_unlock(&(qp->sq_res.q_lock));	      \
			     return ret;					      \
		     }								      \
		     remaining_reqs -= reqs_this_loop; 				      \
		     send_req_array += reqs_this_loop;                                \
	     }									      \
      }

#if 0
#define XHH_DEV_SPECIFIC_POST_SEND_REQ \
      if (wqe_sz_dwords > (MAX_WQE_SZ>>2)) { \
	  MOSAL_spinlock_unlock(&(qp->sq_res.q_lock)); \
          return HH_EINVAL_AV_HNDL;          \
      }
#endif
      
#define XHH_DEV_SPECIFIC_POST_RECV_REQ  \
      XHH_cidb_qp_update_counter(qp->recvq_db_record_p, ++(qp->wqe_recv_counter)); 
      
#define XHHUL_DEV_SPECIFIC_QPM_POST_RECV_REQS_VARS 
#define XHHUL_DEV_SPECIFIC_QPM_POST_RECV_REQS_INIT
#define XHHUL_DEV_SPECIFIC_QPM_POST_RECV_REQS_LOOP
#define XHHUL_DEV_SPECIFIC_QPM_POST_RECV_REQS_DONE \
		  qp->wqe_recv_counter+= (u_int16_t)num_of_requests;									 \
		  XHH_cidb_qp_update_counter(qp->recvq_db_record_p, (qp->wqe_recv_counter))


 
																									  
#define WQE_RCV_DRAFT_START_COPY  2
#define WQE_SND_DRAFT_START_COPY  2

#define XHHUL_DEV_SPECIFIC_UDAVM(qp) ((qp)->udavm)

						  
#endif /* H_XHHUL_QPM_H */
