/*
  This software is available to you under a choice of one of two
  licenses.  You may choose to be licensed under the terms of the GNU
  General Public License (GPL) Version 2, available at
  <http://www.fsf.org/copyleft/gpl.html>, or the OpenIB.org BSD
  license, available in the LICENSE.TXT file accompanying this
  software.  These details are also available at
  <http://openib.org/license.html>.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
  BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
  ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
  CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
  SOFTWARE.

  Copyright (c) 2004 Mellanox Technologies Ltd.  All rights reserved.
*/

#ifndef CI_XHHUL_QPM_DEV_CI
#define CI_XHHUL_QPM_DEV_CI

#define XHHUL_QPM_END_WQE_LKEY 0x00000100

inline static HH_ret_t init_uar_info(HHUL_hca_hndl_t hca, XHHUL_qp_t qp, XHH_qp_ul_resources_t *ul_res_p)
{
    HH_ret_t rc;

    rc= XHHUL_hob_get_cidb(hca,&(qp->cidb));
    if (rc != HH_OK) {
      MTL_ERROR1(MT_FLFMT("%s: Failed getting XHHUL_hob's cidb (err=%d)."),__func__,rc);
      return rc;
    }
    rc= XHH_cidb_get_uar_index(qp->cidb,&(ul_res_p->uar_index)); 
    if (rc != HH_OK) {
      MTL_ERROR1(MT_FLFMT("%s: Failed getting UAR index (err=%d)"), __func__, rc);
    }
    return rc;
}

inline static HH_ret_t uar_sendq_dbell(XHHUL_qp_t qp, XHH_uar_sendq_dbell_t *sq_dbell)
{
    MTPERF_TIME_START(XHH_cidb_qp_sendq);
    XHH_cidb_qp_sendq(qp->cidb, qp->sendq_db_record_p, qp->qpn, sq_dbell);
    MTPERF_TIME_END(XHH_cidb_qp_sendq);
    qp->wqe_send_counter += sq_dbell->wqe_cnt;
    return HH_OK;
}

inline static HH_ret_t uar_sendq_rd_dbell(XHHUL_qp_t qp, XHH_uar_sendq_dbell_t *sq_dbell, VAPI_sr_desc_t *send_req_p)
{   
    HH_ret_t rc;
    rc = XHH_cidb_sendq_rd(qp->cidb, qp->sendq_db_record_p, qp->qpn, send_req_p->eecn, sq_dbell);
    if (rc == HH_OK) {
        qp->wqe_send_counter += sq_dbell->wqe_cnt;
    }
    return rc;
}

inline static u_int32_t WQE_init_next(u_int32_t *wqe_buf)
{
  /* In Arbel: don't touch ctrl/next segment until linking to a new next */
  /* This is required in order to retain old next info for doorbell recovery */
  return WQE_SEG_SZ_NEXT;
}

inline static u_int32_t WQE_init_next_recv(u_int32_t *wqe_buf)
{
  /* In Arbel RQ NDS/NDA never change */
  return WQE_SEG_SZ_NEXT;
}

/* Build the scatter/gather list (pointer segments) */
inline static u_int32_t WQE_pack_dummy_sg_entry(u_int32_t *segment_p)
{
   u_int32_t *cur_loc_p= segment_p;

   cur_loc_p[MT_BYTE_OFFSET(WQE_SEGMENT_DATA_PTR_ST,byte_count)>>2]= 0;
   cur_loc_p[MT_BYTE_OFFSET(WQE_SEGMENT_DATA_PTR_ST,l_key)>>2]= XHHUL_QPM_END_WQE_LKEY;
   cur_loc_p[MT_BYTE_OFFSET(WQE_SEGMENT_DATA_PTR_ST,local_address_h)>>2]= 0;
   cur_loc_p[MT_BYTE_OFFSET(WQE_SEGMENT_DATA_PTR_ST,local_address_l)>>2]= 0;
   return WQE_SEG_SZ_SG_ENTRY;
}

inline static u_int32_t WQE_initialize_next(u_int32_t *segment_p, 
  u_int32_t next_wqe_32lsb, u_int32_t wqe_sz_16B_chunks)
{
  MT_offset_t  test;
  MT_size_t    test1;
  memset(segment_p,0,WQE_SEG_SZ_NEXT);  /* Clear all "RESERVED" */
  segment_p[MT_BYTE_OFFSET(WQE_SEGMENT_NEXT_ST,nda_31_6)>>2]= ( next_wqe_32lsb & (~MASK32(6)) ); 

  test = MT_BIT_OFFSET(WQE_SEGMENT_NEXT_ST,nds);
  test1 = MT_BIT_SIZE(WQE_SEGMENT_NEXT_ST,nds);
  MT_INSERT_ARRAY32(segment_p,wqe_sz_16B_chunks,
    MT_BIT_OFFSET(WQE_SEGMENT_NEXT_ST,nds),MT_BIT_SIZE(WQE_SEGMENT_NEXT_ST,nds));
  return WQE_SEG_SZ_NEXT;
}

/* Build the scatter/gather list (pointer segments) */
inline static u_int32_t WQE_pack_sg_list_null_entry(u_int32_t *segment_p)
{
   u_int32_t *cur_loc_p= segment_p;

   memset(cur_loc_p, 0, WQE_SEG_SZ_SG_ENTRY);
   cur_loc_p[MT_BYTE_OFFSET(WQE_SEGMENT_DATA_PTR_ST,l_key)>>2]=0x00000100;
   return (u_int32_t)(WQE_SEG_SZ_SG_ENTRY_DW);
}

inline static void initialize_sq_last_posted(queue_res_t *sq_res)
{
  if (sq_res->max_outs > 1) {
      sq_res->last_posted_p = (u_int32_t *) (MT_virt_addr_t)(sq_res->wqe_buf +
                         (MT_virt_addr_t)((sq_res->max_outs - 1)*
			 ((MT_virt_addr_t)(1<<sq_res->log2_max_wqe_sz))));
  } else {
      sq_res->last_posted_p = NULL;
  }

}

#endif /* CI_XHHUL_QPM_DEV_CI */
