/*
  This software is available to you under a choice of one of two
  licenses.  You may choose to be licensed under the terms of the GNU
  General Public License (GPL) Version 2, available at
  <http://www.fsf.org/copyleft/gpl.html>, or the OpenIB.org BSD
  license, available in the LICENSE.TXT file accompanying this
  software.  These details are also available at
  <http://openib.org/license.html>.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
  BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
  ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
  CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
  SOFTWARE.

  Copyright (c) 2004 Mellanox Technologies Ltd.  All rights reserved.
*/

#ifndef H_DBELL_IO_H
#define H_DBELL_IO_H

#include <mosal.h>

/* doorbell ringing for 2 dwords */
#ifdef __MOSAL_MMAP_IO_WRITE_QWORD_ATOMIC__
/* If Qword write is assured to be atomic (MMX or a 64-bit arch) - no need for the spinlock */
#define RING_DBELL_2DW(uar,dword_offset,dbell_draft)                                       \
  dbell_draft[0]= MOSAL_cpu_to_be32(dbell_draft[0]);                                       \
  dbell_draft[1]= MOSAL_cpu_to_be32(dbell_draft[1]);                                       \
  MOSAL_MMAP_IO_WRITE_QWORD(uar->uar_base+dword_offset,*(volatile u_int64_t*)dbell_draft); 
#else
#define RING_DBELL_2DW(uar,dword_offset,dbell_draft)                                       \
  MOSAL_spinlock_irq_lock(&(uar->uar_lock));                                               \
  MOSAL_MMAP_IO_WRITE_DWORD(uar->uar_base+dword_offset,MOSAL_cpu_to_be32(dbell_draft[0])); \
  MOSAL_MMAP_IO_WRITE_DWORD(uar->uar_base+dword_offset+1,MOSAL_cpu_to_be32(dbell_draft[1])); \
  MOSAL_spinlock_unlock(&(uar->uar_lock));
#endif /* Atomic Qword write */

#define RING_DBELL_4DW(uar,dword_offset,dbell_draft)                                  \
  dbell_draft[0]= MOSAL_cpu_to_be32(dbell_draft[0]);                                  \
  dbell_draft[1]= MOSAL_cpu_to_be32(dbell_draft[1]);                                  \
  dbell_draft[2]= MOSAL_cpu_to_be32(dbell_draft[2]);                                  \
  dbell_draft[3]= MOSAL_cpu_to_be32(dbell_draft[3]);                                  \
  MOSAL_spinlock_irq_lock(&(uar->uar_lock));                                          \
  MOSAL_MMAP_IO_WRITE_QWORD(uar->uar_base+dword_offset,((u_int64_t*)dbell_draft)[0]); \
  MOSAL_MMAP_IO_WRITE_QWORD(uar->uar_base+dword_offset+2,((u_int64_t*)dbell_draft)[1]); \
  MOSAL_spinlock_unlock(&(uar->uar_lock));


#endif
