/*
  This software is available to you under a choice of one of two
  licenses.  You may choose to be licensed under the terms of the GNU
  General Public License (GPL) Version 2, available at
  <http://www.fsf.org/copyleft/gpl.html>, or the OpenIB.org BSD
  license, available in the LICENSE.TXT file accompanying this
  software.  These details are also available at
  <http://openib.org/license.html>.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
  BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
  ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
  CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
  SOFTWARE.

  Copyright (c) 2004 Mellanox Technologies Ltd.  All rights reserved.
*/

#include <mtl_types.h>
#include <tavor_dev_defs.h>

#define MLXHH_DEV_TYPE "THH"

/* the following macro defines the initializer for a struct
   that defines which devices will be probed by the pci unit
   to be registered by the driver. The structure itself is
   defined in linux_module.c. Each line defines one device,
   and has the following structure:
   Vendor ID
   Device ID
   Subsytem Vendor ID
   Subsystem Device ID
 */
 
#define HCA_TBL_INITIALIZER                       \
{                                                 \
  {                                               \
    .vendor =    MT_MELLANOX_PCI_VENDOR_ID,       \
    .device =    MT23108_DEV_ID,                  \
    .subvendor = MT_MELLANOX_PCI_VENDOR_ID,       \
    .subdevice = MT23108_DEV_ID                   \
  }, /* Mellanox Tavor */                         \
  {                                               \
    .vendor =    MT_MELLANOX_PCI_VENDOR_ID,       \
    .device =    MT25208_DEV_ID,                  \
    .subvendor = MT_MELLANOX_PCI_VENDOR_ID,       \
    .subdevice = MT25208_DEV_ID                   \
  }, /* Mellanox Arbel */                         \
  {                                               \
    .vendor =    MT_TOPSPIN_PCI_VENDOR_ID,        \
    .device =    MT23108_DEV_ID,                  \
    .subvendor = MT_TOPSPIN_PCI_VENDOR_ID,        \
    .subdevice = MT23108_DEV_ID                   \
  }, /* TopSpin Tavor */                          \
  {                                               \
    .vendor =    MT_TOPSPIN_PCI_VENDOR_ID,        \
    .device =    MT25208_DEV_ID,                  \
    .subvendor = MT_TOPSPIN_PCI_VENDOR_ID,        \
    .subdevice = MT25208_DEV_ID                   \
  }, /* TopSpin Arbel */                          \
  {                                               \
    .vendor =    MT_TOPSPIN_PCI_VENDOR_ID,        \
    .device =    MT25208_DEV_ID,                  \
    .subvendor = 0x103c,                          \
    .subdevice = 0x12ce                           \
  }, /* TopSpin */                                \
  {0, 0, 0, 0}            /* terminate list */    \
}

