/*
  This software is available to you under a choice of one of two
  licenses.  You may choose to be licensed under the terms of the GNU
  General Public License (GPL) Version 2, available at
  <http://www.fsf.org/copyleft/gpl.html>, or the OpenIB.org BSD
  license, available in the LICENSE.TXT file accompanying this
  software.  These details are also available at
  <http://openib.org/license.html>.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
  BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
  ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
  CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
  SOFTWARE.

  Copyright (c) 2004 Mellanox Technologies Ltd.  All rights reserved.
*/

#include <cmdif.h>
#include <cmd_types.h>
#include <cmdif_priv.h>
#include <mtl_common.h>
#include <cmds_wrap_priv.h>
#include <dev_cmds_wrap.h>
 

#if defined(MAX_DEBUG) && XHH_CMDS_WRAP_DEBUG_LEVEL <= MAX_DEBUG
#define XHH_CMD_PRINT_QUERY_DDR(a)     XHH_cmd_print_query_ddr(a)

static void XHH_cmd_print_query_ddr( XHH_ddr_props_t *ddr_props)
{
    CMDS_DBG("QUERY DDR DUMP (XHH_ddr_props_t structure)\n");
    CMDS_DBG( "ddr_props->ddr_start_adr = "U64_FMT"\n", ddr_props->ddr_start_adr);
    CMDS_DBG( "ddr_props->ddr_end_adr = "U64_FMT"\n", ddr_props->ddr_end_adr);
    CMDS_DBG( "\nddr_props->dimm0.di = %d\n", ddr_props->dimm0.di);
    CMDS_DBG( "ddr_props->dimm0.dimmsize = %d\n", ddr_props->dimm0.dimmsize);
    CMDS_DBG( "ddr_props->dimm0.dimmstatus = %d\n", ddr_props->dimm0.dimmstatus);
    CMDS_DBG( "\nddr_props->dimm1.di = %d\n", ddr_props->dimm1.di);
    CMDS_DBG( "ddr_props->dimm1.dimmsize = %d\n", ddr_props->dimm1.dimmsize);
    CMDS_DBG( "ddr_props->dimm1.dimmstatus = %d\n", ddr_props->dimm1.dimmstatus);
    CMDS_DBG( "\nddr_props->dimm2.di = %d\n", ddr_props->dimm2.di);
    CMDS_DBG( "ddr_props->dimm2.dimmsize = %d\n", ddr_props->dimm2.dimmsize);
    CMDS_DBG( "ddr_props->dimm2.dimmstatus = %d\n", ddr_props->dimm2.dimmstatus);
    CMDS_DBG( "\nddr_props->dimm3.di = %d\n", ddr_props->dimm3.di);
    CMDS_DBG( "ddr_props->dimm3.dimmsize = %d\n", ddr_props->dimm3.dimmsize);
    CMDS_DBG( "ddr_props->dimm3.dimmstatus = %d\n", ddr_props->dimm3.dimmstatus);
    CMDS_DBG( "ddr_props->dh = %s\n", (ddr_props->dh ? "TRUE" : "FALSE"));
    CMDS_DBG( "ddr_props->ap = %d\n", ddr_props->ap);
    CMDS_DBG( "ddr_props->di = %d\n", ddr_props->di);
}
#else
#define XHH_CMD_PRINT_QUERY_DDR(a) 

#endif

/*
 *  XHH_cmd_QUERY_DDR
 */ 
XHH_cmd_status_t XHH_cmd_QUERY_DDR(XHH_cmd_t cmd_if, XHH_ddr_props_t *ddr_props)
{
  command_fields_t cmd_desc;
  u_int8_t *outprm;
  XHH_cmd_status_t rc;
  u_int32_t buf_size = PSEUDO_MT_BYTE_SIZE(ADD_DEV_PRM_PREFIX(query_ddr_st));

  FUNC_IN;
  outprm = TNMALLOC(u_int8_t, buf_size);
  if (outprm == NULL) {
    MT_COV_MEM_FLTR_START
    MT_RETURN(XHH_CMD_STAT_EAGAIN);
    MT_COV_MEM_FLTR_END
  }
  memset(outprm, 0, buf_size);
  
  cmd_desc.in_param = 0;
  cmd_desc.in_param_size = 0;
  cmd_desc.in_trans = TRANS_NA;
  cmd_desc.input_modifier = 0;
  cmd_desc.out_param = outprm;
  cmd_desc.out_param_size = buf_size;
  cmd_desc.out_trans = TRANS_MAILBOX;
  cmd_desc.opcode = ADD_DEV_PREFIX(IF_CMD_QUERY_DDR);
  cmd_desc.opcode_modifier = 0;
  cmd_desc.exec_time_micro = ADD_DEV_PREFIX(IF_CMD_ETIME_QUERY_DDR);
  
  rc = cmd_invoke(cmd_if, &cmd_desc);
  if ( rc != XHH_CMD_STAT_OK ) {
    FREE(outprm);
    MT_RETURN(rc);
  }

  if ( ddr_props ) {
    EX_FLD64(ddr_props, outprm, ADD_DEV_PRM_PREFIX(query_ddr_st), ddr_start_adr);
    EX_FLD64(ddr_props, outprm, ADD_DEV_PRM_PREFIX(query_ddr_st), ddr_end_adr);

    EX_BOOL_FLD(ddr_props, outprm, ADD_DEV_PRM_PREFIX(query_ddr_st), dh);
    EX_FLD(ddr_props, outprm, ADD_DEV_PRM_PREFIX(query_ddr_st), di);
    EX_FLD(ddr_props, outprm, ADD_DEV_PRM_PREFIX(query_ddr_st), ap);

    /* TBD handle dimm structs here */
    EX_FLD(ddr_props, outprm, ADD_DEV_PRM_PREFIX(query_ddr_st), dimm0.dimmsize);
    EX_FLD(ddr_props, outprm, ADD_DEV_PRM_PREFIX(query_ddr_st), dimm0.dimmstatus);
    

    EX_FLD(ddr_props, outprm, ADD_DEV_PRM_PREFIX(query_ddr_st), dimm1.dimmsize);
    EX_FLD(ddr_props, outprm, ADD_DEV_PRM_PREFIX(query_ddr_st), dimm1.dimmstatus);
    

    EX_FLD(ddr_props, outprm, ADD_DEV_PRM_PREFIX(query_ddr_st), dimm2.dimmsize);
    EX_FLD(ddr_props, outprm, ADD_DEV_PRM_PREFIX(query_ddr_st), dimm2.dimmstatus);
    

    EX_FLD(ddr_props, outprm, ADD_DEV_PRM_PREFIX(query_ddr_st), dimm3.dimmsize);
    EX_FLD(ddr_props, outprm, ADD_DEV_PRM_PREFIX(query_ddr_st), dimm3.dimmstatus);
    

    XHH_CMD_PRINT_QUERY_DDR(ddr_props);
  }
  FREE(outprm);
  MT_RETURN(rc);
}

/*
 *  XHH_cmd_SYS_DIS
 */ 
XHH_cmd_status_t XHH_cmd_SYS_DIS(XHH_cmd_t cmd_if)
{
  command_fields_t cmd_desc;
  XHH_cmd_status_t rc;

  FUNC_IN;
  cmd_desc.in_param = 0;
  cmd_desc.in_param_size = 0;
  cmd_desc.in_trans = TRANS_NA;
  cmd_desc.input_modifier = 0;
  cmd_desc.out_param = 0;
  cmd_desc.out_param_size = 0;
  cmd_desc.out_trans = TRANS_NA;
  cmd_desc.opcode = ADD_DEV_PREFIX(IF_CMD_SYS_DIS);
  cmd_desc.opcode_modifier = 0;
  cmd_desc.exec_time_micro = ADD_DEV_PREFIX(IF_CMD_ETIME_SYS_DIS);

  rc = cmd_invoke(cmd_if, &cmd_desc);
  MT_RETURN(rc);
}


/*
 *  XHH_cmd_ACCESS_DDR
 */ 
static inline void build_access_hcamem_input_modifier(u_int8_t * inmod,
							XHH_acc_hcamem_inj_err_inp_mod_t *access_hcamem_p)
{
    INS_BOOL_FLD(access_hcamem_p, inmod, tavorprm_access_ddr_inject_errors_input_modifier_st, q0);
    INS_FLD(access_hcamem_p, inmod, tavorprm_access_ddr_inject_errors_input_modifier_st, index0);
    INS_BOOL_FLD(access_hcamem_p, inmod, tavorprm_access_ddr_inject_errors_input_modifier_st, q1);
    INS_FLD(access_hcamem_p, inmod, tavorprm_access_ddr_inject_errors_input_modifier_st, index1);
    INS_BOOL_FLD(access_hcamem_p, inmod, tavorprm_access_ddr_inject_errors_input_modifier_st, q2);
    INS_FLD(access_hcamem_p, inmod, tavorprm_access_ddr_inject_errors_input_modifier_st, index2);
    INS_BOOL_FLD(access_hcamem_p, inmod, tavorprm_access_ddr_inject_errors_input_modifier_st, q3);
    INS_FLD(access_hcamem_p, inmod, tavorprm_access_ddr_inject_errors_input_modifier_st, index3);

}

XHH_cmd_status_t XHH_cmd_INJECT_HCAMEM_ERR(XHH_cmd_t cmd_if, 
				    XHH_acc_hcamem_inj_err_inp_param_t *acc_hcamem_inj_err_param_p, 
				    XHH_acc_hcamem_inj_err_inp_mod_t *acc_hcamem_inj_err_mod_p)
{
  command_fields_t cmd_desc;
  XHH_cmd_status_t rc;
  u_int32_t  inp_param[2] = {0};
  u_int8_t   *inparam = (u_int8_t *)&(inp_param[0]);

  FUNC_IN;

  INS_FLD(acc_hcamem_inj_err_param_p, inparam, tavorprm_access_ddr_inject_errors_input_parameter_st, da);
  INS_FLD(acc_hcamem_inj_err_param_p, inparam, tavorprm_access_ddr_inject_errors_input_parameter_st, ba);
  INS_FLD(acc_hcamem_inj_err_param_p, inparam, tavorprm_access_ddr_inject_errors_input_parameter_st, ca);
  INS_FLD(acc_hcamem_inj_err_param_p, inparam, tavorprm_access_ddr_inject_errors_input_parameter_st, ra);

  cmd_desc.in_param = (u_int8_t *)&(inp_param[0]);
  cmd_desc.in_param_size = sizeof(inp_param);
  cmd_desc.in_trans = TRANS_IMMEDIATE;
  build_access_hcamem_input_modifier((u_int8_t *)&(cmd_desc.input_modifier),acc_hcamem_inj_err_mod_p);
  cmd_desc.out_param = 0;
  cmd_desc.out_param_size = 0;
  cmd_desc.out_trans = TRANS_NA;
  cmd_desc.opcode = ADD_DEV_PREFIX(IF_CMD_ACCESS_DDR);
  cmd_desc.opcode_modifier = 1;
  cmd_desc.exec_time_micro = ADD_DEV_PREFIX(IF_CMD_ETIME_CLASS_C);

  rc = cmd_invoke(cmd_if, &cmd_desc);
  MT_RETURN(rc);
}

