/*
  This software is available to you under a choice of one of two
  licenses.  You may choose to be licensed under the terms of the GNU
  General Public License (GPL) Version 2, available at
  <http://www.fsf.org/copyleft/gpl.html>, or the OpenIB.org BSD
  license, available in the LICENSE.TXT file accompanying this
  software.  These details are also available at
  <http://openib.org/license.html>.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
  BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
  ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
  CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
  SOFTWARE.

  Copyright (c) 2004 Mellanox Technologies Ltd.  All rights reserved.
*/

#ifndef H_DEV_CMD_TYPES_H
#define H_DEV_CMD_TYPES_H

#include <tavor_if_defs.h>

#define XHH_DEV_SPECIFIC_FW_PROPS \
  u_int64_t fw_base_addr;	/* Physical Address of Firmware Area in DDR Memory */ \
  u_int64_t fw_end_addr;	/* End of firmware address in DDR memory */           \
  

enum {
  DDR_STAT_OK = 0,
  DDR_STAT_CAL_1_ERR = 1,
  DDR_STAT_CAL_2_ERR = 2
};
typedef u_int32_t XHH_ddr_cal_status_t;

enum {
  DDR_NO_AUTO=0,
  DDR_AUTO_PRECHARGE_PER_TRANSLATION=1,
  DDR_AUTO_PRECHARGE_PER_64BIT=2
};
typedef u_int32_t XHH_dim_apm_t;

#define XHH_ADAPTER_MEM_PROPS_TYPE XHH_ddr_props_t 

#define XHH_DEV_SPECIFIC_ADAPTER_MEM_PROPS \
  u_int64_t ddr_start_adr; /* DDR memory start address */    \
  u_int64_t ddr_end_adr;	/* DDR memory end address (excluding DDR memory reserved for firmware) */

#define XHH_DEV_SPECIFIC_DEV_LIMS \
  u_int8_t log_max_mtt_seg;	/* Log2 of the Maximum number of MTT segments */\
  u_int8_t  max_sg;\
  u_int16_t max_desc_sz;


#define  XHH_DEV_SPECIFIC_CQC \
  u_int32_t consumer_indx;	/* Contains index to the next entry to be read upon poll for completion. */ \
                            /* The first completion after passing ownership of CQ from software to */   \
                            /* hardware will be reported to value passed in this field. */              \
  u_int32_t producer_indx;	/* Maintained by HW, not to be altered by SW. Points to the next entry to be */  \
                            /* written to by Hardware. CQ overrun is reported if Producer_indx + 1 equals */ \
                            /*to Consumer_indx. */

#define  XHH_DEV_SPECIFIC_QPEE_CONTEXT\
         u_int32_t wqe_base_adr;	/* Bits 63:32 of WQE address for both SQ and RQ. \;Reserved for EE context. */

#define  XHH_DEV_SPECIFIC_SRQ_CONTEXT \
  u_int32_t uar; /* UAR index for doorbells of this SRQ */ \
  u_int32_t ds;         /* Descriptor size for SRQ (divided by 16) */

#define  XHH_DEV_SPECIFIC_HCA_PROPS 
#define  XHH_DEV_SPECIFIC_CONTEXTS_PRMS
#define  XHH_CMD_DEV_SPECIFIC_STATS

#define  XHH_DEV_SPECIFIC_MPT_ENTRY  \
  MT_bool m_io;	/* If set - memory command is used on the uplink bus, */          \
                /* if cleared - IO. If IO configured - PW bit must be cleared. */ \
  u_int64_t mtt_seg_adr;	/* Base (first) address of the MTT segment, aligned on segment_size boundary */
  
#define  XHH_DEV_SPECIFIC_TPT_PRMS       \
  u_int8_t mtt_segment_size;  /* The size of MTT segment is 64*2^MTT_Segment_Size bytes */

#define 	XHH_DEV_SPECIFIC_CMD_STATUS_CASES

typedef struct {
  u_int64_t uar_base_addr; /* UAR Base Address (QUERY_HCA only) */
  u_int8_t uar_page_sz; /* This field defines the size of each UAR page. Size of UAR Page is 4KB*2^UAR_Page_Size */
  u_int64_t uar_scratch_base_addr;      /* Base address of UAR scratchpad. Number of entries in table is UAR BAR size divided by UAR Page Size. Table must be aligned to entry size. */
}
XHH_uar_prms_t;

#endif
