/*
  This software is available to you under a choice of one of two
  licenses.  You may choose to be licensed under the terms of the GNU
  General Public License (GPL) Version 2, available at
  <http://www.fsf.org/copyleft/gpl.html>, or the OpenIB.org BSD
  license, available in the LICENSE.TXT file accompanying this
  software.  These details are also available at
  <http://openib.org/license.html>.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
  BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
  ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
  CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
  SOFTWARE.

  Copyright (c) 2004 Mellanox Technologies Ltd.  All rights reserved.
*/

#ifndef H_DEV_COMMON_H
#define H_DEV_COMMON_H


#include <mtl_common.h>
#include <vapi.h>
#include <hh.h>
#include <log2.h>
#include <MT23108.h>
#include <xhh.h>

#define XHH_NAME "THH"

/* resources reserved in driver */
#define XHH_NUM_RSVD_PD     2
#define XHH_NUM_RSVD_QP     8

#define XHH_MAX_WQE_SZ    1008

#define ADD_DEV_PREFIX(a)       TAVOR_ ## a
#define ADD_DEV_PREFIX_LC(a)    tavor_ ## a
#define ADD_DEV_PRM_PREFIX(a)   tavorprm_ ## a

typedef struct
{
  MT_virt_addr_t   cqe_buf;     /* CQE buffer virtual addr. CQE size aligned */
  MT_size_t        cqe_buf_sz;  /* Buffer size in bytes (mult of CQE size) */
  XHH_uar_index_t  uar_index;   /* Index of UAR used for this CQ. */
  u_int32_t        new_producer_index;     /* New producer index after "resize_cq" (OUT)*/
} XHH_cq_ul_resources_t;

typedef struct 
{
  MT_virt_addr_t   wqes_buf;     /* WQEs buffer virtual address */
  MT_size_t        wqes_buf_sz;  /* Buffer size in bytes */
  XHH_uar_index_t  uar_index;    /* index of UAR used for this QP */
  VAPI_lkey_t qp_mem_lkey;       /* to be used in kernel when buffer is allocated externaly */
} XHH_qp_ul_resources_t;

typedef struct 
{
  MT_virt_addr_t      wqes_buf;     /* WQEs buffer virtual address */
  MT_size_t           wqes_buf_sz;  /* Buffer size in bytes */
  MT_size_t           wqe_sz;       /* WQE (descriptor) size in bytes */
  XHH_uar_index_t     uar_index;    /* index of UAR used for this QP */
  u_int32_t         srq_limit;	   /* SRQ's limit (modify-srq only) */
  VAPI_srq_attr_mask_t srq_attr_mask; /* modify-srq only */
} XHH_srq_ul_resources_t;

typedef struct XHH_uar_st     *XHH_uar_t;

#endif  /* H_DEV_COMMON_H */
