/*
  This software is available to you under a choice of one of two
  licenses.  You may choose to be licensed under the terms of the GNU
  General Public License (GPL) Version 2, available at
  <http://www.fsf.org/copyleft/gpl.html>, or the OpenIB.org BSD
  license, available in the LICENSE.TXT file accompanying this
  software.  These details are also available at
  <http://openib.org/license.html>.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
  BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
  ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
  CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
  SOFTWARE.

  Copyright (c) 2004 Mellanox Technologies Ltd.  All rights reserved.
*/

#ifndef H_XHH_DEV_TPTM_H
#define H_XHH_DEV_TPTM_H

#include <tavor_if_defs.h>
#include <uldm.h>

#define TPTM_CURRENT_MEMKEY(mrwm,mpt_seg,mpt_index) \
  ( ((mrwm)->key_prefix[mpt_seg][mpt_index-(mrwm)->offset[mpt_seg]] << (mrwm)->props.log2_mpt_sz) | (mpt_index))
  
#define TPTM_MEMKEY2INDEX(mrwm,memkey)  ( (memkey) & MASK32((mrwm)->props.log2_mpt_sz) )
#define TPTM_MEMKEY2PREFIX(mrmwm,memkey) ( (memkey) >> ((mrwm)->props.log2_mpt_sz) )

#define TPTM_MEMKEY_HW_FMT(memkey) memkey

#define TPTM_CURRENT_FMR_MEMKEY(mrwm,mpt_seg,mpt_index,fmr_info_p)  TPTM_CURRENT_MEMKEY(mrwm,mpt_seg,mpt_index)
  
#define TPTM_ADVANCE_FMR_MEMKEY  ++(mrwm->key_prefix[MPT_ext][fmr_hndl])
#define TPTM_REVERT_FMR_MEMKEY   (mrwm->key_prefix[MPT_ext][fmr_hndl])--
  

#define TPTM_LOG_MPT_PG_SZ_SHIFT 12

#define XHH_DEV_SPECIFIC_TPTM_INIT_PARAMS \
  u_int64_t  mtt_base;             /* Physical address of MTT */  \
  MT_phys_addr_t  mpt_base;        /* Physical address of MPT */

#define XHH_DEV_SPECIFIC_TPTM_DATA \
        XHH_uldm_t       prot_module;
	
#define XHH_DEV_SPECIFIC_TPTM_FMR_RSC_MAP_TYPE MT_virt_addr_t
#define XHH_DEV_SPECIFIC_TPTM_FMR_INFO	\
            MOSAL_iomem_t mpt_iomem; \
	    MOSAL_iomem_t mtt_iomem;

#define XHH_DEV_SPECIFIC_TPTM_INIT do {} while(0)
#define XHH_DEV_SPECIFIC_TPTM_DONE do {} while(0)
#define XHH_DEV_SPECIFIC_TPTM_CREATE do {} while(0)
#define XHH_DEV_SPECIFIC_TPTM_DESTROY do {} while(0)

#define XHH_DEV_SPECIFIC_TPTM_ALLOC_MPTE do {} while(0)

#define XHH_DEV_SPECIFIC_TPTM_ALLOC_MPTE4FMR									\
{														\
        call_result_t mosal_rc;                                                                                 \
	MT_phys_addr_t mpt_entry_adr= mrwm->props.mpt_base | (mpt_index << ADD_DEV_PREFIX(IF_STRIDE_MPT_BIT));	\
														\
	mosal_rc = MOSAL_iomem_map(MOSAL_get_kernel_prot_ctx(),			                                \
	              MOSAL_MEM_FLAGS_NO_CACHE | MOSAL_MEM_FLAGS_PERM_READ |MOSAL_MEM_FLAGS_PERM_WRITE,         \
	              mpt_entry_adr, (MT_size_t)ADD_DEV_PREFIX(IF_STRIDE_MPT),&(new_fmr_p->mpt_iomem),          \
		      &(new_fmr_p->mpt_entry));	                                                                \
	if (mosal_rc != MT_OK) {									        \
	  MTL_ERROR2(MT_FLFMT("%s: MOSAL_iomem_map("PHYS_ADDR_FMT", 0x%x) for MPT entry failed (%d)"), __func__,\
		     mpt_entry_adr, ADD_DEV_PREFIX(IF_STRIDE_MPT), mosal_rc);					\
	  rc= HH_EAGAIN;											\
	}													\
}

        
#define XHH_DEV_SPECIFIC_TPTM_RETRIEVE_MPTE do {} while(0)

#define XHH_DEV_SPECIFIC_TPTM_RETRIEVE_MPTE4FMR(fmr_p) \
	MOSAL_iomem_unmap(fmr_p->mpt_iomem,MOSAL_get_kernel_prot_ctx())

#define XHH_DEV_SPECIFIC_TPTM_ALLOC_MTT_SEGS rc= HH_OK

#define XHH_DEV_SPECIFIC_TPTM_ALLOC_MTT_SEGS4FMR \
{														\
        call_result_t  mosal_rc;                                                                                \
	MT_phys_addr_t mtt_seg_adr= 										\
	  mrwm->props.mtt_base | (seg_start << (mrwm->props.log2_mtt_seg_sz + MTT_LOG_MTT_ENTRY_SIZE));		\
														\
	mosal_rc= MOSAL_iomem_map(MOSAL_get_kernel_prot_ctx(),                                                  \
	              MOSAL_MEM_FLAGS_NO_CACHE | MOSAL_MEM_FLAGS_PERM_READ |MOSAL_MEM_FLAGS_PERM_WRITE,         \
	              mtt_seg_adr,					                                        \
		      (MT_size_t) (1 << (log2_segs + mrwm->props.log2_mtt_seg_sz + MTT_LOG_MTT_ENTRY_SIZE)),    \
		      &(new_fmr_p->mtt_iomem),                                                                  \
		      &(new_fmr_p->mtt_entries));			                                        \
	if (mosal_rc != MT_OK) {						            		        \
	  MTL_ERROR2(MT_FLFMT("%s: MOSAL_iomem_map("PHYS_ADDR_FMT", 0x%x) failed (%d)"),			\
		      __func__, mtt_seg_adr,									\
		      1 << (log2_segs + mrwm->props.log2_mtt_seg_sz + MTT_LOG_MTT_ENTRY_SIZE),mosal_rc);        \
	  rc= HH_EAGAIN;											\
	}													\
}

#define XHH_DEV_SPECIFIC_TPTM_FREE_MTT_SEGS do {} while(0)

#define XHH_DEV_SPECIFIC_TPTM_FREE_MTT_SEGS4FMR(mrwm,fmr,seg_start,log2_segs)\
        MOSAL_iomem_unmap(fmr->mtt_iomem,MOSAL_get_kernel_prot_ctx())

#define XHH_DEV_SPECIFIC_TPTM_GET_PROT_MODULE \
        XHH_hob_get_uldm(hob, &mrwm->prot_module)

#define XHH_DEV_SPECIFIC_TPTM_MPT_ENTRY_INIT \
        e->m_io = TRUE;

#define XHH_DEV_SPECIFIC_TPTM_FMR_MPT_ENTRY_INIT \
        mpt_entry_p->m_io = TRUE
	
#define XHH_DEV_SPECIFIC_TPTM_FMR_INFO_INIT(fmr_info_p) do {} while(0)

#define MPT_MTT_ADR(a)   (a->mtt_seg_adr)
#define MPT_MTT_ADR_DOT(a)   (a.mtt_seg_adr)

#define XHH_DEV_SPECIFIC_TPTM_COMMIT_MTT_ENTRIES \
	MOSAL_MMAP_IO_WRITE_BUF_QWORD(fmr_info_p->mtt_entries, \
		fmr_info_p->mtt_entries_draft, \
		real_num_of_pages)
				
#define XHH_DEV_SPECIFIC_TPTM_MODIFY_MPT_QWORD(mpt_entry, offset, val) \
	MOSAL_MMAP_IO_WRITE_QWORD(((mpt_entry) + (offset)), val)

#define XHH_DEV_SPECIFIC_TPTM_MODIFY_MPT_DWORD(mpt_entry, offset, val) \
	MOSAL_MMAP_IO_WRITE_DWORD(((mpt_entry) + (offset)), val)

#define XHH_DEV_SPECIFIC_TPTM_MODIFY_MPT_BYTE(mpt_entry, offset, val) \
	MOSAL_MMAP_IO_WRITE_BYTE(((mpt_entry) + (offset)), val)
	
#endif
