/*
  This software is available to you under a choice of one of two
  licenses.  You may choose to be licensed under the terms of the GNU
  General Public License (GPL) Version 2, available at
  <http://www.fsf.org/copyleft/gpl.html>, or the OpenIB.org BSD
  license, available in the LICENSE.TXT file accompanying this
  software.  These details are also available at
  <http://openib.org/license.html>.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
  BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
  ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
  CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
  SOFTWARE.

  Copyright (c) 2004 Mellanox Technologies Ltd.  All rights reserved.
*/

#ifndef CI_XHHUL_CQM_DEV_CI
#define CI_XHHUL_CQM_DEV_CI

/* Given CQ must be locked when calling this function */
inline static void sync_consumer_index(XHHUL_cq_t *cq_p, const char* caller)
{
  HH_ret_t rc;

  if (cq_p->pending_cq_dbell > 0) {
    MTL_DEBUG4(MT_FLFMT("%s: Ringing CQ DB (CQN=0x%X) to increment CI by %u"),caller,
               cq_p->cq_num, cq_p->pending_cq_dbell);
    rc= XHH_uar_cq_cmd(cq_p->uar,TAVOR_IF_UAR_CQ_INC_CI,cq_p->cq_num,cq_p->pending_cq_dbell-1);
    if (rc != HH_OK) {
      MTL_ERROR2(MT_FLFMT("%s: Failed XHH_uar_cq_cmd (%s)"), caller, HH_strerror_sym(rc));
      /* Even though, this is not a show stopper. Let's continue until we get CQ error. */
    } else {
      cq_p->pending_cq_dbell= 0;
    }
  }
}

inline static HH_ret_t init_uar_info(HHUL_hca_hndl_t hca, XHHUL_cq_t *cq, XHH_cq_ul_resources_t *ul_res_p)
{
    HH_ret_t rc;

    rc= XHHUL_hob_get_uar(hca,&(cq->uar));
    if (rc != HH_OK) {
      MTL_ERROR1(MT_FLFMT("%s: Failed getting XHHUL_hob's UAR (err=%d)."),__func__,rc);
      return rc;
    }
    rc= XHH_uar_get_index(cq->uar,&(ul_res_p->uar_index)); 
    if (rc != HH_OK) {
      MTL_ERROR1(MT_FLFMT("%s: Failed getting UAR index (err=%d)"), __func__, rc);
      return rc;
    }
    return HH_OK;
}


inline static HH_ret_t uar_arm_cq_cmd(XHHUL_cq_t *cq, XHH_uar_cq_cmd_t arm_type, u_int32_t param)
{
    return (XHH_uar_cq_cmd(cq->uar, arm_type, cq->cq_num, param));
}

/* Count number of real CQEs (i.e., including CQEs with error) up to given limit */
/* (return the real number of CQEs) */
/* The function must be invoked with CQ lock held */
static VAPI_cqe_num_t count_cqes( 
  /*IN*/ XHHUL_cq_t *cq_p,
  /*IN*/ VAPI_cqe_num_t cqe_cnt_limit, /* Limit count up to given HW CQEs */
  /*OUT*/ VAPI_cqe_num_t *hw_cqe_cnt_p /* HW CQEs count (optional) */
)
{
  volatile u_int32_t *cur_cqe;
  VAPI_cqe_num_t sw_cqe_cntr= 0;
  VAPI_cqe_num_t hw_cqe_cntr= 0;
  u_int32_t wqe_addr_32lsb;
  IB_wqpn_t qpn;
  XHHUL_qp_t qp;
  u_int8_t opcode;
  u_int16_t dbdcnt= 0;
  MT_bool is_sq_completion;

  /* Count CQEs including "external" of CQEs with error */
  while (hw_cqe_cntr < cqe_cnt_limit) {
    /* Find CQE and check ownership */
    cur_cqe= (volatile u_int32_t *)
      (cq_p->cur_buf.cqe_buf_base + 
       (((cq_p->cur_buf.consumer_counter + hw_cqe_cntr) & MASK32(cq_p->cur_buf.log2_num_o_cqes)) 
        << LOG2_CQE_SZ));
    if (is_cqe_hw_own(cur_cqe))  break; /* no more CQEs */

    opcode= ((volatile u_int8_t*)cur_cqe)[CQE_OPCODE_BYTE_OFFSET]; /* get completion status */
    if ((opcode & CQE_ERROR_STATUS_MASK) != CQE_ERROR_STATUS_MASK) {  /* Completed OK */	
      sw_cqe_cntr++;
    } else { /* CQE with error - count external CQEs */
      
      wqe_addr_32lsb= (
        MOSAL_cpu_to_be32(
          cur_cqe[MT_BYTE_OFFSET(ADD_DEV_PRM_PREFIX(completion_queue_entry_st), wqe_adr)>>2])
          & 
          (~MASK32(CQE_WQE_ADR_BIT_SZ)));

      qpn= (
        MOSAL_cpu_to_be32(
          cur_cqe[MT_BYTE_OFFSET(ADD_DEV_PRM_PREFIX(completion_queue_entry_st), my_qpn)>>2]) 
          & 
          MASK32(24) );
      
      dbdcnt= (
        MOSAL_cpu_to_be32(
          cur_cqe[MT_BYTE_OFFSET(ADD_DEV_PRM_PREFIX(completion_queue_entry_st),
                              immediate_ethertype_pkey_indx_eecredits)>>2]) 
          & 
          MASK32(CQE_ERROR_DBDCNT_BIT_SIZE) );

      /* Add total number of WQEs "hang" over given CQE with error */
      qp= find_qp(cq_p,qpn);
      if (qp == INVALID_QP_HNDL) {
	      MTL_ERROR1(MT_FLFMT("%s: Got completion for unknown QPn=0x%X"
				  " (not associated with this CQn=0x%X) - CQE will not be counted !"), 
			 __func__, qpn, cq_p->cq_num);
	      DUMP_CQE(cq_p->cq_num,
		       (cq_p->cur_buf.consumer_counter  + hw_cqe_cntr) & 
			MASK32(cq_p->cur_buf.log2_num_o_cqes),
		       cur_cqe);
      } else {
	      is_sq_completion = (opcode == (u_int8_t)(ADD_DEV_PREFIX(IF_CQE_ERR_OPCODE_SQ)& 0xFF)) ? 
				      TRUE : FALSE;
	      sw_cqe_cntr+= XHHUL_qpm_wqe_cnt(cq_p->qpm, qp, wqe_addr_32lsb, dbdcnt, 
					      is_sq_completion);
      }
    }
    hw_cqe_cntr++; /* Continue to the next HW CQE */
  }
  
  if (hw_cqe_cnt_p)  *hw_cqe_cnt_p= hw_cqe_cntr;
  MTL_DEBUG5(MT_FLFMT("%s: cqe_cnt_limit=%d sw_cqe_cntr=%d hw_cqe_cntr=%d\n"),__func__,
              cqe_cnt_limit,sw_cqe_cntr,hw_cqe_cntr);
  return sw_cqe_cntr;
}

inline static u_int32_t compute_arm_n_param(XHHUL_cq_t *cq_p, u_int32_t cqe_num)
{
	VAPI_cqe_num_t hw_cqe_cnt,sw_cqe_cnt;

	sw_cqe_cnt= count_cqes(cq_p,cqe_num,&hw_cqe_cnt);
        return (sw_cqe_cnt >= cqe_num) ? 1 : hw_cqe_cnt + (cqe_num - sw_cqe_cnt) ;
}

/* Reuse given CQE in order to report "flush-error" for the next WQE in a queue */
inline static void recycle_cqe(volatile u_int32_t* cqe_p, 
  u_int32_t next_wqe_addr_32lsb, u_int32_t new_dbd_cnt)
{
  /* Operations done directly on original "big-endian" CQE */
  /* Set next WQE's address */
  cqe_p[MT_BYTE_OFFSET(ADD_DEV_PRM_PREFIX(completion_queue_entry_st),wqe_adr)>>2]= MOSAL_cpu_to_be32(
      next_wqe_addr_32lsb & (~MASK32(CQE_WQE_ADR_BIT_SZ)) ); /* Mask off "wqe_adr" */
  /* Mark as "ERR_FLUSH" with updated dbd_cnt */
  cqe_p[MT_BYTE_OFFSET(ADD_DEV_PRM_PREFIX(completion_queue_entry_st),immediate_ethertype_pkey_indx_eecredits)>>2]=
    MOSAL_cpu_to_be32(
      (ADD_DEV_PREFIX(IF_COMP_STATUS_ERR_FLUSH) << (CQE_ERROR_SYNDROM_BIT_OFFSET & MASK32(5)) ) |
      new_dbd_cnt);
}

#endif /* CI_XHHUL_CQM_DEV_CI */
