/*
  This software is available to you under a choice of one of two
  licenses.  You may choose to be licensed under the terms of the GNU
  General Public License (GPL) Version 2, available at
  <http://www.fsf.org/copyleft/gpl.html>, or the OpenIB.org BSD
  license, available in the LICENSE.TXT file accompanying this
  software.  These details are also available at
  <http://openib.org/license.html>.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
  BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
  ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
  CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
  SOFTWARE.

  Copyright (c) 2004 Mellanox Technologies Ltd.  All rights reserved.
*/

#ifndef H_XHH_HOB_H
#define H_XHH_HOB_H

#include <vapi.h>
#include <vip_delay_unlock.h>
#include <mosal.h>
#include <hh.h>
#include <xhh.h>
#include <hob_comm.h>

#define XHH_RESERVED_PD(hob)        0

/* Prototypes */

/******************************************************************************
 *  Function:     XHH_hob_get_uldm 
 *
 *  Arguments:
 *        hob
 *        uldm_p - Included XHH_uldm object 
 *
 *  Returns:
 *    HH_OK
 *    HH_EINVAL - Invalid parameters
 *
 *  Description:
 *    Get a handle to associated user-level domain management object.
 */
HH_ret_t XHH_hob_get_uldm ( /*IN*/ XHH_hob_t hob, 
                            /*OUT*/ XHH_uldm_t *uldm_p );


/******************************************************************************
 *  Function:     XHH_hob_get_udavm_info 
 *
 *  Arguments:
 *        hob
 *        udavm_p - Included XHH_udavm object
 *        use_priv_udav - flag:  TRUE if using privileged UDAV mode
 *        av_on_board   - flag:  TRUE if should use DDR SDRAM on Tavor for UDAVs
 *        lkey          - lkey allocated for udav table in privileged UDAV mode 
 *        hide_ddr      - flag:  TRUE if should use host memory for UDAVs
 *
 *  Returns:
 *    HH_OK
 *    HH_EINVAL - Invalid parameters
 *
 *  Description:
 *    return udavm information (needed by uldm object in PD allocation).
 */
HH_ret_t XHH_hob_get_udavm_info ( /*IN*/ XHH_hob_t hob, 
                                  /*OUT*/ XHH_udavm_t *udavm_p,
                                  /*OUT*/ MT_bool *use_priv_udav,
                                  /*OUT*/ MT_bool *av_on_board,
                                  /*OUT*/ VAPI_lkey_t  *lkey,
                                  /*OUT*/ u_int32_t *max_ah_num,
                                  /*OUT*/ MT_bool *hide_ddr);

HH_ret_t XHH_hob_create_ud_av(HH_hca_hndl_t hca_hndl,HH_pd_hndl_t pd,VAPI_ud_av_t *av_p, HH_ud_av_hndl_t *ah_p);
HH_ret_t XHH_hob_modify_ud_av(HH_hca_hndl_t hca_hndl, HH_ud_av_hndl_t ah,VAPI_ud_av_t *av_p);
HH_ret_t XHH_hob_query_ud_av(HH_hca_hndl_t hca_hndl, HH_ud_av_hndl_t ah,VAPI_ud_av_t *av_p);
HH_ret_t XHH_hob_destroy_ud_av(HH_hca_hndl_t hca_hndl, HH_ud_av_hndl_t ah);
HH_ret_t XHH_hob_lamm_alloc(HH_hca_hndl_t  hca_hndl,VAPI_size_t size,u_int8_t align_shift,VAPI_phy_addr_t*  buf_p);
HH_ret_t XHH_hob_lamm_query(HH_hca_hndl_t  hca_hndl,u_int8_t      align_shift,VAPI_size_t*    total_mem,    
                             VAPI_size_t*    free_mem,VAPI_size_t*    largest_chunk,  
                             VAPI_phy_addr_t*  largest_free_addr_p);

HH_ret_t XHH_hob_lamm_free(HH_hca_hndl_t  hca_hndl,VAPI_phy_addr_t  buf, VAPI_size_t size);


#endif  /* H_XHH_H */
