/*
  This software is available to you under a choice of one of two
  licenses.  You may choose to be licensed under the terms of the GNU
  General Public License (GPL) Version 2, available at
  <http://www.fsf.org/copyleft/gpl.html>, or the OpenIB.org BSD
  license, available in the LICENSE.TXT file accompanying this
  software.  These details are also available at
  <http://openib.org/license.html>.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
  BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
  ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
  CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
  SOFTWARE.

  Copyright (c) 2004 Mellanox Technologies Ltd.  All rights reserved.
*/

#ifndef H_XHHUL_QPM_DEV_H
#define H_XHHUL_QPM_DEV_H

#include <uar.h>

/* #define MAX_WQE_SZ 1008 */

#define WQE_SEGMENT_NEXT_ST             wqe_segment_next_st
#define WQE_SEGMENT_CTRL_SEND_ST        wqe_segment_ctrl_send_st
#define WQE_SEGMENT_RD_ST               wqe_segment_rd_st
#define WQE_SEGMENT_UD_ST               wqe_segment_ud_st
#define WQE_SEGMENT_REMOTE_ADDRESS_ST   wqe_segment_remote_address_st
#define WQE_SEGMENT_ATOMIC_ST           wqe_segment_atomic_st
#define WQE_SEGMENT_BIND_ST             wqe_segment_bind_st
#define WQE_SEGMENT_DATA_PTR_ST         wqe_segment_data_ptr_st
#define WQE_SEGMENT_CTRL_MLX_ST         wqe_segment_ctrl_mlx_st
#define WQE_SEGMENT_CTRL_RECV_ST        wqe_segment_ctrl_recv_st

#define XHH_DEV_SPECIFIC_UL_QP_T_FIELDS                                 \
  XHH_uar_t uar;                                                        \
  MT_bool is_priv_ud_av;      /* Privileged UD AVs are enforced */      \
  VAPI_lkey_t   ud_av_memkey; /* Memory key to put for UD AV handles */
  
#define XHH_DEV_SPECIFIC_INIT_QP           \
  {                                        \
      XHHUL_pdm_t pdm;                     \
      MT_bool ok_sqp;                      \
      rc= XHHUL_hob_get_pdm(hca,&pdm);     \
      if (rc != HH_OK) {                   \
        MTL_ERROR2(MT_FLFMT("init_qp: Failed XHHUL_hob_get_pdm (%d:%s)"),rc, HH_strerror_sym(rc)); \
        return rc;                         \
      }                                    \
                                           \
      rc= XHHUL_hob_is_priv_ud_av(hca,&(new_qp->is_priv_ud_av));           \
      if (rc != HH_OK) {                                                   \
        MTL_ERROR2(MT_FLFMT("init_qp: Failed  XHHUL_hob_is_priv_ud_av (%d:%s)"),rc, HH_strerror_sym(rc));   \
        return rc;                                                         \
      }                                                                    \
      rc= XHHUL_pdm_get_ud_av_memkey_sqp_ok(pdm,qp_init_attr_p->pd,&ok_sqp,&(new_qp->ud_av_memkey));        \
      if (rc != HH_OK) {                                                   \
        MTL_ERROR2(MT_FLFMT("init_qp: Failed XHHUL_pdm_get_ud_av_memkey (%d:%s)"),rc, HH_strerror_sym(rc)); \
        return rc;                                                         \
      }                                                                    \
      if (new_qp->sqp_type != VAPI_REGULAR_QP && ok_sqp == FALSE) {         \
          /* the protection domain uses DDR memory for UDAV's -- not good for sqps */ \
          MTL_ERROR2(MT_FLFMT("***WARNING***: AVs for special QPs should use HOST memory; the provided PD has its AVs in DDR memory.\n")); \
      }                                                                    \
  }
  


#define XHH_DEV_SPECIFIC_CREATE_QP_DONE do {} while(0)

#define XHH_DEV_SPECIFIC_RESET_RQ_DONE do {} while(0)
#define XHH_DEV_SPECIFIC_RESET_SQ_DONE do {} while(0)

#define XHH_DEV_SPECIFIC_WQE_INIT_NEXT \
  MT_INSERT_ARRAY32(segment_p,dbd,     \
    MT_BIT_OFFSET(WQE_SEGMENT_NEXT_ST,dbd),MT_BIT_SIZE(WQE_SEGMENT_NEXT_ST,dbd));

#define XHH_DEV_SPECIFIC_PACK_RECV_NEXT \
	  segment_p[MT_BYTE_OFFSET(WQE_SEGMENT_NEXT_ST,nda_31_6)>>2]= ( next_wqe_32lsb & (~MASK32(6)) ) \
		| 1 ;  /* LS-bit is set to work around bug #16159/16160/16161 */;							\
      MT_INSERT_ARRAY32(segment_p,1, /* DBD always '1 for RQ */                                     \
        MT_BIT_OFFSET(WQE_SEGMENT_NEXT_ST,dbd),MT_BIT_SIZE(WQE_SEGMENT_NEXT_ST,dbd));

#define XHHUL_DEV_SPECIFIC_WQE_BUILD_RECV do {} while(0)

#define XHH_DEV_SPECIFIC_WQE_PACK_UD \
      segment_p[MT_BYTE_OFFSET(WQE_SEGMENT_UD_ST,l_key)>>2]= qp->ud_av_memkey;                    \
      segment_p[MT_BYTE_OFFSET(WQE_SEGMENT_UD_ST,av_address_63_32)>>2]= (u_int32_t)(((u_int64_t)ah)>>32);  \
      segment_p[MT_BYTE_OFFSET(WQE_SEGMENT_UD_ST,av_address_31_5)>>2]= ((u_int32_t)ah & (~MASK32(5)) );


#define XHH_DEV_SPECIFIC_EXTRACT_DBD \
      return MT_EXTRACT32(MOSAL_be32_to_cpu(wqe[MT_BYTE_OFFSET(WQE_SEGMENT_NEXT_ST,dbd) >> 2]), \
                       MT_BIT_OFFSET(WQE_SEGMENT_NEXT_ST,dbd) & MASK32(5),                      \
                       MT_BIT_SIZE(WQE_SEGMENT_NEXT_ST,dbd) & MASK32(5));

#define XHHUL_DEV_SPECIFIC_ALLOC_WQE_BUF_INIT do {} while(0)
#define XHHUL_DEV_SPECIFIC_ALLOC_WQE_BUF_DONE do {} while(0)

#define XHH_DEV_SPECIFIC_SQ_TEST_PREV_WQE_NOT_NULL \
	  if (qp->sq_res.last_posted_p != NULL) 


#define XHH_DEV_SPECIFIC_SQ_ALLOC_WQE_LINK_AND_RING \
          sq_dbell.qpn= qp->qpn;                    \
          sq_dbell.next_addr_32lsb= (u_int32_t)((MT_virt_addr_t)next_wqe & 0xFFFFFFFF);
          
#define XHH_DEV_SPECIFIC_POST_SEND_REQS_INTERNAL \
          sq_dbell.qpn= qp->qpn;        \
          sq_dbell.next_addr_32lsb= (u_int32_t)(first_wqe_nda & 0xFFFFFFFF);

/* #define XHH_DEV_SPECIFIC_POST_SEND_REQ do {} while(0) */

#define XHH_DEV_SPECIFIC_POST_SEND_REQS \
	ret = XHHUL_qpm_post_send_reqs_internal(hca, qp, num_of_requests, \
	                               send_req_array, do_sgsize_check);  \
	if (ret) {                                                        \
	     MOSAL_spinlock_unlock(&(qp->sq_res.q_lock));                 \
	     return ret;                                                  \
	}
 
#define XHH_DEV_SPECIFIC_POST_RECV_REQ  \
          {                                                                                              \
	          u_int32_t next_draft[WQE_SEG_SZ_NEXT>>2];                                                  \
		      XHH_uar_recvq_dbell_t rq_dbell;         													 \
	          /* Update "next" segment of previous WQE (if any) */										 \
	          if (qp->rq_res.last_posted_p != NULL) {													 \
		         /* Build linking "next" segment in last posted WQE */									 \
		         WQE_pack_recv_next(next_draft, (u_int32_t)(MT_ulong_ptr_t) next_wqe, wqe_sz_dwords>>2); \
		         for (i= 0;i < (WQE_SEG_SZ_NEXT>>2) ;i++) {  											 \
			        /* This copy assures big-endian as well as that NDS is written last */				 \
			        qp->rq_res.last_posted_p[i]= MOSAL_cpu_to_be32(next_draft[i]);						 \
		         }																						 \
	          }																							 \
              qp->rq_res.last_posted_p= next_wqe;														 \
              																							 \
			  /* Ring doorbell */                                                                        \
              rq_dbell.qpn= qp->qpn;                                                                     \
              rq_dbell.next_addr_32lsb= (u_int32_t)((MT_virt_addr_t)next_wqe & 0xFFFFFFFF);              \
              rq_dbell.next_size= wqe_sz_dwords>>2;                                                      \
              rq_dbell.credits= 1;                                                                       \
              XHH_uar_recvq_dbell(qp->uar,&rq_dbell);                                                    \
          }

#define XHHUL_DEV_SPECIFIC_QPM_POST_RECV_REQS_VARS \
		  u_int32_t next_draft[WQE_SEG_SZ_NEXT>>2]; /* Build "next" segment here */ \
		  XHH_uar_recvq_dbell_t rq_dbell;											\
		  volatile u_int32_t* prev_wqe_p= qp->rq_res.last_posted_p
  
#define XHHUL_DEV_SPECIFIC_QPM_POST_RECV_REQS_INIT \
  rq_dbell.qpn= qp->qpn; /* Fixed for all doorbells */                              \
  rq_dbell.credits= 0 /* During the loop, doorbell is rung every 256 WQEs */
  
#define XHHUL_DEV_SPECIFIC_QPM_POST_RECV_REQS_LOOP \
    if ((reqi & 0xFF) == 0) {                                                                   \
      /* save NDA+NDS of first WQE in each 256 WQEs chain for the doorbell */					\
      rq_dbell.next_addr_32lsb= (u_int32_t)((MT_virt_addr_t)next_wqe & 0xFFFFFFFF);				\
      rq_dbell.next_size= wqe_sz_dwords>>2;														\
    }																							\
																								\
    if (prev_wqe_p != NULL) { /* first in the chain may be the first since reset */				\
      /* Update "next" segment of previous WQE */												\
      /* Build linking "next" segment in last posted WQE */										\
      WQE_pack_recv_next(next_draft, (u_int32_t)(MT_ulong_ptr_t)next_wqe, wqe_sz_dwords>>2);	\
      for (i= 0;i < (WQE_SEG_SZ_NEXT>>2) ;i++) {												\
        /* This copy assures big-endian as well as that NDS is written last */					\
        prev_wqe_p[i]= MOSAL_cpu_to_be32(next_draft[i]);										\
      }																							\
    }																							\
    prev_wqe_p= next_wqe;																		\
																								\
    if ((reqi & 0xFF) == 0xFF) { /* last in 256 WQEs chain - ring doorbell */					\
      /* Ring doorbell on the first WQE only */													\
      XHH_uar_recvq_dbell(qp->uar,&rq_dbell);													\
    }																							
																								
#define XHHUL_DEV_SPECIFIC_QPM_POST_RECV_REQS_DONE \
  if ((reqi & 0xFF) != 0) { /* Doorbel for last WQEs was not rung */  \
    rq_dbell.credits= (reqi & 0xFF);								  \
    XHH_uar_recvq_dbell(qp->uar,&rq_dbell);							  \
  }																	  \
  qp->rq_res.last_posted_p= next_wqe;								  


#define WQE_RCV_DRAFT_START_COPY  0
#define WQE_SND_DRAFT_START_COPY  0

#define XHHUL_DEV_SPECIFIC_UDAVM(qp) (NULL)

#endif /* H_XHHUL_CQM_H */
