/*
  This software is available to you under a choice of one of two
  licenses.  You may choose to be licensed under the terms of the GNU
  General Public License (GPL) Version 2, available at
  <http://www.fsf.org/copyleft/gpl.html>, or the OpenIB.org BSD
  license, available in the LICENSE.TXT file accompanying this
  software.  These details are also available at
  <http://openib.org/license.html>.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
  BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
  ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
  CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
  SOFTWARE.

  Copyright (c) 2004 Mellanox Technologies Ltd.  All rights reserved.
*/

#ifndef CI_XHHUL_QPM_DEV_CI
#define CI_XHHUL_QPM_DEV_CI

inline static HH_ret_t init_uar_info(HHUL_hca_hndl_t hca, XHHUL_qp_t qp, XHH_qp_ul_resources_t *ul_res_p)
{
    HH_ret_t rc;

    rc= XHHUL_hob_get_uar(hca,&(qp->uar));
    if (rc != HH_OK) {
      MTL_ERROR1(MT_FLFMT("%s: Failed getting XHHUL_hob's UAR (err=%d)."),__func__,rc);
      return rc;
    }
    rc= XHH_uar_get_index(qp->uar,&(ul_res_p->uar_index)); 
    if (rc != HH_OK) {
      MTL_ERROR1(MT_FLFMT("%s: Failed getting UAR index (err=%d)"), __func__, rc);
      return rc;
    }
    return HH_OK;
}

inline static HH_ret_t uar_sendq_dbell(XHHUL_qp_t qp, XHH_uar_sendq_dbell_t *sq_dbell)
{
    MTPERF_TIME_START(XHH_uar_sendq_dbell);
    XHH_uar_sendq_dbell(qp->uar,sq_dbell);
    MTPERF_TIME_END(XHH_uar_sendq_dbell);
    return HH_OK;
}

inline static HH_ret_t uar_sendq_rd_dbell(XHHUL_qp_t qp, XHH_uar_sendq_dbell_t *sq_dbell, VAPI_sr_desc_t *send_req_p)
{
    return XHH_uar_sendq_rd_dbell(qp->uar,sq_dbell,send_req_p->eecn);
}

/* Init a not-connected (invalid) "next" segment (i.e. change ONLY NDS=0) */
inline static u_int32_t WQE_init_next(u_int32_t *wqe_buf)
{
  MT_INSERT_ARRAY32(wqe_buf,0,
     MT_BIT_OFFSET(WQE_SEGMENT_NEXT_ST,nds),MT_BIT_SIZE(WQE_SEGMENT_NEXT_ST,nds));
  return WQE_SEG_SZ_NEXT;
}

inline static u_int32_t WQE_init_next_recv(u_int32_t *wqe_buf)
{
  MT_INSERT_ARRAY32(wqe_buf,0,
     MT_BIT_OFFSET(WQE_SEGMENT_NEXT_ST,nds),MT_BIT_SIZE(WQE_SEGMENT_NEXT_ST,nds));
  return WQE_SEG_SZ_NEXT;
}

inline static void initialize_sq_last_posted(queue_res_t *sq_res)
{
  sq_res->last_posted_p = NULL;
}

#endif /* CI_XHHUL_QPM_DEV_CI */
