/*
  This software is available to you under a choice of one of two
  licenses.  You may choose to be licensed under the terms of the GNU
  General Public License (GPL) Version 2, available at
  <http://www.fsf.org/copyleft/gpl.html>, or the OpenIB.org BSD
  license, available in the LICENSE.TXT file accompanying this
  software.  These details are also available at
  <http://openib.org/license.html>.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
  BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
  ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
  CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
  SOFTWARE.

  Copyright (c) 2004 Mellanox Technologies Ltd.  All rights reserved.
*/

#define C_XHHUL_SRQM_COMM_C

#include <mosal.h>
#include <ib_defs.h>
#include <vapi.h>
#include <log2.h>
#include <xhh.h>
#include <ul.h>

#include <ul_hob.h>
#include <ul_pdm.h>
#include <vapi_common.h>

#include "ul_srqm.h"
#include <ul_srqm_priv.h>
#include <dev_ul_srqm_inline.h>
#include <ul_srqm_inline.h>

HH_ret_t XHHUL_srqm_post_recv_reqs(
                                 /*IN*/ HHUL_hca_hndl_t hca, 
                                 /*IN*/ HHUL_srq_hndl_t hhul_srq, 
                                 /*IN*/ u_int32_t num_of_requests,
                                 /*IN*/ VAPI_rr_desc_t *recv_req_array,
                                 /*OUT*/ u_int32_t *posted_requests_p
                                 )
{
  XHHUL_srq_t srq= (XHHUL_srq_t)hhul_srq;
  u_int32_t* wqe_draft= srq->wqe_draft;
  u_int32_t next_draft[WQE_SEG_SZ_NEXT>>2]; /* Build "next" segment here */
  volatile u_int32_t* next_wqe= NULL; /* Actual WQE pointer */
  volatile u_int32_t* resized_next_wqe= NULL; 
  MT_virt_addr_t next_wqe_hca_virt, resized_virt_offset=0;
  volatile u_int32_t* prev_wqe_p;
  volatile u_int32_t* resized_prev_wqe_p= NULL;
  u_int32_t wqe_sz_dwords= 0;
  u_int32_t i,reqi,next2post_index;
  XHH_uar_recvq_dbell_t rq_dbell;
  HH_ret_t ret= HH_OK;

  *posted_requests_p= 0;
  if (num_of_requests == 0)  return HH_OK; /* nothing to do */
  
  /* Init. invariant RQ doorbell fields */
  rq_dbell.qpn= srq->srqn;
  rq_dbell.next_size= 0; /* For SRQs, NDS comes from SRQC */
  rq_dbell.credits= 0;   /* For 256 WQEs quantums */
  
  MOSAL_spinlock_irq_lock(&(srq->q_lock)); /* protect wqe_draft as well as WQE allocation/link */
  
  prev_wqe_p= (srq->last_posted_hca_va == 0) ? 
    NULL : 
    (void*)(srq->last_posted_hca_va + srq->real_virt_offset); 
  rq_dbell.next_addr_32lsb= (u_int32_t)srq->free_wqes_list_top; /* For first chain */

  if (srq->resized_buf_p != NULL) {
    resized_virt_offset= srq->resized_buf_p->wqe_buf - srq->hca_virt_wqe_buf;
    resized_prev_wqe_p= (srq->last_posted_hca_va == 0) ? 
      NULL : 
      (void*)(srq->last_posted_hca_va + resized_virt_offset);
  }

  /* Build and link all WQEs */
  for (reqi= 0; (reqi < num_of_requests) ; reqi++) {
    
    if (srq->free_wqes_list_top == 0) {
      MTL_ERROR2(MT_FLFMT(
        "%s: Posting only %u requests out of %u"), __func__, *posted_requests_p, num_of_requests);
      ret= HH_EAGAIN;
      break;
    }
    
    if (srq->max_sentries < recv_req_array[reqi].sg_lst_len) {
      MTL_ERROR2(MT_FLFMT(
        "%s: Scatter list of req. #%u is too large (%u entries > max_sg_sz=%u)"), __func__,
                reqi,recv_req_array[reqi].sg_lst_len,srq->max_sentries);
      ret= HH_EINVAL_SG_NUM;
      break;
    }

    if (recv_req_array[reqi].opcode != VAPI_RECEIVE) {
      MTL_ERROR2(MT_FLFMT(
        "%s: Invalid opcode (%d=%s)in request #%d"), __func__,
         recv_req_array[reqi].opcode, VAPI_wr_opcode_sym(recv_req_array[reqi].opcode), reqi);
      ret= HH_EINVAL_OPCODE;
      break;
    }

    /* Build WQE */
    wqe_sz_dwords= (WQE_build_recv(srq,recv_req_array+reqi,wqe_draft) >> 2);
  #ifdef MAX_DEBUG
    if ((wqe_sz_dwords<<2) > (1U << srq->log2_max_wqe_sz)) {
      MTL_ERROR1(MT_FLFMT("%s: SRQ 0x%X: WQE too large (%d > max=%d)"), __func__,
                 srq->srqn,(wqe_sz_dwords<<2),(1U << srq->log2_max_wqe_sz));
    	}
  #endif
    
    /* Allocate next WQE */
    next_wqe_hca_virt= srq->free_wqes_list_top ;
    next_wqe= (volatile u_int32_t*)(next_wqe_hca_virt + srq->real_virt_offset);
    srq->free_wqes_list_top = *((MT_virt_addr_t*)next_wqe);/* next WQE is in the WQE (when free) */
    if (srq->free_wqes_list_top == 0) {
	/* list is now empty. This not really needed. */
	srq->free_wqes_list_end = 0;
    }
    /* Save WQE ID */
    next2post_index= (u_int32_t)(((u_int8_t*)next_wqe - (u_int8_t*)srq->cur_buf_p->wqe_buf) >> 
                                 srq->log2_max_wqe_sz);
    MTL_DEBUG6(MT_FLFMT("%s: SRQ 0x%X posting WQE at index %u (real_va=%p, hca_va="
                        VIRT_ADDR_FMT")"), __func__, 
               srq->srqn, next2post_index, next_wqe, next_wqe_hca_virt); //DEBUG
    srq->wqe_id[next2post_index]= recv_req_array[reqi].id;  /* Save WQE ID */

    /* copy (while swapping,if needed) the wqe_draft to the actual WQE */
    /* TBD: for big-endian machines we can optimize here and use memcpy */
    for (i= 0; i < wqe_sz_dwords; i++) {
      next_wqe[i]= MOSAL_cpu_to_be32(wqe_draft[i]);
    }
    
    if (srq->resized_buf_p != NULL) {
      resized_next_wqe= (volatile u_int32_t*)(next_wqe_hca_virt + resized_virt_offset);
      for (i= 0; i < wqe_sz_dwords; i++) { /* Copy WQE to resized buffer */
        resized_next_wqe[i]= MOSAL_cpu_to_be32(wqe_draft[i]);
      }
    }

    if (prev_wqe_p != NULL) { 
      /* Update "next" segment of previous WQE */
      /* Build linking "next" segment in last posted WQE */
      WQE_pack_recv_next(next_draft, (u_int32_t)next_wqe_hca_virt);
      for (i= 0;i < (WQE_SEG_SZ_NEXT>>2) ;i++) {
        /* This copy assures big-endian as well as that DBD/NDS is written last */
        prev_wqe_p[i]= MOSAL_cpu_to_be32(next_draft[i]);
      }
      if (srq->resized_buf_p != NULL) {
        for (i= 0;i < (WQE_SEG_SZ_NEXT>>2) ;i++) { /* Link in resized buffer, too */
          resized_prev_wqe_p[i]= MOSAL_cpu_to_be32(next_draft[i]);
        }
      }
    }

    prev_wqe_p= next_wqe;
    resized_prev_wqe_p= resized_next_wqe;

    (*posted_requests_p)++;

    if (((*posted_requests_p) & 0xFF) == 0) { /* ring RQ doorbell every 256 WQEs */
      XHH_uar_recvq_dbell(srq->uar,&rq_dbell);
      rq_dbell.next_addr_32lsb= (u_int32_t)srq->free_wqes_list_top; /* For next chain */
    }
  }
  
  if (((*posted_requests_p) & 0xFF) != 0) { /* left-overs (less than 256 WQEs) */
    rq_dbell.credits= (*posted_requests_p) & 0xFF;
    XHH_uar_recvq_dbell(srq->uar,&rq_dbell);
  }
  
  srq->last_posted_hca_va= ((MT_virt_addr_t)prev_wqe_p) - srq->real_virt_offset;
  srq->cur_outs+= *posted_requests_p; /* redundant info - for debug */
  
  MOSAL_spinlock_unlock(&(srq->q_lock));
  return ret;

}
