/*
  This software is available to you under a choice of one of two
  licenses.  You may choose to be licensed under the terms of the GNU
  General Public License (GPL) Version 2, available at
  <http://www.fsf.org/copyleft/gpl.html>, or the OpenIB.org BSD
  license, available in the LICENSE.TXT file accompanying this
  software.  These details are also available at
  <http://openib.org/license.html>.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
  BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
  ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
  CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
  SOFTWARE.

  Copyright (c) 2004 Mellanox Technologies Ltd.  All rights reserved.
*/

#include <mosal.h>
#include <dev_common.h>
#include <log2.h>
#include <vip_array.h>
#include <vip_common.h>
#include <udav_comm.h>

  
  
/************ LOCAL FUNCTIONS ************************************************************************/

void XHH_udavm_fill_udav_entry(/*IN */ HH_pd_hndl_t pd, 
                     /*IN */ MT_bool      is_new_pd,
                     /*IN */ VAPI_ud_av_t *av_p, 
                     /*IN */ u_int32_t    *av_entry_p,
		     /*IN */ MT_bool      make_big_endian,
		     /*IN */ MT_bool      use_lam)
{

  int i;
  u_int32_t new_av_arr[8] = {0,0,0,0,0,0,0,0}; /* entry size is 32 bytes */
  u_int32_t *av_buf_p = NULL;
  HH_pd_hndl_t pd_tmp;
  u_int32_t pd_word = 0;
  
  FUNC_IN;
  /*print_udav(av_p);*/
  
  /* if using LAM, create the entry entirely within a temporary buffer, then copy it
   * over to the given udav entry.  If using host memory, create the entry directly in
   * the provided buffer
   */
  if (use_lam) {
      av_buf_p = new_av_arr;
  } else {
      av_buf_p = av_entry_p;
  }
  
  /* if not new PD the we have to clear PD so entry will not considered valid by HW while we modify it*/
  /* need to save the current PD and write it back at the end */
  if (use_lam && !is_new_pd) {
    pd_word = MOSAL_be32_to_cpu(MOSAL_MMAP_IO_READ_DWORD(av_entry_p));
    pd_tmp = MT_EXTRACT32(pd_word, 
                          MT_BIT_OFFSET(PRM_UDAV_ST,pd),
                          MT_BIT_SIZE(PRM_UDAV_ST, pd));
    MOSAL_MMAP_IO_WRITE_DWORD(av_entry_p,0);
  }
  else {
    pd_tmp = pd;
  }
  
  /* PD */
  MT_INSERT_ARRAY32(av_buf_p, pd_tmp, MT_BIT_OFFSET(PRM_UDAV_ST, pd),
                 MT_BIT_SIZE(PRM_UDAV_ST, pd));

  /* port */
  MT_INSERT_ARRAY32(av_buf_p, av_p->port, MT_BIT_OFFSET(PRM_UDAV_ST, port_number),
                 MT_BIT_SIZE(PRM_UDAV_ST, port_number));
  
  /* rlid */
  MT_INSERT_ARRAY32(av_buf_p, av_p->dlid, MT_BIT_OFFSET(PRM_UDAV_ST, rlid),
                 MT_BIT_SIZE(PRM_UDAV_ST, rlid));

  /* mylid_path_bits */
  MT_INSERT_ARRAY32(av_buf_p, av_p->src_path_bits, MT_BIT_OFFSET(PRM_UDAV_ST, my_lid_path_bits),
                 MT_BIT_SIZE(PRM_UDAV_ST, my_lid_path_bits));
  /* grh enable */
  MT_INSERT_ARRAY32(av_buf_p, av_p->grh_flag, MT_BIT_OFFSET(PRM_UDAV_ST, g),
                 MT_BIT_SIZE(PRM_UDAV_ST, g));
  /* hop_limit */
  MT_INSERT_ARRAY32(av_buf_p, av_p->hop_limit, MT_BIT_OFFSET(PRM_UDAV_ST, hop_limit),
                 MT_BIT_SIZE(PRM_UDAV_ST, hop_limit));

  /* max_stat_rate */
  /* 0 - Suited for matched links no flow control
     All other values are transmitted to 1 since this is the only flow control of Tavor */
  MT_INSERT_ARRAY32(av_buf_p, ((av_p->static_rate ==0) ? 0:1), MT_BIT_OFFSET(PRM_UDAV_ST, max_stat_rate),
                 MT_BIT_SIZE(PRM_UDAV_ST, max_stat_rate));
  
  /* msg size - we put allays the max value */
  MT_INSERT_ARRAY32(av_buf_p, 3, MT_BIT_OFFSET(PRM_UDAV_ST, msg),
                 MT_BIT_SIZE(PRM_UDAV_ST, msg));
  
  /* mgid_index (index to port GID table) - 6th (LOG2_PORT_GID_TABLE_SIZE+1) bit is (port-1) */
  MT_INSERT_ARRAY32(av_buf_p, 
    (av_p->sgid_index & MASK32(LOG2_PORT_GID_TABLE_SIZE)) | 
      ((av_p->port - 1) << LOG2_PORT_GID_TABLE_SIZE), 
    MT_BIT_OFFSET(PRM_UDAV_ST, mgid_index), 
    MT_BIT_SIZE(PRM_UDAV_ST, mgid_index));
  
  /* flow_label */
  MT_INSERT_ARRAY32(av_buf_p, av_p->flow_label, MT_BIT_OFFSET(PRM_UDAV_ST, flow_label),
                 MT_BIT_SIZE(PRM_UDAV_ST, flow_label));
  
  /* tclass */
  MT_INSERT_ARRAY32(av_buf_p, av_p->traffic_class, MT_BIT_OFFSET(PRM_UDAV_ST, tclass),
                 MT_BIT_SIZE(PRM_UDAV_ST, tclass));
  
  /* sl */
  MT_INSERT_ARRAY32(av_buf_p, av_p->sl, MT_BIT_OFFSET(PRM_UDAV_ST, sl),
                 MT_BIT_SIZE(PRM_UDAV_ST, sl));
  
  
  /* the gid is coming in BE format so we can insert it directly to the av_entry */
  /* need to fill it only if GRH bit is set */
  if (av_p->grh_flag) {
    /* rgid_127_96 */
    MT_INSERT_ARRAY32(av_buf_p, (make_big_endian==TRUE) ? 
		      ((u_int32_t*)(av_p->dgid))[0] : MOSAL_be32_to_cpu(((u_int32_t*)(av_p->dgid))[0]), 
		      MT_BIT_OFFSET(PRM_UDAV_ST, rgid_127_96),
                      MT_BIT_SIZE(PRM_UDAV_ST, rgid_127_96));
    
    /* rgid_95_64 */
    MT_INSERT_ARRAY32(av_buf_p, (make_big_endian==TRUE) ? 
		      ((u_int32_t*)(av_p->dgid))[1] : MOSAL_be32_to_cpu(((u_int32_t*)(av_p->dgid))[1]), 
		      MT_BIT_OFFSET(PRM_UDAV_ST, rgid_95_64),
                      MT_BIT_SIZE(PRM_UDAV_ST, rgid_95_64));
    
    /* rgid_63_32 */
    MT_INSERT_ARRAY32(av_buf_p, (make_big_endian==TRUE) ?
		      ((u_int32_t*)(av_p->dgid))[2] : MOSAL_be32_to_cpu(((u_int32_t*)(av_p->dgid))[2]), 
		      MT_BIT_OFFSET(PRM_UDAV_ST, rgid_63_32),
                      MT_BIT_SIZE(PRM_UDAV_ST, rgid_63_32));
    
    /* rgid_31_0 */
    MT_INSERT_ARRAY32(av_buf_p, (make_big_endian==TRUE) ?
		      ((u_int32_t*)(av_p->dgid))[3] : MOSAL_be32_to_cpu(((u_int32_t*)(av_p->dgid))[3]), 
		      MT_BIT_OFFSET(PRM_UDAV_ST, rgid_31_0),
                      MT_BIT_SIZE(PRM_UDAV_ST, rgid_31_0));
  }
  else { /* Arbel mode workaround - must give GRH >1 in lowest bits*/
    /* rgid_31_0 */
    MT_INSERT_ARRAY32(av_buf_p, 2, MT_BIT_OFFSET(PRM_UDAV_ST, rgid_31_0),
                   MT_BIT_SIZE(PRM_UDAV_ST, rgid_31_0));

  }
  
  if (use_lam) {
      /* now copy from the temp buffer to the LAM entry in the table.
       * need to write the first DWORDs last since this is the PD that indicates 
       * to HW that this entry is valid.
       * The RGID can be copied as-is, since endianness was handled above.
       * For other fields (which occupy less than 32 bits), endianness can not
       * be handled above -- it must be handled below, in 32-bit copies.
       * For-loop is replicated for efficiency, so that the make_big_endian check is
       * done only once, and not once per loop
       */
      for (i = RGID_OFFSET; i <= RGID_OFFSET_END; i++) {
	    /*MTL_DEBUG1(MT_FLFMT("(i=%d) %p <- 0x%X"),i,av_entry_p+i,new_av_arr[i]);*/
	    MOSAL_MMAP_IO_WRITE_DWORD(av_entry_p+i,new_av_arr[i]); 
      }
      if (make_big_endian==TRUE) {
	  for (i = RGID_OFFSET-1; i >= 0; i--) {
		/*MTL_DEBUG1(MT_FLFMT("(i=%d) %p <- 0x%X"),i,av_entry_p+i,new_av_arr[i]);*/
		MOSAL_MMAP_IO_WRITE_DWORD(av_entry_p+i,MOSAL_cpu_to_be32(new_av_arr[i])); 
	  }
      } else {
	  for (i = RGID_OFFSET-1; i >= 0; i--) {
		/*MTL_DEBUG1(MT_FLFMT("(i=%d) %p <- 0x%X"),i,av_entry_p+i,new_av_arr[i]);*/
		MOSAL_MMAP_IO_WRITE_DWORD(av_entry_p+i,new_av_arr[i]); 
	  }
      }
  } else {	 /* USING HOST MEMORY. */
          /* if is in host memory, only need to adjust non-RGID fields
	   * for the case where want big-endian -- RGID has already been adjusted above.
	   */
      if (make_big_endian==TRUE) {
	  for (i = RGID_OFFSET-1; i >= 0; i--) {
              av_entry_p[i] = MOSAL_cpu_to_be32(av_entry_p[i]);
	  }
      }
  }
 
  FUNC_OUT;

}



/************************************************************************************/
#if 1 <= MAX_DEBUG

void print_udav(VAPI_ud_av_t *av_p)
{
  FUNC_IN;
  MTL_DEBUG1("UDAV values:\n==================\n sl = %d \n dlid = %d\n src_path_bits = %d\n static_rate = %d\n grh_flag = %d\n traffic_class = %d\n"
         " flow_label = %d\n hop_limit = %d\n sgid_index = %d\n port = %d\n, dgid = %d.%d.%d.%d.%d.%d.%d.%d.%d.%d.%d.%d.%d.%d.%d.%d\n",
         av_p->sl, av_p->dlid, av_p->src_path_bits, av_p->static_rate, av_p->grh_flag, av_p->traffic_class, av_p->flow_label,
         av_p->hop_limit, av_p->sgid_index, av_p->port, av_p->dgid[0],av_p->dgid[1],av_p->dgid[2],av_p->dgid[3],av_p->dgid[4],av_p->dgid[5],
         av_p->dgid[6],av_p->dgid[7],av_p->dgid[8],av_p->dgid[9],av_p->dgid[10],av_p->dgid[11],
         av_p->dgid[12],av_p->dgid[13],av_p->dgid[14],av_p->dgid[15]);
  FUNC_OUT;
}
#endif
/************************************************************************************/

HH_ret_t XHH_udavm_parse_udav_entry(u_int32_t    *ud_av_p, 
                         VAPI_ud_av_t *av_p)
{
  u_int32_t i;
  u_int32_t tmp_av_arr[8] = {0,0,0,0,0,0,0,0}; /* entry size is 32 bytes */
  HH_pd_hndl_t pd_tmp = 0;
  
  FUNC_IN;
  if (ud_av_p == NULL || av_p == NULL) {
    MT_COV_ERR_FLTR_START
    MTL_ERROR1(MT_FLFMT("%s: ud_av_p=%p, av_p=%p"),__func__, ud_av_p, av_p);
	return HH_ERR;
    MT_COV_ERR_FLTR_END
  }
  /* read entry to tmp area */
  /* the gid should stay in BE format so we don't change its endianess */
  for (i=0; i<RGID_OFFSET; i++) {
    tmp_av_arr[i] = MOSAL_be32_to_cpu(MOSAL_MMAP_IO_READ_DWORD(ud_av_p+i)); 
  }

  /* PD */
  pd_tmp = MT_EXTRACT_ARRAY32(tmp_av_arr, MT_BIT_OFFSET(PRM_UDAV_ST, pd),
                           MT_BIT_SIZE(PRM_UDAV_ST, pd));

  /* if PD is zero it means that the entry is not valid */
  if (pd_tmp == 0) {
    MT_COV_ERR_FLTR_START
    MT_RETURN(HH_EINVAL);
	MT_COV_ERR_FLTR_END
  }

  /* port */
  av_p->port = MT_EXTRACT_ARRAY32(tmp_av_arr, MT_BIT_OFFSET(PRM_UDAV_ST, port_number),
                  MT_BIT_SIZE(PRM_UDAV_ST, port_number));
  
  /* rlid */
  av_p->dlid = MT_EXTRACT_ARRAY32(tmp_av_arr, MT_BIT_OFFSET(PRM_UDAV_ST, rlid),
                 MT_BIT_SIZE(PRM_UDAV_ST, rlid));

  /* mylid_path_bits */
  av_p->src_path_bits = MT_EXTRACT_ARRAY32(tmp_av_arr, MT_BIT_OFFSET(PRM_UDAV_ST, my_lid_path_bits),
                 MT_BIT_SIZE(PRM_UDAV_ST, my_lid_path_bits));
  /* grh enable */
  av_p->grh_flag = MT_EXTRACT_ARRAY32(tmp_av_arr, MT_BIT_OFFSET(PRM_UDAV_ST, g),
                 MT_BIT_SIZE(PRM_UDAV_ST, g));
  /* hop_limit */
  av_p->hop_limit = MT_EXTRACT_ARRAY32(tmp_av_arr, MT_BIT_OFFSET(PRM_UDAV_ST, hop_limit),
                 MT_BIT_SIZE(PRM_UDAV_ST, hop_limit));

  /* max_stat_rate */
  /* 0 - stay 0, 1 transmitted to 3 as defined in IPD encoding: IB-spec. 9.11.1, table 63 */
  i = MT_EXTRACT_ARRAY32(tmp_av_arr, MT_BIT_OFFSET(PRM_UDAV_ST, max_stat_rate),
                 MT_BIT_SIZE(PRM_UDAV_ST, max_stat_rate));
  av_p->static_rate = (i ? 3:0);
  
  /* msg size - not needed for the av info */
    
  /* mgid_index (index to port GID table)*/
  av_p->sgid_index = MT_EXTRACT_ARRAY32(tmp_av_arr, MT_BIT_OFFSET(PRM_UDAV_ST, mgid_index),
                 MT_BIT_SIZE(PRM_UDAV_ST, mgid_index)) & MASK32(LOG2_PORT_GID_TABLE_SIZE);
  /* TBD: sanity check that 6th bit of mgid_index matches port field (-1) */
  
  /* flow_label */
  av_p->flow_label = MT_EXTRACT_ARRAY32(tmp_av_arr, MT_BIT_OFFSET(PRM_UDAV_ST, flow_label),
                 MT_BIT_SIZE(PRM_UDAV_ST, flow_label));
  
  /* tclass */
  av_p->traffic_class = MT_EXTRACT_ARRAY32(tmp_av_arr, MT_BIT_OFFSET(PRM_UDAV_ST, tclass),
                 MT_BIT_SIZE(PRM_UDAV_ST, tclass));
  
  /* sl */
  av_p->sl = MT_EXTRACT_ARRAY32(tmp_av_arr, MT_BIT_OFFSET(PRM_UDAV_ST, sl),
                 MT_BIT_SIZE(PRM_UDAV_ST, sl));
  
  /* gid stays in BE so it is extracted directly from the ud_av_p */
  /* rgid_127_96 */
  memcpy(av_p->dgid,ud_av_p+(MT_BYTE_OFFSET(PRM_UDAV_ST, rgid_127_96) >> 2),
         sizeof(av_p->dgid));

  
#if 1 <= MAX_DEBUG
  print_udav(av_p);
#endif  
  MT_RETURN(HH_OK);

  
}  

