/*
  This software is available to you under a choice of one of two
  licenses.  You may choose to be licensed under the terms of the GNU
  General Public License (GPL) Version 2, available at
  <http://www.fsf.org/copyleft/gpl.html>, or the OpenIB.org BSD
  license, available in the LICENSE.TXT file accompanying this
  software.  These details are also available at
  <http://openib.org/license.html>.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
  BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
  ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
  CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
  SOFTWARE.

  Copyright (c) 2004 Mellanox Technologies Ltd.  All rights reserved.
*/

#include <dev_common.h>

/*================ macro definitions ===============================================*/
#define PRM_UDAV_ST ADD_DEV_PRM_PREFIX(ud_address_vector_st)
#define UD_AV_ENTRY_SIZE (sizeof(struct PRM_UDAV_ST) / 8)
#define UD_AV_ENTRY_DWORD_SIZE (sizeof(struct PRM_UDAV_ST) / 32)
#define UD_AV_ENTRY_SIZE_LOG2 5
#define UD_AV_ALIGNMENT_MASK (UD_AV_ENTRY_SIZE - 1)
#define RGID_OFFSET (MT_BIT_OFFSET(PRM_UDAV_ST, rgid_127_96) >> 5) /* in DWORDS */
#define RGID_OFFSET_BYTES (MT_BIT_OFFSET(PRM_UDAV_ST, rgid_127_96) >> 3) /* in BYTES */
#define RGID_OFFSET_END (MT_BIT_OFFSET(PRM_UDAV_ST, rgid_31_0) >> 5) /* in DWORDS */
#define RGID_OFFSET_END_BYTES (MT_BIT_OFFSET(PRM_UDAV_ST, rgid_31_0) >> 3) /* in BYTES */
#define LOG2_PORT_GID_TABLE_SIZE 5 /* 32 GID entries per port */

#define XHH_UDAVM_DLID_0(av_p)                                  \
  if (av_p->dlid == 0) {                                        \
    MTL_ERROR4(MT_FLFMT("%s: invalid dlid (ZERO)."),__func__);  \
    MT_RETURN(HH_EINVAL);                                       \
  }

#define XHH_UDAVM_BASIC_VALID_CHECKS(udavm, av_p)          \
  if (udavm == NULL) {                                     \
    MTL_ERROR4(MT_FLFMT("%s: udavm is NULL."),__func__);   \
    MT_RETURN(HH_EINVAL);                                  \
  }                                                        \
  if (av_p == NULL) {                                      \
    MTL_ERROR4(MT_FLFMT("%s: av_p is NULL."),__func__);    \
    MT_RETURN(HH_EINVAL);                                  \
  }                                                        
  
#define XHH_UDAVM_VALID_CHECKS(udavm, av_p)     \
  XHH_UDAVM_BASIC_VALID_CHECKS(udavm, av_p)     \
  XHH_UDAVM_DLID_0(av_p)
               
#define XHH_UDAVM_AH_NULL(ah_p)                            \
  if (ah_p == NULL) {                                      \
    MTL_ERROR4(MT_FLFMT("%s:  ah_p is NULL."),__func__);   \
    MT_RETURN(HH_EINVAL);                                  \
  }                                                      

void XHH_udavm_fill_udav_entry(/*IN */ HH_pd_hndl_t pd, 
                     /*IN */ MT_bool      is_new_pd,
                     /*IN */ VAPI_ud_av_t *av_p, 
                     /*IN */ u_int32_t    *av_entry_p,
		     /*IN */ MT_bool      make_big_endian,
		     /*IN */ MT_bool      use_lam);

HH_ret_t XHH_udavm_parse_udav_entry(u_int32_t    *ud_av_p, VAPI_ud_av_t *av_p);

#if 1 <= MAX_DEBUG
void print_udav(VAPI_ud_av_t *av_p);
#endif    
