/*
  This software is available to you under a choice of one of two
  licenses.  You may choose to be licensed under the terms of the GNU
  General Public License (GPL) Version 2, available at
  <http://www.fsf.org/copyleft/gpl.html>, or the OpenIB.org BSD
  license, available in the LICENSE.TXT file accompanying this
  software.  These details are also available at
  <http://openib.org/license.html>.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
  BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
  ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
  CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
  SOFTWARE.

  Copyright (c) 2004 Mellanox Technologies Ltd.  All rights reserved.
*/

#ifndef H_UL_CQM_PRIV_H
#define H_UL_CQM_PRIV_H

#include <mtperf.h>
MTPERF_NEW_SEGMENT(free_cqe,5000);

/* Uncomment the line below in order to get CQ dump when the same WQE is used twice simultaneously*/
/* #define XHHUL_CQM_DEBUG_WQE_REUSE */

/* Always support fork (assumes number of CQs per process is limited) */
#define MT_FORK_SUPPORT


/* Limit kmalloc to 2 pages (if this fails, vmalloc will fail too) */
#define CQ_KMALLOC_LIMIT (2*MOSAL_SYS_PAGE_SIZE)

/* Maximum CQ doorbell to coalesce/delay (too much is not effective ?) */
#define MAX_CQDB2DELAY 255

/* CQE size */
#define LOG2_CQE_SZ 5   /* 32 bytes */
#define CQE_SZ (1U<<LOG2_CQE_SZ)

/* Assuming that in original CQE (big-endian) the owner field is in the last byte (bit 7) */
#define CQE_OWNER_SHIFT  7          /* Bit of ownership (in the byte) */
#define CQE_OWNER_BYTE_OFFSET 0x1F  /* Last byte of a CQE */
#define CQE_OPCODE_BYTE_OFFSET 0x1C  /* Opcode/status byte */
#define CQE_WQE_ADR_BIT_SZ  6       /* Number of bits in "wqe_adr" field */

#define CQE_IS_SQ_SHIFT  7          /* Bit of ownership (in the byte) */
#define CQE_IS_SQ_BYTE_OFFSET 0x1D  /* in last dword of CQE */

/* new CQE error handling */
#define CQE_ERROR_ON_SQ  0xFF       /* Opcode field value for SQ completion error */
#define CQE_ERROR_ON_RQ  0xFE       /* Opcode field value for RQ completion error */
#define CQE_ERROR_STATUS_MASK 0xFE  /* Common mask for both 0xFF and 0xFE "error status" opcodes */
#define CQE_ERROR_SYNDROM_BIT_OFFSET (MT_BIT_OFFSET(ADD_DEV_PRM_PREFIX(completion_queue_entry_st),\
                                                immediate_ethertype_pkey_indx_eecredits)+24)
#define CQE_ERROR_SYNDROM_BIT_SIZE 8 
#define CQE_ERROR_VENDOR_SYNDROM_BIT_OFFSET (CQE_ERROR_SYNDROM_BIT_OFFSET-8)
#define CQE_ERROR_VENDOR_SYNDROM_BIT_SIZE 8
#define CQE_ERROR_DBDCNT_BIT_OFFSET (MT_BIT_OFFSET(ADD_DEV_PRM_PREFIX(completion_queue_entry_st),\
                                                immediate_ethertype_pkey_indx_eecredits))
#define CQE_ERROR_DBDCNT_BIT_SIZE 16 


#define MAX_CQ_NUM 0x00FFFFFF
#define INVALID_CQ_NUM 0xFFFFFFFF
#define MAX_NCOMP_NOTIF 0x7FFF    /* req_ncomp_notif support up to 2^15-1 */

/* Parameters for the per-CQ QP lookup table */
#define LOG2_QP_HASH_TBL_SZ 8
#define QP_HASH_TBL_SZ  (1<<LOG2_QP_HASH_TBL_SZ)

/* IB opcodes for RQ CQEs (bits 4:0) */
typedef enum {
  IB_OP_SEND_LAST       =2,
  IB_OP_SEND_ONLY       =4,
  IB_OP_SEND_IMM_LAST   =3,
  IB_OP_SEND_IMM_ONLY   =5,
  IB_OP_RDMAW_IMM_LAST  =9,
  IB_OP_RDMAW_IMM_ONLY  =11
} XHHUL_ib_opcode_t;


typedef enum {
    XHHUL_CQ_PREP,        /* Before XHHUL_cqm_create_cq_done */
    XHHUL_CQ_IDLE,        /* No resize is in progress */
    XHHUL_CQ_RESIZE_PREP, /* Allocation and activation of resized buffer is in progress
                             (before return from HH_resize_cq) */
    XHHUL_CQ_RESIZE_DONE  /* In fixed resize-CQ flow, the transition to the resized buffer may 
                           * happen before the return from VIPKL_resize_cq 
                           * (before XHHUL_cqm_resize_cq_done)                                */
} XHHUL_cq_state_t;

typedef struct {
		VAPI_completion_event_handler_t event_handler;
		void *private_data;
}XHH_event_data_t;

typedef struct {
  void* cqe_buf_orig;         /* Address returned from VMALLOC */
  MT_virt_addr_t cqe_buf_base;   /* Aligned to CQE size */
  u_int8_t log2_num_o_cqes;   /* (log2) number of CQEs in the buffer (including one reserved) */
  u_int16_t spare_cqes;       /* CQEs beyond requested number of CQEs for this buffer */
  u_int32_t consumer_counter;
} XHHUL_cqe_buf_t;

typedef struct qp_hash_entry_st {  /* QPN-to-QP hash table entry */
  IB_wqpn_t qpn;
  XHHUL_qp_t qp;
  struct qp_hash_entry_st *next;  /* next in this hash bin */
  struct qp_hash_entry_st *prev;  /* previous in this hash bin */
} qp_hash_entry_t;

typedef struct XHHUL_cq_st {
  XHHUL_cqe_buf_t cur_buf;/* Main CQ buffer properties */
  XHHUL_cqe_buf_t resized_buf;/* Next (resized) CQ buffer properties */
  u_int16_t pending_cq_dbell; /* Pending CQ doorbells 
                               * (DBs to rung when reach spare_cqes or on event req.) */
  u_int16_t cur_spare_cqes;   /* CQEs beyond requested number of CQEs (for CQ DB coalescing) */
  XHHUL_qpm_t qpm;            /* Make this available close to the CQ context */
  qp_hash_entry_t *qpn2qp_hash_tbl[QP_HASH_TBL_SZ]; 
  XHHUL_cq_state_t cq_state;
  XHH_DEV_SPECIFIC_UL_CQ_T_FIELDS
  HH_cq_hndl_t  cq_num;       /* Hardware CQ number */
  XHH_event_data_t event_data;
  MOSAL_spinlock_t cq_lock;
  /* CQ list administration */
  struct XHHUL_cq_st *next;
} XHHUL_cq_t;

struct XHHUL_cqm_st {
  XHHUL_cq_t *cq_list;
  MOSAL_mutex_t cqm_lock; /* This protects list - not time critical (only for adding/removing) */
} /* *XHHUL_cqm_t */;

struct XHHUL_cqe_st {
  IB_wqpn_t qpn;
  IB_eecn_t eecn;
  IB_wqpn_t rqpn;
  IB_sl_t   sl;
  MT_bool grh_present;
  u_int8_t  ml_path;
  IB_lid_t  rlid;

} XHHUL_cqe_t;


#ifdef XHHUL_CQM_DEBUG_WQE_REUSE
#define DUMP_CQE_PRINT_CMD MTL_ERROR4
#else
#define DUMP_CQE_PRINT_CMD MTL_DEBUG4
#endif

#define DUMP_CQE(cq_num,cqe_index,cqe) \
    {DUMP_CQE_PRINT_CMD("CQ[0x%X]:cqe[0x%X] = %08X %08X %08X %08X %08X %08X %08X %08X\n", \
      cq_num,cqe_index,                                                      \
      MOSAL_be32_to_cpu(cqe[0]),                                             \
      MOSAL_be32_to_cpu(cqe[1]),                                             \
      MOSAL_be32_to_cpu(cqe[2]),                                             \
      MOSAL_be32_to_cpu(cqe[3]),                                             \
      MOSAL_be32_to_cpu(cqe[4]),                                             \
      MOSAL_be32_to_cpu(cqe[5]),                                             \
      MOSAL_be32_to_cpu(cqe[6]),                                             \
      MOSAL_be32_to_cpu(cqe[7]));}

#endif
