/*
  This software is available to you under a choice of one of two
  licenses.  You may choose to be licensed under the terms of the GNU
  General Public License (GPL) Version 2, available at
  <http://www.fsf.org/copyleft/gpl.html>, or the OpenIB.org BSD
  license, available in the LICENSE.TXT file accompanying this
  software.  These details are also available at
  <http://openib.org/license.html>.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
  BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
  ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
  CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
  SOFTWARE.

  Copyright (c) 2004 Mellanox Technologies Ltd.  All rights reserved.
*/

#define C_XHHUL_QPM_C

#include <mosal.h>
#include <ib_defs.h>
#include <mpga/nMPGA.h>
#include <vapi.h>
#include <log2.h>
#include <ul.h>
#include <ul_hob.h>
#include <ul_cqm.h>
#include <ul_srqm.h>
#include <ul_pdm.h>
#include <ul_qpm.h>
#include <ul_qpm_priv.h>

/* include device-specific inline functions */ 
#include <ul_qpm_dev_inline.h>

#define CUR_OUTS(qres)  ( ((qres).next2post_index - (qres).next2free_index  + (qres).max_outs) % (qres).max_outs )

/**********************************************************************************************
 *                    Private functions protoypes declarations
 **********************************************************************************************/

static HH_ret_t qp_prep(
  HHUL_hca_hndl_t hca, 
  VAPI_special_qp_t qp_type, 
  HHUL_qp_init_attr_t *qp_init_attr_p, 
  HHUL_qp_hndl_t *qp_hndl_p, 
  VAPI_qp_cap_t *qp_cap_out_p, 
  XHH_qp_ul_resources_t *qp_ul_resources_p, 
  MT_bool in_ddr_mem  /* WQEs buffer allocated in attached DDR mem. or in main memory */
);

static HH_ret_t init_qp(
  HHUL_hca_hndl_t hca, 
  HHUL_qp_init_attr_t *qp_init_attr_p, 
  XHHUL_qp_t new_qp,
  XHH_qp_ul_resources_t *qp_ul_resources_p
);

static HH_ret_t alloc_wqe_buf(
  /*IN*/ XHHUL_qpm_t qpm,
  /*IN*/ MT_bool in_ddr_mem,   /* Allocation of WQEs buffer is requested in attached DDR mem. */
  /*IN*/ u_int32_t max_outs_wqes, /* HCA cap. */
  /*IN*/ u_int32_t max_sg_ent, /* HCA cap. of max.s/g entries */
  /*IN*/ u_int32_t max_wqe_size_sq,    /* from hca_ul_resources */
  /*IN*/ u_int32_t max_wqe_size_rq,    /* from hca_ul_resources */
  /*IN*/ VAPI_lkey_t qp_mem_lkey,
  /*IN/OUT*/ XHHUL_qp_t new_qp,
  /*OUT*/    XHH_qp_ul_resources_t *qp_ul_resources_p
);

static HH_ret_t alloc_aux_data_buf(
  /*IN/OUT*/ XHHUL_qp_t new_qp
);


#ifndef MT_KERNEL
static void* dpool_alloc(XHHUL_qpm_t qpm, u_int8_t buf_size_kb, XHHUL_qpm_dpool_t **dpool_pp);

static void dpool_free(XHHUL_qpm_t qpm, XHHUL_qpm_dpool_t *dpool_p, void* buf);

#else
#define dpool_free(qpm,dpool_p,buf) \
  MTL_ERROR1(MT_FLFMT("%s: Invoked dpool_free in kernel by mistake"), __func__)
   
#endif

/* QPM's QPs list modification functions */
static HH_ret_t insert_to_list(XHHUL_qpm_t qpm, XHHUL_qp_t qp);
static HH_ret_t remove_from_list(XHHUL_qpm_t qpm, XHHUL_qp_t qp);


/**********************************************************************************************
 *                    Private inline functions 
 **********************************************************************************************/

inline static u_int32_t get_wqe_index(
  /*IN*/ queue_res_t *q_res_p, 
  /*IN*/ u_int32_t wqe_addr_32lsb,
  /*OUT*/ u_int32_t *wqe_index_p
)
{
  u_int32_t wqe_buf_base_32lsb;

  /* TBD: On QP resize this will have to be modified (buffers may change during QP life cycle) */
  
  wqe_buf_base_32lsb= (u_int32_t)(q_res_p->wqe_buf);
  if (wqe_addr_32lsb >= wqe_buf_base_32lsb) { /* Assure index computation is positive */
    *wqe_index_p= (wqe_addr_32lsb - wqe_buf_base_32lsb) >> q_res_p->log2_max_wqe_sz;
    if (*wqe_index_p < q_res_p->max_outs)  { /* WQE is within this queue */
      /* TBD: check if given wqe_addr_32lsb is aligned to WQE size */
      return HH_OK;
    }
  } 
  
  return HH_EINVAL; /* WQE is not withing this queue */
}


/* Find the queue context and WQE index from the QP number and WQE address */
inline static HH_ret_t find_wqe(
  /*IN*/ XHHUL_qpm_t qpm, 
  /*IN*/ XHHUL_qp_t qp, 
  /*IN*/ u_int32_t wqe_addr_32lsb,
  /*IN*/ MT_bool is_SQ_completion,
  /*OUT*/ queue_res_t **q_res_pp,
  /*OUT*/ u_int32_t *wqe_index_p,
  /*OUT*/ VAPI_wr_id_t *wqe_id_p
)
{
  HH_ret_t rc;

/* check if this WQE is of SQ */
  if (is_SQ_completion == TRUE) {
      *q_res_pp= &(qp->sq_res);
      rc= get_wqe_index(*q_res_pp,wqe_addr_32lsb,wqe_index_p);
      if (rc == HH_OK)  {
	*wqe_id_p= (*q_res_pp)->wqe_id[*wqe_index_p]; 
	MTL_DEBUG2(MT_FLFMT("%s: WQE is in  SQ (WQE=0x%X QPn=0x%X)"), __func__,
		   wqe_addr_32lsb, qp->qpn);
	return HH_OK;
      } else {
	  MTL_DEBUG2(MT_FLFMT("%s: WQE flagged as SQ completion, but is not in SQ range (WQE=0x%X QPn=0x%X, rc=(%d:%s)"),
		      __func__,wqe_addr_32lsb, qp->qpn, rc, HH_strerror_sym(rc));
	  return rc;
      }
  }
  
  /* WQE is either of RQ or SRQ */
  /* check if this WQE is of RQ */
  if (qp->srq == HHUL_INVAL_SRQ_HNDL) {
    *q_res_pp= &(qp->rq_res);
    rc= get_wqe_index(*q_res_pp,wqe_addr_32lsb,wqe_index_p);
    if (rc == HH_OK)  {
      *wqe_id_p= (*q_res_pp)->wqe_id[*wqe_index_p]; 
      MTL_DEBUG2(MT_FLFMT("%s: WQE is in  RQ (WQE=0x%X QPn=0x%X)"), __func__,
		 wqe_addr_32lsb, qp->qpn);
      return HH_OK;
    }
  } else { /* From SRQ ? */
    *q_res_pp= NULL;
    return HH_OK;
  }
  
  return rc; /* Invalid WQE address for this QP */
}

inline static MT_bool is_qpstate_valid_2send(VAPI_qp_state_t cur_state)
{
    switch (cur_state) {
    case VAPI_RTS:
    case VAPI_SQD:
    case VAPI_ERR:
    case VAPI_SQE:  return TRUE;
                    break;
    default:        return FALSE;
    }

}
inline static MT_bool is_qpstate_valid_2recv(VAPI_qp_state_t cur_state)
{
    switch (cur_state) {
    case VAPI_INIT: 
    case VAPI_RTR:
    case VAPI_RTS:
    case VAPI_SQD:
    case VAPI_ERR:
    case VAPI_SQE:  return TRUE;
                    break;
    default:        return FALSE;
    }

}

inline static ADD_DEV_PREFIX_LC(if_nopcode_t) encode_nopcode(VAPI_wr_opcode_t opcode)
{
  switch (opcode) {
    case VAPI_RDMA_WRITE:
      return ADD_DEV_PREFIX(IF_NOPCODE_RDMAW);
    case VAPI_RDMA_WRITE_WITH_IMM:
      return ADD_DEV_PREFIX(IF_NOPCODE_RDMAW_IMM);
    case VAPI_SEND:
      return  ADD_DEV_PREFIX(IF_NOPCODE_SEND);
    case VAPI_SEND_WITH_IMM:
      return ADD_DEV_PREFIX(IF_NOPCODE_SEND_IMM);
    case VAPI_RDMA_READ:
      return ADD_DEV_PREFIX(IF_NOPCODE_RDMAR);
    case VAPI_ATOMIC_CMP_AND_SWP:
      return ADD_DEV_PREFIX(IF_NOPCODE_ATOM_CMPSWP);
    case VAPI_ATOMIC_FETCH_AND_ADD:
      return ADD_DEV_PREFIX(IF_NOPCODE_ATOM_FTCHADD);
    default:
      return ADD_DEV_PREFIX(IF_NOPCODE_NOP);
  }
}

/*********** WQE building functions ***********/


inline static u_int32_t WQE_pack_send_next(u_int32_t *segment_p, 
  ADD_DEV_PREFIX_LC(if_nopcode_t) nopcode, MT_bool fence, u_int32_t dbd,
  u_int32_t next_wqe_32lsb, u_int32_t wqe_sz_16B_chunks,
  IB_eecn_t eecn)
{
  memset(segment_p,0,WQE_SEG_SZ_NEXT);  /* Clear all "RESERVED" */
  segment_p[MT_BYTE_OFFSET(WQE_SEGMENT_NEXT_ST,nda_31_6)>>2]= next_wqe_32lsb & (~MASK32(6));
  MT_INSERT_ARRAY32(segment_p,nopcode,
    MT_BIT_OFFSET(WQE_SEGMENT_NEXT_ST,nopcode),MT_BIT_SIZE(WQE_SEGMENT_NEXT_ST,nopcode));
  MT_INSERT_ARRAY32(segment_p,fence ? 1 : 0,
    MT_BIT_OFFSET(WQE_SEGMENT_NEXT_ST,f),MT_BIT_SIZE(WQE_SEGMENT_NEXT_ST,f));

  XHH_DEV_SPECIFIC_WQE_INIT_NEXT

  MT_INSERT_ARRAY32(segment_p,wqe_sz_16B_chunks,
   MT_BIT_OFFSET(WQE_SEGMENT_NEXT_ST,nds),MT_BIT_SIZE(WQE_SEGMENT_NEXT_ST,nds));
  MT_INSERT_ARRAY32(segment_p,eecn,
    MT_BIT_OFFSET(WQE_SEGMENT_NEXT_ST,nee),MT_BIT_SIZE(WQE_SEGMENT_NEXT_ST,nee));
  return WQE_SEG_SZ_NEXT;
}

/* Pack Control segment (for sends) */
inline static u_int32_t WQE_pack_ctrl_send(u_int32_t *segment_p,  
    VAPI_comp_type_t comp_type, MT_bool se_bit, u_int32_t event_bit,
    u_int32_t imm_data)
{
  memset(segment_p,0,WQE_SEG_SZ_CTRL);  /* Clear all "RESERVED" */
  MT_INSERT_ARRAY32(segment_p,1,
    MT_BIT_OFFSET(WQE_SEGMENT_CTRL_SEND_ST,always1),MT_BIT_SIZE(WQE_SEGMENT_CTRL_SEND_ST,always1));
  MT_INSERT_ARRAY32(segment_p,(comp_type == VAPI_SIGNALED) ? 1 : 0,
    MT_BIT_OFFSET(WQE_SEGMENT_CTRL_SEND_ST,c),MT_BIT_SIZE(WQE_SEGMENT_CTRL_SEND_ST,c));
  MT_INSERT_ARRAY32(segment_p,se_bit ? 1 : 0,
    MT_BIT_OFFSET(WQE_SEGMENT_CTRL_SEND_ST,s),MT_BIT_SIZE(WQE_SEGMENT_CTRL_SEND_ST,s));
  MT_INSERT_ARRAY32(segment_p,event_bit,
    MT_BIT_OFFSET(WQE_SEGMENT_CTRL_SEND_ST,e),MT_BIT_SIZE(WQE_SEGMENT_CTRL_SEND_ST,e));
  segment_p[MT_BYTE_OFFSET(WQE_SEGMENT_CTRL_SEND_ST,immediate)>>2]= imm_data;
  return WQE_SEG_SZ_CTRL;
}

inline static u_int32_t WQE_pack_ud(u_int32_t *segment_p,
  XHHUL_qp_t  qp,
  VAPI_ud_av_hndl_t ah, 
  IB_wqpn_t destination_qp, IB_qkey_t q_key)
{
  memset(segment_p,0,WQE_SEG_SZ_UD);  /* Clear all "RESERVED" */
  
  XHH_DEV_SPECIFIC_WQE_PACK_UD

  MT_INSERT_ARRAY32(segment_p,destination_qp,
    MT_BIT_OFFSET(WQE_SEGMENT_UD_ST,destination_qp),
    MT_BIT_SIZE(WQE_SEGMENT_UD_ST,destination_qp));
  segment_p[MT_BYTE_OFFSET(WQE_SEGMENT_UD_ST,q_key)>>2]= q_key;
  return WQE_SEG_SZ_UD;
}

inline static u_int32_t WQE_pack_rd(u_int32_t *segment_p,
  IB_wqpn_t destination_qp, IB_qkey_t q_key)
{
  memset(segment_p,0,WQE_SEG_SZ_RD);  /* Clear all "RESERVED" */
  MT_INSERT_ARRAY32(segment_p,destination_qp,
    MT_BIT_OFFSET(WQE_SEGMENT_RD_ST,destination_qp),
    MT_BIT_SIZE(WQE_SEGMENT_RD_ST,destination_qp));
  segment_p[MT_BYTE_OFFSET(WQE_SEGMENT_RD_ST,q_key)>>2]= q_key;
  return WQE_SEG_SZ_RD;
}

inline static u_int32_t WQE_pack_remote_addr(u_int32_t *segment_p,
  IB_virt_addr_t remote_addr, IB_rkey_t remote_rkey)
{
  memset(segment_p,0,WQE_SEG_SZ_RADDR);  /* Clear all "RESERVED" */
  segment_p[MT_BYTE_OFFSET(WQE_SEGMENT_REMOTE_ADDRESS_ST,remote_virt_addr_h)>>2]= 
    (u_int32_t)(remote_addr >> 32);
  segment_p[MT_BYTE_OFFSET(WQE_SEGMENT_REMOTE_ADDRESS_ST,remote_virt_addr_l)>>2]= 
    (u_int32_t)(remote_addr & 0xFFFFFFFF);
  segment_p[MT_BYTE_OFFSET(WQE_SEGMENT_REMOTE_ADDRESS_ST,rkey)>>2]= remote_rkey;
  return WQE_SEG_SZ_RADDR;
}

inline static u_int32_t WQE_pack_recv_next(u_int32_t *segment_p, 
  u_int32_t next_wqe_32lsb, u_int32_t wqe_sz_16B_chunks)
{
  memset(segment_p,0,WQE_SEG_SZ_NEXT);  /* Clear all "RESERVED" */

  XHH_DEV_SPECIFIC_PACK_RECV_NEXT

  MT_INSERT_ARRAY32(segment_p,wqe_sz_16B_chunks,
    MT_BIT_OFFSET(WQE_SEGMENT_NEXT_ST,nds),MT_BIT_SIZE(WQE_SEGMENT_NEXT_ST,nds));
  return WQE_SEG_SZ_NEXT;
}

inline static u_int32_t WQE_pack_atomic_cmpswp(u_int32_t *segment_p,
  u_int64_t cmp_data, u_int64_t swap_data)
{
  segment_p[MT_BYTE_OFFSET(WQE_SEGMENT_ATOMIC_ST,swap_add_h)>>2]= (u_int32_t)(swap_data >> 32);
  segment_p[MT_BYTE_OFFSET(WQE_SEGMENT_ATOMIC_ST,swap_add_l)>>2]= (u_int32_t)(swap_data & 0xFFFFFFFF);
  segment_p[MT_BYTE_OFFSET(WQE_SEGMENT_ATOMIC_ST,compare_h)>>2]= (u_int32_t)(cmp_data >> 32);
  segment_p[MT_BYTE_OFFSET(WQE_SEGMENT_ATOMIC_ST,compare_l)>>2]= (u_int32_t)(cmp_data & 0xFFFFFFFF);
  return WQE_SEG_SZ_ATOMIC;
}

inline static u_int32_t WQE_pack_atomic_fetchadd(u_int32_t *segment_p,u_int64_t add_data)
{
  segment_p[MT_BYTE_OFFSET(WQE_SEGMENT_ATOMIC_ST,swap_add_h)>>2]= (u_int32_t)(add_data >> 32);
  segment_p[MT_BYTE_OFFSET(WQE_SEGMENT_ATOMIC_ST,swap_add_l)>>2]= (u_int32_t)(add_data & 0xFFFFFFFF);
  return WQE_SEG_SZ_ATOMIC;
}

/* Build the scatter/gather list (pointer segments) */
inline static u_int32_t WQE_pack_sg_list(u_int32_t *segment_p,
  u_int32_t sg_lst_len,VAPI_sg_lst_entry_t *sg_lst_p)
{
   u_int32_t i;
   u_int32_t *cur_loc_p= segment_p;

   for (i= 0; i < sg_lst_len; i++ , cur_loc_p+= WQE_SEG_SZ_SG_ENTRY_DW) {
     cur_loc_p[MT_BYTE_OFFSET(WQE_SEGMENT_DATA_PTR_ST,byte_count)>>2]= 
       (sg_lst_p[i].len & MASK32(31));
     cur_loc_p[MT_BYTE_OFFSET(WQE_SEGMENT_DATA_PTR_ST,l_key)>>2]= sg_lst_p[i].lkey;
     cur_loc_p[MT_BYTE_OFFSET(WQE_SEGMENT_DATA_PTR_ST,local_address_h)>>2]= 
       (u_int32_t)(sg_lst_p[i].addr >> 32);
     cur_loc_p[MT_BYTE_OFFSET(WQE_SEGMENT_DATA_PTR_ST,local_address_l)>>2]= 
       (u_int32_t)(sg_lst_p[i].addr & 0xFFFFFFFF);
   }
   return (u_int32_t)(((MT_virt_addr_t)cur_loc_p) - ((MT_virt_addr_t)segment_p));
}


/* Build the WQE in given wqe_buf.
 * Return WQE size.
 */
inline static u_int32_t WQE_build_send(
  XHHUL_qp_t qp,
  VAPI_sr_desc_t *send_req_p,
  volatile u_int32_t *wqe_buf)
{
  u_int8_t *cur_loc_p= (u_int8_t*)wqe_buf; /* Current location in the WQE */

  cur_loc_p+= WQE_init_next((u_int32_t*)cur_loc_p); /* Make "unlinked" "next" segment */
  cur_loc_p+= WQE_pack_ctrl_send((u_int32_t*)cur_loc_p,  /* Pack Control segment */
    send_req_p->comp_type, send_req_p->set_se, 0/*event bit*/,
    ((send_req_p->opcode == VAPI_RDMA_WRITE_WITH_IMM) ||
     (send_req_p->opcode == VAPI_SEND_WITH_IMM) ) ? send_req_p->imm_data : 0);

  /* Transport type checks: Datagram segment */
  switch (qp->ts_type) {
    case VAPI_TS_UD:  /* Check if UD (UD datagram segment) */
      cur_loc_p += WQE_pack_ud((u_int32_t*)cur_loc_p,
        qp,send_req_p->remote_ah,
        send_req_p->remote_qp,send_req_p->remote_qkey);
      break;
    case VAPI_TS_RD:  /* Check if RD (RD datagram segment) */
      cur_loc_p+= WQE_pack_rd((u_int32_t*)cur_loc_p,
        send_req_p->remote_qp,send_req_p->remote_qkey);
      break;
    default:
      break;
  }
  
  /* Opcode checks Remote-address/Atomic segments */
  switch (send_req_p->opcode) {
    /* For RDMA operations: only Remote-address segment */
    case VAPI_RDMA_READ:
    case VAPI_RDMA_WRITE:
    case VAPI_RDMA_WRITE_WITH_IMM:
     cur_loc_p+= WQE_pack_remote_addr((u_int32_t*)cur_loc_p,
       send_req_p->remote_addr,send_req_p->r_key);
     break;
     
    /* Check if Atomic operations (both remote-address and Atomic segments) */
    case VAPI_ATOMIC_CMP_AND_SWP:
      cur_loc_p+= WQE_pack_remote_addr((u_int32_t*)cur_loc_p,send_req_p->remote_addr,
        send_req_p->r_key);
      cur_loc_p+= WQE_pack_atomic_cmpswp((u_int32_t*)cur_loc_p,send_req_p->compare_add,
        send_req_p->swap);
      break;
    case VAPI_ATOMIC_FETCH_AND_ADD:
     cur_loc_p+= WQE_pack_remote_addr((u_int32_t*)cur_loc_p,send_req_p->remote_addr,
       send_req_p->r_key);
     cur_loc_p+= WQE_pack_atomic_fetchadd((u_int32_t*)cur_loc_p,send_req_p->compare_add);
     break;
    default: /*NOP*/
      break;
  }
  
  /* Pack scatter/gather list segments */
  cur_loc_p+= WQE_pack_sg_list((u_int32_t*)cur_loc_p,send_req_p->sg_lst_len,send_req_p->sg_lst_p);
  return (u_int32_t)(((MT_virt_addr_t)cur_loc_p) - ((MT_virt_addr_t)wqe_buf));
}
  
/* Build UD header as inline data for management QPs over MLX "transport" */
inline static u_int32_t WQE_pack_mlx_ud_header(u_int32_t *segment_p,
  XHHUL_qp_t qp, VAPI_sr_desc_t *send_req_p, VAPI_ud_av_t *av_p, HH_hca_hndl_t hh_hndl,
  VAPI_pkey_ix_t pkey_index /* take this index instead of QP's, if not QP1_PKEY_INDEX */)
{
  MPGA_headers_t *hdrs_p = NULL;
  IB_LRH_st *LRH_p;
  IB_BTH_st *BTH_p;
  IB_DETH_st *DETH_p;
  u_int8_t *hdrs_buf_p;
#ifdef MT_LITTLE_ENDIAN
  u_int32_t *hdrs_buf32_p;  /* pointer for endiness swapping */
  u_int16_t i;
#endif
  u_int16_t hdrs_sz;
  MT_bool global= av_p->grh_flag;
#ifdef MT_KERNEL
  IB_port_t num_ports;
  IB_port_t port= (qp->qpn & 0xf);  /* QPN of QP used for port 1 has the even index */
  IB_pkey_t cur_pkey= 0;
  HH_ret_t rc;
  
  rc= XHH_hob_get_num_ports(hh_hndl,&num_ports);
  if (rc != HH_OK) {
    MTL_ERROR1(MT_FLFMT("Could not get number of HCA ports (%s).\n"),HH_strerror_sym(rc));
    return 0;
  }
  port = (port >= num_ports) ? ((port-num_ports)%num_ports)+1 : (port % num_ports)+1;
#endif

  hdrs_p = (MPGA_headers_t *)INTR_MALLOC(sizeof(MPGA_headers_t));
  if (hdrs_p == NULL) {
       MT_COV_MEM_FLTR_START
       MTL_ERROR1(MT_FLFMT("%s: Could not allocate MPGA_headers_t structure."),__func__);
       return 0;
       MT_COV_MEM_FLTR_END
  }


  hdrs_sz= IB_LRH_LEN+IB_BTH_LEN+IB_DETH_LEN;
  if (global)  hdrs_sz+= IB_GRH_LEN;

  /* Set inline entry control */
  *segment_p= ((1<<31) | hdrs_sz) ; /* inline entry | ByteCount */
  hdrs_buf_p= ((u_int8_t*)segment_p) + WQE_INLINE_SZ_BCOUNT /* inline ctrl */ + hdrs_sz;

  /* Put headers data into MPGA structures */  
  if (global) {
    LRH_p= &(hdrs_p->MPGA_G_ud_send_only.IB_LRH);
    BTH_p= &(hdrs_p->MPGA_G_ud_send_only.IB_BTH);
    DETH_p= &(hdrs_p->MPGA_G_ud_send_only.IB_DETH);
    /* Set GRH fields */
    hdrs_p->MPGA_G_ud_send_only.IB_GRH.IPVer= 6; /* ? */
    hdrs_p->MPGA_G_ud_send_only.IB_GRH.TClass= av_p->traffic_class;
    hdrs_p->MPGA_G_ud_send_only.IB_GRH.FlowLabel= av_p->flow_label;
    hdrs_p->MPGA_G_ud_send_only.IB_GRH.PayLen= IB_BTH_LEN+IB_DETH_LEN+IB_MAD_LEN+IB_ICRC_SZ;
    hdrs_p->MPGA_G_ud_send_only.IB_GRH.NxtHdr= 0x1B; /* IB-spec.: compliancy statement C8-7 */
    hdrs_p->MPGA_G_ud_send_only.IB_GRH.HopLmt= av_p->hop_limit; 
    memcpy(&(hdrs_p->MPGA_G_ud_send_only.IB_GRH.DGID),&(av_p->dgid),sizeof(IB_gid_t));
#ifdef MT_KERNEL
    /* SGID field is supported only in kernel space, due to limited access to the GID table */
    rc= XHH_hob_get_sgid(hh_hndl,port,av_p->sgid_index,
      &(hdrs_p->MPGA_G_ud_send_only.IB_GRH.SGID));
    if (rc != HH_OK) {
      MTL_ERROR1(MT_FLFMT("Error in GID table access (%s).\n"),HH_strerror_sym(rc));
      FREE(hdrs_p);
      return 0;
    }
#endif

  } else {  /* local - no GRH */
    LRH_p= &(hdrs_p->MPGA_ud_send_only.IB_LRH);
    BTH_p= &(hdrs_p->MPGA_ud_send_only.IB_BTH);
    DETH_p= &(hdrs_p->MPGA_ud_send_only.IB_DETH);
  }
  
  /* Set LRH fields */
  memset(LRH_p,0,sizeof(IB_LRH_st));
  /* VL must be set for internal loopback ("vl15" bit is ignored) */
  if (qp->sqp_type == VAPI_SMI_QP) LRH_p->VL= 15;
  else LRH_p->VL= 0;
  LRH_p->LVer= 0;
  LRH_p->SL= av_p->sl;
  LRH_p->LNH= global ? IBA_GLOBAL : IBA_LOCAL;

  LRH_p->DLID= av_p->dlid;
  LRH_p->SLID= (av_p->dlid == PERMIS_LID) ? PERMIS_LID : (IB_lid_t) av_p->src_path_bits;
  /* If DLID is permissive LID, we set SLID to the permissive LID too. */
  /* Otherwise, we put in the SLID field the source path bits, and SLR=0, so */ 
  /*   the LID is composed of actual port's LID concatenated with given path bits */

  LRH_p->PktLen= (hdrs_sz+IB_MAD_LEN + IB_ICRC_SZ) >> 2;
  /* Set BTH fields */
  memset(BTH_p,0,sizeof(IB_BTH_st));
  BTH_p->OpCode= UD_SEND_ONLY_OP;
  BTH_p->SE= send_req_p->set_se;
  BTH_p->M= 1;
  BTH_p->PadCnt= 0; /* MADs are always 4byte multiple */
  BTH_p->TVer= 0; 
#ifdef MT_KERNEL
  if (qp->sqp_type == VAPI_GSI_QP) {
    if (pkey_index == QP1_PKEY_INDEX) { /* use QP's pkey */
      rc= XHH_hob_get_qp1_pkey(hh_hndl,port,&cur_pkey);
    } else {
      rc= XHH_hob_get_pkey(hh_hndl,port,pkey_index,&cur_pkey);
    }
    if (rc != HH_OK) {
      MTL_ERROR1(MT_FLFMT("%s: Error in P-key table access (%s) - using pkey_index 0x%X.\n"),__func__,
                 HH_strerror_sym(rc),pkey_index);
      FREE(hdrs_p);
      return 0;
    }
  } else {
      cur_pkey = DEFAULT_PKEY;
  }
  BTH_p->P_KEY= cur_pkey; 
#else
  BTH_p->P_KEY= DEFAULT_PKEY; /* For user space we do not have access to the Pkey table */
#endif
  BTH_p->DestQP= send_req_p->remote_qp;
  /* AckReq and PSN are meaningless for UD */
  /* Set DETH fields */
  memset(DETH_p,0,sizeof(IB_DETH_st));
  DETH_p->SrcQP= (qp->sqp_type == VAPI_SMI_QP) ? 0 : 1; /* invoked only for SMI or GSI */ 
  /* Qkey should be set according to IB-Spec. compliancy statement C10-15, But...      
   * Only QP1/GSI is the special QP which really validates Q-keys and it always uses
   * 0x80010000 (C9-49). So for QP1 we always put this if the high-order bit of the Qkey
   * is set.                                                                             */
  if (qp->sqp_type == VAPI_GSI_QP) {
    DETH_p->Q_Key= (send_req_p->remote_qkey & 0x80000000) ? 0x80010000: send_req_p->remote_qkey;
  } else { /* QP0 */
    /* For QP0 we don't care (QP0 always sends to another QP0 - none of which validates the Qkey) */
    DETH_p->Q_Key= send_req_p->remote_qkey; 
  }

  /* Build the headers */
  if (MPGA_make_headers(hdrs_p,UD_SEND_ONLY_OP,
                        LRH_p->LNH,FALSE,IB_MAD_LEN,&hdrs_buf_p) != MT_OK) {
    FREE(hdrs_p);
    return 0;
  }
  /* Verify headers size */
  if (hdrs_buf_p != (((u_int8_t*)segment_p) + WQE_INLINE_SZ_BCOUNT)) {/*Should be segment begin*/
    MTL_ERROR2(MT_FLFMT("Error in headers size (%d instead of %d).\n"),
     (unsigned) (hdrs_sz - (hdrs_buf_p - (((u_int8_t*)segment_p) + 4))), hdrs_sz);
    FREE(hdrs_p);
    return 0;
  }

#ifdef MT_LITTLE_ENDIAN
  /* MPGA headers returned in BIG endian.  WQE is built in CPU endianess  - so swap bytes */
  for (i= 0 , hdrs_buf32_p= (u_int32_t*)hdrs_buf_p; i < (hdrs_sz>>2); i++) {
    hdrs_buf32_p[i]= MOSAL_cpu_to_be32(hdrs_buf32_p[i]);
  }
#endif

  FREE(hdrs_p);
  return MT_UP_ALIGNX_U32(WQE_INLINE_SZ_BCOUNT + hdrs_sz , 4);  /* Align to WQE segment size */
}
  
/* Build ICRC segment for MLX (UD) */
inline static u_int32_t WQE_pack_mlx_icrc_hw(u_int32_t *segment_p)
{
  segment_p[0]= (1<<31) | 4 ; /* Inline ICRC (32 bits = 4 bytes) */
  segment_p[1]= 0;            /* Hardware generated ICRC */
  /* 2 dwords padded for a single Inline Data segment */
  return WQE_INLINE_ICRC;
}

/* Pack Control segment (for mlx-sends) */
inline static u_int32_t WQE_pack_ctrl_mlx(u_int32_t *segment_p,  
    VAPI_comp_type_t comp_type, MT_bool event_bit,
    IB_sl_t sl, IB_static_rate_t max_statrate, MT_bool slr, MT_bool v15,
    u_int16_t vcrc, IB_lid_t rlid)
{
  memset(segment_p,0,WQE_SEG_SZ_CTRL);  /* Clear all "RESERVED" */
  MT_INSERT_ARRAY32(segment_p,(comp_type == VAPI_SIGNALED) ? 1 : 0,
    MT_BIT_OFFSET(WQE_SEGMENT_CTRL_MLX_ST,c),MT_BIT_SIZE(WQE_SEGMENT_CTRL_MLX_ST,c));
  MT_INSERT_ARRAY32(segment_p,event_bit ? 1 : 0,
    MT_BIT_OFFSET(WQE_SEGMENT_CTRL_MLX_ST,e),MT_BIT_SIZE(WQE_SEGMENT_CTRL_MLX_ST,e));
  MT_INSERT_ARRAY32(segment_p,sl,
    MT_BIT_OFFSET(WQE_SEGMENT_CTRL_MLX_ST,sl),MT_BIT_SIZE(WQE_SEGMENT_CTRL_MLX_ST,sl));
  MT_INSERT_ARRAY32(segment_p,max_statrate > 0 ? 1 : 0,
    MT_BIT_OFFSET(WQE_SEGMENT_CTRL_MLX_ST,max_statrate),
    MT_BIT_SIZE(WQE_SEGMENT_CTRL_MLX_ST,max_statrate));
  MT_INSERT_ARRAY32(segment_p,slr ? 1 : 0,
    MT_BIT_OFFSET(WQE_SEGMENT_CTRL_MLX_ST,slr),MT_BIT_SIZE(WQE_SEGMENT_CTRL_MLX_ST,slr));
  MT_INSERT_ARRAY32(segment_p,v15 ? 1 : 0,
    MT_BIT_OFFSET(WQE_SEGMENT_CTRL_MLX_ST,v15),MT_BIT_SIZE(WQE_SEGMENT_CTRL_MLX_ST,v15));
  MT_INSERT_ARRAY32(segment_p,vcrc,
    MT_BIT_OFFSET(WQE_SEGMENT_CTRL_MLX_ST,vcrc),MT_BIT_SIZE(WQE_SEGMENT_CTRL_MLX_ST,vcrc));
  MT_INSERT_ARRAY32(segment_p,rlid,
    MT_BIT_OFFSET(WQE_SEGMENT_CTRL_MLX_ST,rlid),MT_BIT_SIZE(WQE_SEGMENT_CTRL_MLX_ST,rlid));
  return WQE_SEG_SZ_CTRL;
}

inline static u_int32_t WQE_build_send_mlx(
  HH_hca_hndl_t hh_hndl,
  XHHUL_qp_t qp,
  VAPI_sr_desc_t *send_req_p,
  VAPI_pkey_ix_t pkey_index, /* take this index instead of QP's, if not QP1_PKEY_INDEX */
  volatile u_int32_t *wqe_buf
)  
{
  
  VAPI_ud_av_t *av_p = NULL;
  u_int8_t *cur_loc_p= (u_int8_t*)wqe_buf; /* Current location in the WQE */
  u_int8_t *prev_loc_p;
  HH_ret_t rc;

  av_p = (VAPI_ud_av_t *)INTR_MALLOC(sizeof(VAPI_ud_av_t));
  if (av_p == NULL) {
       MT_COV_MEM_FLTR_START
       MTL_ERROR1(MT_FLFMT("%s: Could not allocate VAPI_ud_av_t structure."),__func__);
       return 0;
       MT_COV_MEM_FLTR_END
  }
  /* no need for memset.  whole structure is copied below */
  rc= XHH_udavm_query_av_internal(XHHUL_DEV_SPECIFIC_UDAVM(qp), send_req_p->remote_ah, av_p);
  if (rc != HH_OK) {
    MTL_ERROR1(MT_FLFMT("Invalid UD AV handle - %s"),
      HH_strerror_sym(rc));
    FREE(av_p);
    return 0;
  }

   
  if ((av_p->dlid == PERMIS_LID) && (qp->sqp_type != VAPI_SMI_QP)) {
    MTL_ERROR1(MT_FLFMT("DLID==Permissive-LID while not an SMI QP.\n"));
    FREE(av_p);
    return 0;
  }

  cur_loc_p+= WQE_init_next((u_int32_t*)cur_loc_p); /* Make "unlinked" "next" segment */
  cur_loc_p+= WQE_pack_ctrl_mlx((u_int32_t*)cur_loc_p,  /* Pack Control segment */
    send_req_p->comp_type, FALSE/*event bit*/,
    av_p->sl,av_p->static_rate,(av_p->dlid == PERMIS_LID),(qp->sqp_type == VAPI_SMI_QP),
    0/*VCRC*/,av_p->dlid);
    
    
  /* Build inline headers */
  switch (qp->sqp_type) {
    case VAPI_SMI_QP:  
    case VAPI_GSI_QP:  
      prev_loc_p= cur_loc_p;
      cur_loc_p+= WQE_pack_mlx_ud_header((u_int32_t*)cur_loc_p,qp,send_req_p,av_p,hh_hndl,pkey_index);
      if (cur_loc_p == prev_loc_p) {
	FREE(av_p);
        return 0;
      }
      /* Pack scatter/gather list segments */
      cur_loc_p+= WQE_pack_sg_list((u_int32_t*)cur_loc_p,send_req_p->sg_lst_len,send_req_p->sg_lst_p);
      cur_loc_p+= WQE_pack_mlx_icrc_hw((u_int32_t*)cur_loc_p);
      break;
    case VAPI_RAW_ETY_QP:
    case VAPI_RAW_IPV6_QP:
    default:
      FREE(av_p);
      return 0;
  }
  

  FREE(av_p);
  return (u_int32_t)(((MT_virt_addr_t)cur_loc_p) - ((MT_virt_addr_t)wqe_buf));
}

/* Pack Control segment (for receive work requests) */
inline static u_int32_t WQE_pack_ctrl_recv(u_int32_t *segment_p,  
    VAPI_comp_type_t comp_type, u_int32_t event_bit)
{
  memset(segment_p,0,WQE_SEG_SZ_CTRL);  /* Clear all "RESERVED" */
  MT_INSERT_ARRAY32(segment_p,(comp_type == VAPI_SIGNALED) ? 1 : 0,
    MT_BIT_OFFSET(WQE_SEGMENT_CTRL_RECV_ST,c),MT_BIT_SIZE(WQE_SEGMENT_CTRL_RECV_ST,c));
  MT_INSERT_ARRAY32(segment_p,event_bit,
    MT_BIT_OFFSET(WQE_SEGMENT_CTRL_RECV_ST,e),MT_BIT_SIZE(WQE_SEGMENT_CTRL_RECV_ST,e));
  return WQE_SEG_SZ_CTRL;
}

inline static u_int32_t WQE_build_recv(
  XHHUL_qp_t qp,
  VAPI_rr_desc_t *recv_req_p,
  volatile u_int32_t *wqe_buf
)
{
  u_int8_t *cur_loc_p= (u_int8_t*)wqe_buf; /* Current location in the WQE */

  cur_loc_p+= WQE_init_next_recv((u_int32_t*)cur_loc_p); /* Make "unlinked" "next" receive segment */
  cur_loc_p+= WQE_pack_ctrl_recv((u_int32_t*)cur_loc_p,
    recv_req_p->comp_type, 0/*event bit*/);
  /* Pack scatter/gather list segments */
  cur_loc_p+= WQE_pack_sg_list((u_int32_t*)cur_loc_p,recv_req_p->sg_lst_len,recv_req_p->sg_lst_p);
  XHHUL_DEV_SPECIFIC_WQE_BUILD_RECV;
  return (u_int32_t)(((MT_virt_addr_t)cur_loc_p) - ((MT_virt_addr_t)wqe_buf));
}

inline static u_int32_t WQE_build_membind(
  HHUL_mw_bind_t *bind_props_p,
  IB_rkey_t new_rkey,
  volatile u_int32_t *wqe_buf
)
{
  volatile u_int32_t *cur_loc_p= wqe_buf; /* Current location in the WQE */

  cur_loc_p+= (WQE_init_next((u_int32_t*)cur_loc_p)>>2);  /* Make "unlinked" "next" segment */
  cur_loc_p+= (WQE_pack_ctrl_send((u_int32_t*)cur_loc_p,  /* Pack Control segment */
    bind_props_p->comp_type, 0/*SE bit*/, 0/*event bit*/,0/*Imm. data*/)>>2);

  memset((void*)cur_loc_p,0,8);  /* clear reserved bits of first 2 dwords */

  /* Set access bits */
  if (bind_props_p->acl & VAPI_EN_REMOTE_READ) {
    MT_INSERT_ARRAY32(cur_loc_p,1,
      MT_BIT_OFFSET(WQE_SEGMENT_BIND_ST,rr),MT_BIT_SIZE(WQE_SEGMENT_BIND_ST,rr));
  }
  if (bind_props_p->acl & VAPI_EN_REMOTE_WRITE) {
    MT_INSERT_ARRAY32(cur_loc_p,1,
      MT_BIT_OFFSET(WQE_SEGMENT_BIND_ST,rw),MT_BIT_SIZE(WQE_SEGMENT_BIND_ST,rw));
  }
  if (bind_props_p->acl & VAPI_EN_REMOTE_ATOM) {
    MT_INSERT_ARRAY32(cur_loc_p,1,
      MT_BIT_OFFSET(WQE_SEGMENT_BIND_ST,a),MT_BIT_SIZE(WQE_SEGMENT_BIND_ST,a));
  }

  cur_loc_p[MT_BYTE_OFFSET(WQE_SEGMENT_BIND_ST,new_rkey)>>2]= new_rkey;
  cur_loc_p[MT_BYTE_OFFSET(WQE_SEGMENT_BIND_ST,region_lkey)>>2]= bind_props_p->mr_lkey;
  cur_loc_p[MT_BYTE_OFFSET(WQE_SEGMENT_BIND_ST,start_address_h)>>2]= 
    (u_int32_t)(bind_props_p->start >> 32);
  cur_loc_p[MT_BYTE_OFFSET(WQE_SEGMENT_BIND_ST,start_address_l)>>2]= 
    (u_int32_t)(bind_props_p->start & 0xFFFFFFFF);
  cur_loc_p[MT_BYTE_OFFSET(WQE_SEGMENT_BIND_ST,length_h)>>2]= 
    (u_int32_t)(bind_props_p->size >> 32);
  cur_loc_p[MT_BYTE_OFFSET(WQE_SEGMENT_BIND_ST,length_l)>>2]= 
    (u_int32_t)(bind_props_p->size & 0xFFFFFFFF);

  return (u_int32_t)(WQE_SEG_SZ_BIND + ((MT_virt_addr_t)cur_loc_p) - ((MT_virt_addr_t)wqe_buf));
}

/* logic for tracking wqes and adjusting nopcode to handle problem where a UD/RD send WQE using
 * QKEY from QPC, which follows a send WQE which used QKEY from WQE did not contain the correct
 * QKEY -- it contained the QKEY from the previous WQE instead of the QKEY from the QPC  */
inline static ADD_DEV_PREFIX_LC(if_nopcode_t) adjust_nopcode_and_last_qkey_flag(
    XHHUL_qp_t qp, u_int32_t qkey_from_qpc, ADD_DEV_PREFIX_LC(if_nopcode_t) nopcode,
    /*IN/OUT*/ MT_bool *last_qkey_from_wqe_p)
{
#ifdef UL_QPM_QKEY_BUG_FM23450_NOT_FIXED
return nopcode;
#else
    MT_bool last_qkey_from_wqe;

    //return nopcode;
    /* we don't support RD as yet.  Also, qkey is a don't-care for QP0 */
    if (((qp->ts_type != IB_TS_UD) && (qp->ts_type != IB_TS_RD))|| (qp->sqp_type == VAPI_SMI_QP)) {
	return nopcode;
    }

    last_qkey_from_wqe = *last_qkey_from_wqe_p;
    *last_qkey_from_wqe_p = (qkey_from_qpc ? FALSE : TRUE);

    if ((last_qkey_from_wqe) && (qkey_from_qpc)) {
       switch (nopcode) {
       case ADD_DEV_PREFIX(IF_NOPCODE_SEND):
	   return ADD_DEV_PREFIX(IF_NOPCODE_RDMAW);
       case ADD_DEV_PREFIX(IF_NOPCODE_SEND_IMM):
	   return ADD_DEV_PREFIX(IF_NOPCODE_RDMAW_IMM);
       default: return nopcode;    
       }
    }
    return nopcode;
#endif
}
/* - Allocate a WQE in given send queue, 
   - put given WQE in it, 
   - link to previos WQE and 
   - ring the doorbell 
   * q_lock must be acquired before invoking this function (to protect WQEs allocation).
 */ 
inline static HH_ret_t sq_alloc_wqe_link_and_ring(XHHUL_qp_t qp, 
  volatile u_int32_t* next_wqe, u_int32_t wqe_sz_dwords, 
#ifdef MT_LITTLE_ENDIAN
  u_int32_t swap_sz_dwords, 
#endif
  VAPI_sr_desc_t *send_req_p, ADD_DEV_PREFIX_LC(if_nopcode_t) nopcode)
{
  u_int32_t next_draft[WQE_SEG_SZ_NEXT>>2]; /* Build "next" segment here */
  u_int32_t i;
  XHH_uar_sendq_dbell_t sq_dbell;
  ADD_DEV_PREFIX_LC(if_nopcode_t) adjusted_opcode;
  
  qp->sq_res.wqe_id[qp->sq_res.next2post_index]= send_req_p->id;  /* Save WQE ID */
  qp->sq_res.next2post_index = (qp->sq_res.next2post_index + 1) % qp->sq_res.max_outs ;
  
  MTPERF_TIME_START(SQ_WQE_copy);
#ifdef MT_LITTLE_ENDIAN
  /* swap WQE's dwords */
  for (i=WQE_SND_DRAFT_START_COPY; i < swap_sz_dwords; i++) {
    next_wqe[i]= MOSAL_cpu_to_be32(next_wqe[i]);
  }
#endif

  MTPERF_TIME_END(SQ_WQE_copy);
#ifdef __KERNEL__
//MTL_ERROR1(MT_FLFMT("%s: qp->sq_res.last_posted_p=%p, max_outs=0x%x, log2_wqe_sz="SIZE_T_DFMT),__func__,
//	   qp->sq_res.last_posted_p, qp->sq_res.max_outs,qp->sq_res.log2_max_wqe_sz);
//MTL_ERROR1(MT_FLFMT("%s: next_wqe=%p"),__func__, next_wqe);
#endif
  
  adjusted_opcode = adjust_nopcode_and_last_qkey_flag(qp, 
		      (u_int32_t) (((u_int32_t) (send_req_p->remote_qkey)) & 0x80000000), 
		      nopcode, &(qp->last_qkey_from_wqe));

  /* Update "next" segment of previous WQE (if any) */
  XHH_DEV_SPECIFIC_SQ_TEST_PREV_WQE_NOT_NULL  {
    /* Build linking "next" segment in last posted WQE*/
    WQE_pack_send_next(next_draft, adjusted_opcode, send_req_p->fence,
      1/*DBD*/, (u_int32_t)(MT_ulong_ptr_t) next_wqe, wqe_sz_dwords>>2, 
      (qp->ts_type==VAPI_TS_RD) ? send_req_p->eecn : 0);
    for (i= 0;i < (WQE_SEG_SZ_NEXT>>2) ;i++) {  
      /* This copy assures big-endian as well as that NDS is written last */
      qp->sq_res.last_posted_p[i]= MOSAL_cpu_to_be32(next_draft[i]);
    }
  }
  qp->sq_res.last_posted_p= next_wqe;
  
  /* Ring doorbell (send or rd-send) */
  sq_dbell.nopcode= adjusted_opcode;
  sq_dbell.fence= send_req_p->fence;
  sq_dbell.next_size= wqe_sz_dwords>>2;
  
  XHH_DEV_SPECIFIC_SQ_ALLOC_WQE_LINK_AND_RING
  
  if (qp->ts_type == VAPI_TS_RD) {
    uar_sendq_rd_dbell(qp,&sq_dbell,send_req_p);
  } else {  /* non-RD send request */
    uar_sendq_dbell(qp,&sq_dbell);
  }

  return HH_OK;
}

/* Extract NDS directly from (big-endian) WQE */
inline static u_int8_t WQE_extract_nds(volatile u_int32_t* wqe)
{
  return MT_EXTRACT32(MOSAL_be32_to_cpu(wqe[MT_BYTE_OFFSET(WQE_SEGMENT_NEXT_ST,nds) >> 2]),
                   MT_BIT_OFFSET(WQE_SEGMENT_NEXT_ST,nds) & MASK32(5),
                   MT_BIT_SIZE(WQE_SEGMENT_NEXT_ST,nds) & MASK32(5));
}

/* Extract NDA directly from (big-endian) WQE */
inline static u_int32_t WQE_extract_nda(volatile u_int32_t* wqe)
{
  return (MOSAL_be32_to_cpu(wqe[MT_BYTE_OFFSET(WQE_SEGMENT_NEXT_ST,nda_31_6) >> 2]) & (~MASK32(6)) );
}

inline static u_int8_t WQE_extract_dbd(volatile u_int32_t* wqe)
{
  XHH_DEV_SPECIFIC_EXTRACT_DBD
}

#ifdef DUMP_SEND_REQ
  static void dump_send_req(XHHUL_qp_t qp, HHUL_send_req_t *sr);
#endif

/**********************************************************************************************
 *                    Public API Functions (defined in xhhul_hob.h)
 **********************************************************************************************/

HH_ret_t XHHUL_qpm_create( 
  XHHUL_hob_t hob, 
  XHHUL_srqm_t srqm,
  XHHUL_qpm_t *qpm_p 
) 
{ 
  int i;

  *qpm_p= (XHHUL_qpm_t) MALLOC(sizeof(struct XHHUL_qpm_st));
  if (*qpm_p == NULL) {
    MTL_ERROR1("XHHUL_qpm_create: Failed allocating XHHUL_qpm_st.\n");
    return HH_EAGAIN;
  }

  /* init internal data structures */
  (*qpm_p)->qp_list= NULL;
  (*qpm_p)->srqm= srqm;
  for (i= XHHUL_DPOOL_SZ_MIN_KB; i <= XHHUL_DPOOL_SZ_MAX_KB; i++) {
    (*qpm_p)->dpool_p[i - XHHUL_DPOOL_SZ_MIN_KB]= NULL;
  }

  MOSAL_mutex_init(&(*qpm_p)->dpool_lock);
  MOSAL_mutex_init(&(*qpm_p)->qpm_lock);

  return HH_OK;
}


HH_ret_t XHHUL_qpm_destroy( 
   XHHUL_qpm_t qpm 
) 
{ 
    qp_list_entry_t *qp_item; 
    XHHUL_qp_t qp;

    while (qpm->qp_list) {
	    qp_item = qpm->qp_list;
	    qpm->qp_list = qp_item->next;
	    qp = qp_item->qp;
	    FREE(qp_item); 

	       /* Clean all CQEs which refer to this QP */
	       XHHUL_cqm_cq_cleanup(qp->rq_cq, qp->qpn, qpm->srqm, qp->srq);
	       if (qp->sq_cq != qp->rq_cq) /* additional cleaning required only if SQ's CQ is different */
	       XHHUL_cqm_cq_cleanup(qp->sq_cq, qp->qpn, qpm->srqm, HHUL_INVAL_SRQ_HNDL);

	       /* Free QP resources: Auxilary buffer + WQEs buffer */
	     if (qp->sq_res.wqe_id != NULL) {
	       XHH_SMART_FREE(qp->sq_res.wqe_id, qp->sq_res.max_outs * sizeof(VAPI_wr_id_t)); 
	     }
	     if (qp->rq_res.wqe_id != NULL) {
	       XHH_SMART_FREE(qp->rq_res.wqe_id, qp->rq_res.max_outs * sizeof(VAPI_wr_id_t)); 
	     }
		 MOSAL_memory_unlock(qp->wq_base2reg, qp->wq_size2reg);
	     if (qp->wqe_buf_orig != NULL) {/* WQEs buffer were allocated in process mem. or by the XHH_qpm ? */ 
	       MTL_DEBUG4(MT_FLFMT("Freeing WQEs buffer at 0x%p"),qp->wqe_buf_orig);
	       if (qp->dpool_p == NULL) { /* direct allocation */
		 if (qp->used_virt_alloc) 
		   MOSAL_pci_virt_free_consistent(qp->wqe_buf_orig, qp->wqe_buf_orig_size);
		 else
		   MOSAL_pci_phys_free_consistent(qp->wqe_buf_orig, qp->wqe_buf_orig_size);    
	       } else { /* used dpool */
		 dpool_free(qpm, qp->dpool_p, qp->wqe_buf_orig);
	       }
	     }
	     FREE(qp);
    }

    
    FREE(qpm);
    return HH_OK;
}


HH_ret_t XHHUL_qpm_create_qp_prep( 
   HHUL_hca_hndl_t hca, 
   HHUL_qp_init_attr_t *qp_init_attr_p, 
   HHUL_qp_hndl_t *qp_hndl_p, 
   VAPI_qp_cap_t *qp_cap_out_p, 
   void/*XHH_qp_ul_resources_t*/ *qp_ul_resources_p 
) 
{ 
  return qp_prep(hca,VAPI_REGULAR_QP,qp_init_attr_p,qp_hndl_p,qp_cap_out_p,
    (XHH_qp_ul_resources_t*)qp_ul_resources_p,
    FALSE);  /* Default is allocation of WQEs buffer in host's mem. */
}

HH_ret_t XHHUL_qpm_special_qp_prep( 
   HHUL_hca_hndl_t hca, 
   VAPI_special_qp_t qp_type, 
   IB_port_t port, 
   HHUL_qp_init_attr_t *qp_init_attr_p, 
   HHUL_qp_hndl_t *qp_hndl_p, 
   VAPI_qp_cap_t *qp_cap_out_p, 
   void/*XHH_qp_ul_resources_t*/ *qp_ul_resources_p 
) 
{ 
  return qp_prep(hca,qp_type,qp_init_attr_p,qp_hndl_p,qp_cap_out_p,
    (XHH_qp_ul_resources_t*)qp_ul_resources_p,
    FALSE);  /* For special QPs no performance issue - WQEs in main memory */
}


HH_ret_t XHHUL_qpm_create_qp_done( 
  HHUL_hca_hndl_t hca, 
  HHUL_qp_hndl_t hhul_qp, 
  IB_wqpn_t hh_qp, 
  void/*XHH_qp_ul_resources_t*/ *qp_ul_resources_p
) 
{ 
  XHHUL_qpm_t qpm;
  XHHUL_qp_t qp= (XHHUL_qp_t)hhul_qp;
  XHH_qp_ul_resources_t *ul_res_p= (XHH_qp_ul_resources_t*)qp_ul_resources_p;
  HH_ret_t rc;
  
  rc= XHHUL_hob_get_qpm(hca,&qpm);
  if (rc != HH_OK) {
    MTL_ERROR4("XHHUL_qpm_create_qp_done: Invalid HCA handle (%p).",hca);
    return HH_EINVAL;
  }
  if (qp == NULL) {
    MTL_ERROR4("XHHUL_qpm_create_qp_done: NULL hhul_qp handle.");
    return HH_EINVAL;
  }
  if ((qp->wqe_buf_orig == NULL) && (qp->wqe_buf_orig_size != 0)) { 
    /* WQEs buffer allocated in DDR mem. by XHH_qpm */
    if (ul_res_p->wqes_buf == 0) {
      MTL_ERROR1(MT_FLFMT("Got NULL WQEs buffer from qp_ul_res for new qpn=%d.\n"),qp->qpn);
      return HH_EINVAL;
    }
    /* Set the per queue resources */
    qp->rq_res.wqe_buf= MT_UP_ALIGNX_VIRT(ul_res_p->wqes_buf,(unsigned long)qp->rq_res.log2_max_wqe_sz);
    if (qp->rq_res.wqe_buf != ul_res_p->wqes_buf) {
      MTL_ERROR1(
        "XHHUL_qpm_create_qp_done: Buffer allocated by XHH_qpm ("VIRT_ADDR_FMT") "
        "is not aligned to RQ WQE size (%d bytes).\n",
        ul_res_p->wqes_buf,1<<qp->rq_res.log2_max_wqe_sz);
      return HH_EINVAL;
    }
    /* SQ is after RQ - aligned to its WQE size */
    qp->sq_res.wqe_buf= MT_UP_ALIGNX_VIRT(qp->rq_res.wqe_buf + 
        (qp->rq_res.max_outs << qp->rq_res.log2_max_wqe_sz), /* End of RQ WQEs buffer */
      (unsigned long)qp->sq_res.log2_max_wqe_sz); 
  }
  
  qp->qpn= hh_qp;

  rc= XHHUL_cqm_attach_qp(qp->rq_cq,qp,qp->qpn);
  if (rc != HH_OK) {
	MTL_ERROR1(MT_FLFMT("%s: Failed attaching QPn=0x%X to its RQ CQ"), __func__, qp->qpn);
	goto attach_rq_cq_fail;
  }

  if (qp->rq_cq != qp->sq_cq) {
	  rc= XHHUL_cqm_attach_qp(qp->sq_cq,qp,qp->qpn);
	  if (rc != HH_OK) {
		MTL_ERROR1(MT_FLFMT("%s: Failed attaching QPn=0x%X to its SQ CQ"), __func__, qp->qpn);
		goto attach_sq_cq_fail;
	  }
  }

  /* Insert QP to the QPs list */
  rc= insert_to_list(qpm,qp);
  if (rc != HH_OK) {
    MTL_ERROR2("XHHUL_qpm_create_qp_done: Failed inserting to QPs list "
               "(QP will remain unusable) !"); 
    goto insert_to_list_fail;
  }

  MTL_DEBUG4(MT_FLFMT("%s: qpn=0x%X rq_res{buf_p=%p, sz=0x%X} sq_res{buf_p=%p, sz=0x%X}"), __func__,
             qp->qpn, 
             (void*)qp->rq_res.wqe_buf, (1 << qp->rq_res.log2_max_wqe_sz) * qp->rq_res.max_outs,
             (void*)qp->sq_res.wqe_buf, (1 << qp->sq_res.log2_max_wqe_sz) * qp->sq_res.max_outs);

  XHH_DEV_SPECIFIC_CREATE_QP_DONE;

  return HH_OK;

  
  insert_to_list_fail:
	  if (qp->rq_cq != qp->sq_cq)  XHHUL_cqm_detach_qp(qp->sq_cq, qp, qp->qpn);
  attach_sq_cq_fail:
	  XHHUL_cqm_detach_qp(qp->rq_cq, qp, qp->qpn);
  attach_rq_cq_fail:
	  qp->qpn= 0xFFFFFFFF; /* Mark that QP initialization was not completed with invalid QP num. */
	  return rc;
}


HH_ret_t XHHUL_qpm_destroy_qp_done( 
   HHUL_hca_hndl_t hca, 
   HHUL_qp_hndl_t hhul_qp 
) 
{ 
  XHHUL_qpm_t qpm;
  XHHUL_qp_t qp= (XHHUL_qp_t)hhul_qp;
  HH_ret_t rc;
  
  MTL_DEBUG1("XHHUL_qpm_destroy_qp_done(hca=%s,hhul_qp=%p) {\n",hca->dev_desc,qp);
  rc= XHHUL_hob_get_qpm(hca,&qpm);
  if (rc != HH_OK) {
    MTL_ERROR4("XHHUL_qpm_destroy_qp_done: Invalid HCA handle (%p).",hca);
    return HH_EINVAL;
  }
  
  if (IS_VALID_QPN(qp->qpn)) {	/* QP has completed XHHUL_qpm_create_qp_done successfully */
    
    /* Clean all CQEs which refer to this QP */
    XHHUL_cqm_cq_cleanup(qp->rq_cq, qp->qpn, qpm->srqm, qp->srq);
    if (qp->sq_cq != qp->rq_cq) /* additional cleaning required only if SQ's CQ is different */
    XHHUL_cqm_cq_cleanup(qp->sq_cq, qp->qpn, qpm->srqm, HHUL_INVAL_SRQ_HNDL);
    
    /* Remove QP from QPs list */
    rc= remove_from_list(qpm,qp);
    if (rc != HH_OK) {
      MTL_ERROR2("XHHUL_qpm_destroy_qp_done: Failed removing qp from QPs list "
                 "(assuming invalid QP handle) !"); 
      return HH_EINVAL_QP_NUM;
    }
    MTL_DEBUG4(MT_FLFMT("QP %d removed from QPs list"),qp->qpn);
  
    if (qp->rq_cq != qp->sq_cq)  XHHUL_cqm_detach_qp(qp->sq_cq, qp, qp->qpn);
    XHHUL_cqm_detach_qp(qp->rq_cq, qp, qp->qpn);
  }

  /* Free QP resources: Auxilary buffer + WQEs buffer */
  MTL_DEBUG4(MT_FLFMT("Freeing user level WQE-IDs auxilary buffers"));
  if (qp->sq_res.wqe_id != NULL) {
    XHH_SMART_FREE(qp->sq_res.wqe_id, qp->sq_res.max_outs * sizeof(VAPI_wr_id_t)); 
  }
  if (qp->rq_res.wqe_id != NULL) {
    XHH_SMART_FREE(qp->rq_res.wqe_id, qp->rq_res.max_outs * sizeof(VAPI_wr_id_t)); 
  }
  if (qp->wqe_buf_orig != NULL) {/* WQEs buffer were allocated in process mem. or by the XHH_qpm ? */ 
    MTL_DEBUG4(MT_FLFMT("Freeing WQEs buffer at 0x%p"),qp->wqe_buf_orig);
	MOSAL_memory_unlock(qp->wq_base2reg, qp->wq_size2reg);
    if (qp->dpool_p == NULL) { /* direct allocation */
      if (qp->used_virt_alloc) 
        MOSAL_pci_virt_free_consistent(qp->wqe_buf_orig, qp->wqe_buf_orig_size);
      else
        MOSAL_pci_phys_free_consistent(qp->wqe_buf_orig, qp->wqe_buf_orig_size);    
    } else { /* used dpool */
      dpool_free(qpm, qp->dpool_p, qp->wqe_buf_orig);
    }
  }
  FREE(qp);
  MTL_DEBUG1("} /* XHHUL_qpm_destroy_qp_done */ \n");
  return HH_OK;  
}

HH_ret_t XHHUL_qpm_modify_qp_done( 
   HHUL_hca_hndl_t hca, 
   HHUL_qp_hndl_t hhul_qp, 
   VAPI_qp_state_t cur_state 
) 
{ 
  XHHUL_qp_t qp= (XHHUL_qp_t)hhul_qp;
  XHHUL_qpm_t qpm;
  HH_ret_t rc;
  
  if (qp == NULL) {
    MTL_ERROR1("XHHUL_qpm_modify_qp_done: NULL hhul_qp.\n");
    return HH_EINVAL;
  }

  rc= XHHUL_hob_get_qpm(hca,&qpm);
  if (rc != HH_OK) {
    MTL_ERROR1(MT_FLFMT("%s: Failed to get QPM handle (%d=%s)"), __func__, rc, HH_strerror_sym(rc));
    return rc;
  }

  /* Update in RQ */
  if (cur_state == VAPI_RESET) {  
    /* Cleanup all CQEs of RQ (flush) when moving to reset state */
    XHHUL_cqm_cq_cleanup(qp->rq_cq, qp->qpn, qpm->srqm, qp->srq);
    MOSAL_spinlock_irq_lock(&(qp->rq_res.q_lock));
    qp->rq_res.next2free_index= qp->rq_res.next2post_index= 0;
    qp->rq_res.last_posted_p= NULL;
    qp->rq_res.qp_state= VAPI_RESET;
    XHH_DEV_SPECIFIC_RESET_RQ_DONE;
    MOSAL_spinlock_unlock(&(qp->rq_res.q_lock));
    
    /* Cleanup all CQEs of SQ (flush) when moving to reset state */
    if (qp->sq_cq != qp->rq_cq) /* additional cleaning required only if SQ's CQ is different */
      XHHUL_cqm_cq_cleanup(qp->sq_cq, qp->qpn, qpm->srqm, HHUL_INVAL_SRQ_HNDL);
    MOSAL_spinlock_irq_lock(&(qp->sq_res.q_lock));
    qp->sq_res.next2free_index= qp->sq_res.next2post_index= 0;
    initialize_sq_last_posted(&(qp->sq_res));
    qp->sq_res.qp_state= VAPI_RESET;
    XHH_DEV_SPECIFIC_RESET_SQ_DONE;
    MOSAL_spinlock_unlock(&(qp->sq_res.q_lock));

  } else {
    qp->rq_res.qp_state= cur_state;
    qp->sq_res.qp_state= cur_state;
  }
  
  return HH_OK;
}



HH_ret_t XHHUL_qpm_post_send_req( 
   HHUL_hca_hndl_t hca, 
   HHUL_qp_hndl_t hhul_qp, 
   VAPI_sr_desc_t *send_req_p 
) 
{ 
  XHHUL_qp_t qp= (XHHUL_qp_t)hhul_qp;
  volatile u_int32_t* next_wqe;
  u_int32_t wqe_sz_dwords;
  HH_hca_hndl_t hh_hca_hndl;
  HH_ret_t rc;

  if (!is_qpstate_valid_2send(qp->sq_res.qp_state)) {
    MTL_ERROR1(MT_FLFMT("%s failed: qp state %d not valid to send \n"),__func__,qp->sq_res.qp_state);
    return HH_EINVAL_QP_STATE;
  }
  
  if (qp->sq_res.max_sg_sz < send_req_p->sg_lst_len) {
    MTL_ERROR2(
      "XHHUL_qpm_post_send_req: Scatter/Gather list is too large (%d entries > max_sg_sz=%d)\n",
      send_req_p->sg_lst_len,qp->sq_res.max_sg_sz);
    return HH_EINVAL_SG_NUM;
  }

#ifdef DUMP_SEND_REQ 
  dump_send_req(qp,send_req_p);
#endif
  
  MOSAL_spinlock_irq_lock(&(qp->sq_res.q_lock)); /* protect WQE allocation/link */

   /* Check if any WQEs are free to be consumed */
  if (qp->sq_res.max_outs <= (CUR_OUTS(qp->sq_res)+1))  {
    MTL_ERROR4("XHHUL_qpm_post_send_req: Send queue is full (%u requests outstanding).\n",
      CUR_OUTS(qp->sq_res));
      MOSAL_spinlock_unlock(&(qp->sq_res.q_lock)); 
    return HH_E2BIG_WR_NUM;
  }

  /* Allocate next WQE */
  next_wqe= (u_int32_t*)(qp->sq_res.wqe_buf + 
                        (qp->sq_res.next2post_index << qp->sq_res.log2_max_wqe_sz) );
 
  if (qp->sqp_type == VAPI_REGULAR_QP)  {
    MTPERF_TIME_START(WQE_build_send);
    wqe_sz_dwords= (WQE_build_send(qp,send_req_p,next_wqe) >> 2);
    MTPERF_TIME_END(WQE_build_send);

    /* XHH_DEV_SPECIFIC_POST_SEND_REQ; */

#ifdef MAX_DEBUG
    if ((wqe_sz_dwords<<2) > (1U << qp->sq_res.log2_max_wqe_sz)) {
      MTL_ERROR1(MT_FLFMT("QP 0x%X: Send WQE too large (%d > max=%d)"),
        qp->qpn,(wqe_sz_dwords<<2),(1U << qp->sq_res.log2_max_wqe_sz));
	}
#endif
  } else { /* special QP */
    if (send_req_p->opcode != VAPI_SEND)  {
      MOSAL_spinlock_unlock(&(qp->sq_res.q_lock)); 
      return HH_EINVAL_OPCODE;
    }
    send_req_p->fence= FALSE; /* required for MLX requests */
    rc= XHHUL_hob_get_hh_hca_hndl(hca,&hh_hca_hndl);
    if (rc != HH_OK) {
      MOSAL_spinlock_unlock(&(qp->sq_res.q_lock)); 
      return HH_EINVAL_HCA_HNDL;
    }
    
    wqe_sz_dwords= 
      (WQE_build_send_mlx(hh_hca_hndl, qp, send_req_p, QP1_PKEY_INDEX, next_wqe) >> 2);
    if (wqe_sz_dwords == 0) {
      MOSAL_spinlock_unlock(&(qp->sq_res.q_lock)); 
      MTL_ERROR1(MT_FLFMT("Failed building MLX headers for special QP.\n"));
      return HH_EINVAL_WQE;
    }
  }

  rc= sq_alloc_wqe_link_and_ring(qp,next_wqe,wqe_sz_dwords,
#ifdef MT_LITTLE_ENDIAN
           wqe_sz_dwords,
#endif
           send_req_p,encode_nopcode(send_req_p->opcode)); 
  MOSAL_spinlock_unlock(&(qp->sq_res.q_lock)); 
  return rc;
}

HH_ret_t XHHUL_qpm_post_inline_send_req( 
   HHUL_hca_hndl_t hca, 
   HHUL_qp_hndl_t hhul_qp, 
   VAPI_sr_desc_t *send_req_p 
) 
{ 
  XHHUL_qp_t qp= (XHHUL_qp_t)hhul_qp;
  u_int32_t* next_wqe;
  u_int8_t *cur_loc_p; /* Current location in the WQE */
  u_int8_t *wqe_edge_p;
  u_int32_t wqe_sz_dwords;
  u_int32_t* inline_p; /* inline control word */
  u_int32_t i;  
  HH_ret_t rc;

#ifdef DUMP_SEND_REQ 
  dump_send_req(qp,send_req_p);
#endif
  
  if (!is_qpstate_valid_2send(qp->sq_res.qp_state)) {
   MTL_ERROR1(MT_FLFMT("%s failed: qp state %d not valid to send \n"),__func__,qp->sq_res.qp_state);
   return HH_EINVAL_QP_STATE;
 }

  MOSAL_spinlock_irq_lock(&(qp->sq_res.q_lock)); /* protect WQE allocation/link */
  
   /* Check if any WQEs are free to be consumed */
  if (qp->sq_res.max_outs <= (CUR_OUTS(qp->sq_res)+1))  {
    MTL_ERROR4("%s: Send queue is full (%u requests outstanding).\n", __func__,
      CUR_OUTS(qp->sq_res));
      MOSAL_spinlock_unlock(&(qp->sq_res.q_lock));
    return HH_E2BIG_WR_NUM;
  }

  /* Allocate next WQE */
  next_wqe= (u_int32_t*)(qp->sq_res.wqe_buf +
                        (qp->sq_res.next2post_index << qp->sq_res.log2_max_wqe_sz) );
  cur_loc_p= (u_int8_t*)next_wqe; 
  wqe_edge_p= ((u_int8_t*)next_wqe)+(1<<qp->sq_res.log2_max_wqe_sz);

  cur_loc_p+= WQE_init_next((u_int32_t*)cur_loc_p); /* Make "unlinked" "next" segment */
  cur_loc_p+= WQE_pack_ctrl_send((u_int32_t*)cur_loc_p,  /* Pack Control segment */
    send_req_p->comp_type, send_req_p->set_se, 0/*event bit*/,
    ((send_req_p->opcode == VAPI_RDMA_WRITE_WITH_IMM) ||
     (send_req_p->opcode == VAPI_SEND_WITH_IMM) ) ? send_req_p->imm_data : 0);

  /* Transport type checks: Datagram segment */
  switch (qp->ts_type) {
    case VAPI_TS_UD:  /* Check if UD (UD datagram segment) */
      cur_loc_p+= WQE_pack_ud((u_int32_t*)cur_loc_p,
        qp,send_req_p->remote_ah,
        send_req_p->remote_qp,send_req_p->remote_qkey);
      break;
    case VAPI_TS_RD:  /* Check if RD (RD datagram segment) */
      cur_loc_p+= WQE_pack_rd((u_int32_t*)cur_loc_p,
        send_req_p->remote_qp,send_req_p->remote_qkey);
      break;
    default:
      break;
  }
  
  /* Opcode checks + Remote-address/Atomic segments */
  switch (send_req_p->opcode) {
    /* For RDMA operations: only Remote-address segment */
    case VAPI_RDMA_WRITE:
    case VAPI_RDMA_WRITE_WITH_IMM:
      cur_loc_p+= WQE_pack_remote_addr((u_int32_t*)cur_loc_p,
        send_req_p->remote_addr,send_req_p->r_key);
      break;
    
    case VAPI_SEND:
    case VAPI_SEND_WITH_IMM:
      break; /* Valid opcodes for "inline" but no extra WQE segment */
    default: 
      MOSAL_spinlock_unlock(&(qp->sq_res.q_lock));
      return HH_EINVAL_OPCODE; /* Invalid opcode */
  }
  
  inline_p= (u_int32_t*)cur_loc_p;
  cur_loc_p+= WQE_INLINE_SZ_BCOUNT;
  /* copy inline data to WQE */
  for (i= 0; i < send_req_p->sg_lst_len; i++) {
    if ((cur_loc_p+send_req_p->sg_lst_p[i].len) > wqe_edge_p) {
      MOSAL_spinlock_unlock(&(qp->sq_res.q_lock));
      MTL_ERROR2(MT_FLFMT("too much inline data for inline send request (qpn=0x%X)"),qp->qpn);
      return HH_EINVAL_SG_NUM;
    }
    if (send_req_p->sg_lst_p[i].addr > (MT_virt_addr_t)MAKE_ULONGLONG(0xFFFFFFFFFFFFFFFF)) {
        MOSAL_spinlock_unlock(&(qp->sq_res.q_lock));
        MTL_ERROR2(MT_FLFMT("sg list addr %d has non-zero upper bits (qpn=0x%X, addr="U64_FMT") \n"),
               i,qp->qpn,send_req_p->sg_lst_p[i].addr );
       return HH_EINVAL_SG_FMT;
    }
    memcpy(cur_loc_p, (void*)(MT_virt_addr_t)(send_req_p->sg_lst_p[i].addr),
           send_req_p->sg_lst_p[i].len); 
    cur_loc_p+= send_req_p->sg_lst_p[i].len;
  }
  *inline_p= 
    (u_int32_t)(0x80000000 | (((MT_virt_addr_t)cur_loc_p) - ((MT_virt_addr_t)inline_p) - 4)); /*inline:size*/

  wqe_sz_dwords= (MT_UP_ALIGNX_U32( (u_int32_t)(((MT_virt_addr_t)cur_loc_p) - ((MT_virt_addr_t)next_wqe)),
                                WQE_SZ_MULTIPLE_SHIFT) >> 2); 
#ifdef MAX_DEBUG
  if ((wqe_sz_dwords<<2) > (1U << qp->sq_res.log2_max_wqe_sz)) {
    MTL_ERROR1(MT_FLFMT("QP 0x%X: Send WQE too large (%d > max=%d) !!!!!!!!!"),
      qp->qpn,(wqe_sz_dwords<<2),(1U << qp->sq_res.log2_max_wqe_sz));
  	}
#endif

  rc= sq_alloc_wqe_link_and_ring(qp,next_wqe,wqe_sz_dwords,
#ifdef MT_LITTLE_ENDIAN
           (u_int32_t)(inline_p - next_wqe + 1), /* swap all up to data */
#endif
           send_req_p,encode_nopcode(send_req_p->opcode)); 
  MOSAL_spinlock_unlock(&(qp->sq_res.q_lock));
  return rc;
}

static inline HH_ret_t XHHUL_qpm_post_send_reqs_internal( 
  /*IN*/ HHUL_hca_hndl_t hca, 
  /*IN*/ XHHUL_qp_t qp, 
  /*IN*/ u_int32_t num_of_requests,
  /*IN*/ VAPI_sr_desc_t *send_req_array,
  /*IN*/ MT_bool do_sgsize_check
)
{
  u_int32_t next_draft[WQE_SEG_SZ_NEXT>>2]; /* Build "next" segment here */
  volatile u_int32_t* next_wqe= NULL;
  volatile u_int32_t* prev_wqe_p= NULL; 
  MT_virt_addr_t first_wqe_nda= 0;
  u_int32_t first_wqe_nds= 0;
  u_int32_t wqe_sz_dwords,i;
  u_int32_t next2post_index,reqi;
  XHH_uar_sendq_dbell_t sq_dbell;
  MT_bool last_qkey_from_wqe;
  ADD_DEV_PREFIX_LC(if_nopcode_t) first_nopcode=0, adjusted_nopcode=0;
  
  last_qkey_from_wqe = qp->last_qkey_from_wqe;
  /* Check for available WQEs */
  /* We hold this value on a seperate var. for easy rollback in case of an error */
  next2post_index= qp->sq_res.next2post_index;
  last_qkey_from_wqe = qp->last_qkey_from_wqe;

  /* Build and link all WQEs */
  for (reqi= 0; reqi < num_of_requests; reqi++) {
    if (do_sgsize_check && (qp->sq_res.max_sg_sz < send_req_array[reqi].sg_lst_len)) {
      MTL_ERROR2(
                "%s: S/G list of request %d is too large (%d entries > max_sg_sz=%d)\n", __func__,
                reqi,send_req_array[reqi].sg_lst_len,qp->sq_res.max_sg_sz);
      return HH_EINVAL_SG_NUM;
    }

    /* Allocate next WQE */
    next_wqe= (u_int32_t*)(qp->sq_res.wqe_buf + 
                          (next2post_index << qp->sq_res.log2_max_wqe_sz) );

    MTPERF_TIME_START(WQE_build_send);
    wqe_sz_dwords= (WQE_build_send(qp,send_req_array+reqi,next_wqe) >> 2);
    MTPERF_TIME_END(WQE_build_send);

    /* XHH_DEV_SPECIFIC_POST_SEND_REQ; */

#ifdef MAX_DEBUG
    if ((wqe_sz_dwords<<2) > (1U << qp->sq_res.log2_max_wqe_sz)) {
      MTL_ERROR1(MT_FLFMT("QP 0x%X: Send WQE too large (%d > max=%d)"),
                 qp->qpn,(wqe_sz_dwords<<2),(1U << qp->sq_res.log2_max_wqe_sz));
    	}
#endif
    qp->sq_res.wqe_id[next2post_index]= send_req_array[reqi].id;  /* Save WQE ID */
    next2post_index = (next2post_index + 1) % qp->sq_res.max_outs ;
#ifdef MT_LITTLE_ENDIAN
  /* swap WQE's dwords */
    for (i= WQE_SND_DRAFT_START_COPY; i < wqe_sz_dwords; i++) {
      next_wqe[i]= MOSAL_cpu_to_be32(next_wqe[i]);
    }
#endif

    if (reqi == 0) { /* For the first WQE save info for linking it later */
      first_wqe_nda= (MT_virt_addr_t)next_wqe;
      first_wqe_nds= (wqe_sz_dwords>>2);
      first_nopcode = adjust_nopcode_and_last_qkey_flag(qp, 
			    (u_int32_t) (((u_int32_t) (send_req_array[0].remote_qkey)) & 0x80000000), 
			    encode_nopcode(send_req_array[0].opcode), &last_qkey_from_wqe);
    
    } else { /* Not first - link to previous with DBD=0 */
      /* Build linking "next" segment in last posted WQE*/
	adjusted_nopcode = adjust_nopcode_and_last_qkey_flag(qp, 
			    (u_int32_t) (((u_int32_t) (send_req_array[reqi].remote_qkey)) & 0x80000000), 
			    encode_nopcode(send_req_array[reqi].opcode), &last_qkey_from_wqe);
      WQE_pack_send_next(next_draft, adjusted_nopcode, 
        send_req_array[reqi].fence,0/*DBD*/, (u_int32_t)(MT_ulong_ptr_t)next_wqe, wqe_sz_dwords>>2, 
        (qp->ts_type==VAPI_TS_RD) ? send_req_array[reqi].eecn : 0 );
      for (i= 0;i < (WQE_SEG_SZ_NEXT>>2) ;i++) {  
        /* This copy assures big-endian as well as that NDS is written last */
        prev_wqe_p[i]= MOSAL_cpu_to_be32(next_draft[i]);
      }
    }
    prev_wqe_p= next_wqe;

  }
  
  XHH_DEV_SPECIFIC_SQ_TEST_PREV_WQE_NOT_NULL  { /* link chain to previous WQE */
    /* Build linking "next" segment with DBD set */
    WQE_pack_send_next(next_draft, first_nopcode, 
      send_req_array[0].fence,1/*DBD*/, (u_int32_t)first_wqe_nda, first_wqe_nds, 
      (qp->ts_type==VAPI_TS_RD) ? send_req_array[0].eecn : 0 );
    for (i= 0;i < (WQE_SEG_SZ_NEXT>>2) ;i++) {  
      /* This copy assures big-endian as well as that NDS is written last */
      qp->sq_res.last_posted_p[i]= MOSAL_cpu_to_be32(next_draft[i]);
    }
  }
  
  /* Update QP status */
  qp->sq_res.last_posted_p= next_wqe; 
  qp->sq_res.next2post_index= next2post_index;
  qp->last_qkey_from_wqe = last_qkey_from_wqe;

  /* Ring doorbell (send or rd-send) */
  sq_dbell.nopcode= first_nopcode;
  sq_dbell.fence= send_req_array[0].fence;
  sq_dbell.next_size= (SENDQ_DBELL_NEXT_SIZE_TYPE)first_wqe_nds;
  
  XHH_DEV_SPECIFIC_POST_SEND_REQS_INTERNAL;

  if (qp->ts_type == VAPI_TS_RD) {
    uar_sendq_rd_dbell(qp, &sq_dbell,&(send_req_array[0]));
  } else {  /* non-RD send request */
    uar_sendq_dbell(qp,&sq_dbell);
  }

  return HH_OK;
}


HH_ret_t XHHUL_qpm_post_send_reqs( 
  /*IN*/ HHUL_hca_hndl_t hca, 
  /*IN*/ HHUL_qp_hndl_t hhul_qp, 
  /*IN*/ u_int32_t num_of_requests,
  /*IN*/ VAPI_sr_desc_t *send_req_array 
)
{
  XHHUL_qp_t qp= (XHHUL_qp_t)hhul_qp;
  MT_bool do_sgsize_check = TRUE;
  HH_ret_t ret = HH_OK;

  if (qp->sqp_type != VAPI_REGULAR_QP) {
    MTL_ERROR4(MT_FLFMT("XHHUL_qpm_post_send_reqs is not supporeted for special QPs"));
    return HH_ENOSYS;
  }
  
  if (num_of_requests == 0) {
    MTL_ERROR4(MT_FLFMT("XHHUL_qpm_post_send_reqs: num_of_requeusts=0 !"));
    return HH_EINVAL_PARAM;
  }
  
  if (!is_qpstate_valid_2send(qp->sq_res.qp_state)) {
   MTL_ERROR1(MT_FLFMT("%s failed: qp state %d not valid to send \n"),__func__,qp->sq_res.qp_state);
   return HH_EINVAL_QP_STATE;
 }

  MOSAL_spinlock_irq_lock(&(qp->sq_res.q_lock)); /* protect WQE allocation/link */
  
  /* Check for available WQEs */
  if (qp->sq_res.max_outs <= (CUR_OUTS(qp->sq_res) + num_of_requests)) {
    MTL_ERROR4("XHHUL_qpm_post_send_reqs: Not enough WQEs for %u requests (%u requests outstanding).\n",
               num_of_requests,CUR_OUTS(qp->sq_res));
    MOSAL_spinlock_unlock(&(qp->sq_res.q_lock));
    return HH_E2BIG_WR_NUM;
  }

  XHH_DEV_SPECIFIC_POST_SEND_REQS;
 
  MOSAL_spinlock_unlock(&(qp->sq_res.q_lock));
  return ret;
}

HH_ret_t XHHUL_qpm_post_gsi_send_req( 
   HHUL_hca_hndl_t hca, 
   HHUL_qp_hndl_t hhul_qp, 
   VAPI_sr_desc_t *send_req_p,
   VAPI_pkey_ix_t pkey_index
) 
{ 
  XHHUL_qp_t qp= (XHHUL_qp_t)hhul_qp;
  u_int32_t* next_wqe;
  u_int32_t wqe_sz_dwords;
  HH_hca_hndl_t hh_hca_hndl;
  HH_ret_t rc;

  if (qp->sqp_type != VAPI_GSI_QP)  {
    MTL_ERROR2(MT_FLFMT("Invoked for non-GSI QP (qpn=0x%X)"),qp->qpn);
    return HH_EINVAL_QP_NUM;
  }
  
  if (qp->sq_res.max_sg_sz < send_req_p->sg_lst_len) {
    MTL_ERROR2(
      "%s: Scatter/Gather list is too large (%d entries > max_sg_sz=%d)\n",__func__,
      send_req_p->sg_lst_len,qp->sq_res.max_sg_sz);
    return HH_EINVAL_SG_NUM;
  }
   
  if (!is_qpstate_valid_2send(qp->sq_res.qp_state)) {
   MTL_ERROR1(MT_FLFMT("%s failed: qp state %d not valid to send \n"),__func__,qp->sq_res.qp_state);
   return HH_EINVAL_QP_STATE;
 }

#ifdef DUMP_SEND_REQ 
  dump_send_req(qp,send_req_p);
#endif
  
  send_req_p->fence= FALSE; /* required for MLX requests */
  rc= XHHUL_hob_get_hh_hca_hndl(hca,&hh_hca_hndl);
  if (rc != HH_OK) {
    return HH_EINVAL_HCA_HNDL;
  }

  MOSAL_spinlock_irq_lock(&(qp->sq_res.q_lock)); /* protect WQE allocation/link */
  
   /* Check if any WQEs are free to be consumed */
  if (qp->sq_res.max_outs <= (CUR_OUTS(qp->sq_res)+1))  {
    MTL_ERROR4("%s: Send queue is full (%u requests outstanding).\n", __func__,
      CUR_OUTS(qp->sq_res));
      MOSAL_spinlock_unlock(&(qp->sq_res.q_lock));
    return HH_E2BIG_WR_NUM;
  }

  /* Allocate next WQE */
  next_wqe= (u_int32_t*)(qp->sq_res.wqe_buf +
                        (qp->sq_res.next2post_index << qp->sq_res.log2_max_wqe_sz) );

  wqe_sz_dwords= 
    (WQE_build_send_mlx(hh_hca_hndl,qp,send_req_p,pkey_index,next_wqe) >> 2);
  if (wqe_sz_dwords == 0) {
    MOSAL_spinlock_unlock(&(qp->sq_res.q_lock)); 
    MTL_ERROR1(MT_FLFMT("Failed building MLX headers for special QP.\n"));
    return HH_EINVAL_WQE;
  }

  rc= sq_alloc_wqe_link_and_ring(qp,next_wqe,wqe_sz_dwords,
#ifdef MT_LITTLE_ENDIAN
           wqe_sz_dwords,
#endif
           send_req_p,encode_nopcode(send_req_p->opcode)); 
  
  MOSAL_spinlock_unlock(&(qp->sq_res.q_lock)); 
  return rc;
}


HH_ret_t XHHUL_qpm_post_recv_req( 
   HHUL_hca_hndl_t hca, 
   HHUL_qp_hndl_t hhul_qp, 
   VAPI_rr_desc_t *recv_req_p 
) 
{ 
  XHHUL_qp_t qp= (XHHUL_qp_t)hhul_qp;
  volatile u_int32_t* next_wqe; /* Actual WQE pointer */
  u_int32_t i, wqe_sz_dwords;

  if (qp->srq != HHUL_INVAL_SRQ_HNDL) {
    MTL_ERROR1(MT_FLFMT("%s: Used for QP 0x%X which is associated with SRQ handle 0x%p"), __func__,
               qp->qpn, qp->srq);
    return HH_EINVAL_SRQ_HNDL;
  }

  if (!is_qpstate_valid_2recv(qp->rq_res.qp_state)) {
   MTL_ERROR1(MT_FLFMT("%s failed: qp state %d not valid to recv \n"),__func__,qp->rq_res.qp_state);
   return HH_EINVAL_QP_STATE;
 }

  if (qp->rq_res.max_sg_sz < recv_req_p->sg_lst_len) {
    MTL_ERROR2(
      "XHHUL_qpm_post_recv_req: Scatter/Gather list is too large (%d entries > max_sg_sz=%d)\n",
      recv_req_p->sg_lst_len,qp->rq_res.max_sg_sz);
    return HH_EINVAL_SG_NUM;
  }
   
  MOSAL_spinlock_irq_lock(&(qp->rq_res.q_lock)); /* protect WQE allocation/link */

   /* Check if any WQEs are free to be consumed */
  if (qp->rq_res.max_outs <= (CUR_OUTS(qp->rq_res) + 1)) {
    MOSAL_spinlock_unlock(&(qp->rq_res.q_lock));
    MTL_ERROR4("XHHUL_qpm_post_recv_req: Receive queue is full (%d requests outstanding).\n",
      CUR_OUTS(qp->rq_res));
    return HH_E2BIG_WR_NUM;
  }
  /* Allocate next WQE */
  next_wqe= (u_int32_t*) (qp->rq_res.wqe_buf + 
                          (qp->rq_res.next2post_index << qp->rq_res.log2_max_wqe_sz) );
 
  /* Build WQE */
  wqe_sz_dwords= (WQE_build_recv(qp,recv_req_p,next_wqe) >> 2);
#ifdef MAX_DEBUG
    if ((wqe_sz_dwords<<2) > (1U << qp->rq_res.log2_max_wqe_sz)) {
      MTL_ERROR1(MT_FLFMT("QP 0x%X: Receive WQE too large (%d > max=%d)"),
        qp->qpn,(wqe_sz_dwords<<2),(1U << qp->rq_res.log2_max_wqe_sz));
    	}
#endif

  qp->rq_res.wqe_id[qp->rq_res.next2post_index]= recv_req_p->id;  /* Save WQE ID */
  qp->rq_res.next2post_index = (qp->rq_res.next2post_index + 1) % qp->rq_res.max_outs ;
  
#ifdef MT_LITTLE_ENDIAN
  /* swap WQE's dwords */
  for (i= WQE_RCV_DRAFT_START_COPY; i < wqe_sz_dwords; i++) {
    next_wqe[i]= MOSAL_cpu_to_be32(next_wqe[i]);
  }
#endif
  
  XHH_DEV_SPECIFIC_POST_RECV_REQ;

  MOSAL_spinlock_unlock(&(qp->rq_res.q_lock));
  return HH_OK;
  
}


HH_ret_t XHHUL_qpm_post_recv_reqs(
                                 /*IN*/ HHUL_hca_hndl_t hca, 
                                 /*IN*/ HHUL_qp_hndl_t hhul_qp, 
                                 /*IN*/ u_int32_t num_of_requests,
                                 /*IN*/ VAPI_rr_desc_t *recv_req_array 
                                 )
{
  XHHUL_qp_t qp= (XHHUL_qp_t)hhul_qp;
  volatile u_int32_t* next_wqe= NULL; /* Actual WQE pointer */
  u_int32_t wqe_sz_dwords= 0;
  u_int32_t i,reqi,next2post_index;
  XHHUL_DEV_SPECIFIC_QPM_POST_RECV_REQS_VARS;
  
  if (qp->srq != HHUL_INVAL_SRQ_HNDL) {
    MTL_ERROR1(MT_FLFMT("%s: Used for QP 0x%X which is associated with SRQ 0x%p"), __func__,
               qp->qpn, qp->srq);
    return HH_EINVAL_SRQ_HNDL;
  }

  if (!is_qpstate_valid_2recv(qp->rq_res.qp_state)) {
    MTL_ERROR1(MT_FLFMT("%s failed: qp state %d not valid to recv \n"),__func__,qp->rq_res.qp_state);
    return HH_EINVAL_QP_STATE;
  }

  if (num_of_requests == 0) {
    MTL_ERROR4(MT_FLFMT("XHHUL_qpm_post_recv_reqs: num_of_requeusts=0 !"));
    return HH_EINVAL_PARAM;
  }

  /* Check parameters of all WQEs first - must assure all posted successfully */
  for (reqi= 0; reqi < num_of_requests; reqi++) {
    if (qp->rq_res.max_sg_sz < recv_req_array[reqi].sg_lst_len) {
      MTL_ERROR2(
        "XHHUL_qpm_post_recv_reqs: S/G list of req. #%d is too large (%d entries > max_sg_sz=%d)\n",
                reqi,recv_req_array[reqi].sg_lst_len,qp->rq_res.max_sg_sz);
      return HH_EINVAL_SG_NUM;
    }
  }

  MOSAL_spinlock_irq_lock(&(qp->rq_res.q_lock)); /* protect WQE allocation/link */
  
  /* Check for available WQEs */
  if (qp->rq_res.max_outs <= (CUR_OUTS(qp->rq_res) + num_of_requests)) {
    MTL_ERROR4("XHHUL_qpm_post_recv_reqs: Not enough WQEs for %d requests (%d requests outstanding).\n",
               num_of_requests,CUR_OUTS(qp->rq_res));
    MOSAL_spinlock_unlock(&(qp->rq_res.q_lock));
    return HH_E2BIG_WR_NUM;
  }

  XHHUL_DEV_SPECIFIC_QPM_POST_RECV_REQS_INIT;

  /* We hold this value on a seperate var. for easy rollback in case of an error */
  next2post_index= qp->rq_res.next2post_index;

  /* Build and link and ring all WQEs */
  for (reqi= 0; reqi < num_of_requests; reqi++) {

     /* Allocate next WQE */
    next_wqe= (u_int32_t*) (qp->rq_res.wqe_buf + 
                            (next2post_index << qp->rq_res.log2_max_wqe_sz) );
   
    /* Build WQE */
    wqe_sz_dwords= (WQE_build_recv(qp,recv_req_array+reqi,next_wqe) >> 2);
  #ifdef MAX_DEBUG
    if ((wqe_sz_dwords<<2) > (1U << qp->rq_res.log2_max_wqe_sz)) {
      MTL_ERROR1(MT_FLFMT("QP 0x%X: Receive WQE too large (%d > max=%d)"),
                 qp->qpn,(wqe_sz_dwords<<2),(1U << qp->rq_res.log2_max_wqe_sz));
    	}
  #endif
    
    qp->rq_res.wqe_id[next2post_index]= recv_req_array[reqi].id;  /* Save WQE ID */
    next2post_index = (next2post_index + 1) % qp->rq_res.max_outs ;

#ifdef MT_LITTLE_ENDIAN
  /* swap WQE's dwords */
    for (i= WQE_RCV_DRAFT_START_COPY; i < wqe_sz_dwords; i++) {
      next_wqe[i]= MOSAL_cpu_to_be32(next_wqe[i]);
    }
#endif

    XHHUL_DEV_SPECIFIC_QPM_POST_RECV_REQS_LOOP;

  }

  XHHUL_DEV_SPECIFIC_QPM_POST_RECV_REQS_DONE;

  /* update producer index + cur. outstanding  (now that no error was found) */
  qp->rq_res.next2post_index = next2post_index;
  
  MOSAL_spinlock_unlock(&(qp->rq_res.q_lock));
  return HH_OK;
}



HH_ret_t XHHUL_qpm_post_bind_req(
  /*IN*/ HHUL_mw_bind_t *bind_props_p,
  /*IN*/ IB_rkey_t new_rkey
)
{
  XHHUL_qp_t qp= (XHHUL_qp_t)bind_props_p->qp;
  volatile u_int32_t* next_wqe= NULL; /* Actual WQE pointer */
  u_int32_t wqe_sz_dwords;
  VAPI_sr_desc_t send_req;
  HH_ret_t rc;

  if ((qp->sqp_type != VAPI_REGULAR_QP) ||
      ((qp->ts_type != VAPI_TS_RC) && (qp->ts_type != VAPI_TS_RD) && (qp->ts_type != VAPI_TS_UC))){
    MTL_ERROR1(MT_FLFMT("Invalid QP type for binding memory windows (qpn=0x%X)."),qp->qpn);
    return HH_EINVAL_SERVICE_TYPE;
  }
  
  
  if (!is_qpstate_valid_2send(qp->sq_res.qp_state)) {
    MTL_ERROR1(MT_FLFMT("%s failed: qp state %d not valid to send \n"),__func__,qp->sq_res.qp_state);
    return HH_EINVAL_QP_STATE;
  }

  MOSAL_spinlock_irq_lock(&(qp->sq_res.q_lock)); /* protect WQE allocation/link */

   /* Check if any WQEs are free to be consumed */
  if (qp->sq_res.max_outs <= (CUR_OUTS(qp->sq_res)+1))  {
    MTL_ERROR4("%s: Send queue is full (%u requests outstanding).\n", __func__,
      CUR_OUTS(qp->sq_res));
      MOSAL_spinlock_unlock(&(qp->sq_res.q_lock));
    return HH_E2BIG_WR_NUM;
  }

  /* Allocate next WQE */
  next_wqe= (u_int32_t*)(qp->sq_res.wqe_buf +
                        (qp->sq_res.next2post_index << qp->sq_res.log2_max_wqe_sz) );
  
  wqe_sz_dwords= (WQE_build_membind(bind_props_p,new_rkey,next_wqe) >> 2);
#ifdef MAX_DEBUG
  if ((wqe_sz_dwords<<2) > (1U << qp->sq_res.log2_max_wqe_sz)) {
    MTL_ERROR1(MT_FLFMT("QP 0x%X: Send WQE too large (%d > max=%d)"),
      qp->qpn,(wqe_sz_dwords<<2),(1U << qp->sq_res.log2_max_wqe_sz));
  	}
#endif

  send_req.id= bind_props_p->id;
  send_req.fence= TRUE; /* just in case, though implicitly fenced */
  if (qp->ts_type == VAPI_TS_RD) {
    send_req.eecn= RESERVED_MEMBIND_EECN;
  }

  rc= sq_alloc_wqe_link_and_ring(qp,next_wqe,wqe_sz_dwords,
#ifdef MT_LITTLE_ENDIAN
           wqe_sz_dwords,
#endif
           &send_req,ADD_DEV_PREFIX(IF_NOPCODE_BIND_MEMWIN));

  MOSAL_spinlock_unlock(&(qp->sq_res.q_lock));
  return rc;
}



HH_ret_t XHHUL_qpm_comp_ok( 
  XHHUL_qpm_t qpm, 
  XHHUL_qp_t qp,
  u_int32_t wqe_addr_32lsb, 
  MT_bool is_SQ_completion,
  VAPI_special_qp_t *qp_type_p,
  IB_ts_t *qp_ts_type_p,
  VAPI_wr_id_t *wqe_id_p,
  u_int32_t *wqes_released_p
) 
{ 
  u_int32_t freed_wqe_index;
  queue_res_t *associated_q= NULL;
  HH_ret_t rc;

  rc= find_wqe(qpm,qp,wqe_addr_32lsb,is_SQ_completion,
	       &associated_q,&freed_wqe_index,wqe_id_p);
  if (rc != HH_OK) {
    MTL_ERROR2("%s: Given QPN/WQE is not associated with any queue (qpn=0x%X,wqe=0x%X).\n", 
               __func__,qp->qpn,wqe_addr_32lsb);
    return HH_EINVAL;
  }

  if ( (qp->ts_type == IB_TS_RD) && (qp->sqp_type == VAPI_REGULAR_QP) ) {
    /* RD is a completely different story due to out of order completion */
    MTL_ERROR4("XHHUL_qpm_comp_ok: RD WQEs tracking not supported, yet.\n");
    return HH_ENOSYS; /* TBD: implement when XHH should support RD */
  }
  
  *qp_type_p= qp->sqp_type;
  *qp_ts_type_p= qp->ts_type;

  if (associated_q != NULL) { /* Release WQEs (if not from SRQ) */
    /* rc is HH_OK from find_wqe above */
    *wqes_released_p= 
      (associated_q->next2free_index <= freed_wqe_index) ? 
      /* Unsigned computation depends on cycic indecies relation (who is the upper index) */
        1+ freed_wqe_index - associated_q->next2free_index :
        1+ associated_q->max_outs - (associated_q->next2free_index - freed_wqe_index);
      /* The +1 results from the fact that next2free_index should be counted as well */
    associated_q->next2free_index= (freed_wqe_index + 1) % associated_q->max_outs;
  } else {
#ifndef NO_UL_SRQM
    rc= XHHUL_srqm_comp(qpm->srqm, qp->srq, wqe_addr_32lsb, wqe_id_p);
#else
    rc = HH_ERR;
#endif
    if (rc != HH_OK) {
        MTL_DEBUG2(MT_FLFMT("%s: Failed to find WQE in SRQ (WQE=0x%X QPn=0x%X)"), __func__,
               wqe_addr_32lsb, qp->qpn);
    }
  }

  return rc;
}


HH_ret_t XHHUL_qpm_comp_err( 
  XHHUL_qpm_t qpm, 
  XHHUL_qp_t qp,
  u_int32_t wqe_addr_32lsb, 
  MT_bool is_SQ_completion,
  VAPI_wr_id_t *wqe_id_p,
  u_int32_t *wqes_released_p, 
  u_int32_t *next_wqe_32lsb_p,
  u_int8_t  *dbd_bit_p
) 
{ 
  u_int32_t freed_wqe_index;
  queue_res_t *associated_q;
  u_int32_t* completed_wqe;
  HH_ret_t rc;

  rc= find_wqe(qpm,qp,wqe_addr_32lsb,is_SQ_completion,
	       &associated_q,&freed_wqe_index,wqe_id_p);
  if (rc != HH_OK) {
    MTL_ERROR2(
      "%s: Given QPN/WQE is not associated with any queue (qpn=0x%X,wqe=0x%X).\n",__func__,
      qp->qpn,wqe_addr_32lsb);
    return HH_EINVAL;
  }

  if ( (qp->ts_type == IB_TS_RD) && (qp->sqp_type == VAPI_REGULAR_QP) ) {
    /* RD is a completely different story due to out of order completion */
    MTL_ERROR4("%s: RD WQEs tracking not supported, yet.\n", __func__);
    return HH_ENOSYS; /* TBD: implement when XHH should support RD */
  }

  if (associated_q != NULL) { /* Not from SRQ */
    /* rc is HH_OK from find_wqe above */
    if (sizeof(MT_virt_addr_t) <= 4) { /* Optimization for 32bit machines */
      completed_wqe= (u_int32_t*)(MT_virt_addr_t) wqe_addr_32lsb;
    } else {
      completed_wqe= (u_int32_t*)(MT_virt_addr_t)
        (((associated_q->wqe_buf) & MAKE_ULONGLONG(0xFFFFFFFF00000000)) | (u_int64_t)wqe_addr_32lsb );
    }
    if (WQE_extract_nds(completed_wqe) == 0) {
      *next_wqe_32lsb_p= XHHUL_QPM_END_OF_WQE_CHAIN;  /* Chain end reached */
    } else {
      *next_wqe_32lsb_p= WQE_extract_nda(completed_wqe);
    }
    *dbd_bit_p= WQE_extract_dbd(completed_wqe);
    
    *wqes_released_p= 
      (associated_q->next2free_index <= freed_wqe_index) ? 
      /* Unsigned computation depends on cycic indecies relation (who is the upper index) */
        1+ freed_wqe_index - associated_q->next2free_index :
        1+ associated_q->max_outs - (associated_q->next2free_index - freed_wqe_index);
      /* The +1 results from the fact that next2free_index should be counted as well */
    associated_q->next2free_index= (freed_wqe_index + 1) % associated_q->max_outs;
  
  } else { /* SRQ - all WQEs generate CQEs... no need to provide NDA */
#ifndef NO_UL_SRQM
    rc= XHHUL_srqm_comp(qpm->srqm, qp->srq, wqe_addr_32lsb, wqe_id_p);
#else
    rc = HH_ERR;
#endif
    if (rc != HH_OK) {
        MTL_DEBUG2(MT_FLFMT("%s: Failed to find WQE in SRQ (WQE=0x%X QPn=0x%X)"), __func__,
               wqe_addr_32lsb, qp->qpn);
    } else {
	*wqes_released_p= 1;
	*next_wqe_32lsb_p= XHHUL_QPM_END_OF_WQE_CHAIN;  /* Chain end reached */
    }
  }

  return rc;
}

u_int32_t XHHUL_qpm_wqe_cnt( 
  /*IN*/XHHUL_qpm_t qpm, 
  /*IN*/XHHUL_qp_t qp, 
  /*IN*/u_int32_t wqe_addr_32lsb, 
  /*IN*/u_int16_t dbd_cnt,
  /*IN*/MT_bool is_SQ_completion)
{
  u_int32_t cur_wqe_index;
  queue_res_t *associated_q;
  volatile u_int32_t *cur_wqe_p;
  u_int32_t wqe_cntr= 0;
  VAPI_wr_id_t wqe_id;
  HH_ret_t rc;

  rc= find_wqe(qpm,qp,wqe_addr_32lsb,is_SQ_completion,
	       &associated_q,&cur_wqe_index,&wqe_id);
  if (rc != HH_OK) {
    MTL_ERROR2(
      "%s: Given QPN/WQE is not associated with any queue (qpn=%d,wqe=0x%X).\n",__func__,
      qp->qpn,wqe_addr_32lsb);
    return 0;
  }

  if ( (qp->ts_type == IB_TS_RD) && (qp->sqp_type == VAPI_REGULAR_QP) ) {
    /* RD is a completely different story due to out of order completion */
    MTL_ERROR4("%s: RD WQEs tracking not supported, yet.\n",__func__);
    return 0; /* TBD: implement when XHH should support RD */
  }

  if (associated_q == NULL) { /* SRQ */
    /* Only one WQE per CQE for SRQs */
    return 1;
  }

  dbd_cnt++;  /* count down to zero (dbd_cnt==0 when waiting for next dbd bit set) */
  MOSAL_spinlock_irq_lock(&(associated_q->q_lock));
  do {
    wqe_cntr++;
    cur_wqe_p= (u_int32_t*)(associated_q->wqe_buf + 
                          (cur_wqe_index << associated_q->log2_max_wqe_sz) );
    dbd_cnt-= WQE_extract_dbd(cur_wqe_p);
    cur_wqe_index= (cur_wqe_index + 1) % associated_q->max_outs;
  } while ((dbd_cnt > 0) && (WQE_extract_nds(cur_wqe_p) != 0));
  MOSAL_spinlock_unlock(&(associated_q->q_lock));

  return wqe_cntr;
}

/**********************************************************************************************
 *                    Private Functions
 **********************************************************************************************/

/* 
 * Prepare QP resources before creation.
 * To be used by both XHH_qpm_create_qp_prep and XHH_qpm_special_qp_prep
 */
static HH_ret_t qp_prep(
  HHUL_hca_hndl_t hca, 
  VAPI_special_qp_t qp_type, 
  HHUL_qp_init_attr_t *qp_init_attr_p, 
  HHUL_qp_hndl_t *qp_hndl_p, 
  VAPI_qp_cap_t *qp_cap_out_p, 
  XHH_qp_ul_resources_t *qp_ul_resources_p,
  MT_bool in_ddr_mem  /* WQEs buffer allocated in attached DDR mem. or in main memory */
)
{
  XHHUL_qpm_t qpm;
  XHH_hca_ul_resources_t *hca_ul_res_p = NULL;
  XHHUL_qp_t new_qp;
  HH_ret_t rc;
  call_result_t mrc;
  
  rc= XHHUL_hob_get_qpm(hca,&qpm);
  if (rc != HH_OK) {
    MTL_ERROR4(MT_FLFMT("qp_prep: Invalid HCA handle (%p)."),hca);
    return HH_EINVAL;
  }
  rc= XHHUL_hob_get_hca_ul_res_p(hca,&hca_ul_res_p);
  if (rc != HH_OK) {
    MTL_ERROR2(MT_FLFMT("qp_prep: Failed XHHUL_hob_get_hca_ul_res (err=%d).\n"),rc);
    return rc;
  }
  
  (new_qp)= (XHHUL_qp_t)MALLOC(sizeof(struct XHHUL_qp_st));
  if (new_qp == NULL) {
    MTL_ERROR1(MT_FLFMT("qp_prep: Failed allocating XHHUL_qp_t.\n"));
    return HH_EAGAIN;
  }
  memset(new_qp,0,sizeof(struct XHHUL_qp_st));

  new_qp->sqp_type= qp_type;
  
  rc= init_qp(hca,qp_init_attr_p,new_qp, qp_ul_resources_p);
  if (rc != HH_OK) {
    goto failed_init_qp;
  }

  rc= alloc_wqe_buf(qpm, in_ddr_mem, hca_ul_res_p->max_qp_ous_wr, hca_ul_res_p->max_num_sg_ent,
					hca_ul_res_p->max_wqe_size_sq, hca_ul_res_p->max_wqe_size_rq, 
					qp_init_attr_p->qp_mem_key,
					new_qp,qp_ul_resources_p);
  if (rc != HH_OK) {
    MTL_ERROR1(MT_FLFMT(": Failed allocating WQEs buffers.\n"));
    goto failed_alloc_wqe;
  }

  rc= alloc_aux_data_buf(new_qp);
  if (rc != HH_OK) {
    MTL_ERROR1(MT_FLFMT(": Failed allocating auxilary buffers.\n"));
    goto failed_alloc_aux;
  }
  
  /* Set output modifiers */
  *qp_hndl_p= new_qp;
  qp_cap_out_p->max_oust_wr_rq= new_qp->rq_res.max_outs ? 
	  new_qp->rq_res.max_outs - 1 : 0; /* One "reserved" WQE */
  qp_cap_out_p->max_oust_wr_sq= new_qp->sq_res.max_outs ? 
	  new_qp->sq_res.max_outs - 1 : 0; /* One "reserved" WQE */
  qp_cap_out_p->max_sg_size_rq= new_qp->rq_res.max_sg_sz;
  qp_cap_out_p->max_sg_size_sq= new_qp->sq_res.max_sg_sz;
  qp_cap_out_p->max_inline_data_sq= new_qp->sq_res.max_inline_data; 
  
  /* wqe_buf data in qp_ul_resources_p is already set in alloc_wqe_buf */

  mrc = MOSAL_memory_lock(new_qp->wq_base2reg, new_qp->wq_size2reg);
  if ( mrc != MT_OK ) {
	rc = MOSAL2HH_RET(mrc);
    goto undo_alloc_aux;
  }
  

  return HH_OK;

  /* Error cleanup */
  undo_alloc_aux:
  failed_alloc_aux:
    if (new_qp->wqe_buf_orig != NULL) {/* WQEs buffer were allocated in process mem. or by the XHH_qpm ? */ 
      /* If allocated here than should be freed */
      if (new_qp->dpool_p == NULL) { /* direct allocation */
        if (new_qp->used_virt_alloc) 
          MOSAL_pci_virt_free_consistent(new_qp->wqe_buf_orig, new_qp->wqe_buf_orig_size);
        else
          MOSAL_pci_phys_free_consistent(new_qp->wqe_buf_orig, new_qp->wqe_buf_orig_size);    
      } else { /* used dpool */
        dpool_free(qpm, new_qp->dpool_p, new_qp->wqe_buf_orig);
      }
    }
  failed_alloc_wqe:
  failed_init_qp:
    FREE(new_qp);
  return rc;
}


/* Allocate XHHUL_qp_t object and initialize it */
static HH_ret_t init_qp(
  HHUL_hca_hndl_t hca, 
  HHUL_qp_init_attr_t *qp_init_attr_p, 
  XHHUL_qp_t new_qp,
  XHH_qp_ul_resources_t *qp_ul_resources_p
)
{
  HH_ret_t rc;

  rc = init_uar_info(hca, new_qp, qp_ul_resources_p);
  if (rc != HH_OK) {
    MTL_ERROR2(MT_FLFMT("init_qp: Failed initializing UAR info (%d:%s)"),rc, HH_strerror_sym(rc));
    return rc;
  }

  XHH_DEV_SPECIFIC_INIT_QP;

  new_qp->dpool_p= NULL;
  new_qp->qpn= 0xFFFFFFFF;  /* Init to invalid QP num. until create_qp_done is invoked */
  new_qp->pd= qp_init_attr_p->pd;
  switch (new_qp->sqp_type) {  /* Set transport type appropriate to QP type */
    case VAPI_REGULAR_QP:
      new_qp->ts_type= qp_init_attr_p->ts_type;
      break;
    
    case VAPI_SMI_QP:
    case VAPI_GSI_QP:
      new_qp->ts_type= VAPI_TS_UD;
      break;
    
    case VAPI_RAW_IPV6_QP:
    case VAPI_RAW_ETY_QP:
      new_qp->ts_type= VAPI_TS_RAW;
      break;

    default:
      MTL_ERROR1(MT_FLFMT("Invalid QP type (sqp_type=%d)"),new_qp->sqp_type);
      return HH_EINVAL;
  }
  
  new_qp->srq= qp_init_attr_p->srq;
  /* Init RQ */
  new_qp->rq_res.qp_state= VAPI_RESET;
  if (qp_init_attr_p->srq == HHUL_INVAL_SRQ_HNDL) {
    new_qp->rq_res.max_outs= qp_init_attr_p->qp_cap.max_oust_wr_rq;
    new_qp->rq_res.max_sg_sz= qp_init_attr_p->qp_cap.max_sg_size_rq;
  } else { /* QP associated with SRQ */
    MTL_DEBUG4(MT_FLFMT("%s: Ignoring RQ attributes for a SRQ associated QP"), __func__);
    new_qp->rq_res.max_outs= 0;
    new_qp->rq_res.max_sg_sz= 0;
  }
  new_qp->rq_res.next2free_index= new_qp->rq_res.next2post_index= 0;
  new_qp->rq_res.last_posted_p= NULL;
  new_qp->rq_cq= qp_init_attr_p->rq_cq;
  MOSAL_spinlock_init(&(new_qp->rq_res.q_lock));

  /* Init SQ */
  new_qp->sq_res.qp_state= VAPI_RESET;
  new_qp->sq_res.max_outs= qp_init_attr_p->qp_cap.max_oust_wr_sq;
  new_qp->sq_res.max_sg_sz= qp_init_attr_p->qp_cap.max_sg_size_sq;
  new_qp->sq_res.max_inline_data= qp_init_attr_p->qp_cap.max_inline_data_sq;
  new_qp->sq_res.next2free_index= new_qp->sq_res.next2post_index= 0;
  new_qp->sq_res.last_posted_p= NULL;
  new_qp->sq_cq= qp_init_attr_p->sq_cq;
  MOSAL_spinlock_init(&(new_qp->sq_res.q_lock));

  return HH_OK;
}

inline static MT_bool within_4GB(void* base, MT_size_t bsize)
{
  u_int64_t start_addr;
  u_int64_t end_addr;

  if (sizeof(MT_virt_addr_t) <=4)  return TRUE;  /* For 32 bits machines no check is required */
  start_addr= (u_int64_t)(MT_virt_addr_t)base;
  end_addr= start_addr+bsize-1;
  return ((start_addr >> 32) == (end_addr >> 32));  /* TRUE if 32 MS-bits equal */

}

inline static void* malloc_within_4GB(MT_size_t bsize, MT_bool *used_virt_alloc_p)
{
  void* buf[MAX_ALLOC_RETRY]={NULL};
  MT_bool used_virt_alloc[MAX_ALLOC_RETRY];
  int i,j;

  for (i= 0; i < MAX_ALLOC_RETRY; i++) {  /* Retry to avoid crossing 4GB */
    MT_IF_PCI_PHYS_ALLOC_CONSISTENT_FAILED
    {
      buf[i]= (void*)MOSAL_pci_virt_alloc_consistent(bsize, 0); //TODO: must pass proper alignment here. For now ul_qpm is unused in Darwin.
      used_virt_alloc[i]= TRUE;
    }
    if (buf[i] == NULL) {
      MTL_ERROR3("malloc_within_4GB: Failed allocating buffer of "SIZE_T_FMT" bytes (iteration %d).\n",
        bsize,i);
    /* Free previously allocated buffers if any*/
      for (j= i; j > 0; j--) {
        if (used_virt_alloc[j-1]) {
          MOSAL_pci_virt_free_consistent(buf[j-1], bsize);
        } else {
          MOSAL_pci_phys_free_consistent(buf[j-1], bsize);
        }
      }
      return NULL;
    }
    if (within_4GB(buf[i],bsize)) break;
  }
  if (i == MAX_ALLOC_RETRY) { /* Failed */
    MTL_ERROR2("malloc_within_4GB: Failed allocating buffer of "SIZE_T_FMT" bytes within 4GB boundry "
      "(%d retries).\n", bsize, MAX_ALLOC_RETRY); 
    /* Free all allocated buffers */
    for (i= 0; i < MAX_ALLOC_RETRY; i++) {
      if (used_virt_alloc[i]) {
        MOSAL_pci_virt_free_consistent(buf[i], bsize);
      } else {
        MOSAL_pci_phys_free_consistent(buf[i], bsize);
      }
    }
    return NULL;
  }
  /* Free disqualified buffers if any */
  for (j= i; j > 0; j--) {
    if (used_virt_alloc[j-1]) {
      MOSAL_pci_virt_free_consistent(buf[j-1], bsize);
    } else {
      MOSAL_pci_phys_free_consistent(buf[j-1], bsize);
    }
  }

  *used_virt_alloc_p= used_virt_alloc[i];
  return  buf[i]; /* This is the one buffer which does not cross 4GB boundry */
}

/* Allocate the WQEs buffer for sendQ and recvQ */
/* This function should be invoked after queue properties are set by alloc_init_qp */
static HH_ret_t alloc_wqe_buf(
  /*IN*/ XHHUL_qpm_t qpm,
  /*IN*/ MT_bool in_ddr_mem,   /* Allocation of WQEs buffer is requested in attached DDR mem. */
  /*IN*/ u_int32_t max_outs_wqes, /* HCA cap. */
  /*IN*/ u_int32_t max_sg_ent, /* HCA cap. of max.s/g entries */
  /*IN*/ u_int32_t max_wqe_size_sq,    /* from hca_ul_resources */
  /*IN*/ u_int32_t max_wqe_size_rq,    /* from hca_ul_resources */
  /*IN*/ VAPI_lkey_t qp_mem_lkey, /* indicate that the QP is going to be allocated not by driver */
  /*IN/OUT*/ XHHUL_qp_t new_qp,
  /*OUT*/    XHH_qp_ul_resources_t *qp_ul_resources_p
)
{
  u_int32_t wqe_sz_rq,buf_sz_rq,rq_wqe_base_sz;
  u_int32_t wqe_sz_sq,buf_sz_sq,sq_wqe_base_sz;
  u_int32_t sq_sg_seg_sz,sq_inline_seg_sz;
  u_int8_t log2_wqe_sz_rq,log2_wqe_sz_sq;
  HH_ret_t ret;

  MTL_DEBUG2(MT_FLFMT(
    "%s: max_wqe_size_sq=0x%x, max_wqe_size_rq = 0x%x"),__func__, max_wqe_size_sq,max_wqe_size_rq);
  
  /* Check requested capabilities */
  if ((new_qp->rq_res.max_outs == 0) && (new_qp->sq_res.max_outs == 0)) {
    if (new_qp->srq == HHUL_INVAL_SRQ_HNDL) {
      MTL_ERROR3(MT_FLFMT("Got a request for a QP with 0 WQEs on both SQ and RQ - rejecting !"));
      return HH_EINVAL_PARAM;
    } else { /* QP has no WQEs buffer - uses SRQ only */
      new_qp->wqe_buf_orig= NULL;
      new_qp->wqe_buf_orig_size= 0;
      qp_ul_resources_p->wqes_buf= 0;   
      qp_ul_resources_p->wqes_buf_sz= 0; /* No WQEs buffer to register */
      return HH_OK;
    }
  }
  
 if ((new_qp->rq_res.max_outs > max_outs_wqes) || (new_qp->sq_res.max_outs > max_outs_wqes)) {
   MTL_ERROR2(MT_FLFMT(
	 "QP cap. requested (rq_res.max_outs=%u, sq_res.max_outs=%u) exceeds HCA cap. (max_qp_ous_wr=%u)"),
	 new_qp->rq_res.max_outs, new_qp->sq_res.max_outs, max_outs_wqes);
   return HH_E2BIG_WR_NUM;
 }
  if ((new_qp->rq_res.max_sg_sz > max_sg_ent) || (new_qp->sq_res.max_sg_sz > max_sg_ent)) {
    MTL_ERROR2(MT_FLFMT(
      "QP cap. requested (rq_res.max_sg_sz=%u, sq_res.max_sg_sz=%u) exceeds HCA cap. (max_sg_ent=%u)"),
      new_qp->rq_res.max_sg_sz, new_qp->sq_res.max_sg_sz, max_sg_ent);
    return HH_E2BIG_SG_NUM;
  }

  /* We need extra entry to manage head/tail of cyclic buffer ("reserved entry" for "full" queue) */
  if (new_qp->rq_res.max_outs > 0)  new_qp->rq_res.max_outs++;
  if (new_qp->sq_res.max_outs > 0)  new_qp->sq_res.max_outs++;

  XHHUL_DEV_SPECIFIC_ALLOC_WQE_BUF_INIT;
 
  /* Compute RQ WQE requirements */
  if (new_qp->rq_res.max_outs == 0) {
    log2_wqe_sz_rq= 0;
    wqe_sz_rq= 0;
    buf_sz_rq= 0;
  } else {
    rq_wqe_base_sz= WQE_SEG_SZ_NEXT + WQE_SEG_SZ_CTRL; 
    wqe_sz_rq= rq_wqe_base_sz + (new_qp->rq_res.max_sg_sz * WQE_SEG_SZ_SG_ENTRY);
    if (wqe_sz_rq > max_wqe_size_rq) {
      MTL_ERROR2(
        MT_FLFMT("required RQ capabilities (max_sg_sz=%d) require a too large WQE (%d bytes)"),
          new_qp->rq_res.max_sg_sz, wqe_sz_rq);
      return HH_E2BIG_SG_NUM;
    }
    log2_wqe_sz_rq= (u_int8_t)ceil_log2(wqe_sz_rq);  /* Align to next power of 2 */
    /* A WQE must be aligned to 64B (WQE_ALIGN_SHIFT) so we take at least this size */
    if (log2_wqe_sz_rq < WQE_ALIGN_SHIFT)  log2_wqe_sz_rq= WQE_ALIGN_SHIFT;
    wqe_sz_rq= (1<<log2_wqe_sz_rq);
    MTL_DEBUG5("alloc_wqe_buf: Allocating RQ WQE of size %d.\n",wqe_sz_rq);
    /* Compute real number of s/g entries based on rounded up WQE size */
    new_qp->rq_res.max_sg_sz= (wqe_sz_rq - rq_wqe_base_sz) / WQE_SEG_SZ_SG_ENTRY;  
    /* Make sure we do not exceed reported HCA cap. */
    new_qp->rq_res.max_sg_sz= (new_qp->rq_res.max_sg_sz > max_sg_ent) ? 
      max_sg_ent : new_qp->rq_res.max_sg_sz;
  }
  
  if (new_qp->sq_res.max_outs == 0) {
    sq_wqe_base_sz= 0;
    log2_wqe_sz_sq= 0;
    wqe_sz_sq= 0;
    buf_sz_sq= 0;
  } else {
    /* Compute SQ WQE requirements */
    wqe_sz_sq= /* "next" and "ctrl" are included in the WQE of any transport */
      WQE_SEG_SZ_NEXT + WQE_SEG_SZ_CTRL;

    switch (new_qp->sqp_type) {
      /* For special QPs additional reservation required for the headers (MLX+inline) */
      case VAPI_SMI_QP:
      case VAPI_GSI_QP:
        /* SMI/GSI ==> UD headers */
        wqe_sz_sq+= WQE_INLINE_SZ_UD_HDR;
        wqe_sz_sq+= WQE_INLINE_ICRC;
        break;
      case VAPI_RAW_ETY_QP:
        /* Raw-Ethertype ==> LRH+RWH */
        wqe_sz_sq+= WQE_INLINE_SZ_RAW_HDR;
        break;
      case VAPI_RAW_IPV6_QP:
        /* IPv6 routing headers are given by the consumer in the gather list (?) */
        break;
      default:  /* Normal QP - add relevant transport WQE segments */
        if (new_qp->ts_type == VAPI_TS_UD) {
          wqe_sz_sq+= WQE_SEG_SZ_UD;
        } else if (new_qp->ts_type == VAPI_TS_RD) {
          wqe_sz_sq+= WQE_SEG_SZ_RD;
        }
        if ((new_qp->ts_type == VAPI_TS_RC) ||
            (new_qp->ts_type == VAPI_TS_RD) ||
            (new_qp->ts_type == VAPI_TS_UC)   ) {
          wqe_sz_sq+= WQE_SEG_SZ_BIND_RADDR_ATOMIC;
        }
    }

    if (wqe_sz_sq > max_wqe_size_sq) {
      MTL_ERROR2(MT_FLFMT("required SQ capabilities(max_sg_sz=%d , max_inline_data=%d) "
                          "require a too large WQE (%d bytes)"),
        new_qp->sq_res.max_sg_sz, new_qp->sq_res.max_inline_data, wqe_sz_sq);
      ret= HH_E2BIG_SG_NUM;
      goto failed_sq2big;
    }

    sq_wqe_base_sz= wqe_sz_sq; /* WQE base without data segments */
    /* Compute data segments size for sendQ */
    sq_sg_seg_sz= new_qp->sq_res.max_sg_sz * WQE_SEG_SZ_SG_ENTRY; /* data pointers segments */
  #ifndef QPM_SUPPORT_INLINE_DATA_SET
    /* max_inline_data from create-qp cap. is not supported by default due to backward compat. */
    new_qp->sq_res.max_inline_data= 64; /* Current default minimum */
  #endif
    sq_inline_seg_sz=  /* Compute inline data segment size */ 
      MT_UP_ALIGNX_U32(WQE_INLINE_SZ_BCOUNT + new_qp->sq_res.max_inline_data,WQE_SZ_MULTIPLE_SHIFT);
    wqe_sz_sq+= ((sq_inline_seg_sz > sq_sg_seg_sz) ? sq_inline_seg_sz : sq_sg_seg_sz); 

    log2_wqe_sz_sq= (u_int8_t)ceil_log2(wqe_sz_sq);  /* Align to next power of 2 */
    /* A WQE must be aligned to 64B (WQE_ALIGN_SHIFT) so we take at least this size */
    if (log2_wqe_sz_sq < WQE_ALIGN_SHIFT)  log2_wqe_sz_sq= WQE_ALIGN_SHIFT;
    wqe_sz_sq= (1<<log2_wqe_sz_sq);
    MTL_DEBUG5("alloc_wqe_buf: Allocating SQ WQE of size %d.\n",wqe_sz_sq);
    /* Compute real number of s/g entries based on rounded up WQE size */
    new_qp->sq_res.max_sg_sz= (wqe_sz_sq - sq_wqe_base_sz) / WQE_SEG_SZ_SG_ENTRY;  
    /* Make sure we do not exceed reported HCA cap. */
    new_qp->sq_res.max_sg_sz= (new_qp->sq_res.max_sg_sz > max_sg_ent) ? 
      max_sg_ent : new_qp->sq_res.max_sg_sz;
  }


  buf_sz_rq= new_qp->rq_res.max_outs * wqe_sz_rq;
  buf_sz_sq= new_qp->sq_res.max_outs * wqe_sz_sq;
  
  
  if ((in_ddr_mem) ||  /* Allocate WQEs buffer by XHH_qpm in the attached DDR memory */
      (buf_sz_rq+buf_sz_sq == 0) /* Or no WQE allocation (possible if SRQ is used) */
	  || (qp_mem_lkey != 0)) /* QP going to be allocated outside */
  {
    new_qp->wqe_buf_orig= NULL;
  } else { /* Allocate WQEs buffer in main memory */
#ifdef MT_KERNEL
    new_qp->wqe_buf_orig_size = 
      buf_sz_rq+((wqe_sz_rq != 0) ? (wqe_sz_rq-1):0)+
      buf_sz_sq+((wqe_sz_sq != 0) ? (wqe_sz_sq-1):0);
    new_qp->wqe_buf_orig= malloc_within_4GB(new_qp->wqe_buf_orig_size, &new_qp->used_virt_alloc);
    /* Make RQ (first WQEs buffer) start at page boundry) */
    new_qp->rq_res.wqe_buf= MT_UP_ALIGNX_VIRT((MT_virt_addr_t)(new_qp->wqe_buf_orig),
                                              log2_wqe_sz_rq);
#else 
/* In user space we need to take care of pages sharing on memory locks (fork issues) */
    /* Allocate one more for each queue in order to make each aligned to its WQE size */
    /* (Assures no WQE crosses a page boundry, since we make WQE size a power of 2)   */ 
    new_qp->wqe_buf_orig_size = buf_sz_rq+buf_sz_sq+((wqe_sz_sq != 0) ? (wqe_sz_sq-1):0);
    if (new_qp->wqe_buf_orig_size > (XHHUL_DPOOL_SZ_MAX_KB << XHHUL_DPOOL_SZ_UNIT_SHIFT)) {
      /* Large WQEs buffer - allocate directly */
      /* Assure the buffer covers whole pages (no sharing of locked memory with other data) */
      new_qp->wqe_buf_orig_size = (MOSAL_SYS_PAGE_SIZE-1) + /* For alignment */
          MT_UP_ALIGNX_U32((u_int32_t)new_qp->wqe_buf_orig_size, MOSAL_SYS_PAGE_SHIFT);
      /* Prevent other data reside in the last page of the buffer... */
      /* cover last page (last WQE can be at last page begin and its size is 64B min.)*/
      new_qp->wqe_buf_orig= malloc_within_4GB(new_qp->wqe_buf_orig_size, &new_qp->used_virt_alloc);
      /* Make RQ (first WQEs buffer) start at page boundry) */
      new_qp->rq_res.wqe_buf= MT_UP_ALIGNX_VIRT((MT_virt_addr_t)(new_qp->wqe_buf_orig),
                                                MOSAL_SYS_PAGE_SHIFT);
    } else { /* small WQEs buffer - use dpool */
      /* Round size up to next KB */
      new_qp->wqe_buf_orig_size= 
        MT_UP_ALIGNX_U32((u_int32_t)new_qp->wqe_buf_orig_size, XHHUL_DPOOL_GRANULARITY_SHIFT);
      new_qp->wqe_buf_orig= dpool_alloc(qpm, (u_int8_t)(new_qp->wqe_buf_orig_size >> XHHUL_DPOOL_SZ_UNIT_SHIFT),
                                        &new_qp->dpool_p);
      new_qp->rq_res.wqe_buf= (MT_virt_addr_t)new_qp->wqe_buf_orig; /* no need to align to WQE size */
      /* All dpool buffers are aligned to at least 1KB - see comment in dpool_create() */
    }
 #endif

    if (new_qp->wqe_buf_orig == NULL) {
      MTL_ERROR2("alloc_wqe_buf: Failed allocation of WQEs buffer of "SIZE_T_FMT" bytes within "
        "4GB boundries.\n",new_qp->wqe_buf_orig_size);
      ret= HH_EAGAIN;
      goto failed_wqe_buf;
    }
  }

  /* Set the per queue resources */
  new_qp->rq_res.log2_max_wqe_sz= log2_wqe_sz_rq;
  /* SQ is after RQ - aligned to its WQE size */
  new_qp->sq_res.wqe_buf= MT_UP_ALIGNX_VIRT((new_qp->rq_res.wqe_buf + buf_sz_rq),log2_wqe_sz_sq); 
  new_qp->sq_res.log2_max_wqe_sz= log2_wqe_sz_sq;
  //MTL_DEBUG5(MT_FLFMT("sq_inline_seg_sz=%d  sq_sg_seg_sz=%d"),sq_inline_seg_sz,sq_sg_seg_sz);
  if (wqe_sz_sq <= max_wqe_size_sq) { /* update actual space for inline data */
    new_qp->sq_res.max_inline_data= wqe_sz_sq - sq_wqe_base_sz - 4; 
  } else { /* Due to alignment we have a WQE of a power-of-2, but actual WQE sz is slightly less */
    new_qp->sq_res.max_inline_data= max_wqe_size_sq - sq_wqe_base_sz - 4; 
  }
  
  /* Set the qp_ul_resources_p */
  qp_ul_resources_p->qp_mem_lkey = qp_mem_lkey;
  if ((in_ddr_mem) || (qp_mem_lkey != 0)){
    qp_ul_resources_p->wqes_buf= 0;   /* Allocate in attached DDR memory */
  } else {
    /* Actual buffer starts at beginning of the RQ WQEs buffer (if exists) */
    qp_ul_resources_p->wqes_buf= (buf_sz_rq != 0) ? new_qp->rq_res.wqe_buf : new_qp->sq_res.wqe_buf;
  }

  if ( buf_sz_rq != 0 ) {
    new_qp->wq_base2reg = new_qp->rq_res.wqe_buf;
  }
  else {
    new_qp->wq_base2reg = new_qp->sq_res.wqe_buf;
  }
  new_qp->wq_size2reg = buf_sz_sq + buf_sz_rq;

  /* Actual buffer size is the difference from the real buffer start to end of SQ buffer */
  /* (even if buffer is allocated in DDR mem. and this computation is done from 0 it is valid) */
  qp_ul_resources_p->wqes_buf_sz= (new_qp->sq_res.wqe_buf + buf_sz_sq) - 
    qp_ul_resources_p->wqes_buf;

  XHHUL_DEV_SPECIFIC_ALLOC_WQE_BUF_DONE;

  return HH_OK;

  failed_wqe_buf:
  failed_sq2big:
    return ret;
}


/* Allocate the auxilary WQEs data 
 * (a software context of a WQE which does not have to be in the registered WQEs buffer) */
static HH_ret_t alloc_aux_data_buf(
  /*IN/OUT*/ XHHUL_qp_t new_qp
)
{
  /* RQ auxilary buffer: WQE ID per WQE */ 
  if (new_qp->rq_res.max_outs > 0) {
    new_qp->rq_res.wqe_id= (VAPI_wr_id_t*)
      XHH_SMART_MALLOC(new_qp->rq_res.max_outs * sizeof(VAPI_wr_id_t)); 
    if (new_qp->rq_res.wqe_id == NULL) {
      MTL_ERROR1("alloc_aux_data_buf: Failed allocating RQ auxilary buffer.\n");
      return HH_EAGAIN;
    }
  }

  /* SQ auxilary buffer: WQE ID per WQE */ 
  if (new_qp->sq_res.max_outs > 0) {
    new_qp->sq_res.wqe_id= (VAPI_wr_id_t*)
      XHH_SMART_MALLOC(new_qp->sq_res.max_outs * sizeof(VAPI_wr_id_t)); 
    if (new_qp->sq_res.wqe_id == NULL) {
      MTL_ERROR1("alloc_aux_data_buf: Failed allocating RQ auxilary buffer.\n");
      /* Free any memory chunk allocated by this function */
      if (new_qp->rq_res.wqe_id != NULL) {
        XHH_SMART_FREE(new_qp->rq_res.wqe_id,new_qp->rq_res.max_outs * sizeof(VAPI_wr_id_t)); 
      }
      return HH_EAGAIN;
    }
  }
  
  return HH_OK;
}



/* Insert given QP to the QPM's QP's list */
static HH_ret_t insert_to_list(XHHUL_qpm_t qpm, XHHUL_qp_t qp)
{
  qp_list_entry_t* new_entry_p;
  
  /* Allocate list entry for the new QP */
  new_entry_p= (qp_list_entry_t*)MALLOC(sizeof(qp_list_entry_t));
  if (new_entry_p == NULL) {
    MTL_ERROR2("insert_to_list: Failed allocating list entry.\n");
    return HH_EAGAIN;
  }
  new_entry_p->qp= qp;

  /* Add to the list */
  if ( MOSAL_mutex_acq(&qpm->qpm_lock, TRUE) == MT_OK ) {
	  new_entry_p->next= qpm->qp_list;
	  qpm->qp_list= new_entry_p;
	  MOSAL_mutex_rel(&qpm->qpm_lock);
  }
  else
	  return HH_EAGAIN;
  
  return HH_OK;
}

/* Remove given QP from the QPM's QPs list */
static HH_ret_t remove_from_list(XHHUL_qpm_t qpm, XHHUL_qp_t qp)
{
  qp_list_entry_t *entry2remove_p;
  qp_list_entry_t *prev_p= NULL;
  HH_ret_t rc;
  
  if ( MOSAL_mutex_acq(&qpm->qpm_lock, TRUE) == MT_OK ) {
	  /* Scan list to find given QP's entry */
	  for (entry2remove_p= qpm->qp_list; entry2remove_p != NULL;
		   entry2remove_p= entry2remove_p->next) {
		if (entry2remove_p->qp == qp) break;
		prev_p= entry2remove_p;
	  }
	  if (entry2remove_p == NULL) {
		MTL_ERROR4("XHHUL_qpm::remove_from_list: qpn=%d not found in the QPs list.\n",
		  qp->qpn);
		MOSAL_mutex_rel(&qpm->qpm_lock);
		return HH_EINVAL;
	  }
	  /* Remove entry */
	  if (prev_p == NULL) {
		qpm->qp_list= entry2remove_p->next;
	  } else { /* else, attach next to prev */
		prev_p->next= entry2remove_p->next;
	  }

	  MOSAL_mutex_rel(&qpm->qpm_lock);
	  FREE(entry2remove_p); 
	  rc = HH_OK;
  }
  else
	  rc = HH_EAGAIN;

  return rc;
}


#ifndef MT_KERNEL
/********************************
 * Descriptors pool functions  - not used in kernel space
 ********************************/

#ifdef XHHUL_QPM_DEBUG_DPOOL

static void dpool_dump_list(XHHUL_qpm_t qpm, MT_size_t size_index, 
                            const char *context_text, XHHUL_qpm_dpool_t *dpool_context)
{
  XHHUL_qpm_dpool_t *cur_dpool_p;
  unsigned long cntr= 0;
  
  MTL_ERROR1(MT_FLFMT("[%s - dpool_p=%p] Found inconsistancy in dpool list for buffers of %u KB:"),
             context_text, dpool_context, size_index + XHHUL_DPOOL_SZ_MIN_KB);
  cur_dpool_p= qpm->dpool_p[size_index];
  while ((cur_dpool_p != NULL) && (cur_dpool_p->next != qpm->dpool_p[size_index]) &&
         (cntr < qpm->dpool_cnt)) {
    MTL_ERROR1("(%p <- %p -> %p) ", cur_dpool_p->prev, cur_dpool_p, cur_dpool_p->next);
    cntr++;
    cur_dpool_p= cur_dpool_p->next;
  }
  MTL_ERROR1("(End of list)\n");
  getchar();
}

static MT_bool dpool_check_consistancy(XHHUL_qpm_t qpm, 
                                       const char *context_text, XHHUL_qpm_dpool_t *dpool_context)
{
  XHHUL_qpm_dpool_t *cur_dpool_p;
  MT_size_t size_index;
  unsigned long cntr= 0;

  for (size_index= 0; 
       size_index < (XHHUL_DPOOL_SZ_MAX_KB - XHHUL_DPOOL_SZ_MIN_KB + 1);
       size_index++) {
    cur_dpool_p= qpm->dpool_p[size_index];
    while ((cur_dpool_p != NULL) && (cur_dpool_p->next != qpm->dpool_p[size_index])) {
      if ((cur_dpool_p->next == NULL) ||  
          (cur_dpool_p->prev == NULL) ||
          (cur_dpool_p->next->prev != cur_dpool_p) ||
          (cur_dpool_p->prev->next != cur_dpool_p))  {
        dpool_dump_list(qpm, size_index, context_text, dpool_context);
        return FALSE;
      }
      cntr++;
      if (cntr > qpm->dpool_cnt) {
        MTL_ERROR1(MT_FLFMT("Reading more dpool objects in list than total (%lu)"), 
                   qpm->dpool_cnt);
        dpool_dump_list(qpm, size_index, context_text, dpool_context);
        return FALSE;
      }
      cur_dpool_p= cur_dpool_p->next;
    }
  }
  return TRUE;
}

#endif /*DEBUG_DPOOL*/


static XHHUL_qpm_dpool_t * dpool_create(XHHUL_qpm_t qpm, u_int8_t buf_size_kb)
{
  XHHUL_qpm_dpool_t *new_dpool_p;
  MT_virt_addr_t cur_buf, full_page_buf_end;
  const MT_size_t size_index= buf_size_kb - XHHUL_DPOOL_SZ_MIN_KB;
  const MT_size_t buf_size= (buf_size_kb << XHHUL_DPOOL_SZ_UNIT_SHIFT);

  new_dpool_p= TMALLOC(XHHUL_qpm_dpool_t);
  if (new_dpool_p == NULL) {
    MTL_ERROR2(MT_FLFMT("%s: Failed allocating XHHUL_qpm_dpool_t"), __func__);
    return NULL;
  }
  
  /* Allocate descriptors pool memory - aligned on page start */
  new_dpool_p->orig_size= (XHHUL_DPOOL_SZ_BASE_BUF_KB << XHHUL_DPOOL_SZ_UNIT_SHIFT) + 
                          (MOSAL_SYS_PAGE_SIZE - 1);
  new_dpool_p->orig_buf= malloc_within_4GB(new_dpool_p->orig_size,
                                           &new_dpool_p->used_virt_alloc) ;
  if (new_dpool_p->orig_buf == NULL) {
    MTL_ERROR2(MT_FLFMT("%s: Failed allocating descriptors pool memory of "SIZE_T_FMT" B"),
               __func__, new_dpool_p->orig_size);
    goto failed_orig_buf;
  }

  new_dpool_p->free_buf_list= NULL;
  /* First buffer starts at page boundry and all buffers are of 1KB size multiples */
  /* So all buffers of the dpool are aligned to 1KB, i.e., aligned to any size of our */
  /* WQEs which are all (stride) of power of 2 */
  for (cur_buf= MT_UP_ALIGNX_VIRT((MT_virt_addr_t)new_dpool_p->orig_buf,MOSAL_SYS_PAGE_SHIFT),
       full_page_buf_end= cur_buf + (XHHUL_DPOOL_SZ_BASE_BUF_KB << XHHUL_DPOOL_SZ_UNIT_SHIFT);
       (cur_buf+buf_size) <= full_page_buf_end ;
       cur_buf+= buf_size ) {
    *(void**)cur_buf= new_dpool_p->free_buf_list; /* link before first */
    new_dpool_p->free_buf_list= (void*)cur_buf;
  }

  new_dpool_p->buf_size_kb= buf_size_kb;
  new_dpool_p->ref_cnt= 0;

  if (qpm->dpool_p[size_index] == NULL) { /* first */
    new_dpool_p->next=  new_dpool_p->prev= new_dpool_p;
  } else {
    new_dpool_p->next= qpm->dpool_p[size_index]; /* link to first */
    new_dpool_p->prev= new_dpool_p->next->prev;  /* reverse-link to last */
    new_dpool_p->prev->next= new_dpool_p->next->prev= new_dpool_p;
  }
  qpm->dpool_p[size_index]= new_dpool_p; /* make first */
  
#ifdef XHHUL_QPM_DEBUG_DPOOL
  qpm->dpool_cnt++; 
  MTL_ERROR1("%s: dpool_cnt=%lu  (%p <- %p -> %p) \n", __func__,
             qpm->dpool_cnt, new_dpool_p->prev, new_dpool_p, new_dpool_p->next);
  dpool_check_consistancy(qpm, "After inserting new dpool", new_dpool_p);
#endif
  return new_dpool_p;

  failed_orig_buf:
    FREE(new_dpool_p);
    return NULL;
}

static void dpool_destroy(XHHUL_qpm_t qpm, XHHUL_qpm_dpool_t *dpool_p)
{
  const MT_size_t size_index= dpool_p->buf_size_kb - XHHUL_DPOOL_SZ_MIN_KB;
  
  /* Assumes ref_cnt==0 */
  /* bypass this item */
  dpool_p->prev->next= dpool_p->next;
  dpool_p->next->prev= dpool_p->prev;
  if (qpm->dpool_p[size_index] == dpool_p) { /* if it was the first */
    if (dpool_p->next == dpool_p) { /* and only... */
      qpm->dpool_p[size_index]= NULL;
    } else {                        /* else, make next be first */
      qpm->dpool_p[size_index]= dpool_p->next;
    }
  }
  
#ifdef XHHUL_QPM_DEBUG_DPOOL
  qpm->dpool_cnt--; 
  MTL_ERROR1(MT_FLFMT("%s: dpool_cnt=%lu  (%p <- %p -> %p) "), __func__,
             qpm->dpool_cnt, dpool_p->prev, dpool_p, dpool_p->next);
  dpool_check_consistancy(qpm, "After removing a dpool", dpool_p);
#endif

  if (dpool_p->used_virt_alloc) 
    MOSAL_pci_virt_free_consistent(dpool_p->orig_buf, dpool_p->orig_size);
  else
    MOSAL_pci_phys_free_consistent(dpool_p->orig_buf, dpool_p->orig_size);    

  FREE(dpool_p);
}

static void* dpool_alloc(XHHUL_qpm_t qpm, u_int8_t buf_size_kb, XHHUL_qpm_dpool_t **dpool_pp)
{
  XHHUL_qpm_dpool_t *dpool_p;
  void* alloc_buf;
  const MT_size_t size_index= buf_size_kb - XHHUL_DPOOL_SZ_MIN_KB;

  if ((buf_size_kb < XHHUL_DPOOL_SZ_MIN_KB) || (buf_size_kb > XHHUL_DPOOL_SZ_MAX_KB)) {
    MTL_ERROR2(MT_FLFMT("%s: Given buf_size_kb=0x%u "
                        "(XHHUL_DPOOL_SZ_MIN_KB=%u , XHHUL_DPOOL_SZ_MAX_KB=%u)"), __func__,
               buf_size_kb, XHHUL_DPOOL_SZ_MIN_KB, XHHUL_DPOOL_SZ_MAX_KB);
    return NULL;
  }

  MOSAL_mutex_acq_ui(&qpm->dpool_lock);
  
  dpool_p= qpm->dpool_p[size_index];
  /* If no dpool for this size or existing dpool is full (empty free list) */
  if ((dpool_p == NULL) || (dpool_p->free_buf_list == NULL)) {
    dpool_p= dpool_create(qpm, buf_size_kb);
    if (dpool_p == NULL)  {
        MOSAL_mutex_rel(&qpm->dpool_lock);
        return NULL;
    }
  }

  alloc_buf= dpool_p->free_buf_list;
  dpool_p->free_buf_list= *(void**)alloc_buf; /* next is embedded in free buffer */
  dpool_p->ref_cnt++;

  if ((dpool_p->free_buf_list == NULL) && (dpool_p->prev != dpool_p)) { 
    /* If emptied and not the only dpool for this size - move to end of dpool list for this size */
    qpm->dpool_p[size_index]= dpool_p->next; /* "shift" first */
  }

#ifdef XHHUL_QPM_DEBUG_DPOOL
  dpool_check_consistancy(qpm, "After moving dpool to end of list", dpool_p);
#endif

  MOSAL_mutex_rel(&qpm->dpool_lock);

  *dpool_pp= dpool_p;
  return alloc_buf;
}

static void dpool_free(XHHUL_qpm_t qpm, XHHUL_qpm_dpool_t *dpool_p, void* buf)
{
  const MT_size_t size_index= dpool_p->buf_size_kb - XHHUL_DPOOL_SZ_MIN_KB;
  /* no check on this - assumes dpool is trusted (value checked on creation) */
 
  MOSAL_mutex_acq_ui(&qpm->dpool_lock);
  /* put in free list of associated dpool */
  *(void**)buf= dpool_p->free_buf_list;
  dpool_p->free_buf_list= buf;
  dpool_p->ref_cnt--;
  if (dpool_p->ref_cnt == 0) {
    /* if reached ref_cnt 0, probably not much of this size - compact dpools list */
    dpool_destroy(qpm,dpool_p);

  } else if (qpm->dpool_p[size_index] != dpool_p)  {
    /* if not the first dpool for this size */
    /* Move to begining of dpool list for this size - it has what to offer... */
    if (dpool_p->next != dpool_p->prev) {
      /* more than 2 items - really need to move */
      /* first disconnect */
      dpool_p->prev->next= dpool_p->next;
      dpool_p->next->prev= dpool_p->prev;
      /* Now connect between first and last */
      dpool_p->next= qpm->dpool_p[size_index]; /* link to first */
      dpool_p->prev= dpool_p->next->prev;       /* reverse-link to last */
      dpool_p->prev->next= dpool_p->next->prev= dpool_p;
    }
    /* (after moved to new location) make first */
    qpm->dpool_p[size_index]= dpool_p; 
  }
  
#ifdef XHHUL_QPM_DEBUG_DPOOL
  dpool_check_consistancy(qpm, "After moving dpool to start of list", dpool_p);
#endif
  
  MOSAL_mutex_rel(&qpm->dpool_lock);
}
#endif

#ifdef DUMP_SEND_REQ
static void dump_send_req(XHHUL_qp_t qp, HHUL_send_req_t *sr)
{
  int i;

  MTL_DEBUG4(MT_FLFMT("QP 0x%X - Send: %d S/G entries"),qp->qpn,sr->sg_lst_len); /* Build WQE */
  for (i= 0; i < sr->sg_lst_len; i++) {
    MTL_DEBUG4(MT_FLFMT("Entry %d: lkey=0x%X va=0x%X len=%d"),i,
      sr->sg_lst_p[i].lkey,(MT_virt_addr_t)sr->sg_lst_p[i].addr,sr->sg_lst_p[i].len);
  }
}
#endif


HH_ret_t XHHUL_qpm_set_wqe_buf( 
  HHUL_hca_hndl_t hca, 
  HHUL_qp_hndl_t hhul_qp,
  MT_size_t wqe_buf_size,
  MT_virt_addr_t wqe_buf) 
{ 
    XHHUL_qp_t qp= (XHHUL_qp_t)hhul_qp;
    XHHUL_qpm_t qpm;
    HH_ret_t rc;
	int rq_size, sq_size;

    if (qp == NULL) {
      MTL_ERROR1("XHHUL_qpm_set_wqe_buf: NULL hhul_qp.\n");
      return HH_EINVAL;
    }

    rc= XHHUL_hob_get_qpm(hca, &qpm);
    if (rc != HH_OK) {
      MTL_ERROR1(MT_FLFMT("%s: Failed to get QPM handle (%d=%s)"), __func__, rc, HH_strerror_sym(rc));
      return rc;
    }
	
    rq_size = qp->rq_res.max_outs * (1<<qp->rq_res.log2_max_wqe_sz); 
	sq_size = qp->sq_res.max_outs * (1<<qp->sq_res.log2_max_wqe_sz);
    if (rq_size + sq_size > wqe_buf_size) {
      MTL_ERROR1("XHHUL_qpm_set_wqe_buf: given buffer size ("SIZE_T_DFMT") is too small. needed (rq=%d, sq=%d)\n",
				 wqe_buf_size, rq_size, sq_size);
      return HH_EINVAL;
	}

    qp->rq_res.wqe_buf=wqe_buf; 
	/* SQ is after RQ - aligned to its WQE size */
	qp->sq_res.wqe_buf= MT_UP_ALIGNX_VIRT((qp->rq_res.wqe_buf + rq_size),
										  qp->sq_res.log2_max_wqe_sz); 

	return HH_OK;
}


