#ifndef H_UL_QPM_PRIV_H
#define H_UL_QPM_PRIV_H

#include <udavm.h>
#include <vapi_common.h>

/* XHH_qpm Pkey and GID table access for usage of the special QPs (mlx IB headers) */
#ifdef MT_KERNEL
#include <hob.h>
#endif

#include <mtperf.h>

MTPERF_NEW_SEGMENT(XHH_uar_sendq_dbell,200);
MTPERF_NEW_SEGMENT(WQE_build_send,2000);
MTPERF_NEW_SEGMENT(SQ_WQE_copy,2000);

#ifndef MT_KERNEL
/* instead of "ifdef"ing all over the code we define an empty macro */
#define MOSAL_pci_phys_free_consistent(addr,sz)  do {} while(0);
#endif


/* Limit kmalloc to 2 pages (if this fails, vmalloc will fail too) */
#define WQ_KMALLOC_LIMIT (4*MOSAL_SYS_PAGE_SIZE)
#define SMALL_VMALLOC_AREA (1<<28)  /* VMALLOC area of 256MB or less is considered a scarce resource */

#define WQE_ALIGN_SHIFT 6        /* WQE address should be aligned to 64 Byte */
#define WQE_SZ_MULTIPLE_SHIFT 4           /* WQE size must be 16 bytes multiple */
/* WQE segments sizes */
#define WQE_SEG_SZ_NEXT (sizeof(struct WQE_SEGMENT_NEXT_ST)/8)            /* NEXT segment */
#define WQE_SEG_SZ_CTRL (sizeof(struct WQE_SEGMENT_CTRL_SEND_ST)/8)       /* CTRL segment */
#define WQE_SEG_SZ_RD   (sizeof(struct WQE_SEGMENT_RD_ST)/8)              /* DATAGRAM:RD */
#define WQE_SEG_SZ_UD   (sizeof(struct WQE_SEGMENT_UD_ST)/8)              /* DATAGRAM:UD */
#define WQE_SEG_SZ_RADDR (sizeof(struct WQE_SEGMENT_REMOTE_ADDRESS_ST)/8) /* Remote address */
#define WQE_SEG_SZ_ATOMIC (sizeof(struct WQE_SEGMENT_ATOMIC_ST)/8)        /* Atomic */
#define WQE_SEG_SZ_BIND (sizeof(struct WQE_SEGMENT_BIND_ST)/8)            /* Bind */
/* There is either BIND or RADDR+ATOMIC */
#define WQE_SEG_SZ_BIND_RADDR_ATOMIC ((WQE_SEG_SZ_RADDR+WQE_SEG_SZ_ATOMIC) >  WQE_SEG_SZ_BIND ? \
   (WQE_SEG_SZ_RADDR+WQE_SEG_SZ_ATOMIC) : WQE_SEG_SZ_BIND )
#define WQE_SEG_SZ_SG_ENTRY (sizeof(struct WQE_SEGMENT_DATA_PTR_ST)/8)/* Scatter/Gather entry(ptr)*/
#define WQE_SEG_SZ_SG_ENTRY_DW (sizeof(struct WQE_SEGMENT_DATA_PTR_ST)/32)/* (same in DWORDs) */
/* INLINE segment for UD headers (SMI/GSI) */
#define IB_RWH_SZ 4
#define IB_ICRC_SZ 4
#define WQE_INLINE_SZ_BCOUNT 4
/* INLINE segment for UD headers (SMI/GSI) */
#define WQE_INLINE_SZ_UD_HDR \
  MT_UP_ALIGNX_U32((WQE_INLINE_SZ_BCOUNT+IB_LRH_LEN+IB_GRH_LEN+IB_BTH_LEN+IB_DETH_LEN),4)
/* INLINE segment for RAW-Ethertype */
#define WQE_INLINE_SZ_RAW_HDR \
  MT_UP_ALIGNX_U32((WQE_INLINE_SZ_BCOUNT+IB_LRH_LEN+IB_RWH_SZ),4)
#define WQE_INLINE_ICRC MT_UP_ALIGNX_U32(WQE_INLINE_SZ_BCOUNT+IB_ICRC_SZ,4)
#define BIND_WQE_SZ (WQE_SEG_SZ_NEXT+WQE_SEG_SZ_CTRL+WQE_SEG_SZ_BIND)

#define MAX_ALLOC_RETRY 3  /* Maximum retries to get WQEs buffer which does not cross 4GB boundry */

#define IS_VALID_QPN(qpn) ((qpn) <=  0x00FFFFFF)
#define DEFAULT_PKEY 0xFFFF
#define QP1_PKEY_INDEX 0xFFFFFFFF

#define RESERVED_MEMBIND_EECN 0  /* Pseudo EE-context reserved for memory binding processing */

/* Dpool in size granularity of 1KB */
#define XHHUL_DPOOL_SZ_MIN_KB 1 /* Minimum WQEs buffer of 1KB */
#define XHHUL_DPOOL_SZ_MAX_KB 64 /* Max. is 64KB */
#define XHHUL_DPOOL_SZ_UNIT_SHIFT 10 /* 1KB units shift */
#define XHHUL_DPOOL_GRANULARITY_SHIFT 10 /* 1KB garnularity - for alignment */
#define XHHUL_DPOOL_SZ_BASE_BUF_KB \
  (XHHUL_DPOOL_SZ_MAX_KB*2) /* Size of buffer shared among dpools*/

/* Descriptors pool for small QPs */
/* This data structure allows sharing of locked pages among QPs in order to reduce amount of
  locked pages and assure they cover full pages (fork support) */
typedef struct XHHUL_qp_dpool_st {
  MT_size_t buf_size_kb;     /* Each buffer in the pool */
  void* free_buf_list;
  unsigned long ref_cnt;  /* When reached zero, may be freed */
  void* orig_buf;  /* Pointer to allocated memory chunk */
  MT_size_t orig_size;
  MT_bool used_virt_alloc;
  struct XHHUL_qp_dpool_st *prev; /* list of dpools of same size */
  struct XHHUL_qp_dpool_st *next; /* list of dpools of same size */
} XHHUL_qpm_dpool_t;

typedef struct { /* Queue resources context */
  MT_virt_addr_t wqe_buf;  /* The buffer for this queue WQEs - aligned to WQE size */ 
  VAPI_wr_id_t *wqe_id; /* Array of max_outs entries for holding each WQE ID (WQE index based) */
  u_int32_t max_sg_sz;  /* Max. Scatter/Gather list size */
  MT_size_t log2_max_wqe_sz; /* WQE size is a power of 2 (software implementation requirement) */
  u_int32_t max_inline_data; /* For send queue only */
  u_int32_t max_outs; /* Num. of WQEs in queue == Max. outstanding + 1 */
  u_int32_t next2post_index; /* Next WQE to use for posting (producer index)*/
  u_int32_t next2free_index; /* Next WQE to use free (consumer index) */
  volatile u_int32_t* last_posted_p;  /* For WQE chain linkage (== NULL if none) */
  VAPI_qp_state_t qp_state; /* User level assumed QP state */
  /* Implementation note:
   * qp_state is held per queue in order to avoid race in qp_state updates 
   * which may result from polling different CQs for each queue.
   * We would also like to keep the common XHHUL_qp_t static during the life
   * of the QP in order to avoid additional synchronization between send and
   * receive queue 
   */
  MOSAL_spinlock_t q_lock;   /* Protect concurrent usage of the queue */
} queue_res_t;

/* HHUL_qp_hndl_t is a pointer to this structure */
struct XHHUL_qp_st {
  VAPI_special_qp_t sqp_type; /* VAPI_REGULAR_QP for non-special QP */
  IB_ts_t ts_type;
  IB_wqpn_t qpn;
  HHUL_pd_hndl_t pd;
  HHUL_cq_hndl_t  sq_cq;
  HHUL_cq_hndl_t  rq_cq;
  void* wqe_buf_orig;   /* Pointer returned by malloc_within_4GB() for WQE buffer */
  MT_bool used_virt_alloc;     /* Used "MOSAL_pci_virt_alloc_consistent" for buffer allocation */
  MT_size_t wqe_buf_orig_size; /* size in bytes of wqe_buf_orig */
  MT_virt_addr_t wq_base2reg;  /* base address to be registered */
  MT_size_t wq_size2reg;       /* size of buffer to register */
  XHHUL_qpm_dpool_t *dpool_p; /* If not NULL, wqe_buf_orig taken from this descriptors pool */
  queue_res_t sq_res;   /* Send queue resources */
  queue_res_t rq_res;   /* Receive queue resources */
  HHUL_srq_hndl_t srq;  /* Set to HHUL_INVAL_SRQ_HNDL if not associated with a SRQ */
  MT_bool  last_qkey_from_wqe; /* TRUE -- qkey in last posted WQE was taken from WQE (not from QPC) */

  XHH_DEV_SPECIFIC_UL_QP_T_FIELDS

};   

typedef struct qp_list_entry_st {  /* QP's linked list */
  XHHUL_qp_t qp;
  struct qp_list_entry_st *next;  
} qp_list_entry_t;


struct XHHUL_qpm_st { /* XHHUL_qpm_t is a pointer to this */
  qp_list_entry_t *qp_list; /* do we really need this ?! */
  MOSAL_mutex_t qpm_lock; /* used for qp_cnt protection, too */
  XHHUL_qpm_dpool_t *dpool_p[XHHUL_DPOOL_SZ_MAX_KB - XHHUL_DPOOL_SZ_MIN_KB + 1];/* KB garanularity */
#ifdef XHHUL_QPM_DEBUG_DPOOL
  unsigned long dpool_cnt; 
#endif
  MOSAL_mutex_t dpool_lock;
  XHHUL_srqm_t srqm;
};

#endif

