/*
  This software is available to you under a choice of one of two
  licenses.  You may choose to be licensed under the terms of the GNU
  General Public License (GPL) Version 2, available at
  <http://www.fsf.org/copyleft/gpl.html>, or the OpenIB.org BSD
  license, available in the LICENSE.TXT file accompanying this
  software.  These details are also available at
  <http://openib.org/license.html>.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
  BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
  ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
  CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
  SOFTWARE.

  Copyright (c) 2004 Mellanox Technologies Ltd.  All rights reserved.
*/

#define C_XHHUL_SRQM_COMM_C

#include <mosal.h>
#include <ib_defs.h>
#include <vapi.h>
#include <log2.h>
#include <xhh.h>
#include <ul.h>

#include <ul_hob.h>
#include <ul_pdm.h>
#include <vapi_common.h>

#include "ul_srqm.h"
#include <ul_srqm_priv.h>
#include <dev_ul_srqm_inline.h>
#include <ul_srqm_inline.h>
/**********************************************************************************************
 *                    Private functions protoypes declarations
 **********************************************************************************************/
static HH_ret_t init_srq(
						HHUL_hca_hndl_t hca,
						HHUL_pd_hndl_t  pd,
						u_int32_t max_sentries,
						XHH_srq_ul_resources_t *ul_res_p,
						XHHUL_srq_t new_srq
						);

static HH_ret_t compute_wqe_sz(
							  /*IN*/ u_int32_t hca_max_sentries,  /* HCA cap. of max.scatter entries for SRQs */
							  /*IN*/ u_int32_t max_srq_wqe_sz,
							  /*IN/OUT*/ XHHUL_srq_t new_srq
							  );

static HH_ret_t alloc_wqe_buf(
							 /*IN*/ MT_bool in_ddr_mem,	  /* Allocation of WQEs buffer is requested in attached DDR mem. */
							 /*IN*/ u_int32_t hca_max_outs,	/* HCA cap. */
							 /*IN*/ u_int32_t req_max_outs,	/* Requested capabilities */
							 /*IN*/ u_int8_t  log2_max_wqe_sz,
							 /*OUT*/    XHHUL_srq_wqe_buf_t **buf_pp,
							 /*OUT*/    XHH_srq_ul_resources_t *srq_ul_resources_p
							 );

static void free_wqe_buf(/*IN*/XHHUL_srq_wqe_buf_t *buf_p);

/* Append WQEs buffer extention to WQEs free list */
/* Must be invoked with SRQ lock held */
static void append_to_free_list(
							   /*IN*/ XHHUL_srq_t srq,
							   /*IN*/ MT_virt_addr_t wqes_buf_extention, /* The part of the new WQEs buf above old part */
							   /*IN*/ MT_size_t extention_sz_in_wqes	 /* Extention size in WQEs */
							   );

static HH_ret_t alloc_aux_data_buf(
								  /*IN/OUT*/ XHHUL_srq_t new_srq
								  );

inline static MT_bool within_4GB(void* base, MT_size_t bsize)
{
	u_int64_t start_addr;
	u_int64_t end_addr;

	if (sizeof(MT_virt_addr_t) <=4)	 return TRUE;  /* For 32 bits machines no check is required */
	start_addr= (u_int64_t)(MT_virt_addr_t)base;
	end_addr= start_addr+bsize-1;
	return((start_addr >> 32) == (end_addr >> 32));	 /* TRUE if 32 MS-bits equal */

}

/**********************************************************************************************
 *                    Public API Functions (defined in xhhul_hob.h)
 **********************************************************************************************/

HH_ret_t XHHUL_srqm_create(
						  XHHUL_hob_t hob, 
						  XHHUL_srqm_t *srqm_p 
						  ) 
{ 
	*srqm_p= (XHHUL_srqm_t) MALLOC(sizeof(struct XHHUL_srqm_st));
	if (*srqm_p == NULL) {
		MTL_ERROR1("%s: Failed allocating XHHUL_srqm_t.\n", __func__);
		return HH_EAGAIN;
	}

	(*srqm_p)->srqs_list= NULL;
	MOSAL_mutex_init(&((*srqm_p)->srqm_lock));

	return HH_OK;
}


HH_ret_t XHHUL_srqm_destroy(
						   XHHUL_srqm_t srqm 
						   ) 
{ 

	XHHUL_srq_t srq;

	while (srqm->srqs_list) {
		srq = srqm->srqs_list;
		srqm->srqs_list= srq->next;
		MTL_ERROR4(MT_FLFMT("%s: Releasing resource left-overs for SRQ 0x%X"), __func__, srq->srqn);
		/* Free QP resources: Auxilary buffer + WQEs buffer */
		XHH_SMART_FREE(srq->wqe_id, srq->cur_buf_p->max_outs * sizeof(VAPI_wr_id_t));
		if ( srq->cur_buf_p->size2reg ) {
			MOSAL_memory_unlock(srq->cur_buf_p->buf2reg, srq->cur_buf_p->size2reg);
		}
		free_wqe_buf(srq->cur_buf_p);
		if (srq->resized_buf_p != NULL) {
			if ( srq->resized_buf_p->size2reg ) {
				MOSAL_memory_unlock(srq->resized_buf_p->buf2reg, srq->resized_buf_p->size2reg);
			}
			free_wqe_buf(srq->resized_buf_p);
		}
		FREE(srq->wqe_draft);
		FREE(srq);
	}

	MOSAL_mutex_free(&(srqm->srqm_lock));
	FREE(srqm);
	return HH_OK; 
}


HH_ret_t XHHUL_srqm_create_srq_prep(
								   /*IN*/
								   HHUL_hca_hndl_t hca, 
								   HHUL_pd_hndl_t  pd,
								   u_int32_t max_outs,
								   u_int32_t max_sentries,
								   /*OUT*/
								   HHUL_srq_hndl_t *srq_hndl_p,
								   u_int32_t *actual_max_outs_p,
								   u_int32_t *actual_max_sentries_p,
								   void	/*XHH_srq_ul_resources_t*/ *srq_ul_resources_p
								   ) 
{ 
	XHHUL_srqm_t srqm;
	XHH_hca_ul_resources_t *hca_ul_res_p = NULL;
	XHH_srq_ul_resources_t *ul_res_p= (XHH_srq_ul_resources_t*)srq_ul_resources_p;
	XHHUL_srq_t new_srq;
	HH_ret_t rc;
	call_result_t mrc;

	rc= XHHUL_hob_get_srqm(hca,&srqm);
	if (rc != HH_OK) {
		MTL_ERROR4(MT_FLFMT("%s: Invalid HCA handle (%p)."), __func__, hca);
		return HH_EINVAL_HCA_HNDL;
	}
	rc= XHHUL_hob_get_hca_ul_res_p(hca,&hca_ul_res_p);
	if (rc != HH_OK) {
		MTL_ERROR2(MT_FLFMT("%s: Failed XHHUL_hob_get_hca_ul_res_p (%d=%s).\n"), __func__,
				   rc,HH_strerror_sym(rc));
		return rc;
	}

	(new_srq)= (XHHUL_srq_t)MALLOC(sizeof(struct XHHUL_srq_st));
	if (new_srq == NULL) {
		MTL_ERROR1(MT_FLFMT("%s: Failed allocating XHHUL_srq_t."), __func__);
		return HH_EAGAIN;
	}

	rc= init_srq(hca,pd,max_sentries,ul_res_p,new_srq);
	if (rc != HH_OK) {
		goto failed_init_srq;
	}

	rc= compute_wqe_sz(hca_ul_res_p->max_num_sg_ent_srq, hca_ul_res_p->max_wqe_size_srq, new_srq);
	if (rc != HH_OK) goto failed_compute_wqe_sz;

	rc= alloc_wqe_buf(FALSE/*not in DDR*/,hca_ul_res_p->max_srq_ous_wr,max_outs,
					  new_srq->log2_max_wqe_sz,&new_srq->cur_buf_p, srq_ul_resources_p);
	if (rc != HH_OK) {
		MTL_ERROR1(MT_FLFMT("%s: Failed allocating WQEs buffers."), __func__);
		goto failed_alloc_wqe;
	}

	new_srq->wqe_draft= (u_int32_t *)MALLOC(1 << new_srq->log2_max_wqe_sz);
	if (new_srq->wqe_draft == NULL) {
		MTL_ERROR2(MT_FLFMT("%s: Failed allocating %u bytes for SRQ's wqe draft"), __func__,
				   1 << new_srq->log2_max_wqe_sz);
		goto failed_wqe_draft;
	}


	rc= alloc_aux_data_buf(new_srq);
	if (rc != HH_OK) {
		MTL_ERROR1(MT_FLFMT("%s: Failed allocating auxilary buffers."), __func__);
		goto failed_alloc_aux;
	}

	/* Set output modifiers */
	*srq_hndl_p= new_srq;
	*actual_max_outs_p= XHH_DEV_SPECIFIC_ADJUST_RET_MAX_OUTS(new_srq->cur_buf_p->max_outs);
	*actual_max_sentries_p= new_srq->max_sentries;

	/* wqe_buf data in srq_ul_resources_p is already set in alloc_wqe_buf */

	/* update SRQs list */
	MOSAL_mutex_acq_ui(&(srqm->srqm_lock));
	new_srq->next= srqm->srqs_list;
	srqm->srqs_list= new_srq;
	MOSAL_mutex_rel(&(srqm->srqm_lock));


	if ( new_srq->cur_buf_p->size2reg > 0 ) {
		/* lock the memory */
		mrc = MOSAL_memory_lock(new_srq->cur_buf_p->buf2reg, new_srq->cur_buf_p->size2reg);
		if ( mrc != MT_OK ) {
			rc = MOSAL2HH_RET(mrc);
			goto undo_alloc_aux;
		}
	}

	return HH_OK;

	/* Error cleanup */
	undo_alloc_aux:
	XHH_SMART_FREE(new_srq->wqe_id, new_srq->cur_buf_p->max_outs * sizeof(VAPI_wr_id_t));
	failed_alloc_aux:
	FREE(new_srq->wqe_draft);
	failed_wqe_draft:
	free_wqe_buf(new_srq->cur_buf_p);
	failed_alloc_wqe:
	failed_compute_wqe_sz:
	failed_init_srq:
	FREE(new_srq);
	return rc;
}


HH_ret_t XHHUL_srqm_create_srq_done(
								   HHUL_hca_hndl_t hca, 
								   HHUL_srq_hndl_t hhul_srq, 
								   HH_srq_hndl_t hh_srq, 
								   void/*XHH_srq_ul_resources_t*/ *srq_ul_resources_p
								   ) 
{ 
	XHHUL_srqm_t srqm;
	XHHUL_srq_t srq= (XHHUL_srq_t)hhul_srq;
	XHH_srq_ul_resources_t *ul_res_p= (XHH_srq_ul_resources_t*)srq_ul_resources_p;
	HH_ret_t rc;

	rc= XHHUL_hob_get_srqm(hca,&srqm);
	if (rc != HH_OK) {
		MTL_ERROR4(MT_FLFMT("%s: Invalid HCA handle (%p)."), __func__, hca);
		return HH_EINVAL_HCA_HNDL;
	}
	if (srq == NULL) {
		MTL_ERROR4(MT_FLFMT("%s: NULL hhul_qp handle."), __func__);
		return HH_EINVAL;
	}

	if (srq->cur_buf_p->wqe_buf_orig == NULL) {	/* WQEs buffer allocated in DDR mem. by XHH_qpm */
		if (ul_res_p->wqes_buf == 0) {
			MTL_ERROR1(MT_FLFMT("%s: Got NULL WQEs buffer from qp_ul_res for new srqn=0x%X."), __func__, 
					   hh_srq);
			return HH_EINVAL;
		}
		/* Set the per queue resources */
		srq->cur_buf_p->wqe_buf= MT_UP_ALIGNX_VIRT(ul_res_p->wqes_buf,srq->log2_max_wqe_sz);
		if (srq->cur_buf_p->wqe_buf != ul_res_p->wqes_buf) {
			MTL_ERROR1(
					  "XHHUL_srqm_create_qp_done: Buffer allocated by XHH_qpm ("VIRT_ADDR_FMT") "
					  "is not aligned to RQ WQE size (%d bytes).\n",
					  ul_res_p->wqes_buf,1<<srq->log2_max_wqe_sz);
			return HH_EINVAL;
		}
	}

	srq->hca_virt_wqe_buf= srq->cur_buf_p->wqe_buf;/* first buffer is the HCA's virt. addr. */

	/* Create free WQEs list of wqe_buf */
	append_to_free_list(srq, srq->cur_buf_p->wqe_buf,srq->cur_buf_p->max_outs);

	srq->srqn= hh_srq;

	XHH_DEV_SPECIFIC_CREATE_SRQ_DONE;

	MTL_DEBUG4(MT_FLFMT("%s: srqn=0x%X  buf_p="VIRT_ADDR_FMT"  sz=0x%X"), __func__,
			   srq->srqn, srq->cur_buf_p->wqe_buf, 
			   (1 << srq->log2_max_wqe_sz) * srq->cur_buf_p->max_outs);

	return HH_OK;
}

HH_ret_t XHHUL_srqm_modify_srq_prep(
								   /*IN*/ HHUL_hca_hndl_t hca, 
								   /*IN*/HHUL_srq_hndl_t hhul_srq,
								   /*IN*/VAPI_srq_attr_t  *srq_attr_p,
								   /*IN*/VAPI_srq_attr_mask_t srq_attr_mask,
								   /*OUT*/void/*XHH_srq_ul_resources_t*/ *srq_ul_resources_p
								   )
{
	XHHUL_srqm_t srqm;
	XHHUL_srq_t srq= (XHHUL_srq_t)hhul_srq;
	XHH_srq_ul_resources_t *srq_ul_res_p= (XHH_srq_ul_resources_t*)srq_ul_resources_p;
	XHH_hca_ul_resources_t *hca_ul_res_p = NULL;
	XHHUL_srq_wqe_buf_t *new_buf_p;
	HH_ret_t rc;
	u_int32_t max_effective_outs_wr; /* max(requested_outs, current_outs) */
	call_result_t mrc;

	/* Parameters check */
    if (srq_attr_mask & ~(((u_int8_t)(VAPI_SRQ_ATTR_MAX_OUTS_WR | VAPI_SRQ_ATTR_LIMIT)))) {
		MTL_ERROR1(MT_FLFMT(
						   "%s: Only 0x%x flags are supported (got srq_attr_mask=0x%X)"),
				   __func__, (VAPI_SRQ_ATTR_MAX_OUTS_WR | VAPI_SRQ_ATTR_LIMIT), srq_attr_mask);
		return HH_ENOSYS;
	}

	rc= XHHUL_hob_get_srqm(hca,&srqm);
	if (rc != HH_OK) {
		MTL_ERROR4(MT_FLFMT("%s: Invalid HCA handle (%p)."), __func__, hca);
		return HH_EINVAL_HCA_HNDL;
	}
	if (srq == NULL) {
		MTL_ERROR4(MT_FLFMT("%s: NULL hhul_qp handle."), __func__);
		return HH_EINVAL;
	}
	rc= XHHUL_hob_get_hca_ul_res_p(hca,&hca_ul_res_p);
	if (rc != HH_OK) {
		MTL_ERROR2(MT_FLFMT("%s: Failed XHHUL_hob_get_hca_ul_res (%d=%s).\n"), __func__,
				   rc,HH_strerror_sym(rc));
		return rc;
	}

	memset(srq_ul_res_p,0,sizeof(XHH_srq_ul_resources_t)); 
	srq_ul_res_p->srq_attr_mask = srq_attr_mask;

	max_effective_outs_wr =  XHH_DEV_SPECIFIC_ADJUST_RET_MAX_OUTS(srq->cur_buf_p->max_outs);
	if (srq_attr_mask & VAPI_SRQ_ATTR_MAX_OUTS_WR) {
		/* For now, since shrinking is not performed, do not need to worry about
		 * IB_spec o10-10.2.2 (page 451)
		 */
		if (srq_attr_p->max_outs_wr > hca_ul_res_p->max_srq_ous_wr) {
			MTL_ERROR1(MT_FLFMT("%s: requested max outs wr(%u) > hca capability max outs wr(%u)"),
					   __func__, srq_attr_p->max_outs_wr,hca_ul_res_p->max_srq_ous_wr);
			return HH_E2BIG_WR_NUM;
		}
		if (srq_attr_p->max_outs_wr > XHH_DEV_SPECIFIC_ADJUST_RET_MAX_OUTS(srq->cur_buf_p->max_outs)) {
			max_effective_outs_wr =  srq_attr_p->max_outs_wr;
		}
	}
	if (srq_attr_mask & VAPI_SRQ_ATTR_LIMIT) {
		/* For now, since shrinking is not performed, do not need to worry about
		 * IB_spec o10-10.2.2 (page 451).  Therefore, change current max_outs only
		 * if requesting to increase the SRQ size.
		 */
		if (srq_attr_p->srq_limit > max_effective_outs_wr) {
			MTL_ERROR1(MT_FLFMT("%s: requested srq limit(%u) > current max outs wr(%u)"),
					   __func__, srq_attr_p->srq_limit,max_effective_outs_wr);
			return HH_E2BIG_SRQ_LIMIT;
		}
		srq_ul_res_p->srq_limit = srq_attr_p->srq_limit;
	}

	if (srq_attr_mask & VAPI_SRQ_ATTR_MAX_OUTS_WR) {
		if (srq_attr_p->max_outs_wr > 
			XHH_DEV_SPECIFIC_ADJUST_RET_MAX_OUTS(srq->cur_buf_p->max_outs)) {
			rc= alloc_wqe_buf(FALSE/*not in DDR*/, hca_ul_res_p->max_srq_ous_wr,srq_attr_p->max_outs_wr,
							  srq->log2_max_wqe_sz,&new_buf_p, srq_ul_res_p);
			if (rc != HH_OK) {
				MTL_ERROR1(MT_FLFMT("%s: Failed allocating WQEs buffers."), __func__);
				return rc;
			}
			/* test that new buffer size can fit into 4G if take the original buffer start
			 * address as the base */
			if (!within_4GB((void *)srq->hca_virt_wqe_buf, srq_ul_res_p->wqes_buf_sz)) {
				MTL_ERROR1(MT_FLFMT("%s: new WQEs buffer would cause 4G overflow (size="SIZE_T_FMT", orig="VIRT_ADDR_FMT")"),
						   __func__, srq_ul_res_p->wqes_buf_sz, srq->hca_virt_wqe_buf);
				if (new_buf_p != NULL) free_wqe_buf(new_buf_p);
				return HH_ERR;
			}

		}
		else { /* Shrinking implies no change for current implementation */
			new_buf_p= NULL;
			srq_ul_res_p->wqes_buf_sz = 0;
			srq_ul_res_p->wqes_buf = 0;
		}

		MOSAL_spinlock_irq_lock(&(srq->q_lock)); 
		if (srq->resize_in_progress) {
			MOSAL_spinlock_unlock(&(srq->q_lock)); 
			MTL_ERROR1(MT_FLFMT("%s: Invoked while resize is in progress (SRQn=0x%X)"), __func__,
					   srq->srqn);
			if (new_buf_p != NULL) free_wqe_buf(new_buf_p);
			return HH_EBUSY;
		}

		srq->resized_buf_p= new_buf_p;
		if (srq->resized_buf_p != NULL) {
			/* Duplicate WQEs buffer image into resized buffer */
			memcpy((void*)srq->resized_buf_p->wqe_buf,(void*)srq->cur_buf_p->wqe_buf,
				   srq->cur_buf_p->max_outs << srq->log2_max_wqe_sz);
			if ( srq->resized_buf_p->size2reg > 0 ) {
				/* lock the new buffer */
				mrc = MOSAL_memory_lock(srq->resized_buf_p->buf2reg, srq->resized_buf_p->size2reg);
				if ( mrc != MT_OK ) {
					MTL_ERROR1(MT_STDFMT("faild locking memory"));
					MOSAL_spinlock_unlock(&(srq->q_lock)); 
					return MOSAL2HH_RET(mrc);
				}
			}
		}
		srq->resize_in_progress= TRUE;

		MOSAL_spinlock_unlock(&(srq->q_lock)); 
	}
	return HH_OK;
}


HH_ret_t XHHUL_srqm_modify_srq_done(
								   /*IN*/HHUL_hca_hndl_t hca, 
								   /*IN*/HHUL_srq_hndl_t hhul_srq, 
								   /*IN*/VAPI_srq_attr_mask_t srq_attr_mask,
								   /*IN*/void/*XHH_srq_ul_resources_t*/ *srq_ul_resources_p,
								   /*OUT*/u_int32_t      *max_outs_wr_p	/* Max. outstanding WQEs */
								   ) 
{ 
	XHHUL_srqm_t srqm;
	XHHUL_srq_t srq= (XHHUL_srq_t)hhul_srq;
	XHH_srq_ul_resources_t *ul_res_p= (XHH_srq_ul_resources_t*)srq_ul_resources_p;
	HH_ret_t rc= HH_OK;
	XHHUL_srq_wqe_buf_t *freed_wqe_buf_p= NULL;
	VAPI_wr_id_t *new_wqe_id_array,*freed_wqe_id_array= NULL;

	rc= XHHUL_hob_get_srqm(hca,&srqm);
	if (rc != HH_OK) {
		MTL_ERROR4(MT_FLFMT("%s: Invalid HCA handle (%p)."), __func__, hca);
		return HH_EINVAL_HCA_HNDL;
	}
	if (srq == NULL) {
		MTL_ERROR4(MT_FLFMT("%s: NULL hhul_qp handle."), __func__);
		return HH_EINVAL_SRQ_HNDL;
	}


	if (srq_attr_mask & VAPI_SRQ_ATTR_MAX_OUTS_WR) {
		if (! srq->resize_in_progress) {
			MOSAL_spinlock_unlock(&(srq->q_lock)); 
			MTL_ERROR1(MT_FLFMT("%s: Invoked while NO resize in progress (SRQn=0x%X)"), __func__,
					   srq->srqn);
			return HH_EFATAL;
		}

		if (srq->resized_buf_p == NULL) { /* Shrinking */
			/* When "shrinking" not resources were allocated - nothing to free */
			goto clean_exit;
		}

		freed_wqe_buf_p= srq->resized_buf_p; /* Clean resized buffer for any of the following errors */

		if (ul_res_p == NULL) {	/* Failure - rollback the modify_prep */
			MTL_ERROR4(MT_FLFMT("%s: Got failure notification from VIPKL_modify_srq (SRQn=0x%X)"),__func__,
					   srq->srqn);
			rc= HH_OK; /* No problem in rolling back the modify_prep... */
			goto clean_exit;
		}

		if (srq->resized_buf_p->wqe_buf_orig == NULL) {	/* WQEs buffer allocated in DDR mem. by XHH_qpm */
			if (ul_res_p->wqes_buf == 0) {
				MTL_ERROR1(MT_FLFMT("%s: Got NULL WQEs buffer from qp_ul_res for new srqn=0x%X."), __func__, 
						   srq->srqn);
				rc= HH_EAGAIN;
				goto clean_exit;
			}
			/* Set the per queue resources */
			srq->resized_buf_p->wqe_buf= MT_UP_ALIGNX_VIRT(ul_res_p->wqes_buf,srq->log2_max_wqe_sz);
			if (srq->resized_buf_p->wqe_buf != ul_res_p->wqes_buf) {
				MTL_ERROR1(
						  "%s: Buffer allocated by XHH_qpm ("VIRT_ADDR_FMT") "
						  "is not aligned to RQ WQE size (%d bytes).\n", __func__,
						  ul_res_p->wqes_buf,1<<srq->log2_max_wqe_sz);
				return HH_EFATAL; /* Free nothing - inconsistancy problem */
			}
		}

		new_wqe_id_array= (VAPI_wr_id_t*)
						  XHH_SMART_MALLOC(srq->resized_buf_p->max_outs * sizeof(VAPI_wr_id_t)); 
		if (new_wqe_id_array == NULL) {
			MTL_ERROR1(MT_FLFMT("%s: Failed allocating SRQ auxilary buffer (for 0x%X WQEs IDs)."), __func__,
					   srq->resized_buf_p->max_outs);
			rc= HH_EAGAIN;
			/* Failing to extend the WQEs ID buffer still requires using the new buffer 
			 * (HCA already moved to new buffer). So we define it with the same size as current */
			srq->resized_buf_p->max_outs= srq->cur_buf_p->max_outs;
			new_wqe_id_array= srq->wqe_id; /* use original */
		}

		MOSAL_spinlock_irq_lock(&(srq->q_lock));
		if (new_wqe_id_array != srq->wqe_id) {
			/* Copy old WQE IDs to new buffer */
			memcpy(new_wqe_id_array, srq->wqe_id, srq->cur_buf_p->max_outs * sizeof(VAPI_wr_id_t));
			freed_wqe_id_array= srq->wqe_id; /* save to free outside of spinlock */
			srq->wqe_id= new_wqe_id_array;
		}

		freed_wqe_buf_p= srq->cur_buf_p; /* save to free outside of spinlock */
		srq->cur_buf_p= srq->resized_buf_p;	/* Set resized as cur_buf */
		srq->real_virt_offset= srq->cur_buf_p->wqe_buf - srq->hca_virt_wqe_buf;

		/* Add to free WQEs the new WQEs */
		append_to_free_list(srq, 
							srq->cur_buf_p->wqe_buf + 
							(freed_wqe_buf_p->max_outs << srq->log2_max_wqe_sz),
							srq->cur_buf_p->max_outs - freed_wqe_buf_p->max_outs);

		MOSAL_spinlock_unlock(&(srq->q_lock)); 

		MOSAL_memory_unlock(freed_wqe_buf_p->buf2reg, freed_wqe_buf_p->size2reg);

		clean_exit:
		/* Set output modifier */
		if (max_outs_wr_p != NULL) *max_outs_wr_p= XHH_DEV_SPECIFIC_ADJUST_RET_MAX_OUTS(srq->cur_buf_p->max_outs);
		/* Free unused resources */
		if (freed_wqe_id_array != NULL) {
			XHH_SMART_FREE(freed_wqe_id_array, freed_wqe_buf_p->max_outs * sizeof(VAPI_wr_id_t));
		}
		if (freed_wqe_buf_p != NULL) {
			free_wqe_buf(freed_wqe_buf_p);
		}
		srq->resized_buf_p= NULL;     
		srq->resize_in_progress= FALSE;

		MTL_DEBUG4(MT_FLFMT("%s: srqn=0x%X  buf_p="VIRT_ADDR_FMT"  sz=0x%X, free_size=%d"), __func__,
				   srq->srqn, srq->cur_buf_p->wqe_buf, 
				   (1 << srq->log2_max_wqe_sz) * srq->cur_buf_p->max_outs, srq->free_wqe_list_sz);
	}

	return rc;
}

HH_ret_t XHHUL_srqm_destroy_srq_done(
									HHUL_hca_hndl_t hca, 
									HHUL_qp_hndl_t hhul_srq 
									) 
{ 
	XHHUL_srqm_t srqm;
	XHHUL_srq_t srq= (XHHUL_srq_t)hhul_srq;
	XHHUL_srq_t cur_srq,prev_srq;
	HH_ret_t rc;

	rc= XHHUL_hob_get_srqm(hca,&srqm);
	if (rc != HH_OK) {
		MTL_ERROR4("%s: Invalid HCA handle (%p).", __func__, hca);
		return HH_EINVAL_HCA_HNDL;
	}

	/* update SRQs list */
	MOSAL_mutex_acq_ui(&(srqm->srqm_lock));
	/* find SRQ in SRQs list */
	for (prev_srq= NULL, cur_srq= srqm->srqs_list; 
		(cur_srq != NULL) && (cur_srq != srq); 
		prev_srq= cur_srq , cur_srq= cur_srq->next);
	if (cur_srq == NULL) {
		MOSAL_mutex_rel(&(srqm->srqm_lock));
		MTL_ERROR2(MT_FLFMT("%s: Could not find given SRQ (hndl=0x%p , srqn=0x%X)"), __func__, 
				   srq, srq->srqn);
		return HH_EINVAL_SRQ_HNDL;
	}
	/* remove SRQ from list */
	if (prev_srq != NULL) prev_srq->next= srq->next;
	else				 srqm->srqs_list= srq->next;
	MOSAL_mutex_rel(&(srqm->srqm_lock));

	/* Free SRQ resources: Auxilary buffer + WQEs buffer + WQE draft + SRQ object */
	MTL_DEBUG4(MT_FLFMT("Freeing user level WQE-IDs auxilary buffers"));
	XHH_SMART_FREE(srq->wqe_id, srq->cur_buf_p->max_outs * sizeof(VAPI_wr_id_t)); 
	if ( srq->cur_buf_p->size2reg ) {
		MOSAL_memory_unlock(srq->cur_buf_p->buf2reg, srq->cur_buf_p->size2reg);
	}
	free_wqe_buf(srq->cur_buf_p);
	FREE(srq->wqe_draft);
	FREE(srq);

	return HH_OK;  
}

/* Release this WQE only and return its WQE ID */
HH_ret_t XHHUL_srqm_comp(
						XHHUL_srqm_t srqm, 
						HHUL_srq_hndl_t hhul_srq,
						u_int32_t wqe_addr_32lsb, 
						VAPI_wr_id_t *wqe_id_p
						) 
{ 
	XHHUL_srq_t srq= (XHHUL_srq_t)hhul_srq;
	u_int32_t wqes_base_32lsb= (u_int32_t)(srq->hca_virt_wqe_buf & 0xFFFFFFFF) ;
	u_int32_t freed_wqe_index;
	MT_virt_addr_t wqe_buf_h= sizeof(u_int32_t*) > 4 ? (srq->hca_virt_wqe_buf >> 32) << 32 : 0;

	/* wqe_addr is wqe HCA virtual address */
	MT_virt_addr_t hca_wqe_addr= wqe_buf_h | wqe_addr_32lsb;

	if (wqe_addr_32lsb < wqes_base_32lsb) {
		MTL_ERROR1(MT_FLFMT("%s: Got wqe_addr_32lsb=0x%X < wqes_base_32lsb=0x%X"), __func__,
				   wqe_addr_32lsb, wqes_base_32lsb);
		return HH_EINVAL;
	}
	if (wqe_addr_32lsb & MASK32(srq->log2_max_wqe_sz)) {
		MTL_ERROR1(MT_FLFMT(
						   "%s: Got wqe_addr_32lsb=0x%X which is not aligned to WQE size/stride 2^%u"),
				   __func__, wqe_addr_32lsb, srq->log2_max_wqe_sz);
		return HH_EINVAL;
	}

	freed_wqe_index= (wqe_addr_32lsb - wqes_base_32lsb) >> srq->log2_max_wqe_sz;

	MOSAL_spinlock_irq_lock(&(srq->q_lock));
	if (freed_wqe_index > srq->cur_buf_p->max_outs) {
		MTL_ERROR1(MT_FLFMT("%s: Got wqe_addr_32lsb=0x%X which is WQE index 0x%X "
							"(max_outs=0x%X , wqes_base_32lsb=0x%X , log2_max_wqe_sz=0x%X)"), 
				   __func__, wqe_addr_32lsb, freed_wqe_index, 
				   srq->cur_buf_p->max_outs, wqes_base_32lsb, srq->log2_max_wqe_sz);
		MOSAL_spinlock_unlock(&(srq->q_lock));
		return HH_EINVAL;
	}

	/* Get WQE ID from auxilary buffer */
	*wqe_id_p= srq->wqe_id[freed_wqe_index]; 

	/* Return WQE to free list */
	srq->cur_outs --;			  /* (for debug purpose)            */
	*((MT_virt_addr_t*)(hca_wqe_addr+srq->real_virt_offset))= 0;
	if (srq->free_wqes_list_top == 0) {
		/*list is empty.  First element is both top and bottom */
		MTL_DEBUG4(MT_FLFMT(
						   "%s: Free wqes list is empty, list_end="
						   VIRT_ADDR_FMT ", curr_outs=0x%x, max_sentries=0x%x,resize_in_progress=%d"), 
				   __func__, srq->free_wqes_list_end,
				   srq->cur_outs, srq->max_sentries,srq->resize_in_progress);
		srq->free_wqes_list_top = srq->free_wqes_list_end = hca_wqe_addr;
		srq->free_wqe_list_sz = 1;
	}
	else {
		/*list has at least one element.  Adding new element to end*/
		/* modify last element in list to point to new wqe */
		*((MT_virt_addr_t*)(srq->free_wqes_list_end + srq->real_virt_offset)) = hca_wqe_addr; 
		srq->free_wqes_list_end = hca_wqe_addr;
		srq->free_wqe_list_sz++;
		MTL_DEBUG4(MT_FLFMT("%s: Added wqe to free list. New size=%d, wqe_addr_32lsb=0x%x"), 
				   __func__, srq->free_wqe_list_sz, wqe_addr_32lsb);
	}
	if (hca_wqe_addr == srq->last_posted_hca_va) {
		/* After WQE put in the free list, we should not link it to next WQE */
		srq->last_posted_hca_va= 0;
	}
	MOSAL_spinlock_unlock(&(srq->q_lock));

	return HH_OK;
}



/**********************************************************************************************
 *                    Private Functions
 **********************************************************************************************/


/* Allocate XHHUL_srq_t object and initialize it */
static HH_ret_t init_srq(
						HHUL_hca_hndl_t hca,
						HHUL_pd_hndl_t  pd,
						u_int32_t max_sentries,
						XHH_srq_ul_resources_t *ul_res_p,
						XHHUL_srq_t new_srq
						)
{
	HH_ret_t rc;

	memset(new_srq,0,sizeof(struct XHHUL_srq_st));

	rc= init_uar_info(hca, new_srq, ul_res_p);
	if (rc != HH_OK) {
		MTL_ERROR2(MT_FLFMT("%s: Failed getting XHHUL_hob's UAR, or UAR index (%d=%s)."),
				   __func__,rc,HH_strerror_sym(rc));
		return rc;
	}

	new_srq->srqn= 0xFFFFFFFF;	/* Init to invalid SRQ num. until create_qp_done is invoked */
	new_srq->pd= pd;
	new_srq->max_sentries= max_sentries;
	MOSAL_spinlock_init(&(new_srq->q_lock));

	return HH_OK;
}


inline static void* malloc_within_4GB(MT_size_t bsize, MT_bool *used_virt_alloc_p)
{
	void* buf[MAX_ALLOC_RETRY]={NULL};
	MT_bool used_virt_alloc[MAX_ALLOC_RETRY];
	int i,j;

	for (i= 0; i < MAX_ALLOC_RETRY; i++) {	/* Retry to avoid crossing 4GB */
		MT_IF_PCI_PHYS_ALLOC_CONSISTENT_FAILED      
		{
			buf[i]= (void*)MOSAL_pci_virt_alloc_consistent(bsize, 0); //TODO: must pass proper alignment here. For now ul_qpm is unused in Darwin.
			used_virt_alloc[i]= TRUE;
		}
		if (buf[i] == NULL) {
			MTL_ERROR3("malloc_within_4GB: Failed allocating buffer of "SIZE_T_FMT" bytes (iteration %d).\n",
					   bsize,i);
			/* Free previously allocated buffers if any*/
			for (j= i; j > 0; j--) {
				if (used_virt_alloc[j-1]) {
					MOSAL_pci_virt_free_consistent(buf[j-1], bsize);
				}
				else {
					MOSAL_pci_phys_free_consistent(buf[j-1], bsize);
				}
			}
			return NULL;
		}
		if (within_4GB(buf[i],bsize)) break;
	}
	if (i == MAX_ALLOC_RETRY) {	/* Failed */
		MTL_ERROR2("malloc_within_4GB: Failed allocating buffer of "SIZE_T_FMT" bytes within 4GB boundry "
				   "(%d retries).\n", bsize, MAX_ALLOC_RETRY); 
		/* Free all allocated buffers */
		for (i= 0; i < MAX_ALLOC_RETRY; i++) {
			if (used_virt_alloc[i]) {
				MOSAL_pci_virt_free_consistent(buf[i], bsize);
			}
			else {
				MOSAL_pci_phys_free_consistent(buf[i], bsize);
			}
		}
		return NULL;
	}
	/* Free disqualified buffers if any */
	for (j= i; j > 0; j--) {
		if (used_virt_alloc[j-1]) {
			MOSAL_pci_virt_free_consistent(buf[j-1], bsize);
		}
		else {
			MOSAL_pci_phys_free_consistent(buf[j-1], bsize);
		}
	}

	*used_virt_alloc_p= used_virt_alloc[i];
	return  buf[i];	/* This is the one buffer which does not cross 4GB boundry */
}

/* Compute needed WQE size and set new_srq->log2_wqe_sz */
static HH_ret_t compute_wqe_sz(
							  /*IN*/ u_int32_t hca_max_sentries,  /* HCA cap. of max.scatter entries for SRQs */
							  /*IN*/ u_int32_t max_srq_wqe_sz,
							  /*IN/OUT*/ XHHUL_srq_t new_srq
							  )
{
	u_int32_t wqe_sz,wqe_base_sz;

	/* Check requested capabilities */
	if (new_srq->max_sentries > hca_max_sentries) {
		MTL_ERROR2(MT_FLFMT(
						   "%s: Got request for %u scatter entries (HCA cap. for SRQ is %u scatter entries)"),
				   __func__, new_srq->max_sentries, hca_max_sentries);
		return HH_E2BIG_SG_NUM;
	}

	/* Compute RQ WQE requirements */
	wqe_base_sz= WQE_SEG_SZ_NEXT + WQE_SEG_SZ_CTRL; 
	wqe_sz= wqe_base_sz + (new_srq->max_sentries * WQE_SEG_SZ_SG_ENTRY);
	if (wqe_sz > max_srq_wqe_sz) {
		MTL_ERROR2(
				  MT_FLFMT("required SRQ capabilities (max_sentries=%d) require a too large WQE (%u bytes)"),
				  new_srq->max_sentries, wqe_sz);
		return HH_E2BIG_SG_NUM;
	}
	new_srq->log2_max_wqe_sz= (u_int8_t)ceil_log2(wqe_sz);	/* Align to next power of 2 */
	/* A WQE must be aligned to 64B (WQE_ALIGN_SHIFT) so we take at least this size */
	if (new_srq->log2_max_wqe_sz < WQE_ALIGN_SHIFT)
		new_srq->log2_max_wqe_sz= WQE_ALIGN_SHIFT;

	wqe_sz= (1 << new_srq->log2_max_wqe_sz);
	MTL_DEBUG4(MT_FLFMT("%s: Allocating SRQ WQE of size %d."), __func__, wqe_sz);

	/* Compute real number of s/g entries based on rounded up WQE size */
	new_srq->max_sentries= (wqe_sz - wqe_base_sz) / WQE_SEG_SZ_SG_ENTRY;  
	/* Make sure we do not exceed reported HCA cap. */
	new_srq->max_sentries= (new_srq->max_sentries > hca_max_sentries) ? 
						   hca_max_sentries : new_srq->max_sentries;

	return HH_OK;
}


/* Allocate the WQEs buffer for sendQ and recvQ */
/* This function should be invoked after queue properties are set by alloc_init_qp */
static HH_ret_t alloc_wqe_buf(
							 /*IN*/ MT_bool in_ddr_mem,	  /* Allocation of WQEs buffer is requested in attached DDR mem. */
							 /*IN*/ u_int32_t hca_max_outs,	/* HCA cap. */
							 /*IN*/ u_int32_t req_max_outs,	/* Requested capabilities */
							 /*IN*/ u_int8_t  log2_max_wqe_sz,
							 /*OUT*/    XHHUL_srq_wqe_buf_t **buf_pp,
							 /*OUT*/    XHH_srq_ul_resources_t *srq_ul_resources_p
							 )
{
	u_int32_t wqe_sz;
	u_int32_t buf_sz;
	XHHUL_srq_wqe_buf_t *new_buf_p;
	u_int32_t adjusted_req_max_outs =  req_max_outs;

	/* Check requested capabilities */
	if ((req_max_outs == 0) || (req_max_outs > hca_max_outs)) {
		MTL_ERROR3(MT_FLFMT("%s: Got a request for a SRQ with %u WQEs - rejecting !"), __func__,
				   req_max_outs);
		return HH_E2BIG_WR_NUM;
	}

	XHHUL_DEV_SPECIFIC_ALLOC_SRQ_WQE_BUF_INIT;

	/* set buf_sz and wqe_sz after possible adjustments by above macro*/
	wqe_sz = 1 << log2_max_wqe_sz;
	buf_sz = adjusted_req_max_outs << log2_max_wqe_sz;

	new_buf_p= TMALLOC(XHHUL_srq_wqe_buf_t);
	if (new_buf_p == NULL) {
		MTL_ERROR2(MT_FLFMT("%s: Failed allocating new WQEs buffer context"), __func__);
		return HH_EAGAIN;
	}

	if (in_ddr_mem) { /* Allocate WQEs buffer by XHH_srqm in the attached DDR memory */
		new_buf_p->wqe_buf_orig= NULL;
		srq_ul_resources_p->wqes_buf= 0;   /* Allocate in attached DDR memory */
		new_buf_p->buf2reg = VA_NULL;
		new_buf_p->size2reg = 0;
	}
	else { /* Allocate WQEs buffer in main memory */
		/* Assure the buffer covers whole pages (no sharing of locked memory with other date) */
		new_buf_p->wqe_buf_orig_size = 
		(MOSAL_SYS_PAGE_SIZE-1)/* For alignment */+MT_UP_ALIGNX_U32(buf_sz, MOSAL_SYS_PAGE_SHIFT);
		/* Prevent other data reside in the last page of the buffer... */
		/* cover last page (last WQE can be at last page begin and its size is 64B min.)*/

		new_buf_p->wqe_buf_orig= malloc_within_4GB(new_buf_p->wqe_buf_orig_size,&new_buf_p->used_virt_alloc);
		if (new_buf_p->wqe_buf_orig == NULL) {
			MTL_ERROR2(MT_FLFMT("%s: Failed allocation of WQEs buffer of "SIZE_T_FMT" bytes within "
								"4GB boundries."), __func__, new_buf_p->wqe_buf_orig_size);
			FREE(new_buf_p);
			return HH_EAGAIN;
		}

		/* Page alignment assures that the WQE buffer is aligned to WQE size.
		 * In addition, after resizing the SRQ the new wqe_buf has the same alignment
		 * in the page (to retain original virtual addresses)                                        
		 */
		new_buf_p->wqe_buf= MT_UP_ALIGNX_VIRT((MT_virt_addr_t)(new_buf_p->wqe_buf_orig),
											  MOSAL_SYS_PAGE_SHIFT);

		srq_ul_resources_p->wqes_buf= new_buf_p->wqe_buf;
		new_buf_p->buf2reg = new_buf_p->wqe_buf;
		new_buf_p->size2reg = buf_sz;
	}

	XHHUL_DEV_SPECIFIC_ALLOC_SRQ_WQE_BUF_INIT_DONE;

	new_buf_p->max_outs= adjusted_req_max_outs; 

	srq_ul_resources_p->wqes_buf_sz= buf_sz;
	srq_ul_resources_p->wqe_sz= wqe_sz;

	*buf_pp= new_buf_p;
	return HH_OK;
}

static void free_wqe_buf(XHHUL_srq_wqe_buf_t *buf_p)
{
	if (buf_p->wqe_buf_orig != NULL) {/* WQEs buffer were allocated in process mem. */ 
		if (buf_p->used_virt_alloc)
			MOSAL_pci_virt_free_consistent(buf_p->wqe_buf_orig, buf_p->wqe_buf_orig_size);
		else
			MOSAL_pci_phys_free_consistent(buf_p->wqe_buf_orig, buf_p->wqe_buf_orig_size);    
	}
	FREE(buf_p);
}

/* Append WQEs buffer extention to WQEs free list */
/* Must be invoked with SRQ lock held */
static void append_to_free_list(
							   /*IN*/ XHHUL_srq_t srq,
							   /*IN*/ MT_virt_addr_t wqes_buf_extention, /* The part of the new WQEs buf above old part */
							   /*IN*/ MT_size_t extention_sz_in_wqes	 /* Extention size in WQEs */
							   )
{
	u_int32_t i;
	MT_virt_addr_t cur_wqe;
	const u_int32_t wqe_sz= 1<<srq->log2_max_wqe_sz;

	MTL_TRACE2(
			  MT_FLFMT("%s(srq=%p SRQn=0x%X , wqe_buf_extention="VIRT_ADDR_FMT
					   ", extention_sz_in_wqes="SIZE_T_DFMT),
			  __func__, srq, srq->srqn, wqes_buf_extention, extention_sz_in_wqes);
	/* Create free WQEs list of wqe_buf */
	/* Note that  srq->free_wqes_list  ALWAYS contains an HCA virtual address, and all the links in
	 * the free list are HCA virtual addresses.  To translate to Host VAs, need to add in the 
	 * real_virt_offset.  Furthermore, new items are always added to the TOP of the list.
	 */
	for (i= 0 , cur_wqe= wqes_buf_extention; i < extention_sz_in_wqes; i++) {
		*((MT_virt_addr_t*)cur_wqe) = 0;
		if (srq->free_wqes_list_top == 0) {
			/*list is empty.  First element is both top and bottom */
			srq->free_wqes_list_top = srq->free_wqes_list_end = cur_wqe - srq->real_virt_offset;
			srq->free_wqe_list_sz = 1;
		}
		else {
			/*list has at least one element.  Adding new element to end*/
			/* modify last element in list to point to new wqe */
			*((MT_virt_addr_t*)(srq->free_wqes_list_end + srq->real_virt_offset)) = cur_wqe - srq->real_virt_offset; 
			srq->free_wqes_list_end = cur_wqe - srq->real_virt_offset;
			srq->free_wqe_list_sz++;
		}
		/* Note that real_virt_offset reduction puts in the free_list the next virt in HCA's space */
		MTL_DEBUG4(MT_FLFMT("%s: Added to srq->free_wqes_list a WQE at "VIRT_ADDR_FMT),__func__,
				   srq->free_wqes_list_end);
		cur_wqe+= wqe_sz; 
	}
	MTL_DEBUG4(
			  MT_FLFMT("%s: SRQn=0x%X: Added "SIZE_T_DFMT" wqes. New size=%d"),
			  __func__, srq->srqn, extention_sz_in_wqes, srq->free_wqe_list_sz);

}

/* Allocate the auxilary WQEs data 
 * (a software context of a WQE which does not have to be in the registered WQEs buffer) */
static HH_ret_t alloc_aux_data_buf(
								  /*IN/OUT*/ XHHUL_srq_t new_srq
								  )
{
	/* RQ auxilary buffer: WQE ID per WQE */ 
	new_srq->wqe_id= (VAPI_wr_id_t*)
					 XHH_SMART_MALLOC(new_srq->cur_buf_p->max_outs * sizeof(VAPI_wr_id_t)); 
	if (new_srq->wqe_id == NULL) {
		MTL_ERROR1(MT_FLFMT("%s: Failed allocating SRQ auxilary buffer (for 0x%X WQEs IDs)."), __func__,
				   new_srq->cur_buf_p->max_outs);
		return HH_EAGAIN;
	}

	return HH_OK;
}

