/*
  This software is available to you under a choice of one of two
  licenses.  You may choose to be licensed under the terms of the GNU
  General Public License (GPL) Version 2, available at
  <http://www.fsf.org/copyleft/gpl.html>, or the OpenIB.org BSD
  license, available in the LICENSE.TXT file accompanying this
  software.  These details are also available at
  <http://openib.org/license.html>.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
  BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
  ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
  CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
  SOFTWARE.

  Copyright (c) 2004 Mellanox Technologies Ltd.  All rights reserved.
*/

#ifndef H_XHHUL_SRQM_INLINE_H
#define H_XHHUL_SRQM_INLINE_H


/**********************************************************************************************
 *                    Private inline functions 
 **********************************************************************************************/

/*********** WQE building functions ***********/

/* Init a not-connected (invalid) "next" segment (i.e. NDS=0) */
inline static u_int32_t WQE_init_next(u_int32_t *wqe_buf)
{
  memset(wqe_buf,0,WQE_SEG_SZ_NEXT);
  return WQE_SEG_SZ_NEXT;
}
/* Build the scatter list (pointer segments) */
inline static u_int32_t WQE_pack_slist(u_int32_t *segment_p,
  u_int32_t sg_lst_len,VAPI_sg_lst_entry_t *sg_lst_p, u_int32_t desc_sentries)
{
   u_int32_t i;
   u_int32_t *cur_loc_p= segment_p;

   for (i= 0; i < sg_lst_len; i++ , cur_loc_p+= WQE_SEG_SZ_SG_ENTRY_DW) {
     cur_loc_p[MT_BYTE_OFFSET(WQE_SEGMENT_DATA_PTR_ST,byte_count)>>2]= 
       (sg_lst_p[i].len & MASK32(31));
     cur_loc_p[MT_BYTE_OFFSET(WQE_SEGMENT_DATA_PTR_ST,l_key)>>2]= sg_lst_p[i].lkey;
     cur_loc_p[MT_BYTE_OFFSET(WQE_SEGMENT_DATA_PTR_ST,local_address_h)>>2]= 
       (u_int32_t)(sg_lst_p[i].addr >> 32);
     cur_loc_p[MT_BYTE_OFFSET(WQE_SEGMENT_DATA_PTR_ST,local_address_l)>>2]= 
       (u_int32_t)(sg_lst_p[i].addr & 0xFFFFFFFF);
   }
   
   if (sg_lst_len < desc_sentries) {
       for (;i < desc_sentries; i++ , cur_loc_p+= WQE_SEG_SZ_SG_ENTRY_DW) {
	   cur_loc_p[MT_BYTE_OFFSET(WQE_SEGMENT_DATA_PTR_ST,byte_count)>>2]= 0;
	   cur_loc_p[MT_BYTE_OFFSET(WQE_SEGMENT_DATA_PTR_ST,l_key)>>2]= SRQ_EMPTY_SENTRY_LKEY;
       }
   }
   return (u_int32_t)(((MT_virt_addr_t)cur_loc_p) - ((MT_virt_addr_t)segment_p));
}


/* Pack Control segment (for receive work requests) */
inline static u_int32_t WQE_pack_ctrl_recv(u_int32_t *segment_p,  
    VAPI_comp_type_t comp_type, u_int32_t event_bit)
{
  memset(segment_p,0,WQE_SEG_SZ_CTRL);  /* Clear all "RESERVED" */
  MT_INSERT_ARRAY32(segment_p,(comp_type == VAPI_SIGNALED) ? 1 : 0,
    MT_BIT_OFFSET(WQE_SEGMENT_CTRL_RECV_ST,c),MT_BIT_SIZE(WQE_SEGMENT_CTRL_RECV_ST,c));
  MT_INSERT_ARRAY32(segment_p,event_bit,
    MT_BIT_OFFSET(WQE_SEGMENT_CTRL_RECV_ST,e),MT_BIT_SIZE(WQE_SEGMENT_CTRL_RECV_ST,e));
  return WQE_SEG_SZ_CTRL;
}

inline static u_int32_t WQE_build_recv(
  XHHUL_srq_t srq,
  VAPI_rr_desc_t *recv_req_p,
  u_int32_t *wqe_buf
)
{
  u_int8_t *cur_loc_p= (u_int8_t*)wqe_buf; /* Current location in the WQE */

  XHH_DEV_SPECIFIC_WQE_BUILD_RECV;

  cur_loc_p+= WQE_pack_ctrl_recv((u_int32_t*)cur_loc_p,
    recv_req_p->comp_type, 0/*event bit*/);
  /* Pack scatter list segments */
  cur_loc_p+= WQE_pack_slist((u_int32_t*)cur_loc_p,recv_req_p->sg_lst_len,recv_req_p->sg_lst_p,
                             srq->max_sentries);
  
  return (u_int32_t)(((MT_virt_addr_t)cur_loc_p) - ((MT_virt_addr_t)wqe_buf));
}

#endif

