#ifndef H_UL_SRQM_PRIV_H
#define H_UL_SRQM_PRIV_H

#ifndef MT_KERNEL
/* instead of "ifdef"ing all over the code we define an empty macro */
#define MOSAL_pci_phys_free_consistent(addr,sz)  do {} while(0);
#endif

#define NULL_WQE_BUF_P ((void*)(-1)) /* To mark resize in progress but on shrinking request */

/* Limit kmalloc to 4 pages  */
#define WQ_KMALLOC_LIMIT (4*MOSAL_SYS_PAGE_SIZE)
#define SMALL_VMALLOC_AREA (1<<28)  /* VMALLOC area of 256MB or less is considered a scarce resource */

#define WQE_ALIGN_SHIFT 6        /* WQE address should be aligned to 64 Byte */
#define WQE_SZ_MULTIPLE_SHIFT 4           /* WQE size must be 16 bytes multiple */
/* WQE segments sizes */
#define WQE_SEG_SZ_NEXT (sizeof(struct WQE_SEGMENT_NEXT_ST)/8)            /* NEXT segment */
#define WQE_SEG_SZ_CTRL (sizeof(struct WQE_SEGMENT_CTRL_SEND_ST)/8)       /* CTRL segment */
#define WQE_SEG_SZ_SG_ENTRY (sizeof(struct WQE_SEGMENT_DATA_PTR_ST)/8)/* Scatter/Gather entry(ptr)*/
#define WQE_SEG_SZ_SG_ENTRY_DW (sizeof(struct WQE_SEGMENT_DATA_PTR_ST)/32)/* (same in DWORDs) */
/* #define MAX_WQE_SZ 1008 */

#define MAX_ALLOC_RETRY 3  /* Maximum retries to get WQEs buffer which does not cross 4GB boundry */

typedef struct XHHUL_srq_wqe_buf_st {
  MT_virt_addr_t wqe_buf;  /* The buffer for this queue WQEs - aligned to WQE size */
  u_int32_t max_outs; /* Max. outstanding (number of WQEs in buffer) */
  /* log2_max_wqe_sz is common to all buffers of a SRQ (no change) */
  void* wqe_buf_orig;   /* If != NULL then resizing is in progress */
  MT_bool used_virt_alloc;
  MT_size_t wqe_buf_orig_size;
  MT_virt_addr_t buf2reg;
  MT_size_t size2reg;
} XHHUL_srq_wqe_buf_t;

struct XHHUL_srq_st { /* SRQ context */
  MT_virt_addr_t hca_virt_wqe_buf;  /* The WQEs buffer base in HCA's virtual addr space */
  /* Note: wqe_buf is in HCA virt. space */ 
  MT_virt_addr_t real_virt_offset; /* Offset of real (OS) WQE virt. address from HCA's (resize)*/
  VAPI_wr_id_t *wqe_id; /* Array of max_outs entries for holding each WQE ID (WQE index based) */
  u_int32_t srqn;     /* SRQ number/index */
  u_int32_t cur_outs; /* Currently outstanding */
  u_int32_t max_sentries;  /* Max. Scatter list size */
  u_int8_t log2_max_wqe_sz; /* WQE size is a power of 2 (software implementation requirement) */
  MT_virt_addr_t free_wqes_list_top;  /* WQEs are taken from list top */
  MT_virt_addr_t free_wqes_list_end;  /* WQEs are returned to free list at bottom */
  int free_wqe_list_sz;
  u_int32_t *wqe_draft;
  MT_virt_addr_t last_posted_hca_va; /* Virtual addr of last posted WQE in HCA's addr space */
  MOSAL_spinlock_t q_lock;   /* Protect concurrent usage of the queue */
  HHUL_pd_hndl_t pd;
  XHHUL_srq_wqe_buf_t *cur_buf_p; /* Current WQEs buffer */
  XHHUL_srq_wqe_buf_t *resized_buf_p; /* Resized WQEs buffer */
  MT_bool resize_in_progress;   /* When set the resizing is in progress */
  /* Note: Resizing may be virtually in progress while resized_buf_p==NULL, when shrinking SRQ */
  struct XHHUL_srq_st *next; /* SRQs list */
  XHH_DEV_SPECIFIC_UL_SRQ_T_FIELDS
};
typedef struct XHHUL_srq_st *XHHUL_srq_t;

struct XHHUL_srqm_st { /* XHHUL_srqm_t is a pointer to this */
  struct XHHUL_srq_st* srqs_list;
  MOSAL_mutex_t srqm_lock;
};

#endif

