/*
  This software is available to you under a choice of one of two
  licenses.  You may choose to be licensed under the terms of the GNU
  General Public License (GPL) Version 2, available at
  <http://www.fsf.org/copyleft/gpl.html>, or the OpenIB.org BSD
  license, available in the LICENSE.TXT file accompanying this
  software.  These details are also available at
  <http://openib.org/license.html>.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
  BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
  ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
  CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
  SOFTWARE.

  Copyright (c) 2004 Mellanox Technologies Ltd.  All rights reserved.
*/

#ifndef H_XHH_H
#define H_XHH_H

#include <mtl_common.h>
#include <xhh_priv.h>

/* VERSION INFORMATION: used in order to retrieve major version numbers 
   in order to deal with differences in different versions. */
typedef struct XHH_ver_info_st {
  u_int32_t hw_ver;  /* HW version (stepping etc.)*/ 
  u_int16_t fw_ver_major;  /* Firmware major version */ 
  u_int16_t fw_ver_minor;  /* Firmware minor version */ 
  u_int16_t fw_ver_subminor;	/* Firmware Sub-minor version (Patch level).  */
  u_int16_t cmd_if_ver;  /* Command interface version */
}XHH_ver_info_t;


typedef u_int32_t             XHH_uar_index_t;

	
#define XHH_FW_VER_VALUE(major,minor,subminor)  \
  ( (((u_int64_t)(major)) << 32) | (((u_int64_t)(minor)) << 16) | ((u_int64_t)(subminor)) )


/* XHH objects handles */
typedef struct XHH_hob_st     *XHH_hob_t;
typedef MT_ulong_ptr_t XHH_cmd_t; /* type to identify the cmdif object */
typedef struct XHH_eventp_st  *XHH_eventp_t;
typedef struct XHH_uldm_st    *XHH_uldm_t; /* TK: for Tavor only - need to change when we start Arbel */
typedef struct XHH_lamm_st   *XHH_lamm_t;
typedef struct XHH_mrwm_st    *XHH_mrwm_t;  
typedef struct XHH_cqm_st     *XHH_cqm_t;
typedef struct XHH_qpm_st     *XHH_qpm_t;
typedef struct XHH_srqm_st     *XHH_srqm_t;
typedef struct XHH_mcgm_st    *XHH_mcgm_t;
typedef struct XHH_sqp_demux_st *XHH_sqp_demux_t;

#include "dev_common.h"

typedef struct {
    HH_hca_hndl_t hh_hca_hndl;
    /* PCI info. */
    u_int32_t     vendor_id;   /* As seen in the PCI config. header */ 
    u_int32_t	    device_id;   /* As seen in the PCI config. header */
    u_int8_t       pci_bus;
    u_int8_t       pci_devfn;
    MOSAL_pci_dev_t pci_device_p;  /* struct pci_dev of HCA */
    
    XHH_ver_info_t version;
    XHH_uar_index_t uar_index;
    MT_virt_addr_t   uar_map;
    /* HCA capabilities to validate or use in XHHUL */
    MT_bool         priv_ud_av; /* Privileged UD AV are enforced ? */
    u_int32_t		    log2_mpt_size;
    u_int32_t       max_qp_ous_wr;       /* Maximum Number of oustanding WR on any WQ.         */             
    u_int32_t       max_srq_ous_wr;       /* Maximum Number of oustanding WR on any WQ.         */             
    u_int32_t       max_num_sg_ent;      /* Max num of scatter/gather entries for desc other than RD */
    u_int32_t       max_num_sg_ent_srq;  /* Max num of scatter/gather entries for SRQs */
    u_int32_t       max_num_sg_ent_rd;   /* Max num of scatter/gather entries for RD desc      */
    u_int32_t       max_num_ent_cq;      /* Max num of supported entries per CQ                */
    u_int32_t       max_wqe_size_sq;      /* Max number of bytes in an SQ wqe */
    u_int32_t       max_wqe_size_rq;      /* Max number of bytes in an RQ wqe   */
    u_int32_t       max_wqe_size_srq;      /* Max number of bytes in an SRQ wqe  */
    u_int32_t       max_kernel_avs;      /* Max num of supported AV entries in kernel space    */
    VAPI_vsd_t	  vsd;	/* vendor specific data (56 dword array) */
} XHH_hca_ul_resources_t;

typedef struct {
    /* if user-level udavm_buf is used (i.e., non-zero), it should be malloc'ed
     * with size = (udavm_buf_sz + (1 udav entry size)), to allow the kernel level
     * to align the start of the udavm table to the entry size -- so some spare is
     * needed.  Therefore, the udavm_buf_size value is the size of the actual udavm
     * table, not the size of the malloc'ed buffer.
     */
    MT_virt_addr_t   udavm_buf;            /* IN */
    MT_size_t     udavm_buf_sz;         /* IN */
    HH_pdm_pd_flags_t  pd_flags;      /* IN - if non-zero, is a PD for a special QP*/
    VAPI_lkey_t   udavm_buf_memkey;     /* OUT - set by XHH_uldm */
} XHH_pd_ul_resources_t;

/* this macro ensure that total buff size is power of 2 and one extra entry for the cyclic buffer */
#define XHH_CYCLIC_BUFF_SIZE(entries) (1<<(floor_log2(entries)+1))

typedef u_int32_t XHH_eqn_t;
#define XHH_INVALID_EQN               0xFFFFFFFF

typedef struct XHH_udavm_st   *XHH_udavm_t; /* type to identify the udav object */


/* event's handlers types */
typedef u_int8_t XHH_event_type_t;
typedef u_int8_t XHH_event_subtype_t;
typedef void (*XHH_mlx_eventh_t)(HH_hca_hndl_t hh_hndl, 
                                 XHH_event_type_t event_type,
                                 XHH_event_subtype_t event_subtype,
                                 void* event_data, 
                                 void* private_data);

typedef union {
  HH_comp_eventh_t comp_event_h;
  HH_async_eventh_t ib_comp_event_h;
  XHH_mlx_eventh_t mlx_event_h;
}XHH_eventp_handler_t;

/* structure for passing module flags or parameters from 'insmod' to XHH_hob_create */
typedef struct XHH_module_flags_st {
    MT_bool legacy_sqp;  /* TRUE if should perform INIT_IB in XHH_hob_open_hca */
    MT_bool av_in_host_mem; /* TRUE if udav's should use host memory. */
                            /* FALSE if udav's should use DDR SDRAM on Tavor */
    MT_bool inifinite_cmd_timeout; /* when TRUE cmdif will wait infinitely for the completion of a command */
    MT_bool fatal_delay_halt; /* when TRUE, HALT_HCA/disable on fatal error will be delayed to before the reset */
    u_int32_t num_cmds_outs; /* max number of outstanding commands that will be used by the driver
                                The real value will not exceed tha value reported by fw */
    u_int32_t async_eq_size; /* The size of the async event queue (max # of outstanding async events) */
    MT_bool cmdif_post_uar0; /* when TRUE cmdif will post commands to uar0 */
    MT_bool legacy_name; /* TRUE if should use "InfiniHost" as device name */
    u_int32_t max_kernel_avs; /* max number of UD AVs in Kernel space.  0xFFFFFFFF indicates to use default */
	MT_bool eneble_sqp_sched; /* TRUE: Enable higher priority scheduling of SQPs in the HCA engines */
	u_int8_t pci_cap_max_read_request_size; /* max read size issued as a requester on the pci bus */
	u_int8_t pci_cap_read_byte_count;   /* max read byte count */
} XHH_module_flags_t;

/*
 * XHH_hob_state_t tracks the status of an HCA -- is it OK, or has a fatal error occurred.
 * Actually, the states used in practice use the FATAL states as modifiers of the base states.
 * Thus, the states we may see in practice are:
 *   XHH_STATE_CREATING, XHH_STATE_OPENING, XHH_STATE_RUNNING,  XHH_STATE_CLOSING, XHH_STATE_DESTROYING
 *     and fatal modifiers on these states:
 *                            
 *      XHH_STATE_CREATING | XHH_STATE_FATAL_HCA_HALTED
 *      XHH_STATE_OPENING | XHH_STATE_FATAL_HCA_HALTED
 * 
 *      XHH_STATE_RUNNING | XHH_STATE_FATAL_START
 *      XHH_STATE_RUNNING | XHH_STATE_FATAL_HCA_HALTED
 *
 *      XHH_STATE_CLOSING | XHH_STATE_FATAL_HCA_HALTED
 *      XHH_STATE_DESTROYING | XHH_STATE_FATAL_HCA_HALTED
 *
 *      Note that in the RUNNING state, have two FATAL possibilities.  When FATAL first occurs,
 *      we enter the RUNNING/FATAL_START state, in which all commands and all calls to XHH
 *      (with very few exceptions) return FATAL.  In addition, we attempt to halt the HCA.
 *      After the halt-hca attempt returns, we enter the RUNNING/FATAL-HCA-HALTED state.
 *                
 */
enum {
    XHH_STATE_NONE      = 0,
    XHH_STATE_CREATING  = 0x1, 
    XHH_STATE_CLOSED    = 0x2, 
    XHH_STATE_OPENING   = 0x4, 
    XHH_STATE_RUNNING   = 0x8, 
    XHH_STATE_CLOSING   = 0x10, 
    XHH_STATE_DESTROYING= 0x20, 
    XHH_STATE_FATAL_START  = 0x40,       /* CATASTROPHIC EVENT has been reported */
    XHH_STATE_FATAL_HCA_HALTED = 0x80   /* Failed HCA has been halted           */
};
typedef u_int32_t XHH_hob_state_t;

#define XHH_STATE_HAVE_ANY_FATAL   (XHH_STATE_FATAL_START | XHH_STATE_FATAL_HCA_HALTED)

/* Fatal event type enumeration, for passing to fatal event handlers */
typedef enum {
    XHH_FATAL_NONE, 
    XHH_FATAL_MASTER_ABORT,  /* detected master abort */
    XHH_FATAL_GOBIT,         /* GO bit of HCR remains set (i.e., stuck) */
    XHH_FATAL_CMD_TIMEOUT,   /* timeout on a command execution */
    XHH_FATAL_EQ_OVF,        /* an EQ has overflowed */
    XHH_FATAL_EVENT,         /* firmware has generated a LOCAL CATASTROPHIC ERR event */
    XHH_FATAL_CR,            /* unexpected read from CR-space */
    XHH_FATAL_TOKEN,         /* invalid token on command completion */
    XHH_FATAL_EXTERNAL,      /* externally-generated artificial fatal via XHH_hob_external_fatal */
    XHH_FATAL_END            /* indicates end of fatal error codes */
} XHH_fatal_err_t;

#define HCA_DEV_ID_LEGACY_PREFIX  "InfiniHost"
#define HCA_MAX_NAME 30

#endif  /* H_XHH_H */
