/*
  This software is available to you under a choice of one of two
  licenses.  You may choose to be licensed under the terms of the GNU
  General Public License (GPL) Version 2, available at
  <http://www.fsf.org/copyleft/gpl.html>, or the OpenIB.org BSD
  license, available in the LICENSE.TXT file accompanying this
  software.  These details are also available at
  <http://openib.org/license.html>.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
  BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
  ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
  CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
  SOFTWARE.

  Copyright (c) 2004 Mellanox Technologies Ltd.  All rights reserved.
*/
#ifndef __CS_MAPPING_H
#define __CS_MAPPING_H

#include <mtl_types.h>


/***************************************************************************************************
 * Function:  MOSAL_8xx_cs_map
 *
 * Description:
 *            Allocate a physical memory region to the given chipselect and memory
 *
 * Parameters:
 *      cs_num  (IN)      short
 *              Chip select to assign to this block.
 *      size    (IN)      u_int32_t
 *              The size of the memory which is to be mapped.   
 *      			
 * Returns: 
 *      physical address for a given chipselect.
 *
 *
 **************************************************************************************************/
MT_phys_addr_t MOSAL_8xx_cs_map(short cs_num, u_int32_t size);


/***************************************************************************************************
 * Function:  MOSAL_8xx_cs_unmap
 *
 * Description:
 *            Performs unmapping of given chip select.
 *
 * Parameters:
 *      cs_num  (IN)      short
 *              Chip select to unmap
 *
 * Returns: 
 *      MT_OK
 *
 **************************************************************************************************/
call_result_t MOSAL_8xx_cs_unmap(short cs_num); 

/***************************************************************************************************
 * Function:  MOSAL_8xx_cs_unmap_all
 *
 * Description:
 *            Performs unmapping of all mapped regions
 *
 * Parameters:
 *
 * Returns: 
 *      MT_OK
 *
 **************************************************************************************************/
call_result_t MOSAL_8xx_cs_unmap_all(void);




#endif /*__PPC_MAPPING_H*/
