#!/bin/sh 

# check that we get the two parameters 
x=${1:? kernel version must be defined}
x=${2:?TOPDIR must be defined}


KERNEL_REVISION=$1
TOPDIR=$2
KERNELOUTPUT=$3
LL_VERSION=`echo $1 | sed -e 's/\(2\.4\...\?\-.[0-9]\?\).*/\1/'`

RH_AS_3_0_LIST="2.4.21-9.EL@ 2.4.21-4.EL@"
RH_AS_3_0_FLAGS="-DRH_AS_3_0"

RH_9_0_LIST="2.4.20-8@"
RH_9_0_FLAGS="-DRH_9_0"


RH_7_3_LIST="2.4.18-14@"
RH_7_3_FLAGS="-DMT_REDHAT_73"

for dist in RH_AS_3_0 RH_9_0 RH_7_3 ; do
  eval echo \$"${dist}_LIST" | grep "$LL_VERSION" >> /dev/null
  if [ $? == 0 ]; then
    FLAGS=$(eval echo \$"${dist}_FLAGS") ;
  fi
done

if grep remap_page_range ${TOPDIR}/include/linux/mm.h | grep vm_area_struct &> /dev/null ; then
  FLAGS="$FLAGS -DEXTRA_REMAP_ARG"
fi

# Where the asm directory is depends on whether the kernel was compiled together with the
# source or output to a separate directory

if [ $# -gt 2 ]; then
    if [ -d ${TOPDIR}/${KERNELOUTPUT}/include2/asm ]; then
        ASMDIR=${TOPDIR}/${KERNELOUTPUT}/include2/asm
    else
        ASMDIR=${KERNELOUTPUT}/include2/asm
    fi
else
    ASMDIR=${TOPDIR}/include/asm
fi

if [ ! -e ${ASMDIR}/pgtable.h ]; then
    echo "${ASMDIR}/pgtable.h does not exist!"
    exit 1
elif grep pte_offset_map ${ASMDIR}/pgtable.h | grep pte_offset_map &> /dev/null ; then
  FLAGS="$FLAGS -DMT_PTE_OFFSET_MAP"
fi

if grep acct ${TOPDIR}include/linux/mm.h &> /dev/null ; then
  FLAGS="$FLAGS -DMT_UNMAP_SUFFIX=,1"
else
  FLAGS="$FLAGS -DMT_UNMAP_SUFFIX=  "
fi

if grep 'EXPORT_SYMBOL(nr_free_pages);' ${TOPDIR}/mm/page_alloc.c &> /dev/null; then
  FLAGS="$FLAGS -DMT_NR_FREE_PAGES"
fi

if grep remap_pfn_range ${TOPDIR}/include/linux/mm.h &> /dev/null ; then
  FLAGS="$FLAGS -DREMAP_VER_PFN"
fi


#find the pointers for the mlock/munlock pointers in System.map-xxx
SYSTEM_MAP_FILE=/boot/System.map-${KERNEL_REVISION}
if [ -r ${SYSTEM_MAP_FILE} ]; then
	mlock_ptr=$(grep 'T sys_mlock' ${SYSTEM_MAP_FILE} | grep -v 'mlockall' | cut -d ' ' -f 1)
	munlock_ptr=$(grep 'T sys_munlock' ${SYSTEM_MAP_FILE} | grep -v 'munlockall' | cut -d ' ' -f 1)
    FLAGS="$FLAGS -DMT_MLOCK_PTR=0x${mlock_ptr} -DMT_MUNLOCK_PTR=0x${munlock_ptr}"
fi

# DO NOT REMOVE CODE BELOW - it is commented out until we fix the efficient BIPPAGE translation flow
#if grep shmem_inode_info ${TOPDIR}include/linux/shmem_fs.h &> /dev/null ; then 
#  FLAGS="$FLAGS -DMT_BP_VIRT_TO_PHYS"
#fi

echo $FLAGS 
