/*
  This software is available to you under a choice of one of two
  licenses.  You may choose to be licensed under the terms of the GNU
  General Public License (GPL) Version 2, available at
  <http://www.fsf.org/copyleft/gpl.html>, or the OpenIB.org BSD
  license, available in the LICENSE.TXT file accompanying this
  software.  These details are also available at
  <http://openib.org/license.html>.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
  BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
  ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
  CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
  SOFTWARE.

  Copyright (c) 2004 Mellanox Technologies Ltd.  All rights reserved.
*/

 #ifndef MOSAL_WRAP_H
#define MOSAL_WRAP_H

#include <mosal.h>

#ifndef MIN
#define MIN(a,b)	((a) < (b) ? (a) : (b))
#endif

#ifndef FIELD_OFFSET
#define FIELD_OFFSET(type, field)    ((long)(int *)&(((type *)0)->field))
#endif

/* OS-dependent stuff, if any */
#include <mosal_wrap_imp.h>		

int mosal_ioctl_wrapper(MOSAL_ops_t fcn,
                        void* pi, u_int32_t pi_sz,
                        void* po, u_int32_t po_sz);


/*************************************************************************
 * Function: MOSAL_MPC860_present
 *************************************************************************/ 
struct i_MOSAL_MPC860_present_ops_t {
    IN int dummy;
  };
struct o_MOSAL_MPC860_present_ops_t {
    OUT bool ret;
  };

/*************************************************************************
 * Function: MOSAL_MPC860_read
 *************************************************************************/ 
   struct i_MOSAL_MPC860_read_ops_t {
    IN u_int32_t addr;
    IN u_int32_t size;
  };
  struct o_MOSAL_MPC860_read_ops_t {
    OUT call_result_t ret;
    OUT u_int8_t data[1];
  };

/*************************************************************************
 * Function: MOSAL_MPC860_write
 *************************************************************************/ 
  struct i_MOSAL_MPC860_write_ops_t {
    IN u_int32_t addr;
    IN u_int32_t size;
    IN u_int8_t data[1];
  };
  struct o_MOSAL_MPC860_write_ops_t {
    OUT call_result_t ret;
  };

/*************************************************************************
 * Function: MOSAL_PCI_find_device
 *************************************************************************/ 
  struct i_MOSAL_PCI_find_device_ops_t {
    IN u_int16_t vendor_id;
    IN u_int16_t dev_id;
    IN u_int16_t index;
  };
  struct o_MOSAL_PCI_find_device_ops_t {
    OUT call_result_t ret;
    OUT u_int8_t bus;
    OUT u_int8_t dev_func;
  };

/*************************************************************************
 * Function: MOSAL_PCI_get_cfg_hdr
 *************************************************************************/ 
 struct i_MOSAL_PCI_get_cfg_hdr_ops_t {
    IN MOSAL_pci_dev_t pci_dev;
  };
  struct o_MOSAL_PCI_get_cfg_hdr_ops_t {
    OUT call_result_t ret;
    OUT MOSAL_PCI_cfg_hdr_t cfg_hdr;
  };
 
/*************************************************************************
 * Function: MOSAL_PCI_present
 *************************************************************************/ 
  struct i_MOSAL_PCI_present_ops_t {
    IN int dummy;
  };
  struct o_MOSAL_PCI_present_ops_t {
    OUT bool ret;
  };

/*************************************************************************
 * Function: MOSAL_PCI_read_config_byte
 *************************************************************************/ 
  struct i_MOSAL_PCI_read_config_byte_ops_t {
    IN u_int8_t bus;
    IN u_int8_t dev_func;
    IN u_int8_t offset;
  };
  struct o_MOSAL_PCI_read_config_byte_ops_t {
    OUT call_result_t ret;
    OUT u_int8_t data;
  };

/*************************************************************************
 * Function: MOSAL_PCI_read_config_word
 *************************************************************************/ 
  struct i_MOSAL_PCI_read_config_word_ops_t {
    IN u_int8_t bus;
    IN u_int8_t dev_func;
    IN u_int8_t offset;
  };
  struct o_MOSAL_PCI_read_config_word_ops_t {
    OUT call_result_t ret;
    OUT u_int16_t data;
  };

/*************************************************************************
 * Function: MOSAL_PCI_read_config_dword
 *************************************************************************/ 
  struct i_MOSAL_PCI_read_config_dword_ops_t {
    IN u_int8_t bus;
    IN u_int8_t dev_func;
    IN u_int8_t offset;
  };
  struct o_MOSAL_PCI_read_config_dword_ops_t {
    OUT call_result_t ret;
    OUT u_int32_t data;
  };

/*************************************************************************
 * Function: MOSAL_PCI_read_io_byte
 *************************************************************************/ 
  struct i_MOSAL_PCI_read_io_byte_ops_t {
    IN u_int32_t addr;
  };
  struct o_MOSAL_PCI_read_io_byte_ops_t {
    OUT call_result_t ret;
    OUT u_int8_t data;
  };

/*************************************************************************
 * Function: MOSAL_PCI_read_io_word
 *************************************************************************/ 
  struct i_MOSAL_PCI_read_io_word_ops_t {
    IN u_int32_t addr;
  };
  struct o_MOSAL_PCI_read_io_word_ops_t {
    OUT call_result_t ret;
    OUT u_int16_t data;
  };

/*************************************************************************
 * Function: MOSAL_PCI_read_io_dword
 *************************************************************************/ 
  struct i_MOSAL_PCI_read_io_dword_ops_t {
    IN u_int32_t addr;
  };
  struct o_MOSAL_PCI_read_io_dword_ops_t {
    OUT call_result_t ret;
    OUT u_int32_t data;
  };

/*************************************************************************
 * Function: MOSAL_PCI_write_config_byte
 *************************************************************************/ 
  struct i_MOSAL_PCI_write_config_byte_ops_t {
    IN u_int8_t data;
    IN u_int8_t bus;
    IN u_int8_t dev_func;
    IN u_int8_t offset;
  };
  struct o_MOSAL_PCI_write_config_byte_ops_t {
    OUT call_result_t ret;
  };

/*************************************************************************
 * Function: MOSAL_PCI_write_config_word
 *************************************************************************/ 
  struct i_MOSAL_PCI_write_config_word_ops_t {
    IN u_int16_t data;
    IN u_int8_t bus;
    IN u_int8_t dev_func;
    IN u_int8_t offset;
  };
  struct o_MOSAL_PCI_write_config_word_ops_t {
    OUT call_result_t ret;
  };

/*************************************************************************
 * Function: MOSAL_PCI_write_config_dword
 *************************************************************************/ 
  struct i_MOSAL_PCI_write_config_dword_ops_t {
    IN u_int32_t data;
    IN u_int8_t bus;
    IN u_int8_t dev_func;
    IN u_int8_t offset;
  };
  struct o_MOSAL_PCI_write_config_dword_ops_t {
    OUT call_result_t ret;
  };

/*************************************************************************
 * Function: MOSAL_PCI_write_io_byte
 *************************************************************************/ 
  struct i_MOSAL_PCI_write_io_byte_ops_t {
    IN u_int32_t addr;
    IN u_int8_t data;
  };
  struct o_MOSAL_PCI_write_io_byte_ops_t {
    OUT call_result_t ret;
  };

  /*************************************************************************
 * Function: MOSAL_PCI_write_io_word
 *************************************************************************/ 
struct i_MOSAL_PCI_write_io_word_ops_t {
    IN u_int32_t addr;
    IN u_int16_t data;
  };
  struct o_MOSAL_PCI_write_io_word_ops_t {
    OUT call_result_t ret;
  };

/*************************************************************************
 * Function: MOSAL_PCI_write_io_dword
 *************************************************************************/ 
struct i_MOSAL_PCI_write_io_dword_ops_t {
    IN u_int32_t addr;
    IN u_int32_t data;
  };
  struct o_MOSAL_PCI_write_io_dword_ops_t {
    OUT call_result_t ret;
  };

/*************************************************************************
 * Function: MOSAL_get_counts_per_sec
 *************************************************************************/ 
  struct i_MOSAL_get_counts_per_sec_ops_t {
    IN int dummy;
  };
struct o_MOSAL_get_counts_per_sec_ops_t {
    OUT u_int64_t ret;
  };

/*************************************************************************
 * Function: MOSAL_get_page_shift
 *************************************************************************/ 
  struct i_MOSAL_get_page_shift_ops_t {
    IN MT_virt_addr_t va;
    IN MOSAL_prot_ctx_t prot_ctx;
  };
  struct o_MOSAL_get_page_shift_ops_t {
    OUT call_result_t ret;
    OUT unsigned int page_shift;
  };

/*************************************************************************
 * Function: MOSAL_is_privileged
 *************************************************************************/ 
  struct i_MOSAL_is_privileged_ops_t {
    IN int dummy;
  };
  struct o_MOSAL_is_privileged_ops_t {
    OUT MT_bool ret;
  };

/*************************************************************************
 * Function: MOSAL_virt_to_phys
 *************************************************************************/ 
  struct i_MOSAL_virt_to_phys_ops_t {
    IN MT_virt_addr_t va;
    IN MOSAL_prot_ctx_t prot_ctx;
  };
  struct o_MOSAL_virt_to_phys_ops_t {
    OUT call_result_t ret;
    OUT MT_phys_addr_t pa;
  };


/*************************************************************************
 * Function: MOSAL_undo_register
 *************************************************************************/ 
  struct i_MOSAL_undo_register_ops_t {
    IN MT_virt_addr_t va;
    IN MT_size_t size;
  };
  struct o_MOSAL_undo_register_ops_t {
    OUT call_result_t ret;
  };

/*************************************************************************
 * Function: k2u_cbk_pollq
 *************************************************************************/ 
  struct i_k2u_cbk_pollq_ops_t {
    IN k2u_cbk_hndl_t k2u_cbk_h;
  };
  struct o_k2u_cbk_pollq_ops_t {
    OUT MT_size_t size;
    OUT call_result_t ret;
    OUT k2u_cbk_id_t cbk_id;
    OUT u_int8_t data[MAX_CBK_DATA_SZ];
  };

/*************************************************************************
 * Function: mtl_log_set
 *************************************************************************/ 
  struct o_mtl_log_set_ops_t {
    IN int dummy;
  };
  struct i_mtl_log_set_ops_t {
    IN char layer[MAX_MTL_LOG_LAYER];
    IN char info[MAX_MTL_LOG_INFO];
  };


/* more OS-dependent stuff, if any */
#include <mosal_wrap_priv.h>		

#endif

