/*
  This software is available to you under a choice of one of two
  licenses.  You may choose to be licensed under the terms of the GNU
  General Public License (GPL) Version 2, available at
  <http://www.fsf.org/copyleft/gpl.html>, or the OpenIB.org BSD
  license, available in the LICENSE.TXT file accompanying this
  software.  These details are also available at
  <http://openib.org/license.html>.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
  BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
  ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
  CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
  SOFTWARE.

  Copyright (c) 2004 Mellanox Technologies Ltd.  All rights reserved.
*/
#define C_HHUL_C

			
#include <mosal.h>
#include <hhul.h>

#define HOB_CREATE_FUNC NULL
#ifdef MT_STATIC_LIB
#include <../mlxhh/ul_create.h>
#undef HOB_CREATE_FUNC
#define HOB_CREATE_FUNC XHHUL_hob_create
#endif
/* add includes for other Mellanox devices here */

HH_ret_t HHUL_alloc_hca_hndl
    (HHUL_init_info_t * init_info_p,
     void *hca_ul_resources_p, HHUL_hca_hndl_t * hhul_hca_hndl_p) {
	HH_ret_t rc = HH_ENODEV;
	HHUL_hob_create_t hob_create_p = HOB_CREATE_FUNC;
#ifndef MT_KERNEL
#ifndef MT_STATIC_LIB
	call_result_t mtrc;
	MOSAL_dll_hndl_t dll_handle;

	mtrc =
	    MOSAL_dll_open(init_info_p->user_devlib_info.devlib_filename, 0,
			   &dll_handle);
	if (mtrc != MT_OK) {
		MTL_ERROR1(MT_FLFMT
			   ("%s: Failed opening device library at %s (%s)"),
			   __func__,
			   init_info_p->user_devlib_info.devlib_filename,
			   mtl_strerror_sym(mtrc));
		return HH_ENOSYS;
	}

	mtrc =
	    MOSAL_dll_get_sym(dll_handle,
			      init_info_p->user_devlib_info.devlib_hob_create,
			      (void **)&hob_create_p);
	if (mtrc != MT_OK) {
		MTL_ERROR1(MT_FLFMT
			   ("%s: Failed finding function %s() in %s (%s)"),
			   __func__,
			   init_info_p->user_devlib_info.devlib_hob_create,
			   init_info_p->user_devlib_info.devlib_filename,
			   mtl_strerror_sym(mtrc));
		MOSAL_dll_close(dll_handle);
		return HH_ENOSYS;
	}
#endif
#else
	hob_create_p = init_info_p->kernel_ul_hob_create_p;

#endif

	if (hob_create_p == NULL) {
		MTL_ERROR1(MT_FLFMT("%s: Given NULL pointer for hob_create()"),
			   __func__);
		return HH_ENOSYS;
	}

	rc = hob_create_p(hca_ul_resources_p, hhul_hca_hndl_p);

	return rc;
}				/* HHUL_alloc_hca_hndl */

#include "hhulenosys.ic"

void HHUL_ifops_tbl_set_enosys(HHUL_if_ops_t * tbl)
{
	enosys_init(tbl);
}
