/*
 * This software is available to you under a choice of one of two
 * licenses.  You may choose to be licensed under the terms of the GNU
 * General Public License (GPL) Version 2, available at
 * <http://www.fsf.org/copyleft/gpl.html>, or the OpenIB.org BSD
 * license, available in the LICENSE.TXT file accompanying this
 * software.  These details are also available at
 * <http://openib.org/license.html>.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 * Copyright (c) 2004 Topspin Communications.  All rights reserved.
 *
 * $Id: tavor_mw.c 426 2004-07-12 04:39:39Z roland $
 */

#include <linux/errno.h>

#include "tavor_priv.h"

int tavor_mw_create(struct ib_pd *pd, struct ib_mw **mw, u32 *rkey)
{
	tTS_IB_TAVOR_PRIVATE priv = pd->device->private;
	VAPI_ret_t ret;

	ret = VAPI_alloc_mw(priv->vapi_handle,
			    *(VAPI_pd_hndl_t *) & pd->private,
			    (VAPI_mw_hndl_t *) & (*mw)->private, rkey);

	if (ret != VAPI_OK)
		TS_REPORT_WARN(MOD_KERNEL_IB,
			       "%s: VAPI_alloc_mw failed, return code = %d (%s)",
			       pd->device->name, ret, VAPI_strerror(ret));

	return (ret == VAPI_OK) ? 0 : -EINVAL;
}

int tavor_mw_destroy(struct ib_mw *mw)
{
	tTS_IB_TAVOR_PRIVATE priv = mw->device->private;
	VAPI_ret_t ret;

	ret = VAPI_dealloc_mw(priv->vapi_handle,
			      *(VAPI_mw_hndl_t *) & mw->private);

	if (ret != VAPI_OK)
		TS_REPORT_WARN(MOD_KERNEL_IB,
			       "%s: VAPI_dealloc_mw failed, return code = %d (%s)",
			       mw->device->name, ret, VAPI_strerror(ret));

	return (ret == VAPI_OK) ? 0 : -EINVAL;
}

int tavor_mw_bind(struct ib_qp *qp,
		  struct ib_mw *mw,
		  struct ib_mw_bind_param *param, u32 *new_rkey)
{
	tTS_IB_TAVOR_QP priv_qp = qp->private;
	tTS_IB_TAVOR_PRIVATE priv = mw->device->private;
	VAPI_ret_t ret;
	VAPI_mw_bind_t bind_prop;

	bind_prop.mr_lkey = param->lkey;
	bind_prop.start = param->address;
	bind_prop.size = param->length;
	bind_prop.acl = tsIbTavorAccessTranslate(param->access);

	/* XXX VAPI does not support fencing bind MW ops */
	ret = VAPI_bind_mw(priv->vapi_handle,
			   *(VAPI_mw_hndl_t *) & mw->private,
			   &bind_prop,
			   priv_qp->qp_handle.kernel_handle,
			   param->work_request_id,
			   param->signaled ? VAPI_SIGNALED : VAPI_UNSIGNALED,
			   new_rkey);

	if (ret != VAPI_OK)
		TS_REPORT_WARN(MOD_KERNEL_IB,
			       "%s: VAPI_bind_mw failed, return code = %d (%s)",
			       mw->device->name, ret, VAPI_strerror(ret));

	return (ret == VAPI_OK) ? 0 : -EINVAL;
}

/*
 * Local Variables:
 * c-file-style: "linux"
 * indent-tabs-mode: t
 * End:
 */
