/*
  This software is available to you under a choice of one of two
  licenses.  You may choose to be licensed under the terms of the GNU
  General Public License (GPL) Version 2, available at
  <http://www.fsf.org/copyleft/gpl.html>, or the OpenIB.org BSD
  license, available in the LICENSE.TXT file accompanying this
  software.  These details are also available at
  <http://openib.org/license.html>.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
  BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
  ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
  CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
  SOFTWARE.

  Copyright (c) 2004 Topspin Communications.  All rights reserved.

  $Id: tavor_priv.h,v 1.10 2004/03/04 02:10:05 roland Exp $
*/

#ifndef _TAVOR_PRIV_H
#define _TAVOR_PRIV_H

#include "ts_ib_tavor_provider.h"

#include <evapi.h>
#include <vapi_common.h>

#define TS_IB_TAVOR_OVERRIDE_NODE_DESCRIPTION 0

#define TS_IB_TAVOR_FMR_UNMAP_CHUNK_SIZE 256

typedef struct tTS_IB_TAVOR_PRIVATE_STRUCT tTS_IB_TAVOR_PRIVATE_STRUCT,
    *tTS_IB_TAVOR_PRIVATE;
typedef struct tTS_IB_TAVOR_CQ_STRUCT tTS_IB_TAVOR_CQ_STRUCT, *tTS_IB_TAVOR_CQ;
typedef struct tTS_IB_TAVOR_QP_STRUCT tTS_IB_TAVOR_QP_STRUCT, *tTS_IB_TAVOR_QP;

struct tTS_IB_TAVOR_PRIVATE_STRUCT {
	VAPI_hca_hndl_t  vapi_handle;
	char             node_desc[IB_NODE_DESC_SIZE];
	EVAPI_fmr_hndl_t fmr_unmap[TS_IB_TAVOR_FMR_UNMAP_CHUNK_SIZE];
	struct semaphore fmr_unmap_mutex;
};

struct tTS_IB_TAVOR_CQ_STRUCT {
	VAPI_cq_hndl_t cq_handle;
	EVAPI_compl_handler_hndl_t handler;
};

struct tTS_IB_TAVOR_QP_STRUCT {
	union {
		VAPI_qp_hndl_t kernel_handle;
		VAPI_k_qp_hndl_t user_handle;
	} qp_handle;
	enum ib_qp_state cached_state;
	int is_user:1;
};

#define TS_IB_MAX_TAVOR 32

void tsIbTavorAsyncHandler(VAPI_hca_hndl_t hca,
			   VAPI_event_record_t * event_record,
			   void *device_ptr);

int tsIbTavorDeviceQuery(struct ib_device *device,
			 struct ib_device_properties *properties);

int tsIbTavorPortQuery(struct ib_device *device,
		       tTS_IB_PORT port, struct ib_port_properties *properties);

int tsIbTavorPortModify(struct ib_device *device,
			tTS_IB_PORT port,
			struct ib_port_changes *properties);

int tsIbTavorPkeyQuery(struct ib_device *device,
		       tTS_IB_PORT port, int index, u16 *pkey);

int tsIbTavorGidQuery(struct ib_device *device,
		      tTS_IB_PORT port, int index, tTS_IB_GID gid);

int tsIbTavorPdCreate(struct ib_device *device,
		      void *device_specific, struct ib_pd *pd);

int tsIbTavorPdDestroy(struct ib_pd *pd);

int tsIbTavorAddressCreate(struct ib_pd *pd,
			   struct ib_address_vector *address_vector,
			   struct ib_address *address);

int tsIbTavorAddressDestroy(struct ib_address *address);

int tsIbTavorInitQp(void);                                      /*added by itayr*/

void tsIbTavorCleanUpQp(void);                                  /*added by itayr*/

int tsIbTavorQpCreate(struct ib_pd *pd, struct ib_qp_create_param *param,
		      struct ib_qp *qp);

int tsIbTavorSpecialQpCreate(struct ib_pd *pd,
			     struct ib_qp_create_param *param,
			     tTS_IB_PORT port,
			     enum ib_special_qp_type qp_type,
                             struct ib_qp *qp);

int tsIbTavorQpModify(struct ib_qp *qp, struct ib_qp_attribute *attr);

int tsIbTavorQpQuery(struct ib_qp *qp, struct ib_qp_attribute *attr);

int tsIbTavorQpDestroy(struct ib_qp *qp);
int tsIbTavorQpDetachDestroy(struct ib_qp *qp);

int tsIbTavorSendPost(struct ib_qp *qp,                                     /*changed by itayr*/
		      struct ib_send_param *param, int num_work_requests, int *actual_num_wr);

int tsIbTavorReceivePost(struct ib_qp *qp,                                  /*changed by itayr*/
		      struct ib_receive_param *param, int num_work_requests, int *actual_num_wr);

int tsIbTavorCqCreate(struct ib_device *device,
		      int *entries, void *device_specific, struct ib_cq *cq);

int tsIbTavorCqDestroy(struct ib_cq *cq);
int tsIbTavorCqStopCompEvents(struct ib_cq *cq);

int tsIbTavorCqResize(struct ib_cq *cq, int *entries);

int tsIbTavorCqPoll(struct ib_cq *cq,
		    int num_entries,
		    struct ib_cq_entry *entry);

int tsIbTavorCqArm(struct ib_cq *cq, int solicited);

int tsIbTavorMemoryRegister(struct ib_pd *pd,
			    void *start_address,
			    uint64_t buffer_size,
			    enum ib_memory_access access,
                            struct ib_mr *mr);

int tsIbTavorMemoryRegisterPhysical(struct ib_pd *pd,
				    struct ib_physical_buffer *buffer_list,
				    int list_len,
				    uint64_t * io_virtual_address,
				    uint64_t buffer_size,
				    uint64_t iova_offset,
				    enum ib_memory_access access,
                                    struct ib_mr *mr);

int tsIbTavorMemoryDeregister(struct ib_mr *mr);

int tavor_mw_create(struct ib_pd *pd, struct ib_mw **mw, u32 *rkey);
int tavor_mw_destroy(struct ib_mw *mw);
int tavor_mw_bind(struct ib_qp *qp, struct ib_mw *mw,
		  struct ib_mw_bind_param *param, u32 *new_rkey);

int tsIbTavorFmrCreate(struct ib_pd *pd, enum ib_memory_access access,
		       int max_pages, int max_remaps, struct ib_fmr *fmr);

int tsIbTavorFmrDestroy(struct ib_fmr *fmr);

int tsIbTavorFmrMap(struct ib_fmr *fmr,
		    uint64_t * page_list,
		    int list_len,
		    uint64_t * io_virtual_address,
		    uint64_t iova_offset,
		    u32 *lkey, u32 *rkey);

int tsIbTavorFmrUnmap(struct ib_device *device, struct list_head *fmr_list);

int tsIbTavorMulticastAttach(struct ib_qp *qp, u16 lid, tTS_IB_GID gid);

int tsIbTavorMulticastDetach(struct ib_qp *qp, u16 lid, tTS_IB_GID gid);

enum ib_mad_result tsIbTavorMadProcess(struct ib_device *device,
                                       int ignore_mkey,
                                       struct ib_mad *in_mad,
                                       struct ib_mad *response_mad);

void tsIbTavorSetNodeDesc(tTS_IB_TAVOR_PRIVATE priv, int index);

static inline VAPI_mrw_acl_t tsIbTavorAccessTranslate(enum ib_memory_access access)
{
	return (!!(access & IB_ACCESS_LOCAL_WRITE) * VAPI_EN_LOCAL_WRITE) |
	    (!!(access & IB_ACCESS_REMOTE_WRITE) * VAPI_EN_REMOTE_WRITE)  |
	    (!!(access & IB_ACCESS_REMOTE_READ) * VAPI_EN_REMOTE_READ)    |
	    (!!(access & IB_ACCESS_REMOTE_ATOMIC) * VAPI_EN_REMOTE_ATOM)  |
	    (!!(access & IB_ACCESS_ENABLE_WINDOW) * VAPI_EN_MEMREG_BIND);
}

#include "tavor_compat.h"

#endif				/* _TAVOR_PRIV_H */
