# environment
MT_KERNEL_REV ?= $(shell uname -r)
KERNEL_SRC_DIR = $(shell test -d /lib/modules/$(MT_KERNEL_REV)/source/ && echo /lib/modules/$(MT_KERNEL_REV)/source || echo /lib/modules/$(MT_KERNEL_REV)/build)
export KERNEL_SRC_DIR
VAPI_KERNEL_SRC_DIR ?= $(KERNEL_SRC_DIR)/drivers/infiniband/ib_verbs/hw/mellanox-hca
# For local development replace the line above with the line below
# VAPI_KERNEL_SRC_DIR = `pwd`/../kernel
export VAPI_KERNEL_SRC_DIR

MT_INSTALL_PREFIX ?= /usr
export MT_INSTALL_PREFIX

MT_KERNEL_MAJOR_DOT_MINOR = $(shell uname -r | cut -b 1-3)

# For now, big endian machines should provide this flag
MT_ENDIANESS ?= -DMT_LITTLE_ENDIAN

# static libs handling
MT_STATIC_LIB ?= 0
ifeq  "${MT_STATIC_LIB}" "1"
  LIB_EXT=a
else 
  LIB_EXT=so
endif

# CFLAGS

CFLAGS += -Wall -g -fno-strict-aliasing -D__LINUX__

# Debug
ifeq  "${MTCONF}" "gdb"
CFLAGS    += -O0 -g -DMAX_ERROR=4
else 
ifeq "${MTCONF}" "debug"
CFLAGS    += -O0 -g -DMAX_ERROR=8 -DMAX_TRACE=8 -DMAX_DEBUG=8
else
ifeq "${MTCONF}" "cover"
CFLAGS    += -O0 -DMAX_ERROR=4 -fprofile-arcs -ftest-coverage 
else
CFLAGS    += -O2 -DMAX_ERROR=4
endif
endif
endif

ifeq "${MT_KERNEL_MAJOR_DOT_MINOR}" "2.4"
	CFLAGS += -DMT_KVER_2_4
else
	ifeq "${MT_KERNEL_MAJOR_DOT_MINOR}" "2.6"
		CFLAGS += -DMT_KVER_2_6
	endif
endif

CFLAGS += -I. -I$(VAPI_KERNEL_SRC_DIR)/include 
CFLAGS += $(MT_ENDIANESS)

ifeq ($(MT_SUSPEND_QP),1)
  # Enable the un/suspend-qp verbes
  CFLAGS += -DMT_SUSPEND_QP
endif

CPU_FLAG_MMX := $(shell cat /proc/cpuinfo | grep -c -e '^flags *.*\<mmx\>.*$$' )
ifneq "$(CPU_FLAG_MMX)" "0"
  CFLAGS += -D__MMX__
endif

CPU_FLAG_SSE := $(shell cat /proc/cpuinfo | grep -c -e '^flags *.*\<sse\>.*$$' )
ifneq "$(CPU_FLAG_SSE)" "0"
  CFLAGS += -D__SSE__
endif
CPU_FLAG_SSE2 := $(shell cat /proc/cpuinfo | grep -c -e '^flags *.*\<sse2\>.*$$' )
ifneq "$(CPU_FLAG_SSE2)" "0"
  CFLAGS += -D__SSE2__
endif


# Architecture info (cross compile via ARCH environment)
SUBARCH := $(shell uname -m | sed -e s/i.86/i386/ -e s/sun4u/sparc64/ \
                                  -e s/arm.*/arm/ -e s/sa110/arm/ \
                                  -e s/s390x/s390/ -e s/parisc64/parisc/ )

ARCH            ?= $(SUBARCH)

# Common rules
# Installation targets
ifneq ($(words $(TAR_BINS)), 0)
  INST_TARGETS += inst_bins
endif
ifneq ($(words $(TAR_ROOT_BINS)), 0)
  INST_TARGETS += inst_root_bins
  inst_root_bins: inst_bins
endif
ifneq ($(words $(EXPORTED_HEADERS)), 0)
  INST_TARGETS += inst_headers
endif

ifeq ($(ARCH),x86_64)
  LIB_DIR_SUFFIX = lib64
else
  LIB_DIR_SUFFIX = lib
endif # x86_64

ifneq ($(words $(TAR_LIBS)), 0)
  INST_TARGETS += inst_libs
  # Library rules
  LIB_TARGETS = $(TAR_LIBS:%=%.$(LIB_EXT))
endif # TAR_LIBS

ifeq ($(ARCH),ia64)
  ifeq ($(shell ( [ -f /lib/ssa/libgcc_s.so ] || [ -f /usr/lib/libgcc_s.so ] ) && echo yes),yes)
    ifneq  "${MT_STATIC_LIB}" "1"
      CFLAGS += -L /lib/ssa -L /usr/lib -lgcc_s
    endif
  endif
endif

TARGETS = $(LIB_TARGETS) $(TAR_BINS) 

all: $(TARGETS)

.Phony: install inst_libs inst_bins inst_headers clean inst_root_bins

install: $(INST_TARGETS)


inst_libs: $(LIB_TARGETS)
	mkdir -p $(MT_INSTALL_PREFIX)/$(LIB_DIR_SUFFIX)
	install -p -m 0755 $^ $(MT_INSTALL_PREFIX)/$(LIB_DIR_SUFFIX)

inst_bins: $(TAR_BINS)
	mkdir -p $(MT_INSTALL_PREFIX)/bin
	install -p -m 0755 $^ $(MT_INSTALL_PREFIX)/bin

inst_headers: $(EXPORTED_HEADERS)
	mkdir -p $(MT_INSTALL_PREFIX)/include
	install -p -m 0644 $^ $(MT_INSTALL_PREFIX)/include

inst_root_bins: $(TAR_ROOT_BINS)
	install -p -m 0544 $(TAR_ROOT_BINS) $(MT_INSTALL_PREFIX)/bin

clean:
	\rm -f $(TARGETS)   &> /dev/null
	\rm -f *.o          &> /dev/null
	\rm -f `find . -type l -name "*.c" ` &> /dev/null


