/*
  This software is available to you under a choice of one of two
  licenses.  You may choose to be licensed under the terms of the GNU
  General Public License (GPL) Version 2, available at
  <http://www.fsf.org/copyleft/gpl.html>, or the OpenIB.org BSD
  license, available in the LICENSE.TXT file accompanying this
  software.  These details are also available at
  <http://openib.org/license.html>.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
  BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
  ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
  CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
  SOFTWARE.

  Copyright (c) 2004 Mellanox Technologies Ltd.  All rights reserved.
*/
#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#define CMD_PARAMS_C_FILE
#include "cmd_pars.h"

static int process_arg( int opt_index, char* equ_ptr, cmd_line_params_t *cmd_line_params_p );
static int should_print(char *str);

int parse_argv( int my_argc, char** my_argv, cmd_line_params_t *cmd_line_params_p)
{
    int i,i2,len;
    int rc;
    char* equ_ptr = 0;

//    for ( i2=0; i2<OPTIONS_N; i2++ ){
//      printf("filed=%s offset=%d\n",usage_descriptor[i2].str_option,usage_descriptor[i2].param_offset);
//    }

    for ( i=1; i<my_argc; i++ ) {
        if ( my_argv[i][0] != '-' ) {

            printf( "Wrong parameter \"%s\"\n", my_argv[i] );
            usage(USAGE_MINIMAL);
            return -1;
        }

        equ_ptr = 0;

        if ( my_argv[i][1] != '-' ) { // option in short form "-x"
            for ( i2=0; i2<OPTIONS_N; i2++ ) {
                if ( my_argv[i][1] == usage_descriptor[i2].char_option ) {
                    if (usage_descriptor[i2].param_type != 'p' &&
                        usage_descriptor[i2].param_type != 'N') {
                        if ( my_argv[i][2] != '\0' ) {
                            equ_ptr = &my_argv[i][2];   // Option in form -xVALUE
                            if ( *equ_ptr == '=' ) {
                                equ_ptr++;   // Option in form -x=VALUE
                            }

                        }
                        else { // Option in form -x VALUE
                            if ( (i+1 == my_argc) || (my_argv[i+1][0] == '-') ) {
                                // This was last argument or next argument starts with '-'
                                printf( "Option %s is not complete\n", my_argv[i] );
                                usage(USAGE_MINIMAL);
                                return -1;
                            }
                            i++;
                            equ_ptr = my_argv[i];
                        }
                    }

                    rc = process_arg( i2, equ_ptr,cmd_line_params_p);
                    if ( rc ) {
                        return rc;
                    }
                    break;
                }
            }
        }
        else {                        // option in verbose form "--option"
            for ( i2=0; i2<OPTIONS_N; i2++ ) {
                len = (int)strlen( usage_descriptor[i2].str_option);
                if ( !strncmp(usage_descriptor[i2].str_option, &my_argv[i][2], len) ) {
                    if (usage_descriptor[i2].param_type != 'p') {
                        if (my_argv[i][2+len]=='=') {
                            equ_ptr = &(my_argv[i][3+len]);
                        }
                        else{
                            i++;
                            equ_ptr = my_argv[i];
                        }
                    }
                    rc = process_arg( i2, equ_ptr,cmd_line_params_p );
                    if ( rc )
                        return rc;
                    break;
                }
            }
        }

        if ( i2 == OPTIONS_N ) {
            printf( "Unknown option %s\n", my_argv[i] );
            usage(USAGE_MINIMAL);
            return -1;
        }
    }/*end for*/

    return 0;
}

int init_cmd_line_params(cmd_line_params_t *cmd_line_params_p)
{
    int i;
    void *field_p;

    for ( i=0; i<OPTIONS_N; i++ ) {
        field_p = (void *) ((char*)cmd_line_params_p + usage_descriptor[i].param_offset);
        switch (usage_descriptor[i].param_type) {
            case 's':
                strcpy((char *)field_p, usage_descriptor[i].default_value);
                break;
            case 'N':
                break;
            default : //integer
                *((int *)field_p) = atoi(usage_descriptor[i].default_value);
                break;
        }
    }
    return 0;
}

int print_cmd_line_params(cmd_line_params_t *cmd_line_params_p)
{
    int i;
    void *field_p;

    for ( i=0; i<OPTIONS_N; i++ ) {
        field_p = (void *) ((char*)cmd_line_params_p + usage_descriptor[i].param_offset);
        printf("%s  ",usage_descriptor[i].str_option);
        switch (usage_descriptor[i].param_type) {
            case 's':
                printf("%s\n",(char *) field_p);
                break;
            default : //integer
                printf("%d\n", *((int *) field_p));
                break;
        }
    }
    return 0;
}


static int process_arg( int opt_index, char* equ_ptr, cmd_line_params_t *cmd_line_params_p )
{
    int rc=0;
    void *field_p;

    field_p = (void *) ((char*)cmd_line_params_p + usage_descriptor[opt_index].param_offset);

    if (usage_descriptor[opt_index].param_type == 'p') {
        *((int *)field_p) = 1;
        return 0;
    }

    switch (usage_descriptor[opt_index].param_type) {
        case 'N':
            if(usage_descriptor[opt_index].char_option == 'h'){
                usage(USAGE_CUSTOMER);
            }
            if(!strcmp(usage_descriptor[opt_index].str_option,"leo")){
                usage(USAGE_ALL);
            }
            break;
        case 's':
            strcpy((char *)field_p, equ_ptr);
            break;
        case 'i':
            *((int *)field_p) = atoi(equ_ptr);
            break;

    }
    return rc;
}

void usage(help_usage_levels_t help_level)
{
    int i;
    int cntr;

    switch (help_level) {
        case USAGE_ALL:
        case USAGE_CUSTOMER:
            printf("Usage:   set_reg_params [options]\n\n"
                   "Example: request card reset\n"
                   "         set_reg_params --reset=1\n\n");
            printf("Options:\n");
            for ( i=0; i<OPTIONS_N; i++ ) {
                if ( usage_descriptor[i].help_desc[0] == '\0' )
                    continue;           // Skip options without help context
                
                if(help_level == USAGE_CUSTOMER){
                    if(!should_print(usage_descriptor[i].str_option)){
                        continue;
                    }
                }

                if ( usage_descriptor[i].char_option != ' ' )
                    printf( "-%c, ",usage_descriptor[i].char_option );
                else
                    printf( "    " );

                printf("--");
                cntr = printf( "%.15s", usage_descriptor[i].str_option);
                if ( usage_descriptor[i].str_value[0] ){
                    cntr+=printf( "=%-7s", usage_descriptor[i].str_value );
                }

                while ( cntr++ < 18 )
                    putchar(' ');

                printf( " %s" ,usage_descriptor[i].help_desc );  
                printf( " (default %s, current %d)\n" ,usage_descriptor[i].default_value ,usage_descriptor[i].current_dw_value );  
            }
            break;
        default:
            printf( "  Type --help|-h to get more information\n" );
            break;
    }
    exit(0);
}

static int should_print(char *str)
{
    
    int i;
    
    int ret=0;
    int num_of_opts = sizeof(help_vect_str)/sizeof(help_vect_str[0]);
    
    for (i=0;i<num_of_opts;i++) {
        if(!(strcmp(str,help_vect_str[i]))){
            ret=1;
            break;
        }
    }
    return ret;

}

int get_params_num() { return OPTIONS_N; }
