/*
  This software is available to you under a choice of one of two
  licenses.  You may choose to be licensed under the terms of the GNU
  General Public License (GPL) Version 2, available at
  <http://www.fsf.org/copyleft/gpl.html>, or the OpenIB.org BSD
  license, available in the LICENSE.TXT file accompanying this
  software.  These details are also available at
  <http://openib.org/license.html>.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
  BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
  ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
  CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
  SOFTWARE.

  Copyright (c) 2004 Mellanox Technologies Ltd.  All rights reserved.
*/

#define C_MOSAL_DLL_C

#ifdef MT_KERNEL
#error MOSAL_dll functions are implemented for user space only
#endif

#include <dlfcn.h>
#include <mtl_common.h>
#include <mosal_dll.h>

call_result_t MOSAL_dll_open(const char* dllpath, 
                             const MOSAL_dll_load_flags_t load_flags,
                             MOSAL_dll_hndl_t *dll_hndl_p)
{
  int flags= 0;

  flags= (load_flags & MOSAL_DLL_LOAD_GLOBAL ? RTLD_GLOBAL : 0) |
         (load_flags & MOSAL_DLL_LOAD_LAZY   ? RTLD_LAZY   : RTLD_NOW);

  *dll_hndl_p= dlopen(dllpath, flags);
  if (*dll_hndl_p == NULL) {
    MTL_ERROR4(MT_FLFMT("%s: Failed loading of %s (%s)"), __func__, dllpath, dlerror());
    return MT_ENOENT;
  }
  
  return MT_OK;
}

call_result_t MOSAL_dll_close(MOSAL_dll_hndl_t dll_hndl)
{
  int ret;

  ret= dlclose(dll_hndl);
  if (ret != 0)  {
    MTL_ERROR4(MT_FLFMT("%s: Failed closing library (%s)"), __func__, dlerror());
    return MT_EINVAL;
  }

  return MT_OK;
}

call_result_t MOSAL_dll_get_sym(MOSAL_dll_hndl_t dll_hndl, const char* sym, void **sym_pp)
{
  char *error;

  *sym_pp= dlsym(dll_hndl, sym);
  if ((error = dlerror()) != NULL) {
    MTL_ERROR4(MT_FLFMT("%s: Failed resolving of symbol %s (%s)"), __func__, sym, error);
    return MT_EINVAL;
  }
  
  return MT_OK;
}

