/*
  This software is available to you under a choice of one of two
  licenses.  You may choose to be licensed under the terms of the GNU
  General Public License (GPL) Version 2, available at
  <http://www.fsf.org/copyleft/gpl.html>, or the OpenIB.org BSD
  license, available in the LICENSE.TXT file accompanying this
  software.  These details are also available at
  <http://openib.org/license.html>.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
  BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
  ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
  CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
  SOFTWARE.

  Copyright (c) 2004 Mellanox Technologies Ltd.  All rights reserved.
*/
 
 
/*
*-------------------------------------------------------------------
* 									   
*-------------------------------------------------------------------								 
* Leonid Keller 								        
*-------------------------------------------------------------------
*/


#include <unistd.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <mtl_common.h>
#include "mosal_wrap.h"


// the name in the /dev dir this should be mknod and then open
static const char*  MOSAL_DEVICE_FILE_NAME = "/dev/mosal";
static int file_desc=-1;
static bool dev_open = FALSE;

int mosal_ioctl_open(void) 
{
    static pthread_mutex_t dev_mutex = PTHREAD_MUTEX_INITIALIZER;
    MTL_DEBUG4(MT_FLFMT("called mosal_ioctl_open"));
    /* when the first time - open the device  */
    pthread_mutex_lock(&dev_mutex);
    if(!dev_open){
        file_desc = open(MOSAL_DEVICE_FILE_NAME,0);
        if(file_desc < 0){
            pthread_mutex_unlock(&dev_mutex);
            MTL_ERROR1("%s: can't open device file %s\n", __func__, MOSAL_DEVICE_FILE_NAME);
            return MT_ERROR;
        }
        dev_open = TRUE;
    }
    pthread_mutex_unlock(&dev_mutex);

    return MT_OK;
}

call_result_t _dev_mosal_open(  int *fd_p )
{
	if(!dev_open){
	    if (mosal_ioctl_open())
	        return MT_ERROR;
	}
	*fd_p = 	file_desc;
	return MT_OK;
}
   



int mosal_ioctl_wrapper(MOSAL_ops_t ops, void *pi, u_int32_t pi_sz, void *po, u_int32_t po_sz )
{     
    int rc = 0;
    mosal_ioctl_wrap_t mosal_ioctl_wrap;

    MTL_DEBUG4(MT_FLFMT("dev_open=%d, fcn=%d, MOSAL_IOCTL_CALL=0x%x"), 
                        dev_open, ops, (int) MT_IOCTL_CALL);
    
    /* fill the parameter structure */
    mosal_ioctl_wrap.ops		= (int)ops;
    mosal_ioctl_wrap.pi 		= pi;
    mosal_ioctl_wrap.pi_sz 		= pi_sz;
    mosal_ioctl_wrap.po 		= po;
    mosal_ioctl_wrap.po_sz 		= po_sz;

    if(!dev_open){
        if (mosal_ioctl_open())
	        return MT_ERROR;
    }

    /* issue the ioctl */
    rc = ioctl(file_desc,MT_IOCTL_CALL,&mosal_ioctl_wrap);

    return rc;
}



int mosal_ioctl_close(void)
{
  int rc=MT_OK, fd;

  if ( dev_open ) {
    fd = file_desc;
    dev_open = FALSE;
    file_desc=-1;
    rc = close(fd);
  }
  return rc;
}

int mosal_dev_filedesc(void)
{
  if ( dev_open ) {
    return file_desc;
  }

  return -1;
}

