/*
  This software is available to you under a choice of one of two
  licenses.  You may choose to be licensed under the terms of the GNU
  General Public License (GPL) Version 2, available at
  <http://www.fsf.org/copyleft/gpl.html>, or the OpenIB.org BSD
  license, available in the LICENSE.TXT file accompanying this
  software.  These details are also available at
  <http://openib.org/license.html>.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
  BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
  ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
  CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
  SOFTWARE.

  Copyright (c) 2004 Mellanox Technologies Ltd.  All rights reserved.
*/



#include <stdio.h>
#include <stdlib.h>
#include <vapi.h>
#include <evapi.h>
#include <vapi_common.h>

void usage(void)
{
    fprintf(stderr, "Usage: check_catastrophic <Device Name>\n");
}

int MT_API main(int argc, char* argv[])
{
    VAPI_ret_t rc;
    VAPI_hca_hndl_t hca_hndl;
    VAPI_hca_vendor_t hca_vendor;
    VAPI_hca_cap_t      hca_cap;

//Check arguments
    if ( argc != 2 ) {
        usage();
        return(1);
    }

//Try to get HCA handle    
    rc = EVAPI_get_hca_hndl (argv[1], &hca_hndl);
    if ( rc != VAPI_OK ) {
        fprintf(stderr, "Could not get HCA hndl for %s\n", argv[1]);
        return rc;
    }
//Get hca properties
    rc = VAPI_query_hca_cap(hca_hndl, &hca_vendor, &hca_cap);
    if ( rc != VAPI_OK ) {
        fprintf(stderr, "Could not query HCA %s\n", argv[1]);
        return rc;
    }
         rc = EVAPI_release_hca_hndl (hca_hndl);
    if ( rc != VAPI_OK ) {
        fprintf(stderr, "Could not release HCA hndl\n");
        return rc;
    }

    return 0;
}
