/*
  This software is available to you under a choice of one of two
  licenses.  You may choose to be licensed under the terms of the GNU
  General Public License (GPL) Version 2, available at
  <http://www.fsf.org/copyleft/gpl.html>, or the OpenIB.org BSD
  license, available in the LICENSE.TXT file accompanying this
  software.  These details are also available at
  <http://openib.org/license.html>.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
  BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
  ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
  CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
  SOFTWARE.

  Copyright (c) 2004 Mellanox Technologies Ltd.  All rights reserved.
*/

#include <stdio.h>
#include <mtl_common.h>
#include <vapi.h>
#include <vapi_common.h>
#include <evapi.h>

int main(int argc, char* argv[])
{
  VAPI_hca_id_t cur_hca, *hca_id_buf_p ;
  VAPI_ret_t vret;
  u_int32_t     num_hcas, returned_num_hcas;
  int i;
  int main_ret = 0;

  if (argc > 1) { /* Close HCAs given in command line */
    for (i= 1; i< argc; i++) {
      strcpy(cur_hca,argv[i]);
      vret= EVAPI_close_hca(cur_hca);
      if (vret == VAPI_EBUSY) {
        printf("HCA [%s] is still in use. Release all allocated resources before closing.\n",cur_hca);
	main_ret++;
      } else if (vret == VAPI_OK) {
        printf("Closed HCA [%s] .\n",cur_hca);
      } else if (vret == VAPI_EINVAL_HCA_ID) {
        printf("HCA [%s] already closed, or invalid HCA ID\n", cur_hca);
      } else {
        printf("Failed closing HCA [%s] - %s .\n",cur_hca,VAPI_strerror_sym(vret));
        /* main_ret++ ;*/
      }
    }
  
  } else { /* Try opening all known devices */
    //printf("Trying to close all known HCAs:\n");

    /* first, find how many HCAs available */
    vret = EVAPI_list_hcas(0,&num_hcas,NULL);
    if (vret != VAPI_EAGAIN && vret != VAPI_OK) {
        printf("Error: could not get number of available hcas (err = %d - %s)\n", 
                vret, VAPI_strerror_sym(vret));
	/* main_ret++; */
        goto done;
    } 
    if (num_hcas == 0) {
        printf("No HCAs are available for closing.\n");
        goto done;
    }

    /* Allocate an HCA ID array with enough entries for all available HCAs*/
    hca_id_buf_p = (VAPI_hca_id_t *) VMALLOC(num_hcas * sizeof(VAPI_hca_id_t));
    if (hca_id_buf_p == NULL) {
        MT_COV_MEM_FLTR_START
        printf("could not allocate memory for the HCA IDs array\n");
	/* main_ret++; */
        goto done;
        MT_COV_MEM_FLTR_END
    }
    vret = EVAPI_list_hcas(num_hcas, &returned_num_hcas, hca_id_buf_p);
    if (vret != VAPI_OK) {
        printf("could not obtain HCA IDs (error = %d - %s)\n", 
                vret, VAPI_strerror_sym(vret));
	/* main_ret++; */
        goto done;
    }
    for (i= 0; i < (int)returned_num_hcas; i++) {
      // printf("hca_id[%d] = %s - ", i, hca_id_buf_p[i]);
      strcpy(cur_hca,hca_id_buf_p[i]);
      vret= EVAPI_close_hca(cur_hca);
      if (vret == VAPI_EBUSY) {
        printf("%s - HCA is still in use. HCA resources must be released before closing it.\n", cur_hca);
	main_ret++;
      } else if (vret == VAPI_OK) {
        printf("%s - Closed.\n",cur_hca);
      } else if (vret == VAPI_EINVAL_HCA_ID) {
        printf("%s - HCA already closed\n", cur_hca);
      } else {
        printf("%s - FAILED to close (%s) \n", cur_hca, VAPI_strerror_sym(vret));
        /* main_ret++ ; */
      }
    } /* for */
    if (i == 0) {
      printf("No HCAs were found\n");
    }
  }
done:
  return main_ret;	
}
