/*
  This software is available to you under a choice of one of two
  licenses.  You may choose to be licensed under the terms of the GNU
  General Public License (GPL) Version 2, available at
  <http://www.fsf.org/copyleft/gpl.html>, or the OpenIB.org BSD
  license, available in the LICENSE.TXT file accompanying this
  software.  These details are also available at
  <http://openib.org/license.html>.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
  BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
  ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
  CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
  SOFTWARE.

  Copyright (c) 2004 Mellanox Technologies Ltd.  All rights reserved.
*/

#include <stdio.h>
#include <mtl_common.h>
#include <vapi.h>
#include <vapi_common.h>
#include <evapi.h>
#include <iniparser.h>
#include <mt_getopt.h>

//#define NUM_O_HCAS 2
//static char* hca_ids[NUM_O_HCAS]= {"InfiniHost0" , "mt21108_pci0"};


#define MT_PATHS_FILE  "/etc/vapi-linux-release"
#define MT_LINESIZE   1024
#define MT_SRC_ENV_VAR  "src_path"

static int get_sec(dictionary *ini, char *name, EVAPI_hca_profile_t *profile) {
    char entry[128];
    if (iniparser_find_entry(ini, name) == 0) {
        return -1;
    }

    strcpy((char *) entry, name); strcat((char *) entry, ":NUM_QP");
    profile->num_qp = iniparser_getint(ini, entry, -1);
    strcpy((char *) entry, name); strcat((char *) entry, ":NUM_CQ");
    profile->num_cq = iniparser_getint(ini, entry, -1);
    strcpy((char *) entry, name); strcat((char *) entry, ":NUM_PD");
    profile->num_pd = iniparser_getint(ini, entry, -1);
    strcpy((char *) entry, name); strcat((char *) entry, ":NUM_MR");
    profile->num_mr = iniparser_getint(ini, entry, -1);
    strcpy((char *) entry, name); strcat((char *) entry, ":NUM_MW");
    profile->num_mw = iniparser_getint(ini, entry, -1);
    strcpy((char *) entry, name); strcat((char *) entry, ":MAX_QP_OUS_RD_ATOM");
    profile->max_qp_ous_rd_atom = iniparser_getint(ini, entry, -1);
    strcpy((char *) entry, name); strcat((char *) entry, ":MAX_MCG");
    profile->max_mcg = iniparser_getint(ini, entry, -1);
    strcpy((char *) entry, name); strcat((char *) entry, ":QP_PER_MCG");
    profile->qp_per_mcg = iniparser_getint(ini, entry, -1);
    
    strcpy((char *) entry, name); strcat((char *) entry, ":require");
    profile->require = iniparser_getboolean(ini, entry, 0);

    return 0;
}

static int get_default_sec(dictionary *ini, EVAPI_hca_profile_t *profile) {
    return get_sec(ini, "InfiniHostDefault", profile);
}

static int print_profile(EVAPI_hca_profile_t *profile) {
        printf("profile.num_qp=%d\nprofile.num_cq=%d\nprofile.num_pd=%d\nprofile.num_mr=%d\nprofile.num_mw=%d\n",
               profile->num_qp,profile->num_cq,profile->num_pd,profile->num_mr,profile->num_mw);
        printf("profile.max_qp_ous_rd_atom=%d\nprofile.max_mcg=%d\nprofile.qp_per_mcg=%d\nprofile.require=%s\n",
               profile->max_qp_ous_rd_atom,profile->max_mcg,profile->qp_per_mcg,
               (profile->require == 0) ? "FALSE":"TRUE");
        return 0;
}

static int get_profile_path_pref(char * path_pref, int buf_size) {
    FILE    *   paths_file ;
    char * line = NULL;
    size_t len = MT_LINESIZE, parse_ix = 0;
    
    MT_bool found = FALSE;

    if ((paths_file=fopen(MT_PATHS_FILE, "r"))==NULL) {
	return -1 ;
    }
    if ((line = MALLOC(len)) == NULL) {
	fclose(paths_file);
	return -1;
    }
    while (fgets(line, len, paths_file) != NULL) {
	 if (!strncmp(line, MT_SRC_ENV_VAR, strlen(MT_SRC_ENV_VAR))) {
	     found = TRUE;
	     break;
	 }
	 //printf("Retrieved line of length %zu :\n", read);
	 //printf("%s", line);
    }
    if (found) {
	/* get rid of newline */
	if (line[strlen(line)-1] == '\n') {line[strlen(line)-1] = 0; }
	/* find '=' sign */
	for (parse_ix = strlen(MT_SRC_ENV_VAR); parse_ix < strlen(line); parse_ix++) {
	    if (line[parse_ix] == '=') { break;}
	}
	if (parse_ix == strlen(line)) { goto err;}
	/* skip any spaces following equals sign */
	for (++parse_ix; parse_ix < strlen(line); parse_ix++) {
	    if (line[parse_ix] != ' ') { break;}
	}
	if (parse_ix == strlen(line)) { goto err;}

	/* copy over the profile file path prefix */
	memset(path_pref, 0, buf_size);
	strcpy(path_pref, &line[parse_ix]);
	free(line);
	return 0;
    }  
err:
    free(line);
    return -1;
}

int main(int argc, char* argv[])
{
  VAPI_hca_id_t cur_hca, *hca_id_buf_p ;
  VAPI_ret_t vret;
  u_int32_t     num_hcas, returned_num_hcas;
  int i;
  int main_ret = 0;
  int ch;
  MT_bool use_profile = FALSE;
  MT_bool have_profile_path = FALSE;
  MT_bool verbose = FALSE;
  char profile_path[256], home[200];
  dictionary *ini = NULL;
  EVAPI_hca_profile_t profile, sugg_profile;



   while ((ch = MTGETOPT_getopt(argc, argv,"vf:p"))!=EOF)
    {
        switch (ch)
        {
        case 'p':
            use_profile = TRUE;
            break;
        case 'v':
            verbose = TRUE;
            break;
        case 'f':
            use_profile = TRUE;
            have_profile_path = TRUE;
            strcpy(profile_path, optarg);
            MTL_DEBUG4("Got profile path =%s\n",profile_path);
            break;
        default:
//            usage();
            ;
        }
    }

    argc -= optind;
    argv += optind;
  if (use_profile) {
      /* get the profile file */
      if (have_profile_path == FALSE) {
	  
          if (get_profile_path_pref(home,200) != 0) {
            strcpy(home,"/usr/mellanox/src");
          }
          strcpy((char *) profile_path, home);
          strcat((char *) profile_path, "/vapi/user/utils/hca_admin/hca_profile.ini");
      }
      ini = iniparser_load(profile_path);
      if (ini==NULL) {
          fprintf(stderr, "cannot parse profile file [%s]\n", profile_path);
          use_profile = FALSE;
      } else {
          if (iniparser_getnsec(ini) == 0) {
             use_profile = FALSE;
             iniparser_freedict(ini);
             ini = NULL;
          } else {
              fprintf(stderr, "Using profile file [%s]\n", profile_path);
          }
      }
  }

  if (argc > 0) { /* Open HCAs given in command line */
    for (i= 0; i< argc; i++) {
      MT_bool have_hca_profile = FALSE;
      strcpy(cur_hca,argv[i]);
      if (use_profile) {
        if (get_sec(ini, argv[i], &profile) == 0 ) {
            have_hca_profile = TRUE;
        } else if (get_default_sec(ini, &profile) == 0){
            have_hca_profile = TRUE;
        } else {
            printf("No profile for %s\n", argv[i]);
        }
      }
      if (verbose && have_hca_profile) {
          printf("Input profile:\n");
          print_profile(&profile);
      }
      vret= EVAPI_open_hca(cur_hca,
                           (have_hca_profile) ? &profile : NULL,
                           (have_hca_profile) ? &sugg_profile : NULL); 
      if (verbose && have_hca_profile) {
          printf("Returned profile:\n");
          print_profile(&sugg_profile);
      }
      if (vret == VAPI_EBUSY) {
        printf("HCA [%s] is already open.\n",cur_hca);
      } else if (vret == VAPI_OK) {
        printf("Opened HCA [%s] .\n",cur_hca);
      } else {
        printf("Failed opening HCA [%s] - %s .\n",cur_hca,VAPI_strerror_sym(vret));
        main_ret++ ;
      }
    }
  
  } else { /* Try opening all known devices */

    /* first, find how many HCAs available */
    vret = EVAPI_list_hcas(0,&num_hcas,NULL);
    if (vret != VAPI_EAGAIN && vret != VAPI_OK) {
        printf("Error: could not get number of available hcas (err = %d - %s)\n", 
                vret, VAPI_strerror_sym(vret));
        main_ret++ ;
		goto done;
    } 
    if (num_hcas == 0) {
        printf("No HCAs are available for opening.\n");
        main_ret++ ;
        goto done;
    }

    /* Allocate an HCA ID array with enough entries for all available HCAs*/
    hca_id_buf_p = (VAPI_hca_id_t *) VMALLOC(num_hcas * sizeof(VAPI_hca_id_t));
    if (hca_id_buf_p == NULL) {
        MT_COV_MEM_FLTR_START
        printf("could not allocate memory for the HCA IDs array\n");
		main_ret++;
        goto done;
        MT_COV_MEM_FLTR_END
    }
    vret = EVAPI_list_hcas(num_hcas, &returned_num_hcas, hca_id_buf_p);
    if (vret != VAPI_OK) {
        printf("could not obtain HCA IDs (error = %d - %s)\n", 
                vret, VAPI_strerror_sym(vret));
		main_ret++;
        goto done;
    }
    for (i= 0; i < (int)returned_num_hcas; i++) {
      MT_bool have_hca_profile = FALSE;
      printf("hca_id[%d] = %s - ", i, hca_id_buf_p[i]);
      strcpy(cur_hca,hca_id_buf_p[i]);
      if (use_profile) {
        if (get_sec(ini, cur_hca, &profile) == 0 ) {
            have_hca_profile = TRUE;
        } else if (get_default_sec(ini, &profile) == 0){
            have_hca_profile = TRUE;
        }
      }
      if (verbose && have_hca_profile) {
          printf("Input profile:\n");
          print_profile(&profile);
      }
      vret= EVAPI_open_hca(cur_hca,
                           (have_hca_profile) ? &profile : NULL,
                           (have_hca_profile) ? &sugg_profile : NULL); 
      if (verbose && have_hca_profile) {
          printf("Returned profile:\n");
          print_profile(&sugg_profile);
      }
      if (vret == VAPI_EBUSY) {
        printf("HCA is already open\n");
      } else if (vret == VAPI_OK) {
        printf("Opened\n");
      } else {
        printf("FAILED to open (%s) \n", VAPI_strerror_sym(vret));
        main_ret++ ;
      }
    } /* for */
    if (i == 0) {
      printf("No HCAs were found\n");
      main_ret++ ;
    }
  }
done:
  if (ini != NULL) {
      iniparser_freedict(ini);
  }
  return main_ret;
}											
