/*
  This software is available to you under a choice of one of two
  licenses.  You may choose to be licensed under the terms of the GNU
  General Public License (GPL) Version 2, available at
  <http://www.fsf.org/copyleft/gpl.html>, or the OpenIB.org BSD
  license, available in the LICENSE.TXT file accompanying this
  software.  These details are also available at
  <http://openib.org/license.html>.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
  BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
  ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
  CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
  SOFTWARE.

  Copyright (c) 2004 Mellanox Technologies Ltd.  All rights reserved.
*/

#ifndef H_ALL_BW_LAT_H
#define H_ALL_BW_LAT_H

#include <vapi.h>
#include <evapi.h>
#include <mosal.h>

#define INVAL_HNDL     0xFFFFFFFF
#define PAGE_SIZE_ALIGN      4096
#define LOG_PAGE_SIZE_ALIGN   12

/* literals */
#define MR_SND_CNT 		8
#define MR_RCV_CNT 		8
#define MR_UNIT				(1024 * 1024)
#define MAX_MR_UNITS_ALLOC   2048L  // That is, no limitation on size - user should suplly enough physycal memory.
#define DFLT_QP_OUS_RD_ATOM		8
#define MAX_PORTS_NUM		2
#define MIN_WAIT_FOR_SYNC_USECS	200000	/* 0.2 SEC */
#define MAX_TRIES_NUM							40
#define MAX_WAIT_FOR_SYNC_USECS	(MAX_TRIES_NUM * MIN_WAIT_FOR_SYNC_USECS)	/* 4 SEC */

/* general parameters values */
#define GEN_PARM_BW_OLD				0	/* use tests from bw_2ports.c to work with 1 port (instead of bw_loops.c) */

/* shortcuts */
#define CUR_PORT			params_p->in_params.port_num
#define CUR_PORT_IX		(params_p->in_params.port_num-1)

#define FOR_ALL_PORTS(_params_p,_ix)	\
{ 	int _ix_limit;	\
	if ((_params_p)->in_params.ports_num > 1) {	\
		_ix = 0;	\
		_ix_limit = (_params_p)->in_params.ports_num;	\
	}	\
	else {	\
		_ix = (_params_p)->in_params.port_num - 1;	\
		_ix_limit = (_params_p)->in_params.port_num;	\
	} \
	for (; _ix < _ix_limit; _ix++) 	{ \

#define END_FOR		 } }

#define CQ_BLOCK_TIMEOUT      0

#ifndef INLINE_DATA
#define EVAPI_post_inline_sr VAPI_post_sr
#endif

#define ARBEL_QPC_SETS        64
#define ARBEL_MPT_SETS        128

#define ASSIGN_MGID_0 \
    gid[0] = 0xff; \
    gid[1] = 0x12; \
    gid[2] = 0xab; \
    gid[3] = 0xab; \
    gid[4] = 0xab; \
    gid[5] = 0xab; \
    gid[6] = 0xab; \
    gid[7] = 0xab; \
    gid[8] = 0xab; \
    gid[9] = 0xab; \
    gid[10] = 0xab; \
    gid[11] = 0xab; \
    gid[12] = 0xab; \
    gid[13] = 0xab; \
    gid[14] = 0xab; \
    gid[15] = 0x00; \

#define ASSIGN_MGID_1 \
    gid[0] = 0xff; \
    gid[1] = 0x12; \
    gid[2] = 0xab; \
    gid[3] = 0xab; \
    gid[4] = 0xab; \
    gid[5] = 0xab; \
    gid[6] = 0xab; \
    gid[7] = 0xab; \
    gid[8] = 0xab; \
    gid[9] = 0xab; \
    gid[10] = 0xab; \
    gid[11] = 0xab; \
    gid[12] = 0xab; \
    gid[13] = 0xab; \
    gid[14] = 0xab; \
    gid[15] = 0xff;

typedef enum {
    RET_OK  = 0,
    RET_ERR = -1
} ret_val_t;

typedef enum {
    LAT_SEND,
    LAT_SEND_NEW,
    LAT_RDMA_W,
    LAT_RDMA_W_POLLCQ,
    LAT_RDMA_R,
    LAT_ATOM_CS,
    BW_ATOM_CS,
    BW_LAT_ATOM_CS,
    LAT_ATOM_FA,
    BW_ATOM_FA,
    BW_SEND,
    BW_RDMA_W,
    BW_RDMA_R,
    BW_MQP,
    BW_BIDIR,
    CHIPSET1,
    SPECIAL,
	BW_SEND_UNI,
	BW_RDMA_W_UNI,
	BW_RDMA_R_UNI,
	BW_SEND_BI,
	BW_RDMA_W_BI,
	BW_RDMA_R_BI,
    BW_SEND_LIST,
    WQE_TEST,
    SPECIAL_2PORT
} test_mode_t;

typedef enum {
    RDMA_NONE = 0,
    RDMA_W,
    RDMA_W_POLLCQ,
    RDMA_R
} rdma_mode_t;


typedef struct {
    VAPI_hca_id_t   hca_id;
	int				num_ah;
    int             socket;
    int             server_socket;
    int             socket_port;
    int             use_events;
    int             is_sender;
    int             is_server;
    rdma_mode_t     rdma_mode;
    VAPI_ts_type_t  transport;
    test_mode_t     test_mode;
    int             size;
    int             iter;      
    int             skip;
    IB_mtu_t        mtu;
    int             port_num;
    int             mr_send_size;
    int             mr_recv_size;
    int             rec_buf_num;
    int             alloc_aligned;
    int             use_send_events;
    int             use_recv_events;
    u_int32_t       num_cqe;
    u_int32_t       num_r_wqe;
    u_int32_t       num_s_wqe;
    u_int32_t       num_rsge;
    u_int32_t       num_ssge;
    MT_bool         use_mwindow;
    int             general_param;
    int             oneside;
    int             use_devmem;
	int				ports_num;
	int				bidirect;
    int             num_of_qps;
//    int             sched_period;
    int             postlist_len;
    int             qpc_sets;    
    int             mpt_sets;   
    int             num_of_keys;
    int             sg_list_len;
//    int             many2one;
    int             multicast;
    int             grh;
    int             srq_num;
    int             comp_inter;
} test_in_params_t;

typedef struct {
  VAPI_qp_hndl_t qp_hndl;
  VAPI_qp_num_t  qp_num;
  IB_lid_t       d_lid;
  VAPI_qp_num_t  d_qpnum;
  VAPI_srq_hndl_t srq_hndl;
  int            outs_wqes;
} qp_entry_t;

typedef struct {
  VAPI_srq_hndl_t srq_hndl;
} srq_entry_t;


typedef struct test_params_tx_s test_params_tx;

typedef ret_val_t (*test_2ports_p)(test_params_tx *params_p, int port_ix);

typedef struct {
   u_int64_t       				t_start;
   u_int64_t       				t_end;
} partner_t;

typedef struct {
    int                         port_num;
    IB_lid_t               		lid;
    IB_lid_t               		d_lid;
	VAPI_cq_hndl_t              s_cq_hndl;
	VAPI_cq_hndl_t              r_cq_hndl;
    VAPI_ud_av_hndl_t		    av_hndl;        //Address vector handle
    MOSAL_semaphore_t           rq_sem;
    MOSAL_semaphore_t           sq_sem; 
    EVAPI_compl_handler_hndl_t  rq_cq_eventh;
    EVAPI_compl_handler_hndl_t  sq_cq_eventh;
    //u_int32_t       			num_of_qps;
    //u_int32_t       			sched_period;
    qp_entry_t                  * qp_vect;
	MOSAL_thread_t			    mto;
	MOSAL_thread_t		        * mto_p;
    partner_t					snd;
    partner_t				    rcv;
    int							size;
    srq_entry_t                 * srq_vect;
} port_params_t;

struct test_params_tx_s {
    test_in_params_t       in_params;
    VAPI_hca_hndl_t        hca_hndl;
    VAPI_pd_hndl_t         pd_hndl;
    VAPI_mr_hndl_t         mr_hndl_s;
    VAPI_mr_hndl_t         mr_hndl_r;
    VAPI_mr_hndl_t *       mr_hndl_s_arr;
    VAPI_mr_hndl_t *       mr_hndl_r_arr;
    VAPI_qp_num_t          qpnum;
    VAPI_qp_num_t          d_qpnum;
    int                    max_inline_size;
    int                    max_cap_inline_size;
    
    VAPI_lkey_t            send_lkey;
    VAPI_lkey_t *          send_lkey_arr;
    u_int8_t               *send_buf;
    u_int8_t               *recv_buf;
    u_int8_t               *real_send_buf;
    u_int8_t               *real_recv_buf;
    VAPI_lkey_t            recv_lkey;
    VAPI_rkey_t            recv_rkey;
    VAPI_lkey_t *          recv_lkey_arr;
    VAPI_rkey_t *          recv_rkey_arr;

    VAPI_rkey_t            remote_rkey;
    VAPI_rkey_t *          remote_rkey_arr;
    VAPI_virt_addr_t       remote_add;

    EVAPI_async_handler_hndl_t async_hndl;

    VAPI_devmem_hndl_t     dm_hndl_s;
    VAPI_devmem_hndl_t     dm_hndl_r;
	VAPI_hca_vendor_t	   hca_vendor;
	VAPI_hca_cap_t		   hca_cap;
	u_int8_t        	   qp_ous_rd_atom;  /* Maximum number of outst. RDMA read/atomic as target */             
	port_params_t		   port[MAX_PORTS_NUM];
	test_2ports_p		   cur_test_p;
	MOSAL_semaphore_t	   sem;
	MOSAL_syncobj_t		   sync;
    u_int32_t              num_of_keys; 
    u_int32_t              mpt_sets; 
};

//typedef struct test_params_tx_s test_params_tx;

typedef struct {
    u_int32_t count;
    u_int32_t ask_comp;
} mqp_t;

typedef struct {
    u_int64_t avg;
    u_int64_t avg_below;
    u_int64_t thresh;
    int num_above;
    u_int64_t min;
    int min_i;
    u_int64_t max;
    int max_i;
} vect_res;

struct qp_list_entry_s;

typedef struct qp_list_entry_s
{
    VAPI_qp_hndl_t qp_hndl;
    struct qp_list_entry_s * next;
} qp_list_entry_t;


struct mr_list_entry_s;

typedef struct mr_list_entry_s
{
    VAPI_mr_hndl_t mr_hndl;
    struct mr_list_entry_s * next;
} mr_list_entry_t;


/* Fllow loops */
ret_val_t rdma_bw_test(test_params_tx *params_p);
ret_val_t atomic_bw_test(test_params_tx *params_p);
ret_val_t atomic_bw_lat_test(test_params_tx *params_p);
ret_val_t send_bw_test(test_params_tx *params_p);
ret_val_t send_bw_list_test(test_params_tx *params_p);
ret_val_t send_lat_test(test_params_tx *params_p);
ret_val_t atomic_swap_lat_test(test_params_tx *params_p);
ret_val_t send_lat_test_new(test_params_tx *params_p);
ret_val_t rdma_lat_test(test_params_tx *params_p);
ret_val_t rdma_lat_test_pollcq(test_params_tx *params_p);
ret_val_t rdma_read_lat_test(test_params_tx *params_p); //Noam
ret_val_t chipset_test1(test_params_tx *params_p);

ret_val_t send_bw_uni_test(test_params_tx *params_p, int port_ix);
ret_val_t send_bw_uni_test_2ports_old(test_params_tx *params_p);  // SHIMON TRY - Do not USE
ret_val_t send_bw_uni_test_2ports(test_params_tx *params_p);  // SHIMON TRY - Do not USE
ret_val_t rdma_w_bw_uni_test(test_params_tx *params_p, int port_ix);
ret_val_t rdma_w_bw_uni_many2one_test(test_params_tx *params_p, int port_ix);
ret_val_t rdma_r_bw_uni_test(test_params_tx *params_p, int port_ix);
ret_val_t send_bw_bi_test(test_params_tx *params_p, int port_ix);
ret_val_t rdma_w_bw_bi_test(test_params_tx *params_p, int port_ix);
ret_val_t rdma_r_bw_bi_test(test_params_tx *params_p, int port_ix);
ret_val_t wqe_test(test_params_tx *params_p, int port_ix);

ret_val_t send_lat_test_delta_sw(test_params_tx *params_p);
ret_val_t test_case_special(test_params_tx *params_p);
ret_val_t test_case_special_xx(test_params_tx *params_p);



ret_val_t create_common_resources(test_params_tx *params);
ret_val_t qp_create(test_params_tx *params, int port_ix);
ret_val_t qp_init2rts_mqp(test_params_tx *params, int port_ix);
ret_val_t lat_test_loop1_inline(test_params_tx *params_p);
ret_val_t rdma_lat_test_loop1(test_params_tx *params_p);
ret_val_t bw_test_loop2(test_params_tx *params_p);
ret_val_t bw_test_loop_sg(test_params_tx *params_p);
ret_val_t bw_test_loop_mqp(test_params_tx *params_p);
ret_val_t bw_bi_direction(test_params_tx *params_p);
ret_val_t bw_desc_list(test_params_tx *params_p);
ret_val_t bw_test_general(test_params_tx *params_p);
ret_val_t lat_post_test(test_params_tx *params_p);
ret_val_t lat_special(test_params_tx *params_p);
ret_val_t test_post_no_db(test_params_tx *params_p);


ret_val_t rdma_bw_test_loop2(test_params_tx *params_p);
ret_val_t create_av(test_params_tx *params, int port_ix);   
ret_val_t PERF_poll_cq(VAPI_hca_hndl_t  hca_hndl, VAPI_cq_hndl_t    cq_hndl,
                       VAPI_wc_desc_t *sc_p, int use_events, MOSAL_semaphore_t *sem_cq);
VAPI_ret_t PERF_poll_cq_nowait(VAPI_hca_hndl_t  hca_hndl, VAPI_cq_hndl_t    cq_hndl,
                       VAPI_wc_desc_t *sc_p, int use_events, MOSAL_semaphore_t *sem_cq_p);

qp_entry_t * search_qp(VAPI_qp_num_t qp_num, qp_entry_t *qp_vect, u_int32_t num_of_qps);

VAPI_ret_t get_list_of_hcas(VAPI_hca_id_t **hca_id_buf_p, u_int32_t *num_of_hcas);
//ret_val_t lat_test_loop1(test_params_tx *params_p);
//ret_val_t rdma_bw_test_loop(test_params_tx *params_p);
//ret_val_t bw_test_loop3(test_params_tx *params_p);
//ret_val_t bw_test_loop_all(test_params_tx *params_p);
#endif
