/*
  This software is available to you under a choice of one of two
  licenses.  You may choose to be licensed under the terms of the GNU
  General Public License (GPL) Version 2, available at
  <http://www.fsf.org/copyleft/gpl.html>, or the OpenIB.org BSD
  license, available in the LICENSE.TXT file accompanying this
  software.  These details are also available at
  <http://openib.org/license.html>.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
  BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
  ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
  CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
  SOFTWARE.

  Copyright (c) 2004 Mellanox Technologies Ltd.  All rights reserved.
*/


#include <stdio.h>
#include <math.h>

#define FALSE 0
#define TRUE  1
#define GAUSS_INTERVAL 1.0           // Min Interval between two Gaussians (in uSec)
#define MIN_INTERVAL   0.5           // Max Interval between two min points (in uSec)
#define ONE_PERCENT    0.01
#define MAX_VEC_SIZE   10
#define ONE_DIST       1
#define TWO_DIST       2
#define TEN_PERCENT    0.1

typedef struct distribution{
    float value;
    int distrib;

} distribution_array;


int find_left_min(int max_ind, int dist_size, int dist_array_size, distribution_array *dist_array)
{
int i, prev_dist=0, min_ind=0, one_down=FALSE;
 prev_dist = dist_array[max_ind].distrib;
 for (i = max_ind - 1 ; i >= 0 ; i--){
     if ((dist_array[i].distrib < prev_dist) && (dist_array[i].distrib <= TWO_DIST)){
         min_ind = i;
         one_down = TRUE;
     }
     else
         if (one_down) {
             break;
         }
     prev_dist = dist_array[i].distrib;     
 }
 return min_ind;
}


int find_right_min(int max_ind, int dist_size, int dist_array_size, distribution_array *dist_array)
{
int i, prev_dist=0, min_ind=0, one_down=FALSE;
 prev_dist = dist_array[max_ind].distrib;
 for (i = max_ind + 1 ; i < dist_array_size ; i++){
     if ((dist_array[i].distrib < prev_dist) && (dist_array[i].distrib <= TWO_DIST)){
         min_ind = i;
         one_down = TRUE;
     }
     else
         if (one_down) {
             break;
         }
     prev_dist = dist_array[i].distrib;     
 }
 return min_ind;
}



int find_max_by_index(int prev_max_ind, int dist_size, int dist_array_size, distribution_array *dist_array)
{
int i, prev_dist=0, max_ind=0, one_up=FALSE;
 
 prev_dist = dist_array[prev_max_ind].distrib;
 for (i = prev_max_ind + 1 ; i < dist_array_size ; i++){
     if ((dist_array[i].distrib > prev_dist) && (dist_array[i].distrib > dist_size*ONE_PERCENT)){
         max_ind = i;
         one_up = TRUE;
     }
     else
         if (one_up) {
             break;
         }
     prev_dist = dist_array[i].distrib;     
 }
 return max_ind;
}



int main(int argc , char **argv)
{

float *dist_val, max_distance=0, min_distance=0, mean, dev, sd, sum, cons;
int dist_size, i=0, j=0, k=0, count_dist=1, dist_array_size;
int ind_max_y=0, current_max_ind, max_vec_ind[MAX_VEC_SIZE];
int cur_ind, min_vec_ind[2*MAX_VEC_SIZE], min_ind, current_min_ind;
int dist_num;


distribution_array *dist_array;

FILE *distrib_fd, *dist_size_fd, *analyses_fd;

    /*
     *  Setting up the file handles
     */
    system("wc -l /tmp/distribution.txt > /tmp/dist_size.txt");
    system("sort -n /tmp/distribution.txt > /tmp/sort_distrib.txt");

    if (!(distrib_fd = fopen("/tmp/sort_distrib.txt","r"))) {
            printf("Error: Can't open for reading /tmp/sort_distrib.txt\n");
            exit(1);
    }        

    if (!(dist_size_fd = fopen("/tmp/dist_size.txt","r"))) {
            printf("Error: Can't open for reading /tmp/dist_size.txt\n");
            exit(1);
    }        
    

    /*
     *  Get the distribution size and allocate arrays accordingly
     */
    if (!(fscanf(dist_size_fd, "%d", &dist_size))) {
            printf("Error: Can't read /tmp/dist_size.txt\n");
            exit(1);
    }        
    

    if (!(dist_val = (float *)calloc(dist_size, sizeof(float)))){
        printf("Error: Can't allocate memory for values array\n");
    }

    if (!(dist_array = (distribution_array *)calloc(dist_size * TEN_PERCENT, sizeof(distribution_array)))){
        printf("Error: Can't allocate memory for values array\n");
    }



    /*
     *  Read and arrange distribution file into array of value,count. 
     */
    for (i=0 ; i < dist_size ; i++){
            if (!(fscanf(distrib_fd, "%f", &dist_val[i]))) {
            printf("Error: Can't read element %d in /tmp/distribution.txt\n", i);
            exit(1);
            }
            
            if ((i == 0) || (i == dist_size-1)){
                dist_array[j].value   = dist_val[i];
                dist_array[j].distrib = count_dist;
                j++;
            }
            else if (dist_val[i] == dist_val[i-1]){
                count_dist++;
            }
            else{
                dist_array[j].value   = dist_val[i-1];
                dist_array[j].distrib = count_dist;
                count_dist = 1;
                j++;
            }                
    }
    dist_array_size = j;

    printf("\n\n *****************************************************************************\n");
    printf(" **************     Performance analyses utility Rev 1.0      ****************\n");
    printf(" *****************************************************************************\n");
    cons = GAUSS_INTERVAL;
    printf(" Analyses constants:\n Min interval between two gaussians - %.1fuS\n", cons);
    cons = MIN_INTERVAL;
    printf(" Max interval between two min points at the shoulders of the gaussian - %.1fuS\n\n", cons);
    

    for (i=0 ; i < j ; i++){
        printf("Element %d - %.1f %d\n", i, dist_array[i].value, dist_array[i].distrib);
    }    

    printf("%d elements read from /tmp/distribution.txt\n", dist_size);

    /*
     *  Finding the Maximum points and placing them in a vector
     *  1. Finding local maximums by a simple algorithm of recording a max whenever the function
     *     change the derivative from positive to negative.
     *  2. When the local max is retrieved check against the previous max, if the retrieved max
     *     is greater then the previous stored max and within a GAUSS_INTERVAL then replace the 
     *     stored max with the new max. if not within GAUSS_INTERVAL store is a new local max.
     */ 
    
    current_max_ind = 0;
    cur_ind = 0;
    max_vec_ind[cur_ind] = find_max_by_index(current_max_ind, dist_size, dist_array_size, dist_array);
    current_max_ind = max_vec_ind[cur_ind];
 
    for (i=0; i < MAX_VEC_SIZE ; i++){
        if (!(ind_max_y = find_max_by_index(current_max_ind, dist_size, dist_array_size, dist_array))){
            break;
        }

        /*
         *  Compare the local maximum to the local maximums on max_vec, once one was found
         *  add or replace element on max_vec.
         */
        
        max_distance = fabs(dist_array[max_vec_ind[cur_ind]].value - dist_array[ind_max_y].value);
        if(max_distance > GAUSS_INTERVAL) {
            cur_ind++;
            max_vec_ind[cur_ind] = ind_max_y;
        }
        else 
            if (dist_array[max_vec_ind[cur_ind]].distrib < dist_array[ind_max_y].distrib) {
                max_vec_ind[cur_ind] = ind_max_y;
            }
        current_max_ind = ind_max_y;
    }
    cur_ind++;


    /*
     *  Finding min points to the sides of each max point:
     *  1. Search min point to the left(search the array in a descending order) of the max point, choose
     *     the minimum of all the local minimums within a GAUSS_INTERVAL.
     *  2. Do the same procedure to the right of the max point.
     */
    j = 0;
    for (i=0; i < cur_ind ; i++){
        min_vec_ind[j] = find_left_min(max_vec_ind[i], dist_size, dist_array_size, dist_array);
        current_min_ind = min_vec_ind[j];
        while(min_ind){
            min_ind = find_left_min(current_min_ind, dist_size, dist_array_size, dist_array);
            min_distance = fabs(dist_array[min_ind].value - dist_array[min_vec_ind[j]].value);
            if (min_distance > MIN_INTERVAL) break;
            else
                if (dist_array[min_vec_ind[j]].distrib > dist_array[min_ind].distrib) {
                    min_vec_ind[j] = min_ind;
                }
            current_min_ind = min_ind; 
        }            
        j++;

        min_vec_ind[j] = find_right_min(max_vec_ind[i], dist_size, dist_array_size, dist_array);
        current_min_ind = min_vec_ind[j];
        while(TRUE){
            min_ind = find_right_min(current_min_ind, dist_size, dist_array_size, dist_array);
            min_distance = fabs(dist_array[min_ind].value - dist_array[min_vec_ind[j]].value);
            if (min_distance > MIN_INTERVAL) break;
            else
                if (dist_array[min_vec_ind[j]].distrib > dist_array[min_ind].distrib) {
                    min_vec_ind[j] = min_ind;
                }
            current_min_ind = min_ind; 
        }
        j++;
    }

    
    for (i=0; i < cur_ind ; i++){
        sum = 0;
        dist_num = 0;
        for (k=min_vec_ind[2*i] ; k <= min_vec_ind[2*i+1] ; k++){
            sum += dist_array[k].value * dist_array[k].distrib;
            dist_num += dist_array[k].distrib;
        }
        mean = sum / dist_num;
        dev = fabs(dist_array[max_vec_ind[i]].value - mean);
        sum = 0;
        for (k=min_vec_ind[2*i] ; k <= min_vec_ind[2*i+1] ; k++){
            sum += (((dist_array[k].value - mean) * dist_array[k].distrib) * 
                    ((dist_array[k].value - mean) * dist_array[k].distrib));
        }                   
        sd = sqrt(sum / dist_num);
        printf("Gaussian %d: Max = %.1fuS, weight = %d dev =  %.2f, sd = %.2f\n", 
               i, 
               dist_array[max_vec_ind[i]].value, 
               dist_array[max_vec_ind[i]].distrib,
               dev,
               sd);
    }             

}







































































