/*
  This software is available to you under a choice of one of two
  licenses.  You may choose to be licensed under the terms of the GNU
  General Public License (GPL) Version 2, available at
  <http://www.fsf.org/copyleft/gpl.html>, or the OpenIB.org BSD
  license, available in the LICENSE.TXT file accompanying this
  software.  These details are also available at
  <http://openib.org/license.html>.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
  BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
  ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
  CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
  SOFTWARE.

  Copyright (c) 2004 Mellanox Technologies Ltd.  All rights reserved.
*/

#include <vapi.h>
#include <evapi.h>
#include <vapi_common.h>
#include <stdio.h>

#define TIME_LEVEL 1
#include "utils.h" 
#include "all_bw_lat.h"
#include "sock.h" 

int tmp_arr[]={50,60,70,80,90,100,2000};

/***********************************************************************************************
* name:      send_bw_test
* args:
*            params_p -- test parameter structure, need to filled before calling
* returns:
*            RET_OK
*            RET_ERR
*
* description:     
*            This is the main loop for the UD and RC send BW test. The test sends N packets
*            then waits for acknowledge from the receiver, and calculate the BW.
*            
************************************************************************************************/  
ret_val_t send_bw_test(test_params_tx *params_p)
{
    VAPI_ret_t             res;
    u_int64_t t_start=0, t_end=0;

    int i,comp_inter,ask_for_comp=0, tmp_cnt=1;

    VAPI_rr_desc_t         rr;
    VAPI_sg_lst_entry_t    sg_entry_r;
    VAPI_sr_desc_t         sr;
    VAPI_sg_lst_entry_t    sg_entry_s;
    VAPI_wc_desc_t         rc;
    VAPI_wc_desc_t         sc;

    u_int32_t              srq_posted_actual;
    /* Copy all input parameters - don't want to rewrite the loop code */
    VAPI_hca_hndl_t        hca_hndl    = params_p->hca_hndl;
    VAPI_qp_hndl_t         qp_hndl     = params_p->port[CUR_PORT_IX].qp_vect[0].qp_hndl;
    VAPI_cq_hndl_t         s_cq_hndl   = params_p->port[CUR_PORT_IX].s_cq_hndl;
    VAPI_cq_hndl_t         r_cq_hndl   = params_p->port[CUR_PORT_IX].r_cq_hndl;
    VAPI_srq_hndl_t        srq_hndl    = params_p->port[CUR_PORT_IX].qp_vect[0].srq_hndl;
    MOSAL_semaphore_t    * sq_sem_p   = &params_p->port[CUR_PORT_IX].sq_sem;
    MOSAL_semaphore_t    * rq_sem_p   = &params_p->port[CUR_PORT_IX].rq_sem;
    u_int8_t               *recv_buf   = params_p->recv_buf;
    u_int8_t               *send_buf   = params_p->send_buf;
    int                    is_sender   = params_p->in_params.is_sender;
    int                    size        = params_p->in_params.size;
    int                    iter        = params_p->in_params.iter;      
    //int                    skip        = params_p->in_params.iter;      
    int use_inline = FALSE;

    int max_msg_size = MIN2(params_p->in_params.mr_recv_size,params_p->in_params.mr_send_size);
    int max_sends = MIN2(params_p->in_params.num_r_wqe,params_p->in_params.num_s_wqe); 


    if (params_p->in_params.transport == IB_TS_UD) {
        max_msg_size = 2048;
        if (size > 2048) {
            printf("Warning!!!: UD max size is 2048\n");
            size = 2048;
        }
    }

    if (size > params_p->in_params.mr_recv_size) {
        printf("Warning!!!: size=%d > mr_recv_size=%d --- changing iter to %d\n", 
               size,params_p->in_params.mr_recv_size,params_p->in_params.mr_recv_size);
        size = params_p->in_params.mr_recv_size;
    }

    rr.opcode = VAPI_RECEIVE;
    rr.comp_type = VAPI_SIGNALED;
    rr.sg_lst_len = 1;
    rr.sg_lst_p  = &sg_entry_r;
    sg_entry_r.lkey = params_p->recv_lkey;
    sg_entry_r.len = params_p->in_params.mr_recv_size;
    sg_entry_r.addr = (VAPI_virt_addr_t)(MT_virt_addr_t)recv_buf;

    /***************************************************************************/
    sr.opcode = VAPI_SEND;
    sr.comp_type = VAPI_SIGNALED;
    sr.set_se = FALSE;
    sr.remote_qkey = 0;
    sr.id = 1;

    /* needed only for UD */
    sr.remote_qp = params_p->port[CUR_PORT_IX].qp_vect[0].d_qpnum;
    sr.remote_ah = params_p->port[CUR_PORT_IX].av_hndl;

    sr.sg_lst_len = 1;
    sg_entry_s.addr = (VAPI_virt_addr_t)(MT_virt_addr_t)send_buf;
    sg_entry_s.len = size;
    sg_entry_s.lkey = params_p->send_lkey;
    sr.sg_lst_p = &sg_entry_s;

    /* Post all possible recive in advance. Sender need 1 while receiver need N*/
    if (!is_sender) {
        for (i=0;i<max_sends;i++) {
            rr.id = i;
            if (params_p->in_params.srq_num > 0)
            {
                res = VAPI_post_srq(hca_hndl, srq_hndl, 1, &rr, &srq_posted_actual);
                //printf("posted %d SRQs\n", srq_posted_actual);
            }
            else
            {
                res = VAPI_post_rr(hca_hndl, qp_hndl, &rr);
            }
            if (res != VAPI_OK) {
                printf("Error: Posting Recv buffer : %s\n", VAPI_strerror(res));
                return(RET_ERR);
            }
        }
    }
    else {
        rr.id = 0;
        if (params_p->in_params.srq_num > 0)
        {
            res = VAPI_post_srq(hca_hndl, srq_hndl, 1, &rr, &srq_posted_actual);
            //printf("posted %d SRQs\n", srq_posted_actual);
        }
        else
        {
            res = VAPI_post_rr(hca_hndl, qp_hndl, &rr);
        }
        if (res != VAPI_OK) {
            printf("Error: Posting Recv buffer : %s\n", VAPI_strerror(res));
            return(RET_ERR);
        }
    }


    /* sync the test start */
    sock_sync_msg(params_p->in_params.socket, is_sender);
    COND_PRINTF1(("\n\n************* %s BW Test started  *********************\n\n",
                  params_p->in_params.transport == IB_TS_UD ? "UD" : "RC"));

    if (iter < max_sends) {
        comp_inter = max_sends; // Ask completion ONLY on last message!!
    }
    else {
        comp_inter = max_sends/2 - 2;
    }

    ask_for_comp = 0;

    if ((size >0 ) &&(size <= params_p->max_inline_size))
    {
        use_inline = TRUE;
    }

    if (is_sender) {
        get_clk64(t_start);
        for (i=0; i< iter; i++, tmp_cnt++) {
            sr.id = i;
            sr.comp_type = VAPI_UNSIGNALED; 


            /* Ask for completion every comp_inter and on last send*/
            //if (((i+1)%comp_inter == 0) || i == (iter-1)) {
            if ((tmp_cnt == comp_inter) || i == (iter-1)) {
                tmp_cnt = 0;
                sr.comp_type = VAPI_SIGNALED;
                ask_for_comp++;
                COND_PRINTF((" **** Posting sr signaled i=%d ask=%d\n",i+1,ask_for_comp));
            }

            /* Use Inline data up to the limit size */
            //if ((size >0 ) &&(size <= params_p->max_inline_size)) {
            if (use_inline == TRUE) {
                res = EVAPI_post_inline_sr(hca_hndl, qp_hndl, &sr);
            }
            else {
                res = VAPI_post_sr(hca_hndl, qp_hndl, &sr);
            }
            if (res != VAPI_OK) {
        		printf("sg_list_len %d\n", sr.sg_lst_len );
                printf("Error: Posting send desc : %s\n", VAPI_strerror(res));
                return(RET_ERR);
            }

            /* Wait for completion every comp_inter starting from 2*comp_inter
             * This applies only when comp_iter < iter */
            if (ask_for_comp > 1) {
                COND_PRINTF2(("**** Polling for sr comp i=%d ask=%d\n",i+1,ask_for_comp));
                if (RET_OK != PERF_poll_cq(hca_hndl,s_cq_hndl ,&sc,params_p->in_params.use_send_events, 
                                           sq_sem_p)) {
                    return RET_ERR;
                }
                ask_for_comp--;
                COND_PRINTF(("Polling for sr comp %d %d\n",i,ask_for_comp));
            }
        }//for i - send packets

        /* Wait for completion on last send */
        while (ask_for_comp>0) {
            COND_PRINTF2(("**** Polling for last sr comp i=%d ask=%d\n",i,ask_for_comp));
            /* Poll for the last send completion */
            if (RET_OK != PERF_poll_cq(hca_hndl,s_cq_hndl ,&sc,params_p->in_params.use_send_events,
                                       sq_sem_p)) {
                return RET_ERR;
            }
            ask_for_comp--;
            COND_PRINTF(("Polling for sr comp %d %d\n",i,ask_for_comp));
        }

        /* Wait for final reply from receiver */
        if (RET_OK != PERF_poll_cq(hca_hndl, r_cq_hndl ,&rc,
                                   params_p->in_params.use_recv_events,  rq_sem_p)) {
            return RET_ERR;
        }
        COND_PRINTF(("Sender came out of Poll msg_size=%u\n",rc.byte_len));
        get_clk64(t_end);

        printf("BW: %.1f MBytes/sec [size: %d bytes]\n", BW_MBSEC(t_end - t_start,1.0*(unsigned)size * iter),size);

    }// Is sender

    else { /* Receiver side */
        for (i=0; i< iter; i++) {
            if (RET_OK != PERF_poll_cq(hca_hndl, r_cq_hndl ,&rc,
                                       params_p->in_params.use_recv_events,  rq_sem_p)) {
                return RET_ERR;
            }
            if (params_p->in_params.srq_num > 0)
            {
                res = VAPI_post_srq(hca_hndl, srq_hndl, 1, &rr, &srq_posted_actual);
                //printf("posted %d SRQs\n", srq_posted_actual);
            }
            else
            {
                res = VAPI_post_rr(hca_hndl, qp_hndl, &rr);
            }
            if (res != VAPI_OK) {
                printf("Error: Posting Recv buffer : %s\n", VAPI_strerror(res));
                return(RET_ERR);
            }
        }

        /* Send reply to sender after last message. Use 0 length message!!! */
        sg_entry_s.addr = (VAPI_virt_addr_t)(MT_virt_addr_t)send_buf;
        sr.sg_lst_len = 0;
        res = VAPI_post_sr(hca_hndl, qp_hndl, &sr);
        if (res != VAPI_OK) {
        		printf("sg_list_len %d\n", sr.sg_lst_len );
            printf("Error: Posting send desc : %s\n", VAPI_strerror(res));
            return(RET_ERR);
        }

        /* Wait for final reply from receiver */
        if (RET_OK != PERF_poll_cq(hca_hndl,s_cq_hndl ,&sc,params_p->in_params.use_send_events, 
                                   sq_sem_p)) {
            return RET_ERR;
        }
    }
    COND_PRINTF1(("\n************* %s BW Test Finished  *********************\n\n",
                  params_p->in_params.transport == IB_TS_UD ? "UD" : "RC"));
    return(RET_OK);
}

/***********************************************************************************************
* name:      rdma_bw_test
* args:
*            params_p -- test parameter structure, need to filled before calling
* returns:
*            RET_OK
*            RET_ERR
*
* description:     
*            This is the main loop for the RDMA (RC) BW test. The test sends N packets
*            then waits for acknowledge from the receiver, and calculate the BW.
*            
************************************************************************************************/  
ret_val_t rdma_bw_test(test_params_tx *params_p)
{
    VAPI_ret_t             res;
    u_int64_t t_start=0, t_end=0;

    int i,comp_inter, ask_for_comp=0, tmp_cnt=1;

    VAPI_rr_desc_t         rr;
    VAPI_sg_lst_entry_t    sg_entry_r;
    VAPI_sr_desc_t         sr;
    VAPI_sg_lst_entry_t    sg_entry_s;

    VAPI_sr_desc_t         rc_sr;
    VAPI_sg_lst_entry_t    rc_sg_entry_s;

    VAPI_wc_desc_t         rc;
    VAPI_wc_desc_t         sc;

    u_int32_t              srq_posted_actual;
    /* Copy all input parameters - don't want to rewrite the loop code */
    VAPI_hca_hndl_t        hca_hndl    = params_p->hca_hndl;
    VAPI_qp_hndl_t         qp_hndl     = params_p->port[CUR_PORT_IX].qp_vect[0].qp_hndl;

    VAPI_cq_hndl_t         s_cq_hndl   = params_p->port[CUR_PORT_IX].s_cq_hndl;
    VAPI_cq_hndl_t         r_cq_hndl   = params_p->port[CUR_PORT_IX].r_cq_hndl;
    VAPI_srq_hndl_t        srq_hndl    = params_p->port[CUR_PORT_IX].qp_vect[0].srq_hndl;
    MOSAL_semaphore_t    * sq_sem_p   = &params_p->port[CUR_PORT_IX].sq_sem;
    MOSAL_semaphore_t    * rq_sem_p   = &params_p->port[CUR_PORT_IX].rq_sem;
    u_int8_t               *recv_buf   = params_p->recv_buf;
    u_int8_t               *send_buf   = params_p->send_buf;
    int                    is_sender   = params_p->in_params.is_sender;
    int                    size        = params_p->in_params.size;
    int                    iter        = params_p->in_params.iter;      
    volatile char          *check_p;
    int index_to_check;
    int use_inline = FALSE;

    int max_msg_size = MIN2(params_p->in_params.mr_recv_size,params_p->in_params.mr_send_size);
    int max_sends = MIN2(params_p->in_params.num_r_wqe,params_p->in_params.num_s_wqe); 

    if (params_p->in_params.transport == IB_TS_UD) {
        max_msg_size = 2048;
        if (size > 2048) {
            printf("Warning!!!: UD max size is 2048\n");
            size = 2048;
        }
    }

    if (size > params_p->in_params.mr_recv_size) {
        printf("Warning!!!: size=%d > mr_recv_size=%d --- changing iter to %d\n", 
               size,params_p->in_params.mr_recv_size,params_p->in_params.mr_recv_size);
        size = params_p->in_params.mr_recv_size;
    }

    rr.opcode = VAPI_RECEIVE;
    rr.comp_type = VAPI_SIGNALED;
    rr.sg_lst_len = 1;
    rr.sg_lst_p  = &sg_entry_r;
    sg_entry_r.lkey = params_p->recv_lkey;
    sg_entry_r.len = params_p->in_params.mr_recv_size;
    sg_entry_r.addr = (VAPI_virt_addr_t)(MT_virt_addr_t)recv_buf;


    /**************    RDMA SR *****************************************************/
    sr.comp_type = VAPI_SIGNALED;
    sr.set_se = FALSE;

    sr.opcode = VAPI_RDMA_WRITE;
    sr.remote_qkey = 0;
    sr.remote_addr = params_p->remote_add;
    sr.r_key = params_p->remote_rkey;
    sr.id = 1;


    sr.sg_lst_len = 1;
    sg_entry_s.len = size;
    sg_entry_s.lkey = params_p->send_lkey;
    sg_entry_s.addr = (VAPI_virt_addr_t)(MT_virt_addr_t)send_buf; 
    sr.sg_lst_p = &sg_entry_s;



    /**************    RC SR *****************************************************/
    rc_sr.comp_type = VAPI_SIGNALED;
    rc_sr.set_se = FALSE;
    rc_sr.opcode = VAPI_SEND;
    rc_sr.remote_qkey = 0;
    rc_sr.id = 1;
    rc_sr.sg_lst_len = 0;

    rc_sg_entry_s.len = size;
    rc_sg_entry_s.lkey = params_p->send_lkey;
    rc_sg_entry_s.addr = (VAPI_virt_addr_t)(MT_virt_addr_t)send_buf; 
    rc_sr.sg_lst_p = &sg_entry_s;

    /* Both Sender  and receiver need 1 for last message*/
    rr.id = 0;
    if (params_p->in_params.srq_num > 0)
    {
        res = VAPI_post_srq(hca_hndl, srq_hndl, 1, &rr, &srq_posted_actual);
        //printf("posted %d SRQs\n", srq_posted_actual);
    }
    else
    {
        res = VAPI_post_rr(hca_hndl, qp_hndl, &rr);
    }
    if (res != VAPI_OK) {
        printf("Error: Posting Recv buffer : %s\n", VAPI_strerror(res));
        return(RET_ERR);
    }

    /* sync the test start */
    sock_sync_msg(params_p->in_params.socket, is_sender);
    COND_PRINTF1(("\n************* RDMA-Write BW Test started  *********************\n\n"));


    if (iter < max_sends) {
        comp_inter = max_sends;
    }
    else {
        comp_inter = max_sends / 2 -2;
    }

    ask_for_comp = 0;

    index_to_check = 0;
    check_p = &recv_buf[index_to_check];

    if ((size >0 ) &&(size <= params_p->max_inline_size))
    {
        use_inline = TRUE;
    }

    if (is_sender) {
        get_clk64(t_start);
        for (i=0; i< iter; i++, tmp_cnt++) {
            sr.id = i;
            sr.comp_type = VAPI_UNSIGNALED; 

            /* Ask for completion every comp_inter */
            //if (((i+1)%comp_inter == 0) || i == (iter-1)) {
            if ((tmp_cnt == comp_inter) || i == (iter-1)) {
                tmp_cnt = 0;
                sr.comp_type = VAPI_SIGNALED;
                ask_for_comp++;
                COND_PRINTF((" **** Posting sr signaled i=%d ask=%d\n",i+1,ask_for_comp));
            }

            //if ((size >0 ) &&(size <= params_p->max_inline_size)) {
            if (use_inline == TRUE) {
                res = EVAPI_post_inline_sr(hca_hndl, qp_hndl, &sr);
            }
            else {
                res = VAPI_post_sr(hca_hndl, qp_hndl, &sr);
            }
            if (res != VAPI_OK) {
                printf("Error: Posting send desc : %s\n", VAPI_strerror(res));
                return(RET_ERR);
            }


            /* Wait for completion every comp_inter starting from 2*comp_inter
             * This applies only when comp_iter > iter */
            if (ask_for_comp > 1) {
                COND_PRINTF(("**** Polling for sr comp i=%d ask=%d\n",i+1,ask_for_comp));
                if (RET_OK != PERF_poll_cq(hca_hndl,s_cq_hndl ,&sc,params_p->in_params.use_send_events, 
                                           sq_sem_p)) {
                    return RET_ERR;
                }
                ask_for_comp--;
            }
        }//for i - send packets

        /* Send an RC send  0 message indicating end of RDMA */
        COND_PRINTF(("Finished RDMA posting, Sending RC-Send message\n"));
        rc_sr.comp_type = VAPI_UNSIGNALED;
        rc_sr.sg_lst_len = 1;
        rc_sg_entry_s.len = 1;
        res = VAPI_post_sr(hca_hndl, qp_hndl, &rc_sr);
        if (res != VAPI_OK) {
            printf("Error: Posting Send buffer : %s\n", VAPI_strerror(res));
            return(RET_ERR);
        }

        /* Wait for completion on last send */
        while (ask_for_comp>0) {
            COND_PRINTF(("**** Polling for last sr comp i=%d ask=%d\n",i,ask_for_comp));

            /* Poll for the last send completion */
            if (RET_OK != PERF_poll_cq(hca_hndl,s_cq_hndl ,&sc,params_p->in_params.use_send_events, 
                                       sq_sem_p)) {
                return RET_ERR;
            }
            ask_for_comp--;
            COND_PRINTF(("Polling for sr comp %d %d\n",i,ask_for_comp));
        }

        /* Wait for final reply from receiver */
        if (RET_OK != PERF_poll_cq(hca_hndl,r_cq_hndl ,&rc,params_p->in_params.use_recv_events, 
                                   rq_sem_p)) {
            return RET_ERR;
        }

        get_clk64(t_end);

        printf("BW: %.1f MBytes/sec [size: %d bytes]\n", BW_MBSEC(t_end - t_start,1.0*(unsigned)size * iter),size);

    }// Is sender
    else {
        /* Wait for the special Send from the sender */
        if (RET_OK != PERF_poll_cq(hca_hndl, r_cq_hndl ,&rc, 
                                   params_p->in_params.use_recv_events,  rq_sem_p)) {
            return RET_ERR;
        }

        /* Send reply to sender after last message. Use 0 length message!!! */
        rc_sr.comp_type = VAPI_SIGNALED;
        rc_sr.sg_lst_len = 0;
        res = VAPI_post_sr(hca_hndl, qp_hndl, &rc_sr);
        if (res != VAPI_OK) {
            printf("Error: Posting send desc : %s\n", VAPI_strerror(res));
            return(RET_ERR);
        }
        COND_PRINTF(("Reciever sent ack, waiting for completion\n"));

        /* Poll for the last send completion */
        if (RET_OK != PERF_poll_cq(hca_hndl, s_cq_hndl ,&sc, 
                                   params_p->in_params.use_send_events,  sq_sem_p)) {
            return RET_ERR;
        }
    }

    COND_PRINTF1(("\n************* RDMA-Write BW Test Finished  *********************\n\n"));
    return(RET_OK);
}


/***********************************************************************************************
* name:      atomic_bw_test
* args:
*            params_p -- test parameter structure, need to filled before calling
* returns:
*            RET_OK
*            RET_ERR
*
* description:     
*            This is the main loop for the RDMA (RC) BW test. The test sends N packets
*            then waits for acknowledge from the receiver, and calculate the BW.
*            
************************************************************************************************/  
ret_val_t atomic_bw_test(test_params_tx *params_p)
{
    VAPI_ret_t             res;
    u_int64_t t_start=0, t_end=0;

    int i,comp_inter, ask_for_comp=0, last, tmp_cnt=1;

    VAPI_sr_desc_t         sr;
    VAPI_sg_lst_entry_t    sg_entry_s;


    VAPI_wc_desc_t         rc;
    VAPI_wc_desc_t         sc;

    /* Copy all input parameters - don't want to rewrite the loop code */
    VAPI_hca_hndl_t        hca_hndl    = params_p->hca_hndl;
    VAPI_qp_hndl_t         qp_hndl     = params_p->port[CUR_PORT_IX].qp_vect[0].qp_hndl;

    VAPI_cq_hndl_t         s_cq_hndl   = params_p->port[CUR_PORT_IX].s_cq_hndl;
    MOSAL_semaphore_t    * sq_sem_p   = &params_p->port[CUR_PORT_IX].sq_sem;
    MOSAL_semaphore_t    * rq_sem_p   = &params_p->port[CUR_PORT_IX].rq_sem;
    volatile u_int8_t               *send_buf   = params_p->send_buf;
    volatile u_int64_t              *recv_buf   = (u_int64_t*)(params_p->recv_buf);
    int                    is_sender   = params_p->in_params.is_sender;
    int                    iter        = params_p->in_params.iter;      
    int max_sends = MIN2(params_p->in_params.num_r_wqe,params_p->in_params.num_s_wqe); 
    int j=0;
    
    /**************    RDMA SR *****************************************************/

    *recv_buf = 84;

    sr.comp_type = VAPI_SIGNALED;
    sr.set_se = FALSE;
    sr.fence = FALSE;

    sr.opcode = VAPI_ATOMIC_CMP_AND_SWP;
    sr.remote_qkey = 0;
    sr.remote_addr = params_p->remote_add;
    sr.r_key = params_p->remote_rkey;
    sr.id = 1;
    sr.compare_add = 84;
    sr.swap = 84;


    sr.sg_lst_len = 1;
    sg_entry_s.len = 8;
    sg_entry_s.lkey = params_p->send_lkey;
    sg_entry_s.addr = (VAPI_virt_addr_t)(MT_virt_addr_t)send_buf; 
    sr.sg_lst_p = &sg_entry_s;

    /* sync the test start */
    sock_sync_msg(params_p->in_params.socket, is_sender);
    COND_PRINTF1(("\n************* Atomic Compare & Swap BW Test started  *********************\n\n"));


    if (iter < max_sends) {
        comp_inter = max_sends;
    }
    else {
        comp_inter = max_sends / 2 -2;
    }

    ask_for_comp = 0;
    //Last iteration number
    last = iter - 1;

    if (is_sender) {
        get_clk64(t_start);
        for (i=0; i< iter; i++, tmp_cnt++) {
        //for (i=0; i< iter; i++) {
            qp_hndl = params_p->port[CUR_PORT_IX].qp_vect[j].qp_hndl;
            sr.id = i;
            sr.comp_type = VAPI_UNSIGNALED; 
            if (i == last) {
                sr.swap = 85; //Sign the last iteration for a remote side. When value changes to 85 - remote side exits
            }

            /* Ask for completion every comp_inter */
            //if (((i+1)%comp_inter == 0) || i == last) {
            if ((tmp_cnt == comp_inter) || i == last) {
                tmp_cnt = 0;
                sr.comp_type = VAPI_SIGNALED;
                ask_for_comp++;
                COND_PRINTF((" **** Posting sr signaled i=%d ask=%d\n",i+1,ask_for_comp));
            }

            
            res = VAPI_post_sr(hca_hndl, qp_hndl, &sr);
            if (res != VAPI_OK) {
                printf("Error: Posting send desc : %s\n", VAPI_strerror(res));
                return(RET_ERR);
            }

            /* Wait for completion every comp_inter starting from 2*comp_inter
             * This applies only when comp_iter > iter */
            if (ask_for_comp > 1) {
                COND_PRINTF(("**** Polling for sr comp i=%d ask=%d\n",i+1,ask_for_comp));
                if (RET_OK != PERF_poll_cq(hca_hndl,s_cq_hndl ,&sc,0, 
                                           sq_sem_p)) {
                    return RET_ERR;
                }
                ask_for_comp--;
            }
            j = (j+1)%params_p->in_params.num_of_qps;            
        }//for i - send packets

        

        /* Wait for final reply from receiver */
        while (ask_for_comp > 0) {
            if (RET_OK != PERF_poll_cq(hca_hndl,s_cq_hndl ,&rc,0, 
                                       rq_sem_p)) {
                return RET_ERR;
            }
            ask_for_comp--;
        } 
        get_clk64(t_end);

        printf("Time: [ %.1f : %.1f ] usec\n", CLK_TO_USEC(t_start), CLK_TO_USEC(t_end));
#if 0
        printf("BW: %.1f ops/sec [totaly %d operations] [Num. of QPs: %d ]\n", BW_BSEC(t_end - t_start,1.0 * iter),iter, params_p->in_params.num_of_qps);
#else
        printf("BW: %.1f ops/sec [totaly %d operations]\n", BW_BSEC(t_end - t_start,1.0 * iter),iter);
#endif
    }// Is sender
    else {
        while (*recv_buf == 84) {
        }
    }
    

    COND_PRINTF1(("\n************* Atomic Compare & Swap BW Test Finished  *********************\n\n"));
    return(RET_OK);
}


/***********************************************************************************************
* name:      atomic_bw_lat_test
* args:
*            params_p -- test parameter structure, need to filled before calling
* returns:
*            RET_OK
*            RET_ERR
*
* description: This test calculates the Latency between sending the atomic to completion
* on the sender side
*            
************************************************************************************************/  
ret_val_t atomic_bw_lat_test(test_params_tx *params_p)
{
    VAPI_ret_t             res;
    u_int64_t t_start=0, t_end=0;

    int i, ask_for_comp=0, last;

    VAPI_sr_desc_t         sr;
    VAPI_sg_lst_entry_t    sg_entry_s;


    //VAPI_wc_desc_t         rc;
    VAPI_wc_desc_t         sc;

    /* Copy all input parameters - don't want to rewrite the loop code */
    VAPI_hca_hndl_t        hca_hndl    = params_p->hca_hndl;
    VAPI_qp_hndl_t         qp_hndl     = params_p->port[CUR_PORT_IX].qp_vect[0].qp_hndl;

    VAPI_cq_hndl_t         s_cq_hndl   = params_p->port[CUR_PORT_IX].s_cq_hndl;
    MOSAL_semaphore_t    * sq_sem_p   = &params_p->port[CUR_PORT_IX].sq_sem;
    //MOSAL_semaphore_t    * rq_sem_p   = &params_p->port[CUR_PORT_IX].rq_sem;
    volatile u_int8_t               *send_buf   = params_p->send_buf;
    volatile u_int64_t              *recv_buf   = (u_int64_t*)(params_p->recv_buf);
    int                    is_sender   = params_p->in_params.is_sender;
    int                    iter        = params_p->in_params.iter;      
    int j=0;
    
    /**************    RDMA SR *****************************************************/

    *recv_buf = 84;

    sr.comp_type = VAPI_SIGNALED;
    sr.set_se = FALSE;
    sr.fence = FALSE;

    sr.opcode = VAPI_ATOMIC_CMP_AND_SWP;
    sr.remote_qkey = 0;
    sr.remote_addr = params_p->remote_add;
    sr.r_key = params_p->remote_rkey;
    sr.id = 1;
    sr.compare_add = 84;
    sr.swap = 84;


    sr.sg_lst_len = 1;
    sg_entry_s.len = 8;
    sg_entry_s.lkey = params_p->send_lkey;
    sg_entry_s.addr = (VAPI_virt_addr_t)(MT_virt_addr_t)send_buf; 
    sr.sg_lst_p = &sg_entry_s;

    /* sync the test start */
    sock_sync_msg(params_p->in_params.socket, is_sender);
    COND_PRINTF1(("\n************* Atomic Compare & Swap BW Test started  *********************\n\n"));


    ask_for_comp = 0;
    //Last iteration number
    last = iter - 1;

    if (is_sender) {
        get_clk64(t_start);
        for (i=0; i< iter; i++) {
            qp_hndl = params_p->port[CUR_PORT_IX].qp_vect[j].qp_hndl;
            sr.id = i;
            if (i == last) {
                sr.swap = 85; //Sign the last iteration for a remote side. When value changes to 85 - remote side exits
            }

            /* Ask for completion every time */
            sr.comp_type = VAPI_SIGNALED;
            //ask_for_comp++;
            COND_PRINTF((" **** Posting sr signaled i=%d ask=%d\n",i+1,ask_for_comp));

            
            res = VAPI_post_sr(hca_hndl, qp_hndl, &sr);
            if (res != VAPI_OK) {
                printf("Error: Posting send desc : %s\n", VAPI_strerror(res));
                return(RET_ERR);
            }

            //if (ask_for_comp > 1) {
                COND_PRINTF(("**** Polling for sr comp i=%d ask=%d\n",i+1,ask_for_comp));
                if (RET_OK != PERF_poll_cq(hca_hndl,s_cq_hndl ,&sc,0, 
                                           sq_sem_p)) {
                    return RET_ERR;
                }
                //ask_for_comp--;
            //}
            j = (j+1)%params_p->in_params.num_of_qps;            
        }//for i - send packets

        

        /* Wait for final reply from receiver */
        //while (ask_for_comp > 0) {
            //if (RET_OK != PERF_poll_cq(hca_hndl,s_cq_hndl ,&rc,0, 
                                       //rq_sem_p)) {
                //return RET_ERR;
            //}
            //ask_for_comp--;
        //} 
        get_clk64(t_end);

        printf("Time: [ %.1f : %.1f ] usec\n", CLK_TO_USEC(t_start), CLK_TO_USEC(t_end));
#if 0
        printf("Latency until completion: %.2f usec/operation [Num. of QPs: %d]\n", CLK_TO_USEC(t_end - t_start) / iter, params_p->in_params.num_of_qps);
#else
        printf("Latency until completion: %.2f usec/operation\n", CLK_TO_USEC(t_end - t_start) / iter);
#endif
    }// Is sender
    else {
        while (*recv_buf == 84) {
        }
    }
    

    COND_PRINTF1(("\n************* Atomic Compare & Swap BW Test Finished  *********************\n\n"));
    return(RET_OK);
}

/***********************************************************************************************
* name:      chipset_bw_test
* args:
*            params_p -- test parameter structure, need to filled before calling
* returns:
*            RET_OK
*            RET_ERR
*
* description:
*
************************************************************************************************/
ret_val_t chipset_test1(test_params_tx *params_p)
{

    VAPI_ret_t             res;
    u_int64_t t_start=0, t_end=0;

    VAPI_sr_desc_t         sr;
    VAPI_sg_lst_entry_t    sg_entry_s;
    VAPI_wc_desc_t         sc;

    /* Copy all input parameters - don't want to rewrite the loop code */
    VAPI_hca_hndl_t        hca_hndl    = params_p->hca_hndl;
    VAPI_qp_hndl_t         qp_hndl     = params_p->port[CUR_PORT_IX].qp_vect[0].qp_hndl;
    VAPI_cq_hndl_t         s_cq_hndl   = params_p->port[CUR_PORT_IX].s_cq_hndl;
    MOSAL_semaphore_t    * sq_sem_p    = &params_p->port[CUR_PORT_IX].sq_sem;
    int                    size        = params_p->in_params.size;
    int                    iter        = params_p->in_params.iter;

    VAPI_phy_buf_t phy_buf;
    VAPI_mrw_t             mr_in, mr_out;
    VAPI_mr_hndl_t         mr_hndl =  VAPI_INVAL_HNDL;

    /* Ignore the MR of create_resources and Register a different physical memory region */
    mr_in.acl =  VAPI_EN_LOCAL_WRITE | VAPI_EN_REMOTE_WRITE;
    mr_in.l_key = 0;
    mr_in.pd_hndl = params_p->pd_hndl;
    mr_in.r_key = 0;
    mr_in.size = 0x80000000; //2GB
    mr_in.start = (VAPI_virt_addr_t)(MT_virt_addr_t)0; //start at 0
    mr_in.type = VAPI_MPR; //Physical Memory Region

    mr_in.iova_offset = 0;
    phy_buf.start=0;
    phy_buf.size=0x80000000;
    mr_in.pbuf_list_len = 1;
    mr_in.pbuf_list_p = &phy_buf;

    res = VAPI_register_mr(hca_hndl, &mr_in, &mr_hndl, &mr_out);
    if (res != VAPI_OK) {
        printf("Error: Registering MR : %s\n", VAPI_strerror(res));
        return(RET_ERR);
    }
    params_p->send_lkey = mr_out.l_key;


    printf("\n\n************* CHIPSET BW Test started  *********************\n\n");

    sr.opcode = VAPI_SEND;
    sr.comp_type = VAPI_SIGNALED;
    sr.set_se = FALSE;
    sr.remote_qkey = 0;
    sr.id = 1;
    sr.sg_lst_len = 1;
    sr.sg_lst_p = &sg_entry_s;

    sg_entry_s.len = size;
    sg_entry_s.lkey = params_p->send_lkey;
    sg_entry_s.addr = (VAPI_virt_addr_t)(MT_virt_addr_t)(0x1<<25);

    while (iter--) {


        get_clk64(t_start);

        res = VAPI_post_sr(hca_hndl, qp_hndl, &sr);
        if (res != VAPI_OK) {
            COND_PRINTF(("Error: Posting send desc(%d): %s\n",iter, VAPI_strerror(res)));
            return(RET_ERR);
        }

        COND_PRINTF(("**** Polling for sr comp i=%d \n",iter+1));
        if (RET_OK != PERF_poll_cq(hca_hndl,s_cq_hndl ,&sc,params_p->in_params.use_send_events,sq_sem_p)) {
            return RET_ERR;
        }

        get_clk64(t_end);

        printf("%.1f [MB/S] - Msg Size %.1f [MB] \n",BW_MBSEC(t_end - t_start,size),(1.0*size)/(1024.0*1024));

    }
    
    /* deregister this new special PHMR */
    if (mr_hndl != VAPI_INVAL_HNDL) {
        VAPI_deregister_mr(hca_hndl,mr_hndl);
    }

    printf("\n************* Test Finished  *********************\n\n");
    return(RET_OK);
}




/***********************************************************************************************
* name:      send_bw_test
* args:
*            params_p -- test parameter structure, need to filled before calling
* returns:
*            RET_OK
*            RET_ERR
*
* description:     
*            This is the main loop for the UD and RC send BW test. The test sends N packets
*            then waits for acknowledge from the receiver, and calculate the BW.
*            
************************************************************************************************/  
ret_val_t send_bw_list_test(test_params_tx *params_p)
{
    VAPI_ret_t             res;
    u_int64_t t_start=0, t_end=0;

    int i,comp_inter,ask_for_comp=0;

    VAPI_rr_desc_t         rr;
    VAPI_sg_lst_entry_t    sg_entry_r;
    VAPI_sr_desc_t         *sr;
    VAPI_sg_lst_entry_t    sg_entry_s;
    VAPI_wc_desc_t         rc;
    VAPI_wc_desc_t         sc;

    u_int32_t              srq_posted_actual;
    /* Copy all input parameters - don't want to rewrite the loop code */
    VAPI_hca_hndl_t        hca_hndl    = params_p->hca_hndl;
    VAPI_qp_hndl_t         qp_hndl     = params_p->port[CUR_PORT_IX].qp_vect[0].qp_hndl;
    VAPI_cq_hndl_t         s_cq_hndl   = params_p->port[CUR_PORT_IX].s_cq_hndl;
    VAPI_cq_hndl_t         r_cq_hndl   = params_p->port[CUR_PORT_IX].r_cq_hndl;
    VAPI_srq_hndl_t        srq_hndl    = params_p->port[CUR_PORT_IX].qp_vect[0].srq_hndl;
    MOSAL_semaphore_t    * sq_sem_p   = &params_p->port[CUR_PORT_IX].sq_sem;
    MOSAL_semaphore_t    * rq_sem_p   = &params_p->port[CUR_PORT_IX].rq_sem;
    u_int8_t               *recv_buf   = params_p->recv_buf;
    u_int8_t               *send_buf   = params_p->send_buf;
    int                    is_sender   = params_p->in_params.is_sender;
    int                    size        = params_p->in_params.size;
    int                    iter        = params_p->in_params.iter;      
    
    int                    send_list_len,num_of_posts;

    int max_msg_size = MIN2(params_p->in_params.mr_recv_size,params_p->in_params.mr_send_size);
    int max_sends = MIN2(params_p->in_params.num_r_wqe,params_p->in_params.num_s_wqe); 

    send_list_len = MAX(1,params_p->in_params.general_param);
    
    sr = VMALLOC(send_list_len*sizeof(VAPI_sr_desc_t));

    if (params_p->in_params.transport == IB_TS_UD) {
        max_msg_size = 2048;
        if (size > 2048) {
            printf("Warning!!!: UD max size is 2048\n");
            size = 2048;
        }
    }

    if (size > params_p->in_params.mr_recv_size) {
        printf("Warning!!!: size=%d > mr_recv_size=%d --- changing iter to %d\n", 
               size,params_p->in_params.mr_recv_size,params_p->in_params.mr_recv_size);
        size = params_p->in_params.mr_recv_size;
    }

    rr.opcode = VAPI_RECEIVE;
    rr.comp_type = VAPI_SIGNALED;
    rr.sg_lst_len = 1;
    rr.sg_lst_p  = &sg_entry_r;
    sg_entry_r.lkey = params_p->recv_lkey;
    sg_entry_r.len = params_p->in_params.mr_recv_size;
    sg_entry_r.addr = (VAPI_virt_addr_t)(MT_virt_addr_t)recv_buf;

    /***************************************************************************/
    sg_entry_s.addr = (VAPI_virt_addr_t)(MT_virt_addr_t)send_buf;
    sg_entry_s.len = size;
    sg_entry_s.lkey = params_p->send_lkey;

    for (i=0 ; i < send_list_len ; ++i) {
        sr[i].comp_type = VAPI_UNSIGNALED;
        if (i == (send_list_len - 1)) {
            sr[i].comp_type = VAPI_SIGNALED;
        }
        sr[i].set_se = FALSE;
        sr[i].opcode = VAPI_SEND;
        sr[i].remote_qkey = 0;
        sr[i].id = i;
        sr[i].remote_qp = params_p->port[CUR_PORT_IX].qp_vect[0].d_qpnum;
        sr[i].remote_ah = params_p->port[CUR_PORT_IX].av_hndl;

        sr[i].sg_lst_len = 1;
        sr[i].sg_lst_p = &sg_entry_s; // use the same sg_ for all list
    }

    /* Post all possible recive in advance. Sender need 1 while receiver need N*/
    if (!is_sender) {
        for (i=0;i<max_sends;i++) {
            rr.id = i;
            if (params_p->in_params.srq_num > 0)
            {
                res = VAPI_post_srq(hca_hndl, srq_hndl, 1, &rr, &srq_posted_actual);
                //printf("posted %d SRQs\n", srq_posted_actual);
            }
            else
            {
                res = VAPI_post_rr(hca_hndl, qp_hndl, &rr);
            }
            if (res != VAPI_OK) {
                printf("Error: Posting Recv buffer : %s\n", VAPI_strerror(res));
                return(RET_ERR);
            }
        }
    }
    else {
        rr.id = 0;
        if (params_p->in_params.srq_num > 0)
        {
            res = VAPI_post_srq(hca_hndl, srq_hndl, 1, &rr, &srq_posted_actual);
            //printf("posted %d SRQs\n", srq_posted_actual);
        }
        else
        {
            res = VAPI_post_rr(hca_hndl, qp_hndl, &rr);
        }
        if (res != VAPI_OK) {
            printf("Error: Posting Recv buffer : %s\n", VAPI_strerror(res));
            return(RET_ERR);
        }
    }


    /* sync the test start */
    sock_sync_msg(params_p->in_params.socket, is_sender);
    COND_PRINTF1(("\n\n************* %s LIST BW Test started  *********************\n\n",
                  params_p->in_params.transport == IB_TS_UD ? "UD" : "RC"));

    comp_inter = send_list_len;
    ask_for_comp = 0;

    num_of_posts = iter / send_list_len;

    if (is_sender) {
        get_clk64(t_start);
        for (i=0; i< num_of_posts; i++) {

            ask_for_comp++;

            res = EVAPI_post_sr_list(hca_hndl, qp_hndl, send_list_len, sr);
            if (res != VAPI_OK) {
                printf("Error: Posting send desc : %s\n", VAPI_strerror(res));
                return(RET_ERR);
            }

            /* Wait for completion every comp_inter starting from 2*comp_inter
             * This applies only when comp_iter < iter */
            if (ask_for_comp > 1) {
                COND_PRINTF(("**** Polling for sr comp i=%d ask=%d\n",i+1,ask_for_comp));
                if (RET_OK != PERF_poll_cq(hca_hndl,s_cq_hndl ,&sc,params_p->in_params.use_send_events, 
                                           sq_sem_p)) {
                    return RET_ERR;
                }
                ask_for_comp--;
                COND_PRINTF(("Polling for sr comp %d %d\n",i,ask_for_comp));
            }
        }//for i - send packets

        /* Wait for completion on last send */
        while (ask_for_comp>0) {
            COND_PRINTF(("**** Polling for last sr comp i=%d ask=%d\n",i,ask_for_comp));
            /* Poll for the last send completion */
            if (RET_OK != PERF_poll_cq(hca_hndl,s_cq_hndl ,&sc,params_p->in_params.use_send_events,
                                       sq_sem_p)) {
                return RET_ERR;
            }
            ask_for_comp--;
            COND_PRINTF(("Polling for sr comp %d %d\n",i,ask_for_comp));
        }

        /* Wait for final reply from receiver */
        if (RET_OK != PERF_poll_cq(hca_hndl, r_cq_hndl ,&rc,
                                   params_p->in_params.use_recv_events,  rq_sem_p)) {
            return RET_ERR;
        }
        COND_PRINTF(("Sender came out of Poll msg_size=%u\n",rc.byte_len));
        get_clk64(t_end);

        printf("BW: %.1f MBytes/sec [size: %d bytes]\n", BW_MBSEC(t_end - t_start,1.0*(unsigned)size * iter),size);

    }// Is sender

    else { /* Receiver side */
        for (i=0; i< iter; i++) {
            //COND_PRINTF(("Receiver - polling i=%u\n",i));
            if (RET_OK != PERF_poll_cq(hca_hndl, r_cq_hndl ,&rc,
                                       params_p->in_params.use_recv_events,  rq_sem_p)) {
                return RET_ERR;
            }
            if (params_p->in_params.srq_num > 0)
            {
                res = VAPI_post_srq(hca_hndl, srq_hndl, 1, &rr, &srq_posted_actual);
                //printf("posted %d SRQs\n", srq_posted_actual);
            }
            else
            {
                res = VAPI_post_rr(hca_hndl, qp_hndl, &rr);
            }
            if (res != VAPI_OK) {
                printf("Error: Posting Recv buffer : %s\n", VAPI_strerror(res));
                return(RET_ERR);
            }
        }

        /* Send reply to sender after last message. Use 0 length message!!! */
        sr[0].comp_type = VAPI_SIGNALED;
        sg_entry_s.addr = (VAPI_virt_addr_t)(MT_virt_addr_t)send_buf;
        sr[0].sg_lst_len = 0;
        res = VAPI_post_sr(hca_hndl, qp_hndl, &sr[0]);
        if (res != VAPI_OK) {
            printf("Error: Posting send desc : %s\n", VAPI_strerror(res));
            return(RET_ERR);
        }

        /* Wait for send comp */
        if (RET_OK != PERF_poll_cq(hca_hndl,s_cq_hndl ,&sc,params_p->in_params.use_send_events, 
                                   sq_sem_p)) {
            return RET_ERR;
        }
    }
    COND_PRINTF1(("\n************* %s BW Test Finished  *********************\n\n",
                  params_p->in_params.transport == IB_TS_UD ? "UD" : "RC"));
    sock_sync_msg(params_p->in_params.socket, is_sender);
    return(RET_OK);
}

