/*
  This software is available to you under a choice of one of two
  licenses.  You may choose to be licensed under the terms of the GNU
  General Public License (GPL) Version 2, available at
  <http://www.fsf.org/copyleft/gpl.html>, or the OpenIB.org BSD
  license, available in the LICENSE.TXT file accompanying this
  software.  These details are also available at
  <http://openib.org/license.html>.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
  BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
  ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
  CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
  SOFTWARE.

  Copyright (c) 2004 Mellanox Technologies Ltd.  All rights reserved.
*/

#include <vapi.h>
#include <evapi.h>
#include <vapi_common.h>
#include <stdio.h>

#define TIME_LEVEL 1
#include "utils.h" 
#include "all_bw_lat.h"
#include "sock.h" 


static ret_val_t special_poll_cq(VAPI_hca_hndl_t  hca_hndl, VAPI_cq_hndl_t    cq_hndl,
                                 VAPI_wc_desc_t *sc_p, int use_events, 
                                 MOSAL_semaphore_t *sem_cq_p, u_int64_t *time_last);

#define EZRA_DELAY(n) \
{                \
    int kkk;           \
    volatile int lll;\
    lll = 0;\
    for (kkk=0;kkk<n;kkk++) {\
       lll += (lll*(kkk+lll-1)); \
    }                         \
}


#define VECT_N 5000

static FILE *out_file = NULL;


static void analayze_vect(u_int64_t *vect, int len, u_int64_t thresh, vect_res *res, float factor)
{
    int i;
    int below_num =0;
    u_int64_t sum_below =0, total_sum=0;
    u_int64_t min_val, max_val;


    res->thresh = thresh;
    min_val = _UI64_MAX;
    max_val = 0;

    for (i=0; i< len; i++) {
        vect[i] = (u_int64_t) (((ADJUST_TYPE)vect[i]) * factor);

        if (vect[i] > max_val){
             max_val = vect[i];
             res->max_i = i;
        }
        if (vect[i] < min_val) {
            min_val = vect[i];
            res->min_i = i;
        }

        total_sum += vect[i];
        if (vect[i] < thresh) {
            sum_below += vect[i];
            below_num++;
        }
    }

    res->avg = total_sum/len;
    res->avg_below = below_num > 0 ? sum_below / below_num : 0;
    res->num_above = len - below_num;
    res->max = max_val;
    res->min = min_val;
}



/* Just test send times */
ret_val_t test_case_special_xx(test_params_tx *params_p)
{

    VAPI_ret_t             res;
    int send_id,i,ask_comp,comp_inter;

    vect_res vect_res;


    u_int64_t t_start=0, t_end=0;
    u_int64_t t_post[VECT_N], t_end_post , t_start_post, t_last_poll[VECT_N];


    VAPI_rr_desc_t         rr;
    VAPI_sg_lst_entry_t    sg_entry_r;
    VAPI_sr_desc_t         sr;
    VAPI_sg_lst_entry_t    sg_entry_s;
    VAPI_wc_desc_t         rc;
    //  VAPI_wc_desc_t         sc;


    u_int32_t              srq_posted_actual;
    // Copy all input parameters - don't want to rewrite the loop code
    VAPI_hca_hndl_t        hca_hndl    = params_p->hca_hndl;
    VAPI_qp_hndl_t         qp_hndl     = params_p->port[CUR_PORT_IX].qp_vect[0].qp_hndl;

//    VAPI_cq_hndl_t         s_cq_hndl   = params_p->port[CUR_PORT_IX].s_cq_hndl;
    VAPI_cq_hndl_t         r_cq_hndl   = params_p->port[CUR_PORT_IX].r_cq_hndl;
    VAPI_srq_hndl_t        srq_hndl    = params_p->port[CUR_PORT_IX].qp_vect[0].srq_hndl;
    MOSAL_semaphore_t    * rq_sem_p   = &params_p->port[CUR_PORT_IX].rq_sem;
    u_int8_t               *recv_buf   = params_p->recv_buf;
    u_int8_t               *send_buf   = params_p->send_buf;
    int                    is_sender   = params_p->in_params.is_sender;
    int                    size        = params_p->in_params.size;
    int                    iter        = params_p->in_params.iter;      
    int                    skip        = params_p->in_params.skip;
    int act_skip, num_of_pre_post;
    int use_inline = FALSE;


    memset(t_post, 0 , VECT_N*sizeof(t_post[0]));
    memset(t_last_poll, 0 , VECT_N*sizeof(t_post[0]));
    t_start_post = t_end_post = 0;

    out_file = fopen("perf_out_data.txt", "w");

    send_id = 0;
    if (params_p->in_params.transport == IB_TS_UD) {
        if (size > 2048) {
            printf("Warning!!!: UD max size is 2048\n");
            size = 2048;
        }
    }

    if (size > params_p->in_params.mr_recv_size) {
        printf("Warning!!!: size=%d > mr_recv_size=%d --- changing iter to %d\n", 
               size,params_p->in_params.mr_recv_size,params_p->in_params.mr_recv_size);
        size = params_p->in_params.mr_recv_size;
    }

    if ((params_p->in_params.transport == IB_TS_UD) &&  params_p->in_params.use_mwindow) {
        printf("Error!!!: Can't bind window in UD!!!!\n");
        return RET_ERR;

    }


    rr.opcode = VAPI_RECEIVE;
    rr.comp_type = VAPI_SIGNALED;
    rr.sg_lst_len = 1;
    sg_entry_r.lkey = params_p->recv_lkey;
    sg_entry_r.len = params_p->in_params.mr_recv_size;
    sg_entry_r.addr = (VAPI_virt_addr_t)(MT_virt_addr_t)recv_buf;
    rr.sg_lst_p  = &sg_entry_r;

    num_of_pre_post = (params_p->in_params.general_param != -1) ? params_p->in_params.general_param :
        params_p->in_params.rec_buf_num;

    for (i=0; i< num_of_pre_post; i++) {
        rr.id = i;
        if (params_p->in_params.srq_num > 0)
        {
            res = VAPI_post_srq(hca_hndl, srq_hndl, 1, &rr, &srq_posted_actual);
            //printf("posted %d SRQs\n", srq_posted_actual);
        }
        else
        {
            res = VAPI_post_rr(hca_hndl, qp_hndl, &rr);
        }
        if (res != VAPI_OK) {
            printf("Error: Posting Recv buffer : %s\n", VAPI_strerror(res));
            return(RET_ERR);
        }
        COND_PRINTF(("Posted RR id=" U64_FMT " i=%d\n",rr.id,i));
    }


/***************************************************************************/
    sr.comp_type = VAPI_UNSIGNALED;
    sr.set_se = FALSE;

    sr.opcode = VAPI_SEND;
    sr.remote_qkey = 0;
    sr.id = 1;

    //needed only for UD
    sr.remote_qp = params_p->port[CUR_PORT_IX].qp_vect[0].d_qpnum; 
    sr.remote_ah = params_p->port[CUR_PORT_IX].av_hndl;

    sr.sg_lst_len = 1;
    sg_entry_s.len = size;
    sg_entry_s.lkey = params_p->send_lkey;
    sg_entry_s.addr = (VAPI_virt_addr_t)(MT_virt_addr_t)send_buf; 
    sr.sg_lst_p = &sg_entry_s;

    /* sync start test */
    sock_sync_msg(params_p->in_params.socket, is_sender);
    COND_PRINTF1(("\n\n************* %s Latency Test started  (Messages up to %d use inline data)  ***************\n\n",
                  params_p->in_params.transport == IB_TS_UD ? "UD" : "RC",params_p->max_inline_size));


    if ((iter+skip) <= (int)params_p->in_params.num_s_wqe) {
        comp_inter = params_p->in_params.num_s_wqe;
    }
    else {
        comp_inter = (params_p->in_params.num_s_wqe / 2) - 1;
    }

    ask_comp =0;

    if (size==0) {
        sr.sg_lst_len = 0;
    }

    // We start the timing AFTER the first send
    act_skip = skip > 0 ? (skip -1) : skip;

    sg_entry_s.len = size;

    send_id = 1;

    if ((size >0 ) &&(size <= params_p->max_inline_size))
    {
        use_inline = TRUE;
    }

    for (i=0; i<iter; i++) {
        if (is_sender) {
            if (i == 0) {
                get_clk64(t_start);
            }
            get_clk64(t_start_post);
            //if ((size >0 ) &&(size <= params_p->max_inline_size)) {
            if (use_inline == TRUE) {
                res = EVAPI_post_inline_sr(hca_hndl, qp_hndl, &sr);
            }
            else {
                res = VAPI_post_sr(hca_hndl, qp_hndl, &sr);
            }
            if (res != VAPI_OK) {
                printf("Error: Posting send desc : %s\n", VAPI_strerror(res));
                return(RET_ERR);
            }
            get_clk64(t_end_post);
            t_post[i] = t_end_post - t_start_post ;
            COND_PRINTF(("Posted send desc %d\n", send_id++));
            EZRA_DELAY(100000);
        }
        else {
            if (RET_OK != special_poll_cq(hca_hndl,r_cq_hndl ,
                                          &rc,params_p->in_params.use_recv_events, 
                                          rq_sem_p,&(t_last_poll[i]))) {
                return RET_ERR;
            }
            rr.id = rc.id;
            if (params_p->in_params.srq_num > 0)
            {
                res = VAPI_post_srq(hca_hndl, srq_hndl, 1, &rr, &srq_posted_actual);
                //printf("posted %d SRQs\n", srq_posted_actual);
            }
            else
            {
                res = VAPI_post_rr(hca_hndl, qp_hndl, &rr);
            }
            if (res != VAPI_OK) {
                printf("Error: Posting Recv buffer : %s\n", VAPI_strerror(res));
                return(RET_ERR);
            }
            COND_PRINTF(("  Posted RR id=" U64_FMT " i=%d\n",rr.id,i));
        }
    }
    get_clk64(t_end);

    printf("Latency: %.2f usec [size: %d bytes]\n",CLK_TO_USEC(t_end - t_start) /( (iter)* 2.0),size);


    if (is_sender) {
        analayze_vect(t_post, i, USEC_TO_CLK(1.5), &vect_res,1.0);
        printf("Post: \n"
               "    avg = %.2f\n"
               "    avg_below = %.2f\n"
               "    min = %.2f [i=%d]\n"
               "    max = %.2f [i=%d]\n",
               CLK_TO_USEC(vect_res.avg),CLK_TO_USEC(vect_res.avg_below),
               CLK_TO_USEC(vect_res.min),vect_res.min_i,
               CLK_TO_USEC(vect_res.max),vect_res.max_i);
        if(out_file != NULL){
            int kk;
            for (kk=0; kk<i; kk++){
                fprintf(out_file,"%.2f \n",CLK_TO_USEC(t_post[kk]));
            }
        }
    }
    else {
        analayze_vect(t_last_poll, i, USEC_TO_CLK(2.0), &vect_res,1.0);
        printf("Poll: \n"
               "    avg = %.2f\n"
               "    avg_below = %.2f\n"
               "    min = %.2f [i=%d]\n"
               "    max = %.2f [i=%d]\n",
               CLK_TO_USEC(vect_res.avg),CLK_TO_USEC(vect_res.avg_below),
               CLK_TO_USEC(vect_res.min),vect_res.min_i,
               CLK_TO_USEC(vect_res.max),vect_res.max_i);
        #if 0
        if(out_file != NULL){
            int kk;
            for (kk=0; kk<i; kk++){
                fprintf(out_file,"%.2f \n",CLK_TO_USEC(t_last_poll[kk]));
            }
        }
        #endif
    }

    printf("Latency = %.2f min = %.2f\n",
             CLK_TO_USEC(t_end - t_start) /( (iter-act_skip)* 2.0),CLK_TO_USEC(vect_res.min));


    COND_PRINTF1(("\n************* %s Latency Test finished  *********************\n\n",
                  params_p->in_params.transport == IB_TS_UD ? "UD" : "RC"));
    sock_sync_msg(params_p->in_params.socket, is_sender);
    if(out_file != NULL){
        fclose(out_file);
    }
    return(RET_OK);
}



static ret_val_t special_poll_cq(VAPI_hca_hndl_t  hca_hndl, VAPI_cq_hndl_t    cq_hndl,
                                 VAPI_wc_desc_t *sc_p, int use_events, MOSAL_semaphore_t *sem_cq_p, u_int64_t *time_last)
{

    VAPI_ret_t             res;

    u_int64_t t_poll_start=0, t_poll_end=0;

    res = VAPI_CQ_EMPTY;
    get_clk64(t_poll_start);
    res = VAPI_poll_cq(hca_hndl, cq_hndl, sc_p);

    while (res == VAPI_CQ_EMPTY) {

#ifdef CQ_BLOCK
        if (use_events) {
            res = EVAPI_poll_cq_block(hca_hndl, cq_hndl, /*CQ_BLOCK_TIMEOUT*/ 0, sc_p);
            if (res != VAPI_CQ_EMPTY) {
                continue;
            }
            if (res != VAPI_OK) {
                printf("Error: Poll CQ Block: %s\n", VAPI_strerror(res));
                return(RET_ERR);
            }
            continue;
        }

#else
        if (use_events) {
            VAPI_req_comp_notif(hca_hndl,cq_hndl,VAPI_NEXT_COMP);
            res = VAPI_poll_cq(hca_hndl, cq_hndl, sc_p);
            if (res != VAPI_CQ_EMPTY) {
                continue;
            }
            if (MOSAL_sem_acq(sem_cq_p,TRUE) != 0) { /* Wait for notification handler */
                printf(MT_FLFMT("Failed acquiring RQ CQ completion semaphore."));
            }
        }
#endif
        if (!use_events) {
            get_clk64(t_poll_start);
            res = VAPI_poll_cq(hca_hndl, cq_hndl, sc_p);
        }
    }
    get_clk64(t_poll_end);

    *time_last = t_poll_end - t_poll_start;
    if (res != VAPI_OK) {
        printf("Error: Polling CQ : %s\n", VAPI_strerror(res));
        return(RET_ERR);
    }
    if (sc_p->status != VAPI_SUCCESS) {
        printf("%s: completion with error (%d) detected\n", __func__, sc_p->status);
        return(RET_ERR);
    }
    return(RET_OK);
}





ret_val_t test_case_special(test_params_tx *params_p)
{

    VAPI_ret_t             res;
    int send_id,i,ask_comp,last_post,comp_inter, tmp_cnt=1;


    u_int64_t t_start=0, t_end=0;
    u_int64_t *t_lat_p,t_lat[VECT_N], t_start_i , t_end_i;
    vect_res vect_res;


    VAPI_rr_desc_t         rr;
    VAPI_sg_lst_entry_t    sg_entry_r;
    VAPI_sr_desc_t         sr;
    VAPI_sg_lst_entry_t    sg_entry_s;
    VAPI_wc_desc_t         rc;
    VAPI_wc_desc_t         sc;


    u_int32_t              srq_posted_actual;
    // Copy all input parameters - don't want to rewrite the loop code
    VAPI_hca_hndl_t        hca_hndl    = params_p->hca_hndl;
    VAPI_qp_hndl_t         qp_hndl     = params_p->port[CUR_PORT_IX].qp_vect[0].qp_hndl;

    VAPI_cq_hndl_t         s_cq_hndl   = params_p->port[CUR_PORT_IX].s_cq_hndl;
    VAPI_cq_hndl_t         r_cq_hndl   = params_p->port[CUR_PORT_IX].r_cq_hndl;
    VAPI_srq_hndl_t        srq_hndl    = params_p->port[CUR_PORT_IX].qp_vect[0].srq_hndl;
    MOSAL_semaphore_t    * sq_sem_p   = &params_p->port[CUR_PORT_IX].sq_sem;
    MOSAL_semaphore_t    * rq_sem_p   = &params_p->port[CUR_PORT_IX].rq_sem;
    u_int8_t               *recv_buf   = params_p->recv_buf;
    u_int8_t               *send_buf   = params_p->send_buf;
    int                    is_sender   = params_p->in_params.is_sender;
    int                    size        = params_p->in_params.size;
    int                    iter        = params_p->in_params.iter;      
    int                    skip        = params_p->in_params.skip;
    int act_skip,num_of_pre_post,samp_i;
    int use_inline = FALSE;

    memset(t_lat, 0 , VECT_N*sizeof(t_lat[0]));
    t_start_i = t_end_i = 0;
    out_file = fopen("perf_out_data.txt", "w");

    send_id = 0;
    if (params_p->in_params.transport == IB_TS_UD) {
        if (size > 2048) {
            printf("Warning!!!: UD max size is 2048\n");
            size = 2048;
        }
    }

    if (size > params_p->in_params.mr_recv_size) {
        printf("Warning!!!: size=%d > mr_recv_size=%d --- changing iter to %d\n", 
               size,params_p->in_params.mr_recv_size,params_p->in_params.mr_recv_size);
        size = params_p->in_params.mr_recv_size;
    }

    if ((params_p->in_params.transport == IB_TS_UD) &&  params_p->in_params.use_mwindow) {
        printf("Error!!!: Can't bind window in UD!!!!\n");
        return RET_ERR;

    }


    rr.opcode = VAPI_RECEIVE;
    rr.comp_type = VAPI_SIGNALED;
    rr.sg_lst_len = 1;
    sg_entry_r.lkey = params_p->recv_lkey;
    sg_entry_r.len = params_p->in_params.mr_recv_size;
    sg_entry_r.addr = (VAPI_virt_addr_t)(MT_virt_addr_t)recv_buf;
    rr.sg_lst_p  = &sg_entry_r;

    num_of_pre_post = (params_p->in_params.general_param != -1) ? params_p->in_params.general_param :
        params_p->in_params.rec_buf_num;

    for (i=0; i< num_of_pre_post; i++) {
        rr.id = i;
        if (params_p->in_params.srq_num > 0)
        {
            res = VAPI_post_srq(hca_hndl, srq_hndl, 1, &rr, &srq_posted_actual);
            //printf("posted %d SRQs\n", srq_posted_actual);
        }
        else
        {
            res = VAPI_post_rr(hca_hndl, qp_hndl, &rr);
        }
        if (res != VAPI_OK) {
            printf("Error: Posting Recv buffer : %s\n", VAPI_strerror(res));
            return(RET_ERR);
        }
        COND_PRINTF(("Posted RR id=" U64_FMT " i=%d\n",rr.id,i));
    }


/***************************************************************************/
    sr.comp_type = VAPI_UNSIGNALED;
    sr.set_se = FALSE;

    sr.opcode = VAPI_SEND;
    sr.remote_qkey = 0;
    sr.id = 1;

    //needed only for UD
    sr.remote_qp = params_p->port[CUR_PORT_IX].qp_vect[0].d_qpnum; 
    sr.remote_ah = params_p->port[CUR_PORT_IX].av_hndl;

    sr.sg_lst_len = 1;
    sg_entry_s.len = size;
    sg_entry_s.lkey = params_p->send_lkey;
    sg_entry_s.addr = (VAPI_virt_addr_t)(MT_virt_addr_t)send_buf; 
    sr.sg_lst_p = &sg_entry_s;

    /* sync start test */
    sock_sync_msg(params_p->in_params.socket, is_sender);

    if ((iter+skip) <= (int)params_p->in_params.num_s_wqe) {
        comp_inter = params_p->in_params.num_s_wqe;
    }
    else {
        comp_inter = (params_p->in_params.num_s_wqe / 2) - 1;
    }

    ask_comp =0;

    if (size==0) {
        sr.sg_lst_len = 0;
    }

    // We start the timing AFTER the first send
    act_skip = skip > 0 ? (skip -1) : skip;

    sg_entry_s.len = size;

    send_id = 1;

    if ((size >0 ) &&(size <= params_p->max_inline_size))
    {
        use_inline = TRUE;
    }

    // Post the first send req
    samp_i = 0;
    get_clk64_c(t_start_i);
    if (is_sender) {
        if (skip == 0) {
            get_clk64(t_start);
        }
        //if ((size >0 ) &&(size <= params_p->max_inline_size)) {
        if (use_inline == TRUE) {
            res = EVAPI_post_inline_sr(hca_hndl, qp_hndl, &sr);
        }
        else {
            res = VAPI_post_sr(hca_hndl, qp_hndl, &sr);
        }
        if (res != VAPI_OK) {
            printf("Error: Posting send desc : %s\n", VAPI_strerror(res));
            return(RET_ERR);
        }
        COND_PRINTF(("Posted send desc %d\n", send_id++));
    }//End of ONLY sender

    for (i=0; i<(iter + act_skip); i++, tmp_cnt++) {

        if (RET_OK != PERF_poll_cq(hca_hndl,r_cq_hndl ,&rc,params_p->in_params.use_recv_events, 
                                   rq_sem_p)) {
            return RET_ERR;
        }
        get_clk64_c(t_end_i);
        t_lat[samp_i++] =  t_end_i - t_start_i;
        //printf("=== dif=%Lu end=%Lu start=%Lu i=%d\n",t_end_i - t_start_i,t_end_i,t_start_i,i);
        get_clk64_c(t_start_i);
        
        if ((i == act_skip) && (skip>0) && is_sender ) {
            get_clk64(t_start);
        }


        last_post = (is_sender && i== (iter + skip -2)) || (!is_sender && (i==(iter + skip -1)));

        sr.comp_type = VAPI_UNSIGNALED;           

        //if (((i+1)%comp_inter == 0) || last_post) {
        if ((tmp_cnt == comp_inter) || last_post) {
            tmp_cnt = 0;
            sr.comp_type = VAPI_SIGNALED;
            ask_comp++;
            COND_PRINTF(("Ask comp [ask=%d i=%d ]\n", ask_comp,i+1));
        }

        // Only receiver post the last one
        if (!(is_sender && (i==(iter + skip -1)))) {
            //if ((size >0) &&(size <= params_p->max_inline_size)) {
            if (use_inline == TRUE) {
                res = EVAPI_post_inline_sr(hca_hndl, qp_hndl, &sr);
            }
            else {
                res = VAPI_post_sr(hca_hndl, qp_hndl, &sr);
            }
            if (res != VAPI_OK) {
                printf("Error: Posting send desc : %s\n", VAPI_strerror(res));
                return(RET_ERR);
            }
            COND_PRINTF(("Posted send desc %d\n" , send_id++));
        }

        rr.id = rc.id;
        if (params_p->in_params.srq_num > 0)
        {
            res = VAPI_post_srq(hca_hndl, srq_hndl, 1, &rr, &srq_posted_actual);
            //printf("posted %d SRQs\n", srq_posted_actual);
        }
        else
        {
            res = VAPI_post_rr(hca_hndl, qp_hndl, &rr);
        }
        if (res != VAPI_OK) {
            printf("Error: Posting Recv buffer : %s\n", VAPI_strerror(res));
            return(RET_ERR);
        }
        COND_PRINTF(("  Posted RR id=" U64_FMT " i=%d\n",rr.id,i));
        //poll for SR
        if (ask_comp >1) {
            COND_PRINTF(("**** Polling for sr comp i=%d ask=%d\n",i+1,ask_comp));
            if (RET_OK != PERF_poll_cq(hca_hndl,s_cq_hndl ,&sc,params_p->in_params.use_send_events, 
                                       sq_sem_p)) {
                return RET_ERR;
            }
            ask_comp--;
        }

    }// end of iteration loop - for i

    if (is_sender) {
        get_clk64(t_end);
        COND_PRINTF(("Latency: %.2f usec [size: %d bytes]\n",
                     CLK_TO_USEC(t_end - t_start) /( (iter-act_skip)* 2.0),size));
        fflush(stdout);
    }
    samp_i -=1;
    t_lat_p = &t_lat[1];

    analayze_vect(t_lat_p, samp_i, USEC_TO_CLK(10.0), &vect_res,0.5);
    printf("Post: \n"
           "    avg = %.2f\n"
           "    avg_below [%.1f] = %.2f\n"
           "    min = %.2f [i=%d]\n"
           "    max = %.2f [i=%d]\n",
           CLK_TO_USEC(vect_res.avg),CLK_TO_USEC(vect_res.thresh),CLK_TO_USEC(vect_res.avg_below),
           CLK_TO_USEC(vect_res.min),vect_res.min_i,
           CLK_TO_USEC(vect_res.max),vect_res.max_i);

    printf("Latency = %.2f min = %.2f\n",
             CLK_TO_USEC(t_end - t_start) /( (iter-act_skip)* 2.0),CLK_TO_USEC(vect_res.min));

    if(out_file != NULL){
        int kk;
        for (kk=0; kk<samp_i; kk++){
            fprintf(out_file,"%.2f \n",CLK_TO_USEC(t_lat_p[kk]));
        }
    }
    // Poll for the last SR complition
    while (ask_comp>0) {
        COND_PRINTF(("**** Polling for last sr comp i=%d ask=%d\n",i,ask_comp));
        if (RET_OK != PERF_poll_cq(hca_hndl,s_cq_hndl ,&sc,params_p->in_params.use_send_events, 
                                   sq_sem_p)) {
            return RET_ERR;
        }
        ask_comp--;
        COND_PRINTF(("**** Polling for sr comp %d %d\n",i,ask_comp));
    }


    if(out_file != NULL){
        fclose(out_file);
    }
    sock_sync_msg(params_p->in_params.socket, is_sender);
    return(RET_OK);
}

