/*
  This software is available to you under a choice of one of two
  licenses.  You may choose to be licensed under the terms of the GNU
  General Public License (GPL) Version 2, available at
  <http://www.fsf.org/copyleft/gpl.html>, or the OpenIB.org BSD
  license, available in the LICENSE.TXT file accompanying this
  software.  These details are also available at
  <http://openib.org/license.html>.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
  BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
  ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
  CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
  SOFTWARE.

  Copyright (c) 2004 Mellanox Technologies Ltd.  All rights reserved.
*/

#ifndef CMD_PARS_H_INCLUDED
#define CMD_PARS_H_INCLUDED 

typedef enum {
    USAGE_ALL,
    USAGE_COSTOMER,
    USAGE_MINIMAL
} help_usage_levels_t;

typedef struct {
    int             is_sender;
    int             mtu;
    char            transport_type[10];  //ud, rc, 
    char            test_type[10];       //bw, lat
    char            rdma[10];            //read, write
    int             msg_size;
    int             iterations;      
    int             hca_port;
    int             app_id;
    char            remote_ip[100];
    int             use_events;
    int             num_ah;
    int             oneside;
    int             general_param;
    char            deviceID[100];
    int             use_devmem;
    int             atomic_mode;        //Atomic mode
    int             wqetest;
	int				ports_num;
    int             comp_inter;      
	int				bidirect;
    int             is_server;
    int             num_of_qps;
    //int             sched_period;
    int             postlist_len;
    int             qpc_sets;
    int             mpt_sets;
    int             num_of_keys;
    int             sg_list_len;
//    int             many2one;
    int             multicast;
    int             grh;
    int             rwrite_pollcq;
    int             srq_num;
} cmd_line_params_t;

#include <mtl_common.h>
#define CMD_LINE_PARAM_OFFSET(field) ((unsigned int)(MT_ulong_ptr_t)&(((cmd_line_params_t *) 0)->field))

//Atomic test modes
#define CMP_AND_SWAP_MODE   1
#define FETCH_AND_ADD_MODE  2



//-----------------------------------------------------------------------------
// Types Definition
//-----------------------------------------------------------------------------
typedef struct {
    char         char_option;       // One character option
    char*        str_option;        // String option
    char*        str_value;         // only a present/not present flag
    char*        help_desc;         // Help string
    char         param_type;        // 'p': present 's':string 'i':integer 'N'-just print
    unsigned int param_offset;      // Offset of parameter in cmd_line_params_t struct
    char         *default_value;
} usage_descriptor_t; 

#ifndef CMD_PARAMS_C_FILE
extern usage_descriptor_t usage_descriptor[]; 
extern char *help_vect_str[];
#else
char *help_vect_str[]={"send","size","iter","transport","test","id","rdma","ibp","s_ip", "device", 
	"atom_fa", "atom_cs", "ports_num", "bidirect", "events", "server", "devmem", "num_qp",
    "oneside", "mtu", "num_srq", "help", "extended",  
};
usage_descriptor_t usage_descriptor[]={

    {   
        ' ', "send", "",
        "Activate a Sender",
        'p',
        CMD_LINE_PARAM_OFFSET(is_sender),
        "0"
    },
    {
        's', "size", "SIZE",
        "Transfer message size in Bytes\n                         (Sender only, ignored for Atomic tests)",
        'i',
        CMD_LINE_PARAM_OFFSET(msg_size),
        "128"
    },
    {   
        'n', "iter", "NUM",
        "Number of iterations (Sender only)",
        'i',
        CMD_LINE_PARAM_OFFSET(iterations),
        "10"
    },

    {   
        't', "transport", "T_TYPE",
        "Transport type, T_TYPE = [\"rc\",\"ud\",\"uc\"] (Sender only)",
        's',
        CMD_LINE_PARAM_OFFSET(transport_type),
        "rc"
    },

    {   
        'm', "test", "MODE",
        "Test type TEST=[\"bw\",\"lat\"] (Sender only)",
        's',
        CMD_LINE_PARAM_OFFSET(test_type),
        "bw"
    },

    {   
        'r', "rdma", "MODE",
        "Optional - MODE=[\"write\",\"read\"] (Sender only)",
        's',
        CMD_LINE_PARAM_OFFSET(rdma),
        "none"
    },
    
    {   
        ' ', "device", "DeviceID",
        "InfiniHost0, InfiniHost_III_Ex0, InfiniHost_III_Lx0 ",
        's',
        CMD_LINE_PARAM_OFFSET(deviceID),
        "none"
    },
    {
        ' ', "atom_cs", "",
        "Atomic Compare & Swap test mode (Sender only)",
        'N',
        CMD_LINE_PARAM_OFFSET(atomic_mode),
        "none"
    },
    /*{
        ' ', "atom_fa", "",
        "Atomic Fetch & Add test mode (Sender only)",
        'N',
        CMD_LINE_PARAM_OFFSET(atomic_mode),
        "none"
    },*/
    {   
        'a', "s_ip", "IP_ADDRESS",
        "IP address of sender (receiver only)",
        's',
        CMD_LINE_PARAM_OFFSET(remote_ip),
        ""
    },
    {   
        'i', "id", "ID",
        //"Application ID",
        "Socket Number",
        'i',
        CMD_LINE_PARAM_OFFSET(app_id),
        "0"
    },
    {   
        'u', "mtu", "MTU",
        "MTU: 1->256, 2->512, 3->1024, 4->2048 (Default 1024)",
        'i',
        CMD_LINE_PARAM_OFFSET(mtu),
        "-1"
    },

    {   
        'p', "ibp", "PORT",
        "HCA port number (default is 1)",
        'i',
        CMD_LINE_PARAM_OFFSET(hca_port),
        "1"
    },
    {   
        'e', "events", "EVENTS",
        "Override the default for Poll CQ\n                         events=1 - with events; events=0 [default] - no events",
        'i',
        CMD_LINE_PARAM_OFFSET(use_events),
        "0"
    },
    {   
        'd', "devmem", "",
        "Register MR from the HCA DDR\n                         devmem=1 register from HCA DDR\n                         devmem=0 [default] from host RAM",
        'i',
        CMD_LINE_PARAM_OFFSET(use_devmem),
        "-1"
    },

    {   
        ' ', "oneside", "",
        "Don't use a receiver",
        'p',
        CMD_LINE_PARAM_OFFSET(oneside),
        "0"
    },
    {   
        'c', "ports_num", "NUM",
        "Send data via NUM ports (sender only)\n                         NUM=[1 (default), 2]",
        'i',
        CMD_LINE_PARAM_OFFSET(ports_num),
        "1"
    },
    {   
        'b', "bidirect", "NUM",
        "Send data in both directions, when NUM=1 (sender only)",
        'i',
        CMD_LINE_PARAM_OFFSET(bidirect),
        "0"
    },
    {   
        ' ', "server", "",
        "Activate a Server on the receiver side",
        'p',
        CMD_LINE_PARAM_OFFSET(is_server),
        "0"
    },
    {   
        'q', "num_qp", "NUM_QP",
        "Number of QPs (Default: NUM_QP=1)",
        'i',
        CMD_LINE_PARAM_OFFSET(num_of_qps),
        "1"
    },
    {   
        ' ', "num_srq", "NUM_SRQ",
        "Number of SRQ's (Shared Receive Queues).\n                         Default: 0 - No SRQ used",
        'i',
        CMD_LINE_PARAM_OFFSET(srq_num),
        "0"
    },
    {   
        'h', "help", "",
        "Prints detailed help",
        'N',
        CMD_LINE_PARAM_OFFSET(is_sender),
        "0"
    },
    {   
        ' ', "extended", "",
        "Prints detailed help, including extended options.\n                         Extended options are experimental and not QAed.",
         
        'N',
        CMD_LINE_PARAM_OFFSET(is_sender),
        "0"
    },
    //{   
    //    'x', "sch_per", "QP_SCHED",
    //    "QP Scheduling period (num. of QP messages sent contiguously)",
    //    'i',
    //    CMD_LINE_PARAM_OFFSET(sched_period),
    //    "10"
    //},
    {   
        'v', "num_ah", "NUM_AH",
        "Number of Address Handles",
        'i',
        CMD_LINE_PARAM_OFFSET(num_ah),
        "256"
    },
    {   
        'l', "postlist", "POST_LIST",
        "Length of Decriptor List posted each time",
        'i',
        CMD_LINE_PARAM_OFFSET(postlist_len),
        "1"
    },
    {   
        ' ', "qpc_sets", "QPC_SETS",
        "Number of QPC sets",
        'i',
        CMD_LINE_PARAM_OFFSET(qpc_sets),
        "64"
    },
    {   
        ' ', "mpt_sets", "MPT_SETS",
        "Number of MPT sets",
        'i',
        CMD_LINE_PARAM_OFFSET(mpt_sets),
        "128"
    },
    {   
        'k', "num_keys", "NUM_KEYS",
        "Number of keys used for memory access",
        'i',
        CMD_LINE_PARAM_OFFSET(num_of_keys),
        "1"
    },
    {   
        'g', "sg_list_len", "SG_LST",
        "Scatter-Gather list length. Affects Receive WQEs only",
        'i',
        CMD_LINE_PARAM_OFFSET(sg_list_len),
        "1"
    },
//    {   
//        ' ', "many2one", "",
//        "Temporary flag for special Many_2_One test",
//        'p',
//        CMD_LINE_PARAM_OFFSET(many2one),
//        "0"
//    },
    {   
        ' ', "multicast", "",
        "multicast - use only with UD",
        'p',
        CMD_LINE_PARAM_OFFSET(multicast),
        "0"
    },
    {   
        ' ', "grh", "",
        "GRH - use only with UD",
        'p',
        CMD_LINE_PARAM_OFFSET(grh),
        "0"
    },
    {   
        ' ', "rpcq", "",
        "Special test - RDMA Write latency from Post until Poll-CQ",
        'p',
        CMD_LINE_PARAM_OFFSET(rwrite_pollcq),
        "0"
    },
    {   
        'j', "Junk", "NUM",
        "Experimental general parameter (sender only)",
        'i',
        CMD_LINE_PARAM_OFFSET(general_param),
        "-1"
    },
    {
        ' ', "wqe", "",
        "wqe test (Sender only, oneside only)",
        'p',
        CMD_LINE_PARAM_OFFSET(wqetest),
        "0"
    },
    {   
        'n', "comp_inter", "NUM",
        "Completion interval (use with --wqe only) (default 256)",
        'i',
        CMD_LINE_PARAM_OFFSET(comp_inter),
        "256"
    },
};
#endif

#define OPTIONS_N	(sizeof(usage_descriptor)/sizeof(usage_descriptor_t))
int parse_argv( int my_argc, char** my_argv, cmd_line_params_t *cmd_line_params_p);
int init_cmd_line_params(cmd_line_params_t *cmd_line_params_p);
int print_cmd_line_params(cmd_line_params_t *cmd_line_params_p);
void usage(help_usage_levels_t help_level);

#endif /*CMD_PARS_H_INCLUDED*/





