/*
  This software is available to you under a choice of one of two
  licenses.  You may choose to be licensed under the terms of the GNU
  General Public License (GPL) Version 2, available at
  <http://www.fsf.org/copyleft/gpl.html>, or the OpenIB.org BSD
  license, available in the LICENSE.TXT file accompanying this
  software.  These details are also available at
  <http://openib.org/license.html>.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
  BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
  ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
  CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
  SOFTWARE.

  Copyright (c) 2004 Mellanox Technologies Ltd.  All rights reserved.
*/

#include <vapi.h>
#include <evapi.h>
#include <vapi_common.h>
#include <stdio.h>

//#define DEBUG_PRINT
#include "utils.h"
#include "all_bw_lat.h"
#include "sock.h"

#ifndef MTPERF
//#define MTPERF
#endif

#include <mtperf.h>

MTPERF_EXTERN_SEGMENT(VAPI_poll_cq);

static void MT_API async_events_handler( VAPI_hca_hndl_t hca_hndl, VAPI_event_record_t *event_p, void* priv_data);

#ifdef CQ_BLOCK
    #define  signal_comp_handler  EVAPI_POLL_CQ_UNBLOCK_HANDLER
#else
static void signal_comp_handler(VAPI_hca_hndl_t hca_hndl, VAPI_cq_hndl_t cq_hndl, void* sem_p)
{
    MOSAL_sem_rel((MOSAL_semaphore_t*) sem_p);
}
//static void signal_comp_handler(VAPI_hca_hndl_t hca_hndl, VAPI_cq_hndl_t cq_hndl, void* sem_p);
#endif

#ifdef CHK_QP_STATE
test_params_tx *g_params_p;
#endif


void sync_rtr2rts(test_params_tx *params);
static ret_val_t qp_init2rts_one_qp(test_params_tx *params,qp_entry_t *qp_e_p, int port_ix);
static ret_val_t qp_init2rts_one_qp_special(test_params_tx *params,qp_entry_t *qp_e_p, int port_ix);
static void set_data_mem(void *inbuff, unsigned long len);
int qp_num_comp(const void * qp_entry1, const void * qp_entry2);
void sort_qps(qp_entry_t *qp_vect, u_int32_t num_of_qps);
int qp_num_check(const void *qp_num, const void *qp_entry);

int qp_num_comp(const void * qp_entry1, const void * qp_entry2)
{
    return (int) (((qp_entry_t *)qp_entry1)->qp_num - ((qp_entry_t *)qp_entry2)->qp_num);
}

void sort_qps(qp_entry_t *qp_vect, u_int32_t num_of_qps)
{
    qsort(qp_vect, num_of_qps, sizeof(qp_entry_t), qp_num_comp);
}

int qp_num_check(const void *qp_num, const void *qp_entry)
{
    return (int) (*(VAPI_qp_num_t *)qp_num - ((qp_entry_t *)qp_entry)->qp_num);
}

qp_entry_t * search_qp(VAPI_qp_num_t qp_num, qp_entry_t *qp_vect, u_int32_t num_of_qps)
{
    return (qp_entry_t *) bsearch(&qp_num, qp_vect, num_of_qps, sizeof(qp_entry_t), qp_num_check);
}

ret_val_t qp_create(test_params_tx *params, int port_ix)
{
    int tmp=0;
    int add_one = 0;
    int jj=0;
    int srq_n=0;
    VAPI_qp_init_attr_t    qp_init_attr, aux_qp_init_attr;
    VAPI_qp_init_attr_ext_t qp_ext_attr;
    VAPI_qp_prop_t         qp_prop;
    VAPI_ret_t             res;
    
    VAPI_qp_hndl_t         qp_hndl =  INVAL_HNDL;
    
    VAPI_srq_attr_t         srq_props;
    VAPI_srq_attr_t         actual_srq_props;
    VAPI_srq_hndl_t         srq_hndl = VAPI_INVAL_SRQ_HNDL;

    qp_list_entry_t *      trash_qp_list = NULL;
    qp_list_entry_t *      trash_qp_entry;

    if (params->in_params.srq_num > 0)
    {
        // Allocate memory for SRQ vect
        params->port[port_ix].srq_vect = (srq_entry_t *) VMALLOC(params->in_params.srq_num * sizeof(srq_entry_t));
        memset( (void*)params->port[port_ix].srq_vect, 0, params->in_params.srq_num * sizeof(srq_entry_t) );
        //
        if ((params->in_params.num_of_qps % params->in_params.srq_num) != 0) add_one = 1;
        srq_props.max_outs_wr = params->in_params.num_r_wqe *
                                 ((params->in_params.num_of_qps / params->in_params.srq_num) + add_one);
        //printf("num_r_wqe %d max %d\n", params->in_params.num_r_wqe, srq_props.max_outs_wr);
        srq_props.max_sentries = params->in_params.num_rsge;
        srq_props.pd_hndl = params->pd_hndl;
        for (jj=0;jj<params->in_params.srq_num;jj++)
        {
            params->port[port_ix].srq_vect[jj].srq_hndl = VAPI_INVAL_SRQ_HNDL;
            res = VAPI_create_srq(params->hca_hndl, &srq_props, &srq_hndl, &actual_srq_props);
            if (res != VAPI_OK)
            {
                printf("Error1: Creating SRQ : %s\n", VAPI_strerror(res));
                params->port[port_ix].srq_vect[jj].srq_hndl = VAPI_INVAL_SRQ_HNDL;
                return(RET_ERR);
            }
            //printf("[*] SRQ handle "MT_ULONG_PTR_FMT" was created, with %u outstanding, %u sentries\n", srq_hndl, 
        //actual_srq_props.max_outs_wr, actual_srq_props.max_sentries);
            params->port[port_ix].srq_vect[jj].srq_hndl = srq_hndl;
        }
    }

    qp_init_attr.cap.max_oust_wr_rq = params->in_params.num_r_wqe;
    qp_init_attr.cap.max_oust_wr_sq = params->in_params.num_s_wqe;
    qp_init_attr.cap.max_sg_size_rq = params->in_params.num_rsge;
    qp_init_attr.cap.max_sg_size_sq = params->in_params.num_ssge;
    qp_init_attr.pd_hndl            = params->pd_hndl;
    qp_init_attr.rdd_hndl           = 0;
    qp_init_attr.sq_sig_type        = VAPI_SIGNAL_REQ_WR;
    qp_init_attr.rq_sig_type        = VAPI_SIGNAL_REQ_WR;
    if (params->in_params.test_mode == BW_SEND_BI ||
		params->in_params.test_mode == BW_RDMA_W_BI ||
		params->in_params.test_mode == BW_RDMA_R_BI )
    {
	    qp_init_attr.rq_cq_hndl     = params->port[port_ix].s_cq_hndl;
	    qp_init_attr.sq_cq_hndl     = params->port[port_ix].s_cq_hndl;
   	}
    else {
	    qp_init_attr.rq_cq_hndl     = params->port[port_ix].r_cq_hndl;
	    qp_init_attr.sq_cq_hndl     = params->port[port_ix].s_cq_hndl;
   	}
    qp_init_attr.ts_type            = params->in_params.transport;

    aux_qp_init_attr.cap.max_oust_wr_rq = 1;
    aux_qp_init_attr.cap.max_oust_wr_sq = 1;
    aux_qp_init_attr.cap.max_sg_size_rq = 1;
    aux_qp_init_attr.cap.max_sg_size_sq = 1;
    aux_qp_init_attr.pd_hndl            = params->pd_hndl;
    aux_qp_init_attr.rdd_hndl           = 0;
    aux_qp_init_attr.sq_sig_type        = VAPI_SIGNAL_REQ_WR;
    aux_qp_init_attr.rq_sig_type        = VAPI_SIGNAL_REQ_WR;
    if (params->in_params.test_mode == BW_SEND_BI ||
		params->in_params.test_mode == BW_RDMA_W_BI ||
		params->in_params.test_mode == BW_RDMA_R_BI )
    {
	    aux_qp_init_attr.rq_cq_hndl     = params->port[port_ix].s_cq_hndl;
	    aux_qp_init_attr.sq_cq_hndl     = params->port[port_ix].s_cq_hndl;
   	}
    else {
	    aux_qp_init_attr.rq_cq_hndl     = params->port[port_ix].r_cq_hndl;
	    aux_qp_init_attr.sq_cq_hndl     = params->port[port_ix].s_cq_hndl;
   	}
    aux_qp_init_attr.ts_type            = params->in_params.transport;

    //printf("QPs = %d\n", params->in_params.num_of_qps);

    for (jj=0;jj<(int)params->in_params.num_of_qps;jj++) {
        params->port[port_ix].qp_vect[jj].qp_hndl = INVAL_HNDL;
        
        if (params->in_params.srq_num > 0)
        {
            srq_n = jj % params->in_params.srq_num;
            //printf("jj = %d  srq_n = %d\n", jj, srq_n);
            params->port[port_ix].qp_vect[jj].srq_hndl = params->port[port_ix].srq_vect[srq_n].srq_hndl;
        }
        
        if (params->in_params.qpc_sets < ARBEL_QPC_SETS) {   
            // First create QP with minimal capabilities.
            // If qp_num matches constraints - re-create it with actual capabilities
            do {
                if (params->in_params.srq_num > 0)
                {
                    qp_ext_attr.srq_hndl = params->port[port_ix].qp_vect[jj].srq_hndl;
                    res = VAPI_create_qp_ext(params->hca_hndl, &aux_qp_init_attr, &qp_ext_attr, &qp_hndl, &qp_prop);
                    if (res != VAPI_OK) {
                        printf("Error1: Creating QP with SRQ: %s\n", VAPI_strerror(res));
                        return(RET_ERR);
                    }
                }
                else
                {
                    res = VAPI_create_qp(params->hca_hndl, &aux_qp_init_attr, &qp_hndl, &qp_prop);
                    if (res != VAPI_OK) {
                        printf("Error1: Creating QP : %s\n", VAPI_strerror(res));
                        return(RET_ERR);
                    }
                }
    
                //printf("Created QP 0x%x, modulo=%d, condition=%d\n", qp_prop.qp_num, 
                //       qp_prop.qp_num % ARBEL_QPC_SETS,
                //       (qp_prop.qp_num % ARBEL_QPC_SETS)/params->in_params.qpc_sets);

                if ((qp_prop.qp_num % ARBEL_QPC_SETS)/params->in_params.qpc_sets == 0) { 

                    //printf("Re-openning QP 0x%x\n", qp_prop.qp_num);

                    res = VAPI_destroy_qp(params->hca_hndl, qp_hndl);
                    if (res != VAPI_OK) {
                        printf("Error: Destroying QP : %s\n", VAPI_strerror(res));
                        return(RET_ERR);
                    }
        
                    if (params->in_params.srq_num > 0)
                    {
                        qp_ext_attr.srq_hndl = params->port[port_ix].qp_vect[jj].srq_hndl;
                        res = VAPI_create_qp_ext(params->hca_hndl, &qp_init_attr, &qp_ext_attr, &qp_hndl, &qp_prop);
                        if (res != VAPI_OK) {
                            printf("Error2: Creating QP with SRQ: %s\n", VAPI_strerror(res));
                            return(RET_ERR);
                        }
                    }
                    else
                    {
                        res = VAPI_create_qp(params->hca_hndl, &qp_init_attr, &qp_hndl, &qp_prop);
                        if (res != VAPI_OK) {
                            printf("Error2: Creating QP : %s\n", VAPI_strerror(res));
                            return(RET_ERR);
                        }
                    }

                    // Sanity check
                    if ((qp_prop.qp_num % ARBEL_QPC_SETS)/params->in_params.qpc_sets != 0) {
                        printf("Error: new-allocated QP does not match QPC set condition\n");
                    }

                    break;

                } else {
                    //printf("Adding QP 0x%x to trash_qp_list\n", qp_prop.qp_num);
                    // Allocate new trash_qp_entry and put it at the head of trash_qp_list
                    trash_qp_entry = VMALLOC(sizeof(qp_list_entry_t));
                    trash_qp_entry->qp_hndl = qp_hndl;
                    trash_qp_entry->next = trash_qp_list;
                    trash_qp_list = trash_qp_entry;
                }
        
            } while (1);//((qp_prop.qp_num % params->in_params.qpc_sets) != 0);
        
        } else {
            if (params->in_params.srq_num > 0)
            {
                qp_ext_attr.srq_hndl = params->port[port_ix].qp_vect[jj].srq_hndl;
                res = VAPI_create_qp_ext(params->hca_hndl, &qp_init_attr, &qp_ext_attr, &qp_hndl, &qp_prop);
                if (res != VAPI_OK) {
                    printf("Error3: Creating QP with SRQ: %s\n", VAPI_strerror(res));
                    return(RET_ERR);
                }
            }
            else
            {
                res = VAPI_create_qp(params->hca_hndl, &qp_init_attr, &qp_hndl, &qp_prop);
                if (res != VAPI_OK) {
                    printf("Error3: Creating QP : %s\n", VAPI_strerror(res));
                    return(RET_ERR);
                }
            }
        }

        //printf("Port %d created QP %d handle 0x%x\n", port_ix, qp_prop.qp_num, qp_hndl);
        params->port[port_ix].qp_vect[jj].qp_hndl = qp_hndl;
        params->port[port_ix].qp_vect[jj].qp_num = qp_prop.qp_num;
        if (jj == 0) {
            params->max_cap_inline_size = qp_prop.cap.max_inline_data_sq;
        }
        // Added for multicast:
        if (params->in_params.multicast > 0)
        {
            IB_gid_t gid;
            IB_lid_t mcg_lid;
            //printf("Adding multicast attributes to QP\n");
            if (params->in_params.is_sender)
            {
                //ASSIGN_MGID_0;
                mcg_lid = MIN_MC_LID;
                params->port[port_ix].qp_vect[jj].d_lid = MIN_MC_LID+1;
            }
            else  // receiver
            {
                //ASSIGN_MGID_1;
                mcg_lid = MIN_MC_LID+1;
                params->port[port_ix].qp_vect[jj].d_lid = MIN_MC_LID;
            }

            params->port[port_ix].qp_vect[jj].d_qpnum = IB_MULTICAST_QP;
            ASSIGN_MGID_0;
            VAPI_attach_to_multicast(params->hca_hndl, gid, qp_hndl, mcg_lid);
        } // end multicast
    }

    // Close all unused QPs
    while (trash_qp_list != NULL) {
        
        trash_qp_entry = trash_qp_list;

        res = VAPI_destroy_qp(params->hca_hndl, trash_qp_entry->qp_hndl);
        if (res != VAPI_OK) {
            printf("Error: Destroying QP : %s\n", VAPI_strerror(res));
            return(RET_ERR);
        }

        trash_qp_list = trash_qp_list->next;

        free(trash_qp_entry);

        tmp++;
    }

    // printf("Destroyed %d QPs\n", tmp);

    sort_qps(params->port[port_ix].qp_vect, params->in_params.num_of_qps);

    return(RET_OK);
}


void sync_rtr2rts(test_params_tx *params)
{
    /* make sure other side is in RTR before moving to RTS */
    char sock_msg[1000];
    int socket = params->in_params.socket;
    if (params->in_params.transport == IB_TS_UD || params->in_params.transport == IB_TS_UC 
        || params->in_params.oneside) {
        return;
    }
    if (params->in_params.is_sender) {
        while (1) {
            COND_PRINTF2(("Waiting to get RTR........"));
            sock_get_msg(socket, sock_msg, 1000);
            if (sock_is_exists(sock_msg,"RTR")) {
                COND_PRINTF2(("Got RTR........"));
                break;
            }
        }
        COND_PRINTF2(("Sending RTR\n"));
        sock_send_msg(socket, "RTR");
    }
    else {
        COND_PRINTF2(("Sending RTR........"));
        sock_send_msg(socket, "RTR");
        while (1) {
            COND_PRINTF2(("Waiting to get RTR........"));
            sock_get_msg(socket, sock_msg, 1000);
            if (sock_is_exists(sock_msg,"RTR")) {
                COND_PRINTF2(("Got RTR\n"));
                break;
            }
        }
    }
}



ret_val_t create_common_resources(test_params_tx *params)   
{
    VAPI_ret_t             res;
    VAPI_hca_hndl_t        hca_hndl;
    VAPI_hca_port_t        hca_port;
    int act_num_cqe;
    VAPI_mrw_t             mr_in, mr_out;
    VAPI_mrw_t             tmp_mr_in;
    u_int8_t               *send_buf=NULL;
    u_int8_t               *recv_buf=NULL;
    u_int8_t               *real_send_buf = NULL;
    u_int8_t               *real_recv_buf = NULL;
    
    u_int8_t               align_shift;
    VAPI_phy_addr_t        buf_p;
    void                   *virt_addr_p=NULL;
    VAPI_phy_buf_t         phy_buf_entry;

    VAPI_pd_hndl_t         pd_hndl   =  INVAL_HNDL; 
    VAPI_cq_hndl_t         s_cq_hndl =  INVAL_HNDL;
    VAPI_cq_hndl_t         r_cq_hndl =  INVAL_HNDL;
    VAPI_mr_hndl_t         mr_hndl =  INVAL_HNDL;
    VAPI_devmem_hndl_t     dm_hndl =  INVAL_HNDL;

    int       mr_send_size = params->in_params.mr_send_size;
    int       mr_recv_size = params->in_params.mr_recv_size;
    u_int32_t num_cqe      = params->in_params.num_cqe;
    EVAPI_devmem_info_t     devmem_info;
    unsigned int page_size;
	int i, jj;
    int tmp;
    int                    is_sender   = params->in_params.is_sender;

    mr_list_entry_t *      trash_mr_list = NULL;
    mr_list_entry_t *      trash_mr_entry;

#ifdef CHK_QP_STATE
	g_params_p = params;
#endif

/********************** HCA handle **********************/
    res = EVAPI_get_hca_hndl(params->in_params.hca_id, &hca_hndl); 
    if (res != VAPI_OK) {
        printf("Error: Opening HCA : %s\n", VAPI_strerror(res));
        return(RET_ERR);
    }
    params->hca_hndl = hca_hndl;

/********************** get some configuration info **********************/
	res = VAPI_query_hca_cap( hca_hndl, &params->hca_vendor, &params->hca_cap);
    if (res != VAPI_OK) {
        printf("Error on VAPI_query_hca_cap : %s\n", VAPI_strerror(res));
        return(RET_ERR);
    }
    
/********************** PD **********************/
    res = EVAPI_alloc_pd(hca_hndl, params->in_params.num_ah, &pd_hndl);
    if (res != VAPI_OK) {
        printf("Error: Allocating PD : %s\n", VAPI_strerror(res));
        return(RET_ERR);
    }
	//printf("Allocated PD with %d AHs\n", params->in_params.num_ah);
    params->pd_hndl = pd_hndl;

/********************** Port Dependent Resources **********************/

	FOR_ALL_PORTS( params, i) 
	
		/********************** get port properties  **********************/
		params->port[i].port_num = (IB_port_t)(i + 1);
	    res = VAPI_query_hca_port_prop(hca_hndl,(IB_port_t)params->port[i].port_num,( VAPI_hca_port_t *)&hca_port);
	    if (res != VAPI_OK) {
	        printf("Error on VAPI_query_hca_port_prop: %s\n", VAPI_strerror(res));
	        return(RET_ERR);
	    }
	    params->port[i].lid = hca_port.lid;
	    
		/********************** allocate memory for QPs  **********************/
	    params->port[i].qp_vect = NULL;
	    if (params->in_params.num_of_qps > 0) {
	        params->port[i].qp_vect = (qp_entry_t *) VMALLOC(params->in_params.num_of_qps * sizeof(qp_entry_t));
	        memset( (void*)params->port[i].qp_vect, 0, params->in_params.num_of_qps * sizeof(qp_entry_t) );
	    }

	    /********************** Send CQ **********************/
		res = VAPI_create_cq(hca_hndl, num_cqe, &s_cq_hndl, &act_num_cqe);
		//res = VAPI_create_cq(hca_hndl, params->in_params.num_of_qps*num_cqe, &s_cq_hndl, &act_num_cqe);
		if (res != VAPI_OK) {
		    printf("Error: Creating CQ : %s\n", VAPI_strerror(res));
		    return(RET_ERR);
		}
		params->port[i].s_cq_hndl = s_cq_hndl;

		if (params->in_params.use_send_events) {

#ifndef CQ_BLOCK
		    MOSAL_sem_init(&(params->port[i].sq_sem),0);
#endif
		    res= EVAPI_set_comp_eventh(hca_hndl,s_cq_hndl,signal_comp_handler, &(params->port[i].sq_sem),&(params->port[i].sq_cq_eventh));
		    if (res != VAPI_OK) {
		        printf("Error: Creating in EVAPI_set_comp_eventh (send) : %s\n", VAPI_strerror(res));
		        return(RET_ERR);
		    }
		}
    
		/********************** RECV CQ **********************/
//		res = VAPI_create_cq(hca_hndl, params->in_params.num_of_qps*num_cqe, &r_cq_hndl, &act_num_cqe);
		res = VAPI_create_cq(hca_hndl, num_cqe, &r_cq_hndl, &act_num_cqe);
        //printf("act_num_cqe = %d\n", act_num_cqe);
		if (res != VAPI_OK) {
		    printf("Error: Creating CQ : %s\n", VAPI_strerror(res));
		    return(RET_ERR);
		}
		params->port[i].r_cq_hndl = r_cq_hndl;

		if (params->in_params.use_recv_events) {

#ifndef CQ_BLOCK
		    MOSAL_sem_init(&(params->port[i].rq_sem),0);
#endif        
		    res= EVAPI_set_comp_eventh(hca_hndl,r_cq_hndl, EVAPI_POLL_CQ_UNBLOCK_HANDLER, NULL, &(params->port[i].rq_cq_eventh));
		    if (res != VAPI_OK) {
		        printf("Error: Creating in EVAPI_set_comp_eventh (send) : %s\n", VAPI_strerror(res));
		        return(RET_ERR);
		    }
		}
		
	END_FOR

		/********************** async event handler **********************/
    res= EVAPI_set_async_event_handler(hca_hndl,async_events_handler,params,&(params->async_hndl));
    if (res != VAPI_OK) {
        printf("Error: Creating in EVAPI_set_comp_eventh (send) : %s\n", VAPI_strerror(res));
        return(RET_ERR);
    }


    /********************** SEND MR **********************/
    // if devmem should be used
    if (!params->in_params.oneside)     
        sock_freeze(params->in_params.socket, is_sender);

    if (params->in_params.use_devmem > 0) {
        if (params->in_params.alloc_aligned)
        {
            align_shift = (u_int8_t)(MOSAL_SYS_PAGE_SHIFT);
        } else {
            align_shift = 0;
        }

        res = EVAPI_query_devmem(hca_hndl, EVAPI_DEVMEM_EXT_DRAM, align_shift, &devmem_info);
        if (res != VAPI_OK) {
            printf("Error: Query DevMem failed : %s\n", VAPI_strerror(res));
            return(RET_ERR);
        }

		/* request memory with adjusting in case of a lack */
		for (i=MR_SND_CNT; i > 0; --i, mr_send_size = i * MR_UNIT) {
	        if (params->hca_cap.max_mr_size >= mr_send_size && devmem_info.largest_chunk >= mr_send_size) 
	        	break;
		}
        if (i <= 0)
        {
            printf("Error: not enough memory in HCA DDR: Larget available chunk: " U64_FMT, devmem_info.largest_chunk);
            printf(" bytes,\nrequested chunk: 0x%x bytes (align shift : 0x%x)\n", MR_UNIT, align_shift);
            return(RET_ERR);
        }
        else {
	        params->in_params.mr_send_size = mr_send_size;
	        if (i<MR_RCV_CNT) {
	        	printf("DDR send MR adjusted size is 0x%x (first requested was 0x%x, max chunk 0x%x)\n",  
	        		mr_send_size, MR_SND_CNT * MR_UNIT, (unsigned)devmem_info.largest_chunk);
        	}
      	}

        res = EVAPI_alloc_map_devmem(hca_hndl, EVAPI_DEVMEM_EXT_DRAM, mr_send_size, align_shift, &buf_p, &virt_addr_p, &dm_hndl);
        if (res != VAPI_OK) {
            printf("Error: Allocating %u bytes from DevMem for send side: %s\n", mr_send_size, VAPI_strerror(res));
            return(RET_ERR);
        }

        params->dm_hndl_s = dm_hndl;
        real_send_buf = virt_addr_p;
        send_buf = virt_addr_p;

        mr_in.start = (VAPI_virt_addr_t)(MT_virt_addr_t)virt_addr_p;
        mr_in.type = VAPI_MR;
        if ( MOSAL_get_page_size(MY_PROT_CTX, (MT_virt_addr_t)virt_addr_p, &page_size) != MT_OK ) {
          return RET_ERR;
        }
        mr_in.iova_offset = (MT_virt_addr_t)virt_addr_p & (page_size-1);
        phy_buf_entry.size = mr_send_size;
        phy_buf_entry.start = buf_p;
        mr_in.pbuf_list_p = &phy_buf_entry;
        mr_in.pbuf_list_len = 1;
    } else
    {
        if (params->in_params.alloc_aligned) {
            real_send_buf = VMALLOC(mr_send_size+PAGE_SIZE_ALIGN);

            send_buf = (void *) MT_UP_ALIGNX_VIRT(real_send_buf, LOG_PAGE_SIZE_ALIGN);
            COND_PRINTF3(("real=%p after=%p\n",real_send_buf,send_buf));
        }
        else {
            real_send_buf = send_buf = VMALLOC(mr_send_size);
        }
        if (send_buf == NULL) {
            printf("Try again..alloc memory failed\n");
            return(RET_ERR);
        }

        mr_in.start = (VAPI_virt_addr_t)(MT_virt_addr_t)send_buf;
        mr_in.type = VAPI_MR;
    }

    memset(send_buf,83,mr_send_size);
    set_data_mem(send_buf, mr_send_size);

    params->send_buf = send_buf;
    params->real_send_buf = real_send_buf;

    mr_in.acl =  VAPI_EN_LOCAL_WRITE | VAPI_EN_REMOTE_WRITE |  VAPI_EN_REMOTE_READ | VAPI_EN_REMOTE_ATOM; //Noam + Vladz
    if ((params->in_params.transport != IB_TS_UD) &&  params->in_params.use_mwindow) {
        mr_in.acl = mr_in.acl |  VAPI_EN_MEMREG_BIND;
    }
    mr_in.l_key = 0;
    mr_in.pd_hndl = pd_hndl;
    mr_in.r_key = 0;
    mr_in.size = mr_send_size;

    if (params->in_params.mpt_sets < ARBEL_MPT_SETS) {   
        
        // Initializing tmp_mr_in - for artificial MPT cache stress
        tmp_mr_in.acl           = mr_in.acl;
        tmp_mr_in.iova_offset   = mr_in.iova_offset;
        tmp_mr_in.l_key         = mr_in.l_key;
        tmp_mr_in.pbuf_list_len = mr_in.pbuf_list_len;
        tmp_mr_in.pbuf_list_p   = mr_in.pbuf_list_p;
        tmp_mr_in.pd_hndl       = mr_in.pd_hndl;
        tmp_mr_in.r_key         = mr_in.r_key;
        tmp_mr_in.start         = mr_in.start;
        tmp_mr_in.type          = mr_in.type;
        tmp_mr_in.size          = 1;
    }

    for (jj=0;jj<(int)params->in_params.num_of_keys;jj++) {

        if (params->in_params.mpt_sets < ARBEL_MPT_SETS) {   

            // First register MR with minimal capabilities.
            // If l_key matches constraints - re-create it with actual capabilities
            do {
                res = VAPI_register_mr(hca_hndl, &tmp_mr_in, &mr_hndl, &mr_out);
                if (res != VAPI_OK) {
                    printf("Error: Registering MR : %s\n", VAPI_strerror(res));
                    return(RET_ERR);
                }
    
                if ((mr_out.l_key % ARBEL_MPT_SETS)/params->in_params.mpt_sets == 0) {
    
                    res = VAPI_deregister_mr(hca_hndl,mr_hndl);
                    if (res != VAPI_OK) {
                        printf("Error: Deregistering MR : %s\n", VAPI_strerror(res));
                        return(RET_ERR);
                    }
    
                    res = VAPI_register_mr(hca_hndl, &mr_in, &mr_hndl, &mr_out);
                    if (res != VAPI_OK) {
                        printf("Error: Registering MR : %s\n", VAPI_strerror(res));
                        return(RET_ERR);
                    }
    
                    // Sanity check
                    if ((mr_out.l_key % ARBEL_MPT_SETS)/params->in_params.mpt_sets != 0) {
                        printf("Warning: new-allocated MR's l_key does not match MPT set condition\n");
                        res = VAPI_deregister_mr(hca_hndl,mr_hndl);
                        if (res != VAPI_OK) {
                            printf("Error: Deregistering MR : %s\n", VAPI_strerror(res));
                            return(RET_ERR);
                        }
                    }
                    else
                        break;
    
                } else {
                    // Allocate new trash_mr_entry and put it at the head of trash_mr_list
                    trash_mr_entry = VMALLOC(sizeof(mr_list_entry_t));
                    trash_mr_entry->mr_hndl = mr_hndl;
                    trash_mr_entry->next = trash_mr_list;
                    trash_mr_list = trash_mr_entry;
                }
    
            } while (1);
        
        } else {    
        
            res = VAPI_register_mr(hca_hndl, &mr_in, &mr_hndl, &mr_out);
            if (res != VAPI_OK) {
                printf("Error: Registering MR : %s\n", VAPI_strerror(res));
                return(RET_ERR);
            }
        }
            
        params->send_lkey_arr[jj] = mr_out.l_key;
        params->mr_hndl_s_arr[jj] = mr_hndl;           
    }

    tmp = 0;
    // Free all unused MRs
    while (trash_mr_list != NULL) {      
        trash_mr_entry = trash_mr_list;
        res = VAPI_deregister_mr(hca_hndl, trash_mr_entry->mr_hndl);
        if (res != VAPI_OK) {
            printf("Error: Deregistering MR : %s\n", VAPI_strerror(res));
            return(RET_ERR);
        }
        trash_mr_list = trash_mr_list->next;
        free(trash_mr_entry);
        tmp++;
    }

    // printf("Deregistered %d Send MRs\n", tmp);

    params->send_lkey = params->send_lkey_arr[0];
    params->mr_hndl_s = params->mr_hndl_s_arr[0];

/**********************  RECV MR  **********************/
    
    // if devmem should be used
    if (params->in_params.use_devmem > 0) {
        if (params->in_params.alloc_aligned)
        {
            align_shift = (u_int8_t)(MOSAL_SYS_PAGE_SHIFT);
        } else {
            align_shift = 0;
        }

        res = EVAPI_query_devmem(hca_hndl, EVAPI_DEVMEM_EXT_DRAM, align_shift, &devmem_info);
        if (res != VAPI_OK) {
            printf("Error: Query DevMem failed : %s\n", VAPI_strerror(res));
            return(RET_ERR);
        }

		/* request memory with adjusting in case of a lack */
		for (i=MR_RCV_CNT; i > 0; --i, mr_recv_size = i * MR_UNIT) {
	        if (params->hca_cap.max_mr_size >= mr_recv_size && devmem_info.largest_chunk >= mr_recv_size) 
	        	break;
		}
        if (i <= 0)
        {
            printf("Error: not enough memory in HCA DDR: Largest available chunk: " U64_FMT, devmem_info.largest_chunk);
            printf(" bytes,\nrequested chunk: 0x%x bytes (align shift : 0x%x)\n", MR_UNIT, align_shift);
            return(RET_ERR);
        }
        else {
	        params->in_params.mr_recv_size = mr_recv_size;
	        if (i<MR_RCV_CNT) {
	        	printf("DDR recv MR adjusted size is 0x%x (first requested was 0x%x, max chunk is 0x%X)\n",  
	        		mr_recv_size, MR_RCV_CNT * MR_UNIT, (unsigned)devmem_info.largest_chunk);
        	}
      	}

        res = EVAPI_alloc_map_devmem(hca_hndl, EVAPI_DEVMEM_EXT_DRAM, mr_recv_size, align_shift, &buf_p, &virt_addr_p, &dm_hndl);
        if (res != VAPI_OK) {
            printf("Error: Allocating %u bytes from DevMem for recv side : %s\n", mr_recv_size, VAPI_strerror(res));
            return(RET_ERR);
        }

        params->dm_hndl_r = dm_hndl;
        real_recv_buf = virt_addr_p;
        recv_buf = virt_addr_p;

        mr_in.start = (VAPI_virt_addr_t)(MT_virt_addr_t)virt_addr_p;
        mr_in.type = VAPI_MR;
        if ( MOSAL_get_page_size(MY_PROT_CTX, (MT_virt_addr_t)virt_addr_p, &page_size) != MT_OK ) {
          return RET_ERR;
        }
        mr_in.iova_offset = (MT_virt_addr_t)virt_addr_p & (page_size-1);
        phy_buf_entry.size = mr_recv_size;
        phy_buf_entry.start = buf_p;
        mr_in.pbuf_list_p = &phy_buf_entry;
        mr_in.pbuf_list_len = 1;
    } else
    {
        if (params->in_params.alloc_aligned) {
            real_recv_buf = VMALLOC(mr_recv_size+PAGE_SIZE_ALIGN);

            recv_buf = (void *) MT_UP_ALIGNX_VIRT(real_recv_buf, LOG_PAGE_SIZE_ALIGN);
           COND_PRINTF(("real=%p after=%p\n",real_recv_buf,recv_buf));
        }
        else {
            real_recv_buf = recv_buf = VMALLOC(mr_recv_size);
        }
        if (recv_buf == NULL) {
            printf("Try again..alloc memory failed\n");
            return(RET_ERR);
        }

        mr_in.start = (VAPI_virt_addr_t)(MT_virt_addr_t)recv_buf;
        mr_in.type = VAPI_MR;
    }

    params->recv_buf = recv_buf;
    params->real_recv_buf = real_recv_buf;
    memset(recv_buf,84,mr_recv_size);

    mr_in.acl =  VAPI_EN_LOCAL_WRITE | VAPI_EN_REMOTE_WRITE | VAPI_EN_REMOTE_READ | VAPI_EN_REMOTE_ATOM; //Noam + Vladz
    if ((params->in_params.transport != IB_TS_UD) &&  params->in_params.use_mwindow) {
        mr_in.acl = mr_in.acl |  VAPI_EN_MEMREG_BIND;
    }
    mr_in.l_key =0;
    mr_in.pd_hndl = pd_hndl;
    mr_in.r_key = 0;
    mr_in.size = mr_recv_size;       

    if (params->in_params.mpt_sets < ARBEL_MPT_SETS) {   
        
        // Initializing tmp_mr_in - for artificial MPT cache stress
        tmp_mr_in.acl           = mr_in.acl;
        tmp_mr_in.iova_offset   = mr_in.iova_offset;
        tmp_mr_in.l_key         = mr_in.l_key;
        tmp_mr_in.pbuf_list_len = mr_in.pbuf_list_len;
        tmp_mr_in.pbuf_list_p   = mr_in.pbuf_list_p;
        tmp_mr_in.pd_hndl       = mr_in.pd_hndl;
        tmp_mr_in.r_key         = mr_in.r_key;
        tmp_mr_in.start         = mr_in.start;
        tmp_mr_in.type          = mr_in.type;
        tmp_mr_in.size          = 1;
    }

    for (jj=0;jj<(int)params->in_params.num_of_keys;jj++) {

        if (params->in_params.mpt_sets < ARBEL_MPT_SETS) {   

            // First register MR with minimal capabilities.
            // If l_key matches constraints - re-create it with actual capabilities
            do {
                res = VAPI_register_mr(hca_hndl, &tmp_mr_in, &mr_hndl, &mr_out);
                if (res != VAPI_OK) {
                    printf("Error: Registering MR : %s\n", VAPI_strerror(res));
                    return(RET_ERR);
                }
    
                if ((mr_out.l_key % ARBEL_MPT_SETS)/params->in_params.mpt_sets == 0) {
    
                    res = VAPI_deregister_mr(hca_hndl,mr_hndl);
                    if (res != VAPI_OK) {
                        printf("Error: Deregistering MR : %s\n", VAPI_strerror(res));
                        return(RET_ERR);
                    }
    
                    res = VAPI_register_mr(hca_hndl, &mr_in, &mr_hndl, &mr_out);
                    if (res != VAPI_OK) {
                        printf("Error: Registering MR : %s\n", VAPI_strerror(res));
                        return(RET_ERR);
                    }
    
                    // Sanity check
                    if ((mr_out.l_key % ARBEL_MPT_SETS)/params->in_params.mpt_sets != 0) {
                        printf("Warning: new-allocated MR's l_key does not match MPT set condition\n");
                        res = VAPI_deregister_mr(hca_hndl,mr_hndl);
                        if (res != VAPI_OK) {
                            printf("Error: Deregistering MR : %s\n", VAPI_strerror(res));
                            return(RET_ERR);
                        }
                    }
                    else
                        break;
    
                } else {
                    // Allocate new trash_mr_entry and put it at the head of trash_mr_list
                    trash_mr_entry = VMALLOC(sizeof(mr_list_entry_t));
                    trash_mr_entry->mr_hndl = mr_hndl;
                    trash_mr_entry->next = trash_mr_list;
                    trash_mr_list = trash_mr_entry;
                }
    
            } while (1);
        
        } else {    
        
            res = VAPI_register_mr(hca_hndl, &mr_in, &mr_hndl, &mr_out);
            if (res != VAPI_OK) {
                printf("Error: Registering MR : %s\n", VAPI_strerror(res));
                return(RET_ERR);
            }
        }
        
        params->recv_lkey_arr[jj] = mr_out.l_key;
        params->recv_rkey_arr[jj] = mr_out.r_key;
        params->mr_hndl_r_arr[jj] = mr_hndl;            
    }
    
    tmp = 0;
    // Free all unused MRs
    while (trash_mr_list != NULL) {      
        trash_mr_entry = trash_mr_list;
        res = VAPI_deregister_mr(hca_hndl, trash_mr_entry->mr_hndl);
        if (res != VAPI_OK) {
            printf("Error: Deregistering MR : %s\n", VAPI_strerror(res));
            return(RET_ERR);
        }
        trash_mr_list = trash_mr_list->next;
        free(trash_mr_entry);
        tmp++;
    }

    // printf("Deregistered %d Receive MRs\n", tmp);

    params->recv_lkey = params->recv_lkey_arr[0];
    params->recv_rkey = params->recv_rkey_arr[0];
    params->mr_hndl_r = params->mr_hndl_r_arr[0];     

    // printf("Receive R_Key = 0x%x\n", params->recv_rkey);

    if (!params->in_params.oneside)     
        sock_unfreeze(params->in_params.socket, is_sender);

    COND_PRINTF(("registered mem\n"));

    return(RET_OK);
}

ret_val_t create_av(test_params_tx *params, int port_ix)   
{
    VAPI_ud_av_t           av;
    VAPI_ret_t             res;
    int i;

    if (params->in_params.multicast > 0)
    {
        IB_gid_t gid;
        
        if (params->in_params.is_sender)
        {
            //ASSIGN_MGID_1;
            av.dlid = MIN_MC_LID+1;
        }
        else // receiver
        {
            //ASSIGN_MGID_0;
            av.dlid = MIN_MC_LID;
        }

        ASSIGN_MGID_0;
        for (i=0;i<16;i++) av.dgid[i] = gid[i];
        av.grh_flag = TRUE;
    }
    else
    {
        av.dgid[0] = 0;
        if (params->in_params.grh) {
            av.grh_flag = TRUE;
        }
        else {
            av.grh_flag = FALSE;
        }
        av.dlid = params->port[port_ix].d_lid;
    }
    av.sl = 0;
    av.src_path_bits = 0;
    av.static_rate = 0;
    av.traffic_class = 0;
    av.port = (IB_port_t)params->port[port_ix].port_num;
    res = VAPI_create_addr_hndl(params->hca_hndl, params->pd_hndl, &av, &(params->port[port_ix].av_hndl));

    if (res != VAPI_OK) {
        printf("Error: Creating address vector : %s\n", VAPI_strerror(res));
        return(RET_ERR);
    }
    COND_PRINTF(("Created addr vector\n")); 
    return(RET_OK);
}

   
ret_val_t qp_init2rts_mqp(test_params_tx *params, int port_ix)
{
    int jj=0;
    ret_val_t ret;
    
    if (params->in_params.test_mode == SPECIAL_2PORT)
    {
        ret = qp_init2rts_one_qp(params,&(params->port[port_ix].qp_vect[0]), port_ix);
        if (ret != RET_OK) {
            return ret;
        }
        ret = qp_init2rts_one_qp_special(params,&(params->port[port_ix].qp_vect[1]), port_ix);
        if (ret != RET_OK) {
            return ret;
        }
        return(RET_OK);
    }

    for (jj=0;jj<(int)params->in_params.num_of_qps;jj++) {
        ret = qp_init2rts_one_qp(params,&(params->port[port_ix].qp_vect[jj]), port_ix);
        if (ret != RET_OK) {
            return ret;
        }
    }
    return(RET_OK);
}

static ret_val_t qp_init2rts_one_qp(test_params_tx *params,qp_entry_t *qp_e_p, int port_ix)
{
    VAPI_qp_attr_mask_t    qp_attr_mask;
    VAPI_qp_attr_t         qp_attr;
    VAPI_qp_cap_t          qp_cap;
    VAPI_ret_t             res;

    // printf("Initializing QP #%d, connected to QP #%d\n", qp_e_p->qp_num, qp_e_p->d_qpnum);

    VAPI_hca_hndl_t hca_hndl  = params->hca_hndl;
    VAPI_qp_hndl_t  qp_hndl   = qp_e_p->qp_hndl;
    int             d_qpnum   = qp_e_p->d_qpnum;
    int             dlid      = params->port[port_ix].d_lid; /*qp_e_p->d_lid;*/
    int             port_num  = params->port[port_ix].port_num;
    IB_mtu_t        mtu       = params->in_params.mtu;

    /***************************************************************************/

    /********************************* INIT ******************************************/
    QP_ATTR_MASK_CLR_ALL(qp_attr_mask);

    qp_attr.qp_state = VAPI_INIT;
    QP_ATTR_MASK_SET(qp_attr_mask,QP_ATTR_QP_STATE);

    qp_attr.pkey_ix  = 0;
    QP_ATTR_MASK_SET(qp_attr_mask,QP_ATTR_PKEY_IX);

    qp_attr.port     = (IB_port_t)port_num;
    QP_ATTR_MASK_SET(qp_attr_mask,QP_ATTR_PORT);

    if (params->in_params.transport == IB_TS_RC) {
        qp_attr.remote_atomic_flags = VAPI_EN_REM_WRITE | VAPI_EN_REM_READ | VAPI_EN_REM_ATOMIC_OP;
        QP_ATTR_MASK_SET(qp_attr_mask,QP_ATTR_REMOTE_ATOMIC_FLAGS);
    }
    else if (params->in_params.transport == IB_TS_UC) {
        qp_attr.remote_atomic_flags = VAPI_EN_REM_WRITE;
        QP_ATTR_MASK_SET(qp_attr_mask,QP_ATTR_REMOTE_ATOMIC_FLAGS);
    }
    else {
        qp_attr.qkey = 0; 
        QP_ATTR_MASK_SET(qp_attr_mask,QP_ATTR_QKEY);
    }


    res = VAPI_modify_qp(hca_hndl, qp_hndl, &qp_attr, &qp_attr_mask, &qp_cap);
    if (res != VAPI_OK) {
        printf("Error: Modifying  QP to INIT: %s\n", VAPI_strerror(res));
        return(RET_ERR);
    }
    COND_PRINTF(("Modified QP to INIT\n"));

    /**************************** RTR ***********************************************/   
    QP_ATTR_MASK_CLR_ALL(qp_attr_mask);

    qp_attr.qp_state = VAPI_RTR;
    QP_ATTR_MASK_SET(qp_attr_mask,QP_ATTR_QP_STATE);


    if (params->in_params.transport == IB_TS_RC) {
        if ((params->in_params.test_mode == BW_ATOM_CS) ||
            (params->in_params.test_mode == BW_LAT_ATOM_CS) ||
            (params->in_params.test_mode == LAT_ATOM_CS) // ||
    //        //(params->in_params.test_mode == LAT_RDMA_R) ||
    //        //(params->in_params.test_mode == BW_RDMA_R) ||
    //        //(params->in_params.test_mode == BW_RDMA_R_UNI)
            )
        {
            qp_attr.qp_ous_rd_atom  = params->hca_cap.max_qp_ous_rd_atom;
    //        qp_attr.qp_ous_rd_atom = 128;
        }
        else {
            qp_attr.qp_ous_rd_atom  = MIN2(DFLT_QP_OUS_RD_ATOM, params->hca_cap.max_qp_ous_rd_atom);
        }
        //printf("ous_rd_atom = %d\n", qp_attr.qp_ous_rd_atom);
        //printf("max_ous_rd_atom = %d\n", params->hca_cap.max_qp_ous_rd_atom);

        QP_ATTR_MASK_SET(qp_attr_mask,QP_ATTR_QP_OUS_RD_ATOM);

        qp_attr.dest_qp_num = d_qpnum; 
        QP_ATTR_MASK_SET(qp_attr_mask, QP_ATTR_DEST_QP_NUM);

        qp_attr.av.sl            = 0;
        qp_attr.av.grh_flag      = FALSE;
        qp_attr.av.dlid          = (IB_lid_t)dlid;
        qp_attr.av.static_rate   = 0;
        qp_attr.av.src_path_bits = 0;
        QP_ATTR_MASK_SET(qp_attr_mask,QP_ATTR_AV);

        qp_attr.path_mtu         = mtu;
        QP_ATTR_MASK_SET(qp_attr_mask,QP_ATTR_PATH_MTU);

        qp_attr.rq_psn           = 0;
        QP_ATTR_MASK_SET(qp_attr_mask,QP_ATTR_RQ_PSN);

        qp_attr.pkey_ix          = 0;
        QP_ATTR_MASK_SET(qp_attr_mask,QP_ATTR_PKEY_IX);
                                
        qp_attr.min_rnr_timer    = 0;
        QP_ATTR_MASK_SET(qp_attr_mask,QP_ATTR_MIN_RNR_TIMER);
    }
    else if (params->in_params.transport == IB_TS_UC) {

        qp_attr.av.sl            = 0;
        qp_attr.av.grh_flag      = FALSE;
        qp_attr.av.dlid          = (IB_lid_t)dlid;
        qp_attr.av.static_rate   = 0;
        qp_attr.av.src_path_bits = 0;
        QP_ATTR_MASK_SET(qp_attr_mask,QP_ATTR_AV);

        qp_attr.dest_qp_num = d_qpnum; 
        QP_ATTR_MASK_SET(qp_attr_mask, QP_ATTR_DEST_QP_NUM);

        qp_attr.path_mtu         = mtu;
        QP_ATTR_MASK_SET(qp_attr_mask,QP_ATTR_PATH_MTU);

        qp_attr.rq_psn           = 0;
        QP_ATTR_MASK_SET(qp_attr_mask,QP_ATTR_RQ_PSN);

        qp_attr.pkey_ix = 0;
        QP_ATTR_MASK_SET(qp_attr_mask,QP_ATTR_PKEY_IX);

    }


    res = VAPI_modify_qp(hca_hndl, qp_hndl, &qp_attr, &qp_attr_mask, &qp_cap);
    if (res != VAPI_OK) {
        printf("Error: Modifying  QP to RTR: %s\n", VAPI_strerror(res));      
        return(RET_ERR);
    }
    COND_PRINTF(("Moodified QP to RTR\n"));


    sync_rtr2rts(params);

    /***************************** RTS **********************************************/   

    QP_ATTR_MASK_CLR_ALL(qp_attr_mask);

    qp_attr.qp_state   = VAPI_RTS;
    QP_ATTR_MASK_SET(qp_attr_mask,QP_ATTR_QP_STATE);

    qp_attr.sq_psn   = 0;
    QP_ATTR_MASK_SET(qp_attr_mask,QP_ATTR_SQ_PSN);

    if (params->in_params.transport == IB_TS_RC) {

        qp_attr.timeout   = 21; //18; /* ~1 sec */
        QP_ATTR_MASK_SET(qp_attr_mask,QP_ATTR_TIMEOUT);

        qp_attr.retry_count   = 2; //1;
        QP_ATTR_MASK_SET(qp_attr_mask,QP_ATTR_RETRY_COUNT);

        qp_attr.rnr_retry     = 10; //3;
        QP_ATTR_MASK_SET(qp_attr_mask,QP_ATTR_RNR_RETRY);

        if ((params->in_params.test_mode == BW_ATOM_CS) ||
            (params->in_params.test_mode == BW_LAT_ATOM_CS) ||
            (params->in_params.test_mode == LAT_ATOM_CS) // ||
        //    //(params->in_params.test_mode == LAT_RDMA_R) ||
        //    //(params->in_params.test_mode == BW_RDMA_R) ||
        //    //(params->in_params.test_mode == BW_RDMA_R_UNI)
            )
        {
            qp_attr.ous_dst_rd_atom  = params->hca_cap.max_qp_ous_rd_atom;
        //    qp_attr.ous_dst_rd_atom = 128;
        }
        else {
            qp_attr.ous_dst_rd_atom  = params->qp_ous_rd_atom;
        }
        //printf("ous_dst_rd_atom = %d\n", qp_attr.ous_dst_rd_atom);
        QP_ATTR_MASK_SET(qp_attr_mask,QP_ATTR_OUS_DST_RD_ATOM);
    }
    else if (params->in_params.transport == IB_TS_UC) {

        qp_attr.sq_psn           = 0;
        QP_ATTR_MASK_SET(qp_attr_mask,QP_ATTR_SQ_PSN);

    }

    res = VAPI_modify_qp(hca_hndl,qp_hndl,&qp_attr,&qp_attr_mask,&qp_cap);
    if (res != VAPI_OK) {
        printf("Error: Modifying  QP to RTS: %s\n", VAPI_strerror(res));
        return(RET_ERR);
    }
    COND_PRINTF(("Modified QP to RTS\n"));
    return(RET_OK);
}

static ret_val_t qp_init2rts_one_qp_special(test_params_tx *params,qp_entry_t *qp_e_p, int port_ix)
{
    VAPI_qp_attr_mask_t    qp_attr_mask;
    VAPI_qp_attr_t         qp_attr;
    VAPI_qp_cap_t          qp_cap;
    VAPI_ret_t             res;

    // printf("Initializing QP #%d, connected to QP #%d\n", qp_e_p->qp_num, qp_e_p->d_qpnum);

    VAPI_hca_hndl_t hca_hndl  = params->hca_hndl;
    VAPI_qp_hndl_t  qp_hndl   = qp_e_p->qp_hndl;
    int             d_qpnum   = qp_e_p->d_qpnum;
    int             dlid      = params->port[port_ix].d_lid; /*qp_e_p->d_lid;*/
    int             port_num  = params->port[port_ix].port_num + 1;
    IB_mtu_t        mtu       = params->in_params.mtu;

    /***************************************************************************/

    /********************************* INIT ******************************************/
    QP_ATTR_MASK_CLR_ALL(qp_attr_mask);

    qp_attr.qp_state = VAPI_INIT;
    QP_ATTR_MASK_SET(qp_attr_mask,QP_ATTR_QP_STATE);

    qp_attr.pkey_ix  = 0;
    QP_ATTR_MASK_SET(qp_attr_mask,QP_ATTR_PKEY_IX);

    qp_attr.port     = (IB_port_t)port_num;
    QP_ATTR_MASK_SET(qp_attr_mask,QP_ATTR_PORT);

    if (params->in_params.transport == IB_TS_RC) {
        qp_attr.remote_atomic_flags = VAPI_EN_REM_WRITE | VAPI_EN_REM_READ | VAPI_EN_REM_ATOMIC_OP;
        QP_ATTR_MASK_SET(qp_attr_mask,QP_ATTR_REMOTE_ATOMIC_FLAGS);
    }
    else if (params->in_params.transport == IB_TS_UC) {
        qp_attr.remote_atomic_flags = VAPI_EN_REM_WRITE;
        QP_ATTR_MASK_SET(qp_attr_mask,QP_ATTR_REMOTE_ATOMIC_FLAGS);
    }
    else {
        qp_attr.qkey = 0; 
        QP_ATTR_MASK_SET(qp_attr_mask,QP_ATTR_QKEY);
    }


    res = VAPI_modify_qp(hca_hndl, qp_hndl, &qp_attr, &qp_attr_mask, &qp_cap);
    if (res != VAPI_OK) {
        printf("Error: Modifying  QP to INIT: %s\n", VAPI_strerror(res));
        return(RET_ERR);
    }
    COND_PRINTF(("Modified QP to INIT\n"));

    /**************************** RTR ***********************************************/   
    QP_ATTR_MASK_CLR_ALL(qp_attr_mask);

    qp_attr.qp_state = VAPI_RTR;
    QP_ATTR_MASK_SET(qp_attr_mask,QP_ATTR_QP_STATE);


    if (params->in_params.transport == IB_TS_RC) {
        if ((params->in_params.test_mode == BW_ATOM_CS) ||
            (params->in_params.test_mode == BW_LAT_ATOM_CS) ||
            (params->in_params.test_mode == LAT_ATOM_CS) // ||
    //        //(params->in_params.test_mode == LAT_RDMA_R) ||
    //        //(params->in_params.test_mode == BW_RDMA_R) ||
    //        //(params->in_params.test_mode == BW_RDMA_R_UNI)
            )
        {
            qp_attr.qp_ous_rd_atom  = params->hca_cap.max_qp_ous_rd_atom;
    //        qp_attr.qp_ous_rd_atom = 128;
        }
        else {
            qp_attr.qp_ous_rd_atom  = MIN2(DFLT_QP_OUS_RD_ATOM, params->hca_cap.max_qp_ous_rd_atom);
        }
        //printf("ous_rd_atom = %d\n", qp_attr.qp_ous_rd_atom);
        //printf("max_ous_rd_atom = %d\n", params->hca_cap.max_qp_ous_rd_atom);

        QP_ATTR_MASK_SET(qp_attr_mask,QP_ATTR_QP_OUS_RD_ATOM);

        qp_attr.dest_qp_num = d_qpnum; 
        QP_ATTR_MASK_SET(qp_attr_mask, QP_ATTR_DEST_QP_NUM);

        qp_attr.av.sl            = 0;
        qp_attr.av.grh_flag      = FALSE;
        qp_attr.av.dlid          = (IB_lid_t)dlid;
        qp_attr.av.static_rate   = 0;
        qp_attr.av.src_path_bits = 0;
        QP_ATTR_MASK_SET(qp_attr_mask,QP_ATTR_AV);

        qp_attr.path_mtu         = mtu;
        QP_ATTR_MASK_SET(qp_attr_mask,QP_ATTR_PATH_MTU);

        qp_attr.rq_psn           = 0;
        QP_ATTR_MASK_SET(qp_attr_mask,QP_ATTR_RQ_PSN);

        qp_attr.pkey_ix          = 0;
        QP_ATTR_MASK_SET(qp_attr_mask,QP_ATTR_PKEY_IX);
                                
        qp_attr.min_rnr_timer    = 0;
        QP_ATTR_MASK_SET(qp_attr_mask,QP_ATTR_MIN_RNR_TIMER);
    }
    else if (params->in_params.transport == IB_TS_UC) {

        qp_attr.av.sl            = 0;
        qp_attr.av.grh_flag      = FALSE;
        qp_attr.av.dlid          = (IB_lid_t)dlid;
        qp_attr.av.static_rate   = 0;
        qp_attr.av.src_path_bits = 0;
        QP_ATTR_MASK_SET(qp_attr_mask,QP_ATTR_AV);

        qp_attr.dest_qp_num = d_qpnum; 
        QP_ATTR_MASK_SET(qp_attr_mask, QP_ATTR_DEST_QP_NUM);

        qp_attr.path_mtu         = mtu;
        QP_ATTR_MASK_SET(qp_attr_mask,QP_ATTR_PATH_MTU);

        qp_attr.rq_psn           = 0;
        QP_ATTR_MASK_SET(qp_attr_mask,QP_ATTR_RQ_PSN);

        qp_attr.pkey_ix = 0;
        QP_ATTR_MASK_SET(qp_attr_mask,QP_ATTR_PKEY_IX);

    }


    res = VAPI_modify_qp(hca_hndl, qp_hndl, &qp_attr, &qp_attr_mask, &qp_cap);
    if (res != VAPI_OK) {
        printf("Error: Modifying  QP to RTR: %s\n", VAPI_strerror(res));      
        return(RET_ERR);
    }
    COND_PRINTF(("Moodified QP to RTR\n"));


    sync_rtr2rts(params);

    /***************************** RTS **********************************************/   

    QP_ATTR_MASK_CLR_ALL(qp_attr_mask);

    qp_attr.qp_state   = VAPI_RTS;
    QP_ATTR_MASK_SET(qp_attr_mask,QP_ATTR_QP_STATE);

    qp_attr.sq_psn   = 0;
    QP_ATTR_MASK_SET(qp_attr_mask,QP_ATTR_SQ_PSN);

    if (params->in_params.transport == IB_TS_RC) {

        qp_attr.timeout   = 21; //18; /* ~1 sec */
        QP_ATTR_MASK_SET(qp_attr_mask,QP_ATTR_TIMEOUT);

        qp_attr.retry_count   = 2; //1;
        QP_ATTR_MASK_SET(qp_attr_mask,QP_ATTR_RETRY_COUNT);

        qp_attr.rnr_retry     = 10; //3;
        QP_ATTR_MASK_SET(qp_attr_mask,QP_ATTR_RNR_RETRY);

        if ((params->in_params.test_mode == BW_ATOM_CS) ||
            (params->in_params.test_mode == BW_LAT_ATOM_CS) ||
            (params->in_params.test_mode == LAT_ATOM_CS) // ||
        //    //(params->in_params.test_mode == LAT_RDMA_R) ||
        //    //(params->in_params.test_mode == BW_RDMA_R) ||
        //    //(params->in_params.test_mode == BW_RDMA_R_UNI)
            )
        {
            qp_attr.ous_dst_rd_atom  = params->hca_cap.max_qp_ous_rd_atom;
        //    qp_attr.ous_dst_rd_atom = 128;
        }
        else {
            qp_attr.ous_dst_rd_atom  = params->qp_ous_rd_atom;
        }
        //printf("ous_dst_rd_atom = %d\n", qp_attr.ous_dst_rd_atom);
        QP_ATTR_MASK_SET(qp_attr_mask,QP_ATTR_OUS_DST_RD_ATOM);
    }
    else if (params->in_params.transport == IB_TS_UC) {

        qp_attr.sq_psn           = 0;
        QP_ATTR_MASK_SET(qp_attr_mask,QP_ATTR_SQ_PSN);

    }

    res = VAPI_modify_qp(hca_hndl,qp_hndl,&qp_attr,&qp_attr_mask,&qp_cap);
    if (res != VAPI_OK) {
        printf("Error: Modifying  QP to RTS: %s\n", VAPI_strerror(res));
        return(RET_ERR);
    }
    COND_PRINTF(("Modified QP to RTS\n"));
    return(RET_OK);
}


#ifdef CHK_QP_STATE

#define SHOW_QP_STATE_STEP		5000000

static void show_qp_state(VAPI_hca_hndl_t  hca_hndl, VAPI_qp_hndl_t qp_hndl, VAPI_qp_num_t  qp_num)
{
	VAPI_qp_attr_t       qp_attr;
	VAPI_qp_attr_mask_t  qp_attr_mask;
	VAPI_qp_init_attr_t  qp_init_attr;
   VAPI_ret_t             res;
	
	res = VAPI_query_qp(hca_hndl, qp_hndl, &qp_attr, &qp_attr_mask, &qp_init_attr );
	if (res == VAPI_OK) {
		COND_PRINTF3(("=== QP handle 0x%x, QP num 0x%x: state %d\n", (unsigned)qp_hndl, (unsigned)qp_num, qp_attr.qp_state));
	}
	else {
		COND_PRINTF3(("Error on VAPI_query_qp: %s\n", VAPI_strerror(res)));
	}
}
#endif

ret_val_t PERF_poll_cq(VAPI_hca_hndl_t  hca_hndl, VAPI_cq_hndl_t    cq_hndl,
                       VAPI_wc_desc_t *sc_p, int use_events, MOSAL_semaphore_t *sem_cq_p)
{
#ifdef CHK_QP_STATE
	static int delay=0;
    VAPI_qp_num_t  qp_num0 = g_params_p->port[0].qp_vect[0].qp_num;
	 VAPI_qp_hndl_t  qp_hndl0 = g_params_p->port[0].qp_vect[0].qp_hndl;
    VAPI_qp_num_t  qp_num1 = g_params_p->port[0].qp_vect[1].qp_num;
	 VAPI_qp_hndl_t  qp_hndl1 = g_params_p->port[0].qp_vect[1].qp_hndl;
#endif
    VAPI_ret_t             res;

    res = VAPI_CQ_EMPTY;
    MTPERF_TIME_START(VAPI_poll_cq);
    res = VAPI_poll_cq(hca_hndl, cq_hndl, sc_p);
    //MTPERF_TIME_END_IF(VAPI_poll_cq, res!=VAPI_CQ_EMPTY);
    MTPERF_TIME_END(VAPI_poll_cq);

    while (res == VAPI_CQ_EMPTY) {

    		
        if (use_events) {
#ifdef CQ_BLOCK
            res = EVAPI_poll_cq_block(hca_hndl, cq_hndl, /* wait forever*/ 0, sc_p);
            if ((res != VAPI_OK) && (res != VAPI_CQ_EMPTY)){
                printf("Error: Poll CQ Block: %s\n", VAPI_strerror(res));
                return(RET_ERR);
            }
            continue;

#else
            VAPI_req_comp_notif(hca_hndl,cq_hndl,VAPI_NEXT_COMP);
            res = VAPI_poll_cq(hca_hndl, cq_hndl, sc_p);
            if (res != VAPI_CQ_EMPTY) {
                continue;
            }
            if (MOSAL_sem_acq(sem_cq_p,TRUE) != 0) { /* Wait for notification handler */
                printf(MT_FLFMT("Failed acquiring RQ CQ completion semaphore."));
            }
#endif
        } // use_events
        else {
    MTPERF_TIME_START(VAPI_poll_cq);
            res = VAPI_poll_cq(hca_hndl, cq_hndl, sc_p);
    //MTPERF_TIME_END_IF(VAPI_poll_cq, res!=VAPI_CQ_EMPTY);
            MTPERF_TIME_END(VAPI_poll_cq);
        }
#ifdef CHK_QP_STATE
		if (res == VAPI_OK) { /* got a completion */
			show_qp_state(hca_hndl, qp_hndl0, qp_num0);
			show_qp_state(hca_hndl, qp_hndl1, qp_num1);
		}
		else { /* no completion */
			if (!(delay++%SHOW_QP_STATE_STEP)) {
				show_qp_state(hca_hndl, qp_hndl0, qp_num0);
				show_qp_state(hca_hndl, qp_hndl1, qp_num1);
			}
		}
#endif		
    }

    if (res != VAPI_OK) {
        printf("Error: Polling CQ : %s\n", VAPI_strerror(res));
        return(RET_ERR);
    }
    if (sc_p->status != VAPI_SUCCESS) {
#ifdef SW_VER_GT_0_2_0
        printf("Completion with error on send queue (syndrome=0x%x=%s , opcode=%d=%s)\n",
               sc_p->vendor_err_syndrome,VAPI_wc_status_sym(sc_p->status),sc_p->opcode,VAPI_cqe_opcode_sym(sc_p->opcode));
#endif
        printf("%s: completion with error (%d) detected\n", __func__, sc_p->status);
        return(RET_ERR);
    }
    return(RET_OK);
}

/*
	Purpose: performs one poll
	Returns:
		VAPI_OK - on success
		VAPI_CQ_EMPTY - while no completion event
		VAPI_EGEN - on error
*/
VAPI_ret_t PERF_poll_cq_nowait(VAPI_hca_hndl_t  hca_hndl, VAPI_cq_hndl_t    cq_hndl,
                       VAPI_wc_desc_t *sc_p, int use_events, MOSAL_semaphore_t *sem_cq_p)
{

#ifdef CHK_QP_STATE
	static int delay=0;
    VAPI_qp_num_t  qp_num0 = g_params_p->port[0].qp_vect[0].qp_num;
	 VAPI_qp_hndl_t  qp_hndl0 = g_params_p->port[0].qp_vect[0].qp_hndl;
    VAPI_qp_num_t  qp_num1 = g_params_p->port[0].qp_vect[1].qp_num;
	 VAPI_qp_hndl_t  qp_hndl1 = g_params_p->port[0].qp_vect[1].qp_hndl;
#endif
    VAPI_ret_t             res;

    res = VAPI_CQ_EMPTY;

	if (use_events) {
#ifdef CQ_BLOCK
	    res = EVAPI_poll_cq_block(hca_hndl, cq_hndl, /* wait forever*/ 0, sc_p);
	    if ((res != VAPI_OK) && (res != VAPI_CQ_EMPTY)){
	        printf("Error: Poll CQ Block: %s\n", VAPI_strerror(res));
	        return(VAPI_EGEN);
	    }
	    goto exit;

#else
	    VAPI_req_comp_notif(hca_hndl,cq_hndl,VAPI_NEXT_COMP);
	    res = VAPI_poll_cq(hca_hndl, cq_hndl, sc_p);
	    if (res != VAPI_CQ_EMPTY) {
	        goto exit;
	    }
	    if (MOSAL_sem_acq(sem_cq_p,TRUE) != 0) { /* Wait for notification handler */
	        printf(MT_FLFMT("Failed acquiring RQ CQ completion semaphore."));
	    }
#endif
	} // use_events
	else {
	    res = VAPI_poll_cq(hca_hndl, cq_hndl, sc_p);
	}

    if (res == VAPI_OK) {
	    if (sc_p->status != VAPI_SUCCESS) {
#ifdef SW_VER_GT_0_2_0
	        printf("Completion with error on send queue (syndrome=0x%x=%s , opcode=%d=%s)\n",
	               sc_p->vendor_err_syndrome,VAPI_wc_status_sym(sc_p->status),sc_p->opcode,VAPI_cqe_opcode_sym(sc_p->opcode));
#endif
	        printf("%s: completion with error (%d) detected\n", __func__, sc_p->status);
	        return(VAPI_EGEN);
	    }
   	}
#ifdef CHK_QP_STATE
		if (res == VAPI_OK) { /* got a completion */
			show_qp_state(hca_hndl, qp_hndl0, qp_num0);
			show_qp_state(hca_hndl, qp_hndl1, qp_num1);
	}
		else { /* no completion */
			if (!(delay++%SHOW_QP_STATE_STEP)) {
				show_qp_state(hca_hndl, qp_hndl0, qp_num0);
				show_qp_state(hca_hndl, qp_hndl1, qp_num1);
			}
		}
#endif		
    
exit:    
    return(res);
}


static void MT_API async_events_handler( VAPI_hca_hndl_t hca_hndl, VAPI_event_record_t *event_p, void* priv_data)
{

  printf("Got async. event: %s (0x%x - %s)\n",
          VAPI_event_record_sym(event_p->type),event_p->syndrome,VAPI_event_syndrome_sym(event_p->syndrome));
  
  printf("CQ handle: 0x%x\n", (u_int32_t) event_p->modifier.cq_hndl);
}


static void set_data_mem(void *inbuff, unsigned long len)
{

    int i,num;
    u_int32_t *buff = (u_int32_t *) inbuff;
    num = len/sizeof(u_int32_t);

    for (i=0;i<num;i++) {
        buff[i] = (u_int32_t) 0x12345678;

    }
}
