/*
  This software is available to you under a choice of one of two
  licenses.  You may choose to be licensed under the terms of the GNU
  General Public License (GPL) Version 2, available at
  <http://www.fsf.org/copyleft/gpl.html>, or the OpenIB.org BSD
  license, available in the LICENSE.TXT file accompanying this
  software.  These details are also available at
  <http://openib.org/license.html>.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
  BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
  ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
  CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
  SOFTWARE.

  Copyright (c) 2004 Mellanox Technologies Ltd.  All rights reserved.
*/

#include <vapi.h>
#include <evapi.h>
#include <vapi_common.h>
#include <stdio.h>

#define NEW_POLL 1

//#define TIME_LEVEL 0
//#define DEBUG_PRINT
#include "utils.h" 
#include "all_bw_lat.h"
#include "sock.h" 

#ifndef MTPERF
//#define MTPERF
#endif

#include <mtperf.h>

MTPERF_NEW_SEGMENT(VAPI_poll_cq,500);
//MTPERF_NEW_SEGMENT(EVAPI_post_inline_sr,20000);
//MTPERF_NEW_SEGMENT(VAPI_post_sr,20000);
//MTPERF_NEW_SEGMENT(test,100);

//MTPERF_EXTERN_SEGMENT(HHUL_post_inline_send_req);
//MTPERF_EXTERN_SEGMENT(HHUL_post_send_req);
//MTPERF_EXTERN_SEGMENT(THH_uar_sendq_dbell);
//MTPERF_EXTERN_SEGMENT(WQE_build_send);
//MTPERF_EXTERN_SEGMENT(SQ_WQE_copy);
//MTPERF_EXTERN_SEGMENT(MOSAL_spinlock_unlock);
//MTPERF_EXTERN_SEGMENT(sq_alloc_wqe_link_and_ring);
//MTPERF_EXTERN_SEGMENT(MOSAL_spinlock_irq_lock);
//MTPERF_EXTERN_SEGMENT(HHUL_poll4cqe);
//MTPERF_EXTERN_SEGMENT(prepare_sq_dbell);
//MTPERF_EXTERN_SEGMENT(WQE_pack_send_next);
//MTPERF_EXTERN_SEGMENT(alloc_next_wqe);
//MTPERF_EXTERN_SEGMENT(GET_QP_INFO);
//MTPERF_EXTERN_SEGMENT(MOSAL_MMAP_IO_WRITE_DWORD);
//MTPERF_EXTERN_SEGMENT(next_draft);

/***********************************************************************************************
 * name:      rdma_read_lat_test
 * args:
 *            params_p -- test parameter structure, need to filled before calling
 * returns:
 *            RET_OK
 *            RET_ERR
 *
 * description:     
 *            This is the main loop for the RDMA read latency test. 
 *            This test, in each iteration loop, send RDMA_READ and poll for completion
 *            
 ************************************************************************************************/  
ret_val_t rdma_read_lat_test(test_params_tx *params_p)
{

    VAPI_ret_t             res;
    int i;
    int j=0;
    //last_post,comp_inter,ask_comp;
    u_int64_t t_start=0, t_end=0;
    u_int64_t t_start_my=0, t_start_my_prev=0, delta_sum=0;
    //int my_iter = 0;
    int my_iter_good;
    float bad_average_double;

    VAPI_sr_desc_t         sr;
    VAPI_sg_lst_entry_t    sg_entry_s;
    VAPI_wc_desc_t         sc;

    VAPI_hca_hndl_t        hca_hndl    = params_p->hca_hndl;
    VAPI_qp_hndl_t         qp_hndl     = params_p->port[CUR_PORT_IX].qp_vect[0].qp_hndl;

    VAPI_cq_hndl_t         s_cq_hndl   = params_p->port[CUR_PORT_IX].s_cq_hndl;
    MOSAL_semaphore_t    * sq_sem_p   = &params_p->port[CUR_PORT_IX].sq_sem;
    //    u_int8_t               *recv_buf   = params_p->recv_buf;
    u_int8_t               *send_buf   = params_p->send_buf;
    u_int64_t               *delta = NULL;
    int                    is_sender   = params_p->in_params.is_sender;
    int                    size        = params_p->in_params.size;
    int                    iter        = params_p->in_params.iter;      
    //u_int32_t * qp_sent_count = calloc(sizeof(u_int32_t), params_p->in_params.num_of_qps);
    //    int                    skip        = params_p->in_params.skip;
    //volatile char          *check_p;
    //int index_to_check=0;
    //    int act_skip;

    //int max_posible = MIN(params_p->in_params.num_s_wqe,params_p->in_params.num_cqe);

    if (size > params_p->in_params.mr_recv_size) {
        printf("Warning!!!: size=%d > mr_recv_size=%d --- changing iter to %d\n", 
               size,params_p->in_params.mr_recv_size,params_p->in_params.mr_recv_size);
        size = params_p->in_params.mr_recv_size;
    }

    if (size==0) {
        printf("For RDMA size must be >= 1. Changing size to 1\n");
        size =1;
    }

/***************************************************************************/
    //sr.comp_type = VAPI_UNSIGNALED;
    sr.comp_type = VAPI_SIGNALED; //Noam
    sr.set_se = FALSE;

    sr.opcode = VAPI_RDMA_READ; //Noam
    sr.remote_qkey = 0;
    sr.remote_addr = params_p->remote_add;
    sr.r_key = params_p->remote_rkey;
    sr.id = 1;
    sr.fence = 0;

    sr.sg_lst_len = 1;
    sg_entry_s.len = size;
    sg_entry_s.lkey = params_p->send_lkey;
    sg_entry_s.addr = (VAPI_virt_addr_t)(MT_virt_addr_t)send_buf; 
    sr.sg_lst_p = &sg_entry_s;


    sock_sync_msg(params_p->in_params.socket, is_sender);
    COND_PRINTF1(("\n\n************* RDMA-Read Latency Test started  (Messages up to %d use inline data)  ***************\n\n",
                  params_p->max_inline_size));

    //if ((iter+skip) <= params_p->in_params.num_s_wqe) {
    //    comp_inter = params_p->in_params.num_s_wqe;
    // }
    //else {
    //    comp_inter = params_p->in_params.num_s_wqe / 2 - 1;
    //}

    // We start the timing AFTER the first send
    //act_skip = skip > 0 ? (skip -1) : skip;

    //ask_comp=0;
    //index_to_check = size -1 ;
    //check_p = &recv_buf[index_to_check];
    sg_entry_s.len = size;

    if (is_sender) {
        delta = malloc(iter * 10);
        get_clk64(t_start);


        for (i=0; i<(iter); i++) {

            qp_hndl = params_p->port[CUR_PORT_IX].qp_vect[j].qp_hndl;
            
            //if (i > 0) {
                t_start_my_prev = t_start_my;
            //}

            get_clk64(t_start_my);

            // Remove the if statement to save cpu time. Later ignore delta[0].
            //if (i > 0) {
                //printf("RoundTrip Latency: %.2f usec\n",CLK_TO_USEC(t_start_my - t_start_my_prev));
                //printf("my_iter %d t_start_my %lld t_start_my_prev %lld\n", my_iter, t_start_my, t_start_my_prev);
                delta[i] = t_start_my - t_start_my_prev;
                //my_iter++;
            //}

            COND_PRINTF(("Posted send desc %d to" U64_FMT " (size=%d val[0]=%d)\n",i,sr.remote_addr,
                         sr.sg_lst_p[0].len, *((int *)((unsigned long)sr.sg_lst_p[0].addr))));
            res = VAPI_post_sr(hca_hndl, qp_hndl, &sr);
            if (res != VAPI_OK) {
                printf("Error: Posting send desc : %s\n", VAPI_strerror(res));
                return(RET_ERR);
            }

            COND_PRINTF(("Polling for sr comp i=%d\n",i));
            //printf(("Polling for sr comp \n"));
            //      printf("Use event = %d \n",params_p->in_params.use_send_events);
            if (RET_OK != PERF_poll_cq(hca_hndl,s_cq_hndl ,&sc,params_p->in_params.use_send_events, 
                                       sq_sem_p)) {
                return RET_ERR;
            }
            j = (j+1)%params_p->in_params.num_of_qps;            
        }/*end of iteration loop*/
    };
    if (is_sender) {
        get_clk64(t_end);
        bad_average_double = 1.5 * (CLK_TO_USEC(t_end - t_start) / iter);
        //my_iter_good = my_iter;
        my_iter_good = iter - 1;
        //for (i=0; i<my_iter; i++)
        for (i=1; i<iter; i++)
        {
            if (CLK_TO_USEC(delta[i]) < bad_average_double)
            {
                delta_sum = delta_sum + delta[i];
                //printf("adding delta = %.2f\n", CLK_TO_USEC(delta[i]));
            }
            else
            {
                //printf("skipping delta = %.2f\n", CLK_TO_USEC(delta[i]));
                my_iter_good--;
            }

        }
        //printf("my_iter_good = %d iter = %d \n", my_iter_good, iter);
        if (my_iter_good == 0)
        {
            printf("Latency cannot be measured. Please use more iterations\n");
        }
        else
        {

#if 0
        //printf("OLD Latency: %.2f usec , NEW Latency: %.2f usec [Num. of QPs: %d, Post List: %d, QPC Sets: %d, MPT Sets: %d, Keys: %d, size: %d bytes]\n",
        printf("Latency: %.2f usec [Num. of QPs: %d, Post List: %d, QPC Sets: %d, MPT Sets: %d, Keys: %d, size: %d bytes]\n",
        //       CLK_TO_USEC(t_end - t_start) /(iter* 1.0),
               CLK_TO_USEC(delta_sum) /(my_iter_good* 1.0),
            params_p->in_params.num_of_qps, params_p->in_params.postlist_len, 
            params_p->in_params.qpc_sets, params_p->in_params.mpt_sets, 
            params_p->in_params.num_of_keys, size);
#else
        //printf("Latency: %.2f usec [size: %d bytes]\n",CLK_TO_USEC(t_end - t_start) /( (iter-act_skip)* 2.0),size);
        //printf("Latency: %.2f usec [size: %d bytes]\n",CLK_TO_USEC(t_end - t_start) /( (iter)* 1.0),size);
        printf("Latency: %.2f usec [size: %d bytes]\n",CLK_TO_USEC(delta_sum) /(my_iter_good * 1.0),size);
#endif 
        }
        fflush(stdout);
        free(delta);
    }
    sock_sync_msg(params_p->in_params.socket, is_sender);
    COND_PRINTF1(("\n************* RDMA-Read Latency Test Finished  *********************\n\n"));
    return(RET_OK);
}


/***********************************************************************************************
 * name:      rdma_lat_test
 * args:
 *            params_p -- test parameter structure, need to filled before calling
 * returns:
 *            RET_OK
 *            RET_ERR
 *
 * description:     
 *            This is the main loop for the RDMA write latency test. T
 *            
 ************************************************************************************************/  
ret_val_t rdma_lat_test(test_params_tx *params_p)
{

    VAPI_ret_t             res;
    int i,k=0,ii=0,j=0,l=0,ask_comp,last_post,comp_inter;
    u_int64_t t_start=0, t_end=0;
    u_int64_t t_start_my=0, t_start_my_prev=0, delta_sum=0;
    //int my_iter = 0;
    int my_iter_good;
    float bad_average_double;

    VAPI_sr_desc_t         * sr;
    VAPI_sg_lst_entry_t    * sg_entry_s;
    VAPI_wc_desc_t         sc;

    VAPI_hca_hndl_t        hca_hndl    = params_p->hca_hndl;
    VAPI_qp_hndl_t         qp_hndl     = params_p->port[CUR_PORT_IX].qp_vect[0].qp_hndl;

    VAPI_cq_hndl_t         s_cq_hndl   = params_p->port[CUR_PORT_IX].s_cq_hndl;
    MOSAL_semaphore_t    * sq_sem_p   = &params_p->port[CUR_PORT_IX].sq_sem;
    u_int8_t               *recv_buf   = params_p->recv_buf;
    u_int8_t               *send_buf   = params_p->send_buf;
    u_int64_t               *delta = NULL;
    int                    is_sender   = params_p->in_params.is_sender;
    int                    size        = params_p->in_params.size;
    int                    iter        = params_p->in_params.iter;      
    int                    skip        = params_p->in_params.skip;
    volatile char          *check_p;
    int index_to_check=0;
    int act_skip;
    int                    list_len    = params_p->in_params.postlist_len;
    u_int32_t * qp_sent_count = calloc(sizeof(u_int32_t), params_p->in_params.num_of_qps);
    int use_inline = FALSE;

    int max_sends = MIN2(params_p->in_params.num_r_wqe,params_p->in_params.num_s_wqe); 
    //int max_posible = MIN(params_p->in_params.num_s_wqe,params_p->in_params.num_cqe);

    if (size > params_p->in_params.mr_recv_size) {
        printf("Warning!!!: size=%d > mr_recv_size=%d --- changing iter to %d\n", 
               size,params_p->in_params.mr_recv_size,params_p->in_params.mr_recv_size);
        size = params_p->in_params.mr_recv_size;
    }

    if (size==0) {
        printf("For RDMA size must be >= 1. Changing size to 1\n");
        size =1;
    }
    
    if (list_len > max_sends)
    {
        printf("Warning!!! Post List Length > WQ size --- reducing it to %d\n", max_sends);
        list_len = max_sends;
    }

    sr = VMALLOC(list_len*sizeof(VAPI_sr_desc_t));
    sg_entry_s = VMALLOC(params_p->in_params.num_of_keys*sizeof(VAPI_sg_lst_entry_t));


/***************************************************************************/
   for (i=0; i<list_len; i++)
    {
        sr[i].comp_type = VAPI_UNSIGNALED;
        sr[i].set_se = FALSE;

        sr[i].opcode = VAPI_RDMA_WRITE;
        sr[i].remote_qkey = 0;
        sr[i].remote_addr = params_p->remote_add;
        //sr[i].r_key = params_p->remote_rkey;
        sr[i].id = 1;
        sr[i].sg_lst_len = 1;
        //sr.sg_lst_p = &sg_entry_s;
    }

    for (i=0; i<params_p->in_params.num_of_keys; i++) {    
        sg_entry_s[i].len = size;
        sg_entry_s[i].lkey = params_p->send_lkey_arr[i];
        sg_entry_s[i].addr = (VAPI_virt_addr_t)(MT_virt_addr_t)send_buf; 
    }

    sock_sync_msg(params_p->in_params.socket, is_sender);
    COND_PRINTF1(("\n\n************* RDMA-Write Latency Test started  (Messages up to %d use inline data)  ***************\n\n",
                  params_p->max_inline_size));

    if ((iter+skip) <= (int)params_p->in_params.num_s_wqe) {
        comp_inter = params_p->in_params.num_s_wqe;
    } else {
        comp_inter = params_p->in_params.num_s_wqe / 2 - 1;
    }

    // We start the timing AFTER the first send
    act_skip = skip > 0 ? (skip -1) : skip;
    //printf("skip = %d , act_skip = %d\n", skip, act_skip);
    ask_comp=0;
    index_to_check = size -1 ;
    check_p = &recv_buf[index_to_check];

    if ((size >0) &&(size <= params_p->max_inline_size) && (list_len == 1))
    {
        use_inline = TRUE;
    }

    //sg_entry_s.len = size;
    // Post the first send req
    if (is_sender) {
   
        delta = malloc((iter+act_skip) * 10);

        //MTPERF_TIME_START(test);
        //MTPERF_TIME_END(test);

        if (skip == 0) {
            get_clk64(t_start);
        }
            for (k=0;k<list_len;k++)
            {
                if (list_len > 1)
                {
                    sr[k].comp_type = VAPI_SIGNALED;
                    ask_comp++;
                    COND_PRINTF2((" **** Posting sr signaled i=%d ask=%d\n",i+1,ask_comp));
                } else
                {
                    sr[k].comp_type = VAPI_UNSIGNALED;
                }
            } // for k

        //if ((size >0) &&(size <= params_p->max_inline_size) && (list_len == 1)) {
        if (use_inline == TRUE) {
            sr[0].r_key = params_p->remote_rkey_arr[l];
            sr[0].sg_lst_p = &sg_entry_s[l];
            res = EVAPI_post_inline_sr(hca_hndl, qp_hndl, sr);
        } else {
            for (ii=0; ii<list_len; ii++) {
                  sr[ii].r_key = params_p->remote_rkey_arr[l];
                  sr[ii].sg_lst_p = &sg_entry_s[l];
            }
            res = EVAPI_post_sr_list(hca_hndl, qp_hndl, list_len, sr);
        }  

        //MTPERF_TIME_START(test);
        //MTPERF_TIME_END(test);

        if (res != VAPI_OK) {
            printf("Error: Posting send desc : %s\n", VAPI_strerror(res));
            return(RET_ERR);
        }
                while (ask_comp > 0)
                {
                    COND_PRINTF2(("**** Polling for sr comp i=%d ask=%d\n",i+1,ask_comp));
                    if (RET_OK != PERF_poll_cq(hca_hndl,s_cq_hndl ,&sc,params_p->in_params.use_send_events, 
                                               sq_sem_p))
                    {
                        printf("Perf_poll_cq failed for QP %d (%d), at sent_count %d\n",
                               j, params_p->port[CUR_PORT_IX].qp_vect[j].qp_num, qp_sent_count[j]);
                        // getchar();
                        return RET_ERR;
                    }

                    ask_comp--;
                }

                qp_sent_count[0] += list_len;
                j = (j+1)%params_p->in_params.num_of_qps;            

                l = (l+1)%params_p->in_params.num_of_keys;
    } // is_sender

    for (i=0; i<(iter + act_skip)/list_len; i++) {
        qp_hndl = params_p->port[CUR_PORT_IX].qp_vect[j].qp_hndl;

        COND_PRINTF(("Checking data at %p (current=%d)\n",check_p,*check_p));
        while (*check_p==84) {
        }
        COND_PRINTF((".................Got data %d at %p\n",*check_p,check_p));
        recv_buf[index_to_check] = 84; 
        if ((i == act_skip) && (skip>0) && is_sender ) {
            get_clk64(t_start);
        }

        //if ((i > act_skip) && is_sender ) {
        if (is_sender) {
            t_start_my_prev = t_start_my;
        }

        //if ((i >= act_skip) && is_sender ) {
        if (is_sender) {
            get_clk64(t_start_my);
        }

        //if ((i > act_skip) && is_sender ) {
        if (is_sender) {
            //printf("RoundTrip Latency: %.2f usec\n",CLK_TO_USEC(t_start_my - t_start_my_prev));
            //printf("my_iter %d t_start_my %lld t_start_my_prev %lld\n", my_iter, t_start_my, t_start_my_prev);
            //delta[my_iter] = t_start_my - t_start_my_prev;
            delta[i] = t_start_my - t_start_my_prev;
            //my_iter++;
        }

        last_post = (is_sender && i== (iter + skip -2)) || (!is_sender && (i==(iter + skip -1)));
        //sr.comp_type = VAPI_UNSIGNALED;           

            for (k=0;k<list_len;k++)
            {
                sr[k].id = list_len * i + k;
                /* Ask for completion every comp_inter */
                if (
                    //((sr[k].id + 1) % comp_inter == 0) || 
                    (sr[k].id == ((iter/list_len)*list_len + k - 1)) ||
                    (((qp_sent_count[j]+k) % comp_inter == 0)) || 
                    (qp_sent_count[j]+k == (iter/params_p->in_params.num_of_qps-1)) ||
                    (list_len > 1) ||
                    (last_post))
                {
                    sr[k].comp_type = VAPI_SIGNALED;
                    ask_comp++;
                    //printf(" **** Posting sr signaled i=%d ask=%d qp=%d qp_sent_count=%d\n",i+1,ask_comp,params_p->port[CUR_PORT_IX].qp_vect[j].qp_num,qp_sent_count[j]);
                    COND_PRINTF2((" **** Posting sr signaled i=%d ask=%d\n",i+1,ask_comp));
                } else
                {
                    sr[k].comp_type = VAPI_UNSIGNALED;
                }
            } // for k

        //if (((i+1)%comp_inter == 0) || last_post) {
            //sr.comp_type = VAPI_SIGNALED;
            //ask_comp++;
            //COND_PRINTF(("Ask comp [ask=%d i=%d ]\n", ask_comp,i+1));
        //}

        if (!(is_sender && (i==(iter + skip -1)))) {
            COND_PRINTF(("Posted send desc %d to" U64_FMT " (size=%d val[0]=%d)\n",i,sr[i].remote_addr, sr[i].sg_lst_p[0].len, *((int *)((unsigned long)sr[i].sg_lst_p[0].addr))));
            //if ((size >0) &&(size <= params_p->max_inline_size) && (list_len==1)) {
            if (use_inline == TRUE) {
                sr[0].r_key = params_p->remote_rkey_arr[l];
                sr[0].sg_lst_p = &sg_entry_s[l];
                //MTPERF_TIME_START(EVAPI_post_inline_sr);
                res = EVAPI_post_inline_sr(hca_hndl, qp_hndl, sr);
                //MTPERF_TIME_END(EVAPI_post_inline_sr);
            } else {
                //MTPERF_TIME_START(VAPI_post_sr);
                for (ii=0; ii<list_len; ii++) {
                    sr[ii].r_key = params_p->remote_rkey_arr[l];
                    sr[ii].sg_lst_p = &sg_entry_s[l];
                }
                res = EVAPI_post_sr_list(hca_hndl, qp_hndl, list_len, sr);
                //MTPERF_TIME_END(VAPI_post_sr);
            }
            if (res != VAPI_OK) {
                printf("Error: Posting send desc : %s\n", VAPI_strerror(res));
                return(RET_ERR);
            }
         qp_sent_count[j] += list_len;
         j = (j+1)%params_p->in_params.num_of_qps;            
   
         l = (l+1)%params_p->in_params.num_of_keys;
        }

        if (ask_comp >1) {
            COND_PRINTF(("Polling for sr comp i=%d ask=%d\n",i+1,ask_comp-1));
            //if (0) printf("Polling for sr comp i=%d ask=%d\n",i+1,ask_comp-1);
            if (RET_OK != PERF_poll_cq(hca_hndl,s_cq_hndl ,&sc,params_p->in_params.use_send_events, 
                                       sq_sem_p)) {
                return RET_ERR;
            }
            ask_comp--;
        }

    }/*end of iteration loop*/


    if (is_sender) {
        get_clk64(t_end);

        bad_average_double = 2.0 * (CLK_TO_USEC(t_end - t_start) / iter);
        //my_iter_good = my_iter;
        my_iter_good = iter - 1;
        //for (i=0; i<my_iter; i++)
        for (i=act_skip+1; i<iter+act_skip; i++)
        {
            if (CLK_TO_USEC(delta[i]) < bad_average_double)
            {
                delta_sum = delta_sum + delta[i];
                //printf("adding delta = %.2f\n", CLK_TO_USEC(delta[i]));
            }
            else
            {
                //printf("skipping delta = %.2f\n", CLK_TO_USEC(delta[i]));
                my_iter_good--;
            }

        }
        //printf("my_iter_good = %d iter = %d\n", my_iter_good, iter);
        if (my_iter_good == 0)
        {
            printf("Latency cannot be measured. Please use more iterations\n");
        }
        else
        {
#if 0
        //printf("OLD Latency: %.2f usec New Latency: %.2f usec [Num. of QPs: %d, Post List: %d, QPC Sets: %d, MPT Sets: %d, Keys: %d, size: %d bytes]\n",
        printf("Latency: %.2f usec [Num. of QPs: %d, Post List: %d, QPC Sets: %d, MPT Sets: %d, Keys: %d, size: %d bytes]\n",
               //CLK_TO_USEC(t_end - t_start) /(iter* 2.0),
               CLK_TO_USEC(delta_sum) / (my_iter_good * 2.0),
            params_p->in_params.num_of_qps, params_p->in_params.postlist_len, 
            params_p->in_params.qpc_sets, params_p->in_params.mpt_sets, 
            params_p->in_params.num_of_keys, size);
#else
        //printf("Latency: %.2f usec [size: %d bytes]\n",CLK_TO_USEC(t_end - t_start) /( (iter)* 2.0),size);
        printf("Latency: %.2f usec [size: %d bytes]\n",CLK_TO_USEC(delta_sum)/(my_iter_good * 2.0), size);
#endif        
        }
        //MTPERF_TIME_START(test);
        //MTPERF_TIME_END(test);

        //MTPERF_REPORT_PRINTF(test);
        //MTPERF_REPORT_PRINTF(EVAPI_post_inline_sr);
        //MTPERF_REPORT_PRINTF(VAPI_post_sr);

        //MTPERF_REPORT_PRINTF(HHUL_post_inline_send_req);
        //MTPERF_REPORT_PRINTF(HHUL_post_send_req);
        //MTPERF_REPORT_PRINTF(THH_uar_sendq_dbell);
        //MTPERF_REPORT_PRINTF(WQE_build_send);
        //MTPERF_REPORT_PRINTF(SQ_WQE_copy);
        //MTPERF_REPORT_PRINTF(MOSAL_spinlock_unlock);
        //MTPERF_REPORT_PRINTF(sq_alloc_wqe_link_and_ring);
        //MTPERF_REPORT_PRINTF(MOSAL_spinlock_irq_lock);
        //MTPERF_REPORT_PRINTF(HHUL_poll4cqe);
        //MTPERF_REPORT_PRINTF(prepare_sq_dbell);
        //MTPERF_REPORT_PRINTF(WQE_pack_send_next);
        //MTPERF_REPORT_PRINTF(alloc_next_wqe);
        //MTPERF_REPORT_PRINTF(GET_QP_INFO);
        //MTPERF_REPORT_PRINTF(MOSAL_MMAP_IO_WRITE_DWORD);
        //MTPERF_REPORT_PRINTF(next_draft);

        fflush(stdout);
        free(delta);
    } // is_sender

    // Poll for the last SR complition
    while (ask_comp>0) {
        COND_PRINTF(("Final: Polling for last sr comp i=%d ask=%d\n",i,ask_comp-1));
        //printf("Final: Polling for last sr comp i=%d ask=%d\n",i,ask_comp-1);
        if (RET_OK != PERF_poll_cq(hca_hndl,s_cq_hndl ,&sc,params_p->in_params.use_send_events, 
                                   sq_sem_p)) {
            return RET_ERR;
        }
        ask_comp--;
    }
    //} // is_sender
    sock_sync_msg(params_p->in_params.socket, is_sender);
    COND_PRINTF1(("\n************* RDMA-Write Latency Test Finished  *********************\n\n"));
    return(RET_OK);
}


/***********************************************************************************************
 * name:      rdma_lat_test_pollcq
 * args:
 *            params_p -- test parameter structure, need to filled before calling
 * returns:
 *            RET_OK
 *            RET_ERR
 *
 * description:     
 *            This test, in each iteration loop, send RDMA_WRITE and poll for completion
 *            
 ************************************************************************************************/  
ret_val_t rdma_lat_test_pollcq(test_params_tx *params_p)
{

    VAPI_ret_t             res;
    int i;
    int j=0;
    //last_post,comp_inter,ask_comp;
    u_int64_t t_start=0, t_end=0;
    u_int64_t t_start_my=0, t_start_my_prev=0, delta_sum=0;
    //int my_iter = 0;
    int my_iter_good;
    float bad_average_double;

    VAPI_sr_desc_t         sr;
    VAPI_sg_lst_entry_t    sg_entry_s;
    VAPI_wc_desc_t         sc;

    VAPI_hca_hndl_t        hca_hndl    = params_p->hca_hndl;
    VAPI_qp_hndl_t         qp_hndl     = params_p->port[CUR_PORT_IX].qp_vect[0].qp_hndl;

    VAPI_cq_hndl_t         s_cq_hndl   = params_p->port[CUR_PORT_IX].s_cq_hndl;
    MOSAL_semaphore_t    * sq_sem_p   = &params_p->port[CUR_PORT_IX].sq_sem;
    //    u_int8_t               *recv_buf   = params_p->recv_buf;
    u_int8_t               *send_buf   = params_p->send_buf;
    u_int64_t               *delta = NULL;
    int                    is_sender   = params_p->in_params.is_sender;
    int                    size        = params_p->in_params.size;
    int                    iter        = params_p->in_params.iter;      
    //u_int32_t * qp_sent_count = calloc(sizeof(u_int32_t), params_p->in_params.num_of_qps);
    //    int                    skip        = params_p->in_params.skip;
    //volatile char          *check_p;
    //int index_to_check=0;
    //    int act_skip;

    //int max_posible = MIN(params_p->in_params.num_s_wqe,params_p->in_params.num_cqe);

    if (size > params_p->in_params.mr_recv_size) {
        printf("Warning!!!: size=%d > mr_recv_size=%d --- changing iter to %d\n", 
               size,params_p->in_params.mr_recv_size,params_p->in_params.mr_recv_size);
        size = params_p->in_params.mr_recv_size;
    }

    if (size==0) {
        printf("For RDMA size must be >= 1. Changing size to 1\n");
        size =1;
    }

/***************************************************************************/
    //sr.comp_type = VAPI_UNSIGNALED;
    sr.comp_type = VAPI_SIGNALED; //Noam
    sr.set_se = FALSE;

    sr.opcode = VAPI_RDMA_WRITE; //Noam
    sr.remote_qkey = 0;
    sr.remote_addr = params_p->remote_add;
    sr.r_key = params_p->remote_rkey;
    sr.id = 1;
    sr.fence = 0;

    sr.sg_lst_len = 1;
    sg_entry_s.len = size;
    sg_entry_s.lkey = params_p->send_lkey;
    sg_entry_s.addr = (VAPI_virt_addr_t)(MT_virt_addr_t)send_buf; 
    sr.sg_lst_p = &sg_entry_s;


    sock_sync_msg(params_p->in_params.socket, is_sender);
    COND_PRINTF1(("\n\n************* RDMA-Write Latency Test from Post to Poll-CQ started  (Messages up to %d use inline data)  ***************\n\n",
                  params_p->max_inline_size));

    //if ((iter+skip) <= params_p->in_params.num_s_wqe) {
    //    comp_inter = params_p->in_params.num_s_wqe;
    // }
    //else {
    //    comp_inter = params_p->in_params.num_s_wqe / 2 - 1;
    //}

    // We start the timing AFTER the first send
    //act_skip = skip > 0 ? (skip -1) : skip;

    //ask_comp=0;
    //index_to_check = size -1 ;
    //check_p = &recv_buf[index_to_check];
    sg_entry_s.len = size;

    if (is_sender) {
        delta = malloc(iter * 10);
        get_clk64(t_start);


        for (i=0; i<(iter); i++) {

            qp_hndl = params_p->port[CUR_PORT_IX].qp_vect[j].qp_hndl;
            
            //if (i > 0) {
                t_start_my_prev = t_start_my;
            //}

            get_clk64(t_start_my);

            // Remove the if statement to save cpu time. Later ignore delta[0].
            //if (i > 0) {
                //printf("RoundTrip Latency: %.2f usec\n",CLK_TO_USEC(t_start_my - t_start_my_prev));
                //printf("my_iter %d t_start_my %lld t_start_my_prev %lld\n", my_iter, t_start_my, t_start_my_prev);
                delta[i] = t_start_my - t_start_my_prev;
                //my_iter++;
            //}

            COND_PRINTF(("Posted send desc %d to" U64_FMT " (size=%d val[0]=%d)\n",i,sr.remote_addr,
                         sr.sg_lst_p[0].len, *((int *)((unsigned long)sr.sg_lst_p[0].addr))));
            res = VAPI_post_sr(hca_hndl, qp_hndl, &sr);
            if (res != VAPI_OK) {
                printf("Error: Posting send desc : %s\n", VAPI_strerror(res));
                return(RET_ERR);
            }

            COND_PRINTF(("Polling for sr comp i=%d\n",i));
            //printf(("Polling for sr comp \n"));
            //      printf("Use event = %d \n",params_p->in_params.use_send_events);
            if (RET_OK != PERF_poll_cq(hca_hndl,s_cq_hndl ,&sc,params_p->in_params.use_send_events, 
                                       sq_sem_p)) {
                return RET_ERR;
            }
            j = (j+1)%params_p->in_params.num_of_qps;            
        }/*end of iteration loop*/
    };
    if (is_sender) {
        get_clk64(t_end);
        bad_average_double = 1.5 * (CLK_TO_USEC(t_end - t_start) / iter);
        //my_iter_good = my_iter;
        my_iter_good = iter - 1;
        //for (i=0; i<my_iter; i++)
        for (i=1; i<iter; i++)
        {
            if (CLK_TO_USEC(delta[i]) < bad_average_double)
            {
                delta_sum = delta_sum + delta[i];
                //printf("adding delta = %.2f\n", CLK_TO_USEC(delta[i]));
            }
            else
            {
                //printf("skipping delta = %.2f\n", CLK_TO_USEC(delta[i]));
                my_iter_good--;
            }

        }
        //printf("my_iter_good = %d iter = %d \n", my_iter_good, iter);
        if (my_iter_good == 0)
        {
            printf("Latency cannot be measured. Please use more iterations\n");
        }
        else
        {

#if 0
        //printf("OLD Latency: %.2f usec , NEW Latency: %.2f usec [Num. of QPs: %d, Post List: %d, QPC Sets: %d, MPT Sets: %d, Keys: %d, size: %d bytes]\n",
        printf("Latency: %.2f usec [Num. of QPs: %d, Post List: %d, QPC Sets: %d, MPT Sets: %d, Keys: %d, size: %d bytes]\n",
        //       CLK_TO_USEC(t_end - t_start) /(iter* 1.0),
               CLK_TO_USEC(delta_sum) /(my_iter_good* 1.0),
            params_p->in_params.num_of_qps, params_p->in_params.postlist_len, 
            params_p->in_params.qpc_sets, params_p->in_params.mpt_sets, 
            params_p->in_params.num_of_keys, size);
#else
        //printf("Latency: %.2f usec [size: %d bytes]\n",CLK_TO_USEC(t_end - t_start) /( (iter-act_skip)* 2.0),size);
        //printf("Latency: %.2f usec [size: %d bytes]\n",CLK_TO_USEC(t_end - t_start) /( (iter)* 1.0),size);
        printf("Latency: %.2f usec [size: %d bytes]\n",CLK_TO_USEC(delta_sum) /(my_iter_good * 1.0),size);
#endif  
        }
        fflush(stdout);
        free(delta);
    }
    sock_sync_msg(params_p->in_params.socket, is_sender);
    COND_PRINTF1(("\n************* RDMA-Write Latency Test Finished  *********************\n\n"));
    return(RET_OK);
}


/***********************************************************************************************
 * name:      atomic_swap_lat_test
 * args:
 *            params_p -- test parameter structure, need to filled before calling
 * returns:
 *            RET_OK
 *            RET_ERR
 *
 * description:     
 *            This is the main loop for the Atomic  Compare & Swap write latency test. T
 *            
 ************************************************************************************************/  
ret_val_t atomic_swap_lat_test(test_params_tx *params_p)
{

    VAPI_ret_t             res;
    int i,ask_comp = 0,last_post,comp_inter, tmp_cnt = 1;
    u_int64_t t_start=0, t_end=0;
    u_int64_t t_start_my=0, t_start_my_prev=0, delta_sum=0;
    int my_iter_good;
    float bad_average_double;

    VAPI_sr_desc_t         sr;
    VAPI_wc_desc_t         sc;

    VAPI_hca_hndl_t        hca_hndl    = params_p->hca_hndl;
    VAPI_qp_hndl_t         qp_hndl     = params_p->port[CUR_PORT_IX].qp_vect[0].qp_hndl;

    VAPI_cq_hndl_t         s_cq_hndl   = params_p->port[CUR_PORT_IX].s_cq_hndl;
    MOSAL_semaphore_t    * sq_sem_p   = &params_p->port[CUR_PORT_IX].sq_sem;
    volatile u_int64_t              *recv_buf   = (u_int64_t*)(params_p->recv_buf);
    volatile u_int64_t              *send_buf   = (u_int64_t*)(params_p->send_buf);
    u_int64_t               *delta = NULL;
    int                    is_sender   = params_p->in_params.is_sender;
    int                    iter        = params_p->in_params.iter;      
    int                    skip        = params_p->in_params.skip;
    int act_skip;
    VAPI_sg_lst_entry_t    sg_entry_s;

    //Value to compare to
    *recv_buf = 84;

    sr.comp_type = VAPI_UNSIGNALED;
    sr.set_se = FALSE;
    sr.fence = FALSE;

    sr.opcode = VAPI_ATOMIC_CMP_AND_SWP;
    sr.remote_qkey = 0;
    sr.remote_addr = params_p->remote_add;
    sr.r_key = params_p->remote_rkey;
    sr.id = 1;


    sr.sg_lst_len = 1;
    sr.compare_add = 84;
    sr.swap = 85;  //Value to change to if equal

    sg_entry_s.len = 8;
    sg_entry_s.lkey = params_p->send_lkey;
    sg_entry_s.addr = (VAPI_virt_addr_t)(MT_virt_addr_t)(send_buf); 

    sr.sg_lst_p = &sg_entry_s;
    //printf("Before:\n");
    //printf("loc_addr="U64_FMT", loc_len=%d, lkey=%d\n", sr.sg_lst_p->addr, sr.sg_lst_p->len, sr.sg_lst_p->lkey);

    sock_sync_msg(params_p->in_params.socket, is_sender);
    COND_PRINTF1(("\n\n************* Atomic Comp & Swap Latensy Test started ***************\n\n"));

    if ((iter+skip) <= (int)params_p->in_params.num_s_wqe) {
        comp_inter = params_p->in_params.num_s_wqe;
    } else {
        comp_inter = params_p->in_params.num_s_wqe / 2 - 1;
    }

    // We start the timing AFTER the first send
    act_skip = skip > 0 ? (skip -1) : skip;


    // Post the first send req
    if (is_sender) {
        delta = malloc((iter+act_skip) * 10);
        if (skip == 0) {
            get_clk64(t_start);
        }
        //printf("Sender: Posting WR\n");
        res = VAPI_post_sr(hca_hndl, qp_hndl, &sr);
        if (res != VAPI_OK) {
            printf("Error: Posting send desc : %s\n", VAPI_strerror(res));
            return(RET_ERR);
        }

        //TEst
        //ask_comp++;
    }

    for (i=0; i<(iter + act_skip); i++, tmp_cnt++) {
        COND_PRINTF(("Checking data at %p (current=%d)\n",recv_buf,*recv_buf));
        while (*recv_buf==84) {
        }
        COND_PRINTF((".................Got data "U64_FMT" at %p\n",*recv_buf, recv_buf));
        *recv_buf = 84; 
        if ((i == act_skip) && (skip>0) && is_sender ) {
            get_clk64(t_start);
        }
        if (is_sender) {
            t_start_my_prev = t_start_my;
        }

        if (is_sender) {
            get_clk64(t_start_my);
        }

        if (is_sender) {
            //printf("RoundTrip Latency: %.2f usec\n",CLK_TO_USEC(t_start_my - t_start_my_prev));
            //printf("my_iter %d t_start_my %lld t_start_my_prev %lld\n", my_iter, t_start_my, t_start_my_prev);
            delta[i] = t_start_my - t_start_my_prev;
        }

        last_post = (is_sender && i== (iter + skip -2)) || (!is_sender && (i==(iter + skip -1)));
        sr.comp_type = VAPI_UNSIGNALED;           

        //if (((i+1)%comp_inter == 0) || last_post) {
        if ((tmp_cnt == comp_inter) || last_post) {
            tmp_cnt = 0;
            sr.comp_type = VAPI_SIGNALED;
            ask_comp++;
            COND_PRINTF(("Ask comp [ask=%d i=%d ]\n", ask_comp,i+1));
        }

        if (!(is_sender && (i==(iter + skip -1)))) {
            //printf("Posting WR %d\n", i);
            res = VAPI_post_sr(hca_hndl, qp_hndl, &sr);
            if (res != VAPI_OK) {
                printf("Error: Posting send desc : %s\n", VAPI_strerror(res));
                return(RET_ERR);
            }

            //Test
            //  ask_comp++;
        }

        if (ask_comp >1) {
            COND_PRINTF(("Polling for sr comp i=%d ask=%d\n",i+1,ask_comp-1));
            if (RET_OK != PERF_poll_cq(hca_hndl,s_cq_hndl ,&sc,0, 
                                       sq_sem_p)) {
                return RET_ERR;
            }

            ask_comp--;
        }
    }/*end of iteration loop*/

    if (is_sender) {
        get_clk64(t_end);
        bad_average_double = 2.0 * (CLK_TO_USEC(t_end - t_start) / iter);
        //my_iter_good = my_iter;
        my_iter_good = iter - 1;
        for (i=act_skip+1; i<iter+act_skip; i++)
        {
            if (CLK_TO_USEC(delta[i]) < bad_average_double)
            {
                delta_sum = delta_sum + delta[i];
                //printf("adding delta = %.2f\n", CLK_TO_USEC(delta[i]));
            }
            else
            {
                //printf("skipping delta = %.2f\n", CLK_TO_USEC(delta[i]));
                my_iter_good--;
            }

        }
        //printf("my_iter_good = %d iter = %d\n", my_iter_good, iter);
        if (my_iter_good == 0)
        {
            printf("Latency cannot be measured. Please use more iterations\n");
        }
        else
        {
#if 0
        //printf("Latency: %.2f usec/operation NEW Latency: %.2f usec/operation [Num. of QPs: %d, Post List: %d, QPC Sets: %d, MPT Sets: %d, Keys: %d, %d operations]\n",
        printf("Latency: %.2f usec/operation [Num. of QPs: %d, Post List: %d, QPC Sets: %d, MPT Sets: %d, Keys: %d, %d operations]\n",
               //CLK_TO_USEC(t_end - t_start) / (iter* 2.0),
               CLK_TO_USEC(delta_sum) / (my_iter_good * 2.0),
            params_p->in_params.num_of_qps, params_p->in_params.postlist_len, 
            params_p->in_params.qpc_sets, params_p->in_params.mpt_sets, 
            params_p->in_params.num_of_keys, (iter-act_skip)* 2);
#else
        //printf("Latency: %.2f usec/operation [performed %d operations] \n",CLK_TO_USEC(t_end - t_start) /( (iter-act_skip)* 2.0), (iter-act_skip)* 2);
        printf("Latency: %.2f usec/operation [performed %d operations] \n",CLK_TO_USEC(delta_sum) / (my_iter_good * 2.0), (iter-act_skip)* 2);
#endif  
        }
        fflush(stdout);
    }

    // Poll for the last SR complition
    while (ask_comp>0) {
        COND_PRINTF(("Final: Polling for last sr comp i=%d ask=%d\n",i,ask_comp-1));
        if (RET_OK != PERF_poll_cq(hca_hndl,s_cq_hndl ,&sc,0, 
                                   sq_sem_p)) {
            while (1) {
            }
            return RET_ERR;
        }
        ask_comp--;
    } 
    sock_sync_msg(params_p->in_params.socket, is_sender);

    COND_PRINTF1(("\n************* Atomic Comp & Swap Latency Test Finished  *********************\n\n"));
    return(RET_OK);
}


/***********************************************************************************************
 * name:      send_lat_test
 * args:
 *            params_p -- test parameter structure, need to filled before calling
 * returns:
 *            RET_OK
 *            RET_ERR
 *
 * description:     
 *            This is the main loop for the Send write latency test. T
 *            
 ************************************************************************************************/  
ret_val_t send_lat_test(test_params_tx *params_p)
{

    VAPI_ret_t             res;
    int send_id,i,ask_comp,last_post,comp_inter;
    int j=0;

    u_int64_t t_start=0, t_end=0;
    u_int64_t t_start_my=0, t_start_my_prev=0, delta_sum=0;
    //int my_iter = 0;
    int my_iter_good;
    float bad_average_double;


    VAPI_rr_desc_t         rr;
    VAPI_sg_lst_entry_t    sg_entry_r;
    VAPI_sr_desc_t         sr;
    VAPI_sg_lst_entry_t    sg_entry_s;
    VAPI_wc_desc_t         rc;
    VAPI_wc_desc_t         sc;

    u_int32_t              srq_posted_actual;

    // Copy all input parameters - don't want to rewrite the loop code
    VAPI_hca_hndl_t        hca_hndl    = params_p->hca_hndl;
    VAPI_qp_hndl_t         qp_hndl     = params_p->port[CUR_PORT_IX].qp_vect[0].qp_hndl;
    VAPI_srq_hndl_t        srq_hndl    = params_p->port[CUR_PORT_IX].qp_vect[0].srq_hndl;

    VAPI_cq_hndl_t         s_cq_hndl   = params_p->port[CUR_PORT_IX].s_cq_hndl;
    VAPI_cq_hndl_t         r_cq_hndl   = params_p->port[CUR_PORT_IX].r_cq_hndl;
    MOSAL_semaphore_t    * sq_sem_p   = &params_p->port[CUR_PORT_IX].sq_sem;
    MOSAL_semaphore_t    * rq_sem_p   = &params_p->port[CUR_PORT_IX].rq_sem;
    u_int8_t               *recv_buf   = params_p->recv_buf;
    u_int8_t               *send_buf   = params_p->send_buf;
    u_int64_t               *delta = NULL;
    int                    is_sender   = params_p->in_params.is_sender;
    int                    size        = params_p->in_params.size;
    int                    iter        = params_p->in_params.iter;      
    int                    rec_buf_num = params_p->in_params.rec_buf_num;      
    int                    skip        = params_p->in_params.skip;
    int act_skip;
    int index_to_check=0;
    volatile char          *check_p;
    u_int32_t * qp_sent_count = calloc(sizeof(u_int32_t), params_p->in_params.num_of_qps);
    int use_inline = FALSE;

    //Value to compare to
    //*recv_buf = 84; // SHIMON

    send_id = 0;
    if (params_p->in_params.transport == IB_TS_UD) {
        if (size > 2048) {
            printf("Warning!!!: UD max size is 2048\n");
            size = 2048;
        }
    }

    if (size > params_p->in_params.mr_recv_size) {
        printf("Warning!!!: size=%d > mr_recv_size=%d --- changing iter to %d\n", 
               size,params_p->in_params.mr_recv_size,params_p->in_params.mr_recv_size);
        size = params_p->in_params.mr_recv_size;
    }

    if ((params_p->in_params.transport == IB_TS_UD) &&  params_p->in_params.use_mwindow) {
        printf("Error!!!: Can't bind window in UD!!!!\n");
        return RET_ERR;

    }

    rr.opcode = VAPI_RECEIVE;
    rr.comp_type = VAPI_SIGNALED;
    rr.sg_lst_len = 1;
    sg_entry_r.lkey = params_p->recv_lkey;
    sg_entry_r.len = params_p->in_params.mr_recv_size;
    //printf("len = %d\n", params_p->in_params.mr_recv_size);
    sg_entry_r.addr = (VAPI_virt_addr_t)(MT_virt_addr_t)recv_buf;
    rr.sg_lst_p  = &sg_entry_r;

    rec_buf_num = MIN2(rec_buf_num,(int)params_p->in_params.num_r_wqe);

    for (j=0;j<(int)params_p->in_params.num_of_qps;j++) 
    {
        qp_hndl = params_p->port[CUR_PORT_IX].qp_vect[j].qp_hndl;
        srq_hndl = params_p->port[CUR_PORT_IX].qp_vect[j].srq_hndl;
        for (i=0; i< rec_buf_num; i++) {
            rr.id = i;
            if (params_p->in_params.srq_num > 0)
            {
                res = VAPI_post_srq(hca_hndl, srq_hndl, 1, &rr, &srq_posted_actual);
                //printf("posted %d SRQs\n", srq_posted_actual);
            }
            else
            {
                res = VAPI_post_rr(hca_hndl, qp_hndl, &rr);
            }
            if (res != VAPI_OK) {
                printf("Error: Posting Recv buffer : %s\n", VAPI_strerror(res));
                return(RET_ERR);
            }
            COND_PRINTF(("Posted RR id=" U64_FMT " i=%d\n",rr.id,i));
        }
    }


/***************************************************************************/
    sr.comp_type = VAPI_UNSIGNALED;
    sr.set_se = FALSE;

    sr.opcode = VAPI_SEND;
    sr.remote_qkey = 0;
    sr.id = 1;

    //needed only for UD
    if (params_p->in_params.multicast > 0)
    {
        sr.remote_qp = IB_MULTICAST_QP;
    }
    else
    {
        sr.remote_qp = params_p->port[CUR_PORT_IX].qp_vect[0].d_qpnum; 
    }
    sr.remote_ah = params_p->port[CUR_PORT_IX].av_hndl;

    sr.sg_lst_len = 1;
    sg_entry_s.len = size;
    sg_entry_s.lkey = params_p->send_lkey;
    sg_entry_s.addr = (VAPI_virt_addr_t)(MT_virt_addr_t)send_buf; 
    sr.sg_lst_p = &sg_entry_s;

    /* sync start test */
    sock_sync_msg(params_p->in_params.socket, is_sender);
    COND_PRINTF1(("\n\n************* %s Latency Test started  (Messages up to %d use inline data)  ***************\n\n",
                  params_p->in_params.transport == IB_TS_UD ? "UD" :
                  params_p->in_params.transport == IB_TS_UC ? "UC": "RC",
                  params_p->max_inline_size));


    if ((iter+skip) <= (int)params_p->in_params.num_s_wqe) {
        comp_inter = params_p->in_params.num_s_wqe;
    } else {
        comp_inter = (params_p->in_params.num_s_wqe / 2) - 1;
    }

    comp_inter = MAX(comp_inter, 1);
    ask_comp =0;

    if (size==0) {
        sr.sg_lst_len = 0;
    }

    // We start the timing AFTER the first send
    act_skip = skip > 0 ? (skip -1) : skip;

    sg_entry_s.len = size;

    send_id = 1;
    
    *send_buf = 55; // SHIMON
    *recv_buf = 0;            // SHIMON
    j=0;
    qp_hndl = params_p->port[CUR_PORT_IX].qp_vect[0].qp_hndl;

    if ((size >0 ) &&(size <= params_p->max_inline_size))
    {
        use_inline = TRUE;
    }
    
    // Post the first send req
    if (is_sender) {
        delta = malloc(iter * 10);
        if (skip == 0) {
            get_clk64(t_start);
        }
        //if ((size >0 ) &&(size <= params_p->max_inline_size)) {
        if (use_inline == TRUE) {
            res = EVAPI_post_inline_sr(hca_hndl, qp_hndl, &sr);
        } else {
            res = VAPI_post_sr(hca_hndl, qp_hndl, &sr);
        }
        if (res != VAPI_OK) {
            printf("Error: Posting FIRST send desc : %s\n", VAPI_strerror(res));
            return(RET_ERR);
        }
        COND_PRINTF(("Posted send desc %d\n", send_id++));
        qp_sent_count[0] += 1;
        j = (j+1)%params_p->in_params.num_of_qps;            
    }//End of ONLY sender

    index_to_check = size -1 ;
    check_p = &recv_buf[index_to_check];
    //recv_buf[index_to_check] = 84;            // SHIMON

    for (i=0; i<(iter + act_skip); i++) {
        qp_hndl = params_p->port[CUR_PORT_IX].qp_vect[j].qp_hndl;
        
        while (1) {
            if (RET_OK != PERF_poll_cq(hca_hndl,r_cq_hndl ,&rc,params_p->in_params.use_recv_events, 
                                    rq_sem_p)) {
                return RET_ERR;
            }

            // For multicast
            if ((params_p->in_params.multicast == 0) ||
                (rc.remote_node_addr.slid != params_p->port[CUR_PORT_IX].lid)){
                break;
            }
        }

        //while (*check_p==84) {    // SHIMON
        //}                          // SHIMON
        //recv_buf[index_to_check] = 84;            // SHIMON


        if ((i == act_skip) && (skip>0) && is_sender ) {
            get_clk64(t_start);
        }
        
        if (is_sender ) {
            t_start_my_prev = t_start_my;
            get_clk64(t_start_my);
            //printf("RoundTrip Latency: %.2f usec\n",CLK_TO_USEC(t_start_my - t_start_my_prev));
            //printf("my_iter %d t_start_my %lld t_start_my_prev %lld\n", my_iter, t_start_my, t_start_my_prev);
            //delta[my_iter] = t_start_my - t_start_my_prev;
            delta[i] = t_start_my - t_start_my_prev;
            //my_iter++;
        }

        last_post = (is_sender && i== (iter + skip -2)) || (!is_sender && (i==(iter + skip -1)));

        sr.comp_type = VAPI_UNSIGNALED;           

        if (((i+1)%comp_inter == 0) || last_post ||
            ((qp_sent_count[j]) % comp_inter == 0) || 
            (qp_sent_count[j] == (iter/params_p->in_params.num_of_qps-1))
            ) {
            sr.comp_type = VAPI_SIGNALED;
            ask_comp++;
            COND_PRINTF(("Ask comp [ask=%d i=%d ]\n", ask_comp,i+1));
        }

        // Only receiver post the last one
        if (!(is_sender && (i==(iter + skip -1)))) {
            //if ((size >0) &&(size <= params_p->max_inline_size)) {
            if (use_inline == TRUE) {
                res = EVAPI_post_inline_sr(hca_hndl, qp_hndl, &sr);
            } else {
                res = VAPI_post_sr(hca_hndl, qp_hndl, &sr);
            }
            if (res != VAPI_OK) {
                printf("Error: Posting send desc : %s\n", VAPI_strerror(res));
                return(RET_ERR);
            }
            COND_PRINTF(("Posted send desc %d\n" , send_id++));
            qp_sent_count[j] += 1;
            j = (j+1)%params_p->in_params.num_of_qps;            
        }

        //if (RET_OK != PERF_poll_cq(hca_hndl,r_cq_hndl ,&rc,params_p->in_params.use_recv_events, 
        //                        rq_sem_p)) {
        //    return RET_ERR;
        //}
        
        rr.id = rc.id;
        if (params_p->in_params.srq_num > 0)
        {
            srq_hndl = params_p->port[CUR_PORT_IX].qp_vect[j].srq_hndl;
            res = VAPI_post_srq(hca_hndl, srq_hndl, 1, &rr, &srq_posted_actual);
            //printf("posted %d SRQs\n", srq_posted_actual);
        }
        else
        {
            res = VAPI_post_rr(hca_hndl, qp_hndl, &rr);
        }
        if (res != VAPI_OK) {
            printf("Error: Posting Recv buffer : %s\n", VAPI_strerror(res));
            return(RET_ERR);
        }
        COND_PRINTF(("  Posted RR id=" U64_FMT " i=%d\n",rr.id,i));

        // In multicast, do an additional post_rr
        if (params_p->in_params.multicast) {
            res = VAPI_post_rr(hca_hndl, qp_hndl, &rr);
            if (res != VAPI_OK) {
                printf("Error: Posting Recv buffer : %s\n", VAPI_strerror(res));
                return(RET_ERR);
            }
            COND_PRINTF(("  Posted RR id=" U64_FMT " i=%d\n",rr.id,i));
        }
        
        //poll for SR
        if (ask_comp >1) {
            COND_PRINTF(("**** Polling for sr comp i=%d ask=%d\n",i+1,ask_comp));
            if (RET_OK != PERF_poll_cq(hca_hndl,s_cq_hndl ,&sc,params_p->in_params.use_send_events, 
                                       sq_sem_p)) {
                return RET_ERR;
            }
            ask_comp--;
        }

    }// end of iteration loop - for i

    if (is_sender) {
        get_clk64(t_end);
        bad_average_double = 1.5 * (CLK_TO_USEC(t_end - t_start) / iter);
        //my_iter_good = my_iter;
        my_iter_good = iter - 1;
        //for (i=0; i<my_iter; i++)
        for (i=act_skip+1; i<iter+act_skip; i++)
        {
            if (CLK_TO_USEC(delta[i]) < bad_average_double)
            {
                delta_sum = delta_sum + delta[i];
                //printf("adding delta = %.2f\n", CLK_TO_USEC(delta[i]));
            }
            else
            {
                //printf("skipping delta = %.2f\n", CLK_TO_USEC(delta[i]));
                my_iter_good--;
            }

        }
        //printf("my_iter_good = %d iter = %d\n", my_iter_good, iter);
        if (my_iter_good == 0)
        {
            printf("Latency cannot be measured. Please use more iterations\n");
        }
        else
        {
#if 0
        //printf("OLD Latency: %.2f usec New Latency: %.2f usec [Num. of QPs: %d, Post List: %d, QPC Sets: %d, MPT Sets: %d, Keys: %d, size: %d bytes]\n",
        printf("Latency: %.2f usec [Num. of QPs: %d, Post List: %d, QPC Sets: %d, MPT Sets: %d, Keys: %d, size: %d bytes]\n",
        //printf("Latency: %.2f usec [Num. of QPs: %d, Post List: %d, QPC Sets: %d, MPT Sets: %d, Keys: %d, size: %d bytes]\n",
               //CLK_TO_USEC(t_end - t_start) /(iter* 2.0),
               CLK_TO_USEC(delta_sum) / (my_iter_good * 2.0),
            params_p->in_params.num_of_qps, params_p->in_params.postlist_len, 
            params_p->in_params.qpc_sets, params_p->in_params.mpt_sets, 
            params_p->in_params.num_of_keys, size);
#else
        //printf("Latency: %.2f usec [size: %d bytes]\n",CLK_TO_USEC(t_end - t_start) /(iter* 2.0),size);
        
        printf("Latency: %.2f usec [size: %d bytes]\n",CLK_TO_USEC(delta_sum) / (my_iter_good * 2.0),size);
#endif        
        }
        MTPERF_REPORT_PRINTF(VAPI_poll_cq);
        fflush(stdout);
    }

    // Poll for the last SR complition
    while (ask_comp>0) {
        COND_PRINTF(("**** Polling for last sr comp i=%d ask=%d\n",i,ask_comp));
        if (RET_OK != PERF_poll_cq(hca_hndl,s_cq_hndl ,&sc,params_p->in_params.use_send_events, 
                                   sq_sem_p)) {
            return RET_ERR;
        }
        ask_comp--;
        COND_PRINTF(("**** Polling for sr comp %d %d\n",i,ask_comp));
    }


    COND_PRINTF1(("\n************* %s Latency Test finished  *********************\n\n",
                  params_p->in_params.transport == IB_TS_UD ? "UD" : 
                  params_p->in_params.transport == IB_TS_UC ? "UC" : "RC"));
    sock_sync_msg(params_p->in_params.socket, is_sender);
    return(RET_OK);
}
