/*
  This software is available to you under a choice of one of two
  licenses.  You may choose to be licensed under the terms of the GNU
  General Public License (GPL) Version 2, available at
  <http://www.fsf.org/copyleft/gpl.html>, or the OpenIB.org BSD
  license, available in the LICENSE.TXT file accompanying this
  software.  These details are also available at
  <http://openib.org/license.html>.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
  BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
  ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
  CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
  SOFTWARE.

  Copyright (c) 2004 Mellanox Technologies Ltd.  All rights reserved.
*/

static char perf_main_cvsid[] = "$Id: main_test.c,v 1.116 2004/05/23 11:32:48 boris Exp $";


#include <stdio.h>

#define DEBUG_PRINT // must be before utils.h
//#define TIME_LEVEL 0
#include "utils.h" 
#include "sock.h" 
#include "cmd_pars.h"
#include "all_bw_lat.h"


static int  get_input_params(int argc , char **argv,test_params_tx *params);
static void init_common_params(test_params_tx *params);
static void init_params(test_params_tx *params);
static void clean_up(test_params_tx *params);
static void clean_up_no_exit(test_params_tx *params);
static void communicate_params_mqp(test_params_tx *params);
static ret_val_t run_2port_test( test_mode_t test_mode, test_2ports_p test_p, test_params_tx *params );

void   print_test_params(test_params_tx *params, char *src);

u_int64_t cpu_CLCKS_PER_SECS;
FILE *res_file = NULL;
int silent_print;
int debug_print;

int MT_API main(int argc , char **argv)
{
    test_params_tx params;
    int socket = 0;
    int rc=0;
    static char *perf_silence;
    int i;
    silent_print = 0;

    perf_silence = getenv("PERF_SILENT");

    if (perf_silence != NULL)
    {
        if (strncmp(perf_silence,"silence", 8)== 0)
        {
            silent_print = 1;
        } else if (strncmp(perf_silence,"debug", 5)== 0)
        {
            debug_print  = 1;
        }
    }

    update_CPU_CLCKS_PER_SECS();

    COND_PRINTF1(("\n********************************************\n"));
    COND_PRINTF1((  "*********  perf_main version 10.3  *********\n"));
    DEBUG_PRINTF((  "*********  cvs id: %s ***\n", perf_main_cvsid )); 
    COND_PRINTF1((  "*********  CPU is: %.2f Mcps    *********\n",
                    (1.0*((ADJUST_TYPE)cpu_CLCKS_PER_SECS))/1000000.0));
#ifdef __i386__
    COND_PRINTF1((   "*********  Architecture X86     *********\n"));
#endif  
#ifdef __ia64__
    COND_PRINTF1((   "*********  Architecture IA64    *********\n"));
#endif
#ifdef __x86_64__
    COND_PRINTF1((   "*********  Architecture X86_64  *********\n"));
#endif
    COND_PRINTF1((  "********************************************\n\n"));


    /* Initialize to deafualt parameters */
    //printf("b init_common\n");
    init_common_params(&params); 
    //printf("a init_common\n");

    do
    {

        /* Get comand line parameters */
        //printf("b get_input\n");
        if (0 != (rc = get_input_params(argc,argv,&params)))
        {
            return -1;
        }
        //printf("a get_input\n");

        /* Initialize parameters according to test type BEFORE creating resources*/
        //printf("b init_param\n");
        init_params(&params);
        //printf("a init_param\n");
        socket = params.in_params.socket;

        /* Create IB resources */
        if (RET_OK != create_common_resources(&params))
        {
            clean_up(&params);
        }

	/* Fix the MTU if the user didn't change the MTU using the command line */
	if (params.in_params.mtu == 0)
	{
		/* for device with ID 23108 use 1K MTU, for other devices use 2K MTU */
		params.in_params.mtu = (params.hca_vendor.vendor_part_id == 23108) ? MTU1024 : MTU2048;
	}

        //printf("a create_common_resource\n");
        
        FOR_ALL_PORTS( &params, i) 
        if (RET_OK != qp_create(&params, i))
        {
            clean_up(&params);
        }
        END_FOR

        /* communicate QP parameters between receiver & sender nodes over IP*/
        communicate_params_mqp(&params);
        //printf("b communicate\n");

        /* Simulate CM using IP sockets */
        FOR_ALL_PORTS( &params, i) 
        if (RET_OK != qp_init2rts_mqp(&params, i))
        {
            clean_up(&params);
        }
        END_FOR


        /* Create address vector in case of UD */
        if (params.in_params.transport == IB_TS_UD)
        {
            FOR_ALL_PORTS( &params, i) 
            if (RET_OK != create_av(&params, i))
            {
                clean_up(&params);
            }
            END_FOR
        }

#ifdef PRINT_PARAMS
        print_test_params(&params, "xxx" );
#endif	

        //printf("Test mode %d\n", params.in_params.test_mode);
        //printf("grh %d\n", params.in_params.grh);
        switch (params.in_params.test_mode)
        {
        case LAT_ATOM_CS:
            atomic_swap_lat_test(&params);
            break;
        case BW_ATOM_CS:
            atomic_bw_test(&params);
            break;
        case BW_LAT_ATOM_CS:
            atomic_bw_lat_test(&params);
            break;
        case LAT_SEND:
            send_lat_test(&params); 
            break;
        case LAT_RDMA_W:
            rdma_lat_test(&params);
            break;
        case LAT_RDMA_W_POLLCQ:
            rdma_lat_test_pollcq(&params);
            break;
        case LAT_RDMA_R:
            rdma_read_lat_test(&params);
            break;
        case BW_SEND:
            send_bw_test(&params);
            break;
        case BW_SEND_LIST:
            send_bw_list_test(&params);
            break;
        case BW_RDMA_W:
            rdma_bw_test(&params);
            break;
        case BW_SEND_UNI:
//            printf("BW_SEND_UNI test\n");        
            run_2port_test( BW_SEND_UNI, send_bw_uni_test, &params );
            break;
        case BW_RDMA_W_UNI:
//            printf("BW_RDMA_W_UNI test\n");     
//            if (params.in_params.many2one) {
//                printf("BW_RDMA_W_UNI Many to One test\n");     
//                run_2port_test( BW_RDMA_W_UNI, rdma_w_bw_uni_many2one_test, &params );
//            }
//            else {
                run_2port_test( BW_RDMA_W_UNI, rdma_w_bw_uni_test, &params );
//            }
            break;
        case WQE_TEST:
            run_2port_test( WQE_TEST, wqe_test, &params );
            break;
        case BW_RDMA_R_UNI:
            run_2port_test( BW_RDMA_R_UNI, rdma_r_bw_uni_test, &params );
            break;
        case BW_SEND_BI:
            run_2port_test( BW_SEND_BI, send_bw_bi_test, &params );
            break;
        case BW_RDMA_W_BI:
            run_2port_test( BW_RDMA_W_BI, rdma_w_bw_bi_test, &params );
            break;
        case BW_RDMA_R_BI:
            run_2port_test( BW_RDMA_R_BI, rdma_r_bw_bi_test, &params );
            break;
        case CHIPSET1:
            chipset_test1(&params);
            break;
        case SPECIAL:
            test_case_special(&params);
            break;
        case BW_MQP:
            break;
        case BW_BIDIR:
            break;
        case SPECIAL_2PORT:
            send_bw_uni_test_2ports(&params);
            break;
        default:
            printf("Error - unsupported test case: %d\n", params.in_params.test_mode);
            break;
        }

        clean_up_no_exit(&params);

    } while (!params.in_params.is_sender && params.in_params.is_server);


    if (res_file != NULL)
    {
        fclose(res_file);
    }
    return(0);
}


/*************************************************************************
 * Function: clean_up
 *
 * Description:
 *   Clean all IB reasources and memory allocations
 *
 *************************************************************************/ 
static void clean_up_no_exit(test_params_tx *params)
{
    int jj;
    int i;
    COND_PRINTF2(("Cleaning up VAPI resources\n\n"));
    FOR_ALL_PORTS( params, i) 
    for (jj=0;jj<(int)params->in_params.num_of_qps;jj++)
    {
        if (params->port[i].qp_vect[jj].qp_hndl != VAPI_INVAL_HNDL)
        {
            if (params->in_params.multicast > 0)
            {
                IB_gid_t gid;
                IB_lid_t mcg_lid;
                
                if (params->in_params.is_sender)
                {
                //    ASSIGN_MGID_0;
                    mcg_lid = MIN_MC_LID;
                }
                else // receiver
                {
                //    ASSIGN_MGID_1;
                    mcg_lid = MIN_MC_LID+1;
                }

                ASSIGN_MGID_0;
                VAPI_detach_from_multicast(params->hca_hndl, gid, params->port[i].qp_vect[jj].qp_hndl, mcg_lid);
            }
            VAPI_destroy_qp(params->hca_hndl,params->port[i].qp_vect[jj].qp_hndl);
        }
    }
    if (params->in_params.srq_num > 0)
    {
        for (jj=0;jj<params->in_params.srq_num;jj++)
        {
            if (params->port[i].srq_vect[jj].srq_hndl != VAPI_INVAL_SRQ_HNDL)
            {
                VAPI_destroy_srq(params->hca_hndl, params->port[i].srq_vect[jj].srq_hndl);
            }
        }
    }
    if (params->port[i].rq_cq_eventh != VAPI_INVAL_HNDL)
    {
        EVAPI_clear_comp_eventh(params->hca_hndl,params->port[i].rq_cq_eventh);
        params->port[i].rq_cq_eventh = VAPI_INVAL_HNDL; // Fixes bug 28945
    }
    if (params->port[i].sq_cq_eventh != VAPI_INVAL_HNDL)
    {
        EVAPI_clear_comp_eventh(params->hca_hndl,params->port[i].sq_cq_eventh);
        params->port[i].sq_cq_eventh = VAPI_INVAL_HNDL; // Fixes bug 28945
    }
    if (params->port[i].r_cq_hndl != VAPI_INVAL_HNDL)
    {
        VAPI_destroy_cq(params->hca_hndl,params->port[i].r_cq_hndl);
    }
    if (params->port[i].s_cq_hndl != VAPI_INVAL_HNDL)
    {
        VAPI_destroy_cq(params->hca_hndl,params->port[i].s_cq_hndl);
    }
    END_FOR

    for (jj=0;jj<(int)params->in_params.num_of_keys;jj++) {

        if (params->mr_hndl_s_arr[jj] != VAPI_INVAL_HNDL)
        {
            VAPI_deregister_mr(params->hca_hndl,params->mr_hndl_s_arr[jj]);
        }
        if (params->mr_hndl_r_arr[jj] != VAPI_INVAL_HNDL)
        {
            VAPI_deregister_mr(params->hca_hndl,params->mr_hndl_r_arr[jj]);
        }
    }

    FOR_ALL_PORTS( params, i) 
    if (params->port[i].av_hndl != VAPI_INVAL_HNDL)
    {
        VAPI_destroy_addr_hndl(params->hca_hndl,params->port[i].av_hndl);
    }
    END_FOR
    if (params->in_params.use_devmem > 0)
    {
        if (params->dm_hndl_r != VAPI_INVAL_HNDL)
        {
            EVAPI_free_unmap_devmem(params->hca_hndl,params->dm_hndl_r);
        }
        if (params->dm_hndl_s != VAPI_INVAL_HNDL)
        {
            EVAPI_free_unmap_devmem(params->hca_hndl,params->dm_hndl_s);
        }
    } else
    {
        if (params->real_send_buf != NULL)
        {
            free(params->real_send_buf);
        }
        if (params->real_recv_buf != NULL)
        {
            free(params->real_recv_buf);
        }
    }
    if (params->pd_hndl != VAPI_INVAL_HNDL)
    {
        VAPI_dealloc_pd(params->hca_hndl,params->pd_hndl);
    }
    if (params->hca_hndl != VAPI_INVAL_HNDL)
    {
        EVAPI_release_hca_hndl(params->hca_hndl);  
    }
    if (params->in_params.socket !=0)
    {
        MOSAL_socket_close(params->in_params.socket);
    }
    FOR_ALL_PORTS( params, i) 
    if (params->port[i].qp_vect != NULL)
    {
        free(params->port[i].qp_vect);
    }
    if ((params->in_params.srq_num > 0) && (params->port[i].srq_vect != NULL))
    {
        free(params->port[i].srq_vect);
    }
    END_FOR
    /* thread stuff */
    if (params->in_params.ports_num > 1)
    {
        for (i=0; i<2; i++)
        {
            if (params->port[i].mto_p != NULL)
            {
                MOSAL_thread_kill( params->port[i].mto_p );
            }
        }
    }

    free(params->mr_hndl_s_arr);
    free(params->mr_hndl_r_arr);
    free(params->recv_lkey_arr);
    free(params->recv_rkey_arr);
    free(params->send_lkey_arr);
    free(params->remote_rkey_arr);
}

static void clean_up(test_params_tx *params) {
    clean_up_no_exit(params);
    exit(1);
}



/*************************************************************************
 * Function: init_common_params
 *
 * Description:
 *   Initialize all IB resources
 *
 *************************************************************************/ 
static void   init_common_params(test_params_tx *params)
{
    int i;

    params->mr_hndl_s   = VAPI_INVAL_HNDL;
    params->mr_hndl_r   = VAPI_INVAL_HNDL;

    params->hca_hndl    = VAPI_INVAL_HNDL;
    params->pd_hndl     = VAPI_INVAL_HNDL;
    params->recv_buf    = NULL;
    params->send_buf    = NULL;
    params->dm_hndl_s   = VAPI_INVAL_HNDL;
    params->dm_hndl_r   = VAPI_INVAL_HNDL;

    params->in_params.socket_port       = 4252;
    params->in_params.server_socket     = 0;
    params->in_params.port_num          = 1;
    params->in_params.skip              = 0;           
    //params->in_params.rec_buf_num       = 50;
    params->in_params.mr_recv_size      = MR_RCV_CNT * MR_UNIT;
    params->in_params.mr_send_size      = MR_SND_CNT * MR_UNIT;
    params->in_params.mtu               = 0;   // 0 mark to use default MTU value
    params->in_params.num_cqe           = 60;
    params->in_params.num_r_wqe         = 50;  // Must be enough for ALL QP'S
    params->in_params.num_s_wqe         = 50;
    params->in_params.num_ssge          = 20; // To support inline data
    params->in_params.num_rsge          = 20;
    params->in_params.alloc_aligned     = 1;
    params->in_params.use_send_events   = 0;
    params->in_params.use_recv_events   = 0;
    params->max_inline_size             = 400; 
    params->in_params.use_mwindow       = FALSE;
    params->in_params.use_devmem        = 0;
    params->in_params.ports_num         = 1;
    params->in_params.bidirect          = 0;
    params->in_params.postlist_len      = 1;
    params->in_params.qpc_sets          = ARBEL_QPC_SETS;
    params->in_params.mpt_sets          = ARBEL_MPT_SETS;
    params->in_params.num_of_keys       = 1;
    params->in_params.sg_list_len       = 1;
    //params->in_params.many2one          = 0;
    
    for (i=0; i<MAX_PORTS_NUM; i++)
    {
        params->port[i].port_num        = i+1;
        params->port[i].av_hndl         = VAPI_INVAL_HNDL;
        params->port[i].lid             = -1;
        params->port[i].d_lid           = -1;
        //params->port[i].num_of_qps      = 1;
        //params->port[i].sched_period    = 10;
        params->port[i].r_cq_hndl       = VAPI_INVAL_HNDL;
        params->port[i].s_cq_hndl       = VAPI_INVAL_HNDL;
        params->port[i].rq_cq_eventh    = VAPI_INVAL_HNDL;
        params->port[i].sq_cq_eventh    = VAPI_INVAL_HNDL;
    }
}

/*************************************************************************
 * Function: init_params
 *
 * Description:
 *   Initialize resources according to specific test
 *
 *************************************************************************/ 
static void   init_params(test_params_tx *params)
{
    int i;
    switch (params->in_params.test_mode)
    {
    case LAT_SEND:
        params->in_params.skip          = 10;
        params->in_params.num_cqe       = 129;
        params->in_params.num_r_wqe     = 50;  
        params->in_params.num_s_wqe     = 50;
        break;
    case LAT_RDMA_W:
        params->in_params.num_cqe       = 129;
        params->in_params.num_r_wqe     = 50;  
        params->in_params.num_s_wqe     = 100;
        params->in_params.skip          = 17;
        break;
    case LAT_ATOM_CS:
        params->in_params.num_cqe       = 129;
        params->in_params.num_r_wqe     = 50;  
        params->in_params.num_s_wqe     = 50;
        params->in_params.skip          = 1;
        break;
    case LAT_ATOM_FA:
        params->in_params.num_cqe       = 129;
        params->in_params.num_r_wqe     = 50;  
        params->in_params.num_s_wqe     = 50;
        params->in_params.skip          = 1;
        break;
    case LAT_RDMA_R:
    case LAT_RDMA_W_POLLCQ:
        //printf("setting parameters for rdma read latency test\n");
        params->in_params.num_cqe       = 129;
        params->in_params.num_r_wqe     = 32;  
        params->in_params.num_s_wqe     = 32;
        params->in_params.skip          = 1;
        break;
    case SPECIAL:
        params->in_params.skip          = 1;
        break;
    case BW_SEND:
#if defined(__WIN__) || defined(__powerpc__)
        params->in_params.num_cqe       = 2020;
        params->in_params.num_r_wqe     = 2000;  
        params->in_params.num_s_wqe     = 2000;
#else
        params->in_params.num_cqe       = 220;
        params->in_params.num_r_wqe     = 200;  
        params->in_params.num_s_wqe     = 200;
#endif
        break;
    case BW_SEND_LIST:
        params->in_params.num_cqe       = 1100;
        params->in_params.num_r_wqe     = 20;  
        params->in_params.num_s_wqe     = 100;
        break;
    case BW_RDMA_W:
#if defined(__WIN__) || defined(__powerpc__)
        params->in_params.num_cqe       = 2020;
        params->in_params.num_r_wqe     = 2000;  
        params->in_params.num_s_wqe     = 2000;
#else
        params->in_params.num_cqe       = 220;
        params->in_params.num_r_wqe     = 200;  
        params->in_params.num_s_wqe     = 200;
#endif
        break;
    case SPECIAL_2PORT:
    case BW_RDMA_R_UNI:
    case BW_RDMA_W_UNI:
    case BW_SEND_UNI:
#if defined(__WIN__) || defined(__powerpc__)
        params->in_params.num_cqe       = 4040;
        params->in_params.num_r_wqe     = 2000;  
        params->in_params.num_s_wqe     = 2000;
#else
        params->in_params.num_cqe       = 440;
        params->in_params.num_r_wqe     = 200;  
        params->in_params.num_s_wqe     = 200;
#endif
        break;
    case WQE_TEST:
#if defined(__WIN__) || defined(__powerpc__)
        params->in_params.num_cqe       = 4040;
        params->in_params.num_r_wqe     = 2000;  
        params->in_params.num_s_wqe     = 2000;
#else
        params->in_params.num_cqe       = 440;
        params->in_params.num_r_wqe     = 512;  
        params->in_params.num_s_wqe     = 512;
#endif
        break;
    case BW_RDMA_W_BI:
    case BW_RDMA_R_BI:
#if defined(__WIN__) || defined(__powerpc__)
        params->in_params.num_cqe       = 4040;
        params->in_params.num_r_wqe     = 2000;  
        params->in_params.num_s_wqe     = 2000;
#else
        params->in_params.num_cqe       = 440;
        params->in_params.num_r_wqe     = 2000;  
        params->in_params.num_s_wqe     = 2000;
#endif
        FOR_ALL_PORTS(params, i) 
//        params->port[i].num_of_qps = 1;
        END_FOR
        break;
    case BW_SEND_BI:
#if defined(__WIN__) || defined(__powerpc__)
        params->in_params.num_cqe       = 4040;
        params->in_params.num_r_wqe     = 2000;  
        params->in_params.num_s_wqe     = 2000;
#else
        params->in_params.num_cqe       = 440;
        params->in_params.num_r_wqe     = 200;  
        params->in_params.num_s_wqe     = 200;
#endif
        FOR_ALL_PORTS(params, i) 
//        params->port[i].num_of_qps = 1;
        END_FOR
        break;
    case BW_MQP:
        break;
    case BW_BIDIR:
        params->in_params.num_r_wqe     = 10000;  
        params->in_params.num_s_wqe     = 10000;
    case CHIPSET1:
        if (params->in_params.transport != IB_TS_UC)
        {
            printf("chipset1 test is only in UC - changing to UC\n");
            params->in_params.transport = IB_TS_UC;
        }
        break;
    default:
        break;
    }

    //Overide in case of explicit request
    if (params->in_params.use_events >= 0)
    {
        params->in_params.use_send_events = params->in_params.use_events;
        params->in_params.use_recv_events = params->in_params.use_events;
    }

    // Set memory alloc size, in MR_UNIT granularity
    // NOTE: I assume recv and send buffer size are the same
    if (params->in_params.size > params->in_params.mr_send_size || 
        params->in_params.size > params->in_params.mr_recv_size     )
    {

        int mr_units_needed = params->in_params.size / MR_UNIT;
        if (params->in_params.size % MR_UNIT)
        {
            mr_units_needed++;
        }

        if (mr_units_needed > MAX_MR_UNITS_ALLOC)
        {
            mr_units_needed = MAX_MR_UNITS_ALLOC;
            printf("Warning: Given size exceeds max size allowed. Changing size to %ld.\n", ((u_int32_t)MR_UNIT * MAX_MR_UNITS_ALLOC)); 
        }
        params->in_params.mr_recv_size = mr_units_needed * MR_UNIT;
        params->in_params.mr_send_size = mr_units_needed * MR_UNIT;
    }

}

/*************************************************************************
 * Function: communicate_params_p_mqp
 *
 * Description:
 * communicate QP parmaters between receiver & sender nodes over IP
 *
 *************************************************************************/ 
static void communicate_params_mqp(test_params_tx *params_p){

    char sock_msg[1000];
    char tmp_msg[1000];
    char tmp[50];
    int socket = params_p->in_params.socket;
    int jj;
    int i;

    if (params_p->in_params.oneside)
    {
        return;
    }

#ifdef SW_VER_GT_0_2_0
    sprintf(tmp_msg, "lid_p1=%d lid_p2=%d addr="U64_FMT "qp_ous=%d", 
            params_p->port[CUR_PORT_IX].lid,params_p->port[1-CUR_PORT_IX].lid,
            (VAPI_virt_addr_t)(MT_virt_addr_t)params_p->recv_buf, 
            (int)MIN2(DFLT_QP_OUS_RD_ATOM, params_p->hca_cap.max_qp_ous_rd_atom));
#else
#ifdef SW_VER_GT_0_2_0
    sprintf(tmp_msg, "lid=%d addr="U64_FMT, 
            params_p->lid,
            (VAPI_virt_addr_t)(MT_virt_addr_t)params_p->recv_buf);
#else
    sprintf(tmp_msg, "lid=%d addr=0x"U64_FMT, 
            params_p->lid,
            (VAPI_virt_addr_t)(MT_virt_addr_t)params_p->recv_buf);
#endif
#endif
    sock_send_msg(socket, tmp_msg);
    
    sock_get_msg(socket, sock_msg, 1000);
    sock_msg2num(sock_msg, "addr=", 0, &(params_p->remote_add), sizeof(params_p->remote_add));
    sock_msg2num(sock_msg, "qp_ous=", 8, &(params_p->qp_ous_rd_atom), sizeof(params_p->qp_ous_rd_atom));
    sock_msg2num(sock_msg, "lid_p1=", -1, &(params_p->port[CUR_PORT_IX].d_lid), sizeof(params_p->port[CUR_PORT_IX].d_lid));
    sock_msg2num(sock_msg, "lid_p2=", -1, &(params_p->port[1-CUR_PORT_IX].d_lid), sizeof(params_p->port[1-CUR_PORT_IX].d_lid));

    for (jj=0;jj<(int)params_p->in_params.num_of_keys;jj++) {    
        sprintf(tmp_msg, "rkey=%d", params_p->recv_rkey_arr[jj]);
        // printf("Receive R_Key = 0x%x\n", params_p->recv_rkey_arr[jj]);
        sock_send_msg(socket, tmp_msg);                     
        sock_get_msg(socket, sock_msg, 1000);
        sock_msg2num(sock_msg, "rkey=", -1, &(params_p->remote_rkey_arr[jj]), sizeof(VAPI_rkey_t));
        // printf("Remote R_Key = 0x%x\n", params_p->remote_rkey_arr[jj]);
    }

    params_p->remote_rkey = params_p->remote_rkey_arr[0];

    FOR_ALL_PORTS(params_p, i) 
    for (jj=0;jj<(int)params_p->in_params.num_of_qps;jj++)
    {
        sprintf(tmp_msg,"qp%d=%d ",jj,params_p->port[i].qp_vect[jj].qp_num);        
        sock_send_msg(socket, tmp_msg);

        sock_get_msg(socket, sock_msg, 1000);
        sprintf(tmp,"qp%d=",jj);
        sock_msg2num(sock_msg,tmp , -1, &(params_p->port[i].qp_vect[jj].d_qpnum), 
                     sizeof(params_p->port[i].qp_vect[jj].d_qpnum));   

        // printf("Source QP 0x%x connected to Dest. QP 0x%x\n", 
        //        params_p->port[i].qp_vect[jj].qp_num,
        //        params_p->port[i].qp_vect[jj].d_qpnum);
        
        if (params_p->port[i].qp_vect[jj].d_qpnum == -1)         
        {        
            printf("Wrong d_qp for QP %d (0x%x), sock_msg = %s\n",
                   jj, params_p->port[i].qp_vect[jj].qp_num, sock_msg);
            printf("sock_msg2num len = %d\n", 
                   (u_int32_t) sizeof(params_p->port[i].qp_vect[jj].d_qpnum));
        }
    }
    END_FOR

/*    
    FOR_ALL_PORTS(params_p, i) 
    for (jj=0;jj<(int)params_p->port[i].num_of_qps;jj++)
    {
        sock_get_msg(socket, sock_msg, 1000);
        sprintf(tmp,"qp%d=",jj);
        sock_msg2num(sock_msg,tmp , -1, &(params_p->port[i].qp_vect[jj].d_qpnum), 
                     sizeof(params_p->port[i].qp_vect[jj].d_qpnum));   
        
        if (params_p->port[i].qp_vect[jj].d_qpnum == -1)         
        {        
            printf("Wrong d_qp for QP %d (0x%x), sock_msg = %s\n",
                   jj, params_p->port[i].qp_vect[jj].qp_num, sock_msg);
            printf("sock_msg2num len = %d\n", 
                   sizeof(params_p->port[i].qp_vect[jj].d_qpnum));
        }
    }
    END_FOR
*/    
/*
    FOR_ALL_PORTS(params_p, i) 
    for (jj=0;jj<(int)params_p->port[i].num_of_qps;jj++)
    {
        printf("QP 0x%x connected to QP 0x%x\n", params_p->port[i].qp_vect[jj].qp_num,
               params_p->port[i].qp_vect[jj].d_qpnum);
    }
    END_FOR
*/
}


/*************************************************************************
 * Function: connect_to_peer
 *
 * Description:
 *      Establish tcp connection with peer, as a server or as a client,
 *      according to the cmd line parameters:
 *      Sender   - if specifies -aIP flag is a client (new)       , otherwise, it is a server (old)
 *      Receiver - if specifies --server option is a server (new) , otherwise, must give -aIP option (old)
 *
 *************************************************************************/ 
static int connect_to_peer(test_params_tx *params, char* remote_ip, int socket_port, int is_sender, int is_server) {
    int socket;
    if (!is_sender && strlen(remote_ip) < 2 && !is_server)
    {
        printf("\nError: Receiver must specify -a[IP] parameter or --server option\n\n");
        printf("Type perf_main --help for information\n\n");
        return(-1);
    } else if (strlen(remote_ip) >= 2 && is_server)
    {
        printf("\nError:  -a[IP] parameter and --server option must be mutualy exlusive\n\n");
        printf("Type perf_main --help for information\n\n");
        return(-1);
    }

    if (strlen(remote_ip) < 2)
    {
        if (params->in_params.server_socket == 0)
        {
            /* init_server() is called only in first time here  */
            params->in_params.server_socket = perf_server_init(socket_port);
            if (params->in_params.server_socket == -1)
            {
                printf("Error: perf_server_init() on port %d failed\n", socket_port);
                return(-1);
            }
        }

        socket = wait_for_connection(params->in_params.server_socket);
    } else
    {
        socket = create_tcp_connection(remote_ip,socket_port);
    }

    if (socket == -1)
    {
        if (is_server )
        {
            printf("Error: Wait connection on port %d failed\n", socket_port);
        } else
        {
            printf("Error: Couldn't connect to the sender socket\n");
        }
    }

    return socket;
}






/*************************************************************************
 * Function: get_input_params
 *
 * Description:
 *      Get command line arguments
 *
 *************************************************************************/ 
static int   get_input_params(int argc, char **argv, test_params_tx *params)
{
    int socket=-1, is_bw,socket_port;
    char sock_msg[1000];

    cmd_line_params_t cmd_line_params;
    IB_ts_t transport;
    rdma_mode_t rdma_mode;
    VAPI_hca_id_t *hca_id_buf;
    u_int32_t num_of_hcas;
    u_int32_t port_num;

    int rc = 0;

    int i;


    is_bw = 0;
    //Init atomic mode param
    cmd_line_params.atomic_mode = 0;
    
    init_cmd_line_params(&cmd_line_params);
    //print_cmd_line_params(&cmd_line_params);

    if ((rc = parse_argv(argc, argv, &cmd_line_params)))
    {
        return rc;
    }

    port_num = cmd_line_params.hca_port - 1;
    params->in_params.num_of_qps = cmd_line_params.num_of_qps;
    //params->port[port_num].sched_period = cmd_line_params.sched_period;

    // print_cmd_line_params(&cmd_line_params);

    socket_port = cmd_line_params.app_id + 4000;
    if (cmd_line_params.is_sender)
    {
        if (!cmd_line_params.oneside)
        {

            socket = connect_to_peer(params, cmd_line_params.remote_ip, socket_port, cmd_line_params.is_sender, cmd_line_params.is_server);
            if (socket == -1)
            {
                return -1;
            }

            sock_send_msg(socket, "size=%d iter=%d mtu=%d transport=%s rdma=%s test=%s events=%d num_ah=%d "
                          "devmem=%d atomic_mode=%d ports_num=%d bidirect=%d general_param=%d num_of_qps=%d "
                          //"sched_period=%d 
                          "postlist_len=%d num_of_keys=%d qpc_sets=%d mpt_sets=%d srq_num=%d "
                          "multicast=%d grh=%d rwrite_pollcq=%d wqetest=%d"
                          ,
                          cmd_line_params.msg_size,
                          cmd_line_params.iterations,
                          cmd_line_params.mtu,
                          cmd_line_params.transport_type,
                          cmd_line_params.rdma,
                          cmd_line_params.test_type,
                          cmd_line_params.use_events,
                          cmd_line_params.num_ah,
                          cmd_line_params.use_devmem,
                          cmd_line_params.atomic_mode,
                          cmd_line_params.ports_num,
                          cmd_line_params.bidirect,
                          cmd_line_params.general_param,
                          cmd_line_params.num_of_qps,
                          //cmd_line_params.sched_period,
                          cmd_line_params.postlist_len,
                          cmd_line_params.num_of_keys,
                          cmd_line_params.qpc_sets,
                          cmd_line_params.mpt_sets,
                          cmd_line_params.srq_num,
                          cmd_line_params.multicast,
                          cmd_line_params.grh,
                          cmd_line_params.rwrite_pollcq,
                          cmd_line_params.wqetest
                          );
        }
    } else
    {
        socket = connect_to_peer(params, cmd_line_params.remote_ip, socket_port, cmd_line_params.is_sender, cmd_line_params.is_server);
        if (socket == -1)
        {
            return -1;
        }

        sock_get_msg(socket, sock_msg, 1000);
        sock_msg2num(sock_msg, "size=", -1,&cmd_line_params.msg_size , sizeof(cmd_line_params.msg_size));
        sock_msg2num(sock_msg, "iter=", -1, &cmd_line_params.iterations, sizeof(cmd_line_params.iterations));
        sock_msg2num(sock_msg, "mtu=", -1, &cmd_line_params.mtu, sizeof(cmd_line_params.mtu));
        sock_msg2str(sock_msg, "transport=",cmd_line_params.transport_type , cmd_line_params.transport_type);
        sock_msg2str(sock_msg, "rdma=","none" , cmd_line_params.rdma);
        sock_msg2str(sock_msg, "test=",cmd_line_params.test_type , cmd_line_params.test_type);
        sock_msg2num(sock_msg, "events=", -1, &cmd_line_params.use_events, sizeof(cmd_line_params.use_events));
        sock_msg2num(sock_msg, "num_ah=", -1, &cmd_line_params.num_ah, sizeof(cmd_line_params.num_ah));
        sock_msg2num(sock_msg, "devmem=", -1, &cmd_line_params.use_devmem, sizeof(cmd_line_params.use_devmem));
        sock_msg2num(sock_msg, "atomic_mode=", -1, &cmd_line_params.atomic_mode, sizeof(cmd_line_params.atomic_mode));
        sock_msg2num(sock_msg, "ports_num=", 1, &cmd_line_params.ports_num, sizeof(cmd_line_params.ports_num));
        sock_msg2num(sock_msg, "bidirect=", 0, &cmd_line_params.bidirect, sizeof(cmd_line_params.bidirect));
        sock_msg2num(sock_msg, "general_param=", -1, &cmd_line_params.general_param, sizeof(cmd_line_params.general_param));
        sock_msg2num(sock_msg, "num_of_qps=", -1, &cmd_line_params.num_of_qps, sizeof(cmd_line_params.num_of_qps));
        params->in_params.num_of_qps = cmd_line_params.num_of_qps;
        //sock_msg2num(sock_msg, "sched_period=", -1, &cmd_line_params.sched_period, sizeof(cmd_line_params.sched_period));
        sock_msg2num(sock_msg, "postlist_len=", -1, &cmd_line_params.postlist_len, sizeof(cmd_line_params.postlist_len));
        sock_msg2num(sock_msg, "num_of_keys=", -1, &cmd_line_params.num_of_keys, sizeof(cmd_line_params.num_of_keys));
        sock_msg2num(sock_msg, "qpc_sets=", -1, &cmd_line_params.qpc_sets, sizeof(cmd_line_params.qpc_sets));
        sock_msg2num(sock_msg, "mpt_sets=", -1, &cmd_line_params.mpt_sets, sizeof(cmd_line_params.mpt_sets));
        sock_msg2num(sock_msg, "srq_num=", -1, &cmd_line_params.srq_num, sizeof(cmd_line_params.srq_num));
        sock_msg2num(sock_msg, "multicast=", -1, &cmd_line_params.multicast, sizeof(cmd_line_params.multicast));
        sock_msg2num(sock_msg, "grh=", -1, &cmd_line_params.grh, sizeof(cmd_line_params.grh));
        sock_msg2num(sock_msg, "rwrite_pollcq=", -1, &cmd_line_params.rwrite_pollcq, sizeof(cmd_line_params.rwrite_pollcq));
        sock_msg2num(sock_msg, "wqetest=", -1, &cmd_line_params.wqetest, sizeof(cmd_line_params.wqetest));
    }

    // Allocate MR and Key arrays
    params->mr_hndl_s_arr   = VMALLOC(cmd_line_params.num_of_keys * sizeof(VAPI_mr_hndl_t));
    params->mr_hndl_r_arr   = VMALLOC(cmd_line_params.num_of_keys * sizeof(VAPI_mr_hndl_t));
    params->recv_lkey_arr   = VMALLOC(cmd_line_params.num_of_keys * sizeof(VAPI_lkey_t));
    params->recv_rkey_arr   = VMALLOC(cmd_line_params.num_of_keys * sizeof(VAPI_rkey_t));
    params->send_lkey_arr   = VMALLOC(cmd_line_params.num_of_keys * sizeof(VAPI_lkey_t));
    params->remote_rkey_arr = VMALLOC(cmd_line_params.num_of_keys * sizeof(VAPI_rkey_t));

    for (i=0;i<(int)cmd_line_params.num_of_keys;i++) {   
        params->mr_hndl_s_arr[i] = VAPI_INVAL_HNDL;
        params->mr_hndl_r_arr[i] = VAPI_INVAL_HNDL;
    }

    /* Now convert the string parameters into the definitions */
    if (strcmp(cmd_line_params.transport_type, "ud") ==0)
    {
        transport = IB_TS_UD;
    } else
    {
        if (strcmp(cmd_line_params.transport_type, "uc") ==0)
        {
            transport = IB_TS_UC;
        } else
        {
            if (strcmp(cmd_line_params.transport_type, "rc") ==0)
            {
                transport = IB_TS_RC;
            } else
            {
                printf("Error - unsuported transport type %s\n", cmd_line_params.transport_type);
                return -1;
            }

        }
    }

    // Check oneside option compatibility
    if (cmd_line_params.oneside && (transport == IB_TS_RC)) { // Add RD check when supported
        printf("Error - 'oneside' option can't be used with reliable transport type\n");
        return -1;
    }

    rdma_mode = RDMA_NONE;
    is_bw = (strcmp(cmd_line_params.test_type, "bw") ==0);
    if (cmd_line_params.rwrite_pollcq == 1)
    {
        rdma_mode = RDMA_W_POLLCQ;
    } else if (strcmp(cmd_line_params.rdma, "write") ==0)
    {
        rdma_mode = RDMA_W;
    } else if (strcmp(cmd_line_params.rdma, "read") ==0)
    {
        rdma_mode = RDMA_R;
    }


    /* define the test mode */
    /* sanity checks */
    if (cmd_line_params.ports_num < 1 || cmd_line_params.ports_num > MAX_PORTS_NUM)
    {
        printf("Error - invalid '--ports_num' parameter (%d). Must be from 1 to %d.\n", 
               cmd_line_params.ports_num, MAX_PORTS_NUM);
        return -1;
    }
    if (cmd_line_params.ports_num < 2 && cmd_line_params.hca_port > MAX_PORTS_NUM)
    {
        printf("Error - invalid '-p' parameter (%d). Must be from 1 to %d.\n", 
               cmd_line_params.ports_num, MAX_PORTS_NUM);
        return -1;
    }
    if (cmd_line_params.bidirect != 0 && cmd_line_params.bidirect != 1)
    {
        printf("Error - invalid '--bidirect' parameter (%d). Must be 0 or 1.\n", 
               cmd_line_params.bidirect);
        return -1;
    }
    if (rdma_mode == RDMA_W_POLLCQ)
    {
        params->in_params.test_mode = LAT_RDMA_W_POLLCQ;
    }
    else if (rdma_mode == RDMA_W)
    {
        if (cmd_line_params.wqetest)
        {
            params->in_params.test_mode = WQE_TEST;
        }
        else if (is_bw)
        {
            if (cmd_line_params.bidirect == 1)
            {
                params->in_params.test_mode = BW_RDMA_W_BI;
            } else
            {
                if ((cmd_line_params.ports_num == 1) && (cmd_line_params.general_param == GEN_PARM_BW_OLD))
                    params->in_params.test_mode = BW_RDMA_W;
                else
                    params->in_params.test_mode = BW_RDMA_W_UNI;
            }
        } else
        {
            params->in_params.test_mode = LAT_RDMA_W;
        }
    }
    // RDMA read latency, added by Noam
    else if (rdma_mode == RDMA_R)
    {
        if (is_bw)
        {
            if (cmd_line_params.bidirect == 1)
            {
                params->in_params.test_mode = BW_RDMA_R_BI;
            } else
            {
                params->in_params.test_mode = BW_RDMA_R_UNI;
            }
        } else
        {
            params->in_params.test_mode = LAT_RDMA_R;
        }
    }

    else if (rdma_mode == RDMA_NONE)
    {
        //Check if it's an Atomic test and update test mode appropriately
        if (cmd_line_params.atomic_mode != 0)
        {

            //Check TS
            if ( (transport != IB_TS_RC) &&  (transport != IB_TS_RD) )
            {
                printf("Error - invalid transport type: Atomic ops are supported only for Reliable TS\n");
                return -1;
            }
            //Set predefined by IB-spec sizes for atomic operational data
            params->in_params.mr_recv_size = 8;
            params->in_params.mr_send_size = 8;

            //Define the mode of the atomic test
            switch (cmd_line_params.atomic_mode)
            {
            case CMP_AND_SWAP_MODE: 
                if (is_bw)
                {
                    params->in_params.test_mode = BW_ATOM_CS;
                    //params->in_params.test_mode = BW_LAT_ATOM_CS;
                } else
                {
                    params->in_params.test_mode = LAT_ATOM_CS;
                }

                break;
            case FETCH_AND_ADD_MODE:
                if (is_bw)
                {
                    params->in_params.test_mode = BW_ATOM_FA;
                } else
                {
                    params->in_params.test_mode = LAT_ATOM_FA;
                }

                break;
            }
        } else if (is_bw)
        {
            if (cmd_line_params.bidirect == 1)
            {
                params->in_params.test_mode = BW_SEND_BI;
            } else
            {
                if ((cmd_line_params.ports_num == 1) && (cmd_line_params.general_param == GEN_PARM_BW_OLD))
                    params->in_params.test_mode = BW_SEND;
                else
                    params->in_params.test_mode = BW_SEND_UNI;
            }
        } else
        {
            params->in_params.test_mode = LAT_SEND;
        }
    } else
    {
        printf("Error - RDMA-read mode is not supported\n"); //Vlad: ???
        return -1;
    }

    if ((strcmp(cmd_line_params.test_type, "mqp") ==0))
    {
        params->in_params.test_mode = BW_MQP;
    }

    if ((strcmp(cmd_line_params.test_type, "bdir") ==0))
    {
        params->in_params.test_mode = BW_BIDIR;
    }

    if ((strcmp(cmd_line_params.test_type, "bw_list") ==0))
    {
        params->in_params.test_mode = BW_SEND_LIST;
    }
    if ((strcmp(cmd_line_params.test_type, "chipset1") ==0))
    {
        params->in_params.test_mode = CHIPSET1;
    }
    if ((strcmp(cmd_line_params.test_type, "ezra") ==0))
    {
        params->in_params.test_mode = SPECIAL;
    }
    if ((strcmp(cmd_line_params.test_type, "shimon") ==0))
    {
        params->in_params.test_mode = SPECIAL_2PORT;
    }

    /* if no dev given - use the first one in the list*/
    if ((strcmp(cmd_line_params.deviceID , "none")==0))
    {
        rc = get_list_of_hcas(&hca_id_buf, &num_of_hcas);
        if ( rc != VAPI_OK )
        {
            printf("failed to retrieve the list of HCAs\n");
            return -1;
        }
        if (num_of_hcas > 1)
        {
            printf("\tFound more than 1 HCAs running on %s \n", hca_id_buf[0]);
        }
        strcpy(params->in_params.hca_id,hca_id_buf[0]);
        if ( hca_id_buf ) free(hca_id_buf);

    } else
    {
        strcpy(params->in_params.hca_id,cmd_line_params.deviceID);
    }

    
    // If tavor mode post 20 .latency inprovments 
    if (!strncmp(params->in_params.hca_id, "InfiniHost_",11) == 0)
    {
        params->in_params.rec_buf_num       = 20;
    } else
    {
        params->in_params.rec_buf_num       = 50;
    }
    //printf("rec_buf_num %d\n", params->in_params.rec_buf_num);

    params->in_params.oneside = cmd_line_params.oneside;
    params->in_params.transport = transport;
    params->in_params.rdma_mode = rdma_mode;

    params->in_params.use_events = cmd_line_params.use_events;
    params->in_params.num_ah = cmd_line_params.num_ah;
    params->in_params.use_devmem = cmd_line_params.use_devmem;
    params->in_params.socket = socket;
    params->in_params.socket_port = socket_port;
    params->in_params.is_sender = cmd_line_params.is_sender;
    params->in_params.is_server = cmd_line_params.is_server;
    params->in_params.iter = cmd_line_params.iterations;
    params->in_params.size = cmd_line_params.msg_size;
    params->in_params.general_param = cmd_line_params.general_param;
    params->in_params.num_of_qps = cmd_line_params.num_of_qps;
    //params->in_params.sched_period = cmd_line_params.sched_period;
    params->in_params.ports_num = cmd_line_params.ports_num;
    params->in_params.postlist_len = cmd_line_params.postlist_len;
    params->in_params.qpc_sets = cmd_line_params.qpc_sets;
    params->in_params.mpt_sets = cmd_line_params.mpt_sets;
    params->in_params.num_of_keys = cmd_line_params.num_of_keys;
    params->in_params.sg_list_len = cmd_line_params.sg_list_len;
    params->in_params.num_rsge = cmd_line_params.sg_list_len;
    //printf("%d\n", params->in_params.num_rsge);
    //params->in_params.many2one = cmd_line_params.many2one;
    params->in_params.multicast = cmd_line_params.multicast;
    params->in_params.grh = cmd_line_params.grh;
    params->in_params.srq_num = cmd_line_params.srq_num;
    params->in_params.comp_inter = cmd_line_params.comp_inter;

    if (params->in_params.ports_num > 1)
    {
        /* while using 2 ports, the port number parameter is ignored and PRESUMED THAT PORT1 IS CONNECTED TO PORT1 */
        params->in_params.port_num = 1;
    } else
    {
        params->in_params.port_num = cmd_line_params.hca_port;
    }
    params->in_params.bidirect = cmd_line_params.bidirect;
    if (cmd_line_params.mtu > 0)
    {
        params->in_params.mtu = (IB_mtu_t) cmd_line_params.mtu;
    }
    
    // print_cmd_line_params(&cmd_line_params);
    // print_test_params(params, "xxx");

    return rc;
}

typedef struct
{
    int pix;
    test_params_tx *params_p;
} thread_params_t;

static void *start_test_in_thread(void * context)
{
    thread_params_t *tp_p = (thread_params_t*)context;
    test_params_tx *params_p = tp_p->params_p;
    ret_val_t rc;
    COND_PRINTF2((" **** In a thread: running test for port %d \n",tp_p->pix+1));
    rc = params_p->cur_test_p( params_p, tp_p->pix );
    return(void*)rc;
}

static ret_val_t run_2port_test( test_mode_t test_mode, test_2ports_p test_p, test_params_tx *params_p )
{
    unsigned int pix, size, iter, i, num_of_qps, num_of_keys; //sched_period
    float ratio;
    int is_sender   = params_p->in_params.is_sender;
    call_result_t mosal_rc;
    u_int64_t t_exec, t0, t1, bytes;
    thread_params_t tp[2];
    char test_name[60];
    char *tstr = (params_p->in_params.transport == IB_TS_UD) ? "UD" : ((params_p->in_params.transport == IB_TS_UC) ? "UC" : "RC");
    ret_val_t rc;


    // SHIMON
    //if (test_mode == SPECIAL_2PORT)
    //{
    //    rc = send_bw_uni_test_2ports(params_p);
    //      return rc;
    //}
    // END SHIMON

    /* running via only one port */
    if (params_p->in_params.ports_num == 1)
    {
        rc = test_p( params_p, CUR_PORT_IX );
        return rc;
    }

    /* define the test name */
    switch (test_mode)
    {
    case BW_SEND_UNI:
        sprintf(test_name, "%s BW Unidirectional Test", tstr);
        break;
    case BW_RDMA_W_UNI:
        sprintf(test_name, "%s RDMA-write BW Unidirectional Test", tstr);
        break;
    case WQE_TEST:
        sprintf(test_name, "%s WQE Test", tstr);
        break;
    case BW_RDMA_R_UNI:
        strcpy(test_name, "RDMA-read BW Unidirectional Test");
        break;
    case BW_SEND_BI:
        sprintf(test_name, "%s BW Bidirectional Test", tstr);
        break;
    case BW_RDMA_W_BI:
        sprintf(test_name, "%s RDMA-write BW Bidirectional Test", tstr);
        break;
    case BW_RDMA_R_BI:
        strcpy(test_name, "RDMA-read BW Bidirectional Test");
        break;
    case SPECIAL_2PORT:
        strcpy(test_name, "Special experiment");
        break;
    default:
        printf( "Unsupported two-port test, exiting\n");
        goto err;
    }

    /* Creating sync objects */
    COND_PRINTF2((" **** Creating sync objects (sync obj %p)\n", &params_p->sync));
    MOSAL_syncobj_init( &params_p->sync );
    MOSAL_sem_init( &params_p->sem, 2 );
//    mosal_rc = MOSAL_sem_acq( &params_p->sem, FALSE );  /* non-blocking */
    for (pix=0; pix < 2; pix++)
    {
        mosal_rc = MOSAL_sem_acq( &params_p->sem, FALSE );  /* non-blocking */
        
        if (mosal_rc != MT_OK)
        {
            printf( "Error on acquiring a semaphore, exiting ...\n");
            goto err;
        }
    }

    /* start threads */
    for (pix=0; pix < 2; pix++)
    {
        tp[pix].pix = pix;
        tp[pix].params_p = params_p;
        params_p->cur_test_p = test_p;
        params_p->port[pix].mto_p = NULL;
        COND_PRINTF2((" **** Creating thread for port %d \n",pix+1));
        if (MOSAL_thread_start( &params_p->port[pix].mto, 0, start_test_in_thread, (void*)&tp[pix] ))
        {
            printf( "Thread failed to start, exiting ...\n");
            if (pix > 0)
                goto err0;
            else
                goto err;
        } else
        {
            params_p->port[pix].mto_p =&params_p->port[pix].mto;
        }
    }

    /* wait for threads to finish the test preparation phase */
    for (pix=i=0; pix < 2; pix++,i=0)
    {
        COND_PRINTF2((" **** Waiting for test to prepare (port %d) \n",pix+1));
        do
        {
            mosal_rc = MOSAL_sem_acq( &params_p->sem, FALSE );  /* non-blocking */
            if (mosal_rc == MT_OK) break;
            MOSAL_delay_execution( MIN_WAIT_FOR_SYNC_USECS );
        } while (++i < MAX_TRIES_NUM);
        if (mosal_rc != MT_OK)
        {
            printf( "Test is stuck on the preparation phase, exiting ...\n");
            goto err1;
        }
    }

    /* sync sender and receiver processes and run the execution phase of the tests */
    sock_sync_msg(params_p->in_params.socket, is_sender);
    COND_PRINTF1(("\n\n************* %s  started  *********************\n\n",    test_name));
    COND_PRINTF2((" **** Signaling tests on %p \n", &params_p->sync));
    MOSAL_syncobj_signal( &params_p->sync );

    /* wait for threads termination*/
    for (pix=0; pix < 2; pix++)
    {
        COND_PRINTF2((" **** Waiting for thread to exit (port %d) \n",pix+1));
        MOSAL_thread_wait_for_term( params_p->port[pix].mto_p, NULL );
        params_p->port[pix].mto_p = NULL;
    }

    /* print the results */
    if (is_sender)
    {
        t0 = params_p->port[0].snd.t_end - params_p->port[0].snd.t_start;
        t1 = params_p->port[1].snd.t_end - params_p->port[1].snd.t_start;
        //printf("t_start0 %lld\n", (long long)CLK_TO_USEC(params_p->port[0].snd.t_start));
        //printf("t_end0 %lld\n", (long long)CLK_TO_USEC(params_p->port[0].snd.t_end));
        //printf("t_start1 %lld\n", (long long)CLK_TO_USEC(params_p->port[1].snd.t_start));
        //printf("t_end1 %lld\n", (long long)CLK_TO_USEC(params_p->port[1].snd.t_end));
        //printf("t0 %lld\n", (long long)CLK_TO_USEC(t0));
        //printf("t1 %lld\n", (long long)CLK_TO_USEC(t1));
        t_exec = MIN2(t0,t1);
        size = params_p->port[0].size;
        iter = params_p->in_params.iter;
        ratio = (float)((params_p->in_params.bidirect) ? 4.0 : 2.0);  
        bytes = (u_int64_t)((ratio * size) * iter);
        num_of_qps = params_p->in_params.num_of_qps;
        num_of_keys = params_p->in_params.num_of_keys;
        //sched_period = params_p->in_params.sched_period;
#if 0   		
        printf("t0 " U64_FMT_SPEC "u, t1 " U64_FMT_SPEC "u, t_exec " U64_FMT_SPEC "u, |t1-t0| %.1f, bytes %.0f, size actual %d\n", 
               t0, t1, t_exec, (t1>t0) ? CLK_TO_MSEC(t1-t0) : CLK_TO_MSEC(t0-t1), (ratio * size) * iter, size );
#endif
        
#if 0   		
        printf("BW: %.1f MBytes/sec [Num. of QPs: %d, Xchng rate: %d, Num. of Keys: %d, size: %d bytes, iter: %d, total %.0f]\n", 
               BW_MBSEC(t_exec, (ratio * size) * iter), num_of_qps, //sched_period,
               num_of_keys, size, iter, (1.0 * size) * ratio * iter);
#endif        
/*
        printf("BW: %.1f MBytes/sec [sg_list_len: %d, Num of QPs: %d, size: %d bytes, iter: %d, total %.0f]\n", 
               BW_MBSEC(t_exec, (1.0 * size) * iter), 
               params_p->in_params.sg_list_len, num_of_qps, 
               size, iter, (1.0 * size) * iter);
*/  
#if 0

        printf("BW0: %.1f MBytes/sec\n BW1: %.1f MBytes/sec\n BW: %.1f MBytes/sec [size: %d bytes, iter: %d, total %.0f]\n", 
               BW_MBSEC(t0, size * iter * ratio/2),
               BW_MBSEC(t1, size * iter * ratio/2),
               BW_MBSEC(t_exec, (ratio * size) * iter),
               size, iter, (1.0 * size) * ratio * iter);
#else
        if (params_p->in_params.test_mode != WQE_TEST) {
            printf("BW: %.1f MBytes/sec [size: %d bytes, iter: %d, total %.0f]\n", 
              BW_MBSEC(t_exec, (ratio * size) * iter),
              size, iter, (1.0 * size) * ratio * iter);
        }
#endif

#if 0   		
        printf("MIN BW: %.1f MBs, AVG BW: %.1f MBs, MAX BW: %.1f MBs \n", 
               BW_MBSEC(MIN2(t0,t1), (ratio * size) * iter),   
               BW_MBSEC((t0+t1)/2, (ratio * size) * iter),     
               BW_MBSEC(MAX(t0,t1), (ratio * size) * iter));   

#endif
    }
    sock_sync_msg(params_p->in_params.socket, is_sender);
    COND_PRINTF1(("\n************* %s finished *********************\n\n", test_name));

    return RET_OK;

    err1:
    MOSAL_thread_kill(  params_p->port[1].mto_p );  
    params_p->port[1].mto_p = NULL;

    err0:
    MOSAL_thread_kill(  params_p->port[0].mto_p );  
    params_p->port[0].mto_p = NULL;

    err:
    return RET_ERR;

}

void   print_test_params(test_params_tx *params, char *src)
{
    test_in_params_t * ti_p = &params->in_params;
    test_params_tx *p = params;
    int i;

    COND_PRINTF(("========== test parameters of %s ==========\n", src));

    COND_PRINTF(("IN: size %d, iter %d, skip %d, mtu %d, general_param %d, oneside %d\n", 
                 ti_p->size, ti_p->iter, ti_p->skip, ti_p->mtu, ti_p->general_param, ti_p->oneside));
    COND_PRINTF(("IN: mr_send_size %d, mr_recv_size %d, rec_buf_num %d, alloc_aligned %d\n", 
                 ti_p->mr_send_size, ti_p->mr_recv_size, ti_p->rec_buf_num, ti_p->alloc_aligned));
    COND_PRINTF(("IN: use: snd_ev %d, rcv_ev %d, mwin %d, port_num %d, bidirect %d, ports_num %d \n", 
                 ti_p->use_send_events, ti_p->use_recv_events,  ti_p->use_mwindow, ti_p->port_num, ti_p->bidirect, ti_p->ports_num));
    COND_PRINTF(("IN: NUMs: cqe %d, r_wqe %d, s_wqe %d, rsge %d, ssge %d\n",
                 ti_p->num_cqe, ti_p->num_r_wqe, ti_p->num_s_wqe, ti_p->num_rsge, ti_p->num_ssge));
    COND_PRINTF(("HANDLEs: hca 0x%x, pd 0x%x, mr_s 0x%x, mr_r 0x%x\n",
                 (unsigned int)p->hca_hndl, (unsigned int)p->pd_hndl, p->mr_hndl_s, p->mr_hndl_r));
    COND_PRINTF(("max_inline_size %d, max_cap_inline_size %d, ous_src_rd_atom %d, ous_dst_rd_atom %d\n",
                 p->max_inline_size, p->max_cap_inline_size, p->hca_cap.max_qp_ous_rd_atom, p->qp_ous_rd_atom ));
    FOR_ALL_PORTS( p, i) 
    COND_PRINTF(("port %d, lid %d, dlid %d, av 0x%x, s_cq 0x%x, r_cq 0x%x \n",
                 p->port[i].port_num, p->port[i].lid, p->port[i].d_lid, (unsigned int)p->port[i].av_hndl, 
                 (unsigned int)p->port[i].s_cq_hndl, (unsigned int)p->port[i].r_cq_hndl 
                 ));      

    COND_PRINTF(("qps %d\n", p->in_params.num_of_qps));                    
    COND_PRINTF(("post_list %d\n", ti_p->postlist_len)); 
//    COND_PRINTF(("qp0 0x%x\n", p->port[i].qp_vect[0].qp_num));
//    COND_PRINTF(("dqp0 0x%x\n",p->port[i].qp_vect[0].d_qpnum));
//    COND_PRINTF(("qp1 0x%x\n", p->port[i].qp_vect[1].qp_num));
//    COND_PRINTF(("dqp1 0x%x\n",p->port[i].qp_vect[1].d_qpnum));

    //COND_PRINTF(("sch_per %d\n", p->port[i].sched_period));
    END_FOR
//    COND_PRINTF(("send_lkey 0x%x, recv_lkey 0x%x, recv_rkey 0x%x, remote_rkey 0x%x\n",
//                 p->send_lkey, p->recv_lkey, p->recv_rkey, p->remote_rkey));
    COND_PRINTF(("============================================\n"));
}

#ifdef __WIN__

    #if 0 /*defined(__i386__)*/
        #include <powrprof.h>
typedef struct _PROCESSOR_POWER_INFORMATION
{
    ULONG Number;  
    ULONG MaxMhz;  
    ULONG CurrentMhz;  
    ULONG MhzLimit;  
    ULONG MaxIdleState;  
    ULONG CurrentIdleState;
} PROCESSOR_POWER_INFORMATION, *PPROCESSOR_POWER_INFORMATION;

unsigned long get_counts_per_sec()
{
    unsigned long clocks = 0;
#define MAX_NUMBER_OF_PROCESSORS		8
    PROCESSOR_POWER_INFORMATION info[MAX_NUMBER_OF_PROCESSORS];
    PPROCESSOR_POWER_INFORMATION info_p = &info[0];
    ULONG ntStatus;
#ifdef PRINT_CPU_FREQUENCY
    SYSTEM_INFO sys_info;
    int i;
    GetSystemInfo( &sys_info );
#endif
    memset( info, 0, sizeof(info) );
    ntStatus = CallNtPowerInformation( ProcessorInformation, NULL, 0, info, sizeof(info) );
    if (!ntStatus)
    {
#ifdef PRINT_CPU_FREQUENCY
        for (i=0; i<(int)sys_info.dwNumberOfProcessors; i++, info_p++)
        {
            printf("CallNtPowerInformation: Number %d, MaxMhz %lu, CurrentMhz %lu, MhzLimit %lu, MaxIdleState %lu, CurrentIdleState %lu\n", 
                   info_p->Number, info_p->MaxMhz, info_p->CurrentMhz, 
                   info_p->MhzLimit, info_p->MaxIdleState, info_p->CurrentIdleState );
        }
#endif		
        // usually, it returns a value by 1MHz less, that MMC. So we correct that
        clocks = (info[0].CurrentMhz + 1) * 1000000;
    } else
    {
        printf("CallNtPowerInformation failed to get CPU frequency (%d). We'll take it from the driver. \n", ntStatus );
        printf("Pay attention to the value of frequency, printed in the header. \n"
               "It might be less than the real one, which will cause the test results to look worse than they really are.\n", ntStatus, ntStatus );
        clocks = (ULONG)MOSAL_get_counts_per_sec();
    }

    return clocks;
}
    #endif
#endif


/*
 *  get_list_of_hcas
 */
VAPI_ret_t get_list_of_hcas(VAPI_hca_id_t **hca_id_buf_p, u_int32_t *num_of_hcas)
{
    VAPI_hca_id_t *hca_id_buf;
    VAPI_ret_t rc;
    u_int32_t local_num_of_hcas;

    *hca_id_buf_p = NULL;
    rc = EVAPI_list_hcas(0, &local_num_of_hcas, NULL);
    switch ( rc )
    {
    case VAPI_OK:
        *num_of_hcas = local_num_of_hcas;
        return VAPI_OK;

    case VAPI_EAGAIN:
        hca_id_buf = malloc(sizeof(VAPI_hca_id_t)*local_num_of_hcas);
        if ( !hca_id_buf )
        {
            return VAPI_ERR;
        }
        rc = EVAPI_list_hcas(local_num_of_hcas, &local_num_of_hcas, hca_id_buf);
        if ( rc != VAPI_OK )
        {
            free(hca_id_buf);
            return VAPI_ERR;
        }
        *num_of_hcas = local_num_of_hcas;
        *hca_id_buf_p = hca_id_buf;
        return VAPI_OK;

    default:
        return VAPI_ERR;
    }
}

