/*
  This software is available to you under a choice of one of two
  licenses.  You may choose to be licensed under the terms of the GNU
  General Public License (GPL) Version 2, available at
  <http://www.fsf.org/copyleft/gpl.html>, or the OpenIB.org BSD
  license, available in the LICENSE.TXT file accompanying this
  software.  These details are also available at
  <http://openib.org/license.html>.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
  BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
  ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
  CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
  SOFTWARE.

  Copyright (c) 2004 Mellanox Technologies Ltd.  All rights reserved.
*/

#include <string.h>
#include "sock.h"
#include "utils.h"


/*
 *  create_tcp_connection
 */ 
int create_tcp_connection(const char *ip, int port)
{
    struct in_addr remote_addr;
    int s, rc,on;
    struct sockaddr_in my_addr, server_addr;

    s = MOSAL_socket_socket(MOSAL_PF_INET, MOSAL_SOCK_STREAM, MOSAL_IPPROTO_TCP);
    if ( s == -1 ) {
        return -1;
    }

    on = 1;
    rc = MOSAL_socket_setsockopt(s, MOSAL_SOL_SOCKET, MOSAL_SO_REUSEADDR, &on, sizeof(on));
    if (-1 == rc)
    {
        return -1;
    }




    my_addr.sin_family = AF_INET;
    my_addr.sin_port = htons((unsigned short)port);
    my_addr.sin_addr.s_addr = INADDR_ANY;

    if ( inet_aton(ip,  &remote_addr) == 0 ) {
        return -1;
    }
    server_addr.sin_family = AF_INET;
    server_addr.sin_port = htons((unsigned short)port);
    server_addr.sin_addr = remote_addr;
    rc = MOSAL_socket_connect(s, (void *)(&server_addr), sizeof(server_addr));
    if ( rc == -1 ) {
        return -1;
    }

    return s;
}


int perf_server_init(int port) {
    int on,s, rc;
    struct sockaddr_in my_addr;

    s = MOSAL_socket_socket(MOSAL_PF_INET, MOSAL_SOCK_STREAM, MOSAL_IPPROTO_TCP);
    if ( s == -1 ) {
        return -1;
    }
    
    on = 1;
    rc = MOSAL_socket_setsockopt(s, MOSAL_SOL_SOCKET, MOSAL_SO_REUSEADDR, &on, sizeof(on));
    if (-1 == rc)
    {
        return -1;
    }

    my_addr.sin_family = AF_INET;
    my_addr.sin_port = htons((unsigned short)port);
    my_addr.sin_addr.s_addr = INADDR_ANY;
    rc = MOSAL_socket_bind(s, (void *)(&my_addr), sizeof(my_addr));
    if ( rc == -1 ) {
        printf("%s\n", strerror(errno));
        return -1;
    }

    rc = MOSAL_socket_listen(s, 1);
    if ( rc == -1 ) {
        printf("%s\n", strerror(errno));
        return -1;
    }

    return s;
}


/* socket server waits for a client to connect */
int wait_for_connection(int socket)
{
    int newsock, addr_len;
    struct sockaddr_in remote_addr;

    addr_len = sizeof(remote_addr);
    newsock = MOSAL_socket_accept(socket, (void *)(&remote_addr), &addr_len);
    if ( newsock == -1 ) {
        printf("%s\n", strerror(errno));
        return -1;
    }
    return newsock;
}

int sock_get_block(int socket, char *msg, int len)
{
    int length = 0;
    do {
        length = MOSAL_socket_recv(socket, msg, len, 0);
        if (length == -1) {
            printf("Error: MOSAL_socket_recv returned -1\n");
            printf("%s\n", strerror(errno));
            return -1;            
        }
        if (length > len) {
            printf("Error: MOSAL_socket_recv read more than required (%d > %d)\n",
                   length, len);
            return -1;
        }
        len -= length;
    } while ( len > 0 );

    return 0;
}

/*
int sock_get_msg(int socket, char *msg, int len)
{
    u_int16_t msg_len =0, net_msg_len = 0;
    
    int length = MOSAL_socket_recv(socket, (void *)&net_msg_len, 2, 0);
    
    if (length != 2) {
        printf("Debug: wrong length returned by MOSAL_socket_recv: %d\n", length);
    }

    // turn from net work ordering to cpu 
    msg_len = MOSAL_be16_to_cpu(net_msg_len);
    
    //printf("Debug: MOSAL_socket_recv got length=%d\n", msg_len);

    if (msg_len > len-1) {
        printf("Debug: sock_get_msg '%s'\n", msg);
        printf("Error: msg_len=%d but buffer is only %d\n",msg_len, len);
        return -1;
    }
    

    memset(msg,0,msg_len);
    length = MOSAL_socket_recv(socket, msg, msg_len, 0);
    
    if (length != msg_len) {
        printf("Debug: wrong length returned by MOSAL_socket_recv: %d\n", length);
    } 

    msg[msg_len] = '\0';
    COND_PRINTF7(("recive: %s [%d]\n", msg,msg_len));
    return length;
    
}
*/

int sock_get_msg(int socket, char *msg, int len)
{
    u_int16_t msg_len =0, net_msg_len = 0;
    
    int length = sock_get_block(socket,(void *)&net_msg_len, 2);
    
    if (length == -1) {
        printf("Error: failed to read Msg length from socket\n");
        return -1;
    }

    /* turn from net work ordering to cpu */
    msg_len = MOSAL_be16_to_cpu(net_msg_len);
    
    //printf("Debug: MOSAL_socket_recv got length=%d\n", msg_len);

    if (msg_len > len-1) {
        printf("Debug: sock_get_msg '%s'\n", msg);
        printf("Error: msg_len=%d but buffer is only %d\n",msg_len, len);
        return -1;
    }
    
    memset(msg,0,msg_len);
    length = sock_get_block(socket, msg, msg_len);
    
    if (length == -1) {
        printf("Error: failed to read Msg from socket\n");
        return -1;
    }

    msg[msg_len] = '\0';
    COND_PRINTF7(("receive: %s [%d]\n", msg, msg_len));
    return msg_len;
    
}

int sock_send_msg(int socket, const char *format, ...)
{
    char msg_out[MAX_MSG_LEN];
    char msg[MAX_MSG_LEN];
    int rc;
    u_int16_t len,net_len;
    va_list arglist;   

    va_start (arglist, format);   
    vsprintf (msg, format, arglist);   
    va_end (arglist);   

    len = (u_int16_t)strlen(msg);

    if (len > (MAX_MSG_LEN-3)) {
        printf("Debug: sock_send_msg got '%s'\n", format);
        printf("Error: Send msg len=%d but buffer is only %d\n",len,MAX_MSG_LEN );
        return -1;
    }

    memset(msg_out,0,MAX_MSG_LEN);
    //turn msg to big endianess - network order
    net_len = MOSAL_cpu_to_be16(len);
    memcpy(msg_out,&net_len,2);
    memcpy(msg_out+2,msg,len);
    
    rc = MOSAL_socket_send(socket, msg_out, len+2, 0);
    if ( rc != (len+2) ) {
        printf("Error: Couldn't send messags=%s\n",msg_out);
        return -1;
    }
    COND_PRINTF7(("send: %s [%d]\n", msg,len+2));
    return 0;
}




void sock_msg2num(char *msg, char *label, long def, void *out, unsigned int len)
{
    char *startstr;
    u_int64_t tmp_val;

//long tmp_val;

    startstr = strstr(msg,label);
    if(NULL == startstr){
        tmp_val = def;
    }
    else {
      //        tmp_val = STR_2_ULL(startstr+strlen(label));
        tmp_val = strtoull(startstr+strlen(label),(char **) NULL,0);
        //tmp_val = atoll(startstr+strlen(label));
    }

    switch (len) {
        case 1:
            *((char *)out) = (char) tmp_val;
            break;
        case 2:
            *((short *)out) = (short ) tmp_val;
            break;
        case 4:
            *((int *)out) = (int ) tmp_val;
            break;
        case 8:
            *((u_int64_t *)out) = (u_int64_t ) tmp_val;
            break;
        default:
            *((char *)out) = (char) tmp_val;
            break;
    }
}
     
void sock_msg2str(char *msg, char *label, char *def, char  *out)
{
    char *startstr;
    
    startstr = strstr(msg,label);
    if(NULL == startstr){
        strcpy(out,def);
        return;
    }
    if(0 == sscanf(startstr+strlen(label),"%s",out)){
        strcpy(out,def);
    }
}

unsigned char sock_is_exists(char *msg, char *label)
{
    return (NULL == strstr(msg,label)) ? 0 : 1;
}


void sock_sync_msg(int socket, int is_sender)
{
    char                   sock_msg[1000];
    if (socket == -1) {
        return;
    }

    memset(sock_msg,0,1000);
    
    /* sync start test */
    if (is_sender) {
        while (1) {
            COND_PRINTF2(("Sender: Waiting to get start........"));
            sock_get_msg(socket, sock_msg, 1000);
            if (sock_is_exists(sock_msg,"start")) {
                COND_PRINTF2(("Sender Got start\n"));
                sock_send_msg(socket, "start");
                COND_PRINTF2(("Sender send start\n"));
                break;
            }
        }
//	sleep(2);
    }
    else {
        sock_send_msg(socket, "start");
        COND_PRINTF2(("Receiver send start\n"));
        while (1) {
            COND_PRINTF2(("Receiver Waiting to get start........"));
            sock_get_msg(socket, sock_msg, 1000);
            if (sock_is_exists(sock_msg,"start")) {
                COND_PRINTF2(("Receiver Got start\n"));
                break;
            }
        }
    }
}

void sock_freeze(int socket, int is_sender)
{
    char                   sock_msg[50];
    memset(sock_msg,0,50);
    
    /* sync start test */
    if (is_sender) {
       COND_PRINTF2(("Sender: Asking receiver to freeze ........\n"));
	sock_send_msg(socket, "freeze");
//	sleep(2);
    }
    else {
	COND_PRINTF2(("Receiver: waiting for freeze ...\n"));
	while (1) {
	    sock_get_msg(socket, sock_msg, 1000);
	    if (sock_is_exists(sock_msg,"freeze")) {
	        COND_PRINTF2(("Receiver got 'freeze'\n"));
	        break;
	    }
	}
	COND_PRINTF2(("Receiver: waiting for unfreeze ...\n"));
	while (1) {
	    sock_get_msg(socket, sock_msg, 1000);
	    if (sock_is_exists(sock_msg,"freeze")) {
	        COND_PRINTF2(("Receiver got 'unfreeze'\n"));
	        break;
	    }
	}
    }
}

void sock_unfreeze(int socket, int is_sender)
{
    char                   sock_msg[50];
    memset(sock_msg,0,50);
    
    /* sync start test */
    if (is_sender) {
       COND_PRINTF2(("Sender: Asking receiver to unfreeze ........\n"));
	sock_send_msg(socket, "freeze");
//	sleep(2);
    }
}


