/*
  This software is available to you under a choice of one of two
  licenses.  You may choose to be licensed under the terms of the GNU
  General Public License (GPL) Version 2, available at
  <http://www.fsf.org/copyleft/gpl.html>, or the OpenIB.org BSD
  license, available in the LICENSE.TXT file accompanying this
  software.  These details are also available at
  <http://openib.org/license.html>.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
  BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
  ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
  CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
  SOFTWARE.

  Copyright (c) 2004 Mellanox Technologies Ltd.  All rights reserved.
*/
#ifndef H_UTILS_H
#define H_UTILS_H

#include <mosal.h>  // for MOSAL_get_counts_per_sec()
#include <stdio.h>
#include "utils_imp.h"

extern u_int64_t cpu_CLCKS_PER_SECS;
extern FILE *res_file;
extern int silent_print;
extern int debug_print;

#define DEBUG_LEVEL 1  //rel=1  //Ezra ???
//#define DEBUG_PRINT  //rel=comment out

#define MIN2(a,b) ((a)>(b) ? (b) : (a))
#define MIN3(a,b,c) MIN2(MIN2(a,b),c)
#ifndef MAX
#define MAX(a,b) ((a)>(b) ? (a) : (b))
#endif

// Only for bw_special!!!
#define get_clk64_c(u64) get_clk64(u64) 

#define __PERF_PRINTF(aa) \
    if(!silent_print) { printf aa; fflush( stdout ); }

#define COND_PRINTF_EZ(aa) printf aa

#ifdef DEBUG_PRINT
#define COND_PRINTF(aa) printf aa
#define DEBUG_PRINTF(aa) if (debug_print) { printf aa; fflush( stdout ); }
#else
#define COND_PRINTF(aa)
#define DEBUG_PRINTF(aa)
#endif

#define COND_PRINTF0(aa) 
#define COND_PRINTF1(aa) 
#define COND_PRINTF2(aa) 
#define COND_PRINTF3(aa) 
#define COND_PRINTF7(aa) 

#if DEBUG_LEVEL >= 0
#undef COND_PRINTF0
#define COND_PRINTF0(aa) printf aa
#endif
#if DEBUG_LEVEL >= 1
#undef COND_PRINTF1
#define COND_PRINTF1(aa) __PERF_PRINTF(aa)
#endif
#if DEBUG_LEVEL >= 2
#undef COND_PRINTF2
#define COND_PRINTF2(aa) __PERF_PRINTF(aa)
#endif
#if DEBUG_LEVEL >= 3
#undef COND_PRINTF3
#define COND_PRINTF3(aa) __PERF_PRINTF(aa)
#endif
#if DEBUG_LEVEL >= 7
#undef COND_PRINTF7
#define COND_PRINTF7(aa) __PERF_PRINTF(aa)
#endif

#define CLK_TO_USEC(__clk) (1000000.0*(((ADJUST_TYPE)(__clk)))/(1.0*((ADJUST_TYPE)cpu_CLCKS_PER_SECS)))
#define USEC_TO_CLK(__usec)   ((ADJUST_TYPE) ((__usec) *  ((1.0*((ADJUST_TYPE) cpu_CLCKS_PER_SECS))/1000000.0)))
#define CLK_TO_MSEC(__clk) (1000.0*(((ADJUST_TYPE)(__clk)))/(1.0*((ADJUST_TYPE)cpu_CLCKS_PER_SECS)))
#define BW_MBSEC(__clk,__bytes) (((ADJUST_TYPE)(__clk))) > 0 ? (1.0*((ADJUST_TYPE)cpu_CLCKS_PER_SECS)*(1.0*((ADJUST_TYPE)(__bytes))))/(1024.0*1024*(1.0*((ADJUST_TYPE)(__clk)))): 0
#define BW_KBSEC(__clk,__bytes) (((ADJUST_TYPE)(__clk))) > 0 ? (1.0*((ADJUST_TYPE)cpu_CLCKS_PER_SECS)*(1.0*((ADJUST_TYPE)(__bytes))))/(1024.0*(1.0*((ADJUST_TYPE)(__clk)))): 0
#define BW_BSEC(__clk,__bytes) (((ADJUST_TYPE)(__clk))) > 0 ? (1.0*((ADJUST_TYPE)cpu_CLCKS_PER_SECS)*(1.0*((ADJUST_TYPE)(__bytes))))/((1.0*((ADJUST_TYPE)(__clk)))): 0


// Backwrd to older versions
#ifndef MOSAL_be16_to_cpu
#ifdef  MT_BIG_ENDIAN
    #define MOSAL_be16_to_cpu(x)         (x)
    #define MOSAL_cpu_to_be16(x)         (x)
#else
    #define MOSAL_be16_to_cpu(x)         ((((x) >> 8)&0xff) | (((x) << 8)&0xff00))
    #define MOSAL_cpu_to_be16(x)         ((((x) >> 8)&0xff) | (((x) << 8)&0xff00))
#endif
#endif

#ifndef MT_DOWN_ALIGNX_VIRT
#define MT_DOWN_ALIGNX_VIRT(value, mask)     ((MT_virt_addr_t)(value) & (~((MT_virt_addr_t)0) << (mask)))
#define MT_UP_ALIGNX_VIRT(value, mask)       MT_DOWN_ALIGNX_VIRT(((value) +  ~(~((MT_virt_addr_t)0) << (mask))), (mask))
#endif


#if 0
#define DELAY_usecs(iter)  \
{ \
    u_int64_t t1,t2;\
    double ff; \
    int jjj;\
    t1=t2=0;\
    get_clk64_1(t1); \
    for(jjj=0;jjj<iter;jjj++){ \
      ff = jjj*jjj/(3.75+jjj*1.27); \
    } \
    get_clk64_1(t2); \
    COND_PRINTF(("********** calc_delay=%.2fusec\n",CLK_TO_USEC(t2-t1))); \
}
#endif 


#endif

