/*
  This software is available to you under a choice of one of two
  licenses.  You may choose to be licensed under the terms of the GNU
  General Public License (GPL) Version 2, available at
  <http://www.fsf.org/copyleft/gpl.html>, or the OpenIB.org BSD
  license, available in the LICENSE.TXT file accompanying this
  software.  These details are also available at
  <http://openib.org/license.html>.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
  BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
  ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
  CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
  SOFTWARE.

  Copyright (c) 2004 Mellanox Technologies Ltd.  All rights reserved.
*/

#ifndef H_UTILS_IMP_H
#define H_UTILS_IMP_H

#include <unistd.h> // for sleep
#include <sys/time.h>
#include <sys/resource.h>


static inline u_int64_t gettime()
{
    u_int64_t ticks;
    struct timeval TsNow;

    gettimeofday(&TsNow,NULL);
    ticks = TsNow.tv_usec + TsNow.tv_sec * 1000000; 

    return ticks;
}

static inline u_int64_t get_cpu_clocks_per_seconds ()
{
  FILE* fp;
  char buffer[1024];
  char *match, *match1;
  float clock_speed = 1000.0;
  int found;

  match1 = match = NULL;



  fp = fopen ("/proc/cpuinfo", "r");
  while(NULL != fgets(buffer, 100, fp)){
      match = strstr (buffer, "cpu MHz");
      if(match) break;
  }
  fclose (fp);
  
  if (match == NULL){
      return 1000000000;
  }
  
  /* Parse the line to extrace the clock speed.  */
  match1 = strchr (match, ':');
  if(match1){
      found = sscanf (match1+1, "%f", &clock_speed);
      if(found == 0){
          clock_speed =  1000.0;
      }
  }


  return (u_int64_t) (clock_speed*1000000);
}




#define get_clk64_timeofday(u64)  u64 = gettime()
#define CLK_TO_USEC_GETTIMEOFDAY(u64)  u64

//#ifdef RDTSC_IA64
#ifdef __ia64__
   #define USE_SPEC_CLOCK_CALC 1
   #define get_clk64(u64) \
    __asm__ __volatile__("mov %0=ar.itc" : "=r" (u64));

    #define update_CPU_CLCKS_PER_SECS() \
        cpu_CLCKS_PER_SECS = get_cpu_clocks_per_seconds();
#endif

//#ifdef RDTSC_X86
#ifdef __i386__
    #define USE_SPEC_CLOCK_CALC 1
    #define update_CPU_CLCKS_PER_SECS() \
        cpu_CLCKS_PER_SECS = MOSAL_get_counts_per_sec()
    #define get_clk64(u64) \
         __asm__ __volatile__("rdtsc" : "=A" (u64))
#endif

#ifdef USE_MOSAL_PPC
    #define USE_SPEC_CLOCK_CALC 1
    #define update_CPU_CLCKS_PER_SECS() \
        cpu_CLCKS_PER_SECS = MOSAL_get_counts_per_sec()
    #define get_clk64(u64) \
         u64 = MOSAL_ppc_cputime_get()
#endif

//#ifdef RDTSC_X86_64
#ifdef __x86_64__    
    #define USE_SPEC_CLOCK_CALC 1
    #define update_CPU_CLCKS_PER_SECS() \
        cpu_CLCKS_PER_SECS = MOSAL_get_counts_per_sec()
    #define get_clk64(u64)                                          \
        do {                                                        \
            unsigned int a,d;                                       \
            asm volatile("rdtsc" : "=a" (a), "=d" (d));             \
            (u64) = ((unsigned long)a) | (((unsigned long)d)<<32);  \
        } while ( 0 )
#endif

/* Use gettimeofday by default */
#ifndef USE_SPEC_CLOCK_CALC
    #define update_CPU_CLCKS_PER_SECS() \
        cpu_CLCKS_PER_SECS = 1000000.0
    #define get_clk64(u64) get_clk64_timeofday(u64)
#endif

#define ADJUST_TYPE		u_int64_t
#define STR_2_ULL(str)	strtoull(str,(char **) NULL,0)

#endif
