/*
  This software is available to you under a choice of one of two
  licenses.  You may choose to be licensed under the terms of the GNU
  General Public License (GPL) Version 2, available at
  <http://www.fsf.org/copyleft/gpl.html>, or the OpenIB.org BSD
  license, available in the LICENSE.TXT file accompanying this
  software.  These details are also available at
  <http://openib.org/license.html>.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
  BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
  ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
  CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
  SOFTWARE.

  Copyright (c) 2004 Mellanox Technologies Ltd.  All rights reserved.
*/
//#define bdebug 1
//#define VERBOSE 1

#include <vapi.h>
#include <evapi.h>
#include <vapi_common.h>
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#if defined(__WIN__) 
#define pthread_mutex_init(mutex_p, b)	MOSAL_mutex_rel(mutex_p)
#define pthread_mutex_lock(mutex_p) 	MOSAL_mutex_acq(mutex_p, TRUE)
#define pthread_mutex_unlock(mutex_p) 	MOSAL_mutex_rel(mutex_p)
#define sleep(secs)						Sleep((secs)*1000)
#undef RAND_MAX
#define RAND_MAX	0x7fffffff
#define MTGETOPT_EXPORTS
#include "mt_getopt.h"
#else
#include <unistd.h>
#include <getopt.h>
#include <readline/readline.h>
#include <signal.h>
#include <pthread.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <sched.h>
#endif
#include <mosal.h>
#include "rctp.h"
#include "sock.h"

#ifdef __WIN__
#define EXIT_CODE	        1
#else
#define EXIT_CODE	        -1
#endif


#define TCP_ERROR_EXIT_CODE	2

//#define SKIP_ATOMIC

/*============== macro definitions ============================*/
#define VAPIERR(rc) printf(__FILE__"#%d - %s: %s\n", __LINE__, VAPI_strerror_sym(rc), VAPI_strerror(rc))
#define GENERR(s) printf(__FILE__"#%d - %s\n", __LINE__, s)
#define CLEAN_EXIT(rc) {                                      \
                         rctp_cleanup(&topt, rc);                  \
                         return rc;                           \
                       }

#define debug_cleanup

#ifdef DEBUG
  #define PRINT_QP_PROPS(qp_props) print_qp_props(qp_props)
  #define PRINT_HCA_CAP(hca_vendor, hca_cap) print_hca_cap(hca_vendor, hca_cap)
#else
  #define PRINT_QP_PROPS(qp_props)
  #define PRINT_HCA_CAP(hca_vendor, hca_cap)
#endif

#define SND_Q_SAFE 2        /* number of free entries that must be in the send Q */
#define RCV_Q_SAFE 2        /* number of free entries that must be in the send Q */

#define MAX_OUST_WR_SQ  (170+SND_Q_SAFE)  /* Max outstanding WR on the SQ */
#define MAX_OUST_WR_RQ  (170+RCV_Q_SAFE)  /* Max outstanding WR on the RQ */
#define MAX_SG_SIZE_SQ  1    /* Max scatter/gather descriptor entries on the SQ */ 
#define MAX_SG_SIZE_RQ  1    /* Max scatter/gather descriptor entries on the RQ */ 

#define SEND_MSG_SIZE (sizeof(struct cmd_msg_st))   /* size of send message */
#define ATOMIC_OP_DATA_SIZE  8  /* size of data returned in SG list buffer by atomic op */
#define USED_SL 0
#define START_PSN 0
#define QP_OUS_RD_ATOM 1

#define TCP_PORT_NO 1407  /* the TCP port number used for communication */

#ifdef __WIN__
#define rctp_endian64_swap(x) letobe64(x)
#else
#define rctp_endian64_swap(x) \
({ \
	u_int64_t __x = (x); \
	((u_int64_t)( \
		(u_int64_t)(((u_int64_t)(__x) & MAKE_ULONGLONG(0x00000000000000ff)) << 56) | \
		(u_int64_t)(((u_int64_t)(__x) & MAKE_ULONGLONG(0x000000000000ff00)) << 40) | \
		(u_int64_t)(((u_int64_t)(__x) & MAKE_ULONGLONG(0x0000000000ff0000)) << 24) | \
		(u_int64_t)(((u_int64_t)(__x) & MAKE_ULONGLONG(0x00000000ff000000)) <<  8) | \
	        (u_int64_t)(((u_int64_t)(__x) & MAKE_ULONGLONG(0x000000ff00000000)) >>  8) | \
		(u_int64_t)(((u_int64_t)(__x) & MAKE_ULONGLONG(0x0000ff0000000000)) >> 24) | \
		(u_int64_t)(((u_int64_t)(__x) & MAKE_ULONGLONG(0x00ff000000000000)) >> 40) | \
		(u_int64_t)(((u_int64_t)(__x) & MAKE_ULONGLONG(0xff00000000000000)) >> 56) )); \
})
#endif

/*============== static functions prototypes ==================*/
static VAPI_ret_t create_qp(test_options_t *topt_p);
static VAPI_ret_t wait_link_up(test_options_t *topt_p);
static int get_remote_params(test_options_t *topt_p);
static int send_local_param(test_options_t *topt_p);
static VAPI_ret_t register_regions(test_options_t *topt_p);
static VAPI_ret_t post_recv_buf(test_options_t *topt_p);
static VAPI_ret_t post_send_buf(test_options_t *topt_p, struct cmd_msg_st *cmd_p, u_int32_t size_to_send);
static VAPI_ret_t post_rdma_request(test_options_t *topt_p, struct cmd_msg_st *cmd_p, MT_bool use_id);
static VAPI_ret_t perform_atomic_test(test_options_t *topt_p, struct cmd_msg_st *cmd_p);
#ifndef SKIP_ATOMIC
static VAPI_ret_t post_atomic_request(test_options_t *topt_p, struct cmd_msg_st *cmd_p,
                                      u_int64_t  fetch_cmp_oper, u_int64_t swap_oper);
#endif
static VAPI_ret_t move_to_rtr(test_options_t *topt_p);
static VAPI_ret_t move_to_rts(test_options_t *topt_p);
static int init_db(test_options_t *topt_p);
static void rctp_cleanup(test_options_t *topt_p, int errcode);
static int create_threads(test_options_t *topt_p);
#ifdef USE_SEND_COMP_THREAD
static void *send_comp_thread_func(void *arg);
#endif
#ifdef USE_RCV_COMP_THREAD
static void *rcv_comp_thread_func(void *arg);
#endif
static void parse_rcv_msg(VAPI_wc_desc_t *comp_desc_p, struct cmd_msg_st *cmd_p);
static VAPI_ret_t get_ib_mtu(u_int32_t mtu, IB_mtu_t *ib_mtu);
static VAPI_ret_t process_command(test_options_t *topt_p, 
                                  VAPI_wc_desc_t *comp_desc_p, struct cmd_msg_st *cmd_p);
static void drain_send_queue(test_options_t *topt_p);
static void drain_rcv_queue(test_options_t *topt_p);
static VAPI_ret_t create_memory_windows(test_options_t *topt_p);
static void cmd_be_to_host(struct cmd_msg_st *cmd_p);
static void cmd_host_to_be(struct cmd_msg_st *cmd_p);

#ifdef __WIN__
BOOL WINAPI clean_handler(  DWORD dwCtrlType );
#else
void clean_handler(int signum);
#endif
static void print_peer_info(test_options_t *topt_p);
#ifdef bdebug
static void print_ctx(test_options_t *topt_p, trans_ctx_t *tctx_p);
#endif

#ifdef im_debug
void print_post_opcode(VAPI_wr_opcode_t opcode, void *context);
void print_completion_opcode(VAPI_cqe_opcode_t opcode);
#endif

static void mem_debugger();
static void dump_mem(void *addr, int count);
typedef void (*sighandler_t)(int);

static test_options_t *glob_opt;

static u_int32_t  sequence_id = 0;
static u_int32_t  highest_cq_seq_id = 0;
static MOSAL_mutex_t glob_mtx;

/*+++++++++++++++++++++ inline functions ++++++++++++++++++++++++*/
/*
 * inc_snd_rcv_balance
 */
static inline void inc_snd_rcv_balance(test_options_t *topt_p)
{
  pthread_mutex_lock(&topt_p->bl_mutex);
  topt_p->snd_rcv_balance++;
  pthread_mutex_unlock(&topt_p->bl_mutex);
}

/*
 *  dec_snd_rcv_balance
 */
static inline void dec_snd_rcv_balance(test_options_t *topt_p)
{
  pthread_mutex_lock(&topt_p->bl_mutex);
  topt_p->snd_rcv_balance--;
  pthread_mutex_unlock(&topt_p->bl_mutex);
}

/*
 *  get_snd_rcv_balance
 */
static inline int get_snd_rcv_balance(test_options_t *topt_p)
{
  int rc;

  /*pthread_mutex_lock(&topt_p->bl_mutex);*/
  rc = topt_p->snd_rcv_balance;
//  printf("snd_rcv_balance=%d\n", topt_p->snd_rcv_balance);
//  fflush(stdout);
  /*pthread_mutex_unlock(&topt_p->bl_mutex);*/

  return rc;
}

#if 0
/*
 * log2()
 */
static int log2(u_int64_t arg)
{
  int i;
  u_int64_t tmp;

  if ( arg == 0 ) {
    return INT_MIN; /* log2(0) = -infinity */
  }

  tmp = 1;
  i = 0;
  while ( tmp < arg ) {
    tmp = tmp << 1;
    ++i;
  }

  return i;
}
#endif

/*
 * inc_cur_send_outstanding
 */
static inline void inc_cur_send_outstanding(test_options_t *topt_p)
{
  pthread_mutex_lock(&topt_p->snd_mutex);
  topt_p->cur_send_outstanding++;
  pthread_mutex_unlock(&topt_p->snd_mutex);
}

/*
 *  dec_cur_send_outstanding
 */
static inline void dec_cur_send_outstanding(test_options_t *topt_p)
{
  pthread_mutex_lock(&topt_p->snd_mutex);
  topt_p->cur_send_outstanding--;
  pthread_mutex_unlock(&topt_p->snd_mutex);
}

/*
 *  get_cur_send_outstanding
 */
static inline int get_cur_send_outstanding(test_options_t *topt_p)
{
  int rc;

  /*pthread_mutex_lock(&topt_p->snd_mutex);*/
  rc = topt_p->cur_send_outstanding;
  /*pthread_mutex_unlock(&topt_p->snd_mutex);*/

  return rc;
}


/*
 * inc_cur_posted_rcv_bufs
 */
static inline void inc_cur_posted_rcv_bufs(test_options_t *topt_p)
{
  pthread_mutex_lock(&topt_p->rcv_mutex);
  topt_p->cur_posted_rcv_bufs++;
  pthread_mutex_unlock(&topt_p->rcv_mutex);
}

/*
 *  dec_cur_posted_rcv_bufs
 */
static inline void dec_cur_posted_rcv_bufs(test_options_t *topt_p)
{
  pthread_mutex_lock(&topt_p->rcv_mutex);
  topt_p->cur_posted_rcv_bufs--;
  pthread_mutex_unlock(&topt_p->rcv_mutex);
}

/*
 *  get_cur_posted_rcv_bufs
 */
static inline int get_cur_posted_rcv_bufs(test_options_t *topt_p)
{
  int rc;

  pthread_mutex_lock(&topt_p->rcv_mutex);
  rc = topt_p->cur_posted_rcv_bufs;
  pthread_mutex_unlock(&topt_p->rcv_mutex);

  return rc;
}

/* gets "ceiling" number of pages required for buf_size */
static inline call_result_t get_num_pages_and_start( MT_virt_addr_t start, u_int32_t buf_size,
                                                   MT_virt_addr_t *actual_start, u_int32_t *num_pgs) 
{
  call_result_t rc;
  unsigned int page_shift;

  rc = MOSAL_get_page_shift(MOSAL_get_current_prot_ctx(), start, &page_shift);
  if ( rc != MT_OK ) {
    return rc;
  }
  *num_pgs = (u_int32_t)(((start + buf_size -1)>> page_shift) + 1 - (start >> page_shift));
  *actual_start = start & (~((MT_virt_addr_t)(1<<page_shift) - 1));
  return rc;
}


static void usage(void)
{
  printf(
        "\nUsage: SERVER SIDE:\n"
        " rctp --server [-h <HCA ID>] [--port <HCA port>] [--r] [--debug] [--nordma]\n"
        "               [common optional parameters]\n\n"
        "Usage: CLIENT SIDE:\n"
        " rctp  --client [-h <HCA ID>] [--port <HCA port>] [--debug]\n"
        "            [-n <transaction count>] [-r <peer's IP addr.>]\n"
        "            [--memwindows | --atomics] [--seed <seed>]\n"
        "            [common optional parameters]\n\n"
        " common optional parameters: \n"
        "            [--immediate] [-s <send msg. size>] [-m <MTU>]\n"
        "            [-u <max. SQ outstanding>] [-v <max. RQ outstanding>]\n"
        "            [--tcpport <TCP/IP port>]\n"
        "     if any common optional parameters are used, they must be the same\n"
        "     for both client and server\n\n");
}

static char * get_cmd_str(u_int32_t cmd)
{
    switch(cmd){
    case  CMD_END:
        return "CMD_END";
    case CMD_READ_BLOCK:
        return "READ(RDMA-W)";
    case CMD_WRITE_BLOCK:
        return "WRITE(RDMA-R)";
    case CMD_ATOMIC_FA:
        return "CMD_ATOMIC_FA";
    case CMD_ATOMIC_CSWP:
        return "CMD_ATOMIC_CSWP";
    case CMD_ACK:
        return "CMD_ACK";
    case CMD_NAK:
        return "CMD_NAK";
    case CMD_NOP:
        return "CMD_NOP";
    default:
        return "UNKNOWN";
    }
}

#if defined(__WIN__) 
char *readline(const char *prompt)
{
	char *line = malloc(120);
	puts(prompt);
	return gets(line);
}
#endif

/*
 *  main
 */
int MT_API main(int argc, char *argv[])
{
  int ch, option_index = 0;
  char *endptr;
  VAPI_ret_t rc;
  test_options_t topt;
  u_int32_t i, free_idx, option_check;
  trans_ctx_t *tctx_p;
#ifdef __WIN__
    SetConsoleCtrlHandler(clean_handler, TRUE);
#else
  u_int32_t signum;
  sighandler_t old_handler;

  signum = SIGINT;
  old_handler = signal(signum, clean_handler);
  if ( old_handler == SIG_ERR ) {
    printf("failed to replace ^C signal handler\n");
    return EXIT_CODE;
  }
#endif  
  glob_opt = &topt;

  if (init_db(&topt) != 0 ) {
    return EXIT_CODE;
  }

  /* parse command line options */
  while ( 1 ) {
    static struct option long_options[] = {
      {"help", 0, 0, 0},
      {"server", 0, 0, 0},
      {"client", 0, 0, 0},
      {"debug", 0, 0, 0},
      {"memwindows", 0, 0, 0},
      {"immediate", 0, 0, 0},
      {"port", required_argument, 0, 0},
      {"port_down", 0, 0, 0},
      {"autordma", 0, 0, 0},
      {"nordma", 0, 0, 0},
      {"atomics", 0, 0, 0},
      {"srq", 0, 0, 0},
      {"r", 0, 0, 0},
      {"tcpport", required_argument, 0, 0},
      {"seed", required_argument, 0, 0},
      {0, 0, 0, 0}
    };

#if defined(__WIN__) 
    ch = MTGETOPT_getopt_long(argc, argv, "n:h:r:m:s:u:v:", long_options, &option_index);
#else
    ch = getopt_long(argc, argv, "n:h:r:m:s:u:v:", long_options, &option_index);
#endif    
    if ( ch == -1 ) {
      break;
    }
    //if (ch != 0) printf("Got short option = %c\n", ch);
    switch ( ch ) {
      case 0:
        //printf("Got long option index=%d\n",option_index);
        /* "Long" options */
        switch (option_index) {
          case 0: /*help*/
            usage();
            CLEAN_EXIT(0);
            break;
          case 1: /*server*/
            topt.mode = MODE_SERVER;
            break;
          case 2: /*client*/
            topt.mode = MODE_CLIENT;
            break;
          case 3: /*debug*/
            topt.debug = TRUE;
            break;
          case 4: /*memwindows*/
            topt.memwindows = TRUE;
            break;
          case 5: /*immediate*/
            topt.immediate = TRUE;
            break;
          case 6: /*port*/
            topt.port = (IB_port_t)strtoul(optarg, &endptr, 0);
            if ( endptr == optarg ) {
              printf("invalid port number specified\n");
              CLEAN_EXIT(EXIT_CODE);
            }
            break;
          case 7: /* port_down */
            topt.port_down= TRUE;
            break;
          case 8: /*autordma*/
            topt.autordma = TRUE;
            break;
          case 9: /*nordma*/
            topt.nordma = TRUE;
            break;
          case 10: /*atomics*/
            topt.atomics = TRUE;
            break;
          case 11: /*srq*/
            topt.srq = TRUE;
            break;
          case 12: /*r*/
            topt.server_auto_connect = TRUE;
            break;
          case 13: /*tcpport*/
            topt.tcp_port = (u_int32_t)strtoul(optarg, &endptr, 0);
            if ( endptr == optarg ) {
              printf("invalid TCP/IP port number specified\n");
              CLEAN_EXIT(TCP_ERROR_EXIT_CODE);
            }
            break;
          case 14:
            topt.seed = strtoul(optarg, &endptr, 0);
            if ( endptr == optarg ) {
              printf("invalid seed specified\n");
              CLEAN_EXIT(TCP_ERROR_EXIT_CODE);
            }
            topt.static_seed = 1;
            break;
          default:
            printf("invalid option argument\n");
            CLEAN_EXIT(EXIT_CODE);
        }
        break;
      
      /* "Short" options */
      case 'n':
        topt.num_trans =  strtoul(optarg, &endptr, 0);
        if ( endptr == optarg ) {
          printf("invalid number of transactions specified\n");
          CLEAN_EXIT(EXIT_CODE);
        }
        break;

      case 'h':
        if ( strlen(optarg) >= sizeof(topt.hca_id) ) {
          printf("hca_id (name) too long\n");
          CLEAN_EXIT(EXIT_CODE);
        }
        strcpy(topt.hca_id, optarg);
        break;

      case 'r': /* remote ip address */
        if ( strlen(optarg) < sizeof(topt.remote_ip) ) {
          strcpy(topt.remote_ip, optarg);
        }
        else {
          printf("invalid remote host address\n");
          CLEAN_EXIT(EXIT_CODE);
        }
        break;

      case 'm': /* mtu */
        {
          u_int32_t mtu;
          mtu =  strtoul(optarg, &endptr, 0);
          if ( endptr == optarg ) {
            printf("invalid MTU specified\n");
            CLEAN_EXIT(EXIT_CODE);
          }
          if ( get_ib_mtu(mtu, &topt.mtu) != VAPI_OK ) {
            printf("invalid MTU specified\n");
            CLEAN_EXIT(EXIT_CODE);
          }
        }
        break;
      case 's':
        {
          u_int32_t msgsz;
          msgsz =  strtoul(optarg, &endptr, 0);
          if ( endptr == optarg ) {
            printf("invalid message size specified specified\n");
            CLEAN_EXIT(EXIT_CODE);
          }
          if ( msgsz < sizeof(struct cmd_msg_st) ) {
            msgsz = sizeof(struct cmd_msg_st);
            printf("Warning: Requested size is too small. Using %d\n", msgsz);
          }
          topt.msgsz = msgsz;
        }
        break;
      case 'u':
        topt.max_oust_wr_sq =  strtoul(optarg, &endptr, 0);
        if ( endptr == optarg ) {
          printf("invalid number of outstanding work requests on the send queue\n");
          CLEAN_EXIT(EXIT_CODE);
        } else if (topt.max_oust_wr_sq <= SND_Q_SAFE) {
            printf("-u value (%d) is too small. Using %d (min allowed by test)\n", topt.max_oust_wr_sq, SND_Q_SAFE+1);
            topt.max_oust_wr_sq = SND_Q_SAFE+1;
        }
        break;

      case 'v':
        topt.max_oust_wr_rq =  strtoul(optarg, &endptr, 0);
        if ( endptr == optarg ) {
          printf("invalid number of outstanding work requests on the receive queue\n");
          CLEAN_EXIT(EXIT_CODE);
        } else if (topt.max_oust_wr_rq <= RCV_Q_SAFE) {
            printf("-v value (%d) is too small. Using %d (min allowed by test)\n", topt.max_oust_wr_rq, RCV_Q_SAFE+1);
            topt.max_oust_wr_rq = RCV_Q_SAFE+1;
        }
        break;
    }
  }

  if ( topt.mode == MODE_NA ) {
    usage();
    CLEAN_EXIT(EXIT_CODE);
  }

  option_check = 0;
  if (topt.memwindows == TRUE) option_check++;
  if (topt.atomics == TRUE) option_check++;

  if (option_check > 1  ) {
    printf("\nYou may specify only ONE of 'memwindows' or 'atomics'\n");
    usage();
    CLEAN_EXIT(EXIT_CODE);
  }

  if ( (topt.mode==MODE_SERVER) ) {
      topt.memwindows = FALSE; topt.atomics = FALSE;
  }


  if ( (topt.mode==MODE_SERVER) && (topt.server_auto_connect==TRUE) ) {
    topt.remote_ip[0] = 'r';
  }
  if ( topt.mode == MODE_CLIENT && topt.remote_ip[0] ) {
    /* client */
    rc = create_tcp_connection(topt.remote_ip, topt.tcp_port);
    if ( rc == -1  ) {
      printf("failed to create tcp connection with peer\n");
      CLEAN_EXIT(TCP_ERROR_EXIT_CODE);
    }
    topt.socket = rc;
  }
  else if ( topt.remote_ip[0] ) {
    /* server */
    rc = wait_for_connection(topt.tcp_port);
    if ( rc == -1  ) {
      printf("failed to accept connection from peer\n");
      CLEAN_EXIT(rc);
    }
    topt.socket = rc;
  }

  /* attempt to open the hca*/
  if ( topt.hca_id[0] == '\0' ) {
    /* hca_id not specified by the user */
    u_int32_t num_of_hcas, j;
    VAPI_hca_id_t *hca_id_buf_p, inst_hca_id;

    rc = EVAPI_list_hcas(1, &num_of_hcas, &inst_hca_id);
    if ( (rc != VAPI_OK) && (rc != VAPI_EAGAIN) ) {
      VAPIERR(rc);
      printf("error in EVAPI_list_hcas\n");
      CLEAN_EXIT(rc);
    }
    switch ( num_of_hcas ) {
      case 0:
        printf("There are no HCAs installed in your system\n");
        CLEAN_EXIT(EXIT_CODE);

      case 1:
        strcpy(topt.hca_id, inst_hca_id);
        printf("Using HCA: %s\n", topt.hca_id);
        break;

      default:
        hca_id_buf_p = malloc(sizeof(VAPI_hca_id_t)*num_of_hcas);
        if ( !hca_id_buf_p ) {
          GENERR("malloc failed");
          CLEAN_EXIT(EXIT_CODE);
        }
        rc = EVAPI_list_hcas(num_of_hcas, &num_of_hcas, hca_id_buf_p);
        if ( rc != VAPI_OK ) {
          VAPIERR(rc);
          free(hca_id_buf_p);
          printf("error in EVAPI_list_hcas\n");
          CLEAN_EXIT(rc);
        }
        printf("The following HCAs are installed in your system. Please use <-h hcaid> to specify one.\n");
        for ( j=0; j<num_of_hcas; ++j ) {
          printf("%s\n", hca_id_buf_p[j]);
        }
        free(hca_id_buf_p);
        CLEAN_EXIT(EXIT_CODE);
    }
  }

  /* try to get a handle to the given hca_id */
  rc = EVAPI_get_hca_hndl(topt.hca_id, &topt.hca_hndl);
  if ( rc != VAPI_OK ) {
    /* specified hca is not already open - try to open it */
    VAPIERR(rc);
    printf("Note! An HCA must be open in order for this program to work\n");
    CLEAN_EXIT(EXIT_CODE);
  }

  rc = VAPI_query_hca_cap(topt.hca_hndl, &topt.hca_vendor, &topt.hca_cap);
  if ( rc != VAPI_OK) {
    VAPIERR(rc);
    CLEAN_EXIT(rc);
  }
  PRINT_HCA_CAP(&topt.hca_vendor, &topt.hca_cap);

  if (topt.port_down) { /* don't wait for link to go up */
    printf("not waiting for port %d to go up -  assuming loopback\n", topt.port);
    if (topt.port == 0)  topt.port= 1;
    rc = VAPI_query_hca_port_prop(topt.hca_hndl, topt.port, &topt.hca_port);
    if (rc != VAPI_OK) {
      printf("Failed getting port properties for port %d (%s)\n", topt.port, VAPI_strerror_sym(rc));
      CLEAN_EXIT(EXIT_CODE);
    }

  } else if ( wait_link_up(&topt) != VAPI_OK ) {
    printf("failed waiting for link to become up\n");
    CLEAN_EXIT(EXIT_CODE);
  }
  printf("using port %d\n", topt.port);

  if ( topt.immediate == TRUE ) {
    topt.max_oust_wr_rq = 2*topt.max_oust_wr_rq;
  }
  if ( create_qp(&topt) != VAPI_OK ) {
    printf("create_qp failed\n");
    CLEAN_EXIT(EXIT_CODE);
  }


  rc = register_regions(&topt);
  if ( rc != VAPI_OK ) {
    printf("register_regions failed\n");
    CLEAN_EXIT(EXIT_CODE);
  }


  if (topt.mode == MODE_CLIENT) {
    if (topt.memwindows ) {
        rc = create_memory_windows(&topt);
        if ( rc != VAPI_OK ) {
          printf("Memory windows allocation failed\n");
          CLEAN_EXIT(EXIT_CODE);
        }
    }
  }

  topt.thread_run_cmd = TRUE;
  if ( create_threads(&topt) != 0 ) {
    return EXIT_CODE;
  }

  for ( i=0; i<topt.max_oust_wr_rq; ++i ) {
    rc = post_recv_buf(&topt);
    if ( rc != VAPI_OK ) {
      printf("post_recv_buf failed\n");
      CLEAN_EXIT(rc);
    }
    inc_cur_posted_rcv_bufs(&topt);
  }

  if ( topt.mode == MODE_CLIENT ) {
    /* client */
    send_local_param(&topt);
    get_remote_params(&topt);
    rc = move_to_rtr(&topt);
    if ( rc != VAPI_OK ) {
      printf("move_to_rtr failed\n");
      CLEAN_EXIT(rc);
    }
  }
  else {
    /* server */
    get_remote_params(&topt);
    rc = move_to_rtr(&topt);
    if ( rc != VAPI_OK ) {
      printf("move_to_rtr failed\n");
      CLEAN_EXIT(rc);
    }
    send_local_param(&topt);
    if ( topt.autordma ) {
      printf("waiting 5 seconds to allow peer to move to rtr\n");
      fflush(stdout);
      sleep(5);
    }
  }

  rc = move_to_rts(&topt);
  if ( rc != VAPI_OK ) {
    printf("move_to_rts failed\n");
    CLEAN_EXIT(rc);
  }

#if defined(__WIN__)
  MOSAL_socket_close(topt.socket);
#else
  close(topt.socket);
#endif  
  topt.socket = -1;

  print_peer_info(&topt);

/*=== main processing loop ===*/
  if ( topt.mode == MODE_CLIENT ) {
    /* client */
    u_int32_t trans_cntr;
    struct cmd_msg_st cmd;
    VAPI_mw_bind_t bind_prop;

    if ( topt.autordma ) {
      printf("client acts as target to rdma write message. press ^C when done\n");
      while ( 1 );
    }
    else {
      trans_cntr = 0;
      if ( !topt.static_seed ) {
        topt.seed = (unsigned int)MOSAL_get_time_counter();
      }
      printf("SEED=%u\n", topt.seed);
      srand(topt.seed);
      do {

        rc = VAPI_EAGAIN;
        if ( trans_cntr<topt.num_trans ) {
            cmd.block_num = (rand() / (RAND_MAX/(MAX_DISK_BLKS-1)));
            if (topt.atomics == FALSE) {
              cmd.command = (rand() > (RAND_MAX/2)) ? CMD_READ_BLOCK:CMD_WRITE_BLOCK;
              cmd.size = 1 + (rand() / (RAND_MAX/MAX_BLK_SZ));
              cmd.va = topt.data_mr.start + cmd.block_num*MAX_BLK_SZ + rand()/(RAND_MAX/MAX_BLK_SZ);;
            } else {
              int cmd_rand = rand();
#if 1
              if (cmd_rand > ((RAND_MAX/6)*5)) {
                  cmd.command = CMD_ATOMIC_FA;
                  cmd.size = ATOMIC_OP_DATA_SIZE;
                  cmd.block_num = 0;
                  cmd.va = (topt.data_mr.start + cmd.block_num*MAX_BLK_SZ
                      + ATOMIC_OP_DATA_SIZE + rand()/(RAND_MAX/MAX_BLK_SZ));
                  cmd.va -= (cmd.va % 8);
              } else if (cmd_rand > ((RAND_MAX/3)*2)) {
                  cmd.command = CMD_ATOMIC_CSWP;
                  cmd.size = ATOMIC_OP_DATA_SIZE;
                  cmd.block_num = 0;
                  cmd.va = (topt.data_mr.start + cmd.block_num*MAX_BLK_SZ
                      + ATOMIC_OP_DATA_SIZE + rand()/(RAND_MAX/MAX_BLK_SZ));
                  cmd.va -= (cmd.va % 8);
              } else if (cmd_rand > (RAND_MAX/3)) {
                  cmd.command =  CMD_READ_BLOCK;
                  cmd.size = 1 + (rand() / (RAND_MAX/MAX_BLK_SZ));
                  if (cmd.block_num == 0) cmd.block_num++;
                  cmd.va = topt.data_mr.start + cmd.block_num*MAX_BLK_SZ + rand()/(RAND_MAX/MAX_BLK_SZ);
              } else {
                  cmd.command =  CMD_WRITE_BLOCK;
                  cmd.size = 1 + (rand() / (RAND_MAX/MAX_BLK_SZ));
                  if (cmd.block_num == 0) cmd.block_num++;
                  cmd.va = topt.data_mr.start + cmd.block_num*MAX_BLK_SZ + rand()/(RAND_MAX/MAX_BLK_SZ);
              }
#else
              if (cmd_rand > (RAND_MAX/2)) {
                  cmd.command = CMD_ATOMIC_FA;
                  cmd.size = ATOMIC_OP_DATA_SIZE;
                  cmd.block_num = 0;
                  cmd.va = (topt.data_mr.start + cmd.block_num*MAX_BLK_SZ
                      + ATOMIC_OP_DATA_SIZE + rand()/(RAND_MAX/MAX_BLK_SZ));
                  cmd.va -= (cmd.va % 8);
              } else {
                  cmd.command = CMD_ATOMIC_CSWP;
                  cmd.size = ATOMIC_OP_DATA_SIZE;
                  cmd.block_num = 0;
                  cmd.va = (topt.data_mr.start + cmd.block_num*MAX_BLK_SZ
                      + ATOMIC_OP_DATA_SIZE + rand()/(RAND_MAX/MAX_BLK_SZ));
                  cmd.va -= (cmd.va % 8);
              }
#endif
		  }
          cmd.tid = trans_cntr;
          //cmd.rkey = topt.data_mr.r_key;
          cmd.wnd_idx = trans_cntr % topt.max_oust_wr_sq;
          free_idx=cmd.wnd_idx;

          if ( topt.memwindows ) {
#ifdef bdebug
            int ttt=0;
#endif
            while ( topt.mw_arr[free_idx].size != 0 ) {
              drain_rcv_queue(&topt);
              drain_send_queue(&topt);

#ifdef bdebug
              if ( ttt == 0 ) {
                printf("^"); fflush(stdout);
                ttt = 1;
              }
#endif
            }
          } 


          if ( topt.memwindows ) {
            bind_prop.mr_lkey = topt.data_mr.l_key;
            bind_prop.start = cmd.va;
            bind_prop.size = cmd.size;
//            bind_prop.acl = cmd.command==CMD_READ_BLOCK ? VAPI_EN_REMOTE_WRITE : VAPI_EN_REMOTE_READ;
            switch(cmd.command) {
                case  CMD_READ_BLOCK:
                    bind_prop.acl = VAPI_EN_REMOTE_WRITE;
                    break;
                case  CMD_WRITE_BLOCK:
                    bind_prop.acl = VAPI_EN_REMOTE_READ;
                    break;
                case  CMD_ATOMIC_FA:
                case  CMD_ATOMIC_CSWP:
                    bind_prop.acl = VAPI_EN_REMOTE_ATOM | VAPI_EN_REMOTE_READ | VAPI_EN_REMOTE_WRITE;
                    break;
                default:
                    bind_prop.acl = 0;
            }

            tctx_p = malloc(sizeof(*tctx_p));
            if ( !tctx_p ) {
              CLEAN_EXIT(EXIT_CODE);
            }

            tctx_p->cqe_opcode = VAPI_CQE_SQ_BIND_MRW;
            tctx_p->ctx.sq_bind_ctx.mw_idx = cmd.wnd_idx;
            tctx_p->ctx.sq_bind_ctx.bind_size = (u_int32_t)bind_prop.size;

            while ( (get_cur_send_outstanding(&topt)>=((int)topt.max_oust_wr_sq-(int)SND_Q_SAFE))  ) {
              drain_send_queue(&topt);
#ifdef bdebug
              printf(")"); fflush(stdout);
#endif
            }
            rc = VAPI_bind_mw(topt.hca_hndl, topt.mw_arr[cmd.wnd_idx].mw_hndl, &bind_prop, topt.qp_hndl,
                              (VAPI_wr_id_t)((MT_virt_addr_t)tctx_p), VAPI_SIGNALED, &cmd.rkey);
            if ( rc != VAPI_OK ) {
              printf("Error binding memory window\n");
              fflush(stdout);
              CLEAN_EXIT(EXIT_CODE);
            }
            else {
#ifdef bdebug
              printf("posted_mem_wnd_idx %d\n", tctx_p->ctx.sq_bind_ctx.mw_idx);
              printf("after post "); print_ctx(&topt, tctx_p);
#endif
              topt.mw_arr[cmd.wnd_idx].size = (u_int32_t)bind_prop.size;
              inc_cur_send_outstanding(&topt);
              topt.inflight_mw_binds++;
            }
            topt.mw_arr[cmd.wnd_idx].rkey = cmd.rkey;
          
          } else {
            /* no use of memory windows */
            cmd.rkey = topt.data_mr.r_key;
          }


          get_cmd_str(cmd.command); // TK - to fix compilation problems
          MTL_DEBUG3("Transaction %d: Asking to %s %d bytes from block %d, va="U64_FMT", rkey=0x%x\n",trans_cntr,
                     get_cmd_str(cmd.command), cmd.size, cmd.block_num, cmd.va, cmd.rkey);
          while ((rc=post_send_buf(&topt, &cmd, topt.msgsz)) != VAPI_OK ) {
            if ( rc!=VAPI_EAGAIN ) {
              CLEAN_EXIT(rc);
            }
            drain_rcv_queue(&topt);
#ifdef bdebug
            printf("+"); fflush(stdout);
#endif
          }

#ifdef bdebug
          printf("xyz wndidx=%d\n", cmd.wnd_idx);
#endif
          trans_cntr++;
        }

        drain_rcv_queue(&topt);
      }
      while ( get_snd_rcv_balance(&topt) || (trans_cntr<topt.num_trans) ); /* end of do block */

      cmd.command = CMD_END;
      rc = post_send_buf(&topt, &cmd, topt.msgsz);
      if ( rc != VAPI_OK ) {
        CLEAN_EXIT(rc);
      }
      do {
        drain_send_queue(&topt);
#ifdef bdebug
        printf("9"); fflush(stdout);
#endif
      }
      while ( get_cur_send_outstanding(&topt) || topt.inflight_mw_binds );
    }
  }
  else {
    /* server */
    struct cmd_msg_st cmd;

    while ( (topt.peer_ended==FALSE) || get_cur_send_outstanding(&topt) ) {
      VAPI_wc_desc_t comp_desc;

      if ( topt.autordma == FALSE ) {
        rc = VAPI_poll_cq(topt.hca_hndl, topt.rqcq_hndl, &comp_desc);
#ifdef im_debug
       if ( rc==VAPI_OK ) {
         if ( comp_desc.status == VAPI_SUCCESS ) {
           print_completion_opcode(comp_desc.opcode);
         }
       }
#endif
      }
      else {
        /* fabricate a message */
        memset(&comp_desc, 0, sizeof(comp_desc));
        cmd.block_num = 0;
        cmd.command = CMD_READ_BLOCK;
        cmd.rkey = topt.auto_rkey;
        cmd.size = MAX_BLK_SZ;
        cmd.tid = 0;
        cmd.va = topt.auto_va;
        cmd.wnd_idx = 0;
        cmd_host_to_be(&cmd);
        comp_desc.id = (VAPI_wr_id_t)(MT_ulong_ptr_t)(&cmd);
        rc = VAPI_OK;
        topt.peer_ended = 1; /* only once */
      }
      switch ( rc ) {
      case VAPI_OK:
          tctx_p = (trans_ctx_t *)(MT_virt_addr_t)comp_desc.id;
          if ( comp_desc.status != VAPI_SUCCESS ) {
            printf("unsuccessful completion on receive queue completion status=%s\n", VAPI_wc_status_sym(comp_desc.status));
            CLEAN_EXIT(-1);
          }
          if ( (comp_desc.opcode==VAPI_CQE_RQ_SEND_DATA) || (comp_desc.opcode==VAPI_CQE_RQ_RDMA_WITH_IMM) ) {
            dec_snd_rcv_balance(&topt);
            if ( comp_desc.imm_data_valid ) {
#ifdef VERBOSE
              fprintf(stderr, "immediate data received %08x\n", comp_desc.imm_data);
#endif
            }
          }
          rc = process_command(&topt, &comp_desc, &cmd);
          if ( rc != MT_OK ) {
            CLEAN_EXIT(rc);
          }
          if ( (topt.autordma == FALSE) && (topt.peer_ended == FALSE) ) {
            /* Return buffer to RQ */
            rc = post_recv_buf(&topt);
            if ( rc != MT_OK ) {
              CLEAN_EXIT(rc);
            }
            cmd.command = CMD_ACK;
#ifdef bdebug
            printf("sending response: memw idx = %d\n", cmd.wnd_idx);
#endif

            /* cmd.wnd_idx is echoed back */
            while ( post_send_buf(&topt, &cmd, comp_desc.byte_len) != VAPI_OK ) {
              drain_send_queue(&topt);
//            printf("&");
            }
          }
          break;

        case VAPI_CQ_EMPTY:
          drain_send_queue(&topt);
          break;

        default:
          printf("poll receive Q failed: %s\n", VAPI_strerror_sym(rc));
          drain_send_queue(&topt);
          CLEAN_EXIT(rc);

      }
    }
  }


  topt.thread_run_cmd = FALSE;
  CLEAN_EXIT(0);
}


/*
 *  create_qp
 */
static VAPI_ret_t create_qp(test_options_t *topt_p)
{
  VAPI_ret_t rc;
  VAPI_qp_init_attr_t  qp_init_attr;
  VAPI_qp_init_attr_ext_t qp_init_attr_ext;
  VAPI_qp_attr_t       qp_attr;
  VAPI_qp_attr_mask_t  qp_attr_mask;
  VAPI_qp_cap_t        qp_cap;
  VAPI_srq_attr_t      srq_attr_in, srq_attr_out;

  /* create completion queue - used for both send and receive queues */
  rc = VAPI_create_cq( topt_p->hca_hndl, topt_p->req_cqe_num, &topt_p->sqcq_hndl, &topt_p->sqcq_cqe_num);
  if ( rc != VAPI_OK) {
    VAPIERR(rc);
    return rc;
  }


  /* create completion queue - used for both send and receive queues */
  rc = VAPI_create_cq( topt_p->hca_hndl, topt_p->req_cqe_num, &topt_p->rqcq_hndl, &topt_p->rqcq_cqe_num);
  if ( rc != VAPI_OK) {
    VAPIERR(rc);
    return rc;
  }

  /* allocate a protection domain to be associated with QP */
  rc = VAPI_alloc_pd(topt_p->hca_hndl, &topt_p->pd_hndl);
  if ( rc != VAPI_OK) {
    VAPIERR(rc);
    return rc;
  }

  if (topt_p->srq) {
    srq_attr_in.max_outs_wr= topt_p->max_oust_wr_rq;
    srq_attr_in.max_sentries= MAX_SG_SIZE_RQ;
    srq_attr_in.pd_hndl= topt_p->pd_hndl;
    rc= VAPI_create_srq(topt_p->hca_hndl, &srq_attr_in, &(topt_p->srq_hndl), &srq_attr_out);
    if ( rc != VAPI_OK) {
      VAPIERR(rc);
      return rc;
    }
    printf("Using SRQ with max_oust_wr_rq=%d.\n", topt_p->max_oust_wr_rq);
  } 

  qp_init_attr.sq_cq_hndl = topt_p->sqcq_hndl;
  qp_init_attr.rq_cq_hndl = topt_p->rqcq_hndl;

  qp_init_attr.cap.max_oust_wr_sq = topt_p->max_oust_wr_sq;
  qp_init_attr.cap.max_oust_wr_rq = topt_p->max_oust_wr_rq;
  qp_init_attr.cap.max_sg_size_sq = MAX_SG_SIZE_SQ;
  qp_init_attr.cap.max_sg_size_rq = MAX_SG_SIZE_RQ;

  qp_init_attr.rdd_hndl = 0; /* N/A for RC transport service */
  qp_init_attr.sq_sig_type = VAPI_SIGNAL_REQ_WR;
  qp_init_attr.rq_sig_type = VAPI_SIGNAL_ALL_WR;
  qp_init_attr.pd_hndl = topt_p->pd_hndl;
  qp_init_attr.ts_type = VAPI_TS_RC; /* RC transport service */

  qp_init_attr_ext.srq_hndl= topt_p->srq_hndl;

  /* create the QP */
  rc = VAPI_create_qp_ext(topt_p->hca_hndl, &qp_init_attr, &qp_init_attr_ext,
                          &topt_p->qp_hndl, &topt_p->qp_prop);
  if ( rc != VAPI_OK ) {
    VAPIERR(rc);
    return rc;
  }
  PRINT_QP_PROPS(&topt_p->qp_prop);


  QP_ATTR_MASK_CLR_ALL(qp_attr_mask);
  qp_attr.qp_state = VAPI_INIT; QP_ATTR_MASK_SET(qp_attr_mask,QP_ATTR_QP_STATE);
  qp_attr.pkey_ix  = 0; QP_ATTR_MASK_SET(qp_attr_mask,QP_ATTR_PKEY_IX);  
  qp_attr.port     = topt_p->port; QP_ATTR_MASK_SET(qp_attr_mask,QP_ATTR_PORT);
  qp_attr.remote_atomic_flags = VAPI_EN_REM_WRITE | VAPI_EN_REM_READ;
      if (topt_p->atomics) {
           qp_attr.remote_atomic_flags |= VAPI_EN_REM_ATOMIC_OP;
      }
      QP_ATTR_MASK_SET(qp_attr_mask, QP_ATTR_REMOTE_ATOMIC_FLAGS);

  rc = VAPI_modify_qp(topt_p->hca_hndl, topt_p->qp_hndl, &qp_attr, &qp_attr_mask, &qp_cap);
  if ( rc != VAPI_OK ) {
    VAPIERR(rc);
    return rc;
  }

  return 0;
}



/*
 *  wait_link_up
 */
static VAPI_ret_t wait_link_up(test_options_t *topt_p)
{
  VAPI_ret_t rc;
  IB_port_t port, start=1, end=topt_p->hca_cap.phys_port_num;
  int first_time=1;
#define SECS_TO_WAIT  			0
#define NANOSECS_TO_WAIT 		500000000 
#if defined(__WIN__) 
  int req = (SECS_TO_WAIT * 1000 + NANOSECS_TO_WAIT / 1000000);
#else
  struct timespec req = {SECS_TO_WAIT, NANOSECS_TO_WAIT};
#endif  

  if ( topt_p->port == 0 ) {
    start=1;
    end=topt_p->hca_cap.phys_port_num;
  }
  else {
    start=topt_p->port;
    end=topt_p->port;
  }

  while ( 1 ) {
    for ( port=start; port<=end; ++port ) {
      rc = VAPI_query_hca_port_prop(topt_p->hca_hndl, port, &topt_p->hca_port);
      if ( rc != VAPI_OK ) {
        if ( !first_time ) {
          printf("\n");
        }
        return rc;
      }
      if ( topt_p->hca_port.state == PORT_ACTIVE ) {
        if ( topt_p->hca_port.max_mtu < topt_p->mtu ) {
          printf("Requested MTU is too large\n");
          return VAPI_ERR;
        }
        topt_p->port = port;
        return VAPI_OK;
      }
    }
    if ( first_time ) {
      printf("waiting for link up ");
      first_time = 0;
    }
    else {
      printf(".");
    }
#if defined(__WIN__)
    sleep(req);
#else
    nanosleep(&req, 0);
#endif    
    fflush(stdout);
  }
}



/*
 *  get_remote_params
 */
static int get_remote_params(test_options_t *topt_p)
{
    char endianness;
#ifdef MT_LITTLE_ENDIAN
    char my_endian = 'l';
#else
    char my_endian = 'b';
#endif     

  if ( topt_p->socket == -1 ) {
    char *line;
    u_int32_t tmp;
    IB_lid_t lid;
    IB_wqpn_t qpn;
    char *endptr;
    
    line = readline("Enter peer's LID: ");
    tmp = strtoul(line, &endptr, 0);
    free(line);
    if ( endptr == line ) {
      return -1;
    }
    lid = (IB_lid_t)tmp;

    line = readline("Enter peer's QP number: ");
    tmp = strtoul(line, &endptr, 0);
    free(line);
    if ( endptr == line ) {
      return -1;
    }
    qpn = tmp;

    line = readline("Enter peer's endianness (b or l): ");
    endianness = *line;
    free(line);
    if ( endianness != 'b' && endianness != 'l' ) {
      return -1;
    }

    if (endianness != my_endian) {
        topt_p->swap_endianness = TRUE;
    } else {
        topt_p->swap_endianness = FALSE;
    }
    topt_p->remote_lid = lid;
    topt_p->remote_qpn = qpn;
    return 0;
  }
  else {
    int rc;

    struct oob_msg_st oob;
#if defined(__WIN__) 
    rc = MOSAL_socket_recv(topt_p->socket, (void *)&oob, sizeof(oob), 0);
#else
    rc = recv(topt_p->socket, (char *)&oob, sizeof(oob), 0);
#endif    
    #ifdef __WIN__
    #ifdef im_debug
    printf( "get_remote_params: rc %d, oob.mode '%c', mode '%c', \n", rc,oob.mode, topt_p->mode + '0');
    #endif
    #endif
    if ( rc != sizeof(oob) ) {
      return -1;
    }
    if ( (oob.mode=='c') && (topt_p->mode==MODE_CLIENT) ) {
      return -1;
    }
    else if ( (oob.mode=='s') && (topt_p->mode==MODE_SERVER) ) {
      return -1;
    }
    endianness = oob.endian;
    if (endianness != 'b' && endianness != 'l') {
        return -1;
    }
    if (endianness != my_endian) {
        topt_p->swap_endianness = TRUE;
    } else {
        topt_p->swap_endianness = FALSE;
    }
    topt_p->remote_lid = ntohs(oob.lid);
    topt_p->remote_qpn = ntohl(oob.wqpn);
    topt_p->auto_rkey = ntohl(oob.rkey);
    topt_p->auto_va = MOSAL_be64_to_cpu(oob.va);
    topt_p->auto_size = ntohl(oob.size);
    #ifdef __WIN__
    #ifdef im_debug
    printf( "get_remote_params: remote_lid %hd, remote_qpn 0x%x, \n\t auto_rkey 0x%x, auto_va 0x%I64x, auto_size %hd \n",
    	topt_p->remote_lid, topt_p->remote_qpn, topt_p->auto_rkey, topt_p->auto_va, topt_p->auto_size ); 
    #endif
    #endif
    return 0;
  }
}


/*
 *  send_local_param
 */
static int send_local_param(test_options_t *topt_p)
{
  int rc;
  struct oob_msg_st oob;
#ifdef MT_LITTLE_ENDIAN
    char my_endian = 'l';
#else
    char my_endian = 'b';
#endif     

    #ifdef __WIN__
    #ifdef im_debug
    printf( "send_local_param: remote_lid %hd, remote_qpn 0x%x, \n\t auto_rkey 0x%x, auto_va 0x%I64x, auto_size %hd \n",
    	topt_p->hca_port.lid, topt_p->qp_prop.qp_num, topt_p->data_mr.r_key, topt_p->data_mr.start, topt_p->data_mr.size ); 
    #endif
    #endif

  if ( topt_p->socket == -1 ) {
    printf("Local LID = 0x%04x\n", topt_p->hca_port.lid);
    printf("Local QP number = 0x%06x\n", topt_p->qp_prop.qp_num);
    printf("Local Endianness = %c\n", my_endian);
    /*printf("Warning! auto rdma params not supported in this mode\n");*/
    return 0;
  }
  else {
    if ( topt_p->mode == MODE_CLIENT ) {
      oob.mode = 'c';
    }
    else 
      if ( topt_p->mode == MODE_SERVER ) {
        oob.mode = 's';
      }
      else {
      	  printf( "ERROR: Illegal mode parameter (topt_p->mode = '%c')\n", topt_p->mode );
      	  return -1;
      	}
    oob.endian = my_endian;
    oob.lid = htons(topt_p->hca_port.lid);
    oob.wqpn = htonl(topt_p->qp_prop.qp_num);
    oob.rkey = htonl(topt_p->data_mr.r_key);
    oob.va = MOSAL_cpu_to_be64(topt_p->data_mr.start);
    oob.size = htonl((u_int32_t)topt_p->data_mr.size);
#if defined(__WIN__) 
    rc = MOSAL_socket_send(topt_p->socket, (void *)&oob, sizeof(oob), 0);
#else
    rc = send(topt_p->socket, (char *)&oob, sizeof(oob), 0);
#endif    
    if ( rc != sizeof(oob) ) {
      return -1;
    }
    return 0;
  }
}


/*
 *  register_regions
 */
static VAPI_ret_t register_regions(test_options_t *topt_p)
{
  VAPI_ret_t rc;
  VAPI_mrw_t mrw;
  void *ptr;
  VAPI_mr_hndl_t  mr_hndl;
  u_int32_t data_blk_sz;

  /* Region for SQ buffers */
  ptr = VMALLOC(topt_p->msgsz*topt_p->max_oust_wr_sq);
  if ( !ptr ) {
    return VAPI_EAGAIN;
  }

  mrw.type = VAPI_MR;
  mrw.start = (VAPI_virt_addr_t)(MT_virt_addr_t)ptr;
  mrw.size = topt_p->msgsz*topt_p->max_oust_wr_sq;
  mrw.pd_hndl = topt_p->pd_hndl;
  mrw.acl = 0; /* read permissions will suffice */
  rc = VAPI_register_mr(topt_p->hca_hndl, &mrw, &mr_hndl, &topt_p->send_mr);
  if ( rc != VAPI_OK ) {
    return rc;
  }
  topt_p->send_mr_hndl = mr_hndl;

  /* Region for RQ buffers */
  ptr = VMALLOC(topt_p->max_oust_wr_rq*topt_p->msgsz);
  if ( !ptr ) {
    return VAPI_EAGAIN;
  }

  mrw.type = VAPI_MR;
  mrw.start = (VAPI_virt_addr_t)(MT_virt_addr_t)ptr;
  mrw.size = topt_p->max_oust_wr_rq*topt_p->msgsz;
  mrw.pd_hndl = topt_p->pd_hndl;
  mrw.acl = VAPI_EN_LOCAL_WRITE;
  rc = VAPI_register_mr(topt_p->hca_hndl, &mrw, &mr_hndl, &topt_p->rcv_mr);
  if ( rc != VAPI_OK ) {
    return rc;
  }
  topt_p->rcv_mr_hndl = mr_hndl;

  /* Region for RDMA accesses */
  if ( topt_p->mode == MODE_CLIENT ) {
    data_blk_sz = DISK_SZ; /* MAX_BLK_SZ; */
  }
  else {
    data_blk_sz = DISK_SZ;
  }


 /* allocate the DMA area. */
  ptr = VMALLOC(data_blk_sz);
  if ( !ptr ) {
    return VAPI_EAGAIN;
  }


  mrw.type = VAPI_MR;
  mrw.start = (VAPI_virt_addr_t)(MT_virt_addr_t)ptr;
  mrw.size = data_blk_sz;
  mrw.pd_hndl = topt_p->pd_hndl;
  if ( topt_p->memwindows ) {
    mrw.acl = VAPI_EN_LOCAL_WRITE | VAPI_EN_MEMREG_BIND;
  }
  else {
    mrw.acl = VAPI_EN_LOCAL_WRITE | VAPI_EN_REMOTE_WRITE | VAPI_EN_REMOTE_READ | VAPI_EN_REMOTE_ATOM;
  }
  rc = VAPI_register_mr(topt_p->hca_hndl, &mrw, &mr_hndl, &topt_p->data_mr);
  if ( rc != VAPI_OK ) {
    return rc;
  }
  topt_p->data_mr_hndl = mr_hndl;

  return VAPI_OK;
}


/*
 *  post_recv_buf
 */
static VAPI_ret_t post_recv_buf(test_options_t *topt_p)
{
  VAPI_ret_t rc;
  VAPI_sg_lst_entry_t sg_lst;
  VAPI_rr_desc_t rr_desc;
  u_int32_t bufptr;

  bufptr = (u_int32_t)(((u_int32_t)topt_p->last_posted_rcv_id)%((u_int32_t)topt_p->max_oust_wr_rq));
  sg_lst.addr = topt_p->rcv_mr.start + bufptr*topt_p->msgsz;
  sg_lst.len = topt_p->msgsz;
  sg_lst.lkey = topt_p->rcv_mr.l_key;

  rr_desc.id = sg_lst.addr;
  rr_desc.opcode = VAPI_RECEIVE;
  rr_desc.comp_type = VAPI_SIGNALED;
  rr_desc.sg_lst_p = &sg_lst;
  rr_desc.sg_lst_len = 1;
  if (!topt_p->srq) {
    rc = VAPI_post_rr(topt_p->hca_hndl, topt_p->qp_hndl, &rr_desc);
  } else {
    rc = VAPI_post_srq(topt_p->hca_hndl, topt_p->srq_hndl, 1, &rr_desc, NULL);
  } 
  MTL_DEBUG3("POSTED RCV BUF, qp=" MT_ULONG_PTR_FMT ", last_posted_rcv_id=" U64_FMT ", rc=%d\n", 
         topt_p->qp_hndl, (u_int64_t )topt_p->last_posted_rcv_id, rc);
  topt_p->last_posted_rcv_id++;
  return rc;
}


/*
 *  post_send_buf
 */
static VAPI_ret_t post_send_buf(test_options_t *topt_p, struct cmd_msg_st *cmd_p, u_int32_t size_to_send)
{
  VAPI_ret_t rc;
  VAPI_sr_desc_t sr_desc;
  VAPI_sg_lst_entry_t sg_lst;
  u_int32_t bufptr;
  struct cmd_msg_st *local_cmd_p;
  trans_ctx_t *tctx_p;


  if ( (get_cur_posted_rcv_bufs(topt_p)<=RCV_Q_SAFE) ||
          (get_snd_rcv_balance(topt_p) >= ((int)topt_p->max_oust_wr_sq-(int)SND_Q_SAFE) )) {
      MTL_DEBUG3("POSTED RCV BUFS = %d, BALANCE=%d\n", 
             get_cur_posted_rcv_bufs(topt_p),
             get_snd_rcv_balance(topt_p));
    return VAPI_EAGAIN;
  }


  tctx_p = malloc(sizeof(*tctx_p));
  if ( !tctx_p ) {
    printf("MALLOC tctx_p failure\n");
    return VAPI_EAGAIN;
  }

  topt_p->fill_val++;
  topt_p->last_posted_send_id++;
  bufptr = (u_int32_t)(topt_p->last_posted_send_id%topt_p->max_oust_wr_sq);

  local_cmd_p = (struct cmd_msg_st *)((MT_virt_addr_t)(topt_p->send_mr.start+topt_p->msgsz*bufptr));
  *local_cmd_p = *cmd_p;

  cmd_host_to_be(local_cmd_p);

  sg_lst.addr = (MT_virt_addr_t)local_cmd_p;
  sg_lst.len = size_to_send;
  sg_lst.lkey = topt_p->send_mr.l_key;

//  sr_desc.id = topt_p->last_posted_send_id;
  sr_desc.opcode = topt_p->immediate==FALSE ? VAPI_SEND : VAPI_SEND_WITH_IMM;
  sr_desc.comp_type = VAPI_SIGNALED;  
  sr_desc.sg_lst_p = &sg_lst;         
  sr_desc.sg_lst_len = 1;             
  sr_desc.imm_data = topt_p->mode==MODE_CLIENT ? 0x55555555 : 0xAAAAAAAA;
  sr_desc.fence = TRUE;   /* In case we are sending a notification after RDMA-R */
  sr_desc.set_se = FALSE;


  while ( (get_cur_send_outstanding(topt_p)>=((int)topt_p->max_oust_wr_sq-(int)SND_Q_SAFE))  ) {
#ifdef VERBOSE
    printf("snd_rcv_balance %d\n", get_snd_rcv_balance(topt_p));
    printf("cur_send_outstanding %d\n", get_cur_send_outstanding(topt_p));
    printf("cur_posted_rcv_bufs %d\n", get_cur_posted_rcv_bufs(topt_p));
    printf("+");
#endif
    //printf("CURR_SEND_OUTST = %d\n",get_cur_send_outstanding(topt_p));
    drain_send_queue(topt_p);
//    return MT_EAGAIN;
  }

#ifdef VERBOSE
  printf("attempt to post send request. wrid="U64_FMT" ...", sr_desc.id);
#endif
  tctx_p->cqe_opcode = VAPI_CQE_SQ_SEND_DATA;
  sr_desc.id = (VAPI_wr_id_t)((MT_virt_addr_t)tctx_p);
#ifdef im_debug
  print_post_opcode(sr_desc.opcode, tctx_p);
#endif
  rc = VAPI_post_sr(topt_p->hca_hndl, topt_p->qp_hndl, &sr_desc);
  if ( rc == VAPI_OK ) {
#ifdef im_debug
  printf("success\n");
#endif
#ifdef VERBOSE
    printf("OK\n");
#endif
#ifdef bdebug
    printf("after post "); print_ctx(topt_p, tctx_p);
    printf("==== address=%p, lkey=%08x, bufptr=%d, mwidx=%d\n", local_cmd_p, sg_lst.lkey, bufptr, cmd_p->wnd_idx);
#endif
    inc_cur_send_outstanding(topt_p);
    if ( topt_p->immediate &&  (cmd_p->command==CMD_READ_BLOCK) ) {
      inc_snd_rcv_balance(topt_p);
    }
    inc_snd_rcv_balance(topt_p);
  }
  else {
#ifdef im_debug
  printf("failed\n");
#endif
  }
  return rc;
}                                     




/*
 *  move_to_rtr
 */
static VAPI_ret_t move_to_rtr(test_options_t *topt_p)
{
  VAPI_qp_attr_t       qp_attr;
  VAPI_qp_attr_mask_t  qp_attr_mask;
  VAPI_qp_cap_t        qp_cap;
  VAPI_ret_t rc;

  QP_ATTR_MASK_CLR_ALL(qp_attr_mask);
  qp_attr.qp_state = VAPI_RTR; QP_ATTR_MASK_SET(qp_attr_mask,QP_ATTR_QP_STATE);
  qp_attr.av.sl            = USED_SL;
  qp_attr.av.grh_flag      = FALSE;
  qp_attr.av.dlid          = topt_p->remote_lid;
  qp_attr.av.static_rate   = 2; /* 1x */
  qp_attr.av.src_path_bits = 0; QP_ATTR_MASK_SET(qp_attr_mask,QP_ATTR_AV);
  qp_attr.path_mtu      = topt_p->mtu; QP_ATTR_MASK_SET(qp_attr_mask,QP_ATTR_PATH_MTU);
  qp_attr.rq_psn           = START_PSN; QP_ATTR_MASK_SET(qp_attr_mask,QP_ATTR_RQ_PSN);
  qp_attr.qp_ous_rd_atom  = QP_OUS_RD_ATOM; QP_ATTR_MASK_SET(qp_attr_mask,QP_ATTR_QP_OUS_RD_ATOM);
  qp_attr.dest_qp_num = topt_p->remote_qpn; QP_ATTR_MASK_SET(qp_attr_mask,QP_ATTR_DEST_QP_NUM);
  qp_attr.min_rnr_timer = 0; QP_ATTR_MASK_SET(qp_attr_mask,QP_ATTR_MIN_RNR_TIMER);

  rc = VAPI_modify_qp(topt_p->hca_hndl, topt_p->qp_hndl, &qp_attr, &qp_attr_mask, &qp_cap);
  return rc;
}


/*
 *  move_to_rts
 */
static VAPI_ret_t move_to_rts(test_options_t *topt_p)
{
  VAPI_qp_attr_t       qp_attr;
  VAPI_qp_attr_mask_t  qp_attr_mask;
  VAPI_qp_cap_t        qp_cap;
  VAPI_ret_t rc;

  QP_ATTR_MASK_CLR_ALL(qp_attr_mask);
  qp_attr.qp_state = VAPI_RTS; QP_ATTR_MASK_SET(qp_attr_mask,QP_ATTR_QP_STATE);

  qp_attr.qp_state         = VAPI_RTS;  QP_ATTR_MASK_SET(qp_attr_mask,QP_ATTR_QP_STATE);
  qp_attr.sq_psn           = START_PSN;    QP_ATTR_MASK_SET(qp_attr_mask,QP_ATTR_SQ_PSN);
  qp_attr.timeout   = 18;     QP_ATTR_MASK_SET(qp_attr_mask,QP_ATTR_TIMEOUT);
  qp_attr.retry_count   = 6;     QP_ATTR_MASK_SET(qp_attr_mask,QP_ATTR_RETRY_COUNT);
  qp_attr.rnr_retry     = 6;     QP_ATTR_MASK_SET(qp_attr_mask,QP_ATTR_RNR_RETRY);
  qp_attr.ous_dst_rd_atom  = QP_OUS_RD_ATOM; QP_ATTR_MASK_SET(qp_attr_mask,QP_ATTR_OUS_DST_RD_ATOM);
  rc = VAPI_modify_qp(topt_p->hca_hndl, topt_p->qp_hndl, &qp_attr, &qp_attr_mask, &qp_cap);
  return rc;
}



/*
 *  init_db
 */
static int init_db(test_options_t *topt_p)
{
  memset(topt_p, 0, sizeof(test_options_t));
  topt_p->num_trans = 1;
  topt_p->hca_id[0] = 0;
  topt_p->req_cqe_num = 1<<15;
  topt_p->inflight_mw_binds = 0;

  topt_p->send_mr_hndl = VAPI_INVAL_HNDL;
  topt_p->rcv_mr_hndl = VAPI_INVAL_HNDL;
  topt_p->data_mr_hndl = VAPI_INVAL_HNDL;
  topt_p->qp_hndl = VAPI_INVAL_HNDL;
  topt_p->srq_hndl = VAPI_INVAL_SRQ_HNDL;
  topt_p->sqcq_hndl = VAPI_INVAL_HNDL;
  topt_p->rqcq_hndl = VAPI_INVAL_HNDL;
  topt_p->pd_hndl = VAPI_INVAL_HNDL;
  topt_p->socket = -1;
  topt_p->remote_ip[0] = '\0';
  topt_p->thread_run_cmd = FALSE;
  topt_p->peer_ended = FALSE;
  topt_p->cur_send_outstanding = 0;
  topt_p->cur_posted_rcv_bufs = 0;
  topt_p->snd_rcv_balance = 0;
#if ! defined(__WIN__) 
  topt_p->scomp_thread = -1; /* is this an invalid thread handle ?? */
  topt_p->rcomp_thread = -1; /* is this an invalid thread handle ?? */
#endif  
  topt_p->imd_ret = FALSE;
  topt_p->mtu = MTU256; /* default mtu */
  topt_p->msgsz = SEND_MSG_SIZE;
  topt_p->max_oust_wr_sq = MAX_OUST_WR_SQ;
  topt_p->max_oust_wr_rq = MAX_OUST_WR_RQ;
  topt_p->port = 0;
  topt_p->port_down= FALSE;
  topt_p->autordma = FALSE;
  topt_p->nordma = FALSE;
  topt_p->debug = FALSE;
  topt_p->immediate = FALSE;
  topt_p->memwindows = FALSE;
  topt_p->server_auto_connect = FALSE;
  topt_p->hca_hndl = -1;
  pthread_mutex_init(&topt_p->snd_mutex, NULL);
  pthread_mutex_init(&topt_p->rcv_mutex, NULL);
  pthread_mutex_init(&topt_p->bl_mutex, NULL);
  pthread_mutex_init(&topt_p->cln_mutex, NULL);
  MOSAL_mutex_init(&glob_mtx);
  topt_p->mw_arr = NULL;
  topt_p->total_send_completions = 0;
  topt_p->tcp_port = TCP_PORT_NO;
  topt_p->static_seed = 0;
  return 0;
}

/*
 *  cleanup
 */
static void rctp_cleanup(test_options_t *topt_p, int errcode)
{
  VAPI_ret_t rc;
  void *p;
  u_int32_t i;

  if ( topt_p->debug == TRUE ) {
    mem_debugger();
  }

#ifdef debug_cleanup
  MTL_DEBUG5("*** 1 ***\n");
  fflush(stdout);
#endif
// pthread_mutex_lock(&topt_p->cln_mutex);
#ifdef debug_cleanup
  MTL_DEBUG5("*** 2 ***\n");
  fflush(stdout);
#endif
  if ( topt_p->socket != -1 ) {
#if defined(__WIN__) 
    MOSAL_socket_close(topt_p->socket);
#else
    close(topt_p->socket);
#endif    
  }
  topt_p->socket = -1;


//  if ( topt_p->scomp_thread != -1 ) {
//    pthread_join(topt_p->scomp_thread, NULL);
//  }
#if ! defined( __WIN__) 
  topt_p->scomp_thread = -1;
#endif  

#ifdef debug_cleanup
  MTL_DEBUG5("*** 3 ***\n");
  fflush(stdout);
#endif

//  if ( topt_p->rcomp_thread != -1 ) {
//    pthread_join(topt_p->rcomp_thread, NULL);
//  }
//#ifndef __WIN__
//  topt_p->rcomp_thread = -1;
//#endif


#ifdef debug_cleanup
  MTL_DEBUG5("*** 4 ***\n");
  fflush(stdout);
#endif

  p = (void *)((MT_virt_addr_t)(topt_p->send_mr.start));
  if ( topt_p->send_mr_hndl != VAPI_INVAL_HNDL ) {
    rc = VAPI_deregister_mr(topt_p->hca_hndl, topt_p->send_mr_hndl);
    if ( rc != VAPI_OK ) {
      printf("VAPI_deregister_mr failed\n");
    }
  }
  VFREE(p);
  topt_p->send_mr_hndl = VAPI_INVAL_HNDL;


#ifdef debug_cleanup
  MTL_DEBUG5("*** 5 ***\n");
  fflush(stdout);
#endif
  p = (void *)((MT_virt_addr_t)(topt_p->rcv_mr.start));
  if ( topt_p->rcv_mr_hndl != VAPI_INVAL_HNDL ) {
    rc = VAPI_deregister_mr(topt_p->hca_hndl, topt_p->rcv_mr_hndl);
    if ( rc != VAPI_OK ) {
      printf("VAPI_deregister_mr failed\n");
    }
    VFREE(p);
    topt_p->rcv_mr_hndl = VAPI_INVAL_HNDL;
  }

#ifdef debug_cleanup
  MTL_DEBUG5("*** 6 ***\n");
  fflush(stdout);
#endif

  if ( topt_p->mw_arr ) {
    for ( i=0; i<topt_p->max_oust_wr_sq; ++i ) {
      if ( (topt_p->mw_arr)[i].mw_hndl != VAPI_INVAL_HNDL ) {
	    if ( (topt_p->mw_arr)[i].size != 0 ) printf("meory window still in use: index=%d\n", i);
        rc = VAPI_dealloc_mw(topt_p->hca_hndl, (topt_p->mw_arr)[i].mw_hndl);
        if ( rc != VAPI_OK ) {
          printf("VAPI_dealloc_mw failed for handle = " MT_ULONG_PTR_FMT "\n", (topt_p->mw_arr)[i].mw_hndl);
        }
        (topt_p->mw_arr)[i].mw_hndl = VAPI_INVAL_HNDL;
      }
      else {
        printf("memory window at index %d not deallocated\n", i);
      }
    }
    FREE(topt_p->mw_arr);
    topt_p->mw_arr = NULL;
  }

  p = (void *)((MT_virt_addr_t)(topt_p->data_mr.start));
  if ( topt_p->data_mr_hndl != VAPI_INVAL_HNDL ) {
    rc = VAPI_deregister_mr(topt_p->hca_hndl, topt_p->data_mr_hndl);
    if ( rc != VAPI_OK ) {
      printf("VAPI_deregister_mr failed: %s\n", VAPI_strerror(rc));
    }
    VFREE(p);
    topt_p->data_mr_hndl = VAPI_INVAL_HNDL;
  }


#ifdef debug_cleanup
  MTL_DEBUG5("*** 7 ***\n");
  fflush(stdout);
#endif

  if ( topt_p->qp_hndl != VAPI_INVAL_HNDL ) {
    rc = VAPI_destroy_qp(topt_p->hca_hndl, topt_p->qp_hndl);
    if ( rc != VAPI_OK ) {
      printf("VAPI_destroy_qp failed\n");
    }
    printf("Destroyed QP.....\n");
  }
  topt_p->qp_hndl = VAPI_INVAL_HNDL;

  if ( topt_p->srq_hndl != VAPI_INVAL_SRQ_HNDL) {
    rc = VAPI_destroy_srq(topt_p->hca_hndl, topt_p->srq_hndl);
    if ( rc != VAPI_OK ) {
      printf("VAPI_destroy_srq failed\n");
    }
  }

#ifdef debug_cleanup
  MTL_DEBUG5("*** 8 ***\n");
  fflush(stdout);
#endif

  if ( topt_p->sqcq_hndl != VAPI_INVAL_HNDL ) {
    rc = VAPI_destroy_cq(topt_p->hca_hndl, topt_p->sqcq_hndl);
    if ( rc != VAPI_OK ) {
      printf("VAPI_destroy_cq failed\n");
    }
  }
  topt_p->sqcq_hndl = VAPI_INVAL_HNDL;


#ifdef debug_cleanup
  MTL_DEBUG5("*** 9 ***\n");
  fflush(stdout);
#endif

  if ( topt_p->rqcq_hndl != VAPI_INVAL_HNDL ) {
    rc = VAPI_destroy_cq(topt_p->hca_hndl, topt_p->rqcq_hndl);
    if ( rc != VAPI_OK ) {
      printf("VAPI_destroy_cq failed\n");
    }
  }
  topt_p->rqcq_hndl = VAPI_INVAL_HNDL;


#ifdef debug_cleanup
  MTL_DEBUG5("*** 10 ***\n");
  fflush(stdout);
#endif

  if ( topt_p->pd_hndl != VAPI_INVAL_HNDL ) {
    rc = VAPI_dealloc_pd(topt_p->hca_hndl, topt_p->pd_hndl);
    if ( rc != VAPI_OK ) {
      printf("VAPI_dealloc_pd failed\n");
    }
  }
  topt_p->pd_hndl = VAPI_INVAL_HNDL;

  if ( errcode != 0 ) {
    printf("test recorded error code %d\n", errcode);
  }
  else {
    printf("\nended successfully\n");
  }
//  pthread_mutex_unlock(&topt_p->cln_mutex);

#ifdef debug_cleanup
  MTL_DEBUG5("*** 11 ***\n");
  fflush(stdout);
#endif
  if ( topt_p->hca_hndl != VAPI_INVAL_HNDL ) {
    rc = EVAPI_release_hca_hndl(topt_p->hca_hndl);
    if ( rc != VAPI_OK ) {
      printf("failed to release HCA hanlde\n");
    }
  }
}


/*
 *  clean_handler
 */
#ifdef __WIN__
BOOL WINAPI clean_handler(  DWORD dwCtrlType )
{
	int signum = (int)dwCtrlType;
#else
void clean_handler(int signum)
{
#endif
  static int called=0;
  int do_it;

  MOSAL_mutex_acq(&glob_mtx, TRUE);
  if ( called == 0 ) {
    called++;
    do_it = 1;
  }
  else {
    do_it = 0;
  }
  MOSAL_mutex_rel(&glob_mtx);
  if ( do_it ) {
    printf("cleaning and exiting\n");
    glob_opt->imd_ret = TRUE;
    rctp_cleanup(glob_opt, -1);
  }
  exit(EXIT_FAILURE);
}


/*
 *  create_threads
 */
static int create_threads(test_options_t *topt_p)
{
#if defined(USE_SEND_COMP_THREAD) || defined(USE_RCV_COMP_THREAD)
  int rc;
  pthread_t th;
#endif

#ifdef USE_SEND_COMP_THREAD
  rc = pthread_create(&th, NULL, send_comp_thread_func, topt_p);
  if ( rc != 0 ) {
    return -1;
  }
  topt_p->scomp_thread = th;
#endif


#ifdef USE_RCV_COMP_THREAD
  if ( topt_p->mode == MODE_CLIENT ) {
    rc = pthread_create(&th, NULL, rcv_comp_thread_func, topt_p);
    if ( rc != 0 ) {
      return -1;
    }
    topt_p->rcomp_thread = th;
  }
#endif

  return 0;
}



/*
 *  send_comp_thread_func
 */
#ifdef USE_SEND_COMP_THREAD
void *send_comp_thread_func(void *arg)
{
  test_options_t *topt_p = (test_options_t *)arg;
  VAPI_wc_desc_t comp_desc;
  VAPI_ret_t rc;

  while ( (topt_p->thread_run_cmd || get_snd_rcv_balance(topt_p) || get_cur_send_outstanding(topt_p)) && (topt_p->imd_ret==FALSE) ) {
    rc = VAPI_poll_cq(topt_p->hca_hndl, topt_p->sqcq_hndl, &comp_desc);
    if ( rc == VAPI_OK ) {
      if ( comp_desc.opcode == VAPI_CQE_SQ_SEND_DATA ) {
        /* SEND completion */
        dec_cur_send_outstanding(topt_p);
#ifdef DEBUG
        printf("received SQ completion\n");
#endif
      }
      else {
        printf("invalid opcode in send queue completion\n");
      }
    }
    else {
      sched_yield();
    }
  }
  pthread_exit(NULL);
}
#endif


/*
 *  rcv_comp_thread_func
 */
#ifdef USE_RCV_COMP_THREAD
void *rcv_comp_thread_func(void *arg)
{
  test_options_t *topt_p = (test_options_t *)arg;
  VAPI_wc_desc_t comp_desc;
  VAPI_ret_t rc;

  while ( (topt_p->thread_run_cmd || get_snd_rcv_balance(topt_p)) && (topt_p->imd_ret==FALSE) ) {
    rc = VAPI_poll_cq(topt_p->hca_hndl, topt_p->rqcq_hndl, &comp_desc);
    if ( rc == VAPI_OK ) {
      if ( comp_desc.opcode == VAPI_CQE_RQ_SEND_DATA ) {
        /* RECEIVE completion */
        MT_virt_addr_t bufptr = comp_desc.id % topt_p->max_oust_wr_rq;
        if ( memcmp((void *)((MT_virt_addr_t)(topt_p->rcv_mr.start+bufptr*topt_p->msgsz)), "END", 3) == 0 ) {
          printf("\nEND received from peer\n");
          fflush(stdout);
          topt_p->peer_ended = TRUE;
        }

        dec_cur_posted_rcv_bufs(topt_p);

        dec_snd_rcv_balance(topt_p);
#ifdef DEBUG
        printf("received RQ completion\n");
#endif
        if ( get_cur_posted_rcv_bufs(topt_p) < (topt_p->max_oust_wr_rq-5) ) {
          rc = post_recv_buf(topt_p);
          if ( rc != VAPI_OK ) {
            printf("post receive buf failed\n");
          }
          else {
            inc_cur_posted_rcv_bufs(topt_p);
          }
        }
      }
      else {
        printf("invalid opcode in RCV queue completion\n");
      }
    }
    else {
      sched_yield();
    }
  }
  pthread_exit(NULL);
}
#endif



/*
 *  get_ib_mtu
 */
static  VAPI_ret_t get_ib_mtu(u_int32_t mtu, IB_mtu_t *ib_mtu)
{
  switch (mtu) {
    case 256:
      *ib_mtu = MTU256;
      return VAPI_OK;
    case 512:
      *ib_mtu = MTU512;
      return VAPI_OK;
    case 1024:
      *ib_mtu = MTU1024;
      return VAPI_OK;
    case 2048:
      *ib_mtu = MTU2048;
      return VAPI_OK;
    case 4096:
      *ib_mtu = MTU4096;
      return VAPI_OK;
    default:
      return VAPI_ERR;
  }
}


/*
 *  process_command
 */
static VAPI_ret_t process_command(test_options_t *topt_p,
                                  VAPI_wc_desc_t *comp_desc_p, struct cmd_msg_st *cmd_p)
{
  struct cmd_msg_st *cmd_buf_p;
  VAPI_ret_t rc= VAPI_OK;

  cmd_buf_p = (struct cmd_msg_st *)(MT_ulong_ptr_t)comp_desc_p->id;

  cmd_be_to_host(cmd_buf_p);
  *cmd_p = *cmd_buf_p;

  switch ( cmd_p->command ) {
    case CMD_END:
      printf("\nEND received from peer\n");
      fflush(stdout);
      topt_p->peer_ended = TRUE;
      break;

    case CMD_WRITE_BLOCK:
    case CMD_READ_BLOCK:
      MTL_DEBUG3("Transaction %d: request for %s from block %d (%d bytes) to "
                 "va=" MT_ULONG_PTR_FMT " (rkey=0x%X).\n",cmd_p->tid,
                 (cmd_p->command == CMD_READ_BLOCK) ? "CMD_READ_BLK" : "CMD_WRITE_BLK",
                 cmd_p->block_num,cmd_p->size,(MT_ulong_ptr_t)(cmd_p->va),cmd_p->rkey);
      if ( topt_p->nordma == FALSE ) {
        rc = post_rdma_request(topt_p, cmd_p, FALSE);
	   //rc = VAPI_OK;
      }
      break;

  case CMD_ATOMIC_FA:
  case CMD_ATOMIC_CSWP:
      MTL_DEBUG2("Transaction %d: request for %s from block %d (%d bytes) to "
               "va=" MT_ULONG_PTR_FMT " (rkey=0x%X).\n",cmd_p->tid,
                 (cmd_p->command == CMD_ATOMIC_FA) ? "CMD_ATOMIC_FA" : "CMD_ATOMIC_CSWP",
                 cmd_p->block_num,cmd_p->size,(MT_ulong_ptr_t)(cmd_p->va),cmd_p->rkey);
      rc = perform_atomic_test(topt_p, cmd_p);
      break;


  case CMD_ACK:
#ifdef VERBOSE
      printf("CMD_ACK received by server - this is impossible\n");
#endif
      break;

    case CMD_NAK:
#ifdef VERBOSE
      printf("CMD_NAK received by server - this is impossible\n");
#endif
      break;

    default:
      printf("invalid command received\n");
      break;
  }

  return rc;
}


#if !defined(__WIN__) 
static char * format_time(char * buf) {
   time_t curr_time;
   struct tm my_tm;
   time(&curr_time);
   memcpy(&my_tm, localtime(&curr_time), sizeof(struct tm));
   sprintf(buf, "%02d.%02d.%d-%02d:%02d:%02d", my_tm.tm_mon+1,my_tm.tm_mday+1, my_tm.tm_year+1900,
         my_tm.tm_hour, my_tm.tm_min, my_tm.tm_sec);
   return buf;
}
#endif
/*
 *  drain_send_queue
 */
static void drain_send_queue(test_options_t *topt_p)
{
  VAPI_wc_desc_t comp_desc;
  trans_ctx_t *tctx_p;
  VAPI_ret_t rc;
  MT_bool comp_error=FALSE;
#if !defined(__WIN__) 
  char asc_time_buf[50];
#endif

  while ( (rc=VAPI_poll_cq(topt_p->hca_hndl, topt_p->sqcq_hndl, &comp_desc))==VAPI_OK ) {
    MTL_DEBUG3("%s: Got completion.  Opcode = %d, status=%d\n", __func__,
            comp_desc.opcode, comp_desc.status);
    topt_p->total_send_completions++;
	if ( (topt_p->total_send_completions%5000)==0 ) {
      printf("+");
      fflush(stdout);
    }
    tctx_p = (trans_ctx_t *)((MT_virt_addr_t)comp_desc.id);
    if ((comp_desc.opcode != VAPI_CQE_SQ_SEND_DATA) && 
        (comp_desc.opcode != VAPI_CQE_SQ_BIND_MRW) &&
        (tctx_p->sequence_id > highest_cq_seq_id)) {
         highest_cq_seq_id = tctx_p->sequence_id ;
         MTL_DEBUG2("%s: Completion Seq ID = %d\n", __func__, highest_cq_seq_id);
    }
    if ( comp_desc.status != VAPI_SUCCESS ) {
#if defined(__WIN__) 
      printf("got send Q completion with error:");
#else
      printf("%s : got send Q completion with error:", format_time(asc_time_buf) );
#endif      
      switch ( tctx_p->cqe_opcode ) {
        case VAPI_CQE_SQ_BIND_MRW:
          printf("bind failed"); // ??
          break;
        case VAPI_CQE_SQ_SEND_DATA:
          printf("send failed"); // ??
          break;
        case VAPI_CQE_SQ_RDMA_WRITE:
          printf("RDMA write failed"); // ??
          break;
        case VAPI_CQE_SQ_RDMA_READ:
          printf("RDMA read failed"); // ??
          break;
      case VAPI_CQE_SQ_FETCH_ADD:
          printf("Atomic FETCH/ADD failed"); // ??
          break;
      case VAPI_CQE_SQ_COMP_SWAP:
          printf("Atomic CMP/SWP failed"); // ??
          break;
        default:
          printf("unexpected completion opcode received in contect: opcode=0x%x", tctx_p->cqe_opcode);
      }
      printf(" - comp status = %s, wr_id=%p, syndrom=%s\n", VAPI_wc_status_sym(comp_desc.status), 
             (void *)(MT_virt_addr_t)comp_desc.id, VAPI_event_syndrome_sym(comp_desc.vendor_err_syndrome));
      comp_error=TRUE;
    }
    else {
#ifdef bdebug
      printf("on completion "); print_ctx(topt_p, tctx_p);
#endif
      if ( comp_desc.opcode == VAPI_CQE_SQ_BIND_MRW ) {
        if ( tctx_p->ctx.sq_bind_ctx.bind_size == 0 ) {
          topt_p->mw_arr[tctx_p->ctx.sq_bind_ctx.mw_idx].size = 0;
        }
#ifdef bdebug
        printf("got memwindow completion: index=%d, size=%d, ctx pointer= %p, rkey=%08x\n", tctx_p->ctx.sq_bind_ctx.mw_idx, tctx_p->ctx.sq_bind_ctx.bind_size, tctx_p, topt_p->mw_arr[tctx_p->ctx.sq_bind_ctx.mw_idx].rkey);
        fflush(stdout);
#endif
      }
  #ifdef VERBOSE
      printf("received SQ completion. tid="U64_FMT", snd_rcv_balance=%d, cur_send_outstanding=%d, cur_posted_rcv_bufs=%d\n",comp_desc.id, get_snd_rcv_balance(topt_p), get_cur_send_outstanding(topt_p), get_cur_posted_rcv_bufs(topt_p));
      printf("completion type = %d\n", comp_desc.opcode);
  #endif
    }
    dec_cur_send_outstanding(topt_p);
    if ( tctx_p->cqe_opcode == VAPI_CQE_SQ_BIND_MRW ) {
      topt_p->inflight_mw_binds--;
    }
    free(tctx_p);
  }
  if ( rc != VAPI_CQ_EMPTY ) {
    printf("VAPI_poll_cq failed: %s\n", VAPI_strerror(rc)); fflush(stdout);
  }
  if ( comp_error ) {
    drain_rcv_queue(topt_p);
    rctp_cleanup(topt_p, EXIT_CODE);
    exit(EXIT_CODE);
  }
}


/*
 *  drain_rcv_queue
 */
static void drain_rcv_queue(test_options_t *topt_p)
{
  VAPI_ret_t rc;
  VAPI_wc_desc_t comp_desc;
  struct cmd_msg_st rcv_cmd;
  VAPI_mw_bind_t bind_prop;
  trans_ctx_t *tctx_p;
  MT_bool comp_error=FALSE;
#if !defined(__WIN__) 
  char asc_time_buf[50];
#endif

  rc = VAPI_poll_cq(topt_p->hca_hndl, topt_p->rqcq_hndl, &comp_desc);
  if ( rc == VAPI_OK ) {
    if ( comp_desc.status != VAPI_SUCCESS ) {
#if defined(__WIN__) 
      printf("unsuccessful completion on receive queue: completion status=%s\n", VAPI_wc_status_sym(comp_desc.status));
#else
      printf("%s : unsuccessful completion on receive queue: completion status=%s\n", 
             format_time(asc_time_buf), VAPI_wc_status_sym(comp_desc.status));
#endif      
      printf("Press <enter> to exit...\n");
      fflush(stdout);
      comp_error = TRUE;
      getchar();
    }
    else {
#ifdef im_debug
      char *comp_str;
      switch ( comp_desc.opcode ) {
        case VAPI_CQE_RQ_SEND_DATA:
          comp_str = "VAPI_CQE_RQ_SEND_DATA";
          break;
        case VAPI_CQE_RQ_RDMA_WITH_IMM:
          comp_str = "VAPI_CQE_RQ_RDMA_WITH_IMM";
          break;
        default:
          comp_str = "unexpected";
      }
      printf("got completion on receive queue: %s\n", comp_str);
#endif

      if ( (comp_desc.opcode==VAPI_CQE_RQ_SEND_DATA) || (comp_desc.opcode==VAPI_CQE_RQ_RDMA_WITH_IMM) ) {
        dec_snd_rcv_balance(topt_p);
      }

      if ( comp_desc.imm_data_valid ) {
  #ifdef VERBOSE
        fprintf(stderr, "immediate data received %08x\n", comp_desc.imm_data);
  #endif
      }
      parse_rcv_msg(&comp_desc, &rcv_cmd);
      if ( (comp_desc.opcode==VAPI_CQE_RQ_SEND_DATA) && (topt_p->mode==MODE_CLIENT)) {
         if (topt_p->memwindows) {
            /* revoke memory region by rebinding to 0 length */
            bind_prop.mr_lkey = topt_p->data_mr.l_key;
            bind_prop.start = 0; /* irrelevant */
            bind_prop.size = 0;
            bind_prop.acl = VAPI_EN_REMOTE_READ; /* irrelevant */
    
            tctx_p = malloc(sizeof(*tctx_p));
            if ( !tctx_p ) {
              return;
            }
            tctx_p->cqe_opcode = VAPI_CQE_SQ_BIND_MRW;
            tctx_p->ctx.sq_bind_ctx.mw_idx = rcv_cmd.wnd_idx;
            tctx_p->ctx.sq_bind_ctx.bind_size = (u_int32_t)bind_prop.size;
    
            while ( (get_cur_send_outstanding(topt_p)>=((int)topt_p->max_oust_wr_sq-(int)SND_Q_SAFE))  ) {
              drain_send_queue(topt_p);
            }
    
    #ifdef bdebug
            printf("after post "); print_ctx(topt_p, tctx_p);
    #endif
            rc = VAPI_bind_mw(topt_p->hca_hndl, topt_p->mw_arr[rcv_cmd.wnd_idx].mw_hndl, &bind_prop, topt_p->qp_hndl,
                              (VAPI_wr_id_t)((MT_virt_addr_t)tctx_p), VAPI_SIGNALED, &rcv_cmd.rkey);
            if ( rc != VAPI_OK ) {
              printf("Error binding memory window\n");
            }
            else {
    #ifdef bdebug
              printf("zero bind: for idx=%d, rkey=%08x\n", rcv_cmd.wnd_idx, rcv_cmd.rkey);
    #endif
              inc_cur_send_outstanding(topt_p);
              topt_p->inflight_mw_binds++;
            }
         }
      }
      if ( rcv_cmd.command == CMD_NAK ) {
        MTL_DEBUG3("Transaction %d: CMD_NAK received from server\n",rcv_cmd.tid);
      }
      else {
        MTL_DEBUG3("Transaction %d: CMD_ACK received from server\n",rcv_cmd.tid);
      }
      /* Return WQE to RQ */
      rc = post_recv_buf(topt_p);
      if ( rc != VAPI_OK ) {
        printf("post_recv_buf failed\n");
        fflush(stdout);
      }
    }
  }
  else if ( rc != VAPI_CQ_EMPTY ) {
    printf("VAPI_poll_cq error on polling receive queue: %s\n", VAPI_strerror(rc));
    fflush(stdout);
  }
  if ( comp_error ) {
    drain_send_queue(topt_p);
    rctp_cleanup(topt_p, EXIT_CODE);
    exit(EXIT_CODE);
  }
}


/*
 *  post_rdma_request
 */
static VAPI_ret_t post_rdma_request(test_options_t *topt_p, struct cmd_msg_st *cmd_p, MT_bool use_id)
{
  VAPI_ret_t rc;
  VAPI_sr_desc_t sr_desc;
  VAPI_wr_opcode_t opcode;
  VAPI_sg_lst_entry_t sg_lst;
  trans_ctx_t *tctx_p;

  switch ( cmd_p->command ) {
    case CMD_READ_BLOCK:
      opcode = topt_p->immediate==FALSE ? VAPI_RDMA_WRITE : VAPI_RDMA_WRITE_WITH_IMM;
      break;
    case CMD_WRITE_BLOCK:
      opcode = VAPI_RDMA_READ;
      break;
    default:
      return VAPI_ERR;
  }

  sg_lst.addr = topt_p->data_mr.start + (cmd_p->block_num*MAX_BLK_SZ)%DISK_SZ;
  sg_lst.len = (cmd_p->size > MAX_BLK_SZ) ? MAX_BLK_SZ : cmd_p->size;
  sg_lst.lkey = topt_p->data_mr.l_key;

  topt_p->last_posted_send_id++;
//  sr_desc.id = topt_p->last_posted_send_id;
  sr_desc.opcode = opcode;
  sr_desc.comp_type = VAPI_SIGNALED;
  sr_desc.sg_lst_p = &sg_lst;
  sr_desc.sg_lst_len = 1;
  if ( sg_lst.len == 0 ) {
    sr_desc.sg_lst_len = 0;
  }
  sr_desc.imm_data = 0x3333cccc;
  sr_desc.fence = TRUE;
  sr_desc.set_se = FALSE;
  sr_desc.remote_addr = cmd_p->va;
  sr_desc.r_key = cmd_p->rkey;


  tctx_p = malloc(sizeof(*tctx_p));
  if ( !tctx_p ) {
    return VAPI_EAGAIN;
  }
  if (use_id == TRUE) {
      tctx_p->sequence_id = ++sequence_id;
  } else {
      tctx_p->sequence_id = 0;
  }

  tctx_p->cqe_opcode = sr_desc.opcode==VAPI_RDMA_WRITE ? VAPI_CQE_SQ_RDMA_WRITE : VAPI_CQE_SQ_RDMA_READ;
  tctx_p->ctx.rdma_ctx.remote_va = sr_desc.remote_addr;
  tctx_p->ctx.rdma_ctx.rkey = sr_desc.r_key;
#ifdef im_debug
  print_post_opcode(sr_desc.opcode, tctx_p);
#endif
  sr_desc.id = (VAPI_wr_id_t)((MT_virt_addr_t)tctx_p);
  while ((rc = VAPI_post_sr(topt_p->hca_hndl, topt_p->qp_hndl, &sr_desc)) != VAPI_OK ) {
    if (rc != VAPI_EAGAIN)  return rc;
    drain_send_queue(topt_p);
  }
#ifdef im_debug
  printf("success\n");
  fflush(stdout);
#endif
  inc_cur_send_outstanding(topt_p);

  return rc;
}

static VAPI_ret_t post_rdma_request_and_wait(test_options_t *topt_p, struct cmd_msg_st *cmd_p) 
{        
    VAPI_ret_t   rc;
    u_int32_t    local_seq_id;
    FUNC_IN;
    rc = post_rdma_request(topt_p,cmd_p, TRUE);
    local_seq_id = sequence_id;
    if (rc != VAPI_OK)
        return rc;

//    MTL_DEBUG4("%s: before drain send queue. seq_id  = %d\n", __func__, local_seq_id);
    while ( local_seq_id > highest_cq_seq_id )
        drain_send_queue(topt_p);
//    MTL_DEBUG4("%s: AFTER drain send queue\n", __func__);
    return VAPI_OK;
}
#ifndef SKIP_ATOMIC
/*
 *  post_atomic_request
 */
static VAPI_ret_t post_atomic_request(test_options_t *topt_p, struct cmd_msg_st *cmd_p,
                                      u_int64_t  fetch_cmp_oper, u_int64_t swap_oper)
{
  VAPI_ret_t rc;
  VAPI_sr_desc_t sr_desc;
  VAPI_wr_opcode_t opcode;
  VAPI_sg_lst_entry_t sg_lst;
  trans_ctx_t *tctx_p;

  memset((void *)&sr_desc, 0, sizeof(VAPI_sr_desc_t));
  switch ( cmd_p->command  ) {
    case CMD_ATOMIC_FA:
      opcode = VAPI_ATOMIC_FETCH_AND_ADD;
	  sr_desc.compare_add = fetch_cmp_oper;
      break;
    case CMD_ATOMIC_CSWP:
      opcode = VAPI_ATOMIC_CMP_AND_SWP;
	  sr_desc.compare_add = fetch_cmp_oper;
	  sr_desc.swap = swap_oper;
      break;
    default:
      return VAPI_ERR;
  }

  sg_lst.addr = topt_p->data_mr.start + (cmd_p->block_num*MAX_BLK_SZ)%DISK_SZ;
  sg_lst.len = ATOMIC_OP_DATA_SIZE;
  sg_lst.lkey = topt_p->data_mr.l_key;
  
  topt_p->last_posted_send_id++;
//  sr_desc.id = topt_p->last_posted_send_id;
  sr_desc.opcode = opcode;
  sr_desc.comp_type = VAPI_SIGNALED;
  sr_desc.sg_lst_p = &sg_lst;
  sr_desc.sg_lst_len = 1;
  sr_desc.imm_data = 0x3333cccc;
  sr_desc.fence = TRUE;
  sr_desc.set_se = TRUE;
  sr_desc.remote_addr = cmd_p->va;
  sr_desc.r_key = cmd_p->rkey;

  tctx_p = malloc(sizeof(*tctx_p));
  if ( !tctx_p ) {
    return VAPI_EAGAIN;
  }
  tctx_p->cqe_opcode = sr_desc.opcode==VAPI_ATOMIC_FETCH_AND_ADD ? VAPI_CQE_SQ_FETCH_ADD : VAPI_CQE_SQ_COMP_SWAP;
  tctx_p->ctx.rdma_ctx.remote_va = sr_desc.remote_addr;
  tctx_p->ctx.rdma_ctx.rkey = sr_desc.r_key;
  tctx_p->sequence_id = ++sequence_id;

  sr_desc.id = (VAPI_wr_id_t)((MT_virt_addr_t)tctx_p);
  while ((rc = VAPI_post_sr(topt_p->hca_hndl, topt_p->qp_hndl, &sr_desc)) != VAPI_OK ) {
    if (rc != VAPI_EAGAIN)  return rc;
    drain_send_queue(topt_p);
  }
  inc_cur_send_outstanding(topt_p);

  return rc;
}

static VAPI_ret_t post_atomic_request_and_wait(test_options_t *topt_p, struct cmd_msg_st *cmd_p,
                                      u_int64_t  fetch_cmp_oper, u_int64_t swap_oper)
{
    VAPI_ret_t   rc;
    u_int32_t    local_seq_id;
    FUNC_IN;
    rc = post_atomic_request(topt_p,cmd_p,fetch_cmp_oper, swap_oper );
    local_seq_id = sequence_id;
    if (rc != VAPI_OK)
        MT_RETURN(rc);

    MTL_DEBUG4("%s: before drain send queue. seq_id  = %d\n", __func__, local_seq_id);
    while ( local_seq_id >  highest_cq_seq_id )
        drain_send_queue(topt_p);
    MTL_DEBUG4("%s: AFTER drain send queue.\n", __func__);

    MT_RETURN(VAPI_OK);
}

#endif

static VAPI_ret_t perform_atomic_test(test_options_t *topt_p, struct cmd_msg_st *cmd_p)
{
    VAPI_virt_addr_t  test_address;
    u_int32_t     atomic_cmd = cmd_p->command;
    u_int64_t     fetch_cmp_oper, swap_oper=0;
    u_int64_t     initial_val, returned_val;
    MT_bool       swap_fail = (rand() > (RAND_MAX/2)) ? TRUE : FALSE ;
    struct cmd_msg_st tmp_cmd = *cmd_p;

    test_address = topt_p->data_mr.start + (cmd_p->block_num*MAX_BLK_SZ)%DISK_SZ;

    MTL_DEBUG4("%s: local buffer address  = "U64_FMT"\n", __func__, test_address);

    switch ( atomic_cmd ) {
      case CMD_ATOMIC_FA:
        fetch_cmp_oper = (((u_int64_t)rand()) * ((u_int64_t)rand()))>>1;
        initial_val = (((u_int64_t)rand()) * ((u_int64_t)rand()))>>1;
		MTL_DEBUG4("%s: initial_val = "U64_FMT", f_cmp = "U64_FMT"\n", __func__, initial_val, fetch_cmp_oper );
        break;
      case CMD_ATOMIC_CSWP:
        fetch_cmp_oper = (((u_int64_t)rand()) * ((u_int64_t)rand()));
        initial_val = fetch_cmp_oper;
        if (swap_fail == TRUE)
             initial_val -= 1;
        swap_oper = (((u_int64_t)rand()) * ((u_int64_t)rand()));
		MTL_DEBUG4("%s: swp_fail=%d, initial_val = "U64_FMT", f_cmp = "U64_FMT", swp="U64_FMT"\n", __func__,
		swap_fail, initial_val, fetch_cmp_oper, swap_oper );
        break;
      default:
        return VAPI_ERR;
    }

    /* first, write initial value to client */
//    *(u_int64_t *)( (MT_virt_addr_t )test_address) = MOSAL_cpu_to_be64(initial_val);
    *(u_int64_t *)( (MT_virt_addr_t )test_address) = topt_p->swap_endianness ? rctp_endian64_swap(initial_val): initial_val;
    tmp_cmd.command = CMD_READ_BLOCK;
    post_rdma_request_and_wait(topt_p, &tmp_cmd);

#ifndef SKIP_ATOMIC

    /* zero out local buffer */
    *(u_int64_t *)( (MT_virt_addr_t )test_address) = (u_int64_t) 0;
    /* Now, do the test */
    post_atomic_request_and_wait(topt_p, cmd_p, fetch_cmp_oper, swap_oper);

    /* Intermediate check on values returned by test */
//    returned_val = MOSAL_be64_to_cpu(*(u_int64_t *)( (MT_virt_addr_t )test_address));
    returned_val = *(u_int64_t *)( (MT_virt_addr_t )test_address);
    if (returned_val != initial_val) {
            /*ERROR */
        MTL_ERROR1("%s: Seq # %d: Initial and returned values don't match(init="U64_FMT", ret="U64_FMT")\n", __func__,
                   sequence_id, initial_val, returned_val);
        return VAPI_ERR;
    }
#else
  atomic_cmd = CMD_ATOMIC_CSWP; 
  swap_fail  = TRUE;
#endif

    /* zero out local buffer */
    *(u_int64_t *)( (MT_virt_addr_t )test_address) = (u_int64_t) 0;
    
    /* Finally, write the new value to the target */
    tmp_cmd.command = CMD_WRITE_BLOCK;
    post_rdma_request_and_wait(topt_p, &tmp_cmd);

    /*final test */
//    returned_val = MOSAL_be64_to_cpu(*(u_int64_t *)( (MT_virt_addr_t )test_address));
    returned_val = *(u_int64_t *)( (MT_virt_addr_t )test_address);
    if (topt_p->swap_endianness) {
        returned_val = rctp_endian64_swap(returned_val);
    }
    switch ( atomic_cmd ) {
      case CMD_ATOMIC_FA:
        if (returned_val != (initial_val + fetch_cmp_oper)) {
            MTL_ERROR1("%s: Seq # %d, FA: remote and local result values don't match(rem="U64_FMT", loc="U64_FMT", init="U64_FMT", Addv="U64_FMT")\n", __func__,
                       sequence_id, returned_val, initial_val + fetch_cmp_oper, initial_val, fetch_cmp_oper);
            MTL_ERROR1("%s: loc addr = "U64_FMT", rem_addr(l)="U64_FMT", rem_addr(r)="U64_FMT", blk(l) = %d, blk(r)=%d\n", __func__,
                       test_address, tmp_cmd.va, cmd_p->va, tmp_cmd.block_num, cmd_p->block_num);
            return VAPI_ERR;
        }
        break;
      case CMD_ATOMIC_CSWP:
        if (swap_fail == TRUE){
            if (returned_val != initial_val) {
                MTL_ERROR1("%s: Seq # %d, CSWP/FAIL: Initial and returned values don't match(init="U64_FMT", ret="U64_FMT", swap="U64_FMT", cmp="U64_FMT")\n", __func__,
                           sequence_id, initial_val, returned_val, swap_oper, fetch_cmp_oper);
                MTL_ERROR1("%s: loc addr = "U64_FMT", rem_addr(l)="U64_FMT", rem_addr(r)="U64_FMT", blk(l) = %d, blk(r)=%d\n", __func__,
                           test_address, tmp_cmd.va, cmd_p->va, tmp_cmd.block_num, cmd_p->block_num);
                return VAPI_ERR;
            }
        } else {
            if (returned_val != swap_oper) {
                MTL_ERROR1("%s: Seq # %d, CSWP/SUCC: swap and returned values don't match(swap="U64_FMT", ret="U64_FMT", init="U64_FMT", cmp="U64_FMT")\n", __func__,
                           sequence_id, swap_oper, returned_val, initial_val, fetch_cmp_oper);
                MTL_ERROR1("%s: loc addr = "U64_FMT", rem_addr(l)="U64_FMT", rem_addr(r)="U64_FMT", blk(l) = %d, blk(r)=%d\n", __func__,
                           test_address, tmp_cmd.va, cmd_p->va, tmp_cmd.block_num, cmd_p->block_num);
                return VAPI_ERR;
             }
        }
        break;
      default:
        return VAPI_ERR;
    }
    return VAPI_OK;
}


static void parse_rcv_msg(VAPI_wc_desc_t *comp_desc_p, struct cmd_msg_st *cmd_p)
{
  struct cmd_msg_st *cmd_buf_p;
  cmd_buf_p = (struct cmd_msg_st *)(MT_ulong_ptr_t)(comp_desc_p->id);

  cmd_be_to_host(cmd_buf_p);
  *cmd_p = *cmd_buf_p;
}



/*
 *  mem_debugger
 */
static void mem_debugger(void)
{
  char *line, cmd;
  u_int8_t *addr;
  int argc, count;

  while ( 1 ) {
    line = readline("Enter mem command: ");
    argc = sscanf(line, "%c %p %d", &cmd, &addr, &count);
    free(line);
    switch ( argc ) {
      case 1:
        if ( cmd == 'q' ) {
          return;
        }

      case 2:
        count = 4;
      case 3:
        switch ( cmd ) {
          case 'd':
            dump_mem(addr, count);
            break;

          case 'q':
            return;

          default:
            printf("unsupported command\n");
        }
    }
  }
}

/*
 *  print_peer_info
 */
static void print_peer_info(test_options_t *topt_p)
{
  printf("=== %s side ===\n", topt_p->mode==MODE_CLIENT ? "Client" : "Server");
  printf("QP number = 0x%06x\n", topt_p->qp_prop.qp_num);
  printf("%s\n", topt_p->swap_endianness? "Changing endianness for atomic tests": "Endianness is the same on client and server");
}


/*
 *  dump_mem
 */
static void dump_mem(void *addr, int count)
{
  u_int32_t *p32 = addr;
  int i;

  for ( i=0; i<=(count>>2); ++i, ++p32 ) {
    if ( (i&3)==0 ) printf("\n");
    printf("%08x ", *p32);
  }
  printf("\n");
}


static VAPI_ret_t create_memory_windows(test_options_t *topt_p)
{
  VAPI_ret_t rc;
  m_windows_ctx_t *mw_arr;
  u_int32_t i;

  /* create array to hold memory windows contexts */
  mw_arr = malloc(sizeof(m_windows_ctx_t)*topt_p->max_oust_wr_sq);
  if ( !mw_arr ) {
    return VAPI_EAGAIN;
  }
  memset(mw_arr, 0, sizeof(m_windows_ctx_t)*topt_p->max_oust_wr_sq);
  for ( i=0; i<topt_p->max_oust_wr_sq; ++i ) {
    mw_arr[i].mw_hndl = VAPI_INVAL_HNDL;
    mw_arr[i].size = 0;
  }
  topt_p->mw_arr = mw_arr;

  for ( i=0; i<topt_p->max_oust_wr_sq; ++i ) {
    rc = VAPI_alloc_mw(topt_p->hca_hndl, topt_p->pd_hndl, &(topt_p->mw_arr)[i].mw_hndl, &(topt_p->mw_arr)[i].rkey);
    if ( rc != VAPI_OK ) {
      (topt_p->mw_arr)[i].mw_hndl = VAPI_INVAL_HNDL;
      return rc;
    }
  }
  return VAPI_OK;
}


/*
 *  cmd_be_to_host
 */
static void cmd_be_to_host(struct cmd_msg_st *cmd_p)
{
  struct cmd_msg_st orig = *cmd_p;

  cmd_p->command = ntohl(orig.command);
  cmd_p->tid = ntohl(orig.tid);
  cmd_p->va = MOSAL_be64_to_cpu(orig.va);
  cmd_p->rkey = ntohl(orig.rkey);
  cmd_p->size = ntohl(orig.size);
  cmd_p->block_num = ntohl(orig.block_num);
  cmd_p->wnd_idx = ntohl(orig.wnd_idx);
}


/*
 *  cmd_host_to_be
 */
static void cmd_host_to_be(struct cmd_msg_st *cmd_p)
{
  struct cmd_msg_st orig = *cmd_p;

  cmd_p->block_num = htonl(orig.block_num);
  cmd_p->command = htonl(orig.command);
  cmd_p->rkey = htonl(orig.rkey);
  cmd_p->size = htonl(orig.size);
  cmd_p->tid = htonl(orig.tid);
  cmd_p->va = MOSAL_cpu_to_be64(orig.va);
  cmd_p->wnd_idx = htonl(orig.wnd_idx);
}

/*%%%%%%%%%%%%% print functions %%%%%%%%%%%%%%*/

/*
 *  print_qp_props
 */
void print_qp_props(VAPI_qp_prop_t *qp_props)
{
  printf("=== QP properties ===\n");
  printf("\tqp_num = 0x%06x\n", qp_props->qp_num);
  printf("\tmax_oust_wr_sq = %d\n", qp_props->cap.max_oust_wr_sq);
  printf("\tmax_oust_wr_rq = %d\n", qp_props->cap.max_oust_wr_rq);
  printf("\tmax_sg_size_sq = %d\n", qp_props->cap.max_sg_size_sq);
  printf("\tmax_sg_size_rq = %d\n", qp_props->cap.max_sg_size_rq);
}


/*
 *
 */
void print_hca_cap(VAPI_hca_vendor_t *hca_vendor_p, VAPI_hca_cap_t *hca_cap_p)
{
  printf("=== HCA vendor info ===\n");
  printf("\tvendor_id = 0x%08x\n", hca_vendor_p->vendor_id);
  printf("\tvendor_part_id = %d\n", hca_vendor_p->vendor_part_id);
  printf("\thw_ver = 0x%08x\n", hca_vendor_p->hw_ver);

  printf("=== HCA capabilities ===\n");


  printf("\tmax_num_qp = %d\n", hca_cap_p->max_num_qp);
  printf("\tmax_qp_ous_wr = %d\n", hca_cap_p->max_qp_ous_wr);
  printf("\tflags = 0x%08x\n", hca_cap_p->flags);
  printf("\tmax_num_sg_ent = %d\n", hca_cap_p->max_num_sg_ent);
  printf("\tmax_num_sg_ent_rd = %d\n", hca_cap_p->max_num_sg_ent_rd);
  printf("\tmax_num_cq = %d\n", hca_cap_p->max_num_cq);
  printf("\tmax_num_ent_cq = %d\n", hca_cap_p->max_num_ent_cq);
  printf("\tmax_num_mr = %d\n", hca_cap_p->max_num_mr);
  printf("\tmax_mr_size = "U64_FMT"\n", hca_cap_p->max_mr_size);
  printf("\tmax_pd_num = %d\n", hca_cap_p->max_pd_num);
  printf("\tpage_size_cap = %d\n", hca_cap_p->page_size_cap);
  printf("\tphys_port_num = %d\n", hca_cap_p->phys_port_num);
  printf("\tmax_pkeys = %d\n", hca_cap_p->max_pkeys);
  printf("\tnode_guid = %02x:%02x:%02x:%02x:%02x:%02x:%02x:%02x\n", hca_cap_p->node_guid[0],
         hca_cap_p->node_guid[1],  
         hca_cap_p->node_guid[2],  
         hca_cap_p->node_guid[3],  
         hca_cap_p->node_guid[4],  
         hca_cap_p->node_guid[5],  
         hca_cap_p->node_guid[6],  
         hca_cap_p->node_guid[7]);
  printf("\tlocal_ca_ack_delay (Log2 4.096usec Max. RX to ACK or NAK delay) = %d\n", hca_cap_p->local_ca_ack_delay);
  printf("\tmax_qp_ous_rd_atom = %d\n", hca_cap_p->max_qp_ous_rd_atom);
  printf("\tmax_ee_ous_rd_atom = %d\n", hca_cap_p->max_ee_ous_rd_atom);
  printf("\tmax_res_rd_atom = %d\n", hca_cap_p->max_res_rd_atom);
  printf("\tmax_qp_init_rd_atom = %d\n", hca_cap_p->max_qp_init_rd_atom);
  printf("\tmax_ee_init_rd_atom = %d\n", hca_cap_p->max_ee_init_rd_atom);
  {
    char s[50];
    switch ( hca_cap_p->atomic_cap) {
      case VAPI_ATOMIC_CAP_NONE:
        strcpy(s, "VAPI_ATOMIC_CAP_NONE");
        break;
      case VAPI_ATOMIC_CAP_HCA:
        strcpy(s, "VAPI_ATOMIC_CAP_HCA");
        break;
      case VAPI_ATOMIC_CAP_GLOB:
        strcpy(s, "VAPI_ATOMIC_CAP_GLOB");
        break;
      default:
        strcpy(s, "invalid capability");
    }
    printf("\tatomic_cap = %s\n", s);
  }
  printf("\tmax_ee_num = %d\n", hca_cap_p->max_ee_num);
  printf("\tmax_rdd_num = %d\n", hca_cap_p->max_rdd_num);
  printf("\tmax_mw_num = %d\n", hca_cap_p->max_mw_num);
  printf("\tmax_raw_ipv6_qp = %d\n", hca_cap_p->max_raw_ipv6_qp);
  printf("\tmax_raw_ethy_qp = %d\n", hca_cap_p->max_raw_ethy_qp);
  printf("\tmax_mcast_grp_num = %d\n", hca_cap_p->max_mcast_grp_num);
  printf("\tmax_mcast_qp_attach_num = %d\n", hca_cap_p->max_mcast_qp_attach_num);
  printf("\tmax_total_mcast_qp_attach_num = %d\n", hca_cap_p->max_total_mcast_qp_attach_num);
  printf("\tmax_ah_num = %d\n", hca_cap_p->max_ah_num);

}

#ifdef bdebug
static void print_ctx(test_options_t *topt_p, trans_ctx_t *tctx_p)
{
  printf("aaa:opcode=");
  switch ( tctx_p->cqe_opcode ) {
    case VAPI_CQE_RQ_SEND_DATA:
      printf("VAPI_CQE_SQ_SEND_DATA");
      break;

    case VAPI_CQE_SQ_SEND_DATA:
      printf("VAPI_CQE_SQ_SEND_DATA");
      break;

    case VAPI_CQE_SQ_RDMA_WRITE:
      printf("VAPI_CQE_SQ_RDMA_WRITE");
      break;

    case VAPI_CQE_SQ_RDMA_READ:
      printf("VAPI_CQE_SQ_RDMA_READ");
      break;

    case VAPI_CQE_SQ_COMP_SWAP:
      printf("VAPI_CQE_SQ_COMP_SWAP");
      break;

    case VAPI_CQE_SQ_FETCH_ADD:
      printf("VAPI_CQE_SQ_FETCH_ADD");
      break;

    case VAPI_CQE_SQ_BIND_MRW:
      printf("VAPI_CQE_SQ_BIND_MRW");
      printf(", bind_size=%d, mw_idx=%d, rkey=%08x", tctx_p->ctx.sq_bind_ctx.bind_size, tctx_p->ctx.sq_bind_ctx.mw_idx, topt_p->mw_arr[tctx_p->ctx.sq_bind_ctx.mw_idx].rkey);
      break;

    case VAPI_CQE_RQ_RDMA_WITH_IMM:
      printf("VAPI_CQE_RQ_RDMA_WITH_IMM");
      break;
      
    default:
      printf("Error: invalid opcode");
  }
  printf("\n");
}
#endif



#ifdef im_debug
void print_post_opcode(VAPI_wr_opcode_t opcode, void *context)
{
  char *poststr;

  switch ( opcode ){
    case VAPI_RDMA_WRITE:
      poststr = "VAPI_RDMA_WRITE";
      break;
    case VAPI_RDMA_WRITE_WITH_IMM:
      poststr = "VAPI_RDMA_WRITE_WITH_IMM";
      break;
    case VAPI_SEND:
      poststr = "VAPI_SEND";
      break;
    case VAPI_SEND_WITH_IMM:
      poststr = "VAPI_SEND_WITH_IMM";
      break;
    case VAPI_RDMA_READ:
      poststr = "VAPI_RDMA_READ";
      break;
    case VAPI_ATOMIC_CMP_AND_SWP:
      poststr = "VAPI_ATOMIC_CMP_AND_SWP";
      break;
    case VAPI_ATOMIC_FETCH_AND_ADD:
      poststr = "VAPI_ATOMIC_FETCH_AND_ADD";
      break;
    case VAPI_RECEIVE:
      poststr = "VAPI_RECEIVE";
      break;
    default:
      poststr = "Unknown opcode";
  }
  printf("posting %s, context=%p ... ", poststr, context);
  fflush(stdout);
}
#endif


#ifdef im_debug
void print_completion_opcode(VAPI_cqe_opcode_t opcode)
{
  char *str;

  switch ( opcode ) {
    case VAPI_CQE_SQ_SEND_DATA:
      str = "VAPI_CQE_SQ_SEND_DATA";
      break;
    case VAPI_CQE_SQ_RDMA_WRITE:
      str = "VAPI_CQE_SQ_RDMA_WRITE";
      break;
    case VAPI_CQE_SQ_RDMA_READ:
      str = "VAPI_CQE_SQ_RDMA_READ";
      break;
    case VAPI_CQE_SQ_COMP_SWAP:
      str = "VAPI_CQE_SQ_COMP_SWAP";
      break;
    case VAPI_CQE_SQ_FETCH_ADD:
      str = "VAPI_CQE_SQ_FETCH_ADD";
      break;
    case VAPI_CQE_SQ_BIND_MRW:
      str = "VAPI_CQE_SQ_BIND_MRW";
      break;
    case VAPI_CQE_RQ_SEND_DATA:
      str = "VAPI_CQE_RQ_SEND_DATA";
      break;
    case VAPI_CQE_RQ_RDMA_WITH_IMM:
      str = "VAPI_CQE_RQ_RDMA_WITH_IMM";
      break;
    default:
      str = "Unknown";
  }
  printf("received completion: %s\n", str);
  fflush(stdout);
}
#endif
