/*
  This software is available to you under a choice of one of two
  licenses.  You may choose to be licensed under the terms of the GNU
  General Public License (GPL) Version 2, available at
  <http://www.fsf.org/copyleft/gpl.html>, or the OpenIB.org BSD
  license, available in the LICENSE.TXT file accompanying this
  software.  These details are also available at
  <http://openib.org/license.html>.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
  BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
  ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
  CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
  SOFTWARE.

  Copyright (c) 2004 Mellanox Technologies Ltd.  All rights reserved.
*/
 
#include <mtl_types.h>
#include <ib_defs.h>
 
#ifndef H_RCTP_H
#define H_RCTP_H

#define MAX_BLK_SZ (1<<17)
#define MAX_DISK_BLKS 16 /* 128 */
#define DISK_SZ  (MAX_BLK_SZ*MAX_DISK_BLKS)


typedef enum {
  MODE_NA = 0,
  MODE_SERVER,
  MODE_CLIENT
}
opmode_t;

typedef struct {
	int		dummy;
}
sq_send_ctx_t;


typedef struct {
	int		dummy;
}
rq_send_ctx_t;


typedef struct {
  VAPI_rkey_t rkey;
  VAPI_virt_addr_t remote_va;
}
rdma_ctx_t;


typedef struct {
  u_int32_t mw_idx;
  u_int32_t bind_size;
}
sq_bind_ctx_t;


typedef struct {
  /* this is the opcode of the completion in case of successfull completion
     It is needed in case of an unsuccessful completion to get details of the 
     operation that failed */
  VAPI_cqe_opcode_t cqe_opcode;
  u_int32_t  sequence_id;
  union {
    sq_send_ctx_t sq_send_ctx;
    sq_send_ctx_t rq_send_ctx;
    sq_bind_ctx_t sq_bind_ctx;
    rdma_ctx_t rdma_ctx;
  }
  ctx;
}
trans_ctx_t;

typedef struct {
  VAPI_mw_hndl_t mw_hndl;
  VAPI_rkey_t rkey;
  u_int32_t size;
}
m_windows_ctx_t;



typedef struct {
  u_int32_t num_trans; /* specifies the number of transactions to do */
  VAPI_hca_id_t hca_id; /* id of the hca used by the test */
  VAPI_hca_hndl_t hca_hndl; /* handle to the hca we use */
  VAPI_hca_vendor_t hca_vendor; /* ?? */
  VAPI_hca_cap_t hca_cap; /* HCA capabilities */
  VAPI_cqe_num_t req_cqe_num; /* required number of entries in a CQ */
  VAPI_cqe_num_t sqcq_cqe_num; /* number of entries in the CQ associated with the send Q */
  VAPI_cqe_num_t rqcq_cqe_num; /* number of entries in the CQ associated with the receive Q */
  VAPI_cq_hndl_t sqcq_hndl; /* handle to cq associated with send Q */
  VAPI_cq_hndl_t rqcq_hndl; /* handle to cq associated with recwive Q */
  VAPI_srq_hndl_t srq_hndl; /* (optional) SRQ handle */
  VAPI_qp_hndl_t qp_hndl; /* handle to the QP used */
  VAPI_qp_prop_t qp_prop; /* QP properties */
  VAPI_pd_hndl_t pd_hndl; /* Protection domain used by the QP */

  IB_mtu_t  mtu;
  u_int32_t msgsz; /* size of send messages */
  u_int32_t max_oust_wr_sq; /* max outstanding send queue work requests */
  u_int32_t max_oust_wr_rq; /* max outstanding receive queue work requests */

  VAPI_mrw_t send_mr;     /* memory region used with sends */
  VAPI_mr_hndl_t  send_mr_hndl; /* handle to memory region used with sends */
  VAPI_mrw_t rcv_mr;      /* memory region used with receives */
  VAPI_mr_hndl_t  rcv_mr_hndl; /* handle to memory region used with receives */
  VAPI_mrw_t data_mr;      /* memory region used for data */
  VAPI_mr_hndl_t  data_mr_hndl; /* handle to memory region used for data */

  VAPI_wr_id_t  last_posted_send_id;
  VAPI_wr_id_t  last_posted_rcv_id;

  IB_port_t port; /* port used for communications */
  MT_bool port_down; /* Don't wait for link to go up */
  VAPI_hca_port_t hca_port; /* params of the used port */


  opmode_t mode;

  MT_bool autordma;  /* when true indicates that the server should send RDMA write
                         requests to the client without the need to get a command */
  IB_virt_addr_t auto_va;
  IB_rkey_t auto_rkey;
  u_int32_t auto_size;

  MT_bool nordma;

  MT_bool debug; /* debug mode */
  MT_bool immediate; /* send immediate data */

  MT_bool memwindows; /* run test with memory windows */
  MT_bool atomics;    /* run with atomic function tests */
  MT_bool srq;  /* run test using SRQ for RQ */

  IB_lid_t remote_lid; /* remote peer's LID */
  IB_wqpn_t remote_qpn; /* remote peer's QPN */

#if defined(__WIN__) 
  MOSAL_thread_t scomp_thread;  /* thread dealing with send completion queue */
  MOSAL_thread_t rcomp_thread;  /* thread dealing with receive completion queue */
#else
  pthread_t scomp_thread;  /* thread dealing with send completion queue */
  pthread_t rcomp_thread;  /* thread dealing with receive completion queue */
#endif  
  volatile MT_bool thread_run_cmd; /* TRUE when thread should run, FALSE when it should exit */
  volatile MT_bool peer_ended; /* set TRUE to notify that the peer wants to end the connection */
  volatile MT_bool imd_ret; /* TRUE id thread should return immediatly */

  volatile int cur_send_outstanding; /* number of outstanding requests on the send Q */
  volatile int cur_posted_rcv_bufs; /* number of posted receive buffers */
  volatile int snd_rcv_balance;
#if defined(__WIN__) 
  MOSAL_mutex_t snd_mutex;
  MOSAL_mutex_t rcv_mutex;
  MOSAL_mutex_t bl_mutex;
  MOSAL_mutex_t cln_mutex;
#else
  pthread_mutex_t snd_mutex;
  pthread_mutex_t rcv_mutex;
  pthread_mutex_t bl_mutex;
  pthread_mutex_t cln_mutex;
#endif

  char fill_val; /* fill value passed in sends */

  char remote_ip[16]; /* remote host ip address */
  int socket; /* socket handle used for communication */


  m_windows_ctx_t *mw_arr; /* array of memory windows contexts */
  u_int32_t inflight_mw_binds; /* number of inflight memory windows */

  MT_bool server_auto_connect; /* set to TRUE by the --r option when the server uses auto connection */

  int dvar;  /*  debug var ?? */
  u_int32_t total_send_completions; /* used for displaying progress dots ..... */
  MT_bool  swap_endianness;  /* if TRUE, client and server have different endianness */
  u_int32_t         tcp_port; /* TCP/IP port that will be used */
  unsigned int seed;  /* seed used by the process */
  int static_seed;  /* if 1 use the above seed. Otherwise use a random one */
}
test_options_t;

#ifdef __WIN__
#pragma pack(push, 1) 	/*  packing is now 1 */
#endif
struct oob_msg_st {
  char mode;
  char endian;     /*  'b' for big endian, 'l' for little endian */
  IB_lid_t lid;
  IB_wqpn_t wqpn;
  IB_virt_addr_t va;
  IB_rkey_t rkey;
  u_int32_t size;
} __attribute__ ((packed));
#ifdef __WIN__
#pragma pack(pop) 		/*  return to default packing */
#endif


/* structure of command messages between the client and the server */
#ifdef __WIN__
#pragma pack(push, 1) 	/*  packing is now 1 */
#endif
struct cmd_msg_st {
  u_int32_t tid;
  u_int32_t command;
  u_int32_t block_num;
  IB_virt_addr_t va;
  IB_rkey_t rkey;
  u_int32_t size;
  u_int32_t wnd_idx; /* index to memory window context element in the array */
} __attribute__ ((packed));
#ifdef __WIN__
#pragma pack(pop) 		/*  return to default packing */
#endif

/*=== commands macro definitions ===*/
enum command_defs {
  CMD_END,
  CMD_READ_BLOCK,
  CMD_WRITE_BLOCK,
  CMD_ATOMIC_FA,
  CMD_ATOMIC_CSWP,
  CMD_ACK,
  CMD_NAK,
  CMD_NOP
};


void print_qp_props(VAPI_qp_prop_t *qp_props);
void print_hca_cap(VAPI_hca_vendor_t *hca_vendor_p, VAPI_hca_cap_t *hca_cap_p);

#endif /* H_RCTP_H */
