#ifndef __TYPES_H__
#define __TYPES_H__

/******************************************************************
                        define section
******************************************************************/
#define MAX_PACKET_SIZE                         256         /* size of a MAD packet */
#define IB_SA_DATA_SIZE                         200         /* SA data in packet */
#define NUM_OF_COMPLETION                       2           /* we'll have 2 completions (for send and recv) */
#define QP_SG_SIZE                              1           /* number of s/g entries in the QP */
#define QP_OUST_WR_SIZE                         1           /* number of outstanding WR in the QP */
#define DEF_PKEY_IX                             0           /* p_key index to use */
#define IB_QP1_WELL_KNOWN_Q_KEY			        0x80010000  /* QKey of SA's QP 1 */
#define IB_QP1								    0x00000001  /* QP1 number */
#define SENSITIVITY                             100         /* the sensitivity level for wating for completions */

#define CHECK_RC(verb, var, cmd)  if (var != VAPI_OK) { \
                                    if (config.verbose == 1) { \
                                       printf("Error, %s failed: %s\n", verb, VAPI_strerror(var));\
                                    } \
                                    cmd; \
                                  }


/******************************************************************
                        typedef section
******************************************************************/
typedef struct
{
    VAPI_hca_hndl_t             hca_hndl;
    VAPI_hca_port_t             hca_port;
    EVAPI_async_handler_hndl_t  async_hndl;
    VAPI_pd_hndl_t              pd_hndl;
    VAPI_cq_hndl_t              cq_hndl;
    u_int32_t                   cq_size;
    VAPI_qp_hndl_t              qp_hndl;
    VAPI_qp_prop_t              qp_prop;
    VAPI_ud_av_hndl_t           av_hndl;
    VAPI_mr_hndl_t              snd_mr_hndl;
    VAPI_mrw_t                  snd_mr_props;
    void*                       snd_buf;
    VAPI_mr_hndl_t              rcv_mr_hndl;
    VAPI_mrw_t                  rcv_mr_props;
    void*                       rcv_buf;

} resource_t;

typedef struct
{
    VAPI_hca_id_t               hca_id;          /* device ID */
    u_int8_t                    port_num;       /* port number to use */
    u_int8_t                    timeout;        /* time (in seconds) to wait for reply */
    u_int8_t                    verbose;        /* verbose mode */

} config_t;


typedef struct _ib_sa_mad
{
	u_int8_t				    base_ver;
	u_int8_t				    mgmt_class;
	u_int8_t				    class_ver;
	u_int8_t				    method;
	u_int16_t				    status;
	u_int16_t				    resv;
	u_int64_t				    trans_id;
	u_int16_t				    attr_id;
	u_int16_t				    resv1;
	u_int32_t				    attr_mod;

	u_int8_t				    rmpp_version;
	u_int8_t				    rmpp_type;
	u_int8_t				    rmpp_flags;
	u_int8_t				    rmpp_status;

	u_int32_t				    seg_num;
	u_int32_t				    paylen_newwin;

	u_int64_t				    sm_key;

	u_int16_t				    attr_offset;
	u_int16_t				    resv3;

	u_int64_t				    comp_mask;

	u_int8_t				    data[IB_SA_DATA_SIZE];

}	ib_sa_mad_t;

#endif
