/*
  This software is available to you under a choice of one of two
  licenses.  You may choose to be licensed under the terms of the GNU
  General Public License (GPL) Version 2, available at
  <http://www.fsf.org/copyleft/gpl.html>, or the OpenIB.org BSD
  license, available in the LICENSE.TXT file accompanying this
  software.  These details are also available at
  <http://openib.org/license.html>.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
  BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
  ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
  CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
  SOFTWARE.

  Copyright (c) 2004 Mellanox Technologies Ltd.  All rights reserved.
*/

#ifndef __WIN__
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/mman.h>
#include <signal.h>
#include <pthread.h>
  
#endif

#include <mtl_common.h>

#include <vapi.h>
#include <evapi.h>
#include <vapi_common.h>
#include <ib_defs.h>

#include "vping.h"
#define MTGETOPT_EXPORTS
#include <mt_getopt.h>
#include <mosal.h>
#include <string.h>

//DIEGO moved the defines below to vping.h so that they can be used in vpingm.c
//#define VPING_MAX_AVS       1000 
//#define VPING_PREALLOC_AVS  500

/* VPING resources contexts */
#define MAX_RESOURCES 16
static vping_res_t     vping_res[MAX_RESOURCES];
static u_int32_t vping_res_num= 0;   /* Number of valid vping_res entries */


static void MT_API signal_comp_handler(
  /*IN*/ VAPI_hca_hndl_t hca_hndl,
  /*IN*/ VAPI_cq_hndl_t cq_hndl,
  /*IN*/ void* sem_p)
{
  MTL_DEBUG1("signal_comp_handler: Got completion event for CQ=" MT_ULONG_PTR_FMT " (sem_p=%p)\n",
    cq_hndl,sem_p);
  MOSAL_sem_rel((MOSAL_semaphore_t*)sem_p);
}

static void MT_API async_events_handler(
  VAPI_hca_hndl_t hca_hndl,
  VAPI_event_record_t *event_p,
  void* priv_data
)
{
  vping_res_t *res= (vping_res_t*)priv_data;

  printf("Got async. event in vping session %u: %s (%s)\n",res->vping_res_num,
          VAPI_event_record_sym(event_p->type),VAPI_event_syndrome_sym(event_p->syndrome));
}

VAPI_ret_t vping_qp_create(vping_res_t *res)
{
  VAPI_ret_t ret;

  VAPI_hca_port_t     hca_port_info;
  VAPI_hca_vendor_t   hca_vendor;
  VAPI_hca_cap_t      hca_cap;

  VAPI_qp_init_attr_t qp_init_attr;
  VAPI_qp_prop_t      qp_prop;
  VAPI_cqe_num_t      act_num;
  unsigned int i;
  VAPI_ud_av_t        av;

  /* Query HCA capabilities */
  ret = VAPI_query_hca_cap (res->hca_hndl, &hca_vendor, &hca_cap); VAPI_CHECK_RET;

  /* Get HCA LID */
  ret = VAPI_query_hca_port_prop(res->hca_hndl, res->port , &hca_port_info); VAPI_CHECK_RET;

  res->slid =  hca_port_info.lid;

  /* Get a PD */
  if (res->vping_max_avs == EVAPI_DEFAULT_AVS_PER_PD) {
      MTL_DEBUG1("Using VAPI_alloc_pd\n");
      ret = VAPI_alloc_pd(res->hca_hndl, &(res->pd_hndl)); VAPI_CHECK_RET;
  } else {
      ret = EVAPI_alloc_pd(res->hca_hndl, res->vping_max_avs,  &(res->pd_hndl)); VAPI_CHECK_RET;
      if (res->vping_prealloc_avs == 0) {
          MTL_DEBUG1("Using EVAPI_alloc_pd. max AVs = %u\n", res->vping_max_avs);
      } else {
          MTL_DEBUG1("Using EVAPI_alloc_pd. max AVs = %u, allocating %u\n", 
                 res->vping_max_avs, res->vping_prealloc_avs);
          res->prealloc_av_hndls = TNVMALLOC(VAPI_ud_av_hndl_t, res->vping_prealloc_avs);
          if (res->prealloc_av_hndls == NULL) {
              ret= VAPI_EAGAIN;
              printf("could not allocate memory for pre-allocated AV handles\n");
              VAPI_CHECK_RET; 
          }
          for (i = 0 ; i < res->vping_prealloc_avs ; i++) {
             (res->prealloc_av_hndls)[i] = VAPI_INVAL_HNDL; 
          }
    
          memset(&av, 0, sizeof(av));
          av.dlid = 2;
          av.port = 1;
          av.grh_flag = FALSE;
        
          for (i = 0 ; i < res->vping_prealloc_avs ; i++) {
            ret =  VAPI_create_addr_hndl(res->hca_hndl,res->pd_hndl,&av, &((res->prealloc_av_hndls)[i]));
            if (ret != VAPI_OK) {
                printf("***ERROR: PRE-ALLOCATED only %d AV HANDLES on PD(ret = %d)\n", i, ret);
                break;
            }
          }
          if (i == res->vping_prealloc_avs) {
              //printf("***SUCCESSFULLY PRE-ALLOCATED %d AV HANDLES on PD\n", i);
          }
      }
  }

  /* Create CQ for RQ and SQ*/  /* TBD - Check we have enough act nums */
  if (res->opts & VPING_OPTS_DAEMON) {
    ret = VAPI_create_cq(res->hca_hndl, MAX_RPT + 1, &(res->sq_cq_hndl), &act_num); 
    VAPI_CHECK_RET; 
    ret = VAPI_create_cq(res->hca_hndl, MAX_OUTSTANDING_PINGS+1, &(res->rq_cq_hndl), &act_num); 
    VAPI_CHECK_RET;
  } else {    /* client */
    ret = VAPI_create_cq(res->hca_hndl, 2 , &(res->sq_cq_hndl), &act_num); 
    VAPI_CHECK_RET; 
    ret = VAPI_create_cq(res->hca_hndl, res->rpt + 1, &(res->rq_cq_hndl), &act_num); 
    VAPI_CHECK_RET;
  }

  /* Initialize completion events semaphores and register event handlers */
  if ((res->opts & VPING_OPTS_EVENT) || (res->opts & VPING_OPTS_NEVENT)) {
    MOSAL_sem_init(&(res->rq_sem),0);
    MOSAL_sem_init(&(res->sq_sem),0);
    ret= EVAPI_set_comp_eventh(res->hca_hndl,res->rq_cq_hndl,signal_comp_handler,
      &(res->rq_sem),&(res->rq_cq_eventh));
    VAPI_CHECK_RET;
    ret= EVAPI_set_comp_eventh(res->hca_hndl,res->sq_cq_hndl,signal_comp_handler,
      &(res->sq_sem),&(res->sq_cq_eventh));
    VAPI_CHECK_RET;
    ret= EVAPI_set_async_event_handler(res->hca_hndl,async_events_handler,res,&res->async_hndl);
    VAPI_CHECK_RET;
  }

  /* register event handlers for polling(block mode) internal use*/
  if (res->opts & VPING_OPTS_BLOCK_POLL) {
    ret= EVAPI_set_comp_eventh(res->hca_hndl,res->rq_cq_hndl,EVAPI_POLL_CQ_UNBLOCK_HANDLER,NULL,&(res->rq_cq_eventh));
    VAPI_CHECK_RET;
    ret= EVAPI_set_comp_eventh(res->hca_hndl,res->sq_cq_hndl,EVAPI_POLL_CQ_UNBLOCK_HANDLER,NULL,&(res->sq_cq_eventh));
    VAPI_CHECK_RET;
    ret= EVAPI_set_async_event_handler(res->hca_hndl,async_events_handler,res,&res->async_hndl);
    VAPI_CHECK_RET;
  }

  /* Create QP */
  if (res->opts & VPING_OPTS_DAEMON) {
    qp_init_attr.cap.max_oust_wr_sq = MAX_RPT + 1;
    qp_init_attr.cap.max_oust_wr_rq = MAX_OUTSTANDING_PINGS + 1; 
  } else { /* client */
    qp_init_attr.cap.max_oust_wr_sq = 2;
    qp_init_attr.cap.max_oust_wr_rq = res->rpt +1; 
  }

  /* Scatter/Gather lists size */
  if (res->opts & VPING_OPTS_INLINE_SEND) { 
    /* For inline data allocate maximum WQE size */
    qp_init_attr.cap.max_sg_size_sq = hca_cap.max_num_sg_ent;
  } else { /* No inline data */
    qp_init_attr.cap.max_sg_size_sq = 1; /* TBD set this to be more dynamic */
  }
  qp_init_attr.cap.max_sg_size_rq = 1; /* TBD set this to be more dynamic */

  qp_init_attr.pd_hndl            = res->pd_hndl;
  qp_init_attr.rdd_hndl           = 0;
  qp_init_attr.rq_cq_hndl         = res->rq_cq_hndl;
  qp_init_attr.rq_sig_type        = VAPI_SIGNAL_ALL_WR; /* default for  IB */
  qp_init_attr.sq_cq_hndl         = res->sq_cq_hndl;
  qp_init_attr.sq_sig_type        = VAPI_SIGNAL_REQ_WR;
  qp_init_attr.ts_type            = VAPI_TS_UD;


  ret = VAPI_create_qp(res->hca_hndl, &qp_init_attr, &(res->qp_hndl), &qp_prop); VAPI_CHECK_RET;
  if (res->opts & VPING_OPTS_INLINE_SEND) { 
    if (qp_prop.cap.max_inline_data_sq < res->payload_size) {
      printf("Requested payload size (%d) is too large for max_inline_data (%d)\n",
        res->payload_size, qp_prop.cap.max_inline_data_sq);
      return VAPI_EGEN;
    }
    printf("Using inline sendQ (max_inline_data=%d bytes)\n",qp_prop.cap.max_inline_data_sq);
  }

  res->local_qpnum = qp_prop.qp_num;

  return(VAPI_OK);
}



VAPI_ret_t   vping_qp2init(vping_res_t *res)
{
  VAPI_ret_t ret;

  VAPI_qp_attr_t      qp_attr;
  VAPI_qp_attr_mask_t qp_attr_mask;
  VAPI_qp_cap_t       qp_cap;


  /* 
   * Change QP to INIT
   *
   */
  QP_ATTR_MASK_CLR_ALL(qp_attr_mask);
  qp_attr.qp_state = VAPI_INIT;          
  QP_ATTR_MASK_SET(qp_attr_mask,QP_ATTR_QP_STATE);
  qp_attr.pkey_ix  = 0;
  QP_ATTR_MASK_SET(qp_attr_mask,QP_ATTR_PKEY_IX);  
  qp_attr.port     = res->port;
  QP_ATTR_MASK_SET(qp_attr_mask,QP_ATTR_PORT);
  qp_attr.qkey = res->qkey; 
  QP_ATTR_MASK_SET(qp_attr_mask,QP_ATTR_QKEY);

  /* If I do not set this mask, I get an error from HH. QPM should catch it */
  ret = VAPI_modify_qp(res->hca_hndl, res->qp_hndl, &qp_attr, &qp_attr_mask, &qp_cap); VAPI_CHECK_RET;
  return VAPI_OK;
}

VAPI_ret_t   vping_qp2rtr(vping_res_t *res)
{
  VAPI_ret_t ret;

  VAPI_qp_attr_t      qp_attr;
  VAPI_qp_attr_mask_t qp_attr_mask;
  VAPI_qp_cap_t       qp_cap;


  /*
   *  Change QP to RTR
   *
   */
  QP_ATTR_MASK_CLR_ALL(qp_attr_mask);
  qp_attr.qp_state = VAPI_RTR;
  QP_ATTR_MASK_SET(qp_attr_mask,QP_ATTR_QP_STATE);

  ret = VAPI_modify_qp(res->hca_hndl, res->qp_hndl, &qp_attr, &qp_attr_mask, &qp_cap); VAPI_CHECK_RET;
  return(ret);
}

VAPI_ret_t   vping_qp2rts(vping_res_t *res)
{
  VAPI_ret_t ret;

  VAPI_qp_attr_t      qp_attr;
  VAPI_qp_attr_mask_t qp_attr_mask;
  VAPI_qp_cap_t       qp_cap;

  /* 
   * Change QP to RTS
   *
   */
  QP_ATTR_MASK_CLR_ALL(qp_attr_mask);
  qp_attr.qp_state   = VAPI_RTS;      
  QP_ATTR_MASK_SET(qp_attr_mask,QP_ATTR_QP_STATE);
  qp_attr.sq_psn   = 0;               
  QP_ATTR_MASK_SET(qp_attr_mask,QP_ATTR_SQ_PSN);

  ret = VAPI_modify_qp(res->hca_hndl,res->qp_hndl,&qp_attr,&qp_attr_mask,&qp_cap); 
  MTL_TRACE1("%s: after VAPI_modify_qp\n", __func__);
  VAPI_CHECK_RET;

  return(ret);

}



VAPI_ret_t  vping_create_mr(vping_res_t *res)
{
    VAPI_ret_t  ret;
    VAPI_mrw_t mr_in, mr_out;

    FUNC_IN;

    if (res->opts & VPING_OPTS_DAEMON)
        res->buf_size = (MAX_PAYLOAD_SZ + GRH_LEN) * (MAX_RPT + MAX_OUTSTANDING_PINGS + 1);
    else
        res->buf_size = (res->payload_size + GRH_LEN) * (res->rpt + 1); /* Add place for grh */

    MTL_TRACE1("%s: before creating mr\n", __func__);
    mr_in.acl = VAPI_EN_LOCAL_WRITE | VAPI_EN_REMOTE_WRITE;            /* Enable local and remote access to memory region */
    mr_in.l_key =0;
    mr_in.pd_hndl = res->pd_hndl;
    mr_in.r_key =0;
    mr_in.size = res->buf_size;


    {
        MTL_DEBUG1("register virtual mr \n");
       /* Register single memory address region */
        MTL_TRACE1("%s: before VMALLOC. res->buf_size=%d. \n", __func__, res->buf_size);
        res->buf_ptr = VMALLOC (res->buf_size);
        MTL_TRACE1("%s: after VMALLOC. res->buf_ptr=%p. \n", __func__, res->buf_ptr);

        if (res->buf_ptr == NULL)
        {
            ret = VAPI_EAGAIN;
            VAPI_CHECK_RET;
        }

        mr_in.start = (VAPI_virt_addr_t)(MT_virt_addr_t)res->buf_ptr;
        mr_in.type = VAPI_MR;
    } 


    ret = VAPI_register_mr(res->hca_hndl, &mr_in, &(res->mr_hndl), &mr_out);
    if ( ret != VAPI_OK) { 
	VAPI_RET_PRINT(ret);
	if (res->buf_ptr) {
	    VFREE(res->buf_ptr);
	    res->buf_ptr = NULL;
	}
	return(ret); 
    }
    res->l_key = mr_out.l_key;
    MTL_DEBUG1("got lkey: 0x%x \n",mr_out.l_key); 
    return(ret);
}


void attach_to_mcg(vping_res_t *res)
{
  unsigned int i;
  VAPI_ret_t ret;

  for (i= 0; i< res->num_of_mcg; i++) {
    ret= VAPI_attach_to_multicast(res->hca_hndl,res->mcgid[i],res->qp_hndl,0);
    if (ret != VAPI_OK) {
      MTL_ERROR2("%s: Failed to attach to multicast group DGID="GID_FMT" (%s).\n", __func__,
        GID_FMT_DATA(res->mcgid[i]),VAPI_strerror_sym(ret));
    } else {
      printf("Attached to multicast GID "GID_FMT"\n",GID_FMT_DATA(res->mcgid[i]));
    }
  }
}


static void detach_of_mcg(vping_res_t *res)
{
  unsigned int i;
  VAPI_ret_t ret;

  for (i= 0; i< res->num_of_mcg; i++) {
    ret= VAPI_detach_from_multicast(res->hca_hndl,res->mcgid[i],res->qp_hndl,0);
    if (ret != VAPI_OK) {
      MTL_ERROR2("%s: Failed to detach from multicast group DGID="GID_FMT" (%s).\n", __func__,
        GID_FMT_DATA(res->mcgid[i]),VAPI_strerror_sym(ret));
    }
  }
}


VAPI_ret_t vping_init(vping_res_t *res)
{
  VAPI_ret_t  ret;

  res->hca_hndl= VAPI_INVAL_HNDL;
  res->pd_hndl= VAPI_INVAL_HNDL;
  res->rq_cq_hndl= VAPI_INVAL_HNDL;
  res->sq_cq_hndl= VAPI_INVAL_HNDL;
  res->sq_cq_eventh= VAPI_INVAL_HNDL;
  res->rq_cq_eventh= VAPI_INVAL_HNDL;
  res->async_hndl= VAPI_INVAL_HNDL;
  res->qp_hndl= VAPI_INVAL_HNDL;
  res->mr_hndl= VAPI_INVAL_HNDL;

#if defined( __LINUX__ ) 
  //DIEGO
  res->socket=-1;
#endif

  MTL_DEBUG1("Initializing resources on HCA=%s ...\n",res->hca_id);
  MTL_DEBUG2("HCA=%s port=%d qkey=%d\n",res->hca_id,res->port,res->qkey);
  /* 
   * Open HCA 
   *
   */
  ret = EVAPI_get_hca_hndl(res->hca_id, &(res->hca_hndl)); 
  if (ret != VAPI_OK) {
    res->hca_hndl= VAPI_INVAL_HNDL;  /* Make sure "cleanup" knows there are no resources to clean */
    printf("Could not find a channel adapter with HCA-ID = %s.  Exiting\n",res->hca_id);
    return ret;
  }
  /*
   * Create QP
   *
   */
  ret = vping_qp_create(res); VAPI_CHECK_RET;

  /*
   * Move to init
   *
   */
  ret = vping_qp2init(res); VAPI_CHECK_RET;


  /*
   * Initialize memory regions
   *
   */
  ret = vping_create_mr(res); VAPI_CHECK_RET;
  

  return(VAPI_OK);
}



VAPI_ret_t vping_cleanup(vping_res_t *res)
{
  unsigned int i;
  printf("Cleanup for session %u - HCA=%s (error count = %d).\n", 
    res->vping_res_num,res->hca_id,res->error_counter);
  
  if (res->prealloc_av_hndls != NULL) {
     for(i=0; i<res->vping_prealloc_avs; i++) {
         if ((res->prealloc_av_hndls)[i] != VAPI_INVAL_HNDL) {
             VAPI_destroy_addr_hndl(res->hca_hndl,(res->prealloc_av_hndls)[i]);
         }
     }
     VFREE(res->prealloc_av_hndls);
  }
  if (res->qp_hndl != VAPI_INVAL_HNDL) {
    detach_of_mcg(res);
    VAPI_destroy_qp(res->hca_hndl,res->qp_hndl);
  }
  if (res->async_hndl != VAPI_INVAL_HNDL) {
    EVAPI_clear_async_event_handler(res->hca_hndl,res->async_hndl);
  }
  if (res->sq_cq_eventh != VAPI_INVAL_HNDL) {
    EVAPI_clear_comp_eventh(res->hca_hndl,res->sq_cq_eventh);
  }
  if (res->rq_cq_eventh != VAPI_INVAL_HNDL) {
    EVAPI_clear_comp_eventh(res->hca_hndl,res->rq_cq_eventh);
  }
  if (res->rq_cq_hndl != VAPI_INVAL_HNDL) {
    VAPI_destroy_cq(res->hca_hndl,res->rq_cq_hndl);
  }
  if (res->sq_cq_hndl != VAPI_INVAL_HNDL) {
    VAPI_destroy_cq(res->hca_hndl,res->sq_cq_hndl);
  }
  if (res->mr_hndl != VAPI_INVAL_HNDL) {
    VAPI_deregister_mr(res->hca_hndl,res->mr_hndl);
  }
  if (res->buf_ptr) {
      VFREE(res->buf_ptr);
      res->buf_ptr = NULL;
  }

  if (res->pd_hndl != VAPI_INVAL_HNDL) {
    VAPI_dealloc_pd(res->hca_hndl,res->pd_hndl);
  }
  /* open/close of HCA should be done system wide - not per application */
  if (res->hca_hndl != VAPI_INVAL_HNDL) {
    EVAPI_release_hca_hndl(res->hca_hndl);  
  }

#if defined( __LINUX__ ) 
  //DIEGO
  if (res->socket != -1) {
    MOSAL_socket_close(res->socket);
  }
#endif
  
  return VAPI_OK;
}


#ifdef __WIN__
BOOL WINAPI sig_handler(  DWORD dwCtrlType )
{
	int signum = (int)dwCtrlType;
#else
void sig_handler(int signum)
{
#endif
  unsigned int i;
  static int got_signal = 0;
  /* prevent handling any signal more than once, so that the cleanup code is executed 
   * only once.  Note that since there is the same behavior for ANY signal handled by
   * this handler (i.e., the process cleans itself up and exits), this code should be 
   * executed only once, even if it is invoked for a different signal number.
   */
  if (!got_signal) {
      got_signal++;
      printf("Got signal %d - exiting.\n",signum);
      for (i= 0; i < vping_res_num; i++) {
        vping_res[i].exit_flag= TRUE;
        if ((vping_res[i].opts & VPING_OPTS_EVENT) || (vping_res[i].opts & VPING_OPTS_NEVENT)) {
            MOSAL_sem_rel(&(vping_res[i].rq_sem));
            MOSAL_sem_rel(&(vping_res[i].sq_sem));
        } else if (vping_res[i].opts & VPING_OPTS_BLOCK_POLL) {
            EVAPI_poll_cq_unblock(vping_res[i].hca_hndl,vping_res[i].rq_cq_hndl);
            EVAPI_poll_cq_unblock(vping_res[i].hca_hndl,vping_res[i].sq_cq_hndl);
        }
      }
  } else {
      MTL_TRACE1("%s:got another signal (%d). signal count = %d. IGNORING\n", __func__,signum, got_signal);
      got_signal++;
  }
#ifdef __WIN__
return TRUE;
#endif
}

void reg_signals(void)
{
#ifdef __WIN__
    SetConsoleCtrlHandler(sig_handler, TRUE);
    SetErrorMode(SEM_NOGPFAULTERRORBOX);
#else    
  struct sigaction sigact;
  sigact.sa_handler = sig_handler;
  sigemptyset(&sigact.sa_mask);
  sigact.sa_flags = 0;
  sigaction(SIGINT, &sigact, NULL);
#endif  
  return;
}

#if defined( __LINUX__ ) 
/* Using void* in parameter in order to match pthread start function prototype */
//DIEGO the function was defined as static. I removed the static so that it can be called
//from the tcpdaemon in vpingm.c
void* start_vping(void *void_res)
#define MAX_GID 256
#else
static void* start_vping(void *void_res)
#define MAX_GID 512
#endif
{
  VAPI_ret_t ret;
  IB_gid_t gid_tbl[MAX_GID];
  u_int16_t out_size;
  int i;
  IB_gid_t null_gid= {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};
  vping_res_t *res= void_res;


#if defined( __LINUX__ ) 
  //DIEGO
  if (res->opts & VPING_OPTS_MANAGER) {
    ret=vping_manager(res);
  } else {
#endif

    if (!(res->rpt)) res->rpt = 1;
    if (!(res->npackets)) res->npackets = 1;


    res->error_counter = 0;


    /*
     * Initialize HCA and QP
     *
     */
    ret = vping_init(res);
    if ( ret != VAPI_OK) {
      res->error_counter++;/* make sure error is propagated to process' exit code*/
      MTL_ERROR1("Failed initializing resources for %s.\n",res->hca_id);
      return (void*)(MT_virt_addr_t)ret;
    }

    reg_signals();  /* register cleanup handler */

    if (res->opts & VPING_OPTS_DAEMON) { /* Ping/Daemon */
      ret= VAPI_query_hca_gid_tbl(res->hca_hndl,res->port,MAX_GID,&out_size,gid_tbl);
      if (ret != VAPI_OK) {
        MTL_ERROR1("! Failed getting port %d's GID table (%s)!\n",
          res->port,VAPI_strerror_sym(ret));
        res->error_counter++;/* make sure error is propagated to process' exit code*/
        return (void*)(MT_virt_addr_t)ret;
      }
      MTL_DEBUG1("vpingd: listening on LID 0x%X and QP 0x%X\n", res->slid, res->local_qpnum);
      printf("vpingd: %s listening on LID 0x%X (port %d) and QP 0x%X . GID table:\n",
        res->hca_id,res->slid,res->port,res->local_qpnum);
      for (i= 0; i < out_size; i++) {
        /* compare only GUID part of GID to 0s */
        if (memcmp(((u_int8_t*)null_gid)+8,((u_int8_t*)(gid_tbl+i))+8,sizeof(IB_gid_t)-8) != 0) {
          printf("  GID[%3d]= " GID_FMT "\n",i,GID_FMT_DATA(gid_tbl[i]));
        }
      }
      fflush( stdout );	/* XOL Need it to enable running vping from script */

#if defined(__LINUX__) 
      //DIEGO
      if (res->cluster_conn_test_fl== TRUE) {
          if (MOSAL_thread_start(&(res->tcpdaemon_thread),MOSAL_THREAD_FLAGS_DETACHED,start_vpingtcpdaemon,res) != 0) {
            printf("Failed creating thread for tcpdaemon session of HCA=%s (%s).\n",vping_res[i].hca_id,strerror(errno));
            res->error_counter++;/* make sure error is propagated to process' exit code*/
            ret=VAPI_EGEN; //REV which error code should I return?
            return (void*)(MT_virt_addr_t)ret;
          }
      }
      attach_to_mcg(res);
      ret= vping_daemon(res);

      //DIEGO
      if (res->cluster_conn_test_fl== TRUE) {
          MOSAL_thread_kill(&(res->tcpdaemon_thread)); //... and vping_cleanup will destroy the socket
      }
#else
      attach_to_mcg(res);
      ret= vping_daemon(res);
#endif

    } else {
      printf("Client session %u on %s port %d (qpn=0x%X) sending to lid=0x%X qp=0x%X ",
	     res->vping_res_num,res->hca_id,res->port,res->local_qpnum,res->dlid, res->remote_qpnum);
      if (res->opts & VPING_OPTS_GRH) {
        printf("dgid=" GID_FMT,GID_FMT_DATA(res->dgid));
      }
      printf("\n");
      ret = vping_client(res);
    }
    vping_cleanup(res);

#if defined( __LINUX__ ) 
  } //DIEGO
#endif

  if ( ret != VAPI_OK) res->error_counter++;/* make sure error is propagated to process' exit code*/
  return (void*)(MT_virt_addr_t)ret;
}


static void setup_default_session(int res_num)
{
  /* Init entry with default */
  vping_res[res_num].exit_flag= FALSE;
  vping_res[res_num].opts= 0;
  vping_res[res_num].num_of_mcg= 0;
  vping_res[res_num].port= 1;
  vping_res[res_num].qkey= VPING_Q_KEY;
  strcpy(vping_res[res_num].hca_id, HCA_ID);
  vping_res[res_num].payload_size= MIN_PAYLOAD_SZ;
  vping_res[res_num].vping_max_avs = VPING_MAX_AVS;
  vping_res[res_num].vping_prealloc_avs = VPING_PREALLOC_AVS;
  vping_res[res_num].vping_res_num = res_num;


#if defined(__LINUX__) 
  //DIEGO
  vping_res[res_num].socket_port=VPING_TCPD_DEFSOCKETPORT;
  vping_res[res_num].simultpingers=MAX_OUTSTANDING_PINGS-1; //so that by def we never overrun the daemon recvQ
  vping_res[res_num].cluster_conn_test_fl=FALSE; // by default, do not run cluster-test version of daemon
#endif
}


static void str2gid(char* gid_str, IB_gid_t *gid)
{
  char *cur_char,*next_char;
  int i;

  for (i= 0, cur_char= gid_str; i < 16; i++) { /* convert string to IB_gid_t */
    (*gid)[i]= (u_int8_t)strtoul(cur_char, &next_char , 10);
    if ((next_char == cur_char) ||
        ((i < 15) && (*next_char != '.')) ) {
      printf("decimal dotted format must be used for DGID, e.g. 10.0.0.229.... (16 numbers):\n-g ");
      cur_char= gid_str;
      while ((*cur_char != ' ') && (*cur_char != 0)) {
        printf("%c",*cur_char); cur_char++;
      }
      printf("\n   ");
      for (; next_char > gid_str; next_char--)  printf(" ");
      printf("^\n");

      exit(1);
    }
    cur_char= next_char+1;  /* skip dot */
  }

}


static void usage(void)
{
  

#if defined( __LINUX__ ) 
  fprintf(stderr,
    "usage 1 (daemon): vping [-h HCA-ID] [-p port] [-C] [-P vping_daemon_tcp_port] [-v|-b] [-I|-l] \n"
    "        -d [-m multicast_GID] [-q qkey]... [-h HCA-ID] [-a max-avs-per-pd] \n"
    "        [-x avs-to-prealloc] -d... [\n" 
    "usage 2 (client): vping [-h HCA-ID] [-p port] [-v|-V|-b] [-I] [-c count] [-i inter-test-delay] \n"
    "        [-s packetsize] [-r reply-repeat-cnt] [-q qkey] [-g DGID]\n"
    "        [-n / -e] [-a max-avs-per-pd] [-x avs-to-prealloc] [-k] DLID QP\n"
    "usage 3 (manager): vping -M hostlistfilename [-P vping_daemon_tcp_port] [-S simult_pingers] \n");
#else    
  fprintf(stderr,
    "usage 1 (daemon): vping [-h HCA-ID] [-p port] [-v|-b] [-I|-l] -d [-m multicast_GID] \n"
    "        [-m multicast_GID]... [-h HCA-ID] [-a max-avs-per-pd] [-x avs-to-prealloc] -d... [\n" 

    "usage 2 (client): vping [-h HCA-ID] [-p port] [-v|-V|-b] [-I] [-c count] [-i inter-test-delay] \n"
    "        [-s packetsize] [-r reply-repeat-cnt] [-q qkey] [-g DGID]\n"
    "        [-n / -e] [-a max-avs-per-pd] [-x avs-to-prealloc] [-k] DLID QP\n");
#endif  
  exit(2);
}


int MT_API main(int argc, char *argv[])
{
  int exit_code= 0; 
  int ch;
  unsigned int i,j;
  int once= 1;


  /* Initialize first/default session */
  vping_res_num= 1;
  memset(vping_res, 0, sizeof(vping_res));
  setup_default_session(0);


#if defined( __LINUX__ ) 
  while ((ch = MTGETOPT_getopt(argc, argv, "P:M:S:Ci:c:s:dq:m:r:g:h:p:a:x:nevVIblk"))!=EOF) {
#else
      while ((ch = MTGETOPT_getopt(argc, argv, "i:c:s:dq:m:r:g:h:p:a:x:nevVIblk"))!=EOF) {
#endif    
   switch (ch) {

#if defined( __LINUX__ ) 
      //DIEGO
      case 'P': //vping_tcp_daemon base tcp port number
        vping_res[vping_res_num-1].socket_port = atoi(optarg);
        break;
      case 'M': //vping_manager mode and hostlist file name
        vping_res[vping_res_num-1].opts |= VPING_OPTS_MANAGER;
        strcpy(vping_res[vping_res_num-1].remote_ip,optarg); //remote_ip holds IP list filename
        break;
      case 'S': //vping_manager max simultaneous pingers started at a time
        vping_res[vping_res_num-1].simultpingers = atoi(optarg);
        break;
       case 'C': //vping daemon is started in cluster connetivity mode, with ability
                 //to work opposite a vping_manager
        vping_res[vping_res_num-1].cluster_conn_test_fl = TRUE;
        break;
#endif    
      case 'i':
        vping_res[vping_res_num-1].intervalsecs = atoi(optarg);
        break;
      case 'I':  /* Inline send */
        vping_res[vping_res_num-1].opts |= VPING_OPTS_INLINE_SEND;
        break;
      case 'l':  /* Chain send request list for daemon */
        vping_res[vping_res_num-1].opts |= VPING_OPTS_CHAIN_SEND;
        break;
      case 'd':
        vping_res[vping_res_num-1].opts |= VPING_OPTS_DAEMON;
        /* set up daemon to work with old VAPI_alloc_pd */
        vping_res[vping_res_num-1].vping_max_avs = EVAPI_DEFAULT_AVS_PER_PD;
        vping_res[vping_res_num-1].vping_prealloc_avs = 0;
        break;
      case 'm': /* multicast groups to attach daemon */
        if (vping_res[vping_res_num-1].num_of_mcg == MAX_MCG) {
          printf("vping supports no more than %d multicast groups.\n",MAX_MCG);
          exit(1);
        }
        str2gid(optarg,&(vping_res[vping_res_num-1].mcgid[vping_res[vping_res_num-1].num_of_mcg]));
        vping_res[vping_res_num-1].num_of_mcg++;
        break;
      case 's':
        vping_res[vping_res_num-1].payload_size =  strtoul(optarg, 0, 0);
        if (vping_res[vping_res_num-1].payload_size > MAX_PAYLOAD_SZ){
          printf("vping supports packet up to %d bytes (Got -s %d).\n",
                 MAX_PAYLOAD_SZ, vping_res[vping_res_num-1].payload_size);
          exit(1);
        }
        if (vping_res[vping_res_num-1].payload_size < MIN_PAYLOAD_SZ)   {
          printf("Warning: Payload is less than %d bytes (Got -s %d). "
             "Daemon will only echo back one copy of the packet - repeat parameters are ignored\n",
             MIN_PAYLOAD_SZ, vping_res[vping_res_num-1].payload_size);
        }
        break;

      case 'q':
        vping_res[vping_res_num-1].qkey = strtoul(optarg, 0, 0);
        break;

      case 'r':
        vping_res[vping_res_num-1].rpt = strtoul(optarg, 0, 0);
        if (vping_res[vping_res_num-1].rpt > 255) {
          printf("vping supports -r option up to 255 only (Got -r %d).\n",vping_res[vping_res_num-1].rpt);
          exit(1);
        }
        break;

      case 'c': 
        vping_res[vping_res_num-1].npackets = strtoul(optarg, 0, 0);
        break;

      case 'g': /* With GRH (DGID as modifier) */
        str2gid(optarg,&(vping_res[vping_res_num-1].dgid));
        vping_res[vping_res_num-1].opts |= VPING_OPTS_GRH;
        break;

      case 'h': /* HCA ID (additional session) */
        if (once) once= 0;  /* First "-h" refers to default entry pointed by vping_res_num=1 */
        else { /* not the first default session */
          vping_res_num++; 
          if (vping_res_num > MAX_RESOURCES) {
            printf("vping supports no more than %d devices/sessions simultaneously.\n",
              MAX_RESOURCES);
            exit(1);
          }
          setup_default_session(vping_res_num-1);
        }
        strcpy(vping_res[vping_res_num-1].hca_id, optarg);
	/*printf("Got HCA-ID[%d]=%s\n",vping_res_num-1,vping_res[vping_res_num-1].hca_id);*/
        break;

      case 'p': /* QP's port */
        vping_res[vping_res_num-1].port = (IB_port_t)strtoul(optarg, 0, 0);
        break;
      
      case 'n':
        vping_res[vping_res_num-1].opts |= VPING_OPTS_RAND_RPT;
        break;

      case 'e': /* echo with rpt id */
        vping_res[vping_res_num-1].opts |= VPING_OPTS_RPT_W_ID;
        break;

      case 'v': /* Use completion events */
      case 'V': /* Use N-completion events */
        if (vping_res[vping_res_num-1].opts & VPING_OPTS_BLOCK_POLL) {
          fprintf(stderr, "vping: -b and -v are mutually exclusive." 
            " Choose only one of them per session.\n"); 
          exit(2);
        }
        vping_res[vping_res_num-1].opts |= ((ch == 'v') ? VPING_OPTS_EVENT : VPING_OPTS_NEVENT);
        break;
      case 'b': /* Use blocking polling on cqes  */
        if (vping_res[vping_res_num-1].opts & VPING_OPTS_EVENT) {
          fprintf(stderr, "vping: -b and -v are mutually exclusive."
                          " Choose only one of them per session.\n"); 
          exit(2);
        }
        vping_res[vping_res_num-1].opts |= VPING_OPTS_BLOCK_POLL;
        break;
    case 'a':
        vping_res[vping_res_num-1].vping_max_avs = strtoul(optarg, 0, 0);
        break;
    case 'x':
        vping_res[vping_res_num-1].vping_prealloc_avs = strtoul(optarg, 0, 0);
        break;
    case 'k':  /* send kill message */
        vping_res[vping_res_num-1].opts |= VPING_OPTS_KILL;
        break;

   default:
        usage();
    }
  }


  argc -= optind;
  argv += optind;

  
  for (i= 0; i< vping_res_num; i++) {
#if defined( __LINUX__ ) 
    //DIEGO
    if (!(vping_res[vping_res_num-1].opts & VPING_OPTS_MANAGER)) { /* not a manager session */
#endif
      if (vping_res[vping_res_num-1].vping_max_avs == 0) {
          fprintf(stderr, "cannot allocate zero vping_max_avs\n");
          usage();
          exit(2);
      }

      if (vping_res[vping_res_num-1].vping_max_avs != EVAPI_DEFAULT_AVS_PER_PD &&
          vping_res[vping_res_num-1].vping_max_avs <= vping_res[vping_res_num-1].vping_prealloc_avs) {
          fprintf(stderr, "vping_max_avs (-a) must be greater than vping_prealloc_avs (-x)\n");
          usage();
          exit(2);
      }

      if (!(vping_res[vping_res_num-1].opts & VPING_OPTS_DAEMON)) { /* client session */
        /* client session - for multiple client session all will get same (dlid,dqp) */
        if (argc != 2) {
            MTL_DEBUG4("argc: %d optind: %d\n",argc,optind);
            usage();
        }

        vping_res[i].dlid = (IB_lid_t)strtoul(*argv, 0, 0);
        if (vping_res[i].dlid == 0) {
          fprintf(stderr, "vping: wrong dlid \n");
          exit(2);
        }


        vping_res[i].remote_qpnum = strtoul(*(argv+1), 0, 0);
        if (vping_res[i].remote_qpnum == 0) {
          fprintf(stderr,
                  "vping: wrong remote queue pair number\n");
          exit(2);
        }

      } else { /* daemon session */
        if (vping_res[vping_res_num-1].opts & VPING_OPTS_NEVENT) {
          fprintf(stderr,"vping: Deamon session does not support -V option, use -v.\n");
          exit(2);
        }
      }
    }
#if defined( __LINUX__ ) 
  }
#endif

  if (vping_res_num == 1) { /* Avoid using threads for a single session (easier to debug) */
    start_vping(&vping_res[0]);
    if (vping_res[0].error_counter > 0)  exit_code= 1;

  } else {                  /* Create sessions threads */
    for (i= 0; i < vping_res_num; i++) {
        if (MOSAL_thread_start(&(vping_res[i].session_thread),0,start_vping,&(vping_res[i])) != 0) {
          printf("Failed creating thread for session of HCA=%s (%s).\n",vping_res[i].hca_id,strerror(errno));
          for (j= i-1; (int)j >= 0; j--) { /* Cancel all previously created threads */
            vping_res[i].exit_flag= TRUE;
            MOSAL_thread_wait_for_term(&(vping_res[i].session_thread),NULL);      
          }
          exit(1);
        }
        
    }

    /* Wait for normal threads termination */
    for (i= 0; i < vping_res_num; i++) {
        MOSAL_thread_wait_for_term(&(vping_res[i].session_thread),NULL);
    if (vping_res[i].error_counter > 0)  exit_code= 1;
    }

  }

  return exit_code;
  
}


