/*
  This software is available to you under a choice of one of two
  licenses.  You may choose to be licensed under the terms of the GNU
  General Public License (GPL) Version 2, available at
  <http://www.fsf.org/copyleft/gpl.html>, or the OpenIB.org BSD
  license, available in the LICENSE.TXT file accompanying this
  software.  These details are also available at
  <http://openib.org/license.html>.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
  BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
  ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
  CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
  SOFTWARE.

  Copyright (c) 2004 Mellanox Technologies Ltd.  All rights reserved.
*/

#ifndef H_VPING_H
#define H_VPING_H

#include <ib_defs.h>
#include <vapi_common.h>
#include <vapi.h>
#include <evapi.h>
#include <mosal.h>


/* HCA Constants */
#define HCA_ID     "InfiniHost0" 
#define GRH_LEN 40

/*#define VPING_DEF_MR_SIZE     (1UL << 11)*/

#define MAGIC_COOKIE 0xA5
#define OPCODE_ECHO  1      /* echo "rpt" times -  3rd byte = rpt cnt  */
#define OPCODE_RAND_RPT_SZ 2  /* echo "rpt" pkts with random sizes (up to sent pkt len) */
#define OPCODE_ECHO_W_RPT_ID 4 /* echo and put rpt id (num) on 4th byte of pkt */

/* DIEGO */
#if defined( __LINUX__ ) 
#define VPING_TCPD_DEFSOCKETPORT 6500

#define VPING_TCPD_MAX_MSG_LEN 0x200

#define VPING_TCPD_ADDRESSSIZE 16 /* L0x0000Q0x000000 */
#define VPING_TCPD_ADDRESSPERMSG (VPING_TCPD_MAX_MSG_LEN / VPING_TCPD_ADDRESSSIZE)

#define VPING_TCPD_RESULTSIZE 5 /* R0x00 */
#define VPING_TCPD_RESULTPERMSG (VPING_TCPD_MAX_MSG_LEN / VPING_TCPD_RESULTSIZE)

#define VPING_TCPD_RESULT_OK      0x10
#define VPING_TCPD_RESULT_BADDEST 0x11
#define VPING_TCPD_RESULT_FAILURE 0x12
#define VPING_TCPD_RESULT_NORES   0xFF

#define VPING_TCPD_REMOTE_COMMAND_DONE 0x40
#define VPING_TCPD_REMOTE_COMMAND_PING 0x41

#endif
/* DIEGO END */




/* UD Constants */
#define VPING_Q_KEY            0xff
#define VPING_FLOW_LABEL          0
#define VPING_HOP_LIMIT          63
#define VPING_SGID_INDEX          0
#define VPING_SL				  0
#define VPING_SRC_PATH_BITS       0
#define VPING_STATIC_RATE         0
#define VPING_TRAFFIC_CLASS       0


/* Running options */
#define VPING_OPTS_DAEMON  (1UL<<0)
#define VPING_OPTS_CLIENT  (1UL<<1)
#define VPING_OPTS_RAND_RPT  (1UL<<2) /* OPCODE_RAND_RPT_SZ */
#define VPING_OPTS_RPT_W_ID  (1UL<<3) /* OPCODE_ECHO_W_RPT_ID */
#define VPING_OPTS_GRH      (1UL<<4)
#define VPING_OPTS_EVENT    (1UL<<5)  /* Use completion notification instead of plain polling */
#define VPING_OPTS_INLINE_SEND (1UL<<7)  /* Use EVAPI_post_inline_sr */
#define VPING_OPTS_BLOCK_POLL (1UL<<8)  /* Use blocking polling on completion (using events) */
#define VPING_OPTS_CHAIN_SEND (1UL<<9) /* Use EVAPI_post_sr_list() for daemon side's response */
#define VPING_OPTS_NEVENT    (1UL<<10)  /* Use N-completion notification instead of plain polling */
#define VPING_OPTS_KILL    (1UL<<11)    /* Send a message that kills the daemon */

#if defined( __LINUX__ ) 
/* DIEGO */
#define VPING_OPTS_MANAGER  (1UL<<12)
#endif

#define VPING_MAX_MTU		2048
#define MAX_PAYLOAD_SZ  VPING_MAX_MTU
#define MIN_PAYLOAD_SZ  3 /* size of vping wire protocol: magic cookie + command + repeat */

#if defined( __LINUX__ ) 
/*  DIEGO */
#define MAX_OUTSTANDING_PINGS  512    /* Max. outstanding pings in daemon RQ */
#define MAX_CONNECTIVITY_POLL_CNT    300
#else
#define MAX_OUTSTANDING_PINGS  127    /* Max. outstanding pings in daemon RQ */
#endif

#define MAX_RPT         255

/* DIEGO - these defines were in vping.c. Moved them here so that they can be used from vpingm.c */
#define VPING_MAX_AVS       1000
#define VPING_PREALLOC_AVS  500

#define MAX_POLL_CNT    2000
#define POLL_SLEEP      1 /* for usleep */

#define POLL_BLOCK_TIMEOUT 10000000 /*usecs */
#define GID_FMT "%02u.%02u.%02u.%02u.%02u.%02u.%02u.%02u.%02u.%02u.%02u.%02u.%02u.%02u.%02u.%02u"
#define GID_FMT_DATA(gid) \
      gid[0],gid[1],gid[2],gid[3],gid[4],gid[5],gid[6],gid[7],      \
      gid[8],gid[9],gid[10],gid[11],gid[12],gid[13],gid[14],gid[15]
#define MAX_MCG 32  /* Max. multicast groups */

/* Verbs Ping resources handles */
typedef struct {
  VAPI_hca_id_t     hca_id;                  
  VAPI_hca_hndl_t     hca_hndl; 
  IB_port_t         port; /* HCA port to bind QP to */
  EVAPI_async_handler_hndl_t async_hndl;
  VAPI_qp_hndl_t      qp_hndl;             
  VAPI_mr_hndl_t      mr_hndl;
  VAPI_cq_hndl_t      rq_cq_hndl, sq_cq_hndl;
#if !defined(MT_KERNEL) 
  MOSAL_semaphore_t   rq_sem,sq_sem; /* For event notification */
  EVAPI_compl_handler_hndl_t rq_cq_eventh,sq_cq_eventh;
#endif
  VAPI_pd_hndl_t      pd_hndl;
  VAPI_ud_av_hndl_t   av_hndl;
  VAPI_ud_av_hndl_t   *prealloc_av_hndls;
  u_int32_t           vping_max_avs;
  u_int32_t           vping_prealloc_avs;

  IB_rkey_t       l_key;
  IB_lid_t        dlid;
  IB_lid_t                slid;
  IB_gid_t        dgid;   /* DGID for GRH header option */
  MT_size_t num_of_mcg;   /* number of multicast groups to attach to */
  IB_gid_t  mcgid[MAX_MCG];

  VAPI_qp_num_t     remote_qpnum;
  VAPI_qp_num_t           local_qpnum;

  VAPI_qkey_t       qkey;

  void            *buf_ptr;
  u_int32_t buf_size;
  u_int32_t       opts;               /* Running options     */

  u_int32_t       rpt;        /* Echo repeat counter */

  u_int32_t       num_of_buffs;
  u_int32_t       intervalsecs;   /* interval between pings in seconds, 0 for no wait */
  u_int32_t       npackets;     /* Number of pings to send                          */

  u_int32_t       payload_size;   /* Size for each packet */

  volatile u_int32_t		  error_counter;
  MOSAL_thread_t  session_thread;
  MT_bool exit_flag;  /* Set to true by signal handler to cause thread to exit properly */
  u_int32_t vping_res_num;

#if defined( __LINUX__ ) 
  /* DIEGO */
  int socket;
  char remote_ip[100];
  int socket_port;
  MOSAL_thread_t  tcpdaemon_thread;
  char tcpdret;
  int simultpingers;
  MT_bool cluster_conn_test_fl;  /* Set to true for use by vping daemon if was started to test
                                  * connectivity of a cluster
                                  */
#endif
} vping_res_t;




VAPI_ret_t  vping_client(vping_res_t  *res);
VAPI_ret_t  vping_daemon(vping_res_t  *res);

#if defined( __LINUX__ ) 
/* DIEGO */
VAPI_ret_t  vping_manager(vping_res_t*);
void* start_vpingtcpdaemon(void *);
void* start_vping(void *); /* need the proto to call it from the tcp daemon in vpingm.c */
#endif

#endif
