/*
  This software is available to you under a choice of one of two
  licenses.  You may choose to be licensed under the terms of the GNU
  General Public License (GPL) Version 2, available at
  <http://www.fsf.org/copyleft/gpl.html>, or the OpenIB.org BSD
  license, available in the LICENSE.TXT file accompanying this
  software.  These details are also available at
  <http://openib.org/license.html>.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
  BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
  ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
  CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
  SOFTWARE.

  Copyright (c) 2004 Mellanox Technologies Ltd.  All rights reserved.
*/



  #include <stdio.h>
  #include <stdlib.h>
  #include <string.h>
  #include <unistd.h>


#include <mosal.h>

#include "vping.h"

#include <mtperf.h>
MTPERF_NEW_SEGMENT(VAPI_poll_cq,3000);
MTPERF_EXTERN_SEGMENT(HHUL_poll4cqe);
MTPERF_EXTERN_SEGMENT(free_cqe);

VAPI_ret_t   vping_qp2rtr(vping_res_t *res);
VAPI_ret_t   vping_qp2rts(vping_res_t *res);
#ifdef __WIN__
call_result_t MOSAL_sem_acq_to(MOSAL_semaphore_t *sem_p, MT_size_t micro_sec);
#endif

#if 0
static MT_bool check_buf_data(void* buf,u_int32_t size, u_int32_t cnt)
{
  void* cur_buf= buf;
  u_int32_t i;
  MT_bool err_pkt= FALSE;

  /*MTL_ERROR1("Checking Magic-cookie in responses:\n");*/
  for (i= 0; i < cnt; i++ , (MT_virt_addr_t)cur_buf+= size) {
    if (((u_int8_t*)cur_buf)[GRH_LEN] != MAGIC_COOKIE) {
      MTL_ERROR1("\n pkt %d has no magic cookie.\n",i);
      err_pkt= TRUE;
    }
  }
  MTL_ERROR1("\n");
  return err_pkt;
}
#endif
/* TRUE for error */
static MT_bool check_buf_cookie_and_id(void* buf,u_int32_t size, u_int32_t cnt)
{
  u_int8_t* cur_buf= buf;
  u_int32_t i,j,num;
  MT_bool err_pkt= FALSE;

  /*MTL_ERROR1("Checking Magic-cookie and IDs in responses:\n");*/
  for (i= 0; i < cnt; i++ , cur_buf+= size) {
    
    if (cur_buf[GRH_LEN] != MAGIC_COOKIE) {  /* check magic cookie */
      MTL_ERROR1("- pkt %d has no magic cookie \n",i);
      err_pkt= TRUE;
    }
    
    for (j= GRH_LEN+3, num= i; j < size; j++, num++) {
      if (cur_buf[j] != (num & 0xFF)) {  /* Pkt ID based pattern */
        MTL_ERROR1("- pkt %u byte %u is %u instead of %u\n",i,j-GRH_LEN,cur_buf[j],(num & 0xFF));
        err_pkt= TRUE;
      }
    } /* j */
  } /* i */
  return err_pkt;
}




VAPI_ret_t  vping_client(vping_res_t  *res)
{
  VAPI_ret_t       ret;
  VAPI_rr_desc_t   rr_desc_array[MAX_RPT];
  VAPI_sr_desc_t   sq_desc;
  VAPI_ud_av_t     av;
  u_int32_t      i, j, k;
  VAPI_wc_desc_t   sq_wc,rq_wc; 
  VAPI_sg_lst_entry_t rr_sg[MAX_RPT], sq_sg;
  u_int32_t id= 0, max_poll_count = MAX_POLL_CNT;
  MT_bool err_pkt= FALSE;


  MTL_ERROR1("vping: client session %u with %d pings of pkt_size=%d rpt=%d, local_qp=0x%x.\n", 
    res->vping_res_num, res->npackets, res->payload_size, res->rpt, res->local_qpnum);
  /* Create Address Vector Handle for remote channel */
  av.dgid[0] = 0;
  av.dlid = res->dlid;
  av.port = res->port;
  /* GRH ? */
  if (res->opts & VPING_OPTS_GRH) {
    av.grh_flag = TRUE;
    av.flow_label = VPING_FLOW_LABEL;
    av.hop_limit = VPING_HOP_LIMIT;
    av.sgid_index = VPING_SGID_INDEX;
    memcpy(av.dgid,res->dgid,sizeof(IB_gid_t)); 

  } else {
    av.grh_flag = FALSE;
  }
  av.sl = VPING_SL;
  av.src_path_bits = VPING_SRC_PATH_BITS;
  av.static_rate = VPING_STATIC_RATE;
  av.traffic_class = VPING_TRAFFIC_CLASS;

  ret = VAPI_create_addr_hndl(res->hca_hndl, res->pd_hndl, &av, &(res->av_hndl)); VAPI_CHECK_RET;

  /* 
   * Prepare descriptor for send queue
   *
   *
   */

  sq_desc.set_se= FALSE;
  sq_desc.fence= FALSE;
  /* Generate completion on each send request */
  sq_desc.comp_type  = VAPI_SIGNALED;
  /* sq_desc.opcode will be modified in the loop (in the last send if needed)*/
  sq_desc.opcode = VAPI_SEND;
  sq_desc.remote_qkey = res->qkey;
  sq_desc.remote_qp = res->remote_qpnum;
  sq_desc.remote_ah= res->av_hndl;

  /* Data for the ping */
  sq_sg.len  = res->payload_size;
  sq_sg.lkey = res->l_key;
  sq_sg.addr = (VAPI_virt_addr_t)(MT_virt_addr_t)res->buf_ptr;
  sq_desc.sg_lst_len = res->payload_size > 0 ? 1 : 0;
  sq_desc.sg_lst_p = &sq_sg;

  /* Add special data in payload */
  if (res->payload_size >= MIN_PAYLOAD_SZ) {
      ((u_int8_t*)(MT_virt_addr_t)sq_sg.addr)[0]= MAGIC_COOKIE;
       /* opcode */
      if (res->opts & VPING_OPTS_RPT_W_ID) {
        ((u_int8_t*)(MT_virt_addr_t)sq_sg.addr)[1]= OPCODE_ECHO_W_RPT_ID;
      } else if (res->opts & VPING_OPTS_RAND_RPT) {
        ((u_int8_t*)(MT_virt_addr_t)sq_sg.addr)[1]= OPCODE_RAND_RPT_SZ;
      } else {
        ((u_int8_t*)(MT_virt_addr_t)sq_sg.addr)[1]= OPCODE_ECHO;
      }
      /* repeat cnt */
      ((u_int8_t*)(MT_virt_addr_t)sq_sg.addr)[2]= (u_int8_t)res->rpt; 
  } else {
    res->rpt=1;
  }

  /*
   * Prepare receive descriptor 
   *
   */
  for (i = 0 ; i < res->rpt; i++) {
    rr_desc_array[i].opcode = VAPI_RECEIVE;
    rr_desc_array[i].sg_lst_len = 1;

    rr_sg[i].len =  res->payload_size + GRH_LEN;  
    rr_sg[i].lkey = res->l_key;
    rr_desc_array[i].sg_lst_p = rr_sg+i;
    rr_desc_array[i].sg_lst_p->addr = (VAPI_virt_addr_t)(MT_virt_addr_t)
      ((char*)res->buf_ptr + (i + 1)*(res->payload_size+GRH_LEN));
    memset((void*)(MT_virt_addr_t)rr_desc_array[i].sg_lst_p->addr,0,res->payload_size+GRH_LEN);
    rr_desc_array[i].id = id++;
  }

  for (j = 0; j < res->npackets; j++) { /* "-c" */
    ret= EVAPI_post_rr_list(res->hca_hndl,res->qp_hndl,res->rpt,rr_desc_array);
    if (ret != VAPI_OK) {
      MTL_ERROR1("Failed posting to RQ (%s). Aborting.\n",VAPI_strerror_sym(ret));
      goto err_exit;
    }
    
    if (j == 0) { /* only once (used to test posting to RQ in INIT state) */
      ret= vping_qp2rtr(res); VAPI_CHECK_RET;
      ret= vping_qp2rts(res); VAPI_CHECK_RET;
    }

    sq_desc.id = j;
    /* if we want to send the "vping of death", post SEND with immediate */
    if ((res->opts & VPING_OPTS_KILL) && (j == res->npackets - 1)) { 
        sq_desc.opcode = VAPI_SEND_WITH_IMM;
    }

    if (res->opts & VPING_OPTS_INLINE_SEND) { 
      ret= EVAPI_post_inline_sr(res->hca_hndl, res->qp_hndl, &sq_desc);
    } else {
      MTL_DEBUG1(MT_FLFMT("%s: session %u: Posting sr"),
		 __func__, res->vping_res_num);
      ret= VAPI_post_sr(res->hca_hndl, res->qp_hndl, &sq_desc);
    }
    if (ret != VAPI_OK) {
      MTL_ERROR1("VAPI_post_sr failed (%s).\n",VAPI_strerror_sym(ret));
      goto err_exit;
    }

    while (1) { /* Wait for SQ completion */

      if (res->opts & VPING_OPTS_BLOCK_POLL) {
        ret= EVAPI_poll_cq_block(res->hca_hndl,res->sq_cq_hndl,POLL_BLOCK_TIMEOUT,&sq_wc);
      }else
      {
            MTPERF_TIME_START(VAPI_poll_cq);
            ret = VAPI_poll_cq(res->hca_hndl, res->sq_cq_hndl, &sq_wc); /*VAPI_CHECK;*/
            MTPERF_TIME_END(VAPI_poll_cq);
      }
      if (res->exit_flag)  {
        ret= VAPI_EGEN;
        goto err_exit;
      }
      if (ret == VAPI_CQ_EMPTY) {
			#ifdef __WIN__      		
			Sleep(1);
			#else
			usleep(1);
			#endif
        continue;

      } else if (ret != (VAPI_OK)) {
        MTL_ERROR1("vping: error while polling completion queue\n");
        goto err_exit;
      }

      break;
    }
    if (sq_wc.status != VAPI_SUCCESS) {
      MTL_ERROR1("vping: Completion with error on send queue (syndrome=%s[vendor_syndrome=0x%X] , opcode=%d=%s)\n",
                 VAPI_wc_status_sym(sq_wc.status),sq_wc.vendor_err_syndrome,sq_wc.opcode,VAPI_cqe_opcode_sym(sq_wc.opcode));
      ret= VAPI_EGEN;
		printf("Press any key to resume\n"); getchar();	
      goto err_exit;
    }
    MTL_DEBUG1(MT_FLFMT("%s: session %u: Got SQ completion"),
	       __func__, res->vping_res_num);

    for (i = 0 ; i < res->rpt; i++) {   /* Wait for all RQ completions */

#if defined( __LINUX__ ) 
if (res->cluster_conn_test_fl == TRUE) {
    max_poll_count = MAX_CONNECTIVITY_POLL_CNT; 
} 
#endif
      for (k= 0; k < max_poll_count; k++) { /* For other methods we may need more than one trial */
          if (res->opts & VPING_OPTS_BLOCK_POLL) {
              ret= EVAPI_poll_cq_block(res->hca_hndl,res->rq_cq_hndl,POLL_BLOCK_TIMEOUT,&rq_wc);
              if (ret != VAPI_CQ_EMPTY) break;
          } else { /* else .. - NOT poll block mode*/
              MTPERF_TIME_START(VAPI_poll_cq);
              ret = VAPI_poll_cq(res->hca_hndl, res->rq_cq_hndl, &rq_wc); 
              MTPERF_TIME_END(VAPI_poll_cq);
              if (ret != VAPI_CQ_EMPTY)  break;
              if ((res->opts & VPING_OPTS_NEVENT) && ((i > 0) || (k > 0))) {
                printf("Got false event on ARM-N\n");
              }
              if ((res->opts & VPING_OPTS_EVENT) || (res->opts & VPING_OPTS_NEVENT)){   /* event mode */
                  ret= (res->opts & VPING_OPTS_EVENT) ? 
                    VAPI_req_comp_notif(res->hca_hndl,res->rq_cq_hndl,VAPI_NEXT_COMP):
                    EVAPI_req_ncomp_notif(res->hca_hndl,res->rq_cq_hndl,res->rpt - i); 
                    if (ret != VAPI_OK) {
                      MTL_ERROR1(MT_FLFMT("Failed (E)VAPI_req_ncomp_notif (%s)"), VAPI_strerror_sym(ret));
                      goto err_exit;
                    }
                  MTPERF_TIME_START(VAPI_poll_cq);
                  ret = VAPI_poll_cq(res->hca_hndl, res->rq_cq_hndl, &rq_wc); /* avoid req. notif. race */
                  MTPERF_TIME_END(VAPI_poll_cq);
                  if (ret != VAPI_CQ_EMPTY)  break;
                 #ifdef __WIN__
	                  if (MOSAL_sem_acq_to(&(res->rq_sem),1000000) != MT_OK) { /* Wait for notification handler */
       	             MTL_ERROR1(MT_FLFMT("Wait on RQ CQ completion semaphore was interrupted."));
                 #else
	                  if (MOSAL_sem_acq(&(res->rq_sem),TRUE) != MT_OK) { /* Wait for notification handler */
       	             MTL_ERROR4(MT_FLFMT("Wait on RQ CQ completion semaphore was interrupted."));
                  #endif
                  }
              } else { /* polling mode */
			#ifdef __WIN__      		
			Sleep(1);
			#else
			usleep(1);
			#endif
            }
          }
          if (res->exit_flag == TRUE)  {
              ret= VAPI_EINTR; /* poll exit flag */
              goto err_exit;
          }
          if ((k & 15) == 15)  {printf("."); fflush(stdout);}
          }
        
          if ((ret != VAPI_OK) && (ret != VAPI_CQ_EMPTY)) {
            MTL_ERROR1(MT_FLFMT("Error while polling RQ completion queue (%s)\n"),VAPI_strerror_sym(ret));
            goto err_exit;
          }
    
          if (ret == VAPI_CQ_EMPTY) {  /* Exhausted all retries of polling (max_poll_count) */
            MTPERF_REPORT_PRINTF(VAPI_poll_cq);
            MTPERF_REPORT_PRINTF(HHUL_poll4cqe);
            MTPERF_REPORT_PRINTF(free_cqe);
        printf("\nFailed getting rpt #%d of cycle %d.\n",i,j);
        if (err_pkt)  res->error_counter++;
        ret= VAPI_EGEN;
        goto err_exit;
      }
      
    } /* for i */
    
    MTPERF_REPORT_PRINTF(VAPI_poll_cq);
    MTPERF_REPORT_PRINTF(HHUL_poll4cqe);
    MTPERF_REPORT_PRINTF(free_cqe);
    
    /* check WQEs data if has stamps */
    if (res->opts & VPING_OPTS_RPT_W_ID) {
      err_pkt= check_buf_cookie_and_id((u_int8_t*)((MT_virt_addr_t)res->buf_ptr+res->payload_size+GRH_LEN),
        res->payload_size+GRH_LEN,res->rpt);
      if (err_pkt)  res->error_counter++;
    }


  } /* npackets */
  
  VAPI_destroy_addr_hndl(res->hca_hndl,res->av_hndl);
  
  MTL_ERROR1("VPING done - OK.\n");
  return(VAPI_OK);

  err_exit:
    VAPI_destroy_addr_hndl(res->hca_hndl,res->av_hndl);
    return ret;
}
