/*
  This software is available to you under a choice of one of two
  licenses.  You may choose to be licensed under the terms of the GNU
  General Public License (GPL) Version 2, available at
  <http://www.fsf.org/copyleft/gpl.html>, or the OpenIB.org BSD
  license, available in the LICENSE.TXT file accompanying this
  software.  These details are also available at
  <http://openib.org/license.html>.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
  BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
  ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
  CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
  SOFTWARE.

  Copyright (c) 2004 Mellanox Technologies Ltd.  All rights reserved.
*/

#define C_VPINGD_C

  #include <stdio.h>
  #include <stdlib.h>
  #include <string.h>
  #include <unistd.h>
  #include <time.h>


#include "vping.h"
#include <mpga/mpga.h>

#define BUF_SIZE (VPING_MAX_MTU+GRH_LEN)

#include <mtperf.h>
MTPERF_NEW_SEGMENT(VAPI_post_sr,10000);
MTPERF_NEW_SEGMENT(EVAPI_post_inline_sr,10000);
MTPERF_EXTERN_SEGMENT(HHUL_post_send_req);   
MTPERF_EXTERN_SEGMENT(HHUL_post_inline_send_req);   
MTPERF_EXTERN_SEGMENT(THH_uar_sendq_dbell);
MTPERF_EXTERN_SEGMENT(WQE_build_send);
MTPERF_EXTERN_SEGMENT(SQ_WQE_copy);


VAPI_ret_t   vping_qp2rtr(vping_res_t *res);
VAPI_ret_t   vping_qp2rts(vping_res_t *res);
	
/************************
 * Function: vping_post_recv_bufs
 * 
 *  Post num_o_bufs buffers of given size at buf_array
 *  to given vres QP RQ. 
 *  WQEs will bare the IDs from start_id to start+id+num_o_bufs+1
 ************************/
static int vping_post_recv_bufs(vping_res_t* vres, void *buf_array, 
                                u_int32_t num_o_bufs,u_int32_t size, VAPI_wr_id_t start_id) 
{
  unsigned int i;
  void* cur_buf;
  VAPI_rr_desc_t rr;
  VAPI_sg_lst_entry_t sg_entry;
  VAPI_ret_t ret;

  rr.opcode= VAPI_RECEIVE;
  rr.comp_type= VAPI_SIGNALED;  /* All with CQE */
  rr.sg_lst_len= 1; /* single buffers */
  rr.sg_lst_p= &sg_entry;
  sg_entry.lkey= vres->l_key;

  for (i= 0, cur_buf= buf_array ; i < num_o_bufs  ; 
       i++ , cur_buf= (void*)((MT_virt_addr_t)cur_buf + size) ) {
    rr.id= start_id+i; /* WQE id used is the index to buffers ptr array */
    sg_entry.addr= (VAPI_virt_addr_t)(MT_virt_addr_t)cur_buf;
    sg_entry.len= size;
    memset(cur_buf,(int)((rr.id+1) & 0xFF),size);
    ret= VAPI_post_rr(vres->hca_hndl,vres->qp_hndl,&rr);
    if (ret != VAPI_OK) {
      MTL_ERROR1("%s: failed posting RQ WQE (%s)\n", __func__,VAPI_strerror_sym(ret));
      return i;
    }
    MTL_DEBUG4("%s: posted buf at %p\n", __func__,cur_buf);
  }

  return i; /* num of buffers posted */
}

static VAPI_ret_t get_next_rq_cqe_block(vping_res_t* vres,VAPI_wc_desc_t *comp_desc_p, void **comp_buf_p, MT_bool *ping_of_death_p)
{
    VAPI_ret_t ret= VAPI_CQ_EMPTY;

    while (ret == VAPI_CQ_EMPTY) {
        ret= EVAPI_poll_cq_block(vres->hca_hndl,vres->rq_cq_hndl,0,comp_desc_p);
        if (vres->exit_flag == TRUE)  return VAPI_EGEN; /* poll exit flag */
        if (ret == VAPI_CQ_EMPTY) {
            continue;
        }
        if (ret != VAPI_OK) {
            MTL_ERROR1("%s: VAPI_poll_cq failed with ret=%s\n", __func__,VAPI_strerror_sym(ret));
            return ret;
        }

        *comp_buf_p= (void*)(((MT_virt_addr_t)vres->buf_ptr) + (u_int32_t)(comp_desc_p->id * BUF_SIZE));  /* pointer to completed buffer */
        MTL_DEBUG4("%s: received buf at %p\n", __func__,*comp_buf_p);

        if (comp_desc_p->status != VAPI_SUCCESS) {
          MTL_ERROR1("%s: completion error (%d) detected\n", __func__,comp_desc_p->status);
          vping_post_recv_bufs(vres,*comp_buf_p,1,BUF_SIZE,comp_desc_p->id); /* post buffer back to RQ */
          continue;
        }
    
        if (comp_desc_p->remote_node_addr.type != VAPI_RNA_UD) {
          MTL_ERROR1("%s: remote node addr type is illegal (%d)\n", __func__,
                     comp_desc_p->remote_node_addr.type);
          vping_post_recv_bufs(vres,*comp_buf_p,1,BUF_SIZE,comp_desc_p->id); /* post buffer back to RQ */
          continue;
        }

        /* we got the ping of death (completion with immediate data) */
        if (comp_desc_p->imm_data_valid) {
            MTL_ERROR1("%s: got vping of death\n", __func__);
            *ping_of_death_p = TRUE;
        } else {
            *ping_of_death_p = FALSE;
        }
    }
    return VAPI_OK; /* got something useful */
}


static VAPI_ret_t get_next_rq_cqe(vping_res_t* vres,VAPI_wc_desc_t *comp_desc_p, void **comp_buf_p, MT_bool *ping_of_death_p)
{
  VAPI_ret_t ret;

  MTL_DEBUG1(MT_FLFMT("%s: Entering"), __func__);
  /* until gets something to pong or real error */
  while (1) {
    ret = VAPI_CQ_EMPTY;

    /* wait for something to arrive */
    ret= VAPI_poll_cq(vres->hca_hndl, vres->rq_cq_hndl,comp_desc_p);
    if (vres->exit_flag == TRUE)  return VAPI_EGEN; /* poll exit flag */
    while (ret == VAPI_CQ_EMPTY) {
      if (vres->opts & VPING_OPTS_EVENT) {
        VAPI_req_comp_notif(vres->hca_hndl,vres->rq_cq_hndl,VAPI_NEXT_COMP);
        if (MOSAL_sem_acq(&(vres->rq_sem),TRUE) != MT_OK) { /* Wait for notification handler */
             MTL_ERROR4(MT_FLFMT("Failed acquiring RQ CQ completion semaphore."));
        }
      }
      if (vres->exit_flag == TRUE)  return VAPI_EGEN; /* poll exit flag */
      ret= VAPI_poll_cq(vres->hca_hndl, vres->rq_cq_hndl,comp_desc_p);
      if (ret != VAPI_CQ_EMPTY)  break;
    }

    if (ret != VAPI_OK) {
      MTL_ERROR1("%s: VAPI_poll_cq failed with ret=%s\n", __func__,
                 VAPI_strerror_sym(ret));
      return ret;
    }

    *comp_buf_p= (void*)(((MT_virt_addr_t)vres->buf_ptr) + (u_int32_t)(comp_desc_p->id * BUF_SIZE));  /* pointer to completed buffer */
    MTL_DEBUG4("%s: received buf at %p\n", __func__,*comp_buf_p);

    if (comp_desc_p->status != VAPI_SUCCESS) {
      MTL_ERROR1("%s: completion error (%d) detected\n", __func__,comp_desc_p->status);
      vping_post_recv_bufs(vres,*comp_buf_p,1,BUF_SIZE,comp_desc_p->id); /* post buffer back to RQ */
      continue;
    }

    if (comp_desc_p->remote_node_addr.type != VAPI_RNA_UD) {
      MTL_ERROR1("%s: remote node addr type is illegal (%d)\n", __func__,
                 comp_desc_p->remote_node_addr.type);
      vping_post_recv_bufs(vres,*comp_buf_p,1,BUF_SIZE,comp_desc_p->id); /* post buffer back to RQ */
      continue;
    }

    /* we got the ping of death (completion with immediate data) */
    if (comp_desc_p->imm_data_valid) {
        MTL_ERROR1("%s: got vping of death\n", __func__);
        *ping_of_death_p = TRUE;
    } else {
        *ping_of_death_p = FALSE;
    }

    return VAPI_OK; /* got something useful */
  }
}

static VAPI_ret_t get_next_sq_cqe_block(vping_res_t* vres,VAPI_wc_desc_t *comp_desc_p)
{
  VAPI_ret_t ret = VAPI_CQ_EMPTY;
  u_int32_t poll_cnt= 0;

  /* wait for something to arrive */
  while ((ret == VAPI_CQ_EMPTY) && (poll_cnt < MAX_POLL_CNT)) {
    ret= EVAPI_poll_cq_block(vres->hca_hndl,vres->sq_cq_hndl,POLL_BLOCK_TIMEOUT,comp_desc_p);
    poll_cnt++;
  }

  if ((poll_cnt == MAX_POLL_CNT) && (ret == VAPI_CQ_EMPTY)) {
    MTL_ERROR1("%s: Failed to get completion after %d polls on sq_cq\n", __func__,MAX_POLL_CNT);
    return VAPI_CQ_EMPTY;
  }
  
  if (ret != VAPI_OK) {
    MTL_ERROR1("%s: VAPI_poll_cq failed with ret=%s on sq_cq\n", __func__,VAPI_strerror_sym(ret));
    return ret;
  }

  if (comp_desc_p->status != VAPI_SUCCESS) {
    MTL_ERROR1("%s: completion error (%d) detected\n", __func__,comp_desc_p->status);
  }

  return VAPI_OK;
}


static VAPI_ret_t get_next_sq_cqe(vping_res_t* vres,VAPI_wc_desc_t *comp_desc_p)
{
  VAPI_ret_t ret = VAPI_CQ_EMPTY;
  u_int32_t poll_cnt= 0;

  /* wait for something to arrive */
  while ((ret == VAPI_CQ_EMPTY) && (poll_cnt < MAX_POLL_CNT)) {
    ret= VAPI_poll_cq(vres->hca_hndl, vres->sq_cq_hndl, comp_desc_p);
	#ifdef __WIN__      		
		Sleep(POLL_SLEEP);
	#else
	    usleep(POLL_SLEEP);
	#endif
    poll_cnt++;
  }

  if ((poll_cnt == MAX_POLL_CNT) && (ret == VAPI_CQ_EMPTY)) {
    MTL_ERROR1("%s: Failed to get completion after %d polls on sq_cq\n", __func__,MAX_POLL_CNT);
    return VAPI_CQ_EMPTY;
  }
  
  if (ret != VAPI_OK) {
    MTL_ERROR1("%s: VAPI_poll_cq failed with ret=%s on sq_cq\n", __func__,VAPI_strerror_sym(ret));
    return ret;
  }

  if (comp_desc_p->status != VAPI_SUCCESS) {
    MTL_ERROR1("%s: completion error (%d) detected\n", __func__,comp_desc_p->status);
  }

#if 0
  /* verify CQ is empty */
  ret= VAPI_poll_cq(vres->hca_hndl, vres->sq_cq_hndl, comp_desc_p);
  if (ret == VAPI_OK) {
    printf("!!!!!!!!!!!!!! I found unexpected CQE !!\n");
  }
#endif
  
  return VAPI_OK;
}


static VAPI_ret_t create_av(vping_res_t* vres,
  VAPI_remote_node_addr_t *dest, u_int8_t* grh_p, VAPI_ud_av_hndl_t *avh_p)
{
  VAPI_ud_av_t av;
  IB_GRH_st grh_data;
  VAPI_ret_t ret;

  av.dlid= dest->slid;
  av.port= vres->port;
  av.sl= dest->sl;
  av.src_path_bits= dest->ee_dlid.dst_path_bits;
  av.static_rate = VPING_STATIC_RATE;
  /* GRH ? */
  av.grh_flag= (grh_p != NULL); 
  if (av.grh_flag) {    /* set global routing data from received GRH */
    extract_GRH(&grh_data,(u_int16_t**)&grh_p);
#ifdef MAX_DEBUG
    grh_p-= IB_GRH_LEN; /* reverse pointer to GRH begin (after moved by extract) */
    MTL_DEBUG4("Got a packet with GRH:\n"
      "%08X %08X %08X %08X %08X %08X %08X %08X %08X %08X\n"
      "From GID:" GID_FMT " )\n"
      "To   GID:" GID_FMT " )\n"
      "flow_label=%d hop_limit=%d Tclass=%d \n",
      ((u_int32_t*)grh_p)[0],((u_int32_t*)grh_p)[1],
      ((u_int32_t*)grh_p)[2],((u_int32_t*)grh_p)[3],
      ((u_int32_t*)grh_p)[4],((u_int32_t*)grh_p)[5],
      ((u_int32_t*)grh_p)[6],((u_int32_t*)grh_p)[7],
      ((u_int32_t*)grh_p)[8],((u_int32_t*)grh_p)[9],
      GID_FMT_DATA(grh_data.SGID),GID_FMT_DATA(grh_data.DGID),
      grh_data.FlowLabel,grh_data.HopLmt,grh_data.TClass);
#endif
    memcpy(av.dgid,grh_data.SGID,sizeof(IB_gid_t));   /* destination is the source of rx.pkt */
    av.sgid_index= 0; /* TBD: lookup index for DGID in incoming pkt (unless multicast) */
    av.flow_label= grh_data.FlowLabel;
    av.hop_limit= grh_data.HopLmt;
    av.traffic_class= grh_data.TClass;
  }
  ret= VAPI_create_addr_hndl(vres->hca_hndl,vres->pd_hndl,&av,avh_p);
  if (ret != VAPI_OK) {
    MTL_ERROR1("%s: failed VAPI_create_addr_hndl (%s)\n", __func__,
               VAPI_strerror_sym(ret));
    return ret;
  }
  return VAPI_OK;
}

static VAPI_ret_t send_buf(vping_res_t* vres, VAPI_wr_id_t id, void* buf, u_int32_t size, 
  u_int8_t opcode, u_int32_t rpt_cnt, VAPI_ud_av_hndl_t avh, VAPI_qp_num_t dest_qp)
{
  VAPI_sr_desc_t sr[MAX_RPT];
  VAPI_sg_lst_entry_t sg_entry[MAX_RPT];
  u_int32_t i,j,num;
  VAPI_ret_t ret;
  u_int32_t sizes[MAX_RPT];
  u_int8_t *cpy_buf;

  if (rpt_cnt == 0)  rpt_cnt= 1;

  sizes[0]= size;
  if (opcode == OPCODE_RAND_RPT_SZ)  {
    for (i= 1; i < rpt_cnt; i++) {
      sizes[i]= 1+ (( (rand()>>8) * (size-1) )/(RAND_MAX>>8));
    }
  }


  /* copy buffers for different IDs */
  cpy_buf= (void*)((MT_virt_addr_t)vres->buf_ptr + ((MAX_OUTSTANDING_PINGS + 1) * BUF_SIZE));
  if (opcode == OPCODE_ECHO_W_RPT_ID) {
    MTL_DEBUG1(MT_FLFMT("%s: ECHO_W_RPT_ID sending %d times pkt_len=%d to qp 0x%x"),
	        __func__,rpt_cnt, size, dest_qp);
    for (i= 0; i < rpt_cnt; i++) {
      memcpy((void*)cpy_buf,buf,3);
      for (j= 3, num= i; j < size; j++, num++) {
        cpy_buf[j]= (num & 0xFF);  /* Pkt ID based pattern */
      }
      cpy_buf += BUF_SIZE;  /* next buf */
    }
    cpy_buf= (void*)((MT_virt_addr_t)vres->buf_ptr + ((MAX_OUTSTANDING_PINGS + 1) * BUF_SIZE));
#if 0
    /* verify ? */
    for (i= 0; i < rpt_cnt; i++) {
      if (memcmp((void*)cpy_buf,buf,3) != 0) 
        MTL_ERROR1("pkt %d header wrong.\n",i);
      for (j= 3, num= i; j < size; j++, num++) {
        if (cpy_buf[j] != (num & 0xFF)) {
          MTL_ERROR1("pkt %d data %d wrong.\n",i,j);  
        	}
      }
      cpy_buf += BUF_SIZE;  /* next buf */
    }
    cpy_buf= (void*)((MT_virt_addr_t)vres->buf_ptr + ((MAX_OUTSTANDING_PINGS + 1) * BUF_SIZE));
#endif
  } else {
      MTL_DEBUG1(MT_FLFMT("%s: ECHO sending %d times pkt_len=%d to qp 0x%x"), 
		 __func__,rpt_cnt,size, dest_qp);
  }

  for (i= 0; i < (rpt_cnt); i++) {
    sr[i].opcode= VAPI_SEND;

    sr[i].remote_ah= avh;
    sr[i].remote_qp= dest_qp;
    sr[i].remote_qkey= vres->qkey;

    sr[i].id= id;
    sr[i].set_se= FALSE;
    sr[i].fence= FALSE;
    sr[i].comp_type= VAPI_UNSIGNALED;  /* Only last completion would be provide completion */
    sr[i].sg_lst_len= size > 0 ? 1 : 0;
    sr[i].sg_lst_p= sg_entry+i;

    sg_entry[i].addr= (VAPI_virt_addr_t)(MT_virt_addr_t)buf;
    sg_entry[i].len= size;
    sg_entry[i].lkey= vres->l_key;
    switch (opcode) {
      case OPCODE_RAND_RPT_SZ:
        sg_entry[i].len= sizes[i];
        break;
      case OPCODE_ECHO_W_RPT_ID:
        sg_entry[i].addr= (VAPI_virt_addr_t)(MT_virt_addr_t)cpy_buf;
	MTL_DEBUG1(MT_FLFMT("%s: cpy_buf=%p, i=%d"),__func__, cpy_buf,i);  
        cpy_buf += BUF_SIZE;
        break;
    default:
	MTL_DEBUG1(MT_FLFMT("%s: buf=%p, i=%d"),__func__, buf,i);  
    }
  }
  sr[rpt_cnt-1].comp_type= VAPI_SIGNALED; /* last one should be signaled */  
  
  if (vres->opts &  VPING_OPTS_CHAIN_SEND) { /* Send requests chaining */
    ret= EVAPI_post_sr_list(vres->hca_hndl,vres->qp_hndl,rpt_cnt,sr);
    if (ret != VAPI_OK) {
      MTL_ERROR1("%s: failed EVAPI_post_sr_list (%s)\n", __func__,
                 VAPI_strerror_sym(ret));
      return ret;
    }

  } else for (i= 0; i < (rpt_cnt); i++) { /* One request at a time */
      if (vres->opts & VPING_OPTS_INLINE_SEND) { 
        MTPERF_TIME_START(EVAPI_post_inline_sr);
        ret= EVAPI_post_inline_sr(vres->hca_hndl,vres->qp_hndl,sr+i);
        MTPERF_TIME_END(EVAPI_post_inline_sr);
      } else {
        MTPERF_TIME_START(VAPI_post_sr);
        ret= VAPI_post_sr(vres->hca_hndl,vres->qp_hndl,sr+i);
        MTPERF_TIME_END(VAPI_post_sr);
      }
      if (ret != VAPI_OK) {
        MTL_ERROR1("%s: failed VAPI_post_sr (%s)\n", __func__,
                   VAPI_strerror_sym(ret));
        return ret;
      }
  }
  MTPERF_REPORT_PRINTF(VAPI_post_sr);
  MTPERF_REPORT_PRINTF(EVAPI_post_inline_sr);
  MTPERF_REPORT_PRINTF(HHUL_post_send_req);
  MTPERF_REPORT_PRINTF(HHUL_post_inline_send_req);
  MTPERF_REPORT_PRINTF(THH_uar_sendq_dbell);
  MTPERF_REPORT_PRINTF(WQE_build_send);
  MTPERF_REPORT_PRINTF(SQ_WQE_copy);
  return VAPI_OK;
}




VAPI_ret_t vping_daemon(vping_res_t* vres)
{
  VAPI_wc_desc_t sq_comp_desc,rq_comp_desc;
  void* comp_buf; /* completed buffer ptr */
  u_int8_t mcookie,opcode;
  u_int32_t rpt_cnt,i;
  VAPI_ret_t ret;
  void *grh_p= NULL;
  MT_bool exit_loop = FALSE;
  VAPI_ud_av_hndl_t avh;

  srand((unsigned int)time(NULL));
  ret= vping_qp2rtr(vres); VAPI_CHECK_RET;
  ret= vping_qp2rts(vres); VAPI_CHECK_RET;
  /*post all availble buffers to RQ*/
  i= vping_post_recv_bufs(vres,vres->buf_ptr, MAX_OUTSTANDING_PINGS + 1 ,BUF_SIZE,0);
  if (i < (MAX_OUTSTANDING_PINGS + 1)) { 
    MTL_ERROR1("Failed posting to RQ (only %d WQEs posted. Aborting.\n",i);
    return VAPI_EAGAIN;
  }


  while (exit_loop == FALSE) {
    
    if (vres->opts & VPING_OPTS_BLOCK_POLL) {
       ret = get_next_rq_cqe_block(vres,&rq_comp_desc,&comp_buf, &exit_loop);
    }
    else 
    {
       ret = get_next_rq_cqe(vres,&rq_comp_desc,&comp_buf, &exit_loop);
    }
    if (ret!= VAPI_OK) {
	MTL_ERROR1(MT_FLFMT("%s: get_next_rq_cqe failed. Aborting"),__func__);
	return VAPI_EGEN; /* Fatal error or exit_flag set */
    }

    if (rq_comp_desc.byte_len >= GRH_LEN+MIN_PAYLOAD_SZ) { 
        mcookie= ((u_int8_t*)comp_buf)[GRH_LEN];
        if (mcookie != MAGIC_COOKIE) {
          MTL_ERROR2("Error in magic-cookie (0x%X).\n",mcookie);
          goto wait4next;
        }

        opcode= ((u_int8_t*)comp_buf)[GRH_LEN+1];
        rpt_cnt= ((u_int8_t*)comp_buf)[GRH_LEN+2];
        /*MTL_ERROR1("Echoing %d times (opcode=%d)...\n",rpt_cnt,opcode);*/
    } else { /* support 0 length send */
      opcode= OPCODE_ECHO;
      rpt_cnt= 1;
      printf("Got a packet of small payload (%d bytes)\n", rq_comp_desc.byte_len-GRH_LEN);
    }
    
    switch (opcode) {
      case OPCODE_ECHO_W_RPT_ID:
      case OPCODE_RAND_RPT_SZ:
      case OPCODE_ECHO:

        grh_p= rq_comp_desc.grh_flag ? comp_buf : NULL;
        /* send and wait for completion of send */
        ret= create_av(vres,&(rq_comp_desc.remote_node_addr ),grh_p, &avh);
        if (ret != VAPI_OK) {
          MTL_ERROR1("Failed creating UD AV handle. Aborting.\n");
          return ret;
        }
        ret= send_buf(vres,rq_comp_desc.id,(void*)((MT_virt_addr_t)comp_buf+GRH_LEN),rq_comp_desc.byte_len - GRH_LEN, 
                     opcode, rpt_cnt, avh, rq_comp_desc.remote_node_addr.qp_ety.qp);
        if (ret == VAPI_OK) {
            if (vres->opts & VPING_OPTS_BLOCK_POLL) {
              ret = get_next_sq_cqe_block(vres,&sq_comp_desc);
            }
            else 
            {
              ret = get_next_sq_cqe(vres,&sq_comp_desc);
            }
            VAPI_destroy_addr_hndl(vres->hca_hndl,avh);
            if (ret == VAPI_CQ_EMPTY) {
                MTL_ERROR1("Failed getting completion for SQ. Aborting.\n");
                return ret;
            }
         } else {
          MTL_ERROR1("Failed send_buf. Aborting.\n");
          return ret;
        }
        break;
      
      default:
        MTL_ERROR2("Error in opcode (%d).\n",opcode);
    }

    wait4next:
    vping_post_recv_bufs(vres,comp_buf,1,BUF_SIZE,rq_comp_desc.id); /* post buffer back to RQ */
  }


  /* We will reach here if got vping of death... */
  /* Cleanup on calling "main" */
  return VAPI_OK;
}

