/*
  This software is available to you under a choice of one of two
  licenses.  You may choose to be licensed under the terms of the GNU
  General Public License (GPL) Version 2, available at
  <http://www.fsf.org/copyleft/gpl.html>, or the OpenIB.org BSD
  license, available in the LICENSE.TXT file accompanying this
  software.  These details are also available at
  <http://openib.org/license.html>.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
  BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
  ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
  CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
  SOFTWARE.

  Copyright (c) 2004 Mellanox Technologies Ltd.  All rights reserved.
*/

#include <unistd.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include "vipkl_wrap.h"


// the name in the /dev dir this should be mknod and then open
static const char*  VIP_DEVICE_FILE_NAME = "/dev/vipkl";
static int file_desc;
static bool dev_open = FALSE;

int vip_ioctl_wrapper(VIPKL_ops_t ops,void *pi, u_int32_t pi_sz, void *po, u_int32_t po_sz )
{     
    int rc = 0;
    vip_ioctl_wrap_t vip_ioctl_wrap;

    MTL_DEBUG4(MT_FLFMT("dev_open=%d, fcn=%d, VIPKL_IOCTL_CALL=0x%x"), 
                        dev_open, ops, (int) VIPKL_IOCTL_CALL);
    
    /* fill the parameter structure */
    vip_ioctl_wrap.ops		= (int)ops;
    vip_ioctl_wrap.pi 		= pi;
    vip_ioctl_wrap.pi_sz 		= pi_sz;
    vip_ioctl_wrap.po 		= po;
    vip_ioctl_wrap.po_sz 		= po_sz;

    if(!dev_open){
        MTL_ERROR1(MT_FLFMT("%s: device file %s not open"), __func__, VIP_DEVICE_FILE_NAME);
        return VIP_ERROR;
    }

    /* issue the ioctl */
    rc = ioctl(file_desc,VIPKL_IOCTL_CALL,&vip_ioctl_wrap);

    MTL_DEBUG4(MT_FLFMT("ioctl_rc=%d, func_rc=%d"), rc, *(int*)po);
    
    return rc;
}


int vip_ioctl_open(void) 
{
    static pthread_mutex_t dev_mutex = PTHREAD_MUTEX_INITIALIZER;
    MTL_DEBUG4(MT_FLFMT("called vip_ioctl_open"));
    /* when the first time - open the device  */
    pthread_mutex_lock(&dev_mutex);
    if(!dev_open){
        file_desc = open(VIP_DEVICE_FILE_NAME,0);
        if(file_desc < 0){
            pthread_mutex_unlock(&dev_mutex);
            MTL_ERROR1("%s: can't open device file %s\n", __func__, VIP_DEVICE_FILE_NAME);
            return VIP_ERROR;
        }
        dev_open = TRUE;
    }
    pthread_mutex_unlock(&dev_mutex);

    return VIP_OK;
}

