#ifndef __mtrack_h_
#define __mtrack_h_

#if !defined(MVAPI_BUILD)

#include <memtrack.h>

#include <linux/slab.h>


	

#define kmalloc(size, flags) ({ \
	void *__memtrack_addr;                 \
                                \
	__memtrack_addr = kmalloc(size, flags); \
	if ( __memtrack_addr ) {                               \
		memtrack_alloc(MEMTRACK_KMALLOC, (unsigned long)(__memtrack_addr), size, __FILE__, __LINE__, flags); \
	}                                                                     \
	__memtrack_addr;			                                                      \
})


#define kfree(addr) ({ \
	if ( addr ) {      \
		memtrack_free(MEMTRACK_KMALLOC, (unsigned long)(addr), __FILE__, __LINE__); \
	}                    \
	kfree(addr); \
})

#define kmem_cache_alloc(cache, flags) ({ \
	void *__memtrack_addr;         \
                                \
	__memtrack_addr = kmem_cache_alloc(cache, flags); \
	if ( __memtrack_addr ) {                               \
		memtrack_alloc(MEMTRACK_KMEM_OBJ, (unsigned long)(__memtrack_addr), 1, __FILE__, __LINE__, flags); \
	}                                                                     \
	__memtrack_addr;			                                                      \
})


#define kmem_cache_free(cache, addr) ({ \
	if ( addr ) {      \
		memtrack_free(MEMTRACK_KMEM_OBJ, (unsigned long)(addr), __FILE__, __LINE__); \
	}                    \
	kmem_cache_free(cache, addr); \
})

#endif

#endif /* __mtrack_h_ */
