/*
  This software is available to you under a choice of one of two
  licenses.  You may choose to be licensed under the terms of the GNU
  General Public License (GPL) Version 2, available at
  <http://www.fsf.org/copyleft/gpl.html>, or the OpenIB.org BSD
  license, available in the LICENSE.TXT file accompanying this
  software.  These details are also available at
  <http://openib.org/license.html>.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
  BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
  ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
  CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
  SOFTWARE.

  Copyright (c) 2004 Topspin Communications.  All rights reserved.

  $Id: ts_ib_client_query_types.h 444 2004-07-13 01:59:33Z roland $
*/

#ifndef _TS_IB_CLIENT_QUERY_TYPES_H
#define _TS_IB_CLIENT_QUERY_TYPES_H

#if defined(__KERNEL__)
#  include <linux/types.h>
#else
#  include <stdint.h>
#  include <stddef.h>             /* for size_t */
#endif

#include "ts_ib_core_types.h"

enum {
        TS_IB_COMMON_NOTICE_GENERIC       = 1,
        TS_IB_COMMON_INFORM_INFO_GENERIC  = 1
};

enum {
        TS_IB_COMMON_INFORM_UNSUBSCRIBE   = 0,
        TS_IB_COMMON_INFORM_SUBSCRIBE     = 1
};

enum {
        TS_IB_COMMON_NOTICE_TYPE_FATAL    = 0x0,
        TS_IB_COMMON_NOTICE_TYPE_URGENT   = 0x1,
        TS_IB_COMMON_NOTICE_TYPE_SECURITY = 0x2,
        TS_IB_COMMON_NOTICE_TYPE_SM       = 0x3,
        TS_IB_COMMON_NOTICE_TYPE_INFO     = 0x4,
        TS_IB_COMMON_NOTICE_TYPE_ALL      = 0xFFFF
};

enum {
        TS_IB_COMMON_TRAP_PRODUCER_TYPE_CA       = 0x1,
        TS_IB_COMMON_TRAP_PRODUCER_TYPE_SW       = 0x2,
        TS_IB_COMMON_TRAP_PRODUCER_TYPE_ROUTER   = 0x3,
        TS_IB_COMMON_TRAP_PRODUCER_TYPE_SM       = 0x4
};

enum {
        TS_IB_GENERIC_TRAP_NUM_IN_SVC                     = 64,
        TS_IB_GENERIC_TRAP_NUM_OUT_OF_SVC                 = 65,
        TS_IB_GENERIC_TRAP_NUM_CREATE_MC_GRP              = 66,
        TS_IB_GENERIC_TRAP_NUM_DELETE_MC_GRP              = 67,
        TS_IB_GENERIC_TRAP_NUM_PORT_CHANGE_STATE          = 128,
        TS_IB_GENERIC_TRAP_NUM_INTEGRITY_TRESHOLD_REACHED = 129,
        TS_IB_GENERIC_TRAP_NUM_EXCESSIVE_BUFFER_OVERRUN   = 130,
        TS_IB_GENERIC_TRAP_NUM_FLOW_CTRL_WATCHDOG_EXPIRED = 131,
        TS_IB_GENERIC_TRAP_NUM_BAD_M_KEY                  = 256,
        TS_IB_GENERIC_TRAP_NUM_BAD_P_KEY                  = 257,
        TS_IB_GENERIC_TRAP_NUM_BAD_Q_KEY                  = 258,
        TS_IB_GENERIC_TRAP_NUM_ALL                        = 0xFFFF
};

/* common attribute */
struct ib_common_attrib_cpi {
        uint8_t         base_version;
        uint8_t         class_version;
        uint16_t        capability_mask;
        uint32_t        resp_time;
        tTS_IB_GID      redirect_gid;
        uint32_t        redirect_tc;
        uint32_t        redirect_sl;
        uint32_t        redirect_fl;
        uint16_t        redirect_lid;
        uint16_t        redirect_pkey;
        uint32_t        redirect_qpn;
        uint32_t        redirect_qkey;
        tTS_IB_GID      trap_gid;
        uint32_t        trap_tc;
        uint32_t        trap_sl;
        uint32_t        trap_fl;
        uint16_t        trap_lid;
        uint16_t        trap_pkey;
        uint32_t        trap_hop_limit;
        uint32_t        trap_qp;
        uint32_t        trap_qkey;
};

struct ib_common_attrib_notice {
        union {
                struct {
                        uint32_t    type;
                        uint32_t    producer_type;
                        uint16_t    trap_num;
                        uint16_t    issuer_lid;
                } generic;
                struct {
                        uint32_t    type;
                        uint32_t    vendor_id;
                        uint16_t    device_type;
                        uint16_t    issuer_lid;
                } vendor;
        } define;
        uint16_t        toggle;
        uint16_t        count;
        union {
                uint8_t       data[54];
                struct {
                        uint8_t     reserved[6];
                        tTS_IB_GID  gid;
                        uint8_t     padding[32];
                } sm_trap;
        } detail;
        tTS_IB_GID      issuer_gid;
};

struct ib_common_attrib_inform {
        tTS_IB_GID      gid;
        uint16_t        lid_range_begin;
        uint16_t        lid_range_end;
        uint16_t        reserved;
        uint8_t         is_generic;
        uint8_t         subscribe;
        uint16_t        type;
        union {
                struct {
                        uint16_t    trap_num;
                        uint32_t    resp_time;
                        uint32_t    producer_type;
                } generic;
                struct {
                        uint16_t    device_id;
                        uint32_t    resp_time;
                        uint32_t    vendor_id;
                } vendor;
        } define;
};

#define IB_SA_SERVICE_COMP_MASK_ID       (1ull <<  0)
#define IB_SA_SERVICE_COMP_MASK_GID      (1ull <<  1)
#define IB_SA_SERVICE_COMP_MASK_PKEY     (1ull <<  2)
/* reserved is bit 3 */
#define IB_SA_SERVICE_COMP_MASK_LEASE    (1ull <<  4)
#define IB_SA_SERVICE_COMP_MASK_KEY      (1ull <<  5)
#define IB_SA_SERVICE_COMP_MASK_NAME     (1ull <<  6)
#define IB_SA_SERVICE_COMP_MASK_DATA8_1  (1ull <<  7)
#define IB_SA_SERVICE_COMP_MASK_DATA8_2  (1ull <<  8)
#define IB_SA_SERVICE_COMP_MASK_DATA8_3  (1ull <<  9)
#define IB_SA_SERVICE_COMP_MASK_DATA8_4  (1ull << 10)
#define IB_SA_SERVICE_COMP_MASK_DATA8_5  (1ull << 11)
#define IB_SA_SERVICE_COMP_MASK_DATA8_6  (1ull << 12)
#define IB_SA_SERVICE_COMP_MASK_DATA8_7  (1ull << 13)
#define IB_SA_SERVICE_COMP_MASK_DATA8_8  (1ull << 14)
#define IB_SA_SERVICE_COMP_MASK_DATA8_9  (1ull << 15)
#define IB_SA_SERVICE_COMP_MASK_DATA8_10 (1ull << 16)
#define IB_SA_SERVICE_COMP_MASK_DATA8_11 (1ull << 17)
#define IB_SA_SERVICE_COMP_MASK_DATA8_12 (1ull << 18)
#define IB_SA_SERVICE_COMP_MASK_DATA8_13 (1ull << 19)
#define IB_SA_SERVICE_COMP_MASK_DATA8_14 (1ull << 20)
#define IB_SA_SERVICE_COMP_MASK_DATA8_15 (1ull << 21)
#define IB_SA_SERVICE_COMP_MASK_DATA8_16 (1ull << 22)
#define IB_SA_SERVICE_COMP_MASK_DATA16_1 (1ull << 23)
#define IB_SA_SERVICE_COMP_MASK_DATA16_2 (1ull << 24)
#define IB_SA_SERVICE_COMP_MASK_DATA16_3 (1ull << 25)
#define IB_SA_SERVICE_COMP_MASK_DATA16_4 (1ull << 26)
#define IB_SA_SERVICE_COMP_MASK_DATA16_5 (1ull << 27)
#define IB_SA_SERVICE_COMP_MASK_DATA16_6 (1ull << 28)
#define IB_SA_SERVICE_COMP_MASK_DATA16_7 (1ull << 29)
#define IB_SA_SERVICE_COMP_MASK_DATA16_8 (1ull << 30)
#define IB_SA_SERVICE_COMP_MASK_DATA32_1 (1ull << 31)
#define IB_SA_SERVICE_COMP_MASK_DATA32_2 (1ull << 32)
#define IB_SA_SERVICE_COMP_MASK_DATA32_3 (1ull << 33)
#define IB_SA_SERVICE_COMP_MASK_DATA32_4 (1ull << 34)
#define IB_SA_SERVICE_COMP_MASK_DATA64_1 (1ull << 35)
#define IB_SA_SERVICE_COMP_MASK_DATA64_2 (1ull << 36)


struct ib_common_attrib_service {
        uint64_t        service_id;
        tTS_IB_GID      service_gid;
        uint16_t        service_pkey;

        uint16_t        reserved;

        uint32_t        service_lease;

        uint8_t         service_key[16];

        uint8_t         service_name[64];

        uint8_t         service_data8[16];
        uint16_t        service_data16[8];
        uint32_t        service_data32[4];
        uint64_t        service_data64[2];
};

#endif /*_TS_IB_CLIENT_QUERY_TYPES_H */
