/*
  This software is available to you under a choice of one of two
  licenses.  You may choose to be licensed under the terms of the GNU
  General Public License (GPL) Version 2, available at
  <http://www.fsf.org/copyleft/gpl.html>, or the OpenIB.org BSD
  license, available in the LICENSE.TXT file accompanying this
  software.  These details are also available at
  <http://openib.org/license.html>.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
  BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
  ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
  CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
  SOFTWARE.

  Copyright (c) 2004 Topspin Communications.  All rights reserved.

  $Id: ts_ib_useraccess.h 432 2004-07-12 17:27:37Z roland $
*/

#ifndef _TS_IB_USERACCESS_H
#define _TS_IB_USERACCESS_H

#include "ts_ib_mad_types.h"

#include <linux/ioctl.h>

typedef uint32_t tTS_IB_USER_MAD_FILTER_HANDLE;

/*
 * Userland compatibility typedefs.
 * These are very temporary.  Please do not use them as they will be going
 * away as soon as opensm and other userland applications update their code.
 */
typedef struct ib_user_mad_filter tTS_IB_USER_MAD_FILTER_STRUCT,
	*tTS_IB_USER_MAD_FILTER;
typedef struct ib_get_port_info_ioctl tTS_IB_GET_PORT_INFO_IOCTL_STRUCT,
	*tTS_IB_GET_PORT_INFO_IOCTL;
typedef struct ib_set_port_info_ioctl tTS_IB_SET_PORT_INFO_IOCTL_STRUCT,
	*tTS_IB_SET_PORT_INFO_IOCTL;
typedef struct ib_mad_process_ioctl tTS_IB_MAD_PROCESS_IOCTL_STRUCT,
	*tTS_IB_MAD_PROCESS_IOCTL;
typedef struct ib_qp_register_ioctl tTS_IB_QP_REGISTER_IOCTL_STRUCT,
	*tTS_IB_QP_REGISTER_IOCTL;
typedef struct ib_path_record_ioctl tTS_IB_PATH_RECORD_IOCTL_STRUCT,
	*tTS_IB_PATH_RECORD_IOCTL;
typedef struct ib_gid_entry_ioctl tTS_IB_GID_ENTRY_IOCTL_STRUCT,
	*tTS_IB_GID_ENTRY_IOCTL;

struct ib_user_mad_filter {
        tTS_IB_PORT                   port;
        __u32                         qpn;
        uint8_t                       mgmt_class;
        uint8_t                       r_method;
        uint16_t                      attribute_id;
        tTS_IB_MAD_DIRECTION          direction;
        tTS_IB_MAD_FILTER_MASK        mask;
        tTS_IB_USER_MAD_FILTER_HANDLE handle;
};

struct ib_get_port_info_ioctl {
        tTS_IB_PORT               port;
        struct ib_port_properties port_info;
};

struct ib_set_port_info_ioctl {
        tTS_IB_PORT            port;
        struct ib_port_changes port_info;
};

struct ib_mad_process_ioctl {
        struct ib_mad      mad;
        enum ib_mad_result result;
};

struct ib_path_record_ioctl {
        uint32_t               dst_addr;
        struct ib_path_record *path_record;
};

struct ib_gid_entry_ioctl {
        tTS_IB_PORT                   port;
        int                           index;
        tTS_IB_GID                    gid_entry;
};

struct ib_get_dev_info_ioctl {
        struct ib_device_properties dev_info;
};

/* Old useraccess module used magic 0xbb; we change it here so
   old binaries don't fail silently in strange ways. */
#define TS_IB_IOCTL_MAGIC 0xbc

/* Add a MAD filter; handle will be filled in by kernel */
#define TS_IB_IOCSMADFILTADD     _IOWR(TS_IB_IOCTL_MAGIC, 1, \
				       struct ib_user_mad_filter *)

/* Remove a MAD filter */
#define TS_IB_IOCSMADFILTDEL     _IOW(TS_IB_IOCTL_MAGIC, 2, \
				      tTS_IB_USER_MAD_FILTER_HANDLE *)

/* Get port info */
#define TS_IB_IOCGPORTINFO       _IOR(TS_IB_IOCTL_MAGIC, 3, \
				      struct ib_get_port_info_ioctl *)

/* Set port info */
#define TS_IB_IOCSPORTINFO       _IOW(TS_IB_IOCTL_MAGIC, 4, \
				      struct ib_set_port_info_ioctl *)

/* Set receive queue size */
#define TS_IB_IOCSRCVQUEUELENGTH _IOW(TS_IB_IOCTL_MAGIC, 5, uint32_t *)

/* Get receive queue size */
#define TS_IB_IOCGRCVQUEUELENGTH _IOR(TS_IB_IOCTL_MAGIC, 6, uint32_t *)

/* Have MAD processed by provider */
#define TS_IB_IOCMADPROCESS      _IOWR(TS_IB_IOCTL_MAGIC, 7, \
				       struct ib_mad_process_ioctl *)

/* Register a QP with the kernel */
#define TS_IB_IOCQPREGISTER      _IOWR(TS_IB_IOCTL_MAGIC, 8, \
				       struct ib_qp_register_ioctl *)

/* Perform path record lookup */
#define TS_IB_IOCGPATHRECORD     _IOWR(TS_IB_IOCTL_MAGIC, 9, \
				       struct ib_path_record_ioctl *)

/* Fetch a GID */
#define TS_IB_IOCGGIDENTRY       _IOWR(TS_IB_IOCTL_MAGIC, 10, \
				       struct ib_gid_entry_ioctl *)

/* Get device info */
#define TS_IB_IOCGDEVINFO        _IOR(TS_IB_IOCTL_MAGIC, 11, \
				     struct ib_get_dev_info_ioctl *)

#endif /* _TS_IB_USERACCESS_H */
