/*
  This software is available to you under a choice of one of two
  licenses.  You may choose to be licensed under the terms of the GNU
  General Public License (GPL) Version 2, available at
  <http://www.fsf.org/copyleft/gpl.html>, or the OpenIB.org BSD
  license, available in the LICENSE.TXT file accompanying this
  software.  These details are also available at
  <http://openib.org/license.html>.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
  BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
  ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
  CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
  SOFTWARE.

  Copyright (c) 2004 Topspin Communications.  All rights reserved.

  $Id: ts_ib_useraccess_cm.h 432 2004-07-12 17:27:37Z roland $
*/

#ifndef _TS_IB_USERACCESS_CM_H
#define _TS_IB_USERACCESS_CM_H

#include "ts_ib_useraccess.h"
#include "ts_ib_cm.h"
#include "ts_ib_cm_user.h"

struct ib_qp_register_ioctl {
        VAPI_k_qp_hndl_t      vk_qp_handle;
        enum ib_qp_state      qp_state;
        __u32                 qpn;
        enum ib_transport     transport;
        struct ib_qp         *qp_handle;		/* OUT */
};

struct ib_cm_connect_ioctl {
        unsigned long                cm_arg;
        uint8_t                      responder_resources;
        uint8_t                      initiator_depth;
        uint8_t                      retry_count;
        uint8_t                      rnr_retry_count;
        uint8_t                      cm_response_timeout;
        uint8_t                      max_cm_retries;
        struct ib_qp                *qp_handle;
        tTS_IB_U_K_QP_HANDLE         k_qp_handle;
        struct ib_path_record       *primary_path;
        struct ib_path_record       *alternate_path;
        tTS_IB_SERVICE_ID            service_id;
        void                        *req_private_data;
        int                          req_private_data_len;
        tTS_IB_CM_COMM_ID            comm_id;			/* OUT */
        enum ib_qp_state             qp_state;		/* OUT */
};

struct ib_cm_listen_ioctl {
        unsigned long         cm_arg;
        tTS_IB_SERVICE_ID     service_id;
        tTS_IB_SERVICE_ID     service_mask;
        tTS_IB_LISTEN_HANDLE  listen_handle;
};

struct ib_cm_alternate_path_load_ioctl {
        tTS_IB_CM_COMM_ID      comm_id;
        struct ib_path_record *alternate_path;
};

struct ib_cm_accept_ioctl {
        unsigned long         cm_arg;
        tTS_IB_CM_COMM_ID     comm_id;
        struct ib_qp         *qp_handle;
        tTS_IB_U_K_QP_HANDLE  k_qp_handle;
        void                 *reply_data;
        int                   reply_size;
        enum ib_qp_state      qp_state;			/* OUT */
};

struct ib_cm_reject_ioctl {
        tTS_IB_CM_COMM_ID    comm_id;
        void                *reply_data;
        int                  reply_size;
};

struct ib_cm_confirm_ioctl {
        tTS_IB_CM_COMM_ID    comm_id;
        struct ib_qp        *qp_handle;
        void                *reply_data;
        int                  reply_size;

        enum ib_qp_state     qp_state;		/* OUT */
};

struct ib_cm_delayresponse_ioctl {
        tTS_IB_CM_COMM_ID    comm_id;
        int                  service_timeout;
        void                *mra_data;
        int                  mra_data_size;
};

struct ib_cm_establist_ioctl {
        tTS_IB_CM_COMM_ID    comm_id;

        enum ib_qp_state     qp_state;		/* OUT */
};

struct ib_u_cm_get_completion_ioctl {
        struct ib_qp          *qp_handle;
        tTS_IB_CM_COMM_ID      comm_id;
        tTS_IB_LISTEN_HANDLE   listen_handle;
        tTS_IB_CM_EVENT        event;
        unsigned long          cm_arg;
        void                  *params;
        ssize_t                params_size;
        enum ib_qp_state       qp_state;			/* OUT */
};

struct ib_cm_get_remote_gid_ioctl {
        tTS_IB_CM_COMM_ID      comm_id;
        tTS_IB_GID            *gid;
};



/* user mode CM filter ioctls */

/* we start numbering at 100 */

/* Make a connection */
#define TS_IB_IOCCMCONNECT       _IOWR(TS_IB_IOCTL_MAGIC, 100, \
				       struct ib_cm_connect_ioctl *)

/* Begin listening */
#define TS_IB_IOCCMLISTEN        _IOWR(TS_IB_IOCTL_MAGIC, 101, \
				       struct ib_cm_listen_ioctl *)

/* Cancel listen */
#define TS_IB_IOCCMLISTENSTOP    _IOW(TS_IB_IOCTL_MAGIC, 102, \
				      struct ib_listen_handle *)

/* Disconnect a connection */
#define TS_IB_IOCCMDISCONNECT    _IOW(TS_IB_IOCTL_MAGIC, 103, \
				      struct ib_cm_comm_id *)

/* Load an alternate path */
#define TS_IB_IOCCMALTPATHLOAD   _IOW(TS_IB_IOCTL_MAGIC, 104, \
				      struct ib_cm_alternate_path_load_ioctl *)

/* Accept a connection (passive) */
#define TS_IB_IOCCMACCEPT        _IOW(TS_IB_IOCTL_MAGIC, 105, \
				      struct ib_cm_accept_ioctl *)

/* Reject a connection (passive) */
#define TS_IB_IOCCMREJECT        _IOW(TS_IB_IOCTL_MAGIC, 106, \
				      struct ib_cm_reject_ioctl *)

/* Confirm a connection (active) */
#define TS_IB_IOCCMCONFIRM       _IOWR(TS_IB_IOCTL_MAGIC, 107, \
				       struct ib_cm_confirm_ioctl *)

/* Notify CM that a message has been received */
#define TS_IB_IOCCMESTABLISH     _IOWR(TS_IB_IOCTL_MAGIC, 109, \
				       struct ib_cm_estable_ioctl *)

/* Drop consumer reference to connect */
#define TS_IB_IOCCMDROPCONSUMER  _IOW(TS_IB_IOCTL_MAGIC, 110, \
				      struct ib_cm_comm_id*)

/* Get next completion buffer or block */
#define TS_IB_IOCCMGETCOMPLETION _IOWR(TS_IB_IOCTL_MAGIC, 111, \
				       struct ib_u_cm_cm_get_completion_ioctl *)

/* Assign a service id */
#define TS_IB_IOCCMSERVIDASSIGN  _IOW(TS_IB_IOCTL_MAGIC, 112, uint64_t *)

/* Delay response (passive) */
#define TS_IB_IOCCMDELAYRESPONSE _IOW(TS_IB_IOCTL_MAGIC, 113, \
				      struct ib_cm_delayresponse_ioctl *)

/* Get connection remote gid */
#define TS_IB_IOCMGETREMOTEGID  _IOW(TS_IB_IOCTL_MAGIC, 114, \
                                     struct ib_cm_get_remote_gid *)



#endif /* _TS_IB_USERACCESS_CM_H */
