#!/bin/bash
#
# A script to uninstall OpenIB modules
#

function usage {
    echo 'uninstall.sh'
    exit 0
}


# Find the directory where this script resides
installdir=`cd ${0%*/*};pwd`
prefix=${installdir}
KVERSION=${KVERSION:-$(uname -r)}
KVERSION_TWO_DIGITS=2$(echo ${KVERSION} | cut -d"." -f2)
KSRC=/lib/modules/${KVERSION}/build
# Case of new kernel source tree
KERNEL_TREE_ROOT=${KSRC%*/*}
if [ -d ${KERNEL_TREE_ROOT}/source/ ]; then
   KSRC=$KERNEL_TREE_ROOT/source
fi

if [ $UID != 0 ]; then
        echo You must be root to install IBGD
        exit 1
fi

cat << EOF

   This will uninstall the following packages:
   	1. HCA HW Driver
   	2. ULPs
   	3. OpenSM
   	4. IBADM tools
   	5. MPI

EOF

echo -n "Do you want to continue (y/n) [y] ?"
read force
if [ "x$force" == "xy" ] \
   || [ "x$force" == "xyes" ] \
   || [ "x$force" == "xY" ] \
   || [ "x$force" == "xYES" ] \
   || [ "x$force" == "x" ]  ; then
	echo "Proceeding uninstall process..."
else
   	echo "Aborting uninstallation"
   	exit 1
fi

if ( /etc/init.d/openibd status > /dev/null 2>&1 ); then
	echo "Trying to unload modules (/etc/init.d/openibd stop). Please wait..."
	/etc/init.d/openibd stop
	if [ $? -ne 0 ]; then
		echo "Failed to unload modules."
		read -p "Do you want to continue uninstall process (y/n) [y] ?" force
		if [ "x$force" == "xy" ] \
		   || [ "x$force" == "xyes" ] \
		   || [ "x$force" == "xY" ] \
		   || [ "x$force" == "xYES" ] \
		   || [ "x$force" == "x" ]  ; then
			echo "Proceeding uninstall process..."
		else
		   	echo "Aborting uninstallation"
		   	echo "First unload modules (/etc/init.d/openibd stop), then rerun uninstall.sh"
		   	exit 1
		fi
	fi	
fi

rm -f /etc/dat.conf

if [ "${KVERSION_TWO_DIGITS}" == "26" ]; then 
	# Remove previous configuration if exist
        perl -ni -e 'if (/\#\# BEGIN OpenIB \#\#$/) { $filter = 1 }' -e 'if (!$filter) { print }' -e 'if (/\#\# END OpenIB \#\#$/){ $filter = 0 }' /etc/modprobe.conf
else
	# Remove previous configuration if exist
        perl -ni -e 'if (/\#\# BEGIN OpenIB \#\#$/) { $filter = 1 }' -e 'if (!$filter) { print }' -e 'if (/\#\# END OpenIB \#\#$/){ $filter = 0 }' /etc/modules.conf
fi
    
perl -i -ne "print unless (m@${prefix}/driver/infinihost/lib@)" /etc/ld.so.conf
perl -i -ne "print unless (m@${prefix}/driver/infinihost/lib64@)" /etc/ld.so.conf
perl -i  -ne "print unless (m@${prefix}/lib@)" /etc/ld.so.conf
perl -i  -ne "print unless (m@${prefix}/apps/osm/lib@)" /etc/ld.so.conf
/sbin/ldconfig

#Clean temporary directories
rm -rf /tmp/ibgd-kernel
rm -rf /tmp/ibgd-osm
rm -rf /tmp/ibgd-user
rm -rf /tmp/ibgd-mpi

#Delete links from /usr/bin
CWD=$(pwd)
bin_dirs=`find ${prefix} -name bin -type d`
for bin_dir in $bin_dirs
do
    cd ${bin_dir}
        for bin in *
        do
           [ "$bin" != "*" ] && rm -f /usr/bin/$bin
        done
done

cd ${CWD}

#Delete kerenl sources
rm -rf ${KSRC}/drivers/infiniband
#Delete kernel modules
if [ "${KVERSION_TWO_DIGITS}" == "24" ]; then
           rm -rf /lib/modules/${KVERSION}/kernel/drivers/infiniband
else 
           rm -rf /lib/modules/${KVERSION}/kernel/drivers/infiniband
fi

#Remove MPI (NCSA)
if [ -d ${prefix}/mpi/ncsa ]; then 
   ${prefix}/mpi/ncsa/mpich-vmi-2.0/sbin/mpiuninstall > /dev/null
fi
rm -f /etc/init.d/vmieyes


rm -f /dev/vipkl_dev
rm -f /dev/vipkl
rm -f /dev/mosal
rm -f /dev/udapl
rm -f /dev/ts_ua_cm*
rm -f /dev/ts_ip2pr
rm -f /dev/ts_ua*

cd ${prefix}
cd ..
rm -rf ${prefix}


NETWORK_CONF_DIR=/etc/infiniband
[ -d $NETWORK_CONF_DIR ] && rm -rf $NETWORK_CONF_DIR

if [[ -f /etc/redhat-release || -f /etc/rocks-release ]]; then        
     /sbin/chkconfig openibd off
fi
if [ -f /etc/SuSE-release ]; then
     /sbin/insserv -r openibd
fi

rm -f /etc/init.d/openibd

rm -f /etc/opensm.conf

# Get list of infiniband source and modules directories
ib_dirs=$(/usr/bin/find /lib/modules/${KVERSION} -name infiniband -type d -follow) # -exec /bin/rm -rf {} > /dev/null 2>&1 \;

for ib_dir in ${ib_dirs}
do
	if [ -d ${ib_dir}/hw/mthca ]; then
		echo "Leaving ${ib_dir} - Gen2 kernel sources"
	elif [ -e ${ib_dir}/hw/mthca/ib_mthca.ko ]; then
		echo "Leaving ${ib_dir} - Gen2 kernel modules"
	else
		# Remove Gen1 sources
		/bin/rm -rf ${ib_dir}
	fi
done

# Restore Gen2 sources
if [ -d ${KSRC}/drivers/infiniband-gen2 ]; then
	echo "Restored Gen2 infiniband kernel sources"
	mv ${KSRC}/drivers/infiniband-gen2 ${KSRC}/drivers/infiniband
fi

# Restore Gen2 modules
if [ -d /lib/modules/${KVERSION}/kernel/drivers/infiniband-gen2 ]; then
	/bin/rm -rf  /lib/modules/${KVERSION}/kernel/drivers/infiniband
	mv /lib/modules/${KVERSION}/kernel/drivers/infiniband-gen2 /lib/modules/${KVERSION}/kernel/drivers/infiniband
	for mod in `find /lib/modules/${KVERSION}/kernel/drivers/infiniband -name "*.ko.gen2"`
	do
		mv ${mod} ${mod%*.*}
	done 
fi

if [ "${KVERSION_TWO_DIGITS}" == "24" ]; then
           /sbin/depmod -ae -F /boot/System.map-${KVERSION} ${KVERSION}
else 
           /sbin/depmod -ae ${KVERSION}
fi

echo 'Done uninstalling IBGD modules'
exit 0
