#!/bin/bash
#
# Sample regression client invocation
#
# replace # with the real number in the next lines
#            NO DEAFAULTS VALUES
# usag : ./Daplregress user #.#.#.# ib# netapp   // for netapp user regression
# usag : ./Daplregress user #.#.#.# ib# mlx      // for mlx user regression
# usag : ./Daplregress kernel #.#.#.# ib# netapp // for netapp kernel regression
# usag : ./Daplregress kernel #.#.#.# ib# netapp // for mlx kernel regression

me=`basename $0`

usage()
{
cat << EOF

    Usage:
            $me user #.#.#.# ib# netapp     // for netapp user regression
            $me user #.#.#.# ib# mlx        // for mlx user regression
            $me user #.#.#.# ib# plugall    // for plugfest user regression
            $me user #.#.#.# ib# quit
            $me kernel #.#.#.# ib# netapp   // for netapp kernel regression
            $me kernel #.#.#.# ib# netapp   // for mlx kernel regression
            $me kernel #.#.#.# ib# plugall  // for plugfest kernel regression
            $me kernel #.#.#.# ib# quit
            
EOF
}


[ -z "$1" ] && usage && exit 1

case $1 in
    -h|--help)
        usage
        exit 1
        shift
        ;;
        *)
        ;;
esac        

tc_cnt=0
pass_cnt=0
fail_cnt=0
      
exe=$1
host=$2
device=$3 
testKind=$4

if [ $exe == "user" ]
then 
    exe_prog="dapltest"
elif [ $exe == "kernel" ]
then
    exe_prog="kdapltest"
else
    usage
    exit 1    
fi


ex()
{
    echo "-----------------------------------------------------------------------------------------"
    echo "Running [$@]"
    echo "-----------------------------------------------------------------------------------------"
    let tc_cnt++
    eval "$@"
    
    if [ $? -ne 0 ]; then
        echo
        echo "Failed executing [$@]"
        echo
        let fail_cnt++
        return 2
    fi

    let pass_cnt++
}


if [ $testKind == "netapp" ]
then
#====================================================================
#client1
#====================================================================
ex "/usr/local/ibgd/bin/${exe_prog} -T T -s ${host} -D ${device} -i 10000 -t 1 -w 1 \
        client SR 256                                                               \
        server SR 256"

#====================================================================
#client2
#====================================================================
ex "/usr/local/ibgd/bin/${exe_prog} -T T -s ${host} -D ${device} -i 10000 -t 1 -w 1 \
        client SR 256                                                               \
        server RW 4096                                                              \
        server SR 256"

#====================================================================
#client3
#====================================================================
ex "/usr/local/ibgd/bin/${exe_prog} -T T -s ${host} -D ${device} -i 10000 -t 1 -w 1 \
	client SR 256 					   \
	server RR 4096 					   \
	server SR 256"

#====================================================================
#client4
#====================================================================
ex "/usr/local/ibgd/bin/${exe_prog} -T T -s ${host} -D ${device} -i 10000 -t 1 -w 1 \
	client SR 256 					   \
	server RW 4096 					   \
	server SR 256 					   \
	client SR 256 					   \
	server RR 4096 					   \
	server SR 256 					   \
	client SR 4096 					   \
	server SR 256"
 
#====================================================================
#client5
#====================================================================
ex "/usr/local/ibgd/bin/${exe_prog} -T T -s ${host} -D ${device} -i 10000 -t 1 -w 8 \
	client SR 256 					   \
	server RW 4096 					   \
	server SR 256 					   \
	client SR 256 					   \
	server RR 4096 					   \
	server SR 256 					   \
	client SR 4096 					   \
	server SR 256"
 
#====================================================================
#client6
#====================================================================
ex "/usr/local/ibgd/bin/${exe_prog} -T T -s ${host} -D ${device} -i 10000 -t 4 -w 8 \
	client SR 256 					   \
	server RW 4096 					   \
	server SR 256 					   \
	client SR 256 					   \
	server RW 4096 					   \
	server SR 256 					   \
	client SR 4096 					   \
	server SR 256"


#====================================================================
# Quit the server
#====================================================================
ex "/usr/local/ibgd/bin/${exe_prog} -T Q -s ${host} -D ${device}"
fi
if [ $testKind == "mlx" ]
then
#====================================================================
#client1
#====================================================================
ex "/usr/local/ibgd/bin/${exe_prog} -T T -s ${host} -D ${device} -i 10000 -t 1 -w 1 \
	client SR 256 -f					   \
	server SR 256 -f"

#====================================================================
#client2
#====================================================================
ex "/usr/local/ibgd/bin/${exe_prog} -T T -s ${host} -D ${device} -i 10000 -t 4 -w 8 \
        -V -P                                              \
	client SR 256 					   \
	server RW 4096 					   \
	server SR 256 					   \
	client SR 256 					   \
	server RR 4096 					   \
	server SR 256 					   \
	client SR 4096 					   \
	server SR 256"

#====================================================================
#client3
#====================================================================
ex "/usr/local/ibgd/bin/${exe_prog} -T T -s ${host} -D ${device} -i 10000 -t 4 -w 8 \
        -V                                                 \
	client SR 256 					   \
	server RW 4096 					   \
	server SR 256 					   \
	client SR 256 					   \
	server RR 4096 					   \
	server SR 256 					   \
	client SR 4096 					   \
	server SR 256"

#====================================================================
#client4
#====================================================================
ex "/usr/local/ibgd/bin/${exe_prog} -T T -s ${host} -D ${device} -i 10000 -t 4 -w 8 \
	client SR 256 2					   \
	server RW 4096 2				   \
	server SR 256 2					   \
	client SR 256 2					   \
	server RR 4096 2				   \
	server SR 256 2					   \
	client SR 4096 2				   \
	server SR 256 2"
#====================================================================
#client5
#====================================================================
ex "/usr/local/ibgd/bin/${exe_prog} -T T -s ${host} -D ${device} -i 10000 -t 4 -w 8 \
	client SR 256 4					   \
	server RW 4096 4				   \
	server SR 256 4					   \
	client SR 256 4					   \
	server RR 4096 4				   \
	server SR 256 4					   \
	client SR 4096 4				   \
	server SR 256 4"
#====================================================================
#client6
#====================================================================
ex "/usr/local/ibgd/bin/${exe_prog} -T P -s ${host} -D ${device} -i 1000 \
        -m p				           \
	RR 4096"
#====================================================================
#client7
#====================================================================
ex "/usr/local/ibgd/bin/${exe_prog} -T P -s ${host} -D ${device} -i 1000 \
        -m b				           \
	RR 4096 4"
#====================================================================
#client8
#====================================================================
ex "/usr/local/ibgd/bin/${exe_prog} -T P -s ${host} -D ${device} -i 1000 \
        -m p				           \
	RW 4096 4"				   
#====================================================================
#client9
#====================================================================
ex "/usr/local/ibgd/bin/${exe_prog} -T P -s ${host} -D ${device} -i 1000 \
        -m b				           \
        RW 4096 4"
	
#====================================================================
# Quit the server
#====================================================================
ex "/usr/local/ibgd/bin/${exe_prog} -T Q -s ${host} -D ${device}"

fi


if [ $testKind == "g1plug1" ] || [ $testKind == "plugall" ]
then
#====================================================================
#plugfest1
#====================================================================
ex "/usr/local/ibgd/bin/${exe_prog} -T T -s ${host} -D ${device} -i 100 -t 1 -w 1 -R BE \
	client SR 256 					   \
	server SR 256"
fi

if [ $testKind == "g1plug2" ] || [  $testKind == "plugall" ]
then
#====================================================================
#plugfest2
#====================================================================
ex "/usr/local/ibgd/bin/${exe_prog} -T T -s ${host} -D ${device} -i 100 -t 1 -w 1 -V -P -R BE \
	client SR 1024 3 -f 					   \
	server SR 1536 2 -f"
fi

if [ $testKind == "g2plug1" ] || [ $testKind == "plugall" ]
then
#====================================================================
#group 2 plugfest1
#====================================================================
ex "/usr/local/ibgd/bin/${exe_prog} -T T -s ${host} -D ${device} -i 100 -t 1 -w 1 -V -R BE \
	client SR 1024 1 					   \
	server SR 1024 1"
fi

if [ $testKind == "g2plug2" ] || [ $testKind == "plugall" ]
then
#====================================================================
#group 2 plugfest2
#====================================================================
ex "/usr/local/ibgd/bin/${exe_prog} -T T -s ${host} -D ${device} -i 100 -t 1 -w 10 -V -P -R BE \
	client SR 1024 3  					   \
	server SR 1536 2"
fi

if [ $testKind == "g2plug3" ] || [ $testKind == "plugall" ]
then
#====================================================================
#group 2 plugfest3
#====================================================================
ex "/usr/local/ibgd/bin/${exe_prog} -T T -s ${host} -D ${device} -i 100 -t 1 -w 1 -V -P -R BE \
        client SR 256 3  					   \
        server RW 4096 1					   \
        server SR 256 1"
fi
if [ $testKind == "g2plug4" ] || [ $testKind == "plugall" ]
then
#====================================================================
#group 2 plugfest4
#====================================================================
ex "/usr/local/ibgd/bin/${exe_prog} -T T -s ${host} -D ${device} -i 100 -t 1 -w 1 -V -P -R BE \
        client SR 256 3  					   \
        server RR 4096 1					   \
        server SR 256 1"
fi

if [ $testKind == "g3plug1" ] || [ $testKind == "plugall" ]
then
#====================================================================
#group 3 plugfest1
#====================================================================
ex "/usr/local/ibgd/bin/${exe_prog} -T T -s ${host} -D ${device} -i 100 -t 4 -w 8 -V -P -R BE\
	client SR 256 1 				   \
	server RR 4096 1				   \
	server SR 256 1					   \
	client SR 256 1					   \
	server RR 4096 1				   \
	server SR 256 1"
fi

if [ $testKind == "g3plug2" ] || [  $testKind == "plugall" ]
then
#====================================================================
#group 3 plugfest2
#====================================================================
ex "/usr/local/ibgd/bin/${exe_prog} -T P -s ${host} -D ${device} -i 1024 -p 64 -m p \
	RW 8192 2"
fi

if [ $testKind == "g3plug3" ] || [ $testKind == "plugall" ]
then
#====================================================================
#group 3 plugfest3
#====================================================================
ex "/usr/local/ibgd/bin/${exe_prog} -T T -s ${host} -D ${device} -i 100 -t 1 -w 10 -V -P -R BE \
	client SR 1024 3 					   \
	server SR 1536 2"

fi
if [ $testKind == "plugall" ] || [ $testKind == "quit" ]
then
#====================================================================
# close
#====================================================================
ex "/usr/local/ibgd/bin/${exe_prog} -T Q -s ${host} -D ${device}"
fi


# Print results statistics
echo
echo "================================================"
echo "  Executed: $tc_cnt                             "
echo "  Passed: $pass_cnt                             "
echo "  Failed: $fail_cnt                             "
echo "================================================"

