#!/bin/bash
#
# A script to install OpenSM over mVAPI
#

function usage {
    echo 'install.sh [-h/--help] [-d/--debug] [-c/--clean] [-t/--tmp_dir dir] [--cover]'
    echo '   [-p/--prefix dir] [-b/--build=ibgd|mtl|sim] [-v/--vendor ts|mtl|sim]'
    echo ' ' 
    echo 'options: ' 
    echo "-c/--clean = cleanup build tree from all object files etc"
    echo "-t/--tmp_dir = temporary directory where the tool is built"
    echo "-p/--prefix = where to install executables and libs"
    echo "-b/--build = use build environment defaults for install and paths"
    echo "-v/--vendor = the API to be used. It is preferable to use -b"
    echo "-r/--build_root = root of where the build is done."
    echo "-o/--build_only = only build the environment. Do not update global links, etc."
    echo "-a/--batch = running in batch mode. No running of fprogbar"
    echo "--cover = running installation in coverage mode , enabled only with TopSpin stack"
    echo ""
    echo "NOTE: If you do not use the -b option you need will to provide the"
    echo "      following environment variables to point to the Mellanox"
    echo "      and Topspin (for vendor=ts only) headers:"
    echo "      1. MTHOME such that the file: \$MTHOME/include/vapi.h exists"
    echo "      2. TSHOME such that the file: \$TSHOME/ts_ib_useraccess.h exists"
    exit 0
}

# set the build options based on the build
function set_build_opts () {
 # Handle the case vendor is specified and given after build option
 if [[ $vset == 1 && $bset == 0 ]]; then
 case "$v" in
    "mtl" )
    b="mtl"
    ;;
    
    "ts" )
    b="ibgd"
    ;;

    "sim" )
    b="sim"
    ;;

    * )
    echo "-I- No vendor specified , using default settings"
    ;;
 esac
 fi
 case "$b" in
   "mtl" )
    if [ $pset == 0 ] ;then p=/usr/mellanox ; fi
    if [ $tset == 0 ] ;then t=/usr/mellanox/osm ;fi
    if [ $vset == 0 ] ;then v=mtl; fi

    if test -z $MTHOME; then 
        # to support vapi4 we test for the vapi.h in the /usr/include first
        if test -f /usr/include/vapi.h; then
            export MTHOME=/usr
        else
            export MTHOME=/usr/mellanox; 
        fi
    fi
    ;;

   "ibgd" ) 
    if [ $pset == 0 ] ;then p=/usr/local/ibgd/apps/osm ;fi
    if [ $tset == 0 ] ;then t=/tmp/osm_build ;fi
    if [ $vset == 0 ] ;then v=ts; fi

    if test -z $MTHOME; then
        export MTHOME=/usr/local/ibgd/driver/infinihost
    fi
    if test -z $TSHOME; then 
	if  test -d /lib/modules/`uname -r`/source/; then
		export TSHOME=/lib/modules/`uname -r`/source/drivers/infiniband/include/
	else
		export TSHOME=/lib/modules/`uname -r`/build/drivers/infiniband/include/
	fi     
    fi
    ;;

    "sim" )
    if [ $pset == 0 ] ;then p=/usr/mellanox/osm-sim ;fi
    if [ $tset == 0 ] ;then t=/tmp/osm_build_sim ;fi
    if [ $vset == 0 ] ;then v=sim; fi

    # we do not want to override any installed opensm
    o=1
    ;;
	*) echo "Unknown Build Type:$b"; exit 1 ;;
   esac
}

# create executable files that do LD_PRELOAD on systems where TLS pthread
# is available
function do_ld_preload_wrapper () {
    # Only do the script if in install mode and not build mode
    # first we move the opensm and osmtest
    mv $p/bin/opensm $p/bin/opensm.bin
    mv $p/bin/osmtest $p/bin/osmtest.bin
    cat > $p/bin/opensm <<EOF
#!/bin/sh -f

# we need to use the corrrect libraries based on the machine
# use 64 bit libs if needed
if test -d /lib/tls || test -d /lib64/tls ; then

   ptl_path=\`gcc -print-file-name=libpthread.so|sed s=[^/]*$==\`
   if test 0 = \`file \$ptl_path/libpthread.so|grep "text"|wc -l\`; then
       ptl_file=\`cd \$ptl_path;pwd\`/libpthread.so
   else
       ptl_file=\`awk '/GROUP/{print \$3}' \$ptl_path/libpthread.so\`
   fi
   glibc_path=\`gcc -print-file-name=libc.so|sed s=[^/]*$==\`
   if test 0 = \`file \$glibc_path/libc.so|grep "text"|wc -l\`; then
      glibc_file=\`cd \$glibc_path;pwd\`/libc.so
   else                               
      glibc_file=\`awk '/GROUP/{print \$3}' \$glibc_path/libc.so\`
   fi
   libm_path=\`gcc -print-file-name=libm.so|sed s=[^/]*$==\`
   if test 0 = \`file \$libm_path/libm.so|grep "text"|wc -l\`; then
      libm_file=\`cd \$libm_path;pwd\`/libm.so
   else
      libm_file=\`awk '/GROUP/{print \$3}' \$libm_path/libm.so\`
   fi
   librt_path=\`gcc -print-file-name=librt.so|sed s=[^/]*$==\`
   if test 0 = \`file \$librt_path/librt.so|grep "text"|wc -l\`; then
      librt_file=\`cd \$librt_path;pwd\`/librt.so
   else
      librt_file=\`awk '/GROUP/{print \$3}' \$librt_path/librt.so\`
   fi
   echo "LD_PRELOAD=\$glibc_file:\$ptl_file:\$libm_file:\$librt_file"
   export LD_PRELOAD="\$glibc_file:\$ptl_file:\$libm_file:\$librt_file"

fi

$orig_p/bin/opensm.bin \$@

EOF
    
    cat > $p/bin/osmtest <<EOF
#!/bin/sh -f

# we need to use the corrrect libraries based on the machine
# use 64 bit libs if needed
if test -d /lib/tls || test -d /lib64/tls ; then
   ptl_path=\`gcc -print-file-name=libpthread.so|sed s=[^/]*$==\`
   if test 0 = \`file \$ptl_path/libpthread.so|grep "text"|wc -l\`; then
       ptl_file=\`cd \$ptl_path;pwd\`/libpthread.so
   else
       ptl_file=\`awk '/GROUP/{print \$3}' \$ptl_path/libpthread.so\`
   fi
fi
glibc_path=\`gcc -print-file-name=libc.so|sed s=[^/]*$==\`
if test 0 = \`file \$glibc_path/libc.so|grep "text"|wc -l\`; then
  glibc_file=\`cd \$glibc_path;pwd\`/libc.so
else
  glibc_file=\`awk '/GROUP/{print \$3}' \$glibc_path/libc.so\`
fi
libm_path=\`gcc -print-file-name=libm.so|sed s=[^/]*$==\`
if test 0 = \`file \$libm_path/libm.so|grep "text"|wc -l\`; then
  libm_file=\`cd \$libm_path;pwd\`/libm.so
else
  libm_file=\`awk '/GROUP/{print \$3}' \$libm_path/libm.so\`
fi
librt_path=\`gcc -print-file-name=librt.so|sed s=[^/]*$==\`
if test 0 = \`file \$librt_path/librt.so|grep "text"|wc -l\`; then
  librt_file=\`cd \$librt_path;pwd\`/librt.so
else
  librt_file=\`awk '/GROUP/{print \$3}' \$librt_path/librt.so\`
fi

echo "LD_PRELOAD=\$glibc_file:\$ptl_file:\$libm_file:\$librt_file"
export LD_PRELOAD=\$glibc_file:\$ptl_file:\$libm_file:\$librt_file
$orig_p/bin/osmtest.bin \$@

EOF

chmod 755 $p/bin/osmtest $p/bin/opensm

}


# Find the directory where this script resides
i=`cd ${0%*/*};pwd`
prog=$0

# set default options
COVER=0
# we build under mellanox build
b=ibgd
bset=0
vset=0
pset=0
tset=0
# is debug version 
d=0
c=0
h=0
v=ts
r=""
o=0
a=0

# Handle the case there is no makedepend...
export PATH=$PATH:$i

# Parse command line
TEMP=`getopt -o t:p:b:v:r:dcoah --long tmp_dir:,prefix:,build:,vendor:,build_root:,debug,clean,build_only,batch,cover,help -n 'install.sh' -- "$@"`
 

if [ $? != 0 ] ; then echo "Terminating..." >&2 ; exit 1 ; fi
 
# Note the quotes around `$TEMP': they are essential!
eval set -- "$TEMP"
 
while true ; do
 	case "$1" in
		-t|--tmp_dir)    t=$2;  tset=1;  shift 2;;
		-p|--prefix)     p=$2;  pset=1;  shift 2;;
		-b|--build)      b=$2;  bset=1;  shift 2;;
		-v|--vendor)     v=$2;  vset=1;  shift 2;;
                -r|--build_root) r=$2;  shift 2;;
		-d|--debug)      d=1;   shift;;
		-c|--clean)      c=1;   shift;;
                -o|--build_only) o=1;  shift;;
                -a|--batch)      a=1;  shift;;
                --cover)      COVER=1;  shift;;
		-h|--help)       usage; shift;;
 		--) shift ; break ;;
 		*) echo "Wrong parameter error!" ; usage; exit 1 ;;
   esac
done

# now it is the time to set the default build type set by user request
set_build_opts $b

echo "Installing OpenSM into Prefix=$p Vendor=$v Stack=$b tmp=$t "

# save the original path. This is what we need to use for PREFIX
orig_p=$p

if [ "$r" != "" ]; then
    p=$r/$p
    t=$r/$t
fi

if [[ "$v" != "mtl" && "$v" != "ts" && "$v" != "sim"  ]]; then
   echo "Error: Provided vendor must be either mtl, ts or sim"
   exit 1
fi

if [ "$v" == "ts" ]; then
    if test -z $TSHOME; then
        echo "-E- Compiling over OpenIB API requires TSHOME env."
        echo "    Please set TSHOME to point to a directory such that"
        echo "    the file \$TSHOME/ts_ib_useraccess.h exists"
        exit 1
    else
        if test ! -f $TSHOME/ts_ib_useraccess.h; then
            echo "-E- Wrong TSHOME env: $TSHOME"
            echo "    Please set TSHOME to point to a directory such that"
            echo "    the file \$TSHOME/ts_ib_useraccess.h exists"
            exit 1
        fi
    fi
fi

if [[ "$v" == "ts" || "$v" == "mtl" ]]; then
    if test -z $MTHOME; then
        echo "-E- Missing MTHOME env variable"
        echo "    Please set MTHOME to point to a directory such that"
        echo "    the file \$MTHOME/include/vapi.h exists"
        exit 1
    else 
        if test ! -f $MTHOME/include/vapi.h; then
            echo "-E- Wrong MTHOME env variable:$MTHOME"
            echo "    Please set MTHOME to point to a directory such that"
            echo "    the file \$MTHOME/include/vapi.h exists"
            exit 1            
        fi
    fi
fi
                                             
if [[ "$v" == "sim" ]]; then
    if test -z $IBMGTSIMHOME; then
        IBMGTSIMHOME=/usr
        export IBMGTSIMHOME;
    fi
    
    if test ! -f $IBMGTSIMHOME/include/ibmgtsim/ibms_client_api.h; then 
        echo "-E- Wrong IBMGTSIMHOME env variable"
        echo "    Please set $IBMGTSIMHOME to point to a directory such that"
        echo "    the file \$IBMGTSIMHOME/include/ibmgtsim/ibms_client_api.h exists"
        exit 1 
    fi
fi

# make sure we have the tmp dir
if [ ! -d $t ]; then
	 echo "Temporary Build dir:$t does not exist. Creating it ..."
	 mkdir -p $t
else
	\rm -rf $t/*
fi

# Open the tar file into the tmp dir:
cd $t 
tar xfz $i/osm-linux.tgz >& /dev/null

# Modify osm_base.h with current change set and BUILD ID:
buildid=`cat $i/BUILD_ID`
if [ $d != 0 ]; then
    buildid="${buildid}-debug"
fi
bf=osm/opensm/osm_base.h
\mv $bf $bf.s
sed 's=define OSM_VERSION "\(.*\)"=define OSM_VERSION "\1 '$buildid'"=' $bf.s > $bf
\rm -f $bf.s

# we rely on the fact we have opened the tar file so 
# have a local base dir:
if [ ! -d osm ]; then 
	 echo "Expected to find a sub-directory named 'base' in current dir"
 	 echo "TAR file probably was not opened locally!"
	 exit
fi
if [[ "$COVER" == 1 ]]; then
    DO_COV=YES
    export DO_COV;
    if [ "$v" != "ts" ]; then 
        echo "-E- Could not continue , coverage mode is enabled with TopSpin stack only"
        exit
    fi
    v=ts_no_vapi
else
    unset DO_COV
fi

export MAKELOG=make.log
export EXPECTED_MAKELOG_SIZE=85233
PLATF=`uname -m`
case $PLATF in
    "i686")
         EXPECTED_MAKELOG_SIZE=90000
         ;;
    "ia64")
         EXPECTED_MAKELOG_SIZE=78000
         ;;
    "x86_64")
         EXPECTED_MAKELOG_SIZE=79000
         ;;
    *)
esac


# the make file uses PREFIX in the rpath (shared object link path)
export PREFIX=$orig_p

cd $t/osm

make clean > /dev/null 2>&1 

rm -f $MAKELOG
if [ $d != 0 ]; then
	 echo "Making DEBUG Version OpenSM (in $t) ...."
	 make VENDOR=$v INSTALL_BASE_LIB=$p/lib INSTALL_BASE=$p/bin debug >> $MAKELOG 2>&1  &
else
	 echo "Making OpenSM (in $t) ...."
	 make VENDOR=$v INSTALL_BASE_LIB=$p/lib INSTALL_BASE=$p/bin free >> $MAKELOG 2>&1  &
fi

make_pid=`ps -eo "%P %p %a" | awk '(($1 == '$$') && ($3 == "make")) {print $2}'`
# if running in batch mode - do not run fprogbar
if [ $a == 0 ]; then
    $i/fprogbar.sh $MAKELOG $EXPECTED_MAKELOG_SIZE $make_pid
fi

wait $make_pid >/dev/null 2>&1 
if [ $? != 0 ]; then 
	 echo "Make failed. For more Information look at: `pwd`/$MAKELOG"
	 exit 1
fi

# if BUILD_ID file exists - copy it into install dir
if [ -f $i/BUILD_ID ]; then
 \cp $i/BUILD_ID $t
fi

# we need to make sure the executables are only evaluated
# after a LD_PRELOAD if TLS pthread is available. So
# we always place a bash script instead of teh compiled code
do_ld_preload_wrapper

# linking should be done only if build_only flag is not set
if [ $o == 0 ]; then
    # try to link in /usr/bin (will fail if not root ...)
    rm -f /usr/bin/opensm >& /dev/null
    ln -s $p/bin/opensm  /usr/bin/opensm >& /dev/null
    rm -f /usr/bin/osmtest >& /dev/null
    ln -s $p/bin/osmtest  /usr/bin/osmtest >& /dev/null
    
    # link the osm lib
    rm -f /usr/lib/osm1.0 >& /dev/null
    if [ -d $p/lib/osm1.0 ]; then    
	ln -s -f $p/lib/osm1.0 /usr/lib/osm1.0 >& /dev/null
    fi
    rm -f /usr/lib/updn1.0 >& /dev/null
    if [ -d $p/lib/updn1.0 ]; then
	ln -s -f $p/lib/updn1.0 /usr/lib/updn1.0 >& /dev/null
    fi
    echo "Installation of OpenSM completed successfuly"
else
    echo "Running with build_only flag. No installation done"
fi


if [ -f /mswg/projects/osm/bld/stats ]; then
    if [ "$d" == 0 ]; then 
	echo 'NO_DEBUG' `date +%H%M-%d%m%y` `wc $MAKELOG` `uname -a`  >> /mswg/projects/osm/bld/stats
    else
	echo 'DEBUG' `date +%H%M-%d%m%y` `wc $MAKELOG` `uname -a` >> /mswg/projects/osm/bld/stats
    fi
fi

if [ $c != 0 ]; then
    echo "Cleaning up temporary build tree"
    cd /tmp
    \rm -rf $t
fi

exit 0

