

/*                  - Mellanox Confidential and Proprietary -
 *
 *  Copyright (C) June 2000, Mellanox Technologies Ltd.  ALL RIGHTS RESERVED.
 *
 *  Except as specifically permitted herein, no portion of the information,
 *  including but not limited to object code and source code, may be reproduced,
 *  modified, distributed, republished or otherwise exploited in any form or by
 *  any means for any purpose without the prior written permission of Mellanox
 *  Technologies Ltd. Use of software subject to the terms and conditions
 *  detailed in the file "LICENSE.txt".
 *
 */                                                                                   
 
 
/*
*-------------------------------------------------------------------
* Description: Command line parsing
*-------------------------------------------------------------------								 
* Version : $Id: cmd_pars.c,v 1.1 2003/10/29 17:10:11 dotanb Exp $									   
* Author  : Dotan Barak
*-------------------------------------------------------------------
*/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <vapi_types.h>

#include "types.h"

#define CMD_PARS_MODULE
 #include "cmd_pars.h"

//-----------------------------------------------------------------------------
// Types Definition
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
// Constants
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
// Global Variables
//-----------------------------------------------------------------------------
extern config_t config;
//-----------------------------------------------------------------------------
// Global Function Definition
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
// Local Function Definition
//-----------------------------------------------------------------------------
static int process_arg( int opt_index, char* equ_ptr );
static void usage( int full_help );

/*
*-----------------------------------------------------------------------------
*								parse_argv
*-----------------------------------------------------------------------------
* Dscr:  parse argv list and call process_arg for each option
*        fill object_name object_argv, object_argc values   
*-----------------------------------------------------------------------------
* Input : argc, argv 
* Output: 0 - continue
*		  <0 - error code to exit 
*         >0 - code to exit
* Notes : call exit(1) in case of cmd line error  
*-----------------------------------------------------------------------------
*/
int parse_argv( int my_argc, char** my_argv )
{
	int i,i2,len;
	int rc;
	char* equ_ptr = 0;

	//
	// Parse command line untill object name found
	//
	for( i=1; i<my_argc; i++ )
	{
		if( my_argv[i][0] != '-' )
		{
			
			printf( "Error, Wrong parameter \"%s\"\n", my_argv[i] );
			usage( 0 );
			return -1;
		}

		equ_ptr = 0;

		if( my_argv[i][1] != '-' )	// option in short form "-x"
		{
			for( i2=0; i2<OPTIONS_N; i2++ )
				if( my_argv[i][1] == usage_descriptor[i2].m_option )
				{

                    // Determine option value 
					if( usage_descriptor[i2].equ_alias[0] != '\0' )
					{
						if( my_argv[i][2] != '\0' )
						{
							equ_ptr = &my_argv[i][2];	// Option in form -xVALUE
							if( *equ_ptr == '=' )
								equ_ptr++;				// Option in form -x=VALUE
						}
						else
						{								// Option in form -x VALUE
							// Take next argument as value
							if( (i+1 == my_argc) || (my_argv[i+1][0] == '-') )
							{
								// This was last argument or next argument starts with '-'
								printf( "Error, Option %s is not complete\n", my_argv[i] );
								usage( 0 );
								return -1;
							}

							i++;
							equ_ptr = my_argv[i];
						}
					}

					rc = process_arg( i2, equ_ptr );
					if( rc )
						return rc;
					break;
				}
		}
		else						// option in verbose form "--option"
		{
			for( i2=0; i2<OPTIONS_N; i2++ )
			{
				len = strlen( usage_descriptor[i2].mm_option );
				if( !strncmp(usage_descriptor[i2].mm_option, &my_argv[i][2], len) )
				{
					if( (my_argv[i][2+len]=='\0') || (my_argv[i][2+len]=='=') )
					{

						// Split argument to option=value if needed

						if( usage_descriptor[i2].equ_alias[0] != '\0' )
						{
							equ_ptr = strchr( my_argv[i], '=' );
							if( !equ_ptr || !strlen( equ_ptr+1 ) )
							{
								// format option=value not found
								printf( "Error, Option %s is not complete\n", my_argv[i] );
								usage( 0 );
								return -1;
							}
							equ_ptr++;
						}

						rc = process_arg( i2, equ_ptr );
						if( rc )
							return rc;
						break;
					}
				}
			}
		}

		if( i2 == OPTIONS_N )
		{
			printf( "Error, Unknown option %s\n", my_argv[i] );
			usage( 0 );
			return -1;
		}
	}/*end for*/

	return 0;
}


/*
*-----------------------------------------------------------------------------
*								usage
*-----------------------------------------------------------------------------
* Dscr:  Output usage message
*-----------------------------------------------------------------------------
* Input : 
* Output:
* Notes :
*-----------------------------------------------------------------------------
*/
void usage( int full_help )
{
	int i;
	int cntr;

    if( full_help )
	{
		for( i=0; i<OPTIONS_N; i++ )
		{
			if( usage_descriptor[i].help_desc[0] == '\0' )
				continue;			// Skip options without help context

			if( usage_descriptor[i].m_option != ' ' )
				printf( "-%c, ",usage_descriptor[i].m_option );
			else 
				printf( "    " );

			printf("--");
			cntr = printf( "%.16s", usage_descriptor[i].mm_option );
			if( usage_descriptor[i].equ_alias[0] )
				cntr+=printf( "=%-10s", usage_descriptor[i].equ_alias );

			while( cntr++ < 18 )
				putchar(' ');

			printf( " %s\n" ,usage_descriptor[i].help_desc );  
		}
	}
	else
		printf( "  Type --help|-h to get more information\n" );
}


/*
*-----------------------------------------------------------------------------
*								process_arg
*-----------------------------------------------------------------------------
* Dscr: process argument  
*-----------------------------------------------------------------------------
* Input : option index, equ_ptr - pointer to option value
* Output: 0 - continue
*		  <0 - error code to exit 
*         >0 - code to exit
* Notes : call exit(1) in case of error in argument
*-----------------------------------------------------------------------------
*/
int process_arg( int opt_index, char* equ_ptr )
{
	int rc=0;

	//
	// process argument
	//
	switch ( usage_descriptor[opt_index].case_code )
	{
	case HELP_CMD_CASE:
		usage( 1 );
		return 1;

    case HCA_N_CMD_CASE:
		strncpy( config.hca_id, equ_ptr, HCA_MAXNAME );
		break;
    
    case DAEMON_CMD_CASE:
		config.is_daemon = 1;
		break;

	case IP_PORT_CMD_CASE:
		config.tcp_port = strtol( equ_ptr, 0, 0 );
		break;

    case IP_ADDR_CMD_CASE:
		config.ip = equ_ptr;
		break;

	case PORT_CMD_CASE:
		config.ib_port = strtoul( equ_ptr, 0, 0 );
		break;

	case ALT_PORT_CMD_CASE:
		config.alt_ib_port = strtoul( equ_ptr, 0, 0 );
		break;

    default:
        printf("Error, unknown parameter in the switch: %s\n", equ_ptr);
        exit(2);
    }/* end case*/

	return rc;
}

