/*
  This software is available to you under a choice of one of two
  licenses.  You may choose to be licensed under the terms of the GNU
  General Public License (GPL) Version 2, available at
  <http://www.fsf.org/copyleft/gpl.html>, or the OpenIB.org BSD
  license, available in the LICENSE.TXT file accompanying this
  software.  These details are also available at
  <http://openib.org/license.html>.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
  BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
  ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
  CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
  SOFTWARE.

  Copyright (c) 2004 Mellanox Technologies Ltd.  All rights reserved.
*/
 
#include <cmdif.h>
#include <cmd_types.h>
#include <cmdif_priv.h>
#include <sm_mad.h>
#include <mtl_common.h>
#include <cmds_wrap_priv.h>
#include <dev_cmds_wrap.h>

#if defined(MAX_DEBUG) && XHH_CMDS_WRAP_DEBUG_LEVEL <= MAX_DEBUG

#define XHH_CMD_PRINT_DEV_LIMS(a)      XHH_cmd_print_dev_lims(a)
#define XHH_CMD_PRINT_HCA_PROPS(a)     XHH_cmd_print_hca_props(a)
#define XHH_CMD_PRINT_INIT_IB(a, b)    XHH_cmd_print_init_ib(a, b)
#define XHH_CMD_PRINT_QUERY_ADAPTER(a) XHH_cmd_print_query_adapter(a)
#define XHH_CMD_PRINT_QUERY_FW(a)      XHH_cmd_print_query_fw(a)
#define XHH_CMD_PRINT_CQ_CONTEXT(a)    XHH_cmd_print_cq_context(a)
#define XHH_CMD_PRINT_QP_CONTEXT(a)    XHH_cmd_print_qp_context(a)
#define XHH_CMD_PRINT_EQ_CONTEXT(a)    XHH_cmd_print_eq_context(a)
#define XHH_CMD_PRINT_MPT_ENTRY(a)     XHH_cmd_print_mpt_entry(a)
#define XHH_CMD_PRINT_MTT_ENTRIES(a,b) XHH_cmd_print_mtt_entries(a,b)
#define XHH_CMD_PRINT_MGM_ENTRY(a)     XHH_cmd_print_mgm_entry(a)
#define XHH_CMD_MAILBOX_PRINT(a,b,c)   XHH_cmd_mailbox_print(a,b,c)


#else
#define XHH_CMD_PRINT_DEV_LIMS(a)      
#define XHH_CMD_PRINT_HCA_PROPS(a)     
#define XHH_CMD_PRINT_INIT_IB(a, b)    
#define XHH_CMD_PRINT_QUERY_ADAPTER(a) 
#define XHH_CMD_PRINT_QUERY_DDR(a)     
#define XHH_CMD_PRINT_QUERY_FW(a) 
#define XHH_CMD_PRINT_CQ_CONTEXT(a)
#define XHH_CMD_PRINT_QP_CONTEXT(a)
#define XHH_CMD_PRINT_EQ_CONTEXT(a)
#define XHH_CMD_PRINT_MPT_ENTRY(a)
#define XHH_CMD_PRINT_MTT_ENTRIES(a,b)
#define XHH_CMD_MAILBOX_PRINT(a,b,c) 
#define XHH_CMD_PRINT_MGM_ENTRY(a)
#endif

/***************************************************** */
/************** CMD INTERFACE  UTILITIES ************* */
/***************************************************** */


static MT_bool XHH_tavor_qpstate_2_vapi_qpstate(ADD_DEV_PREFIX_LC(if_qp_state_t) device_qp_state,
                                                        VAPI_qp_state_t * vapi_qp_state)
{
    switch(device_qp_state) {
    case ADD_DEV_PREFIX(IF_QP_STATE_RESET):
        *vapi_qp_state = VAPI_RESET;
        break;

    case ADD_DEV_PREFIX(IF_QP_STATE_INIT):
        *vapi_qp_state = VAPI_INIT;
        break;

    case ADD_DEV_PREFIX(IF_QP_STATE_RTR):
        *vapi_qp_state = VAPI_RTR;
        break;

    case ADD_DEV_PREFIX(IF_QP_STATE_RTS):
        *vapi_qp_state = VAPI_RTS;
        break;

    case ADD_DEV_PREFIX(IF_QP_STATE_SQER):
        *vapi_qp_state = VAPI_SQE;
        break;

    /* T.D.(matan): change this.*/
	case ADD_DEV_PREFIX(IF_QP_STATE_DRAINING):
        *vapi_qp_state = (VAPI_qp_state_t)(VAPI_SQD | XHH_CMD_QP_DRAINING_FLAG);
		break;
    case ADD_DEV_PREFIX(IF_QP_STATE_SQD):
        *vapi_qp_state = VAPI_SQD;
        break;

    case ADD_DEV_PREFIX(IF_QP_STATE_ERR):
        *vapi_qp_state = VAPI_ERR;
        break;
    default:
        return FALSE;
    }
    return TRUE;
}
static  MT_bool XHH_vapi_qpstate_2_tavor_qpstate(VAPI_qp_state_t vapi_qp_state,
                                         ADD_DEV_PREFIX_LC(if_qp_state_t) *device_qp_state)
{
    switch(vapi_qp_state) {
    case VAPI_RESET:
        *device_qp_state = ADD_DEV_PREFIX(IF_QP_STATE_RESET);
        break;

    case VAPI_INIT:
        *device_qp_state = ADD_DEV_PREFIX(IF_QP_STATE_INIT);
        break;

    case VAPI_RTR:
        *device_qp_state = ADD_DEV_PREFIX(IF_QP_STATE_RTR);
        break;

    case VAPI_RTS:
        *device_qp_state = ADD_DEV_PREFIX(IF_QP_STATE_RTS);
        break;

    case VAPI_SQE:
        *device_qp_state = ADD_DEV_PREFIX(IF_QP_STATE_SQER);
        break;

    case VAPI_SQD:
        *device_qp_state = ADD_DEV_PREFIX(IF_QP_STATE_SQD);
        break;

    case VAPI_ERR:
        *device_qp_state = ADD_DEV_PREFIX(IF_QP_STATE_ERR);
        break;
    default:
        return FALSE;
    }
    return TRUE;
}
                                                      

/***************************************************** */
/********** CMD INTERFACE PRINT UTILITIES **************/
/***************************************************** */

//--------------------------------------------------------------------------------------------------

/*
 *  print_mtt_entries
 */
static void my_print_mtt_entries(u_int32_t elts_this_loop, void *inprm)
{
  u_int32_t *p=(u_int32_t *)((u_int8_t *)inprm+MTT_CMD_PA_PREAMBLE_SIZE), i, 
      *pp = (u_int32_t *)inprm;
  u_int64_t tag;
  int present;
  return;
  MTL_ERROR1("mtt_pa = "U64_FMT"\n", (((u_int64_t)(*pp))<<32)+*(pp+1));
  for ( i=0; i<elts_this_loop; ++i ) {
    tag = ((((u_int64_t)(*p))<<32) + (*(p+1))) & MAKE_LONGLONG(0xfffffffffffff000);
    present = (*(p+1)) & 1;
    p+=2;
    MTL_ERROR1("MTT entry: "U64_FMT", %s\n", tag, present ? "present" : "non present");
  }
}
#if defined(MAX_DEBUG) && XHH_CMDS_WRAP_DEBUG_LEVEL <= MAX_DEBUG

static void XHH_cmd_mailbox_print( void *mailbox, int size, const char * func )
{
    int i, j, maxlines, bytes_left, this_line;
    char linebuf[200], tempout[20];
    u_int8_t *iterator;

    if (mailbox == NULL) return;
    
    iterator = (u_int8_t *)mailbox;
    bytes_left = size;
    MTL_DEBUG4("MailboxPrint from function %s, starting at addr 0x%p, size=%d:\n",
               func, mailbox, size);
    
    if (size <= 0) {
        return;
    }

    maxlines = (size / 16) + ((size % 16) ? 1 : 0);

    for (i = 0; i < maxlines; i++) {
        memset(linebuf, 0, sizeof(linebuf));
        this_line = (bytes_left > 16 ? 16 : bytes_left);

        for (j = 0; j < this_line; j++) {
            if ((j % 4) == 0) {
                strcat(linebuf," ");
            }
            sprintf(tempout, "%02x", *iterator);
            iterator++; bytes_left--;
            strcat(linebuf,tempout);
        }
        MTL_DEBUG4("%s\n", linebuf);
    }
    MTL_DEBUG3("MailboxPrint END\n");
}

static void XHH_cmd_print_mtt_entries(u_int32_t elts_this_loop, void *inprm)
{
  u_int32_t *p=(u_int32_t *)((u_int8_t *)inprm+MTT_CMD_PA_PREAMBLE_SIZE), i, 
      *pp = (u_int32_t *)inprm;
  u_int64_t tag;
  int present;
  return;
  MTL_DEBUG1("mtt_pa = "U64_FMT"\n", (((u_int64_t)(*pp))<<32)+*(pp+1));
  for ( i=0; i<elts_this_loop; ++i ) {
    tag = ((((u_int64_t)(*p))<<32) + (*(p+1))) & MAKE_LONGLONG(0xfffffffffffff000);
    present = (*(p+1)) & 1;
    p+=2;
    MTL_DEBUG1("MTT entry: "U64_FMT", %s\n", tag, present ? "present" : "non present");
  }
}

void XHH_cmd_print_hca_props(XHH_hca_props_t *hca_props)
{
    CMDS_DBG("HCA PROPS DUMP (XHH_hca_props_t structure)\n");

    CMDS_DBG("hca_core_click = %d\n", hca_props->hca_core_clock);

    CMDS_DBG( "he (host endian) = %s\n", hca_props->he ? "Big Endian" : "Little Endian");

    CMDS_DBG( "re (Router Mode Enable) = %s\n", hca_props->re ? "TRUE" :  "FALSE");

    CMDS_DBG( "router_qp = %s\n", hca_props->router_qp ? "TRUE" :  "FALSE");

    CMDS_DBG( "ud (UD address vector protection) = %s\n", hca_props->ud ? "TRUE" :  "FALSE");
    
    CMDS_DBG( "udp (UDP port check enabled) = %s\n", hca_props->udp ? "TRUE" :  "FALSE");
    
    /* multicast parameters */
    CMDS_DBG( "\nmulticast_parameters.log_mc_table_entry_sz = %d\n", hca_props->multicast_parameters.log_mc_table_entry_sz);
    CMDS_DBG( "multicast_parameters.log_mc_table_sz = %d\n", hca_props->multicast_parameters.log_mc_table_sz);
    CMDS_DBG( "multicast_parameters.mc_base_addr = "U64_FMT"\n", hca_props->multicast_parameters.mc_base_addr);
    CMDS_DBG( "multicast_parameters.mc_hash_fn = %d\n", hca_props->multicast_parameters.mc_hash_fn);
    CMDS_DBG( "multicast_parameters.mc_table_hash_sz = %d\n", hca_props->multicast_parameters.mc_table_hash_sz);

    /* QP,EEC, EQC, RDB, CQC parameters */
    CMDS_DBG( "\nqpc_eec_cqc_eqc_rdb_parameters.cqc_base_addr = "U64_FMT"\n", hca_props->qpc_eec_cqc_eqc_rdb_parameters.cqc_base_addr);
    CMDS_DBG( "qpc_eec_cqc_eqc_rdb_parameters.eec_base_addr = "U64_FMT"\n", hca_props->qpc_eec_cqc_eqc_rdb_parameters.eec_base_addr);
    CMDS_DBG( "qpc_eec_cqc_eqc_rdb_parameters.eeec_base_addr = "U64_FMT"\n", hca_props->qpc_eec_cqc_eqc_rdb_parameters.eeec_base_addr);
    CMDS_DBG( "qpc_eec_cqc_eqc_rdb_parameters.eqc_base_addr = "U64_FMT"\n", hca_props->qpc_eec_cqc_eqc_rdb_parameters.eqc_base_addr);
    CMDS_DBG( "qpc_eec_cqc_eqc_rdb_parameters.eqpc_base_addr = "U64_FMT"\n", hca_props->qpc_eec_cqc_eqc_rdb_parameters.eqpc_base_addr);
    CMDS_DBG( "qpc_eec_cqc_eqc_rdb_parameters.qpc_base_addr = "U64_FMT"\n", hca_props->qpc_eec_cqc_eqc_rdb_parameters.qpc_base_addr);
    CMDS_DBG( "qpc_eec_cqc_eqc_rdb_parameters.rdb_base_addr = "U64_FMT"\n", hca_props->qpc_eec_cqc_eqc_rdb_parameters.rdb_base_addr);
    CMDS_DBG( "qpc_eec_cqc_eqc_rdb_parameters.log_num_eq = %d\n", hca_props->qpc_eec_cqc_eqc_rdb_parameters.log_num_eq);
    CMDS_DBG( "qpc_eec_cqc_eqc_rdb_parameters.log_num_of_cq = %d\n", hca_props->qpc_eec_cqc_eqc_rdb_parameters.log_num_of_cq);
    CMDS_DBG( "qpc_eec_cqc_eqc_rdb_parameters.log_num_of_ee = %d\n", hca_props->qpc_eec_cqc_eqc_rdb_parameters.log_num_of_ee);
    CMDS_DBG( "qpc_eec_cqc_eqc_rdb_parameters.log_num_of_qp = %d\n", hca_props->qpc_eec_cqc_eqc_rdb_parameters.log_num_of_qp);

    /* TPT parameters */
    CMDS_DBG( "\ntpt_parameters.mpt_base_adr = "U64_FMT"\n", hca_props->tpt_parameters.mpt_base_adr);
    CMDS_DBG( "tpt_parameters.mtt_base_adr = "U64_FMT"\n", hca_props->tpt_parameters.mtt_base_addr);
    CMDS_DBG( "tpt_parameters.log_mpt_sz = %d\n", hca_props->tpt_parameters.log_mpt_sz);
    CMDS_DBG( "tpt_parameters.pfto = %d\n", hca_props->tpt_parameters.pfto);

    /* UAR parameters */
    CMDS_DBG( "\nuar_parameters.uar_base_addr = "U64_FMT"\n", hca_props->uar_parameters.uar_base_addr);
    CMDS_DBG( "uar_parameters.uar_scratch_base_addr = "U64_FMT"\n", hca_props->uar_parameters.uar_scratch_base_addr);
    CMDS_DBG( "uar_parameters.uar_page_sz = %d\n", hca_props->uar_parameters.uar_page_sz);

    /* UDAV parameters */
    CMDS_DBG( "\nudavtable_memory_parameters.l_key = 0x%x\n", hca_props->udavtable_memory_parameters.l_key);
    CMDS_DBG( "udavtable_memory_parameters.pd = %d\n", hca_props->udavtable_memory_parameters.pd);
    CMDS_DBG( "udavtable_memory_parameters.xlation_en = %s\n", hca_props->udavtable_memory_parameters.xlation_en ? "TRUE" : "FALSE");
    XHH_DEV_SPECIFIC_PRINT_HCA_PROPS
}

void XHH_cmd_print_dev_lims(XHH_dev_lim_t *dev_lim)
{
    CMDS_DBG("QUERY DEV LIMS DUMP (XHH_dev_lim_t structure)\n");

    CMDS_DBG( "dev_lim->apm (Automatic Path Migration) = %s\n", dev_lim->apm ? "TRUE" :  "FALSE");
    CMDS_DBG( "dev_lim->atm (Atomic Operations) = %s\n", dev_lim->atm ? "TRUE" :  "FALSE");
    CMDS_DBG( "dev_lim->avp (Address Vector port checking) = %s\n", dev_lim->avp ? "TRUE" :  "FALSE");
    CMDS_DBG( "dev_lim->udm (UD Multicast support) = %s\n", dev_lim->udm ? "TRUE" :  "FALSE");
    CMDS_DBG( "dev_lim->mw (Memory Windows) = %s\n", dev_lim->mw ? "TRUE" :  "FALSE");
    CMDS_DBG( "dev_lim->pg (Paging on-demand) = %s\n", dev_lim->pg ? "TRUE" :  "FALSE");
    CMDS_DBG( "dev_lim->pkv (PKey Violation Counter) = %s\n", dev_lim->pkv ? "TRUE" :  "FALSE");
    CMDS_DBG( "dev_lim->pkv (PKey Violation Counter) = %s\n", dev_lim->pkv ? "TRUE" :  "FALSE");
    CMDS_DBG( "dev_lim->qkv (QKey Violation Counter) = %s\n", dev_lim->qkv ? "TRUE" :  "FALSE");
    CMDS_DBG( "dev_lim->r (Router Mode) = %s\n", dev_lim->r ? "TRUE" :  "FALSE");
    CMDS_DBG( "dev_lim->raw_ether (Raw Ethernet mode) = %s\n", dev_lim->raw_ether ? "TRUE" :  "FALSE");
    CMDS_DBG( "dev_lim->raw_ipv6 (Raw IpV6 mode) = %s\n", dev_lim->raw_ipv6 ? "TRUE" :  "FALSE");
    CMDS_DBG( "dev_lim->rc (RC Transport) = %s\n", dev_lim->rc ? "TRUE" :  "FALSE");
    CMDS_DBG( "dev_lim->rd (RD Transport) = %s\n", dev_lim->rd ? "TRUE" :  "FALSE");
    CMDS_DBG( "dev_lim->uc (UC Transport) = %s\n", dev_lim->uc ? "TRUE" :  "FALSE");
    CMDS_DBG( "dev_lim->ud (UD Transport) = %s\n", dev_lim->ud ? "TRUE" :  "FALSE");
    CMDS_DBG( "dev_lim->rm (Raw Multicast) = %s\n", dev_lim->rm ? "TRUE" :  "FALSE");
    /* multicast parameters */
    CMDS_DBG( "dev_lim->cqc_entry_sz = %d\n",dev_lim->cqc_entry_sz);
    CMDS_DBG( "dev_lim->eec_entry_sz = %d\n",dev_lim->eec_entry_sz);
    CMDS_DBG( "dev_lim->eeec_entry_sz = %d\n",dev_lim->eeec_entry_sz);
    CMDS_DBG( "dev_lim->eqc_entry_sz = %d\n",dev_lim->eqc_entry_sz);
    CMDS_DBG( "dev_lim->eqpc_entry_sz = %d\n",dev_lim->eqpc_entry_sz);
    CMDS_DBG( "dev_lim->log_max_cq_sz = %d\n",dev_lim->log_max_cq_sz);
    CMDS_DBG( "dev_lim->qpc_entry_sz = %d\n",dev_lim->qpc_entry_sz);
    CMDS_DBG( "dev_lim->max_uar_sz = %d\n",dev_lim->uar_sz);
    CMDS_DBG( "dev_lim->uar_scratch_entry_sz = %d\n",dev_lim->uar_scratch_entry_sz);
    CMDS_DBG( "dev_lim->log_max_cq = %d\n",dev_lim->log_max_cq);
    CMDS_DBG( "dev_lim->log_max_cq_sz = %d\n",dev_lim->log_max_cq_sz);
    CMDS_DBG( "dev_lim->log_max_ee = %d\n",dev_lim->log_max_ee);
    CMDS_DBG( "dev_lim->log_max_eq = %d\n",dev_lim->log_max_eq);
    CMDS_DBG( "dev_lim->log_max_gid = %d\n",dev_lim->log_max_gid);
    CMDS_DBG( "dev_lim->log_max_mcg = %d\n",dev_lim->log_max_mcg);
    CMDS_DBG( "dev_lim->log_max_mpts = %d\n",dev_lim->log_max_mpts);
    CMDS_DBG( "dev_lim->log_max_mrw_sz = %d\n",dev_lim->log_max_mrw_sz);
    CMDS_DBG( "dev_lim->log_max_pd = %d\n",dev_lim->log_max_pd);
    CMDS_DBG( "dev_lim->log_max_pkey = %d\n",dev_lim->log_max_pkey);
    CMDS_DBG( "dev_lim->log_max_qp = %d\n",dev_lim->log_max_qp);
    CMDS_DBG( "dev_lim->log_max_qp_mcg = %d\n",dev_lim->log_max_qp_mcg);
    CMDS_DBG( "dev_lim->log_max_qp_sz = %d\n",dev_lim->log_max_qp_sz);
    CMDS_DBG( "dev_lim->log_max_ra_req_qp = %d\n",dev_lim->log_max_ra_req_qp);
    CMDS_DBG( "dev_lim->log_max_ra_res_qp = %d\n",dev_lim->log_max_ra_res_qp);
    CMDS_DBG( "dev_lim->log_max_ra_res_global = %d\n",dev_lim->log_max_ra_res_global);
    CMDS_DBG( "dev_lim->log_max_rdds = %d\n",dev_lim->log_max_rdds);
    CMDS_DBG( "dev_lim->log_pg_sz = %d\n",dev_lim->log_pg_sz);
    CMDS_DBG( "dev_lim->max_mtu = %d\n",dev_lim->max_mtu);
    CMDS_DBG( "dev_lim->max_port_width = %d\n",dev_lim->max_port_width);
    CMDS_DBG( "dev_lim->max_vl = %d\n",dev_lim->max_vl);
    CMDS_DBG( "dev_lim->num_ports = %d\n",dev_lim->num_ports);
    CMDS_DBG( "dev_lim->log2_rsvd_qps = %d\n",dev_lim->log2_rsvd_qps);
    CMDS_DBG( "dev_lim->log2_rsvd_ees = %d\n",dev_lim->log2_rsvd_ees);
    CMDS_DBG( "dev_lim->log2_rsvd_cqs = %d\n",dev_lim->log2_rsvd_cqs);
    CMDS_DBG( "dev_lim->num_rsvd_eqs = %d\n",dev_lim->num_rsvd_eqs);
    CMDS_DBG( "dev_lim->log2_rsvd_mrws = %d\n",dev_lim->log2_rsvd_mrws);
    CMDS_DBG( "dev_lim->log2_rsvd_mtts = %d\n",dev_lim->log2_rsvd_mtts);
    CMDS_DBG( "dev_lim->num_rsvd_uars = %d\n",dev_lim->num_rsvd_uars);
    CMDS_DBG( "dev_lim->num_rsvd_rdds = %d\n",dev_lim->num_rsvd_rdds);
    CMDS_DBG( "dev_lim->num_rsvd_pds = %d\n",dev_lim->num_rsvd_pds);
    CMDS_DBG( "dev_lim->local_ca_ack_delay = %d\n",dev_lim->local_ca_ack_delay);
    XHH_DEV_SPECIFIC_PRINT_DEV_LIM
}

void XHH_cmd_print_query_fw(XHH_fw_props_t *fw_props)
{
    CMDS_DBG("QUERY FW DUMP (XHH_fw_props_t structure)\n");
    CMDS_DBG( "fw_props->cmd_interface_rev = 0x%x\n", fw_props->cmd_interface_rev);
    CMDS_DBG( "fw_props->fw_rev_major = 0x%x\n", fw_props->fw_rev_major);
    CMDS_DBG( "fw_props->fw_rev_minor = 0x%x\n", fw_props->fw_rev_minor);
    CMDS_DBG( "fw_props->fw_rev_subminor = 0x%x\n", fw_props->fw_rev_subminor);
    CMDS_DBG( "fw_props->error_buf_start = "U64_FMT"\n", fw_props->error_buf_start);
    CMDS_DBG( "fw_props->error_buf_size = %d\n", fw_props->error_buf_size);
    XHH_DEV_SPECIFIC_PRINT_QUERY_FW
}

void XHH_cmd_print_query_adapter( XHH_adapter_props_t *adapter_props)
{
    int i;
    CMDS_DBG("QUERY ADAPTER DUMP (XHH_adapter_props_t structure)\n");
#if 0
    CMDS_DBG( "adapter_props->device_id = %d\n", adapter_props->device_id);
    CMDS_DBG( "adapter_props->vendor_id = %d\n", adapter_props->vendor_id);
    CMDS_DBG( "adapter_props->revision_id = %d\n", adapter_props->revision_id);
#endif    
    CMDS_DBG( "adapter_props->intapin = %d\n", adapter_props->intapin);
    if (XHH_ADAPTER_VSD_DW>0) {
      for (i=0; i<XHH_ADAPTER_VSD_DW; i++) {
	  CMDS_DBG( "adapter_props->vsd[i] = %c\n", adapter_props->vsd[i]);
      }
    }
}


void XHH_cmd_print_init_ib(IB_port_t port, XHH_port_init_props_t *port_init_props)
{
    CMDS_DBG("INIT_IB DUMP (XHH_port_init_props_t structure) for port %d\n", port);
    CMDS_DBG( "port_init_props->max_gid = %d\n", port_init_props->max_gid);
    CMDS_DBG( "port_init_props->max_pkey = %d\n", port_init_props->max_pkey);
    CMDS_DBG( "port_init_props->mtu_cap = 0x%x\n", port_init_props->mtu_cap);
    CMDS_DBG( "port_init_props->port_width_cap = 0x%x\n", port_init_props->port_width_cap);
    CMDS_DBG( "port_init_props->vl_cap = 0x%x\n", port_init_props->vl_cap);
    CMDS_DBG( "port_init_props->g0 = %s\n", port_init_props->g0 ? "TRUE" : "FALSE");
    CMDS_DBG( "port_init_props->guid0 = 0x%2x%2x%2x%2x%2x%2x%2x%2x",
                port_init_props->guid0[0], port_init_props->guid0[1], port_init_props->guid0[2], 
                port_init_props->guid0[3], port_init_props->guid0[4], port_init_props->guid0[5], 
                port_init_props->guid0[6], port_init_props->guid0[7] );
}

void XHH_cmd_print_cq_context(XHH_cqc_t *cqc)
{
    return;
    CMDS_DBG("CQ CONTEXT DUMP (XHH_cqc_t structure)\n");
    CMDS_DBG( "cqc->st = 0x%x\n", cqc->st);
    CMDS_DBG( "cqc->oi (overrun ignore) = %s\n",  cqc->oi ? "TRUE" : "FALSE");
    CMDS_DBG( "cqc->tr (translation required) = %s\n",  cqc->tr ? "TRUE" : "FALSE");
    CMDS_DBG( "cqc->status = 0x%x\n",  cqc->status);
    CMDS_DBG( "cqc->start_address = "U64_FMT"\n",  cqc->start_address);
    CMDS_DBG( "cqc->usr_page = 0x%X\n",  cqc->usr_page);
    CMDS_DBG( "cqc->log_cq_size = %d\n",  cqc->log_cq_size);
    CMDS_DBG( "cqc->c_eqn = 0x%X\n",  cqc->c_eqn);
    CMDS_DBG( "cqc->e_eqn = 0x%X\n",  cqc->e_eqn);
    CMDS_DBG( "cqc->pd = 0x%X\n",  cqc->pd);
    CMDS_DBG( "cqc->l_key = 0x%X\n",  cqc->l_key);
    CMDS_DBG( "cqc->last_notified_indx = 0x%X\n",  cqc->last_notified_indx);
    CMDS_DBG( "cqc->solicit_producer_indx = 0x%X\n",  cqc->solicit_producer_indx);
    CMDS_DBG( "cqc->cqn = 0x%X\n",  cqc->cqn);
    XHH_DEV_SPECIFIC_PRINT_CQ_CONTEXT
}
void XHH_cmd_print_qp_context(XHH_qpee_context_t *qpc)
{
    return;
    CMDS_DBG("QPEE CONTEXT DUMP (XHH_qpee_context_t structure)\n");
    CMDS_DBG( "QPC ver = 0x%x\n", qpc->ver);
    // 'te' field has been removed:
	//CMDS_DBG( "QPC Address Translation Enabled = %s\n", (qpc->te ? "TRUE" : "FALSE"));
    CMDS_DBG( "QPC Descriptor Event Enabled = %s\n", (qpc->de ? "TRUE" : "FALSE"));
    CMDS_DBG( "QPC Path Migration State = 0x%x\n", qpc->pm_state);
    CMDS_DBG( "QPC Service Type = 0x%x\n", qpc->st);
    CMDS_DBG( "QPC VAPI-encoded State = %d\n", qpc->state);
    CMDS_DBG( "QPC Sched Queue = %d\n", qpc->sched_queue);
    CMDS_DBG( "QPC msg_max = %d\n", qpc->msg_max);
    CMDS_DBG( "QPC MTU (encoded) = %d\n", qpc->mtu);
    CMDS_DBG( "QPC usr_page = 0x%x\n", qpc->usr_page);
    CMDS_DBG( "QPC local_qpn_een = 0x%x\n", qpc->local_qpn_een);
    CMDS_DBG( "QPC remote_qpn_een = 0x%x\n", qpc->remote_qpn_een);
    CMDS_DBG( "QPC pd = 0x%x\n", qpc->pd);
    CMDS_DBG( "QPC wqe_lkey = 0x%x\n", qpc->wqe_lkey);
    CMDS_DBG( "QPC ssc = %s\n", (qpc->ssc ? "TRUE" : "FALSE"));
    CMDS_DBG( "QPC sic = %s\n", (qpc->sic ? "TRUE" : "FALSE"));
    CMDS_DBG( "QPC sae = %s\n", (qpc->sae ? "TRUE" : "FALSE"));
    CMDS_DBG( "QPC swe = %s\n", (qpc->swe ? "TRUE" : "FALSE"));
    CMDS_DBG( "QPC sre = %s\n", (qpc->sre ? "TRUE" : "FALSE"));
    CMDS_DBG( "QPC retry_count = 0x%x\n", qpc->retry_count);
    CMDS_DBG( "QPC sra_max = 0x%x\n", qpc->sra_max);
    CMDS_DBG( "QPC flight_lim = 0x%x\n", qpc->flight_lim);
    CMDS_DBG( "QPC ack_req_freq = 0x%x\n", qpc->ack_req_freq);
    CMDS_DBG( "QPC next_send_psn = 0x%x\n", qpc->next_send_psn);
    CMDS_DBG( "QPC cqn_snd = 0x%x\n", qpc->cqn_snd);
    CMDS_DBG( "QPC next_snd_wqe = "U64_FMT"\n", qpc->next_snd_wqe);
    CMDS_DBG( "QPC rsc = %s\n", (qpc->rsc ? "TRUE" : "FALSE"));
    CMDS_DBG( "QPC ric = %s\n", (qpc->ric ? "TRUE" : "FALSE"));
    CMDS_DBG( "QPC rae = %s\n", (qpc->rae ? "TRUE" : "FALSE"));
    CMDS_DBG( "QPC rwe = %s\n", (qpc->rwe ? "TRUE" : "FALSE"));
    CMDS_DBG( "QPC rre = %s\n", (qpc->rre ? "TRUE" : "FALSE"));
    CMDS_DBG( "QPC rra_max = 0x%x\n", qpc->rra_max);
    CMDS_DBG( "QPC next_rcv_psn = 0x%x\n", qpc->next_rcv_psn);
    CMDS_DBG( "QPC min_rnr_nak = 0x%x\n", qpc->min_rnr_nak);
    CMDS_DBG( "QPC ra_buff_indx = 0x%x\n", qpc->ra_buff_indx);
    CMDS_DBG( "QPC cqn_rcv = 0x%x\n", qpc->cqn_rcv);
    CMDS_DBG( "QPC next_rcv_wqe = "U64_FMT"\n", qpc->next_rcv_wqe);
    CMDS_DBG( "QPC q_key = 0x%x\n", qpc->q_key);
    CMDS_DBG( "QPC srqn = 0x%x\n", qpc->srqn);
    CMDS_DBG( "QPC srq = %s\n", (qpc->srq ? "TRUE" : "FALSE"));
    CMDS_DBG( "QPC primary.ack_timeout = %d\n" , qpc->primary_address_path.ack_timeout);
    CMDS_DBG( "QPC primary.max_stat_rate = %d\n" , qpc->primary_address_path.max_stat_rate);
}
void XHH_cmd_print_eq_context(XHH_eqc_t *eqc)
{
    return;
    CMDS_DBG("EQ CONTEXT DUMP (XHH_eqc_t structure)\n");
    CMDS_DBG( "eqc->st = 0x%x\n", eqc->st);
    CMDS_DBG( "eqc->oi (overrun ignore) = %s\n",  eqc->oi ? "TRUE" : "FALSE");
    CMDS_DBG( "eqc->tr (translation required) = %s\n",  eqc->tr ? "TRUE" : "FALSE");
    CMDS_DBG( "eqc->owner = %s\n",  (eqc->owner == XHH_OWNER_SW ? "XHH_OWNER_SW" : "XHH_OWNER_HW"));
    CMDS_DBG( "eqc->status = 0x%x\n",  eqc->status);
    CMDS_DBG( "eqc->start_address = "U64_FMT"\n",  eqc->start_address);
    CMDS_DBG( "eqc->usr_page = 0x%x\n",  eqc->usr_page);
    CMDS_DBG( "eqc->log_eq_size = %d\n",  eqc->log_eq_size);
    CMDS_DBG( "eqc->intr = 0x%x\n",  eqc->intr);
    CMDS_DBG( "eqc->l_key = 0x%X\n",  eqc->lkey);
    CMDS_DBG( "eqc->pd = 0x%X\n",  eqc->pd);
    CMDS_DBG( "eqc->consumer_indx = 0x%X\n",  eqc->consumer_indx);
    CMDS_DBG( "eqc->producer_indx = 0x%X\n",  eqc->producer_indx);
}
void XHH_cmd_print_mpt_entry(XHH_mpt_entry_t *mpt)
{
    CMDS_DBG("MPT ENTRY DUMP (XHH_mpt_entry_t structure)\n");
    CMDS_DBG( "MPT entry type = %s\n", (mpt->r_w ? "REGION" : "WINDOW"));
    CMDS_DBG( "MPT physical addr flag = %s\n", (mpt->pa ? "TRUE" : "FALSE"));
    CMDS_DBG( "MPT Local read access = %s\n", (mpt->lr ? "TRUE" : "FALSE"));
    CMDS_DBG( "MPT Local write access = %s\n", (mpt->lw ? "TRUE" : "FALSE"));
    CMDS_DBG( "MPT Remote read access = %s\n", (mpt->rr ? "TRUE" : "FALSE"));
    CMDS_DBG( "MPT Remote write access = %s\n", (mpt->rw ? "TRUE" : "FALSE"));
    CMDS_DBG( "MPT Atomic access = %s\n", (mpt->a ? "TRUE" : "FALSE"));
    CMDS_DBG( "MPT Atomic access = %s\n", (mpt->a ? "TRUE" : "FALSE"));
    CMDS_DBG( "MPT All writes posted  = %s\n", (mpt->pw ? "TRUE" : "FALSE"));
#ifdef TAVOR		
    CMDS_DBG( "MPT m_io  = %s\n", (mpt->m_io ? "TRUE" : "FALSE"));
#endif
    CMDS_DBG( "MPT Status = 0x%x\n", mpt->status);
    CMDS_DBG( "MPT Page size = %d (Actual size is [4K]*2^Page_size)\n", mpt->page_size);
    CMDS_DBG( "MPT mem key = 0x%x\n", mpt->mem_key);
    CMDS_DBG( "MPT pd = 0x%x\n", mpt->pd);
    CMDS_DBG( "MPT start_address = "U64_FMT"\n", mpt->start_address);
    CMDS_DBG( "MPT length = "U64_FMT"\n", mpt->reg_wnd_len);
    CMDS_DBG( "MPT lkey = 0x%x\n", mpt->lkey);
    CMDS_DBG( "MPT win_cnt = 0x%x\n", mpt->win_cnt);
    XHH_DEV_SPECIFIC_PRINT_MPT_ENTRY
}

void XHH_cmd_print_mgm_entry(XHH_mcg_entry_t *mgm)
{
    IB_wqpn_t *qp_iterator;
    u_int32_t   i;

    CMDS_DBG("MGM ENTRY DUMP (XHH_mcg_entry_t structure)\n");
    CMDS_DBG( "MGM next_gid_index = 0x%x\n", mgm->next_gid_index);
    CMDS_DBG("MGM GID = %d.%d.%d.%d.%d.%d.%d.%d.%d.%d.%d.%d.%d.%d.%d.%d \n",
             mgm->mgid[0], mgm->mgid[1], mgm->mgid[2], mgm->mgid[3]
             , mgm->mgid[4], mgm->mgid[5], mgm->mgid[6], mgm->mgid[7]
             , mgm->mgid[8], mgm->mgid[9], mgm->mgid[10], mgm->mgid[10]
             , mgm->mgid[12], mgm->mgid[13], mgm->mgid[14], mgm->mgid[15]);
    CMDS_DBG( "MGM valid_qps = %d \n", mgm->valid_qps);
    for (qp_iterator = mgm->qps, i = 0; i < mgm->valid_qps; i++, qp_iterator++) {
        CMDS_DBG( "MGM qps[%d] = 0x%x\n", i, *qp_iterator);
    }
}

#else

void XHH_cmd_print_hca_props(XHH_hca_props_t *hca_props) {}
void XHH_cmd_print_dev_lims(XHH_dev_lim_t *dev_lim) {}
void XHH_cmd_print_query_fw(XHH_fw_props_t *fw_props) {}
void XHH_cmd_print_query_adapter( XHH_adapter_props_t *adapter_props) {}
void XHH_cmd_print_init_ib(IB_port_t port, XHH_port_init_props_t *port_init_props) {}
void XHH_cmd_print_cq_context(XHH_cqc_t *cqc) {}
void XHH_cmd_print_qp_context(XHH_qpee_context_t *qpc) {}
void XHH_cmd_print_eq_context(XHH_eqc_t *eqc) {}
void XHH_cmd_print_mpt_entry(XHH_mpt_entry_t *mpt) {}
void XHH_cmd_print_mgm_entry(XHH_mcg_entry_t *mgm) {}

#endif   /* #if XHH_CMDS_WRAP_DEBUG_LEVEL */

#if 0
static void my_cmd_mailbox_print( void *mailbox, int size, const char * func )
{
    int i, j, maxlines, bytes_left, this_line;
    char linebuf[200], tempout[20];
    u_int8_t *iterator;

    if (mailbox == NULL) return;
    
    iterator = (u_int8_t *)mailbox;
    bytes_left = size;
    MTL_ERROR1("my MailboxPrint from function %s, starting at addr 0x%p, size=%d:\n",
               func, mailbox, size);
    
    if (size <= 0) {
        return;
    }

    maxlines = (size / 16) + ((size % 16) ? 1 : 0);

    for (i = 0; i < maxlines; i++) {
        memset(linebuf, 0, sizeof(linebuf));
        this_line = (bytes_left > 16 ? 16 : bytes_left);

        for (j = 0; j < this_line; j++) {
            if ((j % 4) == 0) {
                strcat(linebuf," ");
            }
            sprintf(tempout, "%02x", *iterator);
            iterator++; bytes_left--;
            strcat(linebuf,tempout);
        }
        MTL_ERROR1("%s\n", linebuf);
    }
    MTL_ERROR1("my MailboxPrint END\n");
}
#endif
/***************************************************** */
/************* END of PRINT UTILITIES **************** */
/***************************************************** */


/*
 *  XHH_cmd_QUERY_DEV_LIM
 */ 
XHH_cmd_status_t XHH_cmd_QUERY_DEV_LIM(XHH_cmd_t cmd_if, XHH_dev_lim_t *dev_lim)
{
  command_fields_t cmd_desc;
  u_int8_t *outprm;
  XHH_cmd_status_t rc; 
  u_int32_t buf_size = PSEUDO_MT_BYTE_SIZE(ADD_DEV_PRM_PREFIX(query_dev_lim_st));

  FUNC_IN;
  outprm = TNMALLOC(u_int8_t, buf_size);
  if (outprm == NULL) {
    MT_COV_MEM_FLTR_START
    MT_RETURN(XHH_CMD_STAT_EAGAIN);
    MT_COV_MEM_FLTR_END
  }
  memset(outprm, 0, buf_size);

  cmd_desc.in_param = 0;
  cmd_desc.in_param_size = 0;
  cmd_desc.in_trans = TRANS_NA;
  cmd_desc.input_modifier = 0;
  cmd_desc.out_param = outprm;
  cmd_desc.out_param_size = buf_size;
  cmd_desc.out_trans = TRANS_MAILBOX;
  cmd_desc.opcode = ADD_DEV_PREFIX(IF_CMD_QUERY_DEV_LIM);
  cmd_desc.opcode_modifier = 0;
  cmd_desc.exec_time_micro = ADD_DEV_PREFIX(IF_CMD_ETIME_QUERY_DEV_LIM);
  
  rc = cmd_invoke(cmd_if, &cmd_desc);
  if ( rc != XHH_CMD_STAT_OK ) {
    FREE(outprm);
    MT_RETURN(rc);
  }

  if ( dev_lim ) {
    EX_FLD(dev_lim, outprm, ADD_DEV_PRM_PREFIX(query_dev_lim_st), log_max_qp);
    EX_FLD(dev_lim, outprm, ADD_DEV_PRM_PREFIX(query_dev_lim_st), log2_rsvd_qps);
    EX_FLD(dev_lim, outprm, ADD_DEV_PRM_PREFIX(query_dev_lim_st), log_max_qp_sz);
    EX_FLD(dev_lim, outprm, ADD_DEV_PRM_PREFIX(query_dev_lim_st), log_max_srqs);
    EX_FLD(dev_lim, outprm, ADD_DEV_PRM_PREFIX(query_dev_lim_st), log2_rsvd_srqs);
    EX_FLD(dev_lim, outprm, ADD_DEV_PRM_PREFIX(query_dev_lim_st), log_max_srq_sz);
    EX_FLD(dev_lim, outprm, ADD_DEV_PRM_PREFIX(query_dev_lim_st), log_max_ee);
    EX_FLD(dev_lim, outprm, ADD_DEV_PRM_PREFIX(query_dev_lim_st), log2_rsvd_ees);
    EX_FLD(dev_lim, outprm, ADD_DEV_PRM_PREFIX(query_dev_lim_st), log_max_cq);
    EX_FLD(dev_lim, outprm, ADD_DEV_PRM_PREFIX(query_dev_lim_st), log2_rsvd_cqs);
    EX_FLD(dev_lim, outprm, ADD_DEV_PRM_PREFIX(query_dev_lim_st), log_max_cq_sz);
    EX_FLD(dev_lim, outprm, ADD_DEV_PRM_PREFIX(query_dev_lim_st), log_max_eq);
    EX_FLD(dev_lim, outprm, ADD_DEV_PRM_PREFIX(query_dev_lim_st), num_rsvd_eqs);
    EX_FLD(dev_lim, outprm, ADD_DEV_PRM_PREFIX(query_dev_lim_st), log_max_mpts);
    EX_FLD(dev_lim, outprm, ADD_DEV_PRM_PREFIX(query_dev_lim_st), log2_rsvd_mrws);
    EX_FLD(dev_lim, outprm, ADD_DEV_PRM_PREFIX(query_dev_lim_st), log_max_mrw_sz);
    EX_FLD(dev_lim, outprm, ADD_DEV_PRM_PREFIX(query_dev_lim_st), log2_rsvd_mtts);
    EX_FLD(dev_lim, outprm, ADD_DEV_PRM_PREFIX(query_dev_lim_st), log_max_ra_res_qp);
    EX_FLD(dev_lim, outprm, ADD_DEV_PRM_PREFIX(query_dev_lim_st), log_max_ra_req_qp);
    EX_FLD(dev_lim, outprm, ADD_DEV_PRM_PREFIX(query_dev_lim_st), log_max_ra_res_global);
    EX_FLD(dev_lim, outprm, ADD_DEV_PRM_PREFIX(query_dev_lim_st), num_ports);
    EX_FLD(dev_lim, outprm, ADD_DEV_PRM_PREFIX(query_dev_lim_st), max_vl);
    EX_FLD(dev_lim, outprm, ADD_DEV_PRM_PREFIX(query_dev_lim_st), max_port_width);
    EX_FLD(dev_lim, outprm, ADD_DEV_PRM_PREFIX(query_dev_lim_st), max_mtu);
    EX_FLD(dev_lim, outprm, ADD_DEV_PRM_PREFIX(query_dev_lim_st), local_ca_ack_delay);
    EX_FLD(dev_lim, outprm, ADD_DEV_PRM_PREFIX(query_dev_lim_st), log_max_gid);
    EX_FLD(dev_lim, outprm, ADD_DEV_PRM_PREFIX(query_dev_lim_st), log_max_pkey);

    EX_BOOL_FLD(dev_lim, outprm, ADD_DEV_PRM_PREFIX(query_dev_lim_st), rc);
    EX_BOOL_FLD(dev_lim, outprm, ADD_DEV_PRM_PREFIX(query_dev_lim_st), uc);
    EX_BOOL_FLD(dev_lim, outprm, ADD_DEV_PRM_PREFIX(query_dev_lim_st), ud);
    EX_BOOL_FLD(dev_lim, outprm, ADD_DEV_PRM_PREFIX(query_dev_lim_st), rd);
    EX_BOOL_FLD(dev_lim, outprm, ADD_DEV_PRM_PREFIX(query_dev_lim_st), raw_ipv6);
    EX_BOOL_FLD(dev_lim, outprm, ADD_DEV_PRM_PREFIX(query_dev_lim_st), raw_ether);
    EX_BOOL_FLD(dev_lim, outprm, ADD_DEV_PRM_PREFIX(query_dev_lim_st), srq);
    EX_BOOL_FLD(dev_lim, outprm, ADD_DEV_PRM_PREFIX(query_dev_lim_st), pkv);
    EX_BOOL_FLD(dev_lim, outprm, ADD_DEV_PRM_PREFIX(query_dev_lim_st), qkv);
    EX_BOOL_FLD(dev_lim, outprm, ADD_DEV_PRM_PREFIX(query_dev_lim_st), mw);
    EX_BOOL_FLD(dev_lim, outprm, ADD_DEV_PRM_PREFIX(query_dev_lim_st), apm);
    EX_BOOL_FLD(dev_lim, outprm, ADD_DEV_PRM_PREFIX(query_dev_lim_st), atm);
    EX_BOOL_FLD(dev_lim, outprm, ADD_DEV_PRM_PREFIX(query_dev_lim_st), rm);
    EX_BOOL_FLD(dev_lim, outprm, ADD_DEV_PRM_PREFIX(query_dev_lim_st), avp);
    EX_BOOL_FLD(dev_lim, outprm, ADD_DEV_PRM_PREFIX(query_dev_lim_st), udm);
    EX_BOOL_FLD(dev_lim, outprm, ADD_DEV_PRM_PREFIX(query_dev_lim_st), pg);
    EX_BOOL_FLD(dev_lim, outprm, ADD_DEV_PRM_PREFIX(query_dev_lim_st), r);
    
    EX_FLD(dev_lim, outprm, ADD_DEV_PRM_PREFIX(query_dev_lim_st), log_pg_sz);
    EX_FLD(dev_lim, outprm, ADD_DEV_PRM_PREFIX(query_dev_lim_st), uar_sz);
    EX_FLD(dev_lim, outprm, ADD_DEV_PRM_PREFIX(query_dev_lim_st), num_rsvd_uars);
    EX_FLD(dev_lim, outprm, ADD_DEV_PRM_PREFIX(query_dev_lim_st), log_max_mcg);
    EX_FLD(dev_lim, outprm, ADD_DEV_PRM_PREFIX(query_dev_lim_st), log_max_qp_mcg);
    EX_FLD(dev_lim, outprm, ADD_DEV_PRM_PREFIX(query_dev_lim_st), log_max_rdds);
    EX_FLD(dev_lim, outprm, ADD_DEV_PRM_PREFIX(query_dev_lim_st), num_rsvd_rdds);
    EX_FLD(dev_lim, outprm, ADD_DEV_PRM_PREFIX(query_dev_lim_st), log_max_pd);
    EX_FLD(dev_lim, outprm, ADD_DEV_PRM_PREFIX(query_dev_lim_st), num_rsvd_pds);
    EX_FLD(dev_lim, outprm, ADD_DEV_PRM_PREFIX(query_dev_lim_st), qpc_entry_sz);
    EX_FLD(dev_lim, outprm, ADD_DEV_PRM_PREFIX(query_dev_lim_st), eec_entry_sz);
    EX_FLD(dev_lim, outprm, ADD_DEV_PRM_PREFIX(query_dev_lim_st), eqpc_entry_sz);
    EX_FLD(dev_lim, outprm, ADD_DEV_PRM_PREFIX(query_dev_lim_st), eeec_entry_sz);
    EX_FLD(dev_lim, outprm, ADD_DEV_PRM_PREFIX(query_dev_lim_st), cqc_entry_sz);
    EX_FLD(dev_lim, outprm, ADD_DEV_PRM_PREFIX(query_dev_lim_st), eqc_entry_sz);
    EX_FLD(dev_lim, outprm, ADD_DEV_PRM_PREFIX(query_dev_lim_st), srq_entry_sz);
    EX_FLD(dev_lim, outprm, ADD_DEV_PRM_PREFIX(query_dev_lim_st), uar_scratch_entry_sz);
    XHH_DEV_SPECIFIC_QUERY_DEV_LIM(dev_lim,outprm, query_dev_lim_st)

    XHH_CMD_MAILBOX_PRINT(outprm, buf_size, __func__);

    XHH_CMD_PRINT_DEV_LIMS(dev_lim);
  }

  FREE(outprm);
  MT_RETURN(rc);
}

/*
 *  XHH_cmd_QUERY_FW
 */ 
XHH_cmd_status_t XHH_cmd_QUERY_FW(XHH_cmd_t cmd_if, XHH_fw_props_t *fw_props)
{
  command_fields_t cmd_desc;
  //u_int8_t outprm[PSEUDO_MT_BYTE_SIZE(ADD_DEV_PRM_PREFIX(query_fw_st)];
  u_int8_t *outprm;
  XHH_cmd_status_t rc;
  u_int32_t buf_size = PSEUDO_MT_BYTE_SIZE(ADD_DEV_PRM_PREFIX(query_fw_st));

  FUNC_IN;
  outprm = TNMALLOC(u_int8_t, buf_size);
  if (outprm == NULL) {
    MT_COV_MEM_FLTR_START
    MT_RETURN(XHH_CMD_STAT_EAGAIN);
    MT_COV_MEM_FLTR_END
  }
  memset(outprm, 0, buf_size);
  
  cmd_desc.in_param = 0;
  cmd_desc.in_param_size = 0;
  cmd_desc.in_trans = TRANS_NA;
  cmd_desc.input_modifier = 0;
  cmd_desc.out_param = outprm;
  cmd_desc.out_param_size = buf_size;
  cmd_desc.out_trans = TRANS_MAILBOX;
  cmd_desc.opcode = ADD_DEV_PREFIX(IF_CMD_QUERY_FW);
  cmd_desc.opcode_modifier = 0;
  cmd_desc.exec_time_micro = ADD_DEV_PREFIX(IF_CMD_ETIME_QUERY_FW);
  rc = cmd_invoke(cmd_if, &cmd_desc);
  if ( rc != XHH_CMD_STAT_OK ) {
    FREE(outprm);
    MT_RETURN(rc);
  }

  EX_FLD(fw_props, outprm, ADD_DEV_PRM_PREFIX(query_fw_st), fw_rev_major);
  EX_FLD(fw_props, outprm, ADD_DEV_PRM_PREFIX(query_fw_st), fw_rev_minor);
  EX_FLD(fw_props, outprm, ADD_DEV_PRM_PREFIX(query_fw_st), fw_rev_subminor);
  EX_FLD(fw_props, outprm, ADD_DEV_PRM_PREFIX(query_fw_st), cmd_interface_rev);
  EX_FLD(fw_props, outprm, ADD_DEV_PRM_PREFIX(query_fw_st), log_max_outstanding_cmd);
  EX_FLD64(fw_props, outprm, ADD_DEV_PRM_PREFIX(query_fw_st), error_buf_start);
  EX_FLD(fw_props, outprm, ADD_DEV_PRM_PREFIX(query_fw_st), error_buf_size);
  EX_BOOL_FLD(fw_props, outprm, ADD_DEV_PRM_PREFIX(query_fw_st), cmd_interface_db);
  EX_FLD(fw_props, outprm, ADD_DEV_PRM_PREFIX(query_fw_st), cmd_db_dw0);
  EX_FLD(fw_props, outprm, ADD_DEV_PRM_PREFIX(query_fw_st), cmd_db_dw1);
  EX_FLD(fw_props, outprm, ADD_DEV_PRM_PREFIX(query_fw_st), cmd_db_dw2);
  EX_FLD(fw_props, outprm, ADD_DEV_PRM_PREFIX(query_fw_st), cmd_db_dw3);
  EX_FLD(fw_props, outprm, ADD_DEV_PRM_PREFIX(query_fw_st), cmd_db_dw4);
  EX_FLD(fw_props, outprm, ADD_DEV_PRM_PREFIX(query_fw_st), cmd_db_dw5);
  EX_FLD(fw_props, outprm, ADD_DEV_PRM_PREFIX(query_fw_st), cmd_db_dw6);
  EX_FLD(fw_props, outprm, ADD_DEV_PRM_PREFIX(query_fw_st), cmd_db_dw7);
  EX_FLD64(fw_props, outprm, ADD_DEV_PRM_PREFIX(query_fw_st), cmd_db_addr_base);
 							  
  XHH_DEV_SPECIFIC_QUERY_FW(fw_props,outprm, query_fw_st)
  FREE(outprm);

  rc = XHH_cmd_set_fw_props(cmd_if, fw_props);

  XHH_CMD_PRINT_QUERY_FW(fw_props);
  
  MT_RETURN(rc);
}

/*
 *  XHH_cmd_QUERY_ADAPTER
 */ 

#define ADD_VSD_SUFFIX(i)  vsd.vsd_dw ## i      

#define EXTRACT_VSD(i) \
	  adapter_props->vsd[i] =   \
	      MT_EXTRACT_ARRAY32(outprm, \
	        MT_BIT_OFFSET(ADD_DEV_PRM_PREFIX(query_adapter_st), ADD_VSD_SUFFIX(i)), \
	        MT_BIT_SIZE(ADD_DEV_PRM_PREFIX(query_adapter_st),ADD_VSD_SUFFIX(i)));

XHH_cmd_status_t XHH_cmd_QUERY_ADAPTER(XHH_cmd_t cmd_if, XHH_adapter_props_t *adapter_props)
{
  command_fields_t cmd_desc;
  u_int8_t *outprm;
  XHH_cmd_status_t rc;
  u_int32_t buf_size = PSEUDO_MT_BYTE_SIZE(ADD_DEV_PRM_PREFIX(query_adapter_st));

  FUNC_IN;
  outprm = TNMALLOC(u_int8_t, buf_size);
  if (outprm == NULL) {
    MT_COV_MEM_FLTR_START
    MT_RETURN(XHH_CMD_STAT_EAGAIN);
    MT_COV_MEM_FLTR_END
  }
  memset(outprm, 0, buf_size);
  
  cmd_desc.in_param = 0;
  cmd_desc.in_param_size = 0;
  cmd_desc.in_trans = TRANS_NA;
  cmd_desc.input_modifier = 0;
  cmd_desc.out_param = outprm;
  cmd_desc.out_param_size = buf_size;
  cmd_desc.out_trans = TRANS_MAILBOX;
  cmd_desc.opcode = ADD_DEV_PREFIX(IF_CMD_QUERY_ADAPTER);
  cmd_desc.opcode_modifier = 0;
  cmd_desc.exec_time_micro = ADD_DEV_PREFIX(IF_CMD_ETIME_QUERY_ADAPTER);
  
  rc = cmd_invoke(cmd_if, &cmd_desc);
  if ( rc != XHH_CMD_STAT_OK ) {
    FREE(outprm);
    MT_RETURN(rc);
  }

  if ( adapter_props ) {
    memset(adapter_props, 0, sizeof(XHH_adapter_props_t));
#if 0
    EX_FLD(adapter_props, outprm, ADD_DEV_PRM_PREFIX(query_adapter_st), vendor_id);
    EX_FLD(adapter_props, outprm, ADD_DEV_PRM_PREFIX(query_adapter_st), device_id);
    EX_FLD(adapter_props, outprm, ADD_DEV_PRM_PREFIX(query_adapter_st), revision_id);
#endif    
    EX_FLD(adapter_props, outprm, ADD_DEV_PRM_PREFIX(query_adapter_st), intapin);
    EXTRACT_VSD(0) ; EXTRACT_VSD(1); EXTRACT_VSD(2) ; EXTRACT_VSD(3);
    EXTRACT_VSD(4) ; EXTRACT_VSD(5); EXTRACT_VSD(6) ; EXTRACT_VSD(7);
    EXTRACT_VSD(8) ; EXTRACT_VSD(9); EXTRACT_VSD(10) ; EXTRACT_VSD(11);
    EXTRACT_VSD(12) ; EXTRACT_VSD(13); EXTRACT_VSD(14) ; EXTRACT_VSD(15);
    EXTRACT_VSD(16) ; EXTRACT_VSD(17); EXTRACT_VSD(18) ; EXTRACT_VSD(19);
    EXTRACT_VSD(20) ; EXTRACT_VSD(21); EXTRACT_VSD(22) ; EXTRACT_VSD(23);
    EXTRACT_VSD(24) ; EXTRACT_VSD(25); EXTRACT_VSD(26) ; EXTRACT_VSD(27);
    EXTRACT_VSD(28) ; EXTRACT_VSD(29); EXTRACT_VSD(30) ; EXTRACT_VSD(31);
    EXTRACT_VSD(32) ; EXTRACT_VSD(33); EXTRACT_VSD(34) ; EXTRACT_VSD(35);
    EXTRACT_VSD(36) ; EXTRACT_VSD(37); EXTRACT_VSD(38) ; EXTRACT_VSD(39);
    EXTRACT_VSD(40) ; EXTRACT_VSD(41); EXTRACT_VSD(42) ; EXTRACT_VSD(43);
    EXTRACT_VSD(44) ; EXTRACT_VSD(45); EXTRACT_VSD(46) ; EXTRACT_VSD(47);
    EXTRACT_VSD(48) ; EXTRACT_VSD(49); EXTRACT_VSD(50) ; EXTRACT_VSD(51);
    EXTRACT_VSD(52) ; EXTRACT_VSD(53); EXTRACT_VSD(54) ; EXTRACT_VSD(55);
    
    XHH_CMD_PRINT_QUERY_ADAPTER(adapter_props);
  }
  FREE(outprm);
  MT_RETURN(rc);
}

/*
 *  XHH_cmd_INIT_HCA
 */ 

XHH_cmd_status_t XHH_cmd_INIT_HCA(XHH_cmd_t cmd_if, XHH_hca_props_t *hca_props)
{
  command_fields_t cmd_desc;
  u_int8_t *inprm;
  XHH_cmd_status_t rc;
  u_int32_t buf_size = PSEUDO_MT_BYTE_SIZE(ADD_DEV_PRM_PREFIX(init_hca_st));
  u_int64_t     rdb_base_addr_save;

  FUNC_IN;
  if (hca_props == NULL) {MT_RETURN(XHH_CMD_STAT_EBADARG); }
  inprm = TNMALLOC(u_int8_t, buf_size);
  if (inprm == NULL) {
    MT_COV_MEM_FLTR_START
    MT_RETURN(XHH_CMD_STAT_EAGAIN);
    MT_COV_MEM_FLTR_END
  }
  memset(inprm, 0, buf_size);
  XHH_CMD_PRINT_HCA_PROPS(hca_props);

  cmd_desc.in_param = inprm;
  cmd_desc.in_param_size = buf_size;
  cmd_desc.in_trans = TRANS_MAILBOX;
  cmd_desc.input_modifier = 0;
  cmd_desc.out_param = 0;
  cmd_desc.out_param_size = 0;
  cmd_desc.out_trans = TRANS_NA;
  cmd_desc.opcode = ADD_DEV_PREFIX(IF_CMD_INIT_HCA);
  cmd_desc.opcode_modifier = 0;
  cmd_desc.exec_time_micro = hca_props->qpc_eec_cqc_eqc_rdb_parameters.log_num_of_qp > 18 ?
                             2*(ADD_DEV_PREFIX(IF_CMD_ETIME_INIT_HCA)) : ADD_DEV_PREFIX(IF_CMD_ETIME_INIT_HCA);

  INS_FLD(hca_props, inprm, ADD_DEV_PRM_PREFIX(init_hca_st), router_qp);
  INS_BOOL_FLD(hca_props, inprm, ADD_DEV_PRM_PREFIX(init_hca_st), re);
  INS_BOOL_FLD(hca_props, inprm, ADD_DEV_PRM_PREFIX(init_hca_st), udp);
  INS_BOOL_FLD(hca_props, inprm, ADD_DEV_PRM_PREFIX(init_hca_st), he);

  INS_FLD64_SH(hca_props, inprm, ADD_DEV_PRM_PREFIX(init_hca_st), qpc_eec_cqc_eqc_rdb_parameters.qpc_base_addr);
  INS_FLD(hca_props, inprm, ADD_DEV_PRM_PREFIX(init_hca_st), qpc_eec_cqc_eqc_rdb_parameters.log_num_of_qp);
  INS_FLD64_SH(hca_props, inprm, ADD_DEV_PRM_PREFIX(init_hca_st), qpc_eec_cqc_eqc_rdb_parameters.srqc_base_addr);
  INS_FLD(hca_props, inprm, ADD_DEV_PRM_PREFIX(init_hca_st), qpc_eec_cqc_eqc_rdb_parameters.log_num_of_srq);
  INS_FLD64_SH(hca_props, inprm, ADD_DEV_PRM_PREFIX(init_hca_st), qpc_eec_cqc_eqc_rdb_parameters.eec_base_addr);
  INS_FLD(hca_props, inprm, ADD_DEV_PRM_PREFIX(init_hca_st), qpc_eec_cqc_eqc_rdb_parameters.log_num_of_ee);
  INS_FLD64_SH(hca_props, inprm, ADD_DEV_PRM_PREFIX(init_hca_st), qpc_eec_cqc_eqc_rdb_parameters.cqc_base_addr);
  INS_FLD(hca_props, inprm, ADD_DEV_PRM_PREFIX(init_hca_st), qpc_eec_cqc_eqc_rdb_parameters.log_num_of_cq);
  INS_FLD64(hca_props, inprm, ADD_DEV_PRM_PREFIX(init_hca_st), qpc_eec_cqc_eqc_rdb_parameters.eqpc_base_addr);
  INS_FLD64(hca_props, inprm, ADD_DEV_PRM_PREFIX(init_hca_st), qpc_eec_cqc_eqc_rdb_parameters.eeec_base_addr);
  INS_FLD64_SH(hca_props, inprm, ADD_DEV_PRM_PREFIX(init_hca_st), qpc_eec_cqc_eqc_rdb_parameters.eqc_base_addr);
  INS_FLD(hca_props, inprm, ADD_DEV_PRM_PREFIX(init_hca_st), qpc_eec_cqc_eqc_rdb_parameters.log_num_eq);

  /* zero out low order 32 bits of rdb base addr for passing to device */
  rdb_base_addr_save = hca_props->qpc_eec_cqc_eqc_rdb_parameters.rdb_base_addr;
  hca_props->qpc_eec_cqc_eqc_rdb_parameters.rdb_base_addr &= MAKE_ULONGLONG(0xFFFFFFFF00000000);
  INS_FLD64(hca_props, inprm, ADD_DEV_PRM_PREFIX(init_hca_st), qpc_eec_cqc_eqc_rdb_parameters.rdb_base_addr);
  /*restore original RDB base address */
  hca_props->qpc_eec_cqc_eqc_rdb_parameters.rdb_base_addr =rdb_base_addr_save;

  INS_FLD64(hca_props, inprm, ADD_DEV_PRM_PREFIX(init_hca_st), multicast_parameters.mc_base_addr);
  INS_FLD(hca_props, inprm, ADD_DEV_PRM_PREFIX(init_hca_st), multicast_parameters.log_mc_table_entry_sz);
  INS_FLD(hca_props, inprm, ADD_DEV_PRM_PREFIX(init_hca_st), multicast_parameters.log_mc_table_sz);
  INS_FLD(hca_props, inprm, ADD_DEV_PRM_PREFIX(init_hca_st), multicast_parameters.mc_table_hash_sz);
  INS_FLD(hca_props, inprm, ADD_DEV_PRM_PREFIX(init_hca_st), multicast_parameters.mc_hash_fn);
  
  INS_FLD64(hca_props, inprm, ADD_DEV_PRM_PREFIX(init_hca_st), tpt_parameters.mpt_base_adr);
  INS_FLD(hca_props, inprm, ADD_DEV_PRM_PREFIX(init_hca_st), tpt_parameters.log_mpt_sz);
  INS_FLD(hca_props, inprm, ADD_DEV_PRM_PREFIX(init_hca_st), tpt_parameters.pfto);
  INS_FLD64(hca_props, inprm, ADD_DEV_PRM_PREFIX(init_hca_st), tpt_parameters.mtt_base_addr);
  
  INS_FLD64(hca_props, inprm, ADD_DEV_PRM_PREFIX(init_hca_st), uar_parameters.uar_base_addr);
  INS_FLD(hca_props, inprm, ADD_DEV_PRM_PREFIX(init_hca_st), uar_parameters.uar_page_sz);
  INS_FLD64(hca_props, inprm, ADD_DEV_PRM_PREFIX(init_hca_st), uar_parameters.uar_scratch_base_addr);
  XHH_CMD_DEV_SPECIFIC_INIT_HCA

  XHH_CMD_MAILBOX_PRINT(inprm, buf_size, __func__);
  
  rc = cmd_invoke(cmd_if, &cmd_desc);
  FREE(inprm);
  MT_RETURN(rc);
}


/*
 *  XHH_cmd_CLOSE_HCA
 */ 
#ifdef SIMULATE_HALT_HCA
XHH_cmd_status_t XHH_cmd_CLOSE_HCA(XHH_cmd_t cmd_if)
#else
XHH_cmd_status_t XHH_cmd_CLOSE_HCA(XHH_cmd_t cmd_if, MT_bool do_halt)
#endif
{
  command_fields_t cmd_desc;
  XHH_cmd_status_t rc;

  FUNC_IN;
  cmd_desc.in_param = 0;
  cmd_desc.in_param_size = 0;
  cmd_desc.in_trans = TRANS_NA;
  cmd_desc.input_modifier = 0;
  cmd_desc.out_param = 0;
  cmd_desc.out_param_size = 0;
  cmd_desc.out_trans = TRANS_NA;
  cmd_desc.opcode = ADD_DEV_PREFIX(IF_CMD_CLOSE_HCA);
#ifdef SIMULATE_HALT_HCA
  cmd_desc.opcode_modifier = 0;
#else
  cmd_desc.opcode_modifier = (do_halt == FALSE ? 0 : 1);
#endif
  cmd_desc.exec_time_micro = ADD_DEV_PREFIX(IF_CMD_ETIME_CLOSE_HCA);

  rc = cmd_invoke(cmd_if, &cmd_desc);
  MT_RETURN(rc);
}


/*
 *  XHH_cmd_INIT_IB
 */ 
XHH_cmd_status_t XHH_cmd_INIT_IB(XHH_cmd_t cmd_if, IB_port_t port,
                                 XHH_port_init_props_t *port_init_props)
{
  command_fields_t cmd_desc;
  u_int8_t *inprm;
  XHH_cmd_status_t rc;
  u_int32_t buf_size = PSEUDO_MT_BYTE_SIZE(ADD_DEV_PRM_PREFIX(init_ib_st));
  u_int32_t temp_u32;

  FUNC_IN;
  if (port_init_props == NULL) {MT_RETURN(XHH_CMD_STAT_EBADARG); }
  
  inprm = TNMALLOC(u_int8_t, buf_size);
  if (inprm == NULL) {
    MT_COV_MEM_FLTR_START
    MT_RETURN(XHH_CMD_STAT_EAGAIN);
    MT_COV_MEM_FLTR_END
  }
  memset(inprm, 0, buf_size);
  XHH_CMD_PRINT_INIT_IB(port, port_init_props);
  cmd_desc.in_param = inprm;
  cmd_desc.in_param_size = buf_size;
  cmd_desc.in_trans = TRANS_MAILBOX;
  cmd_desc.input_modifier = port;
  cmd_desc.out_param = 0;
  cmd_desc.out_param_size = 0;
  cmd_desc.out_trans = TRANS_NA;
  cmd_desc.opcode = ADD_DEV_PREFIX(IF_CMD_INIT_IB);
  cmd_desc.opcode_modifier = 0;
  cmd_desc.exec_time_micro = ADD_DEV_PREFIX(IF_CMD_ETIME_INIT_IB);

  INS_FLD(port_init_props, inprm, ADD_DEV_PRM_PREFIX(init_ib_st), vl_cap);
  INS_FLD(port_init_props, inprm, ADD_DEV_PRM_PREFIX(init_ib_st), port_width_cap);
  INS_FLD(port_init_props, inprm, ADD_DEV_PRM_PREFIX(init_ib_st), mtu_cap);
  INS_FLD(port_init_props, inprm, ADD_DEV_PRM_PREFIX(init_ib_st), max_gid);
  INS_FLD(port_init_props, inprm, ADD_DEV_PRM_PREFIX(init_ib_st), max_pkey);

  INS_BOOL_FLD(port_init_props, inprm, ADD_DEV_PRM_PREFIX(init_ib_st), g0);

  /* We get GUID0 in BIG_ENDIAN format.  It needs to be split up into Host-endian format before passing to cmd_invoke */
  /* Note that need to memcpy each 4 bytes to a temporary u_int32_t variable, since there is no guarantee */
  /* that the GUID is 4-byte aligned (it is an array of unsigned chars) */
  memcpy(&temp_u32, &(port_init_props->guid0[0]), sizeof(u_int32_t));
  MT_INSERT_ARRAY32(inprm, MOSAL_be32_to_cpu(temp_u32), 
                 MT_BIT_OFFSET(ADD_DEV_PRM_PREFIX(init_ib_st), guid0_h), MT_BIT_SIZE(ADD_DEV_PRM_PREFIX(init_ib_st), guid0_h));

  memcpy(&temp_u32, &(port_init_props->guid0[4]), sizeof(u_int32_t));
  MT_INSERT_ARRAY32(inprm, MOSAL_be32_to_cpu(temp_u32), 
                 MT_BIT_OFFSET(ADD_DEV_PRM_PREFIX(init_ib_st), guid0_l), MT_BIT_SIZE(ADD_DEV_PRM_PREFIX(init_ib_st), guid0_l));

  rc = cmd_invoke(cmd_if, &cmd_desc);
  FREE(inprm);
  MT_RETURN(rc);
}

/*
 *  XHH_cmd_CLOSE_IB
 */ 
XHH_cmd_status_t XHH_cmd_CLOSE_IB(XHH_cmd_t cmd_if, IB_port_t port)
{
  command_fields_t cmd_desc;
  XHH_cmd_status_t rc;

  FUNC_IN;
  cmd_desc.in_param = 0;
  cmd_desc.in_param_size = 0;
  cmd_desc.in_trans = TRANS_NA;
  cmd_desc.input_modifier = port;
  cmd_desc.out_param = 0;
  cmd_desc.out_param_size = 0;
  cmd_desc.out_trans = TRANS_NA;
  cmd_desc.opcode = ADD_DEV_PREFIX(IF_CMD_CLOSE_IB);
  cmd_desc.opcode_modifier = 0;
  cmd_desc.exec_time_micro = ADD_DEV_PREFIX(IF_CMD_ETIME_CLOSE_IB);

  rc = cmd_invoke(cmd_if, &cmd_desc);
  MT_RETURN(rc);
}

/*
 *  XHH_cmd_QUERY_HCA
 */ 

XHH_cmd_status_t XHH_cmd_QUERY_HCA(XHH_cmd_t cmd_if, XHH_hca_props_t *hca_props)
{
  command_fields_t cmd_desc;
  u_int8_t *outprm;
  XHH_cmd_status_t rc;
  u_int32_t buf_size = PSEUDO_MT_BYTE_SIZE(ADD_DEV_PRM_PREFIX(init_hca_st));

  FUNC_IN;
  outprm = TNMALLOC(u_int8_t, buf_size);
  if (outprm == NULL) {
    MT_COV_MEM_FLTR_START
    MT_RETURN(XHH_CMD_STAT_EAGAIN);
    MT_COV_MEM_FLTR_END
  }
  memset(outprm, 0, buf_size);

  cmd_desc.in_param = 0;
  cmd_desc.in_param_size = 0;
  cmd_desc.in_trans = TRANS_NA;
  cmd_desc.input_modifier = 0;
  cmd_desc.out_param = outprm;
  cmd_desc.out_param_size = buf_size;
  cmd_desc.out_trans = TRANS_MAILBOX;
  cmd_desc.opcode = ADD_DEV_PREFIX(IF_CMD_QUERY_HCA);
  cmd_desc.opcode_modifier = 0;
  cmd_desc.exec_time_micro = ADD_DEV_PREFIX(IF_CMD_ETIME_QUERY_HCA);
  
  rc = cmd_invoke(cmd_if, &cmd_desc);
  if ( rc != XHH_CMD_STAT_OK ) {
    FREE(outprm);
    MT_RETURN(rc);
  }

  if ( hca_props ) {
    EX_FLD(hca_props, outprm, ADD_DEV_PRM_PREFIX(init_hca_st), hca_core_clock);
    EX_FLD(hca_props, outprm, ADD_DEV_PRM_PREFIX(init_hca_st), router_qp);
    EX_BOOL_FLD(hca_props, outprm, ADD_DEV_PRM_PREFIX(init_hca_st), re);
    EX_BOOL_FLD(hca_props, outprm, ADD_DEV_PRM_PREFIX(init_hca_st), udp);
    EX_BOOL_FLD(hca_props, outprm, ADD_DEV_PRM_PREFIX(init_hca_st), he);

    EX_FLD64_SH(hca_props, outprm, ADD_DEV_PRM_PREFIX(init_hca_st), qpc_eec_cqc_eqc_rdb_parameters.qpc_base_addr);
    EX_FLD(hca_props, outprm, ADD_DEV_PRM_PREFIX(init_hca_st), qpc_eec_cqc_eqc_rdb_parameters.log_num_of_qp);
    EX_FLD64_SH(hca_props, outprm, ADD_DEV_PRM_PREFIX(init_hca_st), qpc_eec_cqc_eqc_rdb_parameters.srqc_base_addr);
    EX_FLD(hca_props, outprm, ADD_DEV_PRM_PREFIX(init_hca_st), qpc_eec_cqc_eqc_rdb_parameters.log_num_of_srq);
    EX_FLD64_SH(hca_props, outprm, ADD_DEV_PRM_PREFIX(init_hca_st), qpc_eec_cqc_eqc_rdb_parameters.eec_base_addr);
    EX_FLD(hca_props, outprm, ADD_DEV_PRM_PREFIX(init_hca_st), qpc_eec_cqc_eqc_rdb_parameters.log_num_of_ee);
    EX_FLD64_SH(hca_props, outprm, ADD_DEV_PRM_PREFIX(init_hca_st), qpc_eec_cqc_eqc_rdb_parameters.cqc_base_addr);
    EX_FLD(hca_props, outprm, ADD_DEV_PRM_PREFIX(init_hca_st), qpc_eec_cqc_eqc_rdb_parameters.log_num_of_cq);
    EX_FLD64(hca_props, outprm, ADD_DEV_PRM_PREFIX(init_hca_st), qpc_eec_cqc_eqc_rdb_parameters.eqpc_base_addr);
    EX_FLD64(hca_props, outprm, ADD_DEV_PRM_PREFIX(init_hca_st), qpc_eec_cqc_eqc_rdb_parameters.eeec_base_addr);
    EX_FLD64_SH(hca_props, outprm, ADD_DEV_PRM_PREFIX(init_hca_st), qpc_eec_cqc_eqc_rdb_parameters.eqc_base_addr);
    EX_FLD(hca_props, outprm, ADD_DEV_PRM_PREFIX(init_hca_st), qpc_eec_cqc_eqc_rdb_parameters.log_num_eq);
    EX_FLD64(hca_props, outprm, ADD_DEV_PRM_PREFIX(init_hca_st), qpc_eec_cqc_eqc_rdb_parameters.rdb_base_addr);

    EX_FLD64(hca_props, outprm, ADD_DEV_PRM_PREFIX(init_hca_st), multicast_parameters.mc_base_addr);
    EX_FLD(hca_props, outprm, ADD_DEV_PRM_PREFIX(init_hca_st), multicast_parameters.log_mc_table_entry_sz);
    EX_FLD(hca_props, outprm, ADD_DEV_PRM_PREFIX(init_hca_st), multicast_parameters.log_mc_table_sz);
    EX_FLD(hca_props, outprm, ADD_DEV_PRM_PREFIX(init_hca_st), multicast_parameters.mc_table_hash_sz);
    EX_FLD(hca_props, outprm, ADD_DEV_PRM_PREFIX(init_hca_st), multicast_parameters.mc_hash_fn);

    EX_FLD64(hca_props, outprm, ADD_DEV_PRM_PREFIX(init_hca_st), tpt_parameters.mpt_base_adr);
    EX_FLD(hca_props, outprm, ADD_DEV_PRM_PREFIX(init_hca_st), tpt_parameters.log_mpt_sz);
    EX_FLD(hca_props, outprm, ADD_DEV_PRM_PREFIX(init_hca_st), tpt_parameters.pfto);
    EX_FLD64(hca_props, outprm, ADD_DEV_PRM_PREFIX(init_hca_st), tpt_parameters.mtt_base_addr);

    EX_FLD64_SH(hca_props, outprm, ADD_DEV_PRM_PREFIX(init_hca_st), uar_parameters.uar_base_addr);
    EX_FLD(hca_props, outprm, ADD_DEV_PRM_PREFIX(init_hca_st), uar_parameters.uar_page_sz);
    EX_FLD64(hca_props, outprm, ADD_DEV_PRM_PREFIX(init_hca_st), uar_parameters.uar_scratch_base_addr);
    XHH_CMD_DEV_SPECIFIC_QUERY_HCA

    XHH_CMD_MAILBOX_PRINT(outprm, buf_size, __func__);

    XHH_CMD_PRINT_HCA_PROPS(hca_props);
  }
      
  FREE(outprm);
  MT_RETURN(rc);
}


/*
 *  XHH_cmd_SET_IB
 */ 
XHH_cmd_status_t XHH_cmd_SET_IB(XHH_cmd_t cmd_if, IB_port_t port,
                                 XHH_set_ib_props_t *port_init_props)
{
  command_fields_t cmd_desc;
  u_int8_t *inprm = NULL;
  XHH_cmd_status_t rc;
  u_int32_t buf_size = PSEUDO_MT_BYTE_SIZE(ADD_DEV_PRM_PREFIX(set_ib_st));

  FUNC_IN;
  if (port_init_props == NULL) {MT_RETURN(XHH_CMD_STAT_EBADARG);}
  inprm = (u_int8_t *)MALLOC(PSEUDO_MT_BYTE_SIZE(ADD_DEV_PRM_PREFIX(set_ib_st)));
  if (inprm == NULL) {
     MT_COV_MEM_FLTR_START
     MT_RETURN(XHH_CMD_STAT_EAGAIN);
     MT_COV_MEM_FLTR_END
  }
  memset(inprm, 0, buf_size);

  cmd_desc.in_param = inprm;
  cmd_desc.in_param_size = buf_size;
  cmd_desc.in_trans = TRANS_MAILBOX;
  cmd_desc.input_modifier = port;
  cmd_desc.out_param = 0;
  cmd_desc.out_param_size = 0;
  cmd_desc.out_trans = TRANS_NA;
  cmd_desc.opcode = ADD_DEV_PREFIX(IF_CMD_SET_IB);
  cmd_desc.opcode_modifier = 0;
  cmd_desc.exec_time_micro = ADD_DEV_PREFIX(IF_CMD_ETIME_SET_IB);

  INS_BOOL_FLD(port_init_props, inprm, ADD_DEV_PRM_PREFIX(set_ib_st), rqk);
  INS_FLD(port_init_props, inprm, ADD_DEV_PRM_PREFIX(set_ib_st), capability_mask);
  
  rc = cmd_invoke(cmd_if, &cmd_desc);
  FREE(inprm);
  MT_RETURN(rc);
}

/*
 *  XHH_cmd_SW2HW_MPT
 */ 
XHH_cmd_status_t XHH_cmd_SW2HW_MPT(XHH_cmd_t cmd_if, XHH_mpt_index_t mpt_index,
                                   XHH_mpt_entry_t *mpt_entry)
{
    command_fields_t cmd_desc;
    u_int8_t *inprm = NULL;
    XHH_cmd_status_t rc;
    MT_size_t buf_size = PSEUDO_MT_BYTE_SIZE(ADD_DEV_PRM_PREFIX(mpt_st));

    FUNC_IN;
    MTL_DEBUG4("XHH_cmd_SW2HW_MPT:  mpt_index = "SIZE_T_FMT", buf_size = "SIZE_T_FMT"\n", mpt_index, buf_size); 
    if (mpt_entry == NULL) {MT_RETURN(XHH_CMD_STAT_EBADARG); }
    XHH_CMD_PRINT_MPT_ENTRY(mpt_entry);

    inprm = (u_int8_t *)MALLOC(PSEUDO_MT_BYTE_SIZE(ADD_DEV_PRM_PREFIX(mpt_st)));
    if (inprm == NULL) {
       MT_COV_MEM_FLTR_START
       MT_RETURN(XHH_CMD_STAT_EAGAIN);
       MT_COV_MEM_FLTR_END
    }
    memset(inprm, 0, buf_size);
    
    cmd_desc.in_param = inprm;
    cmd_desc.in_param_size = (u_int32_t)buf_size;
    cmd_desc.in_trans = TRANS_MAILBOX;
    cmd_desc.input_modifier = (u_int32_t)mpt_index;
    cmd_desc.out_param = 0;
    cmd_desc.out_param_size = 0;
    cmd_desc.out_trans = TRANS_NA;
    cmd_desc.opcode = ADD_DEV_PREFIX(IF_CMD_SW2HW_MPT);
    cmd_desc.opcode_modifier = 0;
    cmd_desc.exec_time_micro = ADD_DEV_PREFIX(IF_CMD_ETIME_SW2HW_MPT);

    INS_BOOL_FLD(mpt_entry, inprm, ADD_DEV_PRM_PREFIX(mpt_st), r_w);
    INS_BOOL_FLD(mpt_entry, inprm, ADD_DEV_PRM_PREFIX(mpt_st), pa);
    INS_BOOL_FLD(mpt_entry, inprm, ADD_DEV_PRM_PREFIX(mpt_st), lr);
    INS_BOOL_FLD(mpt_entry, inprm, ADD_DEV_PRM_PREFIX(mpt_st), lw);
    INS_BOOL_FLD(mpt_entry, inprm, ADD_DEV_PRM_PREFIX(mpt_st), rr);
    INS_BOOL_FLD(mpt_entry, inprm, ADD_DEV_PRM_PREFIX(mpt_st), rw);
    INS_BOOL_FLD(mpt_entry, inprm, ADD_DEV_PRM_PREFIX(mpt_st), a);
    INS_BOOL_FLD(mpt_entry, inprm, ADD_DEV_PRM_PREFIX(mpt_st), eb);
    
    INS_FLD(mpt_entry, inprm, ADD_DEV_PRM_PREFIX(mpt_st), status);
    INS_FLD(mpt_entry, inprm, ADD_DEV_PRM_PREFIX(mpt_st), page_size);
    INS_FLD(mpt_entry, inprm, ADD_DEV_PRM_PREFIX(mpt_st), pd);
    INS_FLD64(mpt_entry, inprm, ADD_DEV_PRM_PREFIX(mpt_st), start_address);
    INS_FLD64(mpt_entry, inprm, ADD_DEV_PRM_PREFIX(mpt_st), reg_wnd_len);
    INS_FLD(mpt_entry, inprm, ADD_DEV_PRM_PREFIX(mpt_st), lkey);
    INS_FLD(mpt_entry, inprm, ADD_DEV_PRM_PREFIX(mpt_st), win_cnt);
    XHH_CMD_DEV_SPECIFIC_SW2HW_MPT

    XHH_CMD_MAILBOX_PRINT(inprm, (int)buf_size, __func__);
#if 1
    rc = cmd_invoke(cmd_if, &cmd_desc);
#else
    MTL_DEBUG4("XHH_cmd_SW2HW_MPT:  SKIPPING cmd_invoke !!!!!!!!!\n");
    rc = XHH_CMD_STAT_OK;
#endif
    FREE(inprm);
    MT_RETURN(rc);
}

/*
 *  XHH_cmd_SW2HW_MPT
 */ 
XHH_cmd_status_t XHH_cmd_MODIFY_MPT(XHH_cmd_t cmd_if, XHH_mpt_index_t mpt_index,
                                   XHH_mpt_entry_t *mpt_entry, MT_bool modify_entire_entry)
{
    command_fields_t cmd_desc;
    u_int8_t *inprm = NULL;
    XHH_cmd_status_t rc;
    MT_size_t buf_size = PSEUDO_MT_BYTE_SIZE(ADD_DEV_PRM_PREFIX(mpt_st));

    FUNC_IN;
    MTL_DEBUG4("XHH_cmd_MODIFY_MPT:  mpt_index = "SIZE_T_FMT", (in_param)buf_size = "SIZE_T_FMT"\n", 
               mpt_index, buf_size); 
    if (mpt_entry == NULL) {MT_RETURN(XHH_CMD_STAT_EBADARG); }
    XHH_CMD_PRINT_MPT_ENTRY(mpt_entry);

    inprm = (u_int8_t *)MALLOC(PSEUDO_MT_BYTE_SIZE(ADD_DEV_PRM_PREFIX(mpt_st)));
    if (inprm == NULL) {
       MT_COV_MEM_FLTR_START
       MT_RETURN(XHH_CMD_STAT_EAGAIN);
       MT_COV_MEM_FLTR_END
    }
    memset(inprm, 0, buf_size);
    
    cmd_desc.in_param = inprm;
    cmd_desc.in_param_size = (u_int32_t)buf_size;
    cmd_desc.in_trans = TRANS_MAILBOX;
    cmd_desc.input_modifier = (u_int32_t)mpt_index;
    cmd_desc.out_param = 0;
    cmd_desc.out_param_size = 0;
    cmd_desc.out_trans = TRANS_NA;
    cmd_desc.opcode = ADD_DEV_PREFIX(IF_CMD_MODIFY_MPT) ;
    cmd_desc.opcode_modifier = (modify_entire_entry == TRUE) ? 0x5 : 0x3;
    cmd_desc.exec_time_micro =  ADD_DEV_PREFIX(IF_CMD_ETIME_CLASS_B);

    INS_BOOL_FLD(mpt_entry, inprm, ADD_DEV_PRM_PREFIX(mpt_st), r_w);
    INS_BOOL_FLD(mpt_entry, inprm, ADD_DEV_PRM_PREFIX(mpt_st), pa);
    INS_BOOL_FLD(mpt_entry, inprm, ADD_DEV_PRM_PREFIX(mpt_st), lr);
    INS_BOOL_FLD(mpt_entry, inprm, ADD_DEV_PRM_PREFIX(mpt_st), lw);
    INS_BOOL_FLD(mpt_entry, inprm, ADD_DEV_PRM_PREFIX(mpt_st), rr);
    INS_BOOL_FLD(mpt_entry, inprm, ADD_DEV_PRM_PREFIX(mpt_st), rw);
    INS_BOOL_FLD(mpt_entry, inprm, ADD_DEV_PRM_PREFIX(mpt_st), a);
    INS_BOOL_FLD(mpt_entry, inprm, ADD_DEV_PRM_PREFIX(mpt_st), eb);
    
    INS_FLD(mpt_entry, inprm, ADD_DEV_PRM_PREFIX(mpt_st), status);
    INS_FLD(mpt_entry, inprm, ADD_DEV_PRM_PREFIX(mpt_st), page_size);
    INS_FLD(mpt_entry, inprm, ADD_DEV_PRM_PREFIX(mpt_st), pd);
    INS_FLD64(mpt_entry, inprm, ADD_DEV_PRM_PREFIX(mpt_st), start_address);
    INS_FLD64(mpt_entry, inprm, ADD_DEV_PRM_PREFIX(mpt_st), reg_wnd_len);
    INS_FLD(mpt_entry, inprm, ADD_DEV_PRM_PREFIX(mpt_st), lkey);
    INS_FLD(mpt_entry, inprm, ADD_DEV_PRM_PREFIX(mpt_st), win_cnt);
    XHH_CMD_DEV_SPECIFIC_MODIFY_MPT

    XHH_CMD_MAILBOX_PRINT(inprm, (int)buf_size, __func__);
#if 1
    rc = cmd_invoke(cmd_if, &cmd_desc);
#else
    MTL_DEBUG4("%s:  SKIPPING cmd_invoke !!!!!!!!!\n", __func__);
    rc = XHH_CMD_STAT_OK;
#endif
    FREE(inprm);
    MT_RETURN(rc);
}

/*
 *  XHH_cmd_QUERY_MPT
 */ 
XHH_cmd_status_t XHH_cmd_QUERY_MPT(XHH_cmd_t cmd_if, XHH_mpt_index_t mpt_index,
                                   XHH_mpt_entry_t *mpt_entry)
{
    command_fields_t cmd_desc;
    u_int8_t *outprm = NULL;
    XHH_cmd_status_t rc;
    u_int32_t buf_size = PSEUDO_MT_BYTE_SIZE(ADD_DEV_PRM_PREFIX(mpt_st));

    FUNC_IN;
    outprm = (u_int8_t *)MALLOC(PSEUDO_MT_BYTE_SIZE(ADD_DEV_PRM_PREFIX(mpt_st)));
    if (outprm == NULL) {
       MT_COV_MEM_FLTR_START
       MT_RETURN(XHH_CMD_STAT_EAGAIN);
       MT_COV_MEM_FLTR_END
    }
    memset(outprm, 0, buf_size);

    cmd_desc.in_param = 0;
    cmd_desc.in_param_size = 0;
    cmd_desc.in_trans = TRANS_NA;
    cmd_desc.input_modifier = (u_int32_t)mpt_index;
    cmd_desc.out_param = outprm;
    cmd_desc.out_param_size = buf_size;
    cmd_desc.out_trans = TRANS_MAILBOX;
    cmd_desc.opcode = ADD_DEV_PREFIX(IF_CMD_QUERY_MPT);
    cmd_desc.opcode_modifier = 0;
    cmd_desc.exec_time_micro = ADD_DEV_PREFIX(IF_CMD_ETIME_QUERY_MPT);

    rc = cmd_invoke(cmd_if, &cmd_desc);
    if ( rc != XHH_CMD_STAT_OK ) {
      FREE(outprm);
      MT_RETURN(rc);
    }
    
    if ( mpt_entry ) {
      EX_BOOL_FLD(mpt_entry, outprm, ADD_DEV_PRM_PREFIX(mpt_st), r_w);
      EX_BOOL_FLD(mpt_entry, outprm, ADD_DEV_PRM_PREFIX(mpt_st), pa);
      EX_BOOL_FLD(mpt_entry, outprm, ADD_DEV_PRM_PREFIX(mpt_st), lr);
      EX_BOOL_FLD(mpt_entry, outprm, ADD_DEV_PRM_PREFIX(mpt_st), lw);
      EX_BOOL_FLD(mpt_entry, outprm, ADD_DEV_PRM_PREFIX(mpt_st), rr);
      EX_BOOL_FLD(mpt_entry, outprm, ADD_DEV_PRM_PREFIX(mpt_st), rw);
      EX_BOOL_FLD(mpt_entry, outprm, ADD_DEV_PRM_PREFIX(mpt_st), a);
      EX_BOOL_FLD(mpt_entry, outprm, ADD_DEV_PRM_PREFIX(mpt_st), eb);

      EX_FLD(mpt_entry, outprm, ADD_DEV_PRM_PREFIX(mpt_st), status);
      EX_FLD(mpt_entry, outprm, ADD_DEV_PRM_PREFIX(mpt_st), page_size);
      EX_FLD(mpt_entry, outprm, ADD_DEV_PRM_PREFIX(mpt_st), pd);
      EX_FLD64(mpt_entry, outprm, ADD_DEV_PRM_PREFIX(mpt_st), start_address);
      EX_FLD64(mpt_entry, outprm, ADD_DEV_PRM_PREFIX(mpt_st), reg_wnd_len);
      EX_FLD(mpt_entry, outprm, ADD_DEV_PRM_PREFIX(mpt_st), lkey);
      EX_FLD(mpt_entry, outprm, ADD_DEV_PRM_PREFIX(mpt_st), win_cnt);
      XHH_CMD_DEV_SPECIFIC_QUERY_MPT
    }
    FREE(outprm);
    MT_RETURN(rc);
}

/*
 *  XHH_cmd_HW2SW_MPT
 */ 
XHH_cmd_status_t XHH_cmd_HW2SW_MPT(XHH_cmd_t cmd_if, XHH_mpt_index_t mpt_index,
                                   XHH_mpt_entry_t *mpt_entry)
{
    command_fields_t cmd_desc;
    u_int8_t *outprm = NULL;
    XHH_cmd_status_t rc;
    u_int32_t buf_size = PSEUDO_MT_BYTE_SIZE(ADD_DEV_PRM_PREFIX(mpt_st));

    FUNC_IN;
    outprm = (u_int8_t *)MALLOC(PSEUDO_MT_BYTE_SIZE(ADD_DEV_PRM_PREFIX(mpt_st)));
    if (outprm == NULL) {
       MT_COV_MEM_FLTR_START
       MT_RETURN(XHH_CMD_STAT_EAGAIN);
       MT_COV_MEM_FLTR_END
    }
    memset(outprm, 0, buf_size);

    cmd_desc.in_param = 0;
    cmd_desc.in_param_size = 0;
    cmd_desc.in_trans = TRANS_NA;
    cmd_desc.input_modifier = (u_int32_t)mpt_index;
    cmd_desc.out_param = outprm;
    cmd_desc.out_param_size = buf_size;
    cmd_desc.out_trans = TRANS_MAILBOX;
    cmd_desc.opcode = ADD_DEV_PREFIX(IF_CMD_HW2SW_MPT);
     /* when output is not necessary putting 1 in opcode modifier
        will case the command execute faster */
    cmd_desc.opcode_modifier = mpt_entry ? 0 : 1;
    cmd_desc.exec_time_micro = ADD_DEV_PREFIX(IF_CMD_ETIME_HW2SW_MPT);

    rc = cmd_invoke(cmd_if, &cmd_desc);
    if ( rc != XHH_CMD_STAT_OK ) {
      FREE(outprm);
      MT_RETURN(rc);
    }
    
    if ( mpt_entry ) {
      EX_BOOL_FLD(mpt_entry, outprm, ADD_DEV_PRM_PREFIX(mpt_st), r_w);
      EX_BOOL_FLD(mpt_entry, outprm, ADD_DEV_PRM_PREFIX(mpt_st), pa);
      EX_BOOL_FLD(mpt_entry, outprm, ADD_DEV_PRM_PREFIX(mpt_st), lr);
      EX_BOOL_FLD(mpt_entry, outprm, ADD_DEV_PRM_PREFIX(mpt_st), lw);
      EX_BOOL_FLD(mpt_entry, outprm, ADD_DEV_PRM_PREFIX(mpt_st), rr);
      EX_BOOL_FLD(mpt_entry, outprm, ADD_DEV_PRM_PREFIX(mpt_st), rw);
      EX_BOOL_FLD(mpt_entry, outprm, ADD_DEV_PRM_PREFIX(mpt_st), a);
      EX_BOOL_FLD(mpt_entry, outprm, ADD_DEV_PRM_PREFIX(mpt_st), eb);
      
      EX_FLD(mpt_entry, outprm, ADD_DEV_PRM_PREFIX(mpt_st), status);
      EX_FLD(mpt_entry, outprm, ADD_DEV_PRM_PREFIX(mpt_st), page_size);
      EX_FLD(mpt_entry, outprm, ADD_DEV_PRM_PREFIX(mpt_st), pd);
      EX_FLD64(mpt_entry, outprm, ADD_DEV_PRM_PREFIX(mpt_st), start_address);
      EX_FLD64(mpt_entry, outprm, ADD_DEV_PRM_PREFIX(mpt_st), reg_wnd_len);
      EX_FLD(mpt_entry, outprm, ADD_DEV_PRM_PREFIX(mpt_st), lkey);
      EX_FLD(mpt_entry, outprm, ADD_DEV_PRM_PREFIX(mpt_st), win_cnt);
      XHH_CMD_DEV_SPECIFIC_HW2SW_MPT
    }
    FREE(outprm);
    MT_RETURN(rc);
}

/*
 *  XHH_cmd_READ_MTT
 */ 
XHH_cmd_status_t XHH_cmd_READ_MTT(XHH_cmd_t cmd_if, u_int64_t mtt_pa, MT_size_t num_elems,
                                   XHH_mtt_entry_t *mtt_entry)
{
    command_fields_t *cmd_desc = NULL;
    u_int8_t *outprm, *iterator;
    int i, local_num_elts, elts_this_loop, max_elts_per_buffer;
    u_int32_t buf_size;
    u_int32_t     mtt_pa_transfer[2];
    XHH_cmd_status_t rc = XHH_CMD_STAT_OK;
    MT_bool   buf_align_adjust = TRUE;
    
    /* TBD:  need to:  a. limit number of entries in the mailbox */
    /*                 b. for performance, if the initial addr is odd, */
    /*                    do a loop of a single element, then do the rest */


    FUNC_IN;
    if (!num_elems) {
        MT_RETURN(XHH_CMD_STAT_BAD_PARAM);
    }

    cmd_desc = (command_fields_t *)MALLOC(sizeof(command_fields_t));
    if ( !cmd_desc ) {
      MT_COV_MEM_FLTR_START
      MT_RETURN(XHH_CMD_STAT_EAGAIN);
      MT_COV_MEM_FLTR_END
    }

    outprm = TNMALLOC(u_int8_t, MAX_OUT_PRM_SIZE);
    if ( !outprm ) {
      MT_COV_MEM_FLTR_START
      FREE(cmd_desc);
      MT_RETURN(XHH_CMD_STAT_EAGAIN);
      MT_COV_MEM_FLTR_END
    }
    memset(outprm, 0, MAX_OUT_PRM_SIZE);

    local_num_elts = (int)num_elems;
    max_elts_per_buffer = (MAX_OUT_PRM_SIZE - MTT_CMD_PA_PREAMBLE_SIZE) / ( PSEUDO_MT_BYTE_SIZE(ADD_DEV_PRM_PREFIX(mtt_st)));

    while(local_num_elts > 0) {

        elts_this_loop = (local_num_elts > max_elts_per_buffer ? max_elts_per_buffer : local_num_elts);

        /* if the mtt_pa address is odd (3 LSBs ignored), and we need to use multiple commands */
        /* and we are also reading an odd number of elements, then decrease the elements in this loop */
        /* by one so that on the next go-around, the reading will start at an even mtt_pa address */
        /* If necessary, the adjustment needs to be performed only once */
        if ((buf_align_adjust) && (local_num_elts > max_elts_per_buffer) && ((mtt_pa>>3)& 0x1) && (!(elts_this_loop & 0x1))) {
            elts_this_loop--;
            buf_align_adjust = FALSE;
        }

        buf_size = elts_this_loop*PSEUDO_MT_BYTE_SIZE(ADD_DEV_PRM_PREFIX(mtt_st));
        memset(outprm, 0, buf_size);
    
        iterator = outprm;

        /* The command interface expects the mtt_pa format to be the HIGH order word first, */
        /* then the low-order word.  The command interface adjusts endianness within words */

        mtt_pa_transfer[0]= (u_int32_t)(mtt_pa >> 32);               /* MS-DWORD */
        mtt_pa_transfer[1]= (u_int32_t)(mtt_pa & 0xFFFFFFFF);        /* LS-DWORD */  
    
        cmd_desc->in_param = (u_int8_t *)&(mtt_pa_transfer[0]);
        cmd_desc->in_param_size = sizeof(mtt_pa_transfer);
        cmd_desc->in_trans = TRANS_IMMEDIATE;
        cmd_desc->input_modifier = elts_this_loop;
        cmd_desc->out_param = outprm;
        cmd_desc->out_param_size = buf_size;
        cmd_desc->out_trans = TRANS_MAILBOX;
        cmd_desc->opcode = ADD_DEV_PREFIX(IF_CMD_READ_MTT);
        cmd_desc->opcode_modifier = 0;
        cmd_desc->exec_time_micro = ADD_DEV_PREFIX(IF_CMD_ETIME_READ_MTT);
    
        rc = cmd_invoke(cmd_if, cmd_desc);
        if ( rc != XHH_CMD_STAT_OK ) {
          FREE(cmd_desc);
          FREE(outprm);
          MT_RETURN(rc);
        }
        
        if ( mtt_entry ) {
          for (i = 0; i < elts_this_loop; i++, mtt_entry++, iterator += PSEUDO_MT_BYTE_SIZE(ADD_DEV_PRM_PREFIX(mtt_st)) ) {
              EX_FLD64_SH(mtt_entry, iterator, ADD_DEV_PRM_PREFIX(mtt_st), ptag);
              EX_BOOL_FLD(mtt_entry, iterator, ADD_DEV_PRM_PREFIX(mtt_st), p);
          }
        }
    
        if (rc != XHH_CMD_STAT_OK) {
	    FREE(cmd_desc);
            FREE(outprm);
            MT_RETURN(rc);
        }
        /* update loop parameters */
        mtt_pa += elts_this_loop * PSEUDO_MT_BYTE_SIZE(ADD_DEV_PRM_PREFIX(mtt_st));    /* incr target pointer */
        local_num_elts -= elts_this_loop;
    }
    FREE(cmd_desc);
    FREE(outprm);
    MT_RETURN(rc);
}

/*
 *  XHH_cmd_WRITE_MTT
 */ 
XHH_cmd_status_t XHH_cmd_WRITE_MTT(XHH_cmd_t cmd_if, u_int64_t mtt_pa, MT_size_t num_elems,
                                   XHH_mtt_entry_t *mtt_entry)
{
    command_fields_t *cmd_desc = NULL;
    u_int8_t *inprm, *iterator;
    int i, local_num_elts, elts_this_loop, max_elts_per_buffer;
    u_int32_t buf_size;
    XHH_cmd_status_t rc = XHH_CMD_STAT_OK;
    MT_bool   buf_align_adjust = TRUE;

    /* TBD:  need to:  a. limit number of entries in the mailbox */
    /*                 b. for performance, if the initial addr is odd, */
    /*                    do one loop of an odd number of elements, then do the rest */

    FUNC_IN;
    if ( !STACK_OK ) MT_RETURN(XHH_CMD_STAT_EFATAL);
    if (!num_elems) {
        MT_RETURN(XHH_CMD_STAT_BAD_PARAM);
    }

    cmd_desc = (command_fields_t *)MALLOC(sizeof(command_fields_t));
    if ( !cmd_desc ) {
      MT_COV_MEM_FLTR_START
      MT_RETURN(XHH_CMD_STAT_EAGAIN);
      MT_COV_MEM_FLTR_END
    }

    inprm = (u_int8_t*)MALLOC(MAX_IN_PRM_SIZE);
    if ( !inprm ) {
      MT_COV_MEM_FLTR_START
      FREE(cmd_desc);
      MT_RETURN(XHH_CMD_STAT_EAGAIN);
      MT_COV_MEM_FLTR_END
    }
    local_num_elts = (u_int32_t)num_elems;
    max_elts_per_buffer = (MAX_IN_PRM_SIZE - MTT_CMD_PA_PREAMBLE_SIZE) / ( PSEUDO_MT_BYTE_SIZE(ADD_DEV_PRM_PREFIX(mtt_st)));

    MTL_DEBUG4("XHH_cmd_WRITE_MTT: local_num_elts = %d, max_elts_per_buffer = %d\n",local_num_elts, max_elts_per_buffer);
    while(local_num_elts > 0) {

        elts_this_loop = (local_num_elts > max_elts_per_buffer ? max_elts_per_buffer : local_num_elts);

        /* if the mtt_pa address is odd (3 LSBs ignored), and we need to use multiple commands */
        /* and we are also writing an odd number of elements, then decrease the elements in this loop */
        /* by one so that on the next go-around, the writing will start at an even mtt_pa address */
        /* If necessary, the adjustment needs to be performed only once */
        if ((buf_align_adjust) && (local_num_elts > max_elts_per_buffer) && ((mtt_pa>>3)& 0x1) && (!(elts_this_loop & 0x1))) {
            elts_this_loop--;
            buf_align_adjust = FALSE;
        }
        if (elts_this_loop <= 0) {
            break;
        }

        buf_size = elts_this_loop*PSEUDO_MT_BYTE_SIZE(ADD_DEV_PRM_PREFIX(mtt_st)) + MTT_CMD_PA_PREAMBLE_SIZE;
        memset(inprm, 0, buf_size);
        iterator = inprm;
        
        MTL_DEBUG4("XHH_cmd_WRITE_MTT: elts_this_loop = %d, buf_size = %d, buf_align_adjust = %d\n",elts_this_loop, buf_size, buf_align_adjust);
    
        cmd_desc->in_param = inprm;
        cmd_desc->in_param_size = buf_size;
        cmd_desc->in_trans = TRANS_MAILBOX;
        cmd_desc->input_modifier = elts_this_loop;
        cmd_desc->out_param = 0;
        cmd_desc->out_param_size = 0;
        cmd_desc->out_trans = TRANS_NA;
        cmd_desc->opcode = ADD_DEV_PREFIX(IF_CMD_WRITE_MTT);
        cmd_desc->opcode_modifier = 0;
        cmd_desc->exec_time_micro = ADD_DEV_PREFIX(IF_CMD_ETIME_WRITE_MTT);
    
        /* copy  */
        ((u_int32_t*)iterator)[0]= (u_int32_t)(mtt_pa >> 32);               /* MS-DWORD */
        ((u_int32_t*)iterator)[1]= (u_int32_t)(mtt_pa & 0xFFFFFFFF);        /* LS-DWORD */  

//		MTL_ERROR1(MT_FLFMT("%s: inprm[0] = 0x%x, inprm[1]=0x%x"),__func__,
//							((u_int32_t*)iterator)[0], ((u_int32_t*)iterator)[1]);
        iterator += MTT_CMD_PA_PREAMBLE_SIZE;
    
        for (i = 0; i < elts_this_loop; i++, mtt_entry++, iterator += PSEUDO_MT_BYTE_SIZE(ADD_DEV_PRM_PREFIX(mtt_st))) {
            INS_FLD64_SH(mtt_entry, iterator, ADD_DEV_PRM_PREFIX(mtt_st), ptag);
            INS_BOOL_FLD(mtt_entry, iterator, ADD_DEV_PRM_PREFIX(mtt_st), p);
        }
        

        XHH_CMD_MAILBOX_PRINT(inprm, buf_size, __func__);
		my_print_mtt_entries(elts_this_loop, inprm);
//        XHH_CMD_PRINT_MTT_ENTRIES(elts_this_loop, inprm);
        rc = cmd_invoke(cmd_if, cmd_desc);
        if (rc != XHH_CMD_STAT_OK) {
	    FREE(cmd_desc);
            FREE(inprm);
            MT_RETURN(rc);
        }
        /* update loop parameters */
        mtt_pa += elts_this_loop * PSEUDO_MT_BYTE_SIZE(ADD_DEV_PRM_PREFIX(mtt_st));    /* incr target pointer */
        local_num_elts -= elts_this_loop;
    }
    FREE(cmd_desc);
    FREE(inprm);
    MT_RETURN(rc);
}


XHH_cmd_status_t XHH_cmd_SYNC_TPT(XHH_cmd_t cmd_if)
{
  command_fields_t cmd_prms = {0};
  XHH_cmd_status_t rc;

  FUNC_IN;
  cmd_prms.opcode = ADD_DEV_PREFIX(IF_CMD_SYNC_TPT);
  cmd_prms.in_trans = TRANS_NA;
  cmd_prms.out_trans = TRANS_NA;
  cmd_prms.exec_time_micro = ADD_DEV_PREFIX(IF_CMD_ETIME_SYNC_TPT);
  rc = cmd_invoke(cmd_if, &cmd_prms);
  if ( rc != XHH_CMD_STAT_OK ) {
    MTL_ERROR1(MT_FLFMT("XHH_cmd_SYNC_TPT failed: %s\n"), str_XHH_cmd_status_t(rc));
  }
  MT_RETURN(rc);
}


/*
 *  XHH_cmd_MAP_EQ
 */ 
XHH_cmd_status_t XHH_cmd_MAP_EQ(XHH_cmd_t cmd_if, XHH_eqn_t eqn, u_int64_t event_mask)
{
  command_fields_t cmd_desc;
  XHH_cmd_status_t rc;
  u_int32_t        event_mask_transfer[2];

  FUNC_IN;


  /* The command interface expects the event_mask format to be the HIGH order word first, */
  /* then the low-order word.  The command interface adjusts endianness within words */

  event_mask_transfer[0]= (u_int32_t)(event_mask >> 32);               /* MS-DWORD */
  event_mask_transfer[1]= (u_int32_t)(event_mask & 0xFFFFFFFF);        /* LS-DWORD */  

  MTL_DEBUG4("XHH_cmd_MAP_EQ: eqn = 0x%x, event_mask = "U64_FMT"\n", eqn, event_mask);
  MTL_DEBUG4("XHH_cmd_MAP_EQ: event_mask_transfer [0] = 0x%x, event_mask_transfer [1] = 0x%x\n",
             event_mask_transfer[0], event_mask_transfer[1]);

  cmd_desc.in_param = (u_int8_t *)&(event_mask_transfer[0]);
  cmd_desc.in_param_size = sizeof(event_mask_transfer);
  cmd_desc.in_trans = TRANS_IMMEDIATE;
  cmd_desc.input_modifier = eqn;
  cmd_desc.out_param = 0;
  cmd_desc.out_param_size = 0;
  cmd_desc.out_trans = TRANS_NA;
  cmd_desc.opcode = ADD_DEV_PREFIX(IF_CMD_MAP_EQ);
  cmd_desc.opcode_modifier = 0;
  cmd_desc.exec_time_micro = ADD_DEV_PREFIX(IF_CMD_ETIME_MAP_EQ);

  rc = cmd_invoke(cmd_if, &cmd_desc);
  MT_RETURN(rc);
}

/*
 *  XHH_cmd_SW2HW_EQ
 */ 
XHH_cmd_status_t XHH_cmd_SW2HW_EQ(XHH_cmd_t cmd_if, XHH_eqn_t eqn, XHH_eqc_t *eq_context)
{
    command_fields_t cmd_desc;
    u_int8_t *inprm = NULL;
    XHH_cmd_status_t rc;
    u_int32_t buf_size = PSEUDO_MT_BYTE_SIZE(ADD_DEV_PRM_PREFIX(eqc_st));

    FUNC_IN;
    if (eq_context == NULL) {MT_RETURN(XHH_CMD_STAT_EBADARG); }
    
    inprm = (u_int8_t *)MALLOC(PSEUDO_MT_BYTE_SIZE(ADD_DEV_PRM_PREFIX(eqc_st)));
    if (inprm == NULL) {
       MT_COV_MEM_FLTR_START
       MT_RETURN(XHH_CMD_STAT_EAGAIN);
       MT_COV_MEM_FLTR_END
    }
    memset(inprm, 0, buf_size);

    XHH_CMD_PRINT_EQ_CONTEXT(eq_context);

    cmd_desc.in_param = inprm;
    cmd_desc.in_param_size = buf_size;
    cmd_desc.in_trans = TRANS_MAILBOX;
    cmd_desc.input_modifier = eqn;
    cmd_desc.out_param = 0;
    cmd_desc.out_param_size = 0;
    cmd_desc.out_trans = TRANS_NA;
    cmd_desc.opcode = ADD_DEV_PREFIX(IF_CMD_SW2HW_EQ);
    cmd_desc.opcode_modifier = 0;
    cmd_desc.exec_time_micro = ADD_DEV_PREFIX(IF_CMD_ETIME_SW2HW_EQ);

    INS_FLD(eq_context, inprm, ADD_DEV_PRM_PREFIX(eqc_st), st);

    INS_BOOL_FLD(eq_context, inprm, ADD_DEV_PRM_PREFIX(eqc_st), tr);

    INS_FLD(eq_context, inprm, ADD_DEV_PRM_PREFIX(eqc_st), owner);
    INS_FLD(eq_context, inprm, ADD_DEV_PRM_PREFIX(eqc_st), status);
    INS_FLD64(eq_context, inprm, ADD_DEV_PRM_PREFIX(eqc_st), start_address);
    INS_FLD(eq_context, inprm, ADD_DEV_PRM_PREFIX(eqc_st), log_eq_size);
    INS_FLD(eq_context, inprm, ADD_DEV_PRM_PREFIX(eqc_st), pd);
    INS_FLD(eq_context, inprm, ADD_DEV_PRM_PREFIX(eqc_st), intr);
    INS_FLD(eq_context, inprm, ADD_DEV_PRM_PREFIX(eqc_st), lkey);
    INS_FLD(eq_context, inprm, ADD_DEV_PRM_PREFIX(eqc_st), consumer_indx);
    INS_FLD(eq_context, inprm, ADD_DEV_PRM_PREFIX(eqc_st), producer_indx);
    XHH_CMD_DEV_SPECIFIC_SW2HW_EQ
    
    XHH_CMD_MAILBOX_PRINT(inprm, buf_size, __func__);

    rc = cmd_invoke(cmd_if, &cmd_desc);
    FREE(inprm);
    MT_RETURN(rc);
}

/*
 *  XHH_cmd_HW2SW_EQ
 */ 
XHH_cmd_status_t XHH_cmd_HW2SW_EQ(XHH_cmd_t cmd_if, XHH_eqn_t eqn, XHH_eqc_t *eq_context)
{
    command_fields_t cmd_desc;
    u_int8_t *outprm = NULL;
    XHH_cmd_status_t rc;
    u_int32_t buf_size = PSEUDO_MT_BYTE_SIZE(ADD_DEV_PRM_PREFIX(eqc_st));

    FUNC_IN;
    outprm = (u_int8_t *)MALLOC(PSEUDO_MT_BYTE_SIZE(ADD_DEV_PRM_PREFIX(eqc_st)));
    if (outprm == NULL) {
       MT_COV_MEM_FLTR_START
       MT_RETURN(XHH_CMD_STAT_EAGAIN);
       MT_COV_MEM_FLTR_END
    }
    memset(outprm, 0, buf_size);

    cmd_desc.in_param = 0;
    cmd_desc.in_param_size = 0;
    cmd_desc.in_trans = TRANS_NA;
    cmd_desc.input_modifier = eqn;
    cmd_desc.out_param = outprm;
    cmd_desc.out_param_size = buf_size;
    cmd_desc.out_trans = TRANS_MAILBOX;
    cmd_desc.opcode = ADD_DEV_PREFIX(IF_CMD_HW2SW_EQ);
    cmd_desc.opcode_modifier = 0;
    cmd_desc.exec_time_micro = ADD_DEV_PREFIX(IF_CMD_ETIME_HW2SW_EQ);

    rc = cmd_invoke(cmd_if, &cmd_desc);
    if ( rc != XHH_CMD_STAT_OK ) {
      FREE(outprm);
      MT_RETURN(rc);
    }
    
    if ( eq_context ) {
      EX_FLD(eq_context, outprm, ADD_DEV_PRM_PREFIX(eqc_st), st);
      EX_BOOL_FLD(eq_context, outprm, ADD_DEV_PRM_PREFIX(eqc_st), tr);
      EX_FLD(eq_context, outprm, ADD_DEV_PRM_PREFIX(eqc_st), owner);
      EX_FLD(eq_context, outprm, ADD_DEV_PRM_PREFIX(eqc_st), status);
      EX_FLD64(eq_context, outprm, ADD_DEV_PRM_PREFIX(eqc_st), start_address);
      EX_FLD(eq_context, outprm, ADD_DEV_PRM_PREFIX(eqc_st), log_eq_size);
      EX_FLD(eq_context, outprm, ADD_DEV_PRM_PREFIX(eqc_st), pd);
      EX_FLD(eq_context, outprm, ADD_DEV_PRM_PREFIX(eqc_st), intr);
      EX_FLD(eq_context, outprm, ADD_DEV_PRM_PREFIX(eqc_st), lkey);
      EX_FLD(eq_context, outprm, ADD_DEV_PRM_PREFIX(eqc_st), consumer_indx);
      EX_FLD(eq_context, outprm, ADD_DEV_PRM_PREFIX(eqc_st), producer_indx);
      XHH_CMD_DEV_SPECIFIC_HW2SW_EQ
    }
    FREE(outprm);
    MT_RETURN(rc);
}

/*
 *  XHH_cmd_QUERY_EQ
 */ 
XHH_cmd_status_t XHH_cmd_QUERY_EQ(XHH_cmd_t cmd_if, XHH_eqn_t eqn, XHH_eqc_t *eq_context)
{
    command_fields_t cmd_desc;
    u_int8_t *outprm = NULL;
    XHH_cmd_status_t rc;
    u_int32_t buf_size = PSEUDO_MT_BYTE_SIZE(ADD_DEV_PRM_PREFIX(eqc_st));

    FUNC_IN;
    outprm = (u_int8_t *)MALLOC(PSEUDO_MT_BYTE_SIZE(ADD_DEV_PRM_PREFIX(eqc_st)));
    if (outprm == NULL) {
       MT_COV_MEM_FLTR_START
       MT_RETURN(XHH_CMD_STAT_EAGAIN);
       MT_COV_MEM_FLTR_END
    }
    memset(outprm, 0, buf_size);

    cmd_desc.in_param = 0;
    cmd_desc.in_param_size = 0;
    cmd_desc.in_trans = TRANS_NA;
    cmd_desc.input_modifier = eqn;
    cmd_desc.out_param = outprm;
    cmd_desc.out_param_size = buf_size;
    cmd_desc.out_trans = TRANS_MAILBOX;
    cmd_desc.opcode = ADD_DEV_PREFIX(IF_CMD_HW2SW_EQ);
    cmd_desc.opcode_modifier = 0;
    cmd_desc.exec_time_micro = ADD_DEV_PREFIX(IF_CMD_ETIME_HW2SW_EQ);

    rc = cmd_invoke(cmd_if, &cmd_desc);
    if ( rc != XHH_CMD_STAT_OK ) {
      FREE(outprm);
      MT_RETURN(rc);
    }
    
    if ( eq_context ) {
      EX_FLD(eq_context, outprm, ADD_DEV_PRM_PREFIX(eqc_st), st);

      EX_BOOL_FLD(eq_context, outprm, ADD_DEV_PRM_PREFIX(eqc_st), tr);
      
      EX_FLD(eq_context, outprm, ADD_DEV_PRM_PREFIX(eqc_st), owner);
      EX_FLD(eq_context, outprm, ADD_DEV_PRM_PREFIX(eqc_st), status);
      EX_FLD64(eq_context, outprm, ADD_DEV_PRM_PREFIX(eqc_st), start_address);
      EX_FLD(eq_context, outprm, ADD_DEV_PRM_PREFIX(eqc_st), log_eq_size);
      EX_FLD(eq_context, outprm, ADD_DEV_PRM_PREFIX(eqc_st), pd);
      EX_FLD(eq_context, outprm, ADD_DEV_PRM_PREFIX(eqc_st), intr);
      EX_FLD(eq_context, outprm, ADD_DEV_PRM_PREFIX(eqc_st), lkey);
      EX_FLD(eq_context, outprm, ADD_DEV_PRM_PREFIX(eqc_st), consumer_indx);
      EX_FLD(eq_context, outprm, ADD_DEV_PRM_PREFIX(eqc_st), producer_indx);
      XHH_CMD_DEV_SPECIFIC_QUERY_EQ
    }
    FREE(outprm);
    MT_RETURN(rc);
}

/*
 *  XHH_cmd_SW2HW_EQ
 */ 
XHH_cmd_status_t XHH_cmd_SW2HW_CQ(XHH_cmd_t cmd_if, HH_cq_hndl_t cqn, XHH_cqc_t *cq_context)
{
    command_fields_t cmd_desc;
    u_int8_t *inprm = NULL;
    XHH_cmd_status_t rc;
    u_int32_t  buf_size = PSEUDO_MT_BYTE_SIZE(ADD_DEV_PRM_PREFIX(completion_queue_context_st));

    FUNC_IN;
    if (cq_context == NULL) {MT_RETURN(XHH_CMD_STAT_EBADARG); }
    
    inprm = (u_int8_t *)MALLOC(PSEUDO_MT_BYTE_SIZE(ADD_DEV_PRM_PREFIX(completion_queue_context_st)));
    if (inprm == NULL) {
       MT_COV_MEM_FLTR_START
       MT_RETURN(XHH_CMD_STAT_EAGAIN);
       MT_COV_MEM_FLTR_END
    }
    memset(inprm, 0, buf_size);

    XHH_CMD_PRINT_CQ_CONTEXT(cq_context);

    cmd_desc.in_param = inprm;
    cmd_desc.in_param_size = buf_size;
    cmd_desc.in_trans = TRANS_MAILBOX;
    cmd_desc.input_modifier = cqn;
    cmd_desc.out_param = 0;
    cmd_desc.out_param_size = 0;
    cmd_desc.out_trans = TRANS_NA;
    cmd_desc.opcode = ADD_DEV_PREFIX(IF_CMD_SW2HW_CQ);
    cmd_desc.opcode_modifier = 0;
    cmd_desc.exec_time_micro = ADD_DEV_PREFIX(IF_CMD_ETIME_SW2HW_CQ);

    INS_FLD(cq_context, inprm, ADD_DEV_PRM_PREFIX(completion_queue_context_st), st);

    INS_BOOL_FLD(cq_context, inprm, ADD_DEV_PRM_PREFIX(completion_queue_context_st), oi);
    
    INS_FLD(cq_context, inprm, ADD_DEV_PRM_PREFIX(completion_queue_context_st), status);
    INS_FLD64(cq_context, inprm, ADD_DEV_PRM_PREFIX(completion_queue_context_st), start_address);
    INS_FLD(cq_context, inprm, ADD_DEV_PRM_PREFIX(completion_queue_context_st), usr_page);
    INS_FLD(cq_context, inprm, ADD_DEV_PRM_PREFIX(completion_queue_context_st), log_cq_size);
    INS_FLD(cq_context, inprm, ADD_DEV_PRM_PREFIX(completion_queue_context_st), c_eqn);
    INS_FLD(cq_context, inprm, ADD_DEV_PRM_PREFIX(completion_queue_context_st), pd);
    INS_FLD(cq_context, inprm, ADD_DEV_PRM_PREFIX(completion_queue_context_st), l_key);
    INS_FLD(cq_context, inprm, ADD_DEV_PRM_PREFIX(completion_queue_context_st), last_notified_indx);
    INS_FLD(cq_context, inprm, ADD_DEV_PRM_PREFIX(completion_queue_context_st), solicit_producer_indx);
    INS_FLD(cq_context, inprm, ADD_DEV_PRM_PREFIX(completion_queue_context_st), cqn);
    XHH_CMD_DEV_SPECIFIC_SW2HW_CQ

    XHH_CMD_MAILBOX_PRINT(inprm, buf_size, __func__);
#if 1
    rc = cmd_invoke(cmd_if, &cmd_desc);
#else
    MTL_DEBUG4("XHH_cmd_SW2HW_CG:  SKIPPING cmd_invoke !!!!!!!!!\n");
    rc = XHH_CMD_STAT_OK;
#endif
    FREE(inprm);
    MT_RETURN(rc);
}

/*
 *  XHH_cmd_HW2SW_EQ
 */ 
XHH_cmd_status_t XHH_cmd_HW2SW_CQ(XHH_cmd_t cmd_if,  HH_cq_hndl_t cqn, XHH_cqc_t *cq_context)
{
    command_fields_t cmd_desc;
    u_int8_t *outprm = NULL;
    XHH_cmd_status_t rc;
    u_int32_t  buf_size = PSEUDO_MT_BYTE_SIZE(ADD_DEV_PRM_PREFIX(completion_queue_context_st));

    FUNC_IN;
    MTL_DEBUG4("XHH_cmd_HW2SW_CQ:  cqn = 0x%x, cq_context = 0x%p\n", cqn, cq_context);
    
    outprm = (u_int8_t *)MALLOC(PSEUDO_MT_BYTE_SIZE(ADD_DEV_PRM_PREFIX(completion_queue_context_st)));
    if (outprm == NULL) {
       MT_COV_MEM_FLTR_START
       MT_RETURN(XHH_CMD_STAT_EAGAIN);
       MT_COV_MEM_FLTR_END
    }
    memset(outprm, 0, buf_size);

    cmd_desc.in_param = 0;
    cmd_desc.in_param_size = 0;
    cmd_desc.in_trans = TRANS_NA;
    cmd_desc.input_modifier = cqn;
    cmd_desc.out_param = outprm;
    cmd_desc.out_param_size = buf_size;
    cmd_desc.out_trans = TRANS_MAILBOX;
    cmd_desc.opcode = ADD_DEV_PREFIX(IF_CMD_HW2SW_CQ);
    cmd_desc.opcode_modifier = 0;
    cmd_desc.exec_time_micro = ADD_DEV_PREFIX(IF_CMD_ETIME_HW2SW_CQ);

    rc = cmd_invoke(cmd_if, &cmd_desc);
    if ( rc != XHH_CMD_STAT_OK ) {
      FREE(outprm);
      MT_RETURN(rc);
    }
    

    if ( cq_context ) {
      EX_FLD(cq_context, outprm, ADD_DEV_PRM_PREFIX(completion_queue_context_st), st);
      
      EX_BOOL_FLD(cq_context, outprm, ADD_DEV_PRM_PREFIX(completion_queue_context_st), oi);
      
      EX_FLD(cq_context, outprm, ADD_DEV_PRM_PREFIX(completion_queue_context_st), status);
      EX_FLD64(cq_context, outprm, ADD_DEV_PRM_PREFIX(completion_queue_context_st), start_address);
      EX_FLD(cq_context, outprm, ADD_DEV_PRM_PREFIX(completion_queue_context_st), usr_page);
      EX_FLD(cq_context, outprm, ADD_DEV_PRM_PREFIX(completion_queue_context_st), log_cq_size);
      EX_FLD(cq_context, outprm, ADD_DEV_PRM_PREFIX(completion_queue_context_st), c_eqn);
      EX_FLD(cq_context, outprm, ADD_DEV_PRM_PREFIX(completion_queue_context_st), pd);
      EX_FLD(cq_context, outprm, ADD_DEV_PRM_PREFIX(completion_queue_context_st), l_key);
      EX_FLD(cq_context, outprm, ADD_DEV_PRM_PREFIX(completion_queue_context_st), last_notified_indx);
      EX_FLD(cq_context, outprm, ADD_DEV_PRM_PREFIX(completion_queue_context_st), solicit_producer_indx);
      EX_FLD(cq_context, outprm, ADD_DEV_PRM_PREFIX(completion_queue_context_st), cqn);
      XHH_CMD_DEV_SPECIFIC_HW2SW_CQ

      XHH_CMD_PRINT_CQ_CONTEXT(cq_context);
    }
    FREE(outprm);
    MT_RETURN(rc);
}

/*
 *  XHH_cmd_QUERY_CQ
 */ 
XHH_cmd_status_t XHH_cmd_QUERY_CQ(XHH_cmd_t cmd_if, HH_cq_hndl_t cqn, XHH_cqc_t *cq_context)
{
    command_fields_t cmd_desc;
    u_int8_t *outprm = NULL;
    XHH_cmd_status_t rc;
    u_int32_t buf_size = PSEUDO_MT_BYTE_SIZE(ADD_DEV_PRM_PREFIX(completion_queue_context_st));

    FUNC_IN;
    outprm = (u_int8_t *)MALLOC(PSEUDO_MT_BYTE_SIZE(ADD_DEV_PRM_PREFIX(completion_queue_context_st)));
    if (outprm == NULL) {
       MT_COV_MEM_FLTR_START
       MT_RETURN(XHH_CMD_STAT_EAGAIN);
       MT_COV_MEM_FLTR_END
    }
    memset(outprm, 0, buf_size);

    cmd_desc.in_param = 0;
    cmd_desc.in_param_size = 0;
    cmd_desc.in_trans = TRANS_NA;
    cmd_desc.input_modifier = cqn;
    cmd_desc.out_param = outprm;
    cmd_desc.out_param_size = buf_size;
    cmd_desc.out_trans = TRANS_MAILBOX;
    cmd_desc.opcode = ADD_DEV_PREFIX(IF_CMD_QUERY_CQ);
    cmd_desc.opcode_modifier = 0;
    cmd_desc.exec_time_micro = ADD_DEV_PREFIX(IF_CMD_ETIME_QUERY_CQ);

    rc = cmd_invoke(cmd_if, &cmd_desc);
    if ( rc != XHH_CMD_STAT_OK ) {
      FREE(outprm);
      MT_RETURN(rc);
    }
    
    if ( cq_context ) {
      EX_FLD(cq_context, outprm, ADD_DEV_PRM_PREFIX(completion_queue_context_st), st);
      
      EX_BOOL_FLD(cq_context, outprm, ADD_DEV_PRM_PREFIX(completion_queue_context_st), oi);
      
      EX_FLD(cq_context, outprm, ADD_DEV_PRM_PREFIX(completion_queue_context_st), status);
      EX_FLD64(cq_context, outprm, ADD_DEV_PRM_PREFIX(completion_queue_context_st), start_address);
      EX_FLD(cq_context, outprm, ADD_DEV_PRM_PREFIX(completion_queue_context_st), usr_page);
      EX_FLD(cq_context, outprm, ADD_DEV_PRM_PREFIX(completion_queue_context_st), log_cq_size);
      EX_FLD(cq_context, outprm, ADD_DEV_PRM_PREFIX(completion_queue_context_st), c_eqn);
      EX_FLD(cq_context, outprm, ADD_DEV_PRM_PREFIX(completion_queue_context_st), pd);
      EX_FLD(cq_context, outprm, ADD_DEV_PRM_PREFIX(completion_queue_context_st), l_key);
      EX_FLD(cq_context, outprm, ADD_DEV_PRM_PREFIX(completion_queue_context_st), last_notified_indx);
      EX_FLD(cq_context, outprm, ADD_DEV_PRM_PREFIX(completion_queue_context_st), solicit_producer_indx);
      EX_FLD(cq_context, outprm, ADD_DEV_PRM_PREFIX(completion_queue_context_st), cqn);
      XHH_CMD_DEV_SPECIFIC_QUERY_CQ

      XHH_CMD_PRINT_CQ_CONTEXT(cq_context);
    }
    FREE(outprm);
    MT_RETURN(rc);
}

/*
 *  XHH_cmd_RESIZE_CQ
 */ 
XHH_cmd_status_t XHH_cmd_RESIZE_CQ(XHH_cmd_t cmd_if, HH_cq_hndl_t cqn, 
                                   u_int64_t start_address, u_int32_t l_key, u_int8_t log_cq_size,
                                   u_int32_t *new_producer_index_p)
{
  command_fields_t *cmd_desc = NULL;
  u_int8_t *inprm = NULL;
  XHH_cmd_status_t rc;
  const u_int32_t  buf_size = PSEUDO_MT_BYTE_SIZE(ADD_DEV_PRM_PREFIX(resize_cq_st));
  u_int32_t out_param_tmp[2];

  FUNC_IN;
  cmd_desc = (command_fields_t *)MALLOC(sizeof(command_fields_t));
  if (cmd_desc == NULL) {
      MT_COV_MEM_FLTR_START
      MT_RETURN(XHH_CMD_STAT_EAGAIN);
      MT_COV_MEM_FLTR_END
  }
  inprm = (u_int8_t *)MALLOC(PSEUDO_MT_BYTE_SIZE(ADD_DEV_PRM_PREFIX(resize_cq_st)));
  if (inprm == NULL) {
     MT_COV_MEM_FLTR_START
     FREE(cmd_desc);
     MT_RETURN(XHH_CMD_STAT_EAGAIN);
     MT_COV_MEM_FLTR_END
  }
  memset(inprm, 0, buf_size);

  cmd_desc->in_param = (u_int8_t*)inprm;
  cmd_desc->in_param_size = buf_size;
  cmd_desc->in_trans = TRANS_MAILBOX;
  cmd_desc->input_modifier = cqn;
  cmd_desc->out_trans = TRANS_IMMEDIATE;
  cmd_desc->out_param = (u_int8_t*)out_param_tmp;
  cmd_desc->opcode = ADD_DEV_PREFIX(IF_CMD_RESIZE_CQ);
  cmd_desc->opcode_modifier = new_producer_index_p ? 0 /* legacy mode */: 
                                                    1 /* fixed resize: new_pi= old_pi % new_size */;
  cmd_desc->exec_time_micro = ADD_DEV_PREFIX(IF_CMD_ETIME_RESIZE_CQ);

  memset(inprm,0,buf_size);
  
  MT_INSERT_ARRAY32(inprm, (u_int32_t)(start_address >> 32) , 
                 MT_BIT_OFFSET(ADD_DEV_PRM_PREFIX(resize_cq_st), start_addr_h), 
		 MT_BIT_SIZE(ADD_DEV_PRM_PREFIX(resize_cq_st), start_addr_h)) ;
  MT_INSERT_ARRAY32(inprm, (u_int32_t)(start_address & MAKE_ULONGLONG(0xFFFFFFFF)) , 
                 MT_BIT_OFFSET(ADD_DEV_PRM_PREFIX(resize_cq_st), start_addr_l), 
		 MT_BIT_SIZE(ADD_DEV_PRM_PREFIX(resize_cq_st), start_addr_l)) ;
  MT_INSERT_ARRAY32(inprm, log_cq_size , 
                 MT_BIT_OFFSET(ADD_DEV_PRM_PREFIX(resize_cq_st), log_cq_size), 
		 MT_BIT_SIZE(ADD_DEV_PRM_PREFIX(resize_cq_st), log_cq_size)) ;
  MT_INSERT_ARRAY32(inprm, l_key , 
                 MT_BIT_OFFSET(ADD_DEV_PRM_PREFIX(resize_cq_st), l_key), 
		 MT_BIT_SIZE(ADD_DEV_PRM_PREFIX(resize_cq_st), l_key)) ; 
  
  XHH_CMD_MAILBOX_PRINT(inprm, buf_size, __func__);
  rc = cmd_invoke(cmd_if, cmd_desc);
  if (new_producer_index_p) *new_producer_index_p= out_param_tmp[0]; /* new producer index is in out_param_h */
  FREE(cmd_desc);
  FREE(inprm);
  MT_RETURN(rc);
}


/*
 *  GET_QP_CMD_EXEC_TIME -- returns QP command exec time in microseconds
 */ 
static u_int32_t get_qp_cmd_exec_time(XHH_qpee_transition_t trans)
{
    switch(trans) {
    case QPEE_TRANS_RST2INIT:
        return ADD_DEV_PREFIX(IF_CMD_ETIME_RST2INIT_QPEE);
    case QPEE_TRANS_INIT2INIT:
        return ADD_DEV_PREFIX(IF_CMD_ETIME_INIT2INIT_QPEE);
    case QPEE_TRANS_INIT2RTR:
        return ADD_DEV_PREFIX(IF_CMD_ETIME_INIT2RTR_QPEE);
    case QPEE_TRANS_RTR2RTS:
        return ADD_DEV_PREFIX(IF_CMD_ETIME_RTR2RTS_QPEE);
    case QPEE_TRANS_RTS2RTS:
        return ADD_DEV_PREFIX(IF_CMD_ETIME_RTS2RTS_QPEE);
    case QPEE_TRANS_SQERR2RTS:
        return ADD_DEV_PREFIX(IF_CMD_ETIME_SQERR2RTS_QPEE);
    case QPEE_TRANS_SQD2RTS:
        return ADD_DEV_PREFIX(IF_CMD_ETIME_SQD2RTS_QPEE);
    case QPEE_TRANS_2ERR:
        return ADD_DEV_PREFIX(IF_CMD_ETIME_2ERR_QPEE);
    case QPEE_TRANS_RTS2SQD:
        return ADD_DEV_PREFIX(IF_CMD_ETIME_RTS2SQD);
    case QPEE_TRANS_ERR2RST:
        return ADD_DEV_PREFIX(IF_CMD_ETIME_ERR2RST_QPEE);
    default:
        MTL_ERROR1(MT_FLFMT("no such qp transition exists \n")); 
        return 0;
    }
}
/*
 *  XHH_cmd_MODIFY_QPEE 
 *                   is_ee:  0 = QP, 1 = EE
 */ 
static XHH_cmd_status_t XHH_cmd_MODIFY_QPEE( XHH_cmd_t cmd_if, MT_bool is_ee, u_int32_t qpn, XHH_qpee_transition_t trans,
                                   XHH_qpee_context_t *qp_context,u_int32_t  optparammask)
{
    command_fields_t cmd_desc;
    u_int8_t *inprm = NULL;
    XHH_cmd_status_t rc;
    u_int32_t buf_size, sqd_event_req;
    u_int32_t  temp_u32;
    ADD_DEV_PREFIX_LC(if_qp_state_t) device_if_qp_state;

    FUNC_IN;
    
	/* we save the value of sqd_event bit in xaction field & clr it to receive a normal xaction value. */
	sqd_event_req 	= trans & XHH_CMD_SQD_EVENT_REQ;
	trans			&= ~XHH_CMD_SQD_EVENT_REQ; 
	
	CMDS_DBG("%s: TRANSACTION val = 0x%x\n", __func__, trans);

    /* see which transition was requested */
    switch(trans) {
    /* have input mailbox only */
	case QPEE_TRANS_RST2INIT:
#ifdef MT_QP_STARTS_ARMED
	optparammask = 0;
#endif
    case QPEE_TRANS_INIT2INIT:
    case QPEE_TRANS_INIT2RTR:
    case QPEE_TRANS_RTR2RTS:
    case QPEE_TRANS_RTS2RTS:
    case QPEE_TRANS_SQERR2RTS:
    case QPEE_TRANS_SQD2RTS:
        if (qp_context == NULL) {MT_RETURN(XHH_CMD_STAT_EBADARG); }
    
        inprm = (u_int8_t *)MALLOC(PSEUDO_MT_BYTE_SIZE(ADD_DEV_PRM_PREFIX(qp_ee_state_transitions_st)));
        if (inprm == NULL) {
           MT_COV_MEM_FLTR_START
           MT_RETURN(XHH_CMD_STAT_EAGAIN);
           MT_COV_MEM_FLTR_END
        }
        
        buf_size = PSEUDO_MT_BYTE_SIZE(ADD_DEV_PRM_PREFIX(qp_ee_state_transitions_st));
        memset(inprm, 0, buf_size);

        cmd_desc.in_param = inprm;
        cmd_desc.in_param_size = buf_size;
        cmd_desc.in_trans = TRANS_MAILBOX;
        cmd_desc.input_modifier = qpn | (is_ee ? 0x1000000 : 0);
        cmd_desc.out_param = 0;
        cmd_desc.out_param_size = 0;
        cmd_desc.out_trans = TRANS_NA;
        cmd_desc.opcode = (ADD_DEV_PREFIX_LC(if_cmd_t))trans;
        cmd_desc.opcode_modifier = 0;
        cmd_desc.exec_time_micro = get_qp_cmd_exec_time(trans);

        XHH_CMD_PRINT_QP_CONTEXT(qp_context);

        /* translate VAPI qp state to Tavor qp state, and fail if not valid*/
        if (!XHH_vapi_qpstate_2_tavor_qpstate(qp_context->state, &device_if_qp_state)) {
            CMDS_DBG("%s: VAPI QP state (0x%x) is not valid\n", __func__, qp_context->state);
            rc = XHH_CMD_STAT_EFATAL;
            goto retn;
        }

        MT_INSERT_ARRAY32(inprm, optparammask, MT_BIT_OFFSET(ADD_DEV_PRM_PREFIX(qp_ee_state_transitions_st), opt_param_mask), 
                       MT_BIT_SIZE(ADD_DEV_PRM_PREFIX(qp_ee_state_transitions_st), opt_param_mask));

		// 'te' field has been removed.
        //QP_INS_FLD(qp_context, inprm, ADD_DEV_PRM_PREFIX(qp_ee_state_transitions_st), te);
        QP_INS_BOOL_FLD(qp_context, inprm, ADD_DEV_PRM_PREFIX(qp_ee_state_transitions_st), de);

        QP_INS_FLD(qp_context, inprm, ADD_DEV_PRM_PREFIX(qp_ee_state_transitions_st),pm_state);
        QP_INS_FLD(qp_context, inprm, ADD_DEV_PRM_PREFIX(qp_ee_state_transitions_st), st);

        MT_INSERT_ARRAY32(inprm, device_if_qp_state, MT_BIT_OFFSET(ADD_DEV_PRM_PREFIX(qp_ee_state_transitions_st), qpc_eec_data.state), 
                       MT_BIT_SIZE(ADD_DEV_PRM_PREFIX(qp_ee_state_transitions_st), qpc_eec_data.state));
        
        QP_INS_FLD(qp_context, inprm, ADD_DEV_PRM_PREFIX(qp_ee_state_transitions_st), sched_queue);
        QP_INS_FLD(qp_context, inprm, ADD_DEV_PRM_PREFIX(qp_ee_state_transitions_st), msg_max);
        QP_INS_FLD(qp_context, inprm, ADD_DEV_PRM_PREFIX(qp_ee_state_transitions_st), mtu);
        QP_INS_FLD(qp_context, inprm, ADD_DEV_PRM_PREFIX(qp_ee_state_transitions_st), usr_page);
        QP_INS_FLD(qp_context, inprm, ADD_DEV_PRM_PREFIX(qp_ee_state_transitions_st), local_qpn_een);
        QP_INS_FLD(qp_context, inprm, ADD_DEV_PRM_PREFIX(qp_ee_state_transitions_st), remote_qpn_een);

        QP_INS_FLD(qp_context, inprm, ADD_DEV_PRM_PREFIX(qp_ee_state_transitions_st), primary_address_path.pkey_index);
        QP_INS_FLD(qp_context, inprm, ADD_DEV_PRM_PREFIX(qp_ee_state_transitions_st), primary_address_path.port_number);
        QP_INS_FLD(qp_context, inprm, ADD_DEV_PRM_PREFIX(qp_ee_state_transitions_st), primary_address_path.rlid);
        QP_INS_FLD(qp_context, inprm, ADD_DEV_PRM_PREFIX(qp_ee_state_transitions_st), primary_address_path.my_lid_path_bits);
        
        QP_INS_BOOL_FLD(qp_context, inprm, ADD_DEV_PRM_PREFIX(qp_ee_state_transitions_st), primary_address_path.g);
        
        QP_INS_FLD(qp_context, inprm, ADD_DEV_PRM_PREFIX(qp_ee_state_transitions_st), primary_address_path.rnr_retry);
        QP_INS_FLD(qp_context, inprm, ADD_DEV_PRM_PREFIX(qp_ee_state_transitions_st), primary_address_path.hop_limit);
        QP_INS_FLD(qp_context, inprm, ADD_DEV_PRM_PREFIX(qp_ee_state_transitions_st), primary_address_path.max_stat_rate);
        QP_INS_FLD(qp_context, inprm, ADD_DEV_PRM_PREFIX(qp_ee_state_transitions_st), primary_address_path.mgid_index);
        QP_INS_FLD(qp_context, inprm, ADD_DEV_PRM_PREFIX(qp_ee_state_transitions_st), primary_address_path.ack_timeout);
        QP_INS_FLD(qp_context, inprm, ADD_DEV_PRM_PREFIX(qp_ee_state_transitions_st), primary_address_path.flow_label);
        QP_INS_FLD(qp_context, inprm, ADD_DEV_PRM_PREFIX(qp_ee_state_transitions_st), primary_address_path.tclass);
        QP_INS_FLD(qp_context, inprm, ADD_DEV_PRM_PREFIX(qp_ee_state_transitions_st), primary_address_path.sl);

        memcpy(&temp_u32, &(qp_context->primary_address_path.rgid[0]), sizeof(u_int32_t));
        MT_INSERT_ARRAY32(inprm, MOSAL_be32_to_cpu(temp_u32), 
                       MT_BIT_OFFSET(ADD_DEV_PRM_PREFIX(qp_ee_state_transitions_st), qpc_eec_data.primary_address_path.rgid_127_96),
                       MT_BIT_SIZE(ADD_DEV_PRM_PREFIX(qp_ee_state_transitions_st), qpc_eec_data.primary_address_path.rgid_127_96));
        
        memcpy(&temp_u32, &(qp_context->primary_address_path.rgid[4]), sizeof(u_int32_t));
        MT_INSERT_ARRAY32(inprm, MOSAL_be32_to_cpu(temp_u32), 
                       MT_BIT_OFFSET(ADD_DEV_PRM_PREFIX(qp_ee_state_transitions_st), qpc_eec_data.primary_address_path.rgid_95_64),
                       MT_BIT_SIZE(ADD_DEV_PRM_PREFIX(qp_ee_state_transitions_st), qpc_eec_data.primary_address_path.rgid_95_64));

        memcpy(&temp_u32, &(qp_context->primary_address_path.rgid[8]), sizeof(u_int32_t));
        MT_INSERT_ARRAY32(inprm, MOSAL_be32_to_cpu(temp_u32), 
                       MT_BIT_OFFSET(ADD_DEV_PRM_PREFIX(qp_ee_state_transitions_st), qpc_eec_data.primary_address_path.rgid_63_32),
                       MT_BIT_SIZE(ADD_DEV_PRM_PREFIX(qp_ee_state_transitions_st), qpc_eec_data.primary_address_path.rgid_63_32));

        memcpy(&temp_u32, &(qp_context->primary_address_path.rgid[12]), sizeof(u_int32_t));
        MT_INSERT_ARRAY32(inprm, MOSAL_be32_to_cpu(temp_u32), 
                       MT_BIT_OFFSET(ADD_DEV_PRM_PREFIX(qp_ee_state_transitions_st), qpc_eec_data.primary_address_path.rgid_31_0),
                       MT_BIT_SIZE(ADD_DEV_PRM_PREFIX(qp_ee_state_transitions_st), qpc_eec_data.primary_address_path.rgid_31_0));

        QP_INS_FLD(qp_context, inprm, ADD_DEV_PRM_PREFIX(qp_ee_state_transitions_st), alternative_address_path.pkey_index);
        QP_INS_FLD(qp_context, inprm, ADD_DEV_PRM_PREFIX(qp_ee_state_transitions_st), alternative_address_path.port_number);
        QP_INS_FLD(qp_context, inprm, ADD_DEV_PRM_PREFIX(qp_ee_state_transitions_st), alternative_address_path.rlid);
        QP_INS_FLD(qp_context, inprm, ADD_DEV_PRM_PREFIX(qp_ee_state_transitions_st), alternative_address_path.my_lid_path_bits);
        
        QP_INS_BOOL_FLD(qp_context, inprm, ADD_DEV_PRM_PREFIX(qp_ee_state_transitions_st), alternative_address_path.g);

        QP_INS_FLD(qp_context, inprm, ADD_DEV_PRM_PREFIX(qp_ee_state_transitions_st), alternative_address_path.rnr_retry);

        QP_INS_FLD(qp_context, inprm, ADD_DEV_PRM_PREFIX(qp_ee_state_transitions_st), alternative_address_path.hop_limit);
        QP_INS_FLD(qp_context, inprm, ADD_DEV_PRM_PREFIX(qp_ee_state_transitions_st), alternative_address_path.max_stat_rate);
        QP_INS_FLD(qp_context, inprm, ADD_DEV_PRM_PREFIX(qp_ee_state_transitions_st), alternative_address_path.mgid_index);
        QP_INS_FLD(qp_context, inprm, ADD_DEV_PRM_PREFIX(qp_ee_state_transitions_st), alternative_address_path.ack_timeout);
        QP_INS_FLD(qp_context, inprm, ADD_DEV_PRM_PREFIX(qp_ee_state_transitions_st), alternative_address_path.flow_label);
        QP_INS_FLD(qp_context, inprm, ADD_DEV_PRM_PREFIX(qp_ee_state_transitions_st), alternative_address_path.tclass);
        QP_INS_FLD(qp_context, inprm, ADD_DEV_PRM_PREFIX(qp_ee_state_transitions_st), alternative_address_path.sl);

        memcpy(&temp_u32, &(qp_context->alternative_address_path.rgid[0]), sizeof(u_int32_t));
        MT_INSERT_ARRAY32(inprm, MOSAL_be32_to_cpu(temp_u32), 
                       MT_BIT_OFFSET(ADD_DEV_PRM_PREFIX(qp_ee_state_transitions_st), qpc_eec_data.alternative_address_path.rgid_127_96),
                       MT_BIT_SIZE(ADD_DEV_PRM_PREFIX(qp_ee_state_transitions_st), qpc_eec_data.alternative_address_path.rgid_127_96));
        
        memcpy(&temp_u32, &(qp_context->alternative_address_path.rgid[4]), sizeof(u_int32_t));
        MT_INSERT_ARRAY32(inprm, MOSAL_be32_to_cpu(temp_u32), 
                       MT_BIT_OFFSET(ADD_DEV_PRM_PREFIX(qp_ee_state_transitions_st), qpc_eec_data.alternative_address_path.rgid_95_64),
                       MT_BIT_SIZE(ADD_DEV_PRM_PREFIX(qp_ee_state_transitions_st), qpc_eec_data.alternative_address_path.rgid_95_64));

        memcpy(&temp_u32, &(qp_context->alternative_address_path.rgid[8]), sizeof(u_int32_t));
        MT_INSERT_ARRAY32(inprm, MOSAL_be32_to_cpu(temp_u32), 
                       MT_BIT_OFFSET(ADD_DEV_PRM_PREFIX(qp_ee_state_transitions_st), qpc_eec_data.alternative_address_path.rgid_63_32),
                       MT_BIT_SIZE(ADD_DEV_PRM_PREFIX(qp_ee_state_transitions_st), qpc_eec_data.alternative_address_path.rgid_63_32));

        memcpy(&temp_u32, &(qp_context->alternative_address_path.rgid[12]), sizeof(u_int32_t));
        MT_INSERT_ARRAY32(inprm, MOSAL_be32_to_cpu(temp_u32), 
                       MT_BIT_OFFSET(ADD_DEV_PRM_PREFIX(qp_ee_state_transitions_st), qpc_eec_data.alternative_address_path.rgid_31_0),
                       MT_BIT_SIZE(ADD_DEV_PRM_PREFIX(qp_ee_state_transitions_st), qpc_eec_data.alternative_address_path.rgid_31_0));

        QP_INS_FLD(qp_context, inprm, ADD_DEV_PRM_PREFIX(qp_ee_state_transitions_st), rdd);
        QP_INS_FLD(qp_context, inprm, ADD_DEV_PRM_PREFIX(qp_ee_state_transitions_st), pd);
        QP_INS_FLD(qp_context, inprm, ADD_DEV_PRM_PREFIX(qp_ee_state_transitions_st), wqe_lkey);
        
        QP_INS_BOOL_FLD(qp_context, inprm, ADD_DEV_PRM_PREFIX(qp_ee_state_transitions_st), ssc);
        QP_INS_BOOL_FLD(qp_context, inprm, ADD_DEV_PRM_PREFIX(qp_ee_state_transitions_st), sic);
        QP_INS_BOOL_FLD(qp_context, inprm, ADD_DEV_PRM_PREFIX(qp_ee_state_transitions_st), sae);
        QP_INS_BOOL_FLD(qp_context, inprm, ADD_DEV_PRM_PREFIX(qp_ee_state_transitions_st), swe);
        QP_INS_BOOL_FLD(qp_context, inprm, ADD_DEV_PRM_PREFIX(qp_ee_state_transitions_st), sre);
        
        QP_INS_FLD(qp_context, inprm, ADD_DEV_PRM_PREFIX(qp_ee_state_transitions_st), retry_count);
        QP_INS_FLD(qp_context, inprm, ADD_DEV_PRM_PREFIX(qp_ee_state_transitions_st), sra_max);
        QP_INS_FLD(qp_context, inprm, ADD_DEV_PRM_PREFIX(qp_ee_state_transitions_st), flight_lim);
        QP_INS_FLD(qp_context, inprm, ADD_DEV_PRM_PREFIX(qp_ee_state_transitions_st), ack_req_freq);
        QP_INS_FLD(qp_context, inprm, ADD_DEV_PRM_PREFIX(qp_ee_state_transitions_st), next_send_psn);
        QP_INS_FLD(qp_context, inprm, ADD_DEV_PRM_PREFIX(qp_ee_state_transitions_st), cqn_snd);
        
        QP_INS_BOOL_FLD(qp_context, inprm, ADD_DEV_PRM_PREFIX(qp_ee_state_transitions_st), rsc);
        QP_INS_BOOL_FLD(qp_context, inprm, ADD_DEV_PRM_PREFIX(qp_ee_state_transitions_st), ric);
        QP_INS_BOOL_FLD(qp_context, inprm, ADD_DEV_PRM_PREFIX(qp_ee_state_transitions_st), rae);
        QP_INS_BOOL_FLD(qp_context, inprm, ADD_DEV_PRM_PREFIX(qp_ee_state_transitions_st), rwe);
        QP_INS_BOOL_FLD(qp_context, inprm, ADD_DEV_PRM_PREFIX(qp_ee_state_transitions_st), rre);
        
        QP_INS_FLD(qp_context, inprm, ADD_DEV_PRM_PREFIX(qp_ee_state_transitions_st), rra_max);
        QP_INS_FLD(qp_context, inprm, ADD_DEV_PRM_PREFIX(qp_ee_state_transitions_st), next_rcv_psn);
        QP_INS_FLD(qp_context, inprm, ADD_DEV_PRM_PREFIX(qp_ee_state_transitions_st), min_rnr_nak);
        QP_INS_FLD(qp_context, inprm, ADD_DEV_PRM_PREFIX(qp_ee_state_transitions_st), ra_buff_indx);
        QP_INS_FLD(qp_context, inprm, ADD_DEV_PRM_PREFIX(qp_ee_state_transitions_st), cqn_rcv);
        QP_INS_FLD(qp_context, inprm, ADD_DEV_PRM_PREFIX(qp_ee_state_transitions_st), q_key);
        QP_INS_FLD(qp_context, inprm, ADD_DEV_PRM_PREFIX(qp_ee_state_transitions_st), srqn);
        
        QP_INS_BOOL_FLD(qp_context, inprm, ADD_DEV_PRM_PREFIX(qp_ee_state_transitions_st), srq);
        XHH_CMD_DEV_SPECIFIC_MODIFY_QPEE

        /* my_cmd_mailbox_print(inprm,buf_size,__func__); */
        rc = cmd_invoke(cmd_if, &cmd_desc);
        goto retn;

        /* No mailboxes, and no immed data */
    case QPEE_TRANS_2ERR:
    case QPEE_TRANS_RTS2SQD:
        cmd_desc.in_param = 0;
        cmd_desc.in_param_size = 0;
        cmd_desc.in_trans = TRANS_NA;
        cmd_desc.input_modifier = qpn | ( (sqd_event_req && (trans == QPEE_TRANS_RTS2SQD)) ? ADD_DEV_PREFIX(IF_SQD_EVENT_FLAG):0 );
        cmd_desc.input_modifier |= (is_ee ? 0x1000000 : 0);
        cmd_desc.out_param = 0;
        cmd_desc.out_param_size = 0;
        cmd_desc.out_trans = TRANS_NA;
        cmd_desc.opcode = (ADD_DEV_PREFIX_LC(if_cmd_t))trans;
        cmd_desc.opcode_modifier = 0;
        cmd_desc.exec_time_micro = get_qp_cmd_exec_time(trans);

        rc = cmd_invoke(cmd_if, &cmd_desc);
        goto retn;

        /* output mailbox only */
    /* matan: this is now the ANY2RST xition. */
	case QPEE_TRANS_ERR2RST:

        cmd_desc.in_param = 0;
        cmd_desc.in_param_size = 0;
        cmd_desc.in_trans = TRANS_NA;
        cmd_desc.input_modifier = qpn | (is_ee ? 0x1000000 : 0);
        cmd_desc.out_param = 0;        /* not using outbox */
        cmd_desc.out_param_size = 0;   /* not using outbox */
        cmd_desc.out_trans = TRANS_MAILBOX;
        cmd_desc.opcode = (ADD_DEV_PREFIX_LC(if_cmd_t))trans;
        /* matan: ANY2RST is always called with (opcode_modifier |= 2), meaning no need 
		       to move into ERR before RST. Also, set LSB so that no outbox will be generated */
		    cmd_desc.opcode_modifier = 3 ;  /* bits 0 and 1 set */
        cmd_desc.exec_time_micro = get_qp_cmd_exec_time(trans);

        rc = cmd_invoke(cmd_if, &cmd_desc);
        goto retn;
	default:
		MTL_ERROR1("%s: BAD TRANSACTION val = 0x%x\n", __func__, trans);
        CMDS_DBG("%s: BAD TRANSACTION val = 0x%x\n", __func__, trans);
        MT_RETURN(XHH_CMD_STAT_EBADARG);
    }
    MT_RETURN(XHH_CMD_STAT_EFATAL);  // ??? BAD_ARG ??

retn:
   if (inprm != NULL) {
       FREE(inprm);
   }
   MT_RETURN(rc);
}

/*
 *  XHH_cmd_MODIFY_QP 
 */ 
XHH_cmd_status_t XHH_cmd_MODIFY_QP(XHH_cmd_t cmd_if, IB_wqpn_t qpn,
                                   XHH_qpee_transition_t trans,
                                   XHH_qpee_context_t *qp_context,
                                   u_int32_t           optparammask)
{
  CMDS_DBG("%s: TRANSACTION val = 0x%x\n", __func__, trans);
  return (XHH_cmd_MODIFY_QPEE(cmd_if,0,qpn,trans,qp_context,optparammask));
}

/*
 *  XHH_cmd_MODIFY_EE 
 */ 
XHH_cmd_status_t XHH_cmd_MODIFY_EE(XHH_cmd_t cmd_if, IB_eecn_t eecn,
                                   XHH_qpee_transition_t trans,
                                   XHH_qpee_context_t *ee_context,
                                   u_int32_t           optparammask)
{
  return (XHH_cmd_MODIFY_QPEE(cmd_if,1,eecn,trans,ee_context,optparammask));
}


/*
 *  XHH_cmd_QUERY_QPEE 
 */ 
static XHH_cmd_status_t XHH_cmd_QUERY_QPEE( XHH_cmd_t cmd_if,  MT_bool is_ee, u_int32_t qpn,
                                   XHH_qpee_context_t *qp_context)
{
    command_fields_t cmd_desc;
    u_int8_t *outprm;
    XHH_cmd_status_t rc;
    u_int32_t buf_size = PSEUDO_MT_BYTE_SIZE(ADD_DEV_PRM_PREFIX(qp_ee_state_transitions_st));
    u_int32_t temp_u32;
    ADD_DEV_PREFIX_LC(if_qp_state_t) device_if_qp_state;
    	
    FUNC_IN;
    outprm = TNMALLOC(u_int8_t, buf_size);
    if ( !outprm ) {
      MT_COV_MEM_FLTR_START
      MT_RETURN(XHH_CMD_STAT_EAGAIN);
      MT_COV_MEM_FLTR_END
    }
    memset(outprm, 0, buf_size);

    cmd_desc.in_param = 0;
    cmd_desc.in_param_size = 0;
    cmd_desc.in_trans = TRANS_NA;
    cmd_desc.input_modifier = qpn;
    cmd_desc.out_param = outprm;
    cmd_desc.out_param_size = buf_size;
    cmd_desc.out_trans = TRANS_MAILBOX;
    cmd_desc.opcode = ADD_DEV_PREFIX(IF_CMD_QUERY_QPEE);
    cmd_desc.opcode_modifier = is_ee ? 1 : 0;
    cmd_desc.exec_time_micro = ADD_DEV_PREFIX(IF_CMD_ETIME_QUERY_QPEE);

    rc = cmd_invoke(cmd_if, &cmd_desc);
    if ( rc != XHH_CMD_STAT_OK ) {
      FREE(outprm);
      MT_RETURN(rc);
    }

    if ( qp_context ) {
	  // 'te' field has been removed.
      //QP_EX_FLD(qp_context, outprm, ADD_DEV_PRM_PREFIX(qp_ee_state_transitions_st), te);
      QP_EX_BOOL_FLD(qp_context, outprm, ADD_DEV_PRM_PREFIX(qp_ee_state_transitions_st), de);
      QP_EX_FLD(qp_context, outprm, ADD_DEV_PRM_PREFIX(qp_ee_state_transitions_st), pm_state);
      QP_EX_FLD(qp_context, outprm, ADD_DEV_PRM_PREFIX(qp_ee_state_transitions_st), st);

      QP_EX_FLD(qp_context, outprm, ADD_DEV_PRM_PREFIX(qp_ee_state_transitions_st), sched_queue);
      QP_EX_FLD(qp_context, outprm, ADD_DEV_PRM_PREFIX(qp_ee_state_transitions_st), msg_max);
      QP_EX_FLD(qp_context, outprm, ADD_DEV_PRM_PREFIX(qp_ee_state_transitions_st), mtu);
      QP_EX_FLD(qp_context, outprm, ADD_DEV_PRM_PREFIX(qp_ee_state_transitions_st), usr_page);
      QP_EX_FLD(qp_context, outprm, ADD_DEV_PRM_PREFIX(qp_ee_state_transitions_st), local_qpn_een);
      QP_EX_FLD(qp_context, outprm, ADD_DEV_PRM_PREFIX(qp_ee_state_transitions_st), remote_qpn_een);

      device_if_qp_state = (ADD_DEV_PREFIX_LC(if_qp_state_t)) 
                    MT_EXTRACT_ARRAY32(outprm,  MT_BIT_OFFSET(ADD_DEV_PRM_PREFIX(qp_ee_state_transitions_st), qpc_eec_data.state), 
                                      MT_BIT_SIZE(ADD_DEV_PRM_PREFIX(qp_ee_state_transitions_st), qpc_eec_data.state));
      if(!XHH_tavor_qpstate_2_vapi_qpstate(device_if_qp_state, &(qp_context->state))){
          CMDS_DBG("%s: TAVOR QP state (0x%x) is not valid\n", __func__, device_if_qp_state);
          FREE(outprm);
          return XHH_CMD_STAT_EFATAL;
      }
	  
	  /*T.D.(matan): change along with the rest of SQ_DRAINING improvements.*/
	  qp_context->sq_draining = (qp_context->state & XHH_CMD_QP_DRAINING_FLAG) ? TRUE:FALSE;
	  qp_context->state = (VAPI_qp_state_t) (qp_context->state & ~(XHH_CMD_QP_DRAINING_FLAG));

      QP_EX_FLD(qp_context, outprm, ADD_DEV_PRM_PREFIX(qp_ee_state_transitions_st), primary_address_path.pkey_index);
      QP_EX_FLD(qp_context, outprm, ADD_DEV_PRM_PREFIX(qp_ee_state_transitions_st), primary_address_path.port_number);
      QP_EX_FLD(qp_context, outprm, ADD_DEV_PRM_PREFIX(qp_ee_state_transitions_st), primary_address_path.rlid);
      QP_EX_FLD(qp_context, outprm, ADD_DEV_PRM_PREFIX(qp_ee_state_transitions_st), primary_address_path.my_lid_path_bits);

      QP_EX_BOOL_FLD(qp_context, outprm, ADD_DEV_PRM_PREFIX(qp_ee_state_transitions_st), primary_address_path.g);

      QP_EX_FLD(qp_context, outprm, ADD_DEV_PRM_PREFIX(qp_ee_state_transitions_st), primary_address_path.rnr_retry);
      QP_EX_FLD(qp_context, outprm, ADD_DEV_PRM_PREFIX(qp_ee_state_transitions_st), primary_address_path.hop_limit);
      QP_EX_FLD(qp_context, outprm, ADD_DEV_PRM_PREFIX(qp_ee_state_transitions_st), primary_address_path.max_stat_rate);
      QP_EX_FLD(qp_context, outprm, ADD_DEV_PRM_PREFIX(qp_ee_state_transitions_st), primary_address_path.mgid_index);
      QP_EX_FLD(qp_context, outprm, ADD_DEV_PRM_PREFIX(qp_ee_state_transitions_st), primary_address_path.ack_timeout);
      QP_EX_FLD(qp_context, outprm, ADD_DEV_PRM_PREFIX(qp_ee_state_transitions_st), primary_address_path.flow_label);
      QP_EX_FLD(qp_context, outprm, ADD_DEV_PRM_PREFIX(qp_ee_state_transitions_st), primary_address_path.tclass);
      QP_EX_FLD(qp_context, outprm, ADD_DEV_PRM_PREFIX(qp_ee_state_transitions_st), primary_address_path.sl);

      /* extract RGID.  Note that get the RGID from the command object as 4 double-words, each in CPU-endianness. */
      /* Need to take them one at a time, and convert each to big-endian before storing in the output RGID array */
      /* Note that need to memcpy each 4 bytes to a temporary u_int32_t variable, since there is no guarantee */
      /* that the RGID is 4-byte aligned (it is an array of unsigned chars) */
      temp_u32 = MOSAL_cpu_to_be32(MT_EXTRACT_ARRAY32(outprm, 
                     MT_BIT_OFFSET(ADD_DEV_PRM_PREFIX(qp_ee_state_transitions_st), qpc_eec_data.primary_address_path.rgid_127_96),
                     MT_BIT_SIZE(ADD_DEV_PRM_PREFIX(qp_ee_state_transitions_st), qpc_eec_data.primary_address_path.rgid_127_96)));
      memcpy(&(qp_context->primary_address_path.rgid[0]), &temp_u32, sizeof(u_int32_t));

      temp_u32 = MOSAL_cpu_to_be32(MT_EXTRACT_ARRAY32(outprm, 
                     MT_BIT_OFFSET(ADD_DEV_PRM_PREFIX(qp_ee_state_transitions_st), qpc_eec_data.primary_address_path.rgid_95_64),
                     MT_BIT_SIZE(ADD_DEV_PRM_PREFIX(qp_ee_state_transitions_st), qpc_eec_data.primary_address_path.rgid_95_64)));
      memcpy(&(qp_context->primary_address_path.rgid[4]), &temp_u32, sizeof(u_int32_t));

      temp_u32 = MOSAL_cpu_to_be32(MT_EXTRACT_ARRAY32(outprm, 
                     MT_BIT_OFFSET(ADD_DEV_PRM_PREFIX(qp_ee_state_transitions_st), qpc_eec_data.primary_address_path.rgid_63_32),
                     MT_BIT_SIZE(ADD_DEV_PRM_PREFIX(qp_ee_state_transitions_st), qpc_eec_data.primary_address_path.rgid_63_32)));
      memcpy(&(qp_context->primary_address_path.rgid[8]), &temp_u32, sizeof(u_int32_t));

      temp_u32 = MOSAL_cpu_to_be32(MT_EXTRACT_ARRAY32(outprm, 
                     MT_BIT_OFFSET(ADD_DEV_PRM_PREFIX(qp_ee_state_transitions_st), qpc_eec_data.primary_address_path.rgid_31_0),
                     MT_BIT_SIZE(ADD_DEV_PRM_PREFIX(qp_ee_state_transitions_st), qpc_eec_data.primary_address_path.rgid_31_0)));
      memcpy(&(qp_context->primary_address_path.rgid[12]), &temp_u32, sizeof(u_int32_t));

      QP_EX_FLD(qp_context, outprm, ADD_DEV_PRM_PREFIX(qp_ee_state_transitions_st), alternative_address_path.pkey_index);
      QP_EX_FLD(qp_context, outprm, ADD_DEV_PRM_PREFIX(qp_ee_state_transitions_st), alternative_address_path.port_number);
      QP_EX_FLD(qp_context, outprm, ADD_DEV_PRM_PREFIX(qp_ee_state_transitions_st), alternative_address_path.rlid);
      QP_EX_FLD(qp_context, outprm, ADD_DEV_PRM_PREFIX(qp_ee_state_transitions_st), alternative_address_path.my_lid_path_bits);
      
      QP_EX_BOOL_FLD(qp_context, outprm, ADD_DEV_PRM_PREFIX(qp_ee_state_transitions_st), alternative_address_path.g);
      
      QP_EX_FLD(qp_context, outprm, ADD_DEV_PRM_PREFIX(qp_ee_state_transitions_st), alternative_address_path.rnr_retry);

      QP_EX_FLD(qp_context, outprm, ADD_DEV_PRM_PREFIX(qp_ee_state_transitions_st), alternative_address_path.hop_limit);
      QP_EX_FLD(qp_context, outprm, ADD_DEV_PRM_PREFIX(qp_ee_state_transitions_st), alternative_address_path.max_stat_rate);
      QP_EX_FLD(qp_context, outprm, ADD_DEV_PRM_PREFIX(qp_ee_state_transitions_st), alternative_address_path.mgid_index);
      QP_EX_FLD(qp_context, outprm, ADD_DEV_PRM_PREFIX(qp_ee_state_transitions_st), alternative_address_path.ack_timeout);
      QP_EX_FLD(qp_context, outprm, ADD_DEV_PRM_PREFIX(qp_ee_state_transitions_st), alternative_address_path.flow_label);
      QP_EX_FLD(qp_context, outprm, ADD_DEV_PRM_PREFIX(qp_ee_state_transitions_st), alternative_address_path.tclass);
      QP_EX_FLD(qp_context, outprm, ADD_DEV_PRM_PREFIX(qp_ee_state_transitions_st), alternative_address_path.sl);

      /* extract RGID.  Note that get the RGID from the command object as 4 double-words, each in CPU-endianness. */
      /* Need to take them one at a time, and convert each to big-endian before storing in the output RGID array */
      /* Note that need to memcpy each 4 bytes to a temporary u_int32_t variable, since there is no guarantee */
      /* that the RGID is 4-byte aligned (it is an array of unsigned chars) */
      temp_u32 = MOSAL_cpu_to_be32(MT_EXTRACT_ARRAY32(outprm, 
                     MT_BIT_OFFSET(ADD_DEV_PRM_PREFIX(qp_ee_state_transitions_st), qpc_eec_data.alternative_address_path.rgid_127_96),
                     MT_BIT_SIZE(ADD_DEV_PRM_PREFIX(qp_ee_state_transitions_st), qpc_eec_data.alternative_address_path.rgid_127_96)));
      memcpy(&(qp_context->alternative_address_path.rgid[0]), &temp_u32, sizeof(u_int32_t));

      temp_u32 = MOSAL_cpu_to_be32(MT_EXTRACT_ARRAY32(outprm, 
                     MT_BIT_OFFSET(ADD_DEV_PRM_PREFIX(qp_ee_state_transitions_st), qpc_eec_data.alternative_address_path.rgid_95_64),
                     MT_BIT_SIZE(ADD_DEV_PRM_PREFIX(qp_ee_state_transitions_st), qpc_eec_data.alternative_address_path.rgid_95_64)));
      memcpy(&(qp_context->alternative_address_path.rgid[4]), &temp_u32, sizeof(u_int32_t));

      temp_u32 = MOSAL_cpu_to_be32(MT_EXTRACT_ARRAY32(outprm, 
                     MT_BIT_OFFSET(ADD_DEV_PRM_PREFIX(qp_ee_state_transitions_st), qpc_eec_data.alternative_address_path.rgid_63_32),
                     MT_BIT_SIZE(ADD_DEV_PRM_PREFIX(qp_ee_state_transitions_st), qpc_eec_data.alternative_address_path.rgid_63_32)));
      memcpy(&(qp_context->alternative_address_path.rgid[8]), &temp_u32, sizeof(u_int32_t));

      temp_u32 = MOSAL_cpu_to_be32(MT_EXTRACT_ARRAY32(outprm, 
                     MT_BIT_OFFSET(ADD_DEV_PRM_PREFIX(qp_ee_state_transitions_st), qpc_eec_data.alternative_address_path.rgid_31_0),
                     MT_BIT_SIZE(ADD_DEV_PRM_PREFIX(qp_ee_state_transitions_st), qpc_eec_data.alternative_address_path.rgid_31_0)));
      memcpy(&(qp_context->alternative_address_path.rgid[12]), &temp_u32, sizeof(u_int32_t));

      QP_EX_FLD(qp_context, outprm, ADD_DEV_PRM_PREFIX(qp_ee_state_transitions_st), rdd);
      QP_EX_FLD(qp_context, outprm, ADD_DEV_PRM_PREFIX(qp_ee_state_transitions_st), pd);
      QP_EX_FLD(qp_context, outprm, ADD_DEV_PRM_PREFIX(qp_ee_state_transitions_st), wqe_lkey);
      QP_EX_BOOL_FLD(qp_context, outprm, ADD_DEV_PRM_PREFIX(qp_ee_state_transitions_st), ssc);
      QP_EX_BOOL_FLD(qp_context, outprm, ADD_DEV_PRM_PREFIX(qp_ee_state_transitions_st), sic);
      QP_EX_BOOL_FLD(qp_context, outprm, ADD_DEV_PRM_PREFIX(qp_ee_state_transitions_st), sae);
      QP_EX_BOOL_FLD(qp_context, outprm, ADD_DEV_PRM_PREFIX(qp_ee_state_transitions_st), swe);
      QP_EX_BOOL_FLD(qp_context, outprm, ADD_DEV_PRM_PREFIX(qp_ee_state_transitions_st), sre);
      QP_EX_FLD(qp_context, outprm, ADD_DEV_PRM_PREFIX(qp_ee_state_transitions_st), retry_count);
      QP_EX_FLD(qp_context, outprm, ADD_DEV_PRM_PREFIX(qp_ee_state_transitions_st), sra_max);
      QP_EX_FLD(qp_context, outprm, ADD_DEV_PRM_PREFIX(qp_ee_state_transitions_st), flight_lim);
      QP_EX_FLD(qp_context, outprm, ADD_DEV_PRM_PREFIX(qp_ee_state_transitions_st), ack_req_freq);
      QP_EX_FLD(qp_context, outprm, ADD_DEV_PRM_PREFIX(qp_ee_state_transitions_st), next_send_psn);
      QP_EX_FLD(qp_context, outprm, ADD_DEV_PRM_PREFIX(qp_ee_state_transitions_st), cqn_snd);
      QP_EX_BOOL_FLD(qp_context, outprm, ADD_DEV_PRM_PREFIX(qp_ee_state_transitions_st), rsc);
      QP_EX_BOOL_FLD(qp_context, outprm, ADD_DEV_PRM_PREFIX(qp_ee_state_transitions_st), ric);
      QP_EX_BOOL_FLD(qp_context, outprm, ADD_DEV_PRM_PREFIX(qp_ee_state_transitions_st), rae);
      QP_EX_BOOL_FLD(qp_context, outprm, ADD_DEV_PRM_PREFIX(qp_ee_state_transitions_st), rwe);
      QP_EX_BOOL_FLD(qp_context, outprm, ADD_DEV_PRM_PREFIX(qp_ee_state_transitions_st), rre);
      QP_EX_FLD(qp_context, outprm, ADD_DEV_PRM_PREFIX(qp_ee_state_transitions_st), rra_max);
      QP_EX_FLD(qp_context, outprm, ADD_DEV_PRM_PREFIX(qp_ee_state_transitions_st), next_rcv_psn);
      QP_EX_FLD(qp_context, outprm, ADD_DEV_PRM_PREFIX(qp_ee_state_transitions_st), min_rnr_nak);
      QP_EX_FLD(qp_context, outprm, ADD_DEV_PRM_PREFIX(qp_ee_state_transitions_st), ra_buff_indx);
      QP_EX_FLD(qp_context, outprm, ADD_DEV_PRM_PREFIX(qp_ee_state_transitions_st), cqn_rcv);
      QP_EX_FLD(qp_context, outprm, ADD_DEV_PRM_PREFIX(qp_ee_state_transitions_st), q_key);
      QP_EX_FLD(qp_context, outprm, ADD_DEV_PRM_PREFIX(qp_ee_state_transitions_st), srqn);
      QP_EX_BOOL_FLD(qp_context, outprm, ADD_DEV_PRM_PREFIX(qp_ee_state_transitions_st), srq);
      XHH_CMD_DEV_SPECIFIC_QUERY_QPEE
      
      /* XHH_CMD_PRINT_QP_CONTEXT(qp_context); */
    }

    FREE(outprm);
    MT_RETURN(rc);
}

/*
 *  XHH_cmd_QUERY_QP 
 */ 
XHH_cmd_status_t XHH_cmd_QUERY_QP(XHH_cmd_t cmd_if, IB_wqpn_t qpn,
                                  XHH_qpee_context_t *qp_context)
{
    return (XHH_cmd_QUERY_QPEE(cmd_if,0,qpn,qp_context));
}
/*
 *  XHH_cmd_QUERY_EE 
 */ 
XHH_cmd_status_t XHH_cmd_QUERY_EE(XHH_cmd_t cmd_if, IB_eecn_t eecn,
                                  XHH_qpee_context_t *ee_context)
{
    return (XHH_cmd_QUERY_QPEE(cmd_if,1,eecn,ee_context));
}

/*
 *  XHH_cmd_CONF_SPECIAL_QP 
 */ 
XHH_cmd_status_t XHH_cmd_CONF_SPECIAL_QP(XHH_cmd_t cmd_if, VAPI_special_qp_t qp_type,
                                         IB_wqpn_t base_qpn)
{
  command_fields_t cmd_desc;
  u_int8_t         op_modifier;
  XHH_cmd_status_t rc;

  FUNC_IN;
  MTL_DEBUG4("%s: ENTERING \n", __func__);
  switch(qp_type) {
  case VAPI_SMI_QP:
      op_modifier = 0;
      break;
  case VAPI_GSI_QP:
      op_modifier = 1;
      break;
  case VAPI_RAW_IPV6_QP:
      op_modifier = 2;
      break;
  case VAPI_RAW_ETY_QP:
      op_modifier = 3;
      break;
  default:
      MT_RETURN (XHH_CMD_STAT_EBADARG);
  }
  cmd_desc.in_param = 0;
  cmd_desc.in_param_size = 0;
  cmd_desc.in_trans = TRANS_NA;
  cmd_desc.input_modifier = base_qpn;
  cmd_desc.out_param = 0;
  cmd_desc.out_param_size = 0;
  cmd_desc.out_trans = TRANS_NA;
  cmd_desc.opcode = ADD_DEV_PREFIX(IF_CMD_CONF_SPECIAL_QP);
  cmd_desc.opcode_modifier = op_modifier;
  cmd_desc.exec_time_micro = ADD_DEV_PREFIX(IF_CMD_ETIME_CONF_SPECIAL_QP);

  rc = cmd_invoke(cmd_if, &cmd_desc);
  MT_RETURN(rc);
}

/*
 *  XHH_cmd_MAD_IFC 
 */ 
XHH_cmd_status_t XHH_cmd_MAD_IFC(XHH_cmd_t cmd_if, 
                                 MT_bool mkey_validate, 
                                 IB_lid_t slid, /* SLID is ignored if mkey_validate is false */
                                 IB_port_t port,
                                 void *mad_in, 
                                 void *mad_out)
{
    command_fields_t cmd_desc;
    XHH_cmd_status_t rc;
    u_int32_t       i, *int32_inbuf, *int32_outbuf, *orig_inbuf = NULL, *int32_temp_inbuf = NULL;
    
    /* support NULL mad_out */
    static u_int32_t dummy_mad_out[256/sizeof(u_int32_t)]; /* STATIC ! (not on stack) */
    

    FUNC_IN;
    if (mad_in == NULL)  return XHH_CMD_STAT_EBADARG;
    int32_temp_inbuf = (u_int32_t*)MALLOC(256);
    if (int32_temp_inbuf == NULL) {
        MT_COV_MEM_FLTR_START
        MT_RETURN(XHH_CMD_STAT_EAGAIN);
        MT_COV_MEM_FLTR_END
    }
    int32_inbuf = (u_int32_t *) mad_in;
    int32_outbuf = (u_int32_t *) (mad_out == NULL ? dummy_mad_out : mad_out);
    orig_inbuf = int32_temp_inbuf;

    cmd_desc.in_param = (u_int8_t *) int32_temp_inbuf;
    cmd_desc.in_param_size = 256;
    cmd_desc.in_trans = TRANS_MAILBOX;
    cmd_desc.input_modifier = (port & 3);
    /* For Mkey validation the MAD's source LID is required (upper 16 bits of input mod.) */
    if (mkey_validate)  {
      cmd_desc.input_modifier |= ( ((u_int32_t)slid) << 16 );
    }
    cmd_desc.out_param = (u_int8_t*)int32_outbuf;
    cmd_desc.out_param_size = 256;
    cmd_desc.out_trans = TRANS_MAILBOX;
    cmd_desc.opcode = ADD_DEV_PREFIX(IF_CMD_MAD_IFC);
    cmd_desc.opcode_modifier = mkey_validate ? 0 : 1;
    cmd_desc.exec_time_micro = ADD_DEV_PREFIX(IF_CMD_ETIME_MAD_IFC);
    
    /* reverse endianness to CPU endian, since MAD frames are all BE*/
    for (i = 0; i < 256/(sizeof(u_int32_t)); i++) {
              *int32_temp_inbuf  = MOSAL_be32_to_cpu(*int32_inbuf);
              int32_inbuf++;
              int32_temp_inbuf++;
    }
    rc = cmd_invoke(cmd_if, &cmd_desc);
    /* reverse endianness to big endian, since information is gotten cpu-endian*/
    for (i = 0; i < 256/(sizeof(u_int32_t)); i++) {
              *int32_outbuf  = MOSAL_cpu_to_be32(*int32_outbuf);
              int32_outbuf++;
    }
    FREE(orig_inbuf);
    XHH_CMD_MAILBOX_PRINT(mad_out, 256, __func__);
    MT_RETURN(rc);
}


XHH_cmd_status_t XHH_cmd_SW2HW_SRQ(XHH_cmd_t cmd_if,
                                   u_int32_t srqn,         /* SRQ number/index */
                                   XHH_srq_context_t *srqc_p) /* SRQ context      */
{
  command_fields_t cmd_desc;
  u_int8_t *inprm = NULL;
  XHH_cmd_status_t rc;
  const u_int32_t buf_size = PSEUDO_MT_BYTE_SIZE(ADD_DEV_PRM_PREFIX(srq_context_st));

  FUNC_IN;
  if (srqc_p == NULL) {MT_RETURN(XHH_CMD_STAT_EBADARG); }

  inprm = (u_int8_t *)MALLOC(PSEUDO_MT_BYTE_SIZE(ADD_DEV_PRM_PREFIX(srq_context_st)));
  if (inprm == NULL) {
     MT_COV_MEM_FLTR_START
     MT_RETURN(XHH_CMD_STAT_EAGAIN);
     MT_COV_MEM_FLTR_END
  }
  memset(inprm, 0, buf_size);
  
  INS_FLD(srqc_p, inprm, ADD_DEV_PRM_PREFIX(srq_context_st), pd);
  INS_FLD(srqc_p, inprm, ADD_DEV_PRM_PREFIX(srq_context_st), l_key);
  INS_FLD(srqc_p, inprm, ADD_DEV_PRM_PREFIX(srq_context_st), wqe_addr_h);
  INS_FLD(srqc_p, inprm, ADD_DEV_PRM_PREFIX(srq_context_st), lwm);
  
  XHH_CMD_DEV_SPECIFIC_SW2HW_SRQ

  cmd_desc.in_param = inprm;
  cmd_desc.in_param_size = buf_size;
  cmd_desc.in_trans = TRANS_MAILBOX;
  cmd_desc.input_modifier = srqn;
  cmd_desc.out_param = 0;
  cmd_desc.out_param_size = 0;
  cmd_desc.out_trans = TRANS_NA;
  cmd_desc.opcode = ADD_DEV_PREFIX(IF_CMD_SW2HW_SRQ);
  cmd_desc.opcode_modifier = 0;
  cmd_desc.exec_time_micro = ADD_DEV_PREFIX(IF_CMD_ETIME_CLASS_C);

  /* my_cmd_mailbox_print((void *)inprm,buf_size,"XHH_cmd_SW2HW_SRQ"); */
  rc = cmd_invoke(cmd_if, &cmd_desc);
  FREE(inprm);
  MT_RETURN(rc);
}

XHH_cmd_status_t XHH_cmd_HW2SW_SRQ(XHH_cmd_t cmd_if,
                                   u_int32_t srqn,          /* SRQ number/index */
                                   XHH_srq_context_t *srqc_p) /* SRQ context      */
{
  command_fields_t cmd_desc;
  u_int8_t *outprm = NULL;
  XHH_cmd_status_t rc;
  const u_int32_t buf_size = PSEUDO_MT_BYTE_SIZE(ADD_DEV_PRM_PREFIX(srq_context_st));

  FUNC_IN;

  outprm = (u_int8_t *)MALLOC(PSEUDO_MT_BYTE_SIZE(ADD_DEV_PRM_PREFIX(srq_context_st)));
  if (outprm == NULL) {
     MT_COV_MEM_FLTR_START
     MT_RETURN(XHH_CMD_STAT_EAGAIN);
     MT_COV_MEM_FLTR_END
  }
  memset(outprm, 0, buf_size);

  cmd_desc.in_param = 0;
  cmd_desc.in_param_size = 0;
  cmd_desc.in_trans = TRANS_NA;
  cmd_desc.input_modifier = srqn;
  cmd_desc.out_param = srqc_p != NULL ? outprm : 0;
  cmd_desc.out_param_size = srqc_p != NULL ? buf_size : 0;
  cmd_desc.out_trans = TRANS_MAILBOX;
  cmd_desc.opcode = ADD_DEV_PREFIX(IF_CMD_HW2SW_SRQ);
  cmd_desc.opcode_modifier = srqc_p != NULL ? 1 : 0; /* No need for output if no *srqc_p */
  cmd_desc.exec_time_micro = ADD_DEV_PREFIX(IF_CMD_ETIME_CLASS_C);

  rc = cmd_invoke(cmd_if, &cmd_desc);
  
  if (srqc_p != NULL) {
    EX_FLD(srqc_p, outprm, ADD_DEV_PRM_PREFIX(srq_context_st), pd);
    EX_FLD(srqc_p, outprm, ADD_DEV_PRM_PREFIX(srq_context_st), l_key);
    EX_FLD(srqc_p, outprm, ADD_DEV_PRM_PREFIX(srq_context_st), wqe_addr_h);
    EX_FLD(srqc_p, outprm, ADD_DEV_PRM_PREFIX(srq_context_st), wqe_cnt);
    EX_FLD(srqc_p, outprm, ADD_DEV_PRM_PREFIX(srq_context_st), state);
    EX_FLD(srqc_p, outprm, ADD_DEV_PRM_PREFIX(srq_context_st), lwm);
	XHH_CMD_DEV_SPECIFIC_HW2SW_SRQ

    /* my_cmd_mailbox_print((void *)outprm,buf_size,"XHH_cmd_HW2SW_SRQ"); */
  }
  FREE(outprm);
  MT_RETURN(rc);
}

XHH_cmd_status_t XHH_cmd_QUERY_SRQ(XHH_cmd_t cmd_if,
                                   u_int32_t srqn,          /* SRQ number/index */
                                   XHH_srq_context_t *srqc_p) /* SRQ context      */
{
  command_fields_t cmd_desc;
  u_int8_t *outprm = NULL;
  XHH_cmd_status_t rc;
  const u_int32_t buf_size = PSEUDO_MT_BYTE_SIZE(ADD_DEV_PRM_PREFIX(srq_context_st));

  FUNC_IN;
  if (srqc_p == NULL) {MT_RETURN(XHH_CMD_STAT_EBADARG); }

  outprm = (u_int8_t *)MALLOC(PSEUDO_MT_BYTE_SIZE(ADD_DEV_PRM_PREFIX(srq_context_st)));
  if (outprm == NULL) {
     MT_COV_MEM_FLTR_START
     MT_RETURN(XHH_CMD_STAT_EAGAIN);
     MT_COV_MEM_FLTR_END
  }
  memset(outprm, 0, buf_size);

  cmd_desc.in_param = 0;
  cmd_desc.in_param_size = 0;
  cmd_desc.in_trans = TRANS_NA;
  cmd_desc.input_modifier = srqn;
  cmd_desc.out_param = outprm;
  cmd_desc.out_param_size = buf_size;
  cmd_desc.out_trans = TRANS_MAILBOX;
  cmd_desc.opcode = ADD_DEV_PREFIX(IF_CMD_QUERY_SRQ);
  cmd_desc.opcode_modifier = 0; 
  cmd_desc.exec_time_micro = ADD_DEV_PREFIX(IF_CMD_ETIME_CLASS_C);

  rc = cmd_invoke(cmd_if, &cmd_desc);
  
  if (srqc_p != NULL){
    EX_FLD(srqc_p, outprm, ADD_DEV_PRM_PREFIX(srq_context_st), pd);
    EX_FLD(srqc_p, outprm, ADD_DEV_PRM_PREFIX(srq_context_st), l_key);
    EX_FLD(srqc_p, outprm, ADD_DEV_PRM_PREFIX(srq_context_st), wqe_addr_h);
    EX_FLD(srqc_p, outprm, ADD_DEV_PRM_PREFIX(srq_context_st), wqe_cnt);
    EX_FLD(srqc_p, outprm, ADD_DEV_PRM_PREFIX(srq_context_st), state);
    EX_FLD(srqc_p, outprm, ADD_DEV_PRM_PREFIX(srq_context_st), lwm);
    
	XHH_CMD_DEV_SPECIFIC_QUERY_SRQ
  }
  FREE(outprm);
  MT_RETURN(rc);
}

XHH_cmd_status_t XHH_cmd_ARM_SRQ(XHH_cmd_t cmd_if,
                                   u_int32_t srqn,    /* SRQ number/index */
				   u_int16_t lwm)      /* desired Limit Water Mark */   
{
  command_fields_t cmd_desc;
  XHH_cmd_status_t rc;
  u_int32_t        lwm_transfer[2];

  FUNC_IN;

  MTL_DEBUG4("XHH_cmd_ARM_SRQ: srqn = 0x%x, lwm = 0x%x\n", srqn, lwm);

  /* The command interface expects the lwm to be in the least significant dword. */

  lwm_transfer[0]= (u_int32_t)(0);               /* MS-DWORD */
  lwm_transfer[1]= lwm;                          /* LS-DWORD */  

  MTL_DEBUG4("XHH_cmd_ARM_SRQ: lwm_transfer [0] = 0x%x, lwm_transfer [1] = 0x%x\n",
             lwm_transfer[0], lwm_transfer[1]);

  cmd_desc.in_param = (u_int8_t *)&(lwm_transfer[0]);
  cmd_desc.in_param_size = sizeof(lwm_transfer);
  cmd_desc.in_trans = TRANS_IMMEDIATE;
  cmd_desc.input_modifier = srqn;
  cmd_desc.out_param = 0;
  cmd_desc.out_param_size = 0;
  cmd_desc.out_trans = TRANS_NA;
  cmd_desc.opcode = ADD_DEV_PREFIX(IF_CMD_ARM_SRQ);
  cmd_desc.opcode_modifier = 0;
  cmd_desc.exec_time_micro = ADD_DEV_PREFIX(IF_CMD_ETIME_CLASS_B);

  rc = cmd_invoke(cmd_if, &cmd_desc);
  MT_RETURN(rc);
}


/*
 *  XHH_cmd_READ_MGM 
 */ 
XHH_cmd_status_t XHH_cmd_READ_MGM(XHH_cmd_t cmd_if, u_int32_t mcg_index,
                                  MT_size_t  max_qp_per_mcg, XHH_mcg_entry_t *mcg_entry)
{
    // need to add qps_per_mcg_entry field
    command_fields_t cmd_desc;
    u_int8_t *outprm;
    XHH_cmd_status_t rc;
    IB_wqpn_t  /**qp_buf,*/ *qp_iterator;
    u_int32_t buf_size, i, num_active_qps_found, temp_u32;
    u_int8_t    valid;

    FUNC_IN;

    /* the default mcg_entry structure contains space for 8 qps per mcg */
    /* If HCA is configured for more than 8 qps per group, space for the extra qp entries */
    /* must be allocated as well */
    buf_size = (u_int32_t)(PSEUDO_MT_BYTE_SIZE(ADD_DEV_PRM_PREFIX(mgm_entry_st)) + 
        ((max_qp_per_mcg - 8)*PSEUDO_MT_BYTE_SIZE(ADD_DEV_PRM_PREFIX(mgmqp_st))));

    outprm = (u_int8_t*)MALLOC(buf_size);
    if ( !outprm ) {
      MT_COV_MEM_FLTR_START
      MT_RETURN(XHH_CMD_STAT_EAGAIN);
      MT_COV_MEM_FLTR_END
    }

    memset(outprm, 0, buf_size);

    cmd_desc.in_param = 0;
    cmd_desc.in_param_size = 0;
    cmd_desc.in_trans = TRANS_NA;
    cmd_desc.input_modifier = mcg_index;
    cmd_desc.out_param = outprm;
    cmd_desc.out_param_size = buf_size;
    cmd_desc.out_trans = TRANS_MAILBOX;
    cmd_desc.opcode = ADD_DEV_PREFIX(IF_CMD_READ_MGM);
    cmd_desc.opcode_modifier = 0;
    cmd_desc.exec_time_micro = ADD_DEV_PREFIX(IF_CMD_ETIME_READ_MGM);

    rc = cmd_invoke(cmd_if, &cmd_desc);
    if ( rc != XHH_CMD_STAT_OK ) {
       goto invoke_err;
    }
    
    if ( mcg_entry ) {
      /* allocate memory for the multicast QPs IB_wqpn_t */
      //qp_buf = (IB_wqpn_t  *)MALLOC(sizeof(IB_wqpn_t) * max_qp_per_mcg); 
      //if ( !qp_buf ) {
       // rc = XHH_CMD_STAT_EAGAIN;
        //goto malloc_err;
      //}
      //memset(qp_buf, 0, sizeof(IB_wqpn_t) * max_qp_per_mcg);

      /* get fixed portion of reply */
      EX_FLD(mcg_entry, outprm, ADD_DEV_PRM_PREFIX(mgm_entry_st), next_gid_index);

      /* extract MGID.  Note that get the MGID from the command object as 4 double-words, each in CPU-endianness. */
      /* Need to take them one at a time, and convert each to big-endian before storing in the output MGID array */
      /* Note that need to memcpy each 4 bytes to a temporary u_int32_t variable, since there is no guarantee */
      /* that the MGID is 4-byte aligned (it is an array of unsigned chars) */
      temp_u32 = MOSAL_cpu_to_be32(MT_EXTRACT_ARRAY32(outprm, 
                     MT_BIT_OFFSET(ADD_DEV_PRM_PREFIX(mgm_entry_st), mgid_128_96),
                     MT_BIT_SIZE(ADD_DEV_PRM_PREFIX(mgm_entry_st), mgid_128_96)));
      memcpy(&(mcg_entry->mgid[0]), &temp_u32, sizeof(u_int32_t));

      temp_u32 = MOSAL_cpu_to_be32(MT_EXTRACT_ARRAY32(outprm, 
                     MT_BIT_OFFSET(ADD_DEV_PRM_PREFIX(mgm_entry_st), mgid_95_64),
                     MT_BIT_SIZE(ADD_DEV_PRM_PREFIX(mgm_entry_st), mgid_95_64)));
      memcpy(&(mcg_entry->mgid[4]), &temp_u32, sizeof(u_int32_t));

      temp_u32 = MOSAL_cpu_to_be32(MT_EXTRACT_ARRAY32(outprm, 
                     MT_BIT_OFFSET(ADD_DEV_PRM_PREFIX(mgm_entry_st), mgid_63_32),
                     MT_BIT_SIZE(ADD_DEV_PRM_PREFIX(mgm_entry_st), mgid_63_32)));
      memcpy(&(mcg_entry->mgid[8]), &temp_u32, sizeof(u_int32_t));

      temp_u32 = MOSAL_cpu_to_be32(MT_EXTRACT_ARRAY32(outprm, 
                     MT_BIT_OFFSET(ADD_DEV_PRM_PREFIX(mgm_entry_st), mgid_31_0),
                     MT_BIT_SIZE(ADD_DEV_PRM_PREFIX(mgm_entry_st), mgid_31_0)));
      memcpy(&(mcg_entry->mgid[12]), &temp_u32, sizeof(u_int32_t));

      /* Now, extract the QP entries in the group */
      for (i = 0, num_active_qps_found = 0, qp_iterator = mcg_entry->qps/*qp_buf*/; i < max_qp_per_mcg; i++ ) {
             /* extract VALID bit for each QP.  If valid is set, extract the QP number and insert in */
             /* the QP array returned */
             valid =  (u_int8_t)(MT_EXTRACT_ARRAY32(outprm, 
                         (MT_BIT_OFFSET(ADD_DEV_PRM_PREFIX(mgm_entry_st), mgmqp_0.qi) + i*(MT_BIT_SIZE(ADD_DEV_PRM_PREFIX(mgm_entry_st), mgmqp_0))),
                          MT_BIT_SIZE(ADD_DEV_PRM_PREFIX(mgm_entry_st), mgmqp_0.qi) ));
             if (valid) {
                  /* NULL protection .. */
                  if (mcg_entry->qps) {
                      *((u_int32_t *) qp_iterator) = MT_EXTRACT_ARRAY32(outprm, 
                           (MT_BIT_OFFSET(ADD_DEV_PRM_PREFIX(mgm_entry_st), mgmqp_0.qpn_i) + i*(MT_BIT_SIZE(ADD_DEV_PRM_PREFIX(mgm_entry_st), mgmqp_0))),
                            MT_BIT_SIZE(ADD_DEV_PRM_PREFIX(mgm_entry_st), mgmqp_0.qpn_i) );
                      qp_iterator++;
                  }
                  num_active_qps_found++;
             }
      }
      mcg_entry->valid_qps = num_active_qps_found;
      /* If valid QPs found, return the allocated QP number buffer and number found.  Otherwise */
      /* return 0 QPs and delete the buffer */
      //if (num_active_qps_found) {
        //  mcg_entry->qps = qp_buf;
      //} else {
       //   mcg_entry->qps = NULL;
      //    FREE(qp_buf);
      //}
    }

invoke_err:
    FREE(outprm);
    MT_RETURN(rc);
}

/*
 *  XHH_cmd_WRITE_MGM 
 */ 
XHH_cmd_status_t XHH_cmd_WRITE_MGM(XHH_cmd_t cmd_if, u_int32_t mcg_index,
                                   MT_size_t  max_qp_per_mcg, XHH_mcg_entry_t *mcg_entry)
{
    command_fields_t cmd_desc;
    u_int8_t *inprm = NULL;
    XHH_cmd_status_t rc;
    u_int32_t buf_size = PSEUDO_MT_BYTE_SIZE(ADD_DEV_PRM_PREFIX(mgm_entry_st));
    IB_wqpn_t *qp_iterator;
    u_int32_t   i, temp_u32;

    FUNC_IN;
    
    CMDS_DBG("XHH_cmd_WRITE_MGM: index=%u, max_qp_per_mcg = "SIZE_T_FMT"\n", mcg_index, max_qp_per_mcg);
    if (mcg_entry == NULL) {MT_RETURN(XHH_CMD_STAT_EBADARG); }
    XHH_CMD_PRINT_MGM_ENTRY(mcg_entry);

    inprm = (u_int8_t *)MALLOC(PSEUDO_MT_BYTE_SIZE(ADD_DEV_PRM_PREFIX(mgm_entry_st)));
    if (inprm == NULL) {
       MT_COV_MEM_FLTR_START
       MT_RETURN(XHH_CMD_STAT_EAGAIN);
       MT_COV_MEM_FLTR_END
    }
    memset(inprm, 0, buf_size);

    cmd_desc.in_param = inprm;
    cmd_desc.in_param_size = buf_size;
    cmd_desc.in_trans = TRANS_MAILBOX;
    cmd_desc.input_modifier = mcg_index;
    cmd_desc.out_param = 0;
    cmd_desc.out_param_size = 0;
    cmd_desc.out_trans = TRANS_NA;
    cmd_desc.opcode = ADD_DEV_PREFIX(IF_CMD_WRITE_MGM);
    cmd_desc.opcode_modifier = 0;
    cmd_desc.exec_time_micro = ADD_DEV_PREFIX(IF_CMD_ETIME_WRITE_MGM);

    /* get fixed portion of reply */
    INS_FLD(mcg_entry, inprm, ADD_DEV_PRM_PREFIX(mgm_entry_st), next_gid_index);

    /* insert MGID.  Note that get the MGID from the command object as 4 double-words, each in CPU-endianness. */
    /* Need to take them one at a time, and convert each to big-endian before storing in the output MGID array */
    memcpy(&temp_u32, &(mcg_entry->mgid[0]), sizeof(u_int32_t));
    MT_INSERT_ARRAY32(inprm, MOSAL_be32_to_cpu(temp_u32), 
                   MT_BIT_OFFSET(ADD_DEV_PRM_PREFIX(mgm_entry_st), mgid_128_96),
                   MT_BIT_SIZE(ADD_DEV_PRM_PREFIX(mgm_entry_st), mgid_128_96));
    
    memcpy(&temp_u32, &(mcg_entry->mgid[4]), sizeof(u_int32_t));
    MT_INSERT_ARRAY32(inprm, MOSAL_be32_to_cpu(temp_u32), 
                   MT_BIT_OFFSET(ADD_DEV_PRM_PREFIX(mgm_entry_st), mgid_95_64),
                   MT_BIT_SIZE(ADD_DEV_PRM_PREFIX(mgm_entry_st), mgid_95_64));
    
    memcpy(&temp_u32, &(mcg_entry->mgid[8]), sizeof(u_int32_t));
    MT_INSERT_ARRAY32(inprm, MOSAL_be32_to_cpu(temp_u32), 
                   MT_BIT_OFFSET(ADD_DEV_PRM_PREFIX(mgm_entry_st), mgid_63_32),
                   MT_BIT_SIZE(ADD_DEV_PRM_PREFIX(mgm_entry_st), mgid_63_32));

    memcpy(&temp_u32, &(mcg_entry->mgid[12]), sizeof(u_int32_t));
    MT_INSERT_ARRAY32(inprm, MOSAL_be32_to_cpu(temp_u32), 
                   MT_BIT_OFFSET(ADD_DEV_PRM_PREFIX(mgm_entry_st), mgid_31_0),
                   MT_BIT_SIZE(ADD_DEV_PRM_PREFIX(mgm_entry_st), mgid_31_0));

    /* Now, insert the QP entries in the group */
    for (i = 0, qp_iterator = mcg_entry->qps; i < max_qp_per_mcg; i++, qp_iterator++ ) {
       /* Insert valid entries.  First, insert a VALID bit = 1 for each valid QP number, then insert */
       /* the QP number itself.  If there are no more valid entries, insert only a VALID bit = 0 for each */
       /* invalid entry, up to the maximum allowed QPs */
       if (i < mcg_entry->valid_qps) {
           MT_INSERT_ARRAY32(inprm, 1, 
                       (MT_BIT_OFFSET(ADD_DEV_PRM_PREFIX(mgm_entry_st), mgmqp_0.qi) + i*(MT_BIT_SIZE(ADD_DEV_PRM_PREFIX(mgm_entry_st), mgmqp_0))),
                        MT_BIT_SIZE(ADD_DEV_PRM_PREFIX(mgm_entry_st), mgmqp_0.qi) );
           MT_INSERT_ARRAY32(inprm, *((u_int32_t *) qp_iterator) , 
                       (MT_BIT_OFFSET(ADD_DEV_PRM_PREFIX(mgm_entry_st), mgmqp_0.qpn_i) + i*(MT_BIT_SIZE(ADD_DEV_PRM_PREFIX(mgm_entry_st), mgmqp_0))),
                        MT_BIT_SIZE(ADD_DEV_PRM_PREFIX(mgm_entry_st), mgmqp_0.qpn_i) );
       } else {
           MT_INSERT_ARRAY32(inprm, 0, 
                       (MT_BIT_OFFSET(ADD_DEV_PRM_PREFIX(mgm_entry_st), mgmqp_0.qi) + i*(MT_BIT_SIZE(ADD_DEV_PRM_PREFIX(mgm_entry_st), mgmqp_0))),
                        MT_BIT_SIZE(ADD_DEV_PRM_PREFIX(mgm_entry_st), mgmqp_0.qi) );
       }
    }
    XHH_CMD_MAILBOX_PRINT(inprm, buf_size, __func__); 
    rc = cmd_invoke(cmd_if, &cmd_desc);
    FREE(inprm);
    MT_RETURN(rc);
}

/*
 *  XHH_cmd_HASH 
 */ 
XHH_cmd_status_t XHH_cmd_MGID_HASH(XHH_cmd_t cmd_if, IB_gid_t mgid, XHH_mcg_hash_t *hash_val)
{
    command_fields_t cmd_desc;
    u_int8_t inprm[16];
    XHH_cmd_status_t rc;
    u_int32_t   out_param[2], temp_u32;
    
    FUNC_IN;
    memset(inprm, 0, 16);
    memset(out_param, 0, sizeof(out_param));

    cmd_desc.in_param = inprm;
    cmd_desc.in_param_size = sizeof(IB_gid_t);
    cmd_desc.in_trans = TRANS_MAILBOX;
    cmd_desc.input_modifier = 0;
    cmd_desc.out_param = (u_int8_t *)&(out_param[0]);
    cmd_desc.out_param_size = sizeof(out_param);
    cmd_desc.out_trans = TRANS_IMMEDIATE;
    cmd_desc.opcode = ADD_DEV_PREFIX(IF_CMD_MGID_HASH);
    cmd_desc.opcode_modifier = 0;
    cmd_desc.exec_time_micro = ADD_DEV_PREFIX(IF_CMD_ETIME_MGID_HASH);

    /* insert GID into mailbox, 1 double-word at a time.  Modify byte ordering within doubl-words */
    /* to cpu-endian, so that lower layers will properly process the GID */
    memcpy(&temp_u32, &(mgid[0]), sizeof(u_int32_t));
    MT_INSERT_ARRAY32(inprm, MOSAL_be32_to_cpu(temp_u32),0,32);
    memcpy(&temp_u32, &(mgid[4]), sizeof(u_int32_t));
    MT_INSERT_ARRAY32(inprm, MOSAL_be32_to_cpu(temp_u32),32,32);
    memcpy(&temp_u32, &(mgid[8]), sizeof(u_int32_t));
    MT_INSERT_ARRAY32(inprm, MOSAL_be32_to_cpu(temp_u32),64,32); 
    memcpy(&temp_u32, &(mgid[12]), sizeof(u_int32_t));
    MT_INSERT_ARRAY32(inprm, MOSAL_be32_to_cpu(temp_u32),96,32); 

    rc = cmd_invoke(cmd_if, &cmd_desc);

    /* Note that output result is directly inserted into hash_val parameter, in cpu-endian order */
    /* Only the first of the two output double-words needs to be copied. */
    CMDS_DBG( "XHH_cmd_MGID_HASH:  out_param[0] = 0x%x; out_param[1] = 0x%x\n", 
              out_param[0], out_param[1]);

    *hash_val = (XHH_mcg_hash_t)MT_EXTRACT32(out_param[1],0,16);
    MT_RETURN(rc);
}

#if defined(MT_SUSPEND_QP)
XHH_cmd_status_t XHH_cmd_SUSPEND_QP(XHH_cmd_t cmd_if,  u_int32_t qpn, MT_bool suspend_flag)
{
  command_fields_t cmd_desc;
  XHH_cmd_status_t rc;
  FUNC_IN;
  
  MTL_DEBUG2(MT_FLFMT("%s: qpn = 0x%x, suspend_flag = %s"), 
             __func__, qpn, ((suspend_flag==TRUE) ? "TRUE" : "FALSE" ));
  cmd_desc.in_param = 0;
  cmd_desc.in_param_size = 0;
  cmd_desc.in_trans = TRANS_NA;
  cmd_desc.input_modifier = qpn & 0xFFFFFF;
  //cmd_desc.input_modifier |= (is_ee ? 0x1000000 : 0);
  cmd_desc.out_param = 0;
  cmd_desc.out_param_size = 0;
  cmd_desc.out_trans = TRANS_NA;
  cmd_desc.opcode = (suspend_flag == TRUE) ? ADD_DEV_PREFIX(IF_CMD_SUSPEND_QPEE) : ADD_DEV_PREFIX(IF_CMD_UNSUSPEND_QPEE);
  cmd_desc.opcode_modifier = 1;
  cmd_desc.exec_time_micro = ADD_DEV_PREFIX(IF_CMD_ETIME_CLASS_C);
    
  rc = cmd_invoke(cmd_if, &cmd_desc);
  MTL_DEBUG2(MT_FLFMT("%s: qpn = 0x%x, command returned 0x%x"), __func__,qpn, rc);
  MT_RETURN(rc);
}
#endif
