/*                  - Mellanox Confidential and Proprietary -
 *
 *  Copyright (C) Feb 2004, Mellanox Technologies Ltd.  ALL RIGHTS RESERVED.
 *
 *  Except as specifically permitted herein, no portion of the information,
 *  including but not limited to object code and source code, may be reproduced,
 *  modified, distributed, republished or otherwise exploited in any form or by
 *  any means for any purpose without the prior written permission of Mellanox
 *  Technologies Ltd. Use of software subject to the terms and conditions
 *  detailed in the file "LICENSE.txt".
 *
 *  End of legal section ......................................................
 *
 *  rhh_hob.c - (Arbel/Sinai) hardware abstraction layer
 *
 *  Description: .
 *
 *  Version: $Id: hob.c,v 1.23 2004/06/22 11:03:45 jackm Exp $
 *
 *  Author: Jack Morgenstein
 */
 
#include <mtl_common.h>
#include <cr_types.h>
#include <vapi.h>
#include <vapi_common.h>
#include <hh.h>
#include <hob_priv.h>
#include <ul_hob.h>
#include <sm_mad.h>
#include <log2.h>
#include <default_profile.h>
#include <srqm.h>
#include <hob_imp.h>
#include <hob_pci_caps.h>
#include <hob_comm.h>

    /* the define below is to limit the number of QPs per MCG to 8 due to a bug
       must be fixed later.  Comment out this define when the bug is fixed */
#define ADJUST_DEVLIMS_MAX_QP_PER_MCG

/* #define WAIT_IF_FATAL(hob)    XHH_hob_wait_if_fatal(hob,&have_fatal); */
#define WAIT_IF_FATAL(hob)    XHH_hob_wait_if_fatal(hob,&have_fatal);
#define CMD_EQ_SIZE 250 /* to be smaller then 8K - so it's EQ can get into physical memory */

#if 4 <= MAX_DEBUG
#define XHH_PRINT_PROFILE(a)      XHH_print_profile(a)
#define XHH_PRINT_USR_PROFILE(a)  XHH_print_usr_profile(a)
#else
#define XHH_PRINT_PROFILE(a)      
#define XHH_PRINT_USR_PROFILE(a)
#endif

extern HH_ret_t XHH_hob_waitfor_hca_closed(HH_hca_hndl_t  hca_hndl);

static HH_ret_t  XHH_hob_query_struct_init(XHH_hob_t  hob, MT_bool have_usr_profile, VAPI_hca_cap_t *hca_cap_p);
HH_ret_t  XHH_hob_halt_hca(/*IN*/ XHH_hob_t hob);

/* static void XHH_dummy_async_event(HH_hca_hndl_t hca_hndl, HH_event_record_t  *event_rec_p, void* ptr);*/
static void XHH_dummy_async_event(HH_hca_hndl_t hca_hndl, HH_event_record_t  *event_rec_p, void* ptr)
{
  /* TBD : This should be an error */
  MTL_TRACE1("XHH_dummy_async_event: called for devices %s with event type 0x%x",
             hca_hndl->dev_desc, event_rec_p->etype);
  return;
}

static void XHH_dummy_comp_event(HH_hca_hndl_t hca_hndl, HH_cq_hndl_t cq_num, void* ptr)
{
  /* TBD : This should be an error */
  MTL_TRACE1("XHH_dummy_comp_event: called for device %s and  cq num 0x%x",
                          hca_hndl->dev_desc, cq_num);
  return;
}

int XHH_hob_fatal_err_thread(void  *arg);

/******************************************************************************
******************************************************************************
************************  INTERNAL FUNCTIONS *********************************
******************************************************************************
******************************************************************************/

/******************************************************************************
 *  Function:     XHH_hob_get_node_guid
 *
 *  Description:  Gets node GUID for this HCA.  
 *
 *  input:
 *                hob
 *                port_num - 1 or 2
 *  output: 
 *                node_guid - pointer to a GUID structure
 *  returns:
 *                HH_OK
 *                HH_EINVAL
 *                HH_EINVAL_HCA_HNDL
 *                HH_ERR
 *
 *  Comments:     Does MAD query to get the data in real time. This function is used
 *                in pre-calculation of VAPI_query_hca values (at open_hca time).
 *
 *****************************************************************************/
static HH_ret_t XHH_hob_get_node_guid(XHH_hob_t  hob,
                                     IB_guid_t   *node_guid)
{
  SM_MAD_NodeInfo_t  node_info;
  u_int8_t       *mad_frame_in;
  u_int8_t       *mad_frame_out;
  XHH_cmd_status_t  cmd_ret;

  MT_RETURN_IF_LOW_STACK(XHH_WATERMARK);
  MTL_DEBUG4("==> XHH_hob_get_node_guid\n");
  TEST_RETURN_FATAL(hob);

  mad_frame_in = TNMALLOC(u_int8_t, IB_MAD_SIZE);
  if ( !mad_frame_in ) {
    return HH_EAGAIN;
  }
  mad_frame_out = TNMALLOC(u_int8_t, IB_MAD_SIZE);
  if ( !mad_frame_out ) {
    FREE(mad_frame_in);
    return HH_EAGAIN;
  }
  memset(mad_frame_in, 0, sizeof(mad_frame_in));
  memset(mad_frame_out, 0, sizeof(mad_frame_out));

  /* get PKey table using MAD commands in XHH_cmd object */
  /* First, build the MAD header */
  MADHeaderBuild(IB_CLASS_SMP, 
                    0,
                    IB_METHOD_GET,
                    IB_SMP_ATTRIB_NODEINFO,
                    (u_int32_t)   0,
                    &(mad_frame_in[0]));

  /* issue the query */
  cmd_ret = XHH_cmd_MAD_IFC(hob->cmd, 0, 0, 1, &(mad_frame_in[0]), &(mad_frame_out[0]));
  if (cmd_ret != XHH_CMD_STAT_OK) {
      TEST_CMD_FATAL(cmd_ret);
      MTL_ERROR2( "XHH_hob_get_pkey_tbl: ERROR : Get Node Info command failed (%d) for port 1\n", cmd_ret);
      MTL_DEBUG4("<== XHH_hob_get_node_guid. ERROR\n");
      FREE(mad_frame_out);
      FREE(mad_frame_in);
      return HH_EINVAL;
  }
  MadBufPrint(&(mad_frame_out[0]));
  NodeInfoMADToSt(&node_info, &(mad_frame_out[0]));
  NodeInfoPrint(&node_info);
  
  memcpy((*node_guid), node_info.qwNodeGUID, sizeof(IB_guid_t));
  MTL_DEBUG4("<== XHH_hob_get_node_guid\n");
  FREE(mad_frame_out);
  FREE(mad_frame_in);
  return HH_OK;
}

#if 4 <= MAX_DEBUG
static void XHH_print_profile(XHH_profile_t *profile) 
{
    MTL_DEBUG1("Profile printout\n");

    MTL_DEBUG1("        icm_alloc_vec_size = "SIZE_T_DFMT"\n",profile->icm_alloc_vec_size);
    MTL_DEBUG1("        lam_size = "SIZE_T_XFMT" ("SIZE_T_DFMT")\n", 
               profile->lam_size,profile->lam_size );

    MTL_DEBUG1("        ddr_size_code = %d\n", profile->lam_size_code);
    
    MTL_DEBUG1("        num_external_mem_regions = "SIZE_T_XFMT" ("SIZE_T_DFMT")\n", 
               profile->num_external_mem_regions,profile->num_external_mem_regions );
    MTL_DEBUG1("        num_mem_windows = "SIZE_T_XFMT" ("SIZE_T_DFMT")\n", 
               profile->num_mem_windows,profile->num_mem_windows);

    MTL_DEBUG1("        log2_max_qps = "SIZE_T_DFMT"\n", profile->log2_max_qps);
    MTL_DEBUG1("        max_num_qps = "SIZE_T_XFMT" ("SIZE_T_DFMT")\n", 
               profile->max_num_qps,profile->max_num_qps);
    MTL_DEBUG1("        log2_max_cqs = "SIZE_T_DFMT"\n", profile->log2_max_cqs);
    MTL_DEBUG1("        max_num_cqs = "SIZE_T_XFMT" ("SIZE_T_DFMT")\n", 
               profile->max_num_cqs,profile->max_num_cqs);
    MTL_DEBUG1("        max_num_pds = "SIZE_T_XFMT" ("SIZE_T_DFMT")\n", 
               profile->max_num_pds,profile->max_num_pds);

    MTL_DEBUG1("        log2_max_mpt_entries = "SIZE_T_DFMT"\n", profile->log2_max_mpt_entries);
    MTL_DEBUG1("        log2_max_mtt_entries = "SIZE_T_DFMT"\n", profile->log2_max_mtt_entries);
    MTL_DEBUG1("        log2_mtt_segs_per_region = "SIZE_T_DFMT"\n", profile->log2_mtt_segs_per_region);
    MTL_DEBUG1("        log2_mtt_entries_per_seg = "SIZE_T_DFMT"\n", profile->log2_mtt_entries_per_seg);

    
    MTL_DEBUG1("        log2_max_uar = "SIZE_T_DFMT"\n", profile->log2_max_uar);
    MTL_DEBUG1("        log2_uar_pg_size = %d\n", profile->log2_uar_pg_size);
    
    MTL_DEBUG1("        log2_max_mcgs = "SIZE_T_DFMT"\n", profile->log2_max_mcgs);
    MTL_DEBUG1("        qps_per_mcg = "SIZE_T_DFMT"\n", profile->qps_per_mcg);
    MTL_DEBUG1("        log2_mcg_entry_size = "SIZE_T_DFMT"\n", profile->log2_mcg_entry_size);
    MTL_DEBUG1("        log2_mcg_hash_size = "SIZE_T_DFMT"\n",profile->log2_mcg_hash_size);
    
    MTL_DEBUG1("        log2_max_eecs = "SIZE_T_DFMT"\n",profile->log2_max_eecs);

    MTL_DEBUG1("        log2_max_eqs = %d\n", profile->log2_max_eqs);
	return;
}

static void XHH_print_usr_profile(EVAPI_hca_profile_t *profile) 
{
    MTL_DEBUG1("User Profile printout\n");

    MTL_DEBUG1("        num_qp = %d\n",profile->num_qp);
    MTL_DEBUG1("        num_cq = %d\n",profile->num_cq);
    MTL_DEBUG1("        num_pd = %d\n",profile->num_pd);
    MTL_DEBUG1("        num_mr = %d\n",profile->num_mr);
    MTL_DEBUG1("        num_mw = %d\n",profile->num_mw);
    MTL_DEBUG1("        max_qp_ous_rd_atom = %d\n",profile->max_qp_ous_rd_atom);
    MTL_DEBUG1("        max_mcg = %d\n",profile->max_mcg);
    MTL_DEBUG1("        qp_per_mcg = %d\n",profile->qp_per_mcg);
    MTL_DEBUG1("        require = %s\n",(profile->require == 0) ? "FALSE" : "TRUE");
	return;
}
#endif

static const char* XHH_get_icm_allocation_string (u_int32_t index)
{
    switch(index) {
        case 0: return "uarc_sz";
        case 1: return "mtt sz";
        case 2: return "mpt sz";
        case 3: return "qpc sz";
        case 4: return "eqpc sz";
        case 5: return "srqc sz";
        case 6: return "cqc sz";
        case 7: return "ecqc_sz";
        case 8: return "rdb sz";
        case 9: return "uar scratch sz";
        case 10: return "eqc sz";
        case 11: return "mcg sz";
        case 12: return "eec sz";
        case 13: return "eeec sz";
        default: return "UNKNOWN";

    }
}

/*
 * XHH_print_hw_props
 *
 */
static void XHH_print_hw_props(XHH_hw_props_t   *hw_props_p) 
{
    int i;

    MTL_DEBUG4("%s:        cr_base = " PHYS_ADDR_FMT "\n", __func__, hw_props_p->cr_base);
    MTL_DEBUG4("%s:        uar_base = " PHYS_ADDR_FMT "\n", __func__, hw_props_p->uar_base);
    MTL_DEBUG4("%s:        device_id = 0x%x\n", __func__, hw_props_p->device_id);
    MTL_DEBUG4("%s:        pci_vendor_id = 0x%x\n", __func__, hw_props_p->pci_vendor_id);
    MTL_DEBUG4("%s:        intr_pin = 0x%x\n", __func__, hw_props_p->interrupt_props.intr_pin);

    //MOSAL_IRQ_ID_t does not have to be integer
    MTL_DEBUG4("%s:        bus = %d\n", __func__, hw_props_p->bus);
    MTL_DEBUG4("%s:        dev_func = 0x%x\n", __func__, hw_props_p->dev_func);

    MTL_DEBUG4("%s:        number of allocated vectors = %d\n", __func__, hw_props_p->interrupt_props.num_allocated_vectors);
    for ( i=0; i<(int)hw_props_p->interrupt_props.num_allocated_vectors; ++i ) {
      MTL_DEBUG4("%s:        irq[%d] = 0x%x\n", __func__, i, hw_props_p->interrupt_props.irq[i]);
    }
}

/*
 * XHH_get_ddr_size_code
 *
 */
static XHH_lam_size_enum_t XHH_get_lam_size_code(MT_size_t    lam_size) 
{
    MTL_DEBUG4("XHH_get_lam_size_code: ddr size = "SIZE_T_FMT"\n", lam_size);
    if (lam_size < (1UL<<25)) {
        return XHH_LAM_SIZE_32M;
    } else if (lam_size < (1UL<<26)) {
        return XHH_LAM_SIZE_64M;
    } else if (lam_size < (1UL<<27)) {
        return XHH_LAM_SIZE_128M;
    } else if (lam_size < (1UL<<28)) {
        return XHH_LAM_SIZE_256M;
    } else if (lam_size < (1UL<<29)) {
        return XHH_LAM_SIZE_512M;
    } else if (lam_size < (1UL<<30)) {
        return XHH_LAM_SIZE_1024M;
    } else if (lam_size < (1UL<<31)) {
        return XHH_LAM_SIZE_2048M;
    } else if (lam_size < 0xFFFFFFFF) {
        return XHH_LAM_SIZE_4096M;
    } else {
        return XHH_LAM_SIZE_BIG;
    }
}

/******************************************************************************
 *  Function:     calculate_icm_alloc_vec
 *
 *  Description:  Calculates sizes for ICM area allocation from profile
 *
 *  input:
 *                hob
 *                profile        -- pointer to data structure containing computation input
 *  output: 
 *                alloc_size_vec -- pointer to vector of allocation sizes to compute
 *                
 *  returns:
 *                HH_OK
 *
 *
 *****************************************************************************/
static void calculate_icm_alloc_vec(/*IN*/ XHH_hob_t      hob,
                                    /*IN*/ XHH_profile_t  *profile,
                                    /*IN*/ MT_bool        skip_uar_ctx,
                                    /*OUT*/ MT_size_t     *log2_uar_ctx_size_p,
                                    /*OUT*/XHH_icm_allocation_vector_t *alloc_size_vec)
{
    MT_size_t  *icm_alloc_iterator = (MT_size_t *)alloc_size_vec;
    MT_size_t  log2_uar_ctx_size;
    int        i;

    log2_uar_ctx_size    = ceil_log2(
                                     ((u_int64_t)(1UL << profile->log2_max_uar)) *
                                     ((((u_int64_t)(1UL << profile->log2_max_qps) +
                                      (u_int64_t)(1UL << profile->log2_max_cqs)) * 16) +
                                      (hob->dev_lims.srq ?
                                      (((u_int64_t)(1UL << profile->log2_max_srqs)) * 8):0)));
    if (log2_uar_ctx_size_p != NULL) {
        *log2_uar_ctx_size_p = log2_uar_ctx_size;
    }

    alloc_size_vec->log2_uarc_size  = (skip_uar_ctx == TRUE) ? XHH_ICMM_INVALID_SZ : log2_uar_ctx_size;
    alloc_size_vec->log2_mtt_size = profile->log2_max_mtt_entries + XHH_ICM_LOG2_MTT_ENTRY_SIZE; 
    alloc_size_vec->log2_mpt_size = profile->log2_max_mpt_entries + XHH_ICM_LOG2_MPT_ENTRY_SIZE;   
    alloc_size_vec->log2_qpc_size = profile->log2_max_qps + XHH_ICM_LOG2_QPC_ENTRY_SIZE;
    alloc_size_vec->log2_eqpc_size = profile->log2_max_qps + XHH_ICM_LOG2_EQPC_ENTRY_SIZE; 
    alloc_size_vec->log2_srqc_size = hob->dev_lims.srq ? 
      profile->log2_max_srqs + XHH_ICM_LOG2_SRQC_ENTRY_SIZE : XHH_ICMM_INVALID_SZ;
    alloc_size_vec->log2_cqc_size = profile->log2_max_cqs + XHH_ICM_LOG2_CQC_ENTRY_SIZE;
    alloc_size_vec->log2_ecqc_size = profile->log2_max_cqs;
    alloc_size_vec->log2_rdb_size = profile->log2_max_qps + profile->log2_inflight_rdma_per_qp + 
                                                        XHH_ICM_LOG2_RDB_ENTRY_SIZE;
    alloc_size_vec->log2_uar_scratch_size = profile->log2_max_uar + XHH_ICM_LOG2_UAR_SCR_ENTRY_SIZE;  
    alloc_size_vec->log2_eqc_size = profile->log2_max_eqs + XHH_ICM_LOG2_EQC_ENTRY_SIZE; 
    if (XHH_DEV_LIM_MCG_ENABLED(hob)) {
         alloc_size_vec->log2_mcg_size = profile->log2_max_mcgs + profile->log2_mcg_entry_size;
    } else {
        alloc_size_vec->log2_mcg_size = XHH_ICMM_INVALID_SZ;
    }
//    alloc_size_vec->log2_eec_size = profile->log2_max_eecs + XHH_ICM_LOG2_EEC_ENTRY_SIZE; 
//    alloc_size_vec->log2_eeec_size = profile->log2_max_eecs + XHH_ICM_LOG2_EEEC_ENTRY_SIZE;    /* in-flight rdma */
    alloc_size_vec->log2_eec_size = XHH_ICMM_INVALID_SZ; 
    alloc_size_vec->log2_eeec_size = XHH_ICMM_INVALID_SZ;    /* in-flight rdma */
    
    /* correct the array so that the minimum size value (if should not be skipped)
     * is the arbel page size
     */
    for (i = 0; (MT_size_t)i < hob->profile.icm_alloc_vec_size; i++, icm_alloc_iterator++) {
        if ((*icm_alloc_iterator) == XHH_ICMM_INVALID_SZ)
            continue;  /* no allocation */
        
        if ((*icm_alloc_iterator) < XHH_ARBEL_LOG_PAGE_SIZE) {
            *icm_alloc_iterator = XHH_ARBEL_LOG_PAGE_SIZE;
        }
    }
    return;
}


/******************************************************************************
 *  Function:     XHH_check_profile
 *
 *  Description:  Validates profile values against Arbel max values, as obtained
 *                from GET_DEV_LIM query
 *
 *  Details:
 *
 *****************************************************************************/
static HH_ret_t XHH_check_profile(XHH_hob_t   hob)
{
    
  MT_RETURN_IF_LOW_STACK(XHH_WATERMARK);
    if (hob->profile.log2_uar_pg_size < hob->dev_lims.log_pg_sz) {
        MTL_ERROR1("XHH_calculate_default_profile:  log2 UAR page size(%u) is less than the minimum (%u)\n", 
                   hob->profile.log2_uar_pg_size, hob->dev_lims.log_pg_sz);
        return HH_EAGAIN;
    }
    hob->profile.log2_max_qps             = (hob->profile.log2_max_qps > hob->dev_lims.log_max_qp) ? 
                                                         hob->dev_lims.log_max_qp : hob->profile.log2_max_qps; 
    hob->profile.log2_max_mcgs            = (hob->profile.log2_max_mcgs > hob->dev_lims.log_max_mcg) ? 
                                                         hob->dev_lims.log_max_mcg : hob->profile.log2_max_mcgs;
    hob->profile.log2_max_eecs            = (hob->profile.log2_max_eecs > hob->dev_lims.log_max_ee) ? 
                                                         hob->dev_lims.log_max_ee : hob->profile.log2_max_eecs;
    hob->profile.log2_max_cqs             = (hob->profile.log2_max_cqs > hob->dev_lims.log_max_cq) ? 
                                                         hob->dev_lims.log_max_cq : hob->profile.log2_max_cqs;
    hob->profile.log2_max_uar             = (hob->profile.log2_max_uar > hob->dev_lims.uar_sz + 20UL - hob->profile.log2_uar_pg_size) ? 
                                                         hob->dev_lims.uar_sz + 20UL - hob->profile.log2_uar_pg_size : hob->profile.log2_max_uar;

    hob->profile.log2_max_eqs             = (hob->profile.log2_max_eqs > hob->dev_lims.log_max_eq) ? 
                                                         hob->dev_lims.log_max_eq : hob->profile.log2_max_eqs; 
    hob->profile.max_num_pds              = (hob->profile.max_num_pds > (1UL<<hob->dev_lims.log_max_pd)) ? 
                                                         (1UL<<hob->dev_lims.log_max_pd) : hob->profile.max_num_pds;  

    return HH_OK;
}


#define XHH_PROFILE_CALC_QP_AT_MINIMUM     (1)
#define XHH_PROFILE_CALC_CQ_AT_MINIMUM     (1 << 1)
#define XHH_PROFILE_CALC_PD_AT_MINIMUM     (1 << 2)
#define XHH_PROFILE_CALC_REG_AT_MINIMUM    (1 << 3)
#define XHH_PROFILE_CALC_WIN_AT_MINIMUM    (1 << 4)
#define XHH_PROFILE_CALC_ALL_AT_MINIMUM    (XHH_PROFILE_CALC_QP_AT_MINIMUM | XHH_PROFILE_CALC_CQ_AT_MINIMUM | \
                                            XHH_PROFILE_CALC_PD_AT_MINIMUM | XHH_PROFILE_CALC_REG_AT_MINIMUM | \
                                            XHH_PROFILE_CALC_WIN_AT_MINIMUM )
                                            
static int check_profile_sanity(XHH_hob_t   hob, EVAPI_hca_profile_t *user_profile, XHH_profile_input_t *rhh_profile)
{
    u_int64_t  tmp_calc;
        /* check for bad minimum values */
        if ((user_profile->num_qp == 0) || (user_profile->num_cq == 0) || (user_profile->num_pd == 0) ||
            (user_profile->num_mr == 0) || (user_profile->max_qp_ous_rd_atom == 0) ) {
            MTL_ERROR1(MT_FLFMT("profile: QPs or CQs or PDs or MRs or max_qp_ous_rd_atom equal to 0"));
            return 0;
        }
        if (user_profile->num_qp > (1U<<hob->dev_lims.log_max_qp)) {
            MTL_ERROR1(MT_FLFMT("profile: num QPs more than device limit(%d)"),
                       (1U<<hob->dev_lims.log_max_qp));
            return 0;
        } else if (user_profile->num_qp < 1) {
            MTL_ERROR1(MT_FLFMT("profile: num QPs must be at least 1"));
            return 0;
        }

        if (user_profile->num_cq > (1U<<hob->dev_lims.log_max_cq)) {
            MTL_ERROR1(MT_FLFMT("profile: num CQs more than device limit(%d)"),
                       (1U<<hob->dev_lims.log_max_cq));
            return 0;
        } else if (user_profile->num_cq < 1) {
            MTL_ERROR1(MT_FLFMT("profile: num CQs must be at least 1"));
            return 0;
        }

        if (user_profile->num_pd > (1U<<hob->dev_lims.log_max_pd)) {
            MTL_ERROR1(MT_FLFMT("profile: num PDs more than device limit(%d)"),
                       (1U<<hob->dev_lims.log_max_pd));
            return 0;
        } else if (user_profile->num_pd < 1) {
            MTL_ERROR1(MT_FLFMT("profile: num PDs must be at least 1"));
            return 0;
        }
        if (user_profile->num_mr < 1) {
            MTL_ERROR1(MT_FLFMT("profile: num MRs must be at least 1"));
            return 0;
        }
        if (user_profile->max_qp_ous_rd_atom > (1U<<hob->dev_lims.log_max_ra_res_qp)) {
            MTL_ERROR1(MT_FLFMT("profile: max_qp_ous_rd_atom more than device limit (%d)"),
                       (1U<<hob->dev_lims.log_max_ra_res_qp));
            return 0;
        }

	if (XHH_DEV_LIM_MCG_ENABLED(hob)) {
            if (ceil_log2((u_int64_t)user_profile->max_mcg) > hob->dev_lims.log_max_mcg) {
		MTL_ERROR1(MT_FLFMT("profile: num MCGs more than device limit(%d)"),
                       (1U<<hob->dev_lims.log_max_mcg));
		return 0;
	    }

	    if (ceil_log2((u_int64_t)user_profile->qp_per_mcg) > hob->dev_lims.log_max_qp_mcg) {
		MTL_ERROR1(MT_FLFMT("profile: QPs per multicast group greater than device limit (%d)"),
                       (1U<<hob->dev_lims.log_max_qp_mcg));
		return 0;
	    }
        
	    if (user_profile->qp_per_mcg > user_profile->num_qp) {
		MTL_ERROR1(MT_FLFMT("profile: QPs per multicast group (%d) greater than number of QPs (%d)"),
		       (int)user_profile->qp_per_mcg, (int) user_profile->num_qp);
		return 0;
	    }
	    if ((user_profile->max_mcg > 0) && (user_profile->qp_per_mcg < 8)) {
		MTL_ERROR1(MT_FLFMT("profile: if MCGs not zero, QP_PER_MCG must be >= 8"));
		return 0;
	    }
	}

	if (user_profile->num_cq > (user_profile->num_qp * 2)) {
            MTL_ERROR1(MT_FLFMT("profile: CQs more than twice QPs in hca profile"));
            return 0;
        }
        if (ceil_log2(user_profile->num_mr) > hob->dev_lims.log_max_mpts) {
            MTL_ERROR1("profile:  Requested MRs use more MTTs than HCA provides\n");
            return 0;
        }
        
        tmp_calc = (u_int64_t)((u_int64_t) user_profile->num_qp +
                               (u_int64_t) (unsigned long)XHH_NUM_RSVD_QP +
                               (u_int64_t) user_profile->num_cq + 
                               (u_int64_t) user_profile->num_mr);  
        if ( hob->dev_lims.log_max_mtts < ceil_log2( tmp_calc * (u_int64_t)(1U<<rhh_profile->log2_mtt_segs_per_region) )
                                                     + rhh_profile->log2_mtt_entries_per_seg) {
            MTL_ERROR1("profile:  Requested parameters (CQs + QPs + MRs) use more MTTs than HCA provides\n");
            return 0;
        }


        if (ceil_log2(user_profile->num_mr) > hob->dev_lims.log_max_mpts) {
            MTL_ERROR1("profile:  Requested MRs use more MPTs than HCA provides\n");
            return 0;
        }

        if (ceil_log2(user_profile->num_mw) > hob->dev_lims.log_max_mpts) {
            MTL_ERROR1("profile:  Requested MWs use more MPTs than HCA provides\n");
            return 0;
        }

        tmp_calc = (u_int64_t)((u_int64_t) user_profile->num_qp +
                               (u_int64_t) (unsigned long)XHH_NUM_RSVD_QP +
                               (u_int64_t) user_profile->num_cq + 
                               (u_int64_t) user_profile->num_mr + 
                               (u_int64_t) user_profile->num_mw);

        if ( hob->dev_lims.log_max_mpts < ceil_log2( tmp_calc)) {
            MTL_ERROR1("profile:  Requested parameters (CQs + QPs + MRs + MWs) use more MPTs than HCA provides\n");
            return 0;
        }
        return 1;
}
static void print_icm_address_vec(XHH_icm_base_addr_vector_t *icm_addr_vec_p)
{  XHH_icm_va_t  *icm_addr_iterator;
   int i;
   return;
   icm_addr_iterator =  (XHH_icm_va_t*)icm_addr_vec_p;
   for (i = 0; i < XHH_ICM_ALLOCATION_VEC_SIZE; i++, icm_addr_iterator++) {
       MTL_ERROR1("print_icm_address_vec:ICM: %s = "U64_FMT"\n", 
		  XHH_get_icm_allocation_string(i), *icm_addr_iterator); 
   }
   
}

static u_int64_t calculate_total_icm_size(XHH_icm_allocation_vector_t *icm_alloc_vec_p)
{
    MT_size_t           *icm_alloc_iterator;
    u_int64_t           temp_size, tot_icm_allocs;
    int i;

    /* Add up all the sizes in the icm allocation vector */
    tot_icm_allocs = 0;
    icm_alloc_iterator = (MT_size_t*) icm_alloc_vec_p;
    for (i = 0; i < XHH_ICM_ALLOCATION_VEC_SIZE; i++, icm_alloc_iterator++) {
        if ((*icm_alloc_iterator) == XHH_ICMM_INVALID_SZ) {
            temp_size = 0;  /* no allocation */
        } else  {
            temp_size =  (MAKE_ULONGLONG(1) << (*icm_alloc_iterator));
        }
        MTL_DEBUG1("calculate_total_icm_size:ICM: %s = "U64_FMT"\n", 
                   XHH_get_icm_allocation_string(i), temp_size); 
        tot_icm_allocs += temp_size;
    }
    MTL_DEBUG1("calculate_total_icm_size:ICM size = "U64_FMT"\n", 
               tot_icm_allocs);
    return tot_icm_allocs;
}
/******************************************************************************
 *  Function:     XHH_calculate_profile
 *
 *  Description:  Calculates and installs profile values
 *
 *  input
 *                hob
 *                profile_user_data - pointer to a user override for the data used
 *                                    in calculating the XHH profile.
 *
 *  Details:
 *
 *        All calculations are derived from the following data:
 *
 *        - max QPs = 64k per 128M DDR size (= 2^16)
 *        - max MPT entries per HCA:       1M (= 2^20)
 *        - avg Regions/windows per QP     8  (= 2^3)
 *        - avg Segments per Region        8  (= 2^3)
 *        - avg inflight RDMA per QP       4  (= 2^2)
 *
 *        Calculations are as follows:
 *        Max UARs = 1 per QP
 *        Max CQs  = 1 per QP
 *        Max PDs  = 1 per QP
 *        Max Regions/Wins per QP = 8, divided as follows:
 *             internal regions = 2 per QP  (1 for QP, one for CQ)
 *             external regions = 2 per QP
 *             windows          = 4 per QP
 *
 *        MPT:
 *           Arbel has a max of 1M regions/windows per HCA, and the MPT size must
 *           be a power of 2.  It is pointless to have fewer than 8 regions/windows per QP
 *           (as divided up above).  This means that the maximum number of QPs allowable,
 *           regardless of DDR size, is 128K.  Therefore, the presence of the "min" function
 *           in calculating the max number of MPT entries.  In effect, the 1M table size limitation
 *           means that a DDR larger than 256M will only add to the user-available DDR memory, and
 *           not to the driver's internal tables.
 *
 *        MTT:
 *           The default MTT size allocated has 2 segments per Region, with a segment size of 8 entries.
 *
 *        MCG:  for 128M: 4096 Groups per HCA, with 16 QPs per group (so that entry size is 64 bytes).
 *              for 256M: 8192 Groups per HCA, with 16 QPs per group (so that entry size is 64 bytes).
 *
 *        NOTES:
 *           If the profile_user_data is NULL, default values are used.  After the profile calculation,
 *           a check is done to see that all values are within HCA_DEV_LIM values, and that the result
 *           does not exceed the DDR memory size.  If any violations are encountered, the number of QPs
 *           is reduced by half, and the calculation is redone.
 *           
 *****************************************************************************/
static HH_ret_t XHH_calculate_profile(XHH_hob_t   hob, 
                                      EVAPI_hca_profile_t *profile_user_data,
                                      EVAPI_hca_profile_t  *sugg_profile_p)
{
    u_int8_t            log2_host_pg_size;
    EVAPI_hca_profile_t *local_user_profile_p = NULL;
    XHH_profile_input_t *profile_input_data_p = NULL;
    u_int64_t           tot_icm_allocs;
    XHH_icm_allocation_vector_t *icm_alloc_vec_p = NULL;
    MT_size_t           *icm_alloc_iterator, temp_size;
    u_int32_t           i;
    MT_bool             icm_calc_loop = TRUE, need_to_loop = FALSE;
    u_int32_t           calc_at_minimum = 0;
    MT_size_t           log_uar_ctx_size = 0;
    HH_ret_t            rtn = HH_OK;

    profile_input_data_p = (XHH_profile_input_t *) MALLOC(sizeof(XHH_profile_input_t));
    if (profile_input_data_p == NULL) {
	MTL_ERROR1("XHH_calculate_profile:  MALLOC of profile_input_data_p struct failed\n"); 
	rtn = HH_EAGAIN;
	goto done;
    }

    if (profile_user_data != NULL) {

	local_user_profile_p = (EVAPI_hca_profile_t *) MALLOC(sizeof(EVAPI_hca_profile_t));
	if (local_user_profile_p == NULL) {
	    MTL_ERROR1("XHH_calculate_profile:  MALLOC of local_user_profile struct failed\n"); 
	    rtn =  HH_EAGAIN;
	    goto done;
	}
       
        memcpy(local_user_profile_p,  profile_user_data, sizeof(EVAPI_hca_profile_t));

        /* default value substitutions */
        local_user_profile_p->num_qp = (local_user_profile_p->num_qp == 0xFFFFFFFF) ? 
                       XHH_PROF_MAX_QPS : local_user_profile_p->num_qp; 
        local_user_profile_p->num_cq = (local_user_profile_p->num_cq == 0xFFFFFFFF) ? 
                               XHH_PROF_MAX_CQS  : local_user_profile_p->num_cq; 
        local_user_profile_p->num_pd = (local_user_profile_p->num_pd == 0xFFFFFFFF) ? 
                               XHH_PROF_MAX_PDS : local_user_profile_p->num_pd; 
        local_user_profile_p->num_mr = (local_user_profile_p->num_mr == 0xFFFFFFFF) ? 
                               XHH_PROF_MAX_REGIONS : local_user_profile_p->num_mr; 
        local_user_profile_p->num_mw = (local_user_profile_p->num_mw == 0xFFFFFFFF) ? 
                               XHH_PROF_MAX_WINDOWS : local_user_profile_p->num_mw;
        
        local_user_profile_p->max_qp_ous_rd_atom = (local_user_profile_p->max_qp_ous_rd_atom == 0xFFFFFFFF) ? 
                               (1 << XHH_ICM_LOG2_INFLIGHT_RDMA_PER_QP):
                               local_user_profile_p->max_qp_ous_rd_atom; 

        local_user_profile_p->max_mcg = (local_user_profile_p->max_mcg == 0xFFFFFFFF) ? 
                               (1 << XHH_ICM_LOG2_MAX_MCG):local_user_profile_p->max_mcg; 
        
        local_user_profile_p->qp_per_mcg = (local_user_profile_p->qp_per_mcg == 0xFFFFFFFF) ? 
                               (1 << XHH_ICM_LOG2_MIN_QP_PER_MCG):local_user_profile_p->qp_per_mcg; 
        
        if (sugg_profile_p != NULL) {
            memcpy(sugg_profile_p, local_user_profile_p, sizeof(EVAPI_hca_profile_t));
        }

        profile_input_data_p->max_qps        = local_user_profile_p->num_qp ;
        profile_input_data_p->max_cqs        = local_user_profile_p->num_cq ;
        profile_input_data_p->max_pds        = local_user_profile_p->num_pd ;
        profile_input_data_p->max_regions    = local_user_profile_p->num_mr ;
        profile_input_data_p->max_windows    = local_user_profile_p->num_mw ;

        profile_input_data_p->min_qps        = (1U<<hob->dev_lims.log2_rsvd_qps) + XHH_NUM_RSVD_QP + 1;
        profile_input_data_p->min_cqs        = (1U<<hob->dev_lims.log2_rsvd_cqs) + 1;
        profile_input_data_p->min_pds        = hob->dev_lims.num_rsvd_pds + XHH_NUM_RSVD_PD + 1;
        profile_input_data_p->min_regions    = (1 << hob->dev_lims.log2_rsvd_mtts) + 1;
        profile_input_data_p->min_windows    = (1 << hob->dev_lims.log2_rsvd_mrws);
    
        profile_input_data_p->reduction_pct_qps     = 10;
        profile_input_data_p->reduction_pct_cqs     = 10;
        profile_input_data_p->reduction_pct_pds     = 10;
        profile_input_data_p->reduction_pct_regions = 10;
        profile_input_data_p->reduction_pct_windows = 10;
        
        profile_input_data_p->log2_inflight_rdma_per_qp = ceil_log2(local_user_profile_p->max_qp_ous_rd_atom);
        profile_input_data_p->log2_max_mcg              = ceil_log2(local_user_profile_p->max_mcg);
        profile_input_data_p->log2_min_qp_per_mcg       = ceil_log2(local_user_profile_p->qp_per_mcg);
        profile_input_data_p->suggested_qp_per_mcg      = local_user_profile_p->qp_per_mcg;
        
        profile_input_data_p->log2_max_eq               = XHH_ICM_LOG2_MAX_EQ;
        profile_input_data_p->log2_mcg_hash_proportion  = XHH_ICM_LOG2_MCG_HASH_PROPORTION;
        profile_input_data_p->log2_mtt_entries_per_seg  = XHH_ICM_LOG2_MTT_ENTRIES_PER_SEG;
        profile_input_data_p->log2_mtt_segs_per_region  = XHH_ICM_LOG2_MTT_SEGS_PER_REGION;
        /*sanity checks */
        if (check_profile_sanity(hob,local_user_profile_p, profile_input_data_p) == 0) {
            MTL_ERROR1(MT_FLFMT("XHH_calculate_profile: user profile not valid"));
            rtn = HH_EINVAL_PARAM;
	    goto done;
        }
    } else {
        /* use internally defined default values */
        profile_input_data_p->max_qps        = XHH_PROF_MAX_QPS;
        profile_input_data_p->max_cqs        = XHH_PROF_MAX_CQS;
        profile_input_data_p->max_pds        = XHH_PROF_MAX_PDS;
        profile_input_data_p->max_regions    = XHH_PROF_MAX_REGIONS;
        profile_input_data_p->max_windows    = XHH_PROF_MAX_WINDOWS;
        
        profile_input_data_p->min_qps        = XHH_PROF_MIN_QPS;
        profile_input_data_p->min_cqs        = XHH_PROF_MIN_CQS;
        profile_input_data_p->min_pds        = XHH_PROF_MIN_PDS;
        profile_input_data_p->min_regions    = XHH_PROF_MIN_REGIONS;
        profile_input_data_p->min_windows    = XHH_PROF_MIN_WINDOWS;

        profile_input_data_p->reduction_pct_qps     = XHH_PROF_PCNT_REDUCTION_QPS;
        profile_input_data_p->reduction_pct_cqs     = XHH_PROF_PCNT_REDUCTION_CQS;
        profile_input_data_p->reduction_pct_pds     = XHH_PROF_PCNT_REDUCTION_PDS;
        profile_input_data_p->reduction_pct_regions = XHH_PROF_PCNT_REDUCTION_REGIONS;
        profile_input_data_p->reduction_pct_windows = XHH_PROF_PCNT_REDUCTION_WINDOWS;
        
        profile_input_data_p->log2_inflight_rdma_per_qp = XHH_ICM_LOG2_INFLIGHT_RDMA_PER_QP;
        profile_input_data_p->log2_max_eq               = XHH_ICM_LOG2_MAX_EQ;
        profile_input_data_p->log2_max_mcg              = XHH_ICM_LOG2_MAX_MCG;
        profile_input_data_p->log2_min_qp_per_mcg       = XHH_ICM_LOG2_MIN_QP_PER_MCG;
        profile_input_data_p->suggested_qp_per_mcg      = (1U << XHH_ICM_LOG2_MIN_QP_PER_MCG);
        profile_input_data_p->log2_mcg_hash_proportion  = XHH_ICM_LOG2_MCG_HASH_PROPORTION;
        profile_input_data_p->log2_mtt_entries_per_seg  = XHH_ICM_LOG2_MTT_ENTRIES_PER_SEG;
        profile_input_data_p->log2_mtt_segs_per_region  = XHH_ICM_LOG2_MTT_SEGS_PER_REGION;
    }

    /* need inflight rdma per QP for rdb size in ICM, and for XHH_qpm_create */
    hob->profile.log2_inflight_rdma_per_qp = (u_int8_t) profile_input_data_p->log2_inflight_rdma_per_qp;

    /* manipulate MCG max if have LAM, and if not inputting a profile, or if inputting profile 
     * which allows reduction 
     */
    hob->profile.log2_max_mcgs            = profile_input_data_p->log2_max_mcg;
    if ((hob->have_lam == TRUE) && (hob->profile.lam_size_code < XHH_LAM_SIZE_128M)
        && ((profile_user_data == NULL) || 
	    ((local_user_profile_p != NULL) && (local_user_profile_p->require == FALSE)))) {
            hob->profile.log2_max_mcgs--;
    }

    log2_host_pg_size = MOSAL_SYS_PAGE_SHIFT;
    if (log2_host_pg_size < hob->dev_lims.log_pg_sz) {
        MTL_ERROR1("XHH_calculate_default_profile:  Host min page size(%lu) is too small\n", 
                   (unsigned long ) MOSAL_SYS_PAGE_SIZE);
        rtn = HH_EAGAIN;
	goto done;
    }

    hob->profile.icm_alloc_vec_size = XHH_ICM_ALLOCATION_VEC_SIZE;
    icm_alloc_vec_p = (XHH_icm_allocation_vector_t *) MALLOC(sizeof(XHH_icm_allocation_vector_t));
    if (icm_alloc_vec_p == NULL) {
	  MTL_ERROR1("XHH_calculate_profile:  MALLOC of XHH_icm_allocation_vector_t struct failed\n"); 
          rtn = HH_EAGAIN;
          goto done;
    }

    /* MCG calculations - not in recalculation loop, since the amount of memory involved is very small*/
    /* each MCG entry must be a power-of-2 size. To guarantee a power-of-2, we take a "ceiling" log of the */
    /* MCG entry size(in bytes), and then compute the actual number of QPs per mcg backwards from the mcg_size variable. */
    /* We also require (as a sanity check) that the log2_mcg_hash_size be greater than zero */
    if ((XHH_DEV_LIM_MCG_ENABLED(hob)) &&
        ((int)(hob->profile.log2_max_mcgs + profile_input_data_p->log2_mcg_hash_proportion) > 0)) {
	/* Adjust max QPs per mcg and mcg entry size here, since it affects allocation vector calculation */
	if (profile_input_data_p->suggested_qp_per_mcg > (u_int32_t) (1U<<hob->dev_lims.log_max_qp_mcg)) {
	    hob->profile.qps_per_mcg         =  (u_int32_t) (1U<<hob->dev_lims.log_max_qp_mcg);
	    hob->profile.log2_mcg_entry_size = ceil_log2( (hob->profile.qps_per_mcg * XHH_ICM_MCG_BYTES_PER_QP) + 
							     XHH_ICM_MCG_ENTRY_HEADER_SIZE
							);
	} else {
            hob->profile.log2_mcg_entry_size = ceil_log2((profile_input_data_p->suggested_qp_per_mcg*XHH_ICM_MCG_BYTES_PER_QP) 
                                                     + XHH_ICM_MCG_ENTRY_HEADER_SIZE);
            hob->profile.qps_per_mcg    = ( (1U<<(hob->profile.log2_mcg_entry_size)) - XHH_ICM_MCG_ENTRY_HEADER_SIZE) / 
                                                             XHH_ICM_MCG_BYTES_PER_QP;
	}
    
        /* the hash proportion is the log of the power-of-2 fraction of the total MCG entries used for the hash table. */
        /* Thus, for example, a proportion of (1/2) gets a log2_mcg_hash_proportion = -1 */
        hob->profile.log2_mcg_hash_size  = hob->profile.log2_max_mcgs + profile_input_data_p->log2_mcg_hash_proportion;
    } else {
        /*UD MCGs not available on this HCA*/
        hob->profile.log2_mcg_entry_size = 0;
        hob->profile.qps_per_mcg    = 0;
        hob->profile.log2_mcg_hash_size  = 0;
        hob->profile.log2_max_mcgs = 0;
    }
    
    hob->profile.log2_mtt_entries_per_seg = profile_input_data_p->log2_mtt_entries_per_seg;
    hob->profile.log2_mtt_segs_per_region = profile_input_data_p->log2_mtt_segs_per_region;
    
    hob->profile.log2_uar_pg_size         = log2_host_pg_size;
    hob->profile.log2_max_uar             = hob->dev_lims.uar_sz + 20 - hob->profile.log2_uar_pg_size;
    hob->profile.log2_max_eqs             = (u_int8_t)profile_input_data_p->log2_max_eq;      /* 64 EQs */
    hob->profile.max_num_pds              = profile_input_data_p->max_pds;

    hob->profile.max_num_qps              = profile_input_data_p->max_qps;

    hob->profile.log2_max_qps             = ceil_log2(profile_input_data_p->max_qps+ 
                                               (1U<<hob->dev_lims.log2_rsvd_qps) + XHH_NUM_RSVD_QP); 
    
    /* adjust max QPs downward (if using internal profile, or if user profile permits)
     * if the few reserved QPs cause max qps to go beyond a power-of-2.
     */

    if (hob->profile.log2_max_qps > ceil_log2(profile_input_data_p->max_qps)) {
        MTL_DEBUG1(MT_FLFMT("%s:  reserved qps cause profile qps to jump a power-of-2"),__func__);
        if ((profile_user_data==NULL) || 
	    ((local_user_profile_p != NULL) && (local_user_profile_p->require == FALSE))) {
            hob->profile.log2_max_qps--;
            hob->profile.max_num_qps = (1U<<hob->profile.log2_max_qps) - (1U<<hob->dev_lims.log2_rsvd_qps)
                                                                           - XHH_NUM_RSVD_QP;
            MTL_DEBUG1(MT_FLFMT("%s: Adjusting max qps to "SIZE_T_DFMT),__func__, hob->profile.max_num_qps);
        }
    }

    /* TBD: Expose max_srqs to profile given by user and use MOD_STAT_CFG */
    if (hob->dev_lims.srq) {
      hob->profile.log2_max_srqs            = hob->dev_lims.log_max_srqs;
      hob->profile.max_num_srqs             = 
        (1U << hob->dev_lims.log_max_srqs) - (1 << hob->dev_lims.log2_rsvd_srqs);
    } else {
      hob->profile.log2_max_srqs            = 0;
      hob->profile.max_num_srqs             = 0;
    }
    
    hob->profile.max_num_cqs              = profile_input_data_p->max_cqs;
    hob->profile.log2_max_cqs             = ceil_log2(profile_input_data_p->max_cqs + 
                                                      (1U<<hob->dev_lims.log2_rsvd_cqs)); 
    /* adjust max CQs downward (if using internal profile, or if user profile permits) 
     * if the few reserved CQs cause max cqs to go beyond a power-of-2.
     */

    if (hob->profile.log2_max_cqs > ceil_log2(profile_input_data_p->max_cqs)) {
        MTL_DEBUG1(MT_FLFMT("%s:  reserved cqs cause profile cqs to jump a power-of-2"),__func__);
        if ((profile_user_data == NULL) || 
	    ((local_user_profile_p != NULL) && (local_user_profile_p->require == FALSE))) {
            hob->profile.log2_max_cqs--;
            hob->profile.max_num_cqs = (1U<<hob->profile.log2_max_cqs) - (1U<<hob->dev_lims.log2_rsvd_cqs);
            MTL_DEBUG1(MT_FLFMT("%s: Adjusting max cqs to "SIZE_T_DFMT),__func__, hob->profile.max_num_cqs);
        }
    }
    hob->profile.num_external_mem_regions = profile_input_data_p->max_regions;  /* 2 per QP */
    hob->profile.num_mem_windows          = profile_input_data_p->max_windows;
    hob->profile.log2_max_eecs            = 0;

    while (icm_calc_loop) {
        MT_bool continue_calc_loop;

        continue_calc_loop = FALSE;
    
        MTL_DEBUG4("XHH_calculate_profile: max_qps = "SIZE_T_FMT", max_cqs = "SIZE_T_FMT 
                   ",\nmax_pds="SIZE_T_FMT",max_reg="SIZE_T_FMT", max_win="SIZE_T_FMT"\n",
                        hob->profile.max_num_qps, hob->profile.max_num_cqs, 
                        hob->profile.max_num_pds, hob->profile.num_external_mem_regions, 
                        hob->profile.num_mem_windows); 

        /* add all raw resources without Arbel-reserved quantities */
        temp_size = hob->profile.max_num_qps + XHH_NUM_RSVD_QP + 
                    hob->profile.max_num_cqs + hob->profile.num_external_mem_regions;
        
        hob->profile.log2_max_mtt_entries     = ceil_log2(
                                                   ( temp_size * (1U<<profile_input_data_p->log2_mtt_segs_per_region)
                                                      * (1U<<profile_input_data_p->log2_mtt_entries_per_seg))
                                                   + (1 << hob->dev_lims.log2_rsvd_mtts)
                                                );
    
        /* add all raw resources without Arbel-reserved quantities */
        temp_size = hob->profile.max_num_qps + XHH_NUM_RSVD_QP + hob->profile.max_num_cqs 
            + hob->profile.num_external_mem_regions + hob->profile.num_mem_windows;
        
        hob->profile.log2_max_mpt_entries       = ceil_log2(temp_size + (1 << hob->dev_lims.log2_rsvd_mrws));
    
    
        if (hob->profile.log2_max_mtt_entries > hob->dev_lims.log_max_mtts) {
            continue_calc_loop = TRUE;
            need_to_loop = TRUE;
        }
        
        if (!continue_calc_loop) {
            /* Now, compute the total ICM size not including the UAR context area, 
             * and verify that we have not over-allocated it.  If yes, reduce QPs by half, and 
             * recompute all above parameters starting with log2_max_regions 
             */
            calculate_icm_alloc_vec(hob, &(hob->profile),TRUE, &log_uar_ctx_size,icm_alloc_vec_p);
       
            /* Add up all the sizes in the icm allocation vector */
            tot_icm_allocs = 0;
            icm_alloc_iterator = (MT_size_t *)icm_alloc_vec_p;
            for (i = 0; i < hob->profile.icm_alloc_vec_size; i++, icm_alloc_iterator++) {
                if ((*icm_alloc_iterator) == XHH_ICMM_INVALID_SZ) {
                    temp_size = 0;  /* no allocation */
              //  } else if ((*icm_alloc_iterator) >= ceil_log2(avail_icm_memory)) {
              //      temp_size = avail_icm_memory;
                } else  {
                    temp_size =  (((MT_size_t) 1ul) << (*icm_alloc_iterator));
                }
                MTL_DEBUG4("XHH_calculate_profile:ICM: %s = "SIZE_T_XFMT"("SIZE_T_DFMT")\n", 
                           XHH_get_icm_allocation_string(i), temp_size, temp_size); 
                tot_icm_allocs += temp_size;
            }
            
            /* test against max phys memory available for ICM and max icm virt mem avail for uar ctx*/
            MTL_DEBUG4("XHH_calculate_profile:  total ICM allocs = %d MB (incl reserved areas)\n",(int)(tot_icm_allocs>>20)); 
            if ((log_uar_ctx_size > (MT_ARBEL_LOG2_MAX_ICM_SIZE - 1) ) || /* allow max half of ICM virt addr space for UAR */
                          ((profile_user_data == NULL) && (hob->profile.max_num_qps>(1U<<16)))){ 
                          /*do not want more than 64K QPs if using internal defaults*/
                continue_calc_loop = TRUE;
                need_to_loop = TRUE;
            }

        }
        if (continue_calc_loop) {
            u_int64_t  temp;
            u_int32_t  u32_temp, change_flag;
            /* Reduce flagged profile input params by factor of 10 percent */
            change_flag = 0;
            if ((calc_at_minimum & XHH_PROFILE_CALC_QP_AT_MINIMUM) == 0) {
                change_flag++;
                temp = (u_int64_t)(hob->profile.max_num_qps) * (100 - profile_input_data_p->reduction_pct_qps);
                /*check for overflow. If have overflow, use approximate percentages (divide by 1024) */
                if (temp & MAKE_ULONGLONG(0xFFFFFFFF00000000)) {
                    temp = (u_int64_t)(hob->profile.max_num_qps) * (1024 - (profile_input_data_p->reduction_pct_qps*10));
                    temp >>= 10;
                    u32_temp = (u_int32_t)(temp & 0xFFFFFFFF);
                } else {
                    /* use more exact percentages -- but still not floating point */
                    u32_temp = (u_int32_t)temp;
                    u32_temp /= 100;
                }
                if (u32_temp <= (u_int32_t)profile_input_data_p->min_qps) {
                    calc_at_minimum |= XHH_PROFILE_CALC_QP_AT_MINIMUM;
                    u32_temp = profile_input_data_p->min_qps;
                }
                hob->profile.max_num_qps = u32_temp;
                hob->profile.log2_max_qps = ceil_log2(u32_temp + (1U<<hob->dev_lims.log2_rsvd_qps) 
                                                      + XHH_NUM_RSVD_QP);
            }
            
            
            if ((calc_at_minimum & XHH_PROFILE_CALC_CQ_AT_MINIMUM) == 0) {
                change_flag++;
                temp = (u_int64_t)(hob->profile.max_num_cqs) * (100 - profile_input_data_p->reduction_pct_cqs);
                if (temp & MAKE_ULONGLONG(0xFFFFFFFF00000000)) {
                    temp = (u_int64_t)(hob->profile.max_num_cqs) * (1024 - (profile_input_data_p->reduction_pct_cqs*10));
                    temp >>= 10;
                    u32_temp = (u_int32_t)(temp & 0xFFFFFFFF);
                } else {
                    /* use more exact percentages -- but still not floating point */
                    u32_temp = (u_int32_t)temp;
                    u32_temp /= 100;
                }
                if (u32_temp <= (u_int32_t)profile_input_data_p->min_cqs) {
                    calc_at_minimum |= XHH_PROFILE_CALC_CQ_AT_MINIMUM;
                    u32_temp = profile_input_data_p->min_cqs;
                }
                hob->profile.max_num_cqs = u32_temp;
                hob->profile.log2_max_cqs = ceil_log2(u32_temp + (1U<<hob->dev_lims.log2_rsvd_cqs));
            }

            if ((calc_at_minimum & XHH_PROFILE_CALC_PD_AT_MINIMUM) == 0) {
                change_flag++;
                temp = (u_int64_t)(hob->profile.max_num_pds) * (100 - profile_input_data_p->reduction_pct_pds);
                if (temp & MAKE_ULONGLONG(0xFFFFFFFF00000000)) {
                    temp = (u_int64_t)(hob->profile.max_num_pds) * (1024 - (profile_input_data_p->reduction_pct_pds*10));
                    temp >>= 10;
                    u32_temp = (u_int32_t)(temp & 0xFFFFFFFF);
                } else {
                    /* use more exact percentages -- but still not floating point */
                    u32_temp = (u_int32_t)temp;
                    u32_temp /= 100;
                }
                if (u32_temp <= (u_int32_t)profile_input_data_p->min_pds) {
                    calc_at_minimum |= XHH_PROFILE_CALC_PD_AT_MINIMUM;
                    u32_temp = profile_input_data_p->min_pds;
                }
                hob->profile.max_num_pds = u32_temp;
            }

            if ((calc_at_minimum & XHH_PROFILE_CALC_REG_AT_MINIMUM) == 0) {
                change_flag++;
                temp = (u_int64_t)(hob->profile.num_external_mem_regions) * (100 - profile_input_data_p->reduction_pct_regions);
                if (temp & MAKE_ULONGLONG(0xFFFFFFFF00000000)) {
                    temp = (u_int64_t)(hob->profile.num_external_mem_regions) * (1024 - (profile_input_data_p->reduction_pct_regions*10));
                    temp >>= 10;
                    u32_temp = (u_int32_t)(temp & 0xFFFFFFFF);
                } else {
                    /* use more exact percentages -- but still not floating point */
                    u32_temp = (u_int32_t)temp;
                    u32_temp /= 100;
                }
                if (u32_temp <= (u_int32_t)profile_input_data_p->min_regions) {
                    calc_at_minimum |= XHH_PROFILE_CALC_REG_AT_MINIMUM;
                    u32_temp = profile_input_data_p->min_regions;
                }
                hob->profile.num_external_mem_regions = u32_temp;
            }

            if ((calc_at_minimum & XHH_PROFILE_CALC_WIN_AT_MINIMUM) == 0) {
                change_flag++;
                temp = (u_int64_t)(hob->profile.num_mem_windows) * (100 - profile_input_data_p->reduction_pct_windows);
                if (temp & MAKE_ULONGLONG(0xFFFFFFFF00000000)) {
                    temp = (u_int64_t)(hob->profile.num_mem_windows) * (1024 - (profile_input_data_p->reduction_pct_windows*10));
                    temp >>= 10;
                    u32_temp = (u_int32_t)(temp & 0xFFFFFFFF);
                } else {
                    /* use more exact percentages -- but still not floating point */
                    u32_temp = (u_int32_t)temp;
                    u32_temp /= 100;
                }
                if (u32_temp <= (u_int32_t)profile_input_data_p->min_windows) {
                    calc_at_minimum |= XHH_PROFILE_CALC_WIN_AT_MINIMUM;
                    u32_temp = profile_input_data_p->min_windows;
                }
                hob->profile.num_mem_windows = u32_temp;
            }
            if (hob->profile.log2_inflight_rdma_per_qp > XHH_ICM_LOG2_INFLIGHT_RDMA_PER_QP) {
                 hob->profile.log2_inflight_rdma_per_qp--;
                 change_flag++;
            }

            /* check if we were able to perform any reductions */
            if (change_flag == 0) {
                MTL_ERROR1("XHH_calculate_default_profile:  avail memory too small for MIN profile\n");
                icm_alloc_iterator = (MT_size_t *)icm_alloc_vec_p;
                for (i = 0; i < hob->profile.icm_alloc_vec_size; i++, icm_alloc_iterator++) {
                    if ((*icm_alloc_iterator) == XHH_ICMM_INVALID_SZ) {
                        temp_size = 0;  /* no allocation */
                    } else  {
                        temp_size =  (((MT_size_t) 1ul) << (*icm_alloc_iterator));
                    }
                    MTL_ERROR1(MT_FLFMT("XHH_calculate_profile:ICM: %s = "SIZE_T_XFMT"("SIZE_T_DFMT")"), 
                               XHH_get_icm_allocation_string(i), temp_size, temp_size); 
                }
                rtn = HH_EAGAIN;
		goto done;
            }
        } else {
            icm_calc_loop = FALSE;
        }
    }
    
    XHH_check_profile(hob);  /* final adjustment to catch dev-lim overruns*/

    /* adjust mcg hash table after final adjustment of mcg size */
    if (XHH_DEV_LIM_MCG_ENABLED(hob)) {
        hob->profile.log2_mcg_hash_size  =  (hob->profile.log2_max_mcgs) + profile_input_data_p->log2_mcg_hash_proportion;
        if ((int)hob->profile.log2_mcg_hash_size < 0) {
            hob->profile.log2_mcg_hash_size = 0;
        }
    }

    XHH_PRINT_PROFILE(&(hob->profile));
    MTL_DEBUG4("Leaving XHH_calculate_profile\n");

    if (sugg_profile_p != NULL) { 
            sugg_profile_p->num_mw = (u_int32_t)hob->profile.num_mem_windows;
            sugg_profile_p->num_qp = (u_int32_t)hob->profile.max_num_qps;
            sugg_profile_p->num_cq = (u_int32_t)hob->profile.max_num_cqs;
            sugg_profile_p->num_pd = (u_int32_t)hob->profile.max_num_pds;
            sugg_profile_p->num_mr = (u_int32_t)hob->profile.num_external_mem_regions;
            sugg_profile_p->max_qp_ous_rd_atom = (1U<<hob->profile.log2_inflight_rdma_per_qp);
    }
    if ((profile_user_data != NULL) && (profile_user_data->require != 0) && (need_to_loop == TRUE)) {
        MTL_ERROR1("XHH_calculate_default_profile:  Provided profile requires too many resources\n");
        rtn = HH_ENOMEM;
	goto done;
    }

done:
    if (local_user_profile_p != NULL) { FREE(local_user_profile_p);}
    if (profile_input_data_p != NULL) { FREE(profile_input_data_p);}
    if (icm_alloc_vec_p != NULL) { FREE(icm_alloc_vec_p);}
    return rtn;



}

static void unmap_fw_errbuf(XHH_hob_t rhh_hob_p)
{
    if ((rhh_hob_p->fw_error_buf_start_va != (MT_virt_addr_t)(MT_ulong_ptr_t) NULL)  && 
	(rhh_hob_p->fw_error_iomem_obj != (MOSAL_iomem_t) NULL)){
           MOSAL_iomem_unmap(rhh_hob_p->fw_error_iomem_obj,MOSAL_get_kernel_prot_ctx());
           rhh_hob_p->fw_error_buf_start_va =  (MT_virt_addr_t)(MT_ulong_ptr_t) NULL;
           rhh_hob_p->fw_error_iomem_obj =  (MOSAL_iomem_t) NULL;
    }
}

static void unmap_reserved_fw_resources(XHH_hob_t rhh_hob_p)
{
    XHH_dev_lim_t*              dev_lims_p = &(rhh_hob_p->dev_lims); 
    XHH_icmm_t                  icmm = rhh_hob_p->icmm;
    XHH_icm_base_addr_vector_t  *adrs_vec_p = NULL;
    HH_ret_t ret;
    u_int64_t sz2map = 0;
#if 0
    XHH_icm_allocation_vector_t sz_vec = rhh_hob_p->icm_alloc_size_vec;
    XHH_profile_t* profile_p = &(rhh_hob_p->profile);
    u_int32_t uarc_entry_sz =  (1 << sz_vec.log2_uarc_size) >> profile_p->log2_max_uar ;
#endif    


    adrs_vec_p = (XHH_icm_base_addr_vector_t *)MALLOC(sizeof(XHH_icm_base_addr_vector_t));
    if (adrs_vec_p == NULL) {
       MT_COV_MEM_FLTR_START
       MTL_ERROR1(MT_FLFMT("MALLOC error for adrs_vec_p"));
       return;
       MT_COV_MEM_FLTR_END
    }
    *adrs_vec_p = rhh_hob_p->icm_alloc_base_addrs_vec;
    
    if (dev_lims_p->srq) {
       sz2map = MT_UP_ALIGNX_SIZE(dev_lims_p->srq_entry_sz << dev_lims_p->log2_rsvd_srqs,MOSAL_SYS_PAGE_SHIFT); 
       ret= XHH_icmm_unmap(icmm,adrs_vec_p->srqc_base_addr,sz2map);
       if (ret!= HH_OK) {
           MTL_ERROR1("failed unmapping rsvd srqs \n");
       }
    }

    if (dev_lims_p->num_rsvd_eqs > 0) {
        sz2map = MT_UP_ALIGNX_SIZE(dev_lims_p->num_rsvd_eqs * dev_lims_p->eqc_entry_sz,MOSAL_SYS_PAGE_SHIFT); 
        ret= XHH_icmm_unmap(icmm,adrs_vec_p->eqc_base_addr,sz2map);
        if (ret!= HH_OK) {
          MTL_ERROR1("failed unmapping rsvd eqs \n");
        }
    }
    
    sz2map = MT_UP_ALIGNX_SIZE(dev_lims_p->cqc_entry_sz << dev_lims_p->log2_rsvd_cqs,MOSAL_SYS_PAGE_SHIFT); 
    ret= XHH_icmm_unmap(icmm,adrs_vec_p->cqc_base_addr,sz2map);
    if (ret!= HH_OK) {
      MTL_ERROR1("failed unmapping rsvd cqs \n");
    }

    sz2map = MT_UP_ALIGNX_SIZE(dev_lims_p->qpc_entry_sz << dev_lims_p->log2_rsvd_qps ,MOSAL_SYS_PAGE_SHIFT); 
    ret= XHH_icmm_unmap(icmm, adrs_vec_p->qpc_base_addr,sz2map);
    if (ret!= HH_OK) {
      MTL_ERROR1("failed unmapping rsvd qps \n");
    }

    sz2map = MT_UP_ALIGNX_SIZE(1 << (dev_lims_p->log2_rsvd_mrws + XHH_ICM_LOG2_MPT_ENTRY_SIZE),
                               MOSAL_SYS_PAGE_SHIFT);  
    ret= XHH_icmm_unmap(icmm,adrs_vec_p->mpt_base_addr,sz2map);
    if (ret!= HH_OK) {
      MTL_ERROR1("failed unmapping rsvd mpts \n");
    }

    sz2map = MT_UP_ALIGNX_SIZE(1 << (dev_lims_p->log2_rsvd_mtts + XHH_ICM_LOG2_MTT_ENTRY_SIZE),
                               MOSAL_SYS_PAGE_SHIFT);  
    ret= XHH_icmm_unmap(icmm,adrs_vec_p->mtt_base_addr,sz2map);
    if (ret!= HH_OK) {
      MTL_ERROR1("failed unmapping rsvd mtts \n");
    }

#if 0
    if (dev_lims_p->num_rsvd_uars > 0) {
        sz2map = MT_UP_ALIGNX_SIZE((dev_lims_p->num_rsvd_uars * uarc_entry_sz),MOSAL_SYS_PAGE_SHIFT); 
        ret= XHH_icmm_unmap(icmm, adrs_vec_p->uarc_base_addr,sz2map);
        if (ret!= HH_OK) {
          MTL_ERROR1("failed unmapping rsvd uars \n");
        }
    }
#endif        
    FREE(adrs_vec_p);
    return;
}


/*****************************************************************************
******************************************************************************
************************  HOB Interface FUNCTIONS ****************************
******************************************************************************
******************************************************************************/

/******************************************************************************
 *  Function:     XHH_hob_close_hca
 *
 *  Description:  This function stops HCA hardware activity and frees all associated resources.
 *
 *  input:
 *                hca_hndl
 *  output: 
 *                none
 *  returns:
 *                HH_OK
 *                HH_EINVAL_HCA_HNDL
 *                HH_ERR
 *
 *  Comments:     If any errors occur, continue process of de-allocating resources.  However, log the errors,
 *                and return HH_ERR instead of HH_OK
 *
 *****************************************************************************/
HH_ret_t XHH_hob_close_hca_internal(HH_hca_hndl_t  hca_hndl, MT_bool invoked_from_destroy)
{

    /* TBD - Complete function */
    XHH_cmd_status_t    cmd_ret;
    MT_bool             have_error = FALSE;
    HH_ret_t            ret = HH_OK;
    u_int32_t           i;
    u_int16_t           num_ports;
    XHH_hob_t  rhh_hob_p;
    DECLARE_FATAL_VARS;

    MT_RETURN_IF_LOW_STACK(XHH_WATERMARK);
    if (MOSAL_get_exec_ctx() != MOSAL_IN_TASK) {
        MTL_ERROR1("XHH_hob_close_hca: NOT IN TASK CONTEXT)\n");
        return HH_ERR;
    }

    if (hca_hndl == NULL) {
        MTL_ERROR1("XHH_hob_close_hca : ERROR : Invalid HCA handle\n");
        return HH_EINVAL_HCA_HNDL;
    }
    rhh_hob_p = XHHOBP(hca_hndl);
    
    /* uninterruptible acquire.  Want to be sure to clean up */
    MOSAL_mutex_acq_ui(&(rhh_hob_p->mtx));
    if (hca_hndl->status == HH_HCA_STATUS_CLOSED) {
        MOSAL_mutex_rel(&(rhh_hob_p->mtx));
        MTL_ERROR1("XHH_hob_close_hca: Device already closed\n");
        return HH_EINVAL_HCA_HNDL;
    }
    
    /* move the HCA to CLOSING state, preserving fatal indicators */
    MOSAL_spinlock_dpc_lock(&rhh_hob_p->fatal_spl);
    if ((rhh_hob_p->xhh_state & XHH_STATE_RUNNING) == 0) {
        MOSAL_spinlock_unlock(&rhh_hob_p->fatal_spl);
        MOSAL_mutex_rel(&(rhh_hob_p->mtx));
        MTL_ERROR1(MT_FLFMT("XHH_hob_close_hca:  already invoked"));
        return HH_EBUSY;
    }
    rhh_hob_p->xhh_state &= XHH_STATE_HAVE_ANY_FATAL;
    rhh_hob_p->xhh_state |= XHH_STATE_CLOSING;
    MOSAL_spinlock_unlock(&rhh_hob_p->fatal_spl);

    /* transfer to closing state */    
    WAIT_IF_FATAL(rhh_hob_p);
    if (have_fatal == FALSE) {
        num_ports = rhh_hob_p->dev_lims.num_ports;
        for (i = 1; i <= num_ports; i++) {
            cmd_ret = XHH_cmd_CLOSE_IB(rhh_hob_p->cmd, (IB_port_t) i);
            if (cmd_ret != XHH_CMD_STAT_OK) {
                MTL_ERROR1("XHH_hob_close_hca: XHH_cmd_CLOSE_IB error (%d)\n", cmd_ret);
                have_error = TRUE;
            }
        }
    }
    /* test if a fatal error occurred during CLOSE_IB. */
    if (have_fatal == FALSE) {
        WAIT_IF_FATAL(rhh_hob_p);
    }
    rhh_hob_p->compl_eq = XHH_INVALID_EQN;
    rhh_hob_p->ib_eq = XHH_INVALID_EQN;
    
    {
	call_result_t      mosal_ret;
	MT_USE_VARIABLE(mosal_ret);
	MT_START_FATAL_HANDLING;

	/* release the fatal signalling thread */
	rhh_hob_p->fatal_thread_obj.have_fatal = FALSE;
	MOSAL_syncobj_signal(&rhh_hob_p->fatal_thread_obj.fatal_err_sync);
	mosal_ret = MOSAL_syncobj_waiton(&(rhh_hob_p->fatal_thread_obj.stop_sync), 10000000);
	if (mosal_ret != MT_OK) {
	    if (mosal_ret == MT_EINTR) {
	        MTL_DEBUG1(MT_FLFMT("%s: Received OS interrupt while destroying fatal error thread (err = %d)"), 
	                   __func__,mosal_ret);
	        have_error = TRUE;
	    } else {
	        MTL_ERROR1(MT_FLFMT("%s: Timeout on destroying fatal error thread (err = %d)"), 
	                   __func__,mosal_ret);
	        have_error = TRUE;
	    }
	}
	MT_END_FATAL_HANDLING;
    }

    /* destroy eventq mgr.  Event manager must destroy all EQs */
    ret = XHH_eventp_destroy( rhh_hob_p->eventp );
    if (ret != HH_OK) {
        MTL_ERROR1("XHH_hob_close_hca: XHH_eventp_destroy error (%d:%s)\n", ret, HH_strerror_sym(ret));
        have_error = TRUE;
    }
    rhh_hob_p->eventp = (XHH_eventp_t)XHH_INVALID_HNDL;

    if (rhh_hob_p->mcgm != (XHH_mcgm_t)XHH_INVALID_HNDL) {
        ret = XHH_mcgm_destroy(rhh_hob_p->mcgm );
        if (ret != HH_OK) {
            MTL_ERROR1("XHH_hob_close_hca: XHH_mcgm_destroy error (%d:%s)\n", ret, HH_strerror_sym(ret));
            have_error = TRUE;
        }
        rhh_hob_p->mcgm = (XHH_mcgm_t)XHH_INVALID_HNDL;
    }

    
    /*TBD: do we need kar in arbel??
    rhh_hob_p->kar_addr = (MT_virt_addr_t) 0;
    rhh_hob_p->kar_indx = (u_int32_t) XHH_INVALID_HNDL;*/

    if (rhh_hob_p->srqm != (XHH_srqm_t)XHH_INVALID_HNDL) { /* SRQs are supported - SRQM exists */
      ret = XHH_srqm_destroy( rhh_hob_p->srqm);
      if (ret != HH_OK) {
          MTL_ERROR1("XHH_hob_close_hca: XHH_srqm_destroy error (%d:%s)\n", ret, HH_strerror_sym(ret));
          have_error = TRUE;
      }
      rhh_hob_p->srqm = (XHH_srqm_t)XHH_INVALID_HNDL;
    }
    
    ret = XHH_qpm_destroy( rhh_hob_p->qpm, have_error);
    if (ret != HH_OK) {
        MTL_ERROR1("XHH_hob_close_hca: XHH_qpm_destroy error (%d:%s)\n", ret, HH_strerror_sym(ret));
        have_error = TRUE;
    }
    rhh_hob_p->qpm = (XHH_qpm_t)XHH_INVALID_HNDL;

    FREE(rhh_hob_p->init_ib_props);
    rhh_hob_p->init_ib_props = ( XHH_port_init_props_t *) NULL;

    ret = XHH_cqm_destroy( rhh_hob_p->cqm, have_error);
    if (ret != HH_OK) {
        MTL_ERROR1("XHH_hob_close_hca: XHH_cqm_destroy error (%d:%s)\n", ret, HH_strerror_sym(ret));
        have_error = TRUE;
    }
    rhh_hob_p->cqm = (XHH_cqm_t)XHH_INVALID_HNDL;

    ret = XHH_mrwm_destroy(rhh_hob_p->mrwm, have_error);
    if (ret != HH_OK) {
        MTL_ERROR1("XHH_hob_close_hca: XHH_mrwm_destroy error (%d:%s)\n", ret, HH_strerror_sym(ret));
        have_error = TRUE;
    }
    rhh_hob_p->mrwm = (XHH_mrwm_t)XHH_INVALID_HNDL;

    ret = XHH_pdm_destroy(rhh_hob_p->pdm );
    if (ret != HH_OK) {
        MTL_ERROR1("XHH_hob_close_hca: XHH_pdm_destroy error (%d:%s)\n", ret, HH_strerror_sym(ret));
        have_error = TRUE;
    }
    rhh_hob_p->pdm = (XHH_pdm_t)XHH_INVALID_HNDL;

    ret = XHH_uarm_destroy(rhh_hob_p->uarm );
    if (ret != HH_OK) {
        MTL_ERROR1("XHH_hob_close_hca: XHH_uarm_destroy error (%d:%s)\n", ret, HH_strerror_sym(ret));
        have_error = TRUE;
    }
    rhh_hob_p->uarm = (XHH_uarm_t)XHH_INVALID_HNDL;

    unmap_reserved_fw_resources(rhh_hob_p);
    
    /* test for fatal again here */
    if (have_fatal == FALSE) {
        WAIT_IF_FATAL(rhh_hob_p);
    }
    if (have_fatal == FALSE) {
        MTL_TRACE1("XHH_hob_close_hca: Performing XHH_cmd_CLOSE_HCA (no fatal)\n");
        cmd_ret = XHH_cmd_CLOSE_HCA(rhh_hob_p->cmd, FALSE);
        if (cmd_ret != XHH_CMD_STAT_OK) {
            MTL_ERROR1("XHH_hob_close_hca: XHH_cmd_CLOSE_HCA error (%d)\n", cmd_ret);
            have_error = TRUE;
        }
    }
    hca_hndl->status = HH_HCA_STATUS_CLOSED;

    /* move state to "CLOSED"*/
    MOSAL_spinlock_dpc_lock(&rhh_hob_p->fatal_spl);
    rhh_hob_p->xhh_state &= XHH_STATE_HAVE_ANY_FATAL;
    rhh_hob_p->xhh_state |= XHH_STATE_CLOSED;
    MOSAL_spinlock_unlock(&rhh_hob_p->fatal_spl);
    MTL_TRACE2("XHH_hob_close_hca: device name %s\n", hca_hndl->dev_desc);
    
    ret = XHH_icmm_destroy(rhh_hob_p->icmm);
    if (ret != HH_OK) {
        MTL_ERROR1("XHH_hob_close_hca: XHH_icmm_destroy error (%d:%s)\n", ret, HH_strerror_sym(ret));
        have_error = TRUE;
    }
    rhh_hob_p->icmm = (XHH_icmm_t)XHH_INVALID_HNDL;

    cmd_ret = XHH_cmd_UNMAP_ICM_AUX(rhh_hob_p->cmd);
    if (cmd_ret != XHH_CMD_STAT_OK) {
        MTL_ERROR1("XHH_hob_close_hca: XHH_cmd_UNMAP_ICM_AUX error (%d)\n", cmd_ret);
        have_error = TRUE;
    }

    cmd_ret = XHH_cmd_UNMAP_FA(rhh_hob_p->cmd);
    if (cmd_ret != XHH_CMD_STAT_OK) {
        MTL_ERROR1("XHH_hob_close_hca: XHH_cmd_UNMAP_FA error (%d)\n", cmd_ret);
        have_error = TRUE;
    }
    
    if (rhh_hob_p->have_lam == FALSE) {
        XHH_phys_mem_free(rhh_hob_p->fw_phys_mem_vec_p, rhh_hob_p->fw_phys_mem_vec_size);
        rhh_hob_p->fw_phys_mem_vec_p = NULL;
        rhh_hob_p->fw_phys_mem_vec_size = 0;
        XHH_phys_mem_free(rhh_hob_p->icm_aux_phys_mem_vec_p, rhh_hob_p->icm_aux_phys_mem_vec_size);
        rhh_hob_p->icm_aux_phys_mem_vec_p = NULL;
        rhh_hob_p->icm_aux_phys_mem_vec_size = 0;
    } else {
        ret = XHH_lamm_destroy(rhh_hob_p->lamm);
        if (ret != HH_OK) {
            MTL_ERROR1("XHH_hob_close_hca: XHH_lamm_destroy error (%d:%s)\n", ret, HH_strerror_sym(ret));
            have_error = TRUE;
        }
        rhh_hob_p->lamm = (XHH_lamm_t) XHH_INVALID_HNDL;
        rhh_hob_p->have_lam = FALSE;
        cmd_ret = XHH_cmd_DISABLE_LAM(rhh_hob_p->cmd);
        if (cmd_ret != XHH_CMD_STAT_OK) {
            MTL_ERROR1("XHH_hob_close_hca: XHH_cmd_DISABLE_LAM error (%d)\n", cmd_ret);
            have_error = TRUE;
        }
    }
    
    if (rhh_hob_p->fw_error_buf != NULL) {
	FREE(rhh_hob_p->fw_error_buf);
	rhh_hob_p->fw_error_buf = NULL;
    }
    unmap_fw_errbuf(rhh_hob_p);
    XHH_cmd_destroy(rhh_hob_p->cmd);
    
    MOSAL_mutex_rel(&(rhh_hob_p->mtx));

    if (have_fatal == FALSE) {
        WAIT_IF_FATAL(rhh_hob_p);
    }
    if ((have_fatal == TRUE) && (invoked_from_destroy == FALSE)) {
    	//leo (WINDOWS) restart doesn't work, because the card reset doesn't work
	MT_START_FATAL_HANDLING;
        MTL_ERROR1("XHH_hob_close_hca: HAVE FATAL, restarting\n");
        ret = XHH_hob_restart(hca_hndl);
        if (ret != HH_OK) {
            MTL_ERROR1("XHH_hob_close_hca: XHH_hob_restart error (%d)\n", ret);
            have_error = TRUE;
        }
	MT_END_FATAL_HANDLING;
    }

    if (have_error && (have_fatal == FALSE)) {
        return HH_ERR;
    } else {
        return(HH_OK);
    }
} /* XHH_hob_close_hca_internal */

HH_ret_t XHH_hob_close_hca(HH_hca_hndl_t  hca_hndl)
{
    MT_RETURN_IF_LOW_STACK(XHH_WATERMARK);
    return (XHH_hob_close_hca_internal(hca_hndl,FALSE));
}


/******************************************************************************
 *  Function:     XHH_hob_destroy_internal
 *
 *  Description:  Deregister given device from HH and free the HH object.
 *
 *  input:
 *                hca_hndl
 *                restarting -- TRUE if called as part of fatal error restart.
 *  output: 
 *                none
 *  returns:
 *                HH_OK
 *                HH_EINVAL_HCA_HNDL
 *
 *  Comments:     If HCA is still open,XHH_hob_close_hca() is 
 *                invoked before freeing the XHH_hob.
 *                If called as part of restarting, the HH device is NOT de-registered.
 *
 *                Returns HH_EINVAL_HCA_HNDL if any function called internally fails  
 *
 *****************************************************************************/
HH_ret_t    XHH_hob_destroy_internal(HH_hca_hndl_t hca_hndl, MT_bool restarting)
{
  HH_ret_t  ret, fn_ret = HH_OK;
  XHH_hob_t          hob_p;
  int                int_ret = 0;
  MT_bool            have_fatal = FALSE;

  MT_RETURN_IF_LOW_STACK(XHH_WATERMARK);
  if (MOSAL_get_exec_ctx() != MOSAL_IN_TASK) {
      MTL_ERROR1("XHH_hob_destroy: NOT IN TASK CONTEXT)\n");
      return HH_ERR;
  }

  if (hca_hndl == NULL) {
      MTL_ERROR1("XHH_hob_close_hca : ERROR : Invalid HCA handle\n");
      return HH_EINVAL_HCA_HNDL;
  }
  
  /* return ERROR if device is still open */
  if ((hca_hndl->status) == HH_HCA_STATUS_OPENED) {
      MTL_ERROR1("XHH_hob_destroy:  Unloading device %s: while it is still open.  Attempting to close it.\n", hca_hndl->dev_desc);
      ret = XHH_hob_close_hca_internal(hca_hndl, TRUE);
      if (ret != HH_OK) {
        MTL_ERROR1("XHH_hob_destroy: Could not close device %s: not opened or unknown (err=%d)\n", hca_hndl->dev_desc, ret);
        fn_ret = HH_EINVAL_HCA_HNDL;
      }
      hca_hndl->status = HH_HCA_STATUS_CLOSED;
  }

  MTL_TRACE2("XHH_hob_destroy: removing the device %s\n",  hca_hndl->dev_desc);
  hob_p = XHHOBP(hca_hndl);

  /* move the HCA to DESTROYING state, preserving fatal indicators */
  MOSAL_spinlock_dpc_lock(&hob_p->fatal_spl);
  if ((hob_p->xhh_state & XHH_STATE_DESTROYING) != 0) {
      MOSAL_spinlock_unlock(&hob_p->fatal_spl);
      MTL_ERROR1(MT_FLFMT("XHH_hob_destroy:  already invoked"));
      return HH_EBUSY;
  }
  hob_p->xhh_state &= XHH_STATE_HAVE_ANY_FATAL;
  hob_p->xhh_state |= XHH_STATE_DESTROYING;
  MOSAL_spinlock_unlock(&hob_p->fatal_spl);
  
  if (!restarting) {
      /* Module unloading flow */

      /* if a kernel application is still active, NULL-out its "user-level" ops-table pointer so that
       * attempts to access any "user-level" HHUL functions (in kernel, for a kernel app) will be intercepted
       * by the macros defined in file hhul.h before any actual invocation is attempted 
       */
      if (hca_hndl->ul_if_ops_p) {
	  *(int **)(hca_hndl->ul_if_ops_p) = NULL;
      }
    
      /* unregister the device from HH */
      ret = HH_rem_hca_dev_prep(hca_hndl);
      if (ret != HH_OK) {
	  MTL_ERROR1("XHH_hob_destroy: Could not remove device 0x%p: unknown (%d)\n", hca_hndl, ret);
	  fn_ret = HH_EINVAL_HCA_HNDL;
      }
  } else {
      /* restart flow */
      MTL_DEBUG1(MT_FLFMT("%s: Setting ZOMBIE status. hca_hndl=%p"), __func__, hca_hndl);
      hca_hndl->status = HH_HCA_STATUS_ZOMBIE;
  }
      
  if ((hob_p->xhh_state & XHH_STATE_HAVE_ANY_FATAL) != 0) {
      /* halt the HCA if delayed-halt flag was set, in fatal case only, 
       * to make sure that there is no PCI activity.  
       */
      have_fatal = TRUE;
      if (hob_p->module_flags.fatal_delay_halt != 0) {
	  MTL_DEBUG1(MT_FLFMT("%s: performing delayed halt-hca"), __func__);
	  XHH_hob_halt_hca(hob_p);
      }
   } else {
       have_fatal = FALSE;
   }

  /* do PCI reset here if have a catastrophic error */
  if (have_fatal == TRUE) {
    /* perform sw reset */
    {
	  MT_START_FATAL_HANDLING;
	  /* Do the Arbel RESET */
	  XHH_hob_reset_device(&hob_p->hw_props);
    
	  if (hob_p->pci_hca_info.is_valid == TRUE) {
	      XHH_write_pci_config(hob_p->hw_props.pci_dev,
			       &hob_p->pci_hca_info.config);
	  }
	  MT_END_FATAL_HANDLING;
    }
  }

  int_ret = VIP_delay_unlock_destroy(hob_p->delay_unlocks);
  if (int_ret != 0) {
      MTL_ERROR1("XHH_hob_destroy: Could not destroy delay_unlocks (err = %d)\n", int_ret);
      fn_ret = HH_ERR;
  }
  
  MOSAL_mutex_free(&(hob_p->mtx));
  /* Finally, free the RHH object */
  FREE(hob_p);

  return(fn_ret);
}

/******************************************************************************
 *  Function:     XHH_hob_destroy
 *
 *  Description:  Deregister given device from HH and free the HH object.
 *
 *  input:
 *                hca_hndl
 *  output: 
 *                none
 *  returns:
 *                HH_OK
 *                HH_ERR  (not in task context)
 *                HH_EINVAL_HCA_HNDL (invalid HCA handle, or internal call failure
 *
 *  Comments:     Wait for HCA closed before freeing the XHH_hob (this is only
 *                active if have hotplug support).
 *
 *****************************************************************************/
HH_ret_t    XHH_hob_destroy(HH_hca_hndl_t hca_hndl)
{	    
	HH_ret_t ret, ret1;
        ret = XHH_hob_waitfor_hca_closed(hca_hndl);
        if (ret != HH_OK) {
	    return ret;
	}
	ret = XHH_hob_destroy_internal(hca_hndl, FALSE);
	if (ret != HH_OK) {
		MTL_ERROR1(MT_STDFMT("Failed XHH_hob_destroy for handle=%p (%d,%s)"),
			   hca_hndl, ret, HH_strerror_sym(ret));
	}
        ret1 = HH_notify_rem_hca_dev(hca_hndl);
	if (ret1 != HH_OK) {
		MTL_ERROR1(MT_STDFMT("Failed XHH_hob_destroy for handle=%p (%d,%s)"),
			   hca_hndl, ret1, HH_strerror_sym(ret1));
	}
	return (ret == HH_OK ? ret1 : ret);
}

static void setup_hca_init_props(XHH_hob_t        rhh_hob_p,
                                 XHH_hca_props_t  *local_hca_props_p)
{
    memset(local_hca_props_p, 0, sizeof(XHH_hca_props_t));

    /* */
#ifdef MT_LITTLE_ENDIAN
    local_hca_props_p->he = ARBEL_IF_HOST_LTLENDIAN;
#else
    local_hca_props_p->he = ARBEL_IF_HOST_BIGENDIAN;
#endif
    local_hca_props_p->re = FALSE;   /* not a router */
    local_hca_props_p->wqe_quota_en = FALSE;   
    local_hca_props_p->udp = TRUE;   /* check port in UD AV */ 
    local_hca_props_p->ud =  FALSE;  /* no privileged UDAVs */
    local_hca_props_p->responder_exu = 0; /* How many exec engines are dedicated to the responder. 0 is "auto" */
    local_hca_props_p->time_stamp_granularity = 0;
    local_hca_props_p->ce =  FALSE;
    local_hca_props_p->sph =  FALSE;
    local_hca_props_p->rph =  FALSE;
    local_hca_props_p->bmme =  FALSE;    
    local_hca_props_p->responder_exu = 0; /* How many exec engines are dedicated to the responder. 0 is "auto" */
    local_hca_props_p->wqe_quota = 0;
    local_hca_props_p->wqe_quota_en =  FALSE; 

    local_hca_props_p->qpc_eec_cqc_eqc_rdb_parameters.cqc_base_addr = 
        rhh_hob_p->icm_alloc_base_addrs_vec.cqc_base_addr;
    local_hca_props_p->qpc_eec_cqc_eqc_rdb_parameters.ecqc_base_addr = 
        rhh_hob_p->icm_alloc_base_addrs_vec.ecqc_base_addr;
    local_hca_props_p->qpc_eec_cqc_eqc_rdb_parameters.eec_base_addr = 
        rhh_hob_p->icm_alloc_base_addrs_vec.eec_base_addr;
    local_hca_props_p->qpc_eec_cqc_eqc_rdb_parameters.qpc_base_addr = 
        rhh_hob_p->icm_alloc_base_addrs_vec.qpc_base_addr;
    local_hca_props_p->qpc_eec_cqc_eqc_rdb_parameters.log_num_of_cq = 
      (u_int8_t)rhh_hob_p->profile.log2_max_cqs;
    // local_hca_props_p->qpc_eec_cqc_eqc_rdb_parameters.log_num_of_ee = rhh_hob_p->profile.log2_max_eecs;
    local_hca_props_p->qpc_eec_cqc_eqc_rdb_parameters.log_num_of_ee = 0;
    local_hca_props_p->qpc_eec_cqc_eqc_rdb_parameters.log_num_eq = rhh_hob_p->profile.log2_max_eqs;
    local_hca_props_p->qpc_eec_cqc_eqc_rdb_parameters.log_num_of_qp = (u_int8_t)rhh_hob_p->profile.log2_max_qps;
    local_hca_props_p->qpc_eec_cqc_eqc_rdb_parameters.eqpc_base_addr = 
        rhh_hob_p->icm_alloc_base_addrs_vec.eqpc_base_addr;
    local_hca_props_p->qpc_eec_cqc_eqc_rdb_parameters.log_num_of_srq = (u_int8_t)rhh_hob_p->profile.log2_max_srqs;
    local_hca_props_p->qpc_eec_cqc_eqc_rdb_parameters.srqc_base_addr = 
        rhh_hob_p->icm_alloc_base_addrs_vec.srqc_base_addr;
    local_hca_props_p->qpc_eec_cqc_eqc_rdb_parameters.eqc_base_addr = 
        rhh_hob_p->icm_alloc_base_addrs_vec.eqc_base_addr;
    local_hca_props_p->qpc_eec_cqc_eqc_rdb_parameters.eeec_base_addr = 
        rhh_hob_p->icm_alloc_base_addrs_vec.eeec_base_addr;
    local_hca_props_p->qpc_eec_cqc_eqc_rdb_parameters.rdb_base_addr = 
        (rhh_hob_p->icm_alloc_base_addrs_vec.rdb_base_addr) & MAKE_ULONGLONG(0xFFFFFFFF00000000);
    
    /* This udavtable section is a "don't care", because the "ud" flag above is set to FALSE */
    local_hca_props_p->udavtable_memory_parameters.l_key = XHH_UDAVM_PRIV_RESERVED_LKEY;
    local_hca_props_p->udavtable_memory_parameters.pd    = XHH_RESERVED_PD(rhh_hob_p);
    local_hca_props_p->udavtable_memory_parameters.xlation_en = TRUE;
    
    local_hca_props_p->tpt_parameters.log_mpt_sz = (u_int8_t)rhh_hob_p->profile.log2_max_mpt_entries;
    local_hca_props_p->tpt_parameters.mpt_base_adr = 
        rhh_hob_p->icm_alloc_base_addrs_vec.mpt_base_addr;
    local_hca_props_p->tpt_parameters.mtt_base_addr = 
        rhh_hob_p->icm_alloc_base_addrs_vec.mtt_base_addr;
    local_hca_props_p->tpt_parameters.pfto = 0;   /* TBD -- not yet supported. Page Fault RNR Timeout */
    
    local_hca_props_p->uar_parameters.uar_base_addr = rhh_hob_p->hw_props.uar_base;
	local_hca_props_p->uar_parameters.log_uar_entry_sz = 
		(u_int8_t)(rhh_hob_p->icm_alloc_size_vec.log2_uarc_size - rhh_hob_p->profile.log2_max_uar - 12);
	local_hca_props_p->uar_parameters.log_max_uars = (u_int8_t)rhh_hob_p->profile.log2_max_uar;
    local_hca_props_p->uar_parameters.uar_page_sz   = rhh_hob_p->profile.log2_uar_pg_size - 12;
    local_hca_props_p->uar_parameters.uar_scratch_base_addr = 
        rhh_hob_p->icm_alloc_base_addrs_vec.uar_scratch_base_addr;
	local_hca_props_p->uar_parameters.uar_context_base_addr = 
		rhh_hob_p->icm_alloc_base_addrs_vec.uarc_base_addr;

	MTL_DEBUG1(MT_FLFMT("%s: log_uarc_entry_size=0x%x, log_max_uars=0x%x, uar_context_base_addr="U64_FMT),
			   __func__, local_hca_props_p->uar_parameters.log_uar_entry_sz ,
			   local_hca_props_p->uar_parameters.log_max_uars,
			   local_hca_props_p->uar_parameters.uar_context_base_addr);
    
    local_hca_props_p->multicast_parameters.log_mc_table_sz = (u_int8_t)rhh_hob_p->profile.log2_max_mcgs;
    local_hca_props_p->multicast_parameters.mc_base_addr    = 
        rhh_hob_p->icm_alloc_base_addrs_vec.mcg_base_addr;
    local_hca_props_p->multicast_parameters.mc_hash_fn      = 0;
    local_hca_props_p->multicast_parameters.log_mc_table_entry_sz = (u_int16_t)(rhh_hob_p->profile.log2_mcg_entry_size);
    local_hca_props_p->multicast_parameters.mc_table_hash_sz = 1 << (rhh_hob_p->profile.log2_mcg_hash_size);
}

static HH_ret_t map_reserved_fw_resources(XHH_hob_t rhh_hob_p)
{
    XHH_dev_lim_t*              dev_lims_p = &(rhh_hob_p->dev_lims); 
    XHH_icmm_t                  icmm = rhh_hob_p->icmm;
    XHH_icm_base_addr_vector_t  *adrs_vec_p = NULL;    
    HH_ret_t ret;
    u_int64_t sz2map = 0;
#if 0
    XHH_icm_allocation_vector_t sz_vec = rhh_hob_p->icm_alloc_size_vec;
    XHH_profile_t* profile_p = &(rhh_hob_p->profile);
    u_int32_t uarc_entry_sz =  (u_int32_t)(((u_int64_t)(MAKE_ULONGLONG(1) << sz_vec.log2_uarc_size)) >> profile_p->log2_max_uar);
#endif
    
    MTL_DEBUG1(" map_reserved_fw_resources { \n");
    MTL_DEBUG1("RESERVED FW RESOURCES: \n");
    MTL_DEBUG1("---------------------- \n");
    MTL_DEBUG1("num reserved pds:0x%x   \n",dev_lims_p->num_rsvd_pds);
#if 0
    MTL_DEBUG1("num reserved uars:0x%x   uarc entry sz:0x%x \n",
               dev_lims_p->num_rsvd_uars, uarc_entry_sz  );
#endif
    MTL_DEBUG1("num reserved mtts:0x%x   mtt_entry sz:0x%x \n",
            1 << dev_lims_p->log2_rsvd_mtts, 1 << XHH_ICM_LOG2_MTT_ENTRY_SIZE);
    MTL_DEBUG1("num reserved mpts:0x%x   mpt_entry sz:0x%x \n",
               1 << dev_lims_p->log2_rsvd_mrws, 1 << XHH_ICM_LOG2_MPT_ENTRY_SIZE);
    MTL_DEBUG1("num reserved qps:0x%x    entry sz:0x%x \n",
               1 << dev_lims_p->log2_rsvd_qps, dev_lims_p->qpc_entry_sz);
    MTL_DEBUG1("num reserved cqs:0x%x    entry sz:0x%x \n",
               1 << dev_lims_p->log2_rsvd_cqs, dev_lims_p->cqc_entry_sz);
    MTL_DEBUG1("num reserved srqs:0x%x   entry sz:0x%x \n",
               1 << dev_lims_p->log2_rsvd_srqs,dev_lims_p->srq_entry_sz);
    MTL_DEBUG1("num reserved eqs:0x%x   entry sz:0x%x \n",
               dev_lims_p->num_rsvd_eqs,dev_lims_p->eqc_entry_sz);
    
    /* TBD : rdb */

    adrs_vec_p = (XHH_icm_base_addr_vector_t *)MALLOC(sizeof(XHH_icm_base_addr_vector_t));
    if (adrs_vec_p == NULL) {
       MT_COV_MEM_FLTR_START
       MTL_ERROR1(MT_FLFMT("MALLOC error for adrs_vec_p"));
       return HH_EAGAIN;
       MT_COV_MEM_FLTR_END
    }
    *adrs_vec_p = rhh_hob_p->icm_alloc_base_addrs_vec;

#if 0
    if (dev_lims_p->num_rsvd_uars > 0) {
        sz2map = MT_UP_ALIGNX_SIZE((dev_lims_p->num_rsvd_uars * uarc_entry_sz),MOSAL_SYS_PAGE_SHIFT); 
        ret= XHH_icmm_map(icmm, adrs_vec_p->uarc_base_addr, sz2map, ICMM_MAP_FORCE_NONE, 0, NULL);
        if (ret!= HH_OK) {
            MTL_ERROR1("failed mapping rsvd uars  \n");
            goto map_uarc_err;
        }
    }
#endif
    
    sz2map = MT_UP_ALIGNX_SIZE(1 << (dev_lims_p->log2_rsvd_mtts + XHH_ICM_LOG2_MTT_ENTRY_SIZE),
                               MOSAL_SYS_PAGE_SHIFT); 
    MTL_DEBUG1("mtt tbl base adrs:"U64_FMT " mtt map size:"U64_FMT" \n",adrs_vec_p->mtt_base_addr,sz2map);
    ret= XHH_icmm_map(icmm,adrs_vec_p->mtt_base_addr,sz2map,ICMM_MAP_FORCE_NONE,0,NULL);
    if (ret!= HH_OK) {
        MTL_ERROR1("failed mapping rsvd mtts \n");
        goto map_mtt_err;
    }

    sz2map = MT_UP_ALIGNX_SIZE(1 << (dev_lims_p->log2_rsvd_mrws + XHH_ICM_LOG2_MPT_ENTRY_SIZE),
                               MOSAL_SYS_PAGE_SHIFT); 
    MTL_DEBUG1("mpt tbl base adrs:"U64_FMT " mpt map size:"U64_FMT" \n",adrs_vec_p->mpt_base_addr,sz2map);
    ret= XHH_icmm_map(icmm,adrs_vec_p->mpt_base_addr,sz2map,ICMM_MAP_FORCE_NONE,0,NULL);
    if (ret!= HH_OK) {
        MTL_ERROR1("failed mapping rsvd mpts \n");
        goto map_mpt_err;
    }

    sz2map = MT_UP_ALIGNX_SIZE(dev_lims_p->qpc_entry_sz << dev_lims_p->log2_rsvd_qps,MOSAL_SYS_PAGE_SHIFT); 
    ret= XHH_icmm_map(icmm,adrs_vec_p->qpc_base_addr,sz2map,ICMM_MAP_FORCE_NONE,0,NULL);
    if (ret!= HH_OK) {
        MTL_ERROR1("failed mapping rsvd qps \n");
        goto map_qp_err;
    }

    sz2map = MT_UP_ALIGNX_SIZE(dev_lims_p->eqpc_entry_sz << dev_lims_p->log2_rsvd_qps,MOSAL_SYS_PAGE_SHIFT); 
    ret= XHH_icmm_map(icmm,adrs_vec_p->eqpc_base_addr,sz2map,ICMM_MAP_FORCE_NONE,0,NULL);
    if (ret!= HH_OK) {
        MTL_ERROR1("failed mapping eqpc for rsvd qps \n");
        goto map_qp_err;
    }

    sz2map = MT_UP_ALIGNX_SIZE(dev_lims_p->cqc_entry_sz << dev_lims_p->log2_rsvd_cqs,MOSAL_SYS_PAGE_SHIFT); 
    ret= XHH_icmm_map(icmm,adrs_vec_p->cqc_base_addr,sz2map,ICMM_MAP_FORCE_NONE,0,NULL);
    if (ret!= HH_OK) {
        MTL_ERROR1("failed mapping rsvd cqs \n");
        goto map_cq_err;
    }

    if (dev_lims_p->num_rsvd_eqs > 0) {
        sz2map = MT_UP_ALIGNX_SIZE(dev_lims_p->eqc_entry_sz * dev_lims_p->num_rsvd_eqs,MOSAL_SYS_PAGE_SHIFT); 
        ret= XHH_icmm_map(icmm,adrs_vec_p->eqc_base_addr,sz2map,ICMM_MAP_FORCE_NONE,0,NULL);
        if (ret!= HH_OK) {
            MTL_ERROR1("failed mapping rsvd eqs \n");
            goto map_eq_err;
        }
    }
    

    if (dev_lims_p->srq) {
        sz2map = MT_UP_ALIGNX_SIZE(dev_lims_p->srq_entry_sz << dev_lims_p->log2_rsvd_srqs,MOSAL_SYS_PAGE_SHIFT); 
        ret= XHH_icmm_map(icmm,adrs_vec_p->srqc_base_addr,sz2map,ICMM_MAP_FORCE_NONE,0,NULL);
        if (ret!= HH_OK) {
            MTL_ERROR1("failed mapping rsvd srqs \n");
            goto map_srq_err;
        }
    }

    MTL_DEBUG1(" map_reserved_fw_resources } \n");
    FREE(adrs_vec_p);
    return HH_OK;


map_srq_err:
    if (dev_lims_p->num_rsvd_eqs > 0) {
        sz2map = MT_UP_ALIGNX_SIZE(dev_lims_p->num_rsvd_eqs * dev_lims_p->eqc_entry_sz,MOSAL_SYS_PAGE_SHIFT); 
        XHH_icmm_unmap(icmm,adrs_vec_p->eqc_base_addr,sz2map);
    }
    

map_eq_err:
    sz2map = MT_UP_ALIGNX_SIZE(dev_lims_p->cqc_entry_sz << dev_lims_p->log2_rsvd_cqs,MOSAL_SYS_PAGE_SHIFT); 
    XHH_icmm_unmap(icmm,adrs_vec_p->cqc_base_addr,sz2map);

map_cq_err:
    sz2map = MT_UP_ALIGNX_SIZE(dev_lims_p->qpc_entry_sz << dev_lims_p->log2_rsvd_qps ,MOSAL_SYS_PAGE_SHIFT); 
    XHH_icmm_unmap(icmm, adrs_vec_p->qpc_base_addr,sz2map);

map_qp_err:
    sz2map = MT_UP_ALIGNX_SIZE(1 << (dev_lims_p->log2_rsvd_mrws + XHH_ICM_LOG2_MPT_ENTRY_SIZE),
                               MOSAL_SYS_PAGE_SHIFT);  
    XHH_icmm_unmap(icmm,adrs_vec_p->mpt_base_addr,sz2map);

map_mpt_err:
    sz2map = MT_UP_ALIGNX_SIZE(1 << (dev_lims_p->log2_rsvd_mtts + XHH_ICM_LOG2_MTT_ENTRY_SIZE),
                               MOSAL_SYS_PAGE_SHIFT);  
    XHH_icmm_unmap(icmm,adrs_vec_p->mtt_base_addr,sz2map);

map_mtt_err:
#if 0
    if (dev_lims_p->num_rsvd_uars > 0) {
        sz2map = MT_UP_ALIGNX_SIZE((dev_lims_p->num_rsvd_uars * uarc_entry_sz),MOSAL_SYS_PAGE_SHIFT); 
        XHH_icmm_unmap(icmm, adrs_vec_p->uarc_base_addr,sz2map);
    }
    

map_uarc_err:
#endif
    MTL_ERROR1(" map_reserved_fw_resources } \n");
    FREE(adrs_vec_p);
    return ret;
}

static HH_ret_t version_info_set_and_check(XHH_hob_t hob_p)
{
    u_int64_t       fw_version;
    
    fw_version = hob_p->fw_props.fw_rev_major;
    fw_version = (fw_version <<16) | hob_p->fw_props.fw_rev_minor;
    fw_version = (fw_version <<16) | hob_p->fw_props.fw_rev_subminor;
    /* enter data into version info structure */
    hob_p->version_info.fw_ver_major = hob_p->fw_props.fw_rev_major;
    hob_p->version_info.fw_ver_minor = hob_p->fw_props.fw_rev_minor;
    hob_p->version_info.fw_ver_subminor = hob_p->fw_props.fw_rev_subminor;
    hob_p->version_info.hw_ver       = hob_p->hw_props.hw_ver;
    hob_p->version_info.cmd_if_ver   = hob_p->fw_props.cmd_interface_rev;
    
	/* check that we are NOT in Hermon or mode of ZB mode:
     * Hermon mode. If '1', then WQE and AV format is the advanced format
     * ZB mode: if 1 then ZB mode of WQ and CQ are enabled (i.e. real Memfree PRM is supported) */
	if (hob_p->fw_props.wqe_h_mode) {
      MTL_ERROR1(" FW version is not appropriate for the driver: Hermon mode is used\n");
	  return HH_ERR;
	}  
    if (hob_p->fw_props.zb_wq_cq) {
      MTL_ERROR1(" FW version is not appropriate for the driver: ZB mode of WQ and CQ are enabled\n");
	  return HH_ERR;
	}  
	
    return HH_OK;
}


static HH_ret_t initialize_fw(XHH_hob_t hob_p, MT_bool *used_lam)
{
    XHH_cmd_pa_vec_t   *pa_vec = NULL;
    u_int32_t  pa_vec_size;
    MT_size_t  lam_size;
    HH_ret_t   ret = HH_ERR;
//  HH_ret_t  ret1;
    XHH_cmd_status_t  cmd_ret;

    if (hob_p->have_lam) {
        /* use area at end of LAM as FW area */
        hob_p->fw_phys_start_addr = hob_p->lam_props.lam_end_adr  - 
             (MT_phys_addr_t)((((u_int64_t)(hob_p->fw_props.fw_pages)) << XHH_ARBEL_LOG_PAGE_SIZE) - 1);

        lam_size = (MT_size_t) (hob_p->lam_props.lam_end_adr - hob_p->lam_props.lam_start_adr + 1) ;
        hob_p->profile.lam_size = lam_size - 
	    ( ((MT_size_t)(hob_p->fw_props.fw_pages)) << XHH_ARBEL_LOG_PAGE_SIZE);

        /* DDR size code is used in XHH_calculate_profile to set number of QPs proportionally to size */
        hob_p->profile.lam_size_code = XHH_get_lam_size_code(hob_p->profile.lam_size);
        
        pa_vec_size = 1; 
        pa_vec = TMALLOC(XHH_cmd_pa_vec_t);
        if (pa_vec == NULL) {
            ret = HH_ENOMEM;
            goto alloc_pa_vec_err;
        }
        pa_vec->phys_addr = hob_p->fw_phys_start_addr;
	
	/*********************************************************/
	/*JPM -- FIX THIS BELOW -- NEED TO MAP ALL THE PAGES!!!! */
	/*********************************************************/
        pa_vec->log2size = (u_int8_t)floor_log2(hob_p->fw_props.fw_pages);

        /* print info messages that device is operating in HIDE LAM mode (not an error) */
        if (hob_p->lam_props.dh == TRUE) {
            MTL_ERROR1("Device %s is operating in HIDE_LAM mode.\n", hob_p->dev_name);
        }
        *used_lam = TRUE;
    } else {
        /* using HOST memory for FW -- either no LAM, or LAM has a problem.
         * Need to save the physical address for freeing during close_hca
         */

        ret = XHH_phys_mem_alloc( ((u_int64_t)((u_int64_t)(hob_p->fw_props.fw_pages)<<XHH_ARBEL_LOG_PAGE_SIZE)),
                                       MOSAL_SYS_PAGE_SHIFT,
                                       MOSAL_MEM_ALLOC_FLAGS_NONE,
                                       XHH_MEM_ALLOC_FLAGS_NONE,
                                       &hob_p->fw_phys_mem_vec_p,
                                       &hob_p->fw_phys_mem_vec_size);
        if (ret != HH_OK) {
            MTL_ERROR1(MT_FLFMT("%s: XHH_phys_mem_alloc failed (%d:%s). Could not allocate FW memory"),
                       __func__,ret, HH_strerror_sym(ret));
            goto fw_mem_alloc_err;

        }
#if 0
	ret1 =  XHH_phys_mem_pattern_initialize(hob_p->fw_phys_mem_vec_p, hob_p->fw_phys_mem_vec_size);
        if (ret1 != HH_OK) {
            MTL_ERROR1(MT_FLFMT("%s: XHH_phys_mem_pattern_initialize failed (%d:%s)"),
                       __func__,ret1, HH_strerror_sym(ret1));
        }
#endif
        pa_vec_size = hob_p->fw_phys_mem_vec_size; 
        pa_vec = TNMALLOC(XHH_cmd_pa_vec_t, pa_vec_size);
        if (pa_vec == NULL) {
            ret = HH_ENOMEM;
            goto alloc_pa_vec_err;
        }
        XHH_build_pa_vec(hob_p->fw_phys_mem_vec_p, pa_vec_size, pa_vec); 
        hob_p->profile.lam_size = 0;
        *used_lam = FALSE;
    }

    /* Load the FW */
    cmd_ret = XHH_cmd_MAP_FA(hob_p->cmd, pa_vec_size, pa_vec);
    if (cmd_ret != XHH_CMD_STAT_OK) {
      if (cmd_ret == XHH_CMD_STAT_EFATAL) {
          MTL_ERROR1(MT_FLFMT("%s: FATAL ERROR in XHH_cmd_MAP_FA"),__func__);
          ret = HH_EFATAL;
      } else {
          MTL_ERROR1(MT_FLFMT("%s: CMD_error in XHH_cmd_MAP_FA (%d)"), __func__,cmd_ret);
          ret = HH_ERR;
      }
      goto map_fa_err;
    }
    
    /* Run the FW */
    cmd_ret = XHH_cmd_RUN_FW(hob_p->cmd);
    if (cmd_ret != XHH_CMD_STAT_OK) {
      if (cmd_ret == XHH_CMD_STAT_EFATAL) {
          MTL_ERROR1(MT_FLFMT("%s: FATAL ERROR in XHH_cmd_RUN_FA"),__func__);
          ret = HH_EFATAL;
      } else {
          MTL_ERROR1(MT_FLFMT("%s: CMD_error in XHH_cmd_RUN_FA (%d)"), __func__,cmd_ret);
          ret = HH_ERR;
      }
      goto run_fa_err;
    }

    FREE(pa_vec);
    return HH_OK; 
    

run_fa_err:
    XHH_cmd_UNMAP_FA(hob_p->cmd);

map_fa_err:
alloc_pa_vec_err:
fw_mem_alloc_err:
    if (hob_p->have_lam == FALSE) {
        XHH_phys_mem_free(hob_p->fw_phys_mem_vec_p, hob_p->fw_phys_mem_vec_size);
        hob_p->fw_phys_mem_vec_p = NULL;
        hob_p->fw_phys_mem_vec_size = 0;
    }
    if (pa_vec != NULL ) {FREE(pa_vec);}

    return ret;
}

static void map_fw_errbuf(XHH_hob_t hob_p)
{
    call_result_t rc;
    if ((hob_p->fw_props.error_buf_start != (u_int64_t) 0) &&
        (hob_p->fw_props.error_buf_size != 0)) 
    {
        MTL_DEBUG4(MT_FLFMT("XHH_hob_open_hca: using cat err buf. pa=0x"U64_FMT", sz=%d"),
                   hob_p->fw_props.error_buf_start, hob_p->fw_props.error_buf_size);
        rc = MOSAL_iomem_map(MOSAL_get_kernel_prot_ctx(),
					    MOSAL_MEM_FLAGS_NO_CACHE | MOSAL_MEM_FLAGS_PERM_READ,
					    (MT_phys_addr_t) hob_p->fw_props.error_buf_start, 
                                            (MT_size_t) 4*(hob_p->fw_props.error_buf_size),
				             &(hob_p->fw_error_iomem_obj),
				             &(hob_p->fw_error_buf_start_va));
        if (rc != MT_OK) {
            MTL_ERROR1(MT_FLFMT("%s: Could not map fw error buffer (phys addr = "U64_FMT", size=%d) (%d:%s)"),
                        __func__, hob_p->fw_props.error_buf_start, hob_p->fw_props.error_buf_size,rc,mtl_strerror_sym(rc));
	   hob_p->fw_error_iomem_obj = (MOSAL_iomem_t) NULL;
           hob_p->fw_error_buf_start_va =  (MT_virt_addr_t)(MT_ulong_ptr_t) NULL;
        } 
    }
}

static HH_ret_t initialize_icm_and_aux(XHH_hob_t hob_p)
{
    XHH_cmd_pa_vec_t   *pa_vec = NULL;
    u_int32_t  pa_vec_size;
    MT_size_t  icm_aux_size;
    HH_ret_t   ret = HH_ERR;
    XHH_cmd_status_t  cmd_ret;
    MT_phys_addr_t lamm_buf_addr;


    /* first, set ICM size and get needed aux size */
    cmd_ret = XHH_cmd_SET_ICM_SIZE(hob_p->cmd, hob_p->icm_size, &hob_p->icm_aux_size);
    if (cmd_ret != XHH_CMD_STAT_OK) {
      if (cmd_ret == XHH_CMD_STAT_EFATAL) {
          MTL_ERROR1(MT_FLFMT("%s: FATAL ERROR in XHH_cmd_SET_ICM_SIZE"),__func__);
          ret = HH_EFATAL;
      } else {
          MTL_ERROR1(MT_FLFMT("%s: CMD_error in XHH_cmd_SET_ICM_SIZE (%d)"), __func__,cmd_ret);
          ret = HH_ERR;
      }
      return ret;
    }

    if (hob_p->have_lam) {
        icm_aux_size = (MT_size_t)(hob_p->icm_aux_size << MOSAL_SYS_PAGE_SHIFT);
        ret =   XHH_lamm_alloc(hob_p->lamm, icm_aux_size,MOSAL_SYS_PAGE_SHIFT,&lamm_buf_addr);
        if (ret != HH_OK) {
            MTL_ERROR1(MT_FLFMT("%s: XHH_lamm_alloc failed (%d:%s). Could not allocate LAMM memory for ICM_AUX"),
                       __func__,ret, HH_strerror_sym(ret));
            return ret;

        }
        
        ret = XHH_build_lamm_pa_vec(lamm_buf_addr,(u_int64_t)icm_aux_size,&pa_vec_size, &pa_vec);
        if (ret != HH_OK) {
            MTL_ERROR1(MT_FLFMT("%s: XHH_build_lamm_pa_vec failed (%d:%s) for allocating ICM_AUX memory"),
                       __func__,ret, HH_strerror_sym(ret));
            return ret;

        }

        if (hob_p->lam_props.dh == TRUE) {
            MTL_ERROR1("Device %s is operating in HIDE_LAM mode.\n", hob_p->dev_name);
        }
    } else {
        /* using HOST memory for FW -- either no LAM, or LAM has a problem.
         * Need to save the physical address for freeing during close_hca
         */

        ret = XHH_phys_mem_alloc((hob_p->icm_aux_size) << MOSAL_SYS_PAGE_SHIFT,
                                 MOSAL_SYS_PAGE_SHIFT,
                                 MOSAL_MEM_ALLOC_FLAGS_NONE,
                                 XHH_MEM_ALLOC_FLAGS_NONE,
                                 &hob_p->icm_aux_phys_mem_vec_p,
                                 &hob_p->icm_aux_phys_mem_vec_size);
        if (ret != HH_OK) {
            MTL_ERROR1(MT_FLFMT("%s: XHH_phys_mem_alloc failed (%d:%s). Could not allocate AUX_ICM  memory"),
                       __func__,ret, HH_strerror_sym(ret));
            return ret;

        }
        pa_vec_size = hob_p->icm_aux_phys_mem_vec_size; 
        pa_vec = TNMALLOC(XHH_cmd_pa_vec_t, pa_vec_size);
        if (pa_vec == NULL) {
            ret = HH_ENOMEM;
            goto alloc_pa_vec_err;
        }
        XHH_build_pa_vec(hob_p->icm_aux_phys_mem_vec_p, pa_vec_size, pa_vec); 
        hob_p->profile.lam_size = 0;
    }

    /* Map the auxiliary area */
    cmd_ret = XHH_cmd_MAP_ICM_AUX(hob_p->cmd, pa_vec_size, pa_vec);
    if (cmd_ret != XHH_CMD_STAT_OK) {
      if (cmd_ret == XHH_CMD_STAT_EFATAL) {
          MTL_ERROR1(MT_FLFMT("%s: FATAL ERROR in XHH_cmd_MAP_ICM_AUX"),__func__);
          ret = HH_EFATAL;
      } else {
          MTL_ERROR1(MT_FLFMT("%s: CMD_error in XHH_cmd_MAP_ICM_AUX (%d)"), __func__,cmd_ret);
          ret = HH_ERR;
      }
      goto  icm_aux_mem_alloc_err;
    }
    

    FREE(pa_vec);
    return HH_OK; 

alloc_pa_vec_err:
icm_aux_mem_alloc_err:
    if (hob_p->have_lam == FALSE) {
        XHH_phys_mem_free(hob_p->icm_aux_phys_mem_vec_p, hob_p->icm_aux_phys_mem_vec_size);
        hob_p->icm_aux_phys_mem_vec_p = NULL;
        hob_p->icm_aux_phys_mem_vec_size = 0;
    }
    if (pa_vec != NULL ) {FREE(pa_vec);}

    return ret;
}

static HH_ret_t XHH_hob_cqm_init_create(XHH_hob_t rhh_hob_p)
{
    XHH_cqm_init_t         *rhh_cqm_init_params_p = NULL;
    HH_ret_t               rc;
    
    rhh_cqm_init_params_p = (XHH_cqm_init_t *)MALLOC(sizeof(XHH_cqm_init_t));
    if (rhh_cqm_init_params_p == NULL) {
       MT_COV_MEM_FLTR_START
       MTL_ERROR1(MT_FLFMT("MALLOC rhh_cqm_init_params_p failed"));
       return HH_EAGAIN;
       MT_COV_MEM_FLTR_END
    }
    memset(rhh_cqm_init_params_p, 0, sizeof(XHH_cqm_init_t));

    /* create eventp*/
    rhh_cqm_init_params_p->cqc_tbl_base = rhh_hob_p->icm_alloc_base_addrs_vec.cqc_base_addr;
    rhh_cqm_init_params_p->log2_cqc_entry_sz = (u_int8_t)ceil_log2(rhh_hob_p->dev_lims.cqc_entry_sz);
    rhh_cqm_init_params_p->log2_max_cq =  (u_int8_t) rhh_hob_p->profile.log2_max_cqs;
    rhh_cqm_init_params_p->log2_rsvd_cqs = rhh_hob_p->dev_lims.log2_rsvd_cqs;

    rc = XHH_cqm_create(rhh_hob_p, rhh_cqm_init_params_p, &(rhh_hob_p->cqm));
    FREE(rhh_cqm_init_params_p);
    return rc;

}

static HH_ret_t XHH_hob_mcgm_init_create(XHH_hob_t rhh_hob_p)
{
    XHH_mcgm_init_t        *rhh_mcgm_init_params_p = NULL;
    HH_ret_t               rc;
    
    rhh_mcgm_init_params_p = (XHH_mcgm_init_t *)MALLOC(sizeof(XHH_mcgm_init_t));
    if (rhh_mcgm_init_params_p == NULL) {
       MT_COV_MEM_FLTR_START
       MTL_ERROR1(MT_FLFMT("MALLOC rhh_mcgm_init_params_p failed"));
       return HH_EAGAIN;
       MT_COV_MEM_FLTR_END
    }
    memset(rhh_mcgm_init_params_p, 0, sizeof(XHH_mcgm_init_t));

    /* create mcgm */
    rhh_mcgm_init_params_p->log2_total_entries = rhh_hob_p->hca_props.multicast_parameters.log_mc_table_sz;
    rhh_mcgm_init_params_p->mght_hash_bins = (1UL << (rhh_hob_p->profile.log2_mcg_hash_size));
    rhh_mcgm_init_params_p->max_qp_per_mcg = (u_int16_t)rhh_hob_p->profile.qps_per_mcg;
    rhh_mcgm_init_params_p->mcg_tbl_base = rhh_hob_p->icm_alloc_base_addrs_vec.mcg_base_addr;
    rhh_mcgm_init_params_p->log2_mcg_entry_sz = (u_int8_t)rhh_hob_p->profile.log2_mcg_entry_size;
    
    rc= XHH_mcgm_create(rhh_hob_p, rhh_mcgm_init_params_p, &(rhh_hob_p->mcgm) );
    
    FREE(rhh_mcgm_init_params_p);
    return rc;

}

static HH_ret_t XHH_hob_uarm_init_create(XHH_hob_t rhh_hob_p)
{
    XHH_uarm_init_t        *rhh_uarm_init_params_p = NULL;
    HH_ret_t               rc;
    
    rhh_uarm_init_params_p = (XHH_uarm_init_t *)MALLOC(sizeof(XHH_uarm_init_t));
    if (rhh_uarm_init_params_p == NULL) {
       MT_COV_MEM_FLTR_START
       MTL_ERROR1(MT_FLFMT("MALLOC rhh_uarm_init_params_p failed"));
       return HH_EAGAIN;
       MT_COV_MEM_FLTR_END
    }
    memset(rhh_uarm_init_params_p, 0, sizeof(XHH_uarm_init_t));

    /* create uarm */
    rhh_uarm_init_params_p->uar_base =  rhh_hob_p->hw_props.uar_base;
    rhh_uarm_init_params_p->log2_max_uar = (u_int8_t) rhh_hob_p->profile.log2_max_uar;
    rhh_uarm_init_params_p->log2_uar_pg_sz = (u_int8_t) rhh_hob_p->profile.log2_uar_pg_size;
    rhh_uarm_init_params_p->uar_ctx_base = rhh_hob_p->icm_alloc_base_addrs_vec.uarc_base_addr;
    rhh_uarm_init_params_p->log2_uar_ctx_sz = (u_int8_t)rhh_hob_p->icm_alloc_size_vec.log2_uarc_size;
    rhh_uarm_init_params_p->num_rsvd_uars = rhh_hob_p->dev_lims.num_rsvd_uars;

    
    rc= XHH_uarm_create(rhh_hob_p,rhh_uarm_init_params_p,&(rhh_hob_p->uarm));
    
    FREE(rhh_uarm_init_params_p);
    return rc;

}

static HH_ret_t XHH_hob_srqm_init_create(XHH_hob_t rhh_hob_p)
{
    XHH_srqm_init_t        *rhh_srqm_init_params_p = NULL;
    HH_ret_t               rc;
    
    rhh_srqm_init_params_p = (XHH_srqm_init_t *)MALLOC(sizeof(XHH_srqm_init_t));
    if (rhh_srqm_init_params_p == NULL) {
       MT_COV_MEM_FLTR_START
       MTL_ERROR1(MT_FLFMT("MALLOC rhh_srqm_init_params_p failed"));
       return HH_EAGAIN;
       MT_COV_MEM_FLTR_END
    }
    memset(rhh_srqm_init_params_p, 0, sizeof(XHH_srqm_init_t));

    /* create srqm*/
    rhh_srqm_init_params_p->log2_max_srq = (u_int8_t)rhh_hob_p->profile.log2_max_srqs;
    rhh_srqm_init_params_p->log2_rsvd_srq = rhh_hob_p->dev_lims.log2_rsvd_srqs;
    rhh_srqm_init_params_p->log2_srqc_entry_sz = (u_int8_t)ceil_log2(rhh_hob_p->dev_lims.srq_entry_sz);
    rhh_srqm_init_params_p->srqc_tbl_base = rhh_hob_p->icm_alloc_base_addrs_vec.srqc_base_addr;
    rhh_srqm_init_params_p->ib_eq = rhh_hob_p->ib_eq;
    
    rc= XHH_srqm_create(rhh_hob_p, rhh_srqm_init_params_p, &(rhh_hob_p->srqm));
    
    FREE(rhh_srqm_init_params_p);
    return rc;

}

static HH_ret_t XHH_hob_qpm_init_create(XHH_hob_t rhh_hob_p)
{
    XHH_qpm_init_t         *rhh_qpm_init_params_p = NULL;
    HH_ret_t               rc;
    
    rhh_qpm_init_params_p = (XHH_qpm_init_t *)MALLOC(sizeof(XHH_qpm_init_t));
    if (rhh_qpm_init_params_p == NULL) {
       MT_COV_MEM_FLTR_START
       MTL_ERROR1(MT_FLFMT("MALLOC rhh_qpm_init_params_p failed"));
       return HH_EAGAIN;
       MT_COV_MEM_FLTR_END
    }
    memset(rhh_qpm_init_params_p, 0, sizeof(XHH_qpm_init_t));

    rhh_qpm_init_params_p->rdb_base_index = 
        /* 32 low-order bits, right-shifted by log of size of rdb entry */
      (u_int32_t)(((u_int64_t)(rhh_hob_p->icm_alloc_base_addrs_vec.rdb_base_addr)) & (u_int64_t)0xFFFFFFFF ) 
                                  >> XHH_ICM_LOG2_RDB_ENTRY_SIZE;
    rhh_qpm_init_params_p->log2_max_qp = (u_int8_t) rhh_hob_p->profile.log2_max_qps;
    rhh_qpm_init_params_p->log2_max_outs_rdma_atom = rhh_hob_p->profile.log2_inflight_rdma_per_qp;
    rhh_qpm_init_params_p->log2_max_outs_dst_rd_atom = rhh_hob_p->dev_lims.log_max_ra_req_qp;
    rhh_qpm_init_params_p->n_ports = (u_int8_t)rhh_hob_p->dev_lims.num_ports;
    rhh_qpm_init_params_p->port_props = 
        (rhh_hob_p->module_flags.legacy_sqp == TRUE ? NULL : rhh_hob_p->init_ib_props );
    rhh_qpm_init_params_p->log2_rsvd_qps = rhh_hob_p->dev_lims.log2_rsvd_qps;
    rhh_qpm_init_params_p->log2_qpc_entry_sz = (u_int8_t)ceil_log2(rhh_hob_p->dev_lims.qpc_entry_sz);
    rhh_qpm_init_params_p->qpc_tbl_base = rhh_hob_p->icm_alloc_base_addrs_vec.qpc_base_addr;
    rhh_qpm_init_params_p->log2_eqpc_entry_sz = (u_int8_t)ceil_log2(rhh_hob_p->dev_lims.eqpc_entry_sz);
    rhh_qpm_init_params_p->eqpc_tbl_base = rhh_hob_p->icm_alloc_base_addrs_vec.eqpc_base_addr;
	rhh_qpm_init_params_p->rdb_tbl_base =  rhh_hob_p->icm_alloc_base_addrs_vec.rdb_base_addr;
    rhh_qpm_init_params_p->log2_rdb_entry_sz = rhh_qpm_init_params_p->log2_max_outs_rdma_atom +
                                               XHH_ICM_LOG2_RDB_ENTRY_SIZE;
    rhh_qpm_init_params_p->log2_rsvd_rdbs = rhh_hob_p->dev_lims.log2_rsvd_rdbs;

    rc = XHH_qpm_create(rhh_hob_p, rhh_qpm_init_params_p, &(rhh_hob_p->qpm));
    FREE(rhh_qpm_init_params_p);
    return rc;

}

static HH_ret_t XHH_hob_eventp_init_create(XHH_hob_t rhh_hob_p)
{
    XHH_eventp_res_t       *event_res_p = NULL;
    HH_ret_t               rc;
    
    event_res_p = (XHH_eventp_res_t *)MALLOC(sizeof(XHH_eventp_res_t));
    if (event_res_p == NULL) {
       MT_COV_MEM_FLTR_START
       MTL_ERROR1(MT_FLFMT("MALLOC event_res_p failed"));
       return HH_EAGAIN;
       MT_COV_MEM_FLTR_END
    }
    memset(event_res_p, 0, sizeof(XHH_eventp_res_t));

    /* create eventp*/
    event_res_p->clr_int_base_addr = rhh_hob_p->fw_props.clr_int_base_addr; 
    event_res_p->eq_arm_base_addr = rhh_hob_p->fw_props.eq_arm_base_addr;
    event_res_p->eq_set_ci_base_addr = rhh_hob_p->fw_props.eq_set_ci_base_addr;
    event_res_p->intr_clr_bit = rhh_hob_p->adapter_props.intapin;
    event_res_p->intr_props = &rhh_hob_p->hw_props.interrupt_props;
    event_res_p->is_srq_enable = rhh_hob_p->dev_lims.srq;
    strcpy(event_res_p->name, rhh_hob_p->dev_name);

    rc = XHH_eventp_create (rhh_hob_p, event_res_p, &(rhh_hob_p->eventp));
    FREE(event_res_p);
    return rc;

}
/******************************************************************************
 *  Function:     XHH_hob_open_hca
 *
 *  Description:  This function opens the given HCA and initializes the HCA with 
 *                given properties/ capabilities.  if prop_props_p is NULL a default 
 *                HCA profile will be set up.
 *                    1.Create the XHH_cmd_if object instance. 
 *                    2.QUERY_FW to determince supported features and size of firmware area. 
 *                         Update XHH_cmdif based on reported capabilities.                   
 *                    3.QUERY_LAM to determine LAM availability and its size. 
 *                    4.If LAM available: 
 *                           LAM_EN to allow loading of FW and ICM in the LAM. 
 *                    5.Allocate  firmware area: 
 *                        -If LAM available,locate the firmware area in a 
 *                         physically contiguous chunk on top of the LAM. 
 *                        - Otherwise, Allocate enough physical pages and use MAP_FA to map them. 
 *                    6.Compute HCA profile (See section 3.1.2.2 on page 8.): 
 *                        -Use QUERY_DEV_LIM to receive static configuration limitations and 
 *                         size of resources context. 
 *                        -(optionally)use MOD_STAT_CFG to update static limitations. 
 *                    7.Create XHH_icmm based on required size by the profile. 
 *                        If LAM is available,provide XHH_icmm the physical base address of the 
 *                        LAM,and subtract the FW area size from total size. 
 *                    8.Allocate resources in the ICM based on computed profile. 
 *                    9.Invoke INIT_HCA. 
 *                   10.Create RHH resources management objects. 
 *                   11.Create XHH_eq_doorbell. 
 *                   12.Create XHH_eventp. 
 *                   13.Set up the commands event queue.
 *
 *  input:
 *                hca_hndl
 *                prop_props_p - Proprietary properties (Non IB)
 *  output: 
 *                none
 *  returns:
 *                HH_OK
 *                HH_EINVAL
 *                HH_EBUSY
 *
 *  Comments:     If HCA is still open,XHH_hob_close_hca() is 
 *                invoked before freeing the XHH_hob.
 *
 *                Returns HH_EINVAL_HCA_HNDL if any function called internally fails  
 *
 *****************************************************************************/
HH_ret_t XHH_hob_open_hca(HH_hca_hndl_t  hca_hndl, 
                                 EVAPI_hca_profile_t  *prop_props_p,
                                 EVAPI_hca_profile_t  *sugg_profile_p)
{
    HH_ret_t               ret;
    XHH_cmd_status_t       cmd_ret;
    u_int32_t              i;
    u_int16_t              num_ports, last_port_initialized;
    XHH_hob_t              rhh_hob_p;
    MT_bool                have_fatal = FALSE;
    MT_phys_addr_t         cmdif_uar0_arg;
    XHH_icmm_init_t        rhh_icmm_init_params;
    XHH_pdm_init_t         rhh_pdm_init_params;
    MT_size_t              tmp_log2_uarc_size;
    MT_bool                used_lam_for_fw = TRUE;
    call_result_t	   mosal_ret;

    MT_RETURN_IF_LOW_STACK(1500);
    MTL_DEBUG4("Entering XHH_hob_open_hca\n");
    if (MOSAL_get_exec_ctx() != MOSAL_IN_TASK) {
        MTL_ERROR1("XHH_hob_open_hca: NOT IN TASK CONTEXT)\n");
        return HH_ERR;
    }


    if (hca_hndl == NULL) {
        MTL_ERROR1("XHH_hob_open_hca : ERROR : Invalid HCA handle\n");
        return HH_EINVAL_HCA_HNDL;
    }
    rhh_hob_p = XHHOBP(hca_hndl);

    if (rhh_hob_p == NULL) {
        MTL_ERROR1("XHH_hob_open_hca: ERROR : No device registered\n");
        return HH_EAGAIN;
    }

    /* Test if have fatal error, and move xhh_state to "opening" */
    MOSAL_spinlock_dpc_lock(&rhh_hob_p->fatal_spl);
    if ((rhh_hob_p->xhh_state & XHH_STATE_HAVE_ANY_FATAL) != 0) {
        /* already in FATAL state */
        MOSAL_spinlock_unlock(&rhh_hob_p->fatal_spl);
        MTL_DEBUG4(MT_FLFMT("%s: already in FATAL state"), __func__);  
        MT_RETURN(HH_EFATAL);
    } else if (rhh_hob_p->xhh_state != XHH_STATE_CLOSED) {
        MOSAL_spinlock_unlock(&rhh_hob_p->fatal_spl);
        MTL_ERROR1(MT_FLFMT("XHH_hob_open_hca: ERROR : Device not closed. state = 0x%x"),rhh_hob_p->xhh_state );
        MT_RETURN(HH_EBUSY);
    }
    rhh_hob_p->xhh_state = XHH_STATE_OPENING;
    MOSAL_spinlock_unlock(&rhh_hob_p->fatal_spl);
    
    /* get the MUTEX */
    if (MOSAL_mutex_acq(&(rhh_hob_p->mtx), TRUE) != MT_OK) {
        MTL_ERROR1(MT_FLFMT("XHH_hob_open_hca: received signal. returning"));
        ret = HH_EINTR;
        goto post_state_change_error;
    }
    if (hca_hndl->status == HH_HCA_STATUS_OPENED) {
        MTL_ERROR1("XHH_hob_open_hca: ERROR : Device already open\n");
        ret = HH_EBUSY;
        goto post_mutex_acquire_err;
    }

    /* reset the device. The "FATAL_HANDLING" macros here are mis-named, and should
     * be just "MT_START_DEVICE_RESET/END_DEVICE_RESET"
     */
    {
	 MT_START_FATAL_HANDLING;
	  /* Do the device RESET */
	 XHH_hob_reset_device(&(rhh_hob_p->hw_props));

	  /* now, rewrite the PCI configuration */
	 XHH_write_pci_config(rhh_hob_p->hw_props.pci_dev,&rhh_hob_p->pci_hca_info.config);

///	 XHH_set_capabilities(&(rhh_hob_p->hw_props));

	 MT_END_FATAL_HANDLING;
    }

    /* create the XHH_cmd object so that can initialize and query the adapter */
    /* HCR register offset */

    if ( rhh_hob_p->module_flags.cmdif_post_uar0) {
      cmdif_uar0_arg = rhh_hob_p->hw_props.uar_base;
    }
    else {
      cmdif_uar0_arg = (MT_phys_addr_t) MAKE_ULONGLONG(0xFFFFFFFFFFFFFFFF);
    }
    ret = XHH_cmd_create(rhh_hob_p, rhh_hob_p->hw_props.hw_ver, rhh_hob_p->hw_props.cr_base, 
                         cmdif_uar0_arg, &(rhh_hob_p->cmd),
                         rhh_hob_p->module_flags.inifinite_cmd_timeout, rhh_hob_p->module_flags.num_cmds_outs);
    if (ret != HH_OK) {
      MTL_ERROR1("XHH_hob_open_hca: could not create CMD object (%d)\n", ret);
      ret = HH_ERR;
      goto cmd_create_err;
    }

    /* do query firmware command */
	memset(&(rhh_hob_p->fw_props), 0, sizeof(rhh_hob_p->fw_props));
    cmd_ret = XHH_cmd_QUERY_FW(rhh_hob_p->cmd, &(rhh_hob_p->fw_props));

    if (cmd_ret != XHH_CMD_STAT_OK) {
      if (cmd_ret == XHH_CMD_STAT_EFATAL) {
          MTL_ERROR1(MT_FLFMT("XHH_hob_open_hca: FATAL ERROR in XHH_cmd_QUERY_FW"));
          ret = HH_EFATAL;
      } else {
          MTL_ERROR1(MT_FLFMT("XHH_hob_open_hca: CMD_error in XHH_cmd_QUERY_FW (%d)"), cmd_ret);
          ret = HH_ERR;
      }
      goto post_cmd_create_err;
    }

    if (version_info_set_and_check(rhh_hob_p) != HH_OK) {
        goto post_cmd_create_err;
    }

    /* map the firmware error buffer for access by driver, and prepare a buffer to copy over the error
     * info at catastrophic error time*/
    map_fw_errbuf(rhh_hob_p);
    if (rhh_hob_p->fw_error_buf_start_va != (MT_virt_addr_t) NULL) {
	rhh_hob_p->fw_error_buf = TNMALLOC(u_int32_t,rhh_hob_p->fw_props.error_buf_size);
	if (rhh_hob_p->fw_error_buf == NULL) {
	    MTL_ERROR1(MT_FLFMT("%s: Could not allocate buffer for FW catastrophic error info"),__func__);
	}
    }

    /* launch catastrophic error thread */
    rhh_hob_p->fatal_thread_obj.hob = (struct XHH_hob_st *)rhh_hob_p;
    MOSAL_syncobj_init(&rhh_hob_p->fatal_thread_obj.start_sync);
    MOSAL_syncobj_init(&rhh_hob_p->fatal_thread_obj.stop_sync);
    MOSAL_syncobj_init(&rhh_hob_p->fatal_thread_obj.fatal_err_sync);
    MOSAL_syncobj_init(&rhh_hob_p->xhh_fatal_complete_syncobj);
    rhh_hob_p->fatal_thread_obj.have_fatal = FALSE;
    mosal_ret = MOSAL_thread_start(&rhh_hob_p->fatal_thread_obj.mto, MOSAL_KTHREAD_CLONE_FLAGS,
				   XHH_hob_fatal_err_thread, (void *)(&(rhh_hob_p->fatal_thread_obj)));
  //  if (mosal_ret != MT_OK) {
  //      MTL_ERROR1("XHH_hob_create: could not create fatal error thread (err = %d)\n", mosal_ret);
  //      ret = HH_ERR;
  //      goto fatal_thr_create_err;
  //  }

    /*wait for fatal thread initialization complete */
    mosal_ret = MOSAL_syncobj_waiton(&(rhh_hob_p->fatal_thread_obj.start_sync), 10000000);
    if (mosal_ret != MT_OK) {
	if (mosal_ret == MT_EINTR) {
	    MTL_DEBUG1(MT_FLFMT("%s: Received OS interrupt while initializing fatal error thread (err = %d)"), 
		       __func__,mosal_ret);
	    ret = HH_EINTR;
	} else {
	    MTL_ERROR1(MT_FLFMT("%s: Timeout on initializing fatal error thread (err = %d)"), 
		       __func__,mosal_ret);
	    ret = HH_ERR;
	}
	goto fatal_thr_init_err;
    }
    MTL_DEBUG4("%s: Created error-check thread.\n", __func__);
    /* enable the LAM, and specify area for FW */
    cmd_ret = XHH_cmd_ENABLE_LAM(rhh_hob_p->cmd,&(rhh_hob_p->lam_props));
    if (cmd_ret != XHH_CMD_STAT_OK) {
      if (cmd_ret == XHH_CMD_STAT_LAM_NOT_PRE) {
          MTL_DEBUG1(MT_FLFMT("%s: XHH_cmd_LAM_EN indicates that there is no LAM."),__func__);
          rhh_hob_p->have_lam = FALSE;
          memset(&(rhh_hob_p->lam_props), 0, sizeof(rhh_hob_p->lam_props));
      } else {
          MTL_ERROR1(MT_FLFMT("%s: CMD_error in XHH_cmd_LAM_EN (%d)"), __func__,cmd_ret);
          ret = HH_ERR;
          goto enable_lam_err;
      }
    } else {
        rhh_hob_p->have_lam = TRUE;
    }

    /* INITIALIZE FW */
    ret = initialize_fw(rhh_hob_p, &used_lam_for_fw);
    if (ret != HH_OK) {
      MTL_ERROR1(MT_FLFMT("%s: could not create initialize FW (%d:%s)"),__func__, ret,HH_strerror_sym(ret));
      ret = HH_ERR;
      goto init_fw_err;
    }

    if (rhh_hob_p->have_lam == TRUE) {
        ret = XHH_lamm_create(rhh_hob_p->lam_props.lam_start_adr, 
                             (MT_size_t) (rhh_hob_p->lam_props.lam_end_adr - rhh_hob_p->lam_props.lam_start_adr + 1), 
                             &(rhh_hob_p->lamm));
        if (ret != HH_OK) {
            MTL_ERROR1(MT_FLFMT("%s: could not create lam manager (%d:%s)"),
                       __func__, ret, HH_strerror_sym(ret));
            rhh_hob_p->have_lam = FALSE;
        } else {
            ret = XHH_lamm_reserve(rhh_hob_p->lamm, rhh_hob_p->fw_phys_start_addr,
                        (MT_size_t) ((MT_size_t)(rhh_hob_p->fw_props.fw_pages) << XHH_ARBEL_LOG_PAGE_SIZE));
            if (ret != HH_OK) {
                MTL_ERROR1(MT_FLFMT("%s: could not reserve firmware area in LAM manager (%d:%s)"),
                           __func__, ret, HH_strerror_sym(ret));
                XHH_lamm_destroy(rhh_hob_p->lamm);
                rhh_hob_p->lamm = (XHH_lamm_t) XHH_INVALID_HNDL;
                rhh_hob_p->have_lam = FALSE;
            }
        }
    }

    /* Get device limits */
    cmd_ret = XHH_cmd_QUERY_DEV_LIM(rhh_hob_p->cmd, &(rhh_hob_p->dev_lims));
    if (cmd_ret != XHH_CMD_STAT_OK) {
      if (cmd_ret == XHH_CMD_STAT_EFATAL) {
          MTL_ERROR1(MT_FLFMT("%s: FATAL ERROR in XHH_cmd_QUERY_DEV_LIM"),__func__);
          ret = HH_EFATAL;
      } else {
          MTL_ERROR1(MT_FLFMT("%s: CMD_error in XHH_cmd_QUERY_DEV_LIM (%d)"), __func__,cmd_ret);
          ret = HH_ERR;
      }
      goto query_devlim_err;
    }

#if 0
    MTL_ERROR1(MT_FLFMT("%s: log_max_srq=%u  log2_rsvd_srqs=%u  srq_entry_sz=%u  srq=%ssupported"), __func__,
               rhh_hob_p->dev_lims.log_max_srqs, rhh_hob_p->dev_lims.log2_rsvd_srqs, 
               rhh_hob_p->dev_lims.srq_entry_sz, rhh_hob_p->dev_lims.srq ? " ":"NOT-");
#endif
#ifdef ADJUST_DEVLIMS_MAX_QP_PER_MCG
     rhh_hob_p->dev_lims.log_max_qp_mcg = (rhh_hob_p->dev_lims.log_max_qp_mcg > 3) ? 
	 3 : rhh_hob_p->dev_lims.log_max_qp_mcg;
#endif

    /* query arbel for adapter data */
    cmd_ret = XHH_cmd_QUERY_ADAPTER(rhh_hob_p->cmd, &(rhh_hob_p->adapter_props));
    if (cmd_ret != XHH_CMD_STAT_OK) {
      if (cmd_ret == XHH_CMD_STAT_EFATAL) {
          MTL_ERROR1(MT_FLFMT("%s: FATAL ERROR in XHH_cmd_QUERY_ADAPTER"),__func__);
          ret = HH_EFATAL;
      } else {
          MTL_ERROR1(MT_FLFMT("%s: CMD_error in XHH_cmd_QUERY_ADAPTER (%d)"),__func__, cmd_ret);
          ret = HH_ERR;
      }
      goto query_adapter_err;
    }

    
    if (prop_props_p != NULL) {
       XHH_PRINT_USR_PROFILE(prop_props_p);
    }
    ret = XHH_calculate_profile(rhh_hob_p, prop_props_p, sugg_profile_p);
    if (ret != HH_OK) {
      MTL_ERROR1(MT_FLFMT("%s: could not create internal profile (%d)"), __func__, ret);
      if (sugg_profile_p != NULL) {
          XHH_PRINT_USR_PROFILE(sugg_profile_p);
      }
      /* ret = HH_ERR; */
      goto profile_err;
    }
    if (sugg_profile_p != NULL) {
        XHH_PRINT_USR_PROFILE(sugg_profile_p);
    }

    /* check profile against QUERY_DEV_LIMS data*/
    ret = XHH_check_profile(rhh_hob_p);
    if (ret != HH_OK) {
      MTL_ERROR1("%s: Profile check failed (%d)\n", __func__, ret);
      /* ret = HH_ERR;*/
      goto profile_err;
    }

  /* Calculate required ICM size */
    calculate_icm_alloc_vec(rhh_hob_p, &(rhh_hob_p->profile), FALSE,
                            /* NULL*/ &tmp_log2_uarc_size,&(rhh_hob_p->icm_alloc_size_vec));
    rhh_hob_p->icm_size = calculate_total_icm_size(&(rhh_hob_p->icm_alloc_size_vec));
    
    ret = initialize_icm_and_aux(rhh_hob_p);
    if (ret != HH_OK) {
      MTL_ERROR1("%s: initialize_icm_and_aux failed (%d)\n", __func__, ret);
      /*ret = HH_ERR;*/
      goto profile_err;
    }

    rhh_icmm_init_params.icm_byte_sz = rhh_hob_p->icm_size; 
  
    ret = XHH_icmm_create(rhh_hob_p, &(rhh_icmm_init_params),&(rhh_hob_p->icmm));
    if (ret != HH_OK) {
       MTL_ERROR1("XHH_hob_create: could not create ICMM object (%d)\n", ret);
       goto icmm_create_err;
    }

    /*  Do icmm allocation here.  Note that we are only allocating areas of ICM
     *  virtual address space here.  NO physical-memory mapping is done as yet;
     *  Physical mapping is done ONLY by the various resource managers. */

    /* Allocate all required ICM areas */
    ret = XHH_icmm_alloc_sz_aligned(rhh_hob_p->icmm, 
                             rhh_hob_p->profile.icm_alloc_vec_size,      /*number of chunks */
                             (MT_size_t *) &(rhh_hob_p->icm_alloc_size_vec), /* IN  */
                             (XHH_icm_va_t *)&(rhh_hob_p->icm_alloc_base_addrs_vec) );  /* OUT */
    if (ret != HH_OK) {
        MTL_ERROR1("%s: could not allocate required areas in ICM (%s)\n", __func__, HH_strerror_sym(ret));
        ret = HH_ERR;
        goto alloc_size_aligned_err;
    }
    /* JPM -- THE ALLOCATED ICM MEMORY WILL BE MAPPED BY THE VARIOUS MANAGERS */
    
    /* set up parameters for INIT HCA */
    setup_hca_init_props(rhh_hob_p,&(rhh_hob_p->hca_props));

    ret= map_reserved_fw_resources(rhh_hob_p);
    if (ret != HH_OK) {
        MTL_ERROR1(MT_FLFMT("problems mapping fw reserved resources in ICM \n"));
        goto map_fw_rscs_err;
    }
    print_icm_address_vec(&rhh_hob_p->icm_alloc_base_addrs_vec);
    /* INIT_HCA command */
    cmd_ret = XHH_cmd_INIT_HCA(rhh_hob_p->cmd,&(rhh_hob_p->hca_props));
    if (cmd_ret != XHH_CMD_STAT_OK) {
        MTL_ERROR1("%s: CMD_error in XHH_cmd_INIT_HCA (%d)\n", __func__, cmd_ret);
        //MOSAL_delay_execution(2000000000);
        ret = HH_EAGAIN;
        goto init_hca_err;
    }

  /* Now, query HCA to get actual allocated parameters */
    cmd_ret = XHH_cmd_QUERY_HCA(rhh_hob_p->cmd, &(rhh_hob_p->hca_props));
    if (cmd_ret != XHH_CMD_STAT_OK) {
        MTL_ERROR1("%s: CMD_error in XHH_cmd_QUERY_HCA (%d)\n", __func__, cmd_ret);
        ret = HH_EAGAIN;
        goto query_hca_err;
    }
    
    /* create uarm */
   ret = XHH_hob_uarm_init_create(rhh_hob_p);                      
   if (ret != HH_OK) {
       MTL_ERROR1("%s: could not create uarm (%d:%s)\n", __func__, ret,HH_strerror_sym(ret));
        goto uarm_create_err;
    }

    /* create pdm */
    rhh_pdm_init_params.max_pds = (u_int32_t) (rhh_hob_p->profile.max_num_pds + rhh_hob_p->dev_lims.num_rsvd_pds + XHH_NUM_RSVD_PD);
    rhh_pdm_init_params.num_rsvd_pds = rhh_hob_p->dev_lims.num_rsvd_pds;
    
    ret = XHH_pdm_create(rhh_hob_p,
                         &rhh_pdm_init_params,
                         &(rhh_hob_p->pdm));                      
    if (ret != HH_OK) {
        MTL_ERROR1("%s: could not create pdm (%d:%s)\n", __func__, ret,HH_strerror_sym(ret));
        goto pdm_create_err;
    }

    /* create mrwm */
    rhh_hob_p->mrwm_props.mtt_base = rhh_hob_p->icm_alloc_base_addrs_vec.mtt_base_addr;
    rhh_hob_p->mrwm_props.mpt_base = rhh_hob_p->icm_alloc_base_addrs_vec.mpt_base_addr;
    rhh_hob_p->mrwm_props.log2_mpt_sz = (u_int8_t)rhh_hob_p->profile.log2_max_mpt_entries;
    rhh_hob_p->mrwm_props.log2_mtt_sz = (u_int8_t)rhh_hob_p->profile.log2_max_mtt_entries;
    rhh_hob_p->mrwm_props.log2_mtt_seg_sz = (u_int8_t)rhh_hob_p->profile.log2_mtt_entries_per_seg;
    rhh_hob_p->mrwm_props.max_mem_reg = rhh_hob_p->profile.num_external_mem_regions;
    rhh_hob_p->mrwm_props.max_mem_reg_internal = rhh_hob_p->profile.max_num_qps + rhh_hob_p->profile.max_num_cqs;
    rhh_hob_p->mrwm_props.max_mem_win          = rhh_hob_p->profile.num_mem_windows;
    rhh_hob_p->mrwm_props.log2_max_mtt_segs    = (u_int8_t)(rhh_hob_p->profile.log2_max_mtt_entries - 
                                                           rhh_hob_p->mrwm_props.log2_mtt_seg_sz);
    rhh_hob_p->mrwm_props.log2_rsvd_mpts       = rhh_hob_p->dev_lims.log2_rsvd_mrws;
    rhh_hob_p->mrwm_props.log2_rsvd_mtt_segs   = rhh_hob_p->dev_lims.log2_rsvd_mtts - 
                                                    rhh_hob_p->mrwm_props.log2_mtt_seg_sz;
    
    ret = XHH_mrwm_create(rhh_hob_p, &(rhh_hob_p->mrwm_props), &(rhh_hob_p->mrwm));
    if (ret != HH_OK) {
        MTL_ERROR1("%s: could not create mrwm (%d)\n", __func__, ret);
        goto mrwm_create_err;
    }
    
    /* Create cqm object */
    ret = XHH_hob_cqm_init_create(rhh_hob_p);
    if (ret != HH_OK) {
        MTL_ERROR1("%s: could not create cqm (%d)\n", __func__, ret);
        goto cqm_create_err;
    }
    
    /* create qpm object here */

    /* initialize INIT_IB parameters here for possible use by qpm */
    num_ports = rhh_hob_p->dev_lims.num_ports;

    rhh_hob_p->init_ib_props = (XHH_port_init_props_t *) MALLOC(num_ports * sizeof(XHH_port_init_props_t));
    if (!(rhh_hob_p->init_ib_props)) {
        MTL_ERROR1( "%s: ERROR : cannot allocate memory for port init props)\n", __func__);
        goto init_ib_props_malloc_err;
    }
    for (i = 1; i <= num_ports; i++) {
        /* redundant for now. However, leaving option for setting different properties per port */
        rhh_hob_p->init_ib_props[i-1].e = TRUE;
        rhh_hob_p->init_ib_props[i-1].g0 = FALSE;
        rhh_hob_p->init_ib_props[i-1].max_gid = (1 << (rhh_hob_p->dev_lims.log_max_gid));
        rhh_hob_p->init_ib_props[i-1].mtu_cap = rhh_hob_p->dev_lims.max_mtu;
        rhh_hob_p->init_ib_props[i-1].max_pkey = (1 << (rhh_hob_p->dev_lims.log_max_pkey));
        rhh_hob_p->init_ib_props[i-1].vl_cap   = rhh_hob_p->dev_lims.max_vl;
        rhh_hob_p->init_ib_props[i-1].port_width_cap = rhh_hob_p->dev_lims.max_port_width;
    }
    
    ret = XHH_hob_qpm_init_create(rhh_hob_p);
    if (ret != HH_OK) {
        MTL_ERROR1("%s: could not create qpm %s(%d)\n", __func__, HH_strerror_sym(ret), ret);
        goto qpm_create_err;
    }


    //sharon: wrote fixed numbers till fw bug fixed
    if (XHH_DEV_LIM_MCG_ENABLED(rhh_hob_p)) {
        ret = XHH_hob_mcgm_init_create(rhh_hob_p);
        if (ret != HH_OK) {
            MTL_ERROR1("%s: could not create mcgm (%d)\n", __func__, ret);
            rhh_hob_p->hca_capabilities.max_mcast_grp_num = 0;            
            rhh_hob_p->hca_capabilities.max_mcast_qp_attach_num = 0;      
            rhh_hob_p->hca_capabilities.max_total_mcast_qp_attach_num = 0;
            rhh_hob_p->mcgm = (XHH_mcgm_t)XHH_INVALID_HNDL;
	    goto mcgm_create_err;
        }
    }

    /* create eventp*/
    ret = XHH_hob_eventp_init_create(rhh_hob_p);
    if (ret != HH_OK) {
        MTL_ERROR1("%s: could not create eventp (%d)\n", __func__, ret);
        goto eventp_create_err;
    }

    /* CREATE THE VARIOUS EVENT QUEUES (eventp object operations) */
    /* register dummy completion and async event handlers */
	/* In Arbel there is no overflow so we can have small EQs */
    ret = XHH_eventp_setup_comp_eq(rhh_hob_p->eventp, 
                                   &XHH_dummy_comp_event , 
                                   NULL, 510, 
                                   &(rhh_hob_p->compl_eq));
    if (ret != HH_OK) {
        MTL_ERROR1( "%s: ERROR : cannot set up completion event queue (%d)\n", __func__, ret);
        goto eventp_compl_err;
    }


    ret = XHH_eventp_setup_ib_eq(rhh_hob_p->eventp, 
                                 &XHH_dummy_async_event,
                                 NULL, 
                                 (MT_size_t)(rhh_hob_p->module_flags.async_eq_size == 0 ? 
                                             XHH_MAX_ASYNC_EQ_SIZE :
                                             rhh_hob_p->module_flags.async_eq_size ), 
                                 &(rhh_hob_p->ib_eq));
    if (ret != HH_OK) {
        MTL_ERROR1(MT_FLFMT("%s: ERROR : cannot set up async event queue for size "SIZE_T_DFMT" (ret=%d)"), __func__, 
                   (MT_size_t)(rhh_hob_p->module_flags.async_eq_size == 0 ? 
                               XHH_MAX_ASYNC_EQ_SIZE :rhh_hob_p->module_flags.async_eq_size ),ret);
        goto eventp_async_err;
    }

    /* create SRQM after setting up ib_eq event -- srqm needs this event queue number for Create SRQ */
    if (rhh_hob_p->dev_lims.srq) {
      ret= XHH_hob_srqm_init_create(rhh_hob_p);
      if (ret != HH_OK) {
          MTL_ERROR1("%s: could not create srqm - %s(%d)\n", __func__, HH_strerror_sym(ret), ret);
          goto srqm_create_err;
      }
    } else {
      rhh_hob_p->srqm= (XHH_srqm_t)XHH_INVALID_HNDL; /* SRQs are not supported */
    }

    /* PERFORM INIT_IB only for legacy SQPs.  Use max values obtained from QUERY_DEV_LIMS */

    last_port_initialized = 0;
    
	if (rhh_hob_p->module_flags.legacy_sqp == TRUE) { 
	    for (i = 1; i <= num_ports; i++) {
			MTL_TRACE2("XHH_hob_open_hca: INIT_IB COMMAND\n");
			cmd_ret = XHH_cmd_INIT_IB(rhh_hob_p->cmd, (IB_port_t) i, &(rhh_hob_p->init_ib_props[i-1]));
			if (cmd_ret != XHH_CMD_STAT_OK) {
				MTL_ERROR1("XHH_hob_open_hca: CMD_error in XHH_cmd_INIT_IB (%d) for port %d\n", cmd_ret, i);
                if (cmd_ret ==XHH_CMD_STAT_EFATAL) {
                    ret = HH_EFATAL;
                } else {
                    ret = HH_EAGAIN;
                }
				goto init_ib_err;
			} else {
				MTL_TRACE2("XHH_hob_open_hca: INIT_IB COMMAND completed successfuly\n");
			}
			last_port_initialized++;
		}
	}

	/* This must be called after INIT_IB, since it uses the max_pkey value stored in that struct */
    MTL_TRACE2("%s:  Before XHH_hob_query_struct_init\n", __func__);
    ret = XHH_hob_query_struct_init(rhh_hob_p, 
                              ((prop_props_p == NULL)? FALSE : TRUE),
                              &(rhh_hob_p->hca_capabilities));
    if (ret != HH_OK) {
        MTL_ERROR1( "XHH_hob_query_struct_init: ERROR : cannot initialize data for query_hca (%d)\n", ret);
        goto query_struct_init_err;
    }

    /* TK - start events after all CMDs are done */

    ret = XHH_eventp_setup_cmd_eq(rhh_hob_p->eventp, CMD_EQ_SIZE);
    if (ret != HH_OK) {
        MTL_ERROR1( "%s: ERROR : cannot set up command event queue (%d)\n", __func__, ret);
        goto eventp_cmd_err;
    }
    


    /* move the HCA to running state if had no fatal. If had fatal, return HH_EFATAL */
    MOSAL_spinlock_dpc_lock(&rhh_hob_p->fatal_spl);
    if ((rhh_hob_p->xhh_state & XHH_STATE_HAVE_ANY_FATAL) != 0) {
        /* already in FATAL state */
        MTL_DEBUG4(MT_FLFMT("%s: already in FATAL state"), __func__);  
        MOSAL_spinlock_unlock(&rhh_hob_p->fatal_spl);
        ret = HH_EFATAL;
        goto fatal_err_at_end;
    }
    rhh_hob_p->xhh_state = XHH_STATE_RUNNING;
    MOSAL_spinlock_unlock(&rhh_hob_p->fatal_spl);
    
    MTL_TRACE2("%s: device name %s\n", __func__, hca_hndl->dev_desc);
    hca_hndl->status   =   HH_HCA_STATUS_OPENED;
    
    /* free the mutex */
    MOSAL_mutex_rel(&(rhh_hob_p->mtx));

    MTL_ERROR1("%s: open hca for %s succeeded \n", __func__, hca_hndl->dev_desc);
    return(HH_OK);

fatal_err_at_end:
eventp_cmd_err:
query_struct_init_err:
init_ib_err:
    /* see if need to close IB for some ports. Do not close ports on fatal error*/
    MOSAL_spinlock_dpc_lock(&rhh_hob_p->fatal_spl);
    /*test fatal again, here -- may have gotten FATAL before end of OPEN_hca process */
    if ((rhh_hob_p->xhh_state & XHH_STATE_HAVE_ANY_FATAL) != 0) {
        /* got FATAL during OPEN_HCA */
        MTL_DEBUG4(MT_FLFMT("XHH_hob_open_hca: In FATAL state")); 
        have_fatal = TRUE;
    } else {
        have_fatal = FALSE;
    }
    MOSAL_spinlock_unlock(&rhh_hob_p->fatal_spl);
    if ((last_port_initialized) && (have_fatal == FALSE)) {
        for (i = 1; i <= last_port_initialized; i++) {
            MTL_DEBUG4(MT_FLFMT("%s: closing IB port %d"), __func__, i); 
            XHH_cmd_CLOSE_IB(rhh_hob_p->cmd, (IB_port_t) i);
        }
    }
    XHH_srqm_destroy(rhh_hob_p->srqm);
    rhh_hob_p->srqm = (XHH_srqm_t)XHH_INVALID_HNDL;

srqm_create_err:
    rhh_hob_p->ib_eq = XHH_INVALID_EQN;

eventp_async_err:
    rhh_hob_p->compl_eq = XHH_INVALID_EQN;

eventp_compl_err:
    XHH_eventp_destroy(rhh_hob_p->eventp);

eventp_create_err:
    if (rhh_hob_p->mcgm != (XHH_mcgm_t)XHH_INVALID_HNDL) {
        XHH_mcgm_destroy(rhh_hob_p->mcgm );
        rhh_hob_p->mcgm = (XHH_mcgm_t)XHH_INVALID_HNDL;
    }

mcgm_create_err:
    XHH_qpm_destroy( rhh_hob_p->qpm, TRUE);
    rhh_hob_p->qpm = (XHH_qpm_t)XHH_INVALID_HNDL;

qpm_create_err:
    FREE(rhh_hob_p->init_ib_props);
    rhh_hob_p->init_ib_props = ( XHH_port_init_props_t *) NULL;

init_ib_props_malloc_err:
    XHH_cqm_destroy( rhh_hob_p->cqm, TRUE);
    rhh_hob_p->cqm = (XHH_cqm_t)XHH_INVALID_HNDL;

cqm_create_err:
    XHH_mrwm_destroy(rhh_hob_p->mrwm, TRUE);
    rhh_hob_p->mrwm = (XHH_mrwm_t)XHH_INVALID_HNDL;

mrwm_create_err:
    XHH_pdm_destroy(rhh_hob_p->pdm );
    rhh_hob_p->pdm = (XHH_pdm_t)XHH_INVALID_HNDL;

pdm_create_err:
    XHH_uarm_destroy(rhh_hob_p->uarm);
    rhh_hob_p->uarm = (XHH_uarm_t)XHH_INVALID_HNDL;

uarm_create_err:
query_hca_err:
    MOSAL_spinlock_dpc_lock(&rhh_hob_p->fatal_spl);
    /*test fatal again, here -- may have gotten FATAL before end of OPEN_hca process */
    if ((rhh_hob_p->xhh_state & XHH_STATE_HAVE_ANY_FATAL) != 0) {
        /* got FATAL during OPEN_HCA */
        MTL_DEBUG4(MT_FLFMT("XHH_hob_open_hca: In FATAL state")); 
        have_fatal = TRUE;
    } else {
        have_fatal = FALSE;
    }
    MOSAL_spinlock_unlock(&rhh_hob_p->fatal_spl);
    if (have_fatal) {
        if (rhh_hob_p->module_flags.fatal_delay_halt == 0) {
            MTL_ERROR1(MT_FLFMT("%s: Fatal State: halting the HCA"), __func__);
            XHH_cmd_CLOSE_HCA(rhh_hob_p->cmd, TRUE);
        }
    } else {
        MTL_ERROR1(MT_FLFMT("%s: closing the HCA on non-fatal error %d"), __func__, ret);
        XHH_cmd_CLOSE_HCA(rhh_hob_p->cmd, FALSE);
    }



init_hca_err:
    unmap_reserved_fw_resources(rhh_hob_p);

map_fw_rscs_err:        
    XHH_icmm_free_sz_aligned(rhh_hob_p->icmm, 
                             rhh_hob_p->profile.icm_alloc_vec_size,      /*number of chunks */
                             (MT_size_t *) &(rhh_hob_p->icm_alloc_size_vec), /* IN  */
                             (XHH_icm_va_t *)&(rhh_hob_p->icm_alloc_base_addrs_vec) );  /* IN */

alloc_size_aligned_err:
    XHH_icmm_destroy(rhh_hob_p->icmm);

icmm_create_err:
    cmd_ret = XHH_cmd_UNMAP_ICM_AUX(rhh_hob_p->cmd);
    if (cmd_ret != XHH_CMD_STAT_OK) {
      if (cmd_ret == XHH_CMD_STAT_EFATAL) {
          MTL_ERROR1(MT_FLFMT("%s: FATAL ERROR in XHH_cmd_UNMAP_ICM_AUX"),__func__);
      } else {
          MTL_ERROR1(MT_FLFMT("%s: CMD_error in XHH_cmd_UNMAP_ICM_AUX (%d)"), __func__,cmd_ret);
      }
    }
    if ((cmd_ret == XHH_CMD_STAT_OK)&&(rhh_hob_p->have_lam == FALSE)) {
        XHH_phys_mem_free(rhh_hob_p->icm_aux_phys_mem_vec_p, rhh_hob_p->icm_aux_phys_mem_vec_size);
        rhh_hob_p->icm_aux_phys_mem_vec_p = NULL;
        rhh_hob_p->icm_aux_phys_mem_vec_size = 0;
    }
profile_err:
query_adapter_err:
query_devlim_err:
    XHH_cmd_UNMAP_FA(rhh_hob_p->cmd);
    if (used_lam_for_fw == FALSE) {
        XHH_phys_mem_free(rhh_hob_p->fw_phys_mem_vec_p, rhh_hob_p->fw_phys_mem_vec_size);
        rhh_hob_p->fw_phys_mem_vec_p = NULL;
        rhh_hob_p->fw_phys_mem_vec_size = 0;
    }
	if (rhh_hob_p->lamm != (XHH_lamm_t) XHH_INVALID_HNDL) {
        XHH_lamm_destroy(rhh_hob_p->lamm);
        rhh_hob_p->lamm = (XHH_lamm_t) XHH_INVALID_HNDL;
        rhh_hob_p->have_lam = FALSE;
    }
init_fw_err:
     if (rhh_hob_p->have_lam == TRUE) {
         XHH_cmd_DISABLE_LAM(rhh_hob_p->cmd);
     }
enable_lam_err:
    unmap_fw_errbuf(rhh_hob_p);

fatal_thr_init_err:
  /* signal the waited-on sync object, so that thread exits */
    {
	call_result_t      mosal_ret1;
	MT_USE_VARIABLE(mosal_ret1);
	MT_START_FATAL_HANDLING;

	 /* release the fatal signalling thread */
	MTL_ERROR1(MT_FLFMT("%s: Releasing cleanup thread"),__func__); 
	rhh_hob_p->fatal_thread_obj.have_fatal = FALSE;
	MOSAL_syncobj_signal(&rhh_hob_p->fatal_thread_obj.fatal_err_sync);
	mosal_ret1 = MOSAL_syncobj_waiton(&(rhh_hob_p->fatal_thread_obj.stop_sync), 10000000);
	if (mosal_ret1 != MT_OK) {
	    if (mosal_ret1 == MT_EINTR) {
	        MTL_DEBUG1(MT_FLFMT("%s: Received OS interrupt while destroying fatal error thread (err = %d)"), 
	                     __func__,mosal_ret1);
	    } else {
	        MTL_ERROR1(MT_FLFMT("%s: Timeout on destroying fatal error thread (err = %d)"), 
	                     __func__,mosal_ret1);
	    }
	}
	MT_END_FATAL_HANDLING;
    }
  if (rhh_hob_p->fw_error_buf != NULL) {
      FREE(rhh_hob_p->fw_error_buf);
      rhh_hob_p->fw_error_buf = NULL;
  }
  unmap_fw_errbuf(rhh_hob_p);

post_cmd_create_err:
    XHH_cmd_destroy(rhh_hob_p->cmd);

cmd_create_err:
post_mutex_acquire_err:
    MOSAL_mutex_rel(&(rhh_hob_p->mtx));

post_state_change_error:
    MOSAL_spinlock_dpc_lock(&rhh_hob_p->fatal_spl);
    /*test fatal again, here -- may have gotten FATAL before end of OPEN_hca process */
    if ((rhh_hob_p->xhh_state & XHH_STATE_HAVE_ANY_FATAL) != 0) {
        /* got FATAL during OPEN_HCA */
        have_fatal = TRUE;
    } else {
        /* restore the state to closed */
        have_fatal = FALSE;
        rhh_hob_p->xhh_state = XHH_STATE_CLOSED;
    }
    MOSAL_spinlock_unlock(&rhh_hob_p->fatal_spl);
    if (have_fatal) {
        MTL_ERROR1(MT_FLFMT("XHH_hob_open_hca: Received FATAL error while opening.  Attempting restart.")); 
        XHH_hob_restart(hca_hndl);
    }
    return ret;
}

/******************************************************************************
 *  Function:     XHH_hob_query
 *
 *  Description:  Implements VAPI_query_hca verb.  Data is already stored in HOB object.
 *
 *  input:
 *                hca_hndl
 *  output: 
 *                hca_cap_p -- pointer to output structure
 *  returns:
 *                HH_OK
 *                HH_EINVAL_HCA_HNDL
 *
 *****************************************************************************/
 HH_ret_t XHH_hob_query(HH_hca_hndl_t  hca_hndl, 
			MOSAL_protection_ctx_t prot_ctx,
                        VAPI_hca_cap_t *hca_cap_p)
{
    XHH_hob_t             hob;

    MT_RETURN_IF_LOW_STACK(XHH_WATERMARK);
    if (MOSAL_get_exec_ctx() != MOSAL_IN_TASK) {
        MTL_ERROR1("XHH_hob_query: NOT IN TASK CONTEXT)\n");
        return HH_ERR;
    }

    if (hca_hndl == NULL) {
        MTL_ERROR1("XHH_hob_query : ERROR : Invalid HCA handle\n");
        return HH_EINVAL_HCA_HNDL;
    }
    
    hob = XHHOBP(hca_hndl);

    if (hob == NULL) {
        MTL_ERROR1("XHH_hob_query : ERROR : No device registered\n");
        return HH_EINVAL;
    }
    
    TEST_RETURN_FATAL(hob);

    /* check if ib_init_props have been created -- indicates that open_hca called for this device */
    if (hob->init_ib_props == (XHH_port_init_props_t *)NULL) {
        MTL_ERROR1("XHH_hob_query: ERROR : HCA device has not yet been opened\n");
        return HH_EINVAL;
    }

    memcpy(hca_cap_p, &(hob->hca_capabilities), sizeof(VAPI_hca_cap_t));

    /* for user space apps, number of udavs is limited only by host memory */
    if (prot_ctx != MOSAL_get_kernel_prot_ctx()) {
	hca_cap_p->max_ah_num = 0x7FFFFFFF;
    }
    return HH_OK;
}

#define PRM_WQE_SEG_SZ_NEXT (sizeof(struct arbelprm_wqe_segment_next_st)/8)            /* NEXT segment */
#define PRM_WQE_SEG_SZ_CTRL (sizeof(struct arbelprm_wqe_segment_ctrl_send_st)/8)       /* CTRL segment */
#define PRM_WQE_SEG_SZ_UD   (sizeof(struct arbelprm_wqe_segment_ud_st)/8)              /* DATAGRAM:UD */
#define PRM_WQE_SEG_SZ_SG_ENTRY (sizeof(struct arbelprm_wqe_segment_data_ptr_st)/8)/* Scatter/Gather entry(ptr)*/
#define PRM_MAX_SQ_HDR  (PRM_WQE_SEG_SZ_NEXT+PRM_WQE_SEG_SZ_CTRL+PRM_WQE_SEG_SZ_UD)
#define PRM_MAX_RQ_HDR  (PRM_WQE_SEG_SZ_NEXT+PRM_WQE_SEG_SZ_CTRL)


#define HOB_MIN(a,b)	((a) < (b) ? (a) : (b))
#define SET_MAX_SG(a)  (((a) < (u_int32_t)HOB_MIN(((hob->dev_lims.max_desc_sz_sq-PRM_MAX_SQ_HDR)/PRM_WQE_SEG_SZ_SG_ENTRY),\
                        ((hob->dev_lims.max_desc_sz_rq-PRM_MAX_RQ_HDR)/PRM_WQE_SEG_SZ_SG_ENTRY))) ? \
	                (a) : (u_int32_t)HOB_MIN(((hob->dev_lims.max_desc_sz_sq-PRM_MAX_SQ_HDR)/PRM_WQE_SEG_SZ_SG_ENTRY),\
                                                ((hob->dev_lims.max_desc_sz_rq-PRM_MAX_RQ_HDR)/PRM_WQE_SEG_SZ_SG_ENTRY)))

#define SET_MAX_SG_RQ(a)  (((a) < (u_int32_t)((hob->dev_lims.max_desc_sz_rq-PRM_MAX_RQ_HDR)/PRM_WQE_SEG_SZ_SG_ENTRY)) ? \
                                   (a) : ((hob->dev_lims.max_desc_sz_rq-PRM_MAX_RQ_HDR)/PRM_WQE_SEG_SZ_SG_ENTRY))

/******************************************************************************
 *  Function:     XHH_hob_query_struct_init
 *
 *  Description:  Pre-computes the data for VAPI_query_hca.  Called during XHH_hob_open_hca()
 *
 *  input:
 *                hob
 *  output: 
 *                hca_cap_p -- pointer to output structure
 *  returns:
 *                HH_OK
 *                HH_EINVAL_HCA_HNDL
 *
 *  Comments:     Needs to use a MAD query for some of the parameters
 *
 *****************************************************************************/
static  HH_ret_t XHH_hob_query_struct_init(XHH_hob_t  hob,
                               MT_bool have_usr_profile, 
                               VAPI_hca_cap_t *hca_cap_p)
{
    HH_ret_t  ret;
    u_int32_t flags = 0;

    MT_RETURN_IF_LOW_STACK(XHH_WATERMARK);
  MTL_DEBUG4("Entering XHH_hob_query_struct_init\n");
        /* Maximum Number of QPs supported.                   */
  hca_cap_p->max_num_qp = (u_int32_t)hob->profile.max_num_qps;
  /*  JPM: FOR NOW, ARBEL DRIVER DOES NOT SUPPORT SRQ */
  hca_cap_p->max_num_srq = (u_int32_t)hob->profile.max_num_srqs;
  /* hca_cap_p->max_num_srq = 0; */
  hca_cap_p->srq_resize_supported= TRUE;
        /* Maximum Number of oustanding WR on any WQ.         */
  hca_cap_p->max_qp_ous_wr = (1 << (hob->dev_lims.log_max_qp_sz)) - 1;                    
  hca_cap_p->max_wqe_per_srq = (1 << (hob->dev_lims.log_max_srq_sz)) - 1;                    
        
  /* Various flags (VAPI_hca_cap_flags_t)               */

//  VAPI_RESIZE_OUS_WQE_CAP     = 1  /* Not currently supported */

  flags =  (hob->hca_props.udp ? VAPI_UD_AV_PORT_ENFORCE_CAP : 0) |
           (hob->dev_lims.apm ? VAPI_AUTO_PATH_MIG_CAP : 0 )      |
           (hob->dev_lims.rm  ? VAPI_RAW_MULTI_CAP     : 0)       |
           (hob->dev_lims.pkv ? VAPI_BAD_PKEY_COUNT_CAP : 0)      |
           (hob->dev_lims.qkv ? VAPI_BAD_QKEY_COUNT_CAP : 0)      |
           VAPI_CHANGE_PHY_PORT_CAP | VAPI_RC_RNR_NAK_GEN_CAP | VAPI_PORT_ACTIVE_EV_CAP;
           

  hca_cap_p->flags = flags;               

        /* Max num of scatter/gather entries for desc other than RD 
         * NOTE: max number differs from Tavor, since we don't need 
         * a workaround in Arbel memfree 
	 */
  hca_cap_p->max_num_sg_ent = SET_MAX_SG(59);       
        /* Max num of scatter entries for SRQs */
  hca_cap_p->max_srq_sentries = SET_MAX_SG_RQ(59);       
        /* Max num of scatter/gather entries for RD desc      */
  hca_cap_p->max_num_sg_ent_rd = 0 ;   
        /* Max num of supported CQs                           */
  hca_cap_p->max_num_cq = (u_int32_t)hob->profile.max_num_cqs;
  /* Max num of supported entries per CQ                */
  hca_cap_p->max_num_ent_cq = (1 << (hob->dev_lims.log_max_cq_sz)) - 1 /*for extra cqe needed */;      
        /* Maximum number of memory region supported.         */
  hca_cap_p->max_num_mr = (u_int32_t)hob->profile.num_external_mem_regions;          
        /* Maximum  memory region size supported.         */
  hca_cap_p->max_mr_size = MAKE_ULONGLONG(0xFFFFFFFFFFFFFFFF);
        /* Maximum number of protection domains supported.    */
  hca_cap_p->max_pd_num = (u_int32_t)hob->profile.max_num_pds;
  /* Largest page size supported by this HCA            */
  hca_cap_p->page_size_cap = (1 << (hob->dev_lims.log_pg_sz));       
        /* Number of physical ports of the HCA.                */
  hca_cap_p->phys_port_num = hob->dev_lims.num_ports;       
        /* Maximum number of partitions supported .           */
  hca_cap_p->max_pkeys = hob->init_ib_props[0].max_pkey;           
        /* Maximum number of oust. RDMA read/atomic as target */
  hca_cap_p->max_qp_ous_rd_atom = 1 << (hob->profile.log2_inflight_rdma_per_qp);  
        /* EE Maximum number of outs. RDMA read/atomic as target -- NOT YET SUPPORTED  */
  hca_cap_p->max_ee_ous_rd_atom = 0;  
        /* Max. Num. of resources used for RDMA read/atomic as target */
  hca_cap_p->max_res_rd_atom = ((1 << (hob->icm_alloc_size_vec.log2_rdb_size - XHH_ICM_LOG2_RDB_ENTRY_SIZE)) > 255 ? 
                                    255 :(1 << (hob->icm_alloc_size_vec.log2_rdb_size - XHH_ICM_LOG2_RDB_ENTRY_SIZE))) ;     
        /* Max. Num. of outs. RDMA read/atomic as initiator. Note that 255 is the max in the struct  */
  hca_cap_p->max_qp_init_rd_atom = ((1 << (hob->dev_lims.log_max_ra_req_qp)) > 255 ? 
                                    255 :  (1 << (hob->dev_lims.log_max_ra_req_qp))); 
        /* EE Max. Num. of outs. RDMA read/atomic as initiator  -- NOT YET SUPPORTED   */
  hca_cap_p->max_ee_init_rd_atom = 0;
        /* Level of Atomicity supported:  if supported, is only within this HCA*/
  hca_cap_p->atomic_cap = (hob->dev_lims.atm ? VAPI_ATOMIC_CAP_HCA : VAPI_ATOMIC_CAP_NONE);        
        /* Maximum number of EEC supported.   -- NOT YET SUPPORTED  */
  hca_cap_p->max_ee_num = 0;
        /* Maximum number of IB_RDD supported  -- NOT YET SUPPORTED */
  hca_cap_p-> max_rdd_num = 0;                 
        /* Maximum Number of memory windows supported  */
  hca_cap_p->max_mw_num = (u_int32_t)hob->profile.num_mem_windows;                   
        /* Maximum number of Raw IPV6 QPs supported  -- NOT YET SUPPORTED */ 
  hca_cap_p->max_raw_ipv6_qp = 0;              
        /* Maximum number of Raw Ethertypes QPs supported  -- NOT YET SUPPORTED */
  hca_cap_p->max_raw_ethy_qp = 0;              
        /* Maximum Number of multicast groups  */
  hca_cap_p->max_mcast_grp_num = 1 << (hob->hca_props.multicast_parameters.log_mc_table_sz);            
        /* Maximum number of QP per multicast group    */
  hca_cap_p->max_mcast_qp_attach_num = ( (1U<<(hob->hca_props.multicast_parameters.log_mc_table_entry_sz)) 
                                               -  XHH_ICM_MCG_ENTRY_HEADER_SIZE) / 
                                                             XHH_ICM_MCG_BYTES_PER_QP;
        /* Maximum number of QPs which can be attached to a mcast grp */
  hca_cap_p->max_total_mcast_qp_attach_num = hca_cap_p->max_mcast_grp_num * hca_cap_p->max_mcast_qp_attach_num;
        /* Maximum number of address handles. No real max -- so just use what is in the default profile file */
  hca_cap_p->max_ah_num = ((hob->module_flags.max_kernel_avs == 0xFFFFFFFF) ?
                       (u_int32_t) RHH_KERNEL_MAX_AV : hob->module_flags.max_kernel_avs ); 
        /* max number of fmrs for the use is the number of user entries in MPT */
  hca_cap_p->max_num_fmr    = (u_int32_t)(hob->profile.num_external_mem_regions); 
/* max maps per fmr: in Arbel we use the 8-LSbits for FMR memkey modification */
  hca_cap_p->max_num_map_per_fmr = 255;
  
        /* Log2 4.096usec Max. RX to ACK or NAK delay */
  hca_cap_p->local_ca_ack_delay = hob->dev_lims.local_ca_ack_delay;

  /* Node GUID for this hca */
  ret = XHH_hob_get_node_guid(hob,&(hca_cap_p->node_guid));           

  return(ret);
}

/******************************************************************************
 *  Function:     XHH_hob_create_internal
 *
 *  Description:  Creates the HOB object for an HCA, and registers it in HH  
 *
 *  input:
 *                hw_props_p -- PCI properties (BARs, etc)
 *                hca_seq_num  - a sequence number assigned to this HCA to differentiate it 
 *                               from other HCAs on this host
 *                mod_flags   - flags passed in at module initialization (e.g., insmod)
 *                curr_hh_hndl  - current HH handle - valid for fatal error restart. NULL otherwise
 *                                (NULL value indicates new creation)
 *  output: 
 *                hh_hndl_p  - returned HH handle
 *  returns:
 *                HH_OK
 *                HH_EAGAIN
 *                HH_ERR -- other errors
 *
 *  Comments:     This function involves the following steps: 
 *                    1.Allocate XHH_hob data context. 
 *                    2.Register HCA in HH (i.e.call HH_add_hca_dev()) for a new registration, or set
 *                      the HH object state to HH_HCA_STATUS_CLOSED for a re-activation in fatal error
 *                      restart
 *
 *                Also initializes the HOB mutex for controlling Open HCA and Close HCA 
 *
 *****************************************************************************/
HH_ret_t    XHH_hob_create_internal(/*IN*/  XHH_hw_props_t   *hw_props_p,
                           /*IN*/  u_int32_t         hca_seq_num,
                           /*IN*/  XHH_module_flags_t *mod_flags,
                           /*IN*/  HH_hca_hndl_t    curr_hh_hndl,
                           /*OUT*/ HH_hca_hndl_t    *hh_hndl_p )
{

  //HH_hca_hndl_t  hca_hndl = 0;
  HH_ret_t        ret;
  HH_if_ops_t     *if_ops_p = 0;
  XHH_hob_t       hob_p;
  HH_hca_hndl_t   new_hh_hndl;
  HH_hca_dev_t    *rdev_p = NULL;
  int             int_ret = 0;

  MT_RETURN_IF_LOW_STACK(XHH_WATERMARK);
  MTL_DEBUG4("Entering XHH_hob_create\nhca_seq_num = %d, legacy_flag = %s, av_in_host_mem = %s\n",  
             hca_seq_num, (mod_flags->legacy_sqp == FALSE ? "FALSE" : "TRUE"),
              (mod_flags->av_in_host_mem == FALSE ? "FALSE" : "TRUE"));
  XHH_print_hw_props(hw_props_p);

  if (MOSAL_get_exec_ctx() != MOSAL_IN_TASK) {
      MTL_ERROR1("XHH_hob_create: NOT IN TASK CONTEXT)\n");
      return HH_ERR;
  }

  XHH_set_capabilities(hw_props_p, mod_flags);

  /* Allocate RHH hob structure  */                                     
  hob_p = (XHH_hob_t)MALLOC(sizeof(struct XHH_hob_st));

  if (hob_p == 0) {
    MTL_ERROR1("XHH_hob_create: could not allocate memory for RHH hob\n");
    return HH_EAGAIN;
  }

  MTL_DEBUG1("XHH_hob_create: HOB address = 0x%p\n", hob_p);
  memset(hob_p, 0, sizeof(struct XHH_hob_st));


  rdev_p = (HH_hca_dev_t *)MALLOC(sizeof(HH_hca_dev_t));
  if (rdev_p == NULL) {
     MT_COV_MEM_FLTR_START
     FREE(hob_p);
     return HH_EAGAIN;
     MT_COV_MEM_FLTR_END
  }
  memset(rdev_p, 0, sizeof(HH_hca_dev_t));

    
  /* initialize the HOB mutex */
  MOSAL_mutex_init(&(hob_p->mtx));

  /* set device name */
  if (mod_flags->legacy_name == FALSE) {
	 if (hw_props_p->device_id == MT25218_DEV_ID) { /* Arbel */
             sprintf(hob_p->dev_name,HCA_ARBEL_DEV_ID_PREFIX"%d",hca_seq_num); 
	 }
	 else {	 /* Sinay */
             sprintf(hob_p->dev_name,HCA_SINAI_DEV_ID_PREFIX"%d",hca_seq_num); 
	 }
  } else {
     sprintf(hob_p->dev_name,HCA_DEV_ID_LEGACY_PREFIX"%d",hca_seq_num); 
  }

  printk("\nMellanox Device Driver is creating device \"%s\" (bus=%02x, devfn=%02x)\n\n", 
         hob_p->dev_name,hw_props_p->bus,hw_props_p->dev_func);

  /* set embedded object handles to invalid */
  
  hob_p->cmd  = (XHH_cmd_t)XHH_INVALID_HNDL;
  hob_p->pdm = (XHH_pdm_t)XHH_INVALID_HNDL;
  hob_p->uarm = (XHH_uarm_t)XHH_INVALID_HNDL;
  hob_p->mrwm = (XHH_mrwm_t)XHH_INVALID_HNDL;
  hob_p->cqm = (XHH_cqm_t)XHH_INVALID_HNDL;
//hob_p->eecm = (XHH_eecm_t)XHH_INVALID_HNDL;  /* JPM -- EECM ADDITIONS HERE */
  hob_p->qpm = (XHH_qpm_t)XHH_INVALID_HNDL;
  hob_p->mcgm = (XHH_mcgm_t)XHH_INVALID_HNDL;
  hob_p->eventp = (XHH_eventp_t)XHH_INVALID_HNDL;
  hob_p->lamm = (XHH_lamm_t)XHH_INVALID_HNDL;
  hob_p->icmm = (XHH_icmm_t)XHH_INVALID_HNDL;
  hob_p->srqm = (XHH_srqm_t)XHH_INVALID_HNDL;
  
  /* initialize EQ handles to all EQs invalid */
  hob_p->compl_eq = XHH_INVALID_EQN;
  hob_p->ib_eq    = XHH_INVALID_EQN;

  /* initialize fatal error handling fields */
  memcpy(&(hob_p->hw_props), hw_props_p, sizeof(XHH_hw_props_t));
  memcpy(&(hob_p->module_flags), mod_flags, sizeof(XHH_module_flags_t));
  hob_p->hca_seq_num = hca_seq_num;
  hob_p->xhh_state = XHH_STATE_CREATING;
  if (MOSAL_spinlock_init(&(hob_p->fatal_spl)) != MT_OK){
    MTL_ERROR4(MT_FLFMT("%s: Failed to initializing fatal error spinlock"), __func__);  
    ret= HH_ERR;
    goto err_free_hob;
  }
  if (MOSAL_spinlock_init(&(hob_p->async_spl)) != MT_OK){
    MTL_ERROR4(MT_FLFMT("%s: Failed to initializing async handler tracking spinlock"), __func__);  
    ret= HH_ERR;
    goto err_free_hob;
  }
  /* dummy async event handler is passed to XHH_eventp when initializing ib_eq */
  hob_p->async_eventh = &XHH_dummy_async_event;
  hob_p->async_ev_private_context = NULL;

  /* get hca config info */
  ret = XHH_read_pci_config(hob_p->hw_props.pci_dev, &hob_p->pci_hca_info.config);
  if ( ret != HH_OK ) {
    MTL_ERROR1(MT_STDFMT("XHH_read_pci_config failed"));
    ret = HH_ERR;
    goto err_free_hob;
  }

  hob_p->pci_hca_info.bus = hw_props_p->bus;
  hob_p->pci_hca_info.dev_func = hw_props_p->dev_func;
  hob_p->pci_hca_info.is_valid = TRUE;
  
  /*create the delay unlock object for catastrophic error use */
  int_ret = VIP_delay_unlock_create(&hob_p->delay_unlocks);
  if (int_ret != 0) {
      MTL_ERROR1("XHH_hob_create: could create delay unlock object (err = %d)\n", int_ret);
      ret = HH_ENOMEM;
      goto delay_unlock_err;
  }
  
  /* set up the procedure mapping table and register the arbel device */
  if_ops_p = &(hob_p->if_ops);

  HH_ifops_tbl_set_enosys(if_ops_p); /* by default, all retuen HH_ENOSYS */

  /* HCA Calls */
  if_ops_p->HHIF_open_hca           = &XHH_hob_open_hca;
  if_ops_p->HHIF_close_hca          = &XHH_hob_close_hca;
  if_ops_p->HHIF_alloc_ul_resources = &XHH_hob_alloc_ul_res;
  if_ops_p->HHIF_free_ul_resources  = &XHH_hob_free_ul_res;
  if_ops_p->HHIF_query_hca          = &XHH_hob_query;
  if_ops_p->HHIF_modify_hca         = &XHH_hob_modify;

  /* Misc HCA Operations*/
  if_ops_p->HHIF_query_port_prop        = &XHH_hob_query_port_prop;
  if_ops_p->HHIF_get_pkey_tbl           = &XHH_hob_get_pkey_tbl;
  if_ops_p->HHIF_get_gid_tbl            = &XHH_hob_get_gid_tbl;

  /* Protection Domain Calls */
  if_ops_p->HHIF_alloc_pd               = &XHH_hob_alloc_pd;
  if_ops_p->HHIF_free_pd                = &XHH_hob_free_pd;
  if_ops_p->HHIF_alloc_rdd              = &XHH_hob_alloc_rdd;
  if_ops_p->HHIF_free_rdd               = &XHH_hob_free_rdd;

  /* Memory Registration */
  if_ops_p->HHIF_register_mr     = &XHH_hob_register_mr;
  if_ops_p->HHIF_reregister_mr   = &XHH_hob_reregister_mr;
  if_ops_p->HHIF_register_smr    = &XHH_hob_register_smr;
  if_ops_p->HHIF_query_mr        = &XHH_hob_query_mr;
  if_ops_p->HHIF_deregister_mr   = &XHH_hob_deregister_mr;

  if_ops_p->HHIF_alloc_mw        = &XHH_hob_alloc_mw;
  if_ops_p->HHIF_query_mw        = &XHH_hob_query_mw;
  if_ops_p->HHIF_free_mw         = &XHH_hob_free_mw;

  /* Fast memory regions */
  if_ops_p->HHIF_alloc_fmr       = &XHH_hob_alloc_fmr;
  if_ops_p->HHIF_map_fmr         = &XHH_hob_map_fmr;
  if_ops_p->HHIF_unmap_fmr       = &XHH_hob_unmap_fmr;
  if_ops_p->HHIF_free_fmr        = &XHH_hob_free_fmr;

  /* Completion Queues */
  if_ops_p->HHIF_create_cq       = &XHH_hob_create_cq;
  if_ops_p->HHIF_resize_cq       = &XHH_hob_resize_cq;
  if_ops_p->HHIF_query_cq        = &XHH_hob_query_cq;
  if_ops_p->HHIF_destroy_cq      = &XHH_hob_destroy_cq;

  /* Queue Pair */
  if_ops_p->HHIF_create_qp       = &XHH_hob_create_qp;
  if_ops_p->HHIF_get_special_qp  = &XHH_hob_get_special_qp;
  if_ops_p->HHIF_modify_qp       = &XHH_hob_modify_qp;
  if_ops_p->HHIF_query_qp        = &XHH_hob_query_qp;
  if_ops_p->HHIF_destroy_qp      = &XHH_hob_destroy_qp;
#if defined(MT_SUSPEND_QP)
  if_ops_p->HHIF_suspend_qp      = &XHH_hob_suspend_qp;
  if_ops_p->HHIF_suspend_cq      = &XHH_hob_suspend_cq;
#endif
  /* SRQ */

#if 1
  if_ops_p->HHIF_create_srq      = &XHH_hob_create_srq;
  if_ops_p->HHIF_query_srq       = &XHH_hob_query_srq;
  if_ops_p->HHIF_modify_srq      = &XHH_hob_modify_srq;
  if_ops_p->HHIF_destroy_srq     = &XHH_hob_destroy_srq;
#endif
  /* EEC */
  if_ops_p->HHIF_create_eec      = &XHH_hob_create_eec;
  if_ops_p->HHIF_modify_eec      = &XHH_hob_modify_eec;
  if_ops_p->HHIF_query_eec       = &XHH_hob_query_eec;
  if_ops_p->HHIF_destroy_eec     = &XHH_hob_destroy_eec;

  if_ops_p->HHIF_set_comp_eventh  = &XHH_hob_set_comp_eventh;
  if_ops_p->HHIF_set_async_eventh = &XHH_hob_set_async_eventh;
  
  
  /* Multicast groups */
  if_ops_p->HHIF_attach_to_multicast   = &XHH_hob_attach_to_multicast;
  if_ops_p->HHIF_detach_from_multicast = &XHH_hob_detach_from_multicast;
  
  /* Process local MAD */
  if_ops_p->HHIF_process_local_mad = &XHH_hob_process_local_mad;
  
#if 0
  /* no separate ddrmm object in Arbel. The LAM is managed exclusively
   * for ICM virtual addr space in Arbel (and its object is contained
   * within ICMM
   */
  if_ops_p->HHIF_ddrmm_alloc = &XHH_hob_ddrmm_alloc;
  if_ops_p->HHIF_ddrmm_query = &XHH_hob_ddrmm_query;
  if_ops_p->HHIF_ddrmm_free = &XHH_hob_ddrmm_free;
#endif

  /*
   *  Register device in the init structure
   *
   */

  if (curr_hh_hndl == NULL) {
      rdev_p->ul_if_ops_p = NULL;
      strncpy(rdev_p->dev_desc, hob_p->dev_name, sizeof(rdev_p->dev_desc)-1);
      strcpy(rdev_p->ul_init_info.user_devlib_info.devlib_filename, MT_RHH_DLL_NAME);
      strcpy(rdev_p->ul_init_info.user_devlib_info.devlib_hob_create, "XHHUL_hob_create");
      rdev_p->ul_init_info.kernel_ul_hob_create_p  = &XHHUL_hob_create;
      rdev_p->vendor_id = MT_MELLANOX_IEEE_VENDOR_ID;
      rdev_p->dev_id    = (u_int32_t)hw_props_p->device_id;
      rdev_p->hw_ver    = hob_p->hw_props.hw_ver; 
      rdev_p->if_ops    = if_ops_p;
      rdev_p->hca_ul_resources_sz = sizeof(XHH_hca_ul_resources_t);
      rdev_p->pd_ul_resources_sz = sizeof(XHH_pd_ul_resources_t);
      rdev_p->cq_ul_resources_sz = sizeof(XHH_cq_ul_resources_t);
      rdev_p->srq_ul_resources_sz = sizeof(XHH_srq_ul_resources_t);
      rdev_p->qp_ul_resources_sz = sizeof(XHH_qp_ul_resources_t);
      rdev_p->device = (void *) hob_p;
      rdev_p->status = HH_HCA_STATUS_ADD_PREP;
      /* Grab the mutex now, just before adding the device to HH */
      MTL_DEBUG4(MT_FLFMT("XHH_hob_create:  about to grab mutex"));
      if (MOSAL_mutex_acq(&(hob_p->mtx), TRUE) != MT_OK) {
	  MTL_DEBUG1(MT_FLFMT("%s: Received signal. returning HH_EINTR"), __func__);
	  ret = HH_EINTR;
	  goto err_acq_mutex;
      }
    
      MTL_DEBUG4("XHH_hob_create:  Before HH_add_hca_dev\n");
      ret = HH_add_hca_dev_prep(rdev_p, &new_hh_hndl);
      if (ret != HH_OK) {
	MTL_ERROR1("XHH_hob_create: could not register device %s in HCA HAL\n",
				hob_p->dev_name);
	goto err_release_mutex;
      }
  } else {
    /* Grab the mutex now, just before adding the device to HH */
    MTL_DEBUG4(MT_FLFMT("XHH_hob_create:  about to grab mutex"));
    if (MOSAL_mutex_acq(&(hob_p->mtx), TRUE) != MT_OK) {
      MTL_DEBUG1(MT_FLFMT("%s: Received signal. returning HH_EINTR"), __func__);
      ret = HH_EINTR;
      goto err_acq_mutex;
    }
    MTL_DEBUG1(MT_FLFMT("XHH_hob_create:  Attaching XHH and reactivating HH object. hndl=%p"), curr_hh_hndl);
    new_hh_hndl = curr_hh_hndl;
    new_hh_hndl->device = (void *) hob_p;
    new_hh_hndl->if_ops = if_ops_p;
    new_hh_hndl->hw_ver = hob_p->hw_props.hw_ver;
    new_hh_hndl->status = HH_HCA_STATUS_CLOSED;
  }

  /* insert HH hca handle into device structure */
  ((XHH_hob_t)(new_hh_hndl->device))->hh_hca_hndl =  new_hh_hndl;
  MTL_TRACE1("XHH_hob_create: hh_hca_hndl created = %p\n", (void *) new_hh_hndl);

  if (hh_hndl_p != NULL) {
      *hh_hndl_p =  new_hh_hndl;
  }
  MOSAL_spinlock_dpc_lock(&hob_p->fatal_spl);
  hob_p->xhh_state = XHH_STATE_CLOSED;
  MOSAL_spinlock_unlock(&hob_p->fatal_spl);
  
  MOSAL_mutex_rel(&(hob_p->mtx));
  FREE(rdev_p);
  return(HH_OK);

  /* ERROR HANDLING: undoes previous steps in reverse order, as needed */
err_release_mutex:  
  MOSAL_mutex_rel(&(hob_p->mtx));
err_acq_mutex:  
  VIP_delay_unlock_destroy(hob_p->delay_unlocks);

delay_unlock_err:

err_free_hob:
  MOSAL_mutex_free(&(hob_p->mtx));
  FREE(rdev_p);
  FREE(hob_p);
  return (ret);
}

/******************************************************************************
 *  Function:     XHH_hob_create
 *
 *  Description:  Creates the HOB object for an HCA, and registers it in HH  
 *
 *  input:
 *                hw_props_p -- PCI properties (BARs, etc)
 *                hca_seq_num  - a sequence number assigned to this HCA to differentiate it 
 *                               from other HCAs on this host
 *                mod_flags   - flags passed in at module initialization (e.g., insmod)
 *  output: 
 *                hh_hndl_p  - size of returned table (in pkeys)
 *  returns:
 *                HH_OK
 *                HH_EAGAIN
 *                HH_ERR -- other errors
 *
 *****************************************************************************/
HH_ret_t    XHH_hob_create(/*IN*/  XHH_hw_props_t   *hw_props_p,
                           /*IN*/  u_int32_t         hca_seq_num,
                           /*IN*/  XHH_module_flags_t *mod_flags,
                           /*OUT*/ HH_hca_hndl_t    *hh_hndl_p )
{
    HH_ret_t ret;
    ret = XHH_hob_create_internal(hw_props_p, hca_seq_num, mod_flags, NULL,hh_hndl_p );
    if (ret != HH_OK) {
	    MTL_ERROR1(MT_FLFMT("%s: XHH_hob_create_internal failed (%d:%s)"),
		       __func__, ret, HH_strerror_sym(ret));
	    return ret;
    }
    ret= HH_notify_add_hca_dev(*hh_hndl_p);
    if (ret != HH_OK) {
	    MTL_ERROR1(MT_FLFMT("%s: HH_notify_add_hca_dev failed (%d:%s)"),
		       __func__, ret, HH_strerror_sym(ret));
	    XHH_hob_destroy_internal(*hh_hndl_p, FALSE);
    }
    return ret;
}

/*****************************************************************************
******************************************************************************
************** EXTERNALLY VISIBLE FUNCTIONS, WITH PROTOTYPES IN XHH_HOB.H ****
******************************************************************************
*****************************************************************************/

HH_ret_t XHH_hob_get_pdm ( /*IN*/ XHH_hob_t hob, 
                            /*OUT*/ XHH_pdm_t *pdm_p )
{
  MT_RETURN_IF_LOW_STACK(XHH_WATERMARK);
    if (hob == NULL) {
        MTL_ERROR1("XHH_hob_get_pdm: ERROR : No device registered\n");
        return HH_EINVAL;
    }


    if (hob->pdm == (XHH_pdm_t)XHH_INVALID_HNDL) {
        MTL_ERROR1("XHH_hob_get_pdm: ERROR : HCA device has not yet been opened\n");
        return HH_EINVAL;
    }

    *pdm_p = hob->pdm;
    return HH_OK;
}

HH_ret_t XHH_hob_get_icmm ( /*IN*/ XHH_hob_t hob, 
                            /*OUT*/ XHH_icmm_t *icmm_p )
{
  MT_RETURN_IF_LOW_STACK(XHH_WATERMARK);
    if (hob == NULL) {
        MTL_ERROR1("XHH_hob_get_icmm: ERROR : No device registered\n");
        return HH_EINVAL;
    }


    if (hob->icmm == (XHH_icmm_t)XHH_INVALID_HNDL) {
        MTL_ERROR1("XHH_hob_get_icmm: ERROR : HCA device has not yet been opened\n");
        return HH_EINVAL;
    }

    *icmm_p = hob->icmm;
    return HH_OK;
}

HH_ret_t XHH_hob_get_uarm ( /*IN*/ XHH_hob_t hob, 
                            /*OUT*/ XHH_uarm_t *uarm_p )
{
  MT_RETURN_IF_LOW_STACK(XHH_WATERMARK);
    if (hob == NULL) {
        MTL_ERROR1("XHH_hob_get_uarm: ERROR : No device registered\n");
        return HH_EINVAL;
    }


    if (hob->uarm == (XHH_uarm_t)XHH_INVALID_HNDL) {
        MTL_ERROR1("XHH_hob_get_uarm: ERROR : HCA device has not yet been opened\n");
        return HH_EINVAL;
    }

    *uarm_p = hob->uarm;
    return HH_OK;
}

/* Used in restarting HCA on fatal error, in function XHH_hob_restart */
HH_ret_t   XHH_hob_get_init_params(/*IN*/ XHH_hob_t  xhh_hob_p, 
                   /*OUT*/  XHH_hw_props_t     *hw_props_p,
                   /*OUT*/  u_int32_t          *hca_seq_num,
                   /*OUT*/  XHH_module_flags_t *mod_flags)
{

  MT_RETURN_IF_LOW_STACK(XHH_WATERMARK);
    if (xhh_hob_p == NULL) {
        MTL_ERROR1(MT_FLFMT("%s: ERROR : No device registered"), __func__);  
        return HH_EAGAIN;
    }
    memcpy(hw_props_p, &xhh_hob_p->hw_props, sizeof(XHH_hw_props_t));
    *hca_seq_num = xhh_hob_p->hca_seq_num;
    memcpy(mod_flags, &xhh_hob_p->module_flags, sizeof(XHH_module_flags_t));
    return HH_OK;
}

HH_ret_t  XHH_hob_get_state(XHH_hob_t xhh_hob_p, XHH_hob_state_t *fatal_state)
{
  MT_RETURN_IF_LOW_STACK(XHH_WATERMARK);
    if (xhh_hob_p == NULL) {
        MTL_ERROR1(MT_FLFMT("%s: ERROR : No device registered"), __func__);  
        return HH_EAGAIN;
    }
    
    if (fatal_state == NULL) {
        MTL_ERROR1(MT_FLFMT("%s: ERROR : NULL fatal_state parameter"), __func__);  
        return HH_EINVAL;
    }

    MOSAL_spinlock_dpc_lock(&xhh_hob_p->fatal_spl);
    *fatal_state = xhh_hob_p->xhh_state;
    MOSAL_spinlock_unlock(&xhh_hob_p->fatal_spl);
    return HH_OK;
}

/*****************************************************************************
******************************************************************************
************** PASS-THROUGH FUNCTIONS  ********************************** ****
******************************************************************************
*****************************************************************************/


/******************************************************************************
 *  Function:     XHH_hob_alloc_ul_res <==> XHH_uldm_alloc_ul_res
 *****************************************************************************/
HH_ret_t XHH_hob_alloc_ul_res(HH_hca_hndl_t      hca_hndl,
                                 MOSAL_protection_ctx_t prot_ctx,
                                 void                   *hca_ul_resources_p)
{
  XHH_hca_ul_resources_t*  res = (XHH_hca_ul_resources_t *)hca_ul_resources_p;
  XHH_hob_t  xhh_hob_p;

  MT_RETURN_IF_LOW_STACK(XHH_WATERMARK);
  if (MOSAL_get_exec_ctx() != MOSAL_IN_TASK) {
      MTL_ERROR1("XHH_hob_alloc_ul_res: NOT IN TASK CONTEXT)\n");
      return HH_ERR;
  }

  if (hca_hndl == NULL) {
      MTL_ERROR1("XHH_hob_alloc_ul_res : ERROR : Invalid HCA handle\n");
      return HH_EINVAL_HCA_HNDL;
  }
  xhh_hob_p = XHHOBP(hca_hndl);

  /* Need to see that uldm object has been allocated.  Then, need to invoke
   * the alloc_ul_resources method of the uldm here. 
   * NOTE: may want the constructor to already pre-allocate the ul resources based upon
   *       configuration info obtained via query.
   */
  memset(res, 0, sizeof(XHH_hca_ul_resources_t));

  if (xhh_hob_p == NULL) {
      MTL_ERROR1("XHH_hob_alloc_ul_res: ERROR : No device registered\n");
      return HH_EAGAIN;
  }
  TEST_RETURN_FATAL(xhh_hob_p);

  if (xhh_hob_p->uarm == (XHH_uarm_t)XHH_INVALID_HNDL) {
      MTL_ERROR1("XHH_hob_alloc_ul_res: ERROR : HCA device has not yet been opened\n");
      return HH_EAGAIN;
  }

  /* Set XHH_hob's information in given hca_ul_res_p buffer */
  ((XHH_hca_ul_resources_t*)hca_ul_resources_p)->hh_hca_hndl= hca_hndl;
  ((XHH_hca_ul_resources_t*)hca_ul_resources_p)->vendor_id= xhh_hob_p->hw_props.pci_vendor_id;
  ((XHH_hca_ul_resources_t*)hca_ul_resources_p)->device_id= xhh_hob_p->hh_hca_hndl->dev_id;
  ((XHH_hca_ul_resources_t*)hca_ul_resources_p)->pci_bus= xhh_hob_p->pci_hca_info.bus;
  ((XHH_hca_ul_resources_t*)hca_ul_resources_p)->pci_devfn= xhh_hob_p->pci_hca_info.dev_func;
  ((XHH_hca_ul_resources_t*)hca_ul_resources_p)->pci_device_p= xhh_hob_p->pci_hca_info.pci_dev;
  memcpy(&(((XHH_hca_ul_resources_t*)hca_ul_resources_p)->version),
    &(xhh_hob_p->version_info),sizeof(XHH_ver_info_t));
  ((XHH_hca_ul_resources_t*)hca_ul_resources_p)->log2_mpt_size = (u_int32_t)xhh_hob_p->profile.log2_max_mpt_entries;
  ((XHH_hca_ul_resources_t*)hca_ul_resources_p)->max_qp_ous_wr= xhh_hob_p->hca_capabilities.max_qp_ous_wr;
  ((XHH_hca_ul_resources_t*)hca_ul_resources_p)->max_srq_ous_wr= xhh_hob_p->hca_capabilities.max_wqe_per_srq;
  ((XHH_hca_ul_resources_t*)hca_ul_resources_p)->max_num_sg_ent= xhh_hob_p->hca_capabilities.max_num_sg_ent;
  ((XHH_hca_ul_resources_t*)hca_ul_resources_p)->max_num_sg_ent_srq= xhh_hob_p->hca_capabilities.max_srq_sentries;
  ((XHH_hca_ul_resources_t*)hca_ul_resources_p)->max_num_sg_ent_rd= xhh_hob_p->hca_capabilities.max_num_sg_ent_rd;
  ((XHH_hca_ul_resources_t*)hca_ul_resources_p)->max_num_ent_cq= xhh_hob_p->hca_capabilities.max_num_ent_cq;
  ((XHH_hca_ul_resources_t*)hca_ul_resources_p)->max_kernel_avs= xhh_hob_p->hca_capabilities.max_ah_num;

  ((XHH_hca_ul_resources_t*)hca_ul_resources_p)->max_wqe_size_sq = xhh_hob_p->dev_lims.max_desc_sz_sq;
  ((XHH_hca_ul_resources_t*)hca_ul_resources_p)->max_wqe_size_rq = xhh_hob_p->dev_lims.max_desc_sz_rq;
  ((XHH_hca_ul_resources_t*)hca_ul_resources_p)->max_wqe_size_srq = xhh_hob_p->dev_lims.max_desc_sz_rq;
  MTL_DEBUG1(MT_FLFMT("%s: max_wqe_size_sq=0x%x, max_wqe_size_rq=0x%x, max_wqe_size_srq = 0x%x"),
	     __func__, ((XHH_hca_ul_resources_t*)hca_ul_resources_p)->max_wqe_size_sq,
	     ((XHH_hca_ul_resources_t*)hca_ul_resources_p)->max_wqe_size_rq,
	     ((XHH_hca_ul_resources_t*)hca_ul_resources_p)->max_wqe_size_srq);

  memcpy(((XHH_hca_ul_resources_t*)hca_ul_resources_p)->vsd, xhh_hob_p->adapter_props.vsd, (XHH_ADAPTER_VSD_DW << 2));

  /* Invoke XHH_uldm in order to get a UAR resource */
  return XHH_uarm_alloc_ul_res(xhh_hob_p->uarm, prot_ctx,  
		  (XHH_hca_ul_resources_t *)hca_ul_resources_p);
} /* XHH_alloc_ul_resources */

/******************************************************************************
 *  Function:     XHH_hob_free_ul_res <==> XHH_uldm_free_ul_res
 *****************************************************************************/
HH_ret_t XHH_hob_free_ul_res(HH_hca_hndl_t  hca_hndl,
                                    void           *hca_ul_resources_p)
{
    XHH_hca_ul_resources_t*  res = (XHH_hca_ul_resources_t *)hca_ul_resources_p;
    XHH_hob_t  xhh_hob_p;

    MT_RETURN_IF_LOW_STACK(XHH_WATERMARK);
    if (MOSAL_get_exec_ctx() != MOSAL_IN_TASK) {
        MTL_ERROR1("XHH_hob_free_ul_res: NOT IN TASK CONTEXT)\n");
        return HH_ERR;
    }

    if (hca_hndl == NULL) {
        MTL_ERROR1("XHH_hob_free_ul_res : ERROR : Invalid HCA handle\n");
        return HH_EINVAL_HCA_HNDL;
    }
    xhh_hob_p = XHHOBP(hca_hndl);

    if (xhh_hob_p == NULL) {
        MTL_ERROR1("XHH_hob_free_ul_res: ERROR : No device registered\n");
        return HH_EAGAIN;
    }

    if (xhh_hob_p->uarm == (XHH_uarm_t)XHH_INVALID_HNDL) {
        MTL_ERROR1("XHH_hob_free_ul_res: ERROR : HCA device has not yet been opened\n");
        return HH_EAGAIN;
    }

    return XHH_uarm_free_ul_res(xhh_hob_p->uarm, res);
} /* XHH_free_ul_resources */




/******************************************************************************
 *  Function:     XHH_hob_alloc_pd <==> XHH_uldm_alloc_pd
 *****************************************************************************/
HH_ret_t XHH_hob_alloc_pd(HH_hca_hndl_t hca_hndl, 
                             MOSAL_protection_ctx_t prot_ctx, 
                             void * pd_ul_resources_p,
                             HH_pd_hndl_t *pd_num_p)
{
  XHH_hob_t xhh_hob_p;
  HH_ret_t  ret;

  MT_RETURN_IF_LOW_STACK(XHH_WATERMARK);
  if (MOSAL_get_exec_ctx() != MOSAL_IN_TASK) {
      MTL_ERROR1("XHH_hob_alloc_pd: NOT IN TASK CONTEXT)\n");
      return HH_ERR;
  }

  if (hca_hndl == NULL) {
      MTL_ERROR1("XHH_hob_alloc_pd : ERROR : Invalid HCA handle\n");
        return HH_EINVAL_HCA_HNDL;
  }
  xhh_hob_p = XHHOBP(hca_hndl);

  if (xhh_hob_p == NULL) {
      MTL_ERROR1("XHH_hob_alloc_pd: ERROR : No device registered\n");
      return HH_EAGAIN;
  }
  TEST_RETURN_FATAL(xhh_hob_p);


  if (xhh_hob_p->pdm == (XHH_pdm_t)XHH_INVALID_HNDL) {
      MTL_ERROR1("XHH_hob_alloc_pd: ERROR : HCA device has not yet been opened\n");
      return HH_EAGAIN;
  }

  ret = XHH_pdm_alloc_pd(xhh_hob_p->pdm, prot_ctx, 
		  (XHH_pd_ul_resources_t *)pd_ul_resources_p, pd_num_p);
//  MTL_DEBUG4("XHH_hob_alloc_pd: ret = %d\n", ret);
  return ret;
}

/******************************************************************************
 *  Function:     XHH_hob_free_pd <==> XHH_uldm_free_pd
 *****************************************************************************/
HH_ret_t XHH_hob_free_pd(HH_hca_hndl_t hca_hndl, HH_pd_hndl_t pd_num)
{
  u_int32_t   max_pd;
  XHH_hob_t   xhh_hob_p;
  HH_ret_t    rc = HH_OK;

  MT_RETURN_IF_LOW_STACK(XHH_WATERMARK);
  if (MOSAL_get_exec_ctx() != MOSAL_IN_TASK) {
      MTL_ERROR1("XHH_hob_free_pd: NOT IN TASK CONTEXT)\n");
      return HH_ERR; 
  }

  if (hca_hndl == NULL) {
      MTL_ERROR1("XHH_hob_free_pd : ERROR : Invalid HCA handle\n");
      return HH_EINVAL_HCA_HNDL; 
  }
  xhh_hob_p = XHHOBP(hca_hndl);

  if (xhh_hob_p == NULL) {
      MTL_ERROR1("XHH_hob_free_pd: ERROR : No device registered\n");
      return HH_EAGAIN;
  }

  max_pd = (1 << xhh_hob_p->dev_lims.log_max_pd);
  if (pd_num > max_pd  - 1) {
    MTL_ERROR1("XHH_hob_free_pd: ERROR : PD number (%d) is greater than max allowed (%d)\n", pd_num, max_pd);
    return HH_EAGAIN;
  }

  if (xhh_hob_p->pdm == (XHH_pdm_t)XHH_INVALID_HNDL) {
      MTL_ERROR1("XHH_hob_free_pd: ERROR : HCA device has not yet been opened\n");
      return HH_EAGAIN;
  }
  
  rc = XHH_pdm_free_pd(xhh_hob_p->pdm, pd_num);
  return rc; 
}


/******************************************************************************
 *  Function:     XHH_hob_alloc_rdd <==> XHH_eecm_alloc_rdd
 *****************************************************************************/
HH_ret_t XHH_hob_alloc_rdd(HH_hca_dev_t *hh_dev_p, 
                                 HH_rdd_hndl_t *rdd_p)
{
#if 0
  XHH_hob_t  xhh_hob_p;

  MT_RETURN_IF_LOW_STACK(XHH_WATERMARK);
  if (MOSAL_get_exec_ctx() != MOSAL_IN_TASK) {
      MTL_ERROR1("XHH_hob_alloc_rdd: NOT IN TASK CONTEXT)\n");
      return HH_ERR;
  }

  if (hca_hndl == NULL) {
      MTL_ERROR1("XHH_hob_alloc_rdd : ERROR : Invalid HCA handle\n");
      return HH_EINVAL_HCA_HNDL;
  }
  xhh_hob_p = XHHOBP(hca_hndl);

  if (xhh_hob_p == NULL) {
      MTL_ERROR1("XHH_hob_alloc_rdd: ERROR : No device registered\n");
      return HH_EAGAIN;
  }
  TEST_RETURN_FATAL(xhh_hob_p);


  if (xhh_hob_p->eecm == (XHH_eecm_t)XHH_INVALID_HNDL) {
      MTL_ERROR1("XHH_hob_alloc_rdd: ERROR : HCA device has not yet been opened\n");
      return HH_EAGAIN;
  }

  return XHH_eecm_alloc_rdd(xhh_hob_p->uldm, rdd_p);
#else
  return HH_ENOSYS;
#endif
}

/******************************************************************************
 *  Function:     XHH_hob_free_rdd <==> XHH_eecm_free_rdd
 *****************************************************************************/
HH_ret_t XHH_hob_free_rdd(HH_hca_dev_t *hh_dev_p, HH_rdd_hndl_t rdd)
{
#if 0
  XHH_hob_t  xhh_hob_p;
  HH_ret_t   rc = HH_OK;

  if (MOSAL_get_exec_ctx() != MOSAL_IN_TASK) {
      MTL_ERROR1("XHH_hob_free_rdd: NOT IN TASK CONTEXT)\n");
      return HH_ERR;
  }

  if (hca_hndl == NULL) {
      MTL_ERROR1("XHH_hob_free_rdd : ERROR : Invalid HCA handle\n");
      return HH_EINVAL_HCA_HNDL;
  }
  xhh_hob_p = XHHOBP(hca_hndl);

  if (xhh_hob_p == NULL) {
      MTL_ERROR1("XHH_hob_free_rdd: ERROR : No device registered\n");
      return HH_EAGAIN;
  }


  if (xhh_hob_p->eecm == (XHH_eecm_t)XHH_INVALID_HNDL) {
      MTL_ERROR1("XHH_hob_free_rdd: ERROR : HCA device has not yet been opened\n");
      return HH_EAGAIN;
  }

  rc = XHH_eecm_free_rdd(xhh_hob_p->uldm, rdd);
  return rc;
#else
  return HH_ENOSYS;
#endif
}

  /* Fast Memory Regions */
  /***********************/
/******************************************************************************
 *  Function:     XHH_hob_create_cq <==> XHH_mrwm_alloc_fmr
 *****************************************************************************/
HH_ret_t  XHH_hob_alloc_fmr(HH_hca_hndl_t  hca_hndl,
                            HH_pd_hndl_t   pd,
                            VAPI_mrw_acl_t acl, 
                            MT_size_t      max_pages,      /* Maximum number of pages that can be mapped using this region */
                            u_int8_t       log2_page_sz,	 /* Fixed page size for all maps on a given FMR */
                            VAPI_lkey_t*   last_lkey_p)    /* To be used as the initial FMR handle */
{
  XHH_hob_t  xhh_hob_p;

  MT_RETURN_IF_LOW_STACK(XHH_WATERMARK);
  if (hca_hndl == NULL) {
      MTL_ERROR1("XHH_hob_alloc_fmr : ERROR : Invalid HCA handle\n");
      return HH_EINVAL_HCA_HNDL;
  }
  xhh_hob_p = XHHOBP(hca_hndl);

  if (xhh_hob_p == NULL) {
      MTL_ERROR1("XHH_hob_alloc_fmr: ERROR : No device registered\n");
      return HH_EAGAIN;
  }
  TEST_RETURN_FATAL(xhh_hob_p);


  if (xhh_hob_p->mrwm == (XHH_mrwm_t)XHH_INVALID_HNDL) {
      MTL_ERROR1("XHH_hob_alloc_fmr: ERROR : HCA device has not yet been opened\n");
      return HH_EAGAIN;
  }

  return XHH_mrwm_alloc_fmr(xhh_hob_p->mrwm,pd,acl,max_pages,log2_page_sz,last_lkey_p);
}

/******************************************************************************
 *  Function:     XHH_hob_map_fmr <==> XHH_mrwm_map_fmr
 *****************************************************************************/
HH_ret_t  XHH_hob_map_fmr(HH_hca_hndl_t  hca_hndl,
                        VAPI_lkey_t      last_lkey,
                        EVAPI_fmr_map_t* map_p,
                        VAPI_lkey_t*     lkey_p,
                        IB_rkey_t*       rkey_p)

{
  XHH_hob_t  xhh_hob_p;

  MT_RETURN_IF_LOW_STACK(XHH_WATERMARK);
  if (hca_hndl == NULL) {
      MTL_ERROR1("XHH_hob_map_fmr : ERROR : Invalid HCA handle\n");
      return HH_EINVAL_HCA_HNDL;
  }
  xhh_hob_p = XHHOBP(hca_hndl);

  if (xhh_hob_p == NULL) {
      MTL_ERROR1("XHH_hob_map_fmr: ERROR : No device registered\n");
      return HH_EAGAIN;
  }
  TEST_RETURN_FATAL(xhh_hob_p);


  if (xhh_hob_p->mrwm == (XHH_mrwm_t)XHH_INVALID_HNDL) {
      MTL_ERROR1("XHH_hob_map_fmr: ERROR : HCA device has not yet been opened\n");
      return HH_EAGAIN;
  }

  return XHH_mrwm_map_fmr(xhh_hob_p->mrwm,last_lkey,map_p,lkey_p,rkey_p);
}
  
/******************************************************************************
 *  Function:     XHH_hob_unmap_fmr <==> XHH_mrwm_unmap_fmr
 *****************************************************************************/
HH_ret_t  XHH_hob_unmap_fmr(HH_hca_hndl_t hca_hndl,
                          u_int32_t     num_of_fmrs_to_unmap,
                          VAPI_lkey_t*  last_lkeys_array)

{
  XHH_hob_t  xhh_hob_p;

  MT_RETURN_IF_LOW_STACK(XHH_WATERMARK);
  if (hca_hndl == NULL) {
      MTL_ERROR1("XHH_hob_unmap_fmr : ERROR : Invalid HCA handle\n");
      return HH_EINVAL_HCA_HNDL;
  }
  xhh_hob_p = XHHOBP(hca_hndl);

  if (xhh_hob_p == NULL) {
      MTL_ERROR1("XHH_hob_unmap_fmr: ERROR : No device registered\n");
      return HH_EAGAIN;
  }
  TEST_RETURN_FATAL(xhh_hob_p);


  if (xhh_hob_p->mrwm == (XHH_mrwm_t)XHH_INVALID_HNDL) {
      MTL_ERROR1("XHH_hob_unmap_fmr: ERROR : HCA device has not yet been opened\n");
      return HH_EAGAIN;
  }

  return XHH_mrwm_unmap_fmr(xhh_hob_p->mrwm,num_of_fmrs_to_unmap,last_lkeys_array);
}

/******************************************************************************
 *  Function:     XHH_hob_free_fmr <==> XHH_mrwm_free_fmr
 *****************************************************************************/
HH_ret_t  XHH_hob_free_fmr(HH_hca_hndl_t  hca_hndl,
                           VAPI_lkey_t      last_lkey)   /* as returned on last successful mapping operation */

{
  XHH_hob_t  xhh_hob_p;
  HH_ret_t   rc = HH_OK;

  MT_RETURN_IF_LOW_STACK(XHH_WATERMARK);
  if (hca_hndl == NULL) {
      MTL_ERROR1("XHH_hob_free_fmr : ERROR : Invalid HCA handle\n");
      return HH_EINVAL_HCA_HNDL;
  }
  xhh_hob_p = XHHOBP(hca_hndl);

  if (xhh_hob_p == NULL) {
      MTL_ERROR1("XHH_hob_free_fmr: ERROR : No device registered\n");
      return HH_EAGAIN;
  }


  if (xhh_hob_p->mrwm == (XHH_mrwm_t)XHH_INVALID_HNDL) {
      MTL_ERROR1("XHH_hob_free_fmr: ERROR : HCA device has not yet been opened\n");
      return HH_EAGAIN;
  }

  rc = XHH_mrwm_free_fmr(xhh_hob_p->mrwm,last_lkey);
  return rc;
}

/******************************************************************************
 *  Function:     XHH_hob_set_comp_eventh <==> 
 *****************************************************************************/
HH_ret_t XHH_hob_set_comp_eventh(HH_hca_hndl_t      hca_hndl,
                                        HH_comp_eventh_t   event,
                                        void*              private_data)
{
  XHH_eventp_handler_t ev_hndlr;
  
  XHH_hob_t  xhh_hob_p;

  MT_RETURN_IF_LOW_STACK(XHH_WATERMARK);
  if (MOSAL_get_exec_ctx() != MOSAL_IN_TASK) {
      MTL_ERROR1("XHH_hob_set_comp_eventh: NOT IN TASK CONTEXT)\n");
      return HH_ERR;
  }

  if (hca_hndl == NULL) {
      MTL_ERROR1("XHH_hob_set_comp_eventh : ERROR : Invalid HCA handle\n");
      return HH_EINVAL_HCA_HNDL;
  }
  xhh_hob_p = XHHOBP(hca_hndl);
  if (xhh_hob_p == NULL) {
      MTL_ERROR1("XHH_hob_set_comp_eventh: ERROR : No device registered\n");
      return HH_EAGAIN;
  }

  if (event == NULL) {
      event = XHH_dummy_comp_event;
  } else {
      TEST_RETURN_FATAL(xhh_hob_p);
  }
  ev_hndlr.comp_event_h = event;


  if (xhh_hob_p->eventp == (XHH_eventp_t)XHH_INVALID_HNDL) {
      MTL_ERROR1("XHH_hob_set_comp_eventh: ERROR : HCA device has not yet been opened\n");
      return HH_EAGAIN;
  }

  return XHH_eventp_replace_handler(xhh_hob_p->eventp,xhh_hob_p->compl_eq, ev_hndlr, private_data);

}

HH_ret_t XHH_hob_set_async_eventh( HH_hca_hndl_t      hca_hndl,
                                          HH_async_eventh_t  event,
                                          void*              private_data)
{
  XHH_eventp_handler_t ev_hndlr;
  XHH_hob_t  xhh_hob_p;
  HH_ret_t ret = HH_OK;

  MT_RETURN_IF_LOW_STACK(XHH_WATERMARK);
  if (MOSAL_get_exec_ctx() != MOSAL_IN_TASK) {
      MTL_ERROR1("XHH_hob_set_async_eventh: NOT IN TASK CONTEXT)\n");
      return HH_ERR;
  }

  if (hca_hndl == NULL) {
      MTL_ERROR1("XHH_hob_set_async_eventh : ERROR : Invalid HCA handle\n");
      return HH_EINVAL_HCA_HNDL;
  }
  xhh_hob_p = XHHOBP(hca_hndl);

  if (event == NULL) {event = &XHH_dummy_async_event;
  } else {
      TEST_RETURN_FATAL(xhh_hob_p);
  }
  
  ev_hndlr.ib_comp_event_h = event;

  if (xhh_hob_p == NULL) {
      MTL_ERROR1("XHH_hob_set_async_eventh: ERROR : No device registered\n");
      return HH_EAGAIN;
  }

  if (xhh_hob_p->eventp == (XHH_eventp_t)XHH_INVALID_HNDL) {
      MTL_ERROR1("XHH_hob_set_async_eventh: ERROR : HCA device has not yet been opened\n");
      return HH_EAGAIN;
  }

  ret = XHH_eventp_replace_handler(xhh_hob_p->eventp,xhh_hob_p->ib_eq, ev_hndlr, private_data);
  if (ret != HH_OK) {
      MTL_ERROR1( "XHH_hob_set_async_eventh: ERROR : cannot register async event handler (%d)\n", ret);
      return ret;
  }


  /* track async event handler setting for use in fatal error handling */
  MOSAL_spinlock_irq_lock(&xhh_hob_p->async_spl);
  xhh_hob_p->async_eventh = event;
  xhh_hob_p->async_ev_private_context = private_data;
  MOSAL_spinlock_unlock(&xhh_hob_p->async_spl);

  return HH_OK;
}

HH_ret_t XHH_get_debug_info(
	HH_hca_hndl_t hca_hndl, 		/*IN*/
	XHH_debug_info_t *debug_info_p	/*OUT*/
)
{
    XHH_hob_t  hob_p;
    HH_ret_t   rc = HH_OK;
    MT_bool    have_error = FALSE;

    MT_RETURN_IF_LOW_STACK(XHH_WATERMARK);
    if (MOSAL_get_exec_ctx() != MOSAL_IN_TASK) {
        MTL_ERROR1("XHH_get_debug_info: NOT IN TASK CONTEXT)\n");
        return HH_ERR;
    }

    if (hca_hndl == NULL) {
        MTL_ERROR1("XHH_get_debug_info : ERROR : Invalid HCA handle\n");
        return HH_EINVAL_HCA_HNDL;
    }
    hob_p = XHHOBP(hca_hndl);

    if (hob_p == NULL) {
        MTL_ERROR1("XHH_get_debug_info: ERROR : No device registered\n");
        return HH_EAGAIN;
    }

    memset(debug_info_p, 0, sizeof(XHH_debug_info_t));
    
    memcpy(&(debug_info_p->hw_props), &(hob_p->hw_props), sizeof(XHH_hw_props_t));
    memcpy(&(debug_info_p->profile), &(hob_p->profile), sizeof(XHH_profile_t));
    memcpy(&(debug_info_p->icm_addr_vec), &(hob_p->icm_alloc_base_addrs_vec), 
           sizeof(XHH_icm_base_addr_vector_t));
    memcpy(&(debug_info_p->icm_size_vec), &(hob_p->icm_alloc_size_vec), 
           sizeof(XHH_icm_allocation_vector_t));
    debug_info_p->num_icm_addrs = XHH_ICM_ALLOCATION_VEC_SIZE;
    memcpy(&(debug_info_p->mrwm_props), &(hob_p->mrwm_props), sizeof(XHH_mrwm_props_t));

/*    debug_info_p->hide_ddr = hob_p->lam_props.dh; */

#if 0
    rc = XHH_uarm_get_num_objs(hob_p->uldm,&(debug_info_p->allocated_ul_res),&(debug_info_p->allocated_pd));
    if (rc != HH_OK) {
        MTL_ERROR1(MT_FLFMT("XHH_get_debug_info:  proc XHH_uldm_get_num_of_objs returned ERROR"));
        have_error = TRUE;
    }
#endif
    
    rc = XHH_mrwm_get_num_objs(hob_p->mrwm,&(debug_info_p->allocated_mr_int),
                                  &(debug_info_p->allocated_mr_ext), &(debug_info_p->allocated_mw));
    if (rc != HH_OK) {
        MTL_ERROR1(MT_FLFMT("XHH_get_debug_info:  proc XHH_mrwm_get_num_of_objs returned ERROR"));
        have_error = TRUE;
    }

    rc = XHH_qpm_get_num_qps(hob_p->qpm,&(debug_info_p->allocated_qp));
    if (rc != HH_OK) {
        MTL_ERROR1(MT_FLFMT("XHH_get_debug_info:  proc XHH_tqpm_get_num_of_qps returned ERROR"));
        have_error = TRUE;
    }
    
    rc = XHH_cqm_get_num_cqs(hob_p->cqm,&(debug_info_p->allocated_cq));
    if (rc != HH_OK) {
        MTL_ERROR1(MT_FLFMT("XHH_get_debug_info:  proc XHH_tcqm_get_num_of_cqs returned ERROR"));
        have_error = TRUE;
    }
    
    rc = XHH_mcgm_get_num_mcgs(hob_p->mcgm,&(debug_info_p->allocated_mcg));
    if (rc != HH_OK) {
        MTL_ERROR1(MT_FLFMT("XHH_get_debug_info:  proc XHH_mcgm_get_num_of_mcgs returned ERROR"));
        have_error = TRUE;
    }

    return (have_error ? HH_ERR : HH_OK);
}


HH_ret_t XHH_hob_get_lamm ( /*IN*/ XHH_hob_t hob, 
                            /*OUT*/ XHH_lamm_t *lamm_p )
{
  MT_RETURN_IF_LOW_STACK(XHH_WATERMARK);
    if (hob == NULL) {
        MTL_ERROR1("XHH_hob_get_lamm: ERROR : No device registered\n");
        return HH_EINVAL;
    }

    *lamm_p = hob->lamm;
    return HH_OK;
}

/* the PD used by the driver is the first PD following the reserved PDs (which start numbering at 0) */
u_int32_t XHH_get_rsvd_pd_num(XHH_hob_t hob)
{
    return ( (u_int32_t) hob->dev_lims.num_rsvd_pds);
}
HH_ret_t XHH_hob_check_qp_init_attrs ( /*IN*/ XHH_hob_t hob,
                                       /*IN*/ HH_qp_init_attr_t * init_attr_p,
                                       /*IN*/ MT_bool is_special_qp )
{
  MT_RETURN_IF_LOW_STACK(XHH_WATERMARK);
    if (hob == NULL) {
        MTL_ERROR1("XHH_hob_get_check_qp_init_attrs: ERROR : No device registered\n");
        return HH_EINVAL;
    }


    if (init_attr_p == NULL) {
        MTL_ERROR1("XHH_hob_get_check_qp_init_attrs: ERROR : null attributes\n");
        return HH_EINVAL;
    }

    if (init_attr_p->qp_cap.max_oust_wr_rq > (u_int32_t)(1<<hob->dev_lims.log_max_qp_sz)  ||
        init_attr_p->qp_cap.max_oust_wr_sq > (u_int32_t)(1<<hob->dev_lims.log_max_qp_sz)) {
        MTL_ERROR1(MT_FLFMT("%s : max outs work requests too large. RQ out=%u, SQ out=%u, max=%d"),
			     __func__, init_attr_p->qp_cap.max_oust_wr_rq, 
			    init_attr_p->qp_cap.max_oust_wr_sq, (1<<hob->dev_lims.log_max_qp_sz));
        return HH_E2BIG_WR_NUM;
    }

    if (is_special_qp  || init_attr_p->ts_type != VAPI_TS_RD) {
        if (init_attr_p->qp_cap.max_sg_size_rq > hob->hca_capabilities.max_num_sg_ent ||
                   init_attr_p->qp_cap.max_sg_size_sq > hob->hca_capabilities.max_num_sg_ent) {
            MTL_ERROR1("%s : max s/g list size more than HCA maximum\n", __func__);
            return HH_E2BIG_SG_NUM;
        }
    } else {
        /* is RD */
        if (init_attr_p->qp_cap.max_sg_size_rq > hob->hca_capabilities.max_num_sg_ent_rd ||
                   init_attr_p->qp_cap.max_sg_size_sq > hob->hca_capabilities.max_num_sg_ent_rd) {
            MTL_ERROR1("%s : max s/g list size more than HCA maximum\n", __func__);
            return HH_E2BIG_SG_NUM;
        } 
    }
    return HH_OK;
}

