/*                  - Mellanox Confidential and Proprietary -
 *
 *  Copyright (C) April 2002, Mellanox Technologies Ltd.  ALL RIGHTS RESERVED.
 *
 *  Except as specifically permitted herein, no portion of the information,
 *  including but not limited to object code and source code, may be reproduced,
 *  modified, distributed, republished or otherwise exploited in any form or by
 *  any means for any purpose without the prior written permission of Mellanox
 *  Technologies Ltd. Use of software subject to the terms and conditions
 *  detailed in the file "LICENSE.txt".
 *
 *  End of legal section ......................................................
 *
 */
 
/*
*-------------------------------------------------------------------
* Description: CM functions over TCP/IP Socket
*-------------------------------------------------------------------								 
* Version : $Id: sock.c,v 1.20 2003/12/03 09:44:04 leonid Exp $
* Author  : Kholodenko Alex  ( xol@mellanox.co.il)
*-------------------------------------------------------------------
*/
 
#include <unistd.h>
#include <signal.h>

#include <stdlib.h>

#ifdef WIN32
#include <WINSOCK.h>
#else
#include <sys/types.h>
#include <sys/socket.h>
#include <netdb.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <arpa/inet.h>
#endif

#include <string.h>
#include <stdio.h>
#include <errno.h>

#include "sock.h"
#include <mtl_common.h>
#include <mosal.h>
s_int_ptr_t socket_desc;
int wait_accept;

/*
*-----------------------------------------------------------------------------
*		 		sock_cm_connect
*-----------------------------------------------------------------------------
* Dscr  :  Establish tcpip socket connection and exchange CM info
*-----------------------------------------------------------------------------
* Input : is_daemon		- daemon mode
*		: tcp_port	   	- socket port
*		: ip			- ip address of daemon ( clien only )
*		: local			- local IB CM parameters	
* Output: remote		- remote peer IB CM parameters
*		: p_cm_sock		- created socket handler
* Notes : 
*-----------------------------------------------------------------------------
*/
int sock_cm_connect( int is_daemon, int tcp_port, char *ip, 
				struct test_cm_st* local, struct test_cm_st* remote ,
				s_int_ptr_t* p_cm_sock )
{
	int rc=0;

	if( is_daemon )	
	{
		//
		// Daemon mode
		//

		*p_cm_sock = wait_for_connection( tcp_port );
		if( *p_cm_sock == -1 )
		{
			printf( "Can't connect to TCPIP Socket\n" );
			return -1;
		}

		// Exchange CM info
		rc = sock_cm_recv( *p_cm_sock, (char*)remote, sizeof( struct test_cm_st ) );
		if( rc )
			return rc;
		rc = sock_cm_send( *p_cm_sock, (char*)local,  sizeof( struct test_cm_st ) );
		if( rc )						 
			return rc;
	}
	else
	{
		//
		// Client Mode
		//

		*p_cm_sock = create_tcp_connection( ip, tcp_port );
		if( *p_cm_sock == -1 )
		{
			printf( "Can't Create TCPIP Socket\n" );
			return -1;
		}

		// Exchange CM info
		rc = sock_cm_send( *p_cm_sock, (char*)local, sizeof( struct test_cm_st ) );
		if( rc )
			return rc;
		rc = sock_cm_recv( *p_cm_sock, (char*)remote, sizeof( struct test_cm_st ) );
		if( rc )
			return rc;
	}

    return rc;
}


/*
*-----------------------------------------------------------------------------
*		 		sock_cm_ready_sync
*-----------------------------------------------------------------------------
* Dscr  :  Sync sides
*-----------------------------------------------------------------------------
* Input : is_daemon	- daemon mode
* Output: 0-OK, else error code
* Notes : 
*-----------------------------------------------------------------------------
*/
int sock_cm_ready_sync( s_int_ptr_t cm_sock, int is_daemon )
{
	char cm_buf[1];
	int rc=0;

	if( is_daemon )
	{
		cm_buf[0] = 'R';
		rc = sock_cm_send( cm_sock, cm_buf, 1 );	
		if( rc )
			return rc;
		rc = sock_cm_recv( cm_sock, cm_buf, 1 ); 	
		if( rc )
			return rc;
	}
	else
	{
		rc = sock_cm_recv( cm_sock, cm_buf, 1 ); 	
		if( rc )
			return rc;
		rc = sock_cm_send( cm_sock, cm_buf, 1 );	
		if( rc )
			return rc;
	}
	return rc;
}


/*
*-----------------------------------------------------------------------------
*		 		sock_cm_addr_sync
*-----------------------------------------------------------------------------
* Dscr  :  Exchange address information 
*-----------------------------------------------------------------------------
* Input : is_daemon	- daemon mode
*
* Output: 0-OK, else error code
* Notes : 
*-----------------------------------------------------------------------------
*/
int sock_cm_addr_sync( s_int_ptr_t cm_sock, int is_daemon, 
									VAPI_mrw_t* p_mrw_local, VAPI_mrw_t* p_mrw_remote )
{
	int rc=0;

    u_int64_t tmp64;
    u_int32_t tmp32;

    //MR Size
    tmp64 = p_mrw_local->size;
    rc = sock_cm_uint64_swap(cm_sock, is_daemon, &tmp64);
    if( rc ) {
        return rc;
    }

    p_mrw_remote->size = tmp64;

    //MR Start
    tmp64 = p_mrw_local->start;
    rc = sock_cm_uint64_swap(cm_sock, is_daemon, &tmp64);
    if( rc ) {
        return rc;
    }

    p_mrw_remote->start = tmp64;

    //L_Key
    tmp32 = p_mrw_local->l_key;
    rc = sock_cm_uint32_swap(cm_sock, is_daemon, &tmp32);
    if( rc ) {
        return rc;
    }

    p_mrw_remote->l_key = tmp32;

    //R_Key
    tmp32 = p_mrw_local->r_key;
    rc = sock_cm_uint32_swap(cm_sock, is_daemon, &tmp32);
    if( rc ) {
        return rc;
    }

    p_mrw_remote->r_key = tmp32;

	return rc;
}

int sock_cm_uint8_swap(s_int_ptr_t cm_sock, int is_daemon, 
                       u_int8_t* buff)
{
    int rc=0;

    if( is_daemon )
    {
        rc = sock_cm_send( cm_sock, (void*)buff, 1);	
        if( rc )
            return rc;
        rc = sock_cm_recv( cm_sock, (void*)buff, 1); 	
        if( rc )
            return rc;
    }
    else
    {
        u_int8_t tmp;

        memcpy(&tmp, buff, 1);

        rc = sock_cm_recv( cm_sock, (void*)buff, 1 ); 	
        if( rc )
            return rc;
        rc = sock_cm_send( cm_sock, (void*)(&tmp), 1);	
        if( rc )
            return rc;
    }

    return(rc);
}

int sock_cm_uint16_swap(s_int_ptr_t cm_sock, int is_daemon, 
                        u_int16_t* buff)
{
    int rc=0;

    if( is_daemon )
    {
        u_int16_t tmp = MOSAL_cpu_to_be16(*buff);

        rc = sock_cm_send( cm_sock, (void*)(&tmp), 2);	
        if( rc )
            return rc;

        rc = sock_cm_recv( cm_sock, (void*)(&tmp), 2); 	
        if( rc )
            return rc;

        *buff = MOSAL_be16_to_cpu(tmp);
    }
    else
    {
        u_int16_t tmp = MOSAL_cpu_to_be16(*buff);

        rc = sock_cm_recv( cm_sock, (void*)buff, 2 ); 	
        if( rc )
            return rc;

        *buff = MOSAL_be16_to_cpu(*buff);

        rc = sock_cm_send( cm_sock, (void*)(&tmp), 2);	
        if( rc )
            return rc;
    }

    return(rc);
}


int sock_cm_uint32_swap(s_int_ptr_t cm_sock, int is_daemon, 
                        u_int32_t* buff)
{
    int rc=0;

    if( is_daemon )
    {
        u_int32_t tmp = MOSAL_cpu_to_be32(*buff);

        rc = sock_cm_send( cm_sock, (void*)(&tmp), 4);	
        if( rc )
            return rc;

        rc = sock_cm_recv( cm_sock, (void*)(&tmp), 4); 	
        if( rc )
            return rc;

        *buff = MOSAL_be32_to_cpu(tmp);
    }
    else
    {
        u_int32_t tmp = MOSAL_cpu_to_be32(*buff);

        rc = sock_cm_recv( cm_sock, (void*)buff, 4 ); 	
        if( rc )
            return rc;

        *buff = MOSAL_be32_to_cpu(*buff);

        rc = sock_cm_send( cm_sock, (void*)(&tmp), 4);	
        if( rc )
            return rc;
    }

    return(rc);
}


int sock_cm_uint64_swap(s_int_ptr_t cm_sock, int is_daemon, 
                        u_int64_t* buff)
{
    int rc=0;

    if( is_daemon )
    {
        u_int64_t be64 = MOSAL_cpu_to_be64(*buff);
        //u_int64_t be64 = *buff;
        
        rc = sock_cm_send( cm_sock, (void*)(&be64), 8);	
        if( rc )
            return rc;

        rc = sock_cm_recv( cm_sock, (void*)(&be64), 8); 	
        if( rc )
            return rc;

        *buff = MOSAL_be64_to_cpu(be64);
        //*buff = be64;

    }
    else
    {
        u_int64_t be64 = MOSAL_cpu_to_be64(*buff), be64_rec;
        //u_int64_t be64 = *buff, be64_rec;


        rc = sock_cm_recv( cm_sock, (void*)(&be64_rec), 8 ); 	
        if( rc )
            return rc;

        *buff = MOSAL_be64_to_cpu(be64_rec);

        rc = sock_cm_send( cm_sock, (void*)(&be64), 8);	
        if( rc )
            return rc;
    }

    return(rc);
}

/*
 *  create_tcp_connection
 */ 
s_int_ptr_t create_tcp_connection(const char *ip, int port)
{
	struct in_addr remote_addr;
	s_int_ptr_t s;
	int rc;
	struct sockaddr_in my_addr, server_addr;
    int retry_count = 0;
	
    s = socket(PF_INET, SOCK_STREAM, IPPROTO_TCP);
	if ( s == -1 )
	{
		return -1;
	}

	my_addr.sin_family = PF_INET;
	my_addr.sin_port = htons(port);
	my_addr.sin_addr.s_addr = INADDR_ANY;
#ifdef WIN32
    if ( (remote_addr.S_un.S_addr = inet_addr(ip)) == INADDR_NONE ) 
#else
	if ( inet_aton(ip,  &remote_addr) == 0 )
#endif
	{
		printf("SOCK: inet_aton failed: %s\n", strerror(errno));
		return -1;
	}
	server_addr.sin_family = PF_INET;
	server_addr.sin_port = htons(port);
	server_addr.sin_addr = remote_addr;

	do
    {
        rc = connect(s, (struct sockaddr *)(&server_addr), sizeof(server_addr));
        if (rc == -1)
        {
            usleep(200000);
            retry_count ++;
        }
    } while (( rc == -1 ) && (retry_count < SOCKET_RETRY));
    
	if ( rc == -1 )
	{
		printf("SOCK: connect failed: %s\n", strerror(errno));
		return -1;
	}

    return s;
}



s_int_ptr_t wait_for_connection(int port)
{
	int rc, addr_len;
	s_int_ptr_t newsock;
	struct sockaddr_in my_addr, remote_addr;
	struct protoent *protoent;
	int tmp;

	socket_desc = socket(PF_INET, SOCK_STREAM, 0);
	if ( socket_desc == -1 )
	{
		return -1;
	}

	tmp = 1;
	setsockopt (socket_desc, SOL_SOCKET, SO_REUSEADDR, (char *) &tmp, sizeof (tmp));

	my_addr.sin_family = PF_INET;
	my_addr.sin_port = htons(port);
	my_addr.sin_addr.s_addr = INADDR_ANY;

	rc = bind(socket_desc, (struct sockaddr *)(&my_addr), sizeof(my_addr));
	if ( rc == -1 )
	{
		printf("SOCK: bind failed: %s\n", strerror(errno));
		return -1;
	}

	rc = listen(socket_desc, 1);
	if ( rc == -1 )
	{
		printf("SOCK: listen failed: %s\n", strerror(errno));
		return -1;
	}

	wait_accept = 1;
    
	addr_len = sizeof(remote_addr);
	newsock = accept(socket_desc, (struct sockaddr *)(&remote_addr), &addr_len);
	if ( rc == -1 )
	{
		printf("SOCK: accept failed: %s\n", strerror(errno));
		wait_accept = 0;
		return -1;
	}

	wait_accept = 0;
    

	protoent = getprotobyname ("tcp");
	if (!protoent)
	{
		printf( "SOCK: getprotobyname error\n" );
		return 5;
	}

	/* Tell TCP not to delay small packets.  This greatly speeds up
	   interactive response. */
	tmp = 1;
	setsockopt (newsock, protoent->p_proto, TCP_NODELAY,
											(char *) &tmp, sizeof (tmp));
#ifdef WIN32
    closesocket (socket_desc);		/* No longer need this */
#else
	close (socket_desc);		/* No longer need this */
#endif

#ifndef WIN32
	signal (SIGPIPE, SIG_IGN);	/* If we don't do this, then gdbserver simply
											exits when the remote side dies.  */
#endif

	return newsock;
}


int sock_cm_send( s_int_ptr_t sock, char* buf, unsigned int size )
{
	int rc;

    
    
retry_after_signal:
	rc = send( sock, buf, size, 0 );
	if( rc != size )
	{
		if( (errno == 4) && (rc != 0) )
			goto retry_after_signal;	// Interrupted system call

		printf( "SOCK: send failed(%d):\"%s\" rc=%d\n", errno, strerror(errno), rc );
		if( rc )
			return rc;
		else 
			return -1;
	}
	
    return 0;
}

int sock_cm_recv( s_int_ptr_t sock, char* buf, unsigned int size )
{
	int rc;

    
retry_after_signal:
#ifndef WIN32
	rc = recv( sock, buf, size, MSG_WAITALL );
#else 
    rc = recv( sock, buf, size, 0);
#endif
	if( rc != size )
	{
		printf( "SOCK: recv failed(%d):\"%s\" rc=%d\n", errno, strerror(errno), rc );

		if( (errno == 4) && (rc != 0) )
			goto retry_after_signal;	// Interrupted system call

		if( rc )
			return rc;
		else 
			return -1;
	}

    return 0;
}


/*
*-----------------------------------------------------------------------------
*		 		sock_daemon_wait
*-----------------------------------------------------------------------------
* Dscr  :  Wait connect from client
*-----------------------------------------------------------------------------
* Input : port - TCP/IP port
* Output: socket or error code <0
* Notes : 
*-----------------------------------------------------------------------------
*/
s_int_ptr_t sock_daemon_wait(int port)
{
	int rc, addr_len;
	s_int_ptr_t newsock;
	struct sockaddr_in my_addr, remote_addr;
	struct protoent *protoent;
	int tmp;

	socket_desc = socket(PF_INET, SOCK_STREAM, 0);
	if ( socket_desc == -1 )
	{
		return -1;
	}

	tmp = 1;
	setsockopt (socket_desc, SOL_SOCKET, SO_REUSEADDR, (char *) &tmp, sizeof (tmp));

	my_addr.sin_family = PF_INET;
	my_addr.sin_port = htons(port);
	my_addr.sin_addr.s_addr = INADDR_ANY;

	rc = bind(socket_desc, (struct sockaddr *)(&my_addr), sizeof(my_addr));
	if ( rc == -1 )
	{
		printf("SOCK: bind failed: %s\n", strerror(errno));
		return -1;
	}

    rc = listen(socket_desc, 1);
	if ( rc == -1 )
	{
		printf("SOCK: listen failed: %s\n", strerror(errno));
		return -1;
	}

	wait_accept = 1;
    
	addr_len = sizeof(remote_addr);
	newsock = accept(socket_desc, (struct sockaddr *)(&remote_addr), &addr_len);
	if ( rc == -1 )
	{
		printf("SOCK: accept failed: %s\n", strerror(errno));
		wait_accept = 0;
		return -1;
	}

	wait_accept = 0;
    

	protoent = getprotobyname ("tcp");
	if (!protoent)
	{
		printf( "SOCK: getprotobyname error\n" );
		return 5;
	}

	/* Tell TCP not to delay small packets.  This greatly speeds up
	   interactive response. */
	tmp = 1;
	setsockopt (newsock, protoent->p_proto, TCP_NODELAY,
											(char *) &tmp, sizeof (tmp));

#ifdef WIN32
    closesocket (socket_desc);		/* No longer need this */
#else
	close (socket_desc);		/* No longer need this */
#endif

#ifndef WIN32
	signal (SIGPIPE, SIG_IGN);	/* If we don't do this, then gdbserver simply
											exits when the remote side dies.  */
#endif


	return newsock;
}

/*
*-----------------------------------------------------------------------------
*		 		sock_clien_connect
*-----------------------------------------------------------------------------
* Dscr  :  Connet to socket
*-----------------------------------------------------------------------------
* Input : ip - ip addr of daemon's peer
*		  port - TCP/IP port
* Output: socket or error code <0
* Notes : 
*-----------------------------------------------------------------------------
*/
s_int_ptr_t sock_client_connect(const char *ip, int port)
{
	struct in_addr remote_addr;
	int rc = -1;
	s_int_ptr_t s;
    int i;
	struct sockaddr_in my_addr, server_addr;
	
    s = socket(PF_INET, SOCK_STREAM, IPPROTO_TCP);
	if ( s == -1 )
	{
		return -1;
	}

	my_addr.sin_family = PF_INET;
	my_addr.sin_port = htons(port);
	my_addr.sin_addr.s_addr = INADDR_ANY;


#ifdef WIN32
    if ( (remote_addr.S_un.S_addr = inet_addr(ip)) == INADDR_NONE ) 
#else
	if ( inet_aton(ip,  &remote_addr) == 0 )
#endif
	{
		printf("SOCK: inet_aton failed: %s\n", strerror(errno));
		return -1;
	}
	server_addr.sin_family = PF_INET;
	server_addr.sin_port = htons(port);
	server_addr.sin_addr = remote_addr;

    /*Try 100 seconds to connect*/
    for (i = 0; i < 100 ; i++) {
        rc = connect(s, (struct sockaddr *)(&server_addr), sizeof(server_addr));
    	if ( rc == -1 )
    	{
            MOSAL_sleep(1);
            printf("SOCK: connect[%d] failed: %s\n",i, strerror(errno));
    	} else {
            break;
        }
    }

    if ( rc == -1 )
    {
        printf("SOCK: connect failed: %s\n", strerror(errno));
        return -1;
    }

    return s;
}

/*
*-----------------------------------------------------------------------------
*		 		sock_cm_data_sync
*-----------------------------------------------------------------------------
* Dscr  :  Sync sides and exchange data
*-----------------------------------------------------------------------------
* Input : is_daemon	- daemon mode
* Output: 0-OK, else error code
* Notes : 
*-----------------------------------------------------------------------------
*/
int sock_cm_data_sync( s_int_ptr_t cm_sock, int is_daemon, char* out_buf, char* in_buf, int size )
{
	int rc=0;

	if( is_daemon )
	{
		rc = sock_cm_send( cm_sock, out_buf, size );	
		if( rc )
			return rc;
		rc = sock_cm_recv( cm_sock, in_buf, size ); 	
		if( rc )
			return rc;
	}
	else
	{
		rc = sock_cm_recv( cm_sock, in_buf, size ); 	
		if( rc )
			return rc;
		rc = sock_cm_send( cm_sock, out_buf, size );	
		if( rc )
			return rc;
	}
	return rc;
}

