/*
  This software is available to you under a choice of one of two
  licenses.  You may choose to be licensed under the terms of the GNU
  General Public License (GPL) Version 2, available at
  <http://www.fsf.org/copyleft/gpl.html>, or the OpenIB.org BSD
  license, available in the LICENSE.TXT file accompanying this
  software.  These details are also available at
  <http://openib.org/license.html>.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
  BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
  ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
  CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
  SOFTWARE.

  Copyright (c) 2004 Mellanox Technologies Ltd.  All rights reserved.
*/
 
#include <mtl_common.h>
#include <cr_types.h>
#include <vapi.h>
#include <vapi_common.h>
#include <hh.h>
#include <hob_priv.h>
#include <ul_hob.h>
#include <sm_mad.h>
#include <log2.h>
#include <default_profile.h>
#include <srqm.h>
#include <hob_imp.h>
#include <hob_pci_caps.h>
#include <hob_comm.h>

    /* the define below is to limit the number of QPs per MCG to 8 due to a bug
       must be fixed later.  Comment out this define when the bug is fixed */
#define ADJUST_DEVLIMS_MAX_QP_PER_MCG

#define WAIT_IF_FATAL(hob)    XHH_hob_wait_if_fatal(hob,&have_fatal);
#define CMD_EQ_SIZE 250 /* to be smaller then 8K - so it's EQ can get into physical memory */

#if 4 <= MAX_DEBUG
#define XHH_PRINT_PROFILE(a)      XHH_print_profile(a)
#define XHH_PRINT_USR_PROFILE(a)  XHH_print_usr_profile(a)
#else
#define XHH_PRINT_PROFILE(a)      
#define XHH_PRINT_USR_PROFILE(a)
#endif

/* global reference here for cmdif when putting Outbox in DDR memory*/
#ifdef DEBUG_MEM_OV
#define CMDIF_SIZE_IN_DDR   0x100000    /* allocate 1M in DDR memory*/ 
MT_phys_addr_t cmdif_dbg_ddr; /* address in ddr used for out params in debug mode */
#endif

#define GET_DDR_ADDR(phys_addr,hide_ddr,ddr_base)   ((u_int64_t) ((((hide_ddr) == TRUE)&&(sizeof(MT_phys_addr_t)<=4)) ? \
                                  (((u_int64_t)(phys_addr)) | ((ddr_base) & MAKE_ULONGLONG(0xFFFFFFFF00000000))) : phys_addr))

extern HH_ret_t XHH_hob_waitfor_hca_closed(HH_hca_hndl_t  hca_hndl);

                                                      
static HH_ret_t XHH_hob_query_struct_init(XHH_hob_t  hob, MT_bool have_usr_profile, VAPI_hca_cap_t *hca_cap_p);
HH_ret_t   XHH_hob_halt_hca(/*IN*/ XHH_hob_t hob);


static void XHH_dummy_async_event(HH_hca_hndl_t hca_hndl, HH_event_record_t  *event_rec_p, void* ptr)
{
  /* TBD : This should be an error */
  MTL_TRACE1("XHH_dummy_async_event: called for devices %s with event type 0x%x",
             hca_hndl->dev_desc, event_rec_p->etype);
  return;
}

static void XHH_dummy_comp_event(HH_hca_hndl_t hca_hndl, HH_cq_hndl_t cq_num, void* ptr)
{
  /* TBD : This should be an error */
  MTL_TRACE1("XHH_dummy_comp_event: called for device %s and  cq num 0x%x",
                          hca_hndl->dev_desc, cq_num);
  return;
}

int XHH_hob_fatal_err_thread(void  *arg);

/******************************************************************************
******************************************************************************
************************  INTERNAL FUNCTIONS *********************************
******************************************************************************
******************************************************************************/
#if 4 <= MAX_DEBUG
static void XHH_print_profile(XHH_profile_t *profile) 
{
    MTL_DEBUG1("Profile printout\n");

    MTL_DEBUG1("        ddr_alloc_vec_size = "SIZE_T_DFMT"\n",profile->ddr_alloc_vec_size);
    MTL_DEBUG1("        ddr_size = "SIZE_T_XFMT" ("SIZE_T_DFMT")\n", 
               profile->ddr_size,profile->ddr_size );

    MTL_DEBUG1("        ddr_size_code = %d\n", profile->ddr_size_code);
    
    MTL_DEBUG1("        num_external_mem_regions = "SIZE_T_XFMT" ("SIZE_T_DFMT")\n", 
               profile->num_external_mem_regions,profile->num_external_mem_regions );
    MTL_DEBUG1("        num_mem_windows = "SIZE_T_XFMT" ("SIZE_T_DFMT")\n", 
               profile->num_mem_windows,profile->num_mem_windows);

    MTL_DEBUG1("        log2_max_qps = "SIZE_T_DFMT"\n", profile->log2_max_qps);
    MTL_DEBUG1("        max_num_qps = "SIZE_T_XFMT" ("SIZE_T_DFMT")\n", 
               profile->max_num_qps,profile->max_num_qps);
    MTL_DEBUG1("        log2_max_cqs = "SIZE_T_DFMT"\n", profile->log2_max_cqs);
    MTL_DEBUG1("        max_num_cqs = "SIZE_T_XFMT" ("SIZE_T_DFMT")\n", 
               profile->max_num_cqs,profile->max_num_cqs);
    MTL_DEBUG1("        max_num_pds = "SIZE_T_XFMT" ("SIZE_T_DFMT")\n", 
               profile->max_num_pds,profile->max_num_pds);

    MTL_DEBUG1("        log2_max_mpt_entries = "SIZE_T_DFMT"\n", profile->log2_max_mpt_entries);
    MTL_DEBUG1("        log2_max_mtt_entries = "SIZE_T_DFMT"\n", profile->log2_max_mtt_entries);
    MTL_DEBUG1("        log2_mtt_segs_per_region = "SIZE_T_DFMT"\n", profile->log2_mtt_segs_per_region);
    MTL_DEBUG1("        log2_mtt_entries_per_seg = "SIZE_T_DFMT"\n", profile->log2_mtt_entries_per_seg);

    
    MTL_DEBUG1("        log2_max_uar = "SIZE_T_DFMT"\n", profile->log2_max_uar);
    MTL_DEBUG1("        log2_uar_pg_size = %d\n", profile->log2_uar_pg_size);
    
    MTL_DEBUG1("        log2_wqe_ddr_space_per_qp = "SIZE_T_DFMT"\n",profile->log2_wqe_ddr_space_per_qp);

    MTL_DEBUG1("        use_priv_udav  = %s\n", (profile->use_priv_udav ? "TRUE" : "FALSE"));
    MTL_DEBUG1("        max_priv_udavs = "SIZE_T_XFMT" ("SIZE_T_DFMT")\n", 
               profile->max_priv_udavs,profile->max_priv_udavs);
    
    MTL_DEBUG1("        log2_max_mcgs = "SIZE_T_DFMT"\n", profile->log2_max_mcgs);
    MTL_DEBUG1("        qps_per_mcg = "SIZE_T_DFMT"\n", profile->qps_per_mcg);
    MTL_DEBUG1("        log2_mcg_entry_size = "SIZE_T_DFMT"\n", profile->log2_mcg_entry_size);
    MTL_DEBUG1("        log2_mcg_hash_size = "SIZE_T_DFMT"\n",profile->log2_mcg_hash_size);


    MTL_DEBUG1("        log2_max_eecs = "SIZE_T_DFMT"\n",profile->log2_max_eecs);

    MTL_DEBUG1("        log2_max_eqs = %d\n", profile->log2_max_eqs);
	return;
}

static void XHH_print_usr_profile(EVAPI_hca_profile_t *profile) 
{
    MTL_DEBUG1("User Profile printout\n");

    MTL_DEBUG1("        num_qp = %d\n",profile->num_qp);
    MTL_DEBUG1("        num_cq = %d\n",profile->num_cq);
    MTL_DEBUG1("        num_pd = %d\n",profile->num_pd);
    MTL_DEBUG1("        num_mr = %d\n",profile->num_mr);
    MTL_DEBUG1("        num_mw = %d\n",profile->num_mw);
    MTL_DEBUG1("        max_qp_ous_rd_atom = %d\n",profile->max_qp_ous_rd_atom);
    MTL_DEBUG1("        max_mcg = %d\n",profile->max_mcg);
    MTL_DEBUG1("        qp_per_mcg = %d\n",profile->qp_per_mcg);
    MTL_DEBUG1("        require = %s\n",(profile->require == 0) ? "FALSE" : "TRUE");
	return;
}
#endif

static const char* XHH_get_ddr_allocation_string (u_int32_t index)
{
    switch(index) {
        case 0: return "mtt sz";
        case 1: return "mpt sz";
        case 2: return "qpc sz";
        case 3: return "eqpc sz";
        case 4: return "srqc sz";
        case 5: return "cqc sz";
        case 6: return "rdb sz";
        case 7: return "uar scratch sz";
        case 8: return "eqc sz";
        case 9: return "mcg sz";
        case 10: return "eec sz";
        case 11: return "eeec sz";
    #if 0
        case 12: return "wqe pool sz";
        case 13: return "uplink qp sz";
        case 14: return "uplink mem sz";
    #endif
        default: return "UNKNOWN";

    }
}


/*
 * XHH_print_hw_props
 *
 */
static void XHH_print_hw_props(XHH_hw_props_t   *hw_props_p) 
{
    int i;

    MTL_DEBUG4("%s:        cr_base = " PHYS_ADDR_FMT "\n", __func__, hw_props_p->cr_base);
//    MTL_DEBUG4("%s:        ddr_base = " PHYS_ADDR_FMT "\n", __func__hw_props_p->ddr_base);
    MTL_DEBUG4("%s:        uar_base = " PHYS_ADDR_FMT "\n", __func__, hw_props_p->uar_base);
    MTL_DEBUG4("%s:        device_id = 0x%x\n", __func__, hw_props_p->device_id);
    MTL_DEBUG4("%s:        pci_vendor_id = 0x%x\n", __func__, hw_props_p->pci_vendor_id);
    MTL_DEBUG4("%s:        intr_pin = 0x%x\n", __func__, hw_props_p->interrupt_props.intr_pin);

    //MOSAL_IRQ_ID_t does not have to be integer
    MTL_DEBUG4("%s:        bus = %d\n", __func__, hw_props_p->bus);
    MTL_DEBUG4("%s:        dev_func = 0x%x\n", __func__, hw_props_p->dev_func);

    MTL_DEBUG4("%s:        number of allocated vectors = %d\n", __func__, hw_props_p->interrupt_props.num_allocated_vectors);
    for ( i=0; i<(int)hw_props_p->interrupt_props.num_allocated_vectors; ++i ) {
      MTL_DEBUG4("%s:        irq[%d] = 0x%x\n", __func__, i, hw_props_p->interrupt_props.irq[i]);
    }
}

/*
 * XHH_get_ddr_size_code
 *
 */
static XHH_ddr_size_enum_t XHH_get_ddr_size_code(MT_size_t    ddr_size) 
{
    MTL_DEBUG4("XHH_get_ddr_size_code: ddr size = "SIZE_T_FMT"\n", ddr_size);
    if (ddr_size < (1UL<<25)) {
        return XHH_DDR_SIZE_32M;
    } else if (ddr_size < (1UL<<26)) {
        return XHH_DDR_SIZE_64M;
    } else if (ddr_size < (1UL<<27)) {
        return XHH_DDR_SIZE_128M;
    } else if (ddr_size < (1UL<<28)) {
        return XHH_DDR_SIZE_256M;
    } else if (ddr_size < (1UL<<29)) {
        return XHH_DDR_SIZE_512M;
    } else if (ddr_size < (1UL<<30)) {
        return XHH_DDR_SIZE_1024M;
    } else if (ddr_size < (1UL<<31)) {
        return XHH_DDR_SIZE_2048M;
    } else if (ddr_size < 0xFFFFFFFF) {
        return XHH_DDR_SIZE_4096M;
    } else {
        return XHH_DDR_SIZE_BIG;
    }
}

/******************************************************************************
 *  Function:     XHH_hob_get_pci_br_config
 *
 *  Description:  Gets p2p bridge configuration for this hca's bridge  
 *
 *  input:
 *                hob
 *  output: 
 *                ack_timeout_p 
 *  returns:
 *                HH_OK
 *                HH_EINVAL
 *                HH_EINVAL_HCA_HNDL
 *                HH_ERR
 *
 *  Comments:     Does MAD query to get the data in real time. This function is used
 *                in pre-calculation of VAPI_query_hca values (at open_hca time).
 *
 *****************************************************************************/
static HH_ret_t  XHH_hob_get_pci_br_config(XHH_hob_t  hob)
{
    call_result_t  rc;
    u_int16_t      index=0;
    u_int8_t       bus;
    u_int8_t       dev_func;
    MOSAL_PCI_cfg_hdr_t *cfg_hdr_p = NULL;
    MOSAL_pci_dev_t pci_dev = NULL;

    MT_RETURN_IF_LOW_STACK(XHH_WATERMARK);

    cfg_hdr_p = (MOSAL_PCI_cfg_hdr_t *)MALLOC(sizeof(MOSAL_PCI_cfg_hdr_t));
    if (cfg_hdr_p == NULL) {
       MT_COV_MEM_FLTR_START
       return HH_EAGAIN;
       MT_COV_MEM_FLTR_END
    }
    memset(cfg_hdr_p, 0, sizeof(MOSAL_PCI_cfg_hdr_t));
    
    /* scan all bridges to find mellanox pci bridge which belongs to this hca */
    while (TRUE) {
      /*1. find device */
      rc = MOSAL_PCI_find_dev(hob->hw_props.pci_vendor_id, (hob->hw_props.device_id)+2,
                              pci_dev, &pci_dev, &bus, &dev_func);

      index++;
      if (rc != MT_OK) {
        MTL_DEBUG4(MT_FLFMT("%s: No more InfiniBridges."), __func__); 
        break;
      }
      MTL_DEBUG4(MT_FLFMT("%s: InfiniBridge %d: pci_find_device returned: bus=%d, dev_func=%d"), 
                 __func__, index, bus, dev_func);
      
      /*2. get pci header */
      rc = MOSAL_PCI_get_cfg_hdr(pci_dev, cfg_hdr_p);
      if (rc != MT_OK) {
        MTL_ERROR4(MT_FLFMT("%s: Could not get header for device bus %d, dev_func 0x%x"),
                   __func__, bus, dev_func); 
        continue;
      }

      if ((cfg_hdr_p->type1.header_type & 0x7F) != MOSAL_PCI_HEADER_TYPE1) {
        MTL_DEBUG1(MT_FLFMT("%s: Wrong PCI header type (0x%02X). Should be type 1. Device ignored."),
                   __func__, cfg_hdr_p->type0.header_type);
        continue; 
      }

      /*3. check if this is our bridge */
      if (cfg_hdr_p->type1.sec_bus != hob->hw_props.bus) {
        MTL_DEBUG1(MT_FLFMT("%s: Not our bridge. bus = %d, dev_num=%d"),
                   __func__, bus, dev_func );
        continue; 
      }

      /* found our bridge.  Read and save its configuration */
      MTL_DEBUG1(MT_FLFMT("%s: found bridge. bus = %d, dev_num=%d"),
                 __func__, bus, dev_func );
      if (XHH_read_pci_config(pci_dev, &hob->pci_bridge_info.config) != MT_OK) {
	  FREE(cfg_hdr_p);
          return (HH_ERR);
      } else {
          hob->pci_bridge_info.bus = bus;
          hob->pci_bridge_info.dev_func = dev_func;
          hob->pci_bridge_info.pci_dev = pci_dev;
          hob->pci_bridge_info.is_valid = TRUE;
	  FREE(cfg_hdr_p);
          return HH_OK;
      }
    }

    FREE(cfg_hdr_p);
    return HH_ERR; // did not find bridge
}

/******************************************************************************
 *  Function:     XHH_hob_get_node_guid
 *
 *  Description:  Gets node GUID for this HCA.  
 *
 *  input:
 *                hob
 *                port_num - 1 or 2
 *  output: 
 *                node_guid - pointer to a GUID structure
 *  returns:
 *                HH_OK
 *                HH_EINVAL
 *                HH_EINVAL_HCA_HNDL
 *                HH_ERR
 *
 *  Comments:     Does MAD query to get the data in real time. This function is used
 *                in pre-calculation of VAPI_query_hca values (at open_hca time).
 *
 *****************************************************************************/
static HH_ret_t XHH_hob_get_node_guid(XHH_hob_t  hob,
                                     IB_guid_t   *node_guid)
{
  SM_MAD_NodeInfo_t  node_info;
  u_int8_t       *mad_frame_in;
  u_int8_t       *mad_frame_out;
  XHH_cmd_status_t  cmd_ret;

  MT_RETURN_IF_LOW_STACK(XHH_WATERMARK);
  MTL_DEBUG4("==> XHH_hob_get_node_guid\n");
  TEST_RETURN_FATAL(hob);

  mad_frame_in = TNMALLOC(u_int8_t, IB_MAD_SIZE);
  if ( !mad_frame_in ) {
    return HH_EAGAIN;
  }
  mad_frame_out = TNMALLOC(u_int8_t, IB_MAD_SIZE);
  if ( !mad_frame_out ) {
    FREE(mad_frame_in);
    return HH_EAGAIN;
  }
  memset(mad_frame_in, 0, sizeof(mad_frame_in));
  memset(mad_frame_out, 0, sizeof(mad_frame_out));

  /* get PKey table using MAD commands in XHH_cmd object */
  /* First, build the MAD header */
  MADHeaderBuild(IB_CLASS_SMP, 
                    0,
                    IB_METHOD_GET,
                    IB_SMP_ATTRIB_NODEINFO,
                    (u_int32_t)   0,
                    &(mad_frame_in[0]));

  /* issue the query */
  cmd_ret = XHH_cmd_MAD_IFC(hob->cmd, 0, 0, 1, &(mad_frame_in[0]), &(mad_frame_out[0]));
  if (cmd_ret != XHH_CMD_STAT_OK) {
      TEST_CMD_FATAL(cmd_ret);
      MTL_ERROR2( "XHH_hob_get_pkey_tbl: ERROR : Get Node Info command failed (%d) for port 1\n", cmd_ret);
      MTL_DEBUG4("<== XHH_hob_get_node_guid. ERROR\n");
      FREE(mad_frame_out);
      FREE(mad_frame_in);
      return HH_EINVAL;
  }
  MadBufPrint(&(mad_frame_out[0]));
  NodeInfoMADToSt(&node_info, &(mad_frame_out[0]));
  NodeInfoPrint(&node_info);
  
//  guid = node_info.qwNodeGUID;
//  MTL_DEBUG4("XHH_hob_get_node_guid: Node GUID = 0x%Lx\n", guid);
//  for (i = 7; i >= 0 ; --i) {
//     (*node_guid)[i] = (u_int8_t) (guid & 0x0FF);
//     guid >>= 8;
//  }
  memcpy((*node_guid), node_info.qwNodeGUID, sizeof(IB_guid_t));
  MTL_DEBUG4("<== XHH_hob_get_node_guid\n");
  FREE(mad_frame_out);
  FREE(mad_frame_in);
  return HH_OK;
}

/******************************************************************************
 *  Function:     calculate_ddr_alloc_vec
 *
 *  Description:  Calculates sizes for DDR area allocation from profile
 *
 *  input:
 *                hob
 *                profile        -- pointer to data structure containing computation input
 *  output: 
 *                alloc_size_vec -- pointer to vector of allocation sizes to compute
 *                
 *  returns:
 *                HH_OK
 *
 *
 *****************************************************************************/
static void calculate_ddr_alloc_vec(/*IN*/ XHH_hob_t     hob,
                                            /*IN*/ XHH_profile_t *profile,  
                                            /*OUT*/XHH_ddr_allocation_vector_t *alloc_size_vec)
{

    alloc_size_vec->log2_mtt_size = profile->log2_max_mtt_entries + XHH_DDR_LOG2_MTT_ENTRY_SIZE; 
    alloc_size_vec->log2_mpt_size = profile->log2_max_mpt_entries + XHH_DDR_LOG2_MPT_ENTRY_SIZE;   
    alloc_size_vec->log2_qpc_size = profile->log2_max_qps + XHH_DDR_LOG2_QPC_ENTRY_SIZE;
    alloc_size_vec->log2_eqpc_size = profile->log2_max_qps + XHH_DDR_LOG2_EQPC_ENTRY_SIZE; 
    alloc_size_vec->log2_srqc_size = hob->dev_lims.srq ? 
      profile->log2_max_srqs + XHH_DDR_LOG2_SRQC_ENTRY_SIZE : XHH_DDRMM_INVALID_SZ;
    alloc_size_vec->log2_cqc_size = profile->log2_max_cqs + XHH_DDR_LOG2_CQC_ENTRY_SIZE;
    alloc_size_vec->log2_rdb_size = profile->log2_max_qps + profile->log2_inflight_rdma_per_qp + 
                                                        XHH_DDR_LOG2_RDB_ENTRY_SIZE;
    alloc_size_vec->log2_uar_scratch_size = profile->log2_max_uar + XHH_DDR_LOG2_UAR_SCR_ENTRY_SIZE;  
    alloc_size_vec->log2_eqc_size = profile->log2_max_eqs + XHH_DDR_LOG2_EQC_ENTRY_SIZE; 
    if (XHH_DEV_LIM_MCG_ENABLED(hob)) {
         alloc_size_vec->log2_mcg_size = profile->log2_max_mcgs + profile->log2_mcg_entry_size;
    } else {
        alloc_size_vec->log2_mcg_size = 0;
    }
    alloc_size_vec->log2_eec_size = profile->log2_max_eecs + XHH_DDR_LOG2_EEC_ENTRY_SIZE; 
    alloc_size_vec->log2_eeec_size = profile->log2_max_eecs + XHH_DDR_LOG2_EEEC_ENTRY_SIZE;    /* in-flight rdma */

    return;
}


/******************************************************************************
 *  Function:     XHH_check_profile
 *
 *  Description:  Validates profile values against tavor max values, as obtained
 *                from GET_DEV_LIM query
 *
 *  Details:
 *
 *****************************************************************************/
static HH_ret_t XHH_check_profile(XHH_hob_t   hob)
{
    
  MT_RETURN_IF_LOW_STACK(XHH_WATERMARK);
    if (hob->profile.log2_uar_pg_size < hob->dev_lims.log_pg_sz) {
        MTL_ERROR1("XHH_calculate_default_profile:  log2 UAR page size(%u) is less than the Tavor minimum (%u)\n", 
                   hob->profile.log2_uar_pg_size, hob->dev_lims.log_pg_sz);
        return HH_EAGAIN;
    }
    hob->profile.log2_max_qps             = (hob->profile.log2_max_qps > hob->dev_lims.log_max_qp) ? 
                                                         hob->dev_lims.log_max_qp : hob->profile.log2_max_qps; 
    hob->profile.log2_max_mcgs            = (hob->profile.log2_max_mcgs > hob->dev_lims.log_max_mcg) ? 
                                                         hob->dev_lims.log_max_mcg : hob->profile.log2_max_mcgs;
    hob->profile.log2_max_eecs            = (hob->profile.log2_max_eecs > hob->dev_lims.log_max_ee) ? 
                                                         hob->dev_lims.log_max_ee : hob->profile.log2_max_eecs;
    hob->profile.log2_max_cqs             = (hob->profile.log2_max_cqs > hob->dev_lims.log_max_cq) ? 
                                                         hob->dev_lims.log_max_cq : hob->profile.log2_max_cqs;
    hob->profile.log2_max_uar             = (hob->profile.log2_max_uar > hob->dev_lims.uar_sz + 20UL - hob->profile.log2_uar_pg_size) ? 
                                                         hob->dev_lims.uar_sz + 20UL - hob->profile.log2_uar_pg_size : hob->profile.log2_max_uar;

    hob->profile.log2_max_eqs             = (hob->profile.log2_max_eqs > hob->dev_lims.log_max_eq) ? 
                                                         hob->dev_lims.log_max_eq : hob->profile.log2_max_eqs; 
    hob->profile.max_num_pds              = (hob->profile.max_num_pds > (1UL<<hob->dev_lims.log_max_pd)) ? 
                                                         (1UL<<hob->dev_lims.log_max_pd) : hob->profile.max_num_pds;  

    if (XHH_DEV_LIM_MCG_ENABLED(hob)) {
        hob->profile.qps_per_mcg         = (hob->profile.qps_per_mcg > (MT_size_t) (1U<<hob->dev_lims.log_max_qp_mcg)) ? 
                                                   (MT_size_t) (1U<<hob->dev_lims.log_max_qp_mcg) : hob->profile.qps_per_mcg;
        hob->profile.log2_max_mcgs =  (hob->profile.log2_max_mcgs > hob->dev_lims.log_max_mcg) ? 
                                                         hob->dev_lims.log_max_mcg : hob->profile.log2_max_mcgs; 
    }

    return HH_OK;
}


#define XHH_PROFILE_CALC_QP_AT_MINIMUM     (1)
#define XHH_PROFILE_CALC_CQ_AT_MINIMUM     (1 << 1)
#define XHH_PROFILE_CALC_PD_AT_MINIMUM     (1 << 2)
#define XHH_PROFILE_CALC_REG_AT_MINIMUM    (1 << 3)
#define XHH_PROFILE_CALC_WIN_AT_MINIMUM    (1 << 4)
#define XHH_PROFILE_CALC_ALL_AT_MINIMUM    (XHH_PROFILE_CALC_QP_AT_MINIMUM | XHH_PROFILE_CALC_CQ_AT_MINIMUM | \
                                            XHH_PROFILE_CALC_PD_AT_MINIMUM | XHH_PROFILE_CALC_REG_AT_MINIMUM | \
                                            XHH_PROFILE_CALC_WIN_AT_MINIMUM )
                                            
static int check_profile_sanity(XHH_hob_t   hob, EVAPI_hca_profile_t *user_profile, XHH_profile_input_t *xhh_profile)
{
    u_int64_t  tmp_calc;
        /* check for bad minimum values */
        if ((user_profile->num_qp == 0) || (user_profile->num_cq == 0) || (user_profile->num_pd == 0) ||
            (user_profile->num_mr == 0) || (user_profile->max_qp_ous_rd_atom == 0) ) {
            MTL_ERROR1(MT_FLFMT("profile: QPs or CQs or PDs or MRs or max_qp_ous_rd_atom equal to 0"));
            return 0;
        }
        if (user_profile->num_qp > (1U<<hob->dev_lims.log_max_qp)) {
            MTL_ERROR1(MT_FLFMT("profile: num QPs more than device limit(%d)"),
                       (1U<<hob->dev_lims.log_max_qp));
            return 0;
        } else if (user_profile->num_qp < 1) {
            MTL_ERROR1(MT_FLFMT("profile: num QPs must be at least 1"));
            return 0;
        }

        if (user_profile->num_cq > (1U<<hob->dev_lims.log_max_cq)) {
            MTL_ERROR1(MT_FLFMT("profile: num CQs more than device limit(%d)"),
                       (1U<<hob->dev_lims.log_max_cq));
            return 0;
        } else if (user_profile->num_cq < 1) {
            MTL_ERROR1(MT_FLFMT("profile: num CQs must be at least 1"));
            return 0;
        }

        if (user_profile->num_pd > (1U<<hob->dev_lims.log_max_pd)) {
            MTL_ERROR1(MT_FLFMT("profile: num PDs more than device limit(%d)"),
                       (1U<<hob->dev_lims.log_max_pd));
            return 0;
        } else if (user_profile->num_pd < 1) {
            MTL_ERROR1(MT_FLFMT("profile: num PDs must be at least 1"));
            return 0;
        }
        if (user_profile->num_mr < 1) {
            MTL_ERROR1(MT_FLFMT("profile: num MRs must be at least 1"));
            return 0;
        }
        if (user_profile->max_qp_ous_rd_atom > (1U<<hob->dev_lims.log_max_ra_res_qp)) {
            MTL_ERROR1(MT_FLFMT("profile: max_qp_ous_rd_atom more than device limit (%d)"),
                       (1U<<hob->dev_lims.log_max_ra_res_qp));
            return 0;
        }

	if (XHH_DEV_LIM_MCG_ENABLED(hob)) {
	    if (ceil_log2((u_int64_t)user_profile->max_mcg) > hob->dev_lims.log_max_mcg) {
		MTL_ERROR1(MT_FLFMT("profile: num MCGs more than device limit(%d)"),
			   (1U<<hob->dev_lims.log_max_mcg));
		return 0;
	    }
    
	    if (ceil_log2((u_int64_t)user_profile->qp_per_mcg) > hob->dev_lims.log_max_qp_mcg) {
		MTL_ERROR1(MT_FLFMT("profile: QPs per multicast group greater than device limit (%d)"),
			   (1U<<hob->dev_lims.log_max_qp_mcg));
		return 0;
	    }
	    if (user_profile->qp_per_mcg > user_profile->num_qp) {
		MTL_ERROR1(MT_FLFMT("profile: QPs per multicast group (%d) greater than number of QPs (%d)"),
		       (int)user_profile->qp_per_mcg, (int) user_profile->num_qp);
		return 0;
	    }
	    if ((user_profile->max_mcg > 0) && (user_profile->qp_per_mcg < 8)) {
		MTL_ERROR1(MT_FLFMT("profile: if MCGs not zero, QP_PER_MCG must be >= 8"));
		return 0;
	    }
	}
        if (user_profile->num_cq > (user_profile->num_qp * 2)) {
            MTL_ERROR1(MT_FLFMT("profile: CQs more than twice QPs in hca profile"));
            return 0;
        }
        if (ceil_log2(user_profile->num_mr) > hob->dev_lims.log_max_mpts) {
            MTL_ERROR1("profile:  Requested MRs use more MTTs than HCA provides\n");
            return 0;
        }
        
        tmp_calc = (u_int64_t)((u_int64_t) user_profile->num_qp +
                               (u_int64_t) (unsigned long)XHH_NUM_RSVD_QP +
                               (u_int64_t) user_profile->num_cq + 
                               (u_int64_t) user_profile->num_mr);
        if ( hob->dev_lims.log_max_mtt_seg < ceil_log2( tmp_calc * (u_int64_t)(1U<<xhh_profile->log2_mtt_segs_per_region) )) {
            MTL_ERROR1("profile:  Requested parameters (CQs + QPs + MRs) use more MTTs than HCA provides\n");
            return 0;
        }


        if (ceil_log2(user_profile->num_mr) > hob->dev_lims.log_max_mpts) {
            MTL_ERROR1("profile:  Requested MRs use more MPTs than HCA provides\n");
            return 0;
        }

        if (ceil_log2(user_profile->num_mw) > hob->dev_lims.log_max_mpts) {
            MTL_ERROR1("profile:  Requested MWs use more MPTs than HCA provides\n");
            return 0;
        }

        tmp_calc = (u_int64_t)((u_int64_t) user_profile->num_qp +
                               (u_int64_t) (unsigned long)XHH_NUM_RSVD_QP +
                               (u_int64_t) user_profile->num_cq + 
                               (u_int64_t) user_profile->num_mr + 
                               (u_int64_t) user_profile->num_mw);

        if ( hob->dev_lims.log_max_mpts < ceil_log2( tmp_calc)) {
            MTL_ERROR1("profile:  Requested parameters (CQs + QPs + MRs + MWs) use more MPTs than HCA provides\n");
            return 0;
        }
        return 1;
}
/******************************************************************************
 *  Function:     XHH_calculate_profile
 *
 *  Description:  Calculates and installs profile values
 *
 *  input
 *                hob
 *                profile_user_data - pointer to a user override for the data used
 *                                    in calculating the XHH profile.
 *
 *  Details:
 *
 *        All calculations are derived from the following data:
 *
 *        - max QPs = 64k per 128M DDR size (= 2^16)
 *        - max MPT entries per HCA:       1M (= 2^20)
 *        - avg Regions/windows per QP     8  (= 2^3)
 *        - avg Segments per Region        8  (= 2^3)
 *        - avg inflight RDMA per QP       4  (= 2^2)
 *
 *        Calculations are as follows:
 *        Max UARs = 1 per QP
 *        Max CQs  = 1 per QP
 *        Max PDs  = 1 per QP
 *        Max Regions/Wins per QP = 8, divided as follows:
 *             internal regions = 2 per QP  (1 for QP, one for CQ)
 *             external regions = 2 per QP
 *             windows          = 4 per QP
 *
 *        MPT:
 *           Tavor has a max of 1M regions/windows per HCA, and the MPT size must
 *           be a power of 2.  It is pointless to have fewer than 8 regions/windows per QP
 *           (as divided up above).  This means that the maximum number of QPs allowable,
 *           regardless of DDR size, is 128K.  Therefore, the presence of the "min" function
 *           in calculating the max number of MPT entries.  In effect, the 1M table size limitation
 *           means that a DDR larger than 256M will only add to the user-available DDR memory, and
 *           not to the driver's internal tables.
 *
 *        MTT:
 *           The default MTT size allocated has 2 segments per Region, with a segment size of 8 entries.
 *
 *        MCG:  for 128M: 4096 Groups per HCA, with 16 QPs per group (so that entry size is 64 bytes).
 *              for 256M: 8192 Groups per HCA, with 16 QPs per group (so that entry size is 64 bytes).
 *
 *        NOTES:
 *           If the profile_user_data is NULL, default values are used.  After the profile calculation,
 *           a check is done to see that all values are within HCA_DEV_LIM values, and that the result
 *           does not exceed the DDR memory size.  If any violations are encountered, the number of QPs
 *           is reduced by half, and the calculation is redone.
 *           
 *****************************************************************************/
static HH_ret_t XHH_calculate_profile(XHH_hob_t   hob, 
                                      EVAPI_hca_profile_t *profile_user_data,
                                      EVAPI_hca_profile_t  *sugg_profile_p)
{
    u_int8_t            log2_host_pg_size;
    EVAPI_hca_profile_t *local_user_profile_p = NULL;
    XHH_profile_input_t *profile_input_data_p = NULL;
    u_int64_t           tot_ddr_allocs;
    XHH_ddr_allocation_vector_t *ddr_alloc_vec_p = NULL;
    MT_size_t           *ddr_alloc_iterator, temp_size;
    u_int32_t           i;
    MT_bool             ddr_calc_loop = TRUE, need_to_loop = FALSE;
    u_int32_t           calc_at_minimum = 0;
//    EVAPI_hca_profile_t hca_profile;

    local_user_profile_p = (EVAPI_hca_profile_t *)MALLOC(sizeof(EVAPI_hca_profile_t));
    if (local_user_profile_p == NULL) {
       MT_COV_MEM_FLTR_START
       MT_RETURN(HH_EAGAIN);
       MT_COV_MEM_FLTR_END
    }
    memset(local_user_profile_p, 0, sizeof(EVAPI_hca_profile_t));

    profile_input_data_p = (XHH_profile_input_t *)MALLOC(sizeof(XHH_profile_input_t));
    if (profile_input_data_p == NULL) {
       MT_COV_MEM_FLTR_START
       FREE(local_user_profile_p);
       MT_RETURN(HH_EAGAIN);
       MT_COV_MEM_FLTR_END
    }
    memset(profile_input_data_p, 0, sizeof(XHH_profile_input_t));

    ddr_alloc_vec_p = (XHH_ddr_allocation_vector_t *)MALLOC(sizeof(XHH_ddr_allocation_vector_t));
    if (ddr_alloc_vec_p == NULL) {
       MT_COV_MEM_FLTR_START
       FREE(local_user_profile_p);
       FREE(profile_input_data_p);
       MT_RETURN(HH_EAGAIN);
       MT_COV_MEM_FLTR_END
    }
    memset(ddr_alloc_vec_p, 0, sizeof(XHH_ddr_allocation_vector_t));

    if (profile_user_data != NULL) {
       
        memcpy(local_user_profile_p,  profile_user_data, sizeof(EVAPI_hca_profile_t));

        /* default value substitutions */
        local_user_profile_p->num_qp = (local_user_profile_p->num_qp == 0xFFFFFFFF) ? 
                       XHH_PROF_MAX_QPS : local_user_profile_p->num_qp; 
        local_user_profile_p->num_cq = (local_user_profile_p->num_cq == 0xFFFFFFFF) ? 
                               XHH_PROF_MAX_CQS  : local_user_profile_p->num_cq; 
        local_user_profile_p->num_pd = (local_user_profile_p->num_pd == 0xFFFFFFFF) ? 
                               XHH_PROF_MAX_PDS : local_user_profile_p->num_pd; 
        local_user_profile_p->num_mr = (local_user_profile_p->num_mr == 0xFFFFFFFF) ? 
                               XHH_PROF_MAX_REGIONS : local_user_profile_p->num_mr; 
        local_user_profile_p->num_mw = (local_user_profile_p->num_mw == 0xFFFFFFFF) ? 
                               XHH_PROF_MAX_WINDOWS : local_user_profile_p->num_mw;
        
        local_user_profile_p->max_qp_ous_rd_atom = (local_user_profile_p->max_qp_ous_rd_atom == 0xFFFFFFFF) ? 
                               (1 << XHH_DDR_LOG2_INFLIGHT_RDMA_PER_QP):
                               local_user_profile_p->max_qp_ous_rd_atom; 

        local_user_profile_p->max_mcg = (local_user_profile_p->max_mcg == 0xFFFFFFFF) ? 
                               (1 << XHH_DDR_LOG2_MAX_MCG):local_user_profile_p->max_mcg; 
        
        local_user_profile_p->qp_per_mcg = (local_user_profile_p->qp_per_mcg == 0xFFFFFFFF) ? 
                               (1 << XHH_DDR_LOG2_MIN_QP_PER_MCG):local_user_profile_p->qp_per_mcg; 
        
        if (sugg_profile_p != NULL) {
            memcpy(sugg_profile_p, local_user_profile_p, sizeof(EVAPI_hca_profile_t));
        }

        profile_input_data_p->max_qps        = local_user_profile_p->num_qp ;
        profile_input_data_p->max_cqs        = local_user_profile_p->num_cq ;
        profile_input_data_p->max_pds        = local_user_profile_p->num_pd ;
        profile_input_data_p->max_regions    = local_user_profile_p->num_mr ;
        profile_input_data_p->max_windows    = local_user_profile_p->num_mw ;

        profile_input_data_p->min_qps        = (1U<<hob->dev_lims.log2_rsvd_qps) + XHH_NUM_RSVD_QP + 1;
        profile_input_data_p->min_cqs        = (1U<<hob->dev_lims.log2_rsvd_cqs) + 1;
        profile_input_data_p->min_pds        = hob->dev_lims.num_rsvd_pds + XHH_NUM_RSVD_PD + 1;
        profile_input_data_p->min_regions    = (1 << hob->dev_lims.log2_rsvd_mtts) + 1;
        profile_input_data_p->min_windows    = (1 << hob->dev_lims.log2_rsvd_mrws);
    
        profile_input_data_p->reduction_pct_qps     = 10;
        profile_input_data_p->reduction_pct_cqs     = 10;
        profile_input_data_p->reduction_pct_pds     = 10;
        profile_input_data_p->reduction_pct_regions = 10;
        profile_input_data_p->reduction_pct_windows = 10;
        
        profile_input_data_p->log2_inflight_rdma_per_qp = ceil_log2(local_user_profile_p->max_qp_ous_rd_atom);
        profile_input_data_p->log2_max_mcg              = ceil_log2(local_user_profile_p->max_mcg);
        profile_input_data_p->log2_min_qp_per_mcg       = ceil_log2(local_user_profile_p->qp_per_mcg);
        profile_input_data_p->suggested_qp_per_mcg      = local_user_profile_p->qp_per_mcg;
        
        profile_input_data_p->log2_max_eq               = XHH_DDR_LOG2_MAX_EQ;
        profile_input_data_p->log2_mcg_hash_proportion  = XHH_DDR_LOG2_MCG_HASH_PROPORTION;
        profile_input_data_p->log2_mtt_entries_per_seg  = XHH_DDR_LOG2_MTT_ENTRIES_PER_SEG;
        profile_input_data_p->log2_mtt_segs_per_region  = XHH_DDR_LOG2_MTT_SEGS_PER_REGION;
        profile_input_data_p->use_priv_udav             = XHH_USE_PRIV_UDAV;
        profile_input_data_p->log2_wqe_ddr_space_per_qp = XHH_LOG2_WQE_DDR_SPACE_PER_QP;
        /*sanity checks */
        if (check_profile_sanity(hob,local_user_profile_p, profile_input_data_p) == 0) {
            MTL_ERROR1(MT_FLFMT("XHH_calculate_profile: user profile not valid"));
	    FREE(local_user_profile_p);
	    FREE(profile_input_data_p);
	    FREE(ddr_alloc_vec_p);
            return HH_EINVAL_PARAM;
        }
    } else {
        /* use internally defined default values */
        profile_input_data_p->max_qps        = XHH_PROF_MAX_QPS;
        profile_input_data_p->max_cqs        = XHH_PROF_MAX_CQS;
        profile_input_data_p->max_pds        = XHH_PROF_MAX_PDS;
        profile_input_data_p->max_regions    = XHH_PROF_MAX_REGIONS;
        profile_input_data_p->max_windows    = XHH_PROF_MAX_WINDOWS;
        profile_input_data_p->max_priv_udavs = XHH_DDR_MAX_PRIV_UDAVS;
        
        profile_input_data_p->min_qps        = XHH_PROF_MIN_QPS;
        profile_input_data_p->min_cqs        = XHH_PROF_MIN_CQS;
        profile_input_data_p->min_pds        = XHH_PROF_MIN_PDS;
        profile_input_data_p->min_regions    = XHH_PROF_MIN_REGIONS;
        profile_input_data_p->min_windows    = XHH_PROF_MIN_WINDOWS;

        profile_input_data_p->reduction_pct_qps     = XHH_PROF_PCNT_REDUCTION_QPS;
        profile_input_data_p->reduction_pct_cqs     = XHH_PROF_PCNT_REDUCTION_CQS;
        profile_input_data_p->reduction_pct_pds     = XHH_PROF_PCNT_REDUCTION_PDS;
        profile_input_data_p->reduction_pct_regions = XHH_PROF_PCNT_REDUCTION_REGIONS;
        profile_input_data_p->reduction_pct_windows = XHH_PROF_PCNT_REDUCTION_WINDOWS;
        
        profile_input_data_p->log2_inflight_rdma_per_qp = XHH_DDR_LOG2_INFLIGHT_RDMA_PER_QP;
        profile_input_data_p->log2_max_eq               = XHH_DDR_LOG2_MAX_EQ;
        profile_input_data_p->log2_max_mcg              = XHH_DDR_LOG2_MAX_MCG;
        profile_input_data_p->log2_min_qp_per_mcg       = XHH_DDR_LOG2_MIN_QP_PER_MCG;
        profile_input_data_p->suggested_qp_per_mcg      = (1U << XHH_DDR_LOG2_MIN_QP_PER_MCG);
        profile_input_data_p->log2_mcg_hash_proportion  = XHH_DDR_LOG2_MCG_HASH_PROPORTION;
        profile_input_data_p->log2_mtt_entries_per_seg  = XHH_DDR_LOG2_MTT_ENTRIES_PER_SEG;
        profile_input_data_p->log2_mtt_segs_per_region  = XHH_DDR_LOG2_MTT_SEGS_PER_REGION;
        profile_input_data_p->use_priv_udav             = XHH_USE_PRIV_UDAV;
        profile_input_data_p->log2_wqe_ddr_space_per_qp = XHH_LOG2_WQE_DDR_SPACE_PER_QP;
    }

    hob->profile.use_priv_udav = profile_input_data_p->use_priv_udav;
    hob->profile.max_priv_udavs = profile_input_data_p->max_priv_udavs;
    
    /* need inflight rdma per QP for rdb size in DDR, and for XHH_qpm_create */
    hob->profile.log2_inflight_rdma_per_qp = (u_int8_t) profile_input_data_p->log2_inflight_rdma_per_qp;

    /* manipulate MCG max if not inputting a profile, or if inputting profile which allows reduction */
    /* Reduce the number of MCGs for smaller DDR memories */
    hob->profile.log2_max_mcgs            = profile_input_data_p->log2_max_mcg;
    if ((hob->profile.ddr_size_code < XHH_DDR_SIZE_128M)
        && ((profile_user_data == NULL)|| (local_user_profile_p->require == FALSE))) {
            hob->profile.log2_max_mcgs--;
    }

    log2_host_pg_size = MOSAL_SYS_PAGE_SHIFT;
    if (log2_host_pg_size < hob->dev_lims.log_pg_sz) {
        MTL_ERROR1("XHH_calculate_default_profile:  Host min page size(%lu) is too small\n", 
                   (unsigned long ) MOSAL_SYS_PAGE_SIZE);
	FREE(local_user_profile_p);
	FREE(profile_input_data_p);
	FREE(ddr_alloc_vec_p);
        return HH_EAGAIN;
    }

    /* do not allocate DDR memory for MCGs if MCG is not enabled in dev_limits  */
    hob->profile.ddr_alloc_vec_size = (XHH_DEV_LIM_MCG_ENABLED(hob) ? 
                                       XHH_DDR_ALLOCATION_VEC_SIZE : XHH_DDR_ALLOCATION_VEC_SIZE - 1) ;  /* no eec as yet */
    hob->profile.log2_wqe_ddr_space_per_qp = profile_input_data_p->log2_wqe_ddr_space_per_qp;
    
    /* MCG calculations - not in recalculation loop, since the amount of memory involved is very small*/
    /* each MCG entry must be a power-of-2 size. To guarantee a power-of-2, we take a "ceiling" log of the */
    /* MCG entry size(in bytes), and then compute the actual number of QPs per mcg backwards from the mcg_size variable. */
    /* We also require (as a sanity check) that the log2_mcg_hash_size be greater than zero */
    if ((XHH_DEV_LIM_MCG_ENABLED(hob)) &&
        ((int)(hob->profile.log2_max_mcgs + profile_input_data_p->log2_mcg_hash_proportion) > 0)) {
	/* Adjust max QPs per mcg and mcg entry size here, since it affects allocation vector calculation */
	if (profile_input_data_p->suggested_qp_per_mcg > (u_int32_t) (1U<<hob->dev_lims.log_max_qp_mcg)) {
	    hob->profile.qps_per_mcg         =  (u_int32_t) (1U<<hob->dev_lims.log_max_qp_mcg);
	    hob->profile.log2_mcg_entry_size = ceil_log2( (hob->profile.qps_per_mcg * XHH_DDR_MCG_BYTES_PER_QP) + 
							     XHH_DDR_MCG_ENTRY_HEADER_SIZE
							);
	} else {
            hob->profile.log2_mcg_entry_size = ceil_log2((profile_input_data_p->suggested_qp_per_mcg*XHH_DDR_MCG_BYTES_PER_QP) 
                                                     + XHH_DDR_MCG_ENTRY_HEADER_SIZE);
            hob->profile.qps_per_mcg    = ( (1U<<(hob->profile.log2_mcg_entry_size)) - XHH_DDR_MCG_ENTRY_HEADER_SIZE) / 
                                                             XHH_DDR_MCG_BYTES_PER_QP;
	}
    
        /* the hash proportion is the log of the power-of-2 fraction of the total MCG entries used for the hash table. */
        /* Thus, for example, a proportion of (1/2) gets a log2_mcg_hash_proportion = -1 */
        hob->profile.log2_mcg_hash_size  = hob->profile.log2_max_mcgs + profile_input_data_p->log2_mcg_hash_proportion;
    } else {
        /*UD MCGs not available on this HCA*/
        hob->profile.log2_mcg_entry_size = 0;
        hob->profile.qps_per_mcg    = 0;
        hob->profile.log2_mcg_hash_size  = 0;
        hob->profile.log2_max_mcgs = 0;
        }
    
    hob->profile.log2_mtt_entries_per_seg = profile_input_data_p->log2_mtt_entries_per_seg;
    hob->profile.log2_mtt_segs_per_region = profile_input_data_p->log2_mtt_segs_per_region;
    
    hob->profile.log2_uar_pg_size         = log2_host_pg_size;
    hob->profile.log2_max_uar             = hob->dev_lims.uar_sz + 20 - hob->profile.log2_uar_pg_size;
    hob->profile.log2_max_eqs             = (u_int8_t)profile_input_data_p->log2_max_eq;      /* 64 EQs */
    hob->profile.max_num_pds              = profile_input_data_p->max_pds;

    hob->profile.max_num_qps              = profile_input_data_p->max_qps;

    hob->profile.log2_max_qps             = ceil_log2(profile_input_data_p->max_qps+ 
                                               (1U<<hob->dev_lims.log2_rsvd_qps) + XHH_NUM_RSVD_QP); 
    
    /* adjust max QPs downward (if using internal profile, or if user profile permits)
     * if the few reserved QPs cause max qps to go beyond a power-of-2.
     */

    if (hob->profile.log2_max_qps > ceil_log2(profile_input_data_p->max_qps)) {
        MTL_DEBUG1(MT_FLFMT("%s:  reserved qps cause profile qps to jump a power-of-2"),__func__);
        if ((profile_user_data==NULL) || (local_user_profile_p->require == FALSE)) {
            hob->profile.log2_max_qps--;
            hob->profile.max_num_qps = (1U<<hob->profile.log2_max_qps) - (1U<<hob->dev_lims.log2_rsvd_qps)
                                                                           - XHH_NUM_RSVD_QP;
            MTL_DEBUG1(MT_FLFMT("%s: Adjusting max qps to "SIZE_T_DFMT),__func__, hob->profile.max_num_qps);
        }
    }

    /* TBD: Expose max_srqs to profile given by user and use MOD_STAT_CFG */
    if (hob->dev_lims.srq) {
      hob->profile.log2_max_srqs            = hob->dev_lims.log_max_srqs;
      hob->profile.max_num_srqs             = 
        (1U << hob->dev_lims.log_max_srqs) - (1 << hob->dev_lims.log2_rsvd_srqs);
    } else {
      hob->profile.log2_max_srqs            = 0;
      hob->profile.max_num_srqs             = 0;
    }
    
    hob->profile.max_num_cqs              = profile_input_data_p->max_cqs;
    hob->profile.log2_max_cqs             = ceil_log2(profile_input_data_p->max_cqs + 
                                                      (1U<<hob->dev_lims.log2_rsvd_cqs)); 
    /* adjust max CQs downward (if using internal profile, or if user profile permits) 
     * if the few reserved CQs cause max cqs to go beyond a power-of-2.
     */

    if (hob->profile.log2_max_cqs > ceil_log2(profile_input_data_p->max_cqs)) {
        MTL_DEBUG1(MT_FLFMT("%s:  reserved cqs cause profile cqs to jump a power-of-2"),__func__);
        if ((profile_user_data == NULL) || (local_user_profile_p->require == FALSE)) {
            hob->profile.log2_max_cqs--;
            hob->profile.max_num_cqs = (1U<<hob->profile.log2_max_cqs) - (1U<<hob->dev_lims.log2_rsvd_cqs);
            MTL_DEBUG1(MT_FLFMT("%s: Adjusting max cqs to "SIZE_T_DFMT),__func__, hob->profile.max_num_cqs);
        }
    }
    hob->profile.num_external_mem_regions = profile_input_data_p->max_regions;  /* 2 per QP */
    hob->profile.num_mem_windows          = profile_input_data_p->max_windows;
    hob->profile.log2_max_eecs            = 0;

    while (ddr_calc_loop) {
        MT_bool continue_calc_loop;

        continue_calc_loop = FALSE;
    
        MTL_DEBUG4("XHH_calculate_profile: max_qps = "SIZE_T_FMT", max_cqs = "SIZE_T_FMT", max_priv_udav="SIZE_T_FMT
                   ",\nmax_pds="SIZE_T_FMT",max_reg="SIZE_T_FMT", max_win="SIZE_T_FMT"\n",
                        hob->profile.max_num_qps, hob->profile.max_num_cqs, hob->profile.max_priv_udavs,
                        hob->profile.max_num_pds, hob->profile.num_external_mem_regions, 
                        hob->profile.num_mem_windows); 

        /* add all raw resources without Tavor-reserved quantities */
        temp_size = hob->profile.max_num_qps + XHH_NUM_RSVD_QP + 
                    hob->profile.max_num_cqs + hob->profile.num_external_mem_regions;
        
        hob->profile.log2_max_mtt_entries     = ceil_log2(
                                                   ( temp_size * (1U<<profile_input_data_p->log2_mtt_segs_per_region)
                                                      * (1U<<profile_input_data_p->log2_mtt_entries_per_seg))
                                                   + (1 << hob->dev_lims.log2_rsvd_mtts)
                                                );
    
        /* add all raw resources without Tavor-reserved quantities */
        temp_size = hob->profile.max_num_qps + XHH_NUM_RSVD_QP + hob->profile.max_num_cqs 
            + hob->profile.num_external_mem_regions + hob->profile.num_mem_windows;
        
        hob->profile.log2_max_mpt_entries       = ceil_log2(temp_size + (1 << hob->dev_lims.log2_rsvd_mrws));
    
    
        if (hob->profile.log2_max_mtt_entries - profile_input_data_p->log2_mtt_entries_per_seg 
             > hob->dev_lims.log_max_mtt_seg) {
            continue_calc_loop = TRUE;
            need_to_loop = TRUE;
        }
        
        if (!continue_calc_loop) {
            /* Now, compute the total DDR size, and verify that we have not over-allocated it.  If yes, reduce QPs by half, and */
            /* recompute all above parameters starting with log2_max_regions */
            calculate_ddr_alloc_vec(hob, &(hob->profile),ddr_alloc_vec_p);
        
            /* Add up all the sizes in the ddr allocation vector */
            tot_ddr_allocs = 0;
            ddr_alloc_iterator = (MT_size_t *)ddr_alloc_vec_p;
            for (i = 0; i < hob->profile.ddr_alloc_vec_size; i++, ddr_alloc_iterator++) {
                if ((*ddr_alloc_iterator) == XHH_DDRMM_INVALID_SZ) {
                    temp_size = 0;  /* no allocation */
                } else if ((*ddr_alloc_iterator) >= ceil_log2(hob->profile.ddr_size)) {
                    temp_size = hob->profile.ddr_size;
                } else  {
                    temp_size =  (((MT_size_t) 1ul) << (*ddr_alloc_iterator));
                }
                MTL_DEBUG4("XHH_calculate_profile:DDR: %s = "SIZE_T_XFMT"("SIZE_T_DFMT")\n", 
                           XHH_get_ddr_allocation_string(i), temp_size, temp_size); 
                tot_ddr_allocs += temp_size;
            }
            
            /* see if need to reserve space for WQEs in DDR */
            if (hob->profile.log2_wqe_ddr_space_per_qp != 0) {
                temp_size =  (((MT_size_t) 1ul) << (hob->profile.log2_max_qps + hob->profile.log2_wqe_ddr_space_per_qp));
                MTL_DEBUG4("XHH_calculate_profile:  WQEs ddr area = "SIZE_T_XFMT" ("SIZE_T_DFMT")\n", 
                           temp_size, temp_size); 
                tot_ddr_allocs += temp_size;
            }
        
            /* see if need to reserve space for privileged UDAVs in DDR */
            if (hob->profile.use_priv_udav) {
                temp_size =  hob->profile.max_priv_udavs * (sizeof(struct tavorprm_ud_address_vector_st) / 8);
                MTL_DEBUG4("XHH_calculate_profile:  privileged UDAVs ddr area = "SIZE_T_XFMT" ("SIZE_T_DFMT")\n",
                            temp_size, temp_size); 
                tot_ddr_allocs += temp_size;
            }
        
            /* test against DDR size */
            MTL_DEBUG4("XHH_calculate_profile:  total DDR allocs = %d MB (incl reserved areas)\n",(int)(tot_ddr_allocs>>20)); 
            if ((hob->profile.ddr_size < tot_ddr_allocs) || 
                          ((profile_user_data == NULL) && (hob->profile.max_num_qps>(1U<<16)))){ 
                          /*do not want more than 64K QPs if using internal defaults*/
                continue_calc_loop = TRUE;
                need_to_loop = TRUE;
            }

        }
        if (continue_calc_loop) {
            u_int64_t  temp;
            u_int32_t  u32_temp, change_flag;
            /* Reduce flagged profile input params by factor of 10 percent */
            change_flag = 0;
            if ((calc_at_minimum & XHH_PROFILE_CALC_QP_AT_MINIMUM) == 0) {
                change_flag++;
                temp = (u_int64_t)(hob->profile.max_num_qps) * (100 - profile_input_data_p->reduction_pct_qps);
                /*check for overflow. If have overflow, use approximate percentages (divide by 1024) */
                if (temp & MAKE_ULONGLONG(0xFFFFFFFF00000000)) {
                    temp = (u_int64_t)(hob->profile.max_num_qps) * (1024 - (profile_input_data_p->reduction_pct_qps*10));
                    temp >>= 10;
                    u32_temp = (u_int32_t)(temp & 0xFFFFFFFF);
                } else {
                    /* use more exact percentages -- but still not floating point */
                    u32_temp = (u_int32_t)temp;
                    u32_temp /= 100;
                }
                if (u32_temp <= (u_int32_t)profile_input_data_p->min_qps) {
                    calc_at_minimum |= XHH_PROFILE_CALC_QP_AT_MINIMUM;
                    u32_temp = profile_input_data_p->min_qps;
                }
                hob->profile.max_num_qps = u32_temp;
                hob->profile.log2_max_qps = ceil_log2(u32_temp + (1U<<hob->dev_lims.log2_rsvd_qps) 
                                                      + XHH_NUM_RSVD_QP);
            }
            
            
            if ((calc_at_minimum & XHH_PROFILE_CALC_CQ_AT_MINIMUM) == 0) {
                change_flag++;
                temp = (u_int64_t)(hob->profile.max_num_cqs) * (100 - profile_input_data_p->reduction_pct_cqs);
                if (temp & MAKE_ULONGLONG(0xFFFFFFFF00000000)) {
                    temp = (u_int64_t)(hob->profile.max_num_cqs) * (1024 - (profile_input_data_p->reduction_pct_cqs*10));
                    temp >>= 10;
                    u32_temp = (u_int32_t)(temp & 0xFFFFFFFF);
                } else {
                    /* use more exact percentages -- but still not floating point */
                    u32_temp = (u_int32_t)temp;
                    u32_temp /= 100;
                }
                if (u32_temp <= (u_int32_t)profile_input_data_p->min_cqs) {
                    calc_at_minimum |= XHH_PROFILE_CALC_CQ_AT_MINIMUM;
                    u32_temp = profile_input_data_p->min_cqs;
                }
                hob->profile.max_num_cqs = u32_temp;
                hob->profile.log2_max_cqs = ceil_log2(u32_temp + (1U<<hob->dev_lims.log2_rsvd_cqs));
            }

            if ((calc_at_minimum & XHH_PROFILE_CALC_PD_AT_MINIMUM) == 0) {
                change_flag++;
                temp = (u_int64_t)(hob->profile.max_num_pds) * (100 - profile_input_data_p->reduction_pct_pds);
                if (temp & MAKE_ULONGLONG(0xFFFFFFFF00000000)) {
                    temp = (u_int64_t)(hob->profile.max_num_pds) * (1024 - (profile_input_data_p->reduction_pct_pds*10));
                    temp >>= 10;
                    u32_temp = (u_int32_t)(temp & 0xFFFFFFFF);
                } else {
                    /* use more exact percentages -- but still not floating point */
                    u32_temp = (u_int32_t)temp;
                    u32_temp /= 100;
                }
                if (u32_temp <= (u_int32_t)profile_input_data_p->min_pds) {
                    calc_at_minimum |= XHH_PROFILE_CALC_PD_AT_MINIMUM;
                    u32_temp = profile_input_data_p->min_pds;
                }
                hob->profile.max_num_pds = u32_temp;
            }

            if ((calc_at_minimum & XHH_PROFILE_CALC_REG_AT_MINIMUM) == 0) {
                change_flag++;
                temp = (u_int64_t)(hob->profile.num_external_mem_regions) * (100 - profile_input_data_p->reduction_pct_regions);
                if (temp & MAKE_ULONGLONG(0xFFFFFFFF00000000)) {
                    temp = (u_int64_t)(hob->profile.num_external_mem_regions) * (1024 - (profile_input_data_p->reduction_pct_regions*10));
                    temp >>= 10;
                    u32_temp = (u_int32_t)(temp & 0xFFFFFFFF);
                } else {
                    /* use more exact percentages -- but still not floating point */
                    u32_temp = (u_int32_t)temp;
                    u32_temp /= 100;
                }
                if (u32_temp <= (u_int32_t)profile_input_data_p->min_regions) {
                    calc_at_minimum |= XHH_PROFILE_CALC_REG_AT_MINIMUM;
                    u32_temp = profile_input_data_p->min_regions;
                }
                hob->profile.num_external_mem_regions = u32_temp;
            }

            if ((calc_at_minimum & XHH_PROFILE_CALC_WIN_AT_MINIMUM) == 0) {
                change_flag++;
                temp = (u_int64_t)(hob->profile.num_mem_windows) * (100 - profile_input_data_p->reduction_pct_windows);
                if (temp & MAKE_ULONGLONG(0xFFFFFFFF00000000)) {
                    temp = (u_int64_t)(hob->profile.num_mem_windows) * (1024 - (profile_input_data_p->reduction_pct_windows*10));
                    temp >>= 10;
                    u32_temp = (u_int32_t)(temp & 0xFFFFFFFF);
                } else {
                    /* use more exact percentages -- but still not floating point */
                    u32_temp = (u_int32_t)temp;
                    u32_temp /= 100;
                }
                if (u32_temp <= (u_int32_t)profile_input_data_p->min_windows) {
                    calc_at_minimum |= XHH_PROFILE_CALC_WIN_AT_MINIMUM;
                    u32_temp = profile_input_data_p->min_windows;
                }
                hob->profile.num_mem_windows = u32_temp;
            }
            if (hob->profile.log2_inflight_rdma_per_qp > XHH_DDR_LOG2_INFLIGHT_RDMA_PER_QP) {
                 hob->profile.log2_inflight_rdma_per_qp--;
                 change_flag++;
            }

            /* check if we were able to perform any reductions */
            if (change_flag == 0) {
                MTL_ERROR1("XHH_calculate_default_profile:  DDR memory to small for MIN profile\n");
                ddr_alloc_iterator = (MT_size_t *)ddr_alloc_vec_p;
                for (i = 0; i < hob->profile.ddr_alloc_vec_size; i++, ddr_alloc_iterator++) {
                    if ((*ddr_alloc_iterator) == XHH_DDRMM_INVALID_SZ) {
                        temp_size = 0;  /* no allocation */
                    } else if ((*ddr_alloc_iterator) >= ceil_log2(hob->profile.ddr_size)) {
                        temp_size = hob->profile.ddr_size;
                        MTL_ERROR1(MT_FLFMT("XHH_calculate_profile: %s uses ALL available DDR memory"), 
                               XHH_get_ddr_allocation_string(i)); 
                    } else  {
                        temp_size =  (((MT_size_t) 1ul) << (*ddr_alloc_iterator));
                    }
                    MTL_ERROR1(MT_FLFMT("XHH_calculate_profile:DDR: %s = "SIZE_T_XFMT"("SIZE_T_DFMT")"), 
                               XHH_get_ddr_allocation_string(i), temp_size, temp_size); 
                }
		FREE(local_user_profile_p);
		FREE(profile_input_data_p);
		FREE(ddr_alloc_vec_p);
                return HH_EAGAIN;
            }
        } else {
            ddr_calc_loop = FALSE;
        }
    }
    
    XHH_check_profile(hob);  /* final adjustment to catch dev-lim overruns*/

    /* adjust mcg hash table after final adjustment of mcg size */
    if (XHH_DEV_LIM_MCG_ENABLED(hob)) {
        hob->profile.log2_mcg_hash_size  =  (hob->profile.log2_max_mcgs) + profile_input_data_p->log2_mcg_hash_proportion;
        if ((int)hob->profile.log2_mcg_hash_size < 0) {
            hob->profile.log2_mcg_hash_size = 0;
        }
    }

    XHH_PRINT_PROFILE(&(hob->profile));
    MTL_DEBUG4("Leaving XHH_calculate_profile\n");

    if (sugg_profile_p != NULL) { 
            sugg_profile_p->num_mw = (u_int32_t)hob->profile.num_mem_windows;
            sugg_profile_p->num_qp = (u_int32_t)hob->profile.max_num_qps;
            sugg_profile_p->num_cq = (u_int32_t)hob->profile.max_num_cqs;
            sugg_profile_p->num_pd = (u_int32_t)hob->profile.max_num_pds;
            sugg_profile_p->num_mr = (u_int32_t)hob->profile.num_external_mem_regions;
            sugg_profile_p->max_qp_ous_rd_atom = (1U<<hob->profile.log2_inflight_rdma_per_qp);
    }
    if ((profile_user_data != NULL) && (profile_user_data->require != 0) && (need_to_loop == TRUE)) {
            MTL_ERROR1("XHH_calculate_default_profile:  Provided profile requires too many resources\n");
	    FREE(local_user_profile_p);
	    FREE(profile_input_data_p);
	    FREE(ddr_alloc_vec_p);
            return HH_ENOMEM;
    }
    FREE(local_user_profile_p);
    FREE(profile_input_data_p);
    FREE(ddr_alloc_vec_p);
    return HH_OK;
}

/*****************************************************************************
******************************************************************************
************************  HOB Interface FUNCTIONS ****************************
******************************************************************************
******************************************************************************/

/******************************************************************************
 *  Function:     XHH_hob_close_hca
 *
 *  Description:  This function stops HCA hardware activity and frees all associated resources.
 *
 *  input:
 *                hca_hndl
 *  output: 
 *                none
 *  returns:
 *                HH_OK
 *                HH_EINVAL_HCA_HNDL
 *                HH_ERR
 *
 *  Comments:     If any errors occur, continue process of de-allocating resources.  However, log the errors,
 *                and return HH_ERR instead of HH_OK
 *
 *****************************************************************************/
HH_ret_t XHH_hob_close_hca_internal(HH_hca_hndl_t  hca_hndl, MT_bool invoked_from_destroy)
{
  /* TBD - Complete function */
    XHH_cmd_status_t    cmd_ret;
    MT_bool             have_error = FALSE;
    HH_ret_t            ret = HH_OK;
    MT_phys_addr_t         *ddr_alloc_area;
    MT_size_t           *ddr_alloc_size;
    u_int32_t           i;
    u_int16_t           num_ports;
    call_result_t       res;
    XHH_hob_t  xhh_hob_p;
    DECLARE_FATAL_VARS;

    MT_RETURN_IF_LOW_STACK(XHH_WATERMARK);
    if (MOSAL_get_exec_ctx() != MOSAL_IN_TASK) {
        MTL_ERROR1("XHH_hob_close_hca: NOT IN TASK CONTEXT)\n");
        return HH_ERR;
    }

    if (hca_hndl == NULL) {
        MTL_ERROR1("XHH_hob_close_hca : ERROR : Invalid HCA handle\n");
        return HH_EINVAL_HCA_HNDL;
    }
    xhh_hob_p = XHHOBP(hca_hndl);
    
    /* uninterruptible acquire.  Want to be sure to clean up */
    MOSAL_mutex_acq_ui(&(xhh_hob_p->mtx));
    if (hca_hndl->status == HH_HCA_STATUS_CLOSED) {
        MOSAL_mutex_rel(&(xhh_hob_p->mtx));
        MTL_ERROR1("XHH_hob_close_hca: Device already closed\n");
        return HH_EINVAL_HCA_HNDL;
    }
    
    /* move the HCA to CLOSING state, preserving fatal indicators */
    MOSAL_spinlock_dpc_lock(&xhh_hob_p->fatal_spl);
    if ((xhh_hob_p->xhh_state & XHH_STATE_RUNNING) == 0) {
        MOSAL_spinlock_unlock(&xhh_hob_p->fatal_spl);
        MOSAL_mutex_rel(&(xhh_hob_p->mtx));
        MTL_ERROR1(MT_FLFMT("XHH_hob_close_hca:  already invoked"));
        return HH_EBUSY;
    }
    xhh_hob_p->xhh_state &= XHH_STATE_HAVE_ANY_FATAL;
    xhh_hob_p->xhh_state |= XHH_STATE_CLOSING;
    MOSAL_spinlock_unlock(&xhh_hob_p->fatal_spl);

    /* transfer to closing state */    
    WAIT_IF_FATAL(xhh_hob_p);
    if (have_fatal == FALSE) {
        num_ports = xhh_hob_p->dev_lims.num_ports;
        for (i = 1; i <= num_ports; i++) {
            cmd_ret = XHH_cmd_CLOSE_IB(xhh_hob_p->cmd, (IB_port_t) i);
            if (cmd_ret != XHH_CMD_STAT_OK) {
                MTL_ERROR1("XHH_hob_close_hca: XHH_cmd_CLOSE_IB error (%d)\n", cmd_ret);
                have_error = TRUE;
            }
        }
    }

    /* test if a fatal error occurred during CLOSE_IB. */
    if (have_fatal == FALSE) {
        WAIT_IF_FATAL(xhh_hob_p);
    }
    xhh_hob_p->compl_eq = XHH_INVALID_EQN;
    xhh_hob_p->ib_eq = XHH_INVALID_EQN;
    /* destroy eventq mgr.  Event manager must destroy all EQs */
    ret = XHH_eventp_destroy( xhh_hob_p->eventp );
    if (ret != HH_OK) {
        MTL_ERROR1("XHH_hob_close_hca: XHH_eventp_destroy error (%d)\n", ret);
        have_error = TRUE;
    }
    xhh_hob_p->eventp = (XHH_eventp_t)XHH_INVALID_HNDL;

    if (xhh_hob_p->mcgm != (XHH_mcgm_t)XHH_INVALID_HNDL) {
        ret = XHH_mcgm_destroy(xhh_hob_p->mcgm );
        if (ret != HH_OK) {
            MTL_ERROR1("XHH_hob_close_hca: XHH_mcgm_destroy error (%d)\n", ret);
            have_error = TRUE;
        }
        xhh_hob_p->mcgm = (XHH_mcgm_t)XHH_INVALID_HNDL;
    }

    MTL_DEBUG4("%s: calling MOSAL_iomem_unmap FOR KAR = " VIRT_ADDR_FMT "\n", __func__,
               (MT_virt_addr_t) xhh_hob_p->kar_addr);
    if ((res = (MOSAL_iomem_unmap(xhh_hob_p->kar_iomem_obj, MOSAL_get_kernel_prot_ctx()))) != MT_OK) {
        MTL_ERROR1("XHH_hob_close_hca: MOSAL_iomem_unmap error for kar (%d:%s)\n", res, mtl_strerror_sym(res));
        have_error = TRUE;
    }
    xhh_hob_p->kar_addr = (MT_virt_addr_t) 0;
    
    ret = XHH_uar_destroy(xhh_hob_p->kar);
    if (ret != HH_OK) {
        MTL_ERROR1("XHH_hob_close_hca: XHH_uar_destroy error (%d)\n", ret);
        have_error = TRUE;
    }
    xhh_hob_p->kar = (XHH_uar_t)XHH_INVALID_HNDL;


    if (xhh_hob_p->udavm_use_priv) {
        ret = XHH_udavm_destroy(xhh_hob_p->udavm);
        if (ret != HH_OK) {
            MTL_ERROR1("XHH_hob_close_hca: XHH_udavm_destroy error (%d)\n", ret);
            have_error = TRUE;
        }
        xhh_hob_p->udavm = (XHH_udavm_t)XHH_INVALID_HNDL;

        ret = XHH_mrwm_deregister_mr(xhh_hob_p->mrwm, xhh_hob_p->udavm_lkey);
        if (ret != HH_OK) {
            MTL_ERROR1("XHH_hob_close_hca: XHH_mrwm_deregister_mr error (%d)\n", ret);
            have_error = TRUE;
        }

        if ((res = MOSAL_iomem_unmap(xhh_hob_p->udavm_iomem_obj,MOSAL_get_kernel_prot_ctx())) != MT_OK) {
            MTL_ERROR1("XHH_hob_close_hca: MOSAL_iomem_unmap error for udavm (%d:%s)\n", res, mtl_strerror_sym(res));
            have_error = TRUE;
        }
        xhh_hob_p->udavm_table = (MT_virt_addr_t) NULL;

        ret = XHH_lamm_free(xhh_hob_p->lamm, xhh_hob_p->udavm_table_ddr, xhh_hob_p->udavm_table_size);
        if (ret != HH_OK) {
            MTL_ERROR1("XHH_hob_close_hca: XHH_lamm_free error (%d)\n", ret);
            have_error = TRUE;
        }
 
    }
    xhh_hob_p->udavm = (XHH_udavm_t)XHH_INVALID_HNDL;
    xhh_hob_p->udavm_table = (MT_virt_addr_t) NULL;
    xhh_hob_p->udavm_table_ddr  = (MT_phys_addr_t) 0;
    xhh_hob_p->udavm_table_size = 0;
    xhh_hob_p->udavm_lkey = 0;

    if (xhh_hob_p->srqm != (XHH_srqm_t)XHH_INVALID_HNDL) { /* SRQs are supported - SRQM exists */
      ret = XHH_srqm_destroy( xhh_hob_p->srqm);
      if (ret != HH_OK) {
          MTL_ERROR1("XHH_hob_close_hca: XHH_srqm_destroy error %s(%d)\n", HH_strerror_sym(ret), ret);
          have_error = TRUE;
      }
      xhh_hob_p->srqm = (XHH_srqm_t)XHH_INVALID_HNDL;
    }
    
    ret = XHH_qpm_destroy( xhh_hob_p->qpm, have_error);
    if (ret != HH_OK) {
        MTL_ERROR1("XHH_hob_close_hca: XHH_qpm_destroy error (%d)\n", ret);
        have_error = TRUE;
    }
    xhh_hob_p->qpm = (XHH_qpm_t)XHH_INVALID_HNDL;

    FREE(xhh_hob_p->init_ib_props);
    xhh_hob_p->init_ib_props = ( XHH_port_init_props_t *) NULL;

    ret = XHH_cqm_destroy( xhh_hob_p->cqm, have_error);
    if (ret != HH_OK) {
        MTL_ERROR1("XHH_hob_close_hca: XHH_cqm_destroy error (%d)\n", ret);
        have_error = TRUE;
    }
    xhh_hob_p->cqm = (XHH_cqm_t)XHH_INVALID_HNDL;

    ret = XHH_mrwm_destroy(xhh_hob_p->mrwm, have_error);
    if (ret != HH_OK) {
        MTL_ERROR1("XHH_hob_close_hca: XHH_mrwm_destroy error (%d)\n", ret);
        have_error = TRUE;
    }
    xhh_hob_p->mrwm = (XHH_mrwm_t)XHH_INVALID_HNDL;

    ret = XHH_uldm_destroy(xhh_hob_p->uldm );
    if (ret != HH_OK) {
        MTL_ERROR1("XHH_hob_close_hca: XHH_uldm_destroy error (%d)\n", ret);
        have_error = TRUE;
    }
    xhh_hob_p->uldm = (XHH_uldm_t)XHH_INVALID_HNDL;

    ret = XHH_cmd_revoke_lamm(xhh_hob_p->cmd);
    if (ret != HH_OK) {
        MTL_ERROR1("XHH_hob_close_hca: XHH_cmd_revoke_lamm error (%d)\n", ret);
        have_error = TRUE;
    }

    ddr_alloc_area = (MT_phys_addr_t *) &(xhh_hob_p->ddr_alloc_base_addrs_vec);
    ddr_alloc_size = (MT_size_t *)&(xhh_hob_p->ddr_alloc_size_vec);
    for (i = 0; i < xhh_hob_p->profile.ddr_alloc_vec_size; i++, ddr_alloc_area++, ddr_alloc_size++) {
        ret = *ddr_alloc_area != XHH_DDRMM_INVALID_PHYS_ADDR ?
          XHH_lamm_free(xhh_hob_p->lamm,*ddr_alloc_area, (1U<< (*ddr_alloc_size))) : HH_OK;
        if (ret != HH_OK) {
            MTL_ERROR1("XHH_hob_close_hca: XHH_lamm_free error (%d). i = %d\n", ret, i);
            have_error = TRUE;
        }
    }

    /* test for fatal again here */
    if (have_fatal == FALSE) {
        WAIT_IF_FATAL(xhh_hob_p);
    }
    if (have_fatal == FALSE) {
        MTL_TRACE1("XHH_hob_close_hca: Performing XHH_cmd_CLOSE_HCA (no fatal)\n");
#ifdef SIMULATE_HALT_HCA
        cmd_ret = XHH_cmd_CLOSE_HCA(xhh_hob_p->cmd);
#else
        cmd_ret = XHH_cmd_CLOSE_HCA(xhh_hob_p->cmd, FALSE);
#endif
        if (cmd_ret != XHH_CMD_STAT_OK) {
            MTL_ERROR1("XHH_hob_close_hca: XHH_cmd_CLOSE_HCA error (%d)\n", cmd_ret);
            have_error = TRUE;
        }
    }
    hca_hndl->status = HH_HCA_STATUS_CLOSED;

    /* move state to "CLOSED"*/
    MOSAL_spinlock_dpc_lock(&xhh_hob_p->fatal_spl);
    xhh_hob_p->xhh_state &= XHH_STATE_HAVE_ANY_FATAL;
    xhh_hob_p->xhh_state |= XHH_STATE_CLOSED;
    MOSAL_spinlock_unlock(&xhh_hob_p->fatal_spl);
    MOSAL_mutex_rel(&(xhh_hob_p->mtx));

    MTL_TRACE2("XHH_hob_close_hca: device name %s\n", hca_hndl->dev_desc);

    if (have_fatal == FALSE) {
        WAIT_IF_FATAL(xhh_hob_p);
    }
    if ((have_fatal == TRUE) && (invoked_from_destroy == FALSE)) {
	MT_START_FATAL_HANDLING;
        MTL_TRACE1("XHH_hob_close_hca: HAVE FATAL, restarting\n");
        ret = XHH_hob_restart(hca_hndl);
        if (ret != HH_OK) {
            MTL_ERROR1("XHH_hob_close_hca: XHH_hob_restart error (%d)\n", ret);
            have_error = TRUE;
        }
	MT_END_FATAL_HANDLING;
    }
    if (have_error && (have_fatal == FALSE)) {
        return HH_ERR;
    } else {
        return(HH_OK);
    }
} /* XHH_hob_close_hca_internal */

HH_ret_t XHH_hob_close_hca(HH_hca_hndl_t  hca_hndl)
{
    MT_RETURN_IF_LOW_STACK(XHH_WATERMARK);
    return (XHH_hob_close_hca_internal(hca_hndl,FALSE));
}


/******************************************************************************
 *  Function:     XHH_hob_destroy
 *
 *  Description:  Deregister given device from HH and free the HH object.
 *
 *  input:
 *                hca_hndl
 *  output: 
 *                none
 *  returns:
 *                HH_OK
 *                HH_EINVAL_HCA_HNDL
 *
 *  Comments:     If HCA is still open,XHH_hob_close_hca() is 
 *                invoked before freeing the XHH_hob.
 *
 *                Returns HH_EINVAL_HCA_HNDL if any function called internally fails  
 *
 *****************************************************************************/
HH_ret_t    XHH_hob_destroy_internal(HH_hca_hndl_t hca_hndl, MT_bool restarting)
{
  HH_ret_t  ret, fn_ret = HH_OK;
  XHH_cmd_status_t   cmd_ret;
  XHH_hob_t          hob_p;
  int                int_ret = 0;
  call_result_t      mosal_ret;
  MT_bool            have_fatal = FALSE;

  MT_RETURN_IF_LOW_STACK(XHH_WATERMARK);
  if (MOSAL_get_exec_ctx() != MOSAL_IN_TASK) {
      MTL_ERROR1("XHH_hob_destroy: NOT IN TASK CONTEXT)\n");
      return HH_ERR;
  }

  if (hca_hndl == NULL) {
      MTL_ERROR1("XHH_hob_close_hca : ERROR : Invalid HCA handle\n");
      return HH_EINVAL_HCA_HNDL;
  }
  
  /* return ERROR if device is still open */
  if (hca_hndl->status == HH_HCA_STATUS_OPENED) {
      MTL_ERROR1("XHH_hob_destroy:  Unloading device %s: while it is still open.  Attempting to close it.\n", hca_hndl->dev_desc);
      ret = XHH_hob_close_hca_internal(hca_hndl, TRUE);
      if (ret != HH_OK) {
        MTL_ERROR1("XHH_hob_destroy: Could not close device %s: not opened or unknown (err=%d)\n", hca_hndl->dev_desc, ret);
        fn_ret = HH_EINVAL_HCA_HNDL;
      }
      hca_hndl->status = HH_HCA_STATUS_CLOSED;
  }

  MTL_TRACE2("XHH_hob_destroy: removing the device %s\n",  hca_hndl->dev_desc);
  hob_p = XHHOBP(hca_hndl);

  /* move the HCA to DESTROYING state, preserving fatal indicators */
  MOSAL_spinlock_dpc_lock(&hob_p->fatal_spl);
  if ((hob_p->xhh_state & XHH_STATE_DESTROYING) != 0) {
      MOSAL_spinlock_unlock(&hob_p->fatal_spl);
      MTL_ERROR1(MT_FLFMT("XHH_hob_destroy:  already invoked"));
      return HH_EBUSY;
  }
  hob_p->xhh_state &= XHH_STATE_HAVE_ANY_FATAL;
  hob_p->xhh_state |= XHH_STATE_DESTROYING;
  MOSAL_spinlock_unlock(&hob_p->fatal_spl);
  
  /* release the fatal signalling thread */
  hob_p->fatal_thread_obj.have_fatal = FALSE;
  MOSAL_syncobj_signal(&hob_p->fatal_thread_obj.fatal_err_sync);
  mosal_ret = MOSAL_syncobj_waiton(&(hob_p->fatal_thread_obj.stop_sync), 10000000);
  if (mosal_ret != MT_OK) {
      if (mosal_ret == MT_EINTR) {
          MTL_DEBUG1(MT_FLFMT("%s: Received OS interrupt while initializing fatal error thread (err = %d)"), 
                     __func__,mosal_ret);
          fn_ret = HH_EINTR;
      } else {
          MTL_ERROR1(MT_FLFMT("%s: Timeout on destroying fatal error thread (err = %d)"), 
                     __func__,mosal_ret);
          fn_ret = HH_ERR;
      }
  }

  if (!restarting) {
      /* if a kernel application is still active, NULL-out its "user-level" ops-table pointer so that
       * attempts to access any "user-level" HHUL functions (in kernel, for a kernel app) will be intercepted
       * by the macros defined in file hhul.h before any actual invocation is attempted 
       */
      if (hca_hndl->ul_if_ops_p) {
	  *(int **)(hca_hndl->ul_if_ops_p) = NULL;
      }

      /* unregister the device from HH */
      ret = HH_rem_hca_dev_prep(hca_hndl);
      if (ret != HH_OK) {
	  MTL_ERROR1("XHH_hob_destroy: Could not remove device 0x%p: unknown (%d)\n", hca_hndl, ret);
	  fn_ret = HH_EINVAL_HCA_HNDL;
      }
  
  } else {
      hca_hndl->status = HH_HCA_STATUS_ZOMBIE;
  }

  /* destroy objects created in hob_create, and issue SYS_DIS command to Tavor */
  ret = XHH_lamm_destroy(hob_p->lamm);
  if (ret != HH_OK) {
      MTL_ERROR1("XHH_hob_destroy: Could not destroy lamm object (err = %d)\n", ret);
      fn_ret = HH_ERR;
  }
  /* do SYS_DIS only if do not have a fatal error state */
  if ((hob_p->xhh_state & XHH_STATE_HAVE_ANY_FATAL) == 0) {
      have_fatal = FALSE;
      cmd_ret = XHH_cmd_SYS_DIS(hob_p->cmd);
      if (cmd_ret != XHH_CMD_STAT_OK) {
          MTL_ERROR1("XHH_hob_destroy: SYS_DIS command failed (err = %d)\n", cmd_ret);
          if (cmd_ret == XHH_CMD_STAT_EFATAL) {
              have_fatal = TRUE;
          }
          fn_ret = HH_ERR;
      }
   } else {
      /* halt the HCA if delayed-halt flag was set, in fatal case only, 
       * to make sure that there is no PCI activity.  If fatal occurred during SYS_DIS above,
       * HCA was already closed, so don't need the "halt hca" operation
       */
      have_fatal = TRUE;
      if (hob_p->module_flags.fatal_delay_halt != 0) {
          MTL_DEBUG1(MT_FLFMT("%s: performing delayed halt-hca"), __func__);
          XHH_hob_halt_hca(hob_p);
      }
   }
      
  ret = XHH_cmd_destroy(hob_p->cmd);
  if (ret != HH_OK) {
      MTL_ERROR1("XHH_hob_destroy: Could not destroy cmd object (err = %d)\n", ret);
      fn_ret = HH_ERR;
  }

  /* do PCI reset here if have a catastrophic error */
  if (have_fatal == TRUE) {
    /* perform sw reset */
      MTL_ERROR1(MT_FLFMT("%s: FATAL ERROR "), __func__);

	{
	     MT_START_FATAL_HANDLING;
	      /* Do the Tavor RESET */
	     XHH_hob_reset_device(&(hob_p->hw_props));

	      /* now, rewrite the PCI configuration */
	      if (hob_p->pci_bridge_info.is_valid == TRUE) {
	         XHH_write_pci_config(hob_p->pci_bridge_info.pci_dev,
	                           &hob_p->pci_bridge_info.config);
	      }
	      if (hob_p->pci_hca_info.is_valid == TRUE) {
	          XHH_write_pci_config(hob_p->hw_props.pci_dev,
	                           &hob_p->pci_hca_info.config);
	      }
	      MT_END_FATAL_HANDLING;
	}
  }

  int_ret = VIP_delay_unlock_destroy(hob_p->delay_unlocks);
  if (int_ret != 0) {
      MTL_ERROR1("XHH_hob_destroy: Could not destroy delay_unlocks (err = %d)\n", int_ret);
      fn_ret = HH_ERR;
  }
  
  if (hob_p->fw_error_iomem_obj != (MOSAL_iomem_t) NULL)  {
     MOSAL_iomem_unmap(hob_p->fw_error_iomem_obj, MOSAL_get_kernel_prot_ctx());
  }

  if (hob_p->fw_error_buf != NULL) {
      FREE(hob_p->fw_error_buf);
  }

  MOSAL_mutex_free(&(hob_p->mtx));
  /* Finally, free the XHH object */
  FREE(hob_p);

  return(fn_ret);
}
/******************************************************************************
 *  Function:     XHH_hob_destroy
 *
 *  Description:  Deregister given device from HH and free the HH object.
 *
 *  input:
 *                hca_hndl
 *  output: 
 *                none
 *  returns:
 *                HH_OK
 *                HH_ERR  (not in task context)
 *                HH_EINVAL_HCA_HNDL (invalid HCA handle, or internal call failure
 *
 *  Comments:     Wait for HCA closed before freeing the XHH_hob (this is only
 *                active if have hotplug support).
 *
 *****************************************************************************/
HH_ret_t    XHH_hob_destroy(HH_hca_hndl_t hca_hndl)
{	    
	HH_ret_t ret, ret1;
        ret = XHH_hob_waitfor_hca_closed(hca_hndl);
        if (ret != HH_OK) {
	    return ret;
	}
	ret = XHH_hob_destroy_internal(hca_hndl, FALSE);
	if (ret != HH_OK) {
		MTL_ERROR1(MT_STDFMT("Failed XHH_hob_destroy for handle=%p (%d,%s)"),
			   hca_hndl, ret, HH_strerror_sym(ret));
	}
        ret1 = HH_notify_rem_hca_dev(hca_hndl);
	if (ret1 != HH_OK) {
		MTL_ERROR1(MT_STDFMT("Failed XHH_hob_destroy for handle=%p (%d,%s)"),
			   hca_hndl, ret1, HH_strerror_sym(ret1));
	}
	return (ret == HH_OK ? ret1 : ret);
}
/******************************************************************************
 *  Function:     XHH_hob_open_hca
 *
 *  Description:  This function opens the given HCA and initializes the HCA with 
 *                given properties/ capabilities.  if prop_props_p is NULL a default 
 *                HCA profile will be set up.
 *
 *  input:
 *                hca_hndl
 *                prop_props_p - Proprietary properties (Non IB)
 *  output: 
 *                none
 *  returns:
 *                HH_OK
 *                HH_EINVAL
 *                HH_EBUSY
 *
 *  Comments:     If HCA is still open,XHH_hob_close_hca() is 
 *                invoked before freeing the XHH_hob.
 *
 *                Returns HH_EINVAL_HCA_HNDL if any function called internally fails  
 *
 *****************************************************************************/
HH_ret_t XHH_hob_open_hca(HH_hca_hndl_t  hca_hndl, 
                                 EVAPI_hca_profile_t  *prop_props_p,
                                 EVAPI_hca_profile_t  *sugg_profile_p)
{
    MT_virt_addr_t            kar_addr;
    HH_ret_t               ret;
    XHH_cmd_status_t       cmd_ret;
    XHH_hca_props_t        *local_hca_props_p = NULL;
    MT_phys_addr_t         *ddr_alloc_area;
    MT_size_t              *ddr_alloc_size;
    u_int32_t              i;
    XHH_internal_mr_t      *udav_internal_mr_p = NULL;
    MT_size_t              udav_entry_size = 0, udav_table_size = 0;
    MT_phys_addr_t            udav_phys_addr = 0;
    MT_virt_addr_t            udav_virt_addr = 0;
    VAPI_lkey_t            dummy_key;
    XHH_eventp_res_t       *event_res_p = NULL;
    u_int16_t              num_ports, last_port_initialized;
    XHH_hob_t              xhh_hob_p;
    VAPI_size_t            udav_vapi_size;
    XHH_qpm_init_t         *xhh_qpm_init_params_p;
    XHH_cqm_init_t         xhh_cqm_init_params;
    XHH_srqm_init_t        xhh_srqm_init_params;
    XHH_mcgm_init_t        xhh_mcgm_init_params;
    MT_bool            have_fatal = FALSE;
    call_result_t          rc;

    MT_RETURN_IF_LOW_STACK(1000);
    MTL_DEBUG4("Entering XHH_hob_open_hca\n");
    if (MOSAL_get_exec_ctx() != MOSAL_IN_TASK) {
        MTL_ERROR1("XHH_hob_open_hca: NOT IN TASK CONTEXT)\n");
        return HH_ERR;
    }


    if (hca_hndl == NULL) {
        MTL_ERROR1("XHH_hob_open_hca : ERROR : Invalid HCA handle\n");
        return HH_EINVAL_HCA_HNDL;
    }
    xhh_hob_p = XHHOBP(hca_hndl);

    if (xhh_hob_p == NULL) {
        MTL_ERROR1("XHH_hob_open_hca: ERROR : No device registered\n");
        return HH_EAGAIN;
    }

    local_hca_props_p = (XHH_hca_props_t *)MALLOC(sizeof(XHH_hca_props_t));
    if (local_hca_props_p == NULL) {
       MT_COV_MEM_FLTR_START
       return HH_EAGAIN;
       MT_COV_MEM_FLTR_END
    }

    udav_internal_mr_p = (XHH_internal_mr_t *)MALLOC(sizeof(XHH_internal_mr_t));
    if (udav_internal_mr_p == NULL) {
       MT_COV_MEM_FLTR_START
       FREE(local_hca_props_p);
       return HH_EAGAIN;
       MT_COV_MEM_FLTR_END
    }

    event_res_p = (XHH_eventp_res_t *)MALLOC(sizeof(XHH_eventp_res_t));
    if (event_res_p == NULL) {
       MT_COV_MEM_FLTR_START
       FREE(local_hca_props_p);
       FREE(udav_internal_mr_p);
       return HH_EAGAIN;
       MT_COV_MEM_FLTR_END
    }

  /* Get user profile if available -- if not, use default proportionally to resources. */

  /* DDR parameters.  Get from default profile.  For each one, check that does not exceed */
  /* The maximum resource value supportable by the installed card. */

/*
   Objects:
           cmd;     -- already exists
           lamm;   -- already exists
           uldm;    -- uar, pd:  log2_max_uar, log2_max_pg_sz, max_pd
           mrwm;    -- log2_mpt_sz (log2 of number of entries in MPT)
                       log2_mtt_sz (Log2 of number of entries in the MTT) 
                       max_mem_reg (Maximum memory regions to be allocated in the MPT for external registration only)
                       max_mem_reg_internal (Maximum memory regions to be alloc in the MPT for internal use only (WQEs and CQEs buffers) )
                       max_mem_win (Maximum memory windows to be allocated in the MPT)
           cqm;     -- log2_max_cq,
           eecm;    -- 
           qpm;     -- log2_max_qp,
                       privileged_ud_av (boolean)
           udavm;   -- max_av
           mcgm;    -- num mcg's: IBTA min = 512, 8 QPs/group
           eventp;  -- 64 event queues
           kar;     -- UAR 0 (no extra resources needed)
        
*/

    /* Test if have fatal error, and move xhh_state to "opening" */
    MOSAL_spinlock_dpc_lock(&xhh_hob_p->fatal_spl);
    if ((xhh_hob_p->xhh_state & XHH_STATE_HAVE_ANY_FATAL) != 0) {
        /* already in FATAL state */
        MOSAL_spinlock_unlock(&xhh_hob_p->fatal_spl);
        MTL_DEBUG4(MT_FLFMT("%s: already in FATAL state"), __func__);
	FREE(local_hca_props_p);
	FREE(udav_internal_mr_p);
	FREE(event_res_p);
        MT_RETURN(HH_EFATAL);
    } else if (xhh_hob_p->xhh_state != XHH_STATE_CLOSED) {
        MOSAL_spinlock_unlock(&xhh_hob_p->fatal_spl);
        MTL_ERROR1(MT_FLFMT("XHH_hob_open_hca: ERROR : Device not closed. state = 0x%x"),xhh_hob_p->xhh_state );
	FREE(local_hca_props_p);
	FREE(udav_internal_mr_p);
	FREE(event_res_p);
        MT_RETURN(HH_EBUSY);
    } 
    xhh_hob_p->xhh_state = XHH_STATE_OPENING;
    MOSAL_spinlock_unlock(&xhh_hob_p->fatal_spl);
    
    /* get the MUTEX */
    if (MOSAL_mutex_acq(&(xhh_hob_p->mtx), TRUE) != MT_OK) {
        MTL_ERROR1(MT_FLFMT("XHH_hob_open_hca: received signal. returning"));
        ret = HH_EINTR;
        goto post_state_change_error;
    }
    if (hca_hndl->status == HH_HCA_STATUS_OPENED) {
        MTL_ERROR1("XHH_hob_open_hca: ERROR : Device already open\n");
        ret = HH_EBUSY;
        goto post_mutex_acquire_err;
    }

    if (prop_props_p != NULL) {
        XHH_PRINT_USR_PROFILE(prop_props_p);
    }
    ret = XHH_calculate_profile(xhh_hob_p, prop_props_p, sugg_profile_p);
    if (ret != HH_OK) {
      MTL_ERROR1(MT_FLFMT("XHH_hob_open_hca: could not create internal profile (%d)"), ret);
      if (sugg_profile_p != NULL) {
          XHH_PRINT_USR_PROFILE(sugg_profile_p);
      }
      //ret = HH_ERR;
      goto post_mutex_acquire_err;
    }
    if (sugg_profile_p != NULL) {
        XHH_PRINT_USR_PROFILE(sugg_profile_p);
    }

    /* check profile against QUERY_DEV_LIMS data*/
    ret = XHH_check_profile(xhh_hob_p);
    if (ret != HH_OK) {
      MTL_ERROR1("XHH_hob_open_hca: Profile check failed (%d)\n", ret);
      ret = HH_ERR;
      goto post_mutex_acquire_err;
    }

    /*  Do lamm allocation here, because we need the allocated MCG base address */
    /*  for the INIT_HCA command following the centralized DDR allocation */
    calculate_ddr_alloc_vec(xhh_hob_p, &(xhh_hob_p->profile), &(xhh_hob_p->ddr_alloc_size_vec));

    /* Allocate all required DDR areas */
    ret = XHH_lamm_alloc_sz_aligned(xhh_hob_p->lamm, 
                             xhh_hob_p->profile.ddr_alloc_vec_size,      /*number of chunks */
                             (MT_size_t *) &(xhh_hob_p->ddr_alloc_size_vec), /* IN  */
                             (MT_phys_addr_t *)&(xhh_hob_p->ddr_alloc_base_addrs_vec) );  /* OUT */
    if (ret != HH_OK) {
        MTL_ERROR1("XHH_hob_open_hca: could not allocate required areas in DDR (%s)\n", HH_strerror_sym(ret));
        ret = HH_ERR;
        goto post_mutex_acquire_err;
    }
    
    /* call cmd interface to initialize its mailboxes in DDR */
    ret = XHH_cmd_assign_lamm(xhh_hob_p->cmd, xhh_hob_p->lamm);
    if (ret != HH_OK) {
      MTL_ERROR1("XHH_hob_open_hca: Failed XHH_cmd_assign_lamm (%s)\n",HH_strerror_sym(ret)); 
      goto cmd_assign_lamm_err;
    }
    
    /* set up parameters for INIT HCA */
    memset(local_hca_props_p, 0, sizeof(XHH_hca_props_t));
#ifdef MT_LITTLE_ENDIAN
    local_hca_props_p->he = TAVOR_IF_HOST_LTLENDIAN;
#else
    local_hca_props_p->he = TAVOR_IF_HOST_BIGENDIAN;
#endif
    local_hca_props_p->re = FALSE;   /* not a router */
    local_hca_props_p->udp = TRUE;   /* check port in UD AV */ 
    local_hca_props_p->ud =  xhh_hob_p->profile.use_priv_udav; 
    
    local_hca_props_p->qpc_eec_cqc_eqc_rdb_parameters.cqc_base_addr = 
        GET_DDR_ADDR(xhh_hob_p->ddr_alloc_base_addrs_vec.cqc_base_addr,xhh_hob_p->ddr_props.dh,
                          xhh_hob_p->ddr_props.ddr_start_adr);
    local_hca_props_p->qpc_eec_cqc_eqc_rdb_parameters.eec_base_addr = 
        GET_DDR_ADDR(xhh_hob_p->ddr_alloc_base_addrs_vec.eec_base_addr,xhh_hob_p->ddr_props.dh,
                          xhh_hob_p->ddr_props.ddr_start_adr);
    local_hca_props_p->qpc_eec_cqc_eqc_rdb_parameters.eqc_base_addr = 
        GET_DDR_ADDR(xhh_hob_p->ddr_alloc_base_addrs_vec.eqc_base_addr,xhh_hob_p->ddr_props.dh,
                          xhh_hob_p->ddr_props.ddr_start_adr);
    local_hca_props_p->qpc_eec_cqc_eqc_rdb_parameters.qpc_base_addr = 
        GET_DDR_ADDR(xhh_hob_p->ddr_alloc_base_addrs_vec.qpc_base_addr,xhh_hob_p->ddr_props.dh,
                          xhh_hob_p->ddr_props.ddr_start_adr);
    local_hca_props_p->qpc_eec_cqc_eqc_rdb_parameters.rdb_base_addr = 
        GET_DDR_ADDR(xhh_hob_p->ddr_alloc_base_addrs_vec.rdb_base_addr,xhh_hob_p->ddr_props.dh,
                          xhh_hob_p->ddr_props.ddr_start_adr);
    local_hca_props_p->qpc_eec_cqc_eqc_rdb_parameters.log_num_of_cq = 
      (u_int8_t)xhh_hob_p->profile.log2_max_cqs;
    // local_hca_props_p->qpc_eec_cqc_eqc_rdb_parameters.log_num_of_ee = xhh_hob_p->profile.log2_max_eecs;
    local_hca_props_p->qpc_eec_cqc_eqc_rdb_parameters.log_num_of_ee = 0;
    local_hca_props_p->qpc_eec_cqc_eqc_rdb_parameters.log_num_eq = xhh_hob_p->profile.log2_max_eqs;
    local_hca_props_p->qpc_eec_cqc_eqc_rdb_parameters.log_num_of_qp = (u_int8_t)xhh_hob_p->profile.log2_max_qps;
    local_hca_props_p->qpc_eec_cqc_eqc_rdb_parameters.eqpc_base_addr = 
        GET_DDR_ADDR(xhh_hob_p->ddr_alloc_base_addrs_vec.eqpc_base_addr,xhh_hob_p->ddr_props.dh,
                          xhh_hob_p->ddr_props.ddr_start_adr);
    local_hca_props_p->qpc_eec_cqc_eqc_rdb_parameters.log_num_of_srq = (u_int8_t)xhh_hob_p->profile.log2_max_srqs;
    local_hca_props_p->qpc_eec_cqc_eqc_rdb_parameters.srqc_base_addr = 
        GET_DDR_ADDR(xhh_hob_p->ddr_alloc_base_addrs_vec.srqc_base_addr,xhh_hob_p->ddr_props.dh,
                          xhh_hob_p->ddr_props.ddr_start_adr);
    local_hca_props_p->qpc_eec_cqc_eqc_rdb_parameters.eeec_base_addr = 
        GET_DDR_ADDR(xhh_hob_p->ddr_alloc_base_addrs_vec.eeec_base_addr,xhh_hob_p->ddr_props.dh,
                          xhh_hob_p->ddr_props.ddr_start_adr);
    
    local_hca_props_p->udavtable_memory_parameters.l_key = XHH_UDAVM_PRIV_RESERVED_LKEY;
    local_hca_props_p->udavtable_memory_parameters.pd    = XHH_RESERVED_PD(xhh_hob_p);
    local_hca_props_p->udavtable_memory_parameters.xlation_en = TRUE;
    
    local_hca_props_p->tpt_parameters.log_mpt_sz = (u_int8_t)xhh_hob_p->profile.log2_max_mpt_entries;
    local_hca_props_p->tpt_parameters.mpt_base_adr = 
        GET_DDR_ADDR(xhh_hob_p->ddr_alloc_base_addrs_vec.mpt_base_addr,xhh_hob_p->ddr_props.dh,
                          xhh_hob_p->ddr_props.ddr_start_adr);
    local_hca_props_p->tpt_parameters.mtt_base_addr = 
        GET_DDR_ADDR(xhh_hob_p->ddr_alloc_base_addrs_vec.mtt_base_addr,xhh_hob_p->ddr_props.dh,
                          xhh_hob_p->ddr_props.ddr_start_adr);
    local_hca_props_p->tpt_parameters.pfto = 0;   /* TBD -- not yet supported. Page Fault RNR Timeout */
    local_hca_props_p->tpt_parameters.mtt_segment_size = (u_int8_t)((xhh_hob_p->profile.log2_mtt_entries_per_seg + XHH_DDR_LOG2_MTT_ENTRY_SIZE)
                                                            - XHH_DDR_LOG2_MIN_MTT_SEG_SIZE);
    
    local_hca_props_p->uar_parameters.uar_base_addr = xhh_hob_p->hw_props.uar_base;
    local_hca_props_p->uar_parameters.uar_page_sz   = xhh_hob_p->profile.log2_uar_pg_size - 12;
    local_hca_props_p->uar_parameters.uar_scratch_base_addr = 
        GET_DDR_ADDR(xhh_hob_p->ddr_alloc_base_addrs_vec.uar_scratch_base_addr,xhh_hob_p->ddr_props.dh,
                          xhh_hob_p->ddr_props.ddr_start_adr);
    
    local_hca_props_p->multicast_parameters.log_mc_table_sz = (u_int8_t)xhh_hob_p->profile.log2_max_mcgs;
    local_hca_props_p->multicast_parameters.mc_base_addr    = 
        GET_DDR_ADDR(xhh_hob_p->ddr_alloc_base_addrs_vec.mcg_base_addr,xhh_hob_p->ddr_props.dh,
                          xhh_hob_p->ddr_props.ddr_start_adr);
    local_hca_props_p->multicast_parameters.mc_hash_fn      = 0;
    local_hca_props_p->multicast_parameters.log_mc_table_entry_sz = (u_int16_t)(xhh_hob_p->profile.log2_mcg_entry_size);
    local_hca_props_p->multicast_parameters.mc_table_hash_sz = 1 << (xhh_hob_p->profile.log2_mcg_hash_size);

  /* INIT_HCA command */
    cmd_ret = XHH_cmd_INIT_HCA(xhh_hob_p->cmd,local_hca_props_p);
    if (cmd_ret != XHH_CMD_STAT_OK) {
        MTL_ERROR1("XHH_hob_open_hca: CMD_error in XHH_cmd_INIT_HCA (%d)\n", cmd_ret);
        ret = HH_EAGAIN;
        goto init_hca_err;
    }

  /* Now, query HCA to get actual allocated parameters */
    cmd_ret = XHH_cmd_QUERY_HCA(xhh_hob_p->cmd, &(xhh_hob_p->hca_props));
    if (cmd_ret != XHH_CMD_STAT_OK) {
        MTL_ERROR1("XHH_hob_open_hca: CMD_error in XHH_cmd_QUERY_HCA (%d)\n", cmd_ret);
        ret = HH_EAGAIN;
        goto query_hca_err;
    }
    
    /* create uldm */
    ret = XHH_uldm_create(xhh_hob_p, xhh_hob_p->hw_props.uar_base, (u_int8_t) xhh_hob_p->profile.log2_max_uar,
                           (u_int8_t) xhh_hob_p->profile.log2_uar_pg_size, 
                           (u_int32_t) (xhh_hob_p->profile.max_num_pds + xhh_hob_p->dev_lims.num_rsvd_pds + XHH_NUM_RSVD_PD),
                           &(xhh_hob_p->uldm));                      if (ret != HH_OK) {
        MTL_ERROR1("XHH_hob_open_hca: could not create uldm (%d)\n", ret);
        goto uldm_create_err;
    }

    xhh_hob_p->mrwm_props.mtt_base = GET_DDR_ADDR(xhh_hob_p->ddr_alloc_base_addrs_vec.mtt_base_addr,xhh_hob_p->ddr_props.dh,
                          xhh_hob_p->ddr_props.ddr_start_adr);
    xhh_hob_p->mrwm_props.mpt_base = GET_DDR_ADDR(xhh_hob_p->ddr_alloc_base_addrs_vec.mpt_base_addr,xhh_hob_p->ddr_props.dh,
                          xhh_hob_p->ddr_props.ddr_start_adr);
    xhh_hob_p->mrwm_props.log2_mpt_sz = (u_int8_t)xhh_hob_p->profile.log2_max_mpt_entries;
    xhh_hob_p->mrwm_props.log2_mtt_sz = (u_int8_t)xhh_hob_p->profile.log2_max_mtt_entries;
    xhh_hob_p->mrwm_props.log2_mtt_seg_sz = (u_int8_t)xhh_hob_p->profile.log2_mtt_entries_per_seg;
    xhh_hob_p->mrwm_props.max_mem_reg = xhh_hob_p->profile.num_external_mem_regions;
    xhh_hob_p->mrwm_props.max_mem_reg_internal = xhh_hob_p->profile.max_num_qps + xhh_hob_p->profile.max_num_cqs;
    xhh_hob_p->mrwm_props.max_mem_win          = xhh_hob_p->profile.num_mem_windows;
    xhh_hob_p->mrwm_props.log2_max_mtt_segs    = (u_int8_t)(xhh_hob_p->profile.log2_max_mtt_entries - 
                                                           xhh_hob_p->mrwm_props.log2_mtt_seg_sz);
    xhh_hob_p->mrwm_props.log2_rsvd_mpts       = xhh_hob_p->dev_lims.log2_rsvd_mrws;
    xhh_hob_p->mrwm_props.log2_rsvd_mtt_segs   = xhh_hob_p->dev_lims.log2_rsvd_mtts;

    ret = XHH_mrwm_create(xhh_hob_p, &(xhh_hob_p->mrwm_props), &(xhh_hob_p->mrwm));
    if (ret != HH_OK) {
        MTL_ERROR1("XHH_hob_open_hca: could not create mrwm (%d)\n", ret);
        goto mrwm_create_err;
    }
    
    /* Create objects */
    memset(&xhh_cqm_init_params,0, sizeof(xhh_cqm_init_params));
    xhh_cqm_init_params.log2_max_cq = (u_int8_t) xhh_hob_p->profile.log2_max_cqs;
    xhh_cqm_init_params.log2_rsvd_cqs = xhh_hob_p->dev_lims.log2_rsvd_cqs; 
    ret = XHH_cqm_create(xhh_hob_p, &(xhh_cqm_init_params), &(xhh_hob_p->cqm));
    if (ret != HH_OK) {
        MTL_ERROR1("XHH_hob_open_hca: could not create cqm (%d)\n", ret);
        goto cqm_create_err;
    }
    
    /* create qpm object here */

    /* initialize INIT_IB parameters here for possible use by qpm */
    num_ports = xhh_hob_p->dev_lims.num_ports;

    xhh_hob_p->init_ib_props = (XHH_port_init_props_t *) MALLOC(num_ports * sizeof(XHH_port_init_props_t));
    if (!(xhh_hob_p->init_ib_props)) {
        MTL_ERROR1( "XHH_hob_open_hca: ERROR : cannot allocate memory for port init props)\n");
	ret= HH_EAGAIN;
        goto init_ib_props_malloc_err;
    }
    for (i = 1; i <= num_ports; i++) {
        /* redundant for now. However, leaving option for setting different properties per port */
        xhh_hob_p->init_ib_props[i-1].e = TRUE;
        xhh_hob_p->init_ib_props[i-1].g0 = FALSE;
        xhh_hob_p->init_ib_props[i-1].max_gid = (1 << (xhh_hob_p->dev_lims.log_max_gid));
        xhh_hob_p->init_ib_props[i-1].mtu_cap = xhh_hob_p->dev_lims.max_mtu;
        xhh_hob_p->init_ib_props[i-1].max_pkey = (1 << (xhh_hob_p->dev_lims.log_max_pkey));
        xhh_hob_p->init_ib_props[i-1].vl_cap   = xhh_hob_p->dev_lims.max_vl;
        xhh_hob_p->init_ib_props[i-1].port_width_cap = xhh_hob_p->dev_lims.max_port_width;
    }
    
    xhh_qpm_init_params_p = (XHH_qpm_init_t *)MALLOC(sizeof(XHH_qpm_init_t));
    if (xhh_qpm_init_params_p == NULL) {
       MT_COV_MEM_FLTR_START
       MTL_ERROR1( "XHH_hob_open_hca: ERROR : cannot allocate memory for qpm init params)\n");
       ret= HH_EAGAIN;
       goto init_qpm_params_malloc_err;
       MT_COV_MEM_FLTR_END
    }
    memset(xhh_qpm_init_params_p, 0, sizeof(XHH_qpm_init_t));
    xhh_qpm_init_params_p->rdb_base_index = 
        /* 32 low-order bits, right-shifted by log of size of rdb entry */
      (u_int32_t)(((u_int64_t)(GET_DDR_ADDR(xhh_hob_p->ddr_alloc_base_addrs_vec.rdb_base_addr,xhh_hob_p->ddr_props.dh,
                          xhh_hob_p->ddr_props.ddr_start_adr))) & (u_int64_t)0xFFFFFFFF ) 
                                  >> XHH_DDR_LOG2_RDB_ENTRY_SIZE;
    xhh_qpm_init_params_p->log2_max_qp = (u_int8_t) xhh_hob_p->profile.log2_max_qps;
    xhh_qpm_init_params_p->log2_max_outs_rdma_atom = xhh_hob_p->profile.log2_inflight_rdma_per_qp;
    xhh_qpm_init_params_p->log2_max_outs_dst_rd_atom = xhh_hob_p->dev_lims.log_max_ra_req_qp;
    xhh_qpm_init_params_p->n_ports = (u_int8_t)num_ports;
    xhh_qpm_init_params_p->port_props = 
        (xhh_hob_p->module_flags.legacy_sqp == TRUE ? NULL : xhh_hob_p->init_ib_props );
    xhh_qpm_init_params_p->log2_rsvd_qps = xhh_hob_p->dev_lims.log2_rsvd_qps;
    
    ret = XHH_qpm_create(xhh_hob_p, xhh_qpm_init_params_p, &(xhh_hob_p->qpm));
    if (ret != HH_OK) {
        MTL_ERROR1("XHH_hob_open_hca: could not create qpm %s(%d)\n", HH_strerror_sym(ret), ret);
        goto qpm_create_err;
    }

    if (xhh_hob_p->dev_lims.srq) {
      memset(&xhh_srqm_init_params, 0, sizeof(xhh_srqm_init_params));
      xhh_srqm_init_params.log2_max_srq = (u_int8_t)xhh_hob_p->profile.log2_max_srqs;
      xhh_srqm_init_params.log2_rsvd_srq = xhh_hob_p->dev_lims.log2_rsvd_srqs;
      ret= XHH_srqm_create(xhh_hob_p,&xhh_srqm_init_params, &(xhh_hob_p->srqm));
      if (ret != HH_OK) {
          MTL_ERROR1("XHH_hob_open_hca: could not create srqm - %s(%d)\n", HH_strerror_sym(ret), ret);
          goto srqm_create_err;
      }
    } else {
      xhh_hob_p->srqm= (XHH_srqm_t)XHH_INVALID_HNDL; /* SRQs are not supported */
    }

    /* CREATE ALL CONTAINED OBJECTS  */
    
    /* create UDAVm if privileged UDAV is set */
    if (xhh_hob_p->profile.use_priv_udav) {
        
        xhh_hob_p->udavm_use_priv = TRUE;
        
        /* create the table in DDR memory */
        udav_entry_size = (unsigned)(sizeof(struct tavorprm_ud_address_vector_st) / 8);
        udav_table_size = xhh_hob_p->profile.max_priv_udavs * udav_entry_size;

        ret = XHH_lamm_alloc(xhh_hob_p->lamm, udav_table_size, (u_int8_t)ceil_log2(udav_entry_size),
                          &udav_phys_addr);
        if (ret != HH_OK) {
            MTL_ERROR1("XHH_hob_open_hca: could not allocate protected udavm area in DDR(err = %d)\n", ret);
            goto udavm_lamm_alloc_err;
        }
        rc = MOSAL_iomem_map(MOSAL_get_kernel_prot_ctx(),
                              MOSAL_MEM_FLAGS_NO_CACHE | MOSAL_MEM_FLAGS_PERM_WRITE | MOSAL_MEM_FLAGS_PERM_READ,
                              udav_phys_addr,
                              udav_table_size,
                              &(xhh_hob_p->udavm_iomem_obj),
                              &udav_virt_addr);
                                                     
        if (rc != MT_OK) {
            MTL_ERROR1(MT_FLFMT("%s: MOSAL_iomem_map failed. physical address =" PHYS_ADDR_FMT " (%d:%s)"), 
                        __func__,udav_phys_addr, rc, mtl_strerror_sym(rc));
            goto udavm_mosal_map_err;
        }

        memset(udav_internal_mr_p, 0, sizeof(XHH_internal_mr_t));
        udav_internal_mr_p->force_memkey = TRUE;
        udav_internal_mr_p->memkey       = XHH_UDAVM_PRIV_RESERVED_LKEY;
        udav_internal_mr_p->pd           = XHH_RESERVED_PD(xhh_hob_p);
        udav_internal_mr_p->size         = udav_table_size;
        udav_internal_mr_p->start        = udav_virt_addr;
        udav_internal_mr_p->vm_ctx       = MOSAL_get_kernel_prot_ctx();
        if (udav_phys_addr) {
            VAPI_phy_addr_t udav_phy =  udav_phys_addr;
            udav_internal_mr_p->num_bufs = 1;      /*  != 0   iff   physical buffesrs supplied */
            udav_internal_mr_p->phys_buf_lst = &udav_phy;  /* size = num_bufs */
            udav_vapi_size  = (VAPI_size_t) udav_table_size;
            udav_internal_mr_p->buf_sz_lst = &udav_vapi_size;    /* [num_bufs], corresponds to phys_buf_lst */
        }

        xhh_hob_p->udavm_table_size = udav_table_size;
        xhh_hob_p->udavm_table      = udav_virt_addr;
        xhh_hob_p->udavm_table_ddr  = udav_phys_addr;

        ret = XHH_mrwm_register_internal(xhh_hob_p->mrwm, udav_internal_mr_p, &dummy_key);
        if (ret != HH_OK) {
            MTL_ERROR1("XHH_hob_open_hca: could not register created udavm table (%d)\n", ret);
            goto udavm_table_register_err;
        }
        xhh_hob_p->udavm_lkey = dummy_key;

        ret = XHH_udavm_create(&(xhh_hob_p->version_info), 
                               dummy_key,
                               udav_table_size,
                               udav_virt_addr,
                               &(xhh_hob_p->udavm));
        if (ret != HH_OK) {
            MTL_ERROR1("XHH_hob_open_hca: could not create udavm (%d)\n", ret);
            goto udavm_create_err;
        }

    } else {
        xhh_hob_p->udavm_use_priv = FALSE;
    }

    /* CREATE KAR (kernel UAR), using UAR 1 for this purpose. */
    rc = MOSAL_iomem_map(MOSAL_get_kernel_prot_ctx(),
                          MOSAL_MEM_FLAGS_NO_CACHE | MOSAL_MEM_FLAGS_PERM_WRITE,
                          (xhh_hob_p->hw_props.uar_base + (1UL << xhh_hob_p->profile.log2_uar_pg_size)),
                          (1 << xhh_hob_p->profile.log2_uar_pg_size),
                          &(xhh_hob_p->kar_iomem_obj),
                          &kar_addr);
    if (rc != MT_OK) {
        MTL_ERROR1(MT_FLFMT("%s: MOSAL_iomem_map failed. prot ctx=%d, pa=" PHYS_ADDR_FMT ", size=%d (%d:%s)"), 
                    __func__,MOSAL_get_kernel_prot_ctx(),
                   (MT_phys_addr_t) (xhh_hob_p->hw_props.uar_base),
                   (1 << xhh_hob_p->profile.log2_uar_pg_size),
                    rc, mtl_strerror_sym(rc));
        goto kar_map_phys_addr_err;
    }
    xhh_hob_p->kar_addr = kar_addr;
    MTL_DEBUG4("%s: MOSAL_iomem_map FOR KAR = " VIRT_ADDR_FMT "\n", __func__, kar_addr);
    ret = XHH_uar_create(&(xhh_hob_p->version_info), 1/* Kernel UAR page index */, 
                         (void *) kar_addr, &(xhh_hob_p->kar));
    if (ret != HH_OK) {
        MTL_ERROR1("XHH_hob_open_hca: could not create KAR (%d)\n", ret);
        xhh_hob_p->kar_addr = (MT_virt_addr_t) 0;
        goto kar_create_err;
    }
    
    //sharon: wrote fixed numbers till fw bug fixed
    if (XHH_DEV_LIM_MCG_ENABLED(xhh_hob_p)) {
        memset(&xhh_mcgm_init_params,0,sizeof(xhh_mcgm_init_params));
        xhh_mcgm_init_params.log2_total_entries = 
            xhh_hob_p->hca_props.multicast_parameters.log_mc_table_sz ;
        xhh_mcgm_init_params.mght_hash_bins = 1UL << (xhh_hob_p->profile.log2_mcg_hash_size);
        xhh_mcgm_init_params.max_qp_per_mcg = (u_int16_t)xhh_hob_p->profile.qps_per_mcg;
        ret = XHH_mcgm_create(xhh_hob_p,&xhh_mcgm_init_params,&(xhh_hob_p->mcgm) );
        if (ret != HH_OK) {
            MTL_ERROR1("XHH_hob_open_hca: could not create mcgm (%d)\n", ret);
            xhh_hob_p->hca_capabilities.max_mcast_grp_num = 0;            
            xhh_hob_p->hca_capabilities.max_mcast_qp_attach_num = 0;      
            xhh_hob_p->hca_capabilities.max_total_mcast_qp_attach_num = 0;
            xhh_hob_p->mcgm = (XHH_mcgm_t)XHH_INVALID_HNDL;
        }
    }

    /* CREATE EVENTP*/
    memset(event_res_p, 0, sizeof(XHH_eventp_res_t));
    event_res_p->intr_clr_bit = xhh_hob_p->adapter_props.intapin;
    event_res_p->clr_int_base_addr = xhh_hob_p->hw_props.cr_base + TAVOR_CLR_INT_H_OFFSET_FROM_CR_BASE;
    event_res_p->intr_props = &xhh_hob_p->hw_props.interrupt_props;
    event_res_p->is_srq_enable = xhh_hob_p->dev_lims.srq;
    event_res_p->kar = xhh_hob_p->kar;
    strcpy(event_res_p->name, xhh_hob_p->dev_name);

    ret = XHH_eventp_create (xhh_hob_p, event_res_p, &(xhh_hob_p->eventp));
    if (ret != HH_OK) {
        MTL_ERROR1("XHH_hob_open_hca: could not create eventp (%d)\n", ret);
        goto eventp_create_err;
    }

    /* CREATE THE VARIOUS EVENT QUEUES (eventp object operations) */
    /* register dummy completion and async event handlers */
    /* set max outstanding EQEs to max number of CQs configured */


    ret = XHH_eventp_setup_comp_eq(xhh_hob_p->eventp, 
                                   &XHH_dummy_comp_event , 
                                   NULL,
                                   (MT_size_t)(1 <<(xhh_hob_p->profile.log2_max_cqs)) - 
                                          (MT_size_t)(1ul << (xhh_hob_p->dev_lims.log2_rsvd_cqs)), 
                                   &(xhh_hob_p->compl_eq));
    if (ret != HH_OK) {
        MTL_ERROR1( "XHH_hob_open_hca: ERROR : cannot set up completion event queue (%d)\n", ret);
        goto eventp_compl_err;
    }


    ret = XHH_eventp_setup_ib_eq(xhh_hob_p->eventp, 
                                 &XHH_dummy_async_event,
                                 NULL, 
                                 (MT_size_t)(xhh_hob_p->module_flags.async_eq_size == 0 ? 
                                             XHH_MAX_ASYNC_EQ_SIZE :
                                             xhh_hob_p->module_flags.async_eq_size ), 
                                 &(xhh_hob_p->ib_eq));
    if (ret != HH_OK) {
        MTL_ERROR1(MT_FLFMT("XHH_hob_open_hca: ERROR : cannot set up async event queue for size "SIZE_T_DFMT" (ret=%d)"), 
                   (MT_size_t)(xhh_hob_p->module_flags.async_eq_size == 0 ? 
                               XHH_MAX_ASYNC_EQ_SIZE :xhh_hob_p->module_flags.async_eq_size ),ret);
        goto eventp_async_err;
    }

    /* PERFORM INIT_IB only for legacy SQPs.  Use max values obtained from QUERY_DEV_LIMS */

    last_port_initialized = 0;
    
	if (xhh_hob_p->module_flags.legacy_sqp == TRUE) { 
		for (i = 1; i <= num_ports; i++) {
			MTL_TRACE2("XHH_hob_open_hca: INIT_IB COMMAND\n");
			cmd_ret = XHH_cmd_INIT_IB(xhh_hob_p->cmd, (IB_port_t) i, &(xhh_hob_p->init_ib_props[i-1]));
			if (cmd_ret != XHH_CMD_STAT_OK) {
				MTL_ERROR1("XHH_hob_open_hca: CMD_error in XHH_cmd_INIT_IB (%d) for port %d\n", cmd_ret, i);
                if (cmd_ret ==XHH_CMD_STAT_EFATAL) {
                    ret = HH_EFATAL;
                } else {
                    ret = HH_EAGAIN;
                }
				goto init_ib_err;
			}

			else {
				MTL_TRACE2("XHH_hob_open_hca: INIT_IB COMMAND completed successfuly\n");
			}
			last_port_initialized++;
		}
	}


	/* This must be called after INIT_IB, since it uses the max_pkey value stored in that struct */
    MTL_TRACE2("XHH_hob_open_hca:  Before XHH_hob_query_struct_init\n");
    ret = XHH_hob_query_struct_init(xhh_hob_p, 
                              ((prop_props_p == NULL)? FALSE : TRUE),
                              &(xhh_hob_p->hca_capabilities));
    if (ret != HH_OK) {
        MTL_ERROR1( "XHH_hob_query_struct_init: ERROR : cannot initialize data for query_hca (%d)\n", ret);
        goto query_struct_init_err;
    }

    /* TK - start events after all CMDs are done */

    ret = XHH_eventp_setup_cmd_eq(xhh_hob_p->eventp, CMD_EQ_SIZE /* to overcome a FW bug */
                                  /*(1 << (xhh_hob_p->fw_props.log_max_outstanding_cmd))*/ );
    if (ret != HH_OK) {
        MTL_ERROR1( "XHH_hob_open_hca: ERROR : cannot set up command event queue (%d)\n", ret);
        goto eventp_cmd_err;
    }


//    printk("delay 120 sec\n");
//    MOSAL_delay_execution(120000000);
//    printk("end of delay 120 sec\n");

    /* move the HCA to running state if had no fatal. If had fatal, return HH_EFATAL */
    MOSAL_spinlock_dpc_lock(&xhh_hob_p->fatal_spl);
    if ((xhh_hob_p->xhh_state & XHH_STATE_HAVE_ANY_FATAL) != 0) {
        /* already in FATAL state */
        MTL_DEBUG4(MT_FLFMT("%s: already in FATAL state"), __func__);  
        MOSAL_spinlock_unlock(&xhh_hob_p->fatal_spl);
        ret = HH_EFATAL;
        goto fatal_err_at_end;
    }
    xhh_hob_p->xhh_state = XHH_STATE_RUNNING;
    MOSAL_spinlock_unlock(&xhh_hob_p->fatal_spl);
    
    MTL_TRACE2("XHH_hob_open_hca: device name %s\n", hca_hndl->dev_desc);
    hca_hndl->status   =   HH_HCA_STATUS_OPENED;
    
    /* free the mutex */
    MOSAL_mutex_rel(&(xhh_hob_p->mtx));
    FREE(xhh_qpm_init_params_p);
    FREE(local_hca_props_p);
    FREE(udav_internal_mr_p);
    FREE(event_res_p);
    return(HH_OK);

fatal_err_at_end:
eventp_cmd_err:
query_struct_init_err:
init_ib_err:
    /* see if need to close IB for some ports. Do not close ports on fatal error*/
    MOSAL_spinlock_dpc_lock(&xhh_hob_p->fatal_spl);
    /*test fatal again, here -- may have gotten FATAL before end of OPEN_hca process */
    if ((xhh_hob_p->xhh_state & XHH_STATE_HAVE_ANY_FATAL) != 0) {
        /* got FATAL during OPEN_HCA */
        MTL_DEBUG4(MT_FLFMT("XHH_hob_open_hca: In FATAL state")); 
        have_fatal = TRUE;
    } else {
        have_fatal = FALSE;
    }
    MOSAL_spinlock_unlock(&xhh_hob_p->fatal_spl);
    if ((last_port_initialized) && (have_fatal == FALSE)) {
        for (i = 1; i <= last_port_initialized; i++) {
            MTL_DEBUG4(MT_FLFMT("XHH_hob_open_hca: closing IB port %d"), i); 
            XHH_cmd_CLOSE_IB(xhh_hob_p->cmd, (IB_port_t) i);
        }
    }
    xhh_hob_p->ib_eq = XHH_INVALID_EQN;

eventp_async_err:
    xhh_hob_p->compl_eq = XHH_INVALID_EQN;

eventp_compl_err:
    XHH_eventp_destroy(xhh_hob_p->eventp);

eventp_create_err:
    if (xhh_hob_p->mcgm != (XHH_mcgm_t)XHH_INVALID_HNDL) {
        XHH_mcgm_destroy(xhh_hob_p->mcgm );
        xhh_hob_p->mcgm = (XHH_mcgm_t)XHH_INVALID_HNDL;
    }

    XHH_uar_destroy(xhh_hob_p->kar);
    xhh_hob_p->kar = (XHH_uar_t)XHH_INVALID_HNDL;

kar_create_err:
    MOSAL_iomem_unmap(xhh_hob_p->kar_iomem_obj, MOSAL_get_kernel_prot_ctx());
    xhh_hob_p->kar_addr = (MT_virt_addr_t) 0;

kar_map_phys_addr_err:
    if (xhh_hob_p->profile.use_priv_udav) {
        XHH_udavm_destroy(xhh_hob_p->udavm);
    }
    xhh_hob_p->udavm = (XHH_udavm_t)XHH_INVALID_HNDL;
udavm_create_err:
    if (xhh_hob_p->profile.use_priv_udav) {
        XHH_mrwm_deregister_mr(xhh_hob_p->mrwm, dummy_key);
        xhh_hob_p->udavm_lkey = 0;
    }
udavm_table_register_err:
    if (xhh_hob_p->profile.use_priv_udav) {
        MOSAL_iomem_unmap(xhh_hob_p->udavm_iomem_obj, MOSAL_get_kernel_prot_ctx());
        xhh_hob_p->udavm_table = (MT_virt_addr_t) NULL;
    }

udavm_mosal_map_err:
    if (xhh_hob_p->profile.use_priv_udav) {
        XHH_lamm_free(xhh_hob_p->lamm, udav_phys_addr, udav_table_size);
        xhh_hob_p->udavm_table_ddr  = (MT_phys_addr_t) 0;
        xhh_hob_p->udavm_table_size = 0;
    }

udavm_lamm_alloc_err:
    XHH_srqm_destroy(xhh_hob_p->srqm);
    xhh_hob_p->srqm = (XHH_srqm_t)XHH_INVALID_HNDL;

srqm_create_err:
    XHH_qpm_destroy( xhh_hob_p->qpm, TRUE);
    xhh_hob_p->qpm = (XHH_qpm_t)XHH_INVALID_HNDL;

qpm_create_err:
    FREE(xhh_qpm_init_params_p);

init_qpm_params_malloc_err:
    FREE(xhh_hob_p->init_ib_props);
    xhh_hob_p->init_ib_props = ( XHH_port_init_props_t *) NULL;

init_ib_props_malloc_err:
    XHH_cqm_destroy( xhh_hob_p->cqm, TRUE);
    xhh_hob_p->cqm = (XHH_cqm_t)XHH_INVALID_HNDL;

cqm_create_err:
    XHH_mrwm_destroy(xhh_hob_p->mrwm, TRUE);
    xhh_hob_p->mrwm = (XHH_mrwm_t)XHH_INVALID_HNDL;

mrwm_create_err:
    XHH_uldm_destroy(xhh_hob_p->uldm );
    xhh_hob_p->uldm = (XHH_uldm_t)XHH_INVALID_HNDL;
uldm_create_err:
query_hca_err:
#ifdef SIMULATE_HALT_HCA
    XHH_cmd_CLOSE_HCA(xhh_hob_p->cmd);
#else
    MOSAL_spinlock_dpc_lock(&xhh_hob_p->fatal_spl);
    /*test fatal again, here -- may have gotten FATAL before end of OPEN_hca process */
    if ((xhh_hob_p->xhh_state & XHH_STATE_HAVE_ANY_FATAL) != 0) {
        /* got FATAL during OPEN_HCA */
        MTL_DEBUG4(MT_FLFMT("XHH_hob_open_hca: In FATAL state")); 
        have_fatal = TRUE;
    } else {
        have_fatal = FALSE;
    }
    MOSAL_spinlock_unlock(&xhh_hob_p->fatal_spl);
    if (have_fatal) {
        if (xhh_hob_p->module_flags.fatal_delay_halt == 0) {
            MTL_DEBUG1(MT_FLFMT("%s: halting the HCA"), __func__);
            XHH_cmd_CLOSE_HCA(xhh_hob_p->cmd, TRUE);
        }
    } else {
        MTL_DEBUG1(MT_FLFMT("%s: closing the HCA on non-fatal error %d"), __func__, ret);
        XHH_cmd_CLOSE_HCA(xhh_hob_p->cmd, FALSE);
    }
#endif

init_hca_err:
    XHH_cmd_revoke_lamm(xhh_hob_p->cmd);

cmd_assign_lamm_err:

    ddr_alloc_area = (MT_phys_addr_t *) &(xhh_hob_p->ddr_alloc_base_addrs_vec);
    ddr_alloc_size = (MT_size_t *)&(xhh_hob_p->ddr_alloc_size_vec);
    for (i = 0; i < xhh_hob_p->profile.ddr_alloc_vec_size; i++, ddr_alloc_area++, ddr_alloc_size++) {
        if (*ddr_alloc_area != XHH_DDRMM_INVALID_PHYS_ADDR) 
          /* Do not free in case skipped during allocation (e.g., SRQC) */
          XHH_lamm_free(xhh_hob_p->lamm,*ddr_alloc_area, (1U<< (*ddr_alloc_size)));
    }

post_mutex_acquire_err:
    MOSAL_mutex_rel(&(xhh_hob_p->mtx));
    
post_state_change_error:
    MOSAL_spinlock_dpc_lock(&xhh_hob_p->fatal_spl);
    /*test fatal again, here -- may have gotten FATAL before end of OPEN_hca process */
    if ((xhh_hob_p->xhh_state & XHH_STATE_HAVE_ANY_FATAL) != 0) {
        /* got FATAL during OPEN_HCA */
        MTL_DEBUG4(MT_FLFMT("XHH_hob_open_hca: In FATAL state")); 
        have_fatal = TRUE;
    } else {
        /* restore the state to closed */
        have_fatal = FALSE;
        xhh_hob_p->xhh_state = XHH_STATE_CLOSED;
    }
    MOSAL_spinlock_unlock(&xhh_hob_p->fatal_spl);
    FREE(udav_internal_mr_p);
    FREE(event_res_p);
    FREE(local_hca_props_p);
    if (have_fatal) {
        XHH_hob_restart(hca_hndl);
    }
    return ret;
}

/******************************************************************************
 *  Function:     XHH_hob_query
 *
 *  Description:  Implements VAPI_query_hca verb.  Data is already stored in HOB object.
 *
 *  input:
 *                hca_hndl
 *  output: 
 *                hca_cap_p -- pointer to output structure
 *  returns:
 *                HH_OK
 *                HH_EINVAL_HCA_HNDL
 *
 *****************************************************************************/
 HH_ret_t XHH_hob_query(HH_hca_hndl_t  hca_hndl, 
			MOSAL_protection_ctx_t prot_ctx,
                        VAPI_hca_cap_t *hca_cap_p)
{
    XHH_hob_t             hob;

    MT_RETURN_IF_LOW_STACK(XHH_WATERMARK);
    if (MOSAL_get_exec_ctx() != MOSAL_IN_TASK) {
        MTL_ERROR1("XHH_hob_query: NOT IN TASK CONTEXT)\n");
        return HH_ERR;
    }

    if (hca_hndl == NULL) {
        MTL_ERROR1("XHH_hob_query : ERROR : Invalid HCA handle\n");
        return HH_EINVAL_HCA_HNDL;
    }
    
    hob = XHHOBP(hca_hndl);

    if (hob == NULL) {
        MTL_ERROR1("XHH_hob_query : ERROR : No device registered\n");
        return HH_EINVAL;
    }
    
    TEST_RETURN_FATAL(hob);

    /* check if ib_init_props have been created -- indicates that open_hca called for this device */
    if (hob->init_ib_props == (XHH_port_init_props_t *)NULL) {
        MTL_ERROR1("XHH_hob_query: ERROR : HCA device has not yet been opened\n");
        return HH_EINVAL;
    }

    memcpy(hca_cap_p, &(hob->hca_capabilities), sizeof(VAPI_hca_cap_t));
    return HH_OK;
}

#define SET_MAX_SG(a)  (((a) < (u_int32_t)hob->dev_lims.max_sg) ? (a) : hob->dev_lims.max_sg)
/******************************************************************************
 *  Function:     XHH_hob_query_struct_init
 *
 *  Description:  Pre-computes the data for VAPI_query_hca.  Called during XHH_hob_open_hca()
 *
 *  input:
 *                hob
 *  output: 
 *                hca_cap_p -- pointer to output structure
 *  returns:
 *                HH_OK
 *                HH_EINVAL_HCA_HNDL
 *
 *  Comments:     Needs to use a MAD query for some of the parameters
 *
 *****************************************************************************/
static  HH_ret_t XHH_hob_query_struct_init(XHH_hob_t  hob,
                               MT_bool have_usr_profile, 
                               VAPI_hca_cap_t *hca_cap_p)
{
    HH_ret_t  ret;
    u_int32_t flags = 0;
    u_int32_t log2_num_spare_segs = 0;

    MT_RETURN_IF_LOW_STACK(XHH_WATERMARK);
  MTL_DEBUG4("Entering XHH_hob_query_struct_init\n");
        /* Maximum Number of QPs supported.                   */
#if 0
  hca_cap_p->max_num_qp = hob->profile.max_num_qps - XHH_NUM_RSVD_QP -
                    ((have_usr_profile == FALSE)? 0 : (1U<<hob->dev_lims.log2_rsvd_qps) );
#else
  hca_cap_p->max_num_qp = (u_int32_t)hob->profile.max_num_qps;
#endif
  hca_cap_p->max_num_srq = (u_int32_t)hob->profile.max_num_srqs;
  hca_cap_p->srq_resize_supported= TRUE;
        /* Maximum Number of oustanding WR on any WQ.         */
  hca_cap_p->max_qp_ous_wr = (1 << (hob->dev_lims.log_max_qp_sz)) - 1;                    
  hca_cap_p->max_wqe_per_srq = (1 << (hob->dev_lims.log_max_srq_sz)) - 1;                    
        
  /* Various flags (VAPI_hca_cap_flags_t)               */

//  VAPI_RESIZE_OUS_WQE_CAP     = 1  /* Not currently supported */

  flags =  (hob->hca_props.udp ? VAPI_UD_AV_PORT_ENFORCE_CAP : 0) |
           (hob->dev_lims.apm ? VAPI_AUTO_PATH_MIG_CAP : 0 )      |
           (hob->dev_lims.rm  ? VAPI_RAW_MULTI_CAP     : 0)       |
           (hob->dev_lims.pkv ? VAPI_BAD_PKEY_COUNT_CAP : 0)      |
           (hob->dev_lims.qkv ? VAPI_BAD_QKEY_COUNT_CAP : 0)      |
           VAPI_CHANGE_PHY_PORT_CAP | VAPI_RC_RNR_NAK_GEN_CAP | VAPI_PORT_ACTIVE_EV_CAP;
           

  hca_cap_p->flags = flags;               
        /* Max num of scatter/gather entries for desc other than RD */
  hca_cap_p->max_num_sg_ent = SET_MAX_SG(28);       
        /* Max num of scatter entries for SRQs */
  hca_cap_p->max_srq_sentries = SET_MAX_SG(31);       
        /* Max num of scatter/gather entries for RD desc - not supported  */
  hca_cap_p->max_num_sg_ent_rd = 0 ;   
        /* Max num of supported CQs                           */
#if 0
  hca_cap_p->max_num_cq = hob->profile.max_num_cqs -
             ((have_usr_profile == FALSE)? 0 : (1U<<hob->dev_lims.log2_rsvd_cqs) );
#else
  hca_cap_p->max_num_cq = (u_int32_t)hob->profile.max_num_cqs;
#endif        
  /* Max num of supported entries per CQ                */
  hca_cap_p->max_num_ent_cq = (1 << (hob->dev_lims.log_max_cq_sz)) - 1 /*for extra cqe needed */;      
        /* Maximum number of memory region supported.         */
#if 0
  hca_cap_p->max_num_mr = hob->profile.num_external_mem_regions -           
                    ((have_usr_profile == FALSE)? 0 : (1U<<hob->dev_lims.log2_rsvd_mtts) );
#else
  hca_cap_p->max_num_mr = (u_int32_t)hob->profile.num_external_mem_regions;          
#endif        
        /* Largest contigous block of memory region in bytes. This may be achieved by registering
         * PHYSICAL memory directly for the region, (and using a page size for the region equal to
         * the size of the physical memory block you are registering).  The PRM allocates 5 bytes
         * for registering the log2 of the page size (with a 4K page size having page-size val 0).
         * Thus, the maximum page size per MTT entry is 4k * 2^31 (= 2^(31+12).  A single region 
         * can include multiple entries, and we can use all the spare MTT entries available for
         * this HUGE region.  The driver requires that every memory region have at least a single
         * segment available for registration.  We can thus use all the spare segments (we have
         * allocated 2 segments per region, but only really need one) to this region. If there are
         * no spare MTT entries, we calculate the value based on the usual mtt entries per segment.
         * This will still be a HUGE number (probably 2^46 or greater). 
         */
  if (hob->profile.log2_mtt_segs_per_region == 0)
      log2_num_spare_segs = (u_int32_t)hob->profile.log2_mtt_entries_per_seg;
  else 
      log2_num_spare_segs  = (u_int32_t)(hob->profile.log2_max_mtt_entries - hob->profile.log2_mtt_segs_per_region);

  /* check that do not overflow 2^64 !! */
  if (log2_num_spare_segs >= 21) 
       hca_cap_p->max_mr_size = MAKE_ULONGLONG(0xFFFFFFFFFFFFFFFF);
  else
       hca_cap_p->max_mr_size = (((u_int64_t)1L) << (31 + 12 + log2_num_spare_segs)) ;         
        /* Maximum number of protection domains supported.    */
#if 0
  hca_cap_p->max_pd_num = hob->profile.max_num_pds - XHH_NUM_RSVD_PD - 
      ((have_usr_profile == FALSE)? 0 : hob->dev_lims.num_rsvd_pds );
#else
  hca_cap_p->max_pd_num = (u_int32_t)hob->profile.max_num_pds;
#endif
  /* Largest page size supported by this HCA            */
  hca_cap_p->page_size_cap = (1 << (hob->dev_lims.log_pg_sz));       
        /* Number of physical ports of the HCA.                */
  hca_cap_p->phys_port_num = hob->dev_lims.num_ports;       
        /* Maximum number of partitions supported .           */
  hca_cap_p->max_pkeys = hob->init_ib_props[0].max_pkey;           
        /* Maximum number of oust. RDMA read/atomic as target */
  hca_cap_p->max_qp_ous_rd_atom = 1 << (hob->profile.log2_inflight_rdma_per_qp);  
        /* EE Maximum number of outs. RDMA read/atomic as target -- NOT YET SUPPORTED  */
  hca_cap_p->max_ee_ous_rd_atom = 0;  
        /* Max. Num. of resources used for RDMA read/atomic as target */
  hca_cap_p->max_res_rd_atom = ((1 << (hob->ddr_alloc_size_vec.log2_rdb_size - XHH_DDR_LOG2_RDB_ENTRY_SIZE)) > 255 ? 
                                    255 :(1 << (hob->ddr_alloc_size_vec.log2_rdb_size - XHH_DDR_LOG2_RDB_ENTRY_SIZE))) ;     
        /* Max. Num. of outs. RDMA read/atomic as initiator. Note that 255 is the max in the struct  */
  hca_cap_p->max_qp_init_rd_atom = ((1 << (hob->dev_lims.log_max_ra_req_qp)) > 255 ? 
                                    255 :  (1 << (hob->dev_lims.log_max_ra_req_qp))); 
        /* EE Max. Num. of outs. RDMA read/atomic as initiator  -- NOT YET SUPPORTED   */
  hca_cap_p->max_ee_init_rd_atom = 0;
        /* Level of Atomicity supported:  if supported, is only within this HCA*/
  hca_cap_p->atomic_cap = (hob->dev_lims.atm ? VAPI_ATOMIC_CAP_HCA : VAPI_ATOMIC_CAP_NONE);        
        /* Maximum number of EEC supported.   -- NOT YET SUPPORTED  */
#if 0
  hca_cap_p->max_ee_num = 1 << (hob->hca_props.qpc_eec_cqc_rdb_parameters.log_num_of_ee);
#else
  hca_cap_p->max_ee_num = 0;
#endif
        /* Maximum number of IB_RDD supported  -- NOT YET SUPPORTED */
  hca_cap_p-> max_rdd_num = 0;                 
        /* Maximum Number of memory windows supported  */
#if 0
  hca_cap_p->max_mw_num = hob->profile.num_mem_windows -                    
      ((have_usr_profile == FALSE)? 0 : (1U<<hob->dev_lims.log2_rsvd_mrws) );
#else
  hca_cap_p->max_mw_num = (u_int32_t)hob->profile.num_mem_windows;                   
#endif
        /* Maximum number of Raw IPV6 QPs supported  -- NOT YET SUPPORTED */ 
  hca_cap_p->max_raw_ipv6_qp = 0;              
        /* Maximum number of Raw Ethertypes QPs supported  -- NOT YET SUPPORTED */
  hca_cap_p->max_raw_ethy_qp = 0;              
        /* Maximum Number of multicast groups  */
  hca_cap_p->max_mcast_grp_num = 1 << (hob->hca_props.multicast_parameters.log_mc_table_sz);            
        /* Maximum number of QP per multicast group    */
  hca_cap_p->max_mcast_qp_attach_num = ( (1U<<(hob->hca_props.multicast_parameters.log_mc_table_entry_sz)) 
                                               -  XHH_DDR_MCG_ENTRY_HEADER_SIZE) / 
                                                             XHH_DDR_MCG_BYTES_PER_QP;
        /* Maximum number of QPs which can be attached to a mcast grp */
  hca_cap_p->max_total_mcast_qp_attach_num = hca_cap_p->max_mcast_grp_num * hca_cap_p->max_mcast_qp_attach_num;
        /* Maximum number of address handles */
  hca_cap_p->max_ah_num = (u_int32_t)(hob->profile.use_priv_udav ? hob->profile.max_priv_udavs : 
                                XHHUL_PDM_MAX_UL_UDAV_PER_PD*(hob->profile.max_num_pds+XHH_NUM_RSVD_PD));
        /* max number of fmrs for the use is the number of user entries in MPT */
#if 0
  hca_cap_p->max_num_fmr    = hob->profile.num_external_mem_regions - 
      ((have_usr_profile == FALSE)? 0 : (1U<<hob->dev_lims.log2_rsvd_mtts) );
#else
  hca_cap_p->max_num_fmr    = (u_int32_t)((hob->ddr_props.dh == FALSE) ? hob->profile.num_external_mem_regions : 0);
#endif
/* max maps per fmr is the max number that can be expressed by the MS bits of a u_int32_t
           that are unused for MPT addressing (which will occupy the LS bits of that u_int32_t).*/
  hca_cap_p->max_num_map_per_fmr = (1 << (32 - hob->profile.log2_max_mpt_entries)) - 1;
  
        /* Log2 4.096usec Max. RX to ACK or NAK delay */
  hca_cap_p->local_ca_ack_delay = hob->dev_lims.local_ca_ack_delay;

  /* Node GUID for this hca */
  ret = XHH_hob_get_node_guid(hob,&(hca_cap_p->node_guid));           

  return(ret);
}

/******************************************************************************
 *  Function:     XHH_hob_set_comp_eventh
 *
 *  Description:  Sets completion event handler for VIP layer (below vapi).  Used internally
 *                by VAPI  
 *
 *  input:
 *                hca_hndl
 *                event -  pointer to handler function
 *                private_data - pointer to context data provided to handler
 *  returns:
 *                HH_OK
 *                HH_EAGAIN
 *                HH_EINVAL_HCA_HNDL
 *                HH_ERR
 *
 *  Comments:     Initial (dummy) handler is provided during open_hca.  Therefore,
 *                the function XHH_eventp_replace_handler is used to register the handler.
 *
 *****************************************************************************/
HH_ret_t XHH_hob_set_comp_eventh(HH_hca_hndl_t      hca_hndl,
                                        HH_comp_eventh_t   event,
                                        void*              private_data)
{
  HH_ret_t  ret;
  XHH_eventp_handler_t ev_hndlr;
  
  XHH_hob_t  xhh_hob_p;

  MT_RETURN_IF_LOW_STACK(XHH_WATERMARK);
  if (MOSAL_get_exec_ctx() != MOSAL_IN_TASK) {
      MTL_ERROR1("XHH_hob_set_comp_eventh: NOT IN TASK CONTEXT)\n");
      return HH_ERR;
  }

  if (hca_hndl == NULL) {
      MTL_ERROR1("XHH_hob_set_comp_eventh : ERROR : Invalid HCA handle\n");
      return HH_EINVAL_HCA_HNDL;
  }
  xhh_hob_p = XHHOBP(hca_hndl);
  if (xhh_hob_p == NULL) {
      MTL_ERROR1("XHH_hob_set_comp_eventh: ERROR : No device registered\n");
      return HH_EAGAIN;
  }

  if (event == NULL) {
      event = XHH_dummy_comp_event;
  } else {
      TEST_RETURN_FATAL(xhh_hob_p);
  }
  ev_hndlr.comp_event_h = event;


  if (xhh_hob_p->eventp == (XHH_eventp_t)XHH_INVALID_HNDL) {
      MTL_ERROR1("XHH_hob_set_comp_eventh: ERROR : HCA device has not yet been opened\n");
      return HH_EAGAIN;
  }

  /* neutralizing for VAPI start */
  //  return HH_OK;

  ret = XHH_eventp_replace_handler(xhh_hob_p->eventp,xhh_hob_p->compl_eq, ev_hndlr, private_data);
  if (ret != HH_OK) {
      MTL_ERROR1( "XHH_hob_set_comp_eventh: ERROR : cannot register completion event handler (%d)\n", ret);
      return HH_ERR;
  }
  
  return HH_OK;
}

/******************************************************************************
 *  Function:     XHH_hob_set_async_eventh
 *
 *  Description:  Sets async handler for VIP layer (below vapi).  Used internally
 *                by VAPI  
 *
 *  input:
 *                hca_hndl
 *                event -  pointer to handler function
 *                private_data - pointer to context data provided to handler
 *  returns:
 *                HH_OK
 *                HH_EAGAIN
 *                HH_EINVAL_HCA_HNDL
 *                HH_ERR
 *
 *  Comments:     Initial (dummy) handler is provided during open_hca.  Therefore,
 *                the function XHH_eventp_replace_handler is used to register the handler.
 *
 *****************************************************************************/
HH_ret_t XHH_hob_set_async_eventh( HH_hca_hndl_t      hca_hndl,
                                          HH_async_eventh_t  event,
                                          void*              private_data)
{
  HH_ret_t  ret;
  XHH_eventp_handler_t ev_hndlr;

  XHH_hob_t  xhh_hob_p;

  MT_RETURN_IF_LOW_STACK(XHH_WATERMARK);
  if (MOSAL_get_exec_ctx() != MOSAL_IN_TASK) {
      MTL_ERROR1("XHH_hob_set_async_eventh: NOT IN TASK CONTEXT)\n");
      return HH_ERR;
  }

  if (hca_hndl == NULL) {
      MTL_ERROR1("XHH_hob_set_async_eventh : ERROR : Invalid HCA handle\n");
      return HH_EINVAL_HCA_HNDL;
  }
  xhh_hob_p = XHHOBP(hca_hndl);

  if (event == NULL) {event = &XHH_dummy_async_event;
  } else {
      TEST_RETURN_FATAL(xhh_hob_p);
  }
  
  ev_hndlr.ib_comp_event_h = event;

  if (xhh_hob_p == NULL) {
      MTL_ERROR1("XHH_hob_set_async_eventh: ERROR : No device registered\n");
      return HH_EAGAIN;
  }

  if (xhh_hob_p->eventp == (XHH_eventp_t)XHH_INVALID_HNDL) {
      MTL_ERROR1("XHH_hob_set_async_eventh: ERROR : HCA device has not yet been opened\n");
      return HH_EAGAIN;
  }
// neutralizing for VAPI start
// return HH_OK;
  
  ret = XHH_eventp_replace_handler(xhh_hob_p->eventp,xhh_hob_p->ib_eq, ev_hndlr, private_data);
  if (ret != HH_OK) {
      MTL_ERROR1( "XHH_hob_set_async_eventh: ERROR : cannot register async event handler (%d)\n", ret);
      return HH_ERR;
  }

  /* track async event handler setting for use in fatal error handling */
  MOSAL_spinlock_irq_lock(&xhh_hob_p->async_spl);
  xhh_hob_p->async_eventh = event;
  xhh_hob_p->async_ev_private_context = private_data;
  MOSAL_spinlock_unlock(&xhh_hob_p->async_spl);

  return HH_OK;
}

/******************************************************************************
 *  Function:     XHH_hob_create_internal
 *
 *  Description:  Creates the HOB object for an HCA, and registers it in HH  
 *
 *  input:
 *                hw_props_p -- PCI properties (BARs, etc)
 *                hca_seq_num  - a sequence number assigned to this HCA to differentiate it 
 *                               from other HCAs on this host
 *                mod_flags   - flags passed in at module initialization (e.g., insmod)
 *                curr_hh_hndl  - current HH handle - valid for fatal error restart. NULL otherwise
 *                                (NULL value indicates new creation)
 *  output: 
 *                hh_hndl_p  - size of returned table (in pkeys)
 *  returns:
 *                HH_OK
 *                HH_EAGAIN
 *                HH_ERR -- other errors
 *
 *  Comments:     This function involves the following steps: 
 *                    1.Allocate XHH_hob data context. 
 *                    2.Create the XHH_cmd_if object instance (in order to enable 
 *                       queries of HCA resources even before HCA is opened). 
 *                    3.Invoke  ENABLE_SYS command ((polling mode). 
 *                    4.Query HCA for available DDRmemory resources 
 *                      (use the Command interface in polling mode)and create 
 *                       the XHH_lamm object based on results. 
 *                    5.Query HCA for other capabilties and save them in XHH_hob context. 
 *                    6.Register HCA in HH (i.e.call HH_add_hca_dev()).
 *
 *                Also initializes the HOB mutex for controlling Open HCA and Close HCA 
 *
 *****************************************************************************/
HH_ret_t    XHH_hob_create_internal(/*IN*/  XHH_hw_props_t   *hw_props_p,
                           /*IN*/  u_int32_t         hca_seq_num,
                           /*IN*/  XHH_module_flags_t *mod_flags,
                           /*IN*/  HH_hca_hndl_t      curr_hh_hndl,
                           /*OUT*/ HH_hca_hndl_t    *hh_hndl_p )
{

  //HH_hca_hndl_t  hca_hndl = 0;
  HH_ret_t        ret;
  HH_if_ops_t     *if_ops_p = 0;
  XHH_hob_t       hob_p;
  MT_size_t       ddr_size;
  MT_size_t       fw_size;
  HH_hca_hndl_t   new_hh_hndl;
  XHH_cmd_status_t  cmd_ret;
  HH_hca_dev_t    *tdev_p = NULL;
  u_int64_t       fw_version;
  u_int32_t       req_fw_maj_version=0;
  u_int16_t       req_fw_min_version = 0;
  u_int16_t       req_fw_submin_version = 0;
  int             int_ret = 0;
  call_result_t   mosal_ret = MT_OK;
  MT_phys_addr_t  cmdif_uar0_arg;

  MT_RETURN_IF_LOW_STACK(XHH_WATERMARK);
  MTL_DEBUG4("Entering XHH_hob_create\nhca_seq_num = %d, legacy_flag = %s, av_in_host_mem = %s\n",  
             hca_seq_num, (mod_flags->legacy_sqp == FALSE ? "FALSE" : "TRUE"),
              (mod_flags->av_in_host_mem == FALSE ? "FALSE" : "TRUE"));
  XHH_print_hw_props(hw_props_p);

  if (MOSAL_get_exec_ctx() != MOSAL_IN_TASK) {
      MTL_ERROR1("XHH_hob_create: NOT IN TASK CONTEXT)\n");
      return HH_ERR;
  }

  XHH_set_capabilities(hw_props_p, mod_flags);

  /* Allocate XHH hob structure  */                                     
  hob_p = (XHH_hob_t)MALLOC(sizeof(struct XHH_hob_st));

  if (hob_p == 0) {
    MT_COV_MEM_FLTR_START
    MTL_ERROR1("XHH_hob_create: could not allocate memory for XHH hob\n");
    return HH_EAGAIN;
    MT_COV_MEM_FLTR_END
  }

  MTL_DEBUG1("XHH_hob_create: HOB address = 0x%p\n", hob_p);
  memset(hob_p, 0, sizeof(struct XHH_hob_st));

  tdev_p = (HH_hca_dev_t *)MALLOC(sizeof(HH_hca_dev_t));
  if (tdev_p == NULL) {
     MT_COV_MEM_FLTR_START
     FREE(hob_p);
     return HH_EAGAIN;
     MT_COV_MEM_FLTR_END
  }
  memset(tdev_p, 0, sizeof(HH_hca_dev_t));

    
  /* initialize the HOB mutex */
  MOSAL_mutex_init(&(hob_p->mtx));

  /* set device name */
  sprintf(hob_p->dev_name, "InfiniHost%d",hca_seq_num); 
  printk("\nMellanox Tavor Device Driver is creating device \"%s\" (bus=%02x, devfn=%02x)\n\n", 
         hob_p->dev_name,hw_props_p->bus,hw_props_p->dev_func);

  /* set embedded object handles to invalid */
  
  hob_p->cmd  = (XHH_cmd_t)XHH_INVALID_HNDL;
  hob_p->lamm = (XHH_lamm_t)XHH_INVALID_HNDL;
  hob_p->uldm = (XHH_uldm_t)XHH_INVALID_HNDL;
  hob_p->mrwm = (XHH_mrwm_t)XHH_INVALID_HNDL;
  hob_p->cqm = (XHH_cqm_t)XHH_INVALID_HNDL;
//hob_p->eecm = (XHH_eecm_t)XHH_INVALID_HNDL;  /* JPM -- EECM ADDITIONS HERE */
  hob_p->qpm = (XHH_qpm_t)XHH_INVALID_HNDL;
  hob_p->udavm = (XHH_udavm_t)XHH_INVALID_HNDL;
  hob_p->mcgm = (XHH_mcgm_t)XHH_INVALID_HNDL;
  hob_p->eventp = (XHH_eventp_t)XHH_INVALID_HNDL;
  hob_p->kar = (XHH_uar_t)XHH_INVALID_HNDL;
  hob_p->srqm = (XHH_srqm_t)XHH_INVALID_HNDL;
  
  /* initialize EQ handles to all EQs invalid */
  hob_p->compl_eq = XHH_INVALID_EQN;
  hob_p->ib_eq    = XHH_INVALID_EQN;

  /* initialize fatal error handling fields */
  memcpy(&(hob_p->hw_props), hw_props_p, sizeof(XHH_hw_props_t));
  memcpy(&(hob_p->module_flags), mod_flags, sizeof(XHH_module_flags_t));
  hob_p->hca_seq_num = hca_seq_num;
  hob_p->xhh_state = XHH_STATE_CREATING;
  if (MOSAL_spinlock_init(&(hob_p->fatal_spl)) != MT_OK){
    MTL_ERROR4(MT_FLFMT("%s: Failed to initializing fatal error spinlock"), __func__);  
    ret= HH_ERR;
    goto err_free_hob;
  }
  if (MOSAL_spinlock_init(&(hob_p->async_spl)) != MT_OK){
    MTL_ERROR4(MT_FLFMT("%s: Failed to initializing async handler tracking spinlock"), __func__);  
    ret= HH_ERR;
    goto err_free_hob;
  }
  /* dummy async event handler is passed to XHH_eventp when initializing ib_eq */
  hob_p->async_eventh = &XHH_dummy_async_event;
  hob_p->async_ev_private_context = NULL;

  /* get bridge config info */
  ret = XHH_hob_get_pci_br_config(hob_p);

  /* get hca config info */
  ret = XHH_read_pci_config(hob_p->hw_props.pci_dev, &hob_p->pci_hca_info.config);
  if ( ret != HH_OK ) {
    ret= HH_ERR;
    goto err_free_hob;
  }
  hob_p->pci_hca_info.bus = hw_props_p->bus;
  hob_p->pci_hca_info.dev_func = hw_props_p->dev_func;
  hob_p->pci_hca_info.is_valid = TRUE;

  /* reset the device. The "FATAL_HANDLING" macros here are mis-named, and should
   * be just "MT_START_DEVICE_RESET/END_DEVICE_RESET"
   */
  {
       MT_START_FATAL_HANDLING;
	/* Do the Tavor RESET */
       XHH_hob_reset_device(&(hob_p->hw_props));

	/* now, rewrite the PCI configuration */
	if (hob_p->pci_bridge_info.is_valid == TRUE) {
	   XHH_write_pci_config(hob_p->pci_bridge_info.pci_dev,
			     &hob_p->pci_bridge_info.config);
	}
	if (hob_p->pci_hca_info.is_valid == TRUE) {
	    XHH_write_pci_config(hob_p->hw_props.pci_dev,
			     &hob_p->pci_hca_info.config);
	}
	
	///XHH_set_capabilities(hw_props_p);
	
	MT_END_FATAL_HANDLING;
  }
  
  /* create the XHH_cmd object so that can initialize and query the adapter */
  /* HCR register offset */
  if ( mod_flags->cmdif_post_uar0) {
    cmdif_uar0_arg = hw_props_p->uar_base;
  }
  else {
    cmdif_uar0_arg = (MT_phys_addr_t) MAKE_ULONGLONG(0xFFFFFFFFFFFFFFFF);
  }
  ret = XHH_cmd_create(hob_p, hw_props_p->hw_ver, hw_props_p->cr_base, cmdif_uar0_arg, &(hob_p->cmd),
                       mod_flags->inifinite_cmd_timeout, mod_flags->num_cmds_outs);
  if (ret != HH_OK) {
    MTL_ERROR1("XHH_hob_create: could not create CMD object (%d)\n", ret);
    ret = HH_ERR;
    goto err_free_hob;
  }


  /* invoke SYS_ENA command on tavor to initialize it -- load firmware from flash, etc.*/
  cmd_ret = XHH_cmd_SYS_EN(hob_p->cmd);
  if (cmd_ret != XHH_CMD_STAT_OK) {
    if (cmd_ret == XHH_CMD_STAT_EFATAL) {
        MTL_ERROR1(MT_FLFMT("XHH_hob_create: FATAL ERROR in XHH_cmd_SYS_EN"));
        ret = HH_EFATAL;
    } else {
        MTL_ERROR1(MT_FLFMT("XHH_hob_create: CMD_error in XHH_cmd_SYS_EN (%d)"), cmd_ret);
        ret = HH_ERR;
    }
    goto cmd_err;
  }


  /* do query firmware command */
  cmd_ret = XHH_cmd_QUERY_FW(hob_p->cmd, &(hob_p->fw_props));
  if (cmd_ret != XHH_CMD_STAT_OK) {
    if (cmd_ret == XHH_CMD_STAT_EFATAL) {
        MTL_ERROR1(MT_FLFMT("XHH_hob_create: FATAL ERROR in XHH_cmd_QUERY_FW"));
        ret = HH_EFATAL;
    } else {
        MTL_ERROR1(MT_FLFMT("XHH_hob_create: CMD_error in XHH_cmd_QUERY_FW (%d)"), cmd_ret);
        ret = HH_ERR;
    }
    goto undo_sys_ena;
  }
  fw_version = hob_p->fw_props.fw_rev_major;
  fw_version = (fw_version <<16) | hob_p->fw_props.fw_rev_minor;
  fw_version = (fw_version <<16) | hob_p->fw_props.fw_rev_subminor;
  /* enter data into version info structure */
  hob_p->version_info.fw_ver_major = hob_p->fw_props.fw_rev_major;
  hob_p->version_info.fw_ver_minor = hob_p->fw_props.fw_rev_minor;
  hob_p->version_info.fw_ver_subminor = hob_p->fw_props.fw_rev_subminor;
  hob_p->version_info.hw_ver       = hob_p->hw_props.hw_ver;
  hob_p->version_info.cmd_if_ver   = hob_p->fw_props.cmd_interface_rev;


  if (fw_version < XHH_MIN_FW_VERSION) {
      req_fw_maj_version = (u_int32_t)  ((((u_int64_t)XHH_MIN_FW_VERSION)>>32) & MAKE_ULONGLONG(0xFFFFFFFF));
      req_fw_min_version = (u_int16_t)  ((((u_int64_t)XHH_MIN_FW_VERSION)>>16) & MAKE_ULONGLONG(0xFFFF));
      req_fw_submin_version = (u_int16_t)  (((u_int64_t)XHH_MIN_FW_VERSION) & MAKE_ULONGLONG(0xFFFF));
      MTL_ERROR1("XHH_hob_create: INSTALLED FIRMWARE VERSION IS NOT SUPPORTED:\n                     Installed: %x.%x.%x, Minimum Required: %x.%x.%x\n\n",
                 hob_p->fw_props.fw_rev_major, hob_p->fw_props.fw_rev_minor, hob_p->fw_props.fw_rev_subminor,
                 req_fw_maj_version, req_fw_min_version, req_fw_submin_version);
      ret = HH_ERR;
      goto undo_sys_ena;
  }

  MTL_DEBUG1(MT_FLFMT("%s: fw_version="U64_FMT", errbuf_start="U64_FMT", sz=0x%x"),__func__, fw_version, hob_p->fw_props.error_buf_start, 
				   hob_p->fw_props.error_buf_size);
  /* map the firmware error buffer if the appropriate fw version is installed */
  if ((fw_version >= XHH_MIN_FW_ERRBUF_VERSION) &&
      (hob_p->fw_props.error_buf_start != (u_int64_t) 0) &&
      (hob_p->fw_props.error_buf_size != 0)) 
  {

    /* wa for FW bug number 19695 */
    if ( (hob_p->fw_props.error_buf_start<hw_props_p->cr_base) || (hob_p->fw_props.error_buf_start>(hw_props_p->cr_base+0x100000)) ) {
      MTL_ERROR1(MT_FLFMT("%s: fw_props.error_buf_start is outside of cr-space start="U64_FMT", size=0x%x"),
                 __func__, hob_p->fw_props.error_buf_start, hob_p->fw_props.error_buf_size);
      ret = HH_ERR;
      goto undo_sys_ena;
    }
  
    MTL_DEBUG4(MT_FLFMT("XHH_hob_create: using cat err buf. pa=0x"U64_FMT", sz=%d"),
                 hob_p->fw_props.error_buf_start, hob_p->fw_props.error_buf_size);
    mosal_ret = MOSAL_iomem_map(MOSAL_get_kernel_prot_ctx(),
					    MOSAL_MEM_FLAGS_NO_CACHE | MOSAL_MEM_FLAGS_PERM_READ,
					    (MT_phys_addr_t) hob_p->fw_props.error_buf_start, 
                                            (MT_size_t) 4*(hob_p->fw_props.error_buf_size),
				             &(hob_p->fw_error_iomem_obj),
				             &(hob_p->fw_error_buf_start_va));
    if (mosal_ret != MT_OK) {
         MTL_ERROR1(MT_FLFMT("%s: Could not map fw error buffer (phys addr = "U64_FMT", size=%d) (%d:%s)"),
                        __func__, hob_p->fw_props.error_buf_start, hob_p->fw_props.error_buf_size,mosal_ret,mtl_strerror_sym(mosal_ret));
         hob_p->fw_error_iomem_obj = (MOSAL_iomem_t) NULL;
         hob_p->fw_error_buf_start_va =  (MT_virt_addr_t)(MT_ulong_ptr_t) NULL;
    } else {
         hob_p->fw_error_buf = TNMALLOC(u_int32_t,hob_p->fw_props.error_buf_size);
         if (hob_p->fw_error_buf == NULL) {
              MTL_ERROR1(MT_FLFMT("%s: Could not allocate buffer for FW catastrophic error info"),__func__);
         }
    }
  }

  /* Get device limits */
  cmd_ret = XHH_cmd_QUERY_DEV_LIM(hob_p->cmd, &(hob_p->dev_lims));
  if (cmd_ret != XHH_CMD_STAT_OK) {
    if (cmd_ret == XHH_CMD_STAT_EFATAL) {
        MTL_ERROR1(MT_FLFMT("XHH_hob_create: FATAL ERROR in XHH_cmd_QUERY_DEV_LIM"));
        ret = HH_EFATAL;
    } else {
        MTL_ERROR1(MT_FLFMT("XHH_hob_create: CMD_error in XHH_cmd_QUERY_DEV_LIM (%d)"), cmd_ret);
        ret = HH_ERR;
    }
    goto undo_sys_ena;
  }

  MTL_DEBUG1(MT_FLFMT("%s: log_max_srq=%u  log2_rsvd_srqs=%u  srq_entry_sz=%u  srq=%ssupported"), __func__,
             hob_p->dev_lims.log_max_srqs, hob_p->dev_lims.log2_rsvd_srqs, 
             hob_p->dev_lims.srq_entry_sz, hob_p->dev_lims.srq ? " ":"NOT-");

  /* Enable SRQ only for FW version 3.1 and up */
  if ((hob_p->dev_lims.srq) && 
      (fw_version < XHH_MIN_FW_VERSION_SRQ)) {
      MTL_ERROR1("%s: Disabling SRQ support due to FW version: "
	         "Installed: %x.%x.%x, Minimum Required: 3.1.0\n", __func__,
                 hob_p->fw_props.fw_rev_major, hob_p->fw_props.fw_rev_minor, hob_p->fw_props.fw_rev_subminor);
      hob_p->dev_lims.srq= FALSE;
  }

#ifdef ADJUST_DEVLIMS_MAX_QP_PER_MCG
     hob_p->dev_lims.log_max_qp_mcg = (hob_p->dev_lims.log_max_qp_mcg > 3) ? 
	 3 : hob_p->dev_lims.log_max_qp_mcg;
#endif

  /* query tavor for DDR memory resources data */
  cmd_ret = XHH_cmd_QUERY_DDR(hob_p->cmd, &(hob_p->ddr_props));
  if (cmd_ret != XHH_CMD_STAT_OK) {
    if (cmd_ret == XHH_CMD_STAT_EFATAL) {
        MTL_ERROR1(MT_FLFMT("XHH_hob_create: FATAL ERROR in XHH_cmd_QUERY_DDR"));
        ret = HH_EFATAL;
    } else {
        MTL_ERROR1(MT_FLFMT("XHH_hob_create: CMD_error in XHH_cmd_QUERY_DDR (%d)"), cmd_ret);
        ret = HH_ERR;
    }
    goto undo_sys_ena;
  }

  /* HIDE-DDR set in firmware:  sanity checks */
#if 0
  /* 1. fail if using 32-bit platform (not PAE and not IA64) */
  if ((hob_p->ddr_props.dh == TRUE) && (sizeof(MT_phys_addr_t) <=4)) {
      MTL_ERROR1("XHH_hob_create: HIDE_DDR is not supported on platforms using 32-bit physical addresses\n\n");
      ret = HH_ERR;
      goto undo_sys_ena;
  }
#endif

  /* 2. Fail if firmware version is not recent enough. */
  if ((hob_p->ddr_props.dh == TRUE) && (fw_version < XHH_MIN_FW_HIDE_DDR_VERSION)) {
      req_fw_maj_version = (u_int32_t)  ((((u_int64_t)XHH_MIN_FW_HIDE_DDR_VERSION)>>32) & 0xFFFFFFFF);
      req_fw_min_version = (u_int16_t)  ((((u_int64_t)XHH_MIN_FW_HIDE_DDR_VERSION)>>16) & 0xFFFF);
      req_fw_submin_version = (u_int16_t)  (((u_int64_t)XHH_MIN_FW_HIDE_DDR_VERSION) & 0xFFFF);
      MTL_ERROR1("XHH_hob_create: INSTALLED FIRMWARE VERSION DOES NOT SUPPORT HIDE_DDR:\n                     Installed: %x.%x.%x, Minimum Required: %x.%x.%x\n\n",
                 hob_p->fw_props.fw_rev_major, hob_p->fw_props.fw_rev_minor, hob_p->fw_props.fw_rev_subminor,
                 req_fw_maj_version, req_fw_min_version, req_fw_submin_version);
      ret = HH_ERR;
      goto undo_sys_ena;
  }
  
#ifdef DEBUG_MEM_OV
cmdif_dbg_ddr = hob_p->ddr_props.ddr_start_adr; /* address in ddr used for out params in debug mode */
#endif

  /* print info messages that device is operating in HIDE DDR mode (not an error) */
  if (hob_p->ddr_props.dh == TRUE) {
      MTL_ERROR1("Device %s is operating in HIDE_DDR mode.\n", hob_p->dev_name);
  }
  

/* query tavor for adapter data */
  cmd_ret = XHH_cmd_QUERY_ADAPTER(hob_p->cmd, &(hob_p->adapter_props));
  if (cmd_ret != XHH_CMD_STAT_OK) {
    if (cmd_ret == XHH_CMD_STAT_EFATAL) {
        MTL_ERROR1(MT_FLFMT("XHH_hob_create: FATAL ERROR in XHH_cmd_QUERY_ADAPTER"));
        ret = HH_EFATAL;
    } else {
        MTL_ERROR1(MT_FLFMT("XHH_hob_create: CMD_error in XHH_cmd_QUERY_ADAPTER (%d)"), cmd_ret);
        ret = HH_ERR;
    }
    goto undo_sys_ena;
  }
  if ( (hob_p->fw_props.fw_end_addr <= hob_p->fw_props.fw_base_addr)  ||
        hob_p->fw_props.fw_base_addr < hob_p->ddr_props.ddr_start_adr ||
        hob_p->fw_props.fw_end_addr > hob_p->ddr_props.ddr_end_adr) {
      /* FW region is either improper, or does not lie within bounds of DDR */
      MTL_ERROR1("XHH_hob_create: FW region is either improper, or is outside DDR\nFW end  = "U64_FMT
                 ", FW start = "U64_FMT"\n   DDR end = "U64_FMT", DDR start = "U64_FMT"\n", 
                 hob_p->fw_props.fw_end_addr, hob_p->fw_props.fw_base_addr, 
                 hob_p->ddr_props.ddr_end_adr, hob_p->ddr_props.ddr_start_adr);
      ret = HH_ERR;
      goto undo_sys_ena;

  }
  fw_size = (MT_size_t) (hob_p->fw_props.fw_end_addr - hob_p->fw_props.fw_base_addr + 1);

  if (hob_p->ddr_props.ddr_end_adr < hob_p->ddr_props.ddr_start_adr) {
      MTL_ERROR1("XHH_hob_create: DDR end address ("U64_FMT") is less than DDR base addr (" U64_FMT ")\n", 
                 hob_p->ddr_props.ddr_end_adr, hob_p->ddr_props.ddr_start_adr);
      ret = HH_ERR;
      goto undo_sys_ena;
  }

  ddr_size = (MT_size_t) (hob_p->ddr_props.ddr_end_adr - hob_p->ddr_props.ddr_start_adr + 1) ;
  hob_p->profile.ddr_size = ddr_size - fw_size;

  /* DDR size code is used in XHH_calculate_profile to set number of QPs proportionally to size */
  hob_p->profile.ddr_size_code = XHH_get_ddr_size_code(ddr_size - fw_size);

  ret = XHH_lamm_create((MT_phys_addr_t) (hob_p->ddr_props.ddr_start_adr), ddr_size, &(hob_p->lamm));
  if (ret != HH_OK) {
    MTL_ERROR1("XHH_hob_create: could not create DDRMM object (%d)\n", ret);
    goto undo_sys_ena;
  }
  ret = XHH_lamm_reserve(hob_p->lamm, hob_p->fw_props.fw_base_addr, fw_size);
  if (ret != HH_OK) {
    MTL_ERROR1("XHH_hob_create: could not reserve FW space in DDRMM object (err = %d)\n", ret);
    goto undo_ddrm_create;
  }
  
#ifdef DEBUG_MEM_OV
  ret = XHH_lamm_reserve(hob_p->lamm, hob_p->ddr_props.ddr_start_adr, CMDIF_SIZE_IN_DDR);
  if (ret != HH_OK) {
    MTL_ERROR1("XHH_hob_create: could not reserve DDR Outbox space in DDRMM object (err = %d)\n", ret);
    goto undo_ddrm_create;
  }
#endif

  /*create the delay unlock object for catastrophic error use */
  int_ret = VIP_delay_unlock_create(&hob_p->delay_unlocks);
  if (int_ret != 0) {
      MTL_ERROR1("XHH_hob_create: could create delay unlock object (err = %d)\n", int_ret);
      ret = HH_ENOMEM;
      goto delay_unlock_err;
  }
  

  /* launch catastrophic error thread */
  hob_p->fatal_thread_obj.hob = (struct XHH_hob_st *)hob_p;
  MOSAL_syncobj_init(&hob_p->fatal_thread_obj.start_sync);
  MOSAL_syncobj_init(&hob_p->fatal_thread_obj.stop_sync);
  MOSAL_syncobj_init(&hob_p->fatal_thread_obj.fatal_err_sync);
  MOSAL_syncobj_init(&hob_p->xhh_fatal_complete_syncobj);
  hob_p->fatal_thread_obj.have_fatal = FALSE;
  mosal_ret = MOSAL_thread_start(&hob_p->fatal_thread_obj.mto, MOSAL_KTHREAD_CLONE_FLAGS,
                                 XHH_hob_fatal_err_thread, (void *)(&(hob_p->fatal_thread_obj)));
//  if (mosal_ret != MT_OK) {
//      MTL_ERROR1("XHH_hob_create: could not create fatal error thread (err = %d)\n", mosal_ret);
//      ret = HH_ERR;
//      goto fatal_thr_create_err;
//  }
  
  /*wait for fatal thread initialization complete */
  mosal_ret = MOSAL_syncobj_waiton(&(hob_p->fatal_thread_obj.start_sync), 10000000);
  if (mosal_ret != MT_OK) {
      if (mosal_ret == MT_EINTR) {
          MTL_DEBUG1(MT_FLFMT("%s: Received OS interrupt while initializing fatal error thread (err = %d)"), 
                     __func__,mosal_ret);
          ret = HH_EINTR;
      } else {
          MTL_ERROR1(MT_FLFMT("%s: Timeout on initializing fatal error thread (err = %d)"), 
                     __func__,mosal_ret);
          ret = HH_ERR;
      }
      goto fatal_thr_init_err;
  }
  MTL_DEBUG4("%s: Created send completion thread.\n", __func__);
  /* set up the procedure mapping table and register the tavor device */

  if_ops_p = &(hob_p->if_ops);

  HH_ifops_tbl_set_enosys(if_ops_p); /* by default, all retuen HH_ENOSYS */

  /* HCA Calls */
  if_ops_p->HHIF_open_hca           = &XHH_hob_open_hca;
  if_ops_p->HHIF_close_hca          = &XHH_hob_close_hca;
  if_ops_p->HHIF_alloc_ul_resources = &XHH_hob_alloc_ul_res;
  if_ops_p->HHIF_free_ul_resources  = &XHH_hob_free_ul_res;
  if_ops_p->HHIF_query_hca          = &XHH_hob_query;
  if_ops_p->HHIF_modify_hca         = &XHH_hob_modify;

  /* Misc HCA Operations*/
  if_ops_p->HHIF_query_port_prop        = &XHH_hob_query_port_prop;
  if_ops_p->HHIF_get_pkey_tbl           = &XHH_hob_get_pkey_tbl;
  if_ops_p->HHIF_get_gid_tbl            = &XHH_hob_get_gid_tbl;

  /* Protection Domain Calls */
  if_ops_p->HHIF_alloc_pd               = &XHH_hob_alloc_pd;
  if_ops_p->HHIF_free_pd                = &XHH_hob_free_pd;
  if_ops_p->HHIF_alloc_rdd              = &XHH_hob_alloc_rdd;
  if_ops_p->HHIF_free_rdd               = &XHH_hob_free_rdd;

  /* privileged UD AV */
  if_ops_p->HHIF_create_priv_ud_av      = &XHH_hob_create_ud_av;
  if_ops_p->HHIF_modify_priv_ud_av      = &XHH_hob_modify_ud_av;
  if_ops_p->HHIF_query_priv_ud_av       = &XHH_hob_query_ud_av;
  if_ops_p->HHIF_destroy_priv_ud_av     = &XHH_hob_destroy_ud_av;

  /* Memory Registration */
  if_ops_p->HHIF_register_mr     = &XHH_hob_register_mr;
  if_ops_p->HHIF_reregister_mr   = &XHH_hob_reregister_mr;
  if_ops_p->HHIF_register_smr    = &XHH_hob_register_smr;
  if_ops_p->HHIF_query_mr        = &XHH_hob_query_mr;
  if_ops_p->HHIF_deregister_mr   = &XHH_hob_deregister_mr;

  if_ops_p->HHIF_alloc_mw        = &XHH_hob_alloc_mw;
  if_ops_p->HHIF_query_mw        = &XHH_hob_query_mw;
  if_ops_p->HHIF_free_mw         = &XHH_hob_free_mw;

  /* Fast memory regions */
  if_ops_p->HHIF_alloc_fmr       = &XHH_hob_alloc_fmr;
  if_ops_p->HHIF_map_fmr         = &XHH_hob_map_fmr;
  if_ops_p->HHIF_unmap_fmr       = &XHH_hob_unmap_fmr;
  if_ops_p->HHIF_free_fmr        = &XHH_hob_free_fmr;

  /* Completion Queues */
  if_ops_p->HHIF_create_cq       = &XHH_hob_create_cq;
  if_ops_p->HHIF_resize_cq       = &XHH_hob_resize_cq;
  if_ops_p->HHIF_query_cq        = &XHH_hob_query_cq;
  if_ops_p->HHIF_destroy_cq      = &XHH_hob_destroy_cq;

  /* Queue Pair */
  if_ops_p->HHIF_create_qp       = &XHH_hob_create_qp;
  if_ops_p->HHIF_get_special_qp  = &XHH_hob_get_special_qp;
  if_ops_p->HHIF_modify_qp       = &XHH_hob_modify_qp;
  if_ops_p->HHIF_query_qp        = &XHH_hob_query_qp;
  if_ops_p->HHIF_destroy_qp      = &XHH_hob_destroy_qp;
#if defined(MT_SUSPEND_QP)
  if_ops_p->HHIF_suspend_qp      = &XHH_hob_suspend_qp;
  if_ops_p->HHIF_suspend_cq      = &XHH_hob_suspend_cq;
#endif
  /* SRQ */
  if_ops_p->HHIF_create_srq      = &XHH_hob_create_srq;
  if_ops_p->HHIF_query_srq       = &XHH_hob_query_srq;
  if_ops_p->HHIF_modify_srq      = &XHH_hob_modify_srq;
  if_ops_p->HHIF_destroy_srq     = &XHH_hob_destroy_srq;

  /* EEC */
  if_ops_p->HHIF_create_eec      = &XHH_hob_create_eec;
  if_ops_p->HHIF_modify_eec      = &XHH_hob_modify_eec;
  if_ops_p->HHIF_query_eec       = &XHH_hob_query_eec;
  if_ops_p->HHIF_destroy_eec     = &XHH_hob_destroy_eec;

  if_ops_p->HHIF_set_comp_eventh  = &XHH_hob_set_comp_eventh;
  if_ops_p->HHIF_set_async_eventh = &XHH_hob_set_async_eventh;
  
  
  
  /* Multicast groups */
  if_ops_p->HHIF_attach_to_multicast   = &XHH_hob_attach_to_multicast;
  if_ops_p->HHIF_detach_from_multicast = &XHH_hob_detach_from_multicast;
  
  /* Process local MAD */
  if_ops_p->HHIF_process_local_mad = &XHH_hob_process_local_mad;
  
  if_ops_p->HHIF_ddrmm_alloc = &XHH_hob_lamm_alloc;
  if_ops_p->HHIF_ddrmm_query = &XHH_hob_lamm_query;
  if_ops_p->HHIF_ddrmm_free = &XHH_hob_lamm_free;

  /*
   *  Register device in the init structure
   *
   */
  tdev_p->ul_if_ops_p = NULL;
  strcpy(tdev_p->dev_desc, hob_p->dev_name);
  tdev_p->ul_init_info.kernel_ul_hob_create_p  = &XHHUL_hob_create;
  strcpy(tdev_p->ul_init_info.user_devlib_info.devlib_filename, MT_THH_DLL_NAME);
  strcpy(tdev_p->ul_init_info.user_devlib_info.devlib_hob_create, "XHHUL_hob_create");
  tdev_p->vendor_id = MT_MELLANOX_IEEE_VENDOR_ID;
  tdev_p->dev_id    = (u_int32_t)hw_props_p->device_id;
  MTL_DEBUG1("hw_props_p: device_id = 0x%X, pci_vendor_id=0x%X,hw_ver=0x%X\n",
              hw_props_p->device_id, hw_props_p->pci_vendor_id, hw_props_p->hw_ver);
  tdev_p->hw_ver    = hob_p->hw_props.hw_ver; 
  tdev_p->if_ops    = if_ops_p;
  tdev_p->hca_ul_resources_sz = sizeof(XHH_hca_ul_resources_t);
  tdev_p->pd_ul_resources_sz = sizeof(XHH_pd_ul_resources_t);
  tdev_p->cq_ul_resources_sz = sizeof(XHH_cq_ul_resources_t);
  tdev_p->srq_ul_resources_sz = sizeof(XHH_srq_ul_resources_t);
  tdev_p->qp_ul_resources_sz = sizeof(XHH_qp_ul_resources_t);
  tdev_p->device = (void *) hob_p;
  tdev_p->status = HH_HCA_STATUS_CLOSED;

  /* Grab the mutex now, just before adding the device to HH */
  MTL_DEBUG4("XHH_hob_create:  about to grab mutex\n");
  if (MOSAL_mutex_acq(&(hob_p->mtx), TRUE) != MT_OK) {
      MTL_DEBUG1(MT_FLFMT("%s: Received signal. returning HH_EINTR"), __func__);
      ret = HH_EINTR;
      goto err_acq_mutex;
  }

  if (curr_hh_hndl == NULL) {
      ret = HH_add_hca_dev_prep(tdev_p, &new_hh_hndl);
      if (ret != HH_OK) {
	  MTL_ERROR1("XHH_hob_create: could not register device %s in HCA HAL\n",
                            hob_p->dev_name);
	  goto err_release_mutex;
      }
  } else {
      new_hh_hndl = curr_hh_hndl;
      new_hh_hndl->device = (void *) hob_p;
      new_hh_hndl->if_ops = if_ops_p;
      new_hh_hndl->hw_ver = hob_p->hw_props.hw_ver;
      new_hh_hndl->status = HH_HCA_STATUS_CLOSED;
  }

  /* insert HH hca handle into device structure */
  ((XHH_hob_t)(new_hh_hndl->device))->hh_hca_hndl =  new_hh_hndl;
  MTL_DEBUG1("XHH_hob_create: hh_hca_hndl created = %p\n", (void *) new_hh_hndl);

  if (hh_hndl_p != NULL) {
      *hh_hndl_p =  new_hh_hndl;
  }
  MOSAL_spinlock_dpc_lock(&hob_p->fatal_spl);
  hob_p->xhh_state = XHH_STATE_CLOSED;
  MOSAL_spinlock_unlock(&hob_p->fatal_spl);
  
  MOSAL_mutex_rel(&(hob_p->mtx));

  FREE(tdev_p);
  return(HH_OK);

  /* ERROR HANDLING: undoes previous steps in reverse order, as needed */
err_release_mutex:  
  MOSAL_mutex_rel(&(hob_p->mtx));
err_acq_mutex:  
  VIP_delay_unlock_destroy(hob_p->delay_unlocks);

fatal_thr_init_err:
  /* signal the waited-on sync object, so that thread exits */
  MOSAL_syncobj_signal(&(hob_p->fatal_thread_obj.fatal_err_sync));

//fatal_thr_create_err:
delay_unlock_err:
undo_ddrm_create:
  XHH_lamm_destroy(hob_p->lamm);

undo_sys_ena:
  if ((hob_p->fw_error_iomem_obj != (MOSAL_iomem_t) NULL))  {
     MOSAL_iomem_unmap(hob_p->fw_error_iomem_obj, MOSAL_get_kernel_prot_ctx());
  }

  if (hob_p->fw_error_buf != NULL) {
      FREE(hob_p->fw_error_buf);
  }

cmd_ret = XHH_cmd_SYS_DIS(hob_p->cmd);
if (cmd_ret != XHH_CMD_STAT_OK) {
    if (cmd_ret == XHH_CMD_STAT_EFATAL) {
        MTL_ERROR1(MT_FLFMT("XHH_hob_create: FATAL ERROR in XHH_cmd_SYS_DIS"));
    } else {
        MTL_ERROR1(MT_FLFMT("XHH_hob_create: CMD_error in XHH_cmd_SYS_DIS (%d)"), cmd_ret);
    }
}
cmd_err:
  XHH_cmd_destroy(hob_p->cmd);
  
err_free_hob:
  MOSAL_mutex_free(&(hob_p->mtx));
  FREE(tdev_p);
  FREE(hob_p);
  return (ret);
}

/******************************************************************************
 *  Function:     XHH_hob_create
 *
 *  Description:  Creates the HOB object for an HCA, and registers it in HH  
 *
 *  input:
 *                hw_props_p -- PCI properties (BARs, etc)
 *                hca_seq_num  - a sequence number assigned to this HCA to differentiate it 
 *                               from other HCAs on this host
 *                mod_flags   - flags passed in at module initialization (e.g., insmod)
 *  output: 
 *                hh_hndl_p  - size of returned table (in pkeys)
 *  returns:
 *                HH_OK
 *                HH_EAGAIN
 *                HH_ERR -- other errors
 *
 *  Comments:     This function involves the following steps: 
 *                    1.Allocate XHH_hob data context. 
 *                    2.Create the XHH_cmd_if object instance (in order to enable 
 *                       queries of HCA resources even before HCA is opened). 
 *                    3.Invoke  ENABLE_SYS command ((polling mode). 
 *                    4.Query HCA for available DDRmemory resources 
 *                      (use the Command interface in polling mode)and create 
 *                       the XHH_lamm object based on results. 
 *                    5.Query HCA for other capabilties and save them in XHH_hob context. 
 *                    6.Register HCA in HH (i.e.call HH_add_hca_dev()).
 *
 *                Also initializes the HOB mutex for controlling Open HCA and Close HCA 
 *
 *****************************************************************************/
HH_ret_t    XHH_hob_create(/*IN*/  XHH_hw_props_t   *hw_props_p,
                           /*IN*/  u_int32_t         hca_seq_num,
                           /*IN*/  XHH_module_flags_t *mod_flags,
                           /*OUT*/ HH_hca_hndl_t    *hh_hndl_p )
{
	HH_ret_t ret;
	ret = XHH_hob_create_internal(hw_props_p, hca_seq_num, mod_flags, NULL,hh_hndl_p );
	if (ret != HH_OK) {
		MTL_ERROR1(MT_FLFMT("%s: XHH_hob_create_internal failed (%d:%s)"),
			   __func__, ret, HH_strerror_sym(ret));
		return ret;
	}
	ret= HH_notify_add_hca_dev(*hh_hndl_p);
	if (ret != HH_OK) {
		MTL_ERROR1(MT_FLFMT("%s: HH_notify_add_hca_dev failed (%d:%s)"),
			   __func__, ret, HH_strerror_sym(ret));
		XHH_hob_destroy_internal(*hh_hndl_p, FALSE);
	}
	return ret;
}
/*****************************************************************************
******************************************************************************
************** EXTERNALLY VISIBLE FUNCTIONS, WITH PROTOTYPES IN XHH_HOB.H ****
******************************************************************************
*****************************************************************************/

HH_ret_t XHH_hob_get_uldm ( /*IN*/ XHH_hob_t hob, 
                            /*OUT*/ XHH_uldm_t *uldm_p )
{
  MT_RETURN_IF_LOW_STACK(XHH_WATERMARK);
    if (hob == NULL) {
        MTL_ERROR1("XHH_hob_get_uldm: ERROR : No device registered\n");
        return HH_EINVAL;
    }


    if (hob->uldm == (XHH_uldm_t)XHH_INVALID_HNDL) {
        MTL_ERROR1("XHH_hob_get_uldm: ERROR : HCA device has not yet been opened\n");
        return HH_EINVAL;
    }

    *uldm_p = hob->uldm;
    return HH_OK;
}

HH_ret_t XHH_hob_get_udavm_info ( /*IN*/ XHH_hob_t hob, 
                                  /*OUT*/ XHH_udavm_t *udavm_p,
                                  /*OUT*/ MT_bool *use_priv_udav,
                                  /*OUT*/ MT_bool *av_in_host_mem,
                                  /*OUT*/ VAPI_lkey_t  *lkey ,
                                  /*OUT*/ u_int32_t    *max_ah_num,
                                  /*OUT*/ MT_bool *hide_ddr)
{
  MT_RETURN_IF_LOW_STACK(XHH_WATERMARK);
    if (hob == NULL) {
        MTL_ERROR1("XHH_hob_get_udavm_info: ERROR : No device registered\n");
        return HH_EINVAL;
    }

    *av_in_host_mem = (MT_bool) (hob->module_flags.av_in_host_mem);

    *use_priv_udav = hob->udavm_use_priv;
    *hide_ddr = (hob->ddr_props.dh ? TRUE : FALSE);
    *max_ah_num = hob->hca_capabilities.max_ah_num;
    if (!(hob->udavm_use_priv)) {
        return HH_OK;
    }

    if (hob->udavm == (XHH_udavm_t)XHH_INVALID_HNDL) {
        MTL_ERROR1("XHH_hob_get_udavm_info: ERROR : HCA device has not yet been opened\n");
        return HH_EINVAL;
    }
   
    *udavm_p = hob->udavm;
    *lkey    = hob->udavm_lkey;

   return HH_OK;
}

/* Used in restarting HCA on fatal error, in function XHH_hob_restart */
HH_ret_t   XHH_hob_get_init_params(/*IN*/ XHH_hob_t  xhh_hob_p, 
                   /*OUT*/  XHH_hw_props_t     *hw_props_p,
                   /*OUT*/  u_int32_t          *hca_seq_num,
                   /*OUT*/  XHH_module_flags_t *mod_flags)
{

  MT_RETURN_IF_LOW_STACK(XHH_WATERMARK);
    if (xhh_hob_p == NULL) {
        MTL_ERROR1(MT_FLFMT("%s: ERROR : No device registered"), __func__);  
        return HH_EAGAIN;
    }
    memcpy(hw_props_p, &xhh_hob_p->hw_props, sizeof(XHH_hw_props_t));
    *hca_seq_num = xhh_hob_p->hca_seq_num;
    memcpy(mod_flags, &xhh_hob_p->module_flags, sizeof(XHH_module_flags_t));
    return HH_OK;
}

HH_ret_t  XHH_hob_get_state(XHH_hob_t xhh_hob_p, XHH_hob_state_t *fatal_state)
{
  MT_RETURN_IF_LOW_STACK(XHH_WATERMARK);
    if (xhh_hob_p == NULL) {
        MTL_ERROR1(MT_FLFMT("%s: ERROR : No device registered"), __func__);  
        return HH_EAGAIN;
    }
    
    if (fatal_state == NULL) {
        MTL_ERROR1(MT_FLFMT("%s: ERROR : NULL fatal_state parameter"), __func__);  
        return HH_EINVAL;
    }

    MOSAL_spinlock_dpc_lock(&xhh_hob_p->fatal_spl);
    *fatal_state = xhh_hob_p->xhh_state;
    MOSAL_spinlock_unlock(&xhh_hob_p->fatal_spl);
    return HH_OK;
}

/*****************************************************************************
******************************************************************************
************** PASS-THROUGH FUNCTIONS  ********************************** ****
******************************************************************************
*****************************************************************************/


/******************************************************************************
 *  Function:     XHH_hob_alloc_ul_res <==> XHH_uldm_alloc_ul_res
 *****************************************************************************/
HH_ret_t XHH_hob_alloc_ul_res(HH_hca_hndl_t      hca_hndl,
                                 MOSAL_protection_ctx_t prot_ctx,
                                 void                   *hca_ul_resources_p)
{
  XHH_hca_ul_resources_t*  res = (XHH_hca_ul_resources_t *)hca_ul_resources_p;
  XHH_hob_t  xhh_hob_p;

  MT_RETURN_IF_LOW_STACK(XHH_WATERMARK);
  if (MOSAL_get_exec_ctx() != MOSAL_IN_TASK) {
      MTL_ERROR1("XHH_hob_alloc_ul_res: NOT IN TASK CONTEXT)\n");
      return HH_ERR;
  }

  if (hca_hndl == NULL) {
      MTL_ERROR1("XHH_hob_alloc_ul_res : ERROR : Invalid HCA handle\n");
      return HH_EINVAL_HCA_HNDL;
  }
  xhh_hob_p = XHHOBP(hca_hndl);

  /* Need to see that uldm object has been allocated.  Then, need to invoke
   * the alloc_ul_resources method of the uldm here. 
   * NOTE: may want the constructor to already pre-allocate the ul resources based upon
   *       configuration info obtained via query.
   */
  memset(res, 0, sizeof(XHH_hca_ul_resources_t));

  if (xhh_hob_p == NULL) {
      MTL_ERROR1("XHH_hob_alloc_ul_res: ERROR : No device registered\n");
      return HH_EAGAIN;
  }
  TEST_RETURN_FATAL(xhh_hob_p);

  if (xhh_hob_p->uldm == (XHH_uldm_t)XHH_INVALID_HNDL) {
      MTL_ERROR1("XHH_hob_alloc_ul_res: ERROR : HCA device has not yet been opened\n");
      return HH_EAGAIN;
  }

  /* Set XHH_hob's information in given hca_ul_res_p buffer */
  ((XHH_hca_ul_resources_t*)hca_ul_resources_p)->hh_hca_hndl= hca_hndl;
  memcpy(&(((XHH_hca_ul_resources_t*)hca_ul_resources_p)->version),
    &(xhh_hob_p->version_info),sizeof(XHH_ver_info_t));
  ((XHH_hca_ul_resources_t*)hca_ul_resources_p)->vendor_id= xhh_hob_p->hw_props.pci_vendor_id;
  ((XHH_hca_ul_resources_t*)hca_ul_resources_p)->device_id= xhh_hob_p->hh_hca_hndl->dev_id;
  ((XHH_hca_ul_resources_t*)hca_ul_resources_p)->pci_bus= xhh_hob_p->pci_hca_info.bus;
  ((XHH_hca_ul_resources_t*)hca_ul_resources_p)->pci_devfn= xhh_hob_p->pci_hca_info.dev_func;
  ((XHH_hca_ul_resources_t*)hca_ul_resources_p)->pci_device_p= xhh_hob_p->pci_hca_info.pci_dev;
  ((XHH_hca_ul_resources_t*)hca_ul_resources_p)->priv_ud_av = xhh_hob_p->profile.use_priv_udav;
  ((XHH_hca_ul_resources_t*)hca_ul_resources_p)->log2_mpt_size = (u_int32_t)xhh_hob_p->profile.log2_max_mpt_entries;
  ((XHH_hca_ul_resources_t*)hca_ul_resources_p)->max_qp_ous_wr= xhh_hob_p->hca_capabilities.max_qp_ous_wr;
  ((XHH_hca_ul_resources_t*)hca_ul_resources_p)->max_srq_ous_wr= xhh_hob_p->hca_capabilities.max_wqe_per_srq;
  ((XHH_hca_ul_resources_t*)hca_ul_resources_p)->max_num_sg_ent= xhh_hob_p->hca_capabilities.max_num_sg_ent;
  ((XHH_hca_ul_resources_t*)hca_ul_resources_p)->max_num_sg_ent_srq= xhh_hob_p->hca_capabilities.max_srq_sentries;
  ((XHH_hca_ul_resources_t*)hca_ul_resources_p)->max_num_sg_ent_rd= xhh_hob_p->hca_capabilities.max_num_sg_ent_rd;
  ((XHH_hca_ul_resources_t*)hca_ul_resources_p)->max_num_ent_cq= xhh_hob_p->hca_capabilities.max_num_ent_cq;

  memcpy(((XHH_hca_ul_resources_t*)hca_ul_resources_p)->vsd, xhh_hob_p->adapter_props.vsd, (XHH_ADAPTER_VSD_DW << 2));

  ((XHH_hca_ul_resources_t*)hca_ul_resources_p)->max_wqe_size_sq = XHH_MAX_WQE_SZ;
  ((XHH_hca_ul_resources_t*)hca_ul_resources_p)->max_wqe_size_rq = XHH_MAX_WQE_SZ;
  ((XHH_hca_ul_resources_t*)hca_ul_resources_p)->max_wqe_size_srq = XHH_MAX_WQE_SZ;

  /* Invoke XHH_uldm in order to get a UAR resource */
  return XHH_uldm_alloc_ul_res(xhh_hob_p->uldm, prot_ctx,  
		  (XHH_hca_ul_resources_t *)hca_ul_resources_p);
} /* XHH_alloc_ul_resources */

/******************************************************************************
 *  Function:     XHH_hob_free_ul_res <==> XHH_uldm_free_ul_res
 *****************************************************************************/
HH_ret_t XHH_hob_free_ul_res(HH_hca_hndl_t  hca_hndl,
                                    void           *hca_ul_resources_p)
{
    XHH_hca_ul_resources_t*  res = (XHH_hca_ul_resources_t *)hca_ul_resources_p;
    XHH_hob_t  xhh_hob_p;

    MT_RETURN_IF_LOW_STACK(XHH_WATERMARK);
    if (MOSAL_get_exec_ctx() != MOSAL_IN_TASK) {
        MTL_ERROR1("XHH_hob_free_ul_res: NOT IN TASK CONTEXT)\n");
        return HH_ERR;
    }

    if (hca_hndl == NULL) {
        MTL_ERROR1("XHH_hob_free_ul_res : ERROR : Invalid HCA handle\n");
        return HH_EINVAL_HCA_HNDL;
    }
    xhh_hob_p = XHHOBP(hca_hndl);

    /* Need to see that uldm object has been allocated.  Then, need to invoke
     * the alloc_ul_resources method of the uldm here. 
     * NOTE: may want the constructor to already pre-allocate the ul resources based upon
     *       configuration info obtained via query.
     */

    if (xhh_hob_p == NULL) {
        MTL_ERROR1("XHH_hob_free_ul_res: ERROR : No device registered\n");
        return HH_EAGAIN;
    }

    if (xhh_hob_p->uldm == (XHH_uldm_t)XHH_INVALID_HNDL) {
        MTL_ERROR1("XHH_hob_free_ul_res: ERROR : HCA device has not yet been opened\n");
        return HH_EAGAIN;
    }

    return XHH_uldm_free_ul_res(xhh_hob_p->uldm, res);
} /* XHH_free_ul_resources */




/******************************************************************************
 *  Function:     XHH_hob_alloc_pd <==> XHH_uldm_alloc_pd
 *****************************************************************************/
HH_ret_t XHH_hob_alloc_pd(HH_hca_hndl_t hca_hndl, 
                             MOSAL_protection_ctx_t prot_ctx, 
                             void * pd_ul_resources_p,
                             HH_pd_hndl_t *pd_num_p)
{
  XHH_hob_t xhh_hob_p;
  HH_ret_t  ret;

  MT_RETURN_IF_LOW_STACK(XHH_WATERMARK);
  if (MOSAL_get_exec_ctx() != MOSAL_IN_TASK) {
      MTL_ERROR1("XHH_hob_alloc_pd: NOT IN TASK CONTEXT)\n");
      return HH_ERR;
  }

  if (hca_hndl == NULL) {
      MTL_ERROR1("XHH_hob_alloc_pd : ERROR : Invalid HCA handle\n");
        return HH_EINVAL_HCA_HNDL;
  }
  xhh_hob_p = XHHOBP(hca_hndl);

  if (xhh_hob_p == NULL) {
      MTL_ERROR1("XHH_hob_alloc_pd: ERROR : No device registered\n");
      return HH_EAGAIN;
  }
  TEST_RETURN_FATAL(xhh_hob_p);


  if (xhh_hob_p->uldm == (XHH_uldm_t)XHH_INVALID_HNDL) {
      MTL_ERROR1("XHH_hob_alloc_pd: ERROR : HCA device has not yet been opened\n");
      return HH_EAGAIN;
  }

  ret = XHH_uldm_alloc_pd(xhh_hob_p->uldm, prot_ctx, 
		  (XHH_pd_ul_resources_t *)pd_ul_resources_p, pd_num_p);
//  MTL_DEBUG4("XHH_hob_alloc_pd: ret = %d\n", ret);
  return ret;
}

/******************************************************************************
 *  Function:     XHH_hob_free_pd <==> XHH_uldm_free_pd
 *****************************************************************************/
HH_ret_t XHH_hob_free_pd(HH_hca_hndl_t hca_hndl, HH_pd_hndl_t pd_num)
{
  u_int32_t   max_pd;
  XHH_hob_t   xhh_hob_p;
  HH_ret_t    rc = HH_OK;

  MT_RETURN_IF_LOW_STACK(XHH_WATERMARK);
  if (MOSAL_get_exec_ctx() != MOSAL_IN_TASK) {
      MTL_ERROR1("XHH_hob_free_pd: NOT IN TASK CONTEXT)\n");
      return HH_ERR; 
  }

  if (hca_hndl == NULL) {
      MTL_ERROR1("XHH_hob_free_pd : ERROR : Invalid HCA handle\n");
      return HH_EINVAL_HCA_HNDL; 
  }
  xhh_hob_p = XHHOBP(hca_hndl);

  if (xhh_hob_p == NULL) {
      MTL_ERROR1("XHH_hob_free_pd: ERROR : No device registered\n");
      return HH_EAGAIN;
  }

  max_pd = (1 << xhh_hob_p->dev_lims.log_max_pd);
  if (pd_num > max_pd  - 1) {
    MTL_ERROR1("XHH_hob_free_pd: ERROR : PD number (%d) is greater than max allowed (%d)\n", pd_num, max_pd);
    return HH_EAGAIN;
  }

  if (xhh_hob_p->uldm == (XHH_uldm_t)XHH_INVALID_HNDL) {
      MTL_ERROR1("XHH_hob_free_pd: ERROR : HCA device has not yet been opened\n");
      return HH_EAGAIN;
  }
  
  rc = XHH_uldm_free_pd(xhh_hob_p->uldm, pd_num);
  return rc; 
}


/******************************************************************************
 *  Function:     XHH_hob_alloc_rdd <==> XHH_eecm_alloc_rdd
 *****************************************************************************/
HH_ret_t XHH_hob_alloc_rdd(HH_hca_dev_t *hh_dev_p, 
                                 HH_rdd_hndl_t *rdd_p)
{
#if 0
  XHH_hob_t  xhh_hob_p;

  MT_RETURN_IF_LOW_STACK(XHH_WATERMARK);
  if (MOSAL_get_exec_ctx() != MOSAL_IN_TASK) {
      MTL_ERROR1("XHH_hob_alloc_rdd: NOT IN TASK CONTEXT)\n");
      return HH_ERR;
  }

  if (hca_hndl == NULL) {
      MTL_ERROR1("XHH_hob_alloc_rdd : ERROR : Invalid HCA handle\n");
      return HH_EINVAL_HCA_HNDL;
  }
  xhh_hob_p = XHHOBP(hca_hndl);

  if (xhh_hob_p == NULL) {
      MTL_ERROR1("XHH_hob_alloc_rdd: ERROR : No device registered\n");
      return HH_EAGAIN;
  }
  TEST_RETURN_FATAL(xhh_hob_p);


  if (xhh_hob_p->eecm == (XHH_eecm_t)XHH_INVALID_HNDL) {
      MTL_ERROR1("XHH_hob_alloc_rdd: ERROR : HCA device has not yet been opened\n");
      return HH_EAGAIN;
  }

  return XHH_eecm_alloc_rdd(xhh_hob_p->uldm, rdd_p);
#else
  return HH_ENOSYS;
#endif
}

/******************************************************************************
 *  Function:     XHH_hob_free_rdd <==> XHH_eecm_free_rdd
 *****************************************************************************/
HH_ret_t XHH_hob_free_rdd(HH_hca_dev_t *hh_dev_p, HH_rdd_hndl_t rdd)
{
#if 0
  XHH_hob_t  xhh_hob_p;
  HH_ret_t   rc = HH_OK;

  if (MOSAL_get_exec_ctx() != MOSAL_IN_TASK) {
      MTL_ERROR1("XHH_hob_free_rdd: NOT IN TASK CONTEXT)\n");
      return HH_ERR;
  }

  if (hca_hndl == NULL) {
      MTL_ERROR1("XHH_hob_free_rdd : ERROR : Invalid HCA handle\n");
      return HH_EINVAL_HCA_HNDL;
  }
  xhh_hob_p = XHHOBP(hca_hndl);

  if (xhh_hob_p == NULL) {
      MTL_ERROR1("XHH_hob_free_rdd: ERROR : No device registered\n");
      return HH_EAGAIN;
  }


  if (xhh_hob_p->eecm == (XHH_eecm_t)XHH_INVALID_HNDL) {
      MTL_ERROR1("XHH_hob_free_rdd: ERROR : HCA device has not yet been opened\n");
      return HH_EAGAIN;
  }

  rc = XHH_eecm_free_rdd(xhh_hob_p->uldm, rdd);
  return rc;
#else
  return HH_ENOSYS;
#endif
}


/******************************************************************************
 *  Function:     XHH_hob_create_ud_av <==> XHH_udavm_create_av
 *****************************************************************************/
HH_ret_t XHH_hob_create_ud_av(HH_hca_hndl_t  hca_hndl,
                                     HH_pd_hndl_t    pd,
                                     VAPI_ud_av_t    *av_p, 
                                     HH_ud_av_hndl_t *ah_p)
{
  XHH_hob_t  xhh_hob_p;

  MT_RETURN_IF_LOW_STACK(XHH_WATERMARK);
  if (MOSAL_get_exec_ctx() != MOSAL_IN_TASK) {
      MTL_ERROR1("XHH_hob_create_ud_av: NOT IN TASK CONTEXT)\n");
      return HH_ERR;
  }

  if (hca_hndl == NULL) {
      MTL_ERROR1("XHH_hob_create_ud_av : ERROR : Invalid HCA handle\n");
      return HH_EINVAL_HCA_HNDL;
  }
  xhh_hob_p = XHHOBP(hca_hndl);

  if (xhh_hob_p == NULL) {
      MTL_ERROR1("XHH_hob_create_ud_av: ERROR : No device registered\n");
      return HH_EAGAIN;
  }
  TEST_RETURN_FATAL(xhh_hob_p);


  if (xhh_hob_p->udavm == (XHH_udavm_t)XHH_INVALID_HNDL) {
      MTL_ERROR1("XHH_hob_create_ud_av: ERROR : HCA device has not yet been opened\n");
      return HH_EAGAIN;
  }

  return XHH_udavm_create_av(xhh_hob_p->udavm, pd, av_p, ah_p);
}


/******************************************************************************
 *  Function:     XHH_hob_modify_ud_av <==> XHH_udavm_modify_av
 *****************************************************************************/
HH_ret_t XHH_hob_modify_ud_av(HH_hca_hndl_t  hca_hndl, 
                                     HH_ud_av_hndl_t ah,
                                     VAPI_ud_av_t    *av_p)
{
  XHH_hob_t  xhh_hob_p;

  MT_RETURN_IF_LOW_STACK(XHH_WATERMARK);
  if (MOSAL_get_exec_ctx() != MOSAL_IN_TASK) {
      MTL_ERROR1("XHH_hob_modify_ud_av: NOT IN TASK CONTEXT)\n");
      return HH_ERR;
  }

  if (hca_hndl == NULL) {
      MTL_ERROR1("XHH_hob_modify_ud_av : ERROR : Invalid HCA handle\n");
      return HH_EINVAL_HCA_HNDL;
  }
  xhh_hob_p = XHHOBP(hca_hndl);

  if (xhh_hob_p == NULL) {
      MTL_ERROR1("XHH_hob_modify_ud_av: ERROR : No device registered\n");
      return HH_EAGAIN;
  }
  TEST_RETURN_FATAL(xhh_hob_p);


  if (xhh_hob_p->udavm == (XHH_udavm_t)XHH_INVALID_HNDL) {
      MTL_ERROR1("XHH_hob_modify_ud_av: ERROR : HCA device has not yet been opened\n");
      return HH_EAGAIN;
  }

  return XHH_udavm_modify_av(xhh_hob_p->udavm, ah, av_p);
}


/******************************************************************************
 *  Function:     XHH_hob_query_ud_av <==> XHH_udavm_query_av
 *****************************************************************************/
HH_ret_t XHH_hob_query_ud_av(HH_hca_hndl_t  hca_hndl, 
                                    HH_ud_av_hndl_t ah,
                                    VAPI_ud_av_t    *av_p)
{
  XHH_hob_t  xhh_hob_p;

  MT_RETURN_IF_LOW_STACK(XHH_WATERMARK);
  if (MOSAL_get_exec_ctx() != MOSAL_IN_TASK) {
      MTL_ERROR1("XHH_hob_query_ud_av: NOT IN TASK CONTEXT)\n");
      return HH_ERR;
  }

  if (hca_hndl == NULL) {
      MTL_ERROR1("XHH_hob_query_ud_av : ERROR : Invalid HCA handle\n");
      return HH_EINVAL_HCA_HNDL;
  }
  xhh_hob_p = XHHOBP(hca_hndl);

  if (xhh_hob_p == NULL) {
      MTL_ERROR1("XHH_hob_query_ud_av: ERROR : No device registered\n");
      return HH_EAGAIN;
  }
  TEST_RETURN_FATAL(xhh_hob_p);


  if (xhh_hob_p->udavm == (XHH_udavm_t)XHH_INVALID_HNDL) {
      MTL_ERROR1("XHH_hob_query_ud_av: ERROR : HCA device has not yet been opened\n");
      return HH_EAGAIN;
  }

  return XHH_udavm_query_av(xhh_hob_p->udavm, ah, av_p);
}


/******************************************************************************
 *  Function:     XHH_hob_destroy_ud_av <==> XHH_udavm_destroy_av
 *****************************************************************************/
HH_ret_t XHH_hob_destroy_ud_av(HH_hca_hndl_t  hca_hndl, 
                                    HH_ud_av_hndl_t   ah)
{
  XHH_hob_t  xhh_hob_p;
  HH_ret_t   rc = HH_OK;

  MT_RETURN_IF_LOW_STACK(XHH_WATERMARK);
  if (MOSAL_get_exec_ctx() != MOSAL_IN_TASK) {
      MTL_ERROR1("XHH_hob_destroy_ud_av: NOT IN TASK CONTEXT)\n");
      return HH_ERR;
  }

  if (hca_hndl == NULL) {
      MTL_ERROR1("XHH_hob_destroy_ud_av : ERROR : Invalid HCA handle\n");
      return HH_EINVAL_HCA_HNDL;
  }
  xhh_hob_p = XHHOBP(hca_hndl);

  if (xhh_hob_p == NULL) {
      MTL_ERROR1("XHH_hob_destroy_ud_av: ERROR : No device registered\n");
      return HH_EAGAIN;
  }


  if (xhh_hob_p->udavm == (XHH_udavm_t)XHH_INVALID_HNDL) {
      MTL_ERROR1("XHH_hob_destroy_ud_av: ERROR : HCA device has not yet been opened\n");
      return HH_EAGAIN;
  }
  rc = XHH_udavm_destroy_av(xhh_hob_p->udavm, ah);
  return rc;
}

  /* Fast Memory Regions */
  /***********************/
/******************************************************************************
 *  Function:     XHH_hob_create_cq <==> XHH_mrwm_alloc_fmr
 *****************************************************************************/
HH_ret_t  XHH_hob_alloc_fmr(HH_hca_hndl_t  hca_hndl,
                            HH_pd_hndl_t   pd,
                            VAPI_mrw_acl_t acl, 
                            MT_size_t      max_pages,      /* Maximum number of pages that can be mapped using this region */
                            u_int8_t       log2_page_sz,	 /* Fixed page size for all maps on a given FMR */
                            VAPI_lkey_t*   last_lkey_p)    /* To be used as the initial FMR handle */
{
  XHH_hob_t  xhh_hob_p;

  MT_RETURN_IF_LOW_STACK(XHH_WATERMARK);
  if (hca_hndl == NULL) {
      MTL_ERROR1("XHH_hob_alloc_fmr : ERROR : Invalid HCA handle\n");
      return HH_EINVAL_HCA_HNDL;
  }
  xhh_hob_p = XHHOBP(hca_hndl);

  if (xhh_hob_p == NULL) {
      MTL_ERROR1("XHH_hob_alloc_fmr: ERROR : No device registered\n");
      return HH_EAGAIN;
  }
  TEST_RETURN_FATAL(xhh_hob_p);


  if (xhh_hob_p->mrwm == (XHH_mrwm_t)XHH_INVALID_HNDL) {
      MTL_ERROR1("XHH_hob_alloc_fmr: ERROR : HCA device has not yet been opened\n");
      return HH_EAGAIN;
  }

  if (xhh_hob_p->ddr_props.dh == TRUE) {
      /* Must hide DDR memory. alloc fmr not supported */
      MTL_ERROR1("XHH_hob_alloc_fmr: Device is operating in HIDE_DDR mode.  Cannot alloc fmr\n");
      return HH_ENOSYS;
  }

  return XHH_mrwm_alloc_fmr(xhh_hob_p->mrwm,pd,acl,max_pages,log2_page_sz,last_lkey_p);
}

/******************************************************************************
 *  Function:     XHH_hob_map_fmr <==> XHH_mrwm_map_fmr
 *****************************************************************************/
HH_ret_t  XHH_hob_map_fmr(HH_hca_hndl_t  hca_hndl,
                        VAPI_lkey_t      last_lkey,
                        EVAPI_fmr_map_t* map_p,
                        VAPI_lkey_t*     lkey_p,
                        IB_rkey_t*       rkey_p)

{
  XHH_hob_t  xhh_hob_p;

  MT_RETURN_IF_LOW_STACK(XHH_WATERMARK);
  if (hca_hndl == NULL) {
      MTL_ERROR1("XHH_hob_map_fmr : ERROR : Invalid HCA handle\n");
      return HH_EINVAL_HCA_HNDL;
  }
  xhh_hob_p = XHHOBP(hca_hndl);

  if (xhh_hob_p == NULL) {
      MTL_ERROR1("XHH_hob_map_fmr: ERROR : No device registered\n");
      return HH_EAGAIN;
  }
  TEST_RETURN_FATAL(xhh_hob_p);


  if (xhh_hob_p->mrwm == (XHH_mrwm_t)XHH_INVALID_HNDL) {
      MTL_ERROR1("XHH_hob_map_fmr: ERROR : HCA device has not yet been opened\n");
      return HH_EAGAIN;
  }

  return XHH_mrwm_map_fmr(xhh_hob_p->mrwm,last_lkey,map_p,lkey_p,rkey_p);
}
  
/******************************************************************************
 *  Function:     XHH_hob_unmap_fmr <==> XHH_mrwm_unmap_fmr
 *****************************************************************************/
HH_ret_t  XHH_hob_unmap_fmr(HH_hca_hndl_t hca_hndl,
                          u_int32_t     num_of_fmrs_to_unmap,
                          VAPI_lkey_t*  last_lkeys_array)

{
  XHH_hob_t  xhh_hob_p;

  MT_RETURN_IF_LOW_STACK(XHH_WATERMARK);
  if (hca_hndl == NULL) {
      MTL_ERROR1("XHH_hob_unmap_fmr : ERROR : Invalid HCA handle\n");
      return HH_EINVAL_HCA_HNDL;
  }
  xhh_hob_p = XHHOBP(hca_hndl);

  if (xhh_hob_p == NULL) {
      MTL_ERROR1("XHH_hob_unmap_fmr: ERROR : No device registered\n");
      return HH_EAGAIN;
  }
  TEST_RETURN_FATAL(xhh_hob_p);


  if (xhh_hob_p->mrwm == (XHH_mrwm_t)XHH_INVALID_HNDL) {
      MTL_ERROR1("XHH_hob_unmap_fmr: ERROR : HCA device has not yet been opened\n");
      return HH_EAGAIN;
  }

  return XHH_mrwm_unmap_fmr(xhh_hob_p->mrwm,num_of_fmrs_to_unmap,last_lkeys_array);
}

/******************************************************************************
 *  Function:     XHH_hob_free_fmr <==> XHH_mrwm_free_fmr
 *****************************************************************************/
HH_ret_t  XHH_hob_free_fmr(HH_hca_hndl_t  hca_hndl,
                           VAPI_lkey_t      last_lkey)   /* as returned on last successful mapping operation */

{
  XHH_hob_t  xhh_hob_p;
  HH_ret_t   rc = HH_OK;

  MT_RETURN_IF_LOW_STACK(XHH_WATERMARK);
  if (hca_hndl == NULL) {
      MTL_ERROR1("XHH_hob_free_fmr : ERROR : Invalid HCA handle\n");
      return HH_EINVAL_HCA_HNDL;
  }
  xhh_hob_p = XHHOBP(hca_hndl);

  if (xhh_hob_p == NULL) {
      MTL_ERROR1("XHH_hob_free_fmr: ERROR : No device registered\n");
      return HH_EAGAIN;
  }


  if (xhh_hob_p->mrwm == (XHH_mrwm_t)XHH_INVALID_HNDL) {
      MTL_ERROR1("XHH_hob_free_fmr: ERROR : HCA device has not yet been opened\n");
      return HH_EAGAIN;
  }

  rc = XHH_mrwm_free_fmr(xhh_hob_p->mrwm,last_lkey);
  return rc;
}

/******************************************************************************
 *  Function:     XHH_hob_lamm_alloc <==> XHH_lamm_alloc
 *****************************************************************************/
HH_ret_t XHH_hob_lamm_alloc(
                            HH_hca_hndl_t  hca_hndl,
                            VAPI_size_t     size, 
                            u_int8_t      align_shift,
                            VAPI_phy_addr_t*  buf_p)
{
    XHH_hob_t  xhh_hob_p;
    MT_phys_addr_t adrs;
    HH_ret_t ret = HH_OK;

    MT_RETURN_IF_LOW_STACK(XHH_WATERMARK);
    if (MOSAL_get_exec_ctx() != MOSAL_IN_TASK) {
        MTL_ERROR1(MT_FLFMT(" NOT IN TASK CONTEXT\n"));
        return HH_ERR;
    }

    if (hca_hndl == NULL) {
        MTL_ERROR1(MT_FLFMT("ERROR : Invalid HCA handle\n"));
        return HH_EINVAL_HCA_HNDL;
    }
    xhh_hob_p = XHHOBP(hca_hndl);

    if (xhh_hob_p == NULL) {
        MTL_ERROR1(MT_FLFMT("ERROR : No device registered\n"));
        return HH_EAGAIN;
    }
    TEST_RETURN_FATAL(xhh_hob_p);


    if (xhh_hob_p->lamm == (XHH_lamm_t)XHH_INVALID_HNDL) {
        MTL_ERROR1(MT_FLFMT("ERROR : HCA device has not yet been opened\n"));
        return HH_EAGAIN;
    }

    if (xhh_hob_p->ddr_props.dh == TRUE) {
        /* Must hide DDR memory. alloc fmr not supported */
        MTL_ERROR1(MT_FLFMT("%s: Device is operating in HIDE_DDR mode.  Cannot alloc ddr memory"), __func__);
        return HH_ENOSYS;
    }

    MTL_DEBUG1(MT_FLFMT("before XHH_lamm_alloc \n"));
    /* tavor ALWAYS has DDR, on other devices we should query if ther's DDR */
    ret = XHH_lamm_alloc(xhh_hob_p->lamm,(MT_size_t)size,align_shift,&adrs);
    MTL_DEBUG1(MT_FLFMT("after XHH_lamm_alloc \n"));
    *buf_p = (VAPI_phy_addr_t)adrs;
    return ret;
    
}

/******************************************************************************
 *  Function:     XHH_hob_lamm_query <==> XHH_lamm_query
 *****************************************************************************/
HH_ret_t XHH_hob_lamm_query(
                            HH_hca_hndl_t  hca_hndl,
                            u_int8_t      align_shift,   
                            VAPI_size_t*    total_mem,    
                            VAPI_size_t*    free_mem,     
                            VAPI_size_t*    largest_chunk,  
                            VAPI_phy_addr_t*  largest_free_addr_p)
{
    XHH_hob_t  xhh_hob_p;
        
    MT_RETURN_IF_LOW_STACK(XHH_WATERMARK);
    if (MOSAL_get_exec_ctx() != MOSAL_IN_TASK) {
        MTL_ERROR1(MT_FLFMT(" NOT IN TASK CONTEXT\n"));
        return HH_ERR;
    }

    if (hca_hndl == NULL) {
        MTL_ERROR1(MT_FLFMT("ERROR : Invalid HCA handle\n"));
        return HH_EINVAL_HCA_HNDL;
    }
    xhh_hob_p = XHHOBP(hca_hndl);

    if (xhh_hob_p == NULL) {
        MTL_ERROR1(MT_FLFMT("ERROR : No device registered\n"));
        return HH_EAGAIN;
    }
    TEST_RETURN_FATAL(xhh_hob_p);


    if (xhh_hob_p->lamm == (XHH_lamm_t)XHH_INVALID_HNDL) {
        MTL_ERROR1(MT_FLFMT("ERROR : HCA device has not yet been opened\n"));
        return HH_EAGAIN;
    }

    if (xhh_hob_p->ddr_props.dh == TRUE) {
        /* Must hide DDR memory. alloc fmr not supported */
        MTL_ERROR1(MT_FLFMT("%s: Device is operating in HIDE_DDR mode.  Cannot query ddr memory"), __func__);
        return HH_ENOSYS;
    }

    MTL_DEBUG1(MT_FLFMT("before XHH_lamm_query \n"));
    /* tavor ALWAYS has DDR, on other devices we should query if ther's DDR */
    return XHH_lamm_query(xhh_hob_p->lamm,align_shift,total_mem,free_mem,largest_chunk,largest_free_addr_p);
    
}


/******************************************************************************
 *  Function:     XHH_hob_lamm_free <==> XHH_lamm_free
 *****************************************************************************/
HH_ret_t XHH_hob_lamm_free(
                            HH_hca_hndl_t  hca_hndl,
                            VAPI_phy_addr_t  buf,
                            VAPI_size_t     size)
                            
{
    XHH_hob_t  xhh_hob_p;
    HH_ret_t   rc = HH_OK;

    MT_RETURN_IF_LOW_STACK(XHH_WATERMARK);
    if (MOSAL_get_exec_ctx() != MOSAL_IN_TASK) {
        MTL_ERROR1(MT_FLFMT(" NOT IN TASK CONTEXT\n"));
        return HH_ERR;
    }

    if (hca_hndl == NULL) {
        MTL_ERROR1(MT_FLFMT("ERROR : Invalid HCA handle\n"));
        return HH_EINVAL_HCA_HNDL;
    }
    xhh_hob_p = XHHOBP(hca_hndl);

    if (xhh_hob_p == NULL) {
        MTL_ERROR1(MT_FLFMT("ERROR : No device registered\n"));
        return HH_EAGAIN;
    }


    if (xhh_hob_p->lamm == (XHH_lamm_t)XHH_INVALID_HNDL) {
        MTL_ERROR1(MT_FLFMT("ERROR : HCA device has not yet been opened\n"));
        return HH_EAGAIN;
    }

    rc = XHH_lamm_free(xhh_hob_p->lamm,buf,(MT_size_t)size);
    return rc;
}

HH_ret_t XHH_get_debug_info(
	HH_hca_hndl_t hca_hndl, 		/*IN*/
	XHH_debug_info_t *debug_info_p	/*OUT*/
)
{
    XHH_hob_t  hob_p;
    HH_ret_t   rc = HH_OK;
    MT_bool    have_error = FALSE;

    MT_RETURN_IF_LOW_STACK(XHH_WATERMARK);
    if (MOSAL_get_exec_ctx() != MOSAL_IN_TASK) {
        MTL_ERROR1("XHH_get_debug_info: NOT IN TASK CONTEXT)\n");
        return HH_ERR;
    }

    if (hca_hndl == NULL) {
        MTL_ERROR1("XHH_get_debug_info : ERROR : Invalid HCA handle\n");
        return HH_EINVAL_HCA_HNDL;
    }
    hob_p = XHHOBP(hca_hndl);

    if (hob_p == NULL) {
        MTL_ERROR1("XHH_get_debug_info: ERROR : No device registered\n");
        return HH_EAGAIN;
    }

    memset(debug_info_p, 0, sizeof(XHH_debug_info_t));
    
    memcpy(&(debug_info_p->hw_props), &(hob_p->hw_props), sizeof(XHH_hw_props_t));
    memcpy(&(debug_info_p->profile), &(hob_p->profile), sizeof(XHH_profile_t));
    memcpy(&(debug_info_p->ddr_addr_vec), &(hob_p->ddr_alloc_base_addrs_vec), 
           sizeof(XHH_ddr_base_addr_vector_t));
    memcpy(&(debug_info_p->ddr_size_vec), &(hob_p->ddr_alloc_size_vec), 
           sizeof(XHH_ddr_allocation_vector_t));
    debug_info_p->num_ddr_addrs = XHH_DDR_ALLOCATION_VEC_SIZE;
    memcpy(&(debug_info_p->mrwm_props), &(hob_p->mrwm_props), sizeof(XHH_mrwm_props_t));

    debug_info_p->hide_ddr = hob_p->ddr_props.dh;

    rc = XHH_uldm_get_num_objs(hob_p->uldm,&(debug_info_p->allocated_ul_res),&(debug_info_p->allocated_pd));
    if (rc != HH_OK) {
        MTL_ERROR1(MT_FLFMT("XHH_get_debug_info:  proc XHH_uldm_get_num_of_objs returned ERROR"));
        have_error = TRUE;
    }
    
    rc = XHH_mrwm_get_num_objs(hob_p->mrwm,&(debug_info_p->allocated_mr_int),
                                  &(debug_info_p->allocated_mr_ext), &(debug_info_p->allocated_mw));
    if (rc != HH_OK) {
        MTL_ERROR1(MT_FLFMT("XHH_get_debug_info:  proc XHH_mrwm_get_num_of_objs returned ERROR"));
        have_error = TRUE;
    }

    rc = XHH_qpm_get_num_qps(hob_p->qpm,&(debug_info_p->allocated_qp));
    if (rc != HH_OK) {
        MTL_ERROR1(MT_FLFMT("XHH_get_debug_info:  proc XHH_tqpm_get_num_of_qps returned ERROR"));
        have_error = TRUE;
    }
    
    rc = XHH_cqm_get_num_cqs(hob_p->cqm,&(debug_info_p->allocated_cq));
    if (rc != HH_OK) {
        MTL_ERROR1(MT_FLFMT("XHH_get_debug_info:  proc XHH_tcqm_get_num_of_cqs returned ERROR"));
        have_error = TRUE;
    }
    
    rc = XHH_mcgm_get_num_mcgs(hob_p->mcgm,&(debug_info_p->allocated_mcg));
    if (rc != HH_OK) {
        MTL_ERROR1(MT_FLFMT("XHH_get_debug_info:  proc XHH_mcgm_get_num_of_mcgs returned ERROR"));
        have_error = TRUE;
    }

    return (have_error ? HH_ERR : HH_OK);
}

HH_ret_t XHH_hob_get_lamm ( /*IN*/ XHH_hob_t hob, 
                            /*OUT*/ XHH_lamm_t *lamm_p )
{
  MT_RETURN_IF_LOW_STACK(XHH_WATERMARK);
    if (hob == NULL) {
        MTL_ERROR1("XHH_hob_get_lamm: ERROR : No device registered\n");
        return HH_EINVAL;
    }


    if (hob->lamm == (XHH_lamm_t)XHH_INVALID_HNDL) {
        MTL_ERROR1("XHH_hob_get_lamm: ERROR : HCA device has not yet been opened\n");
        return HH_EINVAL;
    }

    *lamm_p = hob->lamm;
    return HH_OK;
}
HH_ret_t XHH_hob_check_qp_init_attrs ( /*IN*/ XHH_hob_t hob,
                                       /*IN*/ HH_qp_init_attr_t * init_attr_p,
                                       /*IN*/ MT_bool is_special_qp )
{
  MT_RETURN_IF_LOW_STACK(XHH_WATERMARK);
    if (hob == NULL) {
        MTL_ERROR1("XHH_hob_get_check_qp_init_attrs: ERROR : No device registered\n");
        return HH_EINVAL;
    }


    if (init_attr_p == NULL) {
        MTL_ERROR1("XHH_hob_get_check_qp_init_attrs: ERROR : null attributes\n");
        return HH_EINVAL;
    }

    if (init_attr_p->qp_cap.max_oust_wr_rq > hob->hca_capabilities.max_qp_ous_wr  ||
        init_attr_p->qp_cap.max_oust_wr_sq > hob->hca_capabilities.max_qp_ous_wr){
        MTL_ERROR1("%s : max outs work requests more than HCA maximum\n", __func__);
        return HH_E2BIG_WR_NUM;
    }

    if (is_special_qp  || init_attr_p->ts_type != VAPI_TS_RD) {
        if (init_attr_p->qp_cap.max_sg_size_rq > hob->hca_capabilities.max_num_sg_ent ||
                   init_attr_p->qp_cap.max_sg_size_sq > hob->hca_capabilities.max_num_sg_ent) {
            MTL_ERROR1("%s : max s/g list size more than HCA maximum\n", __func__);
            return HH_E2BIG_SG_NUM;
        }
    } else {
        /* is RD */
        if (init_attr_p->qp_cap.max_sg_size_rq > hob->hca_capabilities.max_num_sg_ent_rd ||
                   init_attr_p->qp_cap.max_sg_size_sq > hob->hca_capabilities.max_num_sg_ent_rd) {
            MTL_ERROR1("%s : max s/g list size more than HCA maximum\n", __func__);
            return HH_E2BIG_SG_NUM;
        } 
    }
    return HH_OK;
}

