/*
  This software is available to you under a choice of one of two
  licenses.  You may choose to be licensed under the terms of the GNU
  General Public License (GPL) Version 2, available at
  <http://www.fsf.org/copyleft/gpl.html>, or the OpenIB.org BSD
  license, available in the LICENSE.TXT file accompanying this
  software.  These details are also available at
  <http://openib.org/license.html>.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
  BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
  ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
  CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
  SOFTWARE.

  Copyright (c) 2004 Mellanox Technologies Ltd.  All rights reserved.
*/

#include <vapi.h>
#include <evapi.h>
#include <vapi_common.h>
#include <stdio.h>

#define TIME_LEVEL 1
#include "utils.h" 
#include "all_bw_lat.h"
#include "sock.h" 

#ifndef MTPERF
//#define MTPERF
#endif

#include <mtperf.h>
/*
MTPERF_EXTERN_SEGMENT(HHUL_post_inline_send_req);
MTPERF_EXTERN_SEGMENT(HHUL_post_send_req);
MTPERF_EXTERN_SEGMENT(THH_uar_sendq_dbell);
MTPERF_EXTERN_SEGMENT(WQE_build_send);
MTPERF_EXTERN_SEGMENT(SQ_WQE_copy);
MTPERF_EXTERN_SEGMENT(MOSAL_spinlock_unlock);
MTPERF_EXTERN_SEGMENT(sq_alloc_wqe_link_and_ring);
MTPERF_EXTERN_SEGMENT(MOSAL_spinlock_irq_lock);
MTPERF_EXTERN_SEGMENT(HHUL_poll4cqe);
*/
/***********************************************************************************************
* name:      send_bw_uni_test
* args:
*            params_p -- test parameter structure, need to filled before calling
* returns:
*            RET_OK
*            RET_ERR
*
* description:     
*            This is the main loop for the UD and RC send BW test. The test sends N packets
*            then waits for acknowledge from the receiver, and calculate the BW.
*            
************************************************************************************************/  
ret_val_t send_bw_uni_test(test_params_tx *params_p, int port_ix)
{
    VAPI_ret_t             res;
    u_int64_t t_start=0, t_end=0;

    int i, ii, j, k, l=0, m=0, comp_inter, ask_for_comp=0;

    VAPI_rr_desc_t         rr;
//    VAPI_sg_lst_entry_t  * sg_entry_r;
    VAPI_sr_desc_t       * sr;
    VAPI_sg_lst_entry_t  * sg_entry_s;
    VAPI_wc_desc_t         rc;
    VAPI_wc_desc_t         sc;

    /* Copy all input parameters - don't want to rewrite the loop code */
    VAPI_hca_hndl_t        hca_hndl    = params_p->hca_hndl;
    VAPI_qp_hndl_t         qp_hndl     = params_p->port[port_ix].qp_vect[0].qp_hndl;
    VAPI_cq_hndl_t         s_cq_hndl   = params_p->port[port_ix].s_cq_hndl;
    VAPI_cq_hndl_t         r_cq_hndl   = params_p->port[port_ix].r_cq_hndl;
    VAPI_srq_hndl_t        srq_hndl    = VAPI_INVAL_SRQ_HNDL;
    MOSAL_semaphore_t    * sq_sem_p    = &params_p->port[port_ix].sq_sem;
    MOSAL_semaphore_t    * rq_sem_p    = &params_p->port[port_ix].rq_sem;
    u_int8_t               *recv_buf   = params_p->recv_buf;
//    u_int32_t               *recv_buf   = (u_int32_t *) params_p->recv_buf;
    u_int8_t               *send_buf   = params_p->send_buf;
    int                    is_sender   = params_p->in_params.is_sender;
    u_int32_t            * qp_sent_count;
    int                    size        = params_p->in_params.size;
    int                    iter        = params_p->in_params.iter;      
    int                    list_len    = params_p->in_params.postlist_len;
    call_result_t       mosal_rc;
    char *tstr = (params_p->in_params.transport == IB_TS_UD) ? "UD" : ((params_p->in_params.transport == IB_TS_UC) ? "UC" : "RC");
    u_int32_t tx_overhead = (params_p->in_params.transport == IB_TS_UD) ? 40 : 0;
    int noreply = TRUE, sync_received=0, sync_sent=0, sync_ask_for_comp=0;    
    int comp_qp;    
    int use_inline = FALSE;
    qp_entry_t * comp_qp_entry;
    u_int32_t  srq_posted_actual;
/*
    printf("Send CQ handle = 0x%x, Rcv CQ handle = 0x%x\n", s_cq_hndl, r_cq_hndl);
    printf("QP num 0x%x, Dest QP num = 0x%x\n", 
           params_p->port[port_ix].qp_vect[0].qp_num,
           params_p->port[port_ix].qp_vect[0].d_qpnum);
*/
//    u_int8_t               *user_buf = VMALLOC(params_p->in_params.mr_recv_size); 
//    u_int32_t              *user_buf = VMALLOC(params_p->in_params.mr_recv_size); 

    //int                    skip        = params_p->in_params.iter;      


    int max_msg_size = MIN2(params_p->in_params.mr_recv_size,params_p->in_params.mr_send_size);
    int max_sends = MIN2(params_p->in_params.num_r_wqe,params_p->in_params.num_s_wqe); 


    if (params_p->in_params.transport == IB_TS_UD)
    {
        max_msg_size = 2048;
        if (size > 2048)
        {
            printf("Warning!!!: UD max size is 2048\n");
            size = 2048;
        }
    }

    if (size > params_p->in_params.mr_recv_size)
    {
        printf("Warning!!!: size=%d > mr_recv_size=%d --- changing iter to %d\n", 
               size,params_p->in_params.mr_recv_size,params_p->in_params.mr_recv_size);
        size = params_p->in_params.mr_recv_size;
    }

    if (list_len > max_sends)
    {
        printf("Warning!!! Post List Length > WQ size --- reducing it to %d\n", max_sends);
        list_len = max_sends;
    }

    sr = VMALLOC(list_len*sizeof(VAPI_sr_desc_t));
    sg_entry_s = VMALLOC(params_p->in_params.num_of_keys*sizeof(VAPI_sg_lst_entry_t));

//    rr = VMALLOC(list_len*sizeof(VAPI_rr_desc_t));
//    sg_entry_r = VMALLOC(params_p->in_params.num_of_keys*sizeof(VAPI_sg_lst_entry_t));

    /**************    RC RR *****************************************************/
    rr.opcode = VAPI_RECEIVE;
    rr.comp_type = VAPI_SIGNALED;
    rr.sg_lst_len = params_p->in_params.sg_list_len;

    rr.sg_lst_p = VMALLOC(rr.sg_lst_len*sizeof(VAPI_sg_lst_entry_t));

    rr.sg_lst_p[0].addr = (VAPI_virt_addr_t)(MT_virt_addr_t)recv_buf;
    rr.sg_lst_p[0].len = params_p->in_params.mr_recv_size;

    for (i=1; i<(int)rr.sg_lst_len; i++) {
        rr.sg_lst_p[i].addr = (VAPI_virt_addr_t)(MT_virt_addr_t)recv_buf;
        rr.sg_lst_p[i].len = 4;
        rr.sg_lst_p[i].lkey = params_p->recv_lkey_arr[0];
    }

    /***************************************************************************/
    for (i=0; i<params_p->in_params.num_of_keys; i++) {    
        sg_entry_s[i].len = size;
        sg_entry_s[i].lkey = params_p->send_lkey_arr[i];
        sg_entry_s[i].addr = (VAPI_virt_addr_t)(MT_virt_addr_t)send_buf;
    }
/*
    for (i=0; i<params_p->in_params.num_of_keys; i++) {    
        sg_entry_r[i].len = params_p->in_params.mr_recv_size;
        sg_entry_r[i].lkey = params_p->recv_lkey_arr[i];                   
        sg_entry_r[i].addr = (VAPI_virt_addr_t)(MT_virt_addr_t)recv_buf;
    }
*/
    for (i=0; i<list_len; i++)
    {
        sr[i].comp_type = VAPI_UNSIGNALED; 
        sr[i].set_se = FALSE;

        sr[i].opcode = VAPI_SEND;
        sr[i].remote_qkey = 0;
        sr[i].remote_addr = params_p->remote_add;
        sr[i].id = 1;

        sr[i].sg_lst_len = 1; 
    
        /* Needed only for UD */
        if (params_p->in_params.multicast > 0)
        {
            sr[i].remote_qp = IB_MULTICAST_QP;
        }
        else
        {
            sr[i].remote_qp = params_p->port[port_ix].qp_vect[0].d_qpnum;
        }
        sr[i].remote_ah = params_p->port[port_ix].av_hndl;
    }

    /* Post all possible receive in advance. Sender need 1 while receiver need N*/
    for (j=0;j<(int)params_p->in_params.num_of_qps;j++) 
    {    
        qp_hndl = params_p->port[port_ix].qp_vect[j].qp_hndl;
        srq_hndl = params_p->port[port_ix].qp_vect[j].srq_hndl;

        if (!is_sender)
        {
            for (i=0;i<max_sends;i++)
            {
                rr.id = i;
                rr.sg_lst_p[0].lkey = params_p->recv_lkey_arr[m];
                if (params_p->in_params.srq_num > 0)
                {
                    res = VAPI_post_srq(hca_hndl, srq_hndl, 1, &rr, &srq_posted_actual);
                    //printf("posted %d SRQs\n", srq_posted_actual);
                }
                else
                {
                    res = VAPI_post_rr(hca_hndl, qp_hndl, &rr);
                }
                if (res != VAPI_OK)
                {
                    printf("Error: Pre-posting Recv buffer (port %d) : %s\n", params_p->port[port_ix].port_num, VAPI_strerror(res));
                    return(RET_ERR);
                }
                m = (m+1)%params_p->in_params.num_of_keys;
            }
        } else
        {
            for (i=0; i<max_sends; i++)
            {
                rr.id = iter+i;
                rr.sg_lst_p[0].lkey = params_p->recv_lkey_arr[m];
                if (params_p->in_params.srq_num > 0)
                {
                    res = VAPI_post_srq(hca_hndl, srq_hndl, 1, &rr, &srq_posted_actual);
                    //printf("posted %d SRQs\n", srq_posted_actual);
                }
                else
                {
                    res = VAPI_post_rr(hca_hndl, qp_hndl, &rr);
                }
                if (res != VAPI_OK)
                {
                    printf("Error: Pre-posting Recv buffer (port %d): %s\n", params_p->port[port_ix].port_num, VAPI_strerror(res));
                    return(RET_ERR);
                }
                m = (m+1)%params_p->in_params.num_of_keys;
            }
        }
    }

    /* sync the test start */
    if (params_p->in_params.ports_num > 1)
    {
        COND_PRINTF2((" **** Test: finished the preparation phase for port %d and waiting for signal on %p ... \n", params_p->port[port_ix].port_num, &params_p->sync));
        MOSAL_sem_rel( &params_p->sem );
        mosal_rc = MOSAL_syncobj_waiton( &params_p->sync, MAX_WAIT_FOR_SYNC_USECS );
        if (mosal_rc != MT_OK)
        {
            printf( "Test hasn't received 'run' command from the main thread for port %d, exiting ...\n", params_p->port[port_ix].port_num);
            return(RET_ERR);
        }
    } else
    {
        sock_sync_msg(params_p->in_params.socket, is_sender);
    }
    if (params_p->in_params.ports_num <= 1)
    {
        COND_PRINTF1(("\n\n************* %s BW Unidirection Test started for port %d  *********************\n\n",
                      tstr, params_p->port[port_ix].port_num));
    }

    if (iter < max_sends)
    {
        comp_inter = max_sends; // Ask completion ONLY on last message!!
    } 
    else if (params_p->in_params.num_of_qps > 512) comp_inter = 8;  // UGLY patch
    else comp_inter = max_sends/2 - 2;

    ask_for_comp = 0;

    if (is_sender)
    {
        qp_sent_count = calloc(sizeof(u_int32_t), params_p->in_params.num_of_qps);
        if ((size >0 ) &&(size <= params_p->max_inline_size) && (list_len == 1))
        {
            use_inline = TRUE;
        }
        
        get_clk64(t_start);

        for (i=0,j=0; i<iter/list_len; i++)
        {
            // printf("Changing Send QP to %d\n", params_p->port[port_ix].qp_vect[j].qp_num);
            qp_hndl = params_p->port[port_ix].qp_vect[j].qp_hndl;

            for (k=0;k<list_len;k++)
            {
                sr[k].id = list_len * i + k;

                /* Ask for completion every comp_inter */
                if ((sr[k].id+1 % comp_inter == 0) || 
                    (sr[k].id == ((iter/list_len)*list_len + k - 1)) ||
                    ((qp_sent_count[j]+k) % comp_inter == 0) || 
                    (qp_sent_count[j]+k == (iter/params_p->in_params.num_of_qps-1)) ||
                    (list_len > 1))		    
                {
                    sr[k].comp_type = VAPI_SIGNALED;
                    ask_for_comp++;
                    COND_PRINTF2((" **** Posting sr signaled i=%d ask=%d\n",i+1,ask_for_comp));
                } else
                {
                    sr[k].comp_type = VAPI_UNSIGNALED;
                }

            }
                
            /* Use Inline data up to the limit size */
            //if ((size >0 ) &&(size <= params_p->max_inline_size) && (list_len == 1))
            if (use_inline == TRUE)
            {
                sr[0].r_key = params_p->remote_rkey_arr[l];
                // printf("Remote R_Key = 0x%x\n", sr[0].r_key);
                sr[0].sg_lst_p = &sg_entry_s[l];
                res = EVAPI_post_inline_sr(hca_hndl, qp_hndl, sr);
            } else
            {
                for (ii=0; ii<list_len; ii++) {
                    sr[ii].r_key = params_p->remote_rkey_arr[l];
                    // printf("Remote R_Key = 0x%x\n", sr[ii].r_key);
                    sr[ii].sg_lst_p = &sg_entry_s[l];
                }
                res = EVAPI_post_sr_list(hca_hndl, qp_hndl, list_len, sr);

                COND_PRINTF3(("Posted SR list up to id=%lld, QP=%d, QP handle 0x%x\n", 
                              sr[list_len-1].id, j, qp_hndl));
            }
            if (res != VAPI_OK)
            {
                printf("Error: Posting send desc : %s\n", VAPI_strerror(res));
                return(RET_ERR);
            }

            /* Wait for completion every comp_inter starting from 2*comp_inter
             * This applies only when comp_iter < iter */
            while (ask_for_comp > 0)
            {
                COND_PRINTF2(("**** Polling for sr comp i=%d ask=%d\n",i+1,ask_for_comp));
                if (RET_OK != PERF_poll_cq(hca_hndl,s_cq_hndl ,&sc,params_p->in_params.use_send_events, 
                                           sq_sem_p))
                {
                    return RET_ERR;
                }
                ask_for_comp--;
                COND_PRINTF3(("**** Got a completion of sr %d i=%d ask=%d pix=%d\n", (int)sc.id, i,ask_for_comp, port_ix));
	            if (params_p->in_params.transport != IB_TS_RC && !params_p->in_params.oneside) {
                    if (RET_OK != PERF_poll_cq(hca_hndl, r_cq_hndl ,&rc,
                                               params_p->in_params.use_recv_events,  rq_sem_p)) {
                        return RET_ERR;
                    }

                    comp_qp_entry = search_qp(rc.local_qp_num, params_p->port[port_ix].qp_vect, 
                                              params_p->in_params.num_of_qps);

                    if (rc.byte_len == tx_overhead) {
                        sync_received++;
                        COND_PRINTF3(("**** Got a completion of SYNC rr %d i=%d packets sync_received=%d\n", 
                                      (int)rc.id,i,sync_received));
                    } else {
                        noreply=FALSE;
                        COND_PRINTF3(("**** Got a completion of final receiver's reply rr %d i=%d size %u\n", 
                                      (int)rc.id,i, rc.byte_len));
                    }
                    rr.id = iter+i;
                    rr.sg_lst_p[0].lkey = params_p->recv_lkey_arr[m];
                    if (params_p->in_params.srq_num > 0)
                    {
                        res = VAPI_post_srq(hca_hndl, comp_qp_entry->srq_hndl, 1, &rr, &srq_posted_actual);
                        //printf("posted %d SRQs\n", srq_posted_actual);
                    }
                    else
                    {
                        res = VAPI_post_rr(hca_hndl, comp_qp_entry->qp_hndl, &rr);
                    }
                    if (res != VAPI_OK) {
                        printf("Error: Pre-posting Recv buffer (port %d): %s\n", params_p->port[port_ix].port_num, VAPI_strerror(res));
                        return(RET_ERR);
                    }
                    m = (m+1)%params_p->in_params.num_of_keys;
                }
/*
                if (params_p->in_params.transport != IB_TS_RC)
                {
                    if (RET_OK != PERF_poll_cq(hca_hndl, r_cq_hndl ,&rc,
                                               params_p->in_params.use_recv_events,  rq_sem_p))
                    {
                        return RET_ERR;
                    }
                    if (rc.byte_len == tx_overhead)
                    {
                        sync_received++;
                        COND_PRINTF3(("**** Got a completion of SYNC rr %d i=%d packets sync_received=%d\n", 
                                      (int)rc.id,i,sync_received));
                    } else
                    {
                        noreply=FALSE;
                        COND_PRINTF3(("**** Got a completion of final receiver's reply rr %d i=%d size %u\n", 
                                      (int)rc.id,i, rc.byte_len));
                    }
                    rr.id = iter+i;
                    res = VAPI_post_rr(hca_hndl, qp_hndl, &rr);
                    if (res != VAPI_OK)
                    {
                        printf("Error: Pre-posting Recv buffer (port %d): %s\n", params_p->port[port_ix].port_num, VAPI_strerror(res));
                        return(RET_ERR);
                    }
                }
*/
            }

            qp_sent_count[j] += list_len;
            j = (j+1)%params_p->in_params.num_of_qps;            
        
            l = (l+1)%params_p->in_params.num_of_keys;
        }//for i - send packets

        /* Wait for completion on last send */
        COND_PRINTF2(("**** Polling for last sr comp ask=%d\n",ask_for_comp));
        while (ask_for_comp>0)
        {
            /* Poll for the last send completion */
            if (RET_OK != PERF_poll_cq(hca_hndl,s_cq_hndl ,&sc,params_p->in_params.use_send_events,
                                       sq_sem_p))
            {
                return RET_ERR;
            }
            ask_for_comp--;
            COND_PRINTF3(("**** Got a completion of sr %d pix=%d ask=%d\n", (int)sc.id, port_ix,ask_for_comp));
        }

        /* Wait for final reply from receiver */
        COND_PRINTF2(("**** Waiting for final reply from receiver ask=%d\n",ask_for_comp));
                
        while (noreply && !params_p->in_params.oneside) 
        {
            if (RET_OK != PERF_poll_cq(hca_hndl, r_cq_hndl ,&rc,
                                       params_p->in_params.use_recv_events,  rq_sem_p)) {
                return RET_ERR;
            }

            comp_qp_entry = search_qp(rc.local_qp_num, params_p->port[port_ix].qp_vect, 
                                      params_p->in_params.num_of_qps);

            if (params_p->in_params.transport == IB_TS_RC) {
                COND_PRINTF3(("**** Got a completion of final receiver's reply rr %d i=%d size %u\n", 
                              (int)rc.id,i, rc.byte_len));
                noreply = FALSE;                
            } else {
                if (rc.byte_len == tx_overhead) {
                    sync_received++;
                    COND_PRINTF3(("**** Got a completion of SYNC rr %d i=%d packets sync_received=%d\n", 
                                  (int)rc.id,i,sync_received));
                } else {
                    noreply=FALSE;
                    COND_PRINTF3(("**** Got a completion of final receiver's reply rr %d i=%d size %u\n", 
                                  (int)rc.id,i, rc.byte_len));
                }
                rr.id = iter+i;
                rr.sg_lst_p[0].lkey = params_p->recv_lkey_arr[m];
                if (params_p->in_params.srq_num > 0)
                {
                    res = VAPI_post_srq(hca_hndl, comp_qp_entry->srq_hndl, 1, &rr, &srq_posted_actual);
                    //printf("posted %d SRQs\n", srq_posted_actual);
                }
                else
                {
                    res = VAPI_post_rr(hca_hndl, comp_qp_entry->qp_hndl, &rr);
                }
                if (res != VAPI_OK) {
                    printf("Error: Pre-posting Recv buffer (port %d): %s\n", params_p->port[port_ix].port_num, VAPI_strerror(res));
                    return(RET_ERR);
                }
                m = (m+1)%params_p->in_params.num_of_keys;
            }
        }    

/*
        while (noreply)
        {
            if (RET_OK != PERF_poll_cq(hca_hndl, r_cq_hndl ,&rc,
                                       params_p->in_params.use_recv_events,  rq_sem_p))
            {
                return RET_ERR;
            }
            if (params_p->in_params.transport == IB_TS_RC)
            {
                COND_PRINTF3(("**** Got a completion of final receiver's reply rr %d i=%d size %u\n", 
                              (int)rc.id,i, rc.byte_len));
                noreply = FALSE;
            } else
            {
                if (rc.byte_len == tx_overhead)
                {
                    sync_received++;
                    COND_PRINTF3(("**** Got a completion of SYNC rr %d i=%d packets sync_received=%d\n", 
                                  (int)rc.id,i,sync_received));
                } else
                {
                    noreply=FALSE;
                    COND_PRINTF3(("**** Got a completion of final receiver's reply rr %d i=%d size %u\n", 
                                  (int)rc.id,i, rc.byte_len));
                }
                rr.id = iter+i;
                res = VAPI_post_rr(hca_hndl, qp_hndl, &rr);
                if (res != VAPI_OK)
                {
                    printf("Error: Pre-posting Recv buffer (port %d): %s\n", params_p->port[port_ix].port_num, VAPI_strerror(res));
                    return(RET_ERR);
                }
            }
        }
*/
        COND_PRINTF3(("Sender came out of Poll msg_size=%u\n",rc.byte_len));
        get_clk64(t_end);

        if (params_p->in_params.ports_num > 1)
        {
            params_p->port[port_ix].snd.t_start = t_start;
            params_p->port[port_ix].snd.t_end = t_end;
            params_p->port[port_ix].size = size;
        } else
        {
            //printf("BW: %.1f MBytes/sec [sg_list_len: %d, Num of QPs: %d, size: %d bytes, iter: %d, total %.0f]\n", 
              printf("BW: %.1f MBytes/sec [size: %d bytes, iter: %d, total %.0f]\n", 
                   BW_MBSEC(t_end - t_start, (1.0 * size) * iter), 
              //     params_p->in_params.sg_list_len, params_p->in_params.num_of_qps, 
                   size, iter, (1.0 * size) * iter);
        }
        //printf("BW: %.1f MBytes/sec [sg_list_len: %d, Num of QPs: %d, size: %d bytes, iter: %d, total %.0f]\n", 
        ////  printf("BW: %.1f MBytes/sec [size: %d bytes, iter: %d, total %.0f]\n", 
        //       BW_MBSEC(t_end - t_start, (1.0 * size) * iter), 
        //       params_p->in_params.sg_list_len, params_p->in_params.num_of_qps, 
        //       size, iter, (1.0 * size) * iter);
    }// Is sender

    else
    { /* Receiver side */

        u_int32_t * qp_sync_sent_count = calloc(sizeof(u_int32_t), params_p->in_params.num_of_qps);

        sg_entry_s[0].addr = (VAPI_virt_addr_t)(MT_virt_addr_t)send_buf;
        sr[0].sg_lst_len = 0;
        sr[0].sg_lst_p = &sg_entry_s[0];

        for (i=0,j=0; i<iter; i++)
        {
            if (RET_OK != PERF_poll_cq(hca_hndl, r_cq_hndl, &rc,
                                       params_p->in_params.use_recv_events, rq_sem_p))
            {
                return RET_ERR;
            }
            COND_PRINTF3(("**** Got a completion of rr %d i=%d pix=%d\n", (int)rc.id, i, port_ix));                           
            
            //printf("local_qp_num = 0x%x\n", rc.local_qp_num);
            
            comp_qp_entry = search_qp(rc.local_qp_num, params_p->port[port_ix].qp_vect, 
                                      params_p->in_params.num_of_qps);

            if (comp_qp_entry == NULL) {
                printf("Failed to find QP number\n");
                return RET_ERR;
            }

            comp_qp = (int)(MT_ulong_ptr_t)(comp_qp_entry - params_p->port[port_ix].qp_vect);

            //printf("comp_qp = %d\n", comp_qp);

            qp_hndl = comp_qp_entry->qp_hndl;

            rr.id++;
            rr.sg_lst_p[0].lkey = params_p->recv_lkey_arr[m];
            if (params_p->in_params.srq_num > 0)
            {
                res = VAPI_post_srq(hca_hndl, comp_qp_entry->srq_hndl, 1, &rr, &srq_posted_actual);
                //printf("posted %d SRQs\n", srq_posted_actual);
            }
            else
            {
                res = VAPI_post_rr(hca_hndl, qp_hndl, &rr);
            }
            if (res != VAPI_OK)
            {
                printf("Error: Re-posting Recv buffer (port %d, iter %d) : %s\n", params_p->port[port_ix].port_num, i, VAPI_strerror(res));
                return(RET_ERR);
            }
            m = (m+1)%params_p->in_params.num_of_keys;

            /* send sync msg on every comp_inter and on last send*/
            if (params_p->in_params.transport != IB_TS_RC)
            {
                if ((i%comp_inter == 0) || (i == iter-1) ||
                    (qp_sync_sent_count[comp_qp] % comp_inter == 0) || 
                    (qp_sync_sent_count[comp_qp] == (iter/params_p->in_params.num_of_qps-1)))
                {
                    sync_sent++;
                    sr[0].comp_type = VAPI_UNSIGNALED;
                    if (sync_sent%comp_inter == 0)
                    {
                        sr[0].comp_type =  VAPI_SIGNALED;
                        qp_sync_sent_count[comp_qp]++;
                        sync_ask_for_comp++;
                    }
                    sr[0].id = i;
                    res = VAPI_post_sr(hca_hndl, qp_hndl, &sr[0]);
                    COND_PRINTF3((" **** Posting SYNC sr i=%d ask=%d sync_sent=%d\n",i+1,sync_ask_for_comp, sync_sent));
                    if (res != VAPI_OK)
                    {
                        printf("sg_list_len %d\n", sr[0].sg_lst_len );
                        printf("Error: Posting send desc : %s\n", VAPI_strerror(res));
                        return(RET_ERR);
                    }
                    if (sync_ask_for_comp > 1)
                    {
                        if (sr[0].comp_type == VAPI_SIGNALED)
                        {
                            if (RET_OK != PERF_poll_cq(hca_hndl,s_cq_hndl ,&sc,params_p->in_params.use_send_events, 
                                                       sq_sem_p))
                            {
                                return RET_ERR;
                            }
                        }
                    }
                }
            }
        }

        /* Send reply to sender after last message. Use 0 length message for RC and a short one for others !!! */
        COND_PRINTF2(("**** Send reply to sender after last message, pix %d \n", port_ix ));
        sr[0].comp_type = VAPI_SIGNALED;
        sr[0].id = iter;
        if (params_p->in_params.transport == IB_TS_RC)
        {
            sr[0].sg_lst_len = 0;
        } else
        {
            sr[0].sg_lst_len = 1;
            sg_entry_s[0].len = 4;
        }

        qp_hndl = params_p->port[port_ix].qp_vect[0].qp_hndl;
        
        res = VAPI_post_sr(hca_hndl, qp_hndl, &sr[0]);
        if (res != VAPI_OK)
        {
            printf("sg_list_len %d\n", sr[0].sg_lst_len );
            printf("Error: Posting send desc : %s\n", VAPI_strerror(res));
            return(RET_ERR);
        }

        /* Wait for final reply from receiver */
        if (RET_OK != PERF_poll_cq(hca_hndl,s_cq_hndl ,&sc,params_p->in_params.use_send_events, 
                                   sq_sem_p))
        {
            return RET_ERR;
        }
    }

    if (params_p->in_params.ports_num <= 1)
    {
        /* prevent receiver from exiting before the sender finished it's work */
        sock_sync_msg(params_p->in_params.socket, is_sender);
        COND_PRINTF1(("\n************* %s BW Unidirection Test Finished for port %d *********************\n\n",
                      tstr, params_p->port[port_ix].port_num));
    }
    return(RET_OK);
}


/***********************************************************************************************
* name:      rdma_w_bw_uni_test
* args:
*            params_p -- test parameter structure, need to filled before calling
* returns:
*            RET_OK
*            RET_ERR
*
* description:     
*            This is the main loop for the RDMA (RC) BW test. The test sends N packets via 1 port
*            then waits for acknowledge from the receiver, and calculate the BW.
*            
************************************************************************************************/  
ret_val_t rdma_w_bw_uni_test(test_params_tx *params_p, int port_ix)
{
    VAPI_ret_t             res;
    u_int64_t t_start=0, t_end=0;

    int i, j, k, l=0, comp_inter, ask_for_comp=0, ii;

    VAPI_rr_desc_t         rr;
    VAPI_sg_lst_entry_t    sg_entry_r;
    VAPI_sr_desc_t       * sr;
    VAPI_sg_lst_entry_t  * sg_entry_s;

    VAPI_sr_desc_t         rc_sr;
    VAPI_sg_lst_entry_t    rc_sg_entry_s;

    VAPI_wc_desc_t         rc;
    VAPI_wc_desc_t         sc;

    /* Copy all input parameters - don't want to rewrite the loop code */
    VAPI_hca_hndl_t        hca_hndl    = params_p->hca_hndl;
    VAPI_qp_hndl_t         qp_hndl     = params_p->port[port_ix].qp_vect[0].qp_hndl;
    VAPI_cq_hndl_t         s_cq_hndl   = params_p->port[port_ix].s_cq_hndl;
    VAPI_cq_hndl_t         r_cq_hndl   = params_p->port[port_ix].r_cq_hndl;
    MOSAL_semaphore_t    * sq_sem_p    = &params_p->port[port_ix].sq_sem;
    MOSAL_semaphore_t    * rq_sem_p    = &params_p->port[port_ix].rq_sem;
    u_int8_t             * recv_buf    = params_p->recv_buf;
    u_int8_t             * send_buf    = params_p->send_buf;
    int                    is_sender   = params_p->in_params.is_sender;
    u_int32_t            * qp_sent_count;
    int                    size        = params_p->in_params.size;
    int                    iter        = params_p->in_params.iter;      
    int                    list_len    = params_p->in_params.postlist_len;
    volatile char        * check_p;
    int                    index_to_check;
    call_result_t          mosal_rc;
    char *tstr = (params_p->in_params.transport == IB_TS_UC) ? "UC" : "RC";
    int use_inline = FALSE;
    u_int32_t  srq_posted_actual;
    //printf("Send CQ handle = 0x%x, Rcv CQ handle = 0x%x\n", (u_int32_t) s_cq_hndl, (u_int32_t) r_cq_hndl);

    int max_msg_size = MIN2(params_p->in_params.mr_recv_size,params_p->in_params.mr_send_size);
    int max_sends = MIN2(params_p->in_params.num_r_wqe,params_p->in_params.num_s_wqe); 
    //u_int64_t t_start_my=0, t_start_my_prev=0, delta_sum=0;
    //int my_iter_good;
    //float bad_average_double;
    //u_int64_t               *delta = NULL;

    if (params_p->in_params.transport == IB_TS_UD)
    {
        max_msg_size = 2048;
        if (size > 2048)
        {
            printf("Warning!!!: UD max size is 2048\n");
            size = 2048;
        }
    }

    if (size > params_p->in_params.mr_recv_size)
    {
        printf("Warning!!!: size=%d > mr_recv_size=%d --- changing size to %d\n", 
               size,params_p->in_params.mr_recv_size,params_p->in_params.mr_recv_size);
        size = params_p->in_params.mr_recv_size;
    }

    if (size==0)
    {
        printf("For RDMA size must be >= 1. Changing size to 1\n");
        size =1;
    }

    if (list_len > max_sends)
    {
        printf("Warning!!! Post List Length > WQ size --- reducing it to %d\n", max_sends);
        list_len = max_sends;
    }

    sr = VMALLOC(list_len*sizeof(VAPI_sr_desc_t));
    sg_entry_s = VMALLOC(params_p->in_params.num_of_keys*sizeof(VAPI_sg_lst_entry_t));

    /**************    RC RR *****************************************************/
    rr.opcode = VAPI_RECEIVE;
    rr.comp_type = VAPI_SIGNALED;
    rr.sg_lst_len = 1;
    rr.sg_lst_p  = &sg_entry_r;
    sg_entry_r.lkey = params_p->recv_lkey;
    sg_entry_r.len = params_p->in_params.mr_recv_size;
    sg_entry_r.addr = (VAPI_virt_addr_t)(MT_virt_addr_t)recv_buf;


    /**************    RDMA SR *****************************************************/
    for (i=0; i<params_p->in_params.num_of_keys; i++) {    
        sg_entry_s[i].len = size;
        sg_entry_s[i].lkey = params_p->send_lkey_arr[i];
        sg_entry_s[i].addr = (VAPI_virt_addr_t)(MT_virt_addr_t)send_buf;
    }

    for (i=0; i<list_len; i++)
    {
        sr[i].comp_type = VAPI_UNSIGNALED;
        sr[i].set_se = FALSE;

        sr[i].opcode = VAPI_RDMA_WRITE;
        sr[i].remote_qkey = 0;
        sr[i].remote_addr = params_p->remote_add;
        sr[i].id = 1;

        sr[i].sg_lst_len = 1; 
    }


    /**************    RC SR *****************************************************/
    rc_sr.comp_type = VAPI_SIGNALED;
    rc_sr.set_se = FALSE;
    rc_sr.opcode = VAPI_SEND;
    rc_sr.remote_qkey = 0;
    rc_sr.id = 1;
    rc_sr.sg_lst_len = 0;

    rc_sg_entry_s.len = size;
    rc_sg_entry_s.lkey = params_p->send_lkey;
    rc_sg_entry_s.addr = (VAPI_virt_addr_t)(MT_virt_addr_t)send_buf; 
    rc_sr.sg_lst_p = sg_entry_s;

    /* Both Sender  and receiver need 1 for last message*/
    rr.id = 0;
    if (params_p->in_params.srq_num > 0)
    {
        VAPI_qp_hndl_t         srq_hndl     = params_p->port[port_ix].qp_vect[0].srq_hndl;
        res = VAPI_post_srq(hca_hndl, srq_hndl, 1, &rr, &srq_posted_actual);
        //printf("posted %d SRQs\n", srq_posted_actual);
    }
    else
    {
        res = VAPI_post_rr(hca_hndl, qp_hndl, &rr);
    }
    if (res != VAPI_OK)
    {
        printf("Error: Posting Recv buffer : %s\n", VAPI_strerror(res));
        return(RET_ERR);
    }

    /* sync the test start */
    if (params_p->in_params.ports_num > 1)
    {
        COND_PRINTF2((" **** Test: finished the preparation phase for port %d and waiting for signal on %p ... \n", params_p->port[port_ix].port_num, &params_p->sync));
        MOSAL_sem_rel( &params_p->sem );
        mosal_rc = MOSAL_syncobj_waiton( &params_p->sync, MAX_WAIT_FOR_SYNC_USECS );
        if (mosal_rc != MT_OK)
        {
            printf( "Test hasn't received 'run' command from the main thread for port %d, exiting ...\n", params_p->port[port_ix].port_num);
            return(RET_ERR);
        }
    } else
    {
        sock_sync_msg(params_p->in_params.socket, is_sender);
    }
    if (params_p->in_params.ports_num <= 1)
    {
        COND_PRINTF1(("\n************* %s RDMA-Write BW Unidirectional Test started for port %d *********************\n\n", 
                      tstr, params_p->port[port_ix].port_num));
    }

    if (iter < max_sends)
    {
        comp_inter = max_sends;
    } else
    {
        comp_inter = (max_sends / 2 - list_len) | 1;
    }
    
    ask_for_comp = 0;

    index_to_check = 0;
    check_p = &recv_buf[index_to_check];

    if (is_sender)
    {
        //delta = malloc(20000);
        qp_sent_count = calloc(sizeof(u_int32_t), params_p->in_params.num_of_qps);
        if ((size >0 ) &&(size <= params_p->max_inline_size) && (list_len == 1))
        {
            use_inline = TRUE;
        }
        
        get_clk64(t_start);
        
        for (i=0,j=0; i<iter/list_len; i++)
        {
            //t_start_my_prev = t_start_my;
            //get_clk64(t_start_my);
            //if (i>0) delta[i-1] = t_start_my - t_start_my_prev;
            
            // printf("Changing Send QP to %d\n", params_p->port[port_ix].qp_vect[j].qp_num);
            qp_hndl = params_p->port[port_ix].qp_vect[j].qp_hndl;

            for (k=0;k<list_len;k++)
            {
                sr[k].id = list_len * i + k;
                //printf("id = %d, comp_inter = %d\n", (int)sr[k].id, comp_inter);
                /* Ask for completion every comp_inter */

                if ((sr[k].id % comp_inter == 0) || 
                    (sr[k].id == ((iter/list_len)*list_len + k - 1)) ||
                    ((qp_sent_count[j]+k) % comp_inter == 0) || 
                    (qp_sent_count[j]+k == (iter/params_p->in_params.num_of_qps-1)) ||
                    (list_len > 1))
                {
                    sr[k].comp_type = VAPI_SIGNALED;
                    ask_for_comp++;
                    COND_PRINTF2((" **** Posting sr signaled i=%d ask=%d\n",i+1,ask_for_comp));
                    //printf(" **** Posting sr signaled i=%d ask=%d\n",i+1,ask_for_comp);

                } else
                {
                    sr[k].comp_type = VAPI_UNSIGNALED;
                }

            }
                
            //if ((size >0 ) &&(size <= params_p->max_inline_size) && (list_len == 1))
            if (use_inline == TRUE)
            {
                sr[0].r_key = params_p->remote_rkey_arr[l];
                // printf("Remote R_Key = 0x%x\n", sr[0].r_key);
                sr[0].sg_lst_p = &sg_entry_s[l];
                res = EVAPI_post_inline_sr(hca_hndl, qp_hndl, sr);
            } else
            {
                for (ii=0; ii<list_len; ii++) {
                    sr[ii].r_key = params_p->remote_rkey_arr[l];
                    // printf("Remote R_Key = 0x%x\n", sr[ii].r_key);
                    sr[ii].sg_lst_p = &sg_entry_s[l];
                }
                res = EVAPI_post_sr_list(hca_hndl, qp_hndl, list_len, sr);
            }
            if (res != VAPI_OK)
            {
                printf("Error: Posting send desc : %s\n", VAPI_strerror(res));
                return(RET_ERR);
            }

            /* Wait for completion every comp_inter starting from 2*comp_inter
            * This applies only when comp_iter > iter */
            while (ask_for_comp > 0)
            {
                COND_PRINTF2(("**** Polling for sr comp i=%d ask=%d\n",i+1,ask_for_comp));
                if (RET_OK != PERF_poll_cq(hca_hndl,s_cq_hndl ,&sc,params_p->in_params.use_send_events, 
                                           sq_sem_p))
                {
                    printf("Perf_pll_cq failed for QP %d (%d), at sent_count %d\n",
                           j, params_p->port[port_ix].qp_vect[j].qp_num, qp_sent_count[j]);
                    // getchar();
                    return RET_ERR;
                }
                                    
                ask_for_comp--;
            }
        
            qp_sent_count[j] += list_len;
            j = (j+1)%params_p->in_params.num_of_qps;            
        
            l = (l+1)%params_p->in_params.num_of_keys;

        }//for i - send packets

        qp_hndl = params_p->port[port_ix].qp_vect[0].qp_hndl;
        
        /* Send an RC send 0 message indicating end of RDMA */
        COND_PRINTF2(("Finished RDMA posting, Sending RC-Send message\n"));
        rc_sr.comp_type = VAPI_UNSIGNALED;
        rc_sr.sg_lst_len = 1;
        rc_sg_entry_s.len = 1;
        res = VAPI_post_sr(hca_hndl, qp_hndl, &rc_sr);
        if (res != VAPI_OK)
        {
            printf("Error: Posting Send buffer : %s\n", VAPI_strerror(res));
            return(RET_ERR);
        }

        /* Wait for completion on last send */        
        while (ask_for_comp>0)
        {
            COND_PRINTF2(("**** Polling for last sr comp i=%d ask=%d\n",i,ask_for_comp));

            // printf("Still need to do %d completions\n", ask_for_comp);

            /* Poll for the last send completion */
            if (RET_OK != PERF_poll_cq(hca_hndl,s_cq_hndl ,&sc,params_p->in_params.use_send_events, 
                                       sq_sem_p))
            {
                printf("Perf_poll_cq failed for last send\n");
                return RET_ERR;
            }
            ask_for_comp--;
            COND_PRINTF2(("Polling for sr comp %d %d\n",i,ask_for_comp));
        }

        /* Wait for final reply from receiver */
        if (!params_p->in_params.oneside && 
            RET_OK != PERF_poll_cq(hca_hndl,r_cq_hndl ,&rc,params_p->in_params.use_recv_events, 
                                   rq_sem_p))
        {
            printf("Perf_pll_cq failed for final reply from receiver\n");
            return RET_ERR;
        }

        get_clk64(t_end);

        if (params_p->in_params.ports_num > 1)
        {
            params_p->port[port_ix].snd.t_start = t_start;
            params_p->port[port_ix].snd.t_end = t_end;
            params_p->port[port_ix].size = size;

        } else
            
            //if ((params_p->in_params.qpc_sets < ARBEL_QPC_SETS) ||
            //    (params_p->in_params.mpt_sets < ARBEL_MPT_SETS))
#if 0
                printf("BW: %.1f MBytes/sec [Num. of QPs: %d, Post List: %d, QPC Sets: %d, MPT Sets: %d, Keys: %d, size: %d bytes, iter: %d, total %.0f]\n", 
                      BW_MBSEC(t_end - t_start, (1.0 * size) * iter), 
                       params_p->in_params.num_of_qps, params_p->in_params.postlist_len, 
                       params_p->in_params.qpc_sets, params_p->in_params.mpt_sets, 
                       params_p->in_params.num_of_keys,
                       size, iter, (1.0 * size) * iter);
        //printf("time = %.2f\n", CLK_TO_USEC(t_end - t_start));
#else
            printf("BW: %.1f MBytes/sec [size: %d bytes, iter: %d, total %.0f]\n", 
                   BW_MBSEC(t_end - t_start, (1.0 * size) * iter), 
                   size, iter, (1.0 * size) * iter);
#endif            
            //bad_average_double = 1.1 * (CLK_TO_USEC(t_end - t_start) / iter);
            //my_iter_good = iter - 1;
            //printf("bad_avg_double = %.2f\n", bad_average_double);
            //for (i=0; i<iter-1; i++)
            //{
            //    if (CLK_TO_USEC(delta[i]) < bad_average_double)
            //    {
            //        delta_sum = delta_sum + delta[i];
            //        printf("adding delta = %.2f\n", CLK_TO_USEC(delta[i]));
            //    }
            //    else
            //    {
            //        printf("skipping delta = %.2f\n", CLK_TO_USEC(delta[i]));
            //        my_iter_good--;
            //    }
            //
            //}
            //printf("Filtered BW: %.1f MBytes/sec [size: %d bytes]\n", 
            //       BW_MBSEC(delta_sum, size * my_iter_good), 
            //       size);
            
            //else
            //    printf("BW: %.1f MBytes/sec [Num. of QPs: %d, Post List: %d, size: %d bytes, iter: %d, total %.0f]\n", 
            //           BW_MBSEC(t_end - t_start, (1.0 * size) * iter), 
            //           params_p->in_params.num_of_qps, params_p->in_params.postlist_len,
            //           size, iter, (1.0 * size) * iter);                    

        free(qp_sent_count);
        //free(delta);
/*
        MTPERF_REPORT_PRINTF(HHUL_post_inline_send_req);
        MTPERF_REPORT_PRINTF(HHUL_post_send_req);
        MTPERF_REPORT_PRINTF(THH_uar_sendq_dbell);
        MTPERF_REPORT_PRINTF(WQE_build_send);
        MTPERF_REPORT_PRINTF(SQ_WQE_copy);
        MTPERF_REPORT_PRINTF(MOSAL_spinlock_unlock);
        MTPERF_REPORT_PRINTF(sq_alloc_wqe_link_and_ring);
        MTPERF_REPORT_PRINTF(MOSAL_spinlock_irq_lock);
        MTPERF_REPORT_PRINTF(HHUL_poll4cqe);
*/
    }// Is sender
    else
    {
        get_clk64(t_start);

        /* Wait for the special Send from the sender */
        if (RET_OK != PERF_poll_cq(hca_hndl, r_cq_hndl ,&rc, 
                                   params_p->in_params.use_recv_events,  rq_sem_p))
        {
            printf("Perf_poll_cq failed for the special Send from the sender on RCV side\n");
            return RET_ERR;
        }

        get_clk64(t_end);
        
        //printf("Start at %lld, End at %lld, overall time = %lld\n", 
        //       (long long)CLK_TO_USEC(t_start), (long long)CLK_TO_USEC(t_end),
        //       (long long)(CLK_TO_USEC(t_end) - CLK_TO_USEC(t_start)));


        /* Send reply to sender after last message. Use 0 length message!!! */
        rc_sr.comp_type = VAPI_SIGNALED;
        rc_sr.sg_lst_len = 0;
        res = VAPI_post_sr(hca_hndl, qp_hndl, &rc_sr);
        if (res != VAPI_OK)
        {
            printf("Error: Posting send desc : %s\n", VAPI_strerror(res));
            return(RET_ERR);
        }
        COND_PRINTF2(("Receiver sent ack, waiting for completion\n"));
        /* Poll for the last send completion */
        if (RET_OK != PERF_poll_cq(hca_hndl, s_cq_hndl ,&sc, 
                                   params_p->in_params.use_send_events,  sq_sem_p))
        {
            printf("Perf_poll_cq failed for the last send completion on RCV side\n");
            return RET_ERR;
        }
    }

    free(sr);

    if (params_p->in_params.ports_num <= 1)
    {
        /* prevent receiver from exiting before the sender finished it's work */
        sock_sync_msg(params_p->in_params.socket, is_sender);
        COND_PRINTF1(("\n************* %s RDMA-Write BW Unidirectional Test Finished for port %d *********************\n\n",
                      tstr, params_p->port[port_ix].port_num));
    }
    return(RET_OK);
}


/***********************************************************************************************
* name:      rdma_w_bw_uni_many2one_test
* args:
*            params_p -- test parameter structure, need to filled before calling
* returns:
*            RET_OK
*            RET_ERR
*
* description:     
*            This is the main loop for the RDMA (RC) BW test. The test sends N packets via 1 port
*            then waits for acknowledge from the receiver, and calculate the BW.
*            
************************************************************************************************/  
ret_val_t rdma_w_bw_uni_many2one_test(test_params_tx *params_p, int port_ix)
{
    VAPI_ret_t             res;
    u_int64_t t_start=0, t_end=0;

    int i, j, l=0, m=0, comp_inter, ask_for_comp=0, ii, last_send=0, r_timestamp_num=0;

    VAPI_rr_desc_t         rr;
    VAPI_sg_lst_entry_t    sg_entry_r;
    VAPI_sr_desc_t       * sr;
    VAPI_sg_lst_entry_t  * sg_entry_s;

    VAPI_sr_desc_t         rc_sr;
    VAPI_sg_lst_entry_t    rc_sg_entry_s;

    VAPI_wc_desc_t         rc;
    VAPI_wc_desc_t         sc;

    /* Copy all input parameters - don't want to rewrite the loop code */
    VAPI_hca_hndl_t        hca_hndl    = params_p->hca_hndl;
    VAPI_qp_hndl_t         qp_hndl     = params_p->port[port_ix].qp_vect[0].qp_hndl;
    VAPI_cq_hndl_t         s_cq_hndl   = params_p->port[port_ix].s_cq_hndl;
    VAPI_cq_hndl_t         r_cq_hndl   = params_p->port[port_ix].r_cq_hndl;
    VAPI_srq_hndl_t        srq_hndl    = VAPI_INVAL_SRQ_HNDL;
    MOSAL_semaphore_t    * sq_sem_p    = &params_p->port[port_ix].sq_sem;
    MOSAL_semaphore_t    * rq_sem_p    = &params_p->port[port_ix].rq_sem;
    u_int8_t             * recv_buf    = params_p->recv_buf;
    u_int8_t             * send_buf    = params_p->send_buf;
    int                    is_sender   = params_p->in_params.is_sender;
    u_int32_t            * qp_sent_count;
    int                    size        = params_p->in_params.size;
    int                    iter        = params_p->in_params.iter;      
    int                    list_len    = params_p->in_params.postlist_len;
    volatile char        * check_p;
    int                    index_to_check;
    call_result_t          mosal_rc;
    char *tstr = (params_p->in_params.transport == IB_TS_UC) ? "UC" : "RC";
    int use_inline = FALSE;
    u_int32_t  srq_posted_actual;

    //printf("Send CQ handle = 0x%x, Rcv CQ handle = 0x%x\n", (u_int32_t) s_cq_hndl, (u_int32_t) r_cq_hndl);
    //printf("1\n");
    int max_msg_size = MIN2(params_p->in_params.mr_recv_size,params_p->in_params.mr_send_size);
    int max_sends = MIN2(params_p->in_params.num_r_wqe,params_p->in_params.num_s_wqe); 

    if (params_p->in_params.transport == IB_TS_UD)
    {
        max_msg_size = 2048;
        if (size > 2048)
        {
            printf("Warning!!!: UD max size is 2048\n");
            size = 2048;
        }
    }

    //printf("2\n");
    if (size > params_p->in_params.mr_recv_size)
    {
        printf("Warning!!!: size=%d > mr_recv_size=%d --- changing size to %d\n", 
               size,params_p->in_params.mr_recv_size,params_p->in_params.mr_recv_size);
        size = params_p->in_params.mr_recv_size;
    }

    if (size==0)
    {
        printf("For RDMA size must be >= 1. Changing size to 1\n");
        size =1;
    }

    if (list_len > max_sends)
    {
        printf("Warning!!! Post List Length > WQ size --- reducing it to %d\n", max_sends-1);
        list_len = max_sends-1;
    }

    //printf("3\n");
    sr = VMALLOC((list_len)*sizeof(VAPI_sr_desc_t));
    //printf("4\n");
    sg_entry_s = VMALLOC(params_p->in_params.num_of_keys*sizeof(VAPI_sg_lst_entry_t));
    //printf("5\n");

    /**************    RC RR *****************************************************/
    rr.opcode = VAPI_RECEIVE;
    rr.comp_type = VAPI_SIGNALED;
    rr.sg_lst_len = 1;
    rr.sg_lst_p  = &sg_entry_r;
    sg_entry_r.lkey = params_p->recv_lkey;
    sg_entry_r.len = params_p->in_params.mr_recv_size;
    sg_entry_r.addr = (VAPI_virt_addr_t)(MT_virt_addr_t)recv_buf;


    //printf("6\n");
    /**************    RDMA SR *****************************************************/
    for (i=0; i<params_p->in_params.num_of_keys; i++) {    
        sg_entry_s[i].len = size;
        sg_entry_s[i].lkey = params_p->send_lkey_arr[i];
        sg_entry_s[i].addr = (VAPI_virt_addr_t)(MT_virt_addr_t)send_buf;
    }

    //printf("7\n");
    for (i=0; i<list_len; i++)
    {
        sr[i].comp_type = VAPI_UNSIGNALED;
        sr[i].set_se = FALSE;

        sr[i].opcode = VAPI_RDMA_WRITE;
        sr[i].remote_qkey = 0;
        sr[i].remote_addr = params_p->remote_add;
        sr[i].id = 1;

        sr[i].sg_lst_len = 1; 
    }

    //printf("8\n");
    // Override the last message - special SEND with size 2
    //sr[list_len-1].comp_type = VAPI_SIGNALED;
    //sr[list_len-1].opcode = VAPI_SEND;
    //printf("9\n");
    //sr[list_len-1].sg_lst_p->len = 1;  Do it later because here sg_lst_p is not allocated yet

    /**************    RC SR *****************************************************/
    rc_sr.comp_type = VAPI_SIGNALED;
    rc_sr.set_se = FALSE;
    rc_sr.opcode = VAPI_SEND;
    rc_sr.remote_qkey = 0;
    rc_sr.id = 1;
    rc_sr.sg_lst_len = 0;

    rc_sg_entry_s.len = size;
    rc_sg_entry_s.lkey = params_p->send_lkey;
    rc_sg_entry_s.addr = (VAPI_virt_addr_t)(MT_virt_addr_t)send_buf; 
    rc_sr.sg_lst_p = sg_entry_s;

    /* Sender needs 1 , while receiver needs many */
    rr.id = 0;
    for (j=0;j<(int)params_p->in_params.num_of_qps;j++) 
    {    
        qp_hndl = params_p->port[port_ix].qp_vect[j].qp_hndl;
        srq_hndl = params_p->port[port_ix].qp_vect[j].srq_hndl;

        if (!is_sender)
        {
            for (i=0;i<max_sends;i++)
            {
                rr.id = i;
                rr.sg_lst_p[0].lkey = params_p->recv_lkey_arr[m];
                if (params_p->in_params.srq_num > 0)
                {
                    res = VAPI_post_srq(hca_hndl, srq_hndl, 1, &rr, &srq_posted_actual);
                    //printf("posted %d SRQs\n", srq_posted_actual);
                }
                else
                {
                    res = VAPI_post_rr(hca_hndl, qp_hndl, &rr);
                }
                if (res != VAPI_OK)
                {
                    printf("Error: Pre-posting Recv buffer (port %d) : %s\n", params_p->port[port_ix].port_num, VAPI_strerror(res));
                    return(RET_ERR);
                }
                m = (m+1)%params_p->in_params.num_of_keys;
            }
        } else
        {
            for (i=0; i<max_sends; i++)
            {
                rr.id = iter+i;
                rr.sg_lst_p[0].lkey = params_p->recv_lkey_arr[m];
                if (params_p->in_params.srq_num > 0)
                {
                    res = VAPI_post_srq(hca_hndl, srq_hndl, 1, &rr, &srq_posted_actual);
                    //printf("posted %d SRQs\n", srq_posted_actual);
                }
                else
                {
                    res = VAPI_post_rr(hca_hndl, qp_hndl, &rr);
                }
                if (res != VAPI_OK)
                {
                    printf("Error: Pre-posting Recv buffer (port %d): %s\n", params_p->port[port_ix].port_num, VAPI_strerror(res));
                    return(RET_ERR);
                }
                m = (m+1)%params_p->in_params.num_of_keys;
            }
        }
    }

        /* sync the test start */
    if (params_p->in_params.ports_num > 1)
    {
        COND_PRINTF2((" **** Test: finished the preparation phase for port %d and waiting for signal on %p ... \n", params_p->port[port_ix].port_num, &params_p->sync));
        MOSAL_sem_rel( &params_p->sem );
        mosal_rc = MOSAL_syncobj_waiton( &params_p->sync, MAX_WAIT_FOR_SYNC_USECS );
        if (mosal_rc != MT_OK)
        {
            printf( "Test hasn't received 'run' command from the main thread for port %d, exiting ...\n", params_p->port[port_ix].port_num);
            return(RET_ERR);
        }
    } else
    {
        sock_sync_msg(params_p->in_params.socket, is_sender);
    }
    if (params_p->in_params.ports_num <= 1)
    {
        COND_PRINTF1(("\n************* %s RDMA-Write BW Unidirectional Test started for port %d *********************\n\n", 
                      tstr, params_p->port[port_ix].port_num));
    }

    if (iter < max_sends)
    {
        comp_inter = max_sends;
    } else
    {
        comp_inter = (max_sends / 2 - list_len) | 1;
    }
    
    ask_for_comp = 0;

    index_to_check = 0;
    check_p = &recv_buf[index_to_check];

    if (is_sender)
    {
        qp_sent_count = calloc(sizeof(u_int32_t), params_p->in_params.num_of_qps);
        if ((size >0 ) &&(size <= params_p->max_inline_size) && (list_len == 1))
        {
            use_inline = TRUE;
        }
        
        get_clk64(t_start);
        
        for (i=0,j=0; i<iter/list_len; i++)
        {
            // printf("Changing Send QP to %d\n", params_p->port[port_ix].qp_vect[j].qp_num);
            qp_hndl = params_p->port[port_ix].qp_vect[j].qp_hndl;

            //for (k=0;k<list_len;k++)
            //{
            //    sr[k].id = list_len * i + k;
            //
            //    /* Ask for completion every comp_inter */
            //    
            //    if ((sr[k].id % comp_inter == 0) || 
            //        (sr[k].id == ((iter/list_len)*list_len + k - 1)) ||
            //        ((qp_sent_count[j]+k) % comp_inter == 0) || 
            //        (qp_sent_count[j]+k == (iter/params_p->in_params.num_of_qps-1)) ||
            //        (list_len > 1))
            //    {
            //        sr[k].comp_type = VAPI_SIGNALED;
            //        ask_for_comp++;
            //        COND_PRINTF2((" **** Posting sr signaled i=%d ask=%d\n",i+1,ask_for_comp));
            //    } else
            //    {
            //        sr[k].comp_type = VAPI_UNSIGNALED;
            //    }

            // } 
                
            //if ((size >0 ) &&(size <= params_p->max_inline_size) && (list_len == 1))
            if (use_inline == TRUE)
            {
                sr[0].r_key = params_p->remote_rkey_arr[l];
                // printf("Remote R_Key = 0x%x\n", sr[0].r_key);
                sr[0].sg_lst_p = &sg_entry_s[l];
                res = EVAPI_post_inline_sr(hca_hndl, qp_hndl, sr);
            } else
            {
                for (ii=0; ii<list_len; ii++) {
                    sr[ii].r_key = params_p->remote_rkey_arr[l];
                    // printf("Remote R_Key = 0x%x\n", sr[ii].r_key);
                    sr[ii].sg_lst_p = &sg_entry_s[l];
                }
                //sr[list_len-1].comp_type = VAPI_SIGNALED;
                //ask_for_comp++;
                //sr[list_len-1].opcode = VAPI_SEND;
                //sr[list_len-1].sg_lst_p->len = 1;
                //printf("POSTING sr list to Local QP %d , Remote QP %d\n",
                //       params_p->port[port_ix].qp_vect[j].qp_num,
                //       params_p->port[port_ix].qp_vect[j].d_qpnum);
                res = EVAPI_post_sr_list(hca_hndl, qp_hndl, list_len, sr);
                //printf("Done\n");
            }
            if (res != VAPI_OK)
            {
                printf("Error: Posting send desc : %s\n", VAPI_strerror(res));
                return(RET_ERR);
            }

            /* Send an RC send message after every post list */
            rc_sr.comp_type = VAPI_SIGNALED;
            rc_sr.sg_lst_len = 1;
            rc_sg_entry_s.len = 1;
            rc_sr.sg_lst_p->len = 1;
            rc_sr.opcode = VAPI_SEND;
            ask_for_comp++;
            res = VAPI_post_sr(hca_hndl, qp_hndl, &rc_sr);
            if (res != VAPI_OK)
            {
                printf("Error: Posting Send buffer : %s\n", VAPI_strerror(res));
                return(RET_ERR);
            }

            /* Wait for completion every comp_inter starting from 2*comp_inter
            * This applies only when comp_iter > iter */
            while (ask_for_comp > 0)
            {
                COND_PRINTF2(("**** Polling for sr comp i=%d ask=%d\n",i+1,ask_for_comp));
                if (RET_OK != PERF_poll_cq(hca_hndl,s_cq_hndl ,&sc,params_p->in_params.use_send_events, 
                                           sq_sem_p))
                {
                    printf("Perf_pll_cq failed for QP %d (%d), at sent_count %d\n",
                           j, params_p->port[port_ix].qp_vect[j].qp_num, qp_sent_count[j]);
                    // getchar();
                    return RET_ERR;
                }
                                    
                ask_for_comp--;
            }
        
            qp_sent_count[j] += list_len;
            j = (j+1)%params_p->in_params.num_of_qps;            
        
            l = (l+1)%params_p->in_params.num_of_keys;

        }//for i - send packets

        qp_hndl = params_p->port[port_ix].qp_vect[0].qp_hndl;
        
        /* Send an RC send 0 message indicating end of RDMA */
        COND_PRINTF2(("Finished RDMA posting, Sending RC-Send message\n"));
        rc_sr.comp_type = VAPI_UNSIGNALED;
        rc_sr.sg_lst_len = 1;
        rc_sg_entry_s.len = 2;
        rc_sr.sg_lst_p->len = 2;
        res = VAPI_post_sr(hca_hndl, qp_hndl, &rc_sr);
        //printf("sent final SEND\n");
        if (res != VAPI_OK)
        {
            printf("Error: Posting Send buffer : %s\n", VAPI_strerror(res));
            return(RET_ERR);
        }

        /* Wait for completion on last send */        
        while (ask_for_comp>0)
        {
            COND_PRINTF2(("**** Polling for last sr comp i=%d ask=%d\n",i,ask_for_comp));

            // printf("Still need to do %d completions\n", ask_for_comp);

            /* Poll for the last send completion */
            //printf("Polling ...\n");
            if (RET_OK != PERF_poll_cq(hca_hndl,s_cq_hndl ,&sc,params_p->in_params.use_send_events, 
                                       sq_sem_p))
            {
                printf("Perf_poll_cq failed for last send\n");
                return RET_ERR;
            }
            ask_for_comp--;
            COND_PRINTF2(("Polling for sr comp %d %d\n",i,ask_for_comp));
        }

        /* Wait for final reply from receiver */
        if (!params_p->in_params.oneside && 
            RET_OK != PERF_poll_cq(hca_hndl,r_cq_hndl ,&rc,params_p->in_params.use_recv_events, 
                                   rq_sem_p))
        {
            printf("Perf_pll_cq failed for final reply from receiver\n");
            return RET_ERR;
        }

        //printf("POLLED LAST CQ\n");
        get_clk64(t_end);

        if (params_p->in_params.ports_num > 1)
        {
            params_p->port[port_ix].snd.t_start = t_start;
            params_p->port[port_ix].snd.t_end = t_end;
            params_p->port[port_ix].size = size;

        } else
            
            //if ((params_p->in_params.qpc_sets < ARBEL_QPC_SETS) ||
            //    (params_p->in_params.mpt_sets < ARBEL_MPT_SETS))
#if 0
                printf("BW: %.1f MBytes/sec [Num. of QPs: %d, Post List: %d, QPC Sets: %d, MPT Sets: %d, Keys: %d, size: %d bytes, iter: %d, total %.0f]\n", 
                      BW_MBSEC(t_end - t_start, (1.0 * size) * iter), 
                       params_p->in_params.num_of_qps, params_p->in_params.postlist_len, 
                       params_p->in_params.qpc_sets, params_p->in_params.mpt_sets, 
                       params_p->in_params.num_of_keys,
                       size, iter, (1.0 * size) * iter);
#else
            printf("BW: %.1f MBytes/sec [size: %d bytes, iter: %d, total %.0f]\n", 
                   BW_MBSEC(t_end - t_start, (1.0 * size) * iter), 
                   size, iter, (1.0 * size) * iter);
#endif            
            //else
            //    printf("BW: %.1f MBytes/sec [Num. of QPs: %d, Post List: %d, size: %d bytes, iter: %d, total %.0f]\n", 
            //           BW_MBSEC(t_end - t_start, (1.0 * size) * iter), 
            //           params_p->in_params.num_of_qps, params_p->in_params.postlist_len,
            //           size, iter, (1.0 * size) * iter);                    

        free(qp_sent_count);
/*
        MTPERF_REPORT_PRINTF(HHUL_post_inline_send_req);
        MTPERF_REPORT_PRINTF(HHUL_post_send_req);
        MTPERF_REPORT_PRINTF(THH_uar_sendq_dbell);
        MTPERF_REPORT_PRINTF(WQE_build_send);
        MTPERF_REPORT_PRINTF(SQ_WQE_copy);
        MTPERF_REPORT_PRINTF(MOSAL_spinlock_unlock);
        MTPERF_REPORT_PRINTF(sq_alloc_wqe_link_and_ring);
        MTPERF_REPORT_PRINTF(MOSAL_spinlock_irq_lock);
        MTPERF_REPORT_PRINTF(HHUL_poll4cqe);
*/
    }// Is sender
    else
    {
        //get_clk64(t_start);

        /* Wait for the special Send from the sender */
        while (!last_send)
        {
            for (j=0;j<(int)params_p->in_params.num_of_qps;j++) 
            {
            //printf("Polling ...\n");
            rc.local_qp_num = params_p->port[port_ix].qp_vect[j].qp_num;
            if (RET_OK != PERF_poll_cq(hca_hndl, r_cq_hndl ,&rc, 
                                       params_p->in_params.use_recv_events,  rq_sem_p))
            {
                printf("Perf_poll_cq failed for the special Send from the sender on RCV side\n");
                return RET_ERR;
            }

            get_clk64(t_end);

            //printf("Polled CQ. # %d  len = %d\n", r_timestamp_num, rc.byte_len);

            if (rc.byte_len != 2)
            {
         //       printf("Timestamp # %d at %lld from QP # %d\n", 
         //              r_timestamp_num, (long long)CLK_TO_USEC(t_end), rc.local_qp_num);
                printf("Timestamp # %d at %lld\n", 
                       r_timestamp_num, (long long)CLK_TO_USEC(t_end));
                r_timestamp_num++;
                rr.id++;
                qp_hndl = params_p->port[port_ix].qp_vect[j].qp_hndl;
                if (params_p->in_params.srq_num > 0)
                {
                    srq_hndl = params_p->port[port_ix].qp_vect[j].srq_hndl;
                    res = VAPI_post_srq(hca_hndl, srq_hndl, 1, &rr, &srq_posted_actual);
                    //printf("posted %d SRQs\n", srq_posted_actual);
                }
                else
                {
                    res = VAPI_post_rr(hca_hndl, qp_hndl, &rr);
                }
                if (res != VAPI_OK)
                {
                    printf("Error: Posting Recv buffer : %s\n", VAPI_strerror(res));
                    return(RET_ERR);
                }
            }
            else
            {
                printf("Final Special SEND arrived at %lld\n", 
                       (long long)CLK_TO_USEC(t_end));

                last_send = 1;
                break;
            }
            }
        }

        //printf("Posting Last reply\n");
        /* Send reply to sender after last message. Use 0 length message!!! */
        rc_sr.comp_type = VAPI_SIGNALED;
        rc_sr.sg_lst_len = 0;
        res = VAPI_post_sr(hca_hndl, qp_hndl, &rc_sr);
        if (res != VAPI_OK)
        {
            printf("Error: Posting send desc : %s\n", VAPI_strerror(res));
            return(RET_ERR);
        }
        COND_PRINTF2(("Receiver sent ack, waiting for completion\n"));
        /* Poll for the last send completion */
        if (RET_OK != PERF_poll_cq(hca_hndl, s_cq_hndl ,&sc, 
                                   params_p->in_params.use_send_events,  sq_sem_p))
        {
            printf("Perf_poll_cq failed for the last send completion on RCV side\n");
            return RET_ERR;
        }
    }

    free(sr);

    if (params_p->in_params.ports_num <= 1)
    {
        /* prevent receiver from exiting before the sender finished it's work */
        sock_sync_msg(params_p->in_params.socket, is_sender);
        COND_PRINTF1(("\n************* %s RDMA-Write BW Unidirectional Test Finished for port %d *********************\n\n",
                      tstr, params_p->port[port_ix].port_num));
    }
    return(RET_OK);
}


/***********************************************************************************************
* name:      rdma_r_bw_uni_test
* args:
*            params_p -- test parameter structure, need to filled before calling
* returns:
*            RET_OK
*            RET_ERR
*
* description:     
*            This is the main loop for the RDMA (RC) BW test. The test sends N packets via 1 port
*            then waits for acknowledge from the receiver, and calculate the BW.
*            
************************************************************************************************/  
ret_val_t rdma_r_bw_uni_test(test_params_tx *params_p, int port_ix)
{
    VAPI_ret_t             res;
    u_int64_t t_start=0, t_end=0;

    int i,comp_inter, ask_for_comp=0;
    int j=0;
    VAPI_sr_desc_t         sr;
    VAPI_sg_lst_entry_t    sg_entry_s;
    VAPI_wc_desc_t         sc;

    /* Copy all input parameters - don't want to rewrite the loop code */
    VAPI_hca_hndl_t        hca_hndl    = params_p->hca_hndl;
    VAPI_qp_hndl_t         qp_hndl     = params_p->port[port_ix].qp_vect[0].qp_hndl;
    VAPI_cq_hndl_t         s_cq_hndl   = params_p->port[port_ix].s_cq_hndl;
    MOSAL_semaphore_t    * sq_sem_p   = &params_p->port[port_ix].sq_sem;
    u_int8_t               *send_buf   = params_p->send_buf;
    int                    is_sender   = params_p->in_params.is_sender;
    u_int32_t            * qp_sent_count;
    int                    size        = params_p->in_params.size;
    int                    iter        = params_p->in_params.iter;      
    call_result_t       mosal_rc;

    int max_msg_size = MIN2(params_p->in_params.mr_recv_size,params_p->in_params.mr_send_size);
    int max_sends = MIN2(params_p->in_params.num_r_wqe,params_p->in_params.num_s_wqe); 

    if (params_p->in_params.transport == IB_TS_UD)
    {
        max_msg_size = 2048;
        if (size > 2048)
        {
            printf("Warning!!!: UD max size is 2048\n");
            size = 2048;
        }
    }

    if (size > params_p->in_params.mr_recv_size)
    {
        printf("Warning!!!: size=%d > mr_recv_size=%d --- changing iter to %d\n", 
               size,params_p->in_params.mr_recv_size,params_p->in_params.mr_recv_size);
        size = params_p->in_params.mr_recv_size;
    }

    if (size==0)
    {
        printf("For RDMA size must be >= 1. Changing size to 1\n");
        size =1;
    }


    /**************    RDMA SR *****************************************************/
    sr.comp_type = VAPI_SIGNALED;
    sr.set_se = FALSE;

    sr.opcode = VAPI_RDMA_READ;
    sr.remote_qkey = 0;
    sr.remote_addr = params_p->remote_add;
    sr.r_key = params_p->remote_rkey;
    sr.id = 1;
    sr.fence = 0;

    sr.sg_lst_len = 1;
    sg_entry_s.len = size;
    sg_entry_s.lkey = params_p->send_lkey;
    sg_entry_s.addr = (VAPI_virt_addr_t)(MT_virt_addr_t)send_buf; 
    sr.sg_lst_p = &sg_entry_s;

    /* sync the test start */
    if (params_p->in_params.ports_num > 1)
    {
        COND_PRINTF2((" **** Test: finished the preparation phase for port %d and waiting for signal on %p ... \n", params_p->port[port_ix].port_num, &params_p->sync));
        MOSAL_sem_rel( &params_p->sem );
        mosal_rc = MOSAL_syncobj_waiton( &params_p->sync, MAX_WAIT_FOR_SYNC_USECS );
        if (mosal_rc != MT_OK)
        {
            printf( "Test hasn't received 'run' command from the main thread for port %d, exiting ...\n", params_p->port[port_ix].port_num);
            return(RET_ERR);
        }
    } else
    {
        sock_sync_msg(params_p->in_params.socket, is_sender);
    }
    if (params_p->in_params.ports_num <= 1)
    {
        COND_PRINTF1(("\n************* RDMA-Read BW Unidirectional Test started for port %d *********************\n\n", params_p->port[port_ix].port_num));
    }


    if (iter < max_sends)
    {
        comp_inter = max_sends;
    } else
    {
        comp_inter = max_sends / 2 -2;
    }

    ask_for_comp = 0;

    if (is_sender)
    {
        qp_sent_count = calloc(sizeof(u_int32_t), params_p->in_params.num_of_qps);
        get_clk64(t_start);
        for (i=0; i< iter; i++)
        {
            qp_hndl = params_p->port[port_ix].qp_vect[j].qp_hndl;
            sr.id = i;
            sr.comp_type = VAPI_UNSIGNALED; 

            /* Ask for completion every comp_inter */
            //if (((i+1)%comp_inter == 0) || i == (iter-1))
            if ((sr.id % comp_inter == 0) || 
                (sr.id == iter - 1) ||
                ((qp_sent_count[j]) % comp_inter == 0) || 
                (qp_sent_count[j] == (iter/params_p->in_params.num_of_qps-1)))
            {
                sr.comp_type = VAPI_SIGNALED;
                ask_for_comp++;
                COND_PRINTF2((" **** Posting sr signaled i=%d ask=%d\n",i+1,ask_for_comp));
            }

            res = VAPI_post_sr(hca_hndl, qp_hndl, &sr);
            if (res != VAPI_OK)
            {
                printf("Error: Posting send desc : %s\n", VAPI_strerror(res));
                return(RET_ERR);
            }


            /* Wait for completion every comp_inter starting from 2*comp_inter
             * This applies only when comp_iter > iter */
            if (ask_for_comp > 1)
            {
                COND_PRINTF2(("**** Polling for sr comp i=%d ask=%d\n",i+1,ask_for_comp));
                if (RET_OK != PERF_poll_cq(hca_hndl,s_cq_hndl ,&sc,params_p->in_params.use_send_events, 
                                           sq_sem_p))
                {
                    return RET_ERR;
                }
                ask_for_comp--;
                COND_PRINTF3(("**** Got a completion of sr %d i=%d ask=%d\n", (int)sc.id, i,ask_for_comp));
            }
            qp_sent_count[j] += 1;
            j = (j+1)%params_p->in_params.num_of_qps;            
        }//for i - send packets

        /* Wait for completion on last send */
        COND_PRINTF2(("**** Start polling for the completion of the last sr ask=%d\n",ask_for_comp));
        while (ask_for_comp>0)
        {

            /* Poll for the last send completion */
            if (RET_OK != PERF_poll_cq(hca_hndl,s_cq_hndl ,&sc,params_p->in_params.use_send_events, 
                                       sq_sem_p))
            {
                return RET_ERR;
            }
            ask_for_comp--;
            COND_PRINTF3(("**** Got a completion of sr %d, ask=%d\n", (int)sc.id, ask_for_comp));
        }
        COND_PRINTF2(("Finished polling for the completion if last sr %d\n",ask_for_comp));

        get_clk64(t_end);

        if (params_p->in_params.ports_num > 1)
        {
            params_p->port[port_ix].snd.t_start = t_start;
            params_p->port[port_ix].snd.t_end = t_end;
            params_p->port[port_ix].size = size;
        } else
        {

#if 0
                printf("BW: %.1f MBytes/sec [Num. of QPs: %d, Post List: %d, QPC Sets: %d, MPT Sets: %d, Keys: %d, size: %d bytes, iter: %d, total %.0f]\n", 
                      BW_MBSEC(t_end - t_start, (1.0 * size) * iter), 
                       params_p->in_params.num_of_qps, params_p->in_params.postlist_len, 
                       params_p->in_params.qpc_sets, params_p->in_params.mpt_sets, 
                       params_p->in_params.num_of_keys,
                       size, iter, (1.0 * size) * iter);
#else
            printf("BW: %.1f MBytes/sec [size: %d bytes, iter: %d, total %.0f]\n", 
                   BW_MBSEC(t_end - t_start, (1.0 * size) * iter), 
                   size, iter, (1.0 * size) * iter);
#endif            
        }
    }// Is sender

    if (params_p->in_params.ports_num <= 1)
    {
        /* prevent receiver from exiting before the sender finished it's work */
        sock_sync_msg(params_p->in_params.socket, is_sender);
        COND_PRINTF1(("\n************* RDMA-Read BW Unidirectional Test Finished for port %d *********************\n\n", params_p->port[port_ix].port_num));
    }
    return(RET_OK);
}

/***********************************************************************************************
* name:      send_bw_bi_test
* args:
*            params_p -- test parameter structure, need to filled before calling
* returns:
*            RET_OK
*            RET_ERR
*
* description:     
*            This is the main loop for the UD and RC send BW test. The test sends N packets
*            then waits for acknowledge from the receiver, and calculate the BW.
*            
************************************************************************************************/  
typedef struct
{
    u_int64_t t_exec;
} test_result_t;

ret_val_t send_bw_bi_test(test_params_tx *params_p, int port_ix)
{
    VAPI_ret_t             res;
    u_int64_t t_start=0, t_end=0;

    u_int32_t  srq_posted_actual;
    int i,comp_inter,ask_for_comp=0, tmp_cnt=1;
    int sync_received=0, sync_ask_for_comp=0, sync_sent=0;

    VAPI_rr_desc_t         rr;
    VAPI_sg_lst_entry_t    sg_entry_r;
    VAPI_sr_desc_t         sr, sr_sync;
    VAPI_sg_lst_entry_t    sg_entry_s;
    VAPI_wc_desc_t         sc;
    /* multiple time calculations */
#define MIN_INT_LOOPS		10
#define LOOPS_DELTA		2
#define MAX_INT_LOOPS	(MIN_INT_LOOPS+LOOPS_DELTA*20)
    u_int64_t t_exec, t_total, t_min, t_max=0, t_begin;
    int  iter_loops, iter_cur, ix_cur=0, threshold, j;      
    test_result_t results[MAX_INT_LOOPS];

    /* Copy all input parameters - don't want to rewrite the loop code */
    VAPI_hca_hndl_t        hca_hndl    = params_p->hca_hndl;
    VAPI_qp_hndl_t         qp_hndl0     = params_p->port[port_ix].qp_vect[0].qp_hndl;
    VAPI_cq_hndl_t         s_cq_hndl   = params_p->port[port_ix].s_cq_hndl;
    VAPI_srq_hndl_t        srq_hndl    = params_p->port[port_ix].qp_vect[0].srq_hndl;
    MOSAL_semaphore_t    * sq_sem_p   = &params_p->port[port_ix].sq_sem;
    u_int8_t               *recv_buf   = params_p->recv_buf;
    u_int8_t               *send_buf   = params_p->send_buf;
    int                    is_sender   = params_p->in_params.is_sender;
    int                    size        = params_p->in_params.size;
    int                    iter        = params_p->in_params.iter;      
    call_result_t       mosal_rc;
    int         sent=0, received = 0;
    volatile char          *check_p;
    int index_to_check;
    char *tstr = (params_p->in_params.transport == IB_TS_UD) ? "UD" : ((params_p->in_params.transport == IB_TS_UC) ? "UC" : "RC");
    u_int32_t tx_overhead = (params_p->in_params.transport == IB_TS_UD) ? 40 : 0;
    int use_inline = FALSE;

    //int                    skip        = params_p->in_params.iter;      

    int max_msg_size = MIN2(params_p->in_params.mr_recv_size,params_p->in_params.mr_send_size);
    int max_sends = MIN2(params_p->in_params.num_r_wqe,params_p->in_params.num_s_wqe); 

    if (params_p->in_params.transport == IB_TS_UD)
    {
        max_msg_size = 2048;
        if (size > 2048)
        {
            printf("Warning!!!: UD max size is 2048\n");
            size = 2048;
        }
    }

    if (size > params_p->in_params.mr_recv_size)
    {
        printf("Warning!!!: size=%d > mr_recv_size=%d --- changing iter to %d\n", 
               size,params_p->in_params.mr_recv_size,params_p->in_params.mr_recv_size);
        size = params_p->in_params.mr_recv_size;
    }

    /* multiple time calculaion */
#define MIN_TEST_SIZE_IN_LOOPS		1000	
#define MIN_SIZE_FOR_MULTIPLE_TIMINGS		13
    /* calculate default value of iter_loops - only for sizes <= 2 ** MIN_SIZE_FOR_MULTIPLE_TIMINGS */
    iter_loops = MIN_INT_LOOPS;
    for (i=MIN_SIZE_FOR_MULTIPLE_TIMINGS; i>8;i--)
    {
        if (size > (1<<i))
            iter_loops += LOOPS_DELTA;
    }
    if (size > (1<<MIN_SIZE_FOR_MULTIPLE_TIMINGS))
    {
        iter_loops = 1;
    }

    /* take iter_loops from parameter, if coded */
    if (params_p->in_params.general_param != -1)
    {
        iter_loops = params_p->in_params.general_param;
        if (iter_loops > MAX_INT_LOOPS - 2)
            iter_loops = MAX_INT_LOOPS - 2;
    }

    /* change the number of iterations if it is too small */
    if ( iter < (MIN_TEST_SIZE_IN_LOOPS * iter_loops ))
    {
        iter_cur = MIN_TEST_SIZE_IN_LOOPS;
    } else
    {
        iter_cur = iter / iter_loops;
    }
    iter = iter_cur * iter_loops;
    if (iter != params_p->in_params.iter)
    {
        COND_PRINTF2(("The number of iterations was changed to %d \n",iter));
    }

    t_min = MAKE_ULONGLONG(0x0ffffffffff);
    threshold = (iter_cur<<1) + 2;
    rr.opcode = VAPI_RECEIVE;
    rr.comp_type = VAPI_SIGNALED;
    rr.sg_lst_len = 1;
    rr.sg_lst_p  = &sg_entry_r;
    sg_entry_r.lkey = params_p->recv_lkey;
    sg_entry_r.len = params_p->in_params.mr_recv_size;
    sg_entry_r.addr = (VAPI_virt_addr_t)(MT_virt_addr_t)recv_buf;

    /***************************************************************************/
    sr.opcode = VAPI_SEND;
    sr.comp_type = VAPI_SIGNALED;
    sr.set_se = FALSE;
    sr.remote_qkey = 0;

    /* Needed only for UD */
    if (params_p->in_params.multicast > 0)
    {
        sr.remote_qp = IB_MULTICAST_QP;
    }
    else
    {
        sr.remote_qp = params_p->port[port_ix].qp_vect[0].d_qpnum;
    }
    sr.remote_ah = params_p->port[port_ix].av_hndl;

    sr.sg_lst_len = 1;
    sg_entry_s.addr = (VAPI_virt_addr_t)(MT_virt_addr_t)send_buf;
    sg_entry_s.len = size;
    sg_entry_s.lkey = params_p->send_lkey;
    sr.sg_lst_p = &sg_entry_s;

    /***************************************************************************/
    /* Needed only for UD and UC */
    sr_sync.opcode = VAPI_SEND;
    sr_sync.comp_type = VAPI_UNSIGNALED;
    sr_sync.set_se = FALSE;
    sr_sync.remote_qkey = 0;

    if (params_p->in_params.multicast > 0)
    {
        sr_sync.remote_qp = IB_MULTICAST_QP;
    }
    else
    {
        sr_sync.remote_qp = params_p->port[port_ix].qp_vect[0].d_qpnum;
    }
    sr_sync.remote_ah = params_p->port[port_ix].av_hndl;

    sr_sync.sg_lst_len = 0;
    sr_sync.sg_lst_p = &sg_entry_s;

    /***************************************************************************/

    /* Post all possible receive in advance. Sender need 1 while receiver need N*/
    if (!is_sender)
    {
        for (i=iter;i<max_sends+iter;i++)
        {
            rr.id = i;
            if (params_p->in_params.srq_num > 0)
            {
                res = VAPI_post_srq(hca_hndl, srq_hndl, 1, &rr, &srq_posted_actual);
                //printf("posted %d SRQs\n", srq_posted_actual);
            }
            else
            {
                res = VAPI_post_rr(hca_hndl, qp_hndl0, &rr);
            }
            if (res != VAPI_OK)
            {
                printf("Error: Posting Recv buffer : %s\n", VAPI_strerror(res));
                return(RET_ERR);
            }
        }
    } else
    {
        for (i=iter;i<max_sends+iter;i++)
        {
            rr.id = i;
            if (params_p->in_params.srq_num > 0)
            {
                res = VAPI_post_srq(hca_hndl, srq_hndl, 1, &rr, &srq_posted_actual);
                //printf("posted %d SRQs\n", srq_posted_actual);
            }
            else
            {
                res = VAPI_post_rr(hca_hndl, qp_hndl0, &rr);
            }
            if (res != VAPI_OK)
            {
                printf("Error: Posting Recv buffer : %s\n", VAPI_strerror(res));
                return(RET_ERR);
            }
        }
    }


    /* sync the test start */
    if (params_p->in_params.ports_num > 1)
    {
        COND_PRINTF2((" **** Test: finished the preparation phase for port %d and waiting for signal on %p ... \n", params_p->port[port_ix].port_num, &params_p->sync));
        MOSAL_sem_rel( &params_p->sem );
        mosal_rc = MOSAL_syncobj_waiton( &params_p->sync, MAX_WAIT_FOR_SYNC_USECS );
        if (mosal_rc != MT_OK)
        {
            printf( "Test hasn't received 'run' command from the main thread for port %d, exiting ...\n", params_p->port[port_ix].port_num);
            return(RET_ERR);
        }
    } else
    {
        sock_sync_msg(params_p->in_params.socket, is_sender);
    }
    if (params_p->in_params.ports_num <= 1)
    {
        COND_PRINTF1(("\n\n************* %s BW Bidirection Test started for port %d  *********************\n\n",
                      tstr, params_p->port[port_ix].port_num));
    }

    if (iter < max_sends)
    {
        comp_inter = max_sends; // Ask completion ONLY on last message!!
    } else
    {
        comp_inter = max_sends/2 - 2;
    }
    ask_for_comp = 0;

    index_to_check = port_ix + tx_overhead;
    check_p = &recv_buf[index_to_check];

    if ((size >0 ) &&(size <= params_p->max_inline_size))
    {
        use_inline = TRUE;
    }

    if (is_sender)
    {
        COND_PRINTF2(("Checking data at %p (current=%d)\n",check_p,*check_p));
        while (*check_p==84)
        {
        }
        COND_PRINTF2((".................Got data %d at %p\n",*check_p,check_p));
        get_clk64(t_start);
        t_begin = t_start;
        for (i=0; i< iter; i++, tmp_cnt++)
        //for (i=0; i< iter; i++)
        {
            sr.id = i;
            sr.comp_type = VAPI_UNSIGNALED; 

            /* Ask for completion every comp_inter and on last send*/
            //if (((i+1)%comp_inter == 0) || i == (iter-1))
            if ((tmp_cnt == comp_inter) || i == (iter-1))
            {
                tmp_cnt = 0;
                sr.comp_type = VAPI_SIGNALED;
                ask_for_comp++;
                if (params_p->in_params.transport != IB_TS_RC) sync_ask_for_comp++;
                COND_PRINTF3((" **** Posting sr %d signaled i=%d ask=%d\n",(int)sr.id , i,ask_for_comp));
            } else
            {
                COND_PRINTF3((" **** Posting sr %d unsignaled i=%d ask=%d\n",(int)sr.id ,i,ask_for_comp));
            }

            /* Use Inline data up to the limit size */
            //if ((size >0 ) &&(size <= params_p->max_inline_size))
            if (use_inline == TRUE)
            {
                res = EVAPI_post_inline_sr(hca_hndl, qp_hndl0, &sr);
            } else
            {
                res = VAPI_post_sr(hca_hndl, qp_hndl0, &sr);
            }
            sent++;
            if (res != VAPI_OK)
            {
                printf("sg_list_len %d\n", sr.sg_lst_len );
                printf("Error: Posting send desc : %s\n", VAPI_strerror(res));
                return(RET_ERR);
            }

            /* Wait for completion every comp_inter starting from 2*comp_inter
             * This applies only when comp_iter < iter */
            COND_PRINTF3(("**** Start polling for 1 sr and N rr completions, i=%d ask=%d\n",i,ask_for_comp));
            while (ask_for_comp > 1 || sync_ask_for_comp > 1)
            {
                res = PERF_poll_cq_nowait(hca_hndl, s_cq_hndl ,&sc,
                                          params_p->in_params.use_recv_events,  sq_sem_p);
                if (res == VAPI_EGEN) return RET_ERR;
                if (res == VAPI_CQ_EMPTY) continue;
                if (sc. id < iter)
                { /* it's - send completion */
                    ask_for_comp--;
                    COND_PRINTF3(("**** Got a completion of sr %d i=%d ask=%d\n", (int)sc.id, i,ask_for_comp));
                } else
                { /* it's - receive completion */
                    if ((params_p->in_params.transport != IB_TS_RC) && (sc.byte_len == tx_overhead))
                    {
                        sync_ask_for_comp--;
                        sync_received++;
                        COND_PRINTF3(("**** Got a completion of SYNC rr %d i=%d packets sync_received=%d\n", 
                                      (int)sc.id,i,sync_received));
                    } else
                    {
                        received++;
                        COND_PRINTF3(("**** Got a completion of rr %d i=%d packets received=%d size %u\n", 
                                      (int)sc.id,i,received, sc.byte_len));
                        COND_PRINTF3((".................Got data %d at %p\n",*check_p,check_p));
                        if ((params_p->in_params.transport != IB_TS_RC) && (received%comp_inter == 0 || received >= iter))
                        {
                            sr_sync.id = 2*iter + i;
                            res = VAPI_post_sr(hca_hndl, qp_hndl0, &sr_sync);
                            sync_sent++;
                            if (res != VAPI_OK)
                            {
                                printf("sg_list_len %d\n", sr_sync.sg_lst_len );
                                printf("Error: Posting SYNC sr (sent %d) : %s\n", sync_sent, VAPI_strerror(res));
                                return(RET_ERR);
                            }
                            COND_PRINTF3((" **** Posting SYNC sr %d unsignaled i=%d ask=%d sent=%d\n",(int)sr_sync.id ,i,ask_for_comp, sync_sent));
                        }
                    }
                    rr.id++;
                    if (params_p->in_params.srq_num > 0)
                    {
                        res = VAPI_post_srq(hca_hndl, srq_hndl, 1, &rr, &srq_posted_actual);
                        //printf("posted %d SRQs\n", srq_posted_actual);
                    }
                    else
                    {
                        res = VAPI_post_rr(hca_hndl, qp_hndl0, &rr);
                    }
                    if (res != VAPI_OK)
                    {
                        printf("Error: Posting Recv buffer : %s\n", VAPI_strerror(res));
                        return(RET_ERR);
                    }
                }
                if (sent+received > threshold)
                {
                    /* calculate time */
                    get_clk64(t_end);
                    results[ix_cur].t_exec = t_end - t_start;
                    threshold += iter_cur<<1;
                    ix_cur++;
                    get_clk64(t_start);
                }
            }
            COND_PRINTF3(("Finished polling for 1 sr and N rr completions, i=%d ask=%d\n",i,ask_for_comp));
        }//for i - send packets

        /* Wait for completion on last send while getting also receive completions */
        COND_PRINTF2(("**** Start polling for N rr's and the last sr completions, ask=%d sync_ask=%d\n",
                      ask_for_comp, sync_ask_for_comp));
        while (ask_for_comp>0 || received <= iter || sync_ask_for_comp > 0)
        {
            res = PERF_poll_cq_nowait(hca_hndl, s_cq_hndl ,&sc,
                                      params_p->in_params.use_recv_events,  sq_sem_p);
            if (res == VAPI_EGEN) return RET_ERR;
            if (res == VAPI_CQ_EMPTY) continue;
            if (sc.id < iter)
            { /* it's - send completion */
                ask_for_comp--;
                COND_PRINTF3(("**** Got a completion of  last sr %d comp ask=%d\n", (int)sc.id,ask_for_comp));
                continue;
            } else
            { /* it's - receive completion */
                if ((params_p->in_params.transport != IB_TS_RC) && (sc.byte_len == tx_overhead))
                {
                    sync_ask_for_comp--;
                    sync_received++;
                    COND_PRINTF3(("**** Got a completion of SYNC rr %d i=%d packets sync_received=%d\n", 
                                  (int)sc.id,i,sync_received));
                } else
                {
                    received++;
                    COND_PRINTF3(("**** Got a completion of rr %d i=%d packets received=%d size %u\n", 
                                  (int)sc.id,i,received, sc.byte_len));
                    if ((params_p->in_params.transport != IB_TS_RC) && (received%comp_inter == 0 || received >= iter))
                    {
                        sr_sync.id = 2*iter + i;
                        res = VAPI_post_sr(hca_hndl, qp_hndl0, &sr_sync);
                        sync_sent++;
                        if (res != VAPI_OK)
                        {
                            printf("sg_list_len %d\n", sr_sync.sg_lst_len );
                            printf("Error: Posting SYNC sr (sent %d) : %s\n", sync_sent, VAPI_strerror(res));
                            return(RET_ERR);
                        }
                        COND_PRINTF3((" **** Posting SYNC sr %d unsignaled i=%d ask=%d sent=%d\n",(int)sr_sync.id ,i,ask_for_comp, sync_sent));
                    }
                }
                rr.id++;
                if (sent+received > threshold)
                {
                    /* calculate time */
                    get_clk64(t_end);
                    results[ix_cur].t_exec = t_end - t_start;
                    threshold += iter_cur<<1;
                    ix_cur++;
                    get_clk64(t_start);
                }
                if (received > iter)
                    break;  /* received the last reply from the receiver - one can finish the test ! */
                if (params_p->in_params.srq_num > 0)
                {
                    res = VAPI_post_srq(hca_hndl, srq_hndl, 1, &rr, &srq_posted_actual);
                    //printf("posted %d SRQs\n", srq_posted_actual);
                }
                else
                {
                    res = VAPI_post_rr(hca_hndl, qp_hndl0, &rr);
                }
                if (res != VAPI_OK)
                {
                    printf("Error: Posting Recv buffer : %s\n", VAPI_strerror(res));
                    return(RET_ERR);
                }
            }
        }
        COND_PRINTF3(("Finished polling for N rr's and the last sr completions, ask=%d\n",ask_for_comp));

        COND_PRINTF3(("Sender came out of Poll msg_size=%u, received %d \n",sc.byte_len, received));
        get_clk64(t_end);
        /* calculate time */
        results[ix_cur].t_exec = t_end - t_start;
        t_total = t_end - t_begin;
        /* sort the results */
        for (i=0;i<ix_cur;i++)
        {
            for (j=i+1;j<ix_cur;j++)
            {
                if (results[i].t_exec < results[j].t_exec)
                {
                    t_max = results[j].t_exec;
                    results[j].t_exec = results[i].t_exec;
                    results[i].t_exec = t_max;
                }
            }
        }   
        /* calculate extremums and total */
        t_exec = iter_loops * results[ix_cur>>1].t_exec;
        if (ix_cur >= 5 )
        {
            t_max = results[2].t_exec;
            t_min = results[ix_cur-3].t_exec;
        } else
        {
            t_max= t_min = t_exec;
        }
        COND_PRINTF3((" %.1f(%d) %.1f(%d) %.1f(%d) %.1f(%d) %.1f(%d) %.1f(%d) %.1f(%d) %.1f(%d) %.1f(%d)\n", 
                      BW_MBSEC(results[0].t_exec, (2.0 * size) * iter_cur), (int)CLK_TO_USEC(results[0].t_exec),
                      BW_MBSEC(results[1].t_exec, (2.0 * size) * iter_cur), (int)CLK_TO_USEC(results[1].t_exec),
                      BW_MBSEC(results[2].t_exec, (2.0 * size) * iter_cur), (int)CLK_TO_USEC(results[2].t_exec),
                      BW_MBSEC(results[ix_cur/2 -1].t_exec, (2.0 * size) * iter_cur), (int)CLK_TO_USEC(results[ix_cur/2 -1].t_exec),
                      BW_MBSEC(results[ix_cur/2].t_exec, (2.0 * size) * iter_cur), (int)CLK_TO_USEC(results[ix_cur/2 -1].t_exec),
                      BW_MBSEC(results[ix_cur/2 +1].t_exec, (2.0 * size) * iter_cur), (int)CLK_TO_USEC(results[ix_cur/2 +1].t_exec),
                      BW_MBSEC(results[ix_cur-2].t_exec, (2.0 * size) * iter_cur), (int)CLK_TO_USEC(results[ix_cur-2].t_exec),
                      BW_MBSEC(results[ix_cur-1].t_exec, (2.0 * size) * iter_cur), (int)CLK_TO_USEC(results[ix_cur-1].t_exec),
                      BW_MBSEC(results[ix_cur].t_exec, (2.0 * size) * iter_cur), (int)CLK_TO_USEC(results[ix_cur].t_exec)
                     ));
        COND_PRINTF2(("*** BW (MBytes/sec): MAX %.1f, MED %.1f, MIN %.1f [size: %d bytes, iter: %d, total %.0f]\n", 
                      BW_MBSEC( t_min, (2.0 * size) * iter_cur),
                      BW_MBSEC( t_total, (2.0 * size) * iter),
                      BW_MBSEC( t_max, (2.0 * size) * iter_cur),
                      size, iter, (2.0 * size) * iter));
        //printf("iter_loops %d, ix_cur %d, iter_cur %d, threshold %d\n", iter_loops, ix_cur, iter_cur, threshold );

        if (params_p->in_params.ports_num > 1)
        {
            params_p->in_params.iter = iter                                                                                            ;
            params_p->port[port_ix].snd.t_start = 0;
            params_p->port[port_ix].snd.t_end = t_exec;
            params_p->port[port_ix].size = size;
        } else
        {

#if 0
                printf("BW: %.1f MBytes/sec [Num. of QPs: %d, Post List: %d, QPC Sets: %d, MPT Sets: %d, Keys: %d, size: %d bytes, iter: %d, total %.0f]\n", 
                      BW_MBSEC(t_exec, (2.0 * size) * iter), 
                       params_p->in_params.num_of_qps, params_p->in_params.postlist_len, 
                       params_p->in_params.qpc_sets, params_p->in_params.mpt_sets, 
                       params_p->in_params.num_of_keys,
                       size, iter, (2.0 * size) * iter);
#else
            printf("BW: %.1f MBytes/sec [size: %d bytes, iter: %d, total %.0f]\n", 
                   BW_MBSEC(t_exec, (2.0 * size) * iter), 
                   size, iter, (2.0 * size) * iter);
#endif            
        }
        //printf("BW: %.1f MBytes/sec [size: %d bytes, iter: %d, total %.0f]\n", 
        //       BW_MBSEC(t_exec, (2.0 * size) * iter), 
        //       size, iter, (2.0 * size) * iter);
    }// Is sender

    else
    { /* Receiver side */
        for (i=0; i< iter; i++, tmp_cnt++)
        //for (i=0; i< iter; i++)
        {

            sr.id = i;
            sr.comp_type = VAPI_UNSIGNALED; 

            /* Ask for completion every comp_inter and on last send*/
            //if (((i+1)%comp_inter == 0) || i == (iter-1))
            if ((tmp_cnt == comp_inter) || i == (iter-1))
            {
                tmp_cnt = 0;
                sr.comp_type = VAPI_SIGNALED;
                ask_for_comp++;
                if (params_p->in_params.transport != IB_TS_RC) sync_ask_for_comp++;
                COND_PRINTF3((" **** Posting sr %d signaled i=%d ask=%d\n",(int)sr.id ,i,ask_for_comp));
            } else
            {
                COND_PRINTF3((" **** Posting sr %d unsignaled i=%d ask=%d\n",(int)sr.id ,i,ask_for_comp));
            }

            /* Use Inline data up to the limit size */
            //if ((size >0 ) &&(size <= params_p->max_inline_size))
            if (use_inline == TRUE)
            {
                res = EVAPI_post_inline_sr(hca_hndl, qp_hndl0, &sr);
            } else
            {
                res = VAPI_post_sr(hca_hndl, qp_hndl0, &sr);
            }
            if (res != VAPI_OK)
            {
                printf("sg_list_len %d\n", sr.sg_lst_len );
                printf("Error: Posting send desc : %s\n", VAPI_strerror(res));
                return(RET_ERR);
            }

            /* Wait for completion every comp_inter starting from 2*comp_inter
             * This applies only when comp_iter < iter */
            COND_PRINTF3(("**** Start polling for 1 sr and N rr completions, i=%d ask=%d\n",i,ask_for_comp));
            while (ask_for_comp > 1 || sync_ask_for_comp > 1)
            {
                res = PERF_poll_cq_nowait(hca_hndl, s_cq_hndl ,&sc,
                                          params_p->in_params.use_recv_events,  sq_sem_p);
                if (res == VAPI_EGEN) return RET_ERR;
                if (res == VAPI_CQ_EMPTY) continue;
                if (sc.id < iter)
                { /* it's - send completion */
                    ask_for_comp--;
                    COND_PRINTF3(("**** Got a completion of sr %d i=%d ask=%d\n", (int)sc.id,i,ask_for_comp));
                } else
                { /* it's - receive completion */
                    if ((params_p->in_params.transport != IB_TS_RC) && (sc.byte_len == tx_overhead))
                    {
                        sync_ask_for_comp--;
                        sync_received++;
                        COND_PRINTF3(("**** Got a completion of SYNC rr %d i=%d packets sync_received=%d\n", (int)sc.id,i,sync_received));
                    } else
                    {
                        received++;
                        COND_PRINTF3(("**** Got a completion of rr %d i=%d packets received=%d\n", (int)sc.id,i,received));
                        if ((params_p->in_params.transport != IB_TS_RC) && (received%comp_inter == 0 || received >= iter))
                        {
                            sr_sync.id = 2*iter + i;
                            res = VAPI_post_sr(hca_hndl, qp_hndl0, &sr_sync);
                            sync_sent++;
                            if (res != VAPI_OK)
                            {
                                printf("sg_list_len %d\n", sr_sync.sg_lst_len );
                                printf("Error: Posting SYNC sr (sent %d) : %s\n", sync_sent, VAPI_strerror(res));
                                return(RET_ERR);
                            }
                            COND_PRINTF3((" **** Posting SYNC sr %d unsignaled i=%d ask=%d sent=%d\n",(int)sr_sync.id ,i,ask_for_comp, sync_sent));
                        }
                    }
                    rr.id++;
                    if (params_p->in_params.srq_num > 0)
                    {
                        res = VAPI_post_srq(hca_hndl, srq_hndl, 1, &rr, &srq_posted_actual);
                        //printf("posted %d SRQs\n", srq_posted_actual);
                    }
                    else
                    {
                        res = VAPI_post_rr(hca_hndl, qp_hndl0, &rr);
                    }
                    if (res != VAPI_OK)
                    {
                        printf("Error: Posting Recv buffer : %s\n", VAPI_strerror(res));
                        return(RET_ERR);
                    }
                }
            }
            COND_PRINTF3(("Finished polling for 1 sr and N rr completions, sr.id %d comp i=%d ask=%d\n", (int)sc.id,i,ask_for_comp));
        }//for i - send packets

        /* Wait for completion on last send while getting also receive completions */
        COND_PRINTF2(("**** Start polling for N rr's and the last sr completions, ask=%d sync_ask=%d\n",
                      ask_for_comp, sync_ask_for_comp));
        while (ask_for_comp>0 || received < iter || sync_ask_for_comp > 0)
        {
            res = PERF_poll_cq_nowait(hca_hndl, s_cq_hndl ,&sc,
                                      params_p->in_params.use_recv_events,  sq_sem_p);
            if (res == VAPI_EGEN) return RET_ERR;
            if (res == VAPI_CQ_EMPTY) continue;
            if (sc.id < iter)
            { /* it's - send completion */
                ask_for_comp--;
                COND_PRINTF3(("**** Got a completion of the last sr %d ask=%d\n", (int)sc.id,ask_for_comp));
                continue;
            } else
            { /* it's - receive completion */
                if ((params_p->in_params.transport != IB_TS_RC) && (sc.byte_len == tx_overhead))
                {
                    sync_ask_for_comp--;
                    sync_received++;
                    COND_PRINTF3(("**** Got a completion of SYNC rr %d i=%d packets sync_received=%d\n", 
                                  (int)sc.id,i,sync_received));
                } else
                {
                    received++;
                    COND_PRINTF3(("**** Got a completion of rr %d i=%d packets received=%d size %u\n", 
                                  (int)sc.id,i,received, sc.byte_len));
                    if ((params_p->in_params.transport != IB_TS_RC) && (received%comp_inter == 0 || received >= iter))
                    {
                        sr_sync.id = 2*iter + i;
                        res = VAPI_post_sr(hca_hndl, qp_hndl0, &sr_sync);
                        sync_sent++;
                        if (res != VAPI_OK)
                        {
                            printf("sg_list_len %d\n", sr_sync.sg_lst_len );
                            printf("Error: Posting SYNC sr (sent %d) : %s\n", sync_sent, VAPI_strerror(res));
                            return(RET_ERR);
                        }
                        COND_PRINTF3((" **** Posting SYNC sr %d unsignaled i=%d ask=%d sent=%d\n",(int)sr_sync.id ,i,ask_for_comp, sync_sent));
                    }
                }
                rr.id++;
                if (params_p->in_params.srq_num > 0)
                {
                    res = VAPI_post_srq(hca_hndl, srq_hndl, 1, &rr, &srq_posted_actual);
                    //printf("posted %d SRQs\n", srq_posted_actual);
                }
                else
                {
                    res = VAPI_post_rr(hca_hndl, qp_hndl0, &rr);
                }
                if (res != VAPI_OK)
                {
                    printf("Error: Posting Recv buffer : %s\n", VAPI_strerror(res));
                    return(RET_ERR);
                }
            }
        }
        COND_PRINTF3(("Finished polling for N rr's and the last sr completions, ask=%d\n",ask_for_comp));

        /* Send reply to sender after last message. Use 0 length message!!! */
        COND_PRINTF2(("**** Sending the last reply (received %d, ask %d)\n",received,ask_for_comp));
        sg_entry_s.addr = (VAPI_virt_addr_t)(MT_virt_addr_t)send_buf;
        sr.sg_lst_len = (params_p->in_params.transport == IB_TS_RC) ? 0 : 1;
        sr.id = iter;
        res = VAPI_post_sr(hca_hndl, qp_hndl0, &sr);
        if (res != VAPI_OK)
        {
            printf("sg_list_len %d\n", sr.sg_lst_len );
            printf("Error: Posting send desc : %s\n", VAPI_strerror(res));
            return(RET_ERR);
        }

        /* Wait for final reply from receiver */
        if (RET_OK != PERF_poll_cq(hca_hndl,s_cq_hndl ,&sc,params_p->in_params.use_send_events, 
                                   sq_sem_p))
        {
            return RET_ERR;
        }
    }

    if (params_p->in_params.ports_num <= 1)
    {
        /* prevent receiver from exiting before the sender finished it's work */
        sock_sync_msg(params_p->in_params.socket, is_sender);
        COND_PRINTF1(("\n************* %s BW Bidirection Test Finished for port %d *********************\n\n",
                      tstr, params_p->port[port_ix].port_num));
    }
    return(RET_OK);
}

/***********************************************************************************************
* name:      rdma_w_bw_bi_test
* args:
*            params_p -- test parameter structure, need to filled before calling
* returns:
*            RET_OK
*            RET_ERR
*
* description:     
*            This is the main loop for the RDMA (RC) BW test. The test sends N packets via 1 port
*            then waits for acknowledge from the receiver, and calculate the BW.
*            
************************************************************************************************/  
ret_val_t rdma_w_bw_bi_test(test_params_tx *params_p, int port_ix)
{
    VAPI_ret_t             res;
    u_int64_t t_start=0, t_end=0;

    int i, j, k, comp_inter, ask_for_comp=0;

    VAPI_rr_desc_t         rr;
    VAPI_sg_lst_entry_t    sg_entry_r;
    VAPI_sr_desc_t       * sr;
    VAPI_sg_lst_entry_t    sg_entry_s;

    VAPI_sr_desc_t         rc_sr;
    VAPI_sg_lst_entry_t    rc_sg_entry_s;

    VAPI_wc_desc_t         sc;

    /* Copy all input parameters - don't want to rewrite the loop code */
    VAPI_hca_hndl_t        hca_hndl    = params_p->hca_hndl;
    VAPI_qp_hndl_t         qp_hndl    = params_p->port[port_ix].qp_vect[0].qp_hndl;
    VAPI_cq_hndl_t         s_cq_hndl   = params_p->port[port_ix].s_cq_hndl;
    VAPI_srq_hndl_t        srq_hndl    = params_p->port[port_ix].qp_vect[0].srq_hndl;
    MOSAL_semaphore_t    * sq_sem_p    = &params_p->port[port_ix].sq_sem;
    u_int8_t             * recv_buf    = params_p->recv_buf;
    u_int8_t             * send_buf    = params_p->send_buf;
    int                    is_sender   = params_p->in_params.is_sender;
    int                    size        = params_p->in_params.size;
    int                    iter        = params_p->in_params.iter;      
    int                    list_len    = params_p->in_params.postlist_len;
    volatile char        * check_p;
    int                    index_to_check;
    call_result_t          mosal_rc;
    int                    noreply     = TRUE;
    char *tstr = (params_p->in_params.transport == IB_TS_UC) ? "UC" : "RC";
    u_int32_t            * qp_sent_count;
    int use_inline = FALSE;
    u_int32_t  srq_posted_actual;

    int max_msg_size = MIN2(params_p->in_params.mr_recv_size,params_p->in_params.mr_send_size);
    int max_sends = MIN2(params_p->in_params.num_r_wqe,params_p->in_params.num_s_wqe); 

    if (params_p->in_params.transport == IB_TS_UD)
    {
        max_msg_size = 2048;
        if (size > 2048)
        {
            printf("Warning!!!: UD max size is 2048\n");
            size = 2048;
        }
    }

    if (size > params_p->in_params.mr_recv_size)
    {
        printf("Warning!!!: size=%d > mr_recv_size=%d --- changing iter to %d\n", 
               size,params_p->in_params.mr_recv_size,params_p->in_params.mr_recv_size);
        size = params_p->in_params.mr_recv_size;
    }

    if (size==0)
    {
        printf("For RDMA size must be >= 1. Changing size to 1\n");
        size =1;
    }

    if (list_len > max_sends)
    {
        printf("Warning!!! Post List Length > WQ size --- reducing it to %d\n", max_sends);
        list_len = max_sends;
    }

    sr = VMALLOC(list_len*sizeof(VAPI_sr_desc_t));

    /**************    RC RR *****************************************************/
    rr.opcode = VAPI_RECEIVE;
    rr.comp_type = VAPI_SIGNALED;
    rr.sg_lst_len = 1;
    rr.sg_lst_p  = &sg_entry_r;
    sg_entry_r.lkey = params_p->recv_lkey;
    sg_entry_r.len = params_p->in_params.mr_recv_size;
    sg_entry_r.addr = (VAPI_virt_addr_t)(MT_virt_addr_t)recv_buf;


    /**************    RDMA SR *****************************************************/
    sg_entry_s.len = size;
    sg_entry_s.lkey = params_p->send_lkey;
    sg_entry_s.addr = (VAPI_virt_addr_t)(MT_virt_addr_t)send_buf;

    for (i=0 ; i < list_len ; i++)
    {

        sr[i].comp_type = VAPI_UNSIGNALED;
        sr[i].set_se = FALSE;

        sr[i].opcode = VAPI_RDMA_WRITE;
        sr[i].remote_qkey = 0;
        sr[i].remote_addr = params_p->remote_add;
        sr[i].r_key = params_p->remote_rkey;
        sr[i].id = 1;

        sr[i].sg_lst_len = 1; 
        sr[i].sg_lst_p = &sg_entry_s;
    }


    /**************    RC SR *****************************************************/
    rc_sr.comp_type = VAPI_SIGNALED;
    rc_sr.set_se = FALSE;
    rc_sr.opcode = VAPI_SEND;
    rc_sr.remote_qkey = 0;
    rc_sr.sg_lst_len = 0;

    rc_sg_entry_s.len = size;
    rc_sg_entry_s.lkey = params_p->send_lkey;
    rc_sg_entry_s.addr = (VAPI_virt_addr_t)(MT_virt_addr_t)send_buf; 
    rc_sr.sg_lst_p = &sg_entry_s;

    /* Both Sender  and receiver need 1 for last message*/
    rr.id = iter;
    if (params_p->in_params.srq_num > 0)
    {
        res = VAPI_post_srq(hca_hndl, srq_hndl, 1, &rr, &srq_posted_actual);
        //printf("posted %d SRQs\n", srq_posted_actual);
    }
    else
    {
        res = VAPI_post_rr(hca_hndl, qp_hndl, &rr);
    }
    if (res != VAPI_OK)
    {
        printf("Error: Posting Recv buffer : %s\n", VAPI_strerror(res));
        return(RET_ERR);
    }

    /* sync the test start */
    if (params_p->in_params.ports_num > 1)
    {
        COND_PRINTF2((" **** Test: finished the preparation phase for port %d and waiting for signal on %p ... \n", params_p->port[port_ix].port_num, &params_p->sync));
        MOSAL_sem_rel( &params_p->sem );
        mosal_rc = MOSAL_syncobj_waiton( &params_p->sync, MAX_WAIT_FOR_SYNC_USECS );
        if (mosal_rc != MT_OK)
        {
            printf( "Test hasn't received 'run' command from the main thread for port %d, exiting ...\n", params_p->port[port_ix].port_num);
            return(RET_ERR);
        }
    } else
    {
        sock_sync_msg(params_p->in_params.socket, is_sender);
    }
    if (params_p->in_params.ports_num <= 1)
    {
        COND_PRINTF1(("\n************* %s RDMA-Write BW Bidirectional Test started for port %d *********************\n\n", 
                      tstr, params_p->port[port_ix].port_num));
    }


    if (iter < max_sends)
    {
        comp_inter = max_sends;
    } else
    {
        comp_inter = (max_sends / 2 - list_len) | 1;
    }

    ask_for_comp = 0;

    index_to_check = port_ix;
    check_p = &recv_buf[index_to_check];

    qp_sent_count = calloc(sizeof(u_int32_t), params_p->in_params.num_of_qps);

    if ((size >0 ) &&(size <= params_p->max_inline_size) && (list_len == 1))
    {
        use_inline = TRUE;
    }

    if (is_sender)
    {
        COND_PRINTF2(("Checking data at %p (current=%d)\n",check_p,*check_p));
        while (*check_p==84)
        {
        }
        COND_PRINTF2((".................Got data %d at %p\n",*check_p,check_p));
        get_clk64(t_start);

        for (i=0,j=0; i<iter/list_len; i++)
        {
            // printf("Changing Send QP to %d\n", params_p->port[port_ix].qp_vect[j].qp_num);
            qp_hndl = params_p->port[port_ix].qp_vect[j].qp_hndl;

            for (k=0;k<list_len;k++)
            {
                sr[k].id = list_len * i + k;

                /* Ask for completion every comp_inter */
                if ((sr[k].id % comp_inter == 0) || 
                    (sr[k].id == ((iter/list_len)*list_len + k - 1)) ||
                    ((qp_sent_count[j]+k) % comp_inter == 0) || 
                    (qp_sent_count[j]+k == (iter/params_p->in_params.num_of_qps-1)))		    
                {
/*
                    int reason = 
                        (sr[k].id % comp_inter == 0) ? 1 :
                        (sr[k].id == ((iter/list_len)*list_len + k - 1)) ? 2 :
                        ((qp_sent_count[j]+k) % comp_inter == 0) ? 3 :
                        (qp_sent_count[j]+k == (iter/params_p->in_params.num_of_qps-1)) ? 4 : 0;
*/
                    sr[k].comp_type = VAPI_SIGNALED;
                    ask_for_comp++;
//                    printf("Upon SR id = %d ask_for_comp incremented to %d, reason = %d\n", 
//                           (u_int32_t) sr[k].id, ask_for_comp, reason);
                    COND_PRINTF2((" **** Posting sr signaled i=%d ask=%d\n",i+1,ask_for_comp));
                } else
                {
                    sr[k].comp_type = VAPI_UNSIGNALED;
                }

            }
                
            //if ((size >0 ) &&(size <= params_p->max_inline_size) && (list_len == 1))
            if (use_inline == TRUE)
            {
//                printf("Calling EVAPI_post_inline_sr. id = %d, comp_inter = %d, ask_for_comp = %d\n", 
//                       (u_int32_t) sr[0].id, comp_inter, ask_for_comp);
                res = EVAPI_post_inline_sr(hca_hndl, qp_hndl, sr);
            } else
            {
//                printf("Calling EVAPI_post_sr_list start id = %d, comp_inter = %d, ask_for_comp = %d\n", 
//                       (u_int32_t) sr[0].id, comp_inter, ask_for_comp);
                res = EVAPI_post_sr_list(hca_hndl, qp_hndl, list_len, sr);
            }
            if (res != VAPI_OK)
            {
                printf("Error: Posting send desc : %s\n", VAPI_strerror(res));
                return(RET_ERR);
            }


            /* Wait for completion every comp_inter starting from 2*comp_inter
            * This applies only when comp_iter > iter */
            while (ask_for_comp > 1)
            {
                COND_PRINTF2(("**** Polling for sr comp i=%d ask=%d\n",i+1,ask_for_comp));
                if (RET_OK != PERF_poll_cq(hca_hndl,s_cq_hndl ,&sc,params_p->in_params.use_send_events, 
                                           sq_sem_p))
                {
                    return RET_ERR;
                }
                                    
                ask_for_comp--;
//                printf("Completion succeeded for SR id %d, ask_for_comp decremented to %d\n", 
//                       (u_int32_t) sc.id, ask_for_comp);
            }
        
            qp_sent_count[j] += list_len;
            j = (j+1)%params_p->in_params.num_of_qps;            
        
        }//for i - send packets

        qp_hndl = params_p->port[port_ix].qp_vect[0].qp_hndl;        


//        for (i=0; i< iter; i++)
//        {
//            sr.id = i;
//            sr.comp_type = VAPI_UNSIGNALED; 
//
//            /* Ask for completion every comp_inter */
//            if (((i+1)%comp_inter == 0) || i == (iter-1))
//            {
//                sr.comp_type = VAPI_SIGNALED;
//                ask_for_comp++;
//                COND_PRINTF2((" **** Posting sr signaled i=%d ask=%d pix %d qp 0x%x \n",
//                              i,ask_for_comp, (int)port_ix, (unsigned)qp_hndl));
//            }
//
//            if ((size >0 ) &&(size <= params_p->max_inline_size))
//            {
//                res = EVAPI_post_inline_sr(hca_hndl, qp_hndl, &sr);
//            } else
//            {
//                res = VAPI_post_sr(hca_hndl, qp_hndl, &sr);
//            }
//            if (res != VAPI_OK)
//            {
//                printf("Error: Posting sr %d: %s pix %d qp 0x%x \n", 
//                       (int)sr.id, VAPI_strerror(res), (int)port_ix, (unsigned)qp_hndl);
//                return(RET_ERR);
//            }
//
//
//            /* Wait for completion every comp_inter starting from 2*comp_inter
//             * This applies only when comp_iter > iter */
//            if (ask_for_comp > 1)
//            {
//                COND_PRINTF2(("**** Polling for sr comp i=%d ask=%d pix %d qp 0x%x cq 0x%x\n",
//                              i,ask_for_comp, (int)port_ix, (unsigned)qp_hndl, (unsigned)s_cq_hndl));
//                if (RET_OK != PERF_poll_cq(hca_hndl,s_cq_hndl ,&sc,params_p->in_params.use_send_events, 
//                                           sq_sem_p))
//                {
//                    return RET_ERR;
//                }
//                ask_for_comp--;
//            }
//        }//for i - send packets

        /* Send an RC send  0 message indicating end of RDMA */
        COND_PRINTF2(("Finished RDMA posting, Sending RC-Send message pix %d qp 0x%x cq 0x%x\n", (int)port_ix, (unsigned)qp_hndl, (unsigned)s_cq_hndl));
        rc_sr.comp_type = VAPI_UNSIGNALED;
        rc_sr.sg_lst_len = 1;
        rc_sg_entry_s.len = 1;
        rc_sr.id = iter;
        res = VAPI_post_sr(hca_hndl, qp_hndl, &rc_sr);
        if (res != VAPI_OK)
        {
            printf("Error: Posting sr %d : %s pix %d qp 0x%x\n", 
                   (int)rc_sr.id, VAPI_strerror(res), (int)port_ix, (unsigned)qp_hndl);
            return(RET_ERR);
        }

        /* Wait for completion on last send (optionally) and reply from the receiver */
        COND_PRINTF2(("**** Start polling for last sr and reply from receiver completions, noreply=%d ask=%d pix %d qp 0x%x cq 0x%x\n",
                      noreply,ask_for_comp, (int)port_ix, (unsigned)qp_hndl, (unsigned)s_cq_hndl));
        while (noreply)
        {
            res = PERF_poll_cq_nowait(hca_hndl, s_cq_hndl ,&sc,
                                      params_p->in_params.use_recv_events,  sq_sem_p);
            if (res == VAPI_EGEN) return RET_ERR;
            if (res == VAPI_CQ_EMPTY) continue;
            if (sc.id < iter)
            { /* it's - send completion */
                ask_for_comp--;
                COND_PRINTF3(("**** Got a completion of sr %d noreply=%d ask=%d pix %d qp 0x%x cq 0x%x\n", 
                              (int)sc.id, noreply,ask_for_comp, (int)port_ix, (unsigned)qp_hndl, (unsigned)s_cq_hndl));
            } else
            { /* it's - receive reply completion */
                noreply = FALSE;
                COND_PRINTF3(("**** Got a completion of rr %d noreply=%d  pix %d qp 0x%x cq 0x%x\n", 
                              (int)sc.id,noreply, (int)port_ix, (unsigned)qp_hndl, (unsigned)s_cq_hndl));
            }
        }
        COND_PRINTF3(("Finished polling for last sr and reply from receiver completions, noreply=%d ask=%d pix %d qp 0x%x cq 0x%x\n",noreply,ask_for_comp, (int)port_ix, (unsigned)qp_hndl, (unsigned)s_cq_hndl));

        /* get the time */
        get_clk64(t_end);
        if (params_p->in_params.ports_num > 1)
        {
            params_p->port[port_ix].snd.t_start = t_start;
            params_p->port[port_ix].snd.t_end = t_end;
            params_p->port[port_ix].size = size;
        } else
        {
#if 0
            printf("BW: %.1f MBytes/sec [Num. of QPs: %d, Post List: %d, size: %d bytes, iter: %d, total %.0f]\n", 
                   BW_MBSEC(t_end - t_start, (2.0 * size) * iter), 
                   params_p->in_params.num_of_qps, params_p->in_params.postlist_len,
                   size, iter, (2.0 * size) * iter);
#else
            printf("BW: %.1f MBytes/sec [size: %d bytes, iter: %d, total %.0f]\n", 
                   BW_MBSEC(t_end - t_start, (2.0 * size) * iter), 
                   size, iter, (2.0 * size) * iter);
#endif

            //printf("BW: %.1f MBytes/sec [size: %d bytes, iter: %d, total %.0f] \n", 
            //       BW_MBSEC(t_end - t_start, (2.0 * size) * iter),     size, iter, (2.0 * size) * iter);
        }
    }// Is sender
    else
    {
        /* ---------- receiver ----------*/
        for (i=0,j=0; i<iter/list_len; i++)
        {
            // printf("Changing Send QP to %d\n", params_p->port[port_ix].qp_vect[j].qp_num);
            qp_hndl = params_p->port[port_ix].qp_vect[j].qp_hndl;

            for (k=0;k<list_len;k++)
            {
                sr[k].id = list_len * i + k;

                /* Ask for completion every comp_inter */
                if ((sr[k].id % comp_inter == 0) || 
                    (sr[k].id == ((iter/list_len)*list_len + k - 1)) ||
                    ((qp_sent_count[j]+k) % comp_inter == 0) || 
                    (qp_sent_count[j]+k == (iter/params_p->in_params.num_of_qps-1)))		    
                {
/*
                    int reason = 
                        (sr[k].id % comp_inter == 0) ? 1 :
                        (sr[k].id == ((iter/list_len)*list_len + k - 1)) ? 2 :
                        ((qp_sent_count[j]+k) % comp_inter == 0) ? 3 :
                        (qp_sent_count[j]+k == (iter/params_p->in_params.num_of_qps-1)) ? 4 : 0;
*/
                    sr[k].comp_type = VAPI_SIGNALED;
                    ask_for_comp++;
//                    printf("Upon SR id = %d ask_for_comp incremented to %d, reason = %d\n", 
//                           (u_int32_t) sr[k].id, ask_for_comp, reason);
                    COND_PRINTF2((" **** Posting sr signaled i=%d ask=%d\n",i+1,ask_for_comp));
                } else
                {
                    sr[k].comp_type = VAPI_UNSIGNALED;
                }

            }
                
            //if ((size >0 ) &&(size <= params_p->max_inline_size) && (list_len == 1))
            if (use_inline == TRUE)
            {
//                printf("Calling EVAPI_post_inline_sr. id = %d, comp_inter = %d, ask_for_comp = %d\n", 
//                       (u_int32_t) sr[0].id, comp_inter, ask_for_comp);
                res = EVAPI_post_inline_sr(hca_hndl, qp_hndl, sr);
            } else
            {
//                printf("Calling EVAPI_post_sr_list start id = %d, comp_inter = %d, ask_for_comp = %d\n", 
//                       (u_int32_t) sr[0].id, comp_inter, ask_for_comp);
                res = EVAPI_post_sr_list(hca_hndl, qp_hndl, list_len, sr);
            }
            if (res != VAPI_OK)
            {
                printf("Error: Posting send desc : %s\n", VAPI_strerror(res));
                return(RET_ERR);
            }



//        for (i=0; i< iter; i++)
//        {
//            sr.id = i;
//            sr.comp_type = VAPI_UNSIGNALED; 
//
//            /* Ask for completion every comp_inter */
//            if (((i+1)%comp_inter == 0) || i == (iter-1))
//            {
//                sr.comp_type = VAPI_SIGNALED;
//                ask_for_comp++;
//                COND_PRINTF2((" **** Posting sr signaled i=%d ask=%d pix %d qp 0x%x cq 0x%x\n",
//                              i,ask_for_comp, (int)port_ix, (unsigned)qp_hndl, (unsigned)s_cq_hndl));
//            }
//
//            if ((size >0 ) &&(size <= params_p->max_inline_size))
//            {
//                res = EVAPI_post_inline_sr(hca_hndl, qp_hndl, &sr);
//            } else
//            {
//                res = VAPI_post_sr(hca_hndl, qp_hndl, &sr);
//            }
//            if (res != VAPI_OK)
//            {
//                printf("Error: Posting sr %d (i=%d): %s (pix %d qp 0x%x \n", 
//                       (int)sr.id, i, VAPI_strerror(res), (int)port_ix, (unsigned)qp_hndl);
//                return(RET_ERR);
//            }
//
//
            /* Wait for completion every comp_inter starting from 2*comp_inter
             * This applies only when comp_iter > iter */
            while (ask_for_comp > 1)
            {
                COND_PRINTF2(("**** Polling for sr comp i=%d ask=%d pix %d qp 0x%x cq 0x%x\n",
                              i,ask_for_comp, (int)port_ix, (unsigned)qp_hndl, (unsigned)s_cq_hndl));
                while (1)
                {
                    if (RET_OK != PERF_poll_cq(hca_hndl,s_cq_hndl ,&sc,params_p->in_params.use_send_events, 
                                               sq_sem_p))
                    {
                        return RET_ERR;
                    }
                    if (sc.id < iter)
                    { /* it's - a send request */
                        ask_for_comp--;
                        COND_PRINTF3(("**** Got a completion of sr %d i=%d ask=%d pix %d qp 0x%x cq 0x%x\n", 
                                      (int)sc.id, i,ask_for_comp, (int)port_ix, (unsigned)qp_hndl, (unsigned)s_cq_hndl));
                        break;  
                    } else
                    { /* it's - receive of the special Send from the sender */
                        noreply = FALSE;  
                        COND_PRINTF3(("**** Got a completion of rr %d i=%d ask=%d pix %d qp 0x%x cq 0x%x\n", 
                                      (int)sc.id, i,ask_for_comp, (int)port_ix, (unsigned)qp_hndl, (unsigned)s_cq_hndl));
                    }
                }
            }

            qp_sent_count[j] += list_len;
            j = (j+1)%params_p->in_params.num_of_qps;            

        }//for i - send packets

        qp_hndl = params_p->port[port_ix].qp_vect[0].qp_hndl;
        
        /* Wait for completion on last send AND for the special Send from the sender */
        COND_PRINTF2(("**** Start polling for last sr and  for the special sender's Send completions, noreply=%d ask=%d pix %d qp 0x%x cq 0x%x\n",
                      noreply,ask_for_comp, (int)port_ix, (unsigned)qp_hndl, (unsigned)s_cq_hndl));
        while (ask_for_comp>0 || noreply)
        {
            res = PERF_poll_cq_nowait(hca_hndl, s_cq_hndl ,&sc,
                                      params_p->in_params.use_recv_events,  sq_sem_p);
            if (res == VAPI_EGEN) return RET_ERR;
            if (res == VAPI_CQ_EMPTY) continue;
            if (sc.id < iter)
            { /* it's - send completion */
                ask_for_comp--;
                COND_PRINTF3(("**** Got a completion of sr %d noreply=%d ask=%d pix %d qp 0x%x cq 0x%x\n", 
                              (int)sc.id, noreply,ask_for_comp, (int)port_ix, (unsigned)qp_hndl, (unsigned)s_cq_hndl));
            } else
            { /* it's - the special Send from the sender */
                noreply = FALSE;
                COND_PRINTF3(("**** Got a completion of rr %d noreply=%d  pix %d qp 0x%x cq 0x%x\n", 
                              (int)sc.id,noreply, (int)port_ix, (unsigned)qp_hndl, (unsigned)s_cq_hndl));
            }
        }
        COND_PRINTF2(("Finished polling for last sr and  for the special sender's Send completions, noreply=%d ask=%d pix %d qp 0x%x cq 0x%x\n",
                      noreply,ask_for_comp, (int)port_ix, (unsigned)qp_hndl, (unsigned)s_cq_hndl));

        /* Send reply to sender after last message. Use 0 length message!!! */
        rc_sr.comp_type = VAPI_SIGNALED;
        rc_sr.sg_lst_len = 0;
        rc_sr.id = iter;
        res = VAPI_post_sr(hca_hndl, qp_hndl, &rc_sr);
        if (res != VAPI_OK)
        {
            printf("Error: Posting sr %d : %s pix %d qp 0x%x cq 0x%x\n", 
                   (int)rc_sr.id, VAPI_strerror(res), (int)port_ix, (unsigned)qp_hndl, (unsigned)s_cq_hndl);
            return(RET_ERR);
        }
        COND_PRINTF2(("Receiver sent ack, waiting for completion pix %d qp 0x%x cq 0x%x\n", 
                      (int)port_ix, (unsigned)qp_hndl, (unsigned)s_cq_hndl));

        /* Poll for the last send completion */
        if (RET_OK != PERF_poll_cq(hca_hndl, s_cq_hndl ,&sc, 
                                   params_p->in_params.use_send_events,  sq_sem_p))
        {
            return RET_ERR;
        }
    }

    if (params_p->in_params.ports_num <= 1)
    {
        /* prevent receiver from exiting before the sender finished it's work */
        sock_sync_msg(params_p->in_params.socket, is_sender);
        COND_PRINTF1(("\n************* %s RDMA-Write BW Bidirectional Test Finished for port %d *********************\n\n", 
                      tstr,   params_p->port[port_ix].port_num));
    }

    free(sr);
    free(qp_sent_count);

    return(RET_OK);
}

/***********************************************************************************************
* name:      rdma_r_bw_bi_test
* args:
*            params_p -- test parameter structure, need to filled before calling
* returns:
*            RET_OK
*            RET_ERR
*
* description:     
*            This is the main loop for the RDMA (RC) BW test. The test sends N packets via 1 port
*            then waits for acknowledge from the receiver, and calculate the BW.
*            
************************************************************************************************/  
ret_val_t rdma_r_bw_bi_test(test_params_tx *params_p, int port_ix)
{
    VAPI_ret_t             res;
    u_int64_t t_start=0, t_end=0;
    u_int32_t  srq_posted_actual;

    int i,comp_inter, ask_for_comp=0, tmp_cnt = 1;

    VAPI_rr_desc_t         rr;
    VAPI_sg_lst_entry_t    sg_entry_r;
    VAPI_sr_desc_t         sr;
    VAPI_sg_lst_entry_t    sg_entry_s;

    VAPI_sr_desc_t         rc_sr;
    VAPI_sg_lst_entry_t    rc_sg_entry_s;

    VAPI_wc_desc_t         sc;

    /* Copy all input parameters - don't want to rewrite the loop code */
    VAPI_hca_hndl_t        hca_hndl    = params_p->hca_hndl;
    VAPI_qp_hndl_t         qp_hndl0     = params_p->port[port_ix].qp_vect[0].qp_hndl;
    VAPI_cq_hndl_t         s_cq_hndl   = params_p->port[port_ix].s_cq_hndl;
    VAPI_srq_hndl_t        srq_hndl    = params_p->port[port_ix].qp_vect[0].srq_hndl;
    MOSAL_semaphore_t    * sq_sem_p   = &params_p->port[port_ix].sq_sem;
    u_int8_t               *recv_buf   = params_p->recv_buf;
    u_int8_t               *send_buf   = params_p->send_buf;
    int                    is_sender   = params_p->in_params.is_sender;
    int                    size        = params_p->in_params.size;
    int                    iter        = params_p->in_params.iter;      
    call_result_t       mosal_rc;

    int max_msg_size = MIN2(params_p->in_params.mr_recv_size,params_p->in_params.mr_send_size);
    int max_sends = MIN2(params_p->in_params.num_r_wqe,params_p->in_params.num_s_wqe); 
    int noreply = TRUE;

    if (params_p->in_params.transport == IB_TS_UD)
    {
        max_msg_size = 2048;
        if (size > 2048)
        {
            printf("Warning!!!: UD max size is 2048\n");
            size = 2048;
        }
    }

    if (size > params_p->in_params.mr_recv_size)
    {
        printf("Warning!!!: size=%d > mr_recv_size=%d --- changing iter to %d\n", 
               size,params_p->in_params.mr_recv_size,params_p->in_params.mr_recv_size);
        size = params_p->in_params.mr_recv_size;
    }

    if (size==0)
    {
        printf("For RDMA size must be >= 1. Changing size to 1\n");
        size =1;
    }


    /**************    RC RR *****************************************************/
    rr.opcode = VAPI_RECEIVE;
    rr.comp_type = VAPI_SIGNALED;
    rr.sg_lst_len = 1;
    rr.sg_lst_p  = &sg_entry_r;
    sg_entry_r.lkey = params_p->recv_lkey;
    sg_entry_r.len = params_p->in_params.mr_recv_size;
    sg_entry_r.addr = (VAPI_virt_addr_t)(MT_virt_addr_t)recv_buf;


    /**************    RDMA SR *****************************************************/
    sr.comp_type = VAPI_SIGNALED;
    sr.set_se = FALSE;

    sr.opcode = VAPI_RDMA_READ;
    sr.remote_qkey = 0;
    sr.remote_addr = params_p->remote_add;
    sr.r_key = params_p->remote_rkey;
    sr.id = 1;


    sr.sg_lst_len = 1;
    sg_entry_s.len = size;
    sg_entry_s.lkey = params_p->send_lkey;
    sg_entry_s.addr = (VAPI_virt_addr_t)(MT_virt_addr_t)send_buf; 
    sr.sg_lst_p = &sg_entry_s;


    /**************    RC SR *****************************************************/
    rc_sr.comp_type = VAPI_SIGNALED;
    rc_sr.set_se = FALSE;
    rc_sr.opcode = VAPI_SEND;
    rc_sr.remote_qkey = 0;
    rc_sr.sg_lst_len = 0;

    rc_sg_entry_s.len = size;
    rc_sg_entry_s.lkey = params_p->send_lkey;
    rc_sg_entry_s.addr = (VAPI_virt_addr_t)(MT_virt_addr_t)send_buf; 
    rc_sr.sg_lst_p = &sg_entry_s;

    /* Sender needs 2 for the first and  last messages */
    if (is_sender)
    {
        for (i=0; i< 2; i++)
        {
            rr.id = iter+i;
            if (params_p->in_params.srq_num)
            {
                res = VAPI_post_srq(hca_hndl, srq_hndl, 1, &rr, &srq_posted_actual);
                //printf("posted %d SRQs\n", srq_posted_actual);
            }
            else
            {
                res = VAPI_post_rr(hca_hndl, qp_hndl0, &rr);
            }
            if (res != VAPI_OK)
            {
                printf("Error: Posting Recv buffer : %s\n", VAPI_strerror(res));
                return(RET_ERR);
            }
        }
    }

    /* sync the test start */
    if (params_p->in_params.ports_num > 1)
    {
        COND_PRINTF2((" **** Test: finished the preparation phase for port %d and waiting for signal on %p ... \n", params_p->port[port_ix].port_num, &params_p->sync));
        MOSAL_sem_rel( &params_p->sem );
        mosal_rc = MOSAL_syncobj_waiton( &params_p->sync, MAX_WAIT_FOR_SYNC_USECS );
        if (mosal_rc != MT_OK)
        {
            printf( "Test hasn't received 'run' command from the main thread for port %d, exiting ...\n", params_p->port[port_ix].port_num);
            return(RET_ERR);
        }
    } else
    {
        sock_sync_msg(params_p->in_params.socket, is_sender);
    }
    if (params_p->in_params.ports_num <= 1)
    {
        COND_PRINTF1(("\n************* RDMA-Read BW Bidirectional Test started for port %d *********************\n\n", params_p->port[port_ix].port_num));
    }


    if (iter < max_sends)
    {
        comp_inter = max_sends;
    } else
    {
        comp_inter = max_sends / 2 -2;
    }

    ask_for_comp = 0;

    if (is_sender)
    {
        COND_PRINTF2(("**** Polling for completion of first rr ask=%d\n",ask_for_comp));
        if (RET_OK != PERF_poll_cq(hca_hndl,s_cq_hndl ,&sc,params_p->in_params.use_send_events, 
                                   sq_sem_p))
        {
            return RET_ERR;
        }
        COND_PRINTF2(("**** Received completion of first rr, id=%d ask=%d\n",(int)sc.id, ask_for_comp));
        get_clk64(t_start);

        for (i=0; i< iter; i++, tmp_cnt++)
        {
            sr.id = i;
            sr.comp_type = VAPI_UNSIGNALED; 

            /* Ask for completion every comp_inter */
            //if (((i+1)%comp_inter == 0) || i == (iter-1))
            if ((tmp_cnt == comp_inter) || i == (iter-1))
            {
                tmp_cnt = 0;
                sr.comp_type = VAPI_SIGNALED;
                ask_for_comp++;
                COND_PRINTF2((" **** Posting sr signaled i=%d ask=%d\n",i+1,ask_for_comp));
            }

            res = VAPI_post_sr(hca_hndl, qp_hndl0, &sr);
            if (res != VAPI_OK)
            {
                printf("Error: Posting send desc : %s\n", VAPI_strerror(res));
                return(RET_ERR);
            }

            /* Wait for completion every comp_inter starting from 2*comp_inter
             * This applies only when comp_iter > iter */
            if (ask_for_comp > 1)
            {
                COND_PRINTF2(("**** Polling for sr comp i=%d ask=%d\n",i+1,ask_for_comp));
                while (1)
                {
                    if (RET_OK != PERF_poll_cq(hca_hndl,s_cq_hndl ,&sc,params_p->in_params.use_send_events, 
                                               sq_sem_p))
                    {
                        return RET_ERR;
                    }
                    if (sc.id < iter)
                    { /* it's - a send request */
                        ask_for_comp--;
                        COND_PRINTF3(("**** Got a completion of sr %d i=%d ask=%d pix %d qp 0x%x cq 0x%x\n", 
                                      (int)sc.id, i,ask_for_comp, (int)port_ix, (unsigned)qp_hndl0, (unsigned)s_cq_hndl));
                        break;  
                    } else
                    { /* it's - receive of the special Send from the sender */
                        noreply = FALSE;  
                        COND_PRINTF3(("**** Got a completion of rr %d i=%d ask=%d pix %d qp 0x%x cq 0x%x\n", 
                                      (int)sc.id, i,ask_for_comp, (int)port_ix, (unsigned)qp_hndl0, (unsigned)s_cq_hndl));
                    }
                }
            }
        }//for i - send packets

        /* Wait for completion on last send and reply from the receiver */
        COND_PRINTF2(("**** Start polling for last sr and reply from receiver completions, noreply=%d ask=%d\n",noreply,ask_for_comp));
        while (ask_for_comp>0 || noreply)
        {
            res = PERF_poll_cq_nowait(hca_hndl, s_cq_hndl ,&sc,
                                      params_p->in_params.use_recv_events,  sq_sem_p);
            if (res == VAPI_EGEN) return RET_ERR;
            if (res == VAPI_CQ_EMPTY) continue;
            if (sc. id < iter)
            { /* it's - send completion */
                ask_for_comp--;
                COND_PRINTF3(("**** Got a completion of sr %d noreply=%d ask=%d\n", (int)sc.id, noreply,ask_for_comp));
            } else
            { /* it's - receive reply completion */
                noreply = FALSE;
                COND_PRINTF3(("**** Got a completion of sr rr %d noreply=%d \n", (int)sc.id,noreply));
            }
        }
        COND_PRINTF3(("Finished polling for last sr and reply from receiver completions, noreply=%d ask=%d\n",noreply,ask_for_comp));

        /* get time */
        get_clk64(t_end);
        if (params_p->in_params.ports_num > 1)
        {
            params_p->port[port_ix].snd.t_start = t_start;
            params_p->port[port_ix].snd.t_end = t_end;
            params_p->port[port_ix].size = size;
        } else
        {

#if 0
                printf("BW: %.1f MBytes/sec [Num. of QPs: %d, Post List: %d, QPC Sets: %d, MPT Sets: %d, Keys: %d, size: %d bytes, iter: %d, total %.0f]\n", 
                      BW_MBSEC(t_end - t_start, (2.0 * size) * iter), 
                       params_p->in_params.num_of_qps, params_p->in_params.postlist_len, 
                       params_p->in_params.qpc_sets, params_p->in_params.mpt_sets, 
                       params_p->in_params.num_of_keys,
                       size, iter, (2.0 * size) * iter);
#else
            printf("BW: %.1f MBytes/sec [size: %d bytes, iter: %d, total %.0f]\n", 
                   BW_MBSEC(t_end - t_start, (2.0 * size) * iter), 
                   size, iter, (2.0 * size) * iter);
#endif            
        }
    }// Is sender
    else
    {
        /* ---------- receiver ----------*/

        /* Send message to sender to indicate he start of he work. Use 0 length message!!! */
        rc_sr.comp_type = VAPI_UNSIGNALED;
        rc_sr.sg_lst_len = 0;
        rc_sr.id = iter;
        res = VAPI_post_sr(hca_hndl, qp_hndl0, &rc_sr);
        if (res != VAPI_OK)
        {
            printf("Error: Posting send desc : %s\n", VAPI_strerror(res));
            return(RET_ERR);
        }
        COND_PRINTF2(("Receiver sent start message\n"));

        for (i=0; i< iter; i++, tmp_cnt++)
        {
            sr.id = i;
            sr.comp_type = VAPI_UNSIGNALED; 

            /* Ask for completion every comp_inter */
            //if (((i+1)%comp_inter == 0) || i == (iter-1))
            if ((tmp_cnt == comp_inter) || i == (iter-1))
            {
                tmp_cnt = 0;
                sr.comp_type = VAPI_SIGNALED;
                ask_for_comp++;
                COND_PRINTF2((" **** Posting sr signaled i=%d ask=%d\n",i+1,ask_for_comp));
            }

            res = VAPI_post_sr(hca_hndl, qp_hndl0, &sr);
            if (res != VAPI_OK)
            {
                printf("Error: Posting send desc : %s\n", VAPI_strerror(res));
                return(RET_ERR);
            }


            /* Wait for completion every comp_inter starting from 2*comp_inter
             * This applies only when comp_iter > iter */
            if (ask_for_comp > 1)
            {
                COND_PRINTF2(("**** Polling for sr comp i=%d ask=%d\n",i+1,ask_for_comp));
                if (RET_OK != PERF_poll_cq(hca_hndl,s_cq_hndl ,&sc,params_p->in_params.use_send_events, 
                                           sq_sem_p))
                {
                    return RET_ERR;
                }
                ask_for_comp--;
            }
        }//for i - send packets

        /* Wait for completion on last send */
        while (ask_for_comp>0)
        {
            COND_PRINTF2(("**** Polling for last sr comp i=%d ask=%d\n",i,ask_for_comp));

            /* Poll for the last send completion */
            if (RET_OK != PERF_poll_cq(hca_hndl,s_cq_hndl ,&sc,params_p->in_params.use_send_events, 
                                       sq_sem_p))
            {
                return RET_ERR;
            }
            ask_for_comp--;
            COND_PRINTF2(("Polling for sr comp %d %d\n",i,ask_for_comp));
        }

        /* Send reply to sender after last message. Use 0 length message!!! */
        rc_sr.comp_type = VAPI_SIGNALED;
        rc_sr.sg_lst_len = 0;
        rc_sr.id = iter+1;
        res = VAPI_post_sr(hca_hndl, qp_hndl0, &rc_sr);
        if (res != VAPI_OK)
        {
            printf("Error: Posting send desc : %s\n", VAPI_strerror(res));
            return(RET_ERR);
        }
        COND_PRINTF2(("Receiver sent ack, waiting for completion\n"));

        /* Poll for the last send completion */
        if (RET_OK != PERF_poll_cq(hca_hndl, s_cq_hndl ,&sc, 
                                   params_p->in_params.use_send_events,  sq_sem_p))
        {
            return RET_ERR;
        }
    }

    if (params_p->in_params.ports_num <= 1)
    {
        /* prevent receiver from exiting before the sender finished it's work */
        sock_sync_msg(params_p->in_params.socket, is_sender);
        COND_PRINTF1(("\n************* RDMA-Read BW Bidirectional Test Finished for port %d *********************\n\n", params_p->port[port_ix].port_num));
    }
    return(RET_OK);
}


/***********************************************************************************************
*
*	2-PORT SEND BW TEST W/O THREADS. OLD Version.
*   
*      DO NOT USE THIS TEST !!!!!
*            
************************************************************************************************/  
ret_val_t send_bw_uni_test_2ports_old(test_params_tx *params_p)
{
    VAPI_ret_t             res, res0, res1;
    u_int64_t t_start=0, t_end=0;

    int i, ii, j, k, l=0, m=0, comp_inter, ask_for_comp=0;

    VAPI_rr_desc_t         rr_0;
    VAPI_rr_desc_t         rr_1;
//    VAPI_sg_lst_entry_t  * sg_entry_r;
    VAPI_sr_desc_t       * sr_0;
    VAPI_sr_desc_t       * sr_1;
    VAPI_sg_lst_entry_t  * sg_entry_s;
    VAPI_wc_desc_t         rc_0;
    VAPI_wc_desc_t         rc_1;
    VAPI_wc_desc_t         sc_0;
    VAPI_wc_desc_t         sc_1;

    /* Copy all input parameters - don't want to rewrite the loop code */
    VAPI_hca_hndl_t        hca_hndl    = params_p->hca_hndl;
    VAPI_qp_hndl_t         qp_hndl_0   = params_p->port[0].qp_vect[0].qp_hndl;
    VAPI_qp_hndl_t         qp_hndl_1   = params_p->port[1].qp_vect[0].qp_hndl;
    VAPI_cq_hndl_t         s_cq_hndl_0 = params_p->port[0].s_cq_hndl;
    VAPI_cq_hndl_t         s_cq_hndl_1 = params_p->port[1].s_cq_hndl;
    VAPI_cq_hndl_t         r_cq_hndl_0 = params_p->port[0].r_cq_hndl;
    VAPI_cq_hndl_t         r_cq_hndl_1 = params_p->port[1].r_cq_hndl;
    VAPI_srq_hndl_t        srq_hndl_0  = params_p->port[0].qp_vect[0].srq_hndl;
    VAPI_srq_hndl_t        srq_hndl_1  = params_p->port[1].qp_vect[0].srq_hndl;
    MOSAL_semaphore_t    * sq_sem_p_0  = &params_p->port[0].sq_sem;
    MOSAL_semaphore_t    * sq_sem_p_1  = &params_p->port[1].sq_sem;
    MOSAL_semaphore_t    * rq_sem_p_0  = &params_p->port[0].rq_sem;
    MOSAL_semaphore_t    * rq_sem_p_1  = &params_p->port[1].rq_sem;
    u_int8_t               *recv_buf   = params_p->recv_buf;
//    u_int32_t               *recv_buf   = (u_int32_t *) params_p->recv_buf;
    u_int8_t               *send_buf   = params_p->send_buf;
    int                    is_sender   = params_p->in_params.is_sender;
    u_int32_t            * qp_sent_count;
    int                    size        = params_p->in_params.size;
    int                    iter        = params_p->in_params.iter;      
    int                    list_len    = params_p->in_params.postlist_len;
    //call_result_t       mosal_rc;
    char *tstr = (params_p->in_params.transport == IB_TS_UD) ? "UD" : ((params_p->in_params.transport == IB_TS_UC) ? "UC" : "RC");
    u_int32_t tx_overhead = (params_p->in_params.transport == IB_TS_UD) ? 40 : 0;
    int noreply = TRUE, sync_received=0, sync_sent=0, sync_ask_for_comp=0;    
    int comp_qp;    
    int use_inline = FALSE;
    u_int32_t  srq_posted_actual;
    qp_entry_t * comp_qp_entry_0;
    qp_entry_t * comp_qp_entry_1;

    int max_msg_size = MIN2(params_p->in_params.mr_recv_size,params_p->in_params.mr_send_size);
    int max_sends = MIN2(params_p->in_params.num_r_wqe,params_p->in_params.num_s_wqe); 


    if (params_p->in_params.transport == IB_TS_UD)
    {
        max_msg_size = 2048;
        if (size > 2048)
        {
            printf("Warning!!!: UD max size is 2048\n");
            size = 2048;
        }
    }

    if (size > params_p->in_params.mr_recv_size)
    {
        printf("Warning!!!: size=%d > mr_recv_size=%d --- changing iter to %d\n", 
               size,params_p->in_params.mr_recv_size,params_p->in_params.mr_recv_size);
        size = params_p->in_params.mr_recv_size;
    }

    if (list_len > max_sends)
    {
        printf("Warning!!! Post List Length > WQ size --- reducing it to %d\n", max_sends);
        list_len = max_sends;
    }

    sr_0 = VMALLOC(list_len*sizeof(VAPI_sr_desc_t));
    sr_1 = VMALLOC(list_len*sizeof(VAPI_sr_desc_t));
    sg_entry_s = VMALLOC(params_p->in_params.num_of_keys*sizeof(VAPI_sg_lst_entry_t));

    /**************    RC RR *****************************************************/
    rr_0.opcode = VAPI_RECEIVE;
    rr_1.opcode = VAPI_RECEIVE;
    rr_0.comp_type = VAPI_SIGNALED;
    rr_1.comp_type = VAPI_SIGNALED;
    rr_0.sg_lst_len = params_p->in_params.sg_list_len;
    rr_1.sg_lst_len = params_p->in_params.sg_list_len;

    rr_0.sg_lst_p = VMALLOC(rr_0.sg_lst_len*sizeof(VAPI_sg_lst_entry_t));
    rr_1.sg_lst_p = VMALLOC(rr_1.sg_lst_len*sizeof(VAPI_sg_lst_entry_t));

    rr_0.sg_lst_p[0].addr = (VAPI_virt_addr_t)(MT_virt_addr_t)recv_buf;
    rr_1.sg_lst_p[0].addr = (VAPI_virt_addr_t)(MT_virt_addr_t)recv_buf;
    rr_0.sg_lst_p[0].len = params_p->in_params.mr_recv_size;
    rr_1.sg_lst_p[0].len = params_p->in_params.mr_recv_size;

    for (i=1; i<(int)rr_0.sg_lst_len; i++) {
        rr_0.sg_lst_p[i].addr = (VAPI_virt_addr_t)(MT_virt_addr_t)recv_buf;
        rr_0.sg_lst_p[i].len = 4;
        rr_0.sg_lst_p[i].lkey = params_p->recv_lkey_arr[0];
    }
    for (i=1; i<(int)rr_1.sg_lst_len; i++) {
        rr_1.sg_lst_p[i].addr = (VAPI_virt_addr_t)(MT_virt_addr_t)recv_buf;
        rr_1.sg_lst_p[i].len = 4;
        rr_1.sg_lst_p[i].lkey = params_p->recv_lkey_arr[0];
    }

    /***************************************************************************/
    for (i=0; i<params_p->in_params.num_of_keys; i++) {    
        sg_entry_s[i].len = size;
        sg_entry_s[i].lkey = params_p->send_lkey_arr[i];
        sg_entry_s[i].addr = (VAPI_virt_addr_t)(MT_virt_addr_t)send_buf;
    }

    for (i=0; i<list_len; i++)
    {
        sr_0[i].comp_type = VAPI_UNSIGNALED; 
        sr_0[i].set_se = FALSE;

        sr_0[i].opcode = VAPI_SEND;
        sr_0[i].remote_qkey = 0;
        sr_0[i].remote_addr = params_p->remote_add;
        sr_0[i].id = 1;

        sr_0[i].sg_lst_len = 1; 
    
        /* Needed only for UD */
        if (params_p->in_params.multicast > 0)
        {
            sr_0[i].remote_qp = IB_MULTICAST_QP;
        }
        else
        {
            sr_0[i].remote_qp = params_p->port[0].qp_vect[0].d_qpnum;
        }
        sr_0[i].remote_ah = params_p->port[0].av_hndl;
    }

    for (i=0; i<list_len; i++)
    {
        sr_1[i].comp_type = VAPI_UNSIGNALED; 
        sr_1[i].set_se = FALSE;

        sr_1[i].opcode = VAPI_SEND;
        sr_1[i].remote_qkey = 0;
        sr_1[i].remote_addr = params_p->remote_add;
        sr_1[i].id = 1;

        sr_1[i].sg_lst_len = 1; 
    
        /* Needed only for UD */
        if (params_p->in_params.multicast > 0)
        {
            sr_1[i].remote_qp = IB_MULTICAST_QP;
        }
        else
        {
            sr_1[i].remote_qp = params_p->port[1].qp_vect[0].d_qpnum;
        }
        sr_1[i].remote_ah = params_p->port[1].av_hndl;
    }

    /* Post all possible receive in advance. Sender need 1 while receiver need N*/
    for (j=0;j<(int)params_p->in_params.num_of_qps;j++) 
    {    
        qp_hndl_0 = params_p->port[0].qp_vect[j].qp_hndl;
        qp_hndl_1 = params_p->port[1].qp_vect[j].qp_hndl;
        srq_hndl_0 = params_p->port[0].qp_vect[j].srq_hndl;
        srq_hndl_1 = params_p->port[1].qp_vect[j].srq_hndl;

        if (!is_sender)
        {
            for (i=0;i<max_sends;i++)
            {
                rr_0.id = i;
                rr_1.id = i;
                rr_0.sg_lst_p[0].lkey = params_p->recv_lkey_arr[m];
                rr_1.sg_lst_p[0].lkey = params_p->recv_lkey_arr[m];
                if (params_p->in_params.srq_num > 0)
                {
                    res = VAPI_post_srq(hca_hndl, srq_hndl_0, 1, &rr_0, &srq_posted_actual);
                    //printf("posted %d SRQs\n", srq_posted_actual);
                    if (res != VAPI_OK)
                    {
                        printf("Error: Pre-posting Recv buffer (port 1): %s\n", VAPI_strerror(res));
                        return(RET_ERR);
                    }
                    res = VAPI_post_srq(hca_hndl, srq_hndl_1, 1, &rr_1, &srq_posted_actual);
                    //printf("posted %d SRQs\n", srq_posted_actual);
                    if (res != VAPI_OK)
                    {
                        printf("Error: Pre-posting Recv buffer (port 2): %s\n", VAPI_strerror(res));
                        return(RET_ERR);
                    }
                }
                else
                {
                    res = VAPI_post_rr(hca_hndl, qp_hndl_0, &rr_0);
                    if (res != VAPI_OK)
                    {
                        printf("Error: Pre-posting Recv buffer (port 1) : %s\n", VAPI_strerror(res));
                        return(RET_ERR);
                    }
                    res = VAPI_post_rr(hca_hndl, qp_hndl_1, &rr_1);
                    if (res != VAPI_OK)
                    {
                        printf("Error: Pre-posting Recv buffer (port 2) : %s\n", VAPI_strerror(res));
                        return(RET_ERR);
                    }
                }
                m = (m+1)%params_p->in_params.num_of_keys;
            }
        } else
        {
            for (i=0; i<max_sends; i++)
            {
                rr_0.id = iter+i;
                rr_1.id = iter+i;
                rr_0.sg_lst_p[0].lkey = params_p->recv_lkey_arr[m];
                rr_1.sg_lst_p[0].lkey = params_p->recv_lkey_arr[m];
                if (params_p->in_params.srq_num > 0)
                {
                    res = VAPI_post_srq(hca_hndl, srq_hndl_0, 1, &rr_0, &srq_posted_actual);
                    //printf("posted %d SRQs\n", srq_posted_actual);
                    if (res != VAPI_OK)
                    {
                        printf("Error: Pre-posting Recv buffer (port 1): %s\n", VAPI_strerror(res));
                        return(RET_ERR);
                    }
                    res = VAPI_post_srq(hca_hndl, srq_hndl_1, 1, &rr_1, &srq_posted_actual);
                    //printf("posted %d SRQs\n", srq_posted_actual);
                    if (res != VAPI_OK)
                    {
                        printf("Error: Pre-posting Recv buffer (port 2): %s\n", VAPI_strerror(res));
                        return(RET_ERR);
                    }
                }
                else
                {
                    res = VAPI_post_rr(hca_hndl, qp_hndl_0, &rr_0);
                    if (res != VAPI_OK)
                    {
                        printf("Error: Pre-posting Recv buffer (port 1): %s\n", VAPI_strerror(res));
                        return(RET_ERR);
                    }
                    res = VAPI_post_rr(hca_hndl, qp_hndl_1, &rr_1);
                    if (res != VAPI_OK)
                    {
                        printf("Error: Pre-posting Recv buffer (port 2): %s\n", VAPI_strerror(res));
                        return(RET_ERR);
                    }
                }
                m = (m+1)%params_p->in_params.num_of_keys;
            }
        }
    }

    /* sync the test start */
    //if (params_p->in_params.ports_num > 1)
    //{
        //COND_PRINTF2((" **** Test: finished the preparation phase for port %d and waiting for signal on %p ... \n", params_p->port[port_ix].port_num, &params_p->sync));
        //MOSAL_sem_rel( &params_p->sem );
        //mosal_rc = MOSAL_syncobj_waiton( &params_p->sync, MAX_WAIT_FOR_SYNC_USECS );
        //if (mosal_rc != MT_OK)
        //{
            //printf( "Test hasn't received 'run' command from the main thread for port %d, exiting ...\n", params_p->port[port_ix].port_num);
            //return(RET_ERR);
        //}
    //} else
    //{
        sock_sync_msg(params_p->in_params.socket, is_sender);
    //}
    //if (params_p->in_params.ports_num <= 1)
    //{
        COND_PRINTF1(("\n\n************* %s BW Unidirection Test started for 2 ports *********************\n\n", tstr));
    //}

    if (iter < max_sends)
    {
        comp_inter = max_sends; // Ask completion ONLY on last message!!
    } 
    else if (params_p->in_params.num_of_qps > 512) comp_inter = 8;  // UGLY patch
    else comp_inter = max_sends/2 - 2;

    ask_for_comp = 0;

    if (is_sender)
    {
        qp_sent_count = calloc(sizeof(u_int32_t), params_p->in_params.num_of_qps);
        if ((size >0 ) &&(size <= params_p->max_inline_size) && (list_len == 1))
        {
            use_inline = TRUE;
        }
        
        get_clk64(t_start);

        for (i=0,j=0; i<iter/list_len; i++)
        {
            // printf("Changing Send QP to %d\n", params_p->port[0].qp_vect[j].qp_num);
            qp_hndl_0 = params_p->port[0].qp_vect[j].qp_hndl;
            qp_hndl_1 = params_p->port[1].qp_vect[j].qp_hndl;

            for (k=0;k<list_len;k++)
            {
                sr_0[k].id = list_len * i + k;
                sr_1[k].id = list_len * i + k;

                /* Ask for completion every comp_inter */
                if ((sr_0[k].id+1 % comp_inter == 0) || 
                    (sr_0[k].id == ((iter/list_len)*list_len + k - 1)) ||
                    ((qp_sent_count[j]+k) % comp_inter == 0) || 
                    (qp_sent_count[j]+k == (iter/params_p->in_params.num_of_qps-1)) ||
                    (list_len > 1))		    
                {
                    sr_0[k].comp_type = VAPI_SIGNALED;
                    sr_1[k].comp_type = VAPI_SIGNALED;
                    ask_for_comp++;
                    COND_PRINTF2((" **** Posting sr signaled i=%d ask=%d\n",i+1,ask_for_comp));
                } else
                {
                    sr_0[k].comp_type = VAPI_UNSIGNALED;
                    sr_1[k].comp_type = VAPI_UNSIGNALED;
                }

            }
                
            /* Use Inline data up to the limit size */
            //if ((size >0 ) &&(size <= params_p->max_inline_size) && (list_len == 1))
            if (use_inline == TRUE)
            {
                sr_0[0].r_key = params_p->remote_rkey_arr[l];
                sr_1[0].r_key = params_p->remote_rkey_arr[l];
                // printf("Remote R_Key = 0x%x\n", sr_0[0].r_key);
                sr_0[0].sg_lst_p = &sg_entry_s[l];
                sr_1[0].sg_lst_p = &sg_entry_s[l];
                res0 = EVAPI_post_inline_sr(hca_hndl, qp_hndl_0, sr_0);
                res1 = EVAPI_post_inline_sr(hca_hndl, qp_hndl_1, sr_1);
            } else
            {
                for (ii=0; ii<list_len; ii++) {
                    sr_0[ii].r_key = params_p->remote_rkey_arr[l];
                    sr_1[ii].r_key = params_p->remote_rkey_arr[l];
                    // printf("Remote R_Key = 0x%x\n", sr_0[ii].r_key);
                    sr_0[ii].sg_lst_p = &sg_entry_s[l];
                    sr_1[ii].sg_lst_p = &sg_entry_s[l];
                }
                res0 = EVAPI_post_sr_list(hca_hndl, qp_hndl_0, list_len, sr_0);
                res1 = EVAPI_post_sr_list(hca_hndl, qp_hndl_1, list_len, sr_1);

                COND_PRINTF3(("Posted SR list up to id=%lld, QP=%d, QP handle 0x%x\n", 
                              sr_0[list_len-1].id, j, qp_hndl_0));
                COND_PRINTF3(("Posted SR list up to id=%lld, QP=%d, QP handle 0x%x\n", 
                              sr_1[list_len-1].id, j, qp_hndl_1));
            }
            if (res0 != VAPI_OK)
            {
                printf("Error: Posting send desc : %s\n", VAPI_strerror(res0));
                return(RET_ERR);
            }
            if (res1 != VAPI_OK)
            {
                printf("Error: Posting send desc : %s\n", VAPI_strerror(res1));
                return(RET_ERR);
            }

            /* Wait for completion every comp_inter starting from 2*comp_inter
             * This applies only when comp_iter < iter */
            while (ask_for_comp > 0)
            {
                COND_PRINTF2(("**** Polling for sr comp i=%d ask=%d\n",i+1,ask_for_comp));
                if (RET_OK != PERF_poll_cq(hca_hndl,s_cq_hndl_0,&sc_0,params_p->in_params.use_send_events, sq_sem_p_0))
                {
                    return RET_ERR;
                }
                if (RET_OK != PERF_poll_cq(hca_hndl,s_cq_hndl_1,&sc_1,params_p->in_params.use_send_events, sq_sem_p_1))
                {
                    return RET_ERR;
                }
                ask_for_comp--;
                COND_PRINTF3(("**** Got a completion of sr %d i=%d ask=%d pix=1\n", (int)sc_0.id, i,ask_for_comp));
                COND_PRINTF3(("**** Got a completion of sr %d i=%d ask=%d pix=2\n", (int)sc_1.id, i,ask_for_comp));
	            if (params_p->in_params.transport != IB_TS_RC && !params_p->in_params.oneside) {
                    if (RET_OK != PERF_poll_cq(hca_hndl, r_cq_hndl_0 ,&rc_0, params_p->in_params.use_recv_events, rq_sem_p_0)) {
                        return RET_ERR;
                    }
                    if (RET_OK != PERF_poll_cq(hca_hndl, r_cq_hndl_1 ,&rc_1, params_p->in_params.use_recv_events, rq_sem_p_1)) {
                        return RET_ERR;
                    }

                    comp_qp_entry_0 = search_qp(rc_0.local_qp_num, params_p->port[0].qp_vect, params_p->in_params.num_of_qps);
                    comp_qp_entry_1 = search_qp(rc_1.local_qp_num, params_p->port[1].qp_vect, params_p->in_params.num_of_qps);

                    if (rc_0.byte_len == tx_overhead) {
                        sync_received++;
                        COND_PRINTF3(("**** Got a completion of SYNC rr %d i=%d packets sync_received=%d\n", 
                                      (int)rc_0.id,i,sync_received));
                    } else {
                        noreply=FALSE;
                        COND_PRINTF3(("**** Got a completion of final receiver's reply rr %d i=%d size %u\n", 
                                      (int)rc_0.id,i, rc_0.byte_len));
                    }
                    rr_0.id = iter+i;
                    rr_1.id = iter+i;
                    rr_0.sg_lst_p[0].lkey = params_p->recv_lkey_arr[m];
                    rr_1.sg_lst_p[0].lkey = params_p->recv_lkey_arr[m];
                    if (params_p->in_params.srq_num > 0)
                    {
                        res = VAPI_post_srq(hca_hndl, comp_qp_entry_0->srq_hndl, 1, &rr_0, &srq_posted_actual);
                        //printf("posted %d SRQs\n", srq_posted_actual);
                        if (res != VAPI_OK)
                        {
                            printf("Error: Pre-posting Recv buffer (port 1): %s\n", VAPI_strerror(res));
                            return(RET_ERR);
                        }
                        res = VAPI_post_srq(hca_hndl, comp_qp_entry_1->srq_hndl, 1, &rr_1, &srq_posted_actual);
                        //printf("posted %d SRQs\n", srq_posted_actual);
                        if (res != VAPI_OK)
                        {
                            printf("Error: Pre-posting Recv buffer (port 2): %s\n", VAPI_strerror(res));
                            return(RET_ERR);
                        }
                    }
                    else
                    {
                        res = VAPI_post_rr(hca_hndl, comp_qp_entry_0->qp_hndl, &rr_0);
                        if (res != VAPI_OK) {
                            printf("Error: Pre-posting Recv buffer (port 1): %s\n", VAPI_strerror(res));
                            return(RET_ERR);
                        }
                        res = VAPI_post_rr(hca_hndl, comp_qp_entry_1->qp_hndl, &rr_1);
                        if (res != VAPI_OK) {
                            printf("Error: Pre-posting Recv buffer (port 2): %s\n", VAPI_strerror(res));
                            return(RET_ERR);
                        }
                    }
                    m = (m+1)%params_p->in_params.num_of_keys;
                }
            }

            qp_sent_count[j] += list_len;
            j = (j+1)%params_p->in_params.num_of_qps;            
        
            l = (l+1)%params_p->in_params.num_of_keys;

        }//for i - send packets

        /* Wait for completion on last send */
        COND_PRINTF2(("**** Polling for last sr comp ask=%d\n",ask_for_comp));
        while (ask_for_comp>0)
        {
            /* Poll for the last send completion */
            if (RET_OK != PERF_poll_cq(hca_hndl,s_cq_hndl_0,&sc_0,params_p->in_params.use_send_events, sq_sem_p_0))
            {
                return RET_ERR;
            }
            if (RET_OK != PERF_poll_cq(hca_hndl,s_cq_hndl_1,&sc_1,params_p->in_params.use_send_events, sq_sem_p_1))
            {
                return RET_ERR;
            }
            ask_for_comp--;
            COND_PRINTF3(("**** Got a completion of sr %d pix=1 ask=%d\n", (int)sc_0.id, ask_for_comp));
            COND_PRINTF3(("**** Got a completion of sr %d pix=2 ask=%d\n", (int)sc_1.id, ask_for_comp));
        }

        /* Wait for final reply from receiver */
        COND_PRINTF2(("**** Waiting for final reply from receiver ask=%d\n",ask_for_comp));
                
        while (noreply && !params_p->in_params.oneside) 
        {
            if (RET_OK != PERF_poll_cq(hca_hndl, r_cq_hndl_0 ,&rc_0, params_p->in_params.use_recv_events, rq_sem_p_0)) {
                return RET_ERR;
            }
            if (RET_OK != PERF_poll_cq(hca_hndl, r_cq_hndl_1 ,&rc_1, params_p->in_params.use_recv_events, rq_sem_p_1)) {
                return RET_ERR;
            }

            comp_qp_entry_0 = search_qp(rc_0.local_qp_num, params_p->port[0].qp_vect, params_p->in_params.num_of_qps);
            comp_qp_entry_1 = search_qp(rc_1.local_qp_num, params_p->port[1].qp_vect, params_p->in_params.num_of_qps);

            if (params_p->in_params.transport == IB_TS_RC) {
                COND_PRINTF3(("**** Got a completion of final receiver's reply rr %d i=%d size %u\n", 
                              (int)rc_0.id,i, rc_0.byte_len));
                noreply = FALSE;                
            } else {
                if (rc_0.byte_len == tx_overhead) {
                    sync_received++;
                    COND_PRINTF3(("**** Got a completion of SYNC rr %d i=%d packets sync_received=%d\n", 
                                  (int)rc_0.id,i,sync_received));
                } else {
                    noreply=FALSE;
                    COND_PRINTF3(("**** Got a completion of final receiver's reply rr %d i=%d size %u\n", 
                                  (int)rc_0.id,i, rc_0.byte_len));
                }
                rr_0.id = iter+i;
                rr_1.id = iter+i;
                rr_0.sg_lst_p[0].lkey = params_p->recv_lkey_arr[m];
                rr_1.sg_lst_p[0].lkey = params_p->recv_lkey_arr[m];
                if (params_p->in_params.srq_num > 0)
                {
                    res = VAPI_post_srq(hca_hndl, comp_qp_entry_0->srq_hndl, 1, &rr_0, &srq_posted_actual);
                    //printf("posted %d SRQs\n", srq_posted_actual);
                    if (res != VAPI_OK)
                    {
                        printf("Error: Pre-posting Recv buffer (port 1): %s\n", VAPI_strerror(res));
                        return(RET_ERR);
                    }
                    res = VAPI_post_srq(hca_hndl, comp_qp_entry_1->srq_hndl, 1, &rr_1, &srq_posted_actual);
                    //printf("posted %d SRQs\n", srq_posted_actual);
                    if (res != VAPI_OK)
                    {
                        printf("Error: Pre-posting Recv buffer (port 2): %s\n", VAPI_strerror(res));
                        return(RET_ERR);
                    }
                }
                else
                {
                    res = VAPI_post_rr(hca_hndl, comp_qp_entry_0->qp_hndl, &rr_0);
                    if (res != VAPI_OK) {
                        printf("Error: Pre-posting Recv buffer (port 1): %s\n", VAPI_strerror(res));
                        return(RET_ERR);
                    }
                    res = VAPI_post_rr(hca_hndl, comp_qp_entry_1->qp_hndl, &rr_1);
                    if (res != VAPI_OK) {
                        printf("Error: Pre-posting Recv buffer (port 2): %s\n", VAPI_strerror(res));
                        return(RET_ERR);
                    }
                }
                m = (m+1)%params_p->in_params.num_of_keys;
            }
        }    

        COND_PRINTF3(("Sender came out of Poll msg_size=%u\n",rc_0.byte_len));
        get_clk64(t_end);

        //if (params_p->in_params.ports_num > 1)
        //{
            //params_p->port[port_ix].snd.t_start = t_start;
            //params_p->port[port_ix].snd.t_end = t_end;
            //params_p->port[port_ix].size = size;
        //} else
        //{
            //printf("BW: %.1f MBytes/sec [sg_list_len: %d, Num of QPs: %d, size: %d bytes, iter: %d, total 2 ports %.0f]\n", 
              printf("BW: %.1f MBytes/sec [size: %d bytes, iter: %d, total 2 ports %.0f]\n", 
                   BW_MBSEC(t_end - t_start, (2.0 * size) * iter), 
                   //params_p->in_params.sg_list_len, params_p->in_params.num_of_qps, 
                   size, iter, (2.0 * size) * iter);
        //}
        //printf("BW: %.1f MBytes/sec [sg_list_len: %d, Num of QPs: %d, size: %d bytes, iter: %d, total 2 ports %.0f]\n", 
        ////  printf("BW: %.1f MBytes/sec [size: %d bytes, iter: %d, total 2 ports %.0f]\n", 
        //       BW_MBSEC(t_end - t_start, (2.0 * size) * iter), 
        //       params_p->in_params.sg_list_len, params_p->in_params.num_of_qps, 
        //       size, iter, (2.0 * size) * iter);
    }// Is sender

    else
    { /* Receiver side */

        u_int32_t * qp_sync_sent_count = calloc(sizeof(u_int32_t), params_p->in_params.num_of_qps);

        sg_entry_s[0].addr = (VAPI_virt_addr_t)(MT_virt_addr_t)send_buf;
        sr_0[0].sg_lst_len = 0;
        sr_1[0].sg_lst_len = 0;
        sr_0[0].sg_lst_p = &sg_entry_s[0];
        sr_1[0].sg_lst_p = &sg_entry_s[0];

        for (i=0,j=0; i<iter; i++)
        {
            if (RET_OK != PERF_poll_cq(hca_hndl, r_cq_hndl_0, &rc_0, params_p->in_params.use_recv_events, rq_sem_p_0))
            {
                return RET_ERR;
            }
            COND_PRINTF3(("**** Got a completion of rr %d i=%d pix=1\n", (int)rc_0.id, i));                           
            if (RET_OK != PERF_poll_cq(hca_hndl, r_cq_hndl_1, &rc_1, params_p->in_params.use_recv_events, rq_sem_p_1))
            {
                return RET_ERR;
            }
            COND_PRINTF3(("**** Got a completion of rr %d i=%d pix=2\n", (int)rc_1.id, i));                           
            
            //printf("local_qp_num = 0x%x\n", rc.local_qp_num);
            
            comp_qp_entry_0 = search_qp(rc_0.local_qp_num, params_p->port[0].qp_vect, params_p->in_params.num_of_qps);

            if (comp_qp_entry_0 == NULL) {
                printf("Failed to find QP number\n");
                return RET_ERR;
            }

            comp_qp_entry_1 = search_qp(rc_1.local_qp_num, params_p->port[1].qp_vect, params_p->in_params.num_of_qps);

            if (comp_qp_entry_1 == NULL) {
                printf("Failed to find QP number\n");
                return RET_ERR;
            }

            comp_qp = (int)(MT_ulong_ptr_t)(comp_qp_entry_0 - params_p->port[0].qp_vect);

            //printf("comp_qp = %d\n", comp_qp);

            qp_hndl_0 = comp_qp_entry_0->qp_hndl;
            qp_hndl_1 = comp_qp_entry_1->qp_hndl;
            srq_hndl_0 = comp_qp_entry_0->srq_hndl;
            srq_hndl_1 = comp_qp_entry_1->srq_hndl;

            rr_0.id++;
            rr_1.id++;
            rr_0.sg_lst_p[0].lkey = params_p->recv_lkey_arr[m];
            rr_1.sg_lst_p[0].lkey = params_p->recv_lkey_arr[m];
            if (params_p->in_params.srq_num > 0)
            {
                res = VAPI_post_srq(hca_hndl, srq_hndl_0, 1, &rr_0, &srq_posted_actual);
                //printf("posted %d SRQs\n", srq_posted_actual);
                if (res != VAPI_OK)
                {
                    printf("Error: Pre-posting Recv buffer (port 1): %s\n", VAPI_strerror(res));
                    return(RET_ERR);
                }
                res = VAPI_post_srq(hca_hndl, srq_hndl_1, 1, &rr_1, &srq_posted_actual);
                //printf("posted %d SRQs\n", srq_posted_actual);
                if (res != VAPI_OK)
                {
                    printf("Error: Pre-posting Recv buffer (port 2): %s\n", VAPI_strerror(res));
                    return(RET_ERR);
                }
            }
            else
            {
                res = VAPI_post_rr(hca_hndl, qp_hndl_0, &rr_0);
                if (res != VAPI_OK)
                {
                    printf("Error: Re-posting Recv buffer (port 1, iter %d) : %s\n", i, VAPI_strerror(res));
                    return(RET_ERR);
                }
                res = VAPI_post_rr(hca_hndl, qp_hndl_1, &rr_1);
                if (res != VAPI_OK)
                {
                    printf("Error: Re-posting Recv buffer (port 2, iter %d) : %s\n", i, VAPI_strerror(res));
                    return(RET_ERR);
                }
            }
            m = (m+1)%params_p->in_params.num_of_keys;

            /* send sync msg on every comp_inter and on last send*/
            if (params_p->in_params.transport != IB_TS_RC)
            {
                if ((i%comp_inter == 0) || (i == iter-1) ||
                    (qp_sync_sent_count[comp_qp] % comp_inter == 0) || 
                    (qp_sync_sent_count[comp_qp] == (iter/params_p->in_params.num_of_qps-1)))
                {
                    sync_sent++;
                    sr_0[0].comp_type = VAPI_UNSIGNALED;
                    sr_1[0].comp_type = VAPI_UNSIGNALED;
                    if (sync_sent%comp_inter == 0)
                    {
                        sr_0[0].comp_type =  VAPI_SIGNALED;
                        sr_1[0].comp_type =  VAPI_SIGNALED;
                        qp_sync_sent_count[comp_qp]++;
                        sync_ask_for_comp++;
                    }
                    sr_0[0].id = i;
                    sr_1[0].id = i;
                    res = VAPI_post_sr(hca_hndl, qp_hndl_0, &sr_0[0]);
                    COND_PRINTF3((" **** Posting SYNC sr i=%d ask=%d sync_sent=%d\n",i+1,sync_ask_for_comp, sync_sent));
                    if (res != VAPI_OK)
                    {
                        printf("sg_list_len %d\n", sr_0[0].sg_lst_len );
                        printf("Error: Posting send desc : %s\n", VAPI_strerror(res));
                        return(RET_ERR);
                    }
                    res = VAPI_post_sr(hca_hndl, qp_hndl_1, &sr_1[0]);
                    COND_PRINTF3((" **** Posting SYNC sr i=%d ask=%d sync_sent=%d\n",i+1,sync_ask_for_comp, sync_sent));
                    if (res != VAPI_OK)
                    {
                        printf("sg_list_len %d\n", sr_1[0].sg_lst_len );
                        printf("Error: Posting send desc : %s\n", VAPI_strerror(res));
                        return(RET_ERR);
                    }
                    if (sync_ask_for_comp > 1)
                    {
                        if (sr_0[0].comp_type == VAPI_SIGNALED)
                        {
                            if (RET_OK != PERF_poll_cq(hca_hndl,s_cq_hndl_0 ,&sc_0,params_p->in_params.use_send_events, sq_sem_p_0))
                            {
                                return RET_ERR;
                            }
                        }
                        if (sr_1[0].comp_type == VAPI_SIGNALED)
                        {
                            if (RET_OK != PERF_poll_cq(hca_hndl,s_cq_hndl_1 ,&sc_1,params_p->in_params.use_send_events, sq_sem_p_1))
                            {
                                return RET_ERR;
                            }
                        }
                    }
                }
            }
        }

        /* Send reply to sender after last message. Use 0 length message for RC and a short one for others !!! */
        COND_PRINTF2(("**** Send reply to sender after last message, port 1 \n"));
        COND_PRINTF2(("**** Send reply to sender after last message, port 2 \n"));
        sr_0[0].comp_type = VAPI_SIGNALED;
        sr_1[0].comp_type = VAPI_SIGNALED;
        sr_0[0].id = iter;
        sr_1[0].id = iter;
        if (params_p->in_params.transport == IB_TS_RC)
        {
            sr_0[0].sg_lst_len = 0;
            sr_1[0].sg_lst_len = 0;
        } else
        {
            sr_0[0].sg_lst_len = 1;
            sr_1[0].sg_lst_len = 1;
            sg_entry_s[0].len = 4;
        }

        qp_hndl_0 = params_p->port[0].qp_vect[0].qp_hndl;
        
        res = VAPI_post_sr(hca_hndl, qp_hndl_0, &sr_0[0]);
        if (res != VAPI_OK)
        {
            printf("sg_list_len %d\n", sr_0[0].sg_lst_len );
            printf("Error: Posting send desc : %s\n", VAPI_strerror(res));
            return(RET_ERR);
        }

        qp_hndl_1 = params_p->port[1].qp_vect[0].qp_hndl;
        
        res = VAPI_post_sr(hca_hndl, qp_hndl_1, &sr_1[0]);
        if (res != VAPI_OK)
        {
            printf("sg_list_len %d\n", sr_1[0].sg_lst_len );
            printf("Error: Posting send desc : %s\n", VAPI_strerror(res));
            return(RET_ERR);
        }

        /* Wait for final reply from receiver */
        if (RET_OK != PERF_poll_cq(hca_hndl,s_cq_hndl_0,&sc_0,params_p->in_params.use_send_events, sq_sem_p_0))
        {
            return RET_ERR;
        }
        if (RET_OK != PERF_poll_cq(hca_hndl,s_cq_hndl_1,&sc_1,params_p->in_params.use_send_events, sq_sem_p_1))
        {
            return RET_ERR;
        }
    }

    //if (params_p->in_params.ports_num <= 1)
    //{
        /* prevent receiver from exiting before the sender finished it's work */
        sock_sync_msg(params_p->in_params.socket, is_sender);
        COND_PRINTF1(("\n************* %s BW Unidirection Test Finished for 2 ports *********************\n\n", tstr));
    //}
    return(RET_OK);
}



/***********************************************************************************************
*
*	2-PORT SEND BW TEST W/O THREADS
*   
*      DO NOT USE THIS TEST !!!!!
*            
************************************************************************************************/  
ret_val_t send_bw_uni_test_2ports(test_params_tx *params_p)
{
    VAPI_ret_t             res, res0, res1;
    u_int64_t t_start=0, t_end=0;

    int i, ii, j, k, l=0, m=0, comp_inter, ask_for_comp=0;

    VAPI_rr_desc_t         rr;
//    VAPI_sg_lst_entry_t  * sg_entry_r;
    VAPI_sr_desc_t       * sr;
    VAPI_sg_lst_entry_t  * sg_entry_s;
    VAPI_wc_desc_t         rc;
    VAPI_wc_desc_t         sc;

    /* Copy all input parameters - don't want to rewrite the loop code */
    VAPI_hca_hndl_t        hca_hndl    = params_p->hca_hndl;
    VAPI_qp_hndl_t         qp_hndl_0   = params_p->port[0].qp_vect[0].qp_hndl;
    VAPI_qp_hndl_t         qp_hndl_1   = params_p->port[0].qp_vect[1].qp_hndl;
    VAPI_qp_hndl_t         qp_hndl     = params_p->port[0].qp_vect[0].qp_hndl;
    VAPI_cq_hndl_t         s_cq_hndl = params_p->port[0].s_cq_hndl;
    VAPI_cq_hndl_t         r_cq_hndl = params_p->port[0].r_cq_hndl;
    VAPI_srq_hndl_t        srq_hndl  = params_p->port[0].qp_vect[0].srq_hndl;
    MOSAL_semaphore_t    * sq_sem_p  = &params_p->port[0].sq_sem;
    MOSAL_semaphore_t    * rq_sem_p  = &params_p->port[0].rq_sem;
    u_int8_t               *recv_buf   = params_p->recv_buf;
//    u_int32_t               *recv_buf   = (u_int32_t *) params_p->recv_buf;
    u_int8_t               *send_buf   = params_p->send_buf;
    int                    is_sender   = params_p->in_params.is_sender;
    u_int32_t            * qp_sent_count;
    int                    size        = params_p->in_params.size;
    int                    iter        = params_p->in_params.iter;      
    int                    list_len    = params_p->in_params.postlist_len;
    //call_result_t       mosal_rc;
    char *tstr = (params_p->in_params.transport == IB_TS_UD) ? "UD" : ((params_p->in_params.transport == IB_TS_UC) ? "UC" : "RC");
    u_int32_t tx_overhead = (params_p->in_params.transport == IB_TS_UD) ? 40 : 0;
    int noreply = TRUE, sync_received=0, sync_sent=0, sync_ask_for_comp=0;    
    int comp_qp;    
    int use_inline = FALSE;
    u_int32_t  srq_posted_actual;
    qp_entry_t * comp_qp_entry;

    int max_msg_size = MIN2(params_p->in_params.mr_recv_size,params_p->in_params.mr_send_size);
    int max_sends = MIN2(params_p->in_params.num_r_wqe,params_p->in_params.num_s_wqe); 


    if (params_p->in_params.transport == IB_TS_UD)
    {
        max_msg_size = 2048;
        if (size > 2048)
        {
            printf("Warning!!!: UD max size is 2048\n");
            size = 2048;
        }
    }

    if (size > params_p->in_params.mr_recv_size)
    {
        printf("Warning!!!: size=%d > mr_recv_size=%d --- changing iter to %d\n", 
               size,params_p->in_params.mr_recv_size,params_p->in_params.mr_recv_size);
        size = params_p->in_params.mr_recv_size;
    }

    if (list_len > max_sends)
    {
        printf("Warning!!! Post List Length > WQ size --- reducing it to %d\n", max_sends);
        list_len = max_sends;
    }

    sr = VMALLOC(list_len*sizeof(VAPI_sr_desc_t));
    sg_entry_s = VMALLOC(params_p->in_params.num_of_keys*sizeof(VAPI_sg_lst_entry_t));

    /**************    RC RR *****************************************************/
    rr.opcode = VAPI_RECEIVE;
    rr.comp_type = VAPI_SIGNALED;
    rr.sg_lst_len = params_p->in_params.sg_list_len;

    rr.sg_lst_p = VMALLOC(rr.sg_lst_len*sizeof(VAPI_sg_lst_entry_t));

    rr.sg_lst_p[0].addr = (VAPI_virt_addr_t)(MT_virt_addr_t)recv_buf;
    rr.sg_lst_p[0].len = params_p->in_params.mr_recv_size;

    for (i=1; i<(int)rr.sg_lst_len; i++) {
        rr.sg_lst_p[i].addr = (VAPI_virt_addr_t)(MT_virt_addr_t)recv_buf;
        rr.sg_lst_p[i].len = 4;
        rr.sg_lst_p[i].lkey = params_p->recv_lkey_arr[0];
    }

    /***************************************************************************/
    for (i=0; i<params_p->in_params.num_of_keys; i++) {    
        sg_entry_s[i].len = size;
        sg_entry_s[i].lkey = params_p->send_lkey_arr[i];
        sg_entry_s[i].addr = (VAPI_virt_addr_t)(MT_virt_addr_t)send_buf;
    }

    for (i=0; i<list_len; i++)
    {
        sr[i].comp_type = VAPI_UNSIGNALED; 
        sr[i].set_se = FALSE;

        sr[i].opcode = VAPI_SEND;
        sr[i].remote_qkey = 0;
        sr[i].remote_addr = params_p->remote_add;
        sr[i].id = 1;

        sr[i].sg_lst_len = 1; 
    
        /* Needed only for UD */
        if (params_p->in_params.multicast > 0)
        {
            sr[i].remote_qp = IB_MULTICAST_QP;
        }
        else
        {
            sr[i].remote_qp = params_p->port[0].qp_vect[0].d_qpnum;
        }
        sr[i].remote_ah = params_p->port[0].av_hndl;
    }

    /* Post all possible receive in advance. Sender need 1 while receiver need N*/
    //for (j=0;j<(int)params_p->in_params.num_of_qps;j++) 
    //{    
        qp_hndl_0 = params_p->port[0].qp_vect[0].qp_hndl;
        qp_hndl_1 = params_p->port[0].qp_vect[1].qp_hndl;
        srq_hndl = params_p->port[0].qp_vect[0].srq_hndl;

        if (!is_sender)
        {
            for (i=0;i<max_sends;i++)
            {
                rr.id = i;
                rr.sg_lst_p[0].lkey = params_p->recv_lkey_arr[m];
                if (params_p->in_params.srq_num > 0)
                {
                    res = VAPI_post_srq(hca_hndl, srq_hndl, 2, &rr, &srq_posted_actual);
                    //printf("posted %d SRQs\n", srq_posted_actual);
                    if (res != VAPI_OK)
                    {
                        printf("Error: Pre-posting Recv buffer: %s\n", VAPI_strerror(res));
                        return(RET_ERR);
                    }
                }
                else
                {
                    res = VAPI_post_rr(hca_hndl, qp_hndl_0, &rr);
                    if (res != VAPI_OK)
                    {
                        printf("Error: Pre-posting Recv buffer (port 1) : %s\n", VAPI_strerror(res));
                        return(RET_ERR);
                    }
                    res = VAPI_post_rr(hca_hndl, qp_hndl_1, &rr);
                    if (res != VAPI_OK)
                    {
                        printf("Error: Pre-posting Recv buffer (port 2) : %s\n", VAPI_strerror(res));
                        return(RET_ERR);
                    }
                }
                m = (m+1)%params_p->in_params.num_of_keys;
            }
        } else
        {
            for (i=0; i<max_sends; i++)
            {
                rr.id = iter+i;
                rr.sg_lst_p[0].lkey = params_p->recv_lkey_arr[m];
                if (params_p->in_params.srq_num > 0)
                {
                    res = VAPI_post_srq(hca_hndl, srq_hndl, 2, &rr, &srq_posted_actual);
                    //printf("posted %d SRQs\n", srq_posted_actual);
                    if (res != VAPI_OK)
                    {
                        printf("Error: Pre-posting Recv buffer: %s\n", VAPI_strerror(res));
                        return(RET_ERR);
                    }
                }
                else
                {
                    res = VAPI_post_rr(hca_hndl, qp_hndl_0, &rr);
                    if (res != VAPI_OK)
                    {
                        printf("Error: Pre-posting Recv buffer (port 1): %s\n", VAPI_strerror(res));
                        return(RET_ERR);
                    }
                    res = VAPI_post_rr(hca_hndl, qp_hndl_1, &rr);
                    if (res != VAPI_OK)
                    {
                        printf("Error: Pre-posting Recv buffer (port 2): %s\n", VAPI_strerror(res));
                        return(RET_ERR);
                    }
                }
                m = (m+1)%params_p->in_params.num_of_keys;
            }
        }
    //}

    /* sync the test start */
        sock_sync_msg(params_p->in_params.socket, is_sender);
        COND_PRINTF1(("\n\n************* %s BW Unidirection Test started for 2 ports *********************\n\n", tstr));

    if (iter < max_sends)
    {
        comp_inter = max_sends; // Ask completion ONLY on last message!!
    } 
    else if (params_p->in_params.num_of_qps > 512) comp_inter = 8;  // UGLY patch
    else comp_inter = max_sends/2 - 2;

    ask_for_comp = 0;

    if (is_sender)
    {
        qp_sent_count = calloc(sizeof(u_int32_t), params_p->in_params.num_of_qps);
        if ((size >0 ) &&(size <= params_p->max_inline_size) && (list_len == 1))
        {
            use_inline = TRUE;
        }
        
        get_clk64(t_start);

        for (i=0; i<iter/list_len; i++)
        {
            // printf("Changing Send QP to %d\n", params_p->port[0].qp_vect[j].qp_num);
            qp_hndl_0 = params_p->port[0].qp_vect[0].qp_hndl;
            qp_hndl_1 = params_p->port[0].qp_vect[1].qp_hndl;

            for (k=0;k<list_len;k++)
            {
                sr[k].id = list_len * i + k;

                /* Ask for completion every comp_inter */
                if ((sr[k].id+1 % comp_inter == 0) || 
                    (sr[k].id == ((iter/list_len)*list_len + k - 1)) ||
                    ((qp_sent_count[0]+k) % comp_inter == 0) || 
                    (qp_sent_count[0]+k == (iter/params_p->in_params.num_of_qps-1)) ||
                    (list_len > 1))		    
                {
                    sr[k].comp_type = VAPI_SIGNALED;
                    ask_for_comp = ask_for_comp + 2;
                    COND_PRINTF2((" **** Posting sr signaled i=%d ask=%d\n",i+1,ask_for_comp));
                } else
                {
                    sr[k].comp_type = VAPI_UNSIGNALED;
                }

            }
                
            /* Use Inline data up to the limit size */
            //if ((size >0 ) &&(size <= params_p->max_inline_size) && (list_len == 1))
            if (use_inline == TRUE)
            {
                sr[0].r_key = params_p->remote_rkey_arr[l];
                // printf("Remote R_Key = 0x%x\n", sr_0[0].r_key);
                sr[0].sg_lst_p = &sg_entry_s[l];
                res0 = EVAPI_post_inline_sr(hca_hndl, qp_hndl_0, sr);
                res1 = EVAPI_post_inline_sr(hca_hndl, qp_hndl_1, sr);
            } else
            {
                for (ii=0; ii<list_len; ii++) {
                    sr[ii].r_key = params_p->remote_rkey_arr[l];
                    // printf("Remote R_Key = 0x%x\n", sr_0[ii].r_key);
                    sr[ii].sg_lst_p = &sg_entry_s[l];
                }
                res0 = EVAPI_post_sr_list(hca_hndl, qp_hndl_0, list_len, sr);
                res1 = EVAPI_post_sr_list(hca_hndl, qp_hndl_1, list_len, sr);

                COND_PRINTF3(("Posted SR list up to id=%lld, QP=%d, QP handle 0x%x\n", 
                              sr[list_len-1].id, 0, qp_hndl_0));
                COND_PRINTF3(("Posted SR list up to id=%lld, QP=%d, QP handle 0x%x\n", 
                              sr[list_len-1].id, 1, qp_hndl_1));
            }
            if (res0 != VAPI_OK)
            {
                printf("Error: Posting send desc port 1: %s\n", VAPI_strerror(res0));
                return(RET_ERR);
            }
            if (res1 != VAPI_OK)
            {
                printf("Error: Posting send desc port 2: %s\n", VAPI_strerror(res1));
                return(RET_ERR);
            }

            /* Wait for completion every comp_inter starting from 2*comp_inter
             * This applies only when comp_iter < iter */
            while (ask_for_comp > 0)
            {
                COND_PRINTF2(("**** Polling for sr comp i=%d ask=%d\n",i+1,ask_for_comp));
                if (RET_OK != PERF_poll_cq(hca_hndl,s_cq_hndl,&sc,params_p->in_params.use_send_events, sq_sem_p))
                {
                    return RET_ERR;
                }
                ask_for_comp--;
                COND_PRINTF3(("**** Got a completion of sr %d i=%d ask=%d\n", (int)sc_0.id, i,ask_for_comp));
	            if (params_p->in_params.transport != IB_TS_RC && !params_p->in_params.oneside) {
                    if (RET_OK != PERF_poll_cq(hca_hndl, r_cq_hndl ,&rc, params_p->in_params.use_recv_events, rq_sem_p)) {
                        return RET_ERR;
                    }

                    comp_qp_entry = search_qp(rc.local_qp_num, params_p->port[0].qp_vect, params_p->in_params.num_of_qps);

                    if (rc.byte_len == tx_overhead) {
                        sync_received++;
                        COND_PRINTF3(("**** Got a completion of SYNC rr %d i=%d packets sync_received=%d\n", 
                                      (int)rc.id,i,sync_received));
                    } else {
                        noreply=FALSE;
                        COND_PRINTF3(("**** Got a completion of final receiver's reply rr %d i=%d size %u\n", 
                                      (int)rc.id,i, rc_0.byte_len));
                    }
                    rr.id = iter+i;
                    rr.sg_lst_p[0].lkey = params_p->recv_lkey_arr[m];
                    if (params_p->in_params.srq_num > 0)
                    {
                        res = VAPI_post_srq(hca_hndl, comp_qp_entry->srq_hndl, 1, &rr, &srq_posted_actual);
                        //printf("posted %d SRQs\n", srq_posted_actual);
                        if (res != VAPI_OK)
                        {
                            printf("Error: Pre-posting Recv buffer: %s\n", VAPI_strerror(res));
                            return(RET_ERR);
                        }
                    }
                    else
                    {
                        res = VAPI_post_rr(hca_hndl, comp_qp_entry->qp_hndl, &rr);
                        if (res != VAPI_OK) {
                            printf("Error: Pre-posting Recv buffer: %s\n", VAPI_strerror(res));
                            return(RET_ERR);
                        }
                    }
                    m = (m+1)%params_p->in_params.num_of_keys;
                printf("here\n");
                }
            }

            qp_sent_count[0] += list_len;
            //j = (j+1)%params_p->in_params.num_of_qps;            
        
            l = (l+1)%params_p->in_params.num_of_keys;

        }//for i - send packets

        /* Wait for completion on last send */
        COND_PRINTF2(("**** Polling for last sr comp ask=%d\n",ask_for_comp));
        while (ask_for_comp>0)
        {
            /* Poll for the last send completion */
            if (RET_OK != PERF_poll_cq(hca_hndl,s_cq_hndl,&sc,params_p->in_params.use_send_events, sq_sem_p))
            {
                return RET_ERR;
            }
            ask_for_comp--;
            COND_PRINTF3(("**** Got a completion of sr %d ask=%d\n", (int)sc.id, ask_for_comp));
        }

        /* Wait for final reply from receiver */
        COND_PRINTF2(("**** Waiting for final reply from receiver ask=%d\n",ask_for_comp));
                
        while (noreply && !params_p->in_params.oneside) 
        {
            if (RET_OK != PERF_poll_cq(hca_hndl, r_cq_hndl ,&rc, params_p->in_params.use_recv_events, rq_sem_p)) {
                return RET_ERR;
            }

            comp_qp_entry = search_qp(rc.local_qp_num, params_p->port[0].qp_vect, params_p->in_params.num_of_qps);

            if (params_p->in_params.transport == IB_TS_RC) {
                COND_PRINTF3(("**** Got a completion of final receiver's reply rr %d i=%d size %u\n", 
                              (int)rc.id,i, rc.byte_len));
                noreply = FALSE;                
            } else {
                if (rc.byte_len == tx_overhead) {
                    sync_received++;
                    COND_PRINTF3(("**** Got a completion of SYNC rr %d i=%d packets sync_received=%d\n", 
                                  (int)rc.id,i,sync_received));
                } else {
                    noreply=FALSE;
                    COND_PRINTF3(("**** Got a completion of final receiver's reply rr %d i=%d size %u\n", 
                                  (int)rc.id,i, rc.byte_len));
                }
                rr.id = iter+i;
                rr.sg_lst_p[0].lkey = params_p->recv_lkey_arr[m];
                if (params_p->in_params.srq_num > 0)
                {
                    res = VAPI_post_srq(hca_hndl, comp_qp_entry->srq_hndl, 1, &rr, &srq_posted_actual);
                    //printf("posted %d SRQs\n", srq_posted_actual);
                    if (res != VAPI_OK)
                    {
                        printf("Error: Pre-posting Recv buffer: %s\n", VAPI_strerror(res));
                        return(RET_ERR);
                    }
                }
                else
                {
                    res = VAPI_post_rr(hca_hndl, comp_qp_entry->qp_hndl, &rr);
                    if (res != VAPI_OK) {
                        printf("Error: Pre-posting Recv buffer: %s\n", VAPI_strerror(res));
                        return(RET_ERR);
                    }
                }
                m = (m+1)%params_p->in_params.num_of_keys;
            }
        }    

        COND_PRINTF3(("Sender came out of Poll msg_size=%u\n",rc.byte_len));
        get_clk64(t_end);

            //printf("BW: %.1f MBytes/sec [sg_list_len: %d, Num of QPs: %d, size: %d bytes, iter: %d, total 2 ports %.0f]\n", 
              printf("BW: %.1f MBytes/sec [size: %d bytes, iter: %d, total 2 ports %.0f]\n", 
                   BW_MBSEC(t_end - t_start, (2.0 * size) * iter), 
                   //params_p->in_params.sg_list_len, params_p->in_params.num_of_qps, 
                   size, iter, (2.0 * size) * iter);
    }// Is sender

    else
    { /* Receiver side */

        u_int32_t * qp_sync_sent_count = calloc(sizeof(u_int32_t), params_p->in_params.num_of_qps);

        sg_entry_s[0].addr = (VAPI_virt_addr_t)(MT_virt_addr_t)send_buf;
        sr[0].sg_lst_len = 0;
        sr[0].sg_lst_p = &sg_entry_s[0];

        for (i=0,j=0; i<iter*2; i++)
        {
            if (RET_OK != PERF_poll_cq(hca_hndl, r_cq_hndl, &rc, params_p->in_params.use_recv_events, rq_sem_p))
            {
                return RET_ERR;
            }
            COND_PRINTF3(("**** Got a completion of rr %d i=%d\n", (int)rc.id, i));                           
            
            //printf("local_qp_num = 0x%x\n", rc.local_qp_num);
            
            comp_qp_entry = search_qp(rc.local_qp_num, params_p->port[0].qp_vect, params_p->in_params.num_of_qps);

            if (comp_qp_entry == NULL) {
                printf("Failed to find QP number\n");
                return RET_ERR;
            }

            comp_qp = (int)(MT_ulong_ptr_t)(comp_qp_entry - params_p->port[0].qp_vect);

            //printf("comp_qp = %d\n", comp_qp);

            qp_hndl = comp_qp_entry->qp_hndl;
            srq_hndl = comp_qp_entry->srq_hndl;

            rr.id++;
            rr.sg_lst_p[0].lkey = params_p->recv_lkey_arr[m];
            if (params_p->in_params.srq_num > 0)
            {
                res = VAPI_post_srq(hca_hndl, srq_hndl, 1, &rr, &srq_posted_actual);
                //printf("posted %d SRQs\n", srq_posted_actual);
                if (res != VAPI_OK)
                {
                    printf("Error: Pre-posting Recv buffer: %s\n", VAPI_strerror(res));
                    return(RET_ERR);
                }
            }
            else
            {
                res = VAPI_post_rr(hca_hndl, qp_hndl, &rr);
                if (res != VAPI_OK)
                {
                    printf("Error: Re-posting Recv buffer (iter %d) : %s\n", i, VAPI_strerror(res));
                    return(RET_ERR);
                }
            }
            m = (m+1)%params_p->in_params.num_of_keys;

            /* send sync msg on every comp_inter and on last send*/
            if (params_p->in_params.transport != IB_TS_RC)
            {
                if ((i%comp_inter == 0) || (i == iter-1) ||
                    (qp_sync_sent_count[comp_qp] % comp_inter == 0) || 
                    (qp_sync_sent_count[comp_qp] == (iter/params_p->in_params.num_of_qps-1)))
                {
                    sync_sent++;
                    sr[0].comp_type = VAPI_UNSIGNALED;
                    if (sync_sent%comp_inter == 0)
                    {
                        sr[0].comp_type =  VAPI_SIGNALED;
                        qp_sync_sent_count[comp_qp]++;
                        sync_ask_for_comp++;
                    }
                    sr[0].id = i;
                    res = VAPI_post_sr(hca_hndl, qp_hndl, &sr[0]);
                    COND_PRINTF3((" **** Posting SYNC sr i=%d ask=%d sync_sent=%d\n",i+1,sync_ask_for_comp, sync_sent));
                    if (res != VAPI_OK)
                    {
                        printf("sg_list_len %d\n", sr[0].sg_lst_len );
                        printf("Error: Posting send desc : %s\n", VAPI_strerror(res));
                        return(RET_ERR);
                    }
                    if (sync_ask_for_comp > 1)
                    {
                        if (sr[0].comp_type == VAPI_SIGNALED)
                        {
                            if (RET_OK != PERF_poll_cq(hca_hndl,s_cq_hndl ,&sc,params_p->in_params.use_send_events, sq_sem_p))
                            {
                                return RET_ERR;
                            }
                        }
                    }
                }
            }
        }

        /* Send reply to sender after last message. Use 0 length message for RC and a short one for others !!! */
        COND_PRINTF2(("**** Send reply to sender after last message, port 1 \n"));
        COND_PRINTF2(("**** Send reply to sender after last message, port 2 \n"));
        sr[0].comp_type = VAPI_SIGNALED;
        sr[0].id = iter;
        if (params_p->in_params.transport == IB_TS_RC)
        {
            sr[0].sg_lst_len = 0;
        } else
        {
            sr[0].sg_lst_len = 1;
            sg_entry_s[0].len = 4;
        }

        qp_hndl_0 = params_p->port[0].qp_vect[0].qp_hndl;
        
        res = VAPI_post_sr(hca_hndl, qp_hndl_0, &sr[0]);
        if (res != VAPI_OK)
        {
            printf("sg_list_len %d\n", sr[0].sg_lst_len );
            printf("Error: Posting send desc : %s\n", VAPI_strerror(res));
            return(RET_ERR);
        }

        /* Wait for final reply from receiver */
        if (RET_OK != PERF_poll_cq(hca_hndl,s_cq_hndl,&sc,params_p->in_params.use_send_events, sq_sem_p))
        {
            return RET_ERR;
        }
    }

        /* prevent receiver from exiting before the sender finished it's work */
        sock_sync_msg(params_p->in_params.socket, is_sender);
        COND_PRINTF1(("\n************* %s BW Unidirection Test Finished for 2 ports *********************\n\n", tstr));
    return(RET_OK);
}



/***********************************************************************************************
* name:      wqe_test
* args:
*            params_p -- test parameter structure, need to filled before calling
* returns:
*            RET_OK
*            RET_ERR
*
* description:     
*            Test WQEs/sec
*            
************************************************************************************************/  
ret_val_t wqe_test(test_params_tx *params_p, int port_ix)
{
    VAPI_ret_t             res;
    u_int64_t t_start=0, t_end=0;

    int i,ii, j,comp_inter, ask_for_comp=0;
    int wqe_depth, totaldone=0, postlist_num=1;

    VAPI_rr_desc_t         rr;
    VAPI_sg_lst_entry_t    sg_entry_r;
    VAPI_sr_desc_t       * sr;
    VAPI_sg_lst_entry_t  * sg_entry_s;

    VAPI_sr_desc_t         rc_sr;
    VAPI_sg_lst_entry_t    rc_sg_entry_s;

    //VAPI_wc_desc_t         rc;
    VAPI_wc_desc_t         sc;

    /* Copy all input parameters - don't want to rewrite the loop code */
    VAPI_hca_hndl_t        hca_hndl    = params_p->hca_hndl;
    VAPI_qp_hndl_t         qp_hndl     = params_p->port[port_ix].qp_vect[0].qp_hndl;
    VAPI_cq_hndl_t         s_cq_hndl   = params_p->port[port_ix].s_cq_hndl;
    //VAPI_cq_hndl_t         r_cq_hndl   = params_p->port[port_ix].r_cq_hndl;
    MOSAL_semaphore_t    * sq_sem_p    = &params_p->port[port_ix].sq_sem;
    //MOSAL_semaphore_t    * rq_sem_p    = &params_p->port[port_ix].rq_sem;
    u_int8_t             * recv_buf    = params_p->recv_buf;
    u_int8_t             * send_buf    = params_p->send_buf;
    int                    is_sender   = params_p->in_params.is_sender;
    //u_int32_t            * qp_sent_count;
    int                    size        = params_p->in_params.size;
    int                    iter        = params_p->in_params.iter;      
    int                    list_len    = params_p->in_params.postlist_len;
    call_result_t          mosal_rc;
    char *tstr = (params_p->in_params.transport == IB_TS_UC) ? "UC" : "RC";
    int use_inline = FALSE;
    qp_entry_t * comp_qp_entry;
    //printf("Send CQ handle = 0x%x, Rcv CQ handle = 0x%x\n", (u_int32_t) s_cq_hndl, (u_int32_t) r_cq_hndl);

    int max_msg_size = MIN2(params_p->in_params.mr_recv_size,params_p->in_params.mr_send_size);
    //int max_sends = MIN2(params_p->in_params.num_r_wqe,params_p->in_params.num_s_wqe); 

    if (params_p->in_params.transport == IB_TS_UD)
    {
        max_msg_size = 2048;
        if (size > 2048)
        {
            printf("Warning!!!: UD max size is 2048\n");
            size = 2048;
        }
    }

    if (size > params_p->in_params.mr_recv_size)
    {
        printf("Warning!!!: size=%d > mr_recv_size=%d --- changing size to %d\n", 
               size,params_p->in_params.mr_recv_size,params_p->in_params.mr_recv_size);
        size = params_p->in_params.mr_recv_size;
    }

    if (size==0)
    {
        printf("For RDMA size must be >= 1. Changing size to 1\n");
        size =1;
    }

    //if (iter < max_sends)
    //{
    //    comp_inter = max_sends;
    //} else
    //{
    //    comp_inter = (max_sends / 2 - list_len) | 1;
    //}
    comp_inter = params_p->in_params.comp_inter;
    //printf("%d\n", comp_inter);
    //comp_inter = 256; // temporary. test 1a
    wqe_depth = MAX(256, comp_inter * 2);
    //wqe_depth = 128;
    // in memfree, for some reason, I can't post-list a list of more than 128 at once
    if (list_len > 1)
    {
        postlist_num = wqe_depth / 128;
        list_len = 128;
    }

    sr = VMALLOC(wqe_depth*sizeof(VAPI_sr_desc_t));
    sg_entry_s = VMALLOC(params_p->in_params.num_of_keys*sizeof(VAPI_sg_lst_entry_t));

    /**************    RC RR *****************************************************/
    rr.opcode = VAPI_RECEIVE;
    rr.comp_type = VAPI_SIGNALED;
    rr.sg_lst_len = 1;
    rr.sg_lst_p  = &sg_entry_r;
    sg_entry_r.lkey = params_p->recv_lkey;
    sg_entry_r.len = params_p->in_params.mr_recv_size;
    sg_entry_r.addr = (VAPI_virt_addr_t)(MT_virt_addr_t)recv_buf;


    /**************    RDMA SR *****************************************************/
    for (i=0; i<params_p->in_params.num_of_keys; i++) {    
        sg_entry_s[i].len = size;
        sg_entry_s[i].lkey = params_p->send_lkey_arr[i];
        sg_entry_s[i].addr = (VAPI_virt_addr_t)(MT_virt_addr_t)send_buf;
    }


    /**************    RC SR *****************************************************/
    rc_sr.comp_type = VAPI_SIGNALED;
    rc_sr.set_se = FALSE;
    rc_sr.opcode = VAPI_SEND;
    rc_sr.remote_qkey = 0;
    rc_sr.id = 1;
    rc_sr.sg_lst_len = 0;

    rc_sg_entry_s.len = size;
    rc_sg_entry_s.lkey = params_p->send_lkey;
    rc_sg_entry_s.addr = (VAPI_virt_addr_t)(MT_virt_addr_t)send_buf; 
    rc_sr.sg_lst_p = sg_entry_s;

    /* Both Sender  and receiver need 1 for last message*/
    //rr.id = 0;
    //res = VAPI_post_rr(hca_hndl, qp_hndl, &rr);
    //if (res != VAPI_OK)
    //{
    //    printf("Error: Posting Recv buffer : %s\n", VAPI_strerror(res));
    //    return(RET_ERR);
    //}

    /* sync the test start */
    if (params_p->in_params.ports_num > 1)
    {
        COND_PRINTF2((" **** Test: finished the preparation phase for port %d and waiting for signal on %p ... \n", params_p->port[port_ix].port_num, &params_p->sync));
        MOSAL_sem_rel( &params_p->sem );
        mosal_rc = MOSAL_syncobj_waiton( &params_p->sync, MAX_WAIT_FOR_SYNC_USECS );
        if (mosal_rc != MT_OK)
        {
            printf( "Test hasn't received 'run' command from the main thread for port %d, exiting ...\n", params_p->port[port_ix].port_num);
            return(RET_ERR);
        }
    } else
    {
        sock_sync_msg(params_p->in_params.socket, is_sender);
    }
    if (params_p->in_params.ports_num <= 1)
    {
        COND_PRINTF1(("\n************* %s WQE Test started for port %d *********************\n\n", 
                      tstr, params_p->port[port_ix].port_num));
    }

    ask_for_comp = 0;

    for (i=0; i<wqe_depth; i++)
    {
        if ((i+1) % comp_inter == 0)
        {
            sr[i].comp_type = VAPI_SIGNALED;
        }
        else
        {
            sr[i].comp_type = VAPI_UNSIGNALED;
        }
        sr[i].set_se = FALSE;

        sr[i].opcode = VAPI_RDMA_WRITE;
        sr[i].remote_qkey = 0;
        sr[i].remote_addr = params_p->remote_add;
        sr[i].id = 1;

        sr[i].sg_lst_len = 1; 
        sr[i].r_key = params_p->remote_rkey_arr[0];
        sr[i].sg_lst_p = &sg_entry_s[0];
    }

    if (is_sender)
    {
        //qp_sent_count = calloc(sizeof(u_int32_t), params_p->in_params.num_of_qps);
        if ((size >0 ) &&(size <= params_p->max_inline_size) && (list_len == 1))
        {
            use_inline = TRUE;
        }
        
        get_clk64(t_start);
        
        //printf("posting %d\n", wqe_depth);
        // post WQEdepth WQEs on each QP 
        if (list_len > 1)
        {
            for (j=0; j<params_p->in_params.num_of_qps; j++)
            {
                qp_hndl = params_p->port[port_ix].qp_vect[j].qp_hndl;
                
                for(ii=0; ii<postlist_num; ii++)
                {
                    res = EVAPI_post_sr_list(hca_hndl, qp_hndl, list_len, &sr[ii*list_len]);
                    if (res != VAPI_OK)
                    {
                        printf("Error: Posting send desc : %s\n", VAPI_strerror(res));
                        return(RET_ERR);
                    }
                }
                params_p->port[port_ix].qp_vect[j].outs_wqes += wqe_depth;
                //printf("posted %d to qp %d\n", wqe_depth, params_p->port[port_ix].qp_vect[j].qp_num);
                ask_for_comp += (wqe_depth / comp_inter);
            }
        }
        else
        {
            for (i=0; i<wqe_depth; i++)
            {
                sr[0].id = i;
                if ((sr[0].id +1) % comp_inter == 0) 
                {
                    sr[0].comp_type = VAPI_SIGNALED;
                    ask_for_comp++;
                    //COND_PRINTF2((" **** Posting sr signaled i=%d ask=%d\n",i+1,ask_for_comp));
                } else
                {
                    sr[0].comp_type = VAPI_UNSIGNALED;
                }

                for (j=0; j<params_p->in_params.num_of_qps; j++)
                {
                    qp_hndl = params_p->port[port_ix].qp_vect[j].qp_hndl;

                    if (use_inline == TRUE)
                    {
                        res = EVAPI_post_inline_sr(hca_hndl, qp_hndl, sr);
                    } else
                    {
                        res = EVAPI_post_sr_list(hca_hndl, qp_hndl, 1, sr);
                    }
                    if (res != VAPI_OK)
                    {
                        printf("Error: Posting send desc : %s\n", VAPI_strerror(res));
                        return(RET_ERR);
                    }
                    params_p->port[port_ix].qp_vect[j].outs_wqes++;
                }
            } // for i < wqe_depth
        } // else not use list

        //printf("done posting %d\n", wqe_depth);

        while (totaldone < iter)
        {
            //COND_PRINTF2(("**** Polling for sr comp i=%d ask=%d\n",i+1,ask_for_comp));
            //printf("polling\n");
            if (RET_OK != PERF_poll_cq(hca_hndl,s_cq_hndl ,&sc,params_p->in_params.use_send_events, 
                                       sq_sem_p))
            {
                printf("Perf_poll_cq failed\n");
                return RET_ERR;
            }
            //printf("Done polling\n");
            //printf("polled cq id %d\n", (int) sc.id);
            comp_qp_entry = search_qp(sc.local_qp_num, params_p->port[port_ix].qp_vect, 
                                      params_p->in_params.num_of_qps);
            //printf("completed %d wqes on qp %d\n", comp_inter, comp_qp_entry->qp_num);                    
            ask_for_comp--;
            totaldone += comp_inter;
            comp_qp_entry->outs_wqes = comp_qp_entry->outs_wqes - comp_inter;

            if ((totaldone < iter) && (comp_qp_entry->outs_wqes <= wqe_depth/2))
            {
                // post WQEdepth/2 WQEs on the QP 
                //printf("posting %d\n", wqe_depth/2);
                if (list_len > 1)
                {
                    //printf("posting list\n");
                    for(ii=0; ii<(postlist_num/2); ii++)
                    {
                        res = EVAPI_post_sr_list(hca_hndl, comp_qp_entry->qp_hndl, list_len, &sr[ii*list_len]);
                        //printf("Done posting list\n");
                        if (res != VAPI_OK)
                        {
                            printf("Error: Posting send desc : %s\n", VAPI_strerror(res));
                            return(RET_ERR);
                        }
                    }
                    //printf("posted %d to qp %d\n", wqe_depth/2, comp_qp_entry->qp_num);
                }
                else
                {
                    for (i=0; i<wqe_depth/2; i++)
                    {
                        sr[0].id = i;
                        if ((sr[0].id + 1) % comp_inter == 0) 
                        {
                            sr[0].comp_type = VAPI_SIGNALED;
                            //ask_for_comp++;
                            //COND_PRINTF2((" **** Posting sr signaled i=%d ask=%d\n",i+1,ask_for_comp));

                        } else
                        {
                            sr[0].comp_type = VAPI_UNSIGNALED;
                        }
                        if (use_inline == TRUE)
                        {
                            res = EVAPI_post_inline_sr(hca_hndl, comp_qp_entry->qp_hndl, sr);
                        } else
                        {
                            res = EVAPI_post_sr_list(hca_hndl, comp_qp_entry->qp_hndl, 1, sr);
                        }
                        if (res != VAPI_OK)
                        {
                            printf("Error: Posting send desc : %s\n", VAPI_strerror(res));
                            return(RET_ERR);
                        }
                    } // for i < wqe_depth/2
                } // else not use list
                //printf("done posting %d\n", wqe_depth/2);
                ask_for_comp += (wqe_depth/2)/comp_inter;
                comp_qp_entry->outs_wqes += wqe_depth/2;
            }
        } // while totaldone < iter

        get_clk64(t_end);

        // Clean left-over CQEs
        while (ask_for_comp > 1)
        {
            if (RET_OK != PERF_poll_cq(hca_hndl,s_cq_hndl ,&sc,params_p->in_params.use_send_events, 
                                       sq_sem_p))
            {
                printf("Perf_poll_cq failed\n");
                return RET_ERR;
            }
            ask_for_comp--;
        }
        //

        //if (params_p->in_params.ports_num > 1)
        //{
        //    params_p->port[port_ix].snd.t_start = t_start;
        //    params_p->port[port_ix].snd.t_end = t_end;
        //    params_p->port[port_ix].size = size;

        //} else
            printf("BW: %.1f WQE/sec [Port: %d  Message size: %d bytes Total WQEs: %d Time: %.4f Sec]\n", 
                   BW_BSEC(t_end - t_start, iter), 
                   port_ix, size, totaldone,((1.0*(ADJUST_TYPE)(t_end - t_start)) / (1.0*(ADJUST_TYPE)cpu_CLCKS_PER_SECS)));

        //free(qp_sent_count);
        //printf("lid %d  dlid %d\n", params_p->port[port_ix].lid, params_p->port[port_ix].d_lid);
    }// Is sender
    else
    {
    }

    free(sr);

    if (params_p->in_params.ports_num <= 1)
    {
        /* prevent receiver from exiting before the sender finished it's work */
        sock_sync_msg(params_p->in_params.socket, is_sender);
        COND_PRINTF1(("\n************* %s WQE Test Finished for port %d *********************\n\n",
                      tstr, params_p->port[port_ix].port_num));
    }
    return(RET_OK);
}


