/*                  - Mellanox Confidential and Proprietary -
 *
 *  Copyright (C) April 2002, Mellanox Technologies Ltd.  ALL RIGHTS RESERVED.
 *
 *  Except as specifically permitted herein, no portion of the information,
 *  including but not limited to object code and source code, may be reproduced,
 *  modified, distributed, republished or otherwise exploited in any form or by
 *  any means for any purpose without the prior written permission of Mellanox
 *  Technologies Ltd. Use of software subject to the terms and conditions
 *  detailed in the file "LICENSE.txt".
 *
 *  End of legal section ......................................................
 *
 *
 *  sock.h - small socket lib
 *
 *  Copyright (C) Mellanox Techniologies Ltd.
 *
 *  Cration date: 04/07/2002
 *
 *  Description: 
 *
 *  Version: $Id: sock.h,v 1.11 2004/03/31 07:58:03 dotanb Exp $
 *
 *  Authors: Eli Cohen, Alex Kholodenko
 *
 *  Changes:
 */
#ifndef H_SOCK_H
#define H_SOCK_H

#include <vapi.h>
#include <vapi_common.h>

#define SOCKET_RETRY 10

//-----------------------------------------------------------------------------
// Types Definition
//-----------------------------------------------------------------------------

struct test_cm_st
{
	IB_port_t 	port;
	IB_mtu_t	mtu_prim, mtu_sec;
	IB_lid_t	lid_prim, lid_sec;
    u_int8_t    sgid[16];

	IB_wqpn_t 		qpn;
};

#ifdef __WIN__
#if defined(_WIN64)
    typedef __int64 s_int_ptr_t;
#else
    typedef int s_int_ptr_t;
#endif
#else
    typedef long s_int_ptr_t;
#endif

//-----------------------------------------------------------------------------
// Global Variables
//-----------------------------------------------------------------------------
extern int wait_accept;
extern s_int_ptr_t socket_desc;

//-----------------------------------------------------------------------------
// Macros
//-----------------------------------------------------------------------------
#define SOCK_D2C( SOCK, DATA, SIZE )	\
	do			 				\
	{							\
	if( config.is_daemon ) 		\
		rc = sock_cm_send( (SOCK), (char*)(&(DATA)), (SIZE) );	\
	else														\
		rc = sock_cm_recv( (SOCK), (char*)(&(DATA)), (SIZE) );	\
	if( rc )					\
		goto cleanup;			\
	} while ( 0 )	


#define SOCK_C2D( SOCK, DATA, SIZE )	\
	do			 				\
	{							\
	if( config.is_daemon ) 		\
		rc = sock_cm_recv( (SOCK), (char*)(&(DATA)), (SIZE) );	\
	else														\
		rc = sock_cm_send( (SOCK), (char*)(&(DATA)), (SIZE) );	\
	if( rc )					\
		goto cleanup;			\
	} while ( 0 )	



//-----------------------------------------------------------------------------
// Global Function Definition
//-----------------------------------------------------------------------------
s_int_ptr_t create_tcp_connection(const char *ip, int port);
s_int_ptr_t wait_for_connection(int port);

int sock_cm_recv( s_int_ptr_t sock, char* buf, unsigned int size );
int sock_cm_send( s_int_ptr_t sock, char* buf, unsigned int size );

int sock_cm_connect( int is_daemon, int tcp_port, char *ip, 
				struct test_cm_st* local, struct test_cm_st* remote,
				s_int_ptr_t* p_cm_sock );

int sock_cm_ready_sync( s_int_ptr_t cm_sock, int is_daemon );
int sock_cm_addr_sync( s_int_ptr_t cm_sock, int is_daemon, 
									VAPI_mrw_t* p_mrw_local, VAPI_mrw_t* p_mrw_remote );

int sock_cm_uint64_swap(s_int_ptr_t cm_sock, int is_daemon, 
                        u_int64_t* buff);
int sock_cm_uint32_swap(s_int_ptr_t cm_sock, int is_daemon, 
                        u_int32_t* buff);
int sock_cm_uint16_swap(s_int_ptr_t cm_sock, int is_daemon, 
                        u_int16_t* buff);
int sock_cm_uint8_swap(s_int_ptr_t cm_sock, int is_daemon, 
                        u_int8_t* buff);
s_int_ptr_t sock_daemon_wait(int port);
s_int_ptr_t sock_client_connect(const char *ip, int port);
int sock_cm_data_sync( s_int_ptr_t cm_sock, int is_daemon, char* out_buf, char* in_buf, int size );


#endif /* H_SOCK_H */
